
/*
 * @copyright
 *
 *  Copyright 2014 Neo Natura
 *
 *  This file is part of the Share Library.
 *  (https://github.com/neonatura/share)
 *        
 *  The Share Library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version. 
 *
 *  The Share Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with The Share Library.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  @endcopyright
 */  

#include <math.h>
#include "shcoind.h"
#include "fractal.h"
#include "spring.h"

/* see end of file */
extern uint64_t _spring_matrix[256][256];

typedef struct spring_bits
{
  shnum_t lat;
  shnum_t lon;
  unsigned int x, y;
  unsigned int lat1, lat2;
  unsigned int lon1, lon2;
} spring_bits;


static unsigned int spring_matrix_bit(uint64_t seed)
{
  sh_sha_t s;
  uint8_t result[64];
  uint32_t val;
  unsigned int *i_ptr;
  int i;

  seed = htonll(seed);
  memset(result, 0, sizeof(result));

  sh_sha1_init(&s);
  sh_sha1_write(&s, (const char *)&seed, sizeof(seed));
  (void)sh_sha1_result(&s, result);

  val = 0;
  i_ptr = (uint32_t *)result;
  for (i = 0; i < 5; i++) {
    val += i_ptr[i];
  }

  return (val % 32);
}

static void spring_loc_bits(spring_bits *bits)
{
  uint32_t abs_lat, abs_lon;
  uint64_t b_lat, b_lon;
  shnum_t lat;
  shnum_t lon;

  lat = bits->lat;
  lon = bits->lon;

  /* collapse hemisphere into 'uni-sphere' */
  if (lat < 0) lat = lat * -1;
  if (lon < 0) lon = lon * -1;

  abs_lat = (uint32_t)lat;
  abs_lon = (uint32_t)lon;

  /* transition offset and ratio used to match matrix size */
  bits->y = (int)((lat - SPRING_OFFSET_LATITUDE) * SPRING_Y_FACTOR) % 256;
  bits->x = (int)((lon - SPRING_OFFSET_LONGITUDE) * SPRING_X_FACTOR) % 256;

  b_lat = (uint64_t)((lat - (shnum_t)abs_lat) * 1000.0);
  bits->lat1 = spring_matrix_bit(b_lat) % 32;
  bits->lat2 = spring_matrix_bit(b_lat/8) % 32;

  b_lon = (uint64_t)((lon - (shnum_t)abs_lon) * 1000.0);
  bits->lon1 = spring_matrix_bit(b_lon) % 32;
  bits->lon2 = spring_matrix_bit(b_lon/8) % 32;
}



void spring_loc_set(double lat, double lon)
{
  spring_bits bits;
  uint32_t *i_ar;

  memset(&bits, 0, sizeof(bits));
  bits.lat = lat;
  bits.lon = lon;
  spring_loc_bits(&bits);

  i_ar = (uint32_t *)&_spring_matrix[bits.y][bits.x];
  i_ar[0] |= (1 << bits.lat1);
  i_ar[0] |= (1 << bits.lat2);
  i_ar[1] |= (1 << bits.lon1);
  i_ar[1] |= (1 << bits.lon2);
}


int is_spring_loc(double lat, double lon)
{
  spring_bits bits;
  uint32_t *i_ar;

  memset(&bits, 0, sizeof(bits));
  bits.lat = lat;
  bits.lon = lon;
  spring_loc_bits(&bits);

  if (bits.lat1 == bits.lat2 || bits.lon1 == bits.lon2)
    return (FALSE);

  i_ar = (uint32_t *)&_spring_matrix[bits.y][bits.x];
  if (
      (i_ar[0] & (1 << bits.lat1)) &&
      (i_ar[0] & (1 << bits.lat2)) &&
      (i_ar[1] & (1 << bits.lon1)) &&
      (i_ar[1] & (1 << bits.lon2))
     ) {
    return (TRUE);
  }

  return (FALSE);
}

void spring_loc_claim(double lat, double lon)
{
  spring_bits bits;
  uint32_t *i_ar;

  memset(&bits, 0, sizeof(bits));
  bits.lat = lat;
  bits.lon = lon;
  spring_loc_bits(&bits);

  if (bits.lat1 == bits.lat2 || bits.lon1 == bits.lon2)
    return;

  /* remove first bits */
  i_ar = (uint32_t *)&_spring_matrix[bits.y][bits.x];
  if (
      (i_ar[0] & (1 << bits.lat1)) &&
      (i_ar[0] & (1 << bits.lat2)) &&
      (i_ar[1] & (1 << bits.lon1)) &&
      (i_ar[1] & (1 << bits.lon2))
     ) {
    i_ar[0] &= ~(1 << bits.lat1);
    i_ar[1] &= ~(1 << bits.lon1);
  }

}

int spring_loc_search(double cur_lat, double cur_lon, double *lat_p, double *lon_p)
{
  double s_lat, e_lat;
  double s_lon, e_lon;
  double lat, lon;
  int total;
  int found;

  s_lat = cur_lat - 0.25;
  s_lon = cur_lon - 0.25;
  e_lat = cur_lat + 0.25;
  e_lon = cur_lon + 0.25;

  for (lat = s_lat; lat <= e_lat; lat += 0.001) {
    for (lon = s_lon; lon <= e_lon; lon += 0.001) {
      if (is_spring_loc(lat, lon)) {
        *lat_p = lat;
        *lon_p = lon;
        return (0);
      }
    }
  }

  return (SHERR_NOENT);
}

void spring_matrix_compress(uint32_t matrix[3][3])
{
  int row, col;
  int y, x;

  for (row = 0; row < 3; row++) {
    for (col = 0; col < 3; col++) {
      matrix[row][col] = 0;
    }
  }

  for (y = 0; y < 256; y++) {
    for (x = 0; x < 256; x++) {
      row = (int)(y / 85) % 3;
      col = (int)(x / 85) % 3;
      matrix[row][col] += (uint32_t)_spring_matrix[y][x];
    }
  }

}

#if 0
int spring_render_fractal(char *img_path, double zoom, double span, double x_of, double y_of)
{
  static const unsigned int width = 256, height = 256;
  BMP *bmp;
  uint32_t m_seed;
  uint32_t val;
  shnum_t seed;
  shnum_t rate;
  shnum_t K;
  shnum_t Z;
  shnum_t C;
  shnum_t min_cord;
  shnum_t max_cord;
  shnum_t dx, dy;
  unsigned int x, y;
  int cval;
  int idx;
  unsigned char r, g, b;
  int px_width, px_height;

  px_width = (int)((shnum_t)256 / span);
  px_height = (int)((shnum_t)256 / span);
  bmp = BMP_Create(px_width, px_height, 32);

  m_seed = 0;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      m_seed += _spring_matrix[y][x];
    }
  }
  //seed = (shnum_t)m_seed / 16;
  seed = (shnum_t)m_seed;
  seed = seed / 65536;

  zoom = MAX(0.001, MIN(100.0, zoom));
  rate = span * zoom;
  min_cord = (-128 * zoom);
  max_cord = (128 * zoom);
  for (dy = min_cord; dy <= max_cord; dy += rate) {
    for (dx = min_cord; dx <= max_cord; dx += rate) {
      C = (dx * dy) + seed;

      Z = 0;
      for (idx = 0; idx < 16; idx++) {
        Z = Z * Z + C;
        if (fabs(Z) >= 2)
          break;
      }
      if (idx == 16) {
        r = g = b = 0;
      } else {
        K = fabs(Z*C);
        val = 4294967296 % (uint32_t)(K+1);
        r = (val >> 16) & 0xff;
        g = (val >> 8) & 0xff;
        b = val & 0xff;

        r = MIN(255, r + idx);
        g = MIN(255, g + idx);
        b = MIN(255, b + idx);
      }
      x = (int)((dx + max_cord) / rate) % px_width;
      y = (int)((dy + max_cord) / rate) % px_height;
      BMP_SetPixelRGB( bmp, x, y, r, g, b);
    }
  }
  BMP_WriteFile(bmp, img_path);
  BMP_CHECK_ERROR( stdout, -2 );
  BMP_Free( bmp );
  return 0;
}
#endif
int spring_render_fractal(char *img_path, double zoom, double span, double x_of, double y_of)
{
  static const int width = 256, height = 256;
  uint32_t m_seed;
  double seed;
  int y, x;

  m_seed = 0;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      m_seed += _spring_matrix[y][x];
    }
  }
  seed = (double)m_seed;

  return (fractal_render(img_path, seed, zoom, span, x_of, y_of));
}




/**
 * A 256x256 matrix with 'dual' bitvectors initialized to common US locations.
 */
uint64_t _spring_matrix[256][256] = {
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200200000000500ULL,0x0,0x4008080000ULL,0x2004008008000ULL,0x9100020000404000ULL,0xc00000080400ULL,0x0,0x0,0x100000880080000ULL,0x1210000008088100ULL,0x0,0x0,0x0,0x0,0x0,0x8000000102000100ULL,0x480000000005000ULL,0x0,0x0,0x0,0x0,0x101005000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc702420a0886600ULL,0x4122784201022700ULL,0x4000000210000400ULL,0x40504001002600ULL,0xa21801b10d482200ULL,0x10154302262400ULL,0x800000280000400ULL,0x0,0x0,0x200040020004000ULL,0x38002010000500ULL,0x2002000005000ULL,0x4002000008080000ULL,0x0,0x2100080000400ULL,0x0,0x0,0x0,0x0,0x0,0x10400000300000ULL,0x9000100c49007000ULL,0x4484018020244000ULL,0x50108000000500ULL,0x0,0x101000002200ULL,0x1004000000000500ULL,0x600e001880400ULL,0x1020c20401882600ULL,0x427e10812a2c00ULL,0x210040809100020ULL,0xe000400005060ULL,0x0,0x0,0x41000000002400ULL,0x0,0x120084000840c000ULL,0x2010000000120ULL,0x410000000300000ULL,0x0,0x0,0x0,0x8040000002200ULL,0x2040002000400ULL,0x0,0x0,0x0,0x0,0x20040008884000ULL,0x80000400000060ULL,0x8002000000000400ULL,0x0,0x0,0x0,0x4000200000300000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002000880000ULL,0x0,0x0,0x0,0x204000004002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200204000222800ULL,0x10400000022000ULL,0x0,0x281c00400022400ULL,0x448610000822a800ULL,0x438008100a80800ULL,0x48000880000ULL,0x20080000022000ULL,0x0,0x10100008004000ULL,0x80010008008000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa00026002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50200082002000ULL,0x4010141a00920120ULL,0xd0100800024520ULL,0x0,0x0,0x0,0x18200120020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8004000000000060ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44000000140ULL,0x0,0x0,0x6060000100460ULL,0x0,0x0,0x8000880000080920ULL,0x0,0x0,0x0,0x804000080020ULL,0x204701020021c0ULL,0x0,0x20800080800ULL,0x8040000000440ULL,0x0,0x4008102000020140ULL,0x4000001012000000ULL,0x0,0x0,0x0,0x0,0xba00244116820560ULL,0x2804010004100400ULL,0x0,0x4200042004060840ULL,0x0,0x0,0x80000412000000ULL,0x8005a0006080d00ULL,0x0,0x0,0x3200002000004440ULL,0x0,0x0,0x40200000420ULL,0x0,0x0,0x100400000000500ULL,0x800000400000140ULL,0x0,0x10402192000440ULL,0x88000000020020ULL,0x0,0x200200000000500ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1800002002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000800040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204000800040ULL,0x0,0x0,0x200000400200400ULL,0x0,0x8102000c00ULL,0x0,0x0,0x0,0x0,0x4000010000800200ULL,0x0,0x0,0x80000200800040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc00000000800ULL,0x208000000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81c454832182404ULL,0x480000002000100ULL,0x0,0x0,0x0,0x800880201082000ULL,0x4000002044000000ULL,0x8000060022140000ULL,0x0,0x0,0x100080000002800ULL,0x4000010002000004ULL,0x0,0x8000400001002000ULL,0x5000200254000004ULL,0x0,0x0,0x0,0x0,0x0,0x101002002000ULL,0xa00005210802100ULL,0x2020080100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x405008c0140040ULL,0x20004002008000ULL,0x4002000000080400ULL,0x0,0x0,0x0,0x1420f440c2005446ULL,0x200240280040040ULL,0x40803042000044ULL,0x200108882800040ULL,0x0,0x8000040002000100ULL,0x210000010040000ULL,0x45000082100000ULL,0x108080001000ULL,0x24048000000ULL,0x100d0084140000ULL,0x400004002040000ULL,0x0,0x0,0x200040020100000ULL,0x2800000108ULL,0x0,0x1000080040000400ULL,0x940020412040044ULL,0x8804052040082440ULL,0x0,0x0,0x8000200044000000ULL,0x2000840000080ULL,0x10040040000080ULL,0x0,0x100002048000000ULL,0x8200000000800040ULL,0x0,0x8400480046040000ULL,0xc12404a2820500c4ULL,0x5001004012000104ULL,0x0,0x0,0x8000080000400004ULL,0x0,0x0,0x0,0x0,0x2200000040004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2002082000000ULL,0x0,0x0,0x0,0x0,0x0,0x210000010000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20020090000000ULL,0x0,0x11000002800ULL,0x0,0x48090000000ULL,0x0,0x0,0x8000004004100000ULL,0x0,0x0,0x0,0x18000002800ULL,0x8100000090000000ULL,0x0,0x0,0x0,0x100002044000000ULL,0x0,0x100080002000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204004100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8010100101104208ULL,0x641808000110c210ULL,0x13206a000b02210ULL,0x8200001000001ULL,0x8000090001100200ULL,0x8080000000c000ULL,0x4000400800004ULL,0x0,0x8010000000004080ULL,0x0,0x0,0x0,0x0,0x1080000000400020ULL,0x0,0x0,0x0,0x0,0x4008004000ULL,0x0,0x8000010008000020ULL,0x0,0x80c10088002000ULL,0xc08020608800030aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2004408002800208ULL,0x400100000800001ULL,0x0,0x6000040000002ULL,0x0,0x0,0x0,0x0,0x0,0x800020000000c000ULL,0x40e0008c0004282ULL,0x98040001020020ULL,0x2020001004000ULL,0x0,0x0,0x0,0x80100800001ULL,0x4040824080044280ULL,0x20240088000002ULL,0x4000000200c00030ULL,0x100020008000100ULL,0x0,0x400006400006480ULL,0x41200d408880260aULL,0x12000002000ULL,0x0,0x840014040800012ULL,0x10002088000a00ULL,0x84000088000000ULL,0x4480004408082100ULL,0x6000000008000020ULL,0x0,0x2020000008004000ULL,0x0,0x8302000008020300ULL,0x0,0x20080000002400ULL,0x0,0x0,0x0,0x1100d2208000e440ULL,0x2000821080002440ULL,0x0,0x2800400020ULL,0x0,0x600000002001ULL,0x0,0x0,0x420080000010ULL,0x0,0x8000020000400020ULL,0x0,0x110c40208c06081ULL,0x841f4e7a0ac0e2a3ULL,0x4000100000000300ULL,0x12000002020ULL,0x4020000004080ULL,0x10000200002020ULL,0x0,0x1004000000020200ULL,0x0,0x10020000020200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800000808000020ULL,0x0,0x41001004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800010000200200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200280000040ULL,0x6000000000200200ULL,0x8100800002ULL,0xc10002080810040ULL,0x0,0x8100000000200200ULL,0x0,0x0,0x1004040808404040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000200200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4092302202402122ULL,0xa88a2d40cb04b10eULL,0x808000680000c00aULL,0x0,0x20000201000200ULL,0x0,0x0,0x2000004000004008ULL,0x0,0x0,0x0,0x0,0x24000004002ULL,0x0,0x0,0x8102040080a100ULL,0x0,0x8010000000440000ULL,0x0,0x1000004001004000ULL,0x0,0x8000140000040ULL,0x80010000008100ULL,0x804080200000ULL,0x0,0xa0000400008ULL,0x0,0x0,0x0,0x0,0x100002101004102ULL,0x0,0x4080001004000ULL,0x10002000008100ULL,0x0,0x0,0x0,0x300000400008ULL,0x0,0x0,0x200101000200ULL,0x802020040000ULL,0x0,0x0,0x100000840004000ULL,0xa00008400000ULL,0x0,0x0,0x0,0x26082004108ULL,0x44000500000ULL,0x800010001004000ULL,0x0,0x0,0x2800018000ULL,0xa00440050004000ULL,0x0,0x80000208000002ULL,0x2100000000060ULL,0x0,0x4000001000400008ULL,0x1000000440004000ULL,0x0,0x0,0x200020080000002ULL,0x0,0x202000604000c100ULL,0x4000000000040004ULL,0x108000004040ULL,0x0,0x0,0x20040000008008ULL,0x0,0x48030004000c108ULL,0x40201000200ULL,0x80040000804000ULL,0x100080001008000ULL,0x240000002100ULL,0x50000000004002ULL,0x5001000200ULL,0x400100001008000ULL,0x200200002100ULL,0x0,0x2045000040019000ULL,0x44e54408045c00aULL,0xb08e56b3eb50c112ULL,0x9001004000848004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x900400882500000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000808000002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10400008400000ULL,0x4000200000000102ULL,0x0,0x0,0x0,0x8040000000060ULL,0x0,0x0,0x0,0x0,0x8080a48408000ULL,0x410402040008000ULL,0x200040000000102ULL,0x0,0x0,0x2000240000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400024020100c000ULL,0x8000050080018000ULL,0x80400000020040ULL,0x201000001020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x120000018000ULL,0x0,0x0,0x0,0x4800001040ULL,0xa88862c8040ca01ULL,0x0,0x0,0x20800005000ULL,0x100020040800000ULL,0x0,0x0,0x0,0x60040000080ULL,0x0,0x0,0x81080280208000ULL,0x4d08100405200ULL,0x8000048000007080ULL,0xa24040280208301ULL,0x830065202203f020ULL,0x801003202084a100ULL,0x20044200038800ULL,0x18005000000d200ULL,0x0,0xc000400280003080ULL,0x0,0x248805010418a00ULL,0x404010400000ULL,0xb18c20804002a149ULL,0x41428020d040ULL,0x120000008200ULL,0x800000400020040ULL,0x80040000018000ULL,0x10001000000a00ULL,0x4102080000005201ULL,0x0,0x840000010400000ULL,0x2100000000001040ULL,0x0,0x2186d50c2223281ULL,0xc006286a81c2e020ULL,0x4602000090400000ULL,0x0,0x0,0x82400488001d000ULL,0x240000020800ULL,0x0,0x0,0x0,0x0,0x0,0x200200005000ULL,0x0,0x0,0x9000000000802000ULL,0x0,0x0,0x0,0x820804004000c004ULL,0x8000002040000020ULL,0x0,0x3004000200808100ULL,0x48000ca0040c820ULL,0x800010010400000ULL,0x0,0x0,0x0,0xa00002080ULL,0x0,0x0,0x300000001020ULL,0x0,0x0,0x8214440000400a00ULL,0xc1845820802da08ULL,0x5080200080041002ULL,0x0,0x4000040000404000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400000008200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50c00040040006ULL,0x0,0x800080000008200ULL,0x0,0x0,0x0,0x0,0x8040000101000ULL,0x0,0x80000008200ULL,0x81000040000004ULL,0x2100000008200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1080100000008200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2200020800808220ULL,0x0,0x8002000400200ULL,0x0,0x0,0x0,0x0,0x100000208000800ULL,0x0,0x8002020100422200ULL,0xc00000808000ULL,0x0,0x48000000001ULL,0x1480244002000a09ULL,0x5880612a88e20201ULL,0x20800022000ULL,0x0,0x0,0x0,0x0,0x800020000010200ULL,0x0,0x0,0x0,0x1004000000000600ULL,0x802200210001201ULL,0x4000400a00000ULL,0x0,0x0,0x808004c81c20000ULL,0x88490244013200ULL,0x0,0x30960108432200ULL,0x4000800000010200ULL,0x8000020200022010ULL,0x80c00408000206ULL,0x0,0x0,0x8200000081000ULL,0x14004020000ULL,0x108000020010ULL,0x0,0x0,0x0,0x288050000022210ULL,0x80800000101000ULL,0xa781048801201ULL,0x18420002000221ULL,0x1000000406000000ULL,0x9fa487047e23801ULL,0x3028044a00a00601ULL,0x8100004002000ULL,0x0,0x0,0x2204100810201ULL,0x4000001000080001ULL,0x820040000001ULL,0x140008000004ULL,0x108904020600ULL,0x8002008000080ULL,0x8000004000004800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x40020008000001ULL,0xa004020000ULL,0x8000002000400200ULL,0x20004000060000ULL,0x0,0x0,0x0,0x0,0x0,0x200011440a00001ULL,0x2200000004800ULL,0x410000000800001ULL,0x0,0x0,0x0,0x842000000e20000ULL,0x480002000001ULL,0x1a0036800905a48ULL,0x8000200000a00000ULL,0x0,0x0,0x0,0x401004800000ULL,0x0,0x4002004020000ULL,0x0,0x0,0x120000000010200ULL,0x0,0x2004000004800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa00008000100ULL,0x200000408000100ULL,0x0,0x0,0x400002000a00000ULL,0x0,0x0,0x0,0x0,0x0,0x8100202001080801ULL,0x0,0x0,0x0,0x0,0x30420000a60000ULL,0x88000000a00000ULL,0x8100000004002000ULL,0x0,0x8088101000800ULL,0x0,0x0,0x4000001000060000ULL,0x0,0xa00080001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000060000ULL,0x0,0x0,0x1a01220010e0801ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000120000080ULL,0x2200020000080ULL,0x200040000000050ULL,0x0,0x0,0x0,0x0,0x0,0x84021000000ULL,0x0,0x0,0x0,0x0,0x20040000000006ULL,0x0,0x240800200010c0ULL,0x8000002000400010ULL,0x0,0xc000000001040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4080000000082ULL,0x0,0x500024000000ULL,0xb00411560741040ULL,0x0,0x8001000400008ULL,0x0,0x1080000000400008ULL,0x7000000220108400ULL,0x0,0x61021021000ULL,0x800010021000000ULL,0x0,0x10001000100200ULL,0x0,0x14000600000ULL,0x80010000000006ULL,0x8010000000003000ULL,0x6000000400010ULL,0x100400000200040ULL,0x8042000020420080ULL,0x401000040040ULL,0x841002020600080ULL,0x284a44028101442ULL,0x8084640020805060ULL,0x802109023400088ULL,0x220c000020808000ULL,0x48000000000000aULL,0x1100040021001020ULL,0x0,0x10800001400ULL,0x0,0x0,0x840000000400010ULL,0x20000222000000ULL,0x0,0x80080000021000ULL,0x10400000001020ULL,0x0,0x4000000200001004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000003000ULL,0x0,0x5000000022000000ULL,0x50020008000ULL,0x6000000000001400ULL,0x0,0x100000200101000ULL,0x0,0x0,0x800400000000050ULL,0x4080000020000400ULL,0x140000001400000ULL,0x500020000200ULL,0x0,0x8000010000001400ULL,0x200400000000000aULL,0x0,0x0,0x0,0x0,0x0,0x8000000200200040ULL,0x0,0x0,0x0,0x0,0x0,0x4040000040040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000000021000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200080460000080ULL,0x0,0x480000000001020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200060000000ULL,0x2090004020420080ULL,0x4020820020000080ULL,0x6000000420000ULL,0x0,0x0,0x18020000080ULL,0x0,0x4000800000021000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200a8020008080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9044040144800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400000060000ULL,0x0,0x0,0x0,0x0,0x202000200800ULL,0x14008000800ULL,0x480000200800ULL,0x0,0x0,0xc00000008002ULL,0x0,0x4400140010c4000ULL,0x0,0x8200000002040000ULL,0x0,0x0,0x10100002000100ULL,0x0,0x804a40213100400ULL,0x4500000101804800ULL,0x4000000000800800ULL,0x0,0x480000004100000ULL,0x1100060002108002ULL,0x72822414130ec580ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000004100000ULL,0x200001004004000ULL,0x402240c2000800ULL,0x400004000800800ULL,0x0,0x2302002084800ULL,0x0,0x1000000400200800ULL,0xa95088300804ULL,0x400201080000ULL,0x4480a010021c0804ULL,0x4400012000000ULL,0x401000082000000ULL,0x0,0x0,0x500008000040ULL,0x0,0x400010001000800ULL,0x0,0x0,0x0,0x0,0x0,0x1000200012000000ULL,0x200101080000ULL,0x200000100004ULL,0x0,0x0,0x40040000040800ULL,0x2100002040000ULL,0x0,0x0,0x2200000000808000ULL,0x840000000040800ULL,0x0,0x8088c8001a042000ULL,0x42200801084800ULL,0x0,0x0,0x100040012000000ULL,0x4000008000040800ULL,0x0,0x6480042080800ULL,0x20004000008002ULL,0x100000282000000ULL,0x0,0x40001000014000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4280000800ULL,0x0,0x0,0x2200002100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42004004000ULL,0x6000000008000040ULL,0x0,0x811020200c084840ULL,0x0,0x81000008000040ULL,0x0,0x0,0x100002004004000ULL,0x8010a00000060000ULL,0x0,0x10100004104000ULL,0x0,0x0,0xa00004004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80004004000ULL,0x0,0x0,0x0,0x140002040000ULL,0x8010200004064000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4100000004000400ULL,0x0,0x0,0x220000000000000aULL,0x0,0x20040008000100ULL,0x4100000000120ULL,0x0,0x0,0x0,0x0,0x0,0x100810010200402ULL,0x0,0x0,0x500202000100408ULL,0x2400201000000aULL,0x20402000010ULL,0x2000800040000100ULL,0x0,0x50210080000300ULL,0x40200000104ULL,0x0,0x0,0x828000200100020ULL,0x0,0x0,0x0,0x8020000000000120ULL,0x0,0x20102000400ULL,0x850680228080501ULL,0x441405e010288613ULL,0x408008000400ULL,0x101000002ULL,0x0,0x0,0x240000008001ULL,0x0,0x0,0x0,0x0,0x0,0x400820000100ULL,0x110020000002ULL,0x8400000000005ULL,0x0,0x404040000020ULL,0x0,0x842000808000601ULL,0x10020000020040ULL,0x804008000100ULL,0x8020240020100400ULL,0x110000020000100ULL,0x60002000010ULL,0x0,0x0,0x8010804001000442ULL,0x0,0x0,0x40000800200010ULL,0x8400004000400ULL,0x0,0x0,0x100800002000010ULL,0x0,0x0,0x1000200000000408ULL,0x0,0x8020000001000002ULL,0x2000008001000002ULL,0x0,0x44021000a000110ULL,0x2004002000400ULL,0x0,0x4000400000120ULL,0x0,0x10001000000440ULL,0x0,0x0,0x41000000040020ULL,0x3084c002100030ULL,0x440000020000002ULL,0x300000080020ULL,0x0,0x0,0x480010000002ULL,0x0,0x0,0x240000100020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800002000000000aULL,0x0,0x0,0x100802000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000000aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8080000000101ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0000000000aULL,0x10100000000408ULL,0x0,0x0,0x0,0x200c0002000501ULL,0x4000400100020ULL,0x0,0x0,0x4100202002000501ULL,0x0,0x280002000400ULL,0x10100000000101ULL,0xa02000400ULL,0x0,0x200540000000509ULL,0x80000000408ULL,0x8082280202080501ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x208440002000400ULL,0x0,0x0,0x0,0x0,0x1200000002000010ULL,0x10000200000000aULL,0x801000000000000aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x860010810b00180ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400004000000280ULL,0x802004400000ULL,0x4800004040ULL,0x849010880400ULL,0x0,0x0,0x0,0xa042a000203020c4ULL,0x0,0x0,0x0,0x8010200000124200ULL,0x8020000000120000ULL,0x0,0x0,0x0,0x2000400000804000ULL,0x6000020000040ULL,0x5000804000ULL,0x0,0x0,0x0,0x0,0x0,0x6020200000ULL,0x0,0x5485202104e802e8ULL,0x4000800000022000ULL,0x0,0x0,0x0,0x42000084000ULL,0x0,0x0,0x0,0x40100000800400ULL,0x0,0x0,0x600200008a0008ULL,0x0,0x0,0x6000000020008ULL,0x1000002000020020ULL,0x0,0x140000000200040ULL,0x0,0x0,0x0,0x200200002040ULL,0x1400000000060ULL,0x0,0x8020000000280ULL,0x0,0x0,0x0,0x0,0x0,0x40400000804000ULL,0x9418400040010e4ULL,0x0,0x0,0x100800000028ULL,0x0,0xc000200004881000ULL,0x0,0x10100000004040ULL,0x0,0x0,0x8000100000000028ULL,0x0,0xa102000000006060ULL,0x0,0x8000004000002040ULL,0x12000000140ULL,0x2000e00001002c0ULL,0x10040000020008ULL,0x14000101000ULL,0x0,0x800002020000040ULL,0x2020004400000ULL,0x0,0x0,0x201004010000ULL,0xa00000084000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000000000060ULL,0x0,0x0,0x88000000084000ULL,0xc00000120000ULL,0x0,0x0,0x40d0740000b82060ULL,0x0,0x8100000000000280ULL,0x0,0x0,0x24000880000ULL,0x0,0x0,0x0,0x0,0x2800120000ULL,0x40000002040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000000060ULL,0x0,0x0,0x0,0x0,0x40000280000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10040000020ULL,0x0,0x0,0x0,0x0,0x0,0x4000000204080000ULL,0xc204000004a8ULL,0x50000000001400ULL,0x0,0x8034260220905030ULL,0x1000010c04040480ULL,0x4000000000100200ULL,0x141080084141400ULL,0x0,0x0,0x4080000000801000ULL,0x0,0x0,0x88810040000460ULL,0x4000001000040080ULL,0x0,0x8010000000000028ULL,0x0,0x21000104000ULL,0x0,0x0,0x0,0x80100000001008ULL,0x0,0x80400000401000ULL,0x300000000021ULL,0x20004000104000ULL,0x0,0x800004000801000ULL,0x0,0x0,0x80000001001ULL,0x0,0x60000000000021ULL,0x0,0x0,0x8c00200000d000ULL,0x0,0x160104040600ULL,0x2004000401000ULL,0x100002004000800ULL,0x0,0x0,0x108000001008ULL,0x1000031004040004ULL,0x0,0x800000204000004ULL,0x10001000401000ULL,0x2000800005000ULL,0x0,0x14000010020ULL,0x20080004080000ULL,0x80000404000004ULL,0x0,0x0,0x2800000000c000ULL,0x820000801000ULL,0x0,0x0,0x0,0x0,0x1002000001400ULL,0x2004020000020ULL,0x0,0x8000100000000040ULL,0x10020000000040ULL,0x0,0x0,0x0,0x0,0x24000801000ULL,0x0,0x8400000020000020ULL,0x1200000080100000ULL,0x8020000010040000ULL,0x0,0x0,0x0,0x0,0x408000000001001ULL,0x200000114000000ULL,0xc00080001000ULL,0x24000010000400ULL,0x0,0x0,0x8004008100000ULL,0x0,0x0,0x800040004000004ULL,0x0,0x0,0x20004000801000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000010000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2800801000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000800008100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2808104000ULL,0x800040000000440ULL,0x0,0x80080080100000ULL,0x24002000040ULL,0x204000004000ULL,0x8001000004000800ULL,0x0,0x0,0x800080040000020ULL,0x204000104000ULL,0x20080040000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500040000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21000000140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204220010000680ULL,0x100400000010040ULL,0x10000220040000ULL,0x8002001000040ULL,0x88000002400000ULL,0x8000010002400000ULL,0x11000004040ULL,0x1114001400a044ULL,0x80100002040ULL,0x0,0x280210080880200ULL,0xc00002400000ULL,0x80000100000140ULL,0x0,0x0,0x0,0x0,0x2080002400000ULL,0x0,0x0,0x0,0x80010000080ULL,0x4000400000002002ULL,0x0,0x0,0x20080010800140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x142810048000ULL,0x0,0x100002000400100ULL,0x0,0x0,0x0,0x100500002400600ULL,0x0,0x700800041002000ULL,0x10000680802000ULL,0x4002001800000ULL,0x1080500044000044ULL,0x6010000080ULL,0x0,0x0,0x0,0x40010001000002ULL,0x0,0x100000100840000ULL,0x180000800ULL,0x0,0x0,0x0,0x8000000110800000ULL,0x200004000400200ULL,0x440000010000080ULL,0x2000002040002000ULL,0x0,0x8400000010040ULL,0x0,0x0,0x20004002400000ULL,0x20004001000400ULL,0x0,0x800100000008040ULL,0x0,0x0,0x1000001080000040ULL,0x200010000400100ULL,0x50000080000040ULL,0x2100000000220ULL,0x0,0x200201000002ULL,0x0,0x0,0x0,0x0,0x0,0x200200040040000ULL,0x0,0x0,0x1000080000008040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000100040000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000800000210ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000800080002000ULL,0x10100001000002ULL,0x0,0x0,0x0,0x0,0x4000001040040000ULL,0x0,0x60080000040ULL,0x0,0x0,0x940800090002080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2130598089402222ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1100008004000ULL,0x6000008010ULL,0x0,0x0,0x10100000300000ULL,0x0,0x1002000000400ULL,0x2010040200000ULL,0x8000014080282000ULL,0x8000000100080100ULL,0x40001001000400ULL,0x0,0x4008012000000ULL,0x0,0x0,0x0,0x800004a080008810ULL,0x120a0090020100ULL,0x0,0x0,0x0,0x20040000020100ULL,0x8010000008008000ULL,0x0,0x8000480800202810ULL,0x0,0x4400000240000ULL,0x0,0x0,0x81000000080100ULL,0x800008002000100ULL,0x400400000080100ULL,0x40100000020100ULL,0x0,0x20090298004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40020000240000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1400100484080000ULL,0x100080012000000ULL,0x408000200400ULL,0x0,0x108180090000400ULL,0x0,0x0,0x0,0x0,0x0,0x211040008008500ULL,0x0,0x0,0x0,0x0,0x0,0x22000008008000ULL,0x0,0x0,0x0,0x800010090000000ULL,0x0,0x0,0x0,0x8000100000400ULL,0x480080000800ULL,0x0,0x8000004000002001ULL,0x0,0x21000084000ULL,0x80004c0080082400ULL,0x24001000400ULL,0x0,0x0,0x0,0x0,0x200020090000000ULL,0x0,0x1004000000008010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x208000008004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x400100000080100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2880100000ULL,0x0,0x0,0x0,0x0,0x0,0x280000880000ULL,0x0,0x6000080100000ULL,0x0,0x80100000000400ULL,0x0,0x8100000000880000ULL,0x0,0x40004090000000ULL,0x0,0x0,0x100000100080100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4480302000880000ULL,0x0,0x80000200200800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000100020ULL,0x0,0x0,0x0,0x0,0x0,0x80940000160020ULL,0x420000022000ULL,0x0,0x200000402000400ULL,0x0,0x800004000008400ULL,0x40800000400ULL,0x0,0x800040009000000ULL,0x0,0x0,0x0,0x1204000000022140ULL,0x1200000000100100ULL,0x10800022000ULL,0x0,0x6002000200800500ULL,0x0,0x500000020040ULL,0x0,0x0,0x0,0x0,0x0,0x800002000000440ULL,0x42000001800000ULL,0x0,0x0,0x108000000000140ULL,0x0,0x0,0x0,0x28000000500ULL,0x0,0x8002001000200ULL,0x0,0x80100000002200ULL,0x0,0x6008008000ULL,0x0,0x0,0x2001040800000ULL,0x0,0x0,0x80010000000140ULL,0x0,0x0,0x0,0x424008000000580ULL,0xc000280009800100ULL,0x8000062000062200ULL,0x8040020000020820ULL,0x800080000840000ULL,0x404000000020400ULL,0x0,0x9100004000120420ULL,0x50620000860120ULL,0x0,0x0,0x8200000000440ULL,0x0,0x0,0x0,0x0,0x0,0x80050000424020ULL,0x0,0x2000800001800000ULL,0x0,0x0,0xa00000000005000ULL,0x0,0x820001000200ULL,0x0,0x0,0x20200000000440ULL,0x0,0x80c49f0a00027642ULL,0x10441002800440ULL,0x2010000840040ULL,0x50000000140ULL,0x4000001020004000ULL,0x0,0x8002000000020400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x408000008008000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42002000400ULL,0x0,0x0,0x0,0x18002000400ULL,0x0,0x0,0x0,0x1080000000004002ULL,0x0,0x0,0x2d00000100100ULL,0x0,0x0,0x4100000008400000ULL,0x20020000022000ULL,0x0,0x0,0x0,0x6000022000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100002000400ULL,0x0,0x0,0x140022002004402ULL,0x208040002004402ULL,0x0,0x500000022000ULL,0x81000002000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000020040000040ULL,0x0,0x0,0x201044000000ULL,0x8200000002000400ULL,0xc00030000000ULL,0x0,0x0,0x0,0x0,0x84010104a000000ULL,0x20004048000000ULL,0x80408448040004ULL,0x0,0x800040002800000ULL,0x0,0x0,0x44000842004ULL,0x0,0x808052042040460ULL,0x164040210180450ULL,0x14082000000ULL,0x0,0x4002820044100580ULL,0x2010000030000000ULL,0x0,0x0,0x0,0x0,0x80010000080800ULL,0x140000020800ULL,0x0,0x0,0x0,0x0,0x0,0x6000000040000040ULL,0x44000012000000ULL,0x8000248000000ULL,0x100804000240404ULL,0x0,0x408000040004ULL,0x400000804100000ULL,0x0,0x0,0x80004000100000ULL,0x0,0x60648420420c0ULL,0x200000500ULL,0xc000000040040ULL,0x0,0x8010000000100000ULL,0x0,0x0,0x80040040000040ULL,0x200040000000420ULL,0x0,0x1800000000420ULL,0x2100002002000ULL,0x4000002a00102800ULL,0x0,0x6104082800000ULL,0x0,0x0,0x0,0x0,0x0,0x8000400000000402ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000000200400ULL,0x1200000040000400ULL,0x0,0x0,0x100400000080400ULL,0x48002000040ULL,0x100080030000000ULL,0x120002040000ULL,0x0,0x0,0x120000000020800ULL,0x0,0x0,0x0,0x0,0x44001002000ULL,0x0,0x0,0x80400020100000ULL,0x8000000240000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20004000082000ULL,0x0,0x0,0x0,0x0,0x10400000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000400030000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000000002800ULL,0x0,0x8000200000002800ULL,0x0,0x0,0x0,0x800088104100800ULL,0x0,0x0,0x8100000000002800ULL,0x0,0x0,0x400010000000420ULL,0x0,0x0,0x0,0x10020000080400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810000044000000ULL,0x0,0x0,0x0,0x0,0x0,0x80000000800ULL,0x8040000200400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802090000000ULL,0x630024200a300ULL,0xa000000082000000ULL,0x11000000000a00ULL,0x0,0x0,0x100020000002400ULL,0x80800040000002ULL,0x0,0x0,0x200180004000ULL,0x40001002800000ULL,0x0,0x820000000000a00ULL,0x4200282040004ULL,0x480060008a000100ULL,0x8000000208000100ULL,0x0,0x0,0x0,0x0,0x24002282801000ULL,0x7800082810000ULL,0x0,0x0,0x0,0x0,0x41000008000100ULL,0x0,0x0,0x0,0x0,0x400000802000004ULL,0x0,0x440000000002400ULL,0x200088000000ULL,0x0,0x8000400008000008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x48008400000ULL,0x100000800040200ULL,0x20040000000300ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80010000002000ULL,0x400004008000020ULL,0x800030802800110ULL,0x50021088081000ULL,0x1004042090004020ULL,0x20080008000008ULL,0x0,0xc00002000004ULL,0x0,0x0,0x200100008200ULL,0x0,0x0,0x0,0x80040010000100ULL,0x8000000000002020ULL,0x0,0x4020010000100ULL,0x0,0x0,0x0,0x0,0x8000001000002001ULL,0x0,0x0,0x0,0x2004000200002ULL,0x0,0x40200080000002ULL,0x0,0x0,0x100000880001000ULL,0x0,0x2080002000004ULL,0x0,0x0,0x0,0x0,0x80000202000004ULL,0x0,0x0,0x24082000000ULL,0x0,0x0,0x180002000ULL,0x5000800002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2000020000002020ULL,0x0,0x0,0x0,0x4000010088000000ULL,0x4000200040002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x28000002000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000200080000040ULL,0x0,0x200000402000004ULL,0x0,0x6010040080810040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80004000810000ULL,0x0,0x0,0x28000090000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x104c0080008240ULL,0x80100008000008ULL,0x2010000000800002ULL,0xc800400a8210c200ULL,0x840000000810000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000000208ULL,0x0,0x0,0x40040050000000ULL,0x800020001004000ULL,0x0,0x0,0x1000540080202020ULL,0x0,0x0,0x0,0x10020000008008ULL,0x0,0xc000000000108ULL,0x0,0x80400000400100ULL,0x840000000008100ULL,0x2000200400020ULL,0x500000004080ULL,0x10000000018000ULL,0x0,0x140000082000000ULL,0x10044000008218ULL,0x1000044400404022ULL,0x0,0x800000800000208ULL,0x120140404020ULL,0x0,0xa050000000ULL,0x0,0x200000482000000ULL,0x0,0x110000000108ULL,0x20400040004000ULL,0x0,0x0,0x0,0x0,0x0,0x20004004004000ULL,0x0,0x8000902080408102ULL,0x0,0x0,0x0,0x0,0x0,0x80080000400020ULL,0x0,0x0,0x0,0x8000400000c000ULL,0x0,0x0,0x8000002000040020ULL,0x400100800001ULL,0x0,0x400000808004000ULL,0x2101403c0004004ULL,0x0,0x0,0x0,0x840000040001000ULL,0x0,0x1001000000400100ULL,0x820000004008ULL,0x10400004080ULL,0x0,0x2542000001300220ULL,0xc00000800004ULL,0x0,0x240000004008ULL,0x0,0x1800040004000ULL,0x0,0x2004000000800001ULL,0x810000000108ULL,0x0,0x0,0x210000000400100ULL,0x60000002020ULL,0x2004080000002ULL,0x81000060c0000042ULL,0x0,0x2000008000004080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4000001000000060ULL,0x800040000000060ULL,0xc00000000102ULL,0x0,0x0,0x500000808000ULL,0x80000008100ULL,0x0,0x0,0x0,0x8000002000200200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8010000000020200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000002000400020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81000000000102ULL,0x6000000000000102ULL,0x0,0x840026800a1c200ULL,0x0,0x20020000200200ULL,0x0,0x0,0x10400000000060ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21000000060ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810400001100102ULL,0x500000000102ULL,0x29101164020c260ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202000000c000ULL,0x0,0x0,0x0,0x284000080803000ULL,0x0,0x0,0x0,0x840000000800002ULL,0x20400018000ULL,0x100040020040000ULL,0x0,0x0,0x0,0x1080000080000000ULL,0x0,0x0,0x0,0x4100000001448000ULL,0x0,0x802001040000ULL,0x80000108400000ULL,0x206080002002082ULL,0x14000040200ULL,0x4002000002020ULL,0x0,0x0,0x0,0x0,0x820000040004ULL,0x0,0x4000002000500000ULL,0x0,0x0,0x40410020042004ULL,0x4000008000020001ULL,0xc00000002002ULL,0x8000180000001ULL,0x0,0x0,0x210000002000002ULL,0xc00001004000ULL,0x2100000001000200ULL,0x0,0x0,0x4800020008ULL,0x20401008000ULL,0x0,0x100001008400000ULL,0x0,0x0,0x0,0x0,0x8002000002000002ULL,0x1000000400020020ULL,0x0,0x200010001000200ULL,0x8000d00000042220ULL,0x204000000008200ULL,0x10001080200000ULL,0x200100020040ULL,0x0,0x0,0x0,0x8000002020040000ULL,0x0,0x0,0x2020001000200ULL,0x0,0x420000500000ULL,0x10400000002002ULL,0x10002000000c000ULL,0x0,0xc0000020020ULL,0x0,0x0,0x1000004002002000ULL,0x20004020040000ULL,0x802000000002080ULL,0x0,0x10002000020040ULL,0x50001000200ULL,0x4000200000002080ULL,0x0,0x2080800000000102ULL,0x48080001000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa00101000ULL,0x0,0x44000500000ULL,0x8000010001000200ULL,0x1000020020042002ULL,0x8010000008100ULL,0x0,0x0,0x0,0x0,0x600101000ULL,0x0,0x0,0x0,0x104000000008100ULL,0x0,0x140000800002ULL,0x0,0x0,0x0,0x480000c000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20104a080100a122ULL,0x0,0x4000001000002080ULL,0x0,0x8004000000020008ULL,0x0,0x0,0x200002000000c000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x102000000500000ULL,0x0,0x440008b00040102ULL,0x0,0x0,0x8290000601109102ULL,0x840000000008200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4020006000000ULL,0x0,0x0,0x0,0x200010400460200ULL,0x0,0x6000000a00000ULL,0x1002004800000ULL,0x0,0x0,0x0,0x4800115004c12a00ULL,0x0,0x0,0x0,0x0,0x80000100022000ULL,0x8002004000400a00ULL,0x0,0x0,0xc00010400000ULL,0x0,0x80800040800000ULL,0x0,0x0,0x0,0x0,0x480010400000ULL,0x0,0x844000805002800ULL,0x4002000000a00ULL,0xc00808081200ULL,0x40800000a00ULL,0x1004000000060000ULL,0x8200000000201ULL,0x0,0x80000400020010ULL,0x0,0x0,0x0,0x4800000000000220ULL,0x5000012001822000ULL,0x0,0x0,0x8410200080420800ULL,0x0,0x0,0x60001800000ULL,0x0,0x10005c4006000201ULL,0x0,0x0,0x2000800000c00ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x420000000005000ULL,0x6001000800ULL,0x0,0x0,0xa10000100a00201ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x400100000000201ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8000020000420000ULL,0x0,0x0,0x0,0x0,0x1000080004002000ULL,0x8000800000404000ULL,0x2000800000404000ULL,0xa030440040101c80ULL,0x5000204040822000ULL,0x0,0x0,0x4140000400208ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x84000800040ULL,0x0,0xc002304040401e28ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400004006000000ULL,0x1800004002000ULL,0x2100000000400200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82010400000ULL,0x2000040000010200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10400000600ULL,0x0,0x0,0x0,0x0,0x0,0x8000100a00000ULL,0x0,0x0,0x0,0x0,0x100800000400200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000a00000ULL,0x0,0x0,0x280000600a60000ULL,0x840200240a60004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd08010420200ULL,0x0,0x8010000000001040ULL,0x4800000000080001ULL,0x0,0x400000821000000ULL,0x0,0x44008000080ULL,0x82000000000aULL,0x0,0x0,0x2000008020008000ULL,0x900208000000008aULL,0x100100000082ULL,0xc3060220200000deULL,0x8010000000081000ULL,0x1004000000000aULL,0x4430042000440050ULL,0x484100000040041ULL,0x4080000008020000ULL,0x0,0x0,0x0,0x0,0x201000400010ULL,0x0,0x40000800001200ULL,0x0,0x8000002020000008ULL,0x0,0x0,0x808668202501201ULL,0x0,0x80100000800001ULL,0x300000021000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000000003000ULL,0x80140001000ULL,0x0,0x2080000600000ULL,0x0,0x20100021000ULL,0x0,0x0,0x100100040001000ULL,0x0,0x44000020020000ULL,0x0,0x0,0x0,0x0,0x2280008000101ULL,0x200000800001200ULL,0x0,0x202402000000004eULL,0x410000002000001ULL,0x408000000000202ULL,0x0,0x0,0x4000208000800ULL,0x80002000001ULL,0x0,0x942030040441001ULL,0x108000000021000ULL,0x8200008000100ULL,0x0,0x0,0x41000008000080ULL,0x0,0x0,0x0,0x0,0x4000100000000aULL,0x0,0x20400020000008ULL,0x20808000100ULL,0xf218257069083043ULL,0x8000100000000050ULL,0x0,0x44000000050ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2080008000001ULL,0x4000240001000ULL,0x80000120000080ULL,0x0,0x4000008040001000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000000000006ULL,0x0,0x0,0x0,0x0,0x0,0x280000000000082ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000010220020080ULL,0x0,0x0,0x0,0x100002000420000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24040000001ULL,0x0,0x208140000000005aULL,0x0,0x8000100003000ULL,0x20004000001040ULL,0x4002000001200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000600000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x101000101000ULL,0x628440000421000ULL,0x4000468604200c0ULL,0x885400084a0101ULL,0x4000001000080001ULL,0x482c80020420080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1014080022840000ULL,0x0,0x0,0x0,0x0,0x8003002000801c00ULL,0x0,0x882100a02101400ULL,0x0,0x0,0x0,0x0,0x70002a0422101054ULL,0x0,0x0,0x0,0x0,0x40200000044000ULL,0x100001024000000ULL,0x200000200001020ULL,0x800010060000000ULL,0x0,0x0,0x110000000001400ULL,0x0,0x0,0x200040000044000ULL,0x0,0x800010000004040ULL,0x0,0x200401802108002ULL,0x24000200040ULL,0x0,0x2010000000100004ULL,0x100040000044000ULL,0x2002000060000000ULL,0x30000101000ULL,0x0,0x200010022000000ULL,0x0,0x0,0x0,0x404000000240000ULL,0x20080024000000ULL,0x0,0xc0801000004840ULL,0x0,0x0,0x0,0x0,0x480020800000ULL,0x0,0x16000300a00ULL,0x0,0x0,0x0,0x140024000100404ULL,0x0,0x200100200800ULL,0x14000004800ULL,0x0,0x400200000000040ULL,0x1400000800800ULL,0x0,0x0,0x0,0x0,0x1400008040000ULL,0x0,0x4002000200800ULL,0x0,0x0,0x0,0x1000800000060000ULL,0x0,0x100001000080800ULL,0x0,0x0,0x0,0x880050000008052ULL,0x5600000020800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x300000008002ULL,0x1400000001020ULL,0x0,0x0,0x0,0x0,0x8080000200040ULL,0x0,0x4400000000004800ULL,0x0,0x0,0x480000044000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1180040020001404ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000200800ULL,0x0,0x0,0x700001101800ULL,0x11000200040ULL,0x0,0x0,0x0,0x0,0x0,0x10100000080800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108004e006160000ULL,0x4042060004800ULL,0x0,0x4050061024960000ULL,0x400400060000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810000000101ULL,0x1000080008000040ULL,0x0,0x0,0x0,0x2400000c000ULL,0x0,0x40200000000104ULL,0x804000008200000ULL,0x200010000000104ULL,0x0,0x2000040008000400ULL,0xc00000000120ULL,0x0,0x0,0x1004048042200010ULL,0x8010000008080ULL,0x0,0x8000000100808000ULL,0x1004020000400ULL,0x6000000201ULL,0x0,0x804002808014100ULL,0x44000000000104ULL,0x0,0x200040000008001ULL,0x0,0x0,0x0,0x0,0x80400002000010ULL,0x0,0x80000008080ULL,0x0,0x0,0x21002000010ULL,0x0,0x4804092009000702ULL,0x0,0x8000000201004000ULL,0x0,0x0,0x4000000200808000ULL,0x0,0x200000000120ULL,0x0,0x400100040000100ULL,0x100080080000100ULL,0x10400000012ULL,0x800010000000aULL,0x2000800003000000ULL,0x80100000000440ULL,0x4080000002000010ULL,0x1000002008002000ULL,0x0,0x0,0x24004004000ULL,0x1000040002000100ULL,0x0,0x0,0x1000020040000100ULL,0xa02000100ULL,0x200201000002ULL,0x0,0x0,0x24000000000101ULL,0x2000010002000010ULL,0x500012000000ULL,0x0,0x0,0x20000400000c000ULL,0x0,0x0,0x0,0x0,0x1000020000000012ULL,0x0,0x2008012000000ULL,0x5080001002200110ULL,0x42004004000ULL,0x20203000000ULL,0x0,0x0,0x0,0x1080000000000012ULL,0x100040020000400ULL,0x0,0x6001004000ULL,0x0,0x0,0x8000080000808000ULL,0x0,0x0,0x0,0x20040000008001ULL,0x0,0x0,0x1200808000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42082000000ULL,0x110c42000808180ULL,0x100080000000101ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10020001000002ULL,0x0,0x0,0x4000820801014000ULL,0x0,0x10020000200010ULL,0x0,0x18000200010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800010000000101ULL,0x0,0x0,0x0,0x0,0x0,0x20040080000100ULL,0x110402000014101ULL,0x0,0x42000008040ULL,0x100800020000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080080000020ULL,0x210020000980000ULL,0x20040000880000ULL,0x0,0x42020000002ULL,0x200200002040ULL,0x0,0x400010000820000ULL,0x0,0x80000200020040ULL,0x0,0x0,0x0,0x404201000004840ULL,0x0,0x200000200080002ULL,0x4066080000804228ULL,0x1410088200988020ULL,0x0,0x0,0x8000020010000400ULL,0x0,0x0,0x0,0x50000004200ULL,0x4000100000200400ULL,0x0,0x0,0x0,0x0,0x4002000020100ULL,0x840000080020000ULL,0x20000080000000aULL,0x0,0x2000002000002040ULL,0x0,0x8010000000084000ULL,0x0,0x0,0x0,0x4804002010300400ULL,0x8010000004001000ULL,0x0,0x480080000400ULL,0x0,0x200010000000440ULL,0x0,0x0,0x0,0x0,0x200000040020ULL,0x0,0x0,0x0,0x2020020000040ULL,0x100800000004800ULL,0x0,0x8000800000200400ULL,0x0,0x48000800400ULL,0x24000000280000ULL,0x4046020204800ULL,0xd000800000880400ULL,0x0,0x8010080000400ULL,0x0,0x4810000400ULL,0x0,0x0,0x410000000080002ULL,0x0,0x200200000000140ULL,0x0,0x0,0x828008004011000ULL,0x0,0x0,0x800040080000020ULL,0x500010000400ULL,0x12000006002800a0ULL,0x40804010000ULL,0x0,0x28002000090000aULL,0x0,0x0,0x22000020200000ULL,0x0,0x0,0x2080000804000ULL,0x0,0x420000011000000ULL,0x184804210904400ULL,0x800044000b80000ULL,0x5000000000820000ULL,0x840000000800400ULL,0x0,0x40040001200000ULL,0x0,0x20400040020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x904000200006060ULL,0x140014003200400ULL,0x0,0x0,0x800040004000100ULL,0x0,0x2000800040000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x832c008880a0ULL,0x4856d2181499446aULL,0x6000000000002040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080000820000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x448400000806040ULL,0x40040000084000ULL,0x30300000024300ULL,0xc800674040084362ULL,0x4004001204012040ULL,0x654040000008400aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc008140100503040ULL,0x24000040000020ULL,0x0,0x0,0x0,0x0,0x1004000001001ULL,0x0,0x200100000040ULL,0x0,0x0,0x60000000000180ULL,0x2010040000020ULL,0x0,0x0,0x144800044000024ULL,0x0,0x0,0x4000800005000ULL,0x200114000000ULL,0x0,0x14000001080ULL,0x0,0x40058000304100ULL,0x8002080000000280ULL,0x0,0x0,0x10400000020020ULL,0x20400040000020ULL,0x0,0x0,0x11000020020ULL,0x42000040000020ULL,0x8000200000084ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802022002004040ULL,0x140000000040ULL,0x28000040000020ULL,0x0,0x12004080000ULL,0x20400010020ULL,0x2002000001400ULL,0x0,0x400200040000020ULL,0x80010000004000ULL,0x0,0x8000400000101000ULL,0x8000040040000020ULL,0x0,0x0,0x100400000401000ULL,0x2004000000001400ULL,0x0,0x0,0x20080080020000ULL,0x0,0x0,0x80000100020020ULL,0x0,0x0,0x0,0x0,0x50000000020004ULL,0x0,0x6000802004000980ULL,0x80010000022000ULL,0x42000120000ULL,0x0,0x200020000001080ULL,0x1080000000000084ULL,0x0,0x4000400020000020ULL,0x0,0x0,0x48002000040ULL,0x1004200010020ULL,0x240000001008ULL,0x0,0x100000084ULL,0x0,0x0,0x0,0x2004000000040ULL,0x0,0x0,0x0,0x0,0x0,0x60004000004ULL,0x100001000300000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x805200400c800ULL,0x0,0x50000000000028ULL,0x0,0x0,0x0,0x0,0x100a00200005000ULL,0x0,0x8002000001008ULL,0x0,0x2004000010020ULL,0x0,0x0,0x0,0x100000800401000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x210000004000800ULL,0x0,0x0,0x0,0x8000000000110000ULL,0x0,0xce22c20b241b42a4ULL,0x1014880000502dULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4042024080020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000100000404000ULL,0x0,0x0,0x0,0xd1a00c40044814e4ULL,0x8100030000104440ULL,0x2080000104000ULL,0x65540c0060120060ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000040008ULL,0x200400050000000ULL,0x0,0x9084000005480ULL,0x0,0x0,0x0,0x0,0x0,0x100010020040000ULL,0x0,0x4000001004000040ULL,0x0,0x0,0x1000080040002000ULL,0x0,0x202080100000ULL,0x0,0x181000000084c0ULL,0x0,0x0,0x8042644180100250ULL,0x0,0x480000000100800ULL,0x0,0x0,0x0,0x6001000400ULL,0x1800000040002ULL,0x0,0x0,0x8040000040008ULL,0x0,0x1800000000600ULL,0x0,0x420000004040ULL,0x100040041040040ULL,0x2010000040000004ULL,0x6000000000801000ULL,0x0,0x0,0x820008100000ULL,0x0,0x0,0x0,0x14050000000ULL,0x0,0x0,0x0,0x1800001000400ULL,0x0,0x0,0x4020000000040008ULL,0x100001020040000ULL,0x0,0x0,0x4008000000240ULL,0x0,0x210000004040000ULL,0x0,0x8001040040000ULL,0x0,0x0,0x1020084054040000ULL,0x0,0x840000050000000ULL,0x0,0x0,0x0,0x1800000100200ULL,0x0,0x10400040000001ULL,0x0,0x800050080100000ULL,0x0,0x0,0x2200000000040008ULL,0x88001000002ULL,0x0,0x42000000600ULL,0x4000040010008000ULL,0x800010050000000ULL,0x40040000840000ULL,0x500000400400ULL,0x0,0x0,0x80804051000400ULL,0x0,0x0,0x4000001000000408ULL,0x0,0x4080000008040ULL,0x1280800001802040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x80010040000ULL,0x42010040000ULL,0x8100000000408ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2050208101008ULL,0x0,0xc302840204040640ULL,0x8000000200001008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc00840000ULL,0x0,0x0,0x0,0x0,0x0,0x2220210008140ULL,0x80101482050004d0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802010000400ULL,0x0,0x0,0x0,0x0,0x0,0x480000000600ULL,0x540010010040400ULL,0x0,0x0,0x10450000000ULL,0x0,0x0,0x0,0x0,0x0,0x4000410800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1000080004040000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1302000041082ULL,0x0,0x44010008000ULL,0x400220040000ULL,0x8181004050803080ULL,0x1068000100040c0ULL,0x8040a19100082ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x5ac0326900006540ULL,0xffb777f318a06570ULL,0xd8000f2a0080a530ULL,0x42021800000540ULL,0xa40008a00004160ULL,0x442610e008804560ULL,0xa440036800002720ULL,0x193400f808000560ULL,0x1e83c24000a02560ULL,0x940401000a00520ULL,0x843201e000002560ULL,0x1000a0080000340ULL,0xd080021000004140ULL,0xc141957cd8202760ULL,0xc14000000540ULL,0x1008098108400500ULL,0xabc5842080006720ULL,0xc006212000400100ULL,0x8840110800002520ULL,0x8002101080000140ULL,0x28a4446a00006500ULL,0x1180126708006520ULL,0x4020258000000160ULL,0x8010962000006520ULL,0x41080800002520ULL,0xab10ca6008a04520ULL,0x607214a608002520ULL,0x8200000000120ULL,0x2908441010802540ULL,0x780f00000004520ULL,0x841585800008530ULL,0x14b0540000204500ULL,0x950eb53408006720ULL,0x8000518008000400ULL,0x4885d22400802520ULL,0xe5383c4008a06501ULL,0x8100000400ULL,0x514b852800a04520ULL,0xa1223e4408202400ULL,0xc00000200400ULL,0x80064a1808002520ULL,0x5986d4a88802540ULL,0x2211000080320ULL,0x4984196008000520ULL,0x8390f24200c04760ULL,0x8004b4100002500ULL,0x588001000000120ULL,0xd00aa20c12200620ULL,0xb02444980006d40ULL,0xabecd2080900e520ULL,0x4402000800200520ULL,0x20e048100000520ULL,0x77020acc0a400560ULL,0x9700ed5300a06520ULL,0x8214100008002500ULL,0x28cc181100206500ULL,0x961cb5090802500ULL,0x4474267008800760ULL,0x1400e1000a04500ULL,0x70000510c0000500ULL,0xc08c00400000520ULL,0x940814008080720ULL,0x132a04a688800560ULL,0x80c0040008000400ULL,0x2032145008802500ULL,0x7903c24008800500ULL,0xc2881008200400ULL,0x2003492000200540ULL,0x938180400020a500ULL,0x8530a64008006540ULL,0x4204830201002560ULL,0x2150176000006520ULL,0x1001848300004520ULL,0x8950964000082720ULL,0x16007008002400ULL,0x8080006480002700ULL,0x28401e8000202520ULL,0x202040200000520ULL,0x313a064a08206500ULL,0x10083000000520ULL,0x0,0x100d00000008500ULL,0x804100000330ULL,0x1421d42000804500ULL,0x9092359392202560ULL,0x9aba0d5008a06560ULL,0x44d3b70a08202530ULL,0x1484000000400ULL,0x142404e408002504ULL,0x8124062080802520ULL,0x0,0x202680000208510ULL,0x150824a00000520ULL,0x2580013200006500ULL,0x8000000008400000ULL,0x0,0xc8a228020020a740ULL,0x0,0x8820050888200d00ULL,0x450012080004300ULL,0x18000004100ULL,0x46000000400ULL,0x802000000000120ULL,0x2024010000804100ULL,0x0,0x0,0x0,0x0,0x4400000002400ULL,0x0,0x799e557198a06560ULL,0xa3383f6000a06564ULL,0x180800000200520ULL,0x14040400400500ULL,0x0,0x0,0xc00010800000500ULL,0x0,0x0,0x12050008000400ULL,0x4004000000000400ULL,0x4000200000000120ULL,0x0,0x2400101000000400ULL,0x104410000002500ULL,0x400026000802500ULL,0x0,0x3000000000000400ULL,0x100100000400ULL,0x0,0x0,0x80008000400ULL,0x50284c0200202400ULL,0x0,0x6a96ce9c90a02560ULL,0x80040000000400ULL,0x0,0x0,0x0,0x200004000800100ULL,0x1014526800004520ULL,0x20100f0000002520ULL,0x280200100002400ULL,0x0,0x20100000120ULL,0x11000000000120ULL,0x6000220000002520ULL,0x8240400000002500ULL,0xe316cf2e08a06560ULL,0x0,0x48000004500ULL,0x1080104008000400ULL,0x50008000400ULL,0x8300d00010000130ULL,0x0,0x2000038008000400ULL,0x0,0x40000400200ULL,0x0,0x2000800800100ULL,0x40004000200400ULL,0x444040000002402ULL,0x0,0x0,0x104800400000520ULL,0x0,0x3008550208000500ULL,0x200240100000520ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800220788412440ULL,0x0,0x10c0064090808010ULL,0x4800000000008200ULL,0x8441308000a014ULL,0x0,0x800480000004720ULL,0x2800000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xff7df77b8f3eff67ULL,0xfffefff7fffeff73ULL,0xdebc6ffe9ffc6f67ULL,0xc1643402c0086462ULL,0xa31bd5508d7c2d23ULL,0xf4b6f6fc89acff73ULL,0x9d6ffe638aace573ULL,0xfb9e7ffaef6eff71ULL,0xcfd6fcafcf6eec72ULL,0x3bf7369bdb3c6d23ULL,0xc0def3b1673eaf60ULL,0x51aebe78b864f673ULL,0xaaa24d64cf366f61ULL,0x8bc4f3e54b686760ULL,0xe1e59ae80aac5513ULL,0xbefb5fbc8f7eff62ULL,0xd4e1f74a2ffe7f73ULL,0xf6767ffcef6e7573ULL,0xcbf1f3efefec7f73ULL,0x6af39fdadd6c7d67ULL,0x9e9fffffef46f723ULL,0xf7feb6ee75627e63ULL,0x697c4dc13bbc7766ULL,0xeb00e76e1a3c3566ULL,0xe4fab35901ec7762ULL,0x2ffedf7afffcf772ULL,0xf59e5efb4baa7767ULL,0xdfb8ff736dfc7f67ULL,0xff947e76dfee7722ULL,0xedf7fe1e89ec7f63ULL,0xab5ecfcb2f7a6763ULL,0x9cf7fee9d9eefe60ULL,0xf59bff7afb6c7f72ULL,0xcbdaffbecffcfc60ULL,0xcdd5ff1e8ffc6d35ULL,0xeefefddbdf6eff63ULL,0xbcb67ffbff6ef763ULL,0xd1f5ae680b6e7f63ULL,0xdfc62d7dfb6e7f77ULL,0xc7eaf2ab3b6e4767ULL,0xb5d5fdf93aefd53ULL,0xb7dfddffdbbc7f77ULL,0xfd9cd860492c4f62ULL,0xbf3c5d2c9238f521ULL,0x5bc5f32fdf6e7d62ULL,0xedfb7f5bdbec7f60ULL,0x91bacd4325eef771ULL,0xd5f7feffcbaef565ULL,0x9ea61de14bc87762ULL,0xf3c6eb368bac6e63ULL,0x9ecb3f51cf2e7661ULL,0x1f9beffb7fec7e70ULL,0xd7b226680b0a4765ULL,0xb771ff70db3e7766ULL,0xfb9cff3b3f487d73ULL,0x63ce763f933e7722ULL,0x665ecedf87fc6563ULL,0xf4befe7dfb7e7732ULL,0x95ffbf7e9feefd65ULL,0xcfae777fabf87763ULL,0xc9d6fe74cf6e6e70ULL,0x7a30061aca404c20ULL,0x8491c4cb287c6e20ULL,0x51b4fee8cf3e2740ULL,0xf7922f539f647562ULL,0xeffaff3febec7766ULL,0xeffe9efe8f4e7573ULL,0xbf3edffbbbfcf777ULL,0xb67d37788a746503ULL,0x8af6755e01ac3e20ULL,0xcbd4f39babceef73ULL,0xe7fffefbdbee7766ULL,0x97ac4de217acef62ULL,0xfff67ebefbae7763ULL,0x4aac7da31bae6e62ULL,0xf974e33ecf4a7f27ULL,0xe8cffe937ffc7d63ULL,0x2b3e4f7ab36e7d61ULL,0x77f776ff7f7c7d76ULL,0x81c8f5674a2cfe63ULL,0xd2994a62812e6430ULL,0x5fdffebd4b7eff73ULL,0x2ffedfd9df7afe32ULL,0x76b4f6e0cf7cec66ULL,0xc9ba7522fb2a2563ULL,0xff9ebe49db7e6572ULL,0xcad7be3359ece663ULL,0xa5c4f1c132e7665ULL,0x75b377e7df6cff63ULL,0x814885201a304443ULL,0xcbbabfe2ba6c6767ULL,0xf2cfda625bfc7555ULL,0x2bfdcfd18b6ea560ULL,0xf77a67bf8a6ced73ULL,0x31c8702080042565ULL,0x70a6387a036a7742ULL,0x4a84e3aa0bea6722ULL,0xe49b3cd74b28e721ULL,0xf50e75282643d60ULL,0x8141106a35082560ULL,0x40a8055012004742ULL,0xf993ff9fcb6eef63ULL,0x830014a000400ULL,0xc0ba4568c122fe71ULL,0x700206acc8006e20ULL,0x2034000008700000ULL,0x12002000100ULL,0x0,0xd6b545e11085e21ULL,0x0,0xefc67ff1db7c6766ULL,0xfab02f758a7c6761ULL,0x100400010000400ULL,0x0,0x205000000ULL,0x0,0x1800220580006822ULL,0xd34409260a7c0d62ULL,0x8b5e5ac02b7c7760ULL,0x8990c02915003462ULL,0x53a8054013083460ULL,0x8a207d62083c7022ULL,0x8138d19a82686660ULL,0xce77e558b4e6766ULL,0xb1bedfeb6ffc7f64ULL,0xf535f7e24bbc7f62ULL,0xe84118010302402ULL,0x8152802c48244042ULL,0x214e120f106ced22ULL,0x308c400018a0000ULL,0x100800000401ULL,0x950000884062ULL,0xc900d5850a4c4c20ULL,0xc42120818884500ULL,0x4f1ddf5a93eef763ULL,0x80188020404020ULL,0x21419b0048487100ULL,0x4000202018400400ULL,0x41a30020b0a4520ULL,0xe936df43daacf572ULL,0xb4bffff77ffe7f63ULL,0xd114f5cfca486760ULL,0x8ba22d4250ec6602ULL,0x9c4020402080501ULL,0x620cd0814f402920ULL,0x9c3a696228c84432ULL,0xd484f24000383160ULL,0xc866354019247420ULL,0xc350ffbb0b3e7f62ULL,0x24931a78b2646624ULL,0x414000002000ULL,0x404004e058402400ULL,0xee8e65600fec7620ULL,0x990e122c9ec7001ULL,0x826adb5e0b4c7f66ULL,0x693eddfb4ffee543ULL,0x71c05a0869487431ULL,0x40200002000400ULL,0x80c77c0b664523ULL,0x8542f21a4b6e4c00ULL,0x108000048000000ULL,0x2024481082000ULL,0x16876010942f402ULL,0x2100c5409420e00ULL,0x2004010208084002ULL,0x2081dfc00b425f21ULL,0x882601ca08384506ULL,0x1148054811280e20ULL,0x8380040080602440ULL,0x800100020000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080000080100ULL,0x8212010c488c0020ULL,0x0,0x81c000880402ULL,0x5426004080882400ULL,0x220244080646000ULL,0x0,0x0,0x8264b460a405d70ULL,0x81b0014882002541ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x14001800000ULL,0x4bf53684e983d40ULL,0xf5f7ffffafbefd68ULL,0xffbebf71c7bb3f60ULL,0xffdeffef5efeff60ULL,0xbfd6febecf9a3f68ULL,0x675dcf47d7b23d68ULL,0xf6feffff7fbf7d60ULL,0xddf7ff7b5eb27f60ULL,0xbb2235785e9eac60ULL,0xfd57ef38df9af760ULL,0xab7d5d5b9f9ebf40ULL,0xbcfb5de9dabb3f68ULL,0x51eeff735fbaff60ULL,0xdaa42f431a966d60ULL,0xe54ef12a1f9e2460ULL,0x6f7fbe5b4faa2f60ULL,0x9ef75ffe17b67f40ULL,0xfc46fceb871f6d68ULL,0x7fbaaf7956fe6740ULL,0xc177fb0a5eb20d60ULL,0xaadc195a67962f40ULL,0x2accffb0e3f7d48ULL,0x9662f6f413bf4f20ULL,0x5bee7dfadfbe6f60ULL,0x8b18e2678e9e7c40ULL,0x8eebbadd2bba7e60ULL,0xe7becfdadfbf7d60ULL,0xd1b7feff87fb5f68ULL,0xcb2bffe01ef65f60ULL,0xffbe7f7d7fff7f60ULL,0xefeeff3bb79b7f60ULL,0xaddacddf4f9e3d68ULL,0x74bdd6eaff966d68ULL,0xb18e5f7258821d40ULL,0xcf1c677edab67c68ULL,0xcdf7ff9e7fbe6d40ULL,0x237f5a504af76f48ULL,0xb57b7fe90fb66760ULL,0xbbe1ee7452baad60ULL,0xfe9fafff0ff23d68ULL,0x61d2f38b07da3d40ULL,0x6b7eddc34af23568ULL,0xa4bacffbd7be3d60ULL,0xddffdffae6beff60ULL,0xdf2e3f74e73aad60ULL,0xab466d6bd6ba1560ULL,0x5edf3eff9e96ed60ULL,0x9fbfdfdb4edadd60ULL,0xdebfe6ec3c8b2f40ULL,0xfffe7ffbdebe7d68ULL,0xfb57df6ed7be7d60ULL,0xeefb9b7fffbe7d60ULL,0x9fafcf5bcffe7d68ULL,0x74dff6eefbfefd60ULL,0xbfbe7d637fbefd40ULL,0xeb26ffffdff63f60ULL,0x7fdeba9bce877f68ULL,0x7bdf9ddb5f9eff60ULL,0xd69cff6b1fbf7d60ULL,0xffee7f625fb77d60ULL,0xcb8621728e329540ULL,0xddcfff8ed6b63f60ULL,0x826ddd535f982540ULL,0x84b3422a92927f20ULL,0x71cc373d82926748ULL,0x7ab6a7eebfba6f60ULL,0x8dd7fa7f5e9eff60ULL,0x6ffedfdfdffe7d68ULL,0x1ebb0fcf8fb2ff40ULL,0x552687f813862e40ULL,0x88be6c7196b23560ULL,0xebcff77f7fbf6d08ULL,0xe7f6967f57f63f60ULL,0xffffffdf5ffeff40ULL,0xd5f737e79fbe7d48ULL,0xc69efdd117965540ULL,0xeb93db8b17d6ad60ULL,0xfffb36becfbe6d60ULL,0xdb3edff285fa7d40ULL,0x34df7febd7aefc68ULL,0xffeed737ff62560ULL,0xc69beb30cada2f40ULL,0xffdffebfdeff7d68ULL,0xab1ecb5606b63f20ULL,0xb4f97e6816fe7d48ULL,0xdbacff72b6beff68ULL,0xec38367a76be7560ULL,0x8ee6a22cd29a7d40ULL,0x63d7528f07da6d20ULL,0x2d3607eb07964d60ULL,0x19581e2a329e2420ULL,0xae9f3d7696b7ed40ULL,0xfdbfff3b17ae3d48ULL,0x6ff6bf4aff863d68ULL,0xf6114fba56125560ULL,0x1200060a08030c00ULL,0x57047579439a7d00ULL,0xcb81eb28869e2d60ULL,0x6000000001100000ULL,0xbbaf5d3996be7d08ULL,0x8432086002ae2400ULL,0x32b625f02ef67d60ULL,0x5200e62414920720ULL,0x840000000900000ULL,0x1602c30a80964c40ULL,0x434244a92822900ULL,0xa90d29200f136c20ULL,0x8004021004800140ULL,0xc42000040120400ULL,0x4018010040200400ULL,0x100305c40b822c60ULL,0x8001246002020560ULL,0x4602000800140ULL,0x0,0x0,0xa000020020ULL,0x0,0xbba0044493820940ULL,0xe4a4600483906600ULL,0x401014804160100ULL,0x207e054901400ULL,0x8850000100440ULL,0x9b9e5c721fba0f60ULL,0xca7ff2468a280d20ULL,0x9c41fd1c2d61d00ULL,0x5fbecd9edbba7760ULL,0xc44cbe216ee63d60ULL,0x1028ad000a362d40ULL,0x4a10131824103c40ULL,0x8114104321040ULL,0x10002000020040ULL,0x4000001000104000ULL,0x20040000200040ULL,0xf802860050a62c08ULL,0x4902940014924040ULL,0xef5fcf5fffbf7f68ULL,0x80040ca002b83400ULL,0x2100000100100ULL,0x2828a000a001420ULL,0xcfcfae3fcebe7d68ULL,0xefffdf557ffbfc60ULL,0xf7bfffefffff3f48ULL,0x15b2f7241b27540ULL,0x54260d600fb26540ULL,0x5788d10920c71d40ULL,0x2966060118923420ULL,0x8fbacff4bfb26f60ULL,0x100cf43b1683bd40ULL,0x400404200e100000ULL,0x89d6fb3d56be3d60ULL,0x76d95a864716cd40ULL,0x1080418110860800ULL,0x610c246050802768ULL,0x620204010900100ULL,0x8294046640944408ULL,0xedf37e9f7f96bf40ULL,0xb364fb246b62f60ULL,0xfdff77ffd7be7d68ULL,0xffffff77ffbf7f20ULL,0x18c523401b04d00ULL,0x40c2122244280440ULL,0x800240800802200ULL,0x11049002902d20ULL,0x8002040000400ULL,0x490024001800240ULL,0x10200800040ULL,0xa000080041100400ULL,0x106018202848340ULL,0x4064030000863000ULL,0xb200080202852c40ULL,0x104c20414000540ULL,0x3088860110901d00ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1400044010800900ULL,0x8800020000020900ULL,0x0,0x880090002082000ULL,0x9111024001960020ULL,0x0,0x0,0x8922100801620ULL,0x2008188002200440ULL,0x1000000200004040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe7bfdffbd7967bdfULL,0xb6ff77ecf7b63fffULL,0xdfb2fd73f32c23ffULL,0xfb8cef2cf6162f7cULL,0xcbffdf1bf7be7bffULL,0xb7f0cdd0d5142ce3ULL,0xf4f7f6fef6ae6fdfULL,0xfcff3773d79c63fcULL,0x8a8b7d2a92a42954ULL,0xffdfffbef63e6fc7ULL,0x6ece5bd1f7ac625bULL,0x1e23effef694236eULL,0xf15ecfe2d224775eULL,0xfebe3feef7b660f5ULL,0xedd6f7bbd7263fd6ULL,0x2fdfbfc5d61e6d81ULL,0xb79a5ff3d68c6de4ULL,0xd53dfee9d62e2bf6ULL,0x9fee7d43d63e2dfeULL,0xc9d6f53af6166bddULL,0xcea0edb52042f84ULL,0x3ff91d81d6a452deULL,0xfcf7be77d62e2ffbULL,0xf3bead70d6262fe7ULL,0x8b08df6af006230cULL,0xcf47bf75d70e2766ULL,0x9bbedd71d68c697dULL,0xd536fefbd79e6becULL,0xbfbaafe1d6a673c7ULL,0xed36ffffd6ae6defULL,0xdff7f787d6b43fecULL,0xaf7fdddfd6062bfeULL,0xf63a76e2d71c6bedULL,0x97fe3778d6166a6fULL,0xff326e5ff79e69f4ULL,0xf7f7bf3bd6be6bffULL,0x4c04090004200ULL,0xbdbd5bf3d79c7fe7ULL,0xdfeea7b8f69e37ceULL,0xdabf3dffd73c27ddULL,0xedfcf1b8d6863bd4ULL,0x236e3b59d79421d7ULL,0xb6bbfffff6ae6feeULL,0xc93abd7ff6967f7cULL,0xa4e70e8b6002944ULL,0x8322d77ad7342848ULL,0x9edf3fd6d6be3b6dULL,0x23bccd73d62623ceULL,0xeb7ff6fbd30e65fcULL,0x7ffe7dfbd63e2fcfULL,0xdbf5ff3ed78e6fffULL,0xefdbfffbd6ae2f5fULL,0xfffedffbd6a47ff7ULL,0xf7f4f6fed63e7fd5ULL,0x9bbeffe3d69c6375ULL,0x93347b7cd7ac63e8ULL,0x59f5e21d70e2105ULL,0xabdfdfdbd6bc6bcdULL,0xb6f9d6e5f4042ff5ULL,0xfd966773d7342befULL,0xbea0577af71650e6ULL,0xc1e3ffb6d62621ddULL,0xa1481a82920c2945ULL,0xaeba75efd61c67d4ULL,0xff69ee30d6943be6ULL,0xdc9a2ff5d28663feULL,0xedf5f235c6867bc5ULL,0xcbde5f79d7162becULL,0x17346ff3b2a02e4dULL,0x70aca64af62421c5ULL,0x1e9c6f6cd62467fdULL,0xa9f7af77d22c2ffdULL,0xeed63bdbd22e2ff4ULL,0xbfbcefdff6a66f7dULL,0xd776f6ffd68e2be7ULL,0xfefe7df3d62e27dfULL,0xf3f6ff3ef72e37efULL,0xce5f97e2d6947ff4ULL,0xf6bfcff7d6162eedULL,0x30b7feebf79e63fcULL,0xdf0aff72d69c73faULL,0x5e83e16ed3022c86ULL,0xc6dffe27f71e6dedULL,0xa7d6cfdb96243d7dULL,0xfcf37e6cd62623ffULL,0x9126afe0d63630bdULL,0xcb9012d9d2a027f4ULL,0x8727ef3ed60c2fadULL,0xdd7d9c4d71225c4ULL,0xb49b5de5d6a02e55ULL,0xbfedfe70d22461ecULL,0xdbbf6fffd7366757ULL,0xedb6f73ed20c3ffbULL,0x6f5f1cd9939e295dULL,0x7287f72928439ecULL,0x102307769210273cULL,0x7694355ed2067e34ULL,0x8bc7a20d95946bc4ULL,0x69a73d80d712414cULL,0xb4d2c38ad21421d7ULL,0x34762666d20420dbULL,0x4e845dc2d51441c2ULL,0xa0c4c10250022f00ULL,0x2870838ac0046140ULL,0x300050054000004ULL,0x3107066e560c2865ULL,0xe5388320d60c2d00ULL,0x99839b76d69065d4ULL,0xc5cfee39b206235dULL,0xb1f475ac2022980ULL,0x8020042040002880ULL,0x8805000000200048ULL,0x4020220010000804ULL,0x0,0x0,0x0,0x0,0x800230090002014ULL,0x4180921050002384ULL,0x2952965190242175ULL,0xa4300560c2002481ULL,0xd2292e70d63021cfULL,0x2e12f562d22c68ebULL,0xebf1f32ed7b673f6ULL,0xe8f35f5dd69473e6ULL,0x7ebfcfdbd72c63fdULL,0x40140280102000ULL,0xc6a82f40560609e4ULL,0x8a102112c2042104ULL,0x90308354002041ULL,0x8080000041000ULL,0x904400c0000102ULL,0x2000882000002801ULL,0x800002002000000ULL,0x8882722982022850ULL,0xcfbe4f53d61e7be5ULL,0x26a897e0d60c2dd4ULL,0x58110000a00ULL,0xf496062cd2102f01ULL,0xcce7af3ed39c2dc6ULL,0xeaffffdff62c7f7fULL,0xfffffffff6be7fffULL,0xfbffff7bf6b62f5fULL,0x9226356092347505ULL,0x88245ea0d4902b41ULL,0xab745ecbd30c25d6ULL,0xb3becff3d7067397ULL,0x91008570d7022a04ULL,0x24042012020200ULL,0x80003000020c0004ULL,0x2fff7ed3d7be6fd7ULL,0xa1ccbd4b2062264ULL,0x100880000010ULL,0x24040000802440ULL,0x804012890002304ULL,0xecf73eabd6346bf6ULL,0x85ff5f74d28c7ff5ULL,0xfffff7fff7be6fffULL,0x97fffff7f7be7fffULL,0x2a10670bc20400baULL,0x29568608d20403c4ULL,0x209480056002020ULL,0x410006f46222002ULL,0x80410500b2140084ULL,0x4010271282142404ULL,0x1068704c4000058ULL,0x802121882302004ULL,0x430142590100058ULL,0x8d0242820c6101ULL,0x0,0x100401830000020ULL,0x2400000000040080ULL,0x22200cd082802c40ULL,0x880080000008ULL,0x0,0x0,0x0,0x0,0x801081000000ULL,0x452c055074042c80ULL,0xc202a56ac6a0654cULL,0x1000020000040080ULL,0x0,0x0,0x94100100c10400a0ULL,0x100008000000000cULL,0x42021012000104ULL,0x0,0xf0142020c6042008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa9421c08813053a8ULL,0xafb7cffb8bbe7bb3ULL,0x70b6fefb8bbc72e6ULL,0x5afefd728bd47bfeULL,0xf626af678ba85beaULL,0xcfc3fe36a9f0767eULL,0xefffddd6abbd53feULL,0x5dbff6e9abbe5bfeULL,0xbf7cff78abba7bfeULL,0x9b6ceb2f8bb463beULL,0xd063ef3aa9f472cfULL,0x7ffedfdfa9fc57bcULL,0x3e9615b588b877a6ULL,0xdde3fdf08bb24b6bULL,0xdebe2f7babde5b25ULL,0xefffff6f8bf55fbfULL,0x3dfefffb8bf77fbfULL,0xbab5cff88bd84726ULL,0xddec37968b3c57aeULL,0xcb5e6f778bbc57aeULL,0xcb92f3eeaabc5364ULL,0x64935ff98aa43306ULL,0xb7b9def28b785beeULL,0xf5fefefe8bbc5fecULL,0x5f46bd3189b45fbfULL,0xe991fb2d89fc73bfULL,0xcfd3ff86899c4ffeULL,0xffbf5fd98bf047abULL,0xd77737ef89bc73a6ULL,0xdb7eeff1abae7f3fULL,0xfbbeff7f8bff5fe3ULL,0xcfcfff9babfd7fffULL,0xabfecfc98b9c535eULL,0xfcbf56728bdd5327ULL,0xbffefff3abbb5b76ULL,0xcfbfbf7fa9b47bbeULL,0xcfeffee8abf47f7aULL,0x699fdf478bf657feULL,0xb2b65fee8bbc7ff5ULL,0x3960ce7c89b8473eULL,0x8fb22f738bb5722eULL,0xc7f7ee35abe42b63ULL,0xaa77ffcb8bf03f7eULL,0xbfbedfff8bbf5bbfULL,0xf9dffff28bd47bf6ULL,0xffd6bf7789fe1777ULL,0x8933fb8a89b8177eULL,0x7dd71fd88bda7b78ULL,0x5bbd6dce8bb63bbeULL,0xf77dbbfd8bfd4bffULL,0xfffe7ffb8bff7fffULL,0xfbffffffabff7fffULL,0x6febfede8bbd7bf2ULL,0x6fbfdffb8bfd7bbfULL,0xf1f6def9abbc73eeULL,0xdabef562abdc73efULL,0xf700af078a945a2aULL,0xccd2b019a9856a6aULL,0xea7f0fdbabbe43f6ULL,0x74efdffd89ec1f6bULL,0xb4c9af708974122cULL,0xf21677b4893c536eULL,0xe7c7f73a89f57727ULL,0x3b765e5fa39e1367ULL,0xfc5eddfaa99c63bfULL,0x1082661a98c0062ULL,0x8af63e53a39a7ad2ULL,0xefcfdf58abf65b6fULL,0x2bffdfe189dc63a6ULL,0x8ebadff98b9c5f26ULL,0xb7cd176ba9b25f34ULL,0x2b1e3c4589846332ULL,0xe1b7ff3dabbb473eULL,0xeffbbfdfabfe6f7fULL,0xb6b545e88bd06336ULL,0xd573fb7babbe6233ULL,0x7bbe7d7389b57bbfULL,0xdbf6df7e89f45fffULL,0xeaf3bfdf89755b6eULL,0xda3eed7889bf1b66ULL,0x5efddfeaa9bd4b63ULL,0x92922d5088500be2ULL,0xce06e52d89b66223ULL,0xfd473b1b89fe4be3ULL,0xbb0ecddb89be53b3ULL,0xf4b51ce20b9843aaULL,0xf7e3ff7b8bd957e7ULL,0xffb267fa8bbe53aeULL,0x49d2b682888c5226ULL,0xa35dce5c0b9c43a1ULL,0xb7f7df6fabdd4b8eULL,0x5967ae7a8bbc6f66ULL,0xfe8e3f2aa991736eULL,0x6fcaf23f89bc7fe6ULL,0x57dedf8881b47ea2ULL,0xb6d84ee109bf432eULL,0x50400080110020cULL,0x18c0010408440120ULL,0x2004000008004000ULL,0x4cd6de5c81935bf6ULL,0x88aca42aa8b856b0ULL,0xd0cfa6f58adc53a0ULL,0xca0204428ba04200ULL,0xd0c1714e0b844308ULL,0x940020000000328ULL,0x97cc0000880a27ULL,0x4486f2d08929436aULL,0x8000040000000120ULL,0xc290732a8b505a02ULL,0xfcd7bebfabdf0bafULL,0x4812094202842300ULL,0x24008080000ULL,0x200010000040200ULL,0x0,0x0,0x100041880005080ULL,0x0,0x8000080001000020ULL,0xfabcbfff8bac63ebULL,0x4000240200140220ULL,0x2220004281340020ULL,0x4a85950a3844a18ULL,0x453e8d6389104be6ULL,0x1a96256009d02a60ULL,0xfbc6eb478bad53beULL,0xaff6dadf89d45bf5ULL,0x9f9ec5cb8bcb23a7ULL,0xd674026881b4032aULL,0xa8003c00a001130aULL,0x0,0x840060000080320ULL,0x0,0x404010800a804000ULL,0x42030000003a0ULL,0x8000000400800020ULL,0x494bbc0b8bc0632eULL,0x69df9d53a9b07727ULL,0xf5fffffb8bf75bffULL,0xd4e8242088c40227ULL,0x9024356489344b00ULL,0x4f8bfbbea99c5fa5ULL,0xfbffdfdfabfe7bffULL,0xbfffffffabff7f7fULL,0xbbefaf7ea95d47e6ULL,0xc012617081b00a13ULL,0xc022081001b00210ULL,0x16d4ce0a8194130aULL,0x7904772893a4ba0ULL,0x2040b37aa100536eULL,0x28221600a8840b0ULL,0x0,0xaf7f13d3abf463aeULL,0xb3dcdf38bec6b3dULL,0x9474c7668bcd332fULL,0x24006009400020ULL,0x314cec088b304626ULL,0xdde3bf18ab8c7732ULL,0xbffecffaabff43eeULL,0xfdfffffbabfd7fffULL,0xd7bfbf77abff7bffULL,0xdc2e6ffe8bb14babULL,0x886098108b200060ULL,0x22e00400b004060ULL,0x4002001100000ULL,0x40000800020ULL,0x8a00000800902320ULL,0xc0000800010ULL,0x40c01810080c0320ULL,0x18002180200002ULL,0x8a64088500300ULL,0x2000240002005200ULL,0x580800080801002ULL,0xc80092030b920202ULL,0x800204800a804000ULL,0x80101a0202840280ULL,0x300080000800ULL,0x0,0x0,0x0,0x1100004803004002ULL,0x82a0c6280100f28ULL,0xc0a0e37209c042a6ULL,0x0,0x80040c8102806020ULL,0x28000100200ULL,0x8081416080a00062ULL,0xc000110208500a00ULL,0x502080000041a0ULL,0x4000000300ULL,0x48000000300ULL,0x120022000300200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2fef1a9bd1d5db6aULL,0xffbecf73c9ddebbaULL,0x75bffefce955f32aULL,0xfffe6ff3fbc5f36aULL,0xfe05ffae7bdcc372ULL,0xce7bbf33fbcde32aULL,0xbbfe4f5fe2d4d16aULL,0xf7f7fee7ebd5f1baULL,0xd9feef7b4bc4d33aULL,0xf2b67fe6cb5cf1daULL,0xffd7ea7ee9d5f37aULL,0xefffffdbf9ddf3aaULL,0xf4fe5fbb89dde3baULL,0xfbe7ef6aebcce16aULL,0xc3ba7472cbd5d30aULL,0xe1f7f7aacbc5e32aULL,0x46fedfdff354e31aULL,0xbf9beff1c1dde33aULL,0x24769f4051c5c302ULL,0x5dbe7e736bc483baULL,0xabec5e167354d32aULL,0xcbc23b5b48c4c13aULL,0x97925fe7d8dcf13aULL,0xdffffffeebddf3faULL,0xfebe3df2e3dddbbaULL,0xbbd4ef7febc5e9baULL,0x6fda860fabc4c12aULL,0xb73fcff3f9c5f13aULL,0x757576ebc3d4e93aULL,0xdffebfe26bddc3faULL,0xfffffffffbddfbfaULL,0xefdeffbff3d4d33aULL,0x8dd64f5fcbd4f3baULL,0xf4f6feaafb5cf19aULL,0xdffbff73fbdddb3aULL,0xefb77f37e9dce3baULL,0xcdf5ff0ef3c4fbbaULL,0xe19edd5b5bc4c1aaULL,0x74fa5bfb7bd4d31aULL,0x5948b5a9c1c5d9aaULL,0xdabbefe2c3dcfbeaULL,0xc7c7fb73fb54d36aULL,0xed7edffffbddd33aULL,0xffbffffffbddebfaULL,0xd57dff7afb54f3daULL,0xfffe7f7ffbd5f3baULL,0xf30effeeabd5ebfaULL,0x4ff77eeacbd5d3aaULL,0xfffffffef9dcf3faULL,0xd7bff2fdfbc4f9daULL,0xfffebdfbfbddf3baULL,0xfbffff3ffbddf3baULL,0xefdfbc3be3d4e3baULL,0xff7fdffff3dcf3faULL,0xfef7fffafbddfbbaULL,0xf7fafff27bdcd3baULL,0xf283e577f144db3aULL,0xc856fe3169ccd00aULL,0xae4c4fdad984e31aULL,0x7cf77fedf3dce17aULL,0xa1fea76ac0ccd30aULL,0xee9e2f7ed354eb02ULL,0xc863ee2ed1d4e3baULL,0x97de5d4b8bd5893aULL,0xfcdb7fa3c1c4d32aULL,0xa54727985150d10aULL,0xeab62f7e4bd4937aULL,0xefa7f76eebc5c32aULL,0xbf7efadefbdcf32aULL,0xb7b7effbebd4fb2aULL,0x94aca760d1c4431aULL,0xdbfe6d54f1c5cbb2ULL,0xd3f3ef2a79c4d33aULL,0x5ed6ba6df354a32aULL,0xbf7e5ffef2c4fb2aULL,0xf7f7f6ffe9d5eb3aULL,0x5fd63dd3e3ccdb3aULL,0xfbfdfffffbddf3baULL,0xcfe3f6e7dbd4e32aULL,0xe7bf4ffaf1c5e3f2ULL,0x7db43ceb6bd5d23aULL,0xf7beeff0ebddf31aULL,0xfe14f25e5154d3b8ULL,0xcedef24260c4c332ULL,0x5f9fdfcff9c5f32aULL,0xf4b5f66ff884eb32ULL,0xcd6e67f1f1ddf32aULL,0xfa97373f49d4d17aULL,0xef77ffbdebddf3daULL,0x1bac9dd88154629aULL,0x347e15eff9dce12aULL,0xd974ad9bfb44c1a2ULL,0xdc063ff452d5813aULL,0x8b563a02c854e82aULL,0x4e76ecc74bddd10aULL,0xb9b34d7d5bdcc23aULL,0x9790ac018000e12aULL,0x5602314648c48122ULL,0x6084e30e80c4c32aULL,0x280222098200a922ULL,0x1fb0ffe8e2d4e07aULL,0xb0c45e4200d4c15aULL,0xfaa40de2d1d4f11aULL,0x49bf532243d4d320ULL,0x40011610e3c04008ULL,0x40180c01e084e000ULL,0xf18277e8e944e12aULL,0x936045209540202ULL,0x5c60020041005108ULL,0x102900040008012ULL,0x8100040010ULL,0x421000080004002ULL,0x8080000001004000ULL,0x0,0x0,0xab7a914a7280d12aULL,0xe148008002ULL,0x8400000040008000ULL,0xf2b6249170018128ULL,0x9c2e4220384c2a0ULL,0x10100000000102ULL,0x24a0040809020ULL,0xd0dcb6c009d4c36aULL,0x17960f648055e92aULL,0x8becc57be9c4eb4aULL,0xfca2e5feb45a36aULL,0xb69c7d400188c31aULL,0x1412e6e4c0c0e10aULL,0x273006010005c30aULL,0x5210410861004000ULL,0xc0001182000802ULL,0x401000008000002ULL,0x140400800080c002ULL,0x2c1008440000ULL,0x0,0x8fee3b3bebc5eb3aULL,0xab7fdfdb69c5d2baULL,0xfeff7fff7bc5c1baULL,0x4000a00040048010ULL,0x2a90400888006012ULL,0x894bfe0b4145813aULL,0x8bdedfdfebc4cb5aULL,0x15bf37e2e344c15aULL,0x91cc87e8e2c4f03aULL,0xc0b401f8f104c322ULL,0x85f6e11a89c5ca0aULL,0x588800900c0c028ULL,0x7feeed7ae3c4f33aULL,0x500000000108ULL,0x5a5e25e361c4d14aULL,0xc5c11001ccc042ULL,0xc31f706145c00aULL,0x861a02100b00802aULL,0x66a31408045c818ULL,0x28ae3c63d3d5ca8aULL,0xc9f4e7aef1c499baULL,0xe6c1065a4184c30aULL,0xefbfdffbfbd5cb6aULL,0x74fffeece1c5f32aULL,0xffffff77fbcdf37aULL,0x5212291500c0c13aULL,0xcc0b9608e101a322ULL,0x2308090000c0400aULL,0x348400408201c002ULL,0xd1a855004045c01aULL,0xa81082008300480aULL,0x10430a004100d020ULL,0x4010000000400002ULL,0xdcd477edc1d4f33aULL,0xc170bf405ac5c13aULL,0x3000012080002000ULL,0x208000008020ULL,0x8000080000008020ULL,0x2080143040004102ULL,0x9000040840008100ULL,0x2004001000200ULL,0x4020080004000ULL,0x0,0x0,0x340424080800812aULL,0x40086d42824cc062ULL,0x1000ebb041808192ULL,0x840000080002000ULL,0x102004000004000aULL,0x0,0xa00010201040280ULL,0x0,0x0,0x418048100008102ULL,0x48000008100ULL,0x88500040008100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xccfb17ffcedbf7f6ULL,0xffbfdffbcedbfffeULL,0xfcfdfff7cecbfafcULL,0x57be3772c6d2327aULL,0xff9ee7becedbf6e0ULL,0xccc22ab946cbfaa0ULL,0xef9edfdfced3fea6ULL,0xf7ffffffcedbbffeULL,0xffffffe3cedbf3f6ULL,0xca2a6f44ced3f324ULL,0xe787ef2accdbf3b8ULL,0x27f6dfdbccdbf6e8ULL,0x75bdfdebc6cbbf2cULL,0xfb4ffef6cecbbef4ULL,0xfa963df8ce92b3fcULL,0xc9f6f623cadbf766ULL,0xc812f6c7c682732eULL,0xfefb7fe9c6daba5cULL,0xa1771e6bcedaba7cULL,0x1976f5714493b300ULL,0x89de5b6bccc3faacULL,0x4ec91a9fc4db9a60ULL,0xdef9fffbcec2b7f2ULL,0xf7ffbfffcedbfff2ULL,0xefbeff73cedbfffcULL,0xfb97eb3fcedbbbfcULL,0xa8c7bfd9cecbbaccULL,0xd1bfcffb4edafff6ULL,0xf4fdf7ebc6cbf972ULL,0xdbbfff72ced2feb2ULL,0xffbdffffced3fffeULL,0xefeffeffcecbfffeULL,0x2f9ffedfce8bf6fcULL,0xf2d2fff9cedab37aULL,0xdbfffffbcedbfff6ULL,0xffbebf5e4edbf7e6ULL,0xfffef6bec8c2bf74ULL,0xc6dfde53cedabe52ULL,0xb5b36de3cad3bf60ULL,0xfbd9f7aacedbba20ULL,0xdfb62f7386d3be32ULL,0xcd8cb3e4ec2ff64ULL,0xdffffedfcedbfffeULL,0xf7bfdffbcedbfffeULL,0xb3f9ee7ace9bb2a6ULL,0xfeb6effec6cabf74ULL,0xe9efdb9fcec2fef4ULL,0xc8dabf698edabac2ULL,0x5fda6ddb4cdbb676ULL,0x9cfe77c38edab3feULL,0xfffe7d72cedbfffeULL,0x8bfdd6bfcec2bf6eULL,0xfddaf75eccd1f7ecULL,0xefffcff7cedbbeeeULL,0x54f7d6fbcecbbfe4ULL,0x7f9efd734ecbbfeeULL,0xf6b6bb764e4bf7aaULL,0x5ade3e9fc6cbf68eULL,0xa34e8f4a4cd2f230ULL,0xbcfe5e654cc3beeeULL,0xdfc66d624ecaf2bcULL,0xdebfbf76cec3b77cULL,0xeac6f71a4ec2b7e8ULL,0xa7ccadec8cab276ULL,0x8838d5cb4c43a220ULL,0xd27c3d180ec2a322ULL,0xe58d3562448ad5a8ULL,0xedc6f2af4ed3bf54ULL,0xafdfbdd9cedbfb3eULL,0xfffbffffcedbfffeULL,0xd5fdbf7dcecafa64ULL,0x1042e56488407024ULL,0x83c7cf03c8d2b924ULL,0x6a733f5bcc43bae4ULL,0x9d9acdfacec2babcULL,0xf477f2f1cadbf3c2ULL,0xebeeff6a4edbfbfeULL,0xfbdfff3fcedbbff6ULL,0xec63bf79ced39f1cULL,0x8b3e4f56c48ab22cULL,0x5a94d7eccec3b33aULL,0xdf03f5514ed2b28cULL,0xc98c742b48cad670ULL,0x4dcfbe494ecabfbcULL,0x2f6adfda4ecab6e6ULL,0xf6fdf6fdced3bbeeULL,0x916eb75ac6c2b896ULL,0x9fba6776ce83f3e2ULL,0xf7e7f33a8edbf3e8ULL,0xa91efddac6d2f676ULL,0xb5f37feacedbfaaeULL,0x9928877286cab2a4ULL,0x1a300f12488b9320ULL,0xa46fb2a448cbbe80ULL,0xd7fde994ecab728ULL,0x96ba7c6d0e5afb6aULL,0x95b49706c0c273e4ULL,0x2023d4004418b08ULL,0x8a2eba3f86839302ULL,0x802021100020140ULL,0xffff7ffbcedbfefeULL,0x81a2d62084c970a0ULL,0xab614200600a30aULL,0x89f6cb36c6cb126cULL,0xc757be8b4ecafe2cULL,0x44880800004ULL,0xf5f756fd8ed3ffe4ULL,0xdb6bbc75cecbba62ULL,0x8280200000420300ULL,0x0,0x800200002002000ULL,0x44048004002080ULL,0x940420844028120ULL,0x8200000002000200ULL,0x8400000400200ULL,0x201100804808040ULL,0x0,0x109c06080c28b00ULL,0xa8282760cc827606ULL,0x5ceca2286c0f282ULL,0x2152c611c8c1b404ULL,0x8064000103020ULL,0x810000080800000ULL,0xd79a2ffc8ecbfbb2ULL,0xe9ffff7bcec3ffe2ULL,0xefdabeffcedbb2feULL,0x169c05e84cc0af08ULL,0x94356270ce80b6e4ULL,0x500002022040ULL,0xdc4526c6c01aa0ULL,0x10022e0002129002ULL,0x8402048001004ULL,0x30108a0004903080ULL,0x303a107004438068ULL,0x0,0xded7be1bc4d3fb2aULL,0x2bfb1b5a46cb934cULL,0x78f636f9c6d3b764ULL,0x8c06242044d13020ULL,0x8800041080002004ULL,0x59c6ea02cccaaf64ULL,0x2f1e9556c6ca3acaULL,0xdb2647e14cdafa30ULL,0xf3ecefee4ed3f7b4ULL,0xca061d6c0e82a700ULL,0x8f439b224ec2a660ULL,0x15ef282448ab078ULL,0x4028674100c2a200ULL,0x10268780c0989104ULL,0xdb3647f6ce82b3f2ULL,0xe9ddd602cc82b234ULL,0x214011c188818200ULL,0xb9320502c4821260ULL,0x9401a07048c18204ULL,0x8b14b521ca82a214ULL,0x4b166ba846c3bfa2ULL,0x10061e0104503600ULL,0xbfffdffbcedbf6e6ULL,0x71ff7fffced37ffeULL,0x9bf7ef53cedbf3feULL,0xeb18776cc69af6e0ULL,0x890e3e1f46dab7e6ULL,0x414a018bc6880600ULL,0x80500084021000ULL,0x10c103704800a632ULL,0x4006202c04822080ULL,0x8000800040020000ULL,0x8040040000004ULL,0x140856eecc81b260ULL,0x2023484002a164ULL,0x1210010000422100ULL,0x8010000000020100ULL,0x80800000018000ULL,0x20204400000a042ULL,0x8034000000121020ULL,0x6000000020100ULL,0x0,0x0,0x0,0x24002180403000ULL,0x5e80244008c9b290ULL,0x302800c00107020ULL,0x0,0x8400080001000ULL,0x402690c820220ULL,0x0,0x0,0x8000000104002000ULL,0x400400040000004ULL,0x1080048004800000ULL,0x120a82044909204ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebd0e92aaa72b39bULL,0xecea3ebb99737b9bULL,0xafbedf77fdf77a9bULL,0x5dd77eefdfd73ebbULL,0x9ffeb5919973da9bULL,0xff94ff7f5ff37fbbULL,0x68cfff225cd2beabULL,0xabe8f5b4b137e8bULL,0xfcf6f6fdd9737fbbULL,0xdfeff7f96f53fb39ULL,0xf3bf1f6f4cd3bf8bULL,0x7f5ef7bfad73be8bULL,0xaff7fa4f88f27e82ULL,0xbeb7dfeb8fe33a0bULL,0xdbfdbfeeaf573abbULL,0xcbaefff32f73fab3ULL,0xe867fe2059523e1bULL,0x9886fe0bafe23f13ULL,0xffbeeffb6dd33f93ULL,0xd1e5566e2ad31b93ULL,0x97ac2bc08b521a01ULL,0xdad77ea609d31fbbULL,0xab775e1218737a11ULL,0x9e1bcdd349567b2bULL,0xfffffffffff7ffbbULL,0xfffefdf3cff77fbbULL,0xbbd5eb2b2ef33b83ULL,0xe8c7ffbdddf3babbULL,0xefbfcffa6ff37eb3ULL,0xfcf676ff8ad3b23bULL,0xdb92bfd17e737eabULL,0xdffdeb7f0e57ff9bULL,0xefdfff3fedf7ffabULL,0xeadddfdf1ef3ff3bULL,0xb09ac7e803645a0bULL,0xdff2ff6fabf3ffbbULL,0xf29a2ff4cb635ea9ULL,0x8bf2a2bc08537b90ULL,0xead77ddd4b76bb9bULL,0xbe77d6ef4a73bb9bULL,0xb3edf77a1af39ab1ULL,0xdeae2ffe0963fe13ULL,0xcfdbfff92dd7ff3bULL,0xbffe9fff6f773e9bULL,0xbeffdfffc9f3ffbbULL,0x927eff7a18735fb1ULL,0xffae7def6ed37b9bULL,0xebcefbddcb73ff8bULL,0xff5b5fdf4bc37b33ULL,0x92bccf6bcc473a83ULL,0xf5efb67bfb73dbb1ULL,0xfffa5d7b6fd3fb89ULL,0xeb94f7369bc7fb9bULL,0xa6ce2ef208d3be93ULL,0x4e96454e0dc73a81ULL,0xb4f5f6c4a9f21a91ULL,0xdbbaff628873b3b3ULL,0x7106a7210844bb11ULL,0x5dce36930c627e2bULL,0x2bdecddb46d37a03ULL,0xf5fff6feedd3ba39ULL,0xfffefdf3fff7ffbbULL,0xffbf3fffeff7ff9bULL,0xffd7ef3e2b733eb3ULL,0x6bdfffdb5bf3bfa9ULL,0xbd3e55fa9a733ea1ULL,0xf97c7f461c77329bULL,0xfe9f28e7cad37b3bULL,0xc1deffef4e53bb9bULL,0x6efede5b6b73df8bULL,0xffffffffeff7ffbbULL,0x75ecf67b09e33ebbULL,0x7f123d67aac37b81ULL,0xe977ffaf3bf3dabbULL,0x4cf71ed32a727e9bULL,0xffbfffffeff7ffbbULL,0xfdf6f7ff5bf3ffabULL,0xdf9e7d710f767bb1ULL,0xf3e6ff6ec9535f9bULL,0xcfdbbbdf4a739e9bULL,0x7bbbcfdb2df7bbb3ULL,0x84e637e8c8931a8bULL,0xeffafd616a77bf9bULL,0xe9b4ef618af3dba9ULL,0xef5fbebfaff33f83ULL,0xab7fff4abb531eabULL,0xd4b7766a0a72bfb1ULL,0xf9a5ef7aefd33fb1ULL,0xbbb437745cf23e01ULL,0xe9fef8a2b1f3ba9bULL,0x2b76194803439a03ULL,0x9cf557ea8bf37a39ULL,0xf9dd76730e723b13ULL,0xaa0016208431b91ULL,0xc95bfeffcbf3bfbbULL,0xc0d33a1908d21a1aULL,0xa9a84d780d037a8bULL,0x90587a2024033a01ULL,0xc10e3ca003921a03ULL,0x121a6040521001ULL,0x8008480108601001ULL,0xffffdfeb9fc77eb1ULL,0x404922805023801ULL,0xcb362c0001177e01ULL,0x83d0b6115852b303ULL,0x9c4bbeebadf36eb1ULL,0x2730140080021602ULL,0x502ca35008035a81ULL,0x80082e5444d03a03ULL,0x820000000202ULL,0x40c00c1048004801ULL,0x4842091208521900ULL,0x0,0x0,0x40000400000000aULL,0x0,0x0,0x0,0x0,0x8000004004002000ULL,0x80032601831e08ULL,0xa8c29a5208633f21ULL,0x402006000008281ULL,0x10502e80ac022210ULL,0xdef66d7c48765a83ULL,0xbfaff6f4dc73ea9ULL,0x69e64e1a05137e9aULL,0x90908242021811ULL,0x9000808400050602ULL,0x4f0264400a521301ULL,0x8982610a41521801ULL,0x110000000201ULL,0x86500010120bULL,0x0,0x0,0x4000800000001800ULL,0x88d7be139ad23a93ULL,0xb25a95cb49521b81ULL,0x61b516b849563e81ULL,0x9dfca7730b535b81ULL,0x72962f7ce87313abULL,0xef77fed669531fbbULL,0xac4ddd88d733a8bULL,0x983e5ffa89137a8bULL,0xfd2eff6aabe75bb3ULL,0xbeb617614b731a8bULL,0x8136d3b809905a8bULL,0xaeda3a9b4d433f8bULL,0x80400000000201ULL,0x1388f0044910181bULL,0x8b32ef7201535a01ULL,0x81a044120800161bULL,0x40023810411a0aULL,0x8cf267e4ac56b289ULL,0x9242124613701801ULL,0xcf523c4a08721f39ULL,0xfb54d33dcb537f31ULL,0xb242a25148521a19ULL,0x8e3f4f734f757e03ULL,0x75ffdff81df3fe9bULL,0xdbaeaf710a77129bULL,0x5200e10b1b021901ULL,0x242220002401001ULL,0x940001008000aa0ULL,0x380aa809021a80ULL,0x28240600084102a0ULL,0x140000000600ULL,0x0,0x90d2c0c0021009ULL,0x444000001000a01ULL,0x84000000500000ULL,0x80000240000001ULL,0xc1000040020aULL,0x0,0x2000010000400001ULL,0x8010240040020011ULL,0x0,0x8100025400031600ULL,0x110000500000ULL,0x34dbefed2e773b91ULL,0xb4a52420085010a1ULL,0x43a83d3009d21a91ULL,0x9292730e0b863a01ULL,0x0,0x0,0x0,0xa00500001101800ULL,0x2c2009000900ULL,0x840000001000800ULL,0x150040000029001ULL,0x28400001008801ULL,0x813258200002900bULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6812247864760954ULL,0xbbfcf7fee67f17daULL,0xffdfbbdf6a7f3cdaULL,0xeffbcff3a36f1ccaULL,0xddf7deece4763c4eULL,0x21b6fd636876acc4ULL,0xffbfef7b6f7f1fdeULL,0xefd7b74fe77735deULL,0xa33a9fcf667d99deULL,0xf4fffefa677fd9ceULL,0xd9cfffd66d7eb5deULL,0xdebf2ffc6777fdd8ULL,0xf7fbf7bae2773ddeULL,0x8f7f5b0ee27e1dd6ULL,0x8dfe77eaaa771dceULL,0xfbfcfefeaf779ddeULL,0xaaaeef4c6a76b4deULL,0xcbf6ffa3677f96deULL,0xeffffedfef7fffdeULL,0xbebfffffe67f9fdeULL,0x9dfcf7caef7fd9daULL,0xd6baff66ea7e3d5cULL,0xc823e352227d154cULL,0xefd79e5a60765fd6ULL,0xbabfcd6b236e15ceULL,0xfdffff7deb7eddceULL,0xf6feffe1ef769fdeULL,0xa39fefdf277ef8deULL,0xdfdffeffef7ffddeULL,0xbebe5ffee36f5cd2ULL,0xd5f676786a7eb5ceULL,0x5fbeffe3277f96deULL,0xffbfff7fe7779ddeULL,0xd9cffe9fe47d1dceULL,0xa8164ed6215fdc44ULL,0xf4d27eec636f11deULL,0xdfeebe62a7761ddcULL,0xff977577e67f5acaULL,0xfdffff3f677f7fdeULL,0xa2dfffdee6771dceULL,0xb6fffeffe67f5ddeULL,0xf76aef496a763358ULL,0xf7b6effee67f7ddaULL,0x67eaff26227eb4ccULL,0xabbfff79a27e14c6ULL,0xf7bb7ff5e05fb4deULL,0xff3d8ff32b7fb8ccULL,0xfef26f74ef7e97ceULL,0xcbfefbbfe27e9fdeULL,0xafff7ebb2f7f7dceULL,0xbdb25f73697d1ddeULL,0xd7fef65eab7f9dc6ULL,0x4f7e6d7a2f7468ceULL,0xc9f7df7fe37f15deULL,0xc9cbff90e24717c2ULL,0x9a9a5570a74413c6ULL,0x1482b6e8267684caULL,0x3bbe67f3ae7611d6ULL,0xf9a47f8f2676900cULL,0xcccab7bb62775ec2ULL,0x3b8cdf572a767dc6ULL,0x36957fe8e37cb5deULL,0xf9eff773ef5ffbceULL,0xfebe7fffa37f13ceULL,0xf554fe04267d79c2ULL,0xb5cdfdae77ecbdeULL,0x6cd25df1a2741dd6ULL,0xf12ebb50227e3c4eULL,0xda8ebb6fe67631daULL,0x89eef7ffef7f9fdeULL,0xd3feff592e7e15caULL,0xffbb7fffe77e95deULL,0xd7fe3fede27619ccULL,0x5f0e2d64623e1656ULL,0xfbf6ffffee7f99deULL,0xae5237d626771958ULL,0x3eff75fe6e7f11ceULL,0xf1f7feffa37f9fdeULL,0x75be7f72e67697deULL,0xdbd7e337e77e1cdeULL,0xceefffff6b7737deULL,0xafb6cfe1e17ea1ceULL,0xcdff7efbeb7695daULL,0x5ba8ef31e9771ed6ULL,0xdbb4dfbe6c7717deULL,0xddd7fe23277493dcULL,0xcbfe5f52637635ceULL,0xf6fffffee7773fdcULL,0xfffdff7f677f59dcULL,0xdbbbef776f7f7fdaULL,0xe5d5f73ee774a1ceULL,0x2a51845a60245850ULL,0xc4f75fe92e769fdeULL,0xb1e9afd823724e4eULL,0xdeae2f4ee37617deULL,0xecf7ff7eef7f9fceULL,0xac363b54a67210c6ULL,0x7005150a2640850ULL,0x98e53e28666259d2ULL,0x68163c63604e145eULL,0x8000e1302234000eULL,0x140021000000ULL,0x829005aa60435910ULL,0x8442046100611440ULL,0x90102140202400c0ULL,0x0,0x2842000e22021016ULL,0x80c00160200800ULL,0x4441628a4700042ULL,0x24402860602010d0ULL,0x8400004000060050ULL,0x490812140a441448ULL,0x0,0x0,0x4224000000ULL,0x8050400520002ULL,0x0,0x0,0x0,0x40020008000ULL,0x10021000200840ULL,0x80000400000140ULL,0x96134322723004ULL,0x28004100001450ULL,0x9201034882161002ULL,0x74b0255a227615deULL,0xa8e430836270154cULL,0x6d901ec26a75204aULL,0x8223444a00471146ULL,0x420000000102000ULL,0x99ea950225485c6ULL,0x50c00c00640048ULL,0xa413289207e084cULL,0x25044000345800ULL,0x900f128000630848ULL,0x4022010020220800ULL,0x0,0x84a3235025e11deULL,0x2330105e027c1c50ULL,0x60ae005022671248ULL,0x70ca8f30a21c114aULL,0x16900350236600c2ULL,0x80d7eb92a57c94deULL,0x8901d4f227619c4ULL,0x1c1f6f77223f11c2ULL,0x73ecb6f5eb7f15d2ULL,0x96962e49227f48ceULL,0x69dfdf7ae2561456ULL,0x3b5a36c7635e95d8ULL,0x80808640000110d0ULL,0x134c97226a74148eULL,0x4a9e29f22b5e14ceULL,0x8202e360201a81ccULL,0x2d0060124720006ULL,0xf29147e0e6571450ULL,0x86fff2eb227f994cULL,0x4082fc10816a0a42ULL,0xe996f336a6779dcaULL,0xbc921e07624681ceULL,0x9bec7ca207658c6ULL,0xf5ffd7ff23775dceULL,0xb600e173227488ceULL,0xcc20022101400ULL,0x4144881422644808ULL,0x8480221620880ULL,0x60201630a270100aULL,0x4100040000301800ULL,0xc220214228620048ULL,0x0,0x800000800102000ULL,0x401012000001448ULL,0x0,0x800050000420240ULL,0x2026000100006ULL,0x8410c00222101050ULL,0x808000900300044ULL,0x0,0x2040020008000ULL,0x8120007021001010ULL,0x4080000004800ULL,0xfffffffbea7ffddeULL,0xa55c3268a256104aULL,0x50626c4060060000ULL,0xc380b215e07400caULL,0x4847b20423509248ULL,0x14026623000240ULL,0x0,0x18000420000ULL,0x80800014a0200800ULL,0x40070a3060248840ULL,0x8a548f1b62348844ULL,0xf0a8c66822761554ULL,0xdffeff7fef7f7fdeULL,0x928c2b40a774034eULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x91fdaf7a9e6cef1aULL,0xdffebd739f7cfef6ULL,0xf9e7f7368f3cddf1ULL,0xedc332b18a3ddf98ULL,0x819dd7f34e75e639ULL,0x35431765001dc791ULL,0x73eaef615318dcf0ULL,0xfbbcfbfddf1cdedcULL,0x8dd3fcbb9a1dfefeULL,0xafdecfdf8b3cfefdULL,0xf4f87e7e4b3dce59ULL,0xfffde7f98f3df79dULL,0xf7920b658b20ce92ULL,0xccfef6be8d19fe96ULL,0x6e5e5ade8b7dffd1ULL,0x5c9f7fe70e7cc6fbULL,0xdd5ef7fd9f1ccfffULL,0xdeae2f42881ccfdeULL,0xefd6cdbe0b5ce7dfULL,0xebfedfffdb3dfff6ULL,0xbeba7ffa9e3dc7fdULL,0x93bbf77e8b1cbe9aULL,0x9efe7ffd1e3dffd8ULL,0xe8c4f26a9b088ff8ULL,0xeabe9ec69f6cdc58ULL,0xb2b3fef2cb3dcedeULL,0xfd7dff65df3ddfffULL,0xcb3e7ff24e7dd7dfULL,0xf9f5fa36ce3dfefbULL,0xdf7fffffcf7ddfffULL,0x7eb84f8a091cfefdULL,0x70f1dee04b5df7f2ULL,0xdbbfff628b7dfefcULL,0xffbdff6b9f3ceffbULL,0xb8c7fe129b197ff0ULL,0xfbfffd3de3ccff8ULL,0x76d586e59f5cef98ULL,0xd97eb77a0e34f7dfULL,0x9bbe7f679f3cdff5ULL,0xd9f7ff3ecf3ccff7ULL,0xff5fdfdacb7dfff9ULL,0xbffe5ffbdf7cfff9ULL,0xffefffffcd7ccfffULL,0xdefe3d7f8f3ceef2ULL,0x83d5efb7c91c8cfdULL,0x39ef3f4d4f1dced4ULL,0x9cdb5fe85718cefbULL,0xedffbf768b3deffdULL,0xfabe47e5db1cefb5ULL,0xe33cfb129e3cdcfeULL,0xefffbbdf9f3decfeULL,0xbfffcffbce7dfffdULL,0x95f6f26f1f7ccbf4ULL,0x7ffefff34f1dcff8ULL,0xfbffff2f9f2dff91ULL,0xde7fbf8f8f3ccf9bULL,0xa69eca028b1cce9eULL,0x30405224060c4600ULL,0x4bbe2df08e3dcae2ULL,0xc190357a891cde98ULL,0xc782aef3097cdad5ULL,0xb7de5f9b8b1cef9dULL,0x74bdde6a4e7cceddULL,0xbdff75638b2cefdeULL,0xbbbb3f7d9b3dffddULL,0xf90ffb8e9b6c3efcULL,0x2fb65f5e1b7ce47fULL,0xb09f5beb855cce97ULL,0x99faefe7da388cf8ULL,0xae9927ad4f1cfff0ULL,0xebdff37fdf1ddffcULL,0x2e7798d91f3dead2ULL,0xb7bbfff89f64ff98ULL,0xf16cfebe8b19ffd0ULL,0xffb625e2421c9297ULL,0xa9c5ffeb9e3d7e7cULL,0xe9f33ed79b3c9eb8ULL,0x1eb5dff98a1cee3dULL,0x55fefee55f7cdf7bULL,0xf2fe7df9cf7defdfULL,0xeb56f77b8f7ddfddULL,0xfa6bfff95f3ddffcULL,0xff9e0ffb873dccfeULL,0xbdb1774e1b5dfed0ULL,0xddf6ff731f3de799ULL,0x4105fe5f9b0ce4d1ULL,0xcfebfebada7ddfbdULL,0xffdbdfdb8f3cffdfULL,0xfeffffffdf7dffffULL,0xfbefffffdf7dffdeULL,0xffffffffdf7cfffdULL,0xefffffbfdf3dffdfULL,0x29bddddc0f01f7daULL,0xb4567ce6cf1cfcdcULL,0xe9fdbf71df1deebdULL,0xff920df78a3dcffdULL,0x89edd33a8b2cef3eULL,0x80000108002000ULL,0x85919250021c80d2ULL,0x9294f70f9b1dc997ULL,0xcf6e0d208b18ebe9ULL,0x8104830013108c80ULL,0x494008201010ULL,0x6984d08811948d0ULL,0x251c906243108a11ULL,0x51244211a002108ULL,0x202020480100480ULL,0x6c8a92010808cad2ULL,0x4610440000c0014ULL,0x40040008002000ULL,0xc2a00840080ca041ULL,0xfb906710cb3cccd0ULL,0x28879e04081c2080ULL,0x261000180800c881ULL,0x10000200000408ULL,0x0,0x8000000400080400ULL,0x6000000002000200ULL,0x2100c00008201010ULL,0x0,0x0,0x8aa60e6082040c98ULL,0x884099b0004c0484ULL,0x8100b4000109032ULL,0x100003882000a00ULL,0x1124042002080430ULL,0x62860ce08a1c4ad0ULL,0xc44310010c4460ULL,0x200895c28a080d18ULL,0x4098442802088098ULL,0x500081c009214000ULL,0x40490000108022ULL,0x880410841108802ULL,0x94c8f243c554c696ULL,0x7a38cfdb5f5cf6d8ULL,0x36d256635b3cdcdcULL,0x0,0x4100008002000ULL,0x54a0e0a4a3090d8ULL,0x234641d088195e90ULL,0x680040009004808ULL,0xbfcfd7b8b2dcebfULL,0x7e982f549f29df5bULL,0x4d298904218c0b0ULL,0xbf1d45b4e3cefc3ULL,0x35ddfdf7cb1d5cdcULL,0x400083a0080c840eULL,0xeaae5de2097cc4e3ULL,0x4cd7f37e1f3ddff8ULL,0x2fda1edb0e1ccdddULL,0x35386a72cb39cc52ULL,0xf0e573f51a2dd4d8ULL,0x4a447524801c9694ULL,0x8800d2000b204900ULL,0x2408b00002089492ULL,0xb49f4fea1f20dcdbULL,0xd577f67edf39ffdbULL,0xcaa23e02823dcd3eULL,0xfbdcf73e5a3cbedfULL,0xe843a2934b1059d8ULL,0xb574df48da389df3ULL,0x5491a66c9b38aa99ULL,0xfbacfc620a0d7fd8ULL,0x6840804808140084ULL,0xc8b980208284c88ULL,0x40054280140a04ULL,0x36340460081860d0ULL,0x1094010808140640ULL,0x0,0x110902080208a10ULL,0x44000000008010ULL,0x10002000080080ULL,0x1080006001080810ULL,0x4400000000008010ULL,0xc00240108080ULL,0x9100100054080ULL,0x0,0x2086000100c001ULL,0xba323d720b28fed8ULL,0x8043a22a0b204a18ULL,0x7eff7e58cf38dffaULL,0xffbfdff3df3cffffULL,0xc00fa0621a2c8514ULL,0x6000440001000a0ULL,0x4940f4228600d450ULL,0xfad73eb7cf11ff58ULL,0x2b19de490201cd18ULL,0x4000002002000010ULL,0x14ebd604b344558ULL,0xff8b6b7bcb7dcfd6ULL,0x60573e210b3cc5ccULL,0xa75edfde5b7deedfULL,0x74bcf5e84b7cafddULL,0xfffefffbdf7dffffULL,0xd6bebffb8b2cdf59ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdffff7ffddbadfffULL,0xdfbe7dfbcdba0fffULL,0xfbbff736dc9adf6bULL,0x8e03b74915ba05a7ULL,0x8e9c4f7a752a0f6bULL,0xfd507eeb6d2a5fe1ULL,0xdaff2f72a4ba57efULL,0xef26ff7f4dba97ebULL,0x8fffffbbfdba8ffbULL,0xef9cffda7dba07eaULL,0xf43efee74d9a5f63ULL,0xd3ab6f5935920fcbULL,0x8bb26a597c980722ULL,0x8cf2b7371c884db2ULL,0x6f5edb1b7d9acd4bULL,0xf7ff5febfcba9febULL,0xf9f3af68dd8a5da6ULL,0xaebc2d7115b80523ULL,0xcfafffbd753a0576ULL,0x4dffbffecd3a0f73ULL,0xffbadfff5dba9febULL,0xf1feff314dba4fefULL,0xffdc7d301dba9747ULL,0x9afd3af2cba87e3ULL,0xbee7bec6649a0fe1ULL,0xdd3e6fef659a09ebULL,0xe0fff3eb3d2a977fULL,0xdbee7df26dba8bebULL,0xfbc4f31f7daaddcbULL,0xccfbb7c9dd3a5ff7ULL,0xaf30df83d11a075fULL,0xf4d1d563c5289fa3ULL,0x99fe7ff3c5badf6bULL,0xfd8ee77ffc2ad7efULL,0xc9effe3b4dba8febULL,0xeedddddafcba1deaULL,0xf4b8ffe8bdb84febULL,0x98ca77374c3a9f6bULL,0xdfbebffa75bad7ffULL,0xeddffc8aad3a0f63ULL,0xef7f5fdffd3a9fefULL,0xfebfffe6fdba4ff3ULL,0xf8efb7fafdba1fe3ULL,0xdef7af537dbacfe3ULL,0x811dd7332c38c7a3ULL,0x297dfccd3caacde7ULL,0xb339cde3dd9a075bULL,0x9d7dffeeddba577aULL,0x7fbe3d7c5cba1f6bULL,0x8976ffaa64bad761ULL,0xabeb3ffbfdaadf63ULL,0x84ceccd6481a077bULL,0xffddf378fdba4f6fULL,0x7afefd737d3acfebULL,0xabd5f732fdbacfa3ULL,0x6c5f9c1af13a852bULL,0xf69855705c8a5762ULL,0x38d2476c243acf63ULL,0xfb7ab9702d1a0f37ULL,0xfbb2e77fddba0ff3ULL,0xcec3feb35d88c5b3ULL,0x950118954820542ULL,0xf6f77ff03dba57afULL,0xd34ef57fedbadfe3ULL,0xdb927bf57d3847ebULL,0xccebd71a74a291ebULL,0xbb7efc5b1cbad7a3ULL,0x347b57eed4929fe3ULL,0xfdc4f74e74aa5febULL,0xfeae3f7eecaa07e7ULL,0xf3f4f37f6cba57fbULL,0xac89a5b05381753ULL,0x359bdffac5badfb3ULL,0xfbfdfef7fdb20ffbULL,0xdfbefffffdba8fefULL,0xcba3f36e65ba57e3ULL,0xedde7e55fda80fabULL,0x43bc4ef1a49a5f62ULL,0xfaf3377a9dba8fe7ULL,0x7da65d6b55ba1ffbULL,0xfbffffff7dbadf7fULL,0xffe7bef9cdba4f7bULL,0xbbbfef787dba4f73ULL,0xdcd6b6eed83acf83ULL,0x93b2ad412dbac7c3ULL,0xfbb5fbffdcba4df3ULL,0x4ddffeadfdaa0f6bULL,0xffffffdffdbadfffULL,0xfffffffffdbadfffULL,0xfffefffbfdbadfffULL,0xfffffffffdbadfffULL,0xffffffbffdbadfffULL,0xcbb6ff5b7daa03ebULL,0xbdb77ff7edbadfe7ULL,0xcd4d676395b857a3ULL,0xfe8e2f70bdba1f63ULL,0x408847121c201702ULL,0xf81a11c4900520ULL,0x930247d2358a15f2ULL,0x53bfaffe5dba17e3ULL,0x74823da18cba4d02ULL,0x8280e56e901a4ce3ULL,0x104c0000280020ULL,0x1c469044b80700ULL,0x9494be4c0cba4b42ULL,0x2004002000020140ULL,0xb890001804085701ULL,0x3880020820100ULL,0x880289100c8001e0ULL,0x71f4b6fe4dbacf7bULL,0x0,0xc28003044c200140ULL,0x848a38008c8a4404ULL,0x2a51d4502cb219e5ULL,0x4000106004300200ULL,0xa84b2f5a5daa4de2ULL,0x0,0x801001020000012aULL,0xb28ecd8008a08cbULL,0x8002000020004ULL,0x2004060818100500ULL,0xfebe3d579dbacfefULL,0x0,0xed563a008cba0504ULL,0x8000442000a80100ULL,0x20261080020452ULL,0x100020000280000ULL,0xc862f7fa15b85563ULL,0x84606000c101100ULL,0x64a05120c2a0906ULL,0x1024008040020060ULL,0x92a445410d1a0921ULL,0x364c62ea12a47e2ULL,0x5c1794840920da3ULL,0xa73e4fea4d3acf2bULL,0xe0b7ffeb6dba9573ULL,0x93b00c5005384702ULL,0x4204020601000502ULL,0x4140001000020840ULL,0x697509427cb80532ULL,0x3420462449800183ULL,0x102514030a00300ULL,0x9baf633f6cba0f3fULL,0x8c81ca080ca85140ULL,0xbeedfd83d3a1febULL,0x14d7fe6bfdbacf2bULL,0x911aa5602da803e3ULL,0xd2b67f768dbadfebULL,0xcdffff264d3ad7abULL,0x23d9dad70db80760ULL,0xfe2a4d7f4dba8fefULL,0xd3229742148a9d23ULL,0xdbde4dd37dba0f63ULL,0x84001080c820000ULL,0xe48306cf041a47b2ULL,0x35f96fd3d5ba4fe3ULL,0xffd7f77a7c3adfbbULL,0x5bbe57f274185daaULL,0xfbdc677ee53897cbULL,0x8cffbf35fdbadf67ULL,0x8bbccf4244ba5f67ULL,0x4dd45668cd08553aULL,0xb0122d050cba0d26ULL,0x8080001410020900ULL,0xca8218060018450aULL,0xeb6ec9cb0cb89b6bULL,0xc4941021481a0d01ULL,0x800481080c020304ULL,0x5a24252044081833ULL,0x141e8200d220246ULL,0xa000841040280120ULL,0x904400010000a2ULL,0x1000040000020040ULL,0x4002008000100ULL,0x0,0x4480001000822ULL,0x20680040001a0ULL,0xd07d2c1231280f0bULL,0x888425202c208141ULL,0xcb847a6e0dba5fe1ULL,0xf3d76ed5fc3a4debULL,0xdfbecfff7dbadfffULL,0x8001004040000022ULL,0x18848400000030bULL,0xf9d6ffbf7dba5ffbULL,0xffffffbfddbadffbULL,0xbfbfdd7fcdaadf6bULL,0x5424c3606420572aULL,0x6918b531593a0d27ULL,0xd8a47be1248a47e3ULL,0x8dd3be1dbd289733ULL,0x6f5fefdfcdba1fe2ULL,0xe6b776f31dba4febULL,0xfffffffffdbadfffULL,0xfffefffffdbadfffULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x94fb5fe0b0136ea6ULL,0xffffffffbcffffeeULL,0xdffeef73b4d3f6eeULL,0xdb96fbf6bc9cefeaULL,0xd947bf5b94bbfbe6ULL,0x8f9ece57b4bdf7eeULL,0x14c777e934135beeULL,0xcbbaff63b4bafce6ULL,0xffaff797b4b6f3eeULL,0xc95ef6b31cfb6ce6ULL,0xaf76ffdb3cf7fbeeULL,0xe4f3f6c634d16de4ULL,0xdd6ff77bbcd7ffe6ULL,0xdfbe3f77b4bfebeeULL,0xcdf7f33e34baf6ecULL,0xafdfdfcb34d37aeaULL,0xfdd71ffba4f2f6e4ULL,0xfb4ff7f334bbffecULL,0xfa9e3fff9cb6fee6ULL,0xe9c4f13630bf7ec4ULL,0x2ffffefeb4b7ffeeULL,0xbfbffffbb433ffeeULL,0xdd7cb77f3cf26eeeULL,0xdfde3f77b4bfceeaULL,0xfbfffb7e3413ffeeULL,0xefffde5f34977f6eULL,0x39bdcf6f94d7f6e6ULL,0xf7f7de6db4baf6eaULL,0xffbe3ffbb4bb7feeULL,0xfb70ff2f3c92feeeULL,0xe7da795904d07ce0ULL,0x7ffbeffbb4b3ffe6ULL,0x7cf55febb4d7f6a6ULL,0x9fbfbf60b493f8eeULL,0xd880db3e10b2feecULL,0xa77ffebf94da6ee6ULL,0x779edfdeb493fee6ULL,0xb43f5ef23495fce6ULL,0xc9cf357bb4137ceeULL,0x1994aff92493e8ecULL,0xeccfea9e34fa6de6ULL,0xe3ffd7dbbcb6fee4ULL,0xfdf65ffbb49bfeeeULL,0xfbedf7ffbc93fae6ULL,0xfe3e3f76b493fce6ULL,0xc7debe2eb4b27eecULL,0xb56fede3cdee2e4ULL,0x9e0a6ffb1492fdecULL,0xf5edaf3a349eefeeULL,0xff9e7f633cffffe6ULL,0xcbfbff6f14f8ffc6ULL,0x7ee6fe1eb49b7fe4ULL,0x8a94cd6bb812e9eeULL,0xd4fef6cb34bf7de6ULL,0xebfe7df234b2f6eeULL,0xcbfee39e3cbafee2ULL,0x8cc2784a249a5aa4ULL,0x1a188d13149a30ecULL,0x556631d0041246acULL,0xef2cad600492eac4ULL,0xffbef3ffb4f3feeeULL,0x5ac3ff133cb666e4ULL,0x2a0f9dcb1898e8ecULL,0xf8f6bffebcdeeeeaULL,0x5bad676b34976eeeULL,0x8b04e26da49412acULL,0xfa5ef7bbb4ba7ce6ULL,0x5e1d7558b0126caeULL,0xc4fed76b149b76e6ULL,0x597c7f0e14b16eeeULL,0xeabf3f7e149efeeeULL,0xafedff3fb4b2fce6ULL,0xec52f24234b2c2e4ULL,0xb7b2ff7b34b3fee6ULL,0x95f5dfebbc9bfce4ULL,0xb9967572bcbadfe4ULL,0xebd6f66ebc9bdee6ULL,0xa0f49fdf3c336ee6ULL,0x6fe26bc394dee7c4ULL,0xd5e5f750a49c6feeULL,0x9ba8bd631cd0fcc6ULL,0xebf4f72f349076e6ULL,0xcfffffff3cdfffeeULL,0x67bdde6814daf2eeULL,0xf4fffee7b0f3f6e2ULL,0x969c2e700412eeacULL,0xdb9dcaab34bb6eecULL,0xdff7b3b1bcb3feaeULL,0xffffffdbbcfffeeeULL,0xf7ffffffbcffffeeULL,0xffffbf7bbcffffeeULL,0xfffe7fffbcffffeeULL,0xffffffffbcffffeeULL,0x2fd67fdb9cbbfee6ULL,0xbdfb7fefbcfffeeeULL,0xd97d7ffa9cb3faeeULL,0xdebe3ffdb4f77eeeULL,0x9c6ef2aac127ceeULL,0x695a3a4994d2f3e2ULL,0x8b20473114123864ULL,0xd189f6df3cb3fbeeULL,0x1d067d2210926a84ULL,0x88001b0010126a40ULL,0x8844a40008a2084ULL,0x210443208024284ULL,0x44406691093a886ULL,0x42000411100220aaULL,0xa314c006b490cc44ULL,0x4244b642a0d04160ULL,0x1880414224904ca0ULL,0x41120220022c000ULL,0x1080240000808280ULL,0x200000400880000ULL,0x940860100901068ULL,0x8102c10204101484ULL,0x3000000000101000ULL,0xc000000020004ULL,0x110000004000020ULL,0x0,0x8080000000c000ULL,0x0,0x2000002000000084ULL,0xe6b2afde3c9a7fecULL,0x4000001010000800ULL,0xef7f7bdf34bb7fe4ULL,0x0,0xc510106b00123aa0ULL,0x8000400000000880ULL,0xc9f6fbabbcffffeeULL,0xeed33ad0b4bffbecULL,0x128b04621482c2e4ULL,0x100807224003160ULL,0x1082040004003040ULL,0xc0d4d82e30125b62ULL,0xeedfd919a4bffde6ULL,0x7ffed6fbb4f37feeULL,0x3df5fffab4bffeeeULL,0xb5192c5534bff6e6ULL,0x4084803028004040ULL,0x8400a0004106040ULL,0x4013c820002060ULL,0x500094c6080062e0ULL,0xc81502280402c484ULL,0x8995777d3c9b6fe4ULL,0xc13ef33834937aeaULL,0x35fc61914d57c86ULL,0xde3e7febb4b26beeULL,0x4f6efffbb49bfeeeULL,0xffbe3fffb4fffeeeULL,0xabb4fbfd3cfe7eeeULL,0x7ffedfff3cfffee6ULL,0x9dbacb31109a7ee4ULL,0xe07da798b49854c4ULL,0xda0425b29492f6a0ULL,0x82328a0218d15042ULL,0x64b05a7814996060ULL,0xfffefffb34b37eeeULL,0xdb70fe7d94b2ffeeULL,0x4fbe3f63b4b67ae4ULL,0xdb9ce33ebc3356e6ULL,0x7ad3f7bbb491fee4ULL,0x25300756a41232c2ULL,0x50d176e224d1a2a4ULL,0x8041400004005040ULL,0x800001480402002cULL,0xa40000100004a80ULL,0xb5e4f5bb412fae2ULL,0x2004c0004304000ULL,0x254080980000ULL,0x1204156000880280ULL,0x8940c80080106160ULL,0xa802000030000280ULL,0x100004e200114a80ULL,0x20080010010000ULL,0xa80000000121004ULL,0x80900000906040ULL,0x50101600111000ULL,0x101001a014200802ULL,0x0,0x800680000004a80ULL,0x418186200088e2c4ULL,0x79d21edf309b75e6ULL,0x18cc69014105c28ULL,0x18105040200048c0ULL,0x0,0xebddf77a1cfafeeeULL,0xefffffffbcfbffeeULL,0xffbfddfabcdfffeeULL,0x982060008a0064ULL,0x22ba6d31340238c6ULL,0x7a0c009a4484ULL,0xdfc33e37b49afceaULL,0xa7d6dfd734bb4ee6ULL,0xf4f5bfedb4fbefe4ULL,0xffffffffbcffffeeULL,0xffbfffffbcfffeeeULL,0x0,0x0,0x0,0x0,0x0,0xc7e6260224102068ULL,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x850840100022c0ULL,0xfe9efffbf0fd57faULL,0xfffffffffafdffffULL,0xfbfefd63f8fcfffeULL,0x2819334100017eaULL,0x8b621eddd095dffaULL,0xbbbfeff9d095d7faULL,0xf0ff7ef79a1dbeebULL,0xfbfeff72f2fddfeaULL,0xc3b7bf7c329d77eeULL,0xdfc7bfbbf0fcd7ebULL,0xaf9fdfd3d035f7edULL,0x9cd4e76852d4b4daULL,0x9bff77e3b25cfffeULL,0xdbb12579f2fcdb70ULL,0xedffeffef8dcfefaULL,0x2ff9dfdff27cf7e8ULL,0x36bb5ff67a9cd6fbULL,0xf7fcbff8f01d5efbULL,0xdabab77ff27cdfe9ULL,0xefdff3bff27dffffULL,0xeffedfdff2f5fff9ULL,0xbfffdffffafcffefULL,0xddffff78929dfeeaULL,0xbfbe3dfed0dcdfdbULL,0xfbffffebf8fcdfefULL,0xefdffedff27cfeffULL,0xdfbffffffaddfeffULL,0xf5fef7eff2fcfffdULL,0xfffe7ffaf2fddffbULL,0xfbfdeb7fd2dcd7fbULL,0x87e7fff9f23cdffaULL,0x2fbb4f5a927dd3e8ULL,0xf9f3fefc90fcf7caULL,0xbefee533f8557eeeULL,0xeb93efb9d09cfef3ULL,0x8dd1380a783198e8ULL,0xeffedf9ff27dfffaULL,0x74fdfefab29c7eedULL,0xc0ea2741c0c09469ULL,0xdbdebaf61874b7d8ULL,0xcec7ff127254be70ULL,0x4ffedd1fd29d9ffcULL,0xcff27ffbf274ffeaULL,0xdbfeef7ff25cd7efULL,0xfebebf7372d5f7eaULL,0xf73affba78d4fee4ULL,0x23563e583030de62ULL,0xf6aee36ab01c37f2ULL,0x536c9773f214b7eaULL,0xcf367f7e925dffeaULL,0x8b82d7af32d4be7aULL,0xeed70b77f0b1deeeULL,0xbffc6fff927cffcaULL,0xd5f7f7ffd2fcfee7ULL,0xffee7de2d0bddefaULL,0xfb86f7367a1c1cfaULL,0xcfd7bbeffadd5ef8ULL,0x23becd31921c367aULL,0xbeb7fedc10dcb6caULL,0xab33af52901cde86ULL,0xd3bee3f2b0fcf5ebULL,0x8bd6fb1638941ef8ULL,0xa94649c2b214d4e8ULL,0x7664fee5b27cbef2ULL,0xd973ee71d27436ceULL,0xffb6eb76529cd6eaULL,0x68d3e73a588096e9ULL,0xeb6e7d5bda945fcaULL,0xb42ed7c3d08c14a8ULL,0xd92237e6d2b4de6bULL,0xfeaa2fdff8fc5ee8ULL,0xe7d2fe725834ffc2ULL,0x6b579ed9d21c1e5aULL,0x96bfcfeffab1f668ULL,0xbb7d9fdb50d5fef5ULL,0x5e82b642d00417eaULL,0x8be6f2cd50b456abULL,0xaef6bf5210b5bff6ULL,0x3efcc5db3014ce7eULL,0xf4ffee79921c17daULL,0x838e7d6150d4f9eaULL,0xd9afeb2e78bc7eeaULL,0x4ffbbe4150d516ecULL,0xe9e5ffb5274f7e2ULL,0xf9f73f7ef21cd6fdULL,0xdffeff73fafddffaULL,0xfafd7fa0f8143673ULL,0xcfdbbabb505c7fdeULL,0xffffffdbfafdffffULL,0xfefffffffafdffffULL,0xffffbffffaf4fffbULL,0xffbf7ffff2fddfffULL,0xefdfffbff8dcd6feULL,0xab7e7ddfd0fcbfd8ULL,0xf4ff77ffbafcfeeaULL,0xb3e22f4b50207466ULL,0x3a960f2292d0dc48ULL,0x4dcadf29fa14d7eaULL,0xffdf7f4ef87c7cfaULL,0x8693757a7034d47aULL,0x11a497f21214dcc2ULL,0x4c800d4230049662ULL,0x7080411200801640ULL,0x61540009104092c2ULL,0xb831457ad01c3442ULL,0x90098640408414a2ULL,0x4aaa2cc0d21c9cd6ULL,0x72806300d01034e0ULL,0x9c22e04200017e8ULL,0x87128f137044d4e6ULL,0x3440526040011668ULL,0x200560200404e0ULL,0x840010080001002ULL,0x600c00812100840ULL,0x800010050000000ULL,0x2220000100800ULL,0x41000040008ULL,0x80000102000040ULL,0x400000800000028ULL,0x0,0x0,0xa080002090000240ULL,0xd2b62d7c7055177bULL,0x110000000001400ULL,0x297d1fcbf2bdd6f8ULL,0x78e45bbd01c16c2ULL,0x0,0x4802240200501340ULL,0x8b3ffbeb72e15768ULL,0x4f4a0e3020215668ULL,0x8880298101c3040ULL,0x1400806010008240ULL,0x44060d22401c0e30ULL,0x2ba4ab0fd004d5c1ULL,0xefffbffff25cfefaULL,0xfbbe5f73da84d7e9ULL,0xf5ff7ffbf2d4ffffULL,0xdbaeed2070b45aeaULL,0x21003040004048aULL,0x840021110d01410ULL,0x2b4034810346640ULL,0x2004024010204400ULL,0x8824200020801424ULL,0xa2144f647060b068ULL,0xd000993850115768ULL,0xca0ddb5b1214c7d0ULL,0xce3bdfe9f26dd6eaULL,0xfffdeff7fafcfefaULL,0xfe9fbffefae4f6e8ULL,0x71d6d7edf295ffe0ULL,0x2bbfdfdff27cfeeaULL,0xc60226e7fa909f61ULL,0x97fdbf9a32347ffaULL,0xdf044d7fb2545c7eULL,0xe924f32f3210de68ULL,0xae4316eaf8101ffbULL,0xbeffefff6afcffebULL,0xff75f67d52b55efaULL,0x6c9631615a1cd542ULL,0xd9d1ff265aa4d7e2ULL,0xfd4bbf9ad014fef6ULL,0x6b1e1f7fb0955761ULL,0x2052fd40141cf8ULL,0x3082002000250062ULL,0x8000010010100140ULL,0x400260140801020ULL,0x8475d529014149eULL,0xf7ed5dea783cfffbULL,0x98e10720921440d2ULL,0x8820090210100480ULL,0x8081020000011022ULL,0x830040020100220ULL,0x8010040020100820ULL,0x24000000000401ULL,0x1020060c10101218ULL,0x4102c30810841008ULL,0x540022818001408ULL,0xa00004000ULL,0x44410000000c240ULL,0x3000000000005000ULL,0x1800080001000ULL,0x44c1ae2830b99ae0ULL,0xa0444000811028ULL,0x8420000000105200ULL,0x10010000000240ULL,0xc9dccbbf787cf6faULL,0xffdfbffbfafdfeeeULL,0xef3fddd3f291f6ebULL,0x0,0x91c023008141ac0ULL,0xfb966d66b8fd17ecULL,0x4fcffeab7af5f6fbULL,0xeffedfdaf0f5d6e9ULL,0xfdfffffef2fcf7faULL,0xdbffeffbf2f5f7fbULL,0xfff7ff7ffabdfffeULL,0x0,0x0,0x0,0x0,0x0,0x2080000001008ULL,0x28021108121012c0ULL,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa8f61bd0a9e4a36cULL,0xdfbafffbf9fd23daULL,0xf5fff76df1ed2b7eULL,0xfffefdf3f9fdaffeULL,0xfb7ce3ae81ed2e5eULL,0x9fdbbe9ff1eda36eULL,0xbf9edfdbf1efaff6ULL,0xd1d0ffecf9eea3c6ULL,0xd7bffd62d1c7af5cULL,0xdfae6d6fd9efa76eULL,0xdfffffbff9eeafdeULL,0x9f776fdfd9eca34eULL,0xf4b5de7cd9e52b74ULL,0xed6cff73596d365eULL,0xbbbcfb7fd1f4ab7aULL,0xfdcffec2f9ed2b6eULL,0x2abedf9be9dfae4cULL,0xb7ffffffd9ff2e76ULL,0xf76f3f6af1652a7eULL,0xfebe7ffef9fc237eULL,0xc58ef63fd1ed261aULL,0xafdbffdbd1ff22d6ULL,0xb7bffffbf1a7bbfaULL,0xdf74dd19d9a4388cULL,0xfbb47775d9fdbbfaULL,0xf9fffbfff1ffb3feULL,0xfef7dfdfd1efb3feULL,0xfabbcdfbf9ffbafaULL,0xd7f4ffe8d9efb34aULL,0x7fbe7f72d9ffa75eULL,0xbbb5f75ed1bda34aULL,0xcfeb7feaf9cfb37aULL,0xebbfdffbe9eda3ceULL,0xfdff57ffd9ffb35eULL,0xfdaeff77d9f7abfeULL,0xeb38677ff11f237eULL,0x4bfcfc27f9e5ba50ULL,0xebfecfdff9ffabfeULL,0xf4faffecd9f5abfaULL,0xd4eebf71d071b34eULL,0xf3bc2f77c9edb77aULL,0xe8f5f902d189aa40ULL,0x2bfedf92d1eda35aULL,0xe5ff5fecf9ecaf7aULL,0xdfbafe5af9cd3bd2ULL,0xdab67fffd9f73bfeULL,0xe1fdff79f1edabfaULL,0x3dfbd759d8cd2a5aULL,0x8382cef3d8cc275cULL,0xd57eff7cd9c4abdaULL,0xef367fe0d1e5a85aULL,0x83eeff76c1f5bbceULL,0xefffbed9d9eda3c6ULL,0xff7f4f62d9ffa35eULL,0x75f6f6f9f9eda766ULL,0x9fbe75f9d1ef3f56ULL,0xf8e4e33ed9dc234aULL,0xee533ffbf9ef235eULL,0x86250970d965a846ULL,0x35bd5fe8f9ec2adeULL,0xc93e4d43f9ed2b86ULL,0xd39e35afd9c53362ULL,0x6ece320ad9c53652ULL,0xa35acfd9f9fdaa4cULL,0x94f97265f13c2016ULL,0x8decffe3f9efa2c2ULL,0xc700eb148145035aULL,0x2587e72ac9cf2352ULL,0x2b37ff5bd1412beeULL,0xbcf977eaf187a3f0ULL,0x510cf170d0ef2848ULL,0x4abd2f32e94ba3f6ULL,0xcdfff374d1e9a266ULL,0xaf5e5e17c8f42b68ULL,0xb9bacfeff9bdabe8ULL,0xd6f5ff7ff9df237eULL,0xfa046534f109a35aULL,0xebf6ebbad9f1bb6cULL,0xc8c71f6fd9f523faULL,0xd8bfdfebf1cfaf7eULL,0xddfb767bd9e5a3deULL,0x7b7efd7af9edb76aULL,0xabf77b7ed9ef3bd2ULL,0x64cb32cac124a24eULL,0xfebecdb8f9fca740ULL,0x9cd5ff62d9f3a37aULL,0xfffafd73d9fdb67eULL,0xc83cf7b3c9f1a2deULL,0xcfdffe3df1cda3faULL,0xffffffdff9edbffeULL,0xfffffffff9ffbffeULL,0xffefeff7f9efaffeULL,0xf3bebf73f9f7a3feULL,0x8dc3ea9ec1dda2deULL,0xafd8cd4bd94d275eULL,0xb6bf5ffbf9edbbfeULL,0xd1b60f32f0ec235cULL,0xd3b22362c8e92b52ULL,0x46d3f3bad19cb24aULL,0x49da9ab198e7a24aULL,0x35ba4168d1c0a2ceULL,0xd9f0a25281692b6eULL,0x1a00652dd1842062ULL,0x88202a408141221eULL,0x23418880c10020e2ULL,0xb0137ee2a167237aULL,0xc32428c809d0258ULL,0x90040002000ULL,0xa0c20201002006ULL,0xc4a22c2d0402390ULL,0x25bb4d60c04c2b1eULL,0x1001000001000008ULL,0x600180002800ULL,0xd8000a240143024aULL,0x1020000082000ULL,0x400400000000210ULL,0x16000044c0482200ULL,0x8014056088800840ULL,0x2080011000000ULL,0x800d00040001140ULL,0x2301108209400212ULL,0xa430f024c0a82042ULL,0xc24040402200ULL,0x12102724d1412ad2ULL,0x5350650aa9c92372ULL,0xeffff95bf1cdabfaULL,0xf7bbc5fbf94dabcaULL,0x41000000000140ULL,0x52a02c4cf16c2252ULL,0x8976f28cd0282042ULL,0x8ff63f5bd1692a46ULL,0x88a8b52c0c02712ULL,0x11635041d0402470ULL,0x262521c0682140ULL,0xdbdcfbbff1efb3feULL,0xcff3febad1dfa35aULL,0x8fbefffff9ff275aULL,0xfdfff3f8f9cebfdaULL,0x9b78fe73d1cfa762ULL,0x1004000000000210ULL,0x2982080510140ULL,0x6e759f4bd9c5ab5cULL,0x24008400000ULL,0x202006524124200aULL,0xcb10767cd1af2a5aULL,0x82ffeb3ed82c3a46ULL,0x63fe9f5be9ec2686ULL,0x1da637e761e7a26aULL,0xd97fffe2e9fdaf7cULL,0xdcbe2f76f9ed23faULL,0xffdefb6ef96ca37eULL,0x2dfedfdaf9ed3adaULL,0x9ea26f7ac96c2a1eULL,0xf5bde472f9cfae6aULL,0xffdeef26c1e52b5aULL,0xc9f7ba67e1ef2356ULL,0xeede1662c1a6a25aULL,0x931dc2fad1c1a3c6ULL,0xd576fefee9ffa3ceULL,0x7c246cf3c94523faULL,0xfbdcff3ef9efabfeULL,0xeecef633d17cb3d4ULL,0xc4760724d9d4231aULL,0x95910c68d0a0a862ULL,0xa80000000012044ULL,0xe0008418d08020c2ULL,0xcbde1213d8f42946ULL,0xaa00ed8bc9f02282ULL,0x95bd4c20616ca332ULL,0x100302088500000ULL,0xca000706912c2800ULL,0x3880000240200ULL,0xc50c210c0412480ULL,0x1b2067b516d03d8ULL,0x5000168040040270ULL,0x800000800400200ULL,0x89461faa718c835aULL,0x140000040008000ULL,0x48000202000ULL,0x43010000c50000ULL,0x8865440c0182800ULL,0x10880000800ULL,0xc873ea0d8082142ULL,0x9094c509004286cULL,0x0,0x400045048402000ULL,0x53d4c42ad1ef3268ULL,0xafffffffd9f7bf7eULL,0xafbcdb50d1feaa54ULL,0x8400102090400100ULL,0x22322500d060a9c0ULL,0xcaabf777f9ef2f52ULL,0xdf46bea7d9ada27cULL,0xaf7f5fdaf9fdbffeULL,0xd677defdd9cf23c2ULL,0xdfeefff3f9fdbfdaULL,0xffbfaffff9efbffeULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xab46fe22d53ca502ULL,0xfff63fda99fcff60ULL,0x16bdcf1b9f38e702ULL,0xf77df3ffddfeff42ULL,0xfffefdfbdffcfff6ULL,0xfbfefb7fdfeaff52ULL,0xeef3b7fbdf6eff36ULL,0x179fc9f3d9ecff30ULL,0x12b3d6e898bce7f2ULL,0x9fee7f71dffcf7a2ULL,0xdbbcff6ddabafd60ULL,0xcfd2773bdfa8e5e2ULL,0xefffcfd7dafcbfa2ULL,0xf4ead6e48f6ce686ULL,0xd9a54e6b9d7ce740ULL,0xd39c7f16dfec7566ULL,0xcf63ffbfdf7ef760ULL,0xfb7dced0d8e8bde2ULL,0xd4f777f6d13cfff0ULL,0x9ffde45b1bbcff42ULL,0xfb9effffdffcfdf0ULL,0xcb5fff2edfccfd30ULL,0xefdffe5bd824ffa0ULL,0x9ede4fbb57acb774ULL,0xddfff7ffdbeefff6ULL,0xfffe3ff7dfecfd74ULL,0xebfffb3fdf7cedb2ULL,0x7edfbffddbfcff74ULL,0xdc9fef7e9a3cff22ULL,0xfdf2ff71dffcf572ULL,0xfffefffadffcfff6ULL,0xfb7cfb6edffeff30ULL,0xeffbfbcedf6ef790ULL,0x3fffcdf3dbfcfd94ULL,0x7df4f7f69beefdb2ULL,0xffbedf71ddbefd20ULL,0xd6945d7f97bcb506ULL,0x45dbbfd118f4e7f0ULL,0x2d9fd7d39fecfff0ULL,0x763df7efdbbce7b6ULL,0xeb6de773df767760ULL,0xff9c7f7ddbe8f7c2ULL,0xc8c3bf3cdfe86764ULL,0xeeffdfdfcfecfda2ULL,0xbffe5ffb9ff8ff30ULL,0xdbedb77ed9f8fd96ULL,0xfaa73f79dafaa530ULL,0xcfe66f261f28f790ULL,0x6ffedfc31a2ebd66ULL,0x7e366dc9fecb772ULL,0xf330377951faf9f0ULL,0xdb86fc4358aeedd0ULL,0xcbdff7ef9a7cff32ULL,0xefff1ffbdffefdd2ULL,0x3ffecf6adbacfff6ULL,0xd5ddf7f8d2aee716ULL,0x69267ee29788e730ULL,0xbf5c30eddfcf160ULL,0x85ef7e99d86cf5e0ULL,0xcff64f6097b8b7e0ULL,0x7765d2699a24e590ULL,0x9bde6d619320af50ULL,0xff86ef7b1b7cbd84ULL,0x4b82a29992e23540ULL,0x3b9c5cd29f64ad02ULL,0x10b87ee91dec3626ULL,0xe9ef6de318fcf742ULL,0xffb88f5e9dec6500ULL,0xc847e22f9f08ad60ULL,0x2d5fff5c5ae8f6a6ULL,0xb4be1bc3dabcb704ULL,0xf1fbfe6a5ffce534ULL,0xdeb0277e58fce9a2ULL,0xefdffbab9a7eff70ULL,0x2eefbecf1b3aff54ULL,0x6bde7e59838ed10ULL,0x947cfdf9dbeca560ULL,0x9b2e242c11388df0ULL,0x6856f35a0fbced44ULL,0x4ede1e1b5ae8f522ULL,0xa3becf77dbb8f770ULL,0xfcfffe77dbaefdf2ULL,0x6ea42ff015acbd90ULL,0xd9a25b7e9d28f552ULL,0xce73ff705328f750ULL,0x68be9f7b9d686774ULL,0xf7f7f77fdfacfdf2ULL,0xdb2afd60ddf0fdd0ULL,0xcca37348187c8d32ULL,0xcb7fbe3b5decfd20ULL,0xafffdfdfdffefff6ULL,0xfeffffffdffefff6ULL,0xffffff7adffcffb6ULL,0xdfb6ef7d9fbcef52ULL,0xc80263985088ac22ULL,0xeafa5dcb9d88fd20ULL,0x963ffde39ba8e5e2ULL,0x990e476a9f40ec24ULL,0x9a8606421ae80522ULL,0x4786de80996e2402ULL,0x297e7e92db6cfd42ULL,0xe825f569a482532ULL,0x803074401200a510ULL,0x6860742102300160ULL,0xabf4eb2edbfc7d20ULL,0xc461a111d406510ULL,0x929ac3889a202540ULL,0x100845800808d00ULL,0x284e156848486522ULL,0xcb81e6a48a44b560ULL,0xa0602c2000884400ULL,0x668058a1f001560ULL,0x44406c010001140ULL,0x19a617d08d642600ULL,0x462226468b244e40ULL,0x810290c018202520ULL,0x490080040003000ULL,0x411000405000120ULL,0x640200002c0100ULL,0x0,0x60008040000ULL,0x0,0x2020000000600ULL,0x50001000c00ULL,0x10042a2c90208980ULL,0xe181e93ad5ac6520ULL,0x4274000910009110ULL,0x1a092a9042a0b0ULL,0x2404002008204400ULL,0x200010020000ULL,0x49d5cf0cd862ff50ULL,0x201002db00006704ULL,0x40100a005084000ULL,0x514446210286400ULL,0x1ac411200a20a440ULL,0xca80a20695a0ed20ULL,0xa846bd489288f780ULL,0x6e9fcbcb9a3cfd74ULL,0x74f7ff6a58fce772ULL,0xfbbcbf7059f8ff64ULL,0x220c24205030a520ULL,0x800000000000a000ULL,0xb65cdc57db38ed52ULL,0x13100000a000500ULL,0x7882460d080f500ULL,0xa31675f85d72e620ULL,0xc0462a3012a28530ULL,0x6ed7da995bb86d60ULL,0xbcbf7febdbe87d22ULL,0xf1618ffc9f28cd60ULL,0xfe872d70cbb8e582ULL,0xed5e7ff29f78f530ULL,0xaefaffc89af4edf0ULL,0xef644f77cb786540ULL,0xb0dcb674db4aef30ULL,0xcb7e5de15bf8fdd6ULL,0xdbb2ffeadf78ef02ULL,0xefd67f5fcbecede0ULL,0x1f98ff83cae8e720ULL,0xd6efaadc1feaf522ULL,0x1000014000088900ULL,0xfbd7e73edafc6576ULL,0x8eebfe719a48fd22ULL,0x3b3c4f32d878ad00ULL,0xf5babefed500ef14ULL,0x80006d010ac08580ULL,0x100400000002400ULL,0x87fefa395facef72ULL,0x2f58c31e1138ad12ULL,0x4210142898c00120ULL,0xd36ea722d9c0af60ULL,0x1b61613d7c8af00ULL,0x20100080020ULL,0xa10040910000410ULL,0xc92178390402430ULL,0xd02020609d200042ULL,0x8a1804300300bd00ULL,0x8300c22180009c00ULL,0x208014108c02500ULL,0x20428008000142ULL,0x0,0x4240d610870c504ULL,0x840800800000602ULL,0x697e1bd693aced00ULL,0x531ccfd94aace726ULL,0x400200040002000ULL,0x4000001000000402ULL,0x415496045508b534ULL,0xebdff6ff9faeff62ULL,0xe79ddfdf9bfefd32ULL,0x2020000240000ULL,0x82900c605d00b500ULL,0xc288631e90c8a510ULL,0x874eb21b12bce530ULL,0x2bfedfd2d8f2ffb2ULL,0xfcf7dfffdffae776ULL,0xfdceff73dabeff72ULL,0xffbfff7fddfefff2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x910f4410a82540ULL,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcbf7f4ab2f926f63ULL,0x7fcf1edaab8e7f60ULL,0x16bdedea0fce7d71ULL,0xfdf7f7f5a7deef63ULL,0xe7fe7fd13bfeef62ULL,0xbbf6ef7e37f67f73ULL,0xbfdabfcf0fcf6d71ULL,0xe7b74fd333b66f61ULL,0xfcff7efdbf37fd71ULL,0xd3bbfd629fb65d43ULL,0xfc9eeb3d9bfe5f63ULL,0x4ff3f75b9fbeff61ULL,0x6dff5dce0ffb6d72ULL,0xbcfe5efcb7de5f63ULL,0xdfffcf60bffedf61ULL,0xca227f260f9e8f70ULL,0x8dd7ef6f9f8f5d73ULL,0x27745f9806de4d62ULL,0xa4f7fde38ade5d70ULL,0xdffffff3bfefff60ULL,0xffbfef7fbfffcf73ULL,0x67a6f72b2f9e7f61ULL,0x2ff67ed393f67b40ULL,0xa6bafffb1fbe7f61ULL,0xfa73ff6b1fde7d62ULL,0xfffe7ff037fe7f61ULL,0xababff3f9fbf7d73ULL,0xeff7ffdf9fff7f63ULL,0x9fbefff73fcf5f63ULL,0xf6d6b37f1fd65f63ULL,0xdbbefd6bbfdedf72ULL,0xfbd6ef7f93df7d63ULL,0xfdf39fff87feff61ULL,0xeef69f79b7d76d61ULL,0x7ff6fe692fde8f60ULL,0xcbfcfd721bce7f73ULL,0xdbb53f2f07efee73ULL,0xfee63e2f06c64f63ULL,0xaf3fdfdb969eef61ULL,0xf4ffffeb8fffff60ULL,0xfda6f76a1f8e7f63ULL,0xf6befff7bf974f63ULL,0xfef7febc9f9edf63ULL,0xabfedf5b9797df73ULL,0x8db9dff307ee6f73ULL,0xd175f66687cef960ULL,0xfeb2bff29edf7d71ULL,0xac4742210a04d40ULL,0xeb7fdfdf97de7f63ULL,0xbeeacdb9879e5d61ULL,0xb1fcf77f979ef962ULL,0xf756ff758fdeef63ULL,0xcac4fb2ea7a6e960ULL,0xebf77fdfafc6df63ULL,0xdfb9f7fb37d65f63ULL,0xfc77f6fd0fde6f63ULL,0xfffeff7b1fdf7d63ULL,0xd9c0edbf8fdecd73ULL,0xfae23ad79aee6d62ULL,0xff3dcfc98b965d72ULL,0x9f21ef88cb86f63ULL,0xbffeff73bf3e6773ULL,0xdbb7f73e9b8e7f63ULL,0x4b423ea319c61460ULL,0xab67fdde83eafd72ULL,0xf4955668a70c3960ULL,0xcdfb77679fdfff60ULL,0xabb16f7d2bcecf61ULL,0x2bd7cf1a0fce5f60ULL,0xbbff5fd8179e7763ULL,0x9093deeb17de5d63ULL,0x95eeffdc3b9f5f63ULL,0xdeb12f6a97866960ULL,0xc7d0fa96177f5d60ULL,0xc09c040396823a60ULL,0xa29a274896d34b42ULL,0xd955b74d33e26f63ULL,0xf6be6dfd9bdf7f71ULL,0xc9f6d7a7af821f61ULL,0xeb469a5406868d61ULL,0x9a9ed77b8a9e5f70ULL,0xf5b7f678179f4d63ULL,0xdfbcfd723fef5f63ULL,0x4b30760a0b087d62ULL,0xca793a9797978960ULL,0xaf3f6f709f877f61ULL,0x71145e88938e4d22ULL,0xdfaffff19fbff963ULL,0xd304af4e09625543ULL,0x5fd73aad9ede7d71ULL,0xbfffdfdb1ff65f63ULL,0xdcfffeeebfffff73ULL,0xfdfdfffabbdfff73ULL,0xf99a777d8bc74f63ULL,0xd5d1b7121ec14f60ULL,0x83efdfd38fae7f73ULL,0x959b7beb23ce5f61ULL,0x95248e600f860760ULL,0xca8520212ac46821ULL,0xdd6e34ba1965f72ULL,0x6fff1fd09f8f7f70ULL,0x411a8ff11ec62960ULL,0x0,0x640244001820160ULL,0xd080020120004461ULL,0x80740e0630c04960ULL,0x1014442000004550ULL,0x10440642000a0860ULL,0x30074001225241ULL,0xe100e67001035a23ULL,0xa480100a83015a02ULL,0x80100000820ULL,0x102cb44004160140ULL,0xcb80c40003904f61ULL,0x800160000c00140ULL,0x80080000020100ULL,0x18104002002400ULL,0x1400000002008000ULL,0x0,0x9004080000080820ULL,0xc00000005000ULL,0x0,0x0,0x0,0x290010401024042ULL,0xdc0c77932de4562ULL,0x0,0x0,0x0,0x0,0xe184900c31861020ULL,0x10000200000060ULL,0x39c5cf005a64361ULL,0xd1e43e6b1f934d42ULL,0x2aa6d401e8c0560ULL,0x8b82043100864763ULL,0x40020000100100ULL,0xff3a097b0f824f61ULL,0xd0d616e8021f7d63ULL,0xdbbcfd7117de7f73ULL,0xe28404e601920560ULL,0x2045b21a00064552ULL,0x8adfd7530ffe7f61ULL,0x14218ce004c05b40ULL,0x42206502012a0060ULL,0xdbfc776f8ff77b43ULL,0xe15cf93a067e4d61ULL,0x13de5f5117877f61ULL,0x76bb6feb1fde7d62ULL,0x506ca4748fa60760ULL,0xc0863f5c1ef66962ULL,0x4dc1f78e87de4d62ULL,0x2d57dfcf058f5f70ULL,0x89e103063d40ULL,0xd231aa5516ee6570ULL,0xd1de7f778fe67f63ULL,0x8bffdb7f3fcf7f62ULL,0x4fca5f971fd34f60ULL,0xbe3d67ec1b8edd63ULL,0xb737f6763bd64761ULL,0x59ca3f420fce6f62ULL,0xd99ccfae03565f63ULL,0xcccadf1f8f375f60ULL,0x299a470a2a8c6d63ULL,0xf85bf25087867f41ULL,0x8622d4104860920ULL,0xf1046f1406122d60ULL,0x201800100c04001ULL,0x804ac158008c0b60ULL,0xf4bdfce08fce6d60ULL,0x8075a500010b1e60ULL,0x21882c15001a6c21ULL,0x80811001020400ULL,0x2266d292169e3120ULL,0x0,0x9de24942a4824860ULL,0x40100000400ULL,0x4480060004020860ULL,0x410020800000d40ULL,0x0,0x1000040000000060ULL,0xa14043000000470ULL,0x0,0x10100008000400ULL,0x19888dc91f825561ULL,0x8001006010880560ULL,0x0,0x4400000a000000ULL,0xefcfbe738fd6ff63ULL,0x2b1ec75b95ea5b61ULL,0xb0feaee806966b40ULL,0x482645c100820740ULL,0xcbe0c538861a4d62ULL,0xdff6da0c16d3ad60ULL,0x6fbfdfdabfcf7f72ULL,0xb4def6e98ac6df63ULL,0xfffeff7b9fffff71ULL,0xffbfff7da7ffcf73ULL,0xe001890884840900ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x461226ccd3047d78ULL,0xcbdbfb3ff7bc7eebULL,0xebef5f1bf7b475e1ULL,0xb7bddff3ffac6f68ULL,0xfd7ffff6f7947ff8ULL,0xcffef7f3f7bc7fe2ULL,0xa354f76ed6bc7472ULL,0xf9e3fffff79e7df8ULL,0xffaecf7fffbc7febULL,0xfbdf96ead7be6df0ULL,0x3bfefff2f61c3ecaULL,0xff957e7ff5b46ef8ULL,0xcf5fff17feac3ffbULL,0xa79e5f42d61c2dd2ULL,0xe4bbf668f4147ee0ULL,0xffeff77bf7967de3ULL,0xffe56f50f5946be2ULL,0xe7dbffabdebc7efaULL,0xebfeffd9f79e7f6aULL,0xf5ef7dfbd5bc7ef2ULL,0xdf7def62d7be37e8ULL,0xdf8fbffff6bc7ff8ULL,0xcf56ea78ff1c75e0ULL,0x8d53be47569e5690ULL,0xb0b6fd63fc307ddaULL,0xfd7cbfffd7b67ffaULL,0xfab67ff7dfbe77e3ULL,0xfbf6fefff7bc7ffbULL,0xefdbffceff9c6de9ULL,0xb6bdccfbff146ea2ULL,0xf7befefef68c7dc1ULL,0xdb767c7bfdbc7fc2ULL,0xf974ef36df3c7ee9ULL,0xefd7bedbffbc7de9ULL,0xbebd4fe8d4bc7dd0ULL,0xd5ff7fefffbc6ffaULL,0xb3bd6de0fd844deaULL,0xfbf7ff3ff79e7ffaULL,0xebdfff7af6bc7de8ULL,0x6f7ecfcef7be7ffaULL,0xf4f6ffebdfbe7fe1ULL,0xd9eff76bde1c7da3ULL,0xfffb7f7ed79c7febULL,0xffdfffbef7bc7ff8ULL,0xefffffdfffbe7ffbULL,0xfefb5ff3f6b475e2ULL,0xd9af6f7add9e3c6aULL,0xcf9a7d79fdbc6de8ULL,0xcbcbd3bbd79c75f8ULL,0xaffedc4bd73845e0ULL,0x969e6f73d5346fc9ULL,0xd1bc97dbd0bc7c70ULL,0xff962f31ff344420ULL,0xe9ffe73a75b43cf3ULL,0xfddfde5ff79e36eaULL,0xddbecffbfdb47debULL,0xd5fbf27fd73e3ef0ULL,0xdbfe7f7af4be6fe8ULL,0xf9c6fb6ad7b47de8ULL,0xdfc7fef7dfbc6dfaULL,0x2fbccffcdebe7c40ULL,0x35e7f66fdf9e6ce1ULL,0xbb4a6d62d79c2c40ULL,0xd8b6b36ddcbc6dc0ULL,0x41569a0452902522ULL,0xa9251cca788c2cf2ULL,0x80f0d4fa80bc6da8ULL,0xf7eb8f7af6bc7ec8ULL,0xcbb43f7cf5bc7ee0ULL,0x68cfefbcff946ce1ULL,0x9fedff5afeb46fe1ULL,0x8b725fe3d6bc6db8ULL,0xf16dff22d5382d70ULL,0xfe9a2f36de343fe8ULL,0xedd4773bf68e6ce1ULL,0x61240041c01401c8ULL,0x90bacfe5fe9c3cc1ULL,0x90511f2fe5b01548ULL,0xc2d62d7652bc3600ULL,0xf890f812d43e5cc8ULL,0x43c95857ddb43600ULL,0x1d8e4987d7b42ccaULL,0xa5fffa5fffb677c2ULL,0xcfaa3d11d6ac6ee2ULL,0xf212614fd51c14f0ULL,0x4e7137fef6be6deaULL,0xa2bedff4febc7ce8ULL,0x30336ae0d13c2c62ULL,0xdfeefff0ffbc3d6aULL,0xdfa0434357b22440ULL,0xddfffef7df3e3fe1ULL,0xefffdfdff7be7f62ULL,0x34367eeadfbc7da2ULL,0xffee7f79f7be7debULL,0xfab6bf7ed7bc7fe2ULL,0x8d6e77bef5b46ef8ULL,0xb5e8259d49c35d0ULL,0x3cffffe3fbbc3f69ULL,0xf490806054003640ULL,0x46843d20821c61c8ULL,0x8a82d3a24e8024e9ULL,0x39240a44c00e0500ULL,0xbfbc0d6ad53e7de0ULL,0xda50120856842240ULL,0x718e1458560424e1ULL,0x180c02352000442ULL,0x800000100200400ULL,0xd2260ce379982641ULL,0x9400366a54a22d50ULL,0xc882a540d4940520ULL,0x1392a332551c2b01ULL,0x80e00010c0802102ULL,0x82ec58fb09c6080ULL,0x1001000080000100ULL,0x20060000143800ULL,0x84805150830c0420ULL,0x8120c40190300a48ULL,0x10100000002001ULL,0xc41010a010242ca0ULL,0x4108200000200460ULL,0xbb2667774ae13e2ULL,0x50000040000080ULL,0x2400000044000000ULL,0x8420202014000240ULL,0x0,0x2240110080420ULL,0x100811000080420ULL,0x208804042022001ULL,0x34bf4df3d7946de0ULL,0x120040044002800ULL,0x0,0xa164e00672222d40ULL,0xe9c61a0b77902e40ULL,0x7214450801a414c0ULL,0x9138ee69749e3c60ULL,0x419c8000c2c01ULL,0xf2f46302f79c4d40ULL,0xedd81151c01c7d80ULL,0xadbe6d22f7902f20ULL,0xb0dc1fe1dcbe67a0ULL,0x13aead70c6b03d00ULL,0x280000140001400ULL,0x802300014100200ULL,0x8f5effd3f7ae2de0ULL,0x400a4e4438c3420ULL,0x8022422440c0080ULL,0xdfb5efffff9c7fe3ULL,0x88d210bc3a082420ULL,0xaddedfd7f5bc2fe3ULL,0xbcbf57ebd6b4377aULL,0xd7682720f8867440ULL,0xffba6d56d59c7fa0ULL,0xe306c72297100760ULL,0xed53f58d59c60e0ULL,0x6982fdfd4a86670ULL,0x41ef77ead6bc2fcaULL,0xdeb6bf7cf7be67f3ULL,0xfbf7ffbfff9e7f79ULL,0xdecf5f5ff7967df8ULL,0x61d4a63889c5460ULL,0xb4ea9e20d4144380ULL,0xc30a3d02b28c3c49ULL,0xfba29b1ed7946dc2ULL,0xeb43b76ffd3c7c70ULL,0xafbf4fd3db3e7d40ULL,0x90f51fead69e2d80ULL,0x9b38ff71d6be3fe0ULL,0xe80844e010043d88ULL,0x802328002ac2400ULL,0xafdcdf5ae4946cf8ULL,0x408c549100180428ULL,0x5320002045100400ULL,0x90810c2250101450ULL,0x840000040000080ULL,0x280400c041900400ULL,0xc705112d6143800ULL,0x8151a109d0046a20ULL,0x8000030410000420ULL,0x8000820400000468ULL,0x0,0x2010420100400ULL,0x2880000100ULL,0x1020042091000400ULL,0x8000000100800040ULL,0x4000001000080400ULL,0x5285440bc300460ULL,0x2040320116140400ULL,0x0,0x8108c002c5040400ULL,0xcfdfbf3df6bc6df0ULL,0x7bbecf52dfbc65f0ULL,0x801042e0009c6840ULL,0x609aadf144b47ce2ULL,0xd924f37c753e0668ULL,0xcfff3e37d33e6de8ULL,0xbdc0d48c6166400ULL,0xbd365efcf69e0ec0ULL,0xebeeff77ffbc7da8ULL,0xffff7fffffbc7ffaULL,0xebfdef3fffbe7ff1ULL,0x0,0x0,0x0,0x0,0x0,0x802100081000000ULL,0x0,0x0,0x0 },
	{ 0x0,0x880118083022002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd63269440bb8270eULL,0xfb87ef73c3b627bfULL,0xafde9e9bbbd6261fULL,0x9fbdd799daa2271dULL,0xd42ff67c8b9a2f7fULL,0xfb9e6762bb9a273dULL,0xcb74fb2eabd0264fULL,0xeee3b6128b9e23beULL,0x7fbadf5b8beb2fefULL,0xf45ffafbabe4231eULL,0xb7beff73cbf2233fULL,0xdfb67ffedbfb267fULL,0x8fdef61b8b9f2bcfULL,0x834fd6428bc32132ULL,0x9cde7ee7dbbc2f7eULL,0xffefef7a8bf927ffULL,0xff9cff7fbbbe277fULL,0xcfffffbf9b9e23ffULL,0x2ffefddeebbc233fULL,0xd7fd7f63dbcb277fULL,0xb3fcf7f7cbb22bbfULL,0xceaf2f778bc02f3fULL,0x6ce237b6cbc22f3eULL,0x4b7f5cdfa3162307ULL,0x97f37ef09bce2fdfULL,0xf5ffb7fb8baf273fULL,0xdbf6ff7fdbbe2b7fULL,0xebffff7febff2fffULL,0xedff7fd7dbf92777ULL,0xbfffdfdb8bf6277fULL,0xf777f6fb8bfe227fULL,0xffbe7ff1dbfd27afULL,0xcbb4f73fdbb223b7ULL,0xeedebb979bf42b7fULL,0xff3fdfdb9ba627bfULL,0x75fff7f9fbbe233fULL,0xdbfb2d63dbf22f7fULL,0xfbbdff7fab9b277fULL,0xdfdeff1febf627feULL,0xaffedfdb9bbe2f7fULL,0xb7b7ffeffbfd2f3fULL,0xf5eeff6b9bba227fULL,0xfbbe7f7edbfc2f3fULL,0xfefffffefbff2f3fULL,0xefffffdfdbfe2fffULL,0xbff69fe79bff2ff7ULL,0xf0fdbfe4cbcd2b7aULL,0xfebf2feb8bdb27f2ULL,0xcbdaff7efb902f9fULL,0xafd63c928392266fULL,0xbfbb5ff99bfa277eULL,0xf7fc17548bba273fULL,0xbabe7d6d8b9e2756ULL,0xc9b1ffea9bcd2f3fULL,0xeddf5e599aec2b15ULL,0x9bbae7bb88dd2f1eULL,0xf4f6fad0ebd82766ULL,0x5aaeff728bd4237eULL,0xdbf6ef77ebff2fb7ULL,0xbfdfba7bdbb22f6eULL,0x6bbdcfff9bfe2b5fULL,0x56ffb6fe8bbb2b37ULL,0x9beef772cb1f260fULL,0xbb906f759bde235fULL,0xeda3a1beb88211fULL,0x8ada19c28286238dULL,0x54957ee6cbf8278dULL,0xad6e77728bf02a1eULL,0xffbeeffeabfe27ffULL,0xcd63ef1b0bbe2369ULL,0xffffffdffbba273fULL,0x96933e7bdb93237fULL,0xf17cafb88bca2676ULL,0xfffeaf7a9bb52fffULL,0xefdff7f78bbd2f1fULL,0xac31e4e0b94022cULL,0x1e047bff0b90232bULL,0xdcfc2f608bed23ffULL,0x7bbe7ff4dbae239dULL,0x83f1e136d3b2279fULL,0x4bf75cd50bba2fafULL,0x9c7fddf3db9b2befULL,0xf487dec9cbc0270fULL,0x420c45628a282204ULL,0xa98dcb4c8bae020eULL,0xe43b600f3962306ULL,0x4194a0041862287ULL,0xf5fcfcee8ba4275fULL,0xe0b9ec619ad2230eULL,0xb3a4d1fb8bd8232fULL,0x58c7ee5b8bfc2bbfULL,0x2f7e4f4bdbf823aeULL,0xb63bffc08be0217fULL,0x59aeb7d783b42effULL,0xffbeaf769bff2fffULL,0xc886fea8a396262aULL,0x66fe5fcd8bfe2f7fULL,0xf5ef5feebb75268eULL,0xd9eca762819c2fa6ULL,0xfabe3e7fdbe62e37ULL,0xd9f5f33e8be62f9aULL,0x4a0e120101b02217ULL,0x36027c801a412686ULL,0xd080300208882420ULL,0x5c302e4c0b2a2383ULL,0x80640260b2000b04ULL,0x122080000302ULL,0x12944c8403860320ULL,0x0,0x40060d208b060302ULL,0x820010000004ULL,0x8421c108ac02101ULL,0x659819128bdc0392ULL,0x43642e21b58022cULL,0x45200840a1142200ULL,0x6000202088082000ULL,0x40d4000410802414ULL,0x2010000802000ULL,0x4806000802402ULL,0x40061040102200ULL,0x4000a4c01802018ULL,0x80800002000004ULL,0x0,0xc00002000106ULL,0x100ed0209480268ULL,0xc88605e20bc30322ULL,0x10841220a480000ULL,0x90c2000b0800006ULL,0x848905a013022a25ULL,0x1200100802802000ULL,0x480204001800012ULL,0x89d4d4869bd0233fULL,0xc95806339a702705ULL,0x68a82d920aa0221dULL,0xf560d2709b902386ULL,0x1c14047003940214ULL,0xc980e33d8bca2abcULL,0x89539b918b90273eULL,0x771bcf1393902766ULL,0x4490f6ebba902327ULL,0xdd2e2d708ba0223eULL,0x100400000002400ULL,0x164a021303822a34ULL,0x8dfedf5c8bb12b96ULL,0x652509a0f840260eULL,0x1002140001302000ULL,0xffb637758bc823ffULL,0x2f5deb06997522dfULL,0x6f7cffdb8b94225fULL,0x9fbf7f63d3ae27afULL,0x996444685bd02027ULL,0x5ebd37fecbc02b6fULL,0xcfebfb16cbf22b27ULL,0x32d49f6a1b8a276fULL,0xef3b4ffabbd823ffULL,0xddfdf67f9bdc2f7fULL,0xffbebff1abbf27ffULL,0xebf6feeffbbf2f7fULL,0x7edfb7d5dbb2270fULL,0xcb0e4f128a902624ULL,0x9175fc4b9bb02305ULL,0x498a3543eb38225bULL,0x62c442069b00272cULL,0x894e901703962b0dULL,0xf28ecdb19b31274eULL,0xfdf5ffeafbf7230fULL,0xdebfbf728b9c237fULL,0xd82e058289c8026aULL,0x9d3920282882607ULL,0xaffecf5b9b9e27ffULL,0xfcff57f9fbfb237fULL,0x8000258209880250ULL,0x92b62e7ccbc023ecULL,0x0,0x808881702202304ULL,0x9492026881802033ULL,0x81382a608280230fULL,0x4084288088802209ULL,0x49d6f0208ba9275aULL,0x0,0x401001000040ULL,0x954105688bc00032ULL,0x821c410081900050ULL,0x8000080000802000ULL,0x284502900300032eULL,0x89a0ed50cb84222bULL,0x100102000004ULL,0x10010000100200ULL,0xd2a0004289882213ULL,0xeddabe4bcbf82f3fULL,0x4014080000002604ULL,0x602c82420b900a0cULL,0x130234600103262aULL,0xdd8027758bc0279cULL,0xcdcafb8ebbdd2717ULL,0xabd8ddd39bba271dULL,0xb6fe57fdbba02a7fULL,0xdafe2f6aeb9c233fULL,0xffb7777f9bf92f3fULL,0xe9ffff7efbba278fULL,0x0,0x20004000800002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf62e3bffeb9cf7feULL,0xabdefb3fdba5dfa2ULL,0xe8fe7f796394d3eaULL,0xb6bdcff2fb34f36eULL,0xfdfff678eba5dbbeULL,0x5f3e6dfbfbb5cbfaULL,0xcbc4ef7ecb95d3aeULL,0xcfefbfcafb8cd1beULL,0xa5af5d52eb04d13eULL,0xddf37de97b05f5aaULL,0xcfbefd72fbbddbeeULL,0xdbffef7f8bb5dfbeULL,0xcd7ffe1ff1b5ff2eULL,0xfa56d9cbd915e76eULL,0xfef4bfeffb3dd77aULL,0xddffefe9f314ff3eULL,0xfbfdefbeebbdf3eeULL,0xcfcfff9edbacd3beULL,0xef5fdfdbeb3cdf6eULL,0xdef77fe3e394dfeaULL,0xd1fff57a8bbddfbaULL,0xd6b23ffefa11d72aULL,0xeb77dfbff3a5df6aULL,0x4ffeedcbf915d72eULL,0xb7bf7f7ffb3dfbbeULL,0xfcfdbff0cbbcf3feULL,0xfeb6fff7ebb5e7beULL,0xcbd7fefefb3cdbeeULL,0xfffb7adffbb5dfeeULL,0xffffeffbebb4dbbeULL,0xded7f6fffb95f7aeULL,0x7bfe7ff0db34d76eULL,0xbb24677ed114573eULL,0xfacfbffbfb34dbaeULL,0xafff4f7bfb25db3eULL,0xbcfdeeeddb30d33aULL,0x9baeff71db95d33eULL,0xd9be6f7ffbb5fb2eULL,0xcfd7ffbfdbb5f7aeULL,0xffffcfdefba5f3beULL,0xfefffeecfbb5f7feULL,0xd7feff7acb95d7aeULL,0xdbb27f76eb35f3eeULL,0x8adfff3feb305baaULL,0xaf7fffdfd9bcf7beULL,0xbcdb7ee3f39ceffeULL,0xddcdfffff3bdd3feULL,0xfe9627104310573aULL,0xd5cfe37bf3b5d7aeULL,0xebfefaeddb0c537eULL,0x95eb6ff3f335d7beULL,0xd0b7bffbfb9cd1beULL,0xebfefd73f9a4ff3eULL,0xcadfffafeb35f3beULL,0xef461e7adb34132eULL,0xaf2d7fb98889f7eeULL,0xed67f66ef334d32cULL,0x5f3eed71ebb4dbbaULL,0xcb95e37fd925f90eULL,0xfefebeffeb3cdb3eULL,0xbfbddff3e325d7feULL,0x5db446fad305c5eeULL,0xdfbffff1dba5d3feULL,0xdbb67f7ffb2cf32eULL,0xab4616338934d22cULL,0xa3ce5f52e32dd1aaULL,0x3491d4eb5b0dc080ULL,0xcd20ff278384d1f2ULL,0xcb976d97eb35eb2eULL,0x5fcfff3eea34d3feULL,0xeffeffdbebbdd3beULL,0x94bb57e9d1adf9baULL,0xf16ceff2fb15dfeaULL,0xfed6bfd7ebbddfb6ULL,0xc310d5704b25c3aaULL,0x98775fdbeb04f19aULL,0xdfbb4f5adbbdd3beULL,0x97bb9f78f234d33aULL,0xb78c7f71cb91d77eULL,0x8b86ff1f7b10d7feULL,0x7cd63f587115c32eULL,0x77bf4ff3fb34fb3eULL,0x857b76e0ea84d1baULL,0x6338bd72ebb4f386ULL,0xe804b2614920c30eULL,0x28529c59430445a2ULL,0xeadedfbeb85d17eULL,0xd0114876a194d128ULL,0x70be7d61cb345302ULL,0xeba1f73beb15d3f6ULL,0xd946fab5db38d3beULL,0xebdfdfdfcb34dbbeULL,0xbc7957e4f230db2eULL,0xdab6a731c39cd5aaULL,0xc7b4ff268ba4d12aULL,0xdcdefeaae110dfa6ULL,0x735f5f596a1dd6a2ULL,0xa39b17e3fb20d39eULL,0x82482708cb8cd1e2ULL,0x5ebeaf6eeb104be2ULL,0x8381e13d73104726ULL,0xa4000140004420ULL,0x410462ac024c1a2ULL,0x60260a8801d900ULL,0x4814201189205080ULL,0x8120882801044182ULL,0xc2ca90110984c13aULL,0x9aaa4799dab0c122ULL,0x800550a0c0208128ULL,0x834f000d1049028ULL,0xc098d110c230510aULL,0xc763b2cb892473aeULL,0x24c000928020c108ULL,0x50404e003006182ULL,0x422204004000912aULL,0x1004018050001000ULL,0x3820040004102ULL,0x28281424104d008ULL,0x402c006088001122ULL,0x120082060004002ULL,0x210000241008080ULL,0x4000800060000000ULL,0x0,0x0,0x100580080200204ULL,0xfabe2ffefb3dffeeULL,0x83f3e379eb9cdbaaULL,0xaffffffffb9df3feULL,0xbfab7f33fb8de9beULL,0x0,0x128008008020d000ULL,0xcbf5f77fdbb5f7beULL,0x66d25e14d105c1a6ULL,0xb121d5a5200932aULL,0xd554ee61d134533aULL,0x421434110184c11aULL,0x810034204904c080ULL,0xe43be7edb145b2aULL,0xffbfddfffbb5d73aULL,0x71b4f6fee314e36eULL,0xd36c7c215b01916aULL,0x8030120080201204ULL,0xdc2ba95e321d328ULL,0xa14f58d0f925f2a2ULL,0x8400100200ULL,0xc044040005080ULL,0xf6866fedf305c1b6ULL,0x684602687100c020ULL,0xae76dfd6eba5d3aaULL,0xcd7b92fe91ddf7eULL,0xb966da608b28616aULL,0xbaa63d20d221d7b2ULL,0xcfdafb1ef994532eULL,0x619ffac56995d32cULL,0xb6bafff2cb3ddbaeULL,0x99fefdf7eb34f33eULL,0xfefe7d77fb15dbbeULL,0x89f3f3ffdb34536eULL,0x38511e94c114f30aULL,0xc08085a08b041062ULL,0xd510d2e06300c182ULL,0x3b304f604b044766ULL,0xe32860268b01c136ULL,0xb7db92a7f3a4d1aaULL,0x9b0cd732eb04e5acULL,0xd5f5f6efebb9d3baULL,0x8ffffff3fbb8d7beULL,0xf7b2af7ffba4d7eeULL,0xfc0b60a6905c082ULL,0x539e5f1beb3453feULL,0xf6fffffffbbdd7feULL,0xfdfffff7f38df9beULL,0x12a404e06a044122ULL,0x604a7e344105c10aULL,0x8435e5250848316ULL,0xb9b25c43f304d162ULL,0x6124b0625301d0e2ULL,0x308001000000c008ULL,0x1270228020c10aULL,0x188080000018204ULL,0x45148008020410aULL,0x914d2e70c214d3aeULL,0x382455e12b084002ULL,0x8020001080001420ULL,0xa000000080000004ULL,0x2f024dcb8300cf2eULL,0x400200000000102ULL,0x800140c0009000ULL,0x100802080005008ULL,0x295b12c00105d32aULL,0x2501424a8020572aULL,0x49612672da905728ULL,0xb01b247088246206ULL,0xc9a5f572f9a5c182ULL,0x8486ba28e9005086ULL,0x3afedf9acbbdeb1eULL,0xfefffefbfa9df1aaULL,0xdffea771e93cd73aULL,0xfa9e3fc74b94d7eaULL,0x57c3ce18e914f33eULL,0x40001002000002ULL,0x0,0x0,0x0,0xe9ddd33af985d1eaULL,0x10e000100c008ULL,0x0,0x0,0x0 },
	{ 0x0,0x0,0x8000100500000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xff142f9409869322ULL,0xcbbdf7ff8bd7f37aULL,0x77c7be79c1d7aa3eULL,0x1a9ceff98ad3b3eaULL,0xff7fba738ad7e3e8ULL,0xc7be7df28ad6f762ULL,0xfaa8ff6f8ad7ffeeULL,0xaddfb7cfcbdff7e6ULL,0x7e3f4b75ca46f3eeULL,0x95fdfedcca5eb7e6ULL,0xffffffe58bdff67eULL,0xefffef7f8bd7f7eaULL,0xcf5ffebb8bdfffeaULL,0x6b7cdf5a88cff3a2ULL,0xd4ff7ffb8bdef7faULL,0xdffeffff8bd7fffaULL,0xffea7fff8bcffffaULL,0xedefffeecbdffaeeULL,0xefdfff5fca5ef3eeULL,0xfdbfffe38adff36aULL,0xc99ff7608ad6b2e2ULL,0x5ab62b7e8ad6f3a6ULL,0xffd6f7bf8ad6f322ULL,0x4a4ffe5588d6f36aULL,0x95bb77f6c25ff3e6ULL,0xfdfff7ffcbdff2e6ULL,0xffb67f7fcbdff7eeULL,0xf9f7ffefcbdffff4ULL,0xffffffffcadfffeeULL,0xbfbdfffb8adff7feULL,0xfe77fefdcbd7fbeeULL,0xedae2fe3cbdef3e6ULL,0xf2e4b22e8893b760ULL,0xfbd7bffccbdff3f6ULL,0x7fbedfe7cbd6f3e6ULL,0xd4f7f6ebcbdfb3aeULL,0xbfbefff3cadeffeeULL,0xfdb7ffe98adfffa8ULL,0xfbd7ffbf88d7fff2ULL,0xafbf5fd3cadef7f6ULL,0xffffffef8bdff7feULL,0xffffff7fcbdffffeULL,0xcea5bf67cadff26eULL,0xebdeff2c89dfbacaULL,0xafffdf1bcadbf7e6ULL,0x35bf7ff2cbdff6e6ULL,0xdfedffe689d6bfe2ULL,0x9abe2f758ad7f7eaULL,0xcdc78f338ac2f366ULL,0xc3dfd7d38bcef3baULL,0xbcfbeff7cad6f3f6ULL,0xd0ccbff9ca86e366ULL,0xf7f67d7c8bd6b66aULL,0xe9dfff3b8bcef3feULL,0xefff5fdfcb5ff6e4ULL,0x81bacc83cad662ecULL,0xdf36befa8bdff352ULL,0x5ffe77e38ad7fa6aULL,0xbbf5e7668bc7fb72ULL,0xeeca939f8ad7f3faULL,0xb3becf5b8bdafaecULL,0xfdfeffff8ad7f7eaULL,0xdffefff3cad7f3feULL,0xfea42f6fcad7f36cULL,0xcf8afa1b8bd7fa4aULL,0x628fcf428a563ab2ULL,0x38b552618adab2e0ULL,0x8b729f628a46b760ULL,0xdf1a4d6289d69324ULL,0xeed3df0f8b8ef662ULL,0xaf7edf5bcadffe6cULL,0xbeb75dda8bd6bacaULL,0xdaf4ff7a88c6a372ULL,0xde3e3f710a56f3eaULL,0x69d0eaf589ceb222ULL,0x3dfebbdf89d7f3aaULL,0x17224ab70acfb332ULL,0xffff7f6fcbdff7f6ULL,0xefdeffffcbdfffeeULL,0xebdedfea82dfb7f0ULL,0x6ca29eafcb5e32a6ULL,0x92bb050b0ad5d208ULL,0xc43ffa6f82c7f326ULL,0xcffe6d718ac7f372ULL,0x8914b72f885fb3e2ULL,0xced630138355ab18ULL,0x3f8c6f5188d4ea22ULL,0x1468124208c2d260ULL,0x5ac4e76289c5a366ULL,0xc6bb67cfca5ebf2eULL,0xdfeebe3b82d6bb3aULL,0xbbbaef5b8a55f6f8ULL,0xa4f357ec8bc4f2c0ULL,0xe9ffff32cadff3a6ULL,0xcf97f7f98954d36eULL,0xe181cf198ac3fae0ULL,0xebffffcbcac7f3e6ULL,0x8c1a21724b56226eULL,0x5dc02e4208560622ULL,0xde2e3f6382def750ULL,0xc4cbc53400d6b3baULL,0x4a0a200085a080ULL,0x25f05f788955f368ULL,0x8400000008040000ULL,0xa50434608208b740ULL,0x80558b000a0a9242ULL,0x40821e1008563022ULL,0x820205000084d002ULL,0xe35640081cb222ULL,0x2b080c208280b880ULL,0x2384ca0b01d4f222ULL,0xcefe3e078883f382ULL,0x3a8e43088256a3a4ULL,0x0,0x8000004000020200ULL,0xc081205109440242ULL,0x4781be81cac9f204ULL,0x1a0c8c4448020604ULL,0x2004000000c00000ULL,0x240268000221a0ULL,0x4340002002020ULL,0x0,0x208000000000600ULL,0x0,0x140724040027006ULL,0x4effa7fe03cff262ULL,0xed92ff3c02def362ULL,0x7f7f3fdd8bcfb7a2ULL,0xdb90dffb89def332ULL,0x2009208200a000ULL,0x9f0606608a07b3a0ULL,0x897efbfa8bd7f7f6ULL,0xffffffbfcbdfffeeULL,0x9ecbcfd988cefe22ULL,0x508580700851d300ULL,0x140000000600ULL,0xfa94ff4f8ad4f2e8ULL,0xeecf3f5b8b5ff3f0ULL,0xbfbfeff38adffba0ULL,0xf1f3f7ee8adff766ULL,0x9591a7518bd5b700ULL,0x4000520408028300ULL,0x98682018846a220ULL,0x8608858005492aaULL,0x10a010480000a260ULL,0x5102a560c050a104ULL,0xff96296c81d3b2a2ULL,0xcdefff388bd7fbbaULL,0xea7e7d5b8ad7aea2ULL,0xf5bb55a8c8ceb36eULL,0xd064764a8145b2e2ULL,0xce9a252080569828ULL,0xc0466ef488c4b796ULL,0xbdde9ffa8b4ef262ULL,0xf7fb6ffe8bdff7eaULL,0xd5ffff6d8bd7ff6aULL,0xfefe7dd68adff7e6ULL,0xc942ef3eca1e7a06ULL,0x468202108194b300ULL,0xc4824ff589d5a202ULL,0xd56c82b88857f262ULL,0x210000000a0a0ULL,0x4302c10800868222ULL,0xa573bb78cacef776ULL,0x6dbdc63c2ddd30cULL,0xfdffffffcbdffbfeULL,0xdfbfff718ad3ffe8ULL,0xfbbeef7e8adffbe2ULL,0xdfd2fe93cadff3ecULL,0xffffffdfcbdff7eeULL,0xfeffffffcbdff7feULL,0xffffffffcbdfffeeULL,0xfffe7f77c2dff7eeULL,0xe5e6e302cace933cULL,0xed5d5fdacac47e9cULL,0x43250a002407040ULL,0x71c1020080587002ULL,0x4002304001848260ULL,0x800853398042e062ULL,0xc00000018000ULL,0x200461200448202ULL,0x76fcf6100a46a62aULL,0x8f6a77e38bdef2eaULL,0x44000000005000ULL,0x0,0x93aa8de28a8cb206ULL,0x9000164082d22020ULL,0x6c0000720a08d000ULL,0x0,0x0,0xb084f5b82ccb22aULL,0x422200000121200ULL,0xd9186d40885e32c2ULL,0x20660228082d200ULL,0xcf53be3d8adff31aULL,0x9d0ccd5b885db7c2ULL,0xf4f6ffea8bd7f6aaULL,0xba4ee77a8bd6f228ULL,0xfef22f6e8bc4fb66ULL,0x580002058142c222ULL,0x2fddde538a4efa72ULL,0x0,0x0,0x0,0xebc7bebb8bd6f366ULL,0x43573e21cade7a16ULL,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfeb63d73eddffeb6ULL,0xe8f5fa6f8dff5b90ULL,0x7b77dffd8ff3be92ULL,0x8f9ecdfecdf3fabfULL,0x9cb2ef7eecd33b33ULL,0x4f7eff63ece733b2ULL,0xa3c4ff25ecd632a2ULL,0x7fdfbb3bcfcfaeb3ULL,0xafbecb7bedcb7badULL,0xf5fdffffecf77fbbULL,0xff9eff72ede23fb3ULL,0xdf927f79cfe73faaULL,0xcffbdebfcdd7fa8aULL,0xfb9aaddc4dd1ffb1ULL,0xf6ffffff4df3ffb1ULL,0xffffff77efffffbfULL,0xfbfebff6cfdbff92ULL,0xcde7ffb7cceb6fb8ULL,0x6edfff5beffbffbbULL,0xbdff5fffcdf37ebaULL,0x456c777a69d01ea0ULL,0x7abebffb4dd7bfb2ULL,0xc992dfb34da75f9aULL,0x4bdcdb4bedc13bb8ULL,0x97bbcffbcddb3fa9ULL,0xf7eff77fcfd33fbeULL,0xfffefdffedffffbbULL,0xebfffffbefffffbfULL,0xffffffdfefffffbbULL,0xffbdcffecdf3ffb8ULL,0xf7f7feffcfd3ffbeULL,0xfffe3f78cd5b3eafULL,0xf9fdebffcffb7fb2ULL,0xe972febf4dc7ae97ULL,0x9fccdff9cfef2fbfULL,0xf4fff6eecdc3fe99ULL,0xdfbf7fe1ecf3bfbfULL,0xffaff7feefeabfbbULL,0xcffffe3fedcbffbfULL,0xbbffdfdfcdebffb9ULL,0xfeffffffefffffbfULL,0xffffff7aeecbffbdULL,0xfbbefffeedfbbbb2ULL,0xfdd4dfbeccdbffb3ULL,0xaffcffd84dd33bb9ULL,0xb7ffb7f3ece38fbdULL,0xfcf0fffbe5c33ea0ULL,0xcc6e966245d3fa32ULL,0xdbfffefeedf77fbfULL,0xaffff75bcdc37ba2ULL,0xb7bb7feacfd7ffb8ULL,0xd3ddbfe8ccdb7bb8ULL,0xfbba7f75cddb7f38ULL,0xebf5f3ff4cc73f93ULL,0xffcbdfd36dd75f9aULL,0xf6bedffbedf3fe3fULL,0xf5ff767fced3bfbcULL,0x7dbe7d72edf37fb7ULL,0xfbd4ef3f6dc6ffb2ULL,0x4bfcffdf4dcbbbbeULL,0x4ebfffffefdfbe94ULL,0x54f572eb6cd33fbbULL,0x9bbfdf72cfd75bb8ULL,0xd99dfffde8d7bebaULL,0x8f69fa08ccd73e90ULL,0xbf6567c0c9533fa0ULL,0x2c1916e90487aa80ULL,0x982aa7620cd37b12ULL,0xf7bf6fd5cdd29a37ULL,0xedcdfc0e8cc6fa98ULL,0x2859dad045ca4a31ULL,0xce35de54d422a92ULL,0xc52caaf2eec22a18ULL,0xdaa66f7fccc2be31ULL,0xc9d7f73aedd33f20ULL,0x4e5afaf9cde33a31ULL,0x7ea2dd100c622b90ULL,0xb1fe277aefd2ffbdULL,0xebeeffffcddbffbcULL,0x8b7ffbfacd86fab8ULL,0x6f5f2e0ce7dbbe38ULL,0x3ed06942aec32ea8ULL,0x1e5b3eeecccbef3bULL,0x6a5c257948033eb6ULL,0x38e900628cc35810ULL,0x2a469feb6dd17e02ULL,0x37b9df6c6d933a8dULL,0xf5b47eeeadc22e99ULL,0x933aee71acc35a91ULL,0xd3b4f7dfcde72f99ULL,0x4febbebaccc3bf96ULL,0xba6d9d5dccc37b9dULL,0xe2315e63c4d3fa24ULL,0xdceeefe0edc33e90ULL,0x5b94678748e35e90ULL,0xcff7ff6f4dcbbea2ULL,0x6ffcdf584cc32f38ULL,0xf7fafff3cec32eb6ULL,0x1334af48ccc222b4ULL,0x58900a4a4cc25a28ULL,0x900f202c4c20204ULL,0x25ca5fd88cc11ab0ULL,0x13204dc3ccc36e14ULL,0x2121066888c21010ULL,0x508605c4858a0808ULL,0x304a72b8580b800ULL,0x340074101030a0aULL,0x82180b428e800302ULL,0xc56b28048030688ULL,0x52221010a00ULL,0x314441004cc20018ULL,0xa513001cd921a80ULL,0x1174a4045800e20ULL,0x143156efccc223b8ULL,0x63055c14c842b00ULL,0x0,0x302a60244c11630ULL,0x1be8ec5b4dc23e34ULL,0x1606248008031a12ULL,0x2e05400c022a81ULL,0x4000040004002000ULL,0x60080800000ULL,0x0,0x0,0x504522045c30830ULL,0xffb6bf73edd27ab4ULL,0x81d7f31eecd3fe00ULL,0x67dfde464dd29a9dULL,0x1ba007a0cc869310ULL,0x288040004022008ULL,0x120620600c0c0688ULL,0x8093f9e36fcb1a2cULL,0x3cdf7f3ecfe3baa5ULL,0x20424041000804ULL,0xf5c416634de2bba1ULL,0x72362d60c8531a94ULL,0xc972e72e2c873b93ULL,0x80e79e114c92ba97ULL,0x9bbfcf7bcff2bd3cULL,0xfcffffedcddfffbfULL,0x8a3c3d632ca3ff12ULL,0x4004028208010281ULL,0x49430e10c8832b20ULL,0x818a0a1200850e00ULL,0x2a068408430288ULL,0x100118000004a01ULL,0x60902f6101c75a10ULL,0xf3eff732efdbfebbULL,0xebfbbbdb4fdbfe31ULL,0x40960fe345c71a0aULL,0x582257c2cc826a22ULL,0xd2a827464dda3a9cULL,0xc0848c3549422e83ULL,0x3899dfc4c4d2fe18ULL,0x975d7f73c58bbe38ULL,0xd75476ddcd9abf01ULL,0xffbe7fffefdbbbb9ULL,0xfbd7fb3ecce71fbfULL,0x8c916e218cc2bb0aULL,0x120488120422200ULL,0x9022664c44c76e10ULL,0x51843dd0c0809606ULL,0x981eb2e64833a32ULL,0xc4a224a85422a00ULL,0x569b5ffbeec6feb9ULL,0xf5feffffefdbffbeULL,0xfffdff73cdebffbbULL,0xffbfffffcffbfebfULL,0xcbfffff9cdf7fe9aULL,0xbf1edfdbcfef7f3eULL,0xf4fffeffeffbffbfULL,0xffffffffefffffbfULL,0xffffffffefffffbbULL,0xfdf7ffefecc3ffbbULL,0x86013514c820281ULL,0x4dc502844003a23ULL,0x1820c242e0420a06ULL,0x2022c4340c21e0aULL,0xc970a7004ce32800ULL,0x204000000020008ULL,0xa2410840832202ULL,0x500096084cc28800ULL,0x14a332048222114ULL,0x108001000800ULL,0x2206000008021281ULL,0x5fbdcfbb6eeb6ba2ULL,0x1001042004800202ULL,0x8000200000800200ULL,0x202222400400a80ULL,0x10100006000000ULL,0x2028034241805800ULL,0x41482c400424a1aULL,0x4182060048a00600ULL,0xd0100e0068808304ULL,0x8162ba014ccae89cULL,0xa91a031bcd8a7f10ULL,0xd4f57ef8cceba2b1ULL,0xdcd32f71c9c33f93ULL,0x6cfe3f6449d71f91ULL,0xc9eddb9fced71fb3ULL,0xcfffdfdbcdfbffbfULL,0xa9aa8ffb8ddf3b9aULL,0x0,0x0,0x0,0xbf571bdac5c3de1cULL,0x0,0x108662081435a18ULL,0xc804200004022600ULL },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc1a8c6202378944eULL,0xfefe7ffffe7eb3ebULL,0xc9d7ff7fa07b93cbULL,0x7d7b9bfdfe5ab7c7ULL,0xbffccde8ef7a9cc7ULL,0xd716767f777f9bcfULL,0x7efe3d7b6f739bdbULL,0xb3f6ef7f7f7b17dfULL,0xea5abfbb7a7a93cbULL,0xfffecf7bea7f99efULL,0xf5fff7fb6f7b3fffULL,0xffdfeff27f7ebbdfULL,0xdbbee77eef7e95cfULL,0xceee7eaaaa5e96cfULL,0xfbfedfdfff7f97ffULL,0xfebffefaef7fbffbULL,0xffffff7ffe7ebfffULL,0xfbbe7f7feb7abfffULL,0x8defeb3bef4a1adbULL,0xea7ffed3fb5a9b4bULL,0xbdf397e7267f355bULL,0xfbfeffede25eb5cbULL,0xdebe3fff6f6fbddfULL,0xed4ff672f756918bULL,0xe3fffe6b7f72b2efULL,0x9fbe4fffee7237fbULL,0xf3ffffffbe7eb5fbULL,0xffbe3f7f6f7a9ecfULL,0xfbffffffff7ebfdfULL,0xefff3fdfef7ebfefULL,0xb79fdfffee7e16dfULL,0xfdbef67feb7e977fULL,0xebbefffbf77eb7ffULL,0xebfcff2eef7b97dfULL,0xcdeaff5beb7797cbULL,0xefbfefdfe67fbbffULL,0xfbd5deeff77595cbULL,0xdff6fd727f52b4cfULL,0xffbff77fff7b17cbULL,0x7fefff3bff7b33dfULL,0xfffeffdfff7ebbffULL,0xfeffffffff7fbfffULL,0x97febf7beb5e1fcfULL,0xff9e67ffea7e16dfULL,0xcd5febdff7721d4eULL,0xeb7fdf5fee7e9ecbULL,0xffff77eebb7a314fULL,0xfbefefda767713dfULL,0xd3be7ef3636e19dbULL,0xcdffffffff7e9eefULL,0x7bfe9fdd6f7299dbULL,0x743e77fbe263b0cfULL,0xd27fd760a6763c4fULL,0xff8efffdbb7f15cfULL,0xd9f7f77fe97a12efULL,0x2efebf5b2e7a9859ULL,0xaf4cef7a28563c5bULL,0xf7fdf7ffea72bedfULL,0xfffebffbea7bb7fbULL,0xf9fcef7efa7395cfULL,0xf9cfff792b7a93cfULL,0xfffccffefa7337dbULL,0x55c5d6eb2b629cdfULL,0x9aaeff70fb72974bULL,0xebd4a7b43f5a13cfULL,0xceebfe07e2773b3bULL,0x2b7fd4de307e10deULL,0x14b4fc4b2b721169ULL,0xbb8eff2bee4b11cbULL,0xbf9a2f726e5296dfULL,0xe55d6f1b227233ceULL,0xfa911b531264108fULL,0xb6ff766fe776b5dfULL,0x954b37d0f25e10cbULL,0xdfb12c7d2a6e9cebULL,0xadcef63ea77234cbULL,0xaddefade6e7a1cefULL,0x579b4fffa37a94cfULL,0xc935ae5263661509ULL,0xde8e7fd23a4394dfULL,0xa9f7f77faa7a10efULL,0xea569fd92e6791c9ULL,0xfd564ce76d661a19ULL,0x53fbfafe637a90cbULL,0xf328f4613777934fULL,0xa84ceb3a3b511602ULL,0xc813b297266793afULL,0xa7dedffbef6217cfULL,0xfcf732f96a7a18ebULL,0x6bfe6f61ad7ab4cbULL,0xff957fffe75714fbULL,0xefd7ffd36b7b1ccbULL,0xa25887d12a52910aULL,0x1496f6ed3e5c94cfULL,0xc16e7fa9785e92cfULL,0xf7b82a4561629880ULL,0x1582ea2b65563bc7ULL,0x67ffdfdb6777bedfULL,0xbe7bffefea7e19ffULL,0x896db7f066769049ULL,0xc6a0250060611049ULL,0xc8c391002340101eULL,0x1c0444220210caULL,0x8044002008001ULL,0x4404959021403380ULL,0x3000000000021000ULL,0x89f0bb7b217615caULL,0x40020000021000ULL,0x1126ea200621097ULL,0xd500280002421143ULL,0xc10024020a021007ULL,0x5200e402a3400023ULL,0x885396002e40160bULL,0xc4441284000a9ULL,0x9470365a6760105aULL,0x9a170d72af5e10d3ULL,0x4002000000400040ULL,0x0,0x888007480052100eULL,0xfcfffeffff7fb7ffULL,0xd96f277beb7a177fULL,0x0,0x2857c2026f4015c0ULL,0x0,0x800010000001800ULL,0x502020000461001ULL,0xdfb425442b5a91cfULL,0xa78c7ca577531b21ULL,0x29de7eae626a1b4fULL,0x13b06e19605a106bULL,0xc088250066020003ULL,0x10120c00080a1041ULL,0xa9c49022225214c8ULL,0x6821201a0701082ULL,0x330cc7102e57108fULL,0x954c8360e342140bULL,0x46ec204a205210cfULL,0x8174c7223264148fULL,0xce53bab0346693b9ULL,0x7bbccffb6e7a97ebULL,0x34beceecb27ab9ebULL,0xb014af71ea4291cbULL,0x80130c246001c2ULL,0x84c9388001731120ULL,0x8810102402001ULL,0x5cb4485000421087ULL,0xd22975023643141ULL,0xa8c32602c421449ULL,0xdec3ffbcac6535cfULL,0xaf56cfd9257a15deULL,0x70ce561a06213c0ULL,0xf19707182c709a81ULL,0xc2b41382aa441243ULL,0x8000f10000401091ULL,0x69ffd9ce777e12cbULL,0x15fecff32a763cefULL,0x9f7c1d0aee7638cbULL,0xfffe7dfeff721debULL,0x49eaff6baa7634cfULL,0x49c35a19e87600cbULL,0x5714d0261029001ULL,0xb2bffe792647158fULL,0x418c3b2020461458ULL,0x8d000220a000049ULL,0xdd435e1ef8721183ULL,0xdffecff3fe7717cfULL,0xf5fffeecbe7abfcfULL,0xfbdeff73ee77b3efULL,0xfbf6776f6a5e15cbULL,0xcfeffbbfbf7b38fbULL,0xaa861fdaeb5e1ccbULL,0xfcb6fefee77a97dfULL,0xffffffffff7fbfffULL,0xffbfffffff7f9fffULL,0x4fdaefaa6e523acfULL,0x294012810a021111ULL,0x24215c612142350aULL,0x4041865861021081ULL,0x14003020029000ULL,0x42000400001ULL,0x10124863000001ULL,0x82010220400081ULL,0xc080501032001001ULL,0x5082044010603210ULL,0x300a00202001540ULL,0x2b4a594b226210cbULL,0x2b2ac91823721401ULL,0x420002822001001ULL,0x0,0x2100c0000040008aULL,0x0,0x8c0c5418a0420288ULL,0x7484960822429b41ULL,0x524b5202c461809ULL,0xe901e40208521249ULL,0x9fee9b136253908bULL,0xb655ddb227b135bULL,0x4480f4c02670121bULL,0x97a6e75ca65a90c3ULL,0xffbe3ff46f7733cfULL,0x8ecffe227f5e11ebULL,0x77595fdefe57156fULL,0xfff37ffbfb7ebfffULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc5cded29cb3dade4ULL,0xf7b67f7b8b3faff5ULL,0xcbdbfb924b3cafd4ULL,0x5cdb37390b3c8bfeULL,0xbebf2ceb097d2fd4ULL,0xdefef67d483e97f5ULL,0xcffc7ff1483cedf7ULL,0xfbd5f77fca3effd4ULL,0xb4ebbff90b3daed5ULL,0x87bedffbeb7fefd6ULL,0x74fffef8e97e9fe6ULL,0xdfbeff72e97ffff9ULL,0xfbb3afef687cbff7ULL,0xcddefecbeb7fbbf5ULL,0x7fffdfdfeb7fafffULL,0xfcf6f7eee83e9dfdULL,0xf9ff7f678b3fedf6ULL,0xffbbfffd6a3dffceULL,0xddcbf33a083ebefdULL,0xeb76fe5fc93eaeddULL,0xb8be57e28b3cd9f6ULL,0xb9ce777b6b3caed4ULL,0xdabe2f5b0b7fbed6ULL,0xcff3ff3ec97dbdd6ULL,0xef9fefffcb7ffdd6ULL,0xb7bb5ffe487fafffULL,0xffffffffca7fffffULL,0xff9effffc87effffULL,0xfbffff7feb7fbfffULL,0xffffffffeb7ffffdULL,0xff9ffffbeb7ebfffULL,0xf67fff7bc93cfff7ULL,0xfffeff7be97fbff7ULL,0xdbd6f9476b3ebacfULL,0x1edf9bcb483eafeeULL,0xb7bfcffbc87eabd7ULL,0xfff5ffffcb7fffffULL,0xd3feff70eb7e9efdULL,0xfbb7ff7feb3ffffeULL,0xdffbffbbca3daff6ULL,0xfbffdfdfeb3efffdULL,0xfcffffe9eb7cdeffULL,0xffff7f7fe93fbff7ULL,0xfbfe7f7e6b7ffcf7ULL,0xcfefff3f6b3feff7ULL,0x5ffe5f5d687dbf53ULL,0xb4f7dfebca7cbef5ULL,0xf57db77ae93cbbd7ULL,0xdbf63df6683c9dceULL,0xc7e7f7fe4a7eabf5ULL,0xebdedfffea3ffffeULL,0xbffbdf796b7fffd7ULL,0xd3fdbef1c93e9e5cULL,0xeed67578ca7e8af1ULL,0xc9fefb7b483e8af5ULL,0x6b573357ca3dadd6ULL,0x179fcfda0a3ca5d5ULL,0xfdf5f7efca7cbeffULL,0xff9e3f6be97dbdf6ULL,0xfbd6df7eea3caedaULL,0xf3f3fffa8977bdd5ULL,0x1efdcf8bea3c6e7cULL,0xfdfef7e9493ebdf6ULL,0xbdbefd636a3ea7d4ULL,0xd716fbfc8b3e9ef1ULL,0xcff7feb3487eafd7ULL,0x6216cfdf683ea25aULL,0xb6b376efc87deff4ULL,0x6deeff6aca3dfff7ULL,0xd6ae7d364a1c11fcULL,0xd67fd91a887d8cf5ULL,0x204c48c8882088d2ULL,0xdcbfff7fea3fffdeULL,0xf7fca7758a3e8ff4ULL,0xd6b67c7f4a1ea6d5ULL,0xc7c7710c083ca0f6ULL,0x7b7f9bd6883efafdULL,0xa61bfefbc97c2df0ULL,0xcd753e5a0b7d8ee4ULL,0xe5be7dff4b3e9ef7ULL,0x8bf0e36b0a356ec4ULL,0x1dd31f82ca3e8c85ULL,0x86d06ec2430c88c6ULL,0x9040974b081402e0ULL,0xfdfe7fe2c93dfdd5ULL,0xfb85f312487c82d8ULL,0xeffe3dcbe83eaaf3ULL,0x2bbcefdac83dbedcULL,0xd0e2a37a083eaad0ULL,0xff7f3f72893cacfbULL,0xfdbe7fffeb3cbf57ULL,0xffdff267c83daef4ULL,0x6fffdfdb483cbff7ULL,0x6201cee80b3d2a64ULL,0xf97adf72693ceae6ULL,0xffbe3f77eb7ecff6ULL,0xffffffd2c97dbbf7ULL,0x6fffffcfe93faff6ULL,0xdfe737aeb35a9e4ULL,0x6104304808103080ULL,0x8b021c00082ca242ULL,0x40324000c84a0ULL,0x39cca0208341250ULL,0x1206e000088091ULL,0x6d7238011ca8c0ULL,0x52510008308401ULL,0x8b9f09176035b7d4ULL,0xc00200400208050ULL,0x9105c00808a082ULL,0xb33d86800834b20bULL,0x452c0450080ca066ULL,0x980e302033c0c80ULL,0x827a1a1801361a80ULL,0x22200012001080a2ULL,0x5570d6e3083d82f2ULL,0x879aa910027e9730ULL,0x8020224008242080ULL,0x100240080060800ULL,0x845008200840ULL,0xfcf7feffe23f7ff4ULL,0x110ba76a00361654ULL,0x1004000000000050ULL,0xe9f5d00a60363744ULL,0x610000000162000ULL,0x0,0x29642a600016ce45ULL,0xfebf3ff4697eafc5ULL,0x8fcaf7b6483e98d7ULL,0xa6141c01090da846ULL,0x32b34f0200369dc6ULL,0x10a4080040048010ULL,0x804002200200940ULL,0xe3d6fa32e834e856ULL,0x2e83be10e0368986ULL,0x103455000834a064ULL,0x84405220401c0620ULL,0x442408c8002c0b40ULL,0x9af49f2e683e59d4ULL,0xc8963ea2493edaf4ULL,0x8ebecdb30a3cccd7ULL,0x55dc5e7b083c96fbULL,0x812abf60603f1060ULL,0x4800d520020e0404ULL,0x22440012001080a2ULL,0x10050208100c40ULL,0x30824248342a00ULL,0x4220e522030f0c44ULL,0xcb922e69e07c1c10ULL,0x8b43fb1a493fc8deULL,0xa5cf858803d8ec4ULL,0x4995ea1083cb0d4ULL,0xa1002a0008340084ULL,0x4081204049140844ULL,0x8683890a033c9856ULL,0xcbce9ddbc87dbed4ULL,0xc233479a022f80c6ULL,0xf5f5b7e8c93c0e74ULL,0xfffe7d730a3fdbedULL,0xf1f5fb8e493ebcd3ULL,0xcdce1299683daed4ULL,0x82c4720825e844ULL,0x4247402081080d2ULL,0x8865541084e8109ULL,0x3900e7254834a944ULL,0x24868609023c00f4ULL,0xc6fecf738a7caff6ULL,0xddf5ffe2c83f9ffeULL,0xe7b6bf63087cbcffULL,0xef95ef6ccb34bcf4ULL,0xcb57ae34e83f7d5cULL,0x891fdd53803edbf5ULL,0xfefffefb6b3e9ff5ULL,0xfffffffbeb3ffffeULL,0xffbe7fffeb7fffffULL,0xefcfff3eeb7cfffcULL,0x6000140000200050ULL,0xba124f408038da65ULL,0x1180816080308890ULL,0x2830040000048c41ULL,0xa010720402140140ULL,0x1009a80002d0040ULL,0x42250c40800ca51ULL,0x5120002400103004ULL,0xcd563d70483cac79ULL,0xe9a4b112893889d0ULL,0x2f7f1fdbcb36dee4ULL,0xa00c34a0234aa41ULL,0x1080807388300ac0ULL,0x202c0000088012ULL,0x68c4c204012c08d4ULL,0xc48a101009088a52ULL,0x20040008000040ULL,0x50a5b06008388270ULL,0x92b00d10012e8956ULL,0xd191633b083ea6c5ULL,0x2e5ffe096a3c0bdcULL,0xefffdfdbcb3da8fbULL,0x74f7e6e40876b9c5ULL,0x651e3760083f83c1ULL,0xfb9e7e7f4b3c9efeULL,0xf7ceef38c93deefbULL,0x9affdfdbeb3fbfffULL,0xffffdfffeb7fffffULL,0xfdffbfffeb3effffULL,0x0,0xc000200068200000ULL,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400000020040ULL,0xfdfcffff5f9acfdbULL,0xfff6efefbf9bcf5fULL,0xa9deff1e618acf57ULL,0xafff7f595f9acf5fULL,0xeebfefff979bcf5fULL,0xf7f5fffd7f9bdf5bULL,0xfffebf73eb9bdf1bULL,0xfbd6fb7e1b1bcfdbULL,0x9bd7feae3f1adf9fULL,0x4fbccfeb4f1bdfdbULL,0x7dfff6ef7f9bcfdbULL,0x9f7eff716f9add9bULL,0xfebffb7f7b9acf53ULL,0xcf56fe9f7b9a9f5bULL,0xaf1fffde7e0bdf5fULL,0x7cfef6efdf9bcf5bULL,0xd9effffbff1bcd5bULL,0xffbe6ff65b9adfd7ULL,0xe3d7fbb73b1add93ULL,0xe7c7f5adb0a9fdbULL,0xfefe4febff12df53ULL,0xf73eaf62d79b5fdfULL,0xf89edffeb79adf5bULL,0xe97ffbb63f8acf57ULL,0x7fffdfddde8bdf5fULL,0xdeffffff3f9acfdbULL,0xf7ffffffff9bdf5bULL,0xfffeffffff9bdfdfULL,0xebffffffdf9bdfdfULL,0xffdf7f7fff9bdfdfULL,0xffbffffb5f9adfdfULL,0xfffff6fd7f9acfdbULL,0xe7fefffb7f1bdfdfULL,0xfbfcc716d39a47dfULL,0xc3dbbb5b5e83555aULL,0xff7fcfff7f9bcfdfULL,0xbfffffefff9bdf5bULL,0xaff77f60ff1a4f5bULL,0xfbf5efffff8bdddbULL,0xcffffe3b3b9adfdfULL,0xabffdfdffb9adfdbULL,0xb7fe5eeb939adf5bULL,0xdfffff734f9adf9bULL,0xfebf7fffdf9bcfdfULL,0xffefffbe7f9bd7dbULL,0xebdfdfde7b93dfdfULL,0xbcfff7ef5f8ad55bULL,0x1bcca64b1318cf1aULL,0xfeb5f7e1959a5f57ULL,0xc59ae139a313cd5fULL,0xd9df7fdf3f9ad7dfULL,0xb5fb5ff3bb9bcf57ULL,0xf1ad777fd79a4d5fULL,0xcfbe6b761319cf19ULL,0xabdedfaa8b035f5bULL,0x2d7ffedc1f8a4fdbULL,0x8cdf07c7779a4f42ULL,0xddffff5a5f1acf41ULL,0x7ffefd73df9bcfdfULL,0xdbf0ffb2338b5fdbULL,0xea5bbbe7d31bdf5fULL,0x779fcfde3f9b4fdbULL,0xfcfffef83f0acf5bULL,0xdbbaff72971a5f1bULL,0xd482ed3ea38a8f1bULL,0xcfd6fe5d7f8bcf5bULL,0x83cfdf550a1a8f19ULL,0xd0b6a4fc13185d1aULL,0x55fe6f623719df5bULL,0x7a9cffc43218cf11ULL,0xe192ea235b8bcd0bULL,0x66dadc433d9b5d5aULL,0xb0b6dce0378bd5dbULL,0xf27cef16039b5f1bULL,0xfef63f531f09cb13ULL,0xafcffe9c3783df5bULL,0xeb7f3ff70f9bdfdbULL,0x85ba7fb95f1b4f4bULL,0xf7ffb568a71bdf5fULL,0xfabafd75379bdf5bULL,0x83b2fba7738bd75bULL,0x4fd55ff79b93cf5bULL,0x4bbec7776b9ad753ULL,0xb487faea5f885753ULL,0xabb4ad623218df1bULL,0xd95ce6281b00df10ULL,0xffffffffbf9bdf53ULL,0xee95cf791b0bdf53ULL,0x37955f62f203cf49ULL,0x9a9ebd737b0bc51eULL,0xdb93f77b1b9bcf93ULL,0xcfeeb62d638adf5bULL,0xfbf7ffdf769acd53ULL,0x74fdfeefd79b4f5bULL,0xe3ec37600782c7dbULL,0x81122c351e084301ULL,0x8ddeeaa9219d71bULL,0x6a8dded1178acd5bULL,0xd0127fe2138bcf52ULL,0x5184c972121a8c1aULL,0xda0a2400320a83d1ULL,0x4582f0e23409cf5bULL,0x8472340268ac312ULL,0x3b30017003088719ULL,0x10300a0405124c00ULL,0x2a2240001820283ULL,0x8920c72803124d12ULL,0x21400a01000a8403ULL,0x410800202004c00ULL,0xb100929882834508ULL,0x480024000300440aULL,0x101820004004440ULL,0x0,0x410000000020040ULL,0x400134800080302ULL,0x42a0006012020340ULL,0xa8048e0c12014510ULL,0x8102f80a0700cd50ULL,0xa448c41a06004710ULL,0x8011c02202114c10ULL,0x9110844030000100ULL,0x800002000004800ULL,0x0,0x140000002000008ULL,0x84e9d460001a0c4eULL,0x21000a2041020102ULL,0xffb06d773f9acf5aULL,0xf4f7f6201a1a075bULL,0xaffee4fc20b4f5bULL,0x310465a5a0a0d41ULL,0x820c850004005750ULL,0x4e462062110a474aULL,0x4097d30813924610ULL,0xaa518eda4703571bULL,0x49bccc9264835912ULL,0xd4e5362807095a1bULL,0x41002c0012001010ULL,0xc9d8ff7a9182df5fULL,0xca66b67bfb888f43ULL,0x17fccfbbdf8a5f5bULL,0x70ce7c6443824b1aULL,0x4a204540120a0617ULL,0xe380c32c0a13cd5bULL,0x100822012001010ULL,0xa854410e0108441aULL,0x422120800021910ULL,0x979a0960070bd2d3ULL,0x820c52e25382c145ULL,0xcfcbffba539b5fdfULL,0xab5a5ed3008bc75fULL,0x30220442000a0c00ULL,0x99430b60001a0cc3ULL,0x5832066b500a175bULL,0xa117d621121a0749ULL,0x29539a521200961bULL,0x14fb7d53491adf57ULL,0xd5edb76a3f9ad7dbULL,0xcfd6e47fb38bcf5fULL,0x89e6ff3b5e934f9fULL,0x6ff77e71771acf5bULL,0x308ac44222115718ULL,0xd110823042088181ULL,0xc49421f1810a4d52ULL,0xb990c20c248a4751ULL,0xc8105923104f03ULL,0xaeffdffbff1bdfdfULL,0xf6f7fefc7b18cf1bULL,0xbf7efd71d79bdf5bULL,0xf79fffff731bdf5bULL,0x4fdff3037c1acfd3ULL,0x2fd35ed1539a4d1aULL,0xfc76bee7b79bcdd3ULL,0xdbfeeff2b71b5f5aULL,0xffbf7fff7f9bdf5fULL,0xffeeff2e5318df53ULL,0x84490004004c41ULL,0x88100081004400ULL,0x6064b49012108710ULL,0x800040002004018ULL,0xe004300001804102ULL,0x281d504802928650ULL,0x40a66200001441aULL,0x4004040202004810ULL,0x9f102fecbe984d99ULL,0xc9c6f36116805d58ULL,0x4283926b1782c75aULL,0x4a10c1480301c518ULL,0x24000080100ULL,0x0,0x4200200410020104ULL,0x42c332414182070bULL,0x2abcdc5e73100783ULL,0x1102060010a0c43ULL,0xcd386c510b12475aULL,0xf7b6ffff7f9adf5fULL,0xeffffecfd71acddfULL,0xef7fdfdb379bdfdbULL,0xf4bffdef5b9bcf5fULL,0xd96fe77e3f9adf53ULL,0xffbedf77f30bdf5fULL,0xc3dfef3b378bdf5bULL,0xeff7dedfff1bcf1eULL,0xfeffffefff9bdfdfULL,0xffffffffff9bdfdfULL,0x22810514408a0146ULL,0xcb34fff4599a0f5eULL,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100000800004040ULL,0x0,0x0,0x0,0xb5ffffecfdf5ffe6ULL,0xf7fffffcfcbdefe6ULL,0xf5befffbb4b97fe6ULL,0xcbf7ef5f749c6fe2ULL,0xfcd7bf53fcbdefe6ULL,0xb29eddde74996ae2ULL,0xd7dd76ebb4b86fc6ULL,0xffbe6ffafcb9f7e6ULL,0xcbd4ef3ff4b9ffe6ULL,0xafdbbbdffdb96ee6ULL,0xbbbfcfdfbcd9dfe0ULL,0xf5feffe7ddfdffe6ULL,0x9fbebff2f9fceee6ULL,0xfb9dee3f7db97fe2ULL,0xdfeffebffcbd7fe2ULL,0x3fdbdfdbbdb97fe6ULL,0xd6fffeeff5fd7fe6ULL,0x9dbfb77b1cb5eec2ULL,0xfbb23ffff5b87fe6ULL,0x8c4fff187da81de2ULL,0xefff1b093c205fe6ULL,0xf57f5ef3f498e5e6ULL,0xd8ec279a101926e6ULL,0xcab62d7694f97fe2ULL,0xedffff7af4dcfde6ULL,0xcfffff4bf5f97fe6ULL,0xdfbecffffdf967e6ULL,0xf7fffffdfcfdffe6ULL,0xfffe7ffffdfdffe6ULL,0xfbfffffffdfdffe6ULL,0xeffffffffdfdffe6ULL,0xfffffffb7df97fe6ULL,0xfff5fefffdb97de6ULL,0xdffe7feaddfcffe6ULL,0xebf3fffe9dfcffe6ULL,0xee6fb769fdac7fe2ULL,0xffbefffffdf97fe2ULL,0xfdf6fffefdfdffe6ULL,0xdbfeff73fcfd6fe6ULL,0xfbb6ff7f5cb96fe6ULL,0xcd5fbf8bfdf8efe6ULL,0xebffffdbdcf9ffe2ULL,0xfeb7f6fbf4bdffe6ULL,0xdbeff7fb7dbdfee0ULL,0xffbf7f7f7df9ffe6ULL,0xef7fffbffdbde7e6ULL,0x6b7edfdfb4f95ee2ULL,0xb5f375f335b9efe6ULL,0xfd6df75cedf94ee6ULL,0xfeb63f5775b9efe2ULL,0xdfd3e57ebcf97fe2ULL,0x265fde6fd5b92d62ULL,0x8efb4fe9d8b967e0ULL,0x97ddff6334f8eee6ULL,0xeffe7c6534fc7fe2ULL,0xe9f4ffef3dfd77e6ULL,0xaed61edf15b95de6ULL,0x9ebb7feb3cf9ffe6ULL,0xfcf6feee34dc6fe2ULL,0xdfbeb7f0f4f9efe6ULL,0xab95c77e7dbd6fe6ULL,0xef733ebbddf1efe6ULL,0xffbf5ffbf5b8ffe2ULL,0xf5ffd7eb7cb9ede6ULL,0xaffe7ff2fcb97fe2ULL,0xff9e3fbfdcfc7d62ULL,0x4fdbbe137dbdc9e2ULL,0xabcfd5dc50d8e422ULL,0x46f0f47898386dc6ULL,0xe943ea6235e8dce2ULL,0xfbb0ee7735b95ee0ULL,0xdedaa2b83db9e6e6ULL,0x2dbeff887cf8cee2ULL,0x259455b23490a4e4ULL,0xd97def7e35bd7fe6ULL,0xde97a7dcc4b96fe0ULL,0xafd7d3fdfdb97ee6ULL,0xfdc1f5a14b0dee2ULL,0x953855fa2cb46dc2ULL,0xf7fdf6ae7c3957e6ULL,0x25967ff950b9e3e6ULL,0xcbc6f336d49c4de2ULL,0xbf21ed75cb8dca4ULL,0xbcbb6dcc14f8eae4ULL,0xfcffd6fa75f9cfe6ULL,0xd2682dc28c3957a0ULL,0x20a4820400952000ULL,0xfdde7f6b79b94ee6ULL,0x8312dfd314f877c6ULL,0x315d9ec8841865e4ULL,0x9beeb772d1a8eee2ULL,0xfcb4ed6675bc41e2ULL,0xcbdfffd355f9fee2ULL,0xaf5fdddeddb8ffe2ULL,0x34fadeea74d97fe2ULL,0x96e4454150106560ULL,0xd6b53f4e7dfcdfe2ULL,0x4ec6dbaa9c99f686ULL,0x2f7e9f5fdcfccfa0ULL,0xdb5317f470b167e2ULL,0x5024c24505b90020ULL,0xa0274018004ca4ULL,0xebdff78814ecffe2ULL,0x8000e7041204a22ULL,0xa222000801280480ULL,0xc07034829db04d40ULL,0xcf34377415b86560ULL,0x89708340083867c2ULL,0xa008000300206842ULL,0x20204c0042012c0ULL,0xc44090000400ULL,0x6a261460340063e2ULL,0x0,0xa448261b34912ae0ULL,0x8012044004004780ULL,0x123226088cc480ULL,0x8018002000180022ULL,0xd010312a04983060ULL,0x800100000000480ULL,0x5500050010020ULL,0x20004000002040ULL,0x20044400c280000ULL,0x4202100112000ULL,0x8012600290080420ULL,0xc80108004480ULL,0x20040040000020ULL,0x696cff6234d96ee0ULL,0xffb62ff7fcbdffe6ULL,0xc8d4f32f94bd5fe0ULL,0x23549ad0b4b416e2ULL,0xb42a4f6914996fe2ULL,0xb8112618449408e2ULL,0xda142130049c6040ULL,0x8bd8df382491dfe2ULL,0x8480000030808000ULL,0x8928902140014e2ULL,0xa414022114205c40ULL,0x408029000080e840ULL,0x9020208000045a0ULL,0x8000621284004540ULL,0xfbbedf3afdb97fe2ULL,0x300c0c4a30006c40ULL,0x4812005000314042ULL,0xe088d30c14907cc0ULL,0x144c00044001120ULL,0x806010000114040ULL,0xe4d47cb4341032e0ULL,0x2260500300000c6ULL,0xa3b43765d0b86ee2ULL,0x8debffbafdb16de2ULL,0xffadfc9a0b96ee6ULL,0x16d2572014b14840ULL,0x630c812014a0cc40ULL,0x92902b50309940c4ULL,0xc9c8652228b962c0ULL,0x63d696b354d07ce2ULL,0x96bbefffbdb97fe6ULL,0xb53577e29da84ee6ULL,0x7fe67dedbcb9ffe6ULL,0xe9d7f7ab55bd6fe0ULL,0xe3f607d730dd65e2ULL,0x1b316f111c084ce4ULL,0xd3edb772b1316dc2ULL,0xf39e255270b8eae0ULL,0xe9ccf71690bc6fe4ULL,0xcfcb9f1399b96ee2ULL,0xfb9ccfb1f098e7e6ULL,0x97b7f6ed54b9efe6ULL,0x922b3de3b0396fe2ULL,0xdabef7695cf16de6ULL,0xdfdfbe9314b1ffe6ULL,0x3c7585830b966c0ULL,0xc01280e44c906060ULL,0xf96dff7b7c98ffe4ULL,0x4b982f26d5f825a2ULL,0x4908d02214a85c40ULL,0x1c21810200061e0ULL,0xb0542a300048e6ULL,0xa0042810008820c4ULL,0xd5a4275410914cc2ULL,0x582400114885420ULL,0xe2d61ac894b9d562ULL,0x10000004800ULL,0x12a044010908002ULL,0x280000004400ULL,0x48402002a0004440ULL,0x408a1209840045a4ULL,0x480001200000ULL,0x804000004400ULL,0x100020000000480ULL,0x8000912c10a80862ULL,0x8480080140106420ULL,0x8000040000084000ULL,0x42204a4d00044a0ULL,0x162204084880540ULL,0xcb9a7fbf953167a2ULL,0xffdfffbffcbd6fe6ULL,0xdfdeff5bf5bdf7e2ULL,0xd6fdefe6fcf9ffe6ULL,0xbfefaf7ffdfdffe2ULL,0xcb36377b35f96fe4ULL,0xccfdf3814b9efe2ULL,0xeffeff5addf9dfe2ULL,0xfffffffffdfdffe6ULL,0xfffffffffdfdffe6ULL,0xffbf3fff3dfdffe6ULL,0xfbfffffffdfdefe6ULL,0xa10010000ULL,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9965a571844412ccULL,0x0,0x0,0xabffffd77efdf6eaULL,0xdffbdff7fefef6eeULL,0xd7ffff7fb6aed7eeULL,0xffdefff7f6f7f6eeULL,0xfbfeff2ffcf6f7eeULL,0xefff3adbb4f6f6eeULL,0xfffffffbbefef6eeULL,0x94f3ffdd24be56eaULL,0xdfaefd62b61677e8ULL,0xfbfce77ef6f6f6eaULL,0xdfffbbfff6f754eeULL,0x7fb7effbf637f7eeULL,0x7495fefbce87f76eULL,0x9ffafff36eff77eaULL,0xffbeef3feaf7f4eaULL,0xcfffff7b3ef677eeULL,0xffffffdffefef6eeULL,0xf6ff7ffefcf6f3eeULL,0xd9eeff63dcf6f7eeULL,0xfbb95fbef49654eeULL,0xeffeef1fee76f5eaULL,0x2b5e17cd6c5e56aaULL,0xb5b657eaa07452eaULL,0xcdecb670f6be7468ULL,0xcc86376b8cd614c6ULL,0xfffffffee2f7f6eeULL,0xefffff7df45ef7eeULL,0xbffb7ffbf6fff6eeULL,0xf7ffffffe6fff4eeULL,0xfffe7ffffefff7eeULL,0xfbfffffffefff7eeULL,0xfffffffffefff7eeULL,0xffbfffffeefe77eeULL,0xf1f7f67ffef6f6ecULL,0x7dfc7ff8b65e37eeULL,0xcb86e7ef3656f66cULL,0xeae2ae75441e366eULL,0x3ebfdffcfc7777eeULL,0xf1fef7fffef6f7eeULL,0xfffe7c63e6def66eULL,0xf9bdff362637f6eeULL,0xdecfff1bb4fe566eULL,0xfa7fdfdbf6f4d7eaULL,0xf7fe7efaf6f676ecULL,0xffffff7bb47ef7eeULL,0xebff7ffff6f7f6ecULL,0xcfdfffbff67e77eeULL,0xfbf3dfcffc16f5eaULL,0xffbf6fffbe5f76eeULL,0xd76fff768cf7d6ecULL,0xfebe3df8ee5651eeULL,0xede3e7ececfe76eaULL,0xef7ffedb7e3e76eeULL,0xb73f7cf2acf674eaULL,0xf7edd67c747fd66eULL,0xf5de7ff7b66f53eeULL,0xe904ffabf476f6eeULL,0xecffbedd661ef3eaULL,0x5ef8cfcaae97d2eeULL,0xf8f3bcfce6b675eaULL,0x7caefd7120f656eeULL,0xdbf2eb7ef67776eeULL,0xf7fa3edbfc7e12eeULL,0xafbfdfffc43ff46eULL,0xd4d5dfeae4b7f6eeULL,0xefbefd62fefff66eULL,0xf180e77c205671eeULL,0x4bdece3f969f56ecULL,0x2bbecf4bb63674eaULL,0x14e76660a4b4746eULL,0xa49ea070221542e8ULL,0xff144d74545630eaULL,0x4ece76ba289e37ecULL,0xa1d08e4216b25566ULL,0x1cc47502a014d02aULL,0xf9f7b770c656546eULL,0xfaad25fdde16776cULL,0xf5c3feb682b2d6eaULL,0xffdfdedae656f7ecULL,0x1eb1cf4a141cd4eaULL,0xd5f7bff9a6f6f4eeULL,0x7b9c7d7cb426d2ecULL,0x81c6f63cb49610ecULL,0x36837aa222a0d668ULL,0xacf7a7adc6327664ULL,0xf477ff7db676566eULL,0x4cbe47d38686f262ULL,0xebb0e72f2cb670eaULL,0xae47fa937eb27468ULL,0xa41eced33456f6e0ULL,0xbcb774e0841e71ecULL,0x1fae2dc12474f06aULL,0xff845f7fe84634eeULL,0xdcffba9a2a92d26cULL,0x7bdedbdfa4f674eeULL,0xb6f6f6e8049776eeULL,0x296c3f608c3476eaULL,0xc2242c63a210556aULL,0xc5d7fb2ca455546cULL,0xeb7ffe5cbe3674ecULL,0x24207480284710a0ULL,0x6008f3612422406cULL,0xfea60f4a647670acULL,0xc154d6043c367664ULL,0xfbfffedfeedcd7eaULL,0x4bb53d2665410ecULL,0x9148f44a2610c468ULL,0x8004000004000020ULL,0x6180e31c2e1e51e6ULL,0x400000820000028ULL,0xde200001068ULL,0xd4a500a0a0c01268ULL,0x54014000000ULL,0x8800042040000060ULL,0x40600011a0120024ULL,0x1142c30000d542ULL,0x304710e568e25220ULL,0x8400014000001008ULL,0x800000800040008ULL,0x1800000101000ULL,0x8040000000c0000ULL,0x86811661661c7464ULL,0x0,0x0,0x4806a08200c5260ULL,0x144044c00163004ULL,0x80400040000020ULL,0x7d64766036d4f5e8ULL,0xdeb7bffff6fef7eeULL,0xc4cfff1fa456d2eeULL,0xae5f5fdef455f4aeULL,0x94e9cdb3cc15f46eULL,0x20100000c000ULL,0x81041900040ed16cULL,0x8bb06763e4bef0eaULL,0x480014000000ULL,0x80a04979460504e0ULL,0x404000000040002ULL,0x280cb52204067662ULL,0x98007fa4400c1462ULL,0xc022960000126240ULL,0x8e99c75a2c1674a8ULL,0x4160e88200414e0ULL,0x9212704004945002ULL,0xcaa2d1262054d2eaULL,0x2c800000010e0ULL,0x18000540001c01e8ULL,0x5455fe6bfe56d66aULL,0x2a0560801c40e0ULL,0x91945f5d024e1646ULL,0x63d39e3a809654eaULL,0x2ac674dda4227768ULL,0x6580a34040062ULL,0xc91ca03824525260ULL,0x54143c52e01412e2ULL,0x8452b290244630a2ULL,0x695f6e45263dd5eeULL,0x1eba7ffba433566eULL,0xf5feff6ef43f7066ULL,0xffb67f7ff6fe76eeULL,0xe9ded72fe67472eeULL,0x6ed33e93a634d668ULL,0x14991ca16204c062ULL,0xd14bb6db34bef06eULL,0x65b2b56a04bf14eeULL,0xf316ef028c6775e2ULL,0xc6da12eba4f05168ULL,0x16140e02a0020024ULL,0x7fbf7feff4f6f6eeULL,0xf5bafd73b63e576eULL,0xcd95bfb92cf6f6eaULL,0xcffffb1fe6f3d66cULL,0xec5fde5076d2f2e8ULL,0x7cf4a452f617d362ULL,0xf1ff7ffe7696f6eeULL,0xabbc0c65b614546aULL,0x81dd4b186416106eULL,0xa3755c8a61434e8ULL,0x14137fa02686d662ULL,0x20050020000260ULL,0x20200000c000ULL,0x81e433322426e066ULL,0x2010108ea0022068ULL,0x800010204001240ULL,0x90145210000cd08aULL,0x40100001008ULL,0x119002014060ULL,0xa5802110290504aULL,0x410028140124024ULL,0x0,0x4800000000801000ULL,0x840000000020080ULL,0x2420a6168005020ULL,0x400100000100020ULL,0x74001290009670a0ULL,0x0,0xc103f92c2602d2eaULL,0xdfdfbe3ffeb6f7eeULL,0xaade97daa436f6ecULL,0x34b97efbf61670eaULL,0x9367fd5834725664ULL,0xd3b67ff58eff576cULL,0x4d4acf2e3ce656eaULL,0xef7fffdffefef5eeULL,0xf4bffffffef7f6eeULL,0xfffffffffefff7eeULL,0xffffbffffefff7eeULL,0xfbffff7ffefff7eeULL,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x408300000090240ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xff7fff77f1feaff8ULL,0xba9a677c71adb3caULL,0xefd7ffbf717db7dfULL,0xefdfdf79f17dafffULL,0xbeb27ffa71fd9fdeULL,0xfbbef77ff1debffbULL,0xcbdeffe5f17d9ffbULL,0xebbbffbb717cb7ffULL,0x6fdebffbf1fdbfdbULL,0xf7bfffeb71fdbffeULL,0xd5aefe77f15ebffbULL,0xeffeede1f1fdafdeULL,0xfbfdff3ff1ffbffeULL,0xddfffbfff1fdaffeULL,0x56bfcfdb71dfa7faULL,0xf7e6f7ef517fbfdaULL,0xffffff7371ffbfffULL,0xdf8eff7b71ffb7feULL,0xffffff3bd1dfbffeULL,0xafdfdfdff1fdbffaULL,0x74f7d6ec71f4aed8ULL,0xffedef3b717db7fbULL,0xfbb77ffef1dda7daULL,0xedffffbb71ff8fdeULL,0x2e7c588a014d074eULL,0xbfde1bf371dd06d2ULL,0xf9dfff1571febfeaULL,0xdeae3f65711caed6ULL,0xffffffff71ffb7fbULL,0xefffffddf17fb7faULL,0xbdbf5f7bd1fd2edaULL,0xf3fffff971ffbfffULL,0xfffefff7f1ffbfffULL,0xebfffffff1ffbfffULL,0xeffffffff1ffbfffULL,0xbffffffbf1ffb7ffULL,0xbfffff7ff1ffbfdfULL,0xdffe7ffbf15f977aULL,0xdbddef7ff17d1ff8ULL,0xe6fbbff950ddbf7eULL,0xffbfdffbf1fdaffaULL,0x7dfffefad17fbfdaULL,0xdfffff62f15fbffbULL,0xfbf7bf7f717dbff2ULL,0xefffff3f71dfbfdfULL,0xffffefd7f17fbffbULL,0xfcf57efbf1fd9fdaULL,0xfdfbfffa71fdafdaULL,0xbf947fbe717daec7ULL,0xfdffffbf71ffbf7fULL,0xebff76cbf1fea7deULL,0xdfbb7ff7d17d87daULL,0xffeeff70717dafd8ULL,0xfa9e2f7ff1fdb7d6ULL,0xd7d7eb6b709c17d0ULL,0xefdfbffff1ffaefaULL,0x96fb6fc0f1bd36feULL,0xfbffbffff17fbfd7ULL,0xbeae3f687115af58ULL,0xebb0fb6af17dafdbULL,0x6cdf139f516d86deULL,0xb7fecfb7f07dbfecULL,0xd5e3bf7a51bd3edeULL,0xef7e7ffb717da6faULL,0xcbc6eb03f1fca2faULL,0x4ecab7735075a3c0ULL,0xbd8efd7bd1dfbbf2ULL,0xb1bff7faf17dbfdeULL,0xaf5eedf5717dafecULL,0xf936f7fa71dd3ff2ULL,0xcfffbe1551bfaefdULL,0xbf5fdfdb51fdbff8ULL,0xf4f6d6e750fd8bfaULL,0x2d682f4a01dd0792ULL,0xfeb2af6a51df87c2ULL,0x8fc6f33ad171afccULL,0xaf7c7f0cf04f86d8ULL,0xf2926fe951debedaULL,0xc97eec50517ea6dcULL,0xee363d7351fdaff0ULL,0x8c5ff3ea51dc8afaULL,0x2b5eded751d4bfdeULL,0x3efdd7f0516da7caULL,0x75a6b77ed1fd9bd2ULL,0xc3debd7451f6077eULL,0xcbd9ef63d1ee37f2ULL,0x4bc89a7871df8358ULL,0x7937dd671f52fd8ULL,0xf7fbf7fc50ddadd8ULL,0x87be7df2517cbfcaULL,0xf9d6ff1e415fb3eaULL,0xe939e8b717cb3d8ULL,0xfadcd5351dfa7d8ULL,0xd8b4874850fc3fd4ULL,0xbf94ef4250cf27faULL,0xdfa977de717f9edaULL,0x8c46be89515d835aULL,0x2b18ff9bf17fa76cULL,0x90b6d6ead0fda6c7ULL,0xe0f6a761716597d6ULL,0xdab47f7e513f17daULL,0x54c26f3b519c0fd0ULL,0xeb3570caf11d3ccbULL,0x92330de7711e06c8ULL,0xc510a843d0500210ULL,0xc11a011180002ULL,0x8f4eb600d11503d8ULL,0xfffffffff1ffbfffULL,0x2fb66f72d07d86d5ULL,0xf17dffe870d78fc9ULL,0xce8020029000a250ULL,0x90c5324902e0370ULL,0x9d0125051a4a240ULL,0x0,0x6000021040410080ULL,0xb0042c2020010350ULL,0x4850800851002082ULL,0xa8620800101c0608ULL,0x0,0x500886e1b1008754ULL,0x90000c4000680300ULL,0x0,0x80200300240008ULL,0x8200100800ULL,0x0,0x0,0x800040000040008ULL,0xf987d52a612ea0eaULL,0x200000850000000ULL,0x80846001410242ULL,0x6144ff6370349743ULL,0xfeb73fff71ffaffbULL,0xefbfffaff1ffb7eaULL,0x2fdefbda715fb6f8ULL,0x8aabcf7b517fb6caULL,0x8002000080200ULL,0x4000280200410140ULL,0x89c2e77b714d07d8ULL,0x4eea7e2040c5365cULL,0x1001800300580308ULL,0x956c924c503c23d8ULL,0x3000246040150840ULL,0x980a303400c2480ULL,0x8000821050000080ULL,0xaa2441a8101716c8ULL,0x7000841001012248ULL,0x5080294041242421ULL,0x438051595061a352ULL,0x8dc7bb83417daf5bULL,0x103494040102270ULL,0xf475566a7055b6d4ULL,0x7986936010451749ULL,0xff946ce3513da65aULL,0xcccfa24e515796d8ULL,0x8bdc5d99513c87d8ULL,0x96d915285128a250ULL,0xd860e36050cf86c0ULL,0xf6b4277551d7afc9ULL,0xcbe4972251d495d1ULL,0xab5bff47d15ca7f0ULL,0x86a3ed5951fdb5e2ULL,0xd1fd777b5177a7f8ULL,0xdfa67f6ed17f37fbULL,0xebf7f7ff51fdafffULL,0xedf7bedf71f7afe8ULL,0x5e8ed56b71efb7caULL,0xd526f3737155a7feULL,0xeeae2570d1741a9aULL,0xfb97ff0751ff37c3ULL,0x449216015108a252ULL,0x27b84f60105d3fc0ULL,0x1140f6f8f114a6dcULL,0xdb8e7d72717d96d9ULL,0xfa34ffeb51b5bff8ULL,0xcffefebbd15cbfdaULL,0xe9deffdfd1feaf7fULL,0xd6f7fcfff17eb7dcULL,0xed5fff7b71bcbeeeULL,0xffb36f7470bdbfdcULL,0xc0d3c33971d51b4aULL,0x69e0f12413c2a22ULL,0x56933fe8017907d2ULL,0x1180a90011400160ULL,0x5c263ee651f5af74ULL,0x8002422040003401ULL,0x48824940111250ULL,0x422000401010060ULL,0x1288c4001ac020aULL,0x0,0x2200800010140ULL,0x4140014e51110240ULL,0x20040000000210ULL,0x440270000700250ULL,0x4002001000020ULL,0x400200000042ULL,0x50100800010060ULL,0x130065250412540ULL,0x417086ca515c2d0cULL,0x12000c4100000342ULL,0xa881c16a5014ac48ULL,0x9ec2bb1ef07f1bd0ULL,0xafdcc952f1bca3caULL,0xfcffffff717fa3faULL,0xdfd5af17716da77aULL,0xfff7ffff71febfffULL,0xffffffbff1ffbffaULL,0xefffffdff1ffaffeULL,0xfff7f7fff1febfffULL,0xfffffffff1ffbfffULL,0xfffffffff1ffbfffULL,0xfbfffffff1ffbfffULL,0x100080041000000ULL,0x0,0x0,0x0 },
	{ 0x94ffa8402a06520ULL,0xac446a460220ad44ULL,0x20aacfd990fd2442ULL,0x8c761a69dab92d30ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd9fdf77adafdbd76ULL,0xdabe35369ab5e5f6ULL,0xdffff7b7daedbf76ULL,0xefdfdfdb9af0bdf6ULL,0xdebaf7fb9af3bff4ULL,0xf3fff7fbdafabfd2ULL,0xfbae7fefdafcbd20ULL,0xdbdff7fe9aecbff0ULL,0xedff1ef992b4bdb2ULL,0xff3b7ffadab5b566ULL,0xfff7f6f6d2a6add6ULL,0xec5c75b390f9af74ULL,0xdbf5fbbe9af5bff2ULL,0xdddfbfdbdaf2bdf4ULL,0xcdfccfffdaf5a756ULL,0xf7fffefbd8f9bff6ULL,0xfffffff7dafffdf6ULL,0xeb9fffffdaffaf36ULL,0xcfdffe5bdaf6af70ULL,0xebdfdfdedaf6aff4ULL,0xd4bbfeee9ab2ad50ULL,0x957f77f9d8e4bd34ULL,0xfb9f6ffc9af4ad74ULL,0xe9cffeffdae5fdf6ULL,0xce1f5f5b92b9bdf4ULL,0xb0b947e0dab42d04ULL,0xbb5e5f6b9af28794ULL,0xdabe2ff79ab2fd36ULL,0xefffff3fdafbafd6ULL,0xeefefed992fdbf76ULL,0xfdffdffbdafcffb6ULL,0xf7ffbffddaffbff6ULL,0xfffe7fffdafdfff6ULL,0xebffffffdaffbff6ULL,0xffffffffdafffff6ULL,0xfffffdffdafeeff6ULL,0xf5fff7fa92efad76ULL,0xfffefde0d8f1bf56ULL,0xcbfcef6f9ab1fd64ULL,0xfffbbeffdaf3bd64ULL,0xb71e46769a69ede4ULL,0xf5fffffadafbffe6ULL,0xfffffff7dafffff6ULL,0xfdb5fb7edaf6fdf6ULL,0xebdffebfdaf1fd36ULL,0xefffdfd7daf7bff2ULL,0x1cbed7e0da60fd54ULL,0xfdfef77adaf5adc6ULL,0xfb9467f7daf0af12ULL,0xedd3ff7f9ae7ad66ULL,0xeefedfd3dab0adf2ULL,0x9dfe5bebd0f9ffd6ULL,0xdb2fffff9afcef64ULL,0xdaef39ea9a20fd66ULL,0xc78af6a3daf5f564ULL,0xbfdffedddaf2fdf6ULL,0xbfefd7f29870b576ULL,0xdff4b7ee9afb2d74ULL,0xb8ac2d669860b564ULL,0xe3fafb3e9ab3bde2ULL,0x6dc63ba592b4e5d6ULL,0xffbffffedaaef534ULL,0xb5f7f7ec9ae4cfe0ULL,0xffee7ff1da78fdf6ULL,0xf3d4ef6edae5ed36ULL,0xcec33e199af13744ULL,0xee3c5f4b98f03466ULL,0x75f7ffe8daf6bd34ULL,0x8f7aff5290f1bfa6ULL,0xff9dff6f9afebd74ULL,0x85dcf62fd238ad44ULL,0xaf5cc35b9aedadb6ULL,0x9ed6f768daf4f560ULL,0xc8a66e6892a0a520ULL,0x4386646490902150ULL,0xc5c7af9e9ae6ad60ULL,0x475cc00a9070ba02ULL,0xb61557639220bc56ULL,0x91000e628064af04ULL,0xfe863f43d834ed24ULL,0x6d83f62bd2a42d14ULL,0x2f56b35f98f2bd50ULL,0xffab3f6f9ae4cd60ULL,0x93f7f769da78bff2ULL,0xf7fe7f6d9af5ad54ULL,0xcbc6e4638a74a504ULL,0xe7e7397918f0ef26ULL,0xb7bfcffd9af4bd44ULL,0xb56fb67b9aeafdd4ULL,0x86a445619800fd20ULL,0xa98de774daf4e566ULL,0xdbdb3fcf9af8fd64ULL,0xcbb4cb6390b4ad26ULL,0x55f2ff7092a0ad14ULL,0x5baaff63dabc2534ULL,0xdfb6676fda70ffe0ULL,0xebf7ffbf98e2fff4ULL,0x75f9ed79aa69d50ULL,0x50b4b2421a883412ULL,0xd9aefff2daf8a5f4ULL,0xde80420292342410ULL,0xa88efb8e12247544ULL,0x2b8a335398a0b416ULL,0x42014a098c02000ULL,0x2040227098608520ULL,0xccf2247fda802d44ULL,0x8f54f02490a0a510ULL,0x87e3c09da81acc0ULL,0x9ee87d7098e12754ULL,0x30ed244a12b02544ULL,0x933c0d61d8c0a100ULL,0x0,0x2042001080800140ULL,0x851a526390202142ULL,0x42054090801000ULL,0x8a02644188882404ULL,0x81004e709a242510ULL,0x450301190642732ULL,0x808000000120ULL,0xd2b07e7012a4cd70ULL,0x9000052098400400ULL,0x6888c132d0a23c30ULL,0x384222398082806ULL,0x8480110002420ULL,0x0,0x540280202000ULL,0x0,0x840280000208c00ULL,0x0,0x47806bac8d12600ULL,0x1f44702288c1bc44ULL,0xfab3bf5fd0e1bf42ULL,0xe9ffff32d274bd44ULL,0x2d108ec992d8a516ULL,0x1aa85f9ed0e2a794ULL,0xd99c040082a42544ULL,0x78a22c420030e584ULL,0xe102b31b9231e512ULL,0xef427e12d8a8a920ULL,0x5218421812a02522ULL,0x50c0143980508a80ULL,0x440205488a008580ULL,0x4002000200040ULL,0x840000000202000ULL,0x218430010222440ULL,0x20880000000ULL,0x4080055000382000ULL,0x300b632c0702446ULL,0x402fea89820a944ULL,0xab5a11489024ad20ULL,0x14d45ed3daa02d40ULL,0xdb44a5c392b4b540ULL,0xf7382f579ae42d26ULL,0x9943ae08c8782c44ULL,0x8846de1398f00120ULL,0x828856b290b4a5a0ULL,0x79cc76269afc2564ULL,0xf2b42576d8f4ed00ULL,0xca41f36bd2b4b570ULL,0xafebeea98f8af60ULL,0x2626d7789a62a552ULL,0x9c57af7898f0eda0ULL,0xfa967d7fd2f0ff64ULL,0xfbfffb1bdab2fd76ULL,0xffdb1fdb9af0eff4ULL,0x86bdd6fe9aa0a572ULL,0xf0dcb75d92fead46ULL,0xc52c7c71d230fda4ULL,0x8a52e5208a65a584ULL,0xcdcb7e59d0b4adf0ULL,0x633a4b53d072ad60ULL,0xfcdffeac90ff6560ULL,0xfdae1d229a70bd56ULL,0xbab4fbfb9ae0af46ULL,0xcdcbfe19d0e0ad56ULL,0x6b7bcf42d2fce5e6ULL,0xdcd75ffd92fff544ULL,0xf168ab739ae0f552ULL,0xdf846b6788f4ab42ULL,0x8fdfef4f9af4bf64ULL,0xa08b108398402256ULL,0xf2b65febd0bdf956ULL,0x500070021080a804ULL,0x8800050008400100ULL,0x4880921880a52404ULL,0xa304108402400ULL,0x8400042008600040ULL,0x8a510d8400400ULL,0x1240400190102012ULL,0x8080339310283404ULL,0xc8d21c0382b42540ULL,0x8002000080800000ULL,0x1100242010308480ULL,0xa00000000008800ULL,0x0,0x0,0x3800010888410000ULL,0x2110000400ULL,0x13380e5012b08520ULL,0xc430637090a62520ULL,0xcfde11019ac32452ULL,0x335e5dca90a0e522ULL,0x34b57ee8d264f732ULL,0x9d6ff763daf4ffe4ULL,0xfbd37f6e903d3d46ULL,0xffffffffdafffff6ULL,0x4fffff579af4bf74ULL,0xffffffffdab6fdf6ULL,0xffffffffdafffff6ULL,0xffffffffdafffff6ULL,0xffffffffdafffff6ULL,0x9edf7fbf9ab6fdf4ULL,0xb4bbe8241a344544ULL,0x0,0x14090000000ULL },
	{ 0x0,0x9540c5080c86600ULL,0xb4485693cdc82e02ULL,0x4130ba49250e4120ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4045029000a0160ULL,0xf9ecf7fefddeff63ULL,0xfeb63f7e6f4a7f62ULL,0xcdded4e05bda4760ULL,0xfbfedfdfedce7d73ULL,0xb6be6fff0f7eff63ULL,0xdfffbdeb6dfeff73ULL,0xdf9e7dfdffdeff63ULL,0xebffff7a1fdeff63ULL,0xfad6fedf9f4eff63ULL,0xfffbcfffdf8efd63ULL,0xffff766c4dfe7f62ULL,0xe39e3ff3cf7aff62ULL,0xf354ef3e1f7eaf70ULL,0xcfdb7effedfe6763ULL,0xeebedff97f1ef762ULL,0xfdfff7feef5eff73ULL,0xdffffff7dffeff63ULL,0xdfbfffff8fde7f73ULL,0xdfeaba0d4d087f62ULL,0xfbdbdddacdfaf763ULL,0xf6f7f56eaf4e7d63ULL,0xdf5b07e04f0e7f62ULL,0xff026bfd5eda6762ULL,0xf5dff7ba8f5e7773ULL,0x6f7fffdf9fde7761ULL,0x34fefeeb37ea7573ULL,0xbbbce76a1dca4d62ULL,0xf2ae3fff7fea7f63ULL,0xcdcfffffdffeff71ULL,0xeffffffbefceed63ULL,0x97bbffebcfde6563ULL,0xdfffb7fbff7eff63ULL,0xfffeffffefdeff73ULL,0xfbffff7ffffeff73ULL,0xffffffffcffeff73ULL,0xbfbfffdbdfeefd73ULL,0x977ef2edff7a6d62ULL,0xdbf67d724f0e7f73ULL,0xfbd6f76fddce7f63ULL,0xeff7bfddcf9ef773ULL,0xadbedf796c7c6f61ULL,0xfdfff6eaddceff72ULL,0xfbffff73efceff63ULL,0xef8fff7f4fe8ff73ULL,0xdfc7fe0f5f5eff73ULL,0x2ffffddbcf7eff63ULL,0xf4bcfff9cfce7771ULL,0xfd7f7f774ffeef63ULL,0xfbbe6f354f4cff73ULL,0xdfc7debceffe7f63ULL,0xff7fdedbbf6e7f73ULL,0x64f696e34f4c7562ULL,0xfdfdbf72cdfe6762ULL,0xfebd6df20d6eff73ULL,0xeff7f3a6435e7f63ULL,0xefffdfff0fde7f72ULL,0x2be5ffa4f6a7762ULL,0xf3fcfffa6dde7763ULL,0xd7ce3fb53dea7f73ULL,0xaa1afa2acd3a6760ULL,0x3dd6be58bb4a7d63ULL,0xced9dfbb2fea7561ULL,0xf17ffeffcfecef63ULL,0xff2eb5eb8f5eff73ULL,0x90dccb66071e0771ULL,0xffcfbfcb5fce7f62ULL,0x7fbf4ff62f7e7f63ULL,0xf5fffee9c9de7f62ULL,0xfbfa3de2fdde6f63ULL,0xfb9ff73e5dfe6f63ULL,0xc7d6bea34f8e5d63ULL,0xa79b9fd7af0cff63ULL,0x7cbd5efb6d4e5560ULL,0xf97fad7a275c5763ULL,0xfda42f67c8f87d61ULL,0x6146e72a05cef960ULL,0x321ed0820d0e4d73ULL,0x101609820d084620ULL,0x9bcdf66a4dda6763ULL,0xfebc8f741f5ef562ULL,0x67ccfb7f4f4a7760ULL,0x6a066e0501a80760ULL,0xedac5ffa3dce7763ULL,0x9cffff69876a7f63ULL,0x7dc635740d1e7761ULL,0xc1e1ef3e8fcc6763ULL,0x7fd637dd4f5af541ULL,0x9ffe6d624ddc7762ULL,0xbdf687f94d0c5f63ULL,0xc8e825018b40f601ULL,0xe3662002059a7b62ULL,0xc87676ba00ece573ULL,0x20b0c7010d8a4162ULL,0x88f35aeb09de2770ULL,0x8ed63d52c8e86451ULL,0xfd80e52449427a23ULL,0x947b28c4d4ee571ULL,0x61d99b54156a4562ULL,0x1495eee21afe4f61ULL,0x813e3562251e4760ULL,0x6e364f2c5baef741ULL,0x8192c918488a4542ULL,0x2b34de1e0c607743ULL,0x28000000004040ULL,0x8008200000000960ULL,0xf6942db1094a7f60ULL,0x20e0815002100ULL,0x35679ff81f3a4f62ULL,0x300858305020541ULL,0x516c34ec092c7761ULL,0x4000008002000400ULL,0x806c902085003440ULL,0x25965c0002c85c20ULL,0x37bc056644b87541ULL,0x4412c4e828405962ULL,0x500160000444022ULL,0x900001800044022ULL,0xe0bda00080a4d60ULL,0x80088c010040e820ULL,0xb15fe24a094c3f60ULL,0x100800000880000ULL,0x210012020402020ULL,0x0,0x804000000440ULL,0x3404046001664062ULL,0x100812000002060ULL,0x8000010000003000ULL,0x4000000000005000ULL,0x2552100140466121ULL,0x400200000000140ULL,0x23486c600d004b60ULL,0xfabe3f7fad5eff73ULL,0xd7fff7a7cf7e6f73ULL,0xa7527fdaac5eed60ULL,0x302069002404461ULL,0x9524046101ce4762ULL,0xc1006c3c0b0a4442ULL,0xe034920c0850f740ULL,0x400636300d080c20ULL,0x801187900f4c0460ULL,0x9258785209c87860ULL,0x100020001080000ULL,0x80800308087020ULL,0x884002124a006402ULL,0x2608cb11410c6721ULL,0x46d02200c406501ULL,0x38c406041087541ULL,0x6480a61002100d60ULL,0x4002380030100120ULL,0x8b514d500e3a6560ULL,0xfc77fefecfde7773ULL,0x90ef2768884eed70ULL,0xdfb73f7f9f5eff63ULL,0xdc5a1d0859084d61ULL,0xa94edf460d485573ULL,0xa47a5be00f0a5760ULL,0x90fcf9634d48f162ULL,0xfef2bf5069227161ULL,0xc7f7f7d61d8a6762ULL,0xeeffdb5b4fee6d63ULL,0xa79a57fb2d7a7f63ULL,0xd0e754c0010a4d61ULL,0xf6fab8728d4a6843ULL,0xebf6ffffddfe7773ULL,0x6ffe9f7f5fda7f63ULL,0xbbbeedf3efdef763ULL,0x7567f7737f6e7f72ULL,0xb667d42150a5942ULL,0x8302eb0f35185960ULL,0xccdab2736fa87462ULL,0x9634442011522750ULL,0x5df7dceacf58e762ULL,0xed7ffdf64fba7763ULL,0xfbbcfe6f1d5e5762ULL,0x57d3ff2fff2e7d62ULL,0x7b52979f9dde6760ULL,0xf436defd1b5e7f62ULL,0x936af77c0f7e7762ULL,0xffba6f7f6d7afd63ULL,0xceffea2add0a7762ULL,0x2b6edbcb4fd8f541ULL,0x901344e2804c2422ULL,0x50008400000ULL,0x4000008001000800ULL,0xd40260008406141ULL,0x2000800000000820ULL,0x2008200000460ULL,0x20100a0020004022ULL,0x0,0x81a0f1031d6c4d42ULL,0x4800060008080432ULL,0xc00005000000ULL,0x0,0x2050200004640ULL,0x8000001000080020ULL,0x0,0x0,0x80a8001000e20ULL,0x41103f8048207e73ULL,0xfa94e4280a4a4561ULL,0xf02db4924487d70ULL,0xe14696cacd4e6160ULL,0xf7b6f6f9cfda7f73ULL,0xfffffffbfdfeff62ULL,0xffbefffffffeff63ULL,0xffffffbffffeff73ULL,0xbfffffdffffe7f73ULL,0xffffffffdffeff73ULL,0xfffffffffffeff73ULL,0xfffffffffffeff73ULL,0xfffffffffffeff73ULL,0xffffffffcfeeff72ULL,0xdfbe8fff4f9e5f63ULL,0xdbfffffb5bfe6772ULL,0x45423f4114300542ULL },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x455766a0a9e7c61ULL,0xf5ef76ea4ffefda1ULL,0xdeb61ff2d7de3fa1ULL,0xc95adb2afaae3cc9ULL,0xefdfffdfdfde7fe1ULL,0xbfbd6ffbffdebea1ULL,0xd7f7f77fffbffde9ULL,0xff9e7ffd7fff7fa1ULL,0xfbf7ffef57df7fe1ULL,0x8fd7fff6dfbefda1ULL,0xfffffffaffbe3da1ULL,0xfcfffffbffbebda8ULL,0xff7ebf23d6bf2e29ULL,0xeb97f77e773f3c21ULL,0xfe8bfadc43b67d80ULL,0x6fbfcbb38aafbde0ULL,0xfdfffffefbffff81ULL,0xff7affb79fbffee9ULL,0xdfb7fb6f5fdefda1ULL,0xcd53fea87e967ea0ULL,0xbbfcdfdb4bde7ea8ULL,0x36b9f6e0dbae3f80ULL,0xdb7ff763c7bf7ea0ULL,0xffbbff7fee9f3da8ULL,0xf7dfffbfffde3da1ULL,0x225b5fd7ffd33e00ULL,0xbeffd7fbd7de7fe1ULL,0xf1b93ffa5fbf7ce1ULL,0xbfb7af7f7bbe3ea8ULL,0xebfffb7fdfff7da1ULL,0x6fdfffdb7fffff60ULL,0xbefbcffbdf9fffe9ULL,0xf7fffffbdfdf6fa1ULL,0xfffefffffffeffe9ULL,0xdbffffffffbfffa1ULL,0xffdf7ffff7bf7fe0ULL,0xbfbbfffbf7dfffa1ULL,0xfdfffef9fef77ea1ULL,0xefbe3d79febf3fa1ULL,0xdbfdff2fdfde7ca9ULL,0xef7efed9de9e7fa0ULL,0x2fdedf7b7ffe3fe8ULL,0xfdffb6fddfdfbd29ULL,0xd7feef725ffe7fe0ULL,0xf9a4ffff5bfe7ee9ULL,0xcfffff6e7b9f7ea9ULL,0xbffdffdbdeff7d81ULL,0xf4bff6efefbf7ea8ULL,0x27feff71b3b62ea9ULL,0xffde7fff5bdf7fe8ULL,0xffdfff3e5fb27ea1ULL,0x5b7fffdbd7beffa1ULL,0xa5ff7eebdeefafe1ULL,0xffdea57effbe3ce1ULL,0x9abe2d7be2fe6ea8ULL,0xcbe0f3bd57de7400ULL,0xadde9fddfab77f28ULL,0xbfba5ffb77dfff81ULL,0xd5f9b7bec79f3fa0ULL,0xf3e46f73cf9a2fa0ULL,0xa127e24c308e3ea0ULL,0xefebfbd3cedefca9ULL,0xffbdeff2fbbefd80ULL,0xfef7d7ed5fbe3fa1ULL,0xffbe4d63df9eae80ULL,0xfbd4af5ffeff7ca0ULL,0xecf3bfcf1f9f3d20ULL,0xf6efdff2c5de5fe0ULL,0xf57ff7eedade7ea8ULL,0xfbbe7df17ffe3fe1ULL,0xdbdfffbfd7be3ee0ULL,0xdbd3febb56be7cc0ULL,0xabfedfdbffffaca8ULL,0xfcfeffe9d7ff3c21ULL,0xdf6fef45d78e3da1ULL,0x42ba2b7e4fde2480ULL,0xc8c2cb18e39efd80ULL,0xc0561606461e2900ULL,0x945117c952be6d80ULL,0xe54ffe73dfdfefc0ULL,0xfebefff3dfbffee8ULL,0xe4ab2a24e182d21ULL,0xa01e5a1550962cc0ULL,0xd7be4ff3579e76a8ULL,0x7af977605ed62f88ULL,0x4f9e07f45b963ca0ULL,0x89c2fb7b4f9d7fa0ULL,0x21d3164802852d60ULL,0x3bbeff59defe3da0ULL,0xdd57b4addaae3da8ULL,0x2cb624725a9e3da0ULL,0x8bd6c36c5d9e6f80ULL,0xfec2121a78b62449ULL,0x89441a102ce1980ULL,0x74fe56f04bfffc80ULL,0xc7aa6f6219bf7c20ULL,0xce8c07338a9e3c80ULL,0xcc8bf1a057833c01ULL,0xab1fd7584eb62ca0ULL,0x96f256e9cbbe3c20ULL,0x2d6f35607a8e1d20ULL,0xdebe7ec45fde7ca0ULL,0xa91d6a0429a7c09ULL,0x2d9e4b000216ac80ULL,0x440010040000400ULL,0x918c8f22482c3481ULL,0x12a40050c0920c40ULL,0xe42e02504a0c2480ULL,0x965e08c68c2c20ULL,0x841817f0561a0c20ULL,0x5124f44856b39c00ULL,0x212023402103400ULL,0xcbc67f3f7f9e7d00ULL,0x644120040043c00ULL,0x24280dc1529e6420ULL,0x10220b1020c1400ULL,0x10041102040480ULL,0x100020000002001ULL,0x6867aa46ce3d2c88ULL,0x200406c004921040ULL,0x1a4040882002000ULL,0x192a454010863e00ULL,0x0,0x4ec2021082182000ULL,0x2d088b9b560c3ca0ULL,0x10806000040980ULL,0x40040000021000ULL,0x0,0x0,0x0,0x10001000024000ULL,0x4d44f66217ffdfe8ULL,0xffff7fffdfbeffe9ULL,0xfffffffefffebfe9ULL,0xeffe3fdff6bfbea0ULL,0x480400000120000ULL,0x819884020a003c00ULL,0x400004000c00000ULL,0x88d09b7156923c00ULL,0x4346223056b2ec20ULL,0x4808c962128e40a0ULL,0x54743660c69e3e80ULL,0x1200240040124000ULL,0x8180a00200c02900ULL,0x2002308818061400ULL,0x480541846162000ULL,0x3006020812000d00ULL,0x20850042002400ULL,0xd082932212ee2c80ULL,0x88f086053a900600ULL,0x8e98015a471e6ca0ULL,0xf8f4beec77be7ea8ULL,0xd088040081860820ULL,0xc332bf6d1bbe7c00ULL,0xac4fe32946d62c20ULL,0x21365e0142963c20ULL,0xce6d7c11abc1d20ULL,0xb63aa7fc6a976e08ULL,0xf73627d66a8e3fe1ULL,0xf9c1fb0ec39e3f28ULL,0x2b9f0bcffbbe2c21ULL,0xa121e88548604a0ULL,0x8100224c02128420ULL,0x681205804a8c0400ULL,0x8bb2fb27bbff3d00ULL,0xeeff7ecf57bffde9ULL,0x3fbeeffbdbbefee1ULL,0xffff36ffdebfff21ULL,0x6d3e1dea5e972ca0ULL,0x252030612863d00ULL,0x86563340ca9c2880ULL,0xe3be89e38f9d3c28ULL,0x51e7d6ed569e2ee1ULL,0xa43a3575529f7e01ULL,0xfb8ffb7ffb9efca8ULL,0xedfef6ffdffe3ea9ULL,0xa36edfdb45be7e80ULL,0xb4f55febdebe7fe0ULL,0xfdd627f85bdaae60ULL,0xdd966f7d73f67ea0ULL,0xdddfffbaf7ffbde9ULL,0x69dfff4f729f7e80ULL,0x1089d5eb608c2da0ULL,0x0,0x0,0x800c32803322c00ULL,0x20110800000ULL,0x422000002082400ULL,0x9021022002002801ULL,0x124246049a812800ULL,0xaba5fbed57946fa8ULL,0x140120002082400ULL,0x3000000000200400ULL,0x0,0x8204002001000420ULL,0xc000c0000ULL,0x0,0x8002000200002800ULL,0x2000020000000408ULL,0x6f4ef7f19bb66ca8ULL,0xdb95fd3f57982d29ULL,0xced7f7b7dfcfec60ULL,0x2948011a05140480ULL,0xfeffffffdffffee1ULL,0x9fbffffafffe7f88ULL,0xffffffffffffffe9ULL,0xffffffffffffffe8ULL,0xffffffdfffffffe8ULL,0xffffffffffffffe9ULL,0xffffffffffbfffe9ULL,0xffffffffffffffe9ULL,0xffffffffffff7fe9ULL,0xefffffff7fdefea1ULL,0xb6ffffefdffe7fe0ULL,0xfdefdf6f7fdebee8ULL,0xfff67dfe77fe7d81ULL },
	{ 0x8ab58326d7142f17ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbefffefff7bc677eULL,0xffffbffff7bc6fffULL,0x9eb0277ce71422dfULL,0xddc7f603b390276eULL,0x2bdecc4af2a42b6eULL,0xbff7dff3d39c6fcfULL,0xfb7da77c921e6f5eULL,0xfffe7d61f7b66fffULL,0xebffff7ff7b66f4fULL,0xafd71f5fb714635fULL,0xbffffffff7be6b7fULL,0xf5fff6f9f7866bffULL,0x7bbe7fe2d6bc236fULL,0xfbafffffd7be6f7fULL,0xccdbffdff6bc2b76ULL,0xefbfdffbf7be6fdeULL,0xfddffffef6b66f7fULL,0xf7fefd67f6b66b7fULL,0xfbbdffefb69c635eULL,0xe7de7ebfd7b00b6eULL,0xbfdfdfdfd696637fULL,0xd4fff6eef79e2bd6ULL,0xfdeb5f60d7822357ULL,0xffb67fffd7be6bffULL,0xcdc7fff7f7942b7fULL,0xeeffdcdbd63c2f4eULL,0x773b7fe7f7a82b7fULL,0xf36bafe09294634fULL,0xdf9737fff7b8677dULL,0xefc7fff9d7b46ffeULL,0x3fffdf7bd7bc67f6ULL,0xbebb7ffbd7b4675fULL,0xfffdfffbd7be677fULL,0xffbe7ffff7bc6fffULL,0xebfdff7ef7b46f77ULL,0xfff73f5ff7bc275fULL,0xdffffff7f7346ffeULL,0xf7fefe7ff6bc6fffULL,0xffbefff3f7be6fffULL,0xf9d9ffffd7b62b7fULL,0xef4ff6fff7b627ffULL,0xb7bfcfdbd7bc2f5eULL,0x76ffdefed7ba67feULL,0x9fdffff2d7ba6b7fULL,0xfbffbf3ff7bc2feeULL,0xdfd6ff3ad794675fULL,0xfffeffdaf6946fcfULL,0xf6fff7fff7b62f7eULL,0xfbf5f779f7946f97ULL,0xff34ff72f79c2f56ULL,0xebd7ff9bf6b0234fULL,0xfffedfdbd7b46fdfULL,0xb5b75f6b938a2fe3ULL,0xf9dff77af79427deULL,0xd8be2f7ef7346b4fULL,0xe777bba3d616673fULL,0xaefffedbd7906f7eULL,0xfffbfffbf7bc6ff7ULL,0xd7f5ff7af7bc6b7fULL,0xffde7f76f7866337ULL,0xe8f6f77bf794236eULL,0x697f3ff2d7b86bdfULL,0xbffecff3f392675eULL,0xd7fefaf9d3342f5dULL,0x7bbc4e7092a467eeULL,0xcb90773cf6166b7fULL,0xefffbfdeb79e6fffULL,0xaf7f5ff3d7946f65ULL,0x317f37fab296232eULL,0x6dfaff71d7962f6eULL,0xff967f7ff7bc275eULL,0xcfebfe93f7162b43ULL,0x3bfee7d3d6b26bd1ULL,0x76f6f6ffd7b467efULL,0xdfffff7bf7be6fffULL,0x9fa8f76fd6342773ULL,0x28c7eb96c780611fULL,0xe8d0b71f928827ccULL,0x1ba5362d3302216ULL,0xdae53543f7302e1eULL,0xba973f61e72c2347ULL,0xeed673dd97166f4eULL,0x88eebf6fb3842bcfULL,0x97baff7bb09c2bdfULL,0x91ee774af4b86347ULL,0xcafe2562d7aa4b4eULL,0xc188e26fd0a82b5aULL,0x617712f6d3b0634eULL,0xedfafd72d7946776ULL,0xd1b6dc6ad62423feULL,0xe3164d62d5106374ULL,0xea906522c4884a42ULL,0xc882fffff7b40366ULL,0x2f9dcbf4b2982f5bULL,0x9df656adf39e431eULL,0x17061df0d602016eULL,0xcdbeff7fd7942fcfULL,0x8ac2330392202941ULL,0xbbcacfdbf7342f7fULL,0x86ff57e7f68a275cULL,0x933e77f8d79e2366ULL,0x5fbe2f7fd7bc6ffeULL,0xf3c7f61bd6322759ULL,0x301e0000a00048ULL,0xa00053a83040382ULL,0xfcccf67182b6275dULL,0xc282354c80042e10ULL,0x8000d00082004104ULL,0x2c206808f4042101ULL,0x828a0d2aa0102b04ULL,0x984026608210428cULL,0x9b04782192a00056ULL,0xa004200c4102400ULL,0x48020100100300ULL,0x59c4c0d54b003d2ULL,0x5e75c26a94902a48ULL,0x1ba43600e6002b01ULL,0x5800612010000160ULL,0x4600040002202100ULL,0x61b88d20d480214cULL,0x64010800342ULL,0x4a98b51293902227ULL,0x4100080002000ULL,0xc10a0010002a20ULL,0x891a8fdbf7b60b7eULL,0xc21444e4b630416cULL,0x800e024287002044ULL,0x2a00400c4100010ULL,0x88109000a0100a00ULL,0xb40834050082215ULL,0x983e0d629200270fULL,0x754cfa63f72c234eULL,0xfebeeffff7b46b7eULL,0xeffffffff7b66fffULL,0xffdf7ffbf7be6fffULL,0x1491640891206100ULL,0x5008046246002004ULL,0xba060da491a02172ULL,0xe3d7e369f6842376ULL,0x5ce61e1496a227e4ULL,0x1011d4d3920c634dULL,0x10c012689200012cULL,0x40000d2090002321ULL,0x8004202002002005ULL,0xc08228210202a00ULL,0xc4534a2082842647ULL,0x500322d6910c0316ULL,0x88a4028220204cULL,0xc1808216820a0a14ULL,0x5b42bb36c6142ad6ULL,0x2beced683802343ULL,0xf09ee2fe96962b64ULL,0x8121841ae4302204ULL,0x59002d3290942642ULL,0x48c2ef7eb786035eULL,0x219180992802107ULL,0x94475763d780434eULL,0xd0daf7fef79443e6ULL,0x74802d62d6900346ULL,0xc8f0e00682106304ULL,0x881a209000480cULL,0x0,0x8102282100200ULL,0x1000004001000002ULL,0x8002100890000042ULL,0x8de5f41f3244372ULL,0x9bafeff3d7942f76ULL,0xfc7bd768c7b42b4eULL,0x5bbc7d32f79c63dfULL,0xf394ef36d714675eULL,0xca5e3287d2b42756ULL,0x8ff5d7f2f6942b7eULL,0x51f43eebd7202696ULL,0xafdaff7392b46b7fULL,0xdb9cfb7ef2ac2f7dULL,0x8fdede9fd69c234eULL,0xbbdedfdfd68e676fULL,0xf6fff6ffd7b26fefULL,0x31be7678f2042b4eULL,0xdbb5b7f7f3942ffeULL,0xfdffff7ed79e6ffeULL,0xeefcdcdff6b46b7eULL,0xbc7bd763d7b467ceULL,0xa1b48228922c2a45ULL,0x40220442a0046648ULL,0x10844000000ULL,0x165a5000202a0dULL,0x800205008000001cULL,0x203c3882002288ULL,0x76343d7cf2942b16ULL,0x8000c00090040040ULL,0x101040000010ULL,0xb010030080046200ULL,0x0,0x18002000004ULL,0x0,0x4040004180002008ULL,0x828000230202000ULL,0x0,0x10ab420d2026210ULL,0x13841210d3002006ULL,0xcc523e43f6a40164ULL,0xeaa287cbd784255eULL,0xf4bffffff7b427ffULL,0xdffffff3d7be6b7fULL,0xffff7f7ff7bc2fffULL,0xedf7fffff7b42f7fULL,0xfbdfdf5ff7be2fffULL,0x37ffffedf736277fULL,0xfff9f77ff79c6f7eULL,0xfeffbffff7be6fffULL,0xfffffffff7b66fffULL,0xfffffffff7be6fffULL,0xdebbffebf7b46f7eULL,0xfffffffbf6b66f7fULL,0xfffefffff7b66f7fULL },
	{ 0xcbbfff7ffbbe77ffULL,0x0,0x0,0x2808000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000c000ULL,0x34bfdfeedbef73bbULL,0xfbffffefcbff77bfULL,0xfef6bffefbf6dfffULL,0xcfd7f7befbfc727aULL,0x2ffffff9f9be73eeULL,0xfdbadffbe391fbffULL,0xfb3edd7efbbc53feULL,0xfbf6fff5fb985bf6ULL,0xebf7eb3bfbfef76dULL,0x6cdb7f7bfb305feeULL,0xbfda87f2dbb453ffULL,0xddf677fffb347ffeULL,0x4fbefff3fb3e57feULL,0xd9b5ff76cbb07b6bULL,0xcbefbfdbc9f87b7fULL,0xffbf5ffffbfb7ffeULL,0x1dff96fbe1865b7fULL,0xcaaefd61cbf0532bULL,0xdbbfef77dbf957ebULL,0x4fdefabbfbb05367ULL,0xffdeffdefbfb73fbULL,0xd5fafe70d998d33fULL,0xdcbf7d6afbddf276ULL,0xefbe3e7eebf4fb74ULL,0xcccaffbaf93dd3eeULL,0xa7ffff59fbbb53fbULL,0xb63f7773d9fcdffdULL,0xfdffffeaebb5c3feULL,0xfabf3f74dbb073feULL,0xeffef7bfebbedfbfULL,0x3fdf9fdfebf37bbfULL,0xf7bb7fffdbffffffULL,0xfdfdf6bbfbd27fbfULL,0xffb67dffdbb4fb7dULL,0xcbdff72edbfe77bcULL,0xefdfbffbebfcd3feULL,0xffbecd73dbfad37bULL,0xd87ff7fec9fc5ffcULL,0x7ffe3ffbfbfffbffULL,0xfbffff7febfef7b9ULL,0x5fcfb6bfebf1d3bfULL,0xb5bfdffffbbef7ebULL,0xddfefefbfbfcf7fbULL,0xbffefff3ebbc7bbdULL,0xffb5ff6ffbfefbffULL,0xfecff7bfebfcd37eULL,0xafffdfdbdbbdf7faULL,0xfcf6fefffbbbf3ffULL,0xf9efffeffbbffb7fULL,0xfbbe7dffdbbc77feULL,0xfde7ff3ffbe0ff7aULL,0xebffffdffbf6ff3eULL,0xfdff7ffffbfe77ffULL,0xddfcfffffbfcdfb9ULL,0xdebc29f78b9877beULL,0xef77fbfeebb7f3fbULL,0xfbffdfdffbffff3bULL,0xffbffffbcbffffffULL,0xd9bcf67bfbf3336eULL,0xda3635fdf9ca77bcULL,0x8bffff7eebedf77eULL,0xeff7bf7ffbfed33aULL,0xbffecffacbfdf7beULL,0xf6f7fffadba8df2eULL,0x7adc7ff9dbae73fcULL,0x9be6fb37fbf1f37fULL,0xffdf3e19dbfef3afULL,0xaf7edf57dbf37fbfULL,0xf4f77ee8dbde7b3eULL,0xdfbeff73ebf4ffffULL,0xfb9efffeebf45ffcULL,0x8f4bfbbaf9b8d7b9ULL,0x5719dfdbeba17b3bULL,0x76fff7fafbff77bfULL,0xbfefffebf9f9dfffULL,0xea9433f4c9805fbcULL,0x83e7ffabd3b07b0eULL,0x25eedbd9cbb47338ULL,0x35a559af5b11d34aULL,0xf47cb6528bd2d328ULL,0xde977dfddbfc433fULL,0xdddbd6b5ebcc533eULL,0x6fdf9fd5e3b27b0aULL,0x97334c5a6ac84721ULL,0xf72df72edbf45bbcULL,0xa212ff648ba8c72eULL,0x8bfdf37ee3b053fcULL,0x7c670e50e99066baULL,0x8b888f4bc3b243eaULL,0x4196c861ca205208ULL,0xe7f63ff0fbbef33aULL,0x1b06e73c8970f3aeULL,0x887e1653c943f20eULL,0xdebeddebebbcdbffULL,0x6cd767e3cb9c57bdULL,0xf38ea470c99e73beULL,0xef95fb7f7bd2533bULL,0xf4a939ac891133fULL,0x2947d993fbb4636cULL,0x76801460e0301320ULL,0xc97c6d23db1f7b6aULL,0xff64fff7dba85bbfULL,0x947afabf8e4d529ULL,0x2a1448d4d888d112ULL,0x240019e1d0a41229ULL,0x91785f70898a53aaULL,0x9c8683708b88517aULL,0x8080c96c81b04b80ULL,0x637810068112498aULL,0x30000b8001004170ULL,0x104c3001518041a1ULL,0x814040000a000102ULL,0x20080100000ULL,0x8e8081181a10323ULL,0xe53d05608b8071f0ULL,0x8010005001100102ULL,0x10c06c800806130ULL,0x5150652c83a20a2cULL,0x121032028100130aULL,0x2904c04001100308ULL,0x400002040001000ULL,0x19ad18051120268ULL,0x3000020002a00200ULL,0x41000040001000ULL,0x0,0x80140801800018ULL,0x808002281105000ULL,0x0,0x508c0088a001208ULL,0x1400000800001ULL,0x100a480008900210ULL,0x682470200310528cULL,0xdab63fd7db91533eULL,0xc95e750b0bc10623ULL,0xa0cac0209f00320ULL,0x957f4da2dbbe53b4ULL,0x52202d1061136202ULL,0x9abe2d56f9e07b1aULL,0x8accf23ad1b441b3ULL,0x4ec221348aa0fa14ULL,0x8080d0800050062ULL,0x0,0x482040508a001120ULL,0x8300c3126b1c4122ULL,0x600300001046020ULL,0x8240028080030aULL,0x949056426081130fULL,0x900205600b983228ULL,0x480005051020000ULL,0x4142600a09400120ULL,0xa5ba9d508b8a1130ULL,0xf5b456e88b335722ULL,0x907d2750f080c3b4ULL,0xb12657a8d110538eULL,0xddc6fe96dbbc6bfcULL,0x398d8528bf0732cULL,0xc4f2176e89d863eeULL,0x53fefef48ba0532eULL,0x8e262f6f5a15536aULL,0xcbc5e603eae8538cULL,0x2292300183804112ULL,0x4a944400bc00228ULL,0x105c260408085382ULL,0x42204080000064ULL,0x153ca0081124124ULL,0x4a461b55e0101b7aULL,0x33b497a259f11363ULL,0xd03ab27561b5033aULL,0xeffcef73dbfd577eULL,0xebffff7ffbbef7ffULL,0xdefb7be7fb1453aaULL,0xcebedffbd3b4733fULL,0xf537fee9fbb453efULL,0x6aaefc73cbb25b2bULL,0xfbffdbfecbc973ffULL,0xffefff3bfbf6d76bULL,0xbbdfdf5febf673baULL,0xf6f7fefffbb2fffeULL,0x136f7d60f924dbf8ULL,0x9ed0ef75cba873bfULL,0xf7f7ff3dfbd3ebfeULL,0xab548c0008bc0330ULL,0x11ba1907e938439cULL,0x400516480b980008ULL,0x9210075cc0340322ULL,0x2802800100320ULL,0xb555d40cbb05728ULL,0x10043930102032aULL,0x80990d408b11030eULL,0x5ef46df18bdc67fcULL,0x20c00280001208ULL,0x4002101001000160ULL,0x94446061820138ULL,0x5000202580b02302ULL,0x4002000800001ULL,0x8010202040000100ULL,0x1800000004080ULL,0x0,0x22088008c00001ULL,0x10085002004108ULL,0x1214000000000520ULL,0xc967a21b8be05328ULL,0xb26055c89c051a1ULL,0xb6f6feedebfb7bbfULL,0xfd7f7f7bfbec7bfbULL,0xff9e3f7fcbfd737fULL,0xfdd0ffb0cbfa7b6fULL,0x6ebfdb95ebfe73b9ULL,0x777bdfe7fbfc73bfULL,0x2966a522cb807779ULL,0xfef67f7eebb5f77fULL,0xffffffbffbffffffULL,0xffffffffebffffffULL,0xffffffffebfff7ffULL,0xfd7cfff98bfd737eULL,0xff3e7fffebfbf3ffULL },
	{ 0xeb9fffe2cbc7fb7eULL,0xdedb3a9bc3c7d372ULL,0x1868b3302028342ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa84cf5a83c4e30eULL,0xf9b80d39e1cde30aULL,0xf9ecff73cbc5c33aULL,0xfffaaffd8bdff37aULL,0xefdeeff68bc7d37aULL,0xaa5e3ddf83cdf3aaULL,0xdfbb6ff7e9dfeb3aULL,0xdf7db7fbcbddfbbaULL,0xfffe3ff9ebd7f0feULL,0xebf7fbffebcffbfaULL,0xdedf5e77e3cee37eULL,0x9b9b47c2eadee31aULL,0xf6fff7fecbdffb3aULL,0x6ffeff7bcbdffb7eULL,0xf3b7f7bf43d7c37aULL,0xfae3b73b8355d3faULL,0xbe9eddb783cfe3eeULL,0x34237c65a1d6d31aULL,0xffbe3df1ebd4f3faULL,0xca6ae77f81ddd13aULL,0xcfeffa3709c7c35aULL,0xb377dfdf83c7f17aULL,0x3497766d89d6e36aULL,0x99c6a763c2d3d85aULL,0xcfa6ffffebd7d37aULL,0xefcfdfbfe3cff37eULL,0xeb7ffdddab5ee3eaULL,0xb4e37feceb85fb6eULL,0xfbfff7facbcfd3faULL,0xfebe3fff8bd7d3eaULL,0xebe6fb7681d5f96eULL,0xc85fdebf8bd4c3feULL,0xbfb7dfffcbd5d31eULL,0xfdfeffefebdff37eULL,0xfbd6ffe6c3d7e3faULL,0xe9f7db2ee10ff37aULL,0xeefabf7fcbdfc3faULL,0xf7ffcffaeb55d36aULL,0xd5fef77fcbddf3feULL,0xfbbe7deaebd7e37eULL,0xcbfdfffdebc7d37eULL,0xcecbbb7da9d5e3faULL,0xff77cf7983cffb7aULL,0x7dfeffe8abc5e36aULL,0xdbfefff2ebd5d3faULL,0xffbfffffebddfbbaULL,0x6fd7ff97cbc7f37aULL,0xaffeffdbcbdff3feULL,0xfffff6fcebc7fbfeULL,0xffeffffbebddd37eULL,0xdebc6fffebdffb3aULL,0xc7cfcb3bcbc5f3eaULL,0xbf5fdfdf8bdfeb7aULL,0xfdfbdffbebdff3feULL,0xdffffffaebd7fb7aULL,0xde9e3ffeeb57f2eaULL,0x8fcbf73bcbd5c97aULL,0xdfff3fdbe3ddd3beULL,0xbffbeffbcbd7fb6eULL,0xdbbf97e3c3c5c36eULL,0x78f6bdf0abccd33aULL,0xebe7ffabebc5db7aULL,0xebdfbfffebddeb7aULL,0x9ebf5fdd01d6c36aULL,0xf5d7d7f9c3ddcbfeULL,0x63aebf6281cfd33aULL,0xfbd4e77ec3d5c36eULL,0xfe637f8b8b55f32eULL,0xaabfeff981d7f3feULL,0xa4bb3af08287b312ULL,0xb3bfbf70c9cfc32aULL,0xc8a0fffccb4ee146ULL,0xe907ee270845837aULL,0x99bf9fd6c3d7c26aULL,0x74f6feebebcdf37aULL,0xfffefff9c3cff36aULL,0xddbd3f77abc5cb3aULL,0x2d8fdf028257935aULL,0x25b17d0c1c5934aULL,0xbcff7ffe8bd5cbfeULL,0xdd7df4714145c13aULL,0xc90604b081d4c220ULL,0xedfbf7b9e954c9faULL,0x2ffeff7fabd7f36aULL,0x76ba4feba3d6fb5aULL,0xfdfffef383d5937eULL,0x7884796281d6d24aULL,0xcbf6f73789cdc32aULL,0x64ab58408284d970ULL,0x938805f3c3c1d32eULL,0x95b5fe5a81c7e3feULL,0x6f8e2e31c305c12eULL,0xbb84c2068bd8d166ULL,0xcc58b3828955c30aULL,0x1e9def5162d5c33aULL,0xb4fdd2df63c4c360ULL,0x5ffa3f718ad4eb72ULL,0xab88476e83c6bb22ULL,0x4fde3f26c8d7d37aULL,0x22f693d3c3d4c3c2ULL,0x342854e08344c042ULL,0xc96875068057f362ULL,0x8102724c83c4530aULL,0xc7deff1bc35ce2aaULL,0xaa044d802b0a400aULL,0x993f1c22e344d14aULL,0x92e1afd5cbd4d22aULL,0x12940b4c08dcc022ULL,0xc596e7a801d6c3f2ULL,0x224000c60244000aULL,0xc3125d2288c4101aULL,0xd2be2d1281c7db76ULL,0xd84028460000c268ULL,0x5bc678212b54c26aULL,0x2f0448000005c112ULL,0x398051a43c8c002ULL,0x0,0x808050000008028ULL,0x200000008100ULL,0x8140000240808000ULL,0x60046108102520aULL,0x2240166003c48182ULL,0x5baebc3101d4c13aULL,0x8104100000818000ULL,0x48180000508008ULL,0x11000020a008002ULL,0x9421462481c44002ULL,0x6000004008ULL,0x8c2c0081804000ULL,0x80100002008000ULL,0x10114080002000ULL,0x0,0x19306e428300c906ULL,0xd68e2f6c8256ea62ULL,0xd01081288081a004ULL,0xa08280009448382ULL,0xc735eff588d5b16aULL,0xa4241d288284f302ULL,0x9614774c0a4cc31aULL,0xc9b4e66ae3c5d10aULL,0x40861e2008c4822aULL,0x0,0x41822608184c200ULL,0x40202400c000a000ULL,0x40031881005082ULL,0x8643b0080087c002ULL,0x280031100440202ULL,0x50350220200c4012ULL,0x17807d408a0dd22aULL,0x4480a2402800c02aULL,0xc316162aab546008ULL,0x5dd6c28843a03aULL,0xa2137768a2c5d11aULL,0x89298862408dc200ULL,0x4ba2272381c5d30aULL,0xffd6fb3ec9d7d36eULL,0xac5c8598b05926aULL,0x1ffcffcaab54d2caULL,0x920424608005b102ULL,0x5c282deecbc5d122ULL,0xe8c2ea0e08d4c10aULL,0x70427b198b84fa2aULL,0x920005a081828306ULL,0x40c0b73388d3c30eULL,0x202344001848020ULL,0xa106aa028150600aULL,0x48422c0002c0802aULL,0x918a4c738082c82aULL,0xb42ca372a3dcc13eULL,0x38b63f31c3d1d3eaULL,0xfb4ec16fc9dde22aULL,0xf95aff5aa285d36eULL,0x7fbec7f9e95ce36eULL,0xb4b8def6cbc7fa6aULL,0xffb6ddf3abdfdafaULL,0xdbbc57ff8bd6e36aULL,0xdfd3be1aabc5d326ULL,0xaf8acdd9c3c5d3beULL,0xd4bffefdc1cdd3eaULL,0xe1efdb6a83d4f84eULL,0xcb9eae728b47f3caULL,0xcfdffffbebd7f36aULL,0xee5f124bcbcfe202ULL,0x10402100848002ULL,0xb300a66003468272ULL,0xceb2414a69c5c22aULL,0x88015230005c03aULL,0x280e38808005a30aULL,0x60266c840468302ULL,0x4020060040008200ULL,0x91802c4009d40102ULL,0x81a000370a868102ULL,0x480000000500000ULL,0x0,0x2020000008040ULL,0x108000000004008ULL,0x4000c02002808002ULL,0x8000400000040002ULL,0x0,0x6000008020ULL,0x0,0x50842402c084a100ULL,0x1800000008020ULL,0x2ab803c14080c10aULL,0x2cf6d6e7a3d5ca0aULL,0xf73e677263d5c37aULL,0xdbbe2a76a8d0d96aULL,0x8044031b81508324ULL,0xebcecdd481d4f3daULL,0x9cbbf7efabc7d36aULL,0xf3fcffef8bddfb7aULL,0xf6b62f564bd4c26aULL,0xebfffbbbabddcb3aULL,0xabdfffdbcbc7db6aULL,0x16b20fa68382f36cULL,0xf17df5fdc3c7e35aULL,0x4d7eff73abd5cbdaULL },
	{ 0xdbffffffcecbfffcULL,0x6fdffefbcedb7bb4ULL,0xffffffffcfdbfff6ULL,0xf77776ffcfd3ffb6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000e1805006a80ULL,0xbabf5f538fcb7aaaULL,0x343676c0cdc22e70ULL,0xffec7f67cfcb7fb4ULL,0xfeb63ffdcf9b7ebcULL,0xebd7f2a7cddafe2aULL,0xeedfdefb4fcbffbcULL,0xf6d87f6b4dcb7afcULL,0xfdfff77fcfcbfffcULL,0xfffeffffcfdbfff8ULL,0xebdfe76fcfda7fb6ULL,0xc84b3e5f4ecb7a34ULL,0x8a1e7b9689d35eb0ULL,0xfbf7fff9cecbfba4ULL,0x2ffe64724fdb7e20ULL,0xdbf0ef2e4fdb7ba4ULL,0xedce12abcfdbbe74ULL,0x8fbecd6e8fcb7bb8ULL,0xd1b3dfeac7cb3eacULL,0x5fbc3fe340d33ed4ULL,0xcf1e6b764fdb3fa2ULL,0xdf57bebbcfdbfab4ULL,0xafffddd80f9bbfc2ULL,0xfefff6ffcfdbff6aULL,0xdb3ebff64edb7e60ULL,0xfbd5ffb34dd33ee6ULL,0xafd7ff2fcfdb7e30ULL,0xae7fffdfcfcb7e3cULL,0xf4bf67ebcfdbba60ULL,0xbbedef7ccfd3fe2cULL,0xdfbe7fffc7c3fef6ULL,0xcfffffffcfdb7fbeULL,0xecfffe9d47dbffb8ULL,0x3bbf4ffbcdc9feb8ULL,0xfdfd27ebcfdb7e76ULL,0xf5fe3ff2cecb7fbcULL,0xebcff7fecfdbbef6ULL,0xeffffe7fcfcbfffcULL,0xff1bdfffcfd3ff30ULL,0xf5fefefecedb7e96ULL,0x5bfe37f34fcbfeeeULL,0xebdfef6bcddbfeb8ULL,0xd9effaffcec27ef4ULL,0xffbe8ffb4fdb7e62ULL,0xdd7f7feecfdbffbaULL,0xfbba3f77cfdb3e3aULL,0xffffffff4fdbfffeULL,0xcff7febfcfcbfeb6ULL,0xffffdfdfcedbfef4ULL,0xfcfdffefcfdb7ff6ULL,0xff7fff7fcfd3fffeULL,0xf7bfffff4fdb7fb2ULL,0xefefffba4fdb7ff6ULL,0xebdbdfdf4fdbfeaaULL,0xf7ffffffcfcbfe7cULL,0xf9ffbf7b4fcbfebeULL,0xfebebf7a4fdbfeb4ULL,0xcfdef7b8cfd3bf84ULL,0xfd6fedec6c97eaaULL,0xe7bb6ffb4fdbfebaULL,0xf6fff7eb4fcb7f86ULL,0xffde7ff64fdb7eb4ULL,0xdbf7ff3b4dd3bfb0ULL,0xeeef7ffb4eda7f7aULL,0xbfbdeffb4fdb7ef0ULL,0xf7ff73bf4cdbfef6ULL,0x5fbe0d7acfd3bbbcULL,0x9bfff76fcfd37eb2ULL,0xefdf7f7fccdb7ef8ULL,0x8f9fdfc2cfdb3e32ULL,0xe8ff6ecacdcbebaaULL,0xdeb42d30cecb6aaeULL,0xffbef7ef4fc3fffcULL,0x4bc3ae974cd27ea2ULL,0x890edd5705d37be0ULL,0xd4f676c4cfdb3b22ULL,0x5deeff774fcbfeb0ULL,0xdaa62fd3cdc2bef8ULL,0xabd7ef2eced3feb0ULL,0x2ebc5fc64ec1fa2cULL,0x4667d7f6cfc3ae84ULL,0xff7dfffb4fdbfe28ULL,0xaebeaf6d4fc37be4ULL,0xe7d3ee7ccfc2ee08ULL,0xafdc6f7fcdcbfeb6ULL,0xbdfedfffcfc3ffa0ULL,0xcd7caeda4dcb7bb0ULL,0xb716efe1cecb7f34ULL,0xe917f2834fda3ab4ULL,0xea4b7a5d4ec936bcULL,0x604e4ca84c27820ULL,0xb5d0066c4ec32244ULL,0x602635f8c49bde64ULL,0xb9ac36e4ecbfa64ULL,0xe561a0f85ca3ec0ULL,0xf39f4ed1c4caaf70ULL,0x517f76e94fc22eb0ULL,0x593abd73479bfab0ULL,0xffa0ef654cd3b6e4ULL,0x9fd7fe0a4edb7ee0ULL,0xa2dcda5a4e5a3eecULL,0x341416e1061236a0ULL,0x89bb6f7e4fdb3ebcULL,0xf63025c04ccb320cULL,0x82020a3046012e10ULL,0xa3d8df59ccc27a04ULL,0x34c054214c083620ULL,0xe0ccb630c5c07e22ULL,0x840087c148423320ULL,0x6144c00000025120ULL,0x686659d84e8abe04ULL,0xbe0a1f1b47887e20ULL,0x640651b106523b20ULL,0x188c49c00c493a80ULL,0x8151a13204c12210ULL,0x68421e10c0807030ULL,0x9b10ed88c4c926a0ULL,0x8000040884802020ULL,0x402004008081000ULL,0xfbd4eb2e4ed87eb4ULL,0x800930006002000ULL,0x8008050180921200ULL,0x66212c00cc27b00ULL,0x2b8e77e1c6d37260ULL,0x8210217004c06a10ULL,0x0,0x401204410208ULL,0x2404102008420200ULL,0x0,0x0,0x4020040800000ULL,0x103600c0820040ULL,0x0,0x0,0xfeb6bf7ccfda7fa4ULL,0x830a42002210ULL,0x5188e408c832202ULL,0xf6ff7ee84edbf7f6ULL,0xd1bcf7eb4ec3fea4ULL,0xb5102e0084c22b00ULL,0x88402a0504020620ULL,0x61480040802024ULL,0x2044a444820024ULL,0x14e0240207022aa0ULL,0x40645041081810ULL,0xc304a91e4ec21a20ULL,0x641f71243022e24ULL,0x4430004402200ULL,0x10406000103002ULL,0xc03aef7747c36aa2ULL,0xc280e64a0ed2b3a8ULL,0xeffff2374cdb7ea0ULL,0x10044204487000ULL,0xf4b3ddebcfcbfe90ULL,0x170257ac1815a14ULL,0x8c51724004332a2ULL,0x7883dfafcecab220ULL,0x7bfe53d9479a3efcULL,0x84b755c888c316e0ULL,0xd108856246cbb2e0ULL,0xe8d62660c5d27876ULL,0xc357f7384593ffbeULL,0x8058ca430c027628ULL,0x1982456a46c23fa0ULL,0x885635080e022700ULL,0x84063040c08804ULL,0x1d02044002010ULL,0xf402341c8c22640ULL,0xd3eea1743c9ba10ULL,0x44000000600ULL,0x2ade75f04dc37aa0ULL,0x9383c92444d0be04ULL,0x865fa10084b73f0ULL,0x8c9ddf784ec27624ULL,0x5565dc6b47caae30ULL,0xddfeedf14fd3bef0ULL,0xffbeed7ecddb3ebaULL,0xedc7f7aecfd2be70ULL,0xffdedf5fccda7ab2ULL,0xf6f7f7fa46cbfeecULL,0xdfefff7bceda3efcULL,0xc59f74724ecab2a0ULL,0xfff7ff2f4ed37feaULL,0x23ef93d3cccb7a24ULL,0xf29c55eacddb7a24ULL,0xd96df7114f93fea2ULL,0xdb342f304fc3afc0ULL,0x8063e90b4c82f610ULL,0xadedc1547c1eea8ULL,0xa844224443200aULL,0x9181874244c1763cULL,0xa00247080889280ULL,0xf9daf76fcfcbfeb8ULL,0x50060000410600ULL,0x0,0x21004604a000204ULL,0x50000020200ULL,0xb100020404802000ULL,0x0,0x10001000800ULL,0x2004001000800ULL,0x0,0xe100510008002280ULL,0x91900180c02020ULL,0x41f5c160ccab260ULL,0x58b28640074b2a10ULL,0x9f677f6acec1ae90ULL,0x9fb2fbed4fdbfea6ULL,0xf7fdf5bacfdbffb0ULL,0x9259dd3c6d23e00ULL,0xb2f7dde04cd27efeULL,0x71efe760cdcbfa80ULL,0xfbf72fee4fcb7ea0ULL,0xe7deffff4fda3ea6ULL,0xefffffffcfdbffb6ULL,0x9fb3fefecfdbfafaULL,0xf12db7fcc582ffe2ULL,0xfafe7debcfcbfe34ULL },
	{ 0xc9bfe72ee817bb9bULL,0x6bd77b5fea36b29fULL,0xb39d5db3a8523abeULL,0x9cff566ca8721aafULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebd7ff0e28723aafULL,0xb42114fa850828bULL,0x726a47b38529283ULL,0xd9feff63ba7239abULL,0xdfbefff778763bbfULL,0xefe9f7ebaa12ba9fULL,0xfb5ff6dbfa76babfULL,0xffbb6ffe7a723b9bULL,0xfffdff7bfa763babULL,0xfff6ffeffa53339fULL,0xe1f7ef7fba56bbbbULL,0xe2ff1b7dfa7232bfULL,0xbaaecfff8872ba9fULL,0x75fcf6eff8729b9bULL,0xfb9e6df26a523abfULL,0xe3d7fb1a7a563bbbULL,0xebcf37bbaa723bbfULL,0xfbbadf7fea72babfULL,0xd4f7f6f8ba7212afULL,0x97a6ff71aa723bbfULL,0xfb9f7f6f6a629abbULL,0x9dd7fe306853ba93ULL,0xabffdfd53a52bb9bULL,0xdcfefff9aa56b3afULL,0xfbfeff7fda779bbfULL,0xfb30ffeee87212bfULL,0xffdffffeea763b8bULL,0xffffffdffa77bbbfULL,0x5fff7fefaa73baafULL,0xdf6e6673ea53b33fULL,0xdbbeaf6eaa731bbbULL,0xfdffffbffa72bbbfULL,0xedffffdffa77bbbfULL,0xf7bbfff7ba72ba9fULL,0xf3fdbf77ba723abfULL,0xffd6fd78fa723b97ULL,0xa9d7fb7eaa523bbbULL,0xfcfbffdbea72bb8fULL,0xbfdfcffbea72babbULL,0xfffffffefa77bbbfULL,0xdcb67d53a8539bbbULL,0xf3b5fb7ffa77b3bfULL,0xcfebb3f9ea733b9fULL,0xcfbfdf7ffa739bbfULL,0xfff5fff67a76bbbfULL,0xfebffd61aa73bb9fULL,0xffccf7fffa723bbbULL,0xdfffffbbea73bbbbULL,0xffffdfdffa76bbbfULL,0xf7fffffffa77bbbfULL,0xdfffff73fa72bbbfULL,0xfbbfff7fb877bbbfULL,0xefffff1ae877bb9fULL,0xbbff5bd9a8523bbbULL,0xfefe7ffbe872bbbfULL,0xfdfffffffa57babfULL,0xfffe7f7ffa77bbbfULL,0xfde7ff3f2a72bbbfULL,0xefdffeeffa731bbfULL,0x7ffa7dfb8a733b9fULL,0xfbfffffaaa533bbfULL,0x7edefff7e876babbULL,0x8bfef7b62a52b38bULL,0xffff3eceaa02bb8fULL,0xbfbdef7baa731bafULL,0xf6bef775ea52199bULL,0xf7ae3ff2e8723b89ULL,0xfbd6fb7fea723bafULL,0xadfbbb1aea53baabULL,0xffbfdfdffa579abfULL,0x31f5eefce873138bULL,0xdd0aed635a52138bULL,0xffa57ffffa72baabULL,0xc5f321f6a52bb83ULL,0xaafe9dcfea73bbbbULL,0xd49578f06873121fULL,0x9bfff773ea57bb8fULL,0xf796766ffa73babbULL,0xeccfff8aba7292bbULL,0x29ffffcaaa729a8fULL,0xfebe4dedaa56329fULL,0xdfdfff73fa739bbfULL,0xfeb63f7bfa53bb8bULL,0xcddafe65fa429a3fULL,0xadfeff5fb852baafULL,0x87bb57f1ea72bbafULL,0xf3d835f86a7299bfULL,0xbfaa77cdb27632afULL,0xabb7ffa7f8723a1fULL,0x49c216946a039b01ULL,0x889a038182131a21ULL,0x75c592fa2a52119eULL,0x47ba2570a8639abbULL,0x8a30610f8072328eULL,0xe3cd5a052a72308bULL,0x9fbacf12aa76339bULL,0x14131ee528533382ULL,0xc87464208a531a83ULL,0xf3b4e667e8729aabULL,0xcfde7ebfaa73338fULL,0x4876054862468289ULL,0x5454748f8a471a9fULL,0xdfffbf77ea73bb9fULL,0xdfdf7f765872bb9fULL,0xef9efb9eaa72b39bULL,0x2beedd572a52bbbbULL,0xb4ff7fefea72b39bULL,0xffffef7ffa53bb9fULL,0xca862e712a529a97ULL,0xc7f2e1aab8763a9bULL,0xe1275ed32a52babfULL,0xadf14e6808431b21ULL,0x99bead68ea721abbULL,0x8a360ce0285192a1ULL,0x84a11b8a850920bULL,0x4a00022228408201ULL,0x8088040000421200ULL,0x522aa0423200ULL,0x98023030001080ULL,0x8080c20a02503083ULL,0xe8ca122020521216ULL,0x4c00002021001ULL,0x741656662a521a3bULL,0xd1ca7460a0569283ULL,0x0,0x460206102002a00eULL,0x10000400000202ULL,0xe4997e68b27232aaULL,0x20120a0000002ULL,0x8000100000021000ULL,0xc478c128a252800bULL,0x0,0xa01004a080121003ULL,0x1000a6602802108bULL,0xd8162d1002039211ULL,0x402130820020212ULL,0x401c040108401b00ULL,0xb6e6cce0e8421a17ULL,0xb5fcaf6aba72389fULL,0xb066d6600523a0bULL,0x61d473aba2629a0bULL,0xa40636002040228aULL,0x100010020000080ULL,0x82108080020201ULL,0x24000000000aULL,0x20020020000200ULL,0x4000000400000201ULL,0x0,0x10001000400200ULL,0x93ba4731a84233bbULL,0x49b4be5b0a423201ULL,0x5f46a28722523083ULL,0x110080420030208ULL,0x949fdfe0ba563b9fULL,0x40ec250168020a9fULL,0x600644028508001ULL,0xc9e6d51868631a8fULL,0xbb46d64268522383ULL,0x9d327fe2aa62128bULL,0x80acf551a8021086ULL,0x9d1a1fc0d8463a83ULL,0xe0f48624c8423893ULL,0x63cbda4028123a29ULL,0x842a0de3a05292aaULL,0x8435743128123a82ULL,0x4804042040400201ULL,0xab75e32a48121b9bULL,0x4802100008021088ULL,0x8ccb8d6242180fULL,0x6026025050021001ULL,0xb2e1e716a721b9bULL,0x220424048a000803ULL,0x159109188701283ULL,0xa647f36052b207ULL,0x5412b2692052121aULL,0x1bfaff72aa523bbfULL,0xfb96ff3faa769bbfULL,0x49d7be1f38723b97ULL,0x9adedf1ee8523a8fULL,0xf49af663aa72babfULL,0xefefb770f8523a9fULL,0xfbb3af7e6a523b9fULL,0xcfffff3efa72bbafULL,0xe7ff5fdbaa763b9bULL,0x19b34de1aa56128bULL,0xd340a5f42046129bULL,0xca94273c3a121011ULL,0xe58ee6a4e252128bULL,0x4af75c91a8520a1bULL,0x1643622202b811ULL,0x20040020000008ULL,0xfa4240c8a8021082ULL,0xebdfff7efa733bbbULL,0x842020030001080ULL,0xa00500000ULL,0x20800008001ULL,0x448e0c408a520207ULL,0x10082200802180aULL,0x4088910108423801ULL,0x0,0x1001000080000002ULL,0x0,0x20004040000001ULL,0x40020080000002ULL,0x208000010ULL,0x2004000008101800ULL,0xa6165610a523a8bULL,0xdf9ffffefa72b3bfULL,0xe0dcf9beca733aabULL,0x0,0x5c0536a2250120bULL,0xfdfcec514a563a1fULL,0xdcb62f24a052a21bULL,0xa5ded3b2a8769b1bULL,0xeefe7e1de8723b9fULL,0x6cbbdfe6e85312afULL,0xd7f4347ab8531b8bULL,0xdd242d6eb85292b2ULL },
	{ 0x89e7c73602361fd0ULL,0xac0713cb62164a84ULL,0x794c5a8237634c4ULL,0xff7ff6fe6f7f1fd0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202100260100084ULL,0x8200100080ULL,0x0,0x0,0xdbb53b65643e3de8ULL,0x8ddffb8fe77f1df8ULL,0x914455200541784ULL,0x943fe4a160340ed0ULL,0xfbac3408677c1348ULL,0xdabe3f67eb760bdcULL,0xcdeff738ee767fd4ULL,0xfdffffdfeb7f7ffcULL,0xfebb6bebe77e3ffcULL,0xd7ffff7feb777fd4ULL,0xfffa6f55a5753fdcULL,0xeae6fbaa256f1f58ULL,0xfef79edfe7771dfcULL,0xafaedff6e3745ff4ULL,0xdebfafff6f7c1fdcULL,0xfbfefff3ee771fdcULL,0xdb9bef7f6f753dfcULL,0x6eebff5f677e4accULL,0xef9fcfd7437f1fd4ULL,0xf5f676ff677f1bfcULL,0xbff92f60297c395cULL,0xd3a4a3fa627e11d8ULL,0xcadf3b9f63777fd4ULL,0xa9bf5fd3e7757ddcULL,0xf4f9ffe7eb7e7fccULL,0xdbfff77867765ffcULL,0xfffd2f7ce27c7becULL,0xeff7ffbee3747fccULL,0xeefeffdb677f1fdcULL,0xf7fb5fa3e5771cfcULL,0xfffcef7f6b7e3bdcULL,0xd6b73ff7ee7e7cf4ULL,0xfff7ff7beb7f7fdcULL,0xfffffefbef7f7ffcULL,0xbfafdfffef7f5efcULL,0xf1feff69a37e7df8ULL,0xfbfeff75ef771dccULL,0x8a78fbc3a57f3cc4ULL,0x6effbfd1a4775ddcULL,0x7bbecf7fe7771ac8ULL,0xf47dfefe67771fdcULL,0xfdbe1d59a27639dcULL,0xfbf4f70fe7757fd4ULL,0xe7fbfe7beb7e3cfcULL,0xffff4ffbef7f5fd0ULL,0xf5f7bfff67763fdcULL,0xdbfefff1e7775ff4ULL,0xdbffffff637e3dfcULL,0xdfffffffe77f3ffcULL,0xbffedfdfef7f3fdcULL,0xfeffffede37f7fccULL,0xffffff7fef7f7ffcULL,0xfbb67fff66767ffcULL,0xeffffffeef7f7ffcULL,0xeffeffdbef7f7fdcULL,0xffbb5ff3e37e5f7cULL,0xffbefffa4b7f1f7cULL,0xfebe6f6ee37f1ff8ULL,0xc7d2e7f9eb773f5cULL,0xeffffdcfe77e1ee4ULL,0x153377fbe7777dfcULL,0xf7ffafd0a77f3b7cULL,0xfffe7ffee97f1dccULL,0xcbf6f73bef7f1fc8ULL,0xedc7bebf677f3aecULL,0x1d8f4fff6e7f3a5cULL,0xfffff67fe75f5efcULL,0xcbfe7d71ea7f1fdcULL,0xfbb7f77f637e6bccULL,0xef7fbe9fe67e7adcULL,0xebbfcff3ef5f3ffcULL,0x76747fd6277e5fc8ULL,0x930ff741e3721d48ULL,0xfbb47fffe37e7f5cULL,0xcf67b29f6d565be8ULL,0x8a1adbdb637703d0ULL,0x5cdf562ee67c5bd8ULL,0xb93eb77a67773cccULL,0xf7be7f7d67775fe4ULL,0xfddfffbbef767ffcULL,0xef7f5fdbef7f5fecULL,0x9696550ae25c19c0ULL,0xd1ec3f5eef765e14ULL,0x46ae3ffca9761dd4ULL,0xefc0f33a257c5c7cULL,0xefdf5ed8ef7f5dfcULL,0xffb7fffbe37f7fe4ULL,0xd5fdb77d6f741fd4ULL,0xfe9e3ff5676e19d8ULL,0xcbffef7f6f775fdcULL,0xdce7379869771de0ULL,0x8a93410a81501c90ULL,0xd6f6be7a2f761fc4ULL,0x8407800240e0040ULL,0xe9d4f723e0351e5cULL,0x46803ec26074104cULL,0x2d46450800745d44ULL,0x945c7626007411dcULL,0xd33a376266351b44ULL,0xda91ef2f6b763554ULL,0xdfd7ff4a6b7e35dcULL,0x271edfdea3771ef4ULL,0x50849449015e0accULL,0xe1ae7fd3e3761f5cULL,0xfbae6c26625610dcULL,0xdff7f72ae3771ffcULL,0xaaeedddbe27f19c4ULL,0xbeff56f2e377184cULL,0xd95ca6e6677419f0ULL,0xfaae1f7f63761fecULL,0xcfffffffe3777efcULL,0x7fdfff7fef751ffcULL,0x87f14778e2744af4ULL,0xd17daf00097c01c4ULL,0x112025000140084ULL,0xc009c11b22141094ULL,0x2ec0081822341004ULL,0xd08290081100880ULL,0x8400200024440040ULL,0x2c35c00072095cULL,0x808e10022720040ULL,0x40120000250040ULL,0xa9c0a0061580240ULL,0x3037c66663740864ULL,0x8f20af6043340fc0ULL,0x2082c0441700800ULL,0x100800000101000ULL,0x810070308600048ULL,0x11404000420048ULL,0x128046025200050ULL,0x8000000400400040ULL,0x8020011802091040ULL,0x0,0x0,0x40032821080048ULL,0x4400000000101000ULL,0x0,0x0,0xfcb5a7ea6f767f54ULL,0xfbfcbf7bef761fe8ULL,0x539257e266341b64ULL,0xc99073762325184cULL,0xa9533ccb21361d58ULL,0x48928b72617708c4ULL,0x0,0x4000000200101000ULL,0x9200420440044880ULL,0x6000001000800ULL,0x0,0x1400000000011000ULL,0x998c5422207415d0ULL,0x8c00b36860363ac8ULL,0xced6b73eef761fdcULL,0x300850122101210ULL,0x36f5f6e2e2761e5cULL,0x12000000240ULL,0x5200008000141004ULL,0x4881d02022660800ULL,0x29ff9f58c0797b64ULL,0xcc9259b7a2723ad8ULL,0x456a752263641954ULL,0x1080821400091440ULL,0x8623eb5001361ce0ULL,0x33c7931a02371474ULL,0x4aa42a401110cc0ULL,0xa14da40023051c04ULL,0xaa844548e2241dc8ULL,0x61346322a3541b00ULL,0x6d4f3d01663648d0ULL,0x321a4f42001d5bc8ULL,0x20020000000048ULL,0x668a0572a31c28f8ULL,0x8304cf2460360ac0ULL,0x2060328000761084ULL,0x98c485a017009c0ULL,0x19e706e86b561e10ULL,0xca5befd0015712d0ULL,0xd1aceb7fa17c19d0ULL,0xcf673c73a37558d4ULL,0x2f1fdf1263741f48ULL,0xb4d55fe92a757dfcULL,0xdfeff76be37f1fdcULL,0xcf8f6fefe7375c7cULL,0xffdfffbfef7f5fdcULL,0xe3ffdfdb6b7e1ff4ULL,0xf77b7f6be3773fecULL,0x59cc0500235608c4ULL,0xf6363fffe77e1ed4ULL,0xcd96b667637e0ec8ULL,0x180112c808551080ULL,0xa0087c3630500d0ULL,0x0,0xb2942d6024771508ULL,0x8142b2a024671848ULL,0x8b424a02a41418ccULL,0x8001000010040ULL,0x40700108440050ULL,0x58a02e52836409d0ULL,0x3200222e005400ccULL,0xc0000901100880ULL,0x2d10482000204940ULL,0x8c0040000106800ULL,0x0,0x1004000000100400ULL,0xa00020200101c00ULL,0x800090000113000ULL,0x0,0x822444000641d00ULL,0x8b846f7ce37e4ae8ULL,0xcbdfbba8677f3fc0ULL,0x8a9edc40017719d4ULL,0x5ca251a4c13409d8ULL,0xc9676e662d5617c8ULL,0x6ebebf70677e1fd0ULL,0xc7fef33f25775fd0ULL,0x6fdf3d5d63765fccULL,0xa2b665f6677f3ec4ULL,0xf3ed9fc8657619d8ULL,0xf01e7d50247f1348ULL },
	{ 0xeba8fb8e1a8de7cbULL,0xc8cb1e5521ecc951ULL,0xd3a496a120dd513ULL,0x7567b2292a6dcdabULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1b7e3f9b2f29ffdbULL,0xa77ddfdb3f28ff5bULL,0xfcd6ea00888301ULL,0xdbeeeff2bfa9dddbULL,0xd6bc6f7e170cc5d9ULL,0xe7feef0f2fadc5bdULL,0x88865ed30221cd52ULL,0x440100212008409ULL,0xd9668f522b6cc75aULL,0xfa9eff7f1badffddULL,0xedfef7fe3b2def52ULL,0x6fdefffc3fecd7d3ULL,0xbfbbeffb1b6ddfd7ULL,0xdfffbf5baa2df7ffULL,0xdb9efd7c3fadd7daULL,0x89fef37f9a2c87f5ULL,0xe2dfbfff9fadffffULL,0xffbe5ffbafedf7f3ULL,0xfdf7fffe3f0cdffbULL,0xfbeefff91f2df5fbULL,0xfbdfffff3fedfff9ULL,0xf3dbbbf91bedefddULL,0xafbeeedbbaa9e7ffULL,0x34bf7fea3aadc7ffULL,0x9feb6d439a0dd7dbULL,0xfff5ff7fbf8df7fbULL,0xaddebe1a1aaddbd3ULL,0xbffedfd3bf2dffdbULL,0xf6bcf7f83badeffbULL,0xdffe7f7f3f2cfffbULL,0xdf937f7f3b6ce7dbULL,0xf5d7ff5aae6cdff3ULL,0xbbdbdfdbbf2cdffbULL,0xf5bbffff3fadffdbULL,0xff7fffffabaddfdfULL,0xffbf3ffe3f2dfffbULL,0xefdeff3ebfecffffULL,0x7ffffe9bab6de7ffULL,0xbffefffbbb6cdffbULL,0xf7ffffffab4dffdbULL,0xfffe7ffe3fedefd9ULL,0xfbfff77f9fadfffbULL,0xff77ff7bbbcdd7ddULL,0xb7d9fee12b4df7d1ULL,0xd7fff9fe9e0dffdbULL,0xfffe7debbf2defddULL,0xbbdcfbffbe6df7fbULL,0xfffbb77fab2cd7f3ULL,0xbeffef783bace7f9ULL,0xfefff7ef9f6dffffULL,0xdffeff733fadfddbULL,0xffbeff7fabadfffbULL,0xffdffeffbf6dffffULL,0xffffdfdebfadfffbULL,0xf7fffffb3fadfffbULL,0xefffff7a1b8ceffdULL,0xffffffffbfedf7dbULL,0xffefefba9f6dcfdbULL,0xabffffdf3f6cffdbULL,0xbfff7ffb3f2dffffULL,0xddbffcec1bedcfdbULL,0xdebffff3bfedffdfULL,0xcffbfbbe1f6dc799ULL,0xebffffdf9f6dfffbULL,0xbcbe7fffafedffdbULL,0xddecb7bb1f2ddf5bULL,0xfbfe7ffebfedc7ffULL,0xebfef73ebb2ccfdbULL,0xeff71f7d3badefdbULL,0xf6dfdf633f2dc7f7ULL,0xfff7fefe3fecd7f9ULL,0x5e7e7d712b29c7ffULL,0xdb95db779b29cf5bULL,0xcfd7bf5e9e6de7dbULL,0xbfddce591b2ce7fbULL,0xfcb43ee313cdcd53ULL,0xcf233d521b28f598ULL,0xfb72d77b874dcfd3ULL,0x8bc2e6732a0ccfd9ULL,0x8b1edfd30b2dc7d3ULL,0xd4f6f7e83b2cffddULL,0xe9fe7769ab0ccfdbULL,0xdfbf7f7f3fedd7dbULL,0xefffffbfbfedffffULL,0xeffb7f5b0e2ce7fbULL,0x90a660230c5310ULL,0x8f75ff6aab2cc7f9ULL,0xcc082cf10f0056d9ULL,0x81dde01633884910ULL,0xebdfdfeb3a2dffd3ULL,0xb7bb6dfbbaeca7dfULL,0xd14d15630d6ce541ULL,0x9b187d702a4ca7daULL,0x89e6feee334cc5dfULL,0x6fd77e7f0b2dc79fULL,0xb2bfdfda1e8de797ULL,0xb655fe72a688cd5bULL,0xcbee4f62bf4c6f48ULL,0x8b16e12e0a4ca5e1ULL,0x9883b7502e25ebd8ULL,0xffff4d401e2cdf53ULL,0x5573dee03a2cedc3ULL,0xc1288c601f096814ULL,0xc2b0e0422b004310ULL,0xcfdbfe9a3b0dffdbULL,0x2f7ffbd8396cdd53ULL,0x84d596e0060c4d1aULL,0x90eb6633aadcd59ULL,0xfebe753fbb2cc5ddULL,0x95cbeb720e2deb98ULL,0xaa7edddf282de5f9ULL,0xbc3b7ded0f2cffb3ULL,0x8d7daf7a0b08e3ddULL,0x4eb42fd60f6cf593ULL,0xefffffffafecffffULL,0xfffffed3bfedf7dbULL,0x83384458080823a1ULL,0x6fd7183a6c83c9ULL,0x12366574094c6523ULL,0x71ddd37e3e6dedffULL,0x615ab4400e09e2d8ULL,0x13904e198009c69aULL,0xf12c92f8310cd5dbULL,0xc242000008010ULL,0x80400000004800ULL,0x121a000a002602ULL,0xdc3d1583300cb10ULL,0x34810ee6a64dc791ULL,0x9006080001084450ULL,0xfa04734e3788d5d5ULL,0x8c080061a0c2410ULL,0x8202000008ULL,0x0,0x20860000008418ULL,0x0,0x800800080010ULL,0x2100024008008080ULL,0xb80209121a0c81d3ULL,0x50000008001ULL,0x0,0x80000001000c0000ULL,0xb5e5e40230c8115ULL,0x95ff5ff73b6cfff9ULL,0xc285534b0e04e7d9ULL,0x3840264000c0190ULL,0xebf5d72e3ecdfdf9ULL,0x402120000000540ULL,0x1800e80001001210ULL,0x201000080080ULL,0x8800040000008102ULL,0x0,0x601080a08014088ULL,0x800100080040ULL,0x32421008042408ULL,0xdf960df21f29fddaULL,0xdc0cd566060cc592ULL,0x4fd77ff68b6cffdbULL,0x100153500a002318ULL,0x74af5ffc3f0dcfdbULL,0x984224010c086011ULL,0x694124418a82140ULL,0x8ddf9e302d0967dcULL,0xe61896dc3a28859dULL,0x4369f7b1a6dd792ULL,0x9720a4721a6dd7d3ULL,0x90a80f02084c4790ULL,0xe844b0843209c701ULL,0x1890da920a084a99ULL,0x54b272708b0cf49aULL,0x18503244088d6dc0ULL,0xe80065620a081719ULL,0xa0000000081ULL,0x204393c00c24c301ULL,0xf9b85180a4cde92ULL,0x4026f4d8200dc587ULL,0x68ea5dc23d0869d0ULL,0x1940800401080150ULL,0x409110000c8121ULL,0x804a00288020ULL,0x1100002824004100ULL,0x332a300200895daULL,0x818cd3270608c73dULL,0x4fe3fe221e0c47d8ULL,0x9a5fdd5b3b2df7ddULL,0xf6b9ffe80c2cc4f9ULL,0xf9ceed63a96cf7dfULL,0x36be2d68ba6cb7d7ULL,0xc8b2f71a1e0cc7d8ULL,0x8f7ddfda022cd5fbULL,0xf6fedcf19a0dd7fbULL,0xf0fdbef6220ddd1dULL,0xf4bd3fd10d0dd5b7ULL,0xc7baff2e1b29f7d1ULL,0xa9c73ec00b08b7d1ULL,0x1c1997e9030cd779ULL,0x2841002009200418ULL,0x14002008000408ULL,0x809030030a080358ULL,0xc4ca4a1a1a21c589ULL,0x0,0x0,0x4480a0101a208201ULL,0x0,0x2461221020084d1ULL,0x481202004808ULL,0x5041000028000140ULL,0x2010240012000408ULL,0x8400001004000ULL,0x803000000c8080ULL,0x8008000000081ULL,0x24f2b770120c85b1ULL,0x8334044a0308c490ULL,0xd38520300804a419ULL,0x808010000a002008ULL,0x8b18064a13448419ULL,0xc43624ac0a0cd6deULL,0x710201a0010d4510ULL,0xfebe3dfdbbadf7fbULL,0xcfcff6dc1f2dffd2ULL,0xeddf7a7a3f8cc5d9ULL,0xb5db5fe21eede5fbULL,0xd47cffd11facdffbULL,0x98b23d320e6c4510ULL },
	{ 0xc208f0234da20b62ULL,0x85817214c904d22ULL,0xb94cff18dfa9f62ULL,0xf2f6a76dcdaa57e3ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4cdb3fbdddba7feeULL,0xefffffdfddbaff7fULL,0xdefbf7eedcbaffe7ULL,0xd9ef277fcdfecf73ULL,0xffba2f3749fa477bULL,0xfff7ffbfddfefffeULL,0xbee17554cba6de0ULL,0x85a24d850d2a6963ULL,0xcdefff6bcdbecfeaULL,0xfb8fff5e49ba4feeULL,0xefdfebf9ddfadf4bULL,0xa97fdef9ddeedfefULL,0xb7fbdfefddbacf33ULL,0x516ef796c5ba67f3ULL,0x5ebaff75ddbadfe3ULL,0xabfbfb2fdcfa5f33ULL,0xefff7f7fddfe5fe3ULL,0xbffeffe3ddfa5fffULL,0xdf7ff7ffcdfecffbULL,0xfefeff73ddba75faULL,0xfbffff3fddbecfe2ULL,0xedc3fe7fccfedfeeULL,0xfffecf7fddfe5feeULL,0xfeffffea55fadfefULL,0x72b6f7f38dbadf4eULL,0xfbbe7f6eddfa5ffbULL,0x5ddebbbe5dfacfe3ULL,0xefdfdfde5dfedf7fULL,0xd4fffffaddfe7fe6ULL,0xdfeefffbddfedffbULL,0xfffe7f7fddbeffebULL,0xefffff9eddfedfe7ULL,0xaff7bfdf5dfadf6bULL,0xb7ffffe3dcfeefe7ULL,0xffffffff5dbeff7bULL,0xffbf3fffddfaef7fULL,0xcfffff7fddde4ffbULL,0xeffffedf0dfeffffULL,0xffbffffeddbeefefULL,0xfffdffffddfedfffULL,0xffdefdffddfe7ffbULL,0xfbfeffffcdfadff6ULL,0xefe73efbddfeff7bULL,0xbfff6fffccbedfebULL,0xf5ffff7fddfeffeeULL,0x6bee7ffaddbafdf3ULL,0xfbfeff37ddba5ffbULL,0xedffff3f5dfedfffULL,0xfffddffbcdbedf73ULL,0xff777eefddfe4deaULL,0xfbbefde35dfadfe3ULL,0xefbfff7fddfeddebULL,0xcff7febfddfedfefULL,0xefffffdfddfeffffULL,0x77bf7efe4dfe5fffULL,0xedcfff7f5dfe5f63ULL,0xfbb77fff5dbe4fe7ULL,0xcfdfffff4dbedff7ULL,0xaffdffdbddfefffbULL,0xfdffffffddbe7fffULL,0xff7fffffcdfeffebULL,0x7ebf2dfd9dfaff6fULL,0xc4cfff3fddbeffebULL,0xfa5e7fd3d5becfe3ULL,0xbebf6ffe5dfa7ffbULL,0xf3fdfffbddfecfe6ULL,0xfffe7ff7ddbadfefULL,0xe9eeebaeddfa5dfaULL,0xcff37ef35dfedc6aULL,0xbbbfcfed4deacbfbULL,0xd1ffb6fa1df85fbbULL,0xf3beff7bddfa7de3ULL,0xfbd5f77e8dfecfffULL,0xeaf3bff71d9aff61ULL,0xffdfdf7bddfefdf6ULL,0xb4ffefea51bc4f5bULL,0x871afff2d8beeda2ULL,0xf884e72e4d8a4f6fULL,0xcff7fe5fcdbaff7bULL,0x234ec81a04307f4bULL,0x500976e309da4fe3ULL,0xd6eff630dba7deaULL,0xffffffffddfeffffULL,0xffffffffddfeffffULL,0xe8ffffd31cba5fefULL,0xbceddb331dea546aULL,0xd968f773099a4933ULL,0x85343c04b84400ULL,0xedc9f626cdb89772ULL,0xef5fdfcf1dfe4c7eULL,0x99bb4fed4dfacb72ULL,0xd4c1d618189a0d40ULL,0xed145d37ddbad426ULL,0x8bf5a76c0da85972ULL,0x6eeb0ed305fed8ebULL,0x911e5f7b4da24962ULL,0xd950b659c5fed46eULL,0xab8e7f73ddbacf7eULL,0x43916b2f4d3a4d6aULL,0x204000c001200002ULL,0x86cecf391cba4f43ULL,0xf317a6ec019a4462ULL,0x9391d7318dae467aULL,0xf8a16d1a0da2c140ULL,0xcfc63e1e4c9a4fbbULL,0xff9fdfdf9dfedf6bULL,0xfcb7f7ec5dbedfe2ULL,0xbdfcff7144ba9ff3ULL,0xdafe7fefddbecfebULL,0xed6cfebe4cda4fabULL,0x92f5c1a0d90472aULL,0x97bfdfedcdaa5f7bULL,0xd17d72d44cae9b67ULL,0x128625670de869e3ULL,0xeff7ff2e5dfe7f73ULL,0xefffffffddfa5ffaULL,0xfffffffbddbecf7fULL,0x955cc61244ba5d2fULL,0x318207f0058281e2ULL,0xc100a122012a0026ULL,0x685516c28d9e5764ULL,0xb2a2cf5a849a47e0ULL,0xc02202f005381162ULL,0x250100926040ULL,0x88000000000000a0ULL,0xa152b04e981a4de6ULL,0x80904f200c8a0102ULL,0xb2a0d0e0008c492aULL,0x88308f02400a63a2ULL,0xbb81df755dea05b6ULL,0x8100220304808160ULL,0x0,0x4c0106181a20402ULL,0x0,0x42000000300ULL,0x0,0x101000000000a0ULL,0x8000040000000801ULL,0x93042760048a4842ULL,0x9802194004ec4060ULL,0x0,0xebde7e4fddfa5ffbULL,0x9efb6ffad1ba4f7fULL,0xf5ec766b8dba6d64ULL,0x7900a9a009a8c060ULL,0x69d0c23e45aa8b6aULL,0xa050300a042a0102ULL,0xc00008004000ULL,0x404144000928404ULL,0x0,0x3004000408820100ULL,0x0,0x0,0x124840000102ULL,0x91ceaf600c8cc362ULL,0x5380a72f0c9e5766ULL,0xcdf6be4fccfe4d63ULL,0xa9124fdb15ba6deaULL,0xf6f8fe61c4bedcfbULL,0x10ca252400a0932ULL,0xc2a4223a498e812aULL,0x2fc7da81889ccfa2ULL,0x88501418401065a0ULL,0x50944ef201ba042eULL,0xf184245a19306d00ULL,0x8243b4c44104f42ULL,0x8bc27a029d920c02ULL,0x200005240200140ULL,0x3aa94f4bc9ba4d22ULL,0x85a4104a05ac8022ULL,0x202000080000400ULL,0x480003020082000aULL,0x43004000080022ULL,0x12004e80888a4500ULL,0x160402e000180f2aULL,0x908552040380460ULL,0x1000000400000201ULL,0x20a1200040a0142ULL,0x822010204106000ULL,0x805061002a4026ULL,0x3232047040900562ULL,0xda9222268db0d7a1ULL,0x9ffabac48cfaece2ULL,0xab6445da85324c64ULL,0xf0ecdc684daacdf2ULL,0xdffeffb3ddfeffebULL,0xebb63ffe9dbaf737ULL,0x5c3fe101db8ddeaULL,0xaa121688448841f2ULL,0xb4ff7f690dba4fefULL,0xf9fc3ed655badf62ULL,0x6c960f345c9acd60ULL,0xcccbba3109124f53ULL,0x12b54a0254940f20ULL,0x61874c448124d2aULL,0xa000502000988420ULL,0x200200001000100ULL,0x8888001040280022ULL,0xc09e321100fa664eULL,0x30404000904020ULL,0x0,0x2016004001100ULL,0x8008a2284940500ULL,0x6388244024122ULL,0x81405a201c8e002ULL,0x2420020000880042ULL,0x210011000880300ULL,0x6002040000820800ULL,0xc00001200000ULL,0x81604c700b84622ULL,0x80884420018a0102ULL,0xa0070041000142ULL,0x400004040000100ULL,0x2380c10a85a84ceaULL,0x10580604000b60ULL,0x1026244201900603ULL,0x100240200084302ULL,0xfbb605e249a8cf6bULL,0xc966f6904dbc0712ULL,0xefff9effddba5fe3ULL,0x87b36fe84dfaf7afULL,0xd1cbf74cd9b84fa3ULL,0xc9a6456198004d00ULL },
	{ 0x8200e21a02112040ULL,0x27711ed034026746ULL,0x958bcffbb632f4eaULL,0xf1f4f6582432e4e4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x5001802040004260ULL,0xffdffef2b4337feeULL,0x83de5f0ab21244eeULL,0xb6e5d6ec3011e7e6ULL,0xd13e2c623222c4c4ULL,0xd7a72debb631eeecULL,0xef4ffe32301366e6ULL,0x2bccdfd3863027eeULL,0xba5668143276c4ULL,0xf16f2f7e7433756aULL,0xfbaebe5fc413f7ecULL,0x8fc3e33f343262e2ULL,0xa9b7bec33413f6c4ULL,0xa6734f6a1633ffe8ULL,0xddbffdffb433feeeULL,0xff9e7ffbf6337eeeULL,0xc83cfb8b94123fe6ULL,0xeffffefbb433efeeULL,0xdf3ecff3b633efe6ULL,0x97d1366e34133ee4ULL,0xcbf6bdf1741076eeULL,0xfbf4fb2f3032f6e2ULL,0xcbc2bedf163377c6ULL,0xebbecfffb433ffeeULL,0xddff7eeb7613f7eeULL,0x76beff63b213feeaULL,0xfb95e73e343376e6ULL,0x8fffff3fb613f6e6ULL,0xeffedfdbf433ffe4ULL,0xfcfdfffe96337feeULL,0xdbefff7ff433fbe6ULL,0xffbeffffb633ffeeULL,0xdfffffff7633f7eeULL,0xffffffdbf633ffeeULL,0xfdfe7febb633f6eeULL,0xddffffefb433f7e4ULL,0xffbf7ffff433ffeeULL,0xfffbfff7f633ffeeULL,0xaffeffdff633ffeeULL,0xbffbfffff633ffeeULL,0xf7fff7fff633feeeULL,0xfffefffff633ffeeULL,0xebffffeff633ffeeULL,0xffffffdbb633feeeULL,0xbed5dfff3613ffeeULL,0xfffffeff7633ffeeULL,0xeffeff6a3633ffeaULL,0xfbd6ff36f433f7eeULL,0xbffffbbf74337fe6ULL,0x6fbfcff33633ffe6ULL,0xfffefefb7633ffecULL,0xffffef73b633efecULL,0xcffe7f3f3633e7eeULL,0xfff7ffbbb63377eeULL,0xbfffdfdeb633ffeeULL,0xf4fffffe3633ffeeULL,0xfffffffff633ffeeULL,0xffbf7ff7f633ffeeULL,0xffdfffbff433ffeeULL,0xafffffdff633ffeeULL,0xffbffffff6337feeULL,0xf9efff7ff6337feeULL,0xffb6bf7ff633f7ecULL,0xffefdffaf633ffeeULL,0xefffffefb4137fecULL,0xbeffdff736337feeULL,0xf57fff7ef6327feeULL,0xdbfe7ff3b433fee2ULL,0xcb97fb7bb613f7e6ULL,0x7fdfbe6ff4136fe6ULL,0xbbbfcdfa3633ffe6ULL,0xde7ff6f6b633f6eeULL,0xfef6777a7633f6e6ULL,0xf3dffbffb632f7e6ULL,0xeff79efef430fdeeULL,0xbbbf4f5a7613f6e6ULL,0x4673fef0163326c6ULL,0x9b04bd52501152e6ULL,0xf8a1f16e743242e6ULL,0xcedffe9ab6337de6ULL,0xbf5eb7da3631f7e8ULL,0x74bcaeec34336feeULL,0xdde7b770a632f6e4ULL,0xffff7ffff633ffeeULL,0xfffffffff633ffeeULL,0xefde5fda3433ffe6ULL,0xb6f7d7eb34137ee0ULL,0x851ad662342366c4ULL,0x9e9e3ffdf412f6eeULL,0xcf52fe0b1433feecULL,0xf9ffdedf9633efe6ULL,0xbefafff3f613f7eeULL,0xedffb77fb4337eeeULL,0xdfee7d6a3433f7eaULL,0x8983836a401144e2ULL,0x15c0020030004042ULL,0x339ecff1b013e4ecULL,0x7655d4eb263376e0ULL,0xcfae7f6834137feeULL,0xc920e7281633f9c4ULL,0x94b1e0f06325664ULL,0x23300610542142a0ULL,0xf4345ef0b432d2a4ULL,0x6198857054124eecULL,0xdd06af65243276c6ULL,0x47cbdea1343177e8ULL,0x8bbecfdab6127fc6ULL,0x362d55ef16132ec2ULL,0xc1e31fd2741017e6ULL,0xfefd2f4f763067e6ULL,0xcfc6efb7f6127fe6ULL,0xadfdff435432e7eeULL,0x948774a2143264ecULL,0xb5bdae283422f6e4ULL,0xfb9e3ff8f430fee8ULL,0xa9d6f3af84136be4ULL,0xb476362340067c2ULL,0xf7fbcdf33032f7e6ULL,0x9053532820122ec0ULL,0xb4100140400062e0ULL,0xc9c083342010f560ULL,0xad9105114004dc0ULL,0x549a156234326cc0ULL,0x11de51aa1431c6a0ULL,0x38452602600006cULL,0x9304a32450010460ULL,0x888180130104402ULL,0xb088e591010746aULL,0x10000fc484131220ULL,0x5822354010126680ULL,0x8000010000110000ULL,0x4082781204104164ULL,0xe802218182105060ULL,0x0,0x100040000020004ULL,0x0,0x8aebfeebf433f6ecULL,0x108000000104000ULL,0x0,0x6840e060243262e4ULL,0x7e9ebf67a633e7ecULL,0x1300002002cULL,0xaff73b4b3433f7eeULL,0xb4f36fe0361372e4ULL,0xd369f77ff231deecULL,0x786516042127460ULL,0x8912011a20006260ULL,0x41000000204000ULL,0x131f480214124bc0ULL,0x2100020204002044ULL,0x800350000136020ULL,0xc058f130143246eeULL,0x4cc21000001124a0ULL,0x480c100020002c0ULL,0x90400620001162c0ULL,0x9f56e4722232e6e0ULL,0xe396f32c5433f5eeULL,0xeecffe8eb433efe6ULL,0xfb9acfdbb43376eeULL,0x16fc74e0b413fbeaULL,0x9010045030104462ULL,0x4510203100136240ULL,0xc0c29f3834323fe6ULL,0xa9c849da3433c260ULL,0x94750ee11412ea6aULL,0xc95c9250342042e0ULL,0x3a83246e10323aecULL,0x120000000084ULL,0x5020402000068ULL,0xbb86170801242e0ULL,0x4626f646361212e4ULL,0x78123e02801066e0ULL,0x40810880000060ULL,0x10100000004040ULL,0x280058800024ac4ULL,0x8c34302112126466ULL,0x8107402103346c0ULL,0x200210110010022ULL,0xa14a90ca040269c4ULL,0x203540c000020cc4ULL,0x50802c0084120444ULL,0x20022c6000204ec0ULL,0xd291a23e343166c2ULL,0xcfebf6fff433ebeeULL,0xba4007db7033f9e8ULL,0xf4fefee6f4316feeULL,0x51aaf160303363e0ULL,0x42a02426200324e4ULL,0x26c5ab18340042eeULL,0x88dfd908941064e2ULL,0x54ff77ed7613feeeULL,0xd9ffaf6cb613e066ULL,0xfb943fc43633476cULL,0xc964f7189422eae6ULL,0x2df0a01341264e4ULL,0xc01014000304060ULL,0x1224272b1413e0e6ULL,0x2aa03c4054127064ULL,0x0,0xaff63f517433e7e6ULL,0x7904a810412c644ULL,0x144204040420c060ULL,0x500000000140ULL,0x20a400000024080ULL,0x2142001000125004ULL,0x1a1e8f53c012e9ecULL,0x0,0x8040200000010260ULL,0x4004023400222440ULL,0x0,0x3594650741270a0ULL,0x401000000004040ULL,0x1100a00000104040ULL,0x8800832014010480ULL,0x0,0x10000200000028ULL,0x90080100004040ULL,0x40a4700024105024ULL,0x52880c0010112142ULL,0x8102601c84300280ULL,0x41810000003c0ULL,0x4a08975b3411c244ULL,0xf03af76bb43364ecULL,0xda362950243362e0ULL },
	{ 0x40000a0400101882ULL,0xb461a595014d43aULL,0xbfbb8de3d8b494fbULL,0x432da60b0949403ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x981d33b50dc96f0ULL,0xcbdffb6ff8bc7efbULL,0x271b4756500ca3c8ULL,0xf6f5fefdf8bcf7dbULL,0x596db76a58b43cebULL,0xdd34270f70dc30f8ULL,0x1ccc0718f09438e0ULL,0xa29edd5fd05c97ecULL,0x3dbb5f6b7094b4dbULL,0xc978bf0bd0d45dceULL,0xdcf4277ef084b5faULL,0xc9eff7fef0dcd6eaULL,0x37ff12cb605c3eb8ULL,0xbfba5ff7d8d4f6eaULL,0xd5edff7ef0fc7feeULL,0xcf327f6ef0943cffULL,0xe994ffaed8bc9ffaULL,0xd9f21ef970b4befaULL,0xfdbfeffbf8fcbefbULL,0xbdf8f66a68dc9ffeULL,0x69fe7df1f0d43cf8ULL,0xeb8ce32cd034b6dcULL,0xc6fb7ad1f0dc96faULL,0x85ffcffef0dcfb6aULL,0xf5f7f7eef85cb6edULL,0xfbffff73f8fcfedaULL,0xff7e7673f81cf4acULL,0x2f57ff2ef0bcb4fbULL,0xabffdfdaf0dcfecaULL,0xfcfffefbf8fcbffbULL,0xdffff763f8dcffe8ULL,0xffbf6ffef8fcfffeULL,0xfffffffff8fcfeffULL,0xfbfedf5ff8dcfefeULL,0xbfff7ffff0dcbefaULL,0xf5feff6ff0fcbffaULL,0xfebf7ff7f8fcbcffULL,0xffffff7ff0fcf7fbULL,0xfffeffdff0fcfffbULL,0xbffffffff0fcffffULL,0xfffffffff8fcbefbULL,0xfffe7ffff0fcbefeULL,0xebf5fffff8fcfefaULL,0xffffbef3f09cffffULL,0x7ffdefeef8dcb5cbULL,0xf7edbe7df0fcbefbULL,0xffee7ff1f0fcb7faULL,0xfbf7fffff8bcbfffULL,0xeefffffff8dcbffeULL,0xbfbf5ffbf094f6fbULL,0xfefffefff0fcffffULL,0xdfffaff7f8bcbffaULL,0xfefefffff8f4fefbULL,0xcfffff3df8fcffffULL,0xffffdfdff8dcfefeULL,0xf6ffffe7f8fcbffbULL,0xdffeff73f8dcbefbULL,0xff973f7ff8fcbffbULL,0xefdffebff0fcffe2ULL,0xafffffdbf8dcfcfaULL,0xfdffffebd8bcb6faULL,0xfffffff6f89cbffbULL,0xdebe3fff78fcf7ffULL,0xfbfef77ef0bcfefbULL,0xfffffffbf8fcfef8ULL,0xbf9ffffff87cbcfbULL,0xf9ebb778d05cb6daULL,0xfbee7f67f0d49eebULL,0xabf67336f8d4bcfbULL,0x69751e8250bc96d9ULL,0xbd999ffbd8dcb3daULL,0xf577cefef8d4bee9ULL,0xeab42db0705cb6f8ULL,0xebfffb3ff07cbfeeULL,0xeedfbe9ff854dee8ULL,0x7bbfcffb70dc96eeULL,0x58ffbef5709416baULL,0xd2babdf06034b5faULL,0xf7a6eb7ff01c1ffbULL,0xeff6ff9bf0d4fefaULL,0x2f7ed6cfe89496feULL,0xf7f6f7e8d0dcf4daULL,0xddffff71f0fcfefbULL,0xfff66f7ff0fcfffaULL,0xa5dffbebf0dcdfeaULL,0xffeff5bf85cd7e3ULL,0xfc9b5ffbf0f49edaULL,0x5f68f7e8509c76daULL,0xdaa67f6ca8a450eaULL,0x8955f32a709cd0a8ULL,0xeafff9dff0dcbefbULL,0xdeba7fb9f09c3de2ULL,0xb265bf79509cbe52ULL,0xf7be3df1f8dcfff7ULL,0xa9b7dfeff0f4b5feULL,0xe9d65f54f0dcbeefULL,0xbfff7ffbf0bc97eaULL,0x963dffe0f084f5c2ULL,0x6ffe2d60f09cb2f2ULL,0xfbb1d36c781cb4f0ULL,0xeaeb9e5f509cbaeaULL,0xf7bddf61708cb6eaULL,0x349d1fe870dcb69aULL,0x9338edf5d05cb6f2ULL,0xe2824063505c327aULL,0xcfc0fd13c0bcb4faULL,0xebd45fd7f0dcd6baULL,0x96395e64f0bc9ee4ULL,0x194a79f9f0449428ULL,0xfa92776e589c16fcULL,0xe4eeef3a70dcbefaULL,0xe97fdf5af01cb6fcULL,0xbfff7feef0fcf6eaULL,0xd9fdbf72d8fc3cfeULL,0xde34755ef8141eaaULL,0xadcfc606d01c9ecaULL,0x1000360040000660ULL,0x85b25638705cb6e3ULL,0x50902050400026c4ULL,0x928809001004148aULL,0xc308e377709498f2ULL,0x6cd21a8778dc1cbeULL,0x10150d4540c01642ULL,0xd4176662005c126aULL,0x48643d2040d4b038ULL,0x8310830400c01402ULL,0x8c3ba005084b1c0ULL,0xa6684e3948141e88ULL,0x30341e7ad024a0c0ULL,0x32204561781416e8ULL,0x1400094850040090ULL,0xd83c0110501c88ULL,0x9b1e1f57f894fff8ULL,0x1000000440008000ULL,0x401a030841000ULL,0x8381050010149042ULL,0x4002002080001400ULL,0x0,0x0,0x600c80201040142aULL,0xde3607fdf0c495ebULL,0x8ccaf912d084a4c0ULL,0x2d4a1e1e40dc36a1ULL,0x46040007000ULL,0x8098be10705cb436ULL,0x35045be0f0701060ULL,0x20410000080ULL,0x1004020000020ULL,0x4308cd41405c92d8ULL,0x9054f249501080d0ULL,0x28a02422101c0090ULL,0x8940e120008016c8ULL,0x8042824000143cc6ULL,0x2800a0c040041403ULL,0x3002906c904430c2ULL,0xf8843d2298141428ULL,0xe1b1ff79d0dcd67bULL,0x4942d810b8141b48ULL,0xcf1bcf53f09cbee6ULL,0x4642960240c41080ULL,0x10225700010a5d0ULL,0x41022d60500070c8ULL,0xad16ea2c501cbe8aULL,0x68be99c0d0941c80ULL,0x444d4e9500434c0ULL,0x8024a67b309c56e8ULL,0x48ac2744701424f0ULL,0x8018532250148490ULL,0x209580190009080ULL,0x80220e81408010a8ULL,0x8068766240a4a441ULL,0x1020c4040841002ULL,0x8800000000000240ULL,0x7407405094a850ULL,0x10808dc0404094c4ULL,0x4002129050048484ULL,0x6a3c2ce05094b7c1ULL,0x30002704501003e0ULL,0x2823021681c34f8ULL,0x0,0x1000160200400a0ULL,0x80890520d0108340ULL,0xc90251e650941e84ULL,0x3fd7371ad09cfedaULL,0xef76dfdad89cfef8ULL,0xb4bfdfedf0dcfefbULL,0xb527f672d0f477d2ULL,0xa3b0767c7014b7e8ULL,0xd5f3ea0e5054b252ULL,0x465800002401c0ULL,0x36b2ffeb70dcd7efULL,0xdd6ea75e70dcbfeaULL,0x4e882f3b7014bdfeULL,0xe9dbf7b860bcf2e8ULL,0x7f567a9fe0bcb6d1ULL,0x80025830708c04a0ULL,0x51715728509cb288ULL,0x32143d31081c189aULL,0x80120b1450048002ULL,0x8a5714465000b4e8ULL,0x5dacb83701c8c68ULL,0x0,0x4406020140010ULL,0x8a10c100400480c8ULL,0x42105a40942050ULL,0x2180c6040549050ULL,0x1004000040008000ULL,0x59c012050108410ULL,0x4000800000040000ULL,0x200200840000ULL,0x2e006300301891e0ULL,0xb4b4f6a270d413e8ULL,0xd9febff2f09c9efeULL,0x90200c4020441020ULL,0xc1084000a000ULL,0x4101001d0001000ULL,0x24040000010ULL,0xa1459f48408cb048ULL,0xde9e2773f0d451e8ULL,0x148612980581250ULL,0x2a1634803004b4e3ULL,0x93a24c72d084b0c8ULL,0x919cbc045034905aULL,0x540e6de8509cd2e8ULL },
	{ 0x8000612401029062ULL,0x485e19549140013aULL,0x3b88cfbaf0f9abf4ULL,0xf5b63022d104ab58ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x240088000000ULL,0xdffefffff1efbf78ULL,0xcffff65bd1e7bb7eULL,0xa7efdfded1edb35aULL,0x949bfeefd16fbb6eULL,0x9566275ad1a02176ULL,0x89102b289124b862ULL,0x5840863001228a62ULL,0x4c0f5e511166b45aULL,0x14af5eebd9e52268ULL,0x8f79772bd8c9bfdaULL,0xbebe3ff491e1bf6cULL,0xf9fcf7bfd967bb9eULL,0x6afafe9bd96bbb52ULL,0xa7fce7f1d9e9af56ULL,0xf7fefffcd9f9af7eULL,0x5fc66f7fd9feaabaULL,0xe9defbdfd1efbf7eULL,0xed5e1edff178abf6ULL,0xffbeeffbf9efbbfeULL,0xd5fefafed959b3f6ULL,0x5bee6fe3f9e9237aULL,0x99b0eb32d9d3ab68ULL,0xebffffdfd9f5abfeULL,0x8fff4f73f9e3abf2ULL,0xf7bdfefbd1fbbbeaULL,0xfffefff3d9ffbf7eULL,0xdfbaeffff9dabbfeULL,0xdfebfe3bd9f7af7cULL,0xefffeddbf9f5bf5cULL,0xf4fefefff9ffbffeULL,0xfffffffff9efbffeULL,0xffbeff7ff9edbf7aULL,0xfff7ff3ff9ffbffaULL,0xabffdf5f99efbbdeULL,0xbeffffeff9fdbbfeULL,0xdffefd7fd9ffbbfeULL,0xffbfbffff9ffbbfeULL,0xefffffbfd9ffbfdeULL,0xfffffffff9ffbbdeULL,0xbffffffff9f7bffeULL,0xffffbffdf9ffbffeULL,0x5fbefff7d9ffbf7eULL,0xcbf6ffffd9fdbb7eULL,0xffffbffff9fdbffeULL,0xfb9eff7998efbffeULL,0xffff7ffef9ffb77aULL,0x7ffefffb99fbaafeULL,0xfbffff7ff9ffbbfeULL,0xdfeffffff9edaf7aULL,0xbfffcffff9efbffeULL,0xf5f7fffdd9ffaffeULL,0xfffeff73d9efbbfeULL,0xffbfff6fd9fdbffeULL,0xdff7be1ff9efab7eULL,0xafffdfdbd9fdafdaULL,0x7efffee7d965bbfeULL,0xffffff7bf9efbbfeULL,0xffb77f7cf9ffabdeULL,0xefffff7ed9fcbf7eULL,0xafde5fd9f97fab7aULL,0xbdbffff3d9f9af78ULL,0xff2dffffd9edbfdeULL,0xffbf3ffff9efbf7eULL,0xfffff62bd9ffbffeULL,0x6fffffdbf9efab7aULL,0xbbbb6fefd1f7affeULL,0xd9fefffff9e9a3eeULL,0xfd7eeffc91f9abfeULL,0x8ba3f73ef9e4bb7eULL,0xeade3e97d9cd2b7aULL,0x979f5ff2d1aaab72ULL,0xf7fefe7ff9e9bfdeULL,0xfbbe2f6bd9ffbb7aULL,0xabfefb3fd9edabfaULL,0x4feb3e69f1ecbb68ULL,0xffbdefffd9ffbffeULL,0x74a577edd1a7ba7eULL,0xa0f86d60c1c18b52ULL,0xdd8eeffef1eebf7eULL,0xcdebf73ff16fab7eULL,0xbefadb5bd16b9bf0ULL,0xd5f576f49965bb78ULL,0xfdefff6bf17fbbeeULL,0xfebe7ffff9eebbfeULL,0xffd7ee7ff9fda77aULL,0x6f5cffd3d1e3bb5eULL,0x15fffefbd9f1a1f6ULL,0xdb707f62f1ed276eULL,0xfeb62f6f89e8a74aULL,0xc3f7fd6cf1ef2374ULL,0xaa7ddfc39172bdf0ULL,0xb7f3ef78d9edbb6eULL,0x268cb768c141ab66ULL,0xb96efd64d1f6a2faULL,0xe884d21a91b5aa4aULL,0x6bc21ab1918a9b82ULL,0xf39feffbf9e7ab7aULL,0xd6b8f2c8d1458118ULL,0x37ee6c7191c13b6aULL,0xcbd4f766d9e6bb5eULL,0xe043162990d70148ULL,0xa6bfdfe9f9e7b77eULL,0xdcff3feff9efab7eULL,0x8018201051c08142ULL,0xfdc4fffbd9eeab64ULL,0xab7fbe13886fab58ULL,0xebfedfd3f9e9a776ULL,0x66ff7ce4d9ecbb4eULL,0x892c656a9120a34cULL,0x66b16450c1642248ULL,0xcff6fea6d9f7bbfeULL,0x637f77dfd1f1a356ULL,0xb477dbf199edabbeULL,0xe9feff6398f9227cULL,0x88342772c1842a1aULL,0xc192dbe6995b8af0ULL,0x84f1a41d840ab34ULL,0x111eedf499e69b1aULL,0x1071bf7d9155bb5eULL,0x500202090008110ULL,0x8bf2f682d9e52958ULL,0x812075088412a04ULL,0x529b079d98c0a146ULL,0x8026726e99e82204ULL,0x224256080c89200ULL,0x8310810c98c09800ULL,0x608010099000a810ULL,0x9c17450810208920ULL,0x1444240801410360ULL,0x8288284088412254ULL,0xf018686089c00630ULL,0x952d080c1650240ULL,0x9398df59d9e9badaULL,0x702696dcd8902066ULL,0x14e334ad8602840ULL,0x84000000010040ULL,0x0,0x0,0x0,0x4d0c2122d0c12046ULL,0xf4a62fdb89cd275eULL,0xc182b1b29909a3f2ULL,0xe1460c02e1a18260ULL,0x3000470281608220ULL,0x807cf742b14eae6aULL,0x2410044190400a00ULL,0x42000841400200ULL,0x808108420000a846ULL,0xbba9cddbd9e5a3eeULL,0x14090000000ULL,0x20080000208000ULL,0x824442401046114aULL,0xc00a0010081200ULL,0xacc54841482602ULL,0x51040062c1202802ULL,0x93b4be70f1aba2e2ULL,0xfb30f7ffd9f7bb7aULL,0x8a467a2fc130b846ULL,0xebdb4f9bd97dbb7aULL,0x5001402080640a00ULL,0xc2052011650140ULL,0x8abd1ec29849a9d2ULL,0xcbd3f31ed9cd2a4aULL,0x69ccdbcbc109b966ULL,0x20b852e2d1ca2742ULL,0x47baf34af8588f50ULL,0x6aae0f94b108b3deULL,0x839acb0ac964abeaULL,0x80a100180400a00ULL,0x202064e91002216ULL,0x4000080200ULL,0x4980265009402446ULL,0x810050f011a08238ULL,0xc0641c039029a320ULL,0x16c1408c81510a60ULL,0x90f41609d0a13264ULL,0x4170261051242008ULL,0x5981d32ad1c52758ULL,0x10513290802c024aULL,0x98ce009902a2806ULL,0x0,0xc00044000400210ULL,0xc010c3a091192070ULL,0xa8defe3391cdbb7aULL,0x9a1fdc4ad1c93066ULL,0xf7fed7eeb967bbdaULL,0x1f4ce61891f72a54ULL,0xc7803365d1062258ULL,0xa167e902d960ab62ULL,0x206c05d8d121a14aULL,0x22a254aa89422d42ULL,0xd126a7eef1caa342ULL,0xc2a63754f1012362ULL,0xc16ee776f1a5abd8ULL,0x4cd2fa4fb976a34aULL,0x939a4dabf12bb352ULL,0x80c4b75a904f224cULL,0x3020004041400400ULL,0xe05070e290e12040ULL,0x23511409110a820ULL,0xc002002010020042ULL,0x0,0x0,0x2000400000400200ULL,0xceca3f1fd1e1aae8ULL,0x2108802110882000ULL,0x1410100a01000272ULL,0x0,0x4000001000002000ULL,0x10616000024a088ULL,0x2f1945c3c1e42618ULL,0xa072e8012da2eeULL,0x4808050040012040ULL,0xd40c284291e40922ULL,0x0,0x28880fd080019940ULL,0x2880100000ULL,0x20ce83601104812aULL,0x148c0140b0800a00ULL,0x950d60000480210ULL,0xa0a989a90402342ULL,0x870ace37d9492646ULL,0x170077991a4886aULL,0x8002204811800042ULL },
	{ 0x8a527aea6728f765ULL,0x84800f4648240400ULL,0x8460100884100ULL,0xbd53e268096ee764ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x69ae4f73e5fe7563ULL,0xa9e3f66f08202744ULL,0xddcf9ebc8ffe7d27ULL,0x9b3a8fd71f3e6f62ULL,0x7cf6f8e25fac6746ULL,0x53fdff79af2c7d22ULL,0xffb3377edffce760ULL,0xe7a6dfaacfea7467ULL,0xed5fdbd71dba6562ULL,0xdcb3f76df74af763ULL,0xd5edb7da5b76ed63ULL,0xdaa6757d0d76f761ULL,0xe5c7ffbe5ffe7560ULL,0xea7fbeddfbae6763ULL,0x3fbbeff98ffce563ULL,0xfffdfffffffeff67ULL,0xd9666c6e432a6402ULL,0x9fdf7379f2e7560ULL,0xecefbffbcbfc7763ULL,0xbffecffb8d7e7562ULL,0xf5d4f6f11b7c7702ULL,0xeb9e3d7117fe7567ULL,0xeb7acb655d7e7762ULL,0xefdbfbfffff87762ULL,0xe7ffcfd7ebfaff62ULL,0xf5f7bfff7ffe7763ULL,0xfdbfff72cbfcf762ULL,0xcbb4ff7fbf6ef767ULL,0xdfdffe37ffeeff62ULL,0xaeffdfdfffeaf763ULL,0xf7fff4eecffeff67ULL,0xfffffff3fffef762ULL,0xfbf7ffff0ffef747ULL,0xeff7feffffecf763ULL,0xffdfdfdfbffe7f67ULL,0xbfff7ffbdffcff67ULL,0xdffaff7acbfa7f63ULL,0xffb7ff7ffffeff63ULL,0xeffffffffffe7767ULL,0xefffffdffffeff67ULL,0xbffbffffdffef767ULL,0xfdfffffbdffeff67ULL,0xfffe7dfffffeff63ULL,0xebf7ff6feffef767ULL,0xffffbffbff7eff62ULL,0xbffeeffbefca7763ULL,0xff7ffeffdbfeff66ULL,0xfffefffbfffeff67ULL,0xfbffffffdffeff67ULL,0xefdfbfdffffe6767ULL,0xffbfeffffffe7763ULL,0xf5f7feeeeffe7f61ULL,0xdfdbff73fb7c7f67ULL,0xfbbffffedffeff63ULL,0xdfdefebfdffe7762ULL,0xffffffdbfffe7f63ULL,0xbffffffedffeff63ULL,0xbfefffffdfeef766ULL,0xdfbe7fff5ffef767ULL,0xcfffef3f7aecf763ULL,0xebffdfdbdffc7f64ULL,0xf4b77fe3cffe7f60ULL,0xffeffffb5bf87766ULL,0xdebfbffecffa7f67ULL,0xcffeff3fcf6ef766ULL,0xbfdfffcfeff8f767ULL,0xafb77ffbeb7c7f63ULL,0xdffcbfffdbfcf763ULL,0xfe9e2f6f9a28f560ULL,0x8be7ffeecb7e7f66ULL,0xefff3fffcffef563ULL,0xbf9ffffbdffe7f63ULL,0xffffffffeffef767ULL,0xfffefff3fffef767ULL,0xfb96bf2e8b6cf763ULL,0xc5e3be900b3c7d63ULL,0xffbfcffbfffe7f62ULL,0xf49dffefdbfa7765ULL,0xa204008600520ULL,0xdb7ceb7f0f7e7f60ULL,0xdfd3ff3f5bfe7722ULL,0x235a9fd98bfc7f64ULL,0xdcfff7fbf8fc7563ULL,0xdbfffff3dbfe7f67ULL,0xfabcfff6ed7e6763ULL,0xffffff9ffd7e6f67ULL,0x6bffd3ddcb68e762ULL,0xb4925ffec9f87f62ULL,0xfb3ef6467bee7762ULL,0xeeb027df1feeee65ULL,0xc7cff3bd4b6aff63ULL,0xcc77a2c0bcc7502ULL,0x463df4e17f206762ULL,0xf1e7ff7ba2fce767ULL,0xfaeeef68ffea7762ULL,0xcaedfbaedf7ee725ULL,0x8675e50096e6626ULL,0xeffccfd97f7e7763ULL,0x90dc766a8f324760ULL,0x5dbe2f62efea6526ULL,0x318ce92e817a7723ULL,0x3120165082020ULL,0x64904d081d4a0524ULL,0x1411502e05b80702ULL,0xb214e9606a486722ULL,0xd3b16e49ad4c6c62ULL,0xc386b3a78f746742ULL,0xaf084b420a2ce562ULL,0xf6a6b6e01f2c6d64ULL,0xf5ebf77bff686f66ULL,0xfbbc673d0f7e7f23ULL,0xef5ee70a7bbe3560ULL,0xcfcdf1a8dc84762ULL,0x1dbe4dbb9f6ab762ULL,0xf96b2ffadfec7f63ULL,0xdaffaf660fca7d62ULL,0xc955eb39caec7566ULL,0x69ccc1401ac4342ULL,0x97234ff6dfa8e541ULL,0x502422c000880160ULL,0x7b63df0b3b82766ULL,0x400c400650002400ULL,0x4818160700704562ULL,0x205c107000500ULL,0x55b452028840a720ULL,0x560472015026402ULL,0x58600002002100ULL,0x800b710100ac6523ULL,0xcc80002404500ULL,0x14e005ec03320720ULL,0xb060420010a4063ULL,0x4100008400000ULL,0x8100000000600ULL,0xb80a8b13c8bc6562ULL,0xf09ce4f305baef23ULL,0x43682e2101885703ULL,0x7020040019084400ULL,0x8000c0408c80000ULL,0x60000008004000ULL,0x2100080100ULL,0x817026010b424160ULL,0x8c008f404dc04140ULL,0x80b28a224d026001ULL,0x284417b0092a4160ULL,0xb47c4ea3cdea7703ULL,0x100502800604402ULL,0x1a0465220b386362ULL,0x81c3b88101324722ULL,0x24c2028000082420ULL,0x21884d5003487c20ULL,0x1a1296e92b70fd21ULL,0xe210296a89603460ULL,0x8100021007000600ULL,0x607a9280198e2720ULL,0x27be4d40ca30a566ULL,0x61c906e00a006743ULL,0x912abd60d9207760ULL,0xadabf3698f727d63ULL,0xffffffbb7dee7f63ULL,0xffffffdffffeff67ULL,0x37b4ffe0e9ec7562ULL,0x2124ae644d8e4242ULL,0xc69009e549466401ULL,0xedfffe9ddf786f62ULL,0x3bde7fdfc76af541ULL,0x553e77bab178f561ULL,0xdbfdfe63bd6e7f67ULL,0x1aa40f48018a4461ULL,0x1d3930a4a684562ULL,0x40923e10e8682520ULL,0x4324af559727462ULL,0x10820069a1007660ULL,0x5112020401024161ULL,0x88008268c0202444ULL,0xa000000000240000ULL,0x209a47c309084163ULL,0x404000000880000ULL,0x0,0xa2141a0000206541ULL,0x40478239ca48e400ULL,0xfbaa8ffb5bbc6724ULL,0x180c44200487120ULL,0x50a67543df046603ULL,0xd800e0200000a520ULL,0x48d7ff8a5f2c6d27ULL,0x6fdf5ddb0fbe6f66ULL,0xfcf77bec9b347502ULL,0xf9ea3f6953e86f62ULL,0xcbac7f5648f86522ULL,0x8222020840a140ULL,0x805e540308342760ULL,0x808054e000304642ULL,0xc072ac5009247702ULL,0xda9e2e3285406762ULL,0xef66273b5d3a6720ULL,0x4ee5fed94f4a7562ULL,0xf30154091ac6500ULL,0xd04496d84be86743ULL,0x202414480404404ULL,0xdbddf19ccbfcfd61ULL,0x62f11fd19bac7f60ULL,0x6038c0800f004400ULL,0x2004000000880000ULL,0x896315008647001ULL,0x2b05c1360d2c7502ULL,0x9ce2f2d28bca6762ULL,0x8008000402000ULL,0x0,0x40008000002ULL,0x529016200ac84522ULL,0x480000000400ULL,0x14440000402000ULL,0x4c0400010404400ULL,0x41b8042009a44900ULL,0x200610001000e00ULL,0x211001003000ULL,0xa344e4889022400ULL,0x5c0209004a00e700ULL,0x0,0x301c3de9cd4efd02ULL,0x6980f60602a4a160ULL,0x95f7ea02bac4760ULL,0x7aa5dfa9b6c7500ULL,0x4426a32919005500ULL,0x4006f4230b342e00ULL },
	{ 0x8110a10a10820560ULL,0xecbbf788fb16e68ULL,0x3410226304080d61ULL,0x5082308004000541ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x517426fc83ba8f20ULL,0xfbdfe47617c62f60ULL,0x0,0x44013202001a0920ULL,0xbfff5fd797fe6d69ULL,0xf2bc5efe1fb66d60ULL,0xe9fe7f7b9fff3d61ULL,0xf6982d3c0f9c8d61ULL,0x4fdefe5a0fb66d61ULL,0xebdcd55396b62f70ULL,0x3df95fff8fbeff60ULL,0xf9f7f7fb9ffee770ULL,0xfabc3ff696f62d60ULL,0xdbe1f39a86b6fd68ULL,0xffffbf5f1fde6d60ULL,0xbe9b4ff912f6a748ULL,0xfffff7fe17feff79ULL,0xba3405621286a460ULL,0xc8d7fb2e07b26d40ULL,0xa7fb3f4b9794ad60ULL,0x96bb47eb87de6d71ULL,0xd6bdfa7897a7ad60ULL,0xfff6fde2169e7d60ULL,0xfbcdf62d97df0f60ULL,0xeae3d3ff16ba5d60ULL,0x2bbecff31befcf78ULL,0xfdffffef97fe7f61ULL,0xfffeff7797f6fd69ULL,0xfbbeff7e8efe6d71ULL,0xcfffffbb8ffeef61ULL,0x7bffdfdb9fbf7f70ULL,0xfffffee99ebfff70ULL,0xffefff7b96b67f71ULL,0xffb73ffe97f7ff61ULL,0xfffffebe9fff6f79ULL,0xefffffdf9fffef69ULL,0xfeffdfff9fff7f69ULL,0xfffff77a9fbfff70ULL,0xdfbe3fff9fff7f69ULL,0xffffffff9fffef78ULL,0xffffffff9ffeff69ULL,0xfeffffff9fbfff61ULL,0xdffffffe1fdfff79ULL,0xfffe7fff8efe6d61ULL,0xdbefffdf9efe7f71ULL,0xefffffdf9eff7f70ULL,0xdfbfffff9fbeff78ULL,0xfdfffffb9fffff79ULL,0x7fbe7ffa1ef7ff79ULL,0xfbffffff9fffff79ULL,0xefffbfff9ffe7f78ULL,0xffff5fff9ffeed61ULL,0xf5f7feff9ffe7d78ULL,0xdfbffff39ff6af69ULL,0xffb7f7fe9fbfef61ULL,0xcffffeff9fffff68ULL,0xbfffdfdb9fffff70ULL,0xffffffef1fbfff79ULL,0xffffffff9efe7f78ULL,0xffb7fffe9fbffd69ULL,0xcdf7ffbf1eff7d68ULL,0x8bfe5fdb9fdf2d70ULL,0xf6bff5ef8eb76d68ULL,0xfffefff79bb6ff79ULL,0xffffffff97befd71ULL,0xefdfffff9fffbf60ULL,0x5fff7ad88ebe2d61ULL,0xb4bb9ff79fdf6f71ULL,0xf6fdbf6d1fbe7f68ULL,0xf9127e3797866d60ULL,0xebdefffa9fbe7f60ULL,0xe9df7fdf1ffe7f70ULL,0xffbfefef9fdfff79ULL,0xffffffff9fffff79ULL,0xfffefff39fffff79ULL,0xdbbcef7f9ff3ef79ULL,0xf5cbbff78fb7bf68ULL,0xfebacfdb9ef7ef68ULL,0xb59e7c688e9e4f60ULL,0x80202c4100022c20ULL,0xda84ffaa97bc6f48ULL,0xcfff7a1b9f9f3f69ULL,0x93fccfdf07fe6f61ULL,0xf4d0fe691fb67d68ULL,0xfba6ff710e96c578ULL,0xff9a7f7f1e972f69ULL,0xffdbfff89fbf7f70ULL,0xabded7db8abe6f68ULL,0xb6bfcdeb0fdedd60ULL,0xfbefee6b9bb76f61ULL,0xdebe3fff16bf7d69ULL,0xc5cfff6f879ead61ULL,0xab7ffedf9eff6f60ULL,0xf7f67de70e366d70ULL,0xf7ffb6de8fdf6f69ULL,0xf7ce3de59eff6d60ULL,0x8bd3e36796b76f79ULL,0x6e5613d61f136d58ULL,0x78ccf80128f2d60ULL,0x7d1620a983960660ULL,0x4692af2086322d61ULL,0xeb50e6390ec22f60ULL,0x4006060800104c60ULL,0x2799cf9392c6df60ULL,0x86b636e106d26d40ULL,0x5ab00420108e0430ULL,0x8aa0a62c029f4e20ULL,0x9e4ffabf879f4d71ULL,0x686d485005954c60ULL,0x14b776e507be2f71ULL,0x81787f7917bf3d60ULL,0x6a99262c00324521ULL,0xfdcffb6e97f76f69ULL,0xdbfabd4b17e60768ULL,0xb4ff5fec8c3f5d60ULL,0x95a704c206da4a60ULL,0xde3eaf730adf2f70ULL,0x6cdfdf7d96977f60ULL,0x6d7efd5a16deaf71ULL,0x3fba7ffa82b71d60ULL,0x756c276016963d61ULL,0x208080c00060840ULL,0x908c6a602324c60ULL,0x118000100000ULL,0x4400008000400ULL,0x80cd030082920d00ULL,0x1e057002928540ULL,0xa100f4000b930c40ULL,0x4002000004000040ULL,0x1244b0003900420ULL,0x420165004942900ULL,0x0,0x0,0xc9c2940190861b60ULL,0x2000800002000400ULL,0xd012f69e12d72d60ULL,0x8002004000100ULL,0xb010010003802200ULL,0x0,0x101000000400ULL,0x2204e004040120ULL,0xb762e7701e96c760ULL,0xfeb6a7f51fff0f70ULL,0x854efb171ad61d70ULL,0x90dbf60ab72e28ULL,0xf7b3e5f11eba3f60ULL,0x9edc76be1fdf5f60ULL,0x19d2f560169f3f40ULL,0x8050a20000040420ULL,0x6f4638000fb32d40ULL,0xa79a45c313962560ULL,0x84645461169e0760ULL,0x7086054202090d20ULL,0x8bced62d8ad22f68ULL,0xc2a3a04000960f48ULL,0x120e032000022960ULL,0x3082062800b02e30ULL,0xcd70fd120bf94d68ULL,0xfebdffff9eb6ef71ULL,0xffffff3f9fffff78ULL,0xffffffdf9fffff79ULL,0xbeb01ee89e9e4d70ULL,0x1cb1fc012964d70ULL,0xaa327fae068a1f60ULL,0xfdafc820bda7f20ULL,0xa70354f829e2d60ULL,0x7cb44be81eb62761ULL,0xd9f6af5c02ba7f60ULL,0xf8beaf360796ae60ULL,0xc593a20b02a66d40ULL,0x714fded98e927920ULL,0xb4ab07fe069a0d61ULL,0x40b44002808e20ULL,0x6000004100000ULL,0x8030200201840400ULL,0x18a016c010923920ULL,0x31101b9608162d50ULL,0x400220800b800000ULL,0x67140118a0c20ULL,0x946a23107304c60ULL,0x60820800300140ULL,0x2020001200122800ULL,0x1425002006060501ULL,0x812885a004d60160ULL,0xc200860c02f20700ULL,0x9c21e0918820f60ULL,0xeb0e5f529fb6dd68ULL,0x946762600e975f40ULL,0xeb6fff7b1fdebd68ULL,0xffaa2e4b06becf69ULL,0x448a080008020c01ULL,0x1204510a82100540ULL,0x8811457012840d40ULL,0x4008d20287880840ULL,0x72362ec207b61540ULL,0x6fccf2068bb72d60ULL,0x63dbbec992aa1d60ULL,0x100a3180100940ULL,0x98bdff680c975f68ULL,0x1204302402860740ULL,0x8944f10612100d70ULL,0xeb77dfde8ef66d61ULL,0x431c4a1004964b40ULL,0x880010800000ULL,0x80200000200040ULL,0x8a50100010860140ULL,0x41c7be5886063c60ULL,0x8006002000024140ULL,0x40528000160120ULL,0x488d56186126c28ULL,0x4200430806800d40ULL,0x0,0x2010028002000540ULL,0x2a020002020500ULL,0xa0360020a1460ULL,0x8840032003802620ULL,0xc00000840000ULL,0x0,0x14200200900140ULL,0x88778e588e1f3d20ULL,0x283200100921540ULL,0xcddff7b617dfef61ULL,0x935fde630ea6bf48ULL,0xb0822a6992b00d40ULL,0x5024100004000448ULL,0x4500480302168040ULL },
	{ 0x4181c0f090080724ULL,0xcfffb40a9c9c7d80ULL,0x46124d82c4206040ULL,0x1088200c0100080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41c0900012041900ULL,0x40184f73b3020304ULL,0xb49057e9548e2cc3ULL,0x0,0x0,0x2723a1396942945ULL,0xffffffdfffbe3fffULL,0x34f374e0d41c3ce0ULL,0x99f6b761d78e2fc2ULL,0xba9c2d79f39e76c4ULL,0xd0420f38d69434f8ULL,0xfffcddded79647c6ULL,0xf4f77dedf79e3f46ULL,0xfdff7f6bf69c6ffeULL,0x1a8e3de5da325cecULL,0xd9f7df69d60e77feULL,0xeb7fdfc7d83c7fe5ULL,0x832207f0d79e3b34ULL,0xd77fff435f3e2fd4ULL,0xde94add7f29e7ff4ULL,0x8adeffeaf6be2fc5ULL,0xffd6fe3bf79c6d46ULL,0x7f9dcff3fe247df6ULL,0xf1fff76af7162fceULL,0x6fa67f419f1c7fe6ULL,0xdb74ef4efe0c378dULL,0xcce3b61bf79e6fc4ULL,0xefffdff3f7be7dfeULL,0xffffffefffbe7fffULL,0x9bb83df2d79c7decULL,0xf39cff7fd79e7fe6ULL,0xffffff1fff9e6fffULL,0xfbffdfdff79e6ff6ULL,0xf4fff6eafe9e77eeULL,0xdffd2772f79e7edfULL,0xfbaefffffe9c7fe4ULL,0xffcfffbbf79e3fffULL,0xefffffdff7be7fefULL,0xfffeffffffbe7fffULL,0x7bfeffffdfbe7fffULL,0xfffebfffffbe7fffULL,0xeffeffffff9e7feeULL,0xffffffdfff9e7ffeULL,0xf7fffffff7be7ffeULL,0xfdfffffff7be7fffULL,0xffde7ff6ffbe7f6fULL,0xfbffffbfd6be7fe6ULL,0xefdfff5ff7be7ffeULL,0xdfbefffbffbe77eeULL,0xfffff7fbffbe7fefULL,0xfffe7ff3ffbe7fffULL,0xfbffef7fffbe7fffULL,0xee5ebeffffbe3feeULL,0x7fbf7ffdd73c7fe6ULL,0xfdffffffff9e7fffULL,0xfffffff2ffbe7fffULL,0xffffef7ff7be6fffULL,0xdffffebfffbe7fefULL,0xbfffffdbffbc7fefULL,0xfcbffffff7be776fULL,0xfbfffffbf7be7ffeULL,0xffe67ffedfae77edULL,0xefd7ffbedfbe67e6ULL,0xbfffdfdadf9c37ffULL,0xfdb77fe7df947df7ULL,0xf36ffffbdfae777eULL,0xffbeff7ffebe7ffeULL,0xfbd7fffff79e7ff6ULL,0x8fff7ffbd7be7ff4ULL,0x9dbb5ff5d39e75dcULL,0xd2ed7ffc57b83f64ULL,0xffbe6f7ffe9a2d67ULL,0xe1f7f3eefebe7ffeULL,0x3fd73fdbffbc7fdeULL,0xffbdfffbfebe77eeULL,0xffffffffffbe7fffULL,0xfffefffbff9e7fffULL,0xbbf5ff3edf9c7fc7ULL,0xcffffedbd71e7fefULL,0xbbbeffd3d79e7df6ULL,0xbabd57c2d7a463e4ULL,0x92bebd10bf1c2f44ULL,0xf9fce3fed79e3fe4ULL,0xeef6bffcf71e7fa5ULL,0x6fdfdfd3f21c6555ULL,0xf4f55e7ed6be376dULL,0xd5beff62d6966fc4ULL,0xffb677fefe0e77eeULL,0xffdfff9ef68e7fceULL,0x6f5ff5d9d71c15deULL,0x8efedffbff3e3de7ULL,0xd37ee778f61e2d9eULL,0xffbe37ffd79e79b6ULL,0xcfc7ff3effbe2ff6ULL,0x6cdcde89f60c3da4ULL,0xf6bbfffbd7be7defULL,0xfd64f7fbd6be6fe7ULL,0x6fbebff2d6ae2dfcULL,0xe9ffff3ef7be6ffcULL,0x4fc71e50d69c3780ULL,0xcdbefefcf39637e4ULL,0xc477becddf0636c5ULL,0x65dc3531e59c6580ULL,0x8080730ef6080d16ULL,0x8d823e85571e04c4ULL,0x4e975d587f9e2464ULL,0x79b3c6e0f2943502ULL,0x2110857056281744ULL,0xf7b4b36ef60e2fc6ULL,0x7bff9f17ff1e5fa0ULL,0xab9cdddefe9e77eeULL,0x521e76a056802946ULL,0x69627763d6063fc4ULL,0xd696ebe6f3a625c7ULL,0xcfdffffed79e7fefULL,0x6fff3fdbfe9e7d84ULL,0x28b05fe5d61e3d44ULL,0xd06b09e0f484664cULL,0xced92b1195180c4cULL,0xcdf2722cd6be2bc4ULL,0x25836a1856063d14ULL,0x741adca3f4942b30ULL,0xdaf6bd76da9c7feeULL,0xdf9e3af7d49a74e2ULL,0x9048622c60c01a4ULL,0xc04b10409a042d02ULL,0x2100630820400a0ULL,0x3d5490609204006cULL,0x20004010000200ULL,0x42942a029202402eULL,0x1680020034000240ULL,0x0,0x6404d0010202220dULL,0x58c890092001506ULL,0x0,0x200082000000ULL,0x1800106c2000090ULL,0xfef7feffffbe7fffULL,0x4565604ad20c1194ULL,0xce10020654040282ULL,0x108040000080ULL,0x0,0x80040000080400ULL,0x100086044802000ULL,0xdfb63f7ed7bc7d7dULL,0xcb624b09c3126704ULL,0xe9775bd6f5b43d60ULL,0xb7bb65eaff2e3ee6ULL,0x11c1540c901c2ca1ULL,0x40200000802000ULL,0x802f800540e0240ULL,0x6110040444802204ULL,0x91144f01548c1564ULL,0x14c860802023c00ULL,0x43fc7d71d71c6bd4ULL,0xd9f0ff3ed79c7b34ULL,0xccd9dfc7d6862b94ULL,0x21044f58d6184784ULL,0xc0140044220040ULL,0x80272326182405ULL,0xfbb8ffdf569c3f75ULL,0xdfffffbfff9e7fffULL,0xffffdfdfff9e7ffeULL,0xd0944ce074823500ULL,0x5900496156043202ULL,0xe8946674f71e2564ULL,0xf907c95ad6be57ecULL,0xaf65d9ad63475c6ULL,0x9cb7dfeaf7983e24ULL,0xd4ccb6d9d68e3ff4ULL,0xb7b7355a77943dc4ULL,0x89c5f973821c7f4dULL,0x834fc84013140184ULL,0x8420141091001420ULL,0x404500040082400ULL,0x92160810502000e0ULL,0xc8a0c328b20625a4ULL,0x908002004c000400ULL,0x48044000000ULL,0x0,0x400200a082000104ULL,0x3546510d2000552ULL,0xa102044802000ULL,0x0,0x4200200040ULL,0x6100002000004ULL,0x1080081012000006ULL,0x40cf3e008b902c01ULL,0xbfffffdfd7be7ffeULL,0xb2b4f668de863d76ULL,0xdb76f772df1c3eecULL,0xdf7e6fd7f2342d66ULL,0x8f468a00c28e0542ULL,0x81e5c0380bc6144ULL,0x2080000040080ULL,0x13a0d658d0b42544ULL,0xab2058000704ULL,0x1503588c7020404ULL,0x4400149c09e0040ULL,0x20406000040080ULL,0xe3a82f32fbb44f4eULL,0x6fe6f7e19aa4276cULL,0xc800800002001404ULL,0xed635fcadb9e7f24ULL,0xc338cf53f63c2cc5ULL,0xc00040020000ULL,0x2008140000081504ULL,0xa155832ed6903d04ULL,0xcc623608ba002d04ULL,0x900485030040480ULL,0x4080942bd20040ecULL,0x202c4120420e3404ULL,0x8000004002000004ULL,0x0,0x18480090100800ULL,0x8401002010001400ULL,0x100100002020000ULL,0x34000244c1802010ULL,0x2480410810000160ULL,0x0,0x94614822b2141624ULL,0x91980000400c00c8ULL,0xca886f16d2946116ULL,0x8208f00073120404ULL,0x3093250db142504ULL,0x1431f21093146b2cULL,0x52658570c7141180ULL,0x74220560d6122540ULL },
	{ 0xcbb1d720a380673aULL,0x4910408000020cULL,0x880c15201126220ULL,0x8600c0002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x88000080000010ULL,0x0,0xa352129b8ba0666bULL,0xa79e8fd4c390271bULL,0x0,0x0,0x0,0x5bcb7ebfabbc6f7aULL,0xffffdfdfabb473efULL,0x94975dea89c262cbULL,0x3228fd728981231bULL,0xda347ebeab8064e2ULL,0xdc4ff61e0bd142e3ULL,0xefdfffdfcbfe67ffULL,0x35feffedaab83777ULL,0xfb6feff0ebd873b7ULL,0xdeb63f7fc1b06feaULL,0xcd87b35a83bc669bULL,0xefdbffdfc3b46ffbULL,0x86bbfefbabfd67abULL,0xdfbfbfeecbfb6ffbULL,0xdee677ffa9d977fbULL,0x8bf5fbf7ebdd7efbULL,0x6bc3dfffcbb86f6bULL,0xcfaedffbcb9277bdULL,0xfffceef9ebf973f7ULL,0x4f3e7d62cbf762bfULL,0xfb8bf37e8bfe277bULL,0xcfdabeba8bdc63fbULL,0xd6ffcd7e8bfd673eULL,0xffffffefebff7fffULL,0xfffefdf3cbbd7bffULL,0xffbfffffabff737fULL,0x4fffff3febfc7fffULL,0xbbffdfdf8bfd7bffULL,0xfeffffeeebfc7fbbULL,0x9bf5f7f24bfc7bbaULL,0xfbd6fffdcbda6bdfULL,0xdfffffbfcbfd77fbULL,0xfbfedfdfabff77ffULL,0xfffffffbebff7ffbULL,0xffffffefebff6fffULL,0xdfb7bfffabf673fbULL,0xefffffff8bfa6ffbULL,0xffffffffebfd7fffULL,0xbfffeffbcbf46fefULL,0xf7fdf7ffabf76fffULL,0xfffe7ff7abfc7fffULL,0xfbfff37febfd7ffbULL,0xefdfbfffcbfe6fffULL,0xffffefffebfb7fefULL,0xfdfff7fbcbfe63fbULL,0xffbe7f7babff7fbfULL,0xfbfbffefabbf7ffbULL,0xeffbff5bcbfd776fULL,0xffffcf7b8bbe7f7bULL,0xfdffffffebfc77ffULL,0xffffff73abff67fbULL,0xfbb6ffffebdd6ffbULL,0xfddfffbfcbff7fffULL,0xffffdfdecbff7fffULL,0xf5bfffffabfc67fbULL,0xdfefffd3ebde77fbULL,0xdffe7fffcbfd7ffbULL,0xffdfffbeebdf6bffULL,0xefdffe9babb07afeULL,0xb7f7ffef8bbb77efULL,0xffefffffebfe777fULL,0xfffebffecbfe7f7fULL,0xf7fffff28bf867fbULL,0xefffbfdfabfd677fULL,0x37f3cffbe9cc6f3fULL,0xf3fdb77a8bfd679bULL,0x1f4c2d2509e466b8ULL,0xdbddff3babbc7b7bULL,0xeefe5ff7abec677bULL,0xbffeeffbebfa679fULL,0xf5fffffeebbb6ffbULL,0xebfeff7acbf56fffULL,0xe9bce37bcbf877b7ULL,0xbfc3ff7aabbc6ffdULL,0xaf3f7ff28bbd6ff3ULL,0xf1933fe2ebf96efbULL,0xdeba95728b9a63ebULL,0xff9e6b5eabfc77fbULL,0xcfd7fe3fcbf877fbULL,0xbfdfdddbabff6efbULL,0xf6fedefdcbf573b2ULL,0xffec77f08bf676baULL,0xdffe7ffecbbf67efULL,0xedffff16ebf977ebULL,0xef9cdf5b89bc63efULL,0xbafa7fe2abf86b7fULL,0xfdfef7728bbd77fbULL,0xfeffbffdcbff77fbULL,0xcfeef3ff8bb873ffULL,0x2fdfff918bb0677bULL,0xb7ba5f76cbb86ffbULL,0xffffbffb8bbe67ffULL,0xef9eef738bf963feULL,0xebf7f77fcbd977faULL,0x6df75eb8abdc6bbfULL,0xf1ce7e1ab950312ULL,0xbd55ae6c8b98637bULL,0xd7be3f704b9d6bb7ULL,0xc2d0e77e0bb56697ULL,0x8fdb1759c9d07213ULL,0xeab6cfe98bf87fedULL,0x20805040a8880010ULL,0x351a7530a9a242afULL,0xf995eb1e88b9727bULL,0xcd4eb621c9ec3239ULL,0x1fdfff5bebbd6fdbULL,0xc8bdd4a98bb267b9ULL,0xb1eeb76acb9967f3ULL,0xb420674ccb9c2a43ULL,0xdfcfffffebbc67fbULL,0xcadedfdecb7e6efbULL,0x127e5feae9b8763cULL,0xb1405648e9922203ULL,0xcd208a22019003a0ULL,0xf156d32d8bb867bcULL,0x2b5d9d50cb146768ULL,0x4f3a4feac99262cfULL,0xd93cf67bcbf367abULL,0x28c3c40898c4218ULL,0x8990f694890866caULL,0xe2801fd989f26369ULL,0xbec88dcb08c063e8ULL,0x4c54130809042322ULL,0x42080c1080002460ULL,0xaa04230f838c2212ULL,0xaec91a0089881302ULL,0xc171e18c190225bULL,0x480164081002028ULL,0x8000000000000208ULL,0x4004100200101200ULL,0x0,0x2f08c0880290220aULL,0xfef7feefebfd7ffbULL,0x177b360abc81310ULL,0x120002800000ULL,0x881900002102201ULL,0x8003400a400000ULL,0x20b036189882042ULL,0x412090000180012aULL,0xf89caf6589b266fbULL,0xc7d4d5b3abf0776fULL,0x5435e188ba067e2ULL,0x1a2000c82802400ULL,0x8108044881004206ULL,0x10a42d40880c6388ULL,0x8012b90282902710ULL,0x40161888192612ULL,0x2080c0188500240ULL,0x5116dccaaaa067a1ULL,0x585ce581abd02619ULL,0xc9f4cf3eabdc43f0ULL,0xc65fb61b89147f32ULL,0x833accf1a19463aaULL,0xa4fe77688bd8437aULL,0x8001a05003016002ULL,0xc79635edabd067bfULL,0xeed6763fcb8b67f3ULL,0xbff7dfdfabfe67ffULL,0x100004600b100000ULL,0xc838a202819042b2ULL,0xcbc7fbf28bb9677bULL,0xe047ee32c9b1720aULL,0x6bffdfceabbc637bULL,0xa5c61563819d2f92ULL,0xdbfebf48abdf63efULL,0x8c2044809044200ULL,0xc1d4d72ca95a7602ULL,0x26a304803ac0272ULL,0x462014003920002ULL,0x8000300a01000060ULL,0xab0245640bb0522aULL,0x8a81134301b22322ULL,0x50001280800010ULL,0x0,0x100100002400ULL,0x1405f080a01269ULL,0xbb04e7298335774bULL,0x2100080000010ULL,0x12c000009080008ULL,0x24014081020008ULL,0x90090560030040a0ULL,0x4582e1488b966631ULL,0x445bda3981a96b9aULL,0xabffdfdfebd56f73ULL,0xd4b07cf98bba225bULL,0xd7f7f773cbd92337ULL,0xfebc3f5fabfc7b9bULL,0xe9c6f62be382760bULL,0x21000000280ULL,0x2a002000202120ULL,0x816407208a806385ULL,0xc600206009604240ULL,0x444c609409984270ULL,0x1100000202000ULL,0x2020000001000020ULL,0xd3ea6f62ab9c2733ULL,0x4b96f5e48bde7699ULL,0xa144d28c0bc841f2ULL,0x2f7717d0abba7f3bULL,0x9f3fd77389f46b0aULL,0x1c56000b40420aULL,0x200200000000120ULL,0x5b0463030b90260bULL,0xa20290e8a390270aULL,0x28ac4f62a9107369ULL,0x201022e080002051ULL,0x294a75600b100228ULL,0x0,0x4100400442802000ULL,0x20c85410be02208ULL,0x0,0x800a240000890303ULL,0x81002a0080006080ULL,0x6000020809004000ULL,0x0,0x2000004000004080ULL,0x2008426080800041ULL,0xfe382f70ab80641bULL,0xcce4d3a709f06231ULL,0xefffbfdb8bfb737bULL,0x35114b12aad1429aULL,0x712cbff08bdb66bbULL,0x4bce3d430b946af8ULL },
	{ 0x10ad2008164c320ULL,0x4682200c01540098ULL,0xf29cc5b2d034513eULL,0x8400d01080400942ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4a8a1205c8d4c30eULL,0x210060080000124ULL,0x4200000100ULL,0x831a6ff009844212ULL,0x8bbe377df9dc439aULL,0xd9d6fa5fabf4db3eULL,0xfbffffdfe3fddffeULL,0xbff5dee9d9f5d77eULL,0xc94822420144420aULL,0xfbbf6d78d9d4d92eULL,0xafefff4a73ccd72aULL,0xbbdfdedbd97ddbfeULL,0xfcfe7ffdcbb5cb3eULL,0xfffcfff8f9f5cffaULL,0xfabc7e6efb2cd38aULL,0xedc6ff7ffbfcd73eULL,0xaffefff5f9f4d5feULL,0xbfaaffdee3fddbeeULL,0xf37ff7f7dbddd7baULL,0xffb6aff7f3dddfaeULL,0xe3b9f3a2e9e5d53aULL,0xe9463e084bc4c03aULL,0xfbbefffbe375d7beULL,0xd4f6d7ebfbfdd7eaULL,0xf9fefffbdb55d77aULL,0xbbf4ef7edbddc3feULL,0xeffbfad3f9ddd32eULL,0xffffc7fed3d5d97eULL,0xf5f7f6fbfbf4d7beULL,0xd7beaf73f3fcdffeULL,0xfdbdff7cebf5df3aULL,0xdbfffe3febfdd7feULL,0x3fffdfdadb7ddb1eULL,0xf5fcf6fffbfdd3faULL,0xff7fb777dbf5d37aULL,0xfb95efffe3d5d73eULL,0xefffffbefbf4dbfaULL,0xffffffdffbfddbbeULL,0xfffffffffbfddfbeULL,0xffeffffffbfdd7feULL,0xfe9f2ffffbfddb7eULL,0xedfffffffbedd76eULL,0xeffffffffbdddfbeULL,0xbffafffffbfddf7eULL,0xfffffffbebfcdfbeULL,0xfffefff7fbf5df3eULL,0xfbf7fffefbfdd3feULL,0xeffffffffbfddf7eULL,0xf79eeff6e175d5feULL,0xf7fffeffcbfcdf7eULL,0xfffefff3f9fddf2eULL,0xfbf5eb0efbd5d37eULL,0xcad3bffff3f4dbfeULL,0xbfefcfffdbf5db7eULL,0xfcfffffffbfddbfaULL,0xffbfff70f9f5df3eULL,0xfbbeeffffbfcd7feULL,0xeff7ff77fbddd7feULL,0xfbdfffdffbedd7beULL,0x7efffefafbedd3beULL,0xfbee7f7bf37cdf3eULL,0xffbeffffdbf5dffeULL,0xefd7ff3edb7cdd3eULL,0xeb9effdffbfddffaULL,0xfdf77fe7fb7ddffeULL,0xfffffffefbfddffeULL,0xfebe7ffffbfddffeULL,0xdff7f7afe3d5db3aULL,0x495eff4fe3d5cb3eULL,0xffbbffb7ebdcc3faULL,0xd1fffeedcbfdc1aeULL,0xce9e6d36ebc5972aULL,0x8bfcfb2feb54db3aULL,0x7fdfff7bfbf5d35aULL,0xffbecfffebf4dbdeULL,0xf5fffffcfbfddf3aULL,0xdffe7ffbfbfdcbbeULL,0xfbc7fd7fd9edd77eULL,0xe4ebffebeb4dd3beULL,0x2b0ac75a0b55c12aULL,0x149e166d6344510aULL,0xb436a531484cd73aULL,0xdba0698c8b04c72eULL,0xcfcbf61bfbf5df7aULL,0x2bffd3dee3e5ddaaULL,0xd4d5bff8e3d4d37aULL,0xbbf6476afbe5d5feULL,0xefbc6fffe314dbbaULL,0xefe6ff3bf9fcd32eULL,0xaffeff53dbddd73eULL,0xb4ff776deb75db7eULL,0xda79fffafbddd33aULL,0xfeb6af7fe9fdd77eULL,0xc9ebff3febfdd5beULL,0x6ffebedb81fdd77eULL,0xdeb37b3acb4dd9aaULL,0xf9fdf5fafbf5d962ULL,0x7fbc3d7cc3d5c77eULL,0xebfffffffbfddf7eULL,0xeede17dde97cd1eaULL,0x21bc4472d360c12eULL,0x77763ae9d9d4d52aULL,0x12a5f605058510aULL,0xa9f4e776db74956eULL,0x2ec39fe3e174d12aULL,0x80a54e52c1e4d23aULL,0x70be44c88b44d56eULL,0xcab79f5282cdd32aULL,0x3a28ef3c61d5c38aULL,0x8dc3b8a0e344d13eULL,0x6f7cc19b0bccc2a2ULL,0xf0f4b6dfebddd1feULL,0xdb6f7f7bfbfdd3feULL,0x72902e145154c19aULL,0xdfdfff9ffbfddbbaULL,0xebffffdbfbfdd3daULL,0xb4b17cebf855d002ULL,0xa920065ac224591aULL,0x8c34097ef954da06ULL,0xa9c4d76be374535aULL,0x4a3e1f4481541336ULL,0x978a6f5188e5c83aULL,0xf7ecdf7ef3dcdb7eULL,0x5fae3dfcfb4dc37eULL,0x8372f3a88955c326ULL,0x27181cc0c950d182ULL,0xcb1cacebf374512eULL,0x84641642d8441122ULL,0x5404000080009008ULL,0x10201000000010aULL,0x431790c0a0d908ULL,0x2104400000804102ULL,0x0,0x1000051082001002ULL,0x20020000040020ULL,0x88e4c00003d4810aULL,0x2004000000420ULL,0x843046e2a3145528ULL,0x102004602004402aULL,0xd5066a4d0155833aULL,0x8210220201440428ULL,0x480000000120ULL,0x34200d60c854180aULL,0xc228a700c944ca8aULL,0xc8be3f6e6344c73aULL,0x4fdbf67ae9d4cffaULL,0xe5575edf6bd5c26aULL,0x86a01f61414493e2ULL,0x92e47464c025d12aULL,0x9c0437e009440392ULL,0x8016066881441212ULL,0x28c0924109d00302ULL,0x1aa084000800411aULL,0x77241258e1c4c13eULL,0x42c046260d44110ULL,0xb444b2380d4d102ULL,0x6cebba17c3d4d39eULL,0x92014ee3d14483f6ULL,0xfaf57eebdbfcdb7eULL,0x9bee7d60f95dd10eULL,0xdb8e2932f125d126ULL,0xdbe7f6b7c9f4db3eULL,0xad7f59dff384dfaeULL,0x9000c6684104d302ULL,0x5b70c7625380c31aULL,0xcfb5677ff1b5d73aULL,0xedcfdffbfbf4d3baULL,0xeadccfdfeac4d17eULL,0x8193ae94b0c410aULL,0x8908266b01d4c280ULL,0x36923104590005aaULL,0xcc832047080c028ULL,0x61f91ed1db4497baULL,0x23a2c501c1049532ULL,0x18341a428041d10aULL,0x2004002060000102ULL,0x8210000200804108ULL,0x0,0x0,0x900200440004c20aULL,0x5f366e60ea44c72eULL,0xcbd2c30ef815c13aULL,0x80cbf8186855c12aULL,0x4402480201000ULL,0x2000046001000320ULL,0x100850000804120ULL,0x891061288081cc24ULL,0x437ed619f154d24aULL,0xdffaefdae95cd7beULL,0x509cf6f8f3548376ULL,0xdbebe543d2e4d362ULL,0xf9280d50d064c128ULL,0xd04a7040304c22aULL,0x800300101040100ULL,0xc1400a00004c428ULL,0x94fa2f6259c5952aULL,0xca022dc28244c902ULL,0x5056d28281041320ULL,0x94e07416200d022ULL,0xd4ca4de0c154c33eULL,0x0,0x2af66d5009ccc760ULL,0x22822c0a04108ULL,0x8eb7ff50cbfdd3beULL,0x84156222c060110aULL,0x183300e0d2040626ULL,0x6aa66c627145c208ULL,0x9b0200044050510aULL,0x842300000000528ULL,0x28a420080040922ULL,0x447190200054030aULL,0x12a104000c44048ULL,0x0,0x0,0x200020000000120ULL,0x5014020201404048ULL,0x8000200001004000ULL,0x8800000000400002ULL,0x609080243044200ULL,0x37882510244c21aULL,0x8140000000ULL,0xc1ac2663f0545248ULL,0xc6862574d1e447b2ULL,0xcfeff3abf9d5dbaeULL,0xefffffdbfbfdd9feULL,0x31a4d3ee1d4532aULL,0xffffaffaebfdd37eULL,0xff76fffbfbdcd3beULL },
	{ 0x8000000200005000ULL,0x6fc97e39c282b7e2ULL,0x284442c280b120ULL,0xc00006c040422c84ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x422211c240e0a6ULL,0x0,0x0,0xa336a7208ad6f3f2ULL,0xfbe2eb7ec8dffe42ULL,0xcfffff9fcad7faf6ULL,0xefffdf5fcadfff66ULL,0xd6f77ce4cacff2f2ULL,0x8d6fef73cad2f7e6ULL,0x5a90af6ac082f2e0ULL,0x57c6b7be4ac7a662ULL,0x7b5fd75dcadebef4ULL,0xbd9f5ffbcac7fbe6ULL,0xfbffefffc2d7fbf6ULL,0x9eb4aff4cad2bba4ULL,0xeff7fa2ecadbfbe0ULL,0xede6fe96cacab3b0ULL,0xbffd5fe7cad3f3e0ULL,0xcd7fa7294a96f3e2ULL,0xfe963f7fcadff7b6ULL,0x89c46605c00ad520ULL,0x4b7c5e564ac38f66ULL,0xbebffdd3c896f762ULL,0xfdb6feefcadfffe6ULL,0xefbe3d7bcac7fff2ULL,0xd314fb36ca96f622ULL,0xefff97eecacffba6ULL,0x7fbfdffecadffff6ULL,0xd9f5fffacadfbfa2ULL,0xf7bae7704adbbef0ULL,0xdbbaf76acad3fb74ULL,0x5fffffa7cad6ff62ULL,0xe7dedfdfcacffff6ULL,0xd6b7d6f9cacff376ULL,0xfdfff779cacffbf2ULL,0xfbf6fff7cadebfe4ULL,0xcfffffffcadffff6ULL,0xffffffdfcadffff6ULL,0xbfbfdfffcacffbf6ULL,0xfbfeffffcacfffe6ULL,0xffbfbfffcadffff6ULL,0xffffffffcadffff6ULL,0xffffff7fcadffff6ULL,0xbfffffffcadffff6ULL,0xf7fffffbcadffbf6ULL,0xffdeffffcadffbe6ULL,0xebfffbffcadfffe6ULL,0xefffffffcadfffb6ULL,0xbfbfffffcadffbf6ULL,0xdfffffffcacffff6ULL,0xfffe7df3cadfff76ULL,0xfbdfff6fcadffbf2ULL,0xcff3ff78cadfbb22ULL,0xf3becff1cac7bfe6ULL,0xfdfff7fac2cbffb4ULL,0xdbbfff77cad7fbf4ULL,0xffffffffcad7fff6ULL,0xeffffeffcadffff6ULL,0xbfffffdfcadfff72ULL,0xfcffffeacad7fff6ULL,0xffffffffcadffff6ULL,0xfe9c7ffecad7f3f6ULL,0xffffff3bc8cfff76ULL,0xefffdfdfcadffff6ULL,0xffffd7ebcacffff2ULL,0xffffb7ffcadffff6ULL,0xffbe7fffcad7ffe6ULL,0xcff6ffffc8dfffe6ULL,0xefdfffdfcad7fff6ULL,0x1bbfefffc8dfffe6ULL,0xdfb4ae6e42cffbf6ULL,0x6ffa7df1cad3fb72ULL,0xcaf2fbeec0db97e4ULL,0xef57bfffcad7fef6ULL,0xff3fdff6c2cfbfe6ULL,0xf5fff6f9c2defff6ULL,0x47b62ff1ca4ef7e6ULL,0xfbfcff77c8dfffe6ULL,0xeafa3f5ecad7bb76ULL,0xadbfd77a8acff3e2ULL,0x553ed7e888c6fba2ULL,0xbd76ad624ac3ab72ULL,0xeb16ff1dca4bfa66ULL,0xfffefe3fcac7ffe6ULL,0xbe9effd7cad2ff76ULL,0xf6f4f6facac3fff2ULL,0xfdee7779cacfff36ULL,0xdbfe2d7ec2cefde2ULL,0xcfc7ffabcac7d774ULL,0x3f7dcbdacac6eb26ULL,0xfdbedfebcadffff2ULL,0x552fffefcacbff60ULL,0xeebeafffcacff3f6ULL,0xedfedf384ad7fbb6ULL,0x4ffe7a41c857f736ULL,0xaff1defccacbffa6ULL,0xf97dae6bcad6ff66ULL,0xfaaeade9cadff7a2ULL,0xebffffffcadffff6ULL,0xceff5e75c2d2fee4ULL,0x879cef8bcacebef6ULL,0x711c71f6cac1f7a0ULL,0xe73e3dfacacbfb20ULL,0xdb9ac7040206f362ULL,0x8d6bb201c8c6dae2ULL,0xfb79d3bc0c6f962ULL,0xf5e436eecadefa72ULL,0xd7baff73cadffef6ULL,0xde3e7f2aca4fbe76ULL,0x9c69e10c09297e0ULL,0x31c5d504a86a612ULL,0xb4b67eeeca8af736ULL,0x8fb6affcc287df36ULL,0xabb50626c0423830ULL,0xffdfffbfcacffff6ULL,0xefffbfdfcadffff6ULL,0xb4ba5dfa8081fb44ULL,0x12410164c280d020ULL,0x803cae794a1e93a2ULL,0x80dde2244287f146ULL,0x8c771a40ca431a64ULL,0x7005779884ffe72ULL,0x95fcb5e8c2d1f7b4ULL,0xfbbe6f7bcadffbe6ULL,0x8973fbf2cadbfff2ULL,0x7ff6be3cc253fea0ULL,0x148c4b3b4243a3d4ULL,0x5100765fc2cb9324ULL,0x8211000000a280ULL,0x19800a0d40d41206ULL,0x64e304688c4d140ULL,0xaa380da04a03a3a4ULL,0x80121480428a3002ULL,0x0,0x5006109048412e24ULL,0x802020400848002ULL,0xa29018400201d200ULL,0x2a380768480e1666ULL,0xc808077248c68270ULL,0x911c1964cacab310ULL,0x8809220800220a0ULL,0x50004448001020ULL,0x4221100c080c200ULL,0x80786770ca405224ULL,0xde122e78424ab2b2ULL,0xfab9f72ecac5fb32ULL,0x6777a09c2c4a2a4ULL,0x1d0203b980c0baa0ULL,0x240000008020ULL,0xdcb65b4e8a57bb62ULL,0x4081b2008080a120ULL,0x40161420c092b126ULL,0x8018010000410800ULL,0x9441a71008859122ULL,0x22a02dc04ac28a30ULL,0x4200f10382018224ULL,0xe6ca1f6bc2dff722ULL,0x9b8a49088087ba22ULL,0x51ffd67fc2c2efa4ULL,0xdfaf7f63cadbfff6ULL,0xc08244880c57002ULL,0xcdcbaab4c0c6eb74ULL,0xa201120042018400ULL,0x4a6d0404802f2e0ULL,0x92ec2739885a9a72ULL,0x8bdc13e9820af766ULL,0xcff6df1cc086aff6ULL,0xa59dd94242c2ae70ULL,0x94b1767942c2bff0ULL,0x586005d2c842e220ULL,0x26942d724a8ab462ULL,0xc7f07b9988cdb2b0ULL,0x2a4e3293020fb062ULL,0x819a42034a423610ULL,0xa9f8bb50ca86b1a6ULL,0x80021900400090a0ULL,0xb92026100022282ULL,0x4d0a1988ca02f3a0ULL,0x210c00000121020ULL,0x90b0064008c8b080ULL,0x69f276428a4ebe72ULL,0x41d292008202a386ULL,0x800093274a41b030ULL,0x130040000808020ULL,0x31000668c0001614ULL,0x0,0x8b08600208181600ULL,0xc152da3d82c39b60ULL,0xa21e5f93ca57bb26ULL,0xd6e77ee5ca53bfe0ULL,0x934f2773c88bf0e0ULL,0xe08034618085da02ULL,0x8080600100038060ULL,0x200000180004000ULL,0x808204400041a200ULL,0x404000008020ULL,0x4800002080ULL,0x4046000848020ULL,0xc84ec404080a4a4ULL,0x488264848c48500ULL,0x1083546040428130ULL,0xf2965d608243f3a0ULL,0x800408208000c040ULL,0x8a44800820292e0ULL,0x5a1ccdd0c047fb72ULL,0x90a0506040422286ULL,0x290204002048022ULL,0x2000400008400000ULL,0x84d392cacac7f050ULL,0x2c140101010ULL,0x4e416c048408b20ULL,0x0,0x1004000002000200ULL,0x0,0x0,0x80040000020200ULL,0x130020000028202ULL,0x820425028000e120ULL,0x880008040000a220ULL,0xa5a0ac3c882b602ULL,0x34e35372ca43da60ULL,0xc1508c1042437230ULL,0xda902e22cac2f292ULL,0xc8c7f97ac0d2bbb4ULL,0xe8fffad6c252faf6ULL,0x973a55304acab242ULL,0xffffbffbcad7fff6ULL,0xdbde7c72c28ffde6ULL },
	{ 0x8b93a37c8cf23f83ULL,0xa870114a07c1aa02ULL,0x830954104865e00ULL,0x95e87f4c0cc31a29ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x66000040022601ULL,0x0,0x0,0xde3dae71cec27f93ULL,0xb00733280871221ULL,0x1af27e178dda7f81ULL,0xaffecfdfeddffb93ULL,0xb4b6feeeeddbff81ULL,0xfbfe5ffbcde37f93ULL,0xd7b22f518e83eb9bULL,0x6d56ffb4ae037e1bULL,0xabdd5d9b8cb6db89ULL,0xf4feffebcdd7fb9bULL,0xf9cdbf72accb3f93ULL,0xff96af6acef3be99ULL,0xeff6fb9acfd3bfb3ULL,0xfe7fded6cfd3beb1ULL,0xaf0adf768cd35bb9ULL,0xddfcbe7ecfc3ff99ULL,0xfff63f7daea77fbbULL,0x89d7f31f8ce36e9bULL,0x395f7ed3ed873fb3ULL,0xe13beff34dd2bf91ULL,0xd77bffe98febafa3ULL,0x7bbe5df2ecd3fbb1ULL,0xfbdcff36cfc27eb1ULL,0x2f7bff520de27fbbULL,0xefbfcffeeddafb9bULL,0xd6ff7ef98fe73f9bULL,0xf3ffee71efa3fbbbULL,0xffbfff7e8ff3ffb9ULL,0x8fcfbf37cdd7ffb1ULL,0x3bfedddfefdbfe9bULL,0xf4ffffed4fdbbf93ULL,0xf7ceef72efc2bf99ULL,0xfff7ff6f8dcbff89ULL,0xdfdffffecfd7ff99ULL,0xefffffdfafd3fb9bULL,0xdfffdffbeffffe99ULL,0xfb7fff7feff3ffb9ULL,0xfeff3f7fafe3ffbbULL,0xeff7ffffeffbff9bULL,0xafffdfdfcfdbffb1ULL,0xbffaeffbeff7ffbbULL,0xffffffffeff6ffabULL,0xfefe7fffcfd3bf9bULL,0xebffff3f6fe3ff9bULL,0xef77ffffefffff3bULL,0xffbffdf7cfdfbfa9ULL,0xfffffeffcfd7bfbbULL,0xfffefd7aefd7ffbbULL,0xebf4f73e4fe7fb9bULL,0x5ec3befacdc3bf03ULL,0x7f7ddf7a6fcb7a93ULL,0xf5fff6ebcfcfffa9ULL,0xeffefff7afd3ffbbULL,0xffbfffffefdfffbbULL,0xdfffffbfefcf7fbbULL,0xeafddfdfcfbefe9bULL,0xffffffffeffbffa9ULL,0xfbffffffcffbffbbULL,0xfebeff7fcffbff13ULL,0xffd7efbecfc3ffb3ULL,0xeffedfdf8ffbffbbULL,0xb5fff7fbcfdffbbbULL,0xffffffffeff7ffbbULL,0xfebe3fffcfdbfebbULL,0xcffdf7fbccc3fab1ULL,0xaf7f7e4fcdf73b8bULL,0x4fff7fffcdd73b91ULL,0xf385f660ccd27a93ULL,0xffce6ffecde6fbb1ULL,0xc3aef7bf8ef3bf11ULL,0xa8cf99c92ec39a83ULL,0xde9efffacff77f81ULL,0xf4fefef7eedbfe83ULL,0xeefaefd34fdfbe89ULL,0xcbf6f37fce93bf19ULL,0xffeb7e696fd3bb19ULL,0xe6bfcffacf433f93ULL,0x70548fc40cc29211ULL,0x45603c0005803801ULL,0xefa6fffb8ed37fbbULL,0x41c3d60b0ec33aa3ULL,0x7bd6dfd6cd837ab9ULL,0xdcfef7fdcedbbb91ULL,0xffafff63efff7fb3ULL,0xfab47f74cdcb3aa9ULL,0xc8fefe284da1be83ULL,0x1b64df598fc3db81ULL,0xb6ffcff72fc3bb13ULL,0xf9bfe7798ed5fb91ULL,0xfebe3f7fcfdbbf89ULL,0xcfdbff7bcfcb7eabULL,0xeffffeef8fc37fb3ULL,0xb73b57fecec37e9bULL,0xf1eca5612dd39b93ULL,0xfbf6ff7e8ff75f9bULL,0xebfff79acfdbffbbULL,0xc7f67c438cc7a281ULL,0xbbbbeffbaffffa1bULL,0xdebe767f6dc3fb83ULL,0xdfbeaf73cccf7f93ULL,0xcb10f72eacc31fa3ULL,0x5b5232970cc3da22ULL,0x4fb6f5fa0cd33b81ULL,0x1dfedee9ee88de03ULL,0x97aced310cc0bea9ULL,0xf100c1f00bc35b11ULL,0xfddfdf3e8cc33fa3ULL,0xab7fd7d3cfb6fa91ULL,0xd4f1dce92cf2ba99ULL,0x19eff7e92e52be89ULL,0xfb9f3ff7aed7ff9bULL,0xffffffffefffbfbbULL,0xffffffdfeff7ffbbULL,0xbeff5fffadd3bbb3ULL,0xddeea7434a503e81ULL,0x66862f3605823e10ULL,0xe9d2ff7befdb7fabULL,0x4efefe47cdf3fe83ULL,0x4d057b18cd0bb0aULL,0xff78ef7fcfd3ff9bULL,0xce2eadf8edc3be89ULL,0xca85d3122ec282b9ULL,0xe3fffef3ced3bebbULL,0xf30481ac4c33a11ULL,0x91b086688c823880ULL,0xa2256000831a28ULL,0x984c432008523a11ULL,0xc8220a2008200b01ULL,0x4400404888000ULL,0x400100000808000ULL,0x840b210008c0a201ULL,0x4012008000001ULL,0x0,0x0,0x74b5cef12e955f91ULL,0x3222d50050f2a02ULL,0x9ec4374c0ad1ba01ULL,0x46e0000c833a09ULL,0x80d0df5888436e91ULL,0xb426d98a2c527a8aULL,0x4179f462cce23b01ULL,0xd69fefd78ddbbe83ULL,0xa9acf31f4dfa3b81ULL,0xbd529b0b8c83be00ULL,0x8e1269988ce66a89ULL,0x8508000028000300ULL,0x9a002f2e4c436ab1ULL,0x8008da5004838a88ULL,0x77d212844c83a281ULL,0x839e46302fc00a01ULL,0x642654c000d23e00ULL,0x2a142c2205023810ULL,0x2984c7280cc332a1ULL,0xc957b2330fc2aa09ULL,0x2b3ccb79ade27a99ULL,0x91d706c8cfa23891ULL,0xfffeff73efdbff9bULL,0x5284ee2e2c923328ULL,0xa24432bbcf8b1b01ULL,0xa12e489acd321b09ULL,0x8699c4ea0842da81ULL,0xe98c476808d25aa9ULL,0xaa9763d48dd35a8bULL,0x5fc3a63e85d67a01ULL,0xa9ee5ed18fd3ba91ULL,0x3ed76be7afe32f9bULL,0x4128a6080c583900ULL,0xd0162e208d430a00ULL,0x6c4e51140cc2a290ULL,0xa2769dd20fe33a89ULL,0x3dba476a8ef0ba83ULL,0xf44dbe618f831ba1ULL,0x410020404420200ULL,0x8280420080131200ULL,0x40a301806103600ULL,0x2090006000202ULL,0x4410800002a001ULL,0xab3c75714cc73f99ULL,0xca54531e89c05e01ULL,0x4cc290388d829991ULL,0x200001c000011a00ULL,0x10020000020008ULL,0x60008000001ULL,0x822002120cc11a00ULL,0x382d60288f23a10ULL,0x8a50040a0c031a00ULL,0xf8f4d26085cb0e8aULL,0xdeec0560c5c67a83ULL,0x8290200080021a20ULL,0x5296000d402b00ULL,0x10e000003000801ULL,0x8810010000808000ULL,0x2010040000001ULL,0x0,0xa002c86048021001ULL,0x19475e0088d61712ULL,0x888294d924971200ULL,0x1000001080020000ULL,0x43a2753184827a18ULL,0xa1049250c8420a03ULL,0x808171005801a81ULL,0x8fa8d75aafaffab1ULL,0x7194d8604dabee01ULL,0x44961d72a4c3a201ULL,0x1280c1220ec00382ULL,0x623f410ec14a11ULL,0x4040008000001ULL,0xa0500080024800ULL,0x400004000010800ULL,0x2020160206103001ULL,0x2100008200000ULL,0x2210000401010a00ULL,0x4008402028000201ULL,0xc08250000115800ULL,0x7000000204800600ULL,0x0,0x6208140004022220ULL,0x19d15dc18cc39a91ULL,0x2204040008000282ULL,0xba0cb7ad8cbaba09ULL,0x8983e72b4d92ba8bULL,0xebdffa5bcdcbbeb1ULL,0x9fa2cd524fd3bf89ULL,0x3935a6788e8a9e09ULL,0x12616400a200801ULL },
	{ 0x6389e50a60711554ULL,0xa000000a004410d0ULL,0x8110400008421000ULL,0x444020020500080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80831e28227b104cULL,0x80000800100400ULL,0x54f17678645c15dcULL,0x10a184060001108ULL,0xb995224da555114aULL,0x4edf7f19e67e15deULL,0x7fffffcff27e97feULL,0x3282d6e826721146ULL,0xdd7fff73626e15ebULL,0xffbf2ffe717b155dULL,0xfffffbafef7797deULL,0xefff17d16a7a15deULL,0xbcbfcfe16e7a13caULL,0xddeee643e45a92ceULL,0xdabc3d78337295d8ULL,0xc9f7ff7b327217ffULL,0xa3ffff78237317d6ULL,0xb0bb6ff3a77b15deULL,0xd176b5d4667a16ceULL,0xff963766677614deULL,0xcbd1ff2ae86211eeULL,0xaf6e0a51a2721258ULL,0xffa9fdf3a957164aULL,0xf5fff47d237e13deULL,0x2e3a6d60f273136aULL,0xdbd4cf1ee37391c8ULL,0xdedbff4fe276156eULL,0x9fbfcff2a37f14dfULL,0xb5bf76eae75f17feULL,0xd9ef6d777f5217daULL,0xffb6ff7f6d7797dfULL,0x4fd7bf0bea5b16faULL,0xfb7cdfdfef7215ceULL,0xedf7fefdfb7615deULL,0xffeff77bff7a13faULL,0xffbe6f7ff37716fbULL,0xefdffe1bff7f17ffULL,0xffffffdfff7f17ffULL,0x9ffffffbf77f97ffULL,0xfbffff7fff7f95ffULL,0xffbfbfffef7f95deULL,0xcfffffbff67f15ffULL,0xebffffdfff7f97ffULL,0xfffbffffff7f97feULL,0xfffffffb6f7f17feULL,0xfffefffff77795deULL,0xebd4feae6f7315dfULL,0xeff75fdfec7797deULL,0xfffffffee77b97feULL,0xfdffffffff7f95ffULL,0xffbefff2f75e15ffULL,0xfbd5ff3ff77615deULL,0xdcf3f29b3172175eULL,0x8fbffffbb67f13ceULL,0xf5f5f6ebf77617e7ULL,0xbff7bd71787f11daULL,0xffbfff7f6f7795feULL,0xefffffbfef7f93ffULL,0xffffdfdff77b17feULL,0xf4fdfee87f7f95ffULL,0xffffe77bbf7f97dfULL,0xdfbfffffe37f17ffULL,0xffdfff3ff77797dfULL,0xefdfffdfe67e95dfULL,0xf7ff77ff377e15deULL,0xfffffffeff7f95ffULL,0xfebe3f7feb7f17feULL,0xffffffb96f7615faULL,0xeb57de5be27395deULL,0xbfb37ffffe7715fbULL,0x50fcffe0277a14d8ULL,0xfefe3feca87f10deULL,0xdbffff7f2f7b155fULL,0xcdcf7fc73f7a11deULL,0xbfbdcffbab7f13deULL,0xfdfef779f77e15faULL,0x7ffefd72e37a95dfULL,0xcb94e7af247b97feULL,0xdafbffdb267e15feULL,0xa79a9fdbab7f13dfULL,0x72d61dce78721478ULL,0xa17ffe22617211caULL,0xfb98ff6fa77a11ceULL,0xcdd2f695377217caULL,0xabfe0b4eaa7210feULL,0x3437d7eefa7e91cfULL,0xd3feeffbfb7b11ceULL,0xdfbf7fffe77795dfULL,0xdddffd32a57a12daULL,0xefdcfe5be77397e3ULL,0xb49f1eee226611deULL,0xff7ff77e6a7e91f3ULL,0xfebe3d66f67e91cfULL,0xcfd6ff7bf37695cfULL,0xefd7fe57bb5f11cbULL,0x3fb27f7efa72154bULL,0xd1ecf4faa27b15eeULL,0xf7b6bf76227f14dfULL,0x89fdffbfb77e14efULL,0xea4f3679ab7a97daULL,0x123addfba27711d2ULL,0xf6bef6f3fe7311efULL,0x5efe7df2277e17c7ULL,0x89c1e1610073155aULL,0x5ec62e3aa45615caULL,0x6d205348005a1446ULL,0x59b48678637213caULL,0x57a8ff60246a91d8ULL,0xffb76d7ee67214deULL,0xfde7b31b267e95ffULL,0xad7f5c922b6e10caULL,0xecf676e1027e17ceULL,0x9d6f6d76237711deULL,0xdb926f7eab7611deULL,0x5fdffebe7b7796ffULL,0x7fffffdfff7f97ffULL,0x9697436031701050ULL,0xdff9a778ba7a17eaULL,0xfabf2d02f073165aULL,0x6182e3386472904aULL,0xa90b1ae0a20214caULL,0x7b0ad7fb325611ceULL,0xdb5c7733235a14dfULL,0xde9eeff5e27693dfULL,0xc2b2fa676f7a91dfULL,0x4adb7f50ea7610cbULL,0x8a3ecd5962561558ULL,0x8010002022400001ULL,0x5a04200100721046ULL,0xe00000001640ULL,0xc8032a303210105cULL,0x0,0x89e9076a237610afULL,0xa519a037ba7e94ceULL,0x1004102000431100ULL,0x82300300001150ULL,0x10bdc4226301052ULL,0xf09deee8b25212ceULL,0x2040000000082ULL,0x88000a2800521021ULL,0xc9334f0e205291cfULL,0x30eded3b263107eULL,0x4000200500000ULL,0xd9aebe30237a13f2ULL,0xfbb22ffff57e11ceULL,0xfdc7ff362f7315dfULL,0x7bfeff48237a15deULL,0x8c306432a3611448ULL,0x9200a640227213d0ULL,0x3200c4030001248ULL,0x6000000050ULL,0x6e877808705b1154ULL,0x8998c08ba362004aULL,0x844030692c621068ULL,0x709c29a060541487ULL,0xf356ef7ce373116fULL,0xc372becaa75613d9ULL,0x37ef4f7a364395d7ULL,0xfcf776fce06a95feULL,0xfffffff7ef7717daULL,0xdb98ffae665915c8ULL,0xea46050f3974105eULL,0xa81eca52615215d0ULL,0x94fbd6e17a73115eULL,0xb0e88f4021461159ULL,0xcbb26ffee87315daULL,0xab47ff1621771552ULL,0xab57dfdf667a174eULL,0xb4bb3de3a77b10cfULL,0x12ca430005814daULL,0x72a227667170117bULL,0x49cac710e0531448ULL,0x83bfee40a47a90ceULL,0xffbbcffb367310feULL,0x91c635007c421000ULL,0x74061560a27406daULL,0x900800820420040ULL,0x0,0x4002a00121004ULL,0x200e00a000621000ULL,0x6830a04264500041ULL,0xcb80f7263163164aULL,0x605236eb6f541089ULL,0x4400020000080ULL,0x811502002011050ULL,0x320d406c410040ULL,0x4f80d24922701160ULL,0x49878e9b296317caULL,0x600e4c12626001caULL,0x909464a060551049ULL,0x41002252220000ceULL,0x85002208600046ULL,0x0,0x835c4e50223010caULL,0x0,0x90492600a542000aULL,0x8000020010001000ULL,0x638c7606325211d6ULL,0x12a20220129000ULL,0x108422026421100ULL,0x0,0x723636d4225710caULL,0x8194877c29531548ULL,0x4802580122021150ULL,0xf8c6dd9677710cfULL,0x4082244a0701000ULL,0xb020d4800721440ULL,0x0,0xa053b0d9e056104eULL,0x60230140a0740006ULL,0x1000208400011000ULL,0x2c0200021021000ULL,0x5004000000400060ULL,0xc1230a04002001d2ULL,0x8108041220700408ULL,0x10400000080040ULL,0x200002000000000aULL,0x80004000000140ULL,0x40108824400020ULL,0x208480121420058ULL,0xc1408a0201000c2ULL,0xa10c2610a610005aULL,0xd28326362472115bULL,0x89f6f326e154905eULL,0xefdeffc1a17315daULL,0x8b825f7a627e10eeULL,0x98fc364110781740ULL,0xa09642052002cULL },
	{ 0x888060460110e205ULL,0x80d01801081e00a0ULL,0x1228c83810c0884ULL,0x942026020211e01aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x98c20f39533de7e7ULL,0x203004a0083401c0ULL,0x5600002c0010ULL,0x1225d500101c002ULL,0x22104306404c8001ULL,0x800820100340080ULL,0xa216ced31a1d60bcULL,0xa4c45429c0342a00ULL,0x86625e05b4d6c10ULL,0xd696f3f60f3eced5ULL,0xedffffbfdf3eeff7ULL,0x277afe51ca7fefedULL,0x859437efce3cefedULL,0xf9ff3f72db7decd6ULL,0xdeb6776e8d7febdaULL,0xefdefbbedf3eebf7ULL,0xbdfffe6b9f3ceff5ULL,0xbfbbffd7de7deefeULL,0x7035b667433ecbd1ULL,0xfbf6fdf69f3ceff0ULL,0xe8fdf206db1cee8dULL,0xe6e31504b3c6e98ULL,0x9f9f5bff9b3deaffULL,0xfddcf67f0f7eeed7ULL,0x71b4fffa8b3ccdfeULL,0xfb84ff3e1f2febb2ULL,0xcfcffef8df3deff7ULL,0xbb1e45fe483dce3cULL,0xbdfeffee4f7defd7ULL,0xdbeebf52df4eed55ULL,0xde96e7771f3deffcULL,0xc9e7ff37df5feef2ULL,0xfbfecfdbcf3fefffULL,0xfdffffe9df7feffeULL,0xb9feff7b5f7defffULL,0xffbfff7fdb3fefffULL,0xfffffffedf7feffbULL,0xffffffdfdb3fefffULL,0xdfffffffdf3eedd7ULL,0xff7ffffe5f3fefffULL,0xfebe7fffdf7fefffULL,0xeffff7ffdf7fefffULL,0xffffffdfdf3fefffULL,0xffbbdfffdf3fefffULL,0xdffdffff5f7fefffULL,0xdffefffe5f3fefffULL,0xcbf2ffbfcb7febfbULL,0xafff7effdf7eeeffULL,0xbfffffffdf3feff7ULL,0xffddf6ffdf3defffULL,0xf7beff7bdf7deffeULL,0xfbfef7ff5f3deedeULL,0xeefab71b9a3eeed4ULL,0x8fbc5f5fdf3f6cb0ULL,0xf57ff6fb5f7ecfb7ULL,0xdbaeff731b3cefd7ULL,0xfbffffffcf7fefffULL,0xdfffff3fdf7eefffULL,0xff5fdfdedf7fefefULL,0xfeffffffdf7fefffULL,0xfffffffedf7fefffULL,0xfbff7fff9f7defffULL,0xffffffbbdb7fefffULL,0xffffdfdbdf7feff7ULL,0xffbaffefdf7deff7ULL,0xdffdff7bdf7eefffULL,0xffbf7f7edf7fefffULL,0xcdffffbfdb7deffeULL,0xcd4f3e5f9b3eafbfULL,0xb4bbffffdb7feffdULL,0xf5fdbf79db7ceffbULL,0xb7be7ffc0f35eed4ULL,0xcbc7ff1eca366ee4ULL,0xfed63fda4d3feed0ULL,0x8fbbdffbdf3dcab6ULL,0xdff7f77ddf7feeffULL,0x4b3ef5728b3fcbf5ULL,0xebf9f33fdf3dedfdULL,0x2a5eaf5b4a3fc6fbULL,0x7ff8efc1de3ecbdcULL,0x3140874a0b1cac81ULL,0xf3bed570cc7eeaccULL,0xdfbbfbff5f3feff6ULL,0xeedeff1f9f3feed1ULL,0x2fff97d98b3eeff1ULL,0x5effffe8cf3deadbULL,0xddfeff7a5f3deff4ULL,0xbffe6f765f7deddfULL,0xecdfff32db3fefd7ULL,0xabda3f978a3dcef4ULL,0x5cf651e31f2cc8e6ULL,0xd92cf7e1df7eeb3eULL,0xf6b6bf7acf3dcbf5ULL,0xf9d7feebdb3ecedfULL,0x2cd7fa6d8c3ceaf6ULL,0xb7bf67fecf3fcda4ULL,0xddfdf7fbdf3cebf6ULL,0x797cfd725b3feec9ULL,0xd1dfdf0e1b6ce9d8ULL,0x8f5317da4b3cebc4ULL,0xfebe5fea8f7deaeeULL,0xe5a7f2ff4b3ecfffULL,0x55be3fe2833eeed6ULL,0x9991fb6c1e26eb58ULL,0x4dd63e06c1144c24ULL,0x8711c5f11f7de468ULL,0xbde957ebcf7ecff4ULL,0x9fefaf77c93ecfacULL,0xfa85b7fece3ecfb7ULL,0xc34ad2290a3c6a9cULL,0x6714d7921b30e2a3ULL,0xf6d4d47d5b2cecd6ULL,0x14a07400026c892ULL,0xf6ac17745f34efc0ULL,0xdfdffebfdf7defffULL,0xefffffdfcf3feffeULL,0x3ef757ed8b7eeed6ULL,0x93c4ae720d5dea84ULL,0x42202c414a242400ULL,0xd151971d487c2ca0ULL,0x4b4e1f0b497c0ce4ULL,0x9cb61eb45e3661c4ULL,0xd3ceb5f24c2ce8d3ULL,0xfe9e5d709f2dedf0ULL,0xcac6fa4e4a7ecae9ULL,0x1ad71f401a57e22aULL,0x1f2adfdf893c6ee4ULL,0xd77ea6a85f3cecd9ULL,0x2221540001c8630ULL,0xc334f102da3dc84dULL,0x8801eea882164a24ULL,0x188090008102088ULL,0x0,0x100800080000800ULL,0x42000004800ULL,0x87cbffd11f256f18ULL,0xa1c0fdd1a3e84fdULL,0x220e8417044404ULL,0x0,0x80000408000080ULL,0x46021002004900ULL,0x2200c00013000800ULL,0x21b2546043344d00ULL,0xfde4ffef1e3fefd8ULL,0xdebe3ff95b7fefdeULL,0xcfd3ffee1b7eecddULL,0x2eded8dc9a3caed4ULL,0xbdca4dba8f2c4cd8ULL,0xd0008a0000208810ULL,0xa40640009042c00ULL,0x83a8d1228924c918ULL,0x95f1003030e0e84ULL,0x1000010008000200ULL,0x0,0x20050008080220ULL,0xdbdeef37cf3deec9ULL,0x8dea3e170a36cfcaULL,0xd1cdd585a3d47c4ULL,0xfdfd17f85f3dee9dULL,0xdfffff77df3fefffULL,0xe6ace56e8e3f4ae0ULL,0x5180a227013d4844ULL,0x8a5a5fc85a3d6c90ULL,0x96bc4eec5b3deb95ULL,0xb9c80fe00d75e045ULL,0xdf8eff2d4a3ccfd7ULL,0xe9d3f3134a3ea88fULL,0xa951845242142004ULL,0x22840309002880ULL,0xaf7d2330c9776884ULL,0xcca62740131c4b94ULL,0x80201362822c0810ULL,0x41185a9981384e90ULL,0xb4aae5ec1f3dcbe4ULL,0x816c42300a3c82c8ULL,0xc2cef5f69e2deaf9ULL,0x100850000040044ULL,0xa080108002240c00ULL,0x28220d400666c200ULL,0x0,0x10087440082c2014ULL,0xa080c026000c44d0ULL,0x4a402c404d10e802ULL,0x0,0x500c002040600ULL,0x9008144006244108ULL,0xd111cb3c037cca92ULL,0x40261081044c00ULL,0xc010091208168222ULL,0x49120210b040844ULL,0xcc35020b040c48ULL,0x2800400080c0008ULL,0x480000000100080ULL,0xa010400002040004ULL,0x90500000008002ULL,0x51480300080940d0ULL,0x3000010408302000ULL,0x802d0088101c880ULL,0x1408001002c0040ULL,0x53ae7654934c2c8ULL,0x0,0xc4222d6403162a80ULL,0x8c4812a401e01b0ULL,0x40020000240000ULL,0xbb817f789b7eeed4ULL,0xc31044600a3c81c8ULL,0x44086c020a2c0080ULL,0x8100041002040140ULL,0x42284004084010ULL,0x8008000008080ULL,0x40040000008002ULL,0x801029400e104160ULL,0x19223051a0e8085ULL,0x0,0xc00000080004ULL,0x420000080000800ULL,0x6008200000ULL,0x10142a640a34282cULL,0x6002020002004110ULL,0x9bb9df5adb7ecfffULL,0x5223e255b2489c3ULL,0xa04225784814ea14ULL,0x9bb22e4ac35ccc5eULL,0xcedaf7fbdf3ceefcULL,0xafdffbfb5b3feefdULL,0xbe3a46f15b3feed2ULL,0x10040618050c4884ULL,0x0 },
	{ 0xc0002a010208853aULL,0x606060801000502ULL,0x61440000122040aULL,0x9290000050080d21ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x488e1285c19acf6fULL,0x5154ab238001799ULL,0x91040000008001ULL,0xdfe82d72b28adf1bULL,0xd944020240908529ULL,0xa10000008030aULL,0x73bdcfc0910a971eULL,0x31a1040040100721ULL,0x94497590901a4fa0ULL,0xdb363faf71bacf77ULL,0x57cbb690c2825d57ULL,0x89dff95b633acf73ULL,0x449a77eb93b287fbULL,0xfdbefffedbba4f7fULL,0xf7bc3f5ee39adfbbULL,0xc9f5f7fffa3adffaULL,0xaf7eff4ffbbadffbULL,0xffbf5ffffbbadfffULL,0xe03717ebdabac7f1ULL,0x7a9674f2291847abULL,0x8afaf3ba909a575fULL,0x6effdeb8f09a57fbULL,0x9fdffddb792adff7ULL,0xfffff6fffbbacfffULL,0xdfee7de3f19add77ULL,0x8bd4ff2cfa3a457eULL,0xfffffffffbbadfffULL,0xe7bccf77fb1a97fbULL,0xff7bfafc0b9acdffULL,0xfefeff70339a9fdbULL,0xffbdf77e79bacffbULL,0x9fffde9bfb9acff3ULL,0x6bffcfdbd9badffbULL,0xf6fff6efdbbacfffULL,0xffffff797b9adfffULL,0xffbf7ffffbbadffbULL,0xffefff7efbbadfffULL,0xefffdfdffbbadfffULL,0xdfb7fff799badffbULL,0xfb7fbffbdbbadfffULL,0xffbeaffffbbacf7bULL,0xeffff7fffbbadfffULL,0xfffffffffbbadfffULL,0xfffffffefbba9fffULL,0xf7dffffbf9badfffULL,0xfff67fecdbba9f7fULL,0xebbfff7bfb3a9fffULL,0xafff7fdffbba5ffbULL,0xffbfdffffbbacffbULL,0xff7ff6fff1badffbULL,0xffbefdf3e9badfffULL,0xdbd5ff3fdbaa97ffULL,0xaccedb5fd13a07fbULL,0xffbbfffefbbacfffULL,0xfdffefebfb8adffbULL,0xdfffeff3dbba9ffbULL,0xfffefffffbbadfffULL,0xcffffe9ff9badffbULL,0xffffdfdffbbadffbULL,0xfefdfffefbbadffbULL,0xfffffffbf9badffbULL,0xffbfff7ffbbadfffULL,0xefdfffbebbbadfffULL,0xebfedfdbdb9acfffULL,0xfdff5fe3fbbadffbULL,0xfdefff7ffbbadfffULL,0xffbebfffd9bacfffULL,0xeffeffbde3badffbULL,0xf35e57db1abacffbULL,0xa6936fee399acfffULL,0xfd69f77f419a57fbULL,0xffde7fedfa9acffbULL,0xebc6efb3a29acffbULL,0xfff7b7dfd93acff3ULL,0x979edfbb39ba97dbULL,0xd77ff7fbf3badff7ULL,0x7ffe7fd2fbbadf7bULL,0xfbdfe76ff19adffbULL,0xefdebefffbb2dfd7ULL,0xb3ccf3f93ba1f5aULL,0x7d97dee8c21a0fbbULL,0x89ae2de2ca9a0f78ULL,0xff9aff7fdbbadffbULL,0xcfdbfeabf9ba9ff7ULL,0xbdfeffdbfbba5ff3ULL,0xfefd5eefd13a4debULL,0xff6eff76ebbadffbULL,0xfbbf77fffbbadffbULL,0xecdff35ae09ac73fULL,0xabfe5fd3c992ddf9ULL,0xe4365767832a8f9bULL,0xd93c2ef2798a113aULL,0x9ca63f7e539a9f7bULL,0xddd7f6aadbbadd7bULL,0xaf5f3fc65abacdfeULL,0x3cf3efe7fb9adffbULL,0xdb79ff7b799a87fbULL,0xfffe7feecb1a9d7bULL,0xe925e377529acf7bULL,0x667f1ef9f112dffbULL,0x9fbf5fdbe38adffbULL,0xf17bff7bd93a5fffULL,0x5fae7f6a439adf77ULL,0xebddff2ff9badff3ULL,0x8957b99ad92acd6bULL,0xefeeecfc3ba957fULL,0x15f75dead0185ff9ULL,0xdfcdff635b9a9fbfULL,0x9bb4ecef592a8ffbULL,0xcdfdfe1719aa9dfbULL,0xfbedfde6a1a8ff1ULL,0xfefddef6fb1acfffULL,0xc1e7ff55109a9f7fULL,0xdf9e677cd3aa5db2ULL,0x7d43cffed33acff7ULL,0xabdedbdfe9b21f79ULL,0x3d3a99f3c9aa4fe9ULL,0x9129ade8001acd0fULL,0x22842c21900a0443ULL,0x8080e620000a456aULL,0xa0181283d0180731ULL,0x95b66e6bc99a8f73ULL,0xd0fc2d74c308877bULL,0x93fc7f1541aa8debULL,0x8880f73a51388dfbULL,0x4b571ec76092dffbULL,0xd3934cfec9aacfd3ULL,0x8df17369d01a1670ULL,0x4bb82d30418a0fabULL,0xcbd0f33b703a0debULL,0xd8829a69409a0d13ULL,0x804884301880102ULL,0x0,0xa00000020000100ULL,0xa900400101200901ULL,0xe9cff3b34b9a5fdfULL,0x2f5fd6d9b18a8fbbULL,0x48000000300ULL,0x4200000120ULL,0x400004000080002ULL,0x0,0x0,0x418523060020542ULL,0xf76d77da6bba8ffbULL,0xca9b2f3fd31a8f3fULL,0xc9fbdf399b3a95d7ULL,0xa9569f6e512a877bULL,0x1bb25f62993a9f11ULL,0x19fab70011184f23ULL,0x902076041320d00ULL,0xe8c672f2839acf53ULL,0xad31e15ea08cf1aULL,0x1084410202000538ULL,0x40800004800ULL,0x303c5440100a04a5ULL,0xbc4e51111888f3bULL,0xccfa36dac82a07f3ULL,0x2cfdd7521b8a97fbULL,0xfdfff7faf39adfffULL,0xdffffff7dbaadf7bULL,0x79c677fea89a9dfbULL,0xcdd7be0cd8bacd72ULL,0xe99a870a4288057bULL,0xe4f67b645b2a0f7bULL,0x354c8762618a1fbfULL,0x8a16330860980523ULL,0xca060e93418807b3ULL,0x8115109200970ULL,0xa022145a2108951bULL,0x8300040000028901ULL,0xf63a2576000a9f73ULL,0x80bafb2849b24d43ULL,0x214ec25050180fb1ULL,0x27b155f24918c523ULL,0x10c0020010080500ULL,0x1dafbe2ba1acdf9ULL,0x8000d08088080502ULL,0xfada5f4f488a8defULL,0xa88c11900289458ULL,0x600e0000003a0ULL,0x290000004500ULL,0x83001119800a0c47ULL,0x110000001010ULL,0x200001a001008011ULL,0x8060176221a20120ULL,0x10982c1040028070ULL,0xdc06534e80180703ULL,0x41101201400001a0ULL,0x44000000408ULL,0x0,0x100236208080540ULL,0x100000140000020ULL,0x2cc1508081824c06ULL,0x804a418108040aULL,0x0,0x48022800080700ULL,0x880829010108480aULL,0x8300240410080402ULL,0x2100088040820100ULL,0x0,0x10c4140080100601ULL,0x484202020012ULL,0x9001218020007b2ULL,0x40020000080002ULL,0xf9ba8fd3891a9f7bULL,0x14764464a8020f13ULL,0x4800000100ULL,0x0,0xa8901287401a4423ULL,0x810036001080002ULL,0x4000800000020800ULL,0x200300009000102ULL,0x420260080000052aULL,0x4c420e011108050bULL,0x1000010000100400ULL,0x8000002000100400ULL,0x62094000080442ULL,0xc282214000109531ULL,0x8000e00010001410ULL,0x1880e12100a0d07ULL,0x8030068108020152ULL,0xd76ffe70d89ad531ULL,0x92a67774381acd39ULL,0xeec7ff3a899a5f7bULL,0xa2ffbe5a590a9f73ULL,0x96fb4f6bc99a8fbfULL,0x50648668620a0732ULL,0x421225200218173aULL },
	{ 0xcab0eb7a3cdefd46ULL,0x10060200828200ULL,0x20000080c000000ULL,0x50c4b7d13c91f5e0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa97ab1de24b27ee4ULL,0x671c4c00a09326e8ULL,0x6014408000122284ULL,0xdb42fc6114afd266ULL,0x200000610014800ULL,0x8c089620008646e4ULL,0x5084020000040ULL,0x5096f6fa35227ac6ULL,0xcfc52e23143a7362ULL,0x9ae22e6514f9c582ULL,0x64ce31221c334ae2ULL,0x6bdeff813cb8eeecULL,0xa61bfee6bc967cc4ULL,0xfb6de77c3dfbffeeULL,0xfebe7f5e3cb2ffeeULL,0xe9d7f72bb43a7feeULL,0xefffffdfbdffffeeULL,0xffbffffbbdffffeeULL,0xd373bff83cbbf8eeULL,0xff9a3ded3cbafbe6ULL,0xabc3c33221ba72e0ULL,0xefdf3edebcfbffe6ULL,0xbdb37df235baf3e6ULL,0xf55ff6e9bcdbeee6ULL,0xcdee7fe114ba7feeULL,0xdbdfffff95bfffeaULL,0x6fdbbffb34feffe4ULL,0xcc7f4548149558e6ULL,0xfdfff7febcff7feeULL,0x9f7fed5334bfffe6ULL,0xd1b8e7febcdbf7e4ULL,0xdfffffbbbdfa7feeULL,0x3fdfffdf3cfbdfe6ULL,0xfe3f56fd3cbf7fe2ULL,0xdbeef77334bf7feeULL,0xffff7fff3dffffeeULL,0xeffffbbfbdbbffe6ULL,0xfffedfdfbcffffeeULL,0xffffdfffbcffffeeULL,0xffefff7fb5ffffeeULL,0xffbfbfffbcffffeeULL,0xffffffffb4fbffeeULL,0xffffffdfbdbfffe6ULL,0xbffffffbbdfffbe6ULL,0xbdffffff3dbfffeeULL,0xcffe7ff7bdff7feeULL,0xfbfffbffbdffffe6ULL,0xefffffff35ffffeeULL,0xfffffffbb4ffffeeULL,0xfdf7ffffb4ffffeeULL,0xfffeff6bbdffffe6ULL,0xbbfee77634bbefeaULL,0xfff3bfbbbc3b7deeULL,0xefbecfdf3dff7fe6ULL,0x7cf7f6f805fff2e6ULL,0xeffffff7bdfbefe6ULL,0xffbcffffbdffffeeULL,0x6fdfffbfbdbfffeeULL,0xeb7bdfda1cffffe6ULL,0xf6ffffff35ffffe6ULL,0xdfffff7b3cfb7fe6ULL,0xfffeffffbcffffeeULL,0xefffffbfbcffffeaULL,0xffffffdfbdffffeeULL,0xf6fbffea1dffffecULL,0xf7bff7fb9dffffe6ULL,0xfebeaff73cbfffe6ULL,0xeff7f3b7b4de7fe6ULL,0xbbd77eef35bf7be6ULL,0x9ffe7fff8dbf7deeULL,0xffeff7fd1c3b7feaULL,0xffee7d453cbaece6ULL,0xeaf9f3293cbe6eecULL,0x6fdeb7df3dfbffe6ULL,0xdffd6ffe3cf9ffe6ULL,0xf53ef67e3cfb7ee6ULL,0xcbae1f60b4bb7ac2ULL,0xf8ddf77f3cb3fdc6ULL,0xee77ff9f15fb7fe2ULL,0xbfffeffbb4ba7eeeULL,0x71ae66d810946ee8ULL,0xdb2bed433cbf5beeULL,0xfb95f7bf3cbbfeeaULL,0xcfdbb73b34bff266ULL,0xfffcffdfb4ffffeeULL,0xf6f7feef3dfb7feeULL,0xd7efed433c9b7ee6ULL,0xbbb7ff7fbcfbffc6ULL,0xafffffbb3dfffee6ULL,0xafffdbdf3db77feeULL,0xb9fe3defbcddf7e2ULL,0xffedff78bcfb37eeULL,0xfa99257f3ca6efe6ULL,0xadf6f77bbcbeeae6ULL,0x2ddffc512c92e7e4ULL,0xbf5fcffbb4bbffeaULL,0xd378ff793dbbffe6ULL,0xfb9eff7e3cbb6fe6ULL,0xc984f7b635bbefeaULL,0x9ef73f5ebcdbfbeeULL,0x47becffbbcbaebeeULL,0xff74beed14eeffe6ULL,0x4fea67eb04bb7ee6ULL,0xebfdff2f3dbfffe6ULL,0xee93feeb3dbffbe4ULL,0x3e6ccd753cbe6ae6ULL,0xf0b065fb1cb726eaULL,0xc3a12d031ca66be6ULL,0xf793f3ff3dfa7fe6ULL,0xcfddfea3b4b336eeULL,0xbecefcb94bb17eeULL,0xb0d1d428809262e4ULL,0x11e20fe334b1d162ULL,0x5bbd6ffd35faf6e6ULL,0xdfcfffae38f2e7eeULL,0xafdc7f93b4972ee2ULL,0xb5d615e20dbcd162ULL,0x114c066880b99280ULL,0x8a04256004126b40ULL,0x82018602002260c0ULL,0x80808e86042027c2ULL,0x232807400c391028ULL,0x6178be180488d6e0ULL,0xd2aefdfd1cfbfee6ULL,0xa9deff2a3dbedae4ULL,0x46d27f1a392a6aeaULL,0x81aa85c208f25aaeULL,0x5062f6f01cfade40ULL,0x5c04a5d004b811a6ULL,0x8000120001220004ULL,0x88018604041002a0ULL,0x401000004800000ULL,0x0,0x102000000220000ULL,0x8000b3804111062ULL,0x0,0x2080000200200ULL,0x8521106214b10386ULL,0x112c202c30b10682ULL,0x0,0x80800000002040ULL,0x100100000084ULL,0x140414c034110462ULL,0xfb64c5b83c966fe4ULL,0x6a0205700891caa8ULL,0x9d0521401b850c4ULL,0xed94100b242a0ULL,0x303208221891cc82ULL,0x9041820020114ac0ULL,0xd6b63fd5349afee4ULL,0x89dee12a3c986ec6ULL,0xac5e5ef81c9c56a6ULL,0xd0acdca301232c2ULL,0xd5b49a5114333ee4ULL,0x41223d6000336a66ULL,0x8b4cfb6a28fe52eeULL,0x6c95b5b18ba6ee2ULL,0x628cccf114f2e6e6ULL,0xf5fff7feb4bfffe6ULL,0xd5baaf71a0fa12eeULL,0xfe9e6bfe349f69eeULL,0x5df7fe8d35bf6de6ULL,0x2afeedd33cfff2eeULL,0x74503df691bae6eaULL,0xf9b0b721343127eeULL,0xbb96b771149f796eULL,0x88d1d303301322c6ULL,0xa51c2488c923004ULL,0x3220816904126064ULL,0x813a776a10117be2ULL,0x4abda7700094e6eeULL,0x2956e71ab4f8256aULL,0x65dfffdb3c396feeULL,0x363a65bf30f2f8e6ULL,0x98501059202a44eeULL,0x5784b7e00cf8f842ULL,0xe938d20224ba6342ULL,0x60869a2115391200ULL,0x8c04000c808000ULL,0xe4346c62049271e4ULL,0x152c1d4000d36244ULL,0xb9acbe6e14fbd2a6ULL,0x5c7ebe86841333e6ULL,0x6000000280000ULL,0x9001806210024e80ULL,0x19640720cd19880ULL,0x6506002a4240ULL,0x4080100100101000ULL,0x870824105a86000ULL,0x810000010010000ULL,0x122584000986068ULL,0x3100000100000484ULL,0x202000100020ULL,0x0,0x1020102000844a0ULL,0x1000040000200002ULL,0x40840020000024c0ULL,0x29ded2321cb9d820ULL,0xa9631800009062e0ULL,0x322101810200482ULL,0x51a82752383864c2ULL,0x900420000112000ULL,0x180001000280200ULL,0x800010000000042ULL,0x7198b7d3b4bf57e4ULL,0x8000000000000084ULL,0x0,0x4800d06810884440ULL,0xe04230d38494606aULL,0x2abec8400c1b6ac6ULL,0x1100860200c02040ULL,0xc0c12000000042e0ULL,0xe312624c14b14162ULL,0x200200000880000ULL,0x410000000120024ULL,0x140010041c000002ULL,0x0,0x280042000006840ULL,0x1c4c00c24ae3242ULL,0x140000000280ULL,0x120000000101000ULL,0xf825228210860bc6ULL,0x7020019201386020ULL,0xafead6a514bb73e4ULL,0xea5f7ed9b4b977e0ULL,0xf81775a1cbbfca2ULL,0x5586176e1c1b7bc4ULL,0x821a2c4018b9c482ULL },
	{ 0x8191c20510d45a28ULL,0x4122400101c0a28ULL,0x22908702145cd242ULL,0x716ca75bf4347768ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2840915024105020ULL,0xefbdcf49f47c5fffULL,0x546c9665bc1cfee2ULL,0x9d40424180205442ULL,0x529160119010168cULL,0xcd73f733fe70726aULL,0x6f9cf53c05456eaULL,0x75b77ff9def474eaULL,0xefef67fbb4fcf6eaULL,0xfa36bf83f43c5e78ULL,0x8681f000141510aaULL,0xebfcdfdbb47c5eedULL,0xa37cddb3a41c7eeaULL,0xdbe5e7fff6bdf6ecULL,0xfe9325e7e65c7efcULL,0xffd4f7be76f076faULL,0xeffffedffefdfeffULL,0xfffbfffbfefddfffULL,0xfd74f75cbc5c5fefULL,0x5fae7f77be547dffULL,0x8be4e352765c14eaULL,0x6ed79fdff4741feeULL,0x5b9ccffb74d5f4eeULL,0xd5f1faff7c305eedULL,0x47be7df3d61d1fffULL,0xdbfbef7faefcff6fULL,0xf9dbfffb34fdfefeULL,0xbfbdc7f33e5cdffeULL,0xd6f77ff2bed4ffeeULL,0xd9fef763fe387eefULL,0xd9f5fffff49cf6eeULL,0xffd7ff17b6fc77feULL,0xaf7fdfdebcbc7efeULL,0xfebafeed54dd76feULL,0xdf77fff3f4f57eefULL,0xffbe7ffefcfdd6ffULL,0xeffffb3efe7cffeeULL,0xefffffdbfeddfffeULL,0xffff7fef7efdfffeULL,0xdfffff7efefc5ffeULL,0xfebebffffefcf7ffULL,0xedffffbefe15fd7eULL,0xbffffffffefd5ffeULL,0xbf7b6fff9ebd7cfaULL,0xf7febffffefdffffULL,0xeffefff7fefcfffeULL,0xfbffff7fbebcdfffULL,0xffffffdffef4feffULL,0x3ffffffbfebdffffULL,0xf5fff6fef6fdfffeULL,0xfffe7ffbb6fdfeffULL,0xfbddefb776bcfffeULL,0xcdfbbbdd769dfffeULL,0xbfffdfdffefd57feULL,0xfb7ff6ffaef55ceaULL,0xfffeef75f63dfeffULL,0xffbffffffef4feffULL,0xffffffaffefdffffULL,0xbfffdfdff6bcfeffULL,0xfeffffe7fefd7ffeULL,0xefffff7d7e7cdffcULL,0xfffe7ffff674dfffULL,0xcfffff3ffefcdfffULL,0xfffeffdffef57fffULL,0xddbefffffefdfffeULL,0xffeffffdfefcdefeULL,0xfff63f7bbc5ddfefULL,0xfffbf37fde14f6efULL,0xffff3f5f34fdfd6eULL,0xbffb4ffafe5d55eaULL,0xd3edbf7ab63477feULL,0xbff67f61bcf5d7faULL,0xdb9fff3af6d4567aULL,0x6f765e5f6e7c54fdULL,0x3abbcd637cbcdefcULL,0xdd74fe7efc5dd6faULL,0xfffe7ffbbef47effULL,0xc1d4ef62b6745ce6ULL,0xfedbbebbbebcf7faULL,0xffbffffbf67dfffeULL,0xb5d17fe2fe985e7dULL,0xdaaa0d519614d66fULL,0xfbbef77ef63cd6eeULL,0xc9eb9a319434549eULL,0xbbbfffdbf4fcf6feULL,0xf6f7f6effe3d7effULL,0xdfffff7bf49d7eeeULL,0xfb967ffef4dc76fbULL,0xe3c7df7afcd4dcebULL,0xebfedfda7efcdffaULL,0x97ba47fbb47c5eebULL,0xfb7dff5edebcd6f7ULL,0xfae2ff65fc7556eaULL,0x8fd8ff37f4dcdeeeULL,0x3ff6bbd8b67cdef4ULL,0xc5bbdefff6d4dfedULL,0xffbefe737cd4dffeULL,0xf89e3f6ef67857f6ULL,0x828ef38ffe155efeULL,0xeedd1f92b49414faULL,0xfbecff594545ff2ULL,0xfd7fbafd7eb9d67cULL,0x7ffcef7afe3ddefeULL,0xdb56fa2fecd47f6fULL,0x84f212589c14367eULL,0xebb0deff1c9057feULL,0x47536e2d61436f2ULL,0xbfdebf74b65577feULL,0xcd86ff3bb430f6acULL,0xcbf7ffbbb6bcfffeULL,0x11c4b4c201c163aULL,0x36d4366a2c10f6e6ULL,0x90ab67411e1452e8ULL,0xdbef7ef6d631d7edULL,0xefdbffbaf6fc5effULL,0x6adfdf51b89cdcbaULL,0xb53adac3549d56ecULL,0xb7ff2f6a72d45dfbULL,0x12940f7c54101666ULL,0x8372fea27011166aULL,0x8080d220005660ULL,0x80804c6020001468ULL,0x933e8760a03cd462ULL,0xb447e6c2615366eULL,0xebb17f27cc81f76aULL,0xee4d4e10161036c4ULL,0xd89ac763a2f0d679ULL,0x312406633004d420ULL,0xe7bc2fe2409c56e2ULL,0xb02820d90141802ULL,0x32a0060000420ULL,0x0,0x2020000004000ULL,0xc2208041402ULL,0xc006912002104840ULL,0x8480c00002805040ULL,0x0,0x48000004000ULL,0x858b2b7fc8d4547aULL,0x0,0x1800000801000ULL,0x604612d000401481ULL,0x4b7c230201c58a8ULL,0x1000260004100154ULL,0x529e3cd12494566eULL,0xc1c4d77e38155ceeULL,0x450a394144146278ULL,0x82222560d2141468ULL,0x452e198000785058ULL,0xd3b4ff74f4195668ULL,0x8b67d2cff6d4556bULL,0x46e636dab2581467ULL,0x994c0290101c28ULL,0x8008a20002000268ULL,0x416e55e8d4141cf8ULL,0xc290fb2e340014a2ULL,0x6bdf36bb2a50dceaULL,0xceafcf7934755c6aULL,0x79e293ec3654576aULL,0xffbefff3341c3effULL,0xdb97fffff47cfffaULL,0xf9deee1bf6bc5ee8ULL,0x43dfcd5eb47d7efbULL,0xb5f85ee8f41474baULL,0x9ce363bb414582aULL,0x3902041c04901a40ULL,0xc023d108009c5e28ULL,0x200001000004000ULL,0xc0507c0001cc340ULL,0x514af7402430f42cULL,0xb62c710434d242ULL,0x4584010614140028ULL,0xcab69c411634c4c2ULL,0x94385f406095166aULL,0xc4f4140a141454ccULL,0x30bcb5c084941456ULL,0xb009322b01450a0ULL,0x80b00a02240410a8ULL,0x8040000000400ULL,0x454d4c0e0001062ULL,0x2920c3014141412ULL,0xcbc0f03a36b456e6ULL,0xd9429b4a74e414eaULL,0x25bec842001c128aULL,0x80808ce028001420ULL,0x2404000140012ULL,0x3000000008001000ULL,0xc061ee2014dc14d0ULL,0x84688410290544aULL,0xec0001006d4ULL,0x10083088000c4a0ULL,0x0,0x800408020000100aULL,0x802000000104000ULL,0x0,0x40a00000000040ULL,0x4080218080204460ULL,0x8110c92834b454b8ULL,0x2219dc8834905460ULL,0x8603433270d01c20ULL,0xb0522570041c3456ULL,0x10a340020104020ULL,0x90083021010cc00ULL,0x0,0x9119cd91a6b1d2e0ULL,0xa020220000105642ULL,0x200240100044080ULL,0x4004000600001481ULL,0x580e16b554305ca8ULL,0x1a020b4000941450ULL,0x24000001008ULL,0x180ad8030801428ULL,0xc9002401a014d022ULL,0x503080801c1002ULL,0x210961b08101748ULL,0x0,0x1180250000101a08ULL,0x228660500145028ULL,0x8b77f6bde0d456eaULL,0x20120450208410a0ULL,0x100000200001008ULL,0x1848a746303456a8ULL,0x1a20074eb019d560ULL,0x2ef87201e414123eULL,0xefded9dbfe1cd4faULL,0xbf3a4fbab4b4d6fbULL,0x24020804000484ULL,0x4a40200414210420ULL },
	{ 0x8b50c000c0c82100ULL,0x826617426124224aULL,0x4c2888001080260ULL,0xfde38e70d1ec327aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbfbfdffff96fb3d2ULL,0xd53e33eef1cd27daULL,0xda263f7099cda3faULL,0xe2a8ff62f044b2d0ULL,0xa8ef3e0bd9e523c0ULL,0xf3f75fdaf1efb3e2ULL,0xd674f4e2d1e7b062ULL,0x206f4768b10923c2ULL,0xd381af7391c4a342ULL,0x8801602a00c2042ULL,0xab5eb7dff1e7b2faULL,0xf5fffff7f9efb7faULL,0xfdf8fff6f1eda2daULL,0xfebf3ffaf9efa7faULL,0xc787f922d9c6b0eaULL,0xefffdfdff9edb3d2ULL,0xefbfcf6ff9cda7e2ULL,0xffffbffff9efb7faULL,0xfff6fff2d9efb3f2ULL,0xc9d6d73ad0e7a372ULL,0xe8ed1ec95166b2eaULL,0x1fdecf93f1cdb3faULL,0xd7fff7faf9edb7faULL,0xffbe7f72f9e7b2d2ULL,0xfa57eb2ed1eda3f8ULL,0xbefeb7eff9e7b7faULL,0xbfbeddcfd9eda7d8ULL,0xd4bcffe0d16da7eaULL,0xfefaeff3d1eda7daULL,0xdf2ee37ff9edb24aULL,0xfefef7b2f1efb3daULL,0xfbfedfdff1efb3daULL,0xf4fdfef6f1efb7faULL,0xffeffff3f1efb7daULL,0xfffe7ffff9efb3faULL,0xfdfffbbff9edb7faULL,0xfbffffdff9efb3faULL,0xf7ffdffff9cfb3faULL,0xfffffffef1efb3faULL,0xffff7ffff9efb7daULL,0xefcfff3ff9efb3faULL,0xeffffffff9efb7faULL,0xbfbfdffff9efb7faULL,0xfffffffff9efb7faULL,0xdffefffff9edb3faULL,0xe9f7fffff9efb2faULL,0xeffffffff9efb7faULL,0xff9feffbf9edb7faULL,0xfffff7faf9edb3daULL,0xffbefffbf9edb3faULL,0xfb52fb2ef9ed2648ULL,0xddff3ffff9edb7faULL,0x3fbfdf77d1efb3faULL,0xfcfeffffd9eda7faULL,0xbf9efff7f9efb7faULL,0xdf87fbfff9edb7faULL,0xfffffefff9edb7faULL,0xffffdfdbf9efb7daULL,0xfeffffebf9efb7faULL,0xfffffffff9efb7faULL,0xffbc7f7fd9edb7faULL,0xffffffbbf9efb3faULL,0xebfe7f93d9eda3daULL,0xbdfffffff1efb7faULL,0xf9fffff2f9edb3faULL,0xdbb6affff9efa3eaULL,0xcb76dfedd9efa7d2ULL,0x6f7efe9df9edb3daULL,0xbefbdffbf9edb3faULL,0xd9fffe7fd1cdb3faULL,0x8f387f6dd9eea2caULL,0xe9fffb1bd9cdb7faULL,0x6ed61f75d1e7a7faULL,0xb39e5efbf9cdb6daULL,0xf5fffefff1edb3faULL,0x7fee3df2f1eeb758ULL,0xf9d0f32ed96ca3eaULL,0x4fdb9ad9f9cdb248ULL,0x7ffddffbf9edb7faULL,0xfdfffe6af1efa3c2ULL,0x888c2d62716d12e0ULL,0xfbbef7dcf9cca372ULL,0xdfdbfed3d9cdb25aULL,0xbbfedf5fd9efb3f2ULL,0xf4f4fffcf1cdb3eaULL,0xdfffff7ff9efb3faULL,0xeebfff34f9eca7faULL,0xcddfd72ad1cea7e2ULL,0xefeffddbf9efb3daULL,0xb7fffffbf9efb7daULL,0xffffff75f1eeb3faULL,0xd8f6bf75d9efb3faULL,0xfd9efbfef1efb3f8ULL,0x2f56abf8f1cc23faULL,0x1fbb6f6fd96cb3faULL,0xf8feff7ed16db3f2ULL,0xfbf6bf7cf9ed23faULL,0xebffffaaf9edb3faULL,0xecc71fdad94db1daULL,0x4a9beb89d0edb2d8ULL,0xd77dbffbf1adb3daULL,0x7ebefffbf9edb7faULL,0xebd6f72fd9edb7faULL,0x8fc3a20bd0cd26caULL,0xafbfeffad96da7daULL,0xdc77bedcf1c333eaULL,0xafaeff73f9ed37d8ULL,0xd792f77f58e7b2faULL,0xcb7ef116f844a602ULL,0x2b95cfc7d947a2daULL,0x16e44d4091252260ULL,0x9da7ce5099ef24caULL,0xdbea257ff1ed83faULL,0x8dc7df8ad9e433d0ULL,0x237ff752d0ed3750ULL,0x189f0febd1c5229aULL,0x7100528891890240ULL,0x82020609080a040ULL,0x886a720a49432242ULL,0x2210040040048040ULL,0x95020566d1c4018aULL,0xc400003090008040ULL,0x4d003f4159e5a140ULL,0x8ab15622484c2658ULL,0x24a3057d9c02350ULL,0x8014449a80042148ULL,0x84917440504283d0ULL,0x4226060260841080ULL,0xa184f930d12ca2d0ULL,0x86603011010a2342ULL,0x249005829008a002ULL,0xd09c120080412440ULL,0x80dee520d184a270ULL,0x820000410000ULL,0x0,0x40200000042ULL,0x80040800000218ULL,0x200140008000ULL,0x0,0x60010400c2060ULL,0x404000174100a04aULL,0xcf1deacd14cb3daULL,0x532ef742d1eca772ULL,0x5a8b2fd4d08eb2d0ULL,0xcfc6ffb7d1edb6f8ULL,0x1c800a0d20c48350ULL,0xa25658818c2068ULL,0xd0e2967b7166127aULL,0xf4123950f94822e8ULL,0xe9c4640091cc02daULL,0x6e5a1682c8cca440ULL,0x33807a2d048a642ULL,0x10008dc0102002c8ULL,0x4884213101668270ULL,0xdbf7f70ef1efa3daULL,0xa9c3bb4981ed2752ULL,0x6b9fef7af16e33f2ULL,0x5dd0377ef8e8b6d2ULL,0xf7b63f61f9eda3daULL,0xdbbcfb7cf965a3eaULL,0xc9d63e9bd9c1236aULL,0x2f1fdbd9d1e6b3eaULL,0x74bd46eed9eca76aULL,0x9b2ef772e14c3378ULL,0xde963f4051c5b36aULL,0x6fc6ef1ed9eca7faULL,0xae7fc0ac061a340ULL,0x242e041410200ULL,0x9400008011400400ULL,0x20108610d0003040ULL,0x0,0x818134050402282ULL,0x45ae7da4d1ada3faULL,0x512a376bd165b7e2ULL,0x4e8e35035864b0c2ULL,0x8176300951240142ULL,0x22f29208500c2042ULL,0x800120000200ULL,0xd858828000cc8648ULL,0x2ae246050a42272ULL,0x4104600240009042ULL,0x800004c01080202ULL,0x8700cc1b71a4a248ULL,0x6411481040a462ULL,0x2200256141488620ULL,0x9a062625c1c103c8ULL,0x900198041048000ULL,0x869d1909d1c9a1c2ULL,0x8890800000640308ULL,0x43894058420250ULL,0x102d1000090340ULL,0x808840008080a040ULL,0x140000000000210ULL,0x0,0x8124202000402640ULL,0x4082004001010060ULL,0x68008328510ca212ULL,0xd0080140012250ULL,0xb6a8545291492542ULL,0x5020054010003082ULL,0x840060400080c0ULL,0x0,0x21000740d0040048ULL,0x1fba45ebd1e4a3d2ULL,0x64317278006ca252ULL,0xce00200251409248ULL,0x8200010461000200ULL,0xb460e00c10020e0ULL,0x20004000000208ULL,0x14050000000ULL,0x0,0x1280254018408288ULL,0x8fc233825184a29aULL,0x421c4748d06122d2ULL,0x84300162400ca140ULL,0x21285d73c0a4b2d8ULL,0x128c8e48d00701c2ULL,0xcf63ee2ad964a0eaULL,0x14e811041002232ULL,0x660045290c08340ULL,0x5120b6ded160234aULL,0xd00c200391410142ULL,0xc9e3d26898ed22caULL,0xef7fffdbf1cdb7eaULL,0x9faeedfbf1e6a3faULL,0x300c260208400350ULL,0x480ea122c0850040ULL },
	{ 0xc182f20b18266144ULL,0x2a7c9ec30a2c4700ULL,0x458852398302c06ULL,0x92cf0648d8306534ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9b8bdffbdabcefe0ULL,0x10001409c0202000ULL,0x550306090009080ULL,0x900401000020a100ULL,0x48b920190302600ULL,0x2b6c17811828c720ULL,0xa48c4628903a85b0ULL,0x8967237292ba0570ULL,0x733064445a202560ULL,0x8fcbfbbfcab8ed26ULL,0xbbafffdeda08edf4ULL,0x3dbbd5fbdbb8bf26ULL,0xfbfdbf2edabefde2ULL,0xdae6bf6ad1baadb6ULL,0x8dd4b3bad22ced56ULL,0x69ffffd3daa8cf46ULL,0xfebfeff39bbefde4ULL,0xfdffffffdbbefff6ULL,0xfffeed69d8b8ff76ULL,0xc9d3fb2b9bbcf760ULL,0x4cd73f659bbc6556ULL,0xdfbbffdfdbbcfff4ULL,0xf7f7ffffdbbcffe6ULL,0xe3fe2d7adbb86d34ULL,0xfbe5e3ae9abcfdf6ULL,0xeddbfffedabcedb6ULL,0x2bbfef5ad9be6552ULL,0xf5d636ee9aacfd66ULL,0xb6bebd719abeef86ULL,0xf636ef5ddbb8ff60ULL,0xd7febf67da38fd64ULL,0xfbffdfdfdbbeffb6ULL,0xdef7fffedbb8fff6ULL,0xffffffffdabcfff6ULL,0xffbeff7fdbbefff6ULL,0xffffffbbdabcfff4ULL,0xefffdfdfdbbefff6ULL,0xfefeffffdabefff6ULL,0xff7f7ffad8bcf7f6ULL,0xffbfbfffdabafff6ULL,0xefffffbfdbbcff76ULL,0xeffeffdf9bbceff6ULL,0xbfbfffffdbbefff6ULL,0xffffffffdabefff6ULL,0xfffeffffdbbefff6ULL,0xebffffffdbbcffe6ULL,0xefffffdfdabefff4ULL,0xbeffeffedabeede4ULL,0xfff7ffffdabcfff6ULL,0x7ffe7df3d3befff6ULL,0xfbf4eb7fdab8af34ULL,0x8dcfb7d9dbbcfd96ULL,0xfffecf7edaacfff6ULL,0xfdfffffbdbbef5f6ULL,0x9ffbfdf7dbbaff74ULL,0xffb0cb7adab86534ULL,0xffdffebfdbbeeff6ULL,0xefffdfdfdabeffe6ULL,0xfefffffcdbbeeff6ULL,0xdbffef7a9bbafd76ULL,0xfb9fefffd2acff34ULL,0xefdfffbedabcfff6ULL,0x7fffffdfd3a8ff66ULL,0xbfbfffffdbbeeff6ULL,0xfffffffadbbeeff2ULL,0xfe96b77f9abae7f4ULL,0xffd7f79fdbacefa0ULL,0x2f7efed9d8b8ef76ULL,0xfefefbffd33cefa4ULL,0xf565a7f9daa82f46ULL,0xdec465ecd8b8adf6ULL,0xe9ffffbfdbb8ef76ULL,0xc2d63e39dba8ed64ULL,0x969fd5eb9abc6534ULL,0x7f76f6feda38ef74ULL,0xefbe7ff292beaf22ULL,0x7385f36e9a386764ULL,0x7bdabfbb9aa8ff30ULL,0xd7becf7adaa6ff74ULL,0x71be76eb5baebd24ULL,0xfbbe6d7392bebf04ULL,0xfebc7ffe9aacf774ULL,0xf7ffe1b9b38edf0ULL,0xbed2cddbd8a2fd80ULL,0xb4d7deeddab0edf4ULL,0xcffffff1dabcfd64ULL,0xd7ae3fffdabcf7a4ULL,0xff8fb2bbd234b760ULL,0xebf9ffd3dabcbd64ULL,0xfc9fffebdabaeeb2ULL,0xdf7faffe9bbafd62ULL,0xfebebfffdabce7e4ULL,0xe9d7f737da3cefe6ULL,0xeb7fbe47daa8e766ULL,0x8772cbfb98b8afb0ULL,0xdffd75eddb3e2d54ULL,0xf63e7f7fdbbcefb6ULL,0xf9f2f73edbba7d22ULL,0xecde1fde91aedf32ULL,0x993e4f52d10a2724ULL,0xfd6da2c898bc6f26ULL,0x13a877e298bab514ULL,0x9a91b77f9aa0ed44ULL,0x94a7b565aa8ef26ULL,0x6aa9dfb9da242d26ULL,0xf5fffffedbbceff6ULL,0xffbffff7dabefff4ULL,0xfaa66ec6d8ba6d14ULL,0xcc6e6e3ddba8f510ULL,0x51b6cb57d1b8afa0ULL,0xf4fdf5eedabcfd44ULL,0x393ec662d2b43524ULL,0x780414a090002e00ULL,0xc8e649b098348550ULL,0x30815d9da226560ULL,0x1c787ce290804530ULL,0x9282ad40da982c26ULL,0x9aa4264cc080ad00ULL,0xc902d62858ace586ULL,0x2100109092084400ULL,0x8012006050802600ULL,0x8018200200888920ULL,0x82122e0408000502ULL,0xe914963a4aa40520ULL,0xd777ea1da28ad40ULL,0xe014c4411a002412ULL,0x402d1e000082524ULL,0x0,0x100030800002100ULL,0xc4230888820a800ULL,0x830e440a92342104ULL,0x440548690a02820ULL,0x10000002004ULL,0x0,0x41020890004120ULL,0x0,0x432202010880520ULL,0x50048000000ULL,0x0,0x800060490000400ULL,0x28821a0149ac2004ULL,0x44a606e790a8ad24ULL,0xd976af64d2884d52ULL,0x88ac213ad030a120ULL,0xffdefe4398befd36ULL,0x4460008002400ULL,0x22804228020bc00ULL,0x6940021880280500ULL,0x71063da01208e5b4ULL,0x808cd2815880f110ULL,0x40801c100820c480ULL,0x5d8a4d03daac6562ULL,0x90c2a20590a40444ULL,0x1c0005000820c100ULL,0xa9d6e37f9abcfda4ULL,0xcdd27adb9b38d760ULL,0x89cef709aa8af66ULL,0xf5fffeffdabebd26ULL,0xbffcff77dbbefff0ULL,0xcb94f3b4cb382666ULL,0xcc5fb29c9388d520ULL,0xb3d7dfdbdaac6764ULL,0xf1ff7eefd8bcf9e0ULL,0x9b60afd5c82c6760ULL,0xd6a4234f10286514ULL,0xc96323101028ec20ULL,0x6a68d94ad8a4bf44ULL,0x48508611022a484ULL,0xd26cbf44d8300f76ULL,0x4606340098086402ULL,0x4110000080080900ULL,0xf0414100248600ULL,0x369a43309a886544ULL,0x8310050890384400ULL,0x4020480000d100ULL,0xe17a92384c20ULL,0xa42124e1822ad80ULL,0x1218c203902085a0ULL,0x4406a0541aa82402ULL,0x5eb06fd19aa4d760ULL,0xca32017a9a3e4500ULL,0x62020810288420ULL,0xcac0c0a022c2520ULL,0x800010600a300402ULL,0x8100a700003881e0ULL,0xc024b33290202600ULL,0xcc40b00010881600ULL,0x0,0xe41450359aa84546ULL,0xb76f2f589ab4af02ULL,0x5384794640a82720ULL,0x800000400000600ULL,0x168001082000700ULL,0x0,0x100080000202000ULL,0x2014c90080500ULL,0x2c88431250326520ULL,0x4856388399880c10ULL,0x14a2446890203700ULL,0x8000200080100000ULL,0x6b3c7d76dabc2542ULL,0xcb81ba6ad838cd76ULL,0x406485888008c02ULL,0xf1bfdffa9bbaf766ULL,0xf7fbfe7bdbbcff74ULL,0xebfe7ff3dbbeed76ULL,0xcade23045838c530ULL,0x44d23a1ad2306744ULL,0x2410008008004600ULL,0x14c1162010882500ULL,0x88053000b02500ULL,0x19a040190242804ULL,0x147000010886004ULL,0x188c90040202604ULL,0x8010002000008c04ULL,0x30802624c0800504ULL,0x82202d4c5a248510ULL,0x80a600498880100ULL,0x84c41290204510ULL,0x8010002200080422ULL,0x4143846012200750ULL,0x8aba0d7a99a8ad24ULL,0xe886d02d9ab8e924ULL,0x2fffde9fdabeef04ULL,0x379b4ff8da38ed24ULL,0xc158572ac2accd00ULL,0x6902140298100c02ULL },
	{ 0xca3cd92e45067360ULL,0x8bf7bfdf57967761ULL,0x37a8d0400dd44342ULL,0x2400106a0128a260ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc30a4d234d067550ULL,0x141326a2b49e2160ULL,0x0,0xe22453f014960160ULL,0xa9467e99178e5762ULL,0xefb7cf55077e7f73ULL,0x1c9554e1450e3621ULL,0xf96e7673b9b66362ULL,0x99d561f11de25571ULL,0xffc7e73391be7f63ULL,0x95e454a1cce6d60ULL,0xfdfeffeb9fa67d61ULL,0xfdfcef7bdffef761ULL,0xdaae3d2c8d866a63ULL,0xefcfff3e7fd67763ULL,0x2f7ede5a7e86ed70ULL,0x96cb6fe2e74e7762ULL,0xf5f7ff7ffffeff73ULL,0xfffe3ff50fee7f73ULL,0xebf7ffaf5ff6ff63ULL,0xfff77f5fdfdeff63ULL,0xffffffffffe67f73ULL,0xf7f7feffbfee7771ULL,0xfff67ff39ffe6773ULL,0xfbf7ffff77feff73ULL,0xcfdbfffbdffeff63ULL,0xb7becff30f9ef763ULL,0xf5dffeefffd67763ULL,0xdb1eef77bf8e7760ULL,0xfbbe7ffbdffe6f63ULL,0xcffbfe1b5df67f71ULL,0xfbffdfdfffdeff73ULL,0xfffffeeffffeff63ULL,0xffefffff5ffeff73ULL,0xffbeff7fdffeff73ULL,0xcffffaffffceff63ULL,0xefffff5fddde7f63ULL,0xbdfffffbdffeff73ULL,0xff7fff7eefdeff73ULL,0xffbebf7fffdeff60ULL,0xedf7fffbdffe7f60ULL,0xef5ffffd9fdeff61ULL,0xffbfffffdffeff73ULL,0xfffffffffffeff73ULL,0xfffe7dfffffeff73ULL,0xebffff7ffff6ff63ULL,0xfff79fdf3ffe7f70ULL,0xbfbfdffbfddeff63ULL,0xd5e7fee7fdee7d71ULL,0xcbee7f7bdff67f72ULL,0xfbd4fb3a13cefc71ULL,0xeecfbfaf5fde6f71ULL,0xaffbdffbeff67f60ULL,0xf5fffffedffeff63ULL,0xdfffef77ddfe7f71ULL,0xd989ff32ebe6ff62ULL,0xddcfffabffdefd61ULL,0xffffffdffffeff63ULL,0xf6ffffffdffeff73ULL,0xfffeaffaffde7f73ULL,0xeffe7ffe1fd6ff61ULL,0xffefffbfbfde7f73ULL,0xefffdfdf7ffeff73ULL,0xffffffffdffe7f73ULL,0xfdfdfffbfff6ff73ULL,0xbeb6bff7fdfefd72ULL,0xefe7feb79fd6ff60ULL,0x9cdbfb5f1db67f61ULL,0x9fbedff61fce6f73ULL,0xf7b7bf7f3dee7e63ULL,0xdfee7b6ad3fcf763ULL,0xcbf7f3fbdfd67f60ULL,0xcfc73fde99f6f563ULL,0xbaa8effbcf9e7762ULL,0xfd7676f79fb67770ULL,0x7fbe5fe2bdee7d62ULL,0xfbe7ff2f3dd6ff73ULL,0xffffffffeffe7f73ULL,0x3fb0fff9ddde7162ULL,0x68775fea0fd67f63ULL,0xb7bff7720ffe6f73ULL,0xf738ede155e23571ULL,0xcfcffe963dd65f62ULL,0xae7edd5c17c2ff61ULL,0x56bbffe8ddfe7f73ULL,0xf1eefff16fce7d63ULL,0xffbe6f6f95967773ULL,0xeceffe0c9dde3761ULL,0xbf5fdfd38ff6fd63ULL,0xfc2bdfe785ce7f61ULL,0xfb5daf7e9ff27573ULL,0xfefb2f7e77de7f61ULL,0xe9f7ff3fcd5e7d63ULL,0xef7e19cbd5b67d40ULL,0x8cfa1dfb1fde6f61ULL,0xb3fdbf7d3be6fd72ULL,0xffbe7fffffeeff73ULL,0xa9dde33f5df66f63ULL,0x3bf37ff097be7f62ULL,0x328a26dc1cc65102ULL,0xbd77e3f01dc65d63ULL,0x9e2805fa0db60560ULL,0xfad7e63a7bce7460ULL,0xc96437df3dc24c62ULL,0x86bacebf4dd67763ULL,0xfdfffffffffeff73ULL,0xffffff773ffe7f73ULL,0xd290e76419864e60ULL,0x1fd6b62a8df66771ULL,0xab5e0fd30ef64760ULL,0x409cfeea4dc63360ULL,0xf9ceff7bbfd6f761ULL,0xd692077a15963160ULL,0xc4d3aa38578e6521ULL,0xa5f5fc663c66d61ULL,0x804a100a43062d40ULL,0x9ac325c295167d21ULL,0x8782a66021860570ULL,0xc824dd3c13ce4e60ULL,0x6bfd1b51c5b2af60ULL,0x8010000005000000ULL,0x9280281722084423ULL,0xc1005100100ULL,0x380508167406420ULL,0x40080603850c2020ULL,0x800084001002200ULL,0x750414d2cd407a00ULL,0x1000040000024022ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x800e0001020100ULL,0x0,0x14000080020ULL,0x8402004005020000ULL,0x0,0x200200004040ULL,0xbe81e51874a7463ULL,0x7832d7e00d845760ULL,0xb124a24881127942ULL,0x8004016008940500ULL,0xc5c0d214040a4560ULL,0x80462a6280866640ULL,0x1408400009400200ULL,0x8010201108403000ULL,0x210001400004821ULL,0xa026ca0684846762ULL,0x443320010424560ULL,0x2330440a108a2960ULL,0x910200600a4e0420ULL,0xc38270780556a160ULL,0xea98e62e1dce6462ULL,0x5dcbbf3f4dee3d70ULL,0xaf3e4ff85fd63560ULL,0xd4d55be88d027551ULL,0xffdeff73bffeff73ULL,0xea98f73dcdaeff61ULL,0xcff7be3305967f41ULL,0xeadecddfd7966f63ULL,0x64befee61cd6f570ULL,0xfbfef671edc87763ULL,0x8400396080002820ULL,0xc842b3b88606f560ULL,0x1200000040002000ULL,0x14066e864d066142ULL,0xa54d27482d527522ULL,0x6c9c2c7324863f70ULL,0x0,0xa20d504000283461ULL,0x23256480f425542ULL,0xc53c3d6b8dc4f360ULL,0x6d3025621d9eed61ULL,0xe980d332078a1760ULL,0x2943361001960640ULL,0xa020e0810860160ULL,0x9c540480810e2860ULL,0x8605c004446060ULL,0xa214e56e02960741ULL,0x8008308100424820ULL,0x8c16474039c40860ULL,0x1180180a09c40640ULL,0xa00000000840000ULL,0xca86e26414b60551ULL,0x100800000000410ULL,0x18002748088c4462ULL,0x54f5fefdaffe7f63ULL,0xbfefffff7fde6f73ULL,0xfb9eff7dbffeff63ULL,0xc040090040823200ULL,0xa0151b5008486c62ULL,0x0,0x8000060001424000ULL,0x8a10024400021540ULL,0x45cc500004424470ULL,0x1632a344a66651ULL,0x120a0f1340c66221ULL,0xa0248d5018f2a640ULL,0xee9609f06d966763ULL,0xc9daea670dd66f63ULL,0xf583e0187066d42ULL,0x8bbb7ffb7fd67760ULL,0xc479f367d586e462ULL,0xfbfefff3fffeff73ULL,0xcbf6ffff9fd67f60ULL,0xc7cffe2bcd7e7563ULL,0x8880812903021740ULL,0x5088826009c04640ULL,0x0,0x8000004002000400ULL,0x8020000000840000ULL,0x12410084002500ULL,0xc15006040126142ULL,0x462a5c088802640ULL,0xf292094c08d66341ULL,0xa0c7a8000a420560ULL,0x290000180426000ULL,0x5a0003a09c4a430ULL,0x84c9221028527860ULL,0xd2a63f6e759a7961ULL,0xc75fe07b95c67b60ULL,0xebdffa5fcfd67f62ULL,0xbeb24cbe55a26160ULL,0xd3befffaffdeff73ULL,0xdf9efe72cf4e7f72ULL },
	{ 0xcbd2ef6ef69e3f4aULL,0x403a126b47d92c08ULL,0x939ec433d7280fc0ULL,0xb0d2463877982648ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42805000a012800ULL,0xbc70da61f71e2c48ULL,0x87163522dd940ec2ULL,0xcf956f77ff9c3ce0ULL,0xafc6febb77bc7df9ULL,0x2b29074244d06460ULL,0xb4f9bce0742c3c42ULL,0xd0c0646212147240ULL,0xd99c6d77573c2fe8ULL,0x4842fe36d63f6e01ULL,0x135552527aac2ce8ULL,0xb4fdd7efdefa6fdaULL,0x91ed2f7a5ebffe08ULL,0xffbc2de4deac2ee0ULL,0x4feef7aadeffeceaULL,0x6afa7ec2f6bd3f50ULL,0x5fba4fffee3d2ca8ULL,0x35ced768df9caee9ULL,0x71b67e55de1e3c50ULL,0xf37bef6bfeff6fe8ULL,0x4dff3f15dedf3cf0ULL,0x9fbbfffefedf7ce2ULL,0xd6fff7fdfeff7de0ULL,0xeffe357bfe9f3dc0ULL,0xfbf6f7ffffbf7fe9ULL,0xffd7becbff3eeeb2ULL,0xfbbecffadeb77ee0ULL,0x75fff6ffdebcfde0ULL,0xf6bafdf3dfbfbee9ULL,0xdfee7f7ffffe7febULL,0xcfffff99df9f7ee3ULL,0xbfffdfdfffff3ffaULL,0xffffffeffebffff9ULL,0xfffffff7ffbe7eebULL,0xffbeffffffff7feaULL,0xffffffbeffbf7ffbULL,0xefffffdf7fff7de8ULL,0xffbfffeff7fe3ffbULL,0xfbffff7a7fdebffbULL,0xfebeffff7ffffdfbULL,0xfffff7ff7eff7ffbULL,0xfffffefffffffff9ULL,0xbfbbfffbfffcfff9ULL,0xfffdbfffffff3df9ULL,0xfffeffffffbf7ff8ULL,0xebf7ffffffbffff9ULL,0xffffbeffdffefffbULL,0xffbfefffffffbeeaULL,0xfdfff67ffffefd70ULL,0xdffefde3d7bc3cf3ULL,0xfbf7e3ffff9ffefaULL,0xefff7e9fffbe7ff9ULL,0xfffffffb7fff7efbULL,0xddffffee7ffe6feaULL,0xf7eef7f3febebefaULL,0xbbefff7effbc7fc8ULL,0xcf5ffe3ffefcacf9ULL,0xffffdfdfffffbffbULL,0xfefffffffffffffbULL,0xefffbfff5efc3e69ULL,0xfbbf7f7d7efe3febULL,0xffcbff3ffef7ffe0ULL,0xff7fffdfffbf7efbULL,0xfefbfffffffeffeaULL,0xd9fde77fffff6efbULL,0xfebe3ffef7bf7eebULL,0xcbf7ff7f5eef3debULL,0xffde1ffff7ee2dfbULL,0x96b3b7ff76be7de0ULL,0xf7edb7794fff7de2ULL,0xc72635f1dedf2cc2ULL,0xeb8fe33ed6e03c61ULL,0xc2b35e40f6a43e40ULL,0x9ffb6ffb77d67de2ULL,0xf776f77ef2be7fe2ULL,0x7ffe7df25ebf7f68ULL,0xdbd2e7ffdeef3d68ULL,0xe9e6b6f3d7df3ee9ULL,0x4e9cd7df5ebe3ec3ULL,0xf5ffffe17eddaf41ULL,0xbfff7f76ffbe3ed2ULL,0xdbacf76edebe3f41ULL,0xd7dbbfbb5fbf2db8ULL,0x93dacf1bce9aad18ULL,0xf4befee5de1ebfa3ULL,0xdbefdf62d69c7ceaULL,0xfbb3ef7bfbfe2ee2ULL,0xdffeffbbdebe3efaULL,0xafffd7dbffbffd21ULL,0xb4e705e8f71a2c4aULL,0xf9bdff7fdedc3f88ULL,0xf6bebf6fdebc3ec8ULL,0xcf5fff365fbf3ceaULL,0x489a3e725eea264bULL,0x96bacef2d6cdbeeaULL,0xf76cfff1ff9a7ce8ULL,0xfffe3ffefebe3ffbULL,0x817eeb6976fe3ec0ULL,0xfbd77ffb5effbddbULL,0xfdbecd71539e3ce8ULL,0x72c4916943143439ULL,0xdf8e3c30de802c60ULL,0xcb9ee37e7fb2ae68ULL,0xc85eb20b759f6ce0ULL,0x27ff5ffa7cfe7ef1ULL,0xfffffffffffffffbULL,0xfffefff7fffe7ffbULL,0xffb7695ac6be7de9ULL,0x8da3f31ad6daa460ULL,0x325313c0e6801450ULL,0xa8500438149648c0ULL,0xfbcf777bdfdfffaaULL,0x4eba2f75762d3f48ULL,0x9145ff224fb42482ULL,0xc5c5048069424e0ULL,0x3b0f7804e952980ULL,0x82a8a56aaa106d08ULL,0xca800520442404c0ULL,0xc112873a43d429a0ULL,0x22319cc182072ca0ULL,0x11841320a842480ULL,0xc1ee06ca579a2e40ULL,0x80244c42902480ULL,0x8c1012002b00400ULL,0x40000800001400ULL,0x88440081b00c00ULL,0x201000200400ULL,0x600044082200400ULL,0x800000844000000ULL,0x8000000102002000ULL,0x0,0x0,0x20040000800040ULL,0xb720094854a00240ULL,0x4080000000002800ULL,0x1000010040000080ULL,0x10001000000420ULL,0x2100044000000ULL,0x84000008000400ULL,0x0,0x17f4c4056303cc0ULL,0x2000002040106000ULL,0x0,0x280c0042000400ULL,0x4080c20444041480ULL,0xe3e83c18e922d00ULL,0x10024206100400ULL,0x9480841040013c00ULL,0xf886fde2422e3c88ULL,0x41c2a042568c2420ULL,0x0,0x211480002802441ULL,0x4400a2e060003c80ULL,0xc9bc37327d9e2cc3ULL,0xa85c420062b06c88ULL,0x9efa3fd7f7be3de2ULL,0xe79f5ffbf6bf3ff9ULL,0xb6f6ffeffead3fdbULL,0xbfbcff77dffc7dd9ULL,0xffb6ff7b7efe2ec3ULL,0xcdc3b733f6367cfbULL,0xdb7fdfdbf7963de0ULL,0xfd76fee4df3e7ee1ULL,0x986aa7e0c6232cd1ULL,0xc9806c2244101410ULL,0x5000110a46122c00ULL,0x210000000900000ULL,0x8c54c8aa020d00ULL,0x4100b46056042880ULL,0x2143468d6081c40ULL,0x3800844000000ULL,0x229398ca74002d48ULL,0x900430006000460ULL,0xd02c35708ac22c20ULL,0x4a222c40562c2d48ULL,0x8970faf07e3e7e68ULL,0x6fff3feb7f1ffd62ULL,0x2040617ca003400ULL,0x2404026e40de4802ULL,0xe6da2120821c2c20ULL,0xc094832c061a3c00ULL,0x1860000144020ULL,0x2622dc103b040448ULL,0x1d5404081a0c0480ULL,0x979c25225e106a00ULL,0xab92ef4fd3fc2d6aULL,0x8160c12a56802240ULL,0x4a00030840822840ULL,0x1cf3f6e07fb4bee9ULL,0xddcfffdb7fdd7fe8ULL,0xdfb8ef7ef6b67da8ULL,0xf9dafb325f3c14caULL,0x8ebcc8508a982d48ULL,0x280040002002800ULL,0x100054001a000040ULL,0x9b008660569c0440ULL,0xc180d122c6982c80ULL,0x9411c064a003410ULL,0x20b0200842042080ULL,0x90682753d6843fc2ULL,0x6aee2d74df1c7f13ULL,0xcbccf646769f2c82ULL,0xeaf6dfce479f3ce8ULL,0xefbecdebffbaaf08ULL,0x943cd642cf1a2d00ULL,0xfffe7ffbffbffffbULL,0xfbfeff7fffff7ffaULL,0xf8c2fe9576b62ca9ULL,0xb317066072a22e41ULL,0x10c1040054100402ULL,0x9238240070222051ULL,0x20000200040080ULL,0x0,0xa210440005140020ULL,0x448524182023c01ULL,0x56352ec2242ULL,0x8200010442003410ULL,0x100a322246c02448ULL,0x1ac495047143d80ULL,0x4143d32354924c68ULL,0xc075aa3054340468ULL,0xdf14bfee76be3ef8ULL,0xedbed275dfb82fc2ULL,0x2affff5ffff63de8ULL,0x9fbf7fefd7bb3de8ULL,0x30251440002c32ULL,0xcd467b61ae906400ULL },
	{ 0xf9feef6a9b966f2eULL,0xe7f62a9a9ac72726ULL,0xc38cfd99ad12f37ULL,0x8c1054e0aa9c2214ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80400080040000ULL,0xf6f7feffebdc27f3ULL,0x9b64ad629b8e2716ULL,0xfd0cf7ff9b8d6727ULL,0xefd73eafbbf82b3eULL,0xbbe2dbdb9b9c639eULL,0xf6bfb4ee9286234fULL,0xe9ceffe48bbc230cULL,0xc98617379ac26b07ULL,0x88c7e30bc3816f8eULL,0xb5cdd4a9b046f04ULL,0xbfff57bbbb892fd7ULL,0xd17aaffa9bdc6bb6ULL,0xfa92abe09bd46bbfULL,0x49d7d6be93942f06ULL,0xae9f7bc39bf52f1eULL,0xf7eb5ff39afc6b3cULL,0xd5e8b7c8ba982717ULL,0xb3362d60cbdc2033ULL,0xcadefbab9bbd6337ULL,0xeed51a559af22334ULL,0xbfaedf7b9a9a6b9fULL,0xf1f5b6fd9b196f7eULL,0xe7feadfafa9f6f3eULL,0xfb9fef6fdbbe6fddULL,0xffea3ffdbbde2f36ULL,0xefaf5dfbd3bf673fULL,0xf5f7feefdbff6fbfULL,0xfffffd70db9c6fbfULL,0xdfe7ffffdbbd6fbfULL,0xc7eebe1fbbdc6fcfULL,0xffdfffdbbbff6ff7ULL,0xfffffffffbff6fbfULL,0xfffffffbbbfc6fbfULL,0xffb67ffffbbf6fffULL,0xdfdfff3edbdf6f3eULL,0xffffdfdfbbfd2f3fULL,0xbefe7ffbdbec2ff6ULL,0xfbfffff6bbdf2f37ULL,0xcebf2ffffbfe6f9fULL,0xffffffffdbdf67feULL,0xefffdfdbbbdf6f3fULL,0xbffbffff9bff6fdfULL,0xf7fffff9dbbf6f77ULL,0xfffeffff9bff6f7eULL,0xfbfffffffbbf6ff7ULL,0x6bdf9ff7fadc6fe6ULL,0x9f9ecfbb9bfd2bb7ULL,0xf5b67effbaaf2b7fULL,0xffbe7d73faec6776ULL,0xfbf6fbef9bbb67ffULL,0xf7fbffdfd39f2f7fULL,0xefbfdf7f93af2fffULL,0xf7fffffefbff27ffULL,0x9ffcff77fb9f671eULL,0xfeb7fffe9b9c6fbeULL,0xafdefa93fac56f7fULL,0xffffdfdbbbff6fbfULL,0xf7ffffeedbfc6fffULL,0xfd6cff3a9b9c6f37ULL,0xdb8fbf7fbbdd2fffULL,0xffc7fffafbdf6ffeULL,0xebffffdffafd6fbeULL,0xfdfffffffbdf6fbeULL,0xfffffffcdbff6f3fULL,0xfffcffffdb9d67ffULL,0xeff6fffd9bdf6feeULL,0xefff5edffabd673fULL,0xbffb7ffbbbed6b7fULL,0x97fe7f4a9b942323ULL,0xbbbe7f7acb992b1eULL,0xc9bef3669b8e6776ULL,0xffda7a8fbb98673fULL,0x3fbbdf7aeadc2bd7ULL,0xd514b6768bc4675aULL,0xf7fefdf39b9f6fb7ULL,0xc3d5ff2f9b8d632fULL,0xf6cabf4b9a3c6355ULL,0xeebfcffb9b1c673dULL,0x7f7ffff58bfe633eULL,0xdfddff77dbfd6777ULL,0xedfbff7fbbd46fb7ULL,0xefd7febfcbb86b5bULL,0x3bdf99429aba6313ULL,0x9dff7eaa8bb12516ULL,0x11ecbf2892862716ULL,0xf99ef77e9bb46f07ULL,0x67def71ffbde2f56ULL,0xab8dffdbba9e273fULL,0xf66357e99b872f56ULL,0xfd7faf3b9bb92f9eULL,0xfebebff3fbff2fffULL,0xefc3e36adaac273eULL,0xafde7e539b886fafULL,0x3e2a6d539bb0623eULL,0xfb3dff7282be2f5cULL,0xfafe7ffddbdc27b7ULL,0xdbb9ff4fbb9d6326ULL,0x9be71bfffada6f1fULL,0x98b9cff29a05430eULL,0xf23672739a5c0306ULL,0x6e862fe99bdc292fULL,0x9a81a7279291233fULL,0x2cc124183840140ULL,0xa3becf72dbd4671eULL,0xfdfd7ef9fbfe2367ULL,0xdfbbaf779bbc6f7fULL,0xd9f1b73792a86e8eULL,0xa34cbc02a2912133ULL,0xbb4c07da19402787ULL,0x740262e082112107ULL,0x656ea77182ba6f15ULL,0xc7ae39678b9e0312ULL,0xced29103a3902712ULL,0x20a1542880c2281ULL,0x49743789a902323ULL,0x8010258092000334ULL,0x44062ec993a02a17ULL,0x6cc4426480042205ULL,0x8f6dac288802312ULL,0x8038082102005ULL,0x4142701b8102215ULL,0x4084018c8a222004ULL,0x280600209002000ULL,0x8080000000800010ULL,0x80800000040200ULL,0x5855300098a02310ULL,0x50a404108b100302ULL,0xdb91ef0682102b14ULL,0x4483fa3388142387ULL,0x16ccc082002102ULL,0x42202a080002402ULL,0x0,0x30a00e629a08092cULL,0x8060a428a802050ULL,0x0,0x298c46089002208ULL,0x2100002000000ULL,0x820000000a00ULL,0x0,0x35354408a080140ULL,0x588080010102ULL,0x0,0x10001000002400ULL,0x8950d2008a040300ULL,0xa44837c81a482e04ULL,0x0,0x60080000100ULL,0xca9eac7092862730ULL,0x890a400098506301ULL,0x60870890930e0b02ULL,0x1900010283000006ULL,0x841422420b000304ULL,0xdfa6fd629b812b57ULL,0x8104ab4cb2002026ULL,0x6c7b3e919a8c678cULL,0xebb6c7f893972336ULL,0xfcfffeee9b966b3eULL,0xa954ba029a99636eULL,0xdbb4f37bbb9e6737ULL,0x4fd7af1edb932336ULL,0x9a0effd3fbd42f96ULL,0xdcbd7ee98b996377ULL,0x690c27e292844b36ULL,0x10306a029000611dULL,0x6400c02888c4200eULL,0x64014498082004ULL,0xc23204620ac12022ULL,0x100140880802002ULL,0x2810255c81242312ULL,0xc0a0880019800822ULL,0x40000800010002ULL,0x3490600282140214ULL,0x149048080000018ULL,0xc9e2254182306304ULL,0xc3fcf59bbb556166ULL,0x2600000090100100ULL,0x1028a900200ULL,0x10e270c08840220fULL,0xcd242c6a8ae0610cULL,0x4000240080102000ULL,0x49c3121093122007ULL,0x8021406082106040ULL,0x10e412088a400014ULL,0x9e0201508280268eULL,0xdb83fb33bb302f0dULL,0x10920000802420ULL,0x300c0202262200ULL,0xdcb6f2f59a962b1dULL,0xb37b7f7b9bde2b17ULL,0xebb2657f988c6b3fULL,0xe7eefe0afb6d6b1fULL,0x2edd5d9abbc62336ULL,0xb8aacbeb9a9d2327ULL,0xca69a542bbcd2317ULL,0x5a0e2c70dbcc6800ULL,0xc8c2820490146381ULL,0xc075168393840324ULL,0xc010052098400842ULL,0x99eeb7499b596b0fULL,0xdfcebd46bbd42327ULL,0xcbb6f1bbbacc6736ULL,0x6fd71ed19a980b26ULL,0x4c86081100000ULL,0xc4222010a0846402ULL,0xfffefffbfbff6fbfULL,0xfbffff7f9bbe6fffULL,0xc5d8102f9b000b36ULL,0xfbcddcbbadd2707ULL,0x2410088a040000ULL,0x9a98ad709ac44b0eULL,0x1804000888004200ULL,0x8320040210090206ULL,0x102014002000006ULL,0x54845420c184630eULL,0x890a5f5a82276a0eULL,0x5090a3449a802316ULL,0x444a7ab8b820231eULL,0xa2bccf4a921c6324ULL,0x110a464293904b12ULL,0xc0c0270282a42404ULL,0xdff63ff6db9c6347ULL,0xadffff33baa4271fULL,0x4efdde4f9bd12fcfULL,0xbd39cf739bbc2b32ULL,0x800024218a0e0000ULL,0x3dfe7d739b1c6f1cULL },
	{ 0xebfffffef1bddf7eULL,0x6d4612996198c3bcULL,0x3fbceffbf1b5df7fULL,0x8c121288f000c00cULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf4ff76effbbaffffULL,0xb3392c55e93457aeULL,0xffff2f7febb4f7efULL,0xa661203d110d32cULL,0xa9da6d5f81b0d7edULL,0x76f7f6e6f1b0d3bfULL,0xfdbbef7efbb5d3efULL,0xd7956975d1a38735ULL,0xa0d2ec42e8885198ULL,0x3b5e595ec084d06cULL,0x28ba4be9e190cb2eULL,0xf9acfe3ac39bdb6dULL,0xde3a2f6ed13bdb2eULL,0xffffffeffbbedfaeULL,0xefffbfdff1b0dfbeULL,0xa3ccfe1cb20fbaaULL,0xd1cdb660f1b2732eULL,0xff9eabf6e1bc533dULL,0xe394b372c080f37eULL,0xea4f5f5ff1b4d32aULL,0xaffbff7be90df72eULL,0xb4a3f3e9e998c338ULL,0xdfbebfd3f1aeffbfULL,0xfbd5f72cf9badbfdULL,0xefeeffffd9bbff6eULL,0x3fbbcddbeb95dfa9ULL,0xf5fdfeefe1b7f33eULL,0xfffe7f72c9bbf3fdULL,0xdfb5ef7fd1b7d3ffULL,0x5fdebe3bebb5db3eULL,0xbffbffdbf1b4ff2fULL,0xfcbfffeffbb5f7bfULL,0xfffffffbfbbfffffULL,0xfffffffffbbfffffULL,0xfffffffffbbfffffULL,0xebffdfdbf3b7dfefULL,0xf7be7fefeba7f7efULL,0xffffff7ffbbdf3fbULL,0xfebe7ffbf9bfd3ffULL,0xffdfffbef9bffbbbULL,0xfffffffff9bfffffULL,0xfffffffff3bdffffULL,0xfdfff7faf9b7d7bfULL,0xffbe7ffffbbffbffULL,0xebfeffeff9bddbffULL,0xfefffffff9bff7ffULL,0xfbbfcffff3bbf7feULL,0xddf7f6fbe1b1f7ffULL,0xdfe63dfae199dbffULL,0xf3bfff3ef9b5fbefULL,0xcfffff7bf9bff3bfULL,0xbfffdff6f9bcdfefULL,0xf5ffffeffbbddfffULL,0xfdfeff73f995ffffULL,0xffbff7fff992ffefULL,0xcff7fe1fe9bdd33fULL,0xffffff5ff3b6dfffULL,0xf6ff7ffafbbeeffdULL,0xdffef7faebb3d7ffULL,0xffbe7ffcf1bdd37fULL,0xffdfffbfe3bdf7bfULL,0xebffdfdffba4fb7dULL,0xbffffffbf1b6dffbULL,0xffffffffd19fdbffULL,0xfab7bffff9b7f7ffULL,0xefffffbef9bfd7ffULL,0x6ffffffbe3bff7beULL,0xffbf7ff3fbbcf7feULL,0xff7df7fde9b7d37bULL,0x7384adefe1a1f17bULL,0xabd6f3fec93edf24ULL,0x6cdb5ff9e1bbd36bULL,0xbb3bcfa9e13cf36aULL,0xf75fdf7ef383d7abULL,0xdba6f5e361b6c3fbULL,0xcbe4ff6ef937d3acULL,0x4dcf3691d191c33eULL,0xbffbdf5bd0b0dbfbULL,0xf5fefffaebbfdbeeULL,0xfbaffdf2d9b0ff7fULL,0xc294ffbef13ed72eULL,0xffdfbe9fe9bed3afULL,0xaa5a97def9a4d33aULL,0x809ae670a3025a68ULL,0xe36a2762d924532eULL,0xfa973fd4d184d3aaULL,0xcdcffe36f8bddbebULL,0xaffd7dfe1b1f329ULL,0xc5f6ffedf19df22dULL,0xd1afff7ff9b87f7aULL,0xee8a0d5e6192f17aULL,0xc7dff6f6fbb3f7bbULL,0x4bdfffd9e9bdc3eeULL,0x9dbf5f6fe9bff32eULL,0xf7febf7af9b6d73bULL,0xfffe7ff3f9b9dfefULL,0xebf0f737c3bdd7baULL,0xeed316cff19dd364ULL,0x85a3cf12f0a0d209ULL,0xd7c0f6784088c33aULL,0x77d625f2d9a5db38ULL,0xda30dbfecb35f7beULL,0xb5b9ec8d198d000ULL,0xffbfdf45f1b0df7fULL,0xaef776fbcbbdd33fULL,0xc88ebc51fa3cc7e4ULL,0x620065002001c108ULL,0x54cbfe38f3305f3aULL,0xa51817d3c0a5d73cULL,0xf43b76f46985c12bULL,0x7faeb768e83c4f2eULL,0xffbf7f74f9b5f3ecULL,0xdfe7ffd9419ff7ebULL,0x7ae09dd189bc5568ULL,0x1811e84041108342ULL,0x5daedcd441b8c369ULL,0xc200260241100302ULL,0x2e726ae110912eULL,0xbc71ab9c992cb7bULL,0x8e122df04181d14aULL,0x822860000108ULL,0xc1a0047ce200d024ULL,0x184c31851900120ULL,0x41180001000060ULL,0x692d7e0d0a0c19eULL,0x3934f2d609964371ULL,0x0,0x200400000000102ULL,0x21000004002ULL,0x100001000800101ULL,0x400002001000020ULL,0x88102012018040a1ULL,0x900400208084000ULL,0x40200000100200ULL,0x14000004002ULL,0x48001000020ULL,0x12043460a080d002ULL,0x4042000000160ULL,0x0,0x0,0x4000200800001ULL,0x200c0000201200ULL,0x0,0xdc20e01e000d004ULL,0x2c48a944a0904384ULL,0xc228067051029228ULL,0x80a42504ca005004ULL,0x960c3c604080d1f2ULL,0x8b00a032d121d342ULL,0x8000a0e61029208ULL,0xb2924000e020c208ULL,0xc022900058004102ULL,0xceaced4881b85727ULL,0x9988d7b0f390d738ULL,0xcefffabbe9b1d37eULL,0x2f9ecff8d191d3efULL,0x30e0c568c190832dULL,0xed4cfd61eba7db7fULL,0xdfac775ee115cb2eULL,0xc6cb5630c1a0f33eULL,0x8a1e4d49c131c328ULL,0x34a12ca950805289ULL,0x1b68e622c180e22dULL,0xcbd2679f7021f30aULL,0xc2a9d215d13058e2ULL,0x1484848c0041922ULL,0x81a806424020c328ULL,0x10800600008c020ULL,0xf88403228280d21cULL,0x4080c08080904008ULL,0x215090400100c10aULL,0xab324ffac9b6f6a4ULL,0x814004400900c120ULL,0xd2a2014e83a453e8ULL,0x80085208020c008ULL,0x22681f532891c31aULL,0x805920181c209ULL,0xd7a69078099065a3ULL,0x25280c0000a06334ULL,0x8953d18c7121c7acULL,0x20421a8a40108328ULL,0xa0048100100c128ULL,0x5dfc7e0bf330d30eULL,0x0,0x4280c20ae1909149ULL,0xcc3af1ae113832aULL,0xc21f52dbf184d128ULL,0x74d550a8c1141566ULL,0xccae2f60e3b0daa9ULL,0x898e7726e1b3eb8fULL,0xeffeff3ae9b7db7bULL,0xbfc5b51dbb4d16bULL,0x97bf5de6d134d728ULL,0x51ccf64a8134c3e8ULL,0xca2c1b4af1a1d32aULL,0x89dbf72a81b2d37eULL,0x31a10c1001020ULL,0x22a1cf8b89384624ULL,0x92082415c131832cULL,0xf3b02c7001a6d3beULL,0xa957f200e9a2c37bULL,0x9c21a01c0a0d308ULL,0x100c410288005120ULL,0x34002042085000ULL,0x50b62c5001901729ULL,0xdadfff7dfbb2d3abULL,0xa8e39710818c51bbULL,0xdb9edffad1b1f7ffULL,0xc0108620010c4128ULL,0x281500190030bULL,0xc00008020000c008ULL,0x18302100848020ULL,0x8802490041105080ULL,0x100452684080d300ULL,0x2804466051040160ULL,0x52022a016111c000ULL,0x90206a02800051a8ULL,0x2752165279804188ULL,0xe082312001804103ULL,0x332ce6415301e328ULL,0xfb982f7ff993cfbfULL,0xc1e7ce2ec1b4db7bULL,0xa1eb5d7c995d36aULL,0x843a6f73e112e378ULL,0x712c26080a88417bULL,0xff3e7f7ce3bdf7feULL },
	{ 0x4bbdf3f20bd4c38aULL,0xeb6606868bc6e06aULL,0x7ab8c9e388d6f96eULL,0xdc766ca689d6cb2aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xeffcf36f8bd6f362ULL,0x9eeafd728357eaf6ULL,0x92b530678a44385aULL,0x0,0xa754401803c48202ULL,0xd6f2b6e583d7e3feULL,0xed0a6f6383d5e3aaULL,0xcd96392303d2e322ULL,0x86cf59d38bd6f372ULL,0xaf18dc8f8354e2a2ULL,0xf69354a18354f3f6ULL,0x716d266083d5e306ULL,0xfadf2f748bd7f3e2ULL,0xe9c5f3a709d6f3eaULL,0xf25eb7d28bd4e3e6ULL,0x969047e081c2d1caULL,0xb7fca75983d7f3f2ULL,0x3fde2fe083c7f3f2ULL,0xeba3f52b8bc6fb6eULL,0xeac31ea78b45a8aaULL,0xf79fdffb8bd7f37eULL,0xfc7fbf7f8bd7f2feULL,0x53a8ef6a8346b3eaULL,0xfbb7ef7e83d7f3eaULL,0xda5bfeff8bd7f3eaULL,0xbfbfcf5a8ad7fb62ULL,0xf5d73f7b8bd7fbfaULL,0xfffefff28bd7f3feULL,0xffacffff8bd6fbfeULL,0xdfd3f6bf8bd7f0feULL,0xcfffdfdb8bd7fbfaULL,0xffffffef8bd7fbfeULL,0xbbffffff8bd7f3faULL,0xffffffff8bd7fbfeULL,0xffffffbf8bd7fbfaULL,0xffffffdf8bd7fbfaULL,0xfdbfffef8bd7fbfeULL,0xffffffff8bd7fbfeULL,0xdebfbfff8bd7fbfaULL,0xefdffffb8bd7f3faULL,0xafffffff8bd7fbfeULL,0xbfffffff8bd7fbeeULL,0xf7feffff8bd7fbfeULL,0xdffe7ff78bd7fbfeULL,0xebfeffbf8bd7fbfaULL,0xffff7fff8bc7f3feULL,0xdfbf6ffb8bd7fbfeULL,0xf57fdeff8bc6f3faULL,0xdfbe757b83d7fb6aULL,0xfbb8ff3f8bc5f376ULL,0xeffbbeff83c7abf6ULL,0xffffefdf8bd7fbfeULL,0xf5ffdfff8bd7ebfeULL,0xffbfff7583c7fbfeULL,0xfbfcff6f8bd7fb7aULL,0xcfd7bf9f8bd7fbfeULL,0xbffeffdf8ac7f3faULL,0x7fffd6ff8bd7fbfaULL,0xdbcafff98bd7fbeaULL,0xffbe7ffe8bd7fbfeULL,0xead6ffb781d7f3eaULL,0xbf7fdfda8bd7ebf2ULL,0xfebfffef8b86fbfaULL,0xfdffffff8bd7fbfeULL,0xfebebfff8bd6f3eeULL,0xefdffffe8bd7f3beULL,0xefffff5b8bd7ebf2ULL,0xf7ffefff8bd6fbfaULL,0xf1ff7ff98bd6f33aULL,0x7f3ebd630bd6a3a2ULL,0xabffffbe8bd7fbfaULL,0xefdfbfff8bc7fbeaULL,0xbb9fcfbb8bd5f1eaULL,0x531ef2f38b44f3eeULL,0xcfbced7b83d7b31eULL,0xfbfffbf682d7f3faULL,0x8e7b7bdf8bc6bbeeULL,0x77bedf758bd7b3a2ULL,0xb5f6ffeb83c7e3faULL,0xdffeff738bd7fbfaULL,0xefbcdfff8bd7ebbeULL,0xcffa9f298b56f9eaULL,0x6f3e5fda8156f3f2ULL,0xb4fefeec83d6f3a2ULL,0x9fecad6383d7b372ULL,0xf3967e768346abe2ULL,0xf5dffefa8bd7ebe2ULL,0xaffedfde83d6bbfeULL,0xc5ffffef8bd7f3faULL,0xdb6fa6db8bd7a3e2ULL,0x7e3c2c7d8b56e1aaULL,0xc1c7f73f8357faeaULL,0xef7fffdb8bd7fbfaULL,0xbfbfefe28bd7f3f2ULL,0xd1fefffa83d7f3feULL,0xfefe7fff8bd6fbfaULL,0xabbcf3be8bd6f3aaULL,0xefff9f7f8bd7f3f2ULL,0xb21ffdc08b55e3a6ULL,0xf534d6f68b57d3b2ULL,0xaaacff2382c2b962ULL,0x8306d26a8b44300aULL,0x81660f4882c4c052ULL,0xb14cfef8ad7e17aULL,0x86f1f6e889d4b2feULL,0x9276ef228bd5e2baULL,0xc906f3388b469260ULL,0xc943f237825672a2ULL,0x80184a420146a0a2ULL,0x94bc8eee83d4f33aULL,0x5940024009c42012ULL,0xcf3e7ffc8bd4f3caULL,0xedf7d8958356f2baULL,0x1abfff598bc7c372ULL,0xfcd777ff82d7fbe6ULL,0xf1fcbefe82c6f366ULL,0xc2362fc48285c902ULL,0xcc2f20301d621b2ULL,0x2c36e2c28b54400aULL,0x101b4de080469132ULL,0x4000a64288542200ULL,0x610004c00068220ULL,0x144a52000c4a022ULL,0xa800e0202522022ULL,0xc10480000568220ULL,0x0,0x0,0x2212400080849002ULL,0x8c83280282859002ULL,0x222004000200a200ULL,0x943364e68a46e16eULL,0x10ac0c608294d042ULL,0x2801014000c0ULL,0x0,0x1a8200a000ULL,0x8000040000020200ULL,0x0,0x8080080000004ULL,0x4020000040200ULL,0x0,0x120100b0300a0a0ULL,0x50a0040003840002ULL,0x440008000101c000ULL,0x40c00800818000ULL,0xa413160108542022ULL,0x0,0x4069052883462100ULL,0x460e3d640306a232ULL,0x0,0x8d41e100145c260ULL,0x81ce8c98211f102ULL,0x712ca2e90856a33aULL,0x73be7f6381d6fbfaULL,0xb90ffb2e83c7e26aULL,0x8ffbbfdb83c6fbeeULL,0x26be4f618bd7fb7eULL,0x1430080046002aULL,0x192ead208ac4b342ULL,0x9a82eb7c8bd6bbfaULL,0x82c3972b80567260ULL,0x8158cc190046a12aULL,0xd4b37ae483c6f3e2ULL,0x4826768254aa02ULL,0x30464310a028122ULL,0xc89db2289867ac2ULL,0x40d0815980862aaaULL,0x2440b380529340ULL,0x1024000400040212ULL,0x90010000060030ULL,0x0,0x941104409460342ULL,0x29b459b88b87a3b2ULL,0xb4a470600356b030ULL,0x645000402102ULL,0xcb803b6f82d6f162ULL,0x60c41fd00bd4800aULL,0x441c44c481446298ULL,0xd5acd2518954aaaaULL,0x398615e38917f332ULL,0xf31477230ac7816aULL,0x802258000002a2e2ULL,0xa020c2230054b030ULL,0x829516420150a1e2ULL,0x508002000040102ULL,0x8184002400062220ULL,0x94410800205a212ULL,0x6b0cd1108204aba2ULL,0x71547882c3e160ULL,0x98025508244832cULL,0x9bbf3f718bc6da6aULL,0x8ddbff3b8bd7fbfeULL,0x2b5ed3d189d7f3aeULL,0x142147e10bc483aaULL,0xc85420420a54900aULL,0x6898afbe0b5660d2ULL,0x894ffb2b83c7a3faULL,0x3886d2f500d6b242ULL,0x87120c360854a2a0ULL,0x2065360c8bd6f122ULL,0x828c388000559012ULL,0xa9d7f92b8b46ebaaULL,0x6c4b82108144f0a4ULL,0x800480000458000ULL,0x4000200002008000ULL,0x4000000200020002ULL,0x4000000200008100ULL,0x142000a00808200ULL,0x3b8442c03d6b372ULL,0x11d206a08a02aae2ULL,0x10b0014000002122ULL,0x491fd3798356b3c2ULL,0x840161003042002ULL,0x1481200c48000ULL,0x45446e38284a3a0ULL,0xc80026008300e202ULL,0x14202e448086a162ULL,0x3038e228804a102ULL,0x108000040200ULL,0x100000200040200ULL,0x14025008200e020ULL,0x9ab493368386e3d2ULL,0xcb97a6b28bc5f3aaULL,0x687e96c58387e3baULL,0xdfba4ffa8b56f3faULL,0xf1e9efeb8896db42ULL,0xdfb63ff68bd7fbeeULL },
	{ 0x68dbf38b0f473e08ULL,0xeaf63f53c5c33a14ULL,0x17dacdfa435a7a34ULL,0x5df7defacfd3fab4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf95d17622fcf3a99ULL,0x5bac2d71cfce3e10ULL,0x700051fc44ca1214ULL,0x6cd92a647cbfa8cULL,0x2bbedfd3c8ca3e3dULL,0xf4bfdee44fc97e10ULL,0x53ff67fb48ca1eb4ULL,0xdeb02f6e47c27e30ULL,0x346a95246871210ULL,0x1affdf9fcfcbfe34ULL,0xbebb15ebc5cb7e1cULL,0xf8c4f67acfcb3a3dULL,0xf2992f5eccdbfe20ULL,0xc9defb9eeecbfe08ULL,0xafd7fb5bedc23a0cULL,0xde1a5fba86ca7e18ULL,0xafffaf7beecabeb4ULL,0x4ea675e2ecd31aa9ULL,0x8303ae6b0d527e10ULL,0x63523feb65dafea9ULL,0xfdbecff7eec3fe35ULL,0xd6f9b27d4dc2beb8ULL,0x447c37304cdebe1dULL,0xfbdefb7fcfd3febdULL,0xffcfffdde6db7e9dULL,0xbffecfebcfcfde35ULL,0xeffee7ffcdcbfeb0ULL,0xeffefd774fcb7e39ULL,0xfbdfffffcddbfe99ULL,0xcddffe3b4fdffabdULL,0xbfdfdfdfefcffe3dULL,0xf7fffeffefdffebdULL,0xffffff73cedbfebdULL,0xdfbfffffcfdffeb5ULL,0xefdfffffcfdffebdULL,0xfbffffdfefdffebdULL,0xbcbfffebefc7febcULL,0xdffdff7befdbfeb5ULL,0xfebfa7ffeddbfe3dULL,0xeff7ffffefdbfeb5ULL,0xefffffd3efcffebcULL,0xbffffffbefdffebdULL,0xfdfffffbcfd7beb5ULL,0xfffeffffefc3febdULL,0xdbf7ff6eefde7eb9ULL,0xeeffffefefd3fe35ULL,0xbfffeffacedbfebdULL,0xd7f77ff2cfdb1e34ULL,0xfffefffb4fdffe31ULL,0xfb96f307cfcbbeb8ULL,0xfe7bb7dbe5da7ab9ULL,0xffffdffbefcf7ebdULL,0xf7f7ffffefcf5ebdULL,0xfffeff73efcffeb0ULL,0xfea6ffffe7cb7ebcULL,0x6fdffa1fefcbfebcULL,0xfffeffdbefcffabdULL,0xf6ffffedefdbfebdULL,0xd9ffff7befcb7e31ULL,0xdf966f7f4ec97ea5ULL,0xe7efee3bcfdffeb0ULL,0xa7defb57cfcbfe38ULL,0xffffffffefdffebdULL,0xff7eefffefcffeb5ULL,0xdfbe7fffcfcf7ebdULL,0xefcffffeedcbfeb8ULL,0xefffffffc4cffebcULL,0xfebfeffbcfc7fe25ULL,0xf77f9ff9cedb7234ULL,0xdbc67dee4fde7eb9ULL,0xfbfdf77ecfdf7ebdULL,0xbbdbbfdbcfcbbe35ULL,0xfbbccfffcf9efe3cULL,0xf721b8fdcfcfbe95ULL,0xfbbefdfaefd7febcULL,0xd9a4ff3bcdc77eb4ULL,0xbdf3b39bcfd37a95ULL,0xaf3fdffa4fdf7e9cULL,0xddf476eeccdb5a14ULL,0xfbfc6d73cfc77ebcULL,0xffbeff7ee7cbfebdULL,0xcdffbe3fcfdb7e39ULL,0x7bdcdddfcfcbfa21ULL,0xd645d6fbcccbbabcULL,0xd1eebfa04fc2aa31ULL,0xfbbebffe4f8e7e3cULL,0x4caec1aecc2b690ULL,0x6fbcfbd3c6cf6a24ULL,0xbcbbffeb4fcffe31ULL,0xf5e7b7f2ccc3eea4ULL,0xfca52feccc837ea1ULL,0xcb5ac77fa4cbfe08ULL,0x2f7f7e6f67db7e18ULL,0xf6faffe9cfda7eb4ULL,0xf3efff7fcfcb7e98ULL,0xd7f27f65cfc35e94ULL,0x7b5fdb2e4fcb4ea4ULL,0x4ff7bfd88fc37ea9ULL,0x7aa0a4e3ee470230ULL,0x98753614244b3a01ULL,0x97bc2d68e7d7feb8ULL,0xa9ccc116cd422e30ULL,0x876716204483ae10ULL,0x8a9e4f52cfc13e84ULL,0x3092547445c20a14ULL,0xdb9877c00c4e1619ULL,0x4a82b724c59b7808ULL,0xc945db21ca430e24ULL,0xaa27dfdacec936a0ULL,0xd414166dc1824e14ULL,0xa5dbe770c5823e31ULL,0xee043f6e85cb4a08ULL,0xc1e6e93ccd8b3e94ULL,0x736116d140c30e35ULL,0x3c751ffb6c8a7691ULL,0xfbfce6caee4a76b4ULL,0xc20404a28ac81280ULL,0x84603084c420020ULL,0xb100184802280ULL,0x14885281ac882200ULL,0x952c977325c37a11ULL,0xd8040710c4c20eb0ULL,0xaafffe6b4ecad6b4ULL,0x6ec10edfc8020634ULL,0x100650040a3000ULL,0x20004000400800ULL,0x1280004040400220ULL,0xe130d120c0c25a00ULL,0x4723082c8c20814ULL,0x0,0x0,0x804092008020608ULL,0x100020540081400ULL,0x142100000121200ULL,0x0,0x0,0x0,0x0,0x0,0x90040140000010ULL,0x100408ae68805600ULL,0x800400000400200ULL,0x540c36f58cc3aa00ULL,0x200200080800000ULL,0x10e000000081280ULL,0x20004831a0828200ULL,0x8500d00284828280ULL,0x7f9e3feeefca7ab4ULL,0xc02020400a400200ULL,0x8050014a048e0210ULL,0x9892ac8245c32a00ULL,0x7621b3f264833a24ULL,0x6ffc6df0cecffe28ULL,0x3d7762e4d4a7a10ULL,0xece6de524cc8de1dULL,0xefa64de0058ffe80ULL,0x4012606440921600ULL,0x1b9a1d70c68b7e2cULL,0xd9b6df3b4fcb7a3cULL,0x4cc3ea134dcb3e18ULL,0x2b1debdd25c36ea0ULL,0xb0919ceaeec24a00ULL,0xd0e30f490d0a3e81ULL,0x85ef2084873a10ULL,0xdcc34e0f41cadea0ULL,0x2bc6d551ec83ba0dULL,0x16420b420e0a3208ULL,0x240080800000ULL,0x80000b0000000680ULL,0x0,0x90d108044027400ULL,0x34a24ec005cb9808ULL,0x5475d058ce4e3684ULL,0xc690090046422600ULL,0xebbefb6f4ddb5a30ULL,0x6ccd1f4804cb5e10ULL,0x23288240c8420400ULL,0x11c2964ac8c08024ULL,0x5214019040826a00ULL,0xb0046f2cccc30e2cULL,0xc0d212b8468a3604ULL,0x148e5008441230ULL,0x3400000003000a00ULL,0x20240306822200ULL,0x102020480810200ULL,0x200200000600ULL,0xa6b1412c5801a84ULL,0x14b056684d836a30ULL,0x97432170cdca9828ULL,0x92222a4044c23498ULL,0xedc6eb3bcf8afe10ULL,0xe41d9042c70a3a14ULL,0x811000041020800ULL,0x808600006001080ULL,0xaa123c1c46c38208ULL,0x8fdbfa0bedcf3a39ULL,0xab7f7ac36cc36e21ULL,0x13880b1020c6aa00ULL,0x20448568440a7600ULL,0x2802d0c46400608ULL,0xe9eee22f4dcb3a10ULL,0x4c4b3a0c86c3a608ULL,0x380c00840c01280ULL,0x424424004503200ULL,0x250108801200ULL,0x8310ef2e6ec20688ULL,0x82328a42101220ULL,0xc4841000a001600ULL,0x64040004422200ULL,0xb58ce5344cc37e08ULL,0xd0b8f31c4e437e1dULL,0x786700a8b805a80ULL,0x800000800800200ULL,0x1436746ac5ca7810ULL,0x1210010004820008ULL,0x3210080442030238ULL,0xe586040046420a18ULL,0xa04860204c3b000ULL,0x40c00a040841280ULL,0x802846014c02220dULL,0xdbb527f4ccc7269dULL,0xeed2e392e4c35e0cULL,0x6ff5dffecfcbbebcULL,0x82aacffb4fd33e90ULL,0xd3bc9e46eecb3e99ULL,0xfa1627e24fcefab4ULL },
	{ 0xc9f7f33ffa7f9affULL,0xeef7fbdaeb7a9befULL,0xefbf4dffe97e39efULL,0xa7040400000188aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd8fcfeda6a7618efULL,0xdabff772eb771bcfULL,0xcdacf441a8650a6bULL,0x4fd9be9f6a569049ULL,0xaf5cdfdc2a761b6fULL,0x74e6d4f9ba731b4bULL,0x936ef7188a4e29cbULL,0xfab421f3a87a92cbULL,0x49c6be323a7a32cbULL,0x8b5ed7db687a9bc9ULL,0xb43777efea731bcbULL,0x8168ff53ea5610cbULL,0xfbbebf72eb7238cfULL,0xcddefbbfeb7610cfULL,0x6df7dedf2a7b9bcbULL,0x2e947de968783bc7ULL,0xf47df67c7a729bebULL,0xd2a6f5ef7a761b8fULL,0xfbdeffaeab7e19dbULL,0xecdb1efb6b7e9befULL,0xbfbfffeeea779bffULL,0xf79cf3f9f876b3dbULL,0xc3ae7dfaea6e1bdfULL,0xdbd6ff7eea7a93cbULL,0xefdabfdb6b773bcfULL,0x2fbfdfdb6a73bbffULL,0xf8f7d6eaba76b9dbULL,0xdfdead723873bbdfULL,0xfb8affbcaa5b9bffULL,0xdbdffebbb87ebbefULL,0xfb7fffdffa76bbffULL,0xfcfdfffafb7f9bffULL,0xfffffffbfa7fbbffULL,0xffbfffffeb7f9befULL,0xffffffbfeb7f9bffULL,0xffffffdffb7f9bffULL,0xfdbf3ffbfa7e9bffULL,0xfbfffffbfb77bbefULL,0xffbebffffb7fbbcfULL,0xfffff7febb7ebbdfULL,0xfffeffdffb7fbbffULL,0xffbffffffa7ebbfbULL,0xffffffff6a7ebbdfULL,0xfffefff76b7f1bcfULL,0xfbf4f77ff97f3adfULL,0xffff3fdfeb77bbdbULL,0xbebf7ffbeb7f1bebULL,0xfffdfffefb7bbbefULL,0xffbe6df2eb7f9bebULL,0xfbf7ef7fab7ebbcbULL,0xfefffeffeb7abbefULL,0xfbbecf73aa7b1bebULL,0xf5f7dfebfb7793cbULL,0xdffeff77fb7abbffULL,0xdfbeffffeb7a9afbULL,0xdb96becfea739aefULL,0xdbfedfdbfb77bbffULL,0xfcfffefeea7fbbffULL,0xffffff7ffb7e9befULL,0xdfbef77f6b7abbcfULL,0xfffdef3fba7a3bebULL,0xeeffffdffa73b3dbULL,0xf7feffff7a7f1bffULL,0xf3bfbfeffa7fbbcfULL,0xbeb6bf7d2a739bcfULL,0xc3e4ef6b2a799acfULL,0xabfffed7ab5ebbf3ULL,0xe6bbc73228521919ULL,0xddfd76fa6a539abbULL,0x71f63f6228539bcbULL,0xc3d0f77afa729aebULL,0x6fff3b5f687e9bdfULL,0x9801dcdc4a5418c3ULL,0xc5f633fd2a6f1acbULL,0xdf9efd796b78b9cfULL,0xfb94e64362509bcfULL,0xebfa3edfe97e385fULL,0x2bbecf7deb5219cfULL,0xf4ff57eeeb7e9bf3ULL,0x951ef7f0e876136fULL,0xfe97ff6feb7e9bdbULL,0xd7dffe3b6a7b1bcfULL,0x8afddfd7aa729a9fULL,0x757dd77faa7e38cfULL,0xddeef7e26b7693c9ULL,0xfbfde77f79723bcfULL,0xefd7fbea7b7f9befULL,0xe37fdcdf2b763acbULL,0x74bf3fe33a7bbbdfULL,0xfb77ee7b3b779bebULL,0xdf8a2f5f7a7810cbULL,0xefdff7ffeb7fbbdbULL,0xbffffef4e977b3cfULL,0xbbbfdff6bb7fbbcfULL,0xd93fbd6a2a721aefULL,0x5ab63c746a72bb8bULL,0xe302f3e77a769987ULL,0xeaf66e8beb7b9acfULL,0x171fcffe2a7e98f7ULL,0xcdd9f6740a7e19cfULL,0xd6b02d40b87212a7ULL,0xe1f4e4200a621007ULL,0xcfda0e3e7a7312cfULL,0x9b4855b2852109eULL,0x64fe066bea73130bULL,0xe49e4d72287c00eeULL,0x7204fc013056114aULL,0xd6443e09285e134bULL,0x431285c329501bcaULL,0x34c00ec4a86419c2ULL,0x330a0761285a91c0ULL,0x200100101000ULL,0xe8d3f3f86a721abbULL,0x2a18dfdb1a5f1bcbULL,0x34235324a86a2a56ULL,0x902425426860808bULL,0x200250200501800ULL,0x47c2073ce852100bULL,0x80a01541810110aULL,0x226680a621107ULL,0x1db404000a561203ULL,0x4870217048621a83ULL,0x42080000002ULL,0x2153320828629006ULL,0x2004bc0a1400001ULL,0x4040821000400242ULL,0x4022080002400009ULL,0x830038009010ULL,0x48400050a8008102ULL,0x802094038001180ULL,0x400200001400000ULL,0xa080052108181147ULL,0x40002401200880caULL,0x623800a8400200ULL,0x8010000000006ULL,0x3022400008701010ULL,0x100002008000100ULL,0x0,0x2001c00001501000ULL,0x0,0x80020e920842108aULL,0x3880502008101302ULL,0x64121b802a62100bULL,0x400212000400007ULL,0x24125f010a600083ULL,0x24000080040ULL,0xa0008b208a080842ULL,0xcb727c5068721ac9ULL,0xe1b0e6222b7e10cfULL,0xa9741dcaa97201ceULL,0x4498c5a1084219cbULL,0xd15dd569a85212cbULL,0xcf9c3dc8ab7a114bULL,0x8b87a7771a7e114fULL,0xcf5bba2b6a7ab04fULL,0xef3e4ff82056988bULL,0x4010504202521041ULL,0x3bb4af303a7b1bcfULL,0xdbbfee3f2a7f9aefULL,0x8beb5e41005218afULL,0x2414465208600140ULL,0x8014002000601882ULL,0xc16ff6306a339303ULL,0xfb226750ba541bcbULL,0xe8acd22e8870b3cbULL,0x165640002244134bULL,0x80023d410060020aULL,0x2200242428500080ULL,0xd0100b520864094eULL,0x602200010104aULL,0x4a080002500002ULL,0xb43854e24a669801ULL,0x18b414025840100fULL,0x160a286008431360ULL,0xe1d6faeaea7c38fbULL,0xaece036128700a8bULL,0x0,0x75602100469029ULL,0x499615621a62110fULL,0x0,0xc806001100109203ULL,0x48840b00a00094aULL,0x34605a822213020fULL,0x202c50a8020081ULL,0x80a00000400068ULL,0xc4c286082a0200cbULL,0x931c4dd16872810bULL,0x9c1272015850198bULL,0x41c8454042551249ULL,0x4692ab5c28761b29ULL,0x659adb38ba7592cbULL,0xa9573aca2872b23fULL,0x8490422119601182ULL,0xb18601088a400a42ULL,0x9191013008521182ULL,0xef45f58ba87a91cfULL,0xa9ca1211887838c3ULL,0x898cbf30872129fULL,0x1000a76020420282ULL,0x1844e9a148521083ULL,0x4828f2224a741981ULL,0xcfc52e18e85693c7ULL,0x34100100a0000086ULL,0x2202402000008aULL,0xc20024010a021850ULL,0x8000113848541881ULL,0x0,0xc80000641806ULL,0x0,0x0,0x8000004002000002ULL,0x800020000000202ULL,0x2000800000000082ULL,0x4451522948423303ULL,0xa500a6100a001901ULL,0x8081498008201a82ULL,0x102801060020001ULL,0xa24e1409005002ceULL,0x1020044002001003ULL,0x8040a8408060000eULL,0xda912f3e2b7f1bcbULL,0xc356ef792b7f92cfULL,0x2bfeda593b7c13efULL,0xbfbb6fefea76b9cfULL,0xf17ffd62aa7e9bcbULL,0xff9e7d63fa7e9befULL },
	{ 0xe9f5fb1ee37cdfc6ULL,0xefffffdee77ffff6ULL,0xdf9ddfffe71f1ff4ULL,0x1425b4ef203fa9f6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x385ca7d8013e3dd4ULL,0xfb3eedf2653f19d4ULL,0xcd86f72fa3665dc0ULL,0x41e55b8a663f14e4ULL,0xe39ecf52223cb7d2ULL,0xd4b47d7ee33c5f74ULL,0x9eefff76a37f9fc6ULL,0xd3928f76a21c9ed6ULL,0xa4d9ea30252d9fd6ULL,0xa82dbd9633d1fc4ULL,0xf69f7fe7e13cfdd6ULL,0xdbfcef7b627f3df0ULL,0xdea67f7ce77e1df4ULL,0xefd6f7f7a77f1fe4ULL,0xefffffdbe77e5fd4ULL,0xcebb5ffa667eff76ULL,0xdfffb75ee37dddf6ULL,0xfd9e7fff657e9fd6ULL,0xebfffffbe77ffff6ULL,0xefffbfffe77fbff6ULL,0xbfbf4efae37f9ef6ULL,0xddf7ffffe63d1ff4ULL,0xfdbe3df2e77f3de4ULL,0xbbffef2f677fbed6ULL,0xeecbf7fbe77f1ff4ULL,0xfbbedfffe33f9ff6ULL,0xffffffe9e77edfd6ULL,0xdbfeef76e47e9fd2ULL,0xfbbefffde77f3ff4ULL,0xdffbff7fe57f7ff4ULL,0xefffffdfe77efff6ULL,0xf6ffffefe77f9ff6ULL,0xffffff76e73fdff6ULL,0xfffe7fffe77ffff6ULL,0xffffffbfe77ffff6ULL,0xffffffdfe77ffff6ULL,0xffb7dfffe77edff6ULL,0xfb7feffee77ffff6ULL,0xdfbf2fff673cfff6ULL,0xefdfffefe37d9ff6ULL,0xffffffff673f7fd4ULL,0xffffffffe77f7ff4ULL,0xf9fdf67aa23d9df6ULL,0xfebe7ff2e77f3ff4ULL,0xeb9bffff677e9ff6ULL,0xefdfbf7fe77ebfd6ULL,0xdfbfefffe33ffff6ULL,0xff7feff9673ffef6ULL,0xfffefdf3e77fffd6ULL,0xfbf7f73ee37ffff6ULL,0xcdfbbf9fe77d1ff4ULL,0xefbe8f7b437ffed6ULL,0x91d5b7ff667f5ed4ULL,0xffbffff0677efbf2ULL,0xfbb6efff677fdff6ULL,0xcfeff6b3673ffdf6ULL,0xffffefdfe77ffff6ULL,0xfeffffede73fbff6ULL,0xfffff77ae77ffff6ULL,0xf3864775643c17d4ULL,0xc7cfffba657f9fd6ULL,0xbefcd3df233e19f4ULL,0x7df7f7ffe23e5ff4ULL,0xf9fdfffee77fdff6ULL,0xfebe3ef2e73c99d6ULL,0xc5fffb3ce33ddcd6ULL,0xee7f7acba77fdbf6ULL,0xc1374ff3e65f3e94ULL,0xd08df66aa53c1cf0ULL,0x8be66f65a77f1dd4ULL,0x8b35fbfb673d9cf6ULL,0x2fdbdfffe43f7dd4ULL,0x4c01493202141580ULL,0xf597a6fea37e3f54ULL,0xc7a87f53a52c9fd2ULL,0xdb96d31e637edce6ULL,0xea67bbdae27e1df4ULL,0xabbe6fffa33d5ff0ULL,0xbfffde78673efff6ULL,0xf3f8af70643ed8f2ULL,0xffb7b3ee617f9ff2ULL,0xcfceff5f673f1e74ULL,0xe3dfdbd2603f30f4ULL,0x14fcfeefe37e7ef4ULL,0xdf7efe7ba77fffd6ULL,0xffbf7f7ce73ffdf6ULL,0xc6f3ef72e63fded2ULL,0x6ffedfdde35d3ff4ULL,0xf6bf5fefe27e9f52ULL,0xd9acbf5ae25e9db6ULL,0xfeb6ffffe77ffff6ULL,0xedffff7fe77fdef6ULL,0x7dfefefc613fdfd6ULL,0xb6b9dffb67769f52ULL,0xe9d597e8277f93d6ULL,0x7ee67ffe677ebff2ULL,0xeb87f76e437e9ff2ULL,0xecbc9e87633e9e46ULL,0xb6bdfefce63e3df4ULL,0xf5f5f67be77f9dd6ULL,0x6eb02d50623c8576ULL,0x1880890c00100470ULL,0x801b98a2007500d0ULL,0x2400001a24041440ULL,0x51507668211d99d6ULL,0xb398ad1022371744ULL,0xba94373663349d86ULL,0x89428010000c8412ULL,0x89020fc382151d94ULL,0xf4106660213c5f14ULL,0x9627a68015d2c54ULL,0x12b40f78403711f0ULL,0x20c47080204c04d0ULL,0x2840d162148042ULL,0x4023306010e0840ULL,0x1106a060e7040800ULL,0x4a002104021c1040ULL,0x450a020005402c0ULL,0x5d4c40621802d4ULL,0x0,0x3564562a635412d0ULL,0x6010b59124461450ULL,0x954d626415c0ef4ULL,0x2a4b9659003591d6ULL,0xb020040040300880ULL,0x8000000000040040ULL,0x60702443802e8852ULL,0x100810440041400ULL,0x4080840001040880ULL,0x808054000160480ULL,0x4b4240001040cd0ULL,0x8308ec01242e14c0ULL,0x30002400c0040c00ULL,0x8020c00000011100ULL,0x9ac8c1200101450ULL,0x80000800100004ULL,0x0,0x31000a14001c16a4ULL,0x0,0x0,0x80100000100004ULL,0x40640124108006ULL,0x12800940211c0200ULL,0x8100106125000140ULL,0xa04005000340e00ULL,0x402002040000ULL,0x91c1224800340e80ULL,0x1a0405cc61140244ULL,0xe946e1ba631d1dd0ULL,0xe5de1bdb633d3dd4ULL,0xe61cf772647f35d4ULL,0x94d61270242c1d44ULL,0x93bc2d61e21cd8d6ULL,0xdbc2bb7e617f9ae6ULL,0xbff2b68b673fdf56ULL,0x1f194d02e2345bc4ULL,0x22034802040e40ULL,0xc99e77b0a57f1de4ULL,0xd91e4336e21c2cd0ULL,0x89acc603607710c0ULL,0x210060a007a1060ULL,0x367736e8c0361fc0ULL,0x8bc2563641893c6ULL,0xfb977f37267fbff6ULL,0xd481eb12003c11e4ULL,0xaa94344ba2240c00ULL,0x8014102022341140ULL,0x4000221060200040ULL,0x9a342c5062521254ULL,0x8000080000080020ULL,0x1100002040000ULL,0x1318450201341960ULL,0xa02496eb647e19e4ULL,0x47e46e2d653c1d64ULL,0xbc3ef06627f15c0ULL,0x9432a38603d1424ULL,0x410030000240c00ULL,0x8400044000101040ULL,0xe3d02d21263c1454ULL,0x80400000100080ULL,0x482200000140420ULL,0xa28c8c381660c40ULL,0x51745466a2164840ULL,0x8064546060103054ULL,0x41200062000880a2ULL,0x4200200400241100ULL,0x83091402002c1480ULL,0xb4914fec677c04c0ULL,0x492c320860380050ULL,0xe292ef86661d1790ULL,0xfadcc11ac11c9df6ULL,0xa6d68656607c1df4ULL,0x4222a6860182c70ULL,0xf164f270006f1950ULL,0xdaaa2940032e8d52ULL,0x6986c31b243d1dd0ULL,0xb361e5a60367f04ULL,0x11200d10603402c0ULL,0x8100044a001413c0ULL,0x1986ade222350ec4ULL,0x280e00200181150ULL,0x24cf1fa1405615e4ULL,0x2840300018042ULL,0x0,0x1a645022421140ULL,0x8390d323625c2664ULL,0x800008080000800ULL,0x6025f1064001510ULL,0x241608000815c0ULL,0x91b8a105001c06f0ULL,0x4000852000181320ULL,0x41a00000089482ULL,0x804a400400c0ULL,0x0,0x203402003400c0ULL,0x520e449000362404ULL,0x44c64026611900c4ULL,0x400813c122208042ULL,0x8400202000101040ULL,0xed42ff63633f0ae0ULL,0xca89273627345bd4ULL,0xfbd5fbbee67dbfd6ULL,0xfff1f63647e17f0ULL,0xdfbefffbe77edfd6ULL,0xf32bdff8637f17d0ULL,0x3bfe7d6de71edad6ULL },
	{ 0x8852f0ca880087d8ULL,0xaff7ffdf7f88cf9bULL,0xf6ffcffeffa9cfdbULL,0xf777feeb7b08dbd3ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf1ed227a4e29e7d9ULL,0xd6beef774f28cf1bULL,0xeb9fff7f6f09efdbULL,0xc9d3f28a530955dbULL,0x6f7fcf5e3f09f59bULL,0x74f6f7ef1f09ef5bULL,0xddffbf6eff29ffdbULL,0xdfdc35723a09ebdbULL,0x51c2f6161320cf5bULL,0xabfabfdfffa9f7dbULL,0xd4fb5feb7f21ffdbULL,0x4964fb606e01cfdbULL,0xdeb63f776b29cf53ULL,0xe7c7ffbf4f21cf5bULL,0xfffeffde7fa8ffd9ULL,0xb6ba5fff3fa9cfdbULL,0xddffbffeff29dfdbULL,0xd7ee3d745aa1d7dbULL,0xa9d6f77f7a29c75bULL,0xef775f3f9f21f7dbULL,0xdf3ddff32fa0cfdbULL,0xfff7ffff3ba9dfdbULL,0xdee677d10f89cf5bULL,0xebf2ef7f3f28ffdbULL,0xefffffff7fa9dfdbULL,0xafbfdfff7fa9ffdbULL,0xf5ffffffffa9ffdbULL,0x97ffff536b28ffdbULL,0xfbf7ff7fefa9dfdbULL,0xffefffbf7fa9ffdbULL,0xbbffffdf7fa9ffdbULL,0xfebffffffb29dfdbULL,0xffffffffffa9dfdbULL,0xdffe7ff7dfa1ffdbULL,0xedcefbfe7a28ffdbULL,0xefffffdfff28efdbULL,0xffff7fff7fa9ffdbULL,0xfdffff7f7f29dfdbULL,0xfebeff6fbf29f7dbULL,0xf9f7ffbfdf28f7dbULL,0x7bdffedf5f29efdbULL,0xb7bb5fffdfa9dfdbULL,0xdffdffefff28dfdbULL,0xfff67ffdff28efdbULL,0xebf7f7ffab21efdbULL,0xafffbfdfdfa9ffdbULL,0xfffffffbbfa9ffdbULL,0xfffffefd7fa9ffdbULL,0x7fbefd7a7fa9ffdbULL,0xfbd7ef76ff29f7dbULL,0xffffbfdb1f29dfdbULL,0xafffcffbce29ffdbULL,0x71febffc5b29e79bULL,0xffb2aff1efa9dfd9ULL,0xeffe6fff7f08dfd3ULL,0xcbfbfe8f9a09ffd1ULL,0x267cc7596a20819bULL,0x6cf4f6632ea1ef8bULL,0x8541047a8008cd03ULL,0xa1942e200820cf9aULL,0x1200f28a0b814b01ULL,0x8efe4a1f1b01675bULL,0x9417d6e82e20cb51ULL,0xfdeff6ebbf29ffdbULL,0xfe1e3f6f3f28e7dbULL,0xef9ef7e92f29c7dbULL,0x2dffaf6f69a8b353ULL,0xcdb9577b3b89dfd1ULL,0x9179bf1a0f08c749ULL,0x468a59ec8f08e582ULL,0x88f6fffa6f08eddbULL,0x6e42b2595b01c55bULL,0x9f44ae4320cf12ULL,0x9cdb70540b01c7d3ULL,0x5a7635d27b28fb9bULL,0xf3d7ffbf3fa8f7dbULL,0xeecf7f5bab21cfd3ULL,0x6efeefebfb20f5dbULL,0x74ffffff7f29efdbULL,0x8efee5713981c79bULL,0xfda7ff6ebf28dfdbULL,0x4fcfeb7b6ba8f75bULL,0x2b5e575b6821d759ULL,0xfef7ffec1f08f7daULL,0xff8f7d639f09ff1bULL,0xffb77f7fcf29ffdbULL,0xcffeffbf7fa8efdbULL,0x8affdfdfcf21cf1bULL,0xd895d5e93b08cfd3ULL,0xdffdfdb24fa8d7dbULL,0xffb73ff77fa9efdbULL,0xeffbfff77fa8ffdbULL,0xbffffec7ffa9dfd9ULL,0x7cba6de64d284d9bULL,0xf462f7fb0ba9f7dbULL,0xfe7e7ff77f29d753ULL,0xab8cf7b61b08cf9bULL,0xefc73ffc6f28efd3ULL,0x93b8efbb5e20cfd3ULL,0xd0f3b6e30aa8c51bULL,0x6eba7f588f20f7dbULL,0xd9ca43650908f599ULL,0xed4a321923084793ULL,0x6f9cd8db2b88d3d1ULL,0x907c564b4928e9d3ULL,0x9eaebff17f20c7daULL,0xc889ceaa0f09cf50ULL,0xcbd7968e3ea9cf53ULL,0x8a8a01420b209312ULL,0x301400a000080301ULL,0xfd8ffffb4d08cf9bULL,0x9a043f6c8a01d758ULL,0xc130dd08830085d8ULL,0x2b64884a198887dbULL,0x15730fc229216902ULL,0x9141366012008513ULL,0x4c062c203100c993ULL,0x400200000800a010ULL,0x8d0849002088311ULL,0x122088240208300ULL,0x406100280a000300ULL,0x8c50284c0400c311ULL,0x804e3660d084102ULL,0x2311184e0b21d392ULL,0x298c80009001250ULL,0x51f586280b08c51aULL,0x2054149002102ULL,0x10020000000300ULL,0x41135008208140ULL,0x200040000000210ULL,0x60c017401400c301ULL,0x82b00d100a214d00ULL,0x402c448203080710ULL,0x9d4ffeb87388d75bULL,0x100041a22000112ULL,0x0,0x1400d2010000708ULL,0x6080240102008018ULL,0x8000000100008001ULL,0x0,0x0,0x10a15221000c51bULL,0x9c0605008a088580ULL,0x8810c06809200883ULL,0x200020200080310ULL,0x4090050003004100ULL,0x48e050001004d08ULL,0x1092f5500c21c281ULL,0xe100a3320820a659ULL,0xc46014009080742ULL,0x294c6f20201d31aULL,0x80405252010885c3ULL,0x18ae64400c20e68bULL,0x69dcd2390e08f7d1ULL,0xfeebff9dfb08ef93ULL,0x6999d7ce3f01cf9bULL,0x1141060c0a0846d0ULL,0xf6d8ef752b08cfdbULL,0xdc8427662d09e79aULL,0x108340208008891ULL,0xa60ecb4983884f93ULL,0x48414024001c112ULL,0xb2c776808005fdbULL,0xfb1eff7dff00efdbULL,0x8cda3b1a3b08e913ULL,0x8054220088583ULL,0x218255fa0b01cb0bULL,0xd5dae772c9080f83ULL,0xf69425f57b89ef5bULL,0xa81050022100010aULL,0x214ad8012c0045dbULL,0x923ec7a3da209d19ULL,0x6d07080f206300ULL,0xed3c0d728a288f8bULL,0x8c0042452004900ULL,0x2006000000008082ULL,0x8480012080400ULL,0x8840024048000189ULL,0x5a8000420000c391ULL,0xb2d4e2624821e7d1ULL,0x42d8b2271a08cd49ULL,0x4214420020008180ULL,0x347d73668f01454aULL,0x86a63c001e006442ULL,0x4000411020008101ULL,0x884040000c004100ULL,0x6009cd4ac200851bULL,0x853406a04208c911ULL,0x90c00620800305aULL,0xc298673608088f83ULL,0xc5bcf9aa1b21dd1bULL,0xabde59995e21c75bULL,0x949606a00880b553ULL,0x4a40f5021a00cb51ULL,0xf6946d602b28c39aULL,0xc1d1c2048820ef59ULL,0x331233424a01e5d0ULL,0xb2024d620a00db10ULL,0x30b800200a802502ULL,0x71284d430a21c759ULL,0x10012000000ULL,0x1710d21a21c810ULL,0xbb0c9d33b015b12ULL,0x20c4144003004093ULL,0x4ce66551c800a79bULL,0xdb92836c5b81c79aULL,0xa000000001004000ULL,0x0,0x440606880808a111ULL,0x38a6103b20050aULL,0x4c08008014080ULL,0x4507ba034909c111ULL,0x2a0009c00c00c180ULL,0x1004000000000100ULL,0x84d042a1700c093ULL,0x100200100209010ULL,0x4000060408200100ULL,0x140014020880540ULL,0x843a04e65400c913ULL,0xb119ca300900c390ULL,0x909209c00a080718ULL,0xe4c6ef3d3928979bULL,0x3f7ffffa2b08efd3ULL,0xa7ba6fff5b29ed9bULL,0xd5fcb67b9f08ffdbULL,0xcad465735ba88f4bULL },
	{ 0xca5a72b23dbe4d40ULL,0x6fcf13fb1cbbed66ULL,0xfffffffbfdffffeeULL,0xfefff6ffddbfffe6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2049035200aa080aULL,0xf7bffff0bcfffde2ULL,0xffbffffffdbffdeeULL,0xbff6f69b34bffde2ULL,0xefffdfdefdfb7feeULL,0x54b67ee934b47de2ULL,0xb33e77e13cfb6ce2ULL,0xfeb53f6bfcbbece6ULL,0xcfc3ffb33dba6846ULL,0xabdffdd3bdba6de2ULL,0xf4f77ffff5befdeaULL,0xfbfde7fafdbe6d6aULL,0xfefe2ffbbdfb6ce6ULL,0xcfdcfb1a709a6c66ULL,0xeadeffd134beed6eULL,0x3ef76ff3bcfa7d6aULL,0xf9ffb7e99dbaed6aULL,0x7f7e3d70f4ba4feaULL,0xc9f3eb2b35ba6de6ULL,0xefeb3bfffdfecfe2ULL,0x9bbfdffbddf87feaULL,0xdff7fff9dcbb7de6ULL,0xdbbeff7335bfede2ULL,0x9bdfff7ffcdaf56eULL,0xefdbbf7ffdbaefe6ULL,0xffbfeffffdbafdeaULL,0xfdf7ff7e7df7fdeeULL,0xfbffff717dfafdeaULL,0xdfbffbfebdffedeeULL,0xffffff9fbdfafde6ULL,0xeffeffdbfdfaffeeULL,0xf4fff7fe7dfe4762ULL,0xffff7ffffcfbefeeULL,0xffbf777dfdbaffe6ULL,0xffffffbffdffffeaULL,0xefffffdffdfffdeeULL,0xfeffffeb7cbefdeeULL,0xff7efffbfdfadfeaULL,0xff9fafff7dbffde2ULL,0xffdfffbebdff5d6eULL,0x2ffeffdffdfbffeeULL,0xbfbffffbfdfefdeeULL,0xfbffbffbfcfbfdeeULL,0xfffefff7f5faffe6ULL,0xebdff7bfbdbf7d6eULL,0xbff73fdffdbefdeaULL,0xffbffffffdffffeeULL,0xfff7fffffdfbfdeeULL,0x5afefdf2fdfbff6aULL,0xdbdeffbff5bbff62ULL,0xeafffed7fdfbfde6ULL,0xcf76e7fb7cbb6c66ULL,0xd5dfffeb6daecb62ULL,0xff9c7d705dbe7f64ULL,0xd7bef7f654faed60ULL,0x8dc2f21b1cba75e2ULL,0x2b5b5ed20c9a4d62ULL,0xd4374a2ed8aacd26ULL,0x940ced75f42a7c62ULL,0xdd84095e51324c60ULL,0x9c423393bc80c520ULL,0x2b7cdfd875badde6ULL,0x6bf0fe270be64e2ULL,0xf77cff52d5bbdeeaULL,0xfab63ff47dfacdeeULL,0xffaef37f7cfeef66ULL,0x3fffff6fd5bbf4eeULL,0x3e24c6bf35ba8546ULL,0xd17f3779fcdbcc62ULL,0xde9e7fee75facf60ULL,0xa3f6db6eb9ba6562ULL,0xafc738c305ba75e2ULL,0xebb439115ba6162ULL,0xf362fa7b34b8f5e2ULL,0x49fe77723cbb7d66ULL,0xfbd4ff3fdcbaeeeaULL,0xeffffffffcbefde6ULL,0x5e95ef2b7dbb5d6aULL,0x7cf7ffeb7dbf65e6ULL,0x973baf72fc964de2ULL,0xfffef37fbdfa7de2ULL,0xedcfff5f5dfafde2ULL,0xbaff9ddbfcba4de2ULL,0xf6fdfefbfdbb7de6ULL,0x9fcfe73139bbcc62ULL,0xffb53f7ed5fb6de2ULL,0xdefeffbbf5bffce2ULL,0xedcfdddf3dfae6e2ULL,0x94f73ffb35be7d62ULL,0xddecfe7e74ba5c62ULL,0xfeae7ffcb8bee66aULL,0x85fbcbabd5b24fe2ULL,0xefdefeff74ba7deaULL,0xffbefffbf5bfede6ULL,0xd756bf6bddbbe760ULL,0xff72be6d1cfbcfe2ULL,0xe3d6df2a9cfa4d6eULL,0x6df7929d3d2b6deeULL,0x9aada7fd74be4f66ULL,0x551060a10c380560ULL,0x5c48694b348b4de2ULL,0xa1144300208608e0ULL,0xa4d396ce50dbcc62ULL,0xc11495804aa5842ULL,0x7deff7fedcbbffe2ULL,0xdffaff55fdbaffeaULL,0xffa6fb367dae5f62ULL,0x40d2d20190024d66ULL,0xae8fc9933d205f40ULL,0x643550733492e042ULL,0xaf6eff73f5ba7f6eULL,0x1b84523414120442ULL,0x1243f30a543a4f62ULL,0x202140000920040ULL,0x4471612040a5046ULL,0xc060a69038ba4c06ULL,0xf294351244b0dc62ULL,0x4580a12234aa0924ULL,0x204490921080880aULL,0x190581014c10440ULL,0xc1180e10049a0440ULL,0xbaae45e03c9d9822ULL,0xa982e13af4906d60ULL,0x22505ed0a0924d44ULL,0x238c4518048a5960ULL,0x400a06014091020ULL,0x52c144011204462ULL,0x8b00b60c14931442ULL,0x80c2101030808042ULL,0xc01004005200100ULL,0x2020030000000ULL,0x110c296015316400ULL,0x80000400080002ULL,0x952ae0004aa00e2ULL,0x200001004001000ULL,0x0,0x8001000000000aULL,0x32010012040ULL,0x1000020000000140ULL,0x488a120b10da0442ULL,0x8010000010000400ULL,0xa104342805284142ULL,0x2a2042000984420ULL,0x4441300014004540ULL,0x2806da800db00160ULL,0x410e452049a4442ULL,0x660ca72205b8f560ULL,0x5b2e7de6c4aa2f62ULL,0xc881836054c80560ULL,0xeee75d523c324d62ULL,0xe194ec8309bece2ULL,0x645292602c98e560ULL,0x220005a0a4a80442ULL,0xb3e6af3614daede4ULL,0xcecb1f95fdb9ec62ULL,0xadbe5f74bdaaff62ULL,0x604124a004a88102ULL,0xaeb26d6254bef8e2ULL,0xd8066e0804b804e2ULL,0x8d4aba3144fa1e60ULL,0x3e49d95a4cba4462ULL,0x24e4516830caa06aULL,0x1068d30b0184560ULL,0xcd8133cd94985422ULL,0xc098580080124540ULL,0x218cd4810284d4aULL,0x76a3742214ba6546ULL,0x142e658109264e2ULL,0xffbe3f77b5bbffe6ULL,0xe9f6fbfefdbb7f60ULL,0x92de1e5774aef062ULL,0x983a4f5a30916c42ULL,0x1008a76214086062ULL,0x200110280002ULL,0xc1aa90183482c440ULL,0x400260000844040ULL,0x7344a00103a040aULL,0x14e120900140ULL,0x92ac0f4194b84d22ULL,0x8972cb6a34924d44ULL,0x43701192208a00eaULL,0x10042100c085002ULL,0x420444030104c40ULL,0x124140000a8e000ULL,0x0,0x40a020020020840ULL,0x129e450940d84420ULL,0xd498cbecd08e4d22ULL,0x8848780214190462ULL,0x88008b08442870e0ULL,0x804509080098cc02ULL,0xaebcd80530ba4f6eULL,0x4220c61000221e2ULL,0x40254050140a1424ULL,0x93ae2c3414ab6866ULL,0x25924207103a2e62ULL,0xf056c65a14929ca0ULL,0x83ba0fe10dfacc42ULL,0x404000000020800ULL,0x728e2d54d0f9c462ULL,0x80000160008a0004ULL,0x0,0x938f89aa748a6462ULL,0x8472f8f03d205d42ULL,0x400c00020082000aULL,0x91040a2411a60444ULL,0x8408000104021800ULL,0x420000084000ULL,0x40440020084040ULL,0x832f35601ca855e2ULL,0x480003a000824042ULL,0xc453b49110ba6966ULL,0x2886814430a90002ULL,0x3004000400a20040ULL,0x8808244211a420eaULL,0x9021c4801aa0800ULL,0x480a22210904042ULL,0x802000000000120ULL,0x20014610020406aULL,0xd771ae08b5325542ULL,0x9aa52c40b4824c40ULL,0x69dfffae3cbefde2ULL,0x2adfffc71cb9d5eaULL,0xb6ba7efe3cfe4d6aULL,0xfdfdfffcfdbb7d6aULL,0xfefe2d709dbc4d6aULL },
	{ 0xc966fb2264d07cecULL,0x6eef3ed2e4125ea4ULL,0xffffffffb6b3ffeeULL,0xf777fef7b6337eeeULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffefff7f6f3ffecULL,0xffbffffff6f3ffeeULL,0xfffffffff6f3ffeeULL,0xebdfdfdea6d3ffeeULL,0xf6f7d7e3f633feeeULL,0xbfec7f4034126fecULL,0x1bb67a7c60f37eecULL,0x70fbeba0053f6e6ULL,0x2b78f7dab472f5eeULL,0x34b6eefaf473f2acULL,0xddf1a7d6e433d6acULL,0xca9e377fa633feecULL,0xcdfef723e2d33ee8ULL,0xedbe5fdd24b377ecULL,0x3ebfcfe39632d6e4ULL,0xe53ce6f7d632e6e6ULL,0xef665deb843076ecULL,0xe9effbebb653feecULL,0x76ff7affe6b377eaULL,0xcfff6febe6f356ecULL,0xf7fdf7fff6f37eecULL,0xfffe7df3e673f7eeULL,0x93d4ea3e647377ecULL,0xafd6dadba0f3f6e6ULL,0x7ffedffbf6f3feeeULL,0xf5fffffff6f3f7ecULL,0xdffefdf2f6f376eeULL,0xff76fffe363376eeULL,0xcdfffebff6f3f7ecULL,0xffffdfdff4f3f7eeULL,0xbcfffee6f6f376eeULL,0xfdfffffff433f6eeULL,0xfbb77f7ff4337eeeULL,0xdff7fffff6b3f7eeULL,0xffffdfdff6f37feeULL,0xffffffefe6f377ecULL,0xd57ffff7b6f3feeeULL,0xffbfff7fe2f3ffeeULL,0xffdff7fff6f3f7eeULL,0xefffdffff6737fecULL,0xbdfbfffaf6b3f6eeULL,0xf5fffffff6f3feecULL,0xfffe3f7fe4f3f7ecULL,0xfbedfb7ef6f3f7ecULL,0xeffffffff6f3ffecULL,0xfffeeffff6f3ffeeULL,0xfffff77ff673ffeeULL,0xdffe7d73b6f3dfecULL,0xfbf7ef7ff4f3f6eeULL,0xffffbfdf76f1d6ecULL,0xffbecff7b6f2ffecULL,0xfcfefefbe633ffeeULL,0x99ceb776a47356ecULL,0xce1ae7ee767277ecULL,0xd574fe042072f160ULL,0x8b5e1e5a941177e4ULL,0xcad694fc06b3f4ecULL,0xd94e6631843356ecULL,0xcb9e752e24f337e8ULL,0x5fd2e73ff4f2f6e8ULL,0x4beabf43249336acULL,0x3c177febe653f2ecULL,0xf79eff7fa6f376ecULL,0xdebe3ff4369377eeULL,0xc7f7fffef6b23eecULL,0x33dcfe99a43377eeULL,0x2cd13000014ecULL,0xa1efbf39a4f336ecULL,0xd7963f7ca0d276ecULL,0xc9c4e72a34d116e8ULL,0x6ed41753a65272e6ULL,0xc2b4dd864306ae8ULL,0x76ebfaff74123eeeULL,0xcebc25e234127aacULL,0xfbd4f76ef453f7eeULL,0xefd3bedff47377ecULL,0x6fbfedd7b613f5e4ULL,0xe5fefbfe34f2f6ecULL,0xbbeefff37673f7ecULL,0xf9b9ff7da4f356eeULL,0xcfdefffb36f377ecULL,0x7b1e1fd7269276eeULL,0xf57154fc467356eeULL,0xdbfe7ff34452f6ecULL,0xfbbe7e7be6f376ecULL,0xeffffe36f672f6eeULL,0xa9fdbcfe4c2f6a4ULL,0xb7be5ffef453feecULL,0xdbf6edfdb4f377e4ULL,0xfca227721692f6ecULL,0xcfca7671c45373ecULL,0x5fde7adfd63376eeULL,0xf7bb7fe36433f2eeULL,0xf9bdfe40e693f7eeULL,0xf7d67df0e672f7e4ULL,0xe9d6f3c9b4f336e8ULL,0xbfee7f5160f3d6e8ULL,0x6d6dfe5c432f6ecULL,0x7424d264a0d1d020ULL,0x52406561041122a6ULL,0x5cc004000202c4ULL,0xa85b94ebd651d468ULL,0x8042412001046e8ULL,0xc973166ac43376e6ULL,0xdffcdf72f472d268ULL,0x40042201000002e0ULL,0xed61e022051126aULL,0xa08080002010060ULL,0xd83ed6e1d632f4ecULL,0x1266a640a05252e4ULL,0x5f387f75b4333eeeULL,0xc4180118244012a0ULL,0xcd8224134334264ULL,0xc53e0d23801336c4ULL,0x93248008a00222a4ULL,0xb104008000110240ULL,0x6080811204007488ULL,0x8e08041005032e0ULL,0x9eab6f5ac01376a2ULL,0xa208060000021640ULL,0x51841810043240a8ULL,0x28c0c10a8413348cULL,0xe4f7941e42175e8ULL,0x3bc4fc1a4007ee4ULL,0x0,0x25a27c00a42050a4ULL,0x202010200026000ULL,0xcbea9c51b43266e4ULL,0x404474300102568ULL,0x811103084001280ULL,0x8601000114020ULL,0x12002d1420c215ecULL,0x810012000c020ULL,0x804000000000060ULL,0x0,0x400010000000240ULL,0xca60000a04121480ULL,0x804100400330000ULL,0xa808104000105008ULL,0x1044108840000168ULL,0x9063322209018eaULL,0xd112025a0022d400ULL,0x4180e016261234e0ULL,0xf21212c2143270acULL,0x30020ac414515020ULL,0xf9d18de2e632f06cULL,0xcf12b52f5633e7ecULL,0xb14238ac20224e0ULL,0xadef169ab4727cacULL,0x8c1043c0020202e8ULL,0x322cfa68201256e6ULL,0x428205314011d020ULL,0x9314d1aa00b372c4ULL,0xac4bb398203332eeULL,0xc39ecf74e6d376eeULL,0x141020800000068ULL,0xe304d62e0823428ULL,0x4f941127001142a8ULL,0x54af281266234e0ULL,0x2620d5c0269016e6ULL,0x94bed621041132e4ULL,0x8d2e54000432d2acULL,0x83bc6734a01156eaULL,0x6044c01434000068ULL,0x8ad7dc82209214aeULL,0x3ebff7eba6d3feecULL,0xb9ece7f1b633feecULL,0xce9a253ea0535ceaULL,0x25d4c18cc41276e0ULL,0x81521a0760401468ULL,0x6aa2d322431c36cULL,0x856eb77a0292f6e4ULL,0x428475301432d028ULL,0x88458228009330a8ULL,0x40c1081020001560ULL,0x809410b04107024ULL,0x40144006200040ULL,0x4d6820600410d028ULL,0x8180e402560004e8ULL,0x101100121200ULL,0x8144e00021054e4ULL,0x51224284422402cULL,0x84950460829252c0ULL,0x82023a0504825644ULL,0x0,0x847e8d4280a27668ULL,0xd41402f14000c2e8ULL,0x85622678043352a4ULL,0x98280f60243242e0ULL,0x308080100020ULL,0x2308129000421be8ULL,0x92431060004012e0ULL,0xe1501e20963118e8ULL,0x46922a6504605ea0ULL,0xcd55f010841274e4ULL,0x304e4ae0a470f8e0ULL,0x131347a286f3d4ecULL,0xd391a16004f054e4ULL,0x77c67c62801376e4ULL,0x0,0x4800c0000021088ULL,0x200c20000102040ULL,0x201042040002a0ULL,0x409a098040331028ULL,0x881410c208216e4ULL,0x1118004011028ULL,0x0,0x2024204330004ULL,0xc1825102009016c2ULL,0x1ac6230524133620ULL,0x8ec3a20bb47252eeULL,0x236c491200123ee8ULL,0x1000040000401000ULL,0xc802204000220640ULL,0x14128c64a01020ULL,0x0,0x0,0x4088456000135200ULL,0x412073512423c0e0ULL,0xf58425778432f6eeULL,0x8f7ee7f3b4f277ecULL,0xa736d54060f3fea0ULL,0xf53acff3f633f8ecULL,0xf5fd7ffee4d2feecULL,0x768e3545042376ecULL },
	{ 0x8b84ebba799c15faULL,0xa48ab207901c148aULL,0x14a9ddeed11e9ac8ULL,0x95041422919c001aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffefff3f9bcbfdfULL,0xfffffffff9bebfffULL,0xdffffefff9bcbfdfULL,0xeb7cefde719eb7daULL,0xd4f76efcd83cbfd8ULL,0x9fefff7a513eb2eaULL,0xdbf37e57511cb7feULL,0x5ec3e332d11c949aULL,0xe7dfd75b599cbfdaULL,0xfcaf7befd91eb7deULL,0xdd6eaffaf93c9fdfULL,0x9eb82df05116befaULL,0x61e0df38591c95d0ULL,0x2bffde595194bfd3ULL,0xf7becfffd91cb7deULL,0x23f33779590cb0caULL,0xda9e7fbc511ebdeaULL,0xe9feffced1be97cfULL,0xfef77efb7994b6faULL,0xffffcffbf19ebfffULL,0xfcfefeff793eb7ffULL,0xfbfefffad9bcbfdaULL,0xdbfcff6b50bebcdaULL,0xe6d232bbd11c9ad2ULL,0xefbfcffff93ebffeULL,0xf5ffffebf99cb7daULL,0xf3feff73593cbfc8ULL,0xfbbfff7ff13cbfefULL,0xeddfffbfd91c1fffULL,0xfbffdfdff1bcbef6ULL,0xf6fefefaf1bebfdbULL,0xffffff73f13ebffeULL,0xfff7fffff1beb7dbULL,0xeffffffff9bebfdfULL,0xafffffdfd1bebfdfULL,0xfdffcfe3d99eb7dfULL,0xfbeffffff99cbfdbULL,0xfebebffff9bebfdeULL,0xffd7f7bfd9bebffbULL,0xff7ffffbf9beb7fbULL,0xf7fffffff9bebfffULL,0xffffffffd9beb7ffULL,0xff9eff7b71beb7faULL,0xfbf7fbfef1b6b7feULL,0xefff7fffd9bebfdfULL,0xcfbffffbf9bebedbULL,0xd5fffffbd1bcbffbULL,0x7d7e7f7bf93c9fdaULL,0xfbd7ff3ff93c9fdfULL,0xca4f7f53d99cbe7bULL,0x8fbfcf50513e9edbULL,0xd7f4bf7a599c9edeULL,0xdeb66d65d13695eeULL,0xae10ef65519eb5e0ULL,0xefc7fe13f10e16c3ULL,0xa29eda1e51b416baULL,0x7cb574a6c01cb5d9ULL,0xfd423fe6593496e3ULL,0xbb163f69519cb6c9ULL,0xefcedeb9d994bfddULL,0x4b5f5f9159be3edaULL,0xaafe57e3d11cbecaULL,0xdf7cbff2f9bc1ecaULL,0xfabe7d7271a4b7cbULL,0xebd7ff7ef9bcbffaULL,0xfbdeff5f59bc97c6ULL,0x36bf4de05994b7deULL,0xf5ed3e79519cbbdaULL,0xff92e576c9b4b5deULL,0xcbeacf7f788cb7daULL,0x7fde1fd2f83c969eULL,0x9eb26c5250143499ULL,0xd5fc9fffd194bbdaULL,0xcbaebd41591cbdf2ULL,0xebb6ffbed1beb7caULL,0xdfdbfeb351bebfdaULL,0xefbfcfebd91abedaULL,0x79ec77ef511c3ecbULL,0xf4b8af7061b43c6bULL,0xdea459fe791c968bULL,0x9fd7f21ad91cbde4ULL,0x8cfecd4ed1bc34c0ULL,0xf4aed77dd13c96caULL,0x1def3fe3d19cbfecULL,0xfebc277ed9bc97fbULL,0xf957ef1ff91ebedeULL,0xaf7fdfded9bebedeULL,0xd7f65ff7d1beb7d3ULL,0xdbe58f7af19cb7caULL,0x8aa73f6e51343edaULL,0xdfebff6851bcbfdaULL,0x6a7f53db7194aadaULL,0xc39aeff6511417caULL,0xf9efa77f71be97daULL,0x4fbaedf95136bfcaULL,0xe9e7eb2ed81cb6caULL,0xeff75f4bd11cb7efULL,0x16310c60512416ceULL,0xfdf57ebbd93cbdfaULL,0x559e0d4298349452ULL,0x9200022440941800ULL,0x8910000100a00ULL,0x213e4c4658143683ULL,0xf13712da10a487c4ULL,0xbbbebd70f934bbcaULL,0xed18ffd25194a41aULL,0x45c23a3108163cdaULL,0x841ace81f02494c0ULL,0x4c11661011217e6ULL,0x856ffe42501cb8daULL,0xc812401c00140688ULL,0xc1f5920791a415ceULL,0x8004400290001008ULL,0x100010090000000ULL,0x69a5f068591c1098ULL,0x42102a0010140484ULL,0x90af128500c954aULL,0x841f4141841248ULL,0x108a2dc5d1009450ULL,0x8149500000260142ULL,0x44020000000c0484ULL,0x103a02051040088ULL,0x8fbfe31d13630f2ULL,0x1621044040342842ULL,0x54e1f4f4411eb2c2ULL,0x64282422111690d8ULL,0x20020050004040cULL,0x1210024000a040ULL,0x1a2000011000400ULL,0x8942402e0041480ULL,0x804000041000000ULL,0x0,0x8a80000140140802ULL,0x204000040040000ULL,0x1401000400040140ULL,0x0,0x0,0x4c22e20400401c8ULL,0xe35012c3d1243442ULL,0x9c11252350b60950ULL,0x9ac8f770519683c8ULL,0x40823dc059049042ULL,0x4dd6d7325116b4e2ULL,0xab5eb647511ca85aULL,0x96a3cf6f5094bed8ULL,0xc577fe7ad1b4b4daULL,0xc6dc6c6551362cdcULL,0x94cd72c4924b6caULL,0xeff7be31501cb7dcULL,0xbdbbdff3518691e8ULL,0xd222c85058101c9aULL,0xcf8e5c72711cb7c2ULL,0x4b96ef36511c36fbULL,0xc6ce164e7184954bULL,0x60bccec2d1b696daULL,0x2034000008441ULL,0x4a82044101100342ULL,0x8f12764ed196324aULL,0x2184b00200041488ULL,0x65c7580850148cc1ULL,0x78713429503ea2c0ULL,0x91bcaf62593490faULL,0x9bb62f7751b4b3d2ULL,0xc7cfc502c184348eULL,0xbc4c958f104348aULL,0x9dbb77ef5194b4faULL,0xd9eff6fbd136b6ebULL,0xff9dbfee7194b7deULL,0xc1f7d30230141edaULL,0x39c71bdab1249442ULL,0x2438f071d118b098ULL,0x846d006810149a4aULL,0xc2825451108c83c8ULL,0x8000c04800940a40ULL,0xa641bc21094a740ULL,0x22206e00100148aULL,0x53242010503030c0ULL,0x7ff67778513cb6cfULL,0x8910842018101410ULL,0x41308040840082ULL,0x886800280001008ULL,0x1000204440000150ULL,0x1008400100c1080ULL,0xc0022041400c3008ULL,0x8eeef8095134b3daULL,0x1c28c52411c1098ULL,0x14314468501698c2ULL,0x812cbba251849062ULL,0xaa98291111148d4bULL,0xae57000c110003c2ULL,0xabf63a5151a494e8ULL,0xb4975ceb613cb572ULL,0x500021a001101202ULL,0xca320a5059149602ULL,0x48d21508701421d2ULL,0x2a171c8151103689ULL,0x98b34528590c91c0ULL,0x31ec140851008742ULL,0x4002041000000ULL,0xc080480051000242ULL,0x4482000040042000ULL,0x640420050101090ULL,0x0,0x84000000000400ULL,0x90002164114143bULL,0x80814019001400ULL,0x1014010001000ULL,0x20610ee0401c85c4ULL,0xd7ee4d60511495caULL,0x42830406412e1040ULL,0x4cc3fa19701c9bdaULL,0x31ecf13d03c32c8ULL,0xf6bef6eed11ebfcaULL,0x2036056048041340ULL,0x80000008040ULL,0x6004008001c0800ULL,0x10000400002040ULL,0x0,0x80a4262870040480ULL,0xdf8e2c75519cbfcaULL,0xceced36e793e9f6aULL,0xaddfbfdf793cbedaULL,0xb7fa5ffb593cbfdaULL,0xdbbc25d87134bccaULL,0x7a523fd67014badaULL },
	{ 0xc8c6e3efd0e8b7a4ULL,0xa0d2b95490e0a772ULL,0x8b8ccffb90a0ef30ULL,0x5114007090432210ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1a868e4090202c00ULL,0xebb4fb3eb879277aULL,0xebffffffd8f3fdf0ULL,0xefffdfdaf8f9bffcULL,0x34b2166790c32644ULL,0xfb6eaf70d8ebbf76ULL,0xf3a65dff90c2aff4ULL,0xdcc2bf6ff069af3eULL,0x6f7e9edbd8692e7aULL,0x33aedffbd8f3be52ULL,0xff69a4f6b0c3bff0ULL,0xfa9727fff0f3bfd0ULL,0xd2d7e34a90e9ef58ULL,0x2b4f36da90f1377aULL,0xe3a4fc798f993c0ULL,0x99ede6e3d071b524ULL,0xcbb2ac608060af8aULL,0xf9d7fe2f98f8abf2ULL,0xeeff3f79d86bbf7eULL,0xffbffffff8fbfff6ULL,0xfff7feeff8fbbffeULL,0xefbefff2b8c9bf76ULL,0xabd4e33ed8e0275eULL,0xdfcbbed3d07bff3aULL,0xaffbdffbd86bbf54ULL,0xfdfff7fe90a8bffeULL,0xeffcfdf5f8e9ef7cULL,0xdbb6fb6ef8d9ef72ULL,0xcf5ffebbf8faaf76ULL,0xafffdfdff8fbff72ULL,0xfcf7ffefd8fbbf7cULL,0xfffffffbd8fbbffeULL,0xffbf7ffff8f9fff6ULL,0xfdfffffff8f9ef76ULL,0xffffffdff8fbfff6ULL,0xbe7f7fe7f869bffaULL,0xfffffffbd8fbbff6ULL,0xdeae3ffff8fbbefaULL,0xcfdfff7fd8fbfffeULL,0xebfffedff8fbfffeULL,0xfefffffff8f9bffeULL,0xf7fff7fbf8f3ff7aULL,0xffb67dffb8e9efeeULL,0xebfffffef0fbef7eULL,0x6bffbedfd8fabf76ULL,0xffbeeffff8e9ff76ULL,0xffffde7ed8eae776ULL,0xfcb62f71f86bbf76ULL,0xd3f5ff3eb8f8fffeULL,0xb7dffedf98f9efeeULL,0x9bbecf5c98eaff8cULL,0x7efef77598c9efd2ULL,0x818cf5709050ab52ULL,0x9cb0a76bd861b364ULL,0xcfeffebad8f3bff2ULL,0x9dfdd5bd0837f70ULL,0x5416fae0d0e1af70ULL,0xd8e7ffdb90f9bf66ULL,0x9bbe7f6e98e8ff16ULL,0xc9d7f23ad8e3c760ULL,0x63fe5b5798e9aef4ULL,0x14fbffc7b8eaf736ULL,0xf9f5ff7a98f9ef4cULL,0xffb73f6d90fbbf7eULL,0xcfdfffec98f1ef6cULL,0xefd7bed9c8e1afe6ULL,0xfaaf77a9041a7f6ULL,0xf77eef77b0d96762ULL,0xf7b63dfff8e1f774ULL,0xa9d0f71e98f07f70ULL,0x19f51a4b90a1cff0ULL,0x15be4fda90f12942ULL,0x96edf67290e9b7f2ULL,0x7f7ebf7bd8f1ff64ULL,0xdbd5ff7ef8f37f3eULL,0xfdfbbfdfb8d8eff6ULL,0xafbe5f1290e86f76ULL,0xf5a6ade2d0f1f77aULL,0xbfbeed73b8ebfffaULL,0xdefcf7fa98fbe7e4ULL,0x8cc6271498682778ULL,0x2b9ec95190e2af00ULL,0x30bac7eab8e0ff60ULL,0x1fe7bd69d8e8aff2ULL,0xfe962f7798f8bf76ULL,0xe847ef5eb869a766ULL,0xebf6fe9f98fbef76ULL,0xfcff5d6798e1ff54ULL,0xfbfc7f7a98e3bfd2ULL,0xf2bf3e76b8f1bf64ULL,0xbdffb2eb8c1afd0ULL,0x3f7dfed19069b77cULL,0xbe9a5ffed8f1af56ULL,0x595df79af861ab5cULL,0x4e8a3d22c8e8af08ULL,0xc9aaf7bfb0ebaf76ULL,0x5afe3e7198c92f72ULL,0x98b9ddf398e02f46ULL,0xa723e261b8403e14ULL,0x10c152090b0a120ULL,0x8944e00218410360ULL,0x8a731280c0612e00ULL,0xaffe9fd2d8e9f560ULL,0x512e066e98810560ULL,0x87931d609089b7c0ULL,0xdbb6aeb698f1b750ULL,0x40c01080010040ULL,0xae6ed9cbd8e92b12ULL,0x34673ec890418f30ULL,0x9b6de16298e06318ULL,0xc800213090410a08ULL,0x8082a22080612600ULL,0x2182315998608c10ULL,0x8010000080000800ULL,0x406ca02090c02214ULL,0x10082b05d0c02660ULL,0x2084020408410040ULL,0x80e085080800160ULL,0x3a41009009a344ULL,0x340800209200ULL,0x2010000080002000ULL,0x10051a080002444ULL,0xa015a508050a286ULL,0x988050288400e10ULL,0x514091c030010270ULL,0x20344008480208ULL,0x5310c00088802004ULL,0x8103be02b040b344ULL,0x0,0x10002000000210ULL,0x100484200480600ULL,0x0,0x0,0x10000280000800ULL,0x1890a6c48820f500ULL,0x20280100002400ULL,0x8000004000000140ULL,0x604000400010248ULL,0x6c5057c0c8612650ULL,0x2474555230e8ae90ULL,0x7b6ceffaf8eab706ULL,0xd6962d739829ba70ULL,0xc0e2db0a98e0a492ULL,0x2b9f9e4098ea2f30ULL,0x3281434a9861b412ULL,0xf1cc96d298219f58ULL,0xcca645e14028ad00ULL,0xa9e7e376b0fabf30ULL,0xe9c2137fd0637766ULL,0x2dbedfe1f0e1eff4ULL,0x604cd66180a02542ULL,0x528a3d400861a750ULL,0xd024036690c0ac80ULL,0x374af21b8879af60ULL,0x2e4a52d0ca2340ULL,0x1040601020b900ULL,0xc58f05288e0c520ULL,0xee94676ed0c1ab60ULL,0xaf46be290861f760ULL,0x820a1c9698e2af02ULL,0x96bf5decd8e9bf60ULL,0x314f75f2d8d2af18ULL,0xcf82676b90e0af12ULL,0x69ced872c040af92ULL,0x6b5616d3b8e0bff4ULL,0x3cd79ffbd0fbaf50ULL,0xda61d7c0d8f08d00ULL,0xf7bc2f7cd843af6aULL,0x8065a210609a10ULL,0x2f89b4398e02644ULL,0x2308c519b0a03602ULL,0x8000246380102460ULL,0x18c409428020b122ULL,0x2350e0028089a500ULL,0x612c08020a500ULL,0x2c5805208001b400ULL,0x9d1c046248602612ULL,0x80866568d0412756ULL,0xd9c5af2c9060bfe0ULL,0x2002208a00202402ULL,0x6008481010082022ULL,0xf1540cc2c030b522ULL,0x100080000400200ULL,0xd3344a0c88492562ULL,0xc8c39e9090733700ULL,0x4010050000002520ULL,0x94241568806087a0ULL,0xf8ef782986aa712ULL,0x12b82f05d0102250ULL,0x6d86831e9060275aULL,0x2f5eb7dd90e93f70ULL,0xbdd457698861af40ULL,0xf5edbed598eabfd0ULL,0xa84236490c12444ULL,0x4d8c110a860a6e4ULL,0xab4e1a569840edc0ULL,0xb8b24fbb98e1ae54ULL,0x1200d10080202900ULL,0x200010000400100ULL,0x100600000400500ULL,0xc48291080212250ULL,0x0,0x5424261090202406ULL,0x82252000400600ULL,0x18202700c0400500ULL,0x2048800900010270ULL,0x0,0x1910000090003000ULL,0x835aad30b0c08a90ULL,0xc810010200216140ULL,0x2dc6c91498298760ULL,0xa10a089288e9ae40ULL,0x845074a010688e10ULL,0x802404000000b002ULL,0x2410000080000000ULL,0xacc0400990009570ULL,0x24000010800000ULL,0x3002010000880300ULL,0x1908426a9020bc00ULL,0xca002585d0212c56ULL,0x8947a228b080b70cULL,0x277fdeff98e9fb76ULL,0x36b2ef5290e9bafeULL,0xfdffb77a98f9bf52ULL,0xefbc7ff38861af50ULL },
	{ 0x2ac7eb20493a2e61ULL,0xeee63f574b7eff72ULL,0x8f1acd53ba646524ULL,0xd1c2957185ac6626ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000000401ULL,0x200040000880000ULL,0xfbbdfffffffe7f67ULL,0xfffffffffffeff77ULL,0xbddedfdb53fe7d63ULL,0xfefff66f9fcc7e67ULL,0xff7fef72df6ef547ULL,0xdf526e3d9b383f61ULL,0xcdffbbbe9bfaed77ULL,0xabdcdfd3fbf87d63ULL,0x863a546b25ac4760ULL,0xdb7dff72d97c7f63ULL,0x763eaf7e4dfc6d62ULL,0xefc3da3adffc6742ULL,0xef7efbda4fcafd63ULL,0xb5b14d2a972c6e62ULL,0xf1ffb7f5876eef63ULL,0xf79a3d71eb2ef663ULL,0xabdfcf3a1f6e4f63ULL,0xefff3fcedffef563ULL,0xfffffffffffe7f73ULL,0xfff7b7fbdffcff73ULL,0xfffeff72dbfcff66ULL,0xcbf7e32fcff8ff73ULL,0xecd79fbedfeef767ULL,0xffff4f72fbecef62ULL,0xffff7ffbeffcff63ULL,0x9b6f7c753f4ce526ULL,0xcbfe7bde5feeee21ULL,0xcddfff332fee7f64ULL,0x6ed6cf5e3bfefe62ULL,0xd6fff7fbdfacff73ULL,0xfdffff7bfffefd73ULL,0xfbfffffffffc7f77ULL,0xffdfffffdffeff77ULL,0xff7fffdfffeeff72ULL,0xbfbffffffbfeff73ULL,0xfbfffff1fbfeff77ULL,0xfabf3fffdfeeff67ULL,0xeffff7fdfffe6f63ULL,0xffffffdbdf7eef67ULL,0xf7be7ffbdffc7f73ULL,0xfdfdfffafbee7f37ULL,0xfff67df5cbfef777ULL,0xfbffffffdffeff67ULL,0xffffdedfff7eef77ULL,0xdfffdffbdfee7f72ULL,0xfcfffefffffeff67ULL,0x6fbe7f7a5fbeed62ULL,0xbbf2eb6f8beeff62ULL,0xea5bfaffcf2e7f76ULL,0x909ecf43cf7c7f71ULL,0xf5f186c81afcf512ULL,0xa51ebd620bac7e22ULL,0xb02476648d7c6c30ULL,0xdddfff2e1bf86f67ULL,0xbb0ecbd79b4e6d70ULL,0xfcf27eecebfc7d63ULL,0xf9cd7f427f8c7f73ULL,0xdaee7feadfaaff72ULL,0x4953b71a03fe7f23ULL,0x2b7f5fc96f4c6773ULL,0x34bedfe75bfe6d61ULL,0xb9ffff76ffeefd63ULL,0xfefeaff6ffeeff71ULL,0xe7f7ff318f2ef771ULL,0xbffffedffffe7f67ULL,0x97bbeffadb6a7766ULL,0xd7cd34fc0beef467ULL,0xefbe6ffe9ffe6d63ULL,0xebf4d6ab4d7c6e35ULL,0x3957fff2cb78ff33ULL,0x9f3fccf36f48f563ULL,0xf5fe96f94f7e6f33ULL,0xeabe6d78ca747775ULL,0xfb9cfbb6c9fe2d62ULL,0xea5bdfe71f7ef773ULL,0xefbedffbebeee773ULL,0x557f5ffdc9bc7561ULL,0xffbdfff2dffe7f67ULL,0xffae3d5f5fbc7f63ULL,0xc7fa96501f064661ULL,0xbffedfdbef7ef563ULL,0x96ff7dea6a687d61ULL,0xa97ee7731aec4f67ULL,0x9b84a675c8ac6560ULL,0xd7d2fe92cac86763ULL,0x2f7fd9de1f2cf762ULL,0xbcf4dfffcbeeef67ULL,0xebedff7bdfee7d73ULL,0xfabe7fde7fecef63ULL,0xeddfff6e4f6efd77ULL,0x2edf57f3077e3f32ULL,0xbfb267ff7d7c7763ULL,0xd95d37b3917e7f62ULL,0xfb966d70bb7ef767ULL,0xa97eaf63c74c6d72ULL,0xffdb3b5ffbfe7d67ULL,0xffbfdffbff7e7d67ULL,0xbe76867c59723e22ULL,0xebfe6ffb8bbef524ULL,0x7350e37e87fa7940ULL,0xedd29a02836a7160ULL,0xcfbfff7bcffe6767ULL,0xf5f7e37a03bc7d62ULL,0xd1b2bd324dee6d33ULL,0xeb1e72cb12340f66ULL,0xe1e7be201ac87d24ULL,0x82fc97d39a7c4761ULL,0xd080b2ec49342260ULL,0xd9eeff73abee7f73ULL,0xc39a29034f406541ULL,0x8a03ef498d44b620ULL,0xa9f61e5449f86f67ULL,0x54508040006472ULL,0x8008014007000400ULL,0x9210221001205461ULL,0x4480112002c4020ULL,0x98462a4002340500ULL,0x820000800300600ULL,0x41212102404401ULL,0x40880d4102240550ULL,0x0,0xb8a61e0071087422ULL,0x3190d0a2490a6822ULL,0x9122887083c85402ULL,0x2250000200444ULL,0x822040042020ULL,0x510020084020ULL,0x10034002080504ULL,0xf1ffb77edfeee767ULL,0x0,0x0,0x0,0x8380f02812c5540ULL,0x34f21ec40b2ef542ULL,0x802f084090006540ULL,0x20242000880820ULL,0x100c00280002060ULL,0x10000400000400ULL,0x4023d2608fc86102ULL,0xd100b48a830a6f27ULL,0x840000008400000ULL,0x80a0000209005401ULL,0xf7d19468fece662ULL,0x388051848006462ULL,0xc4cd241000642600ULL,0xded67ddddb7eef21ULL,0xfbffd76fdafc6f72ULL,0x65e63a190bba6662ULL,0x2eb4c7fec8786521ULL,0x302ea2e4a90a7d62ULL,0x8826056001467860ULL,0x1081010d07300470ULL,0xa690f3ae18747410ULL,0x61f4e4242386426ULL,0x71b7675a0fecf761ULL,0x97bbbd31378a6573ULL,0xd62677ec4b666f61ULL,0x4dd3de18976e6661ULL,0x338ecd438fecfd67ULL,0xf6b777eb6b887f66ULL,0x80eb324005f87240ULL,0xb3b42d6d5b387620ULL,0x8eebaa38193e4760ULL,0x67d21b9b4bec6e23ULL,0x147d51328f74ec62ULL,0xd93caf7acface473ULL,0xcebc2fff9b6e7d60ULL,0x40a002000303000ULL,0x2842809209225402ULL,0x12a80dc80a646161ULL,0x0,0x7000b00001c85000ULL,0x8000824000000660ULL,0x6cc609c0c92c7020ULL,0x0,0x92001408808c2020ULL,0x228e24700aec6502ULL,0xc10c613382a82500ULL,0x8d403e018b4e5420ULL,0x6000080140003000ULL,0x10c6538cd2bc6523ULL,0x9f325d6028747d60ULL,0xf3a07a17d32a6762ULL,0x8e47a6838b2e4c21ULL,0xa378d88689ac7420ULL,0x74c4042841ae6461ULL,0x1100a70281ba2700ULL,0x91811d10892e2502ULL,0x85e6df928f407d62ULL,0xef7e7217da48ed62ULL,0x638cff3807c6d21ULL,0x915c7f60eb6e7560ULL,0x881a093807025e20ULL,0x8102821081021400ULL,0xaeff9fd7872eff63ULL,0x8ab04de823bc5563ULL,0x9090560850887c25ULL,0x6600002408044421ULL,0x48040308050c5161ULL,0x60140801407040ULL,0x808a0520a0086820ULL,0x1090402808480102ULL,0x0,0x100040008004000ULL,0x0,0x2e06040800085d20ULL,0x4100000000000400ULL,0x8c0cac32214a6060ULL,0xb2040508082c2062ULL,0x587380248404461ULL,0x2000010020000060ULL,0x402130080886002ULL,0x4080000000202000ULL,0x0,0x8080600000000406ULL,0x804001020000ULL,0x3000000000000404ULL,0xa124ac004524b000ULL,0xca1224624024fc23ULL,0x4d4ff7a8896cfc60ULL,0x3ffffef2cb6c7d66ULL,0xffbfcffb6f6eff73ULL,0xb3fd7f7ffffeef63ULL,0xd8be7df9afec6d73ULL },
	{ 0xd983f3634f32ed60ULL,0x3fd7fe5dc6963d60ULL,0x119cccb353922d00ULL,0xbdb7d6686ff62d08ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x597fd26ac7be6f60ULL,0x8008a36205b61040ULL,0xfb90fffe9ebe5f60ULL,0xcfffffbfffffff68ULL,0xef7fdfdeffbffd68ULL,0xdcb9fef0fafe7d40ULL,0xd5efad611d9e6d60ULL,0x9bfe677f4fbe7f60ULL,0xeddffa2aeebf7d40ULL,0x2b5efddbdf9b3f68ULL,0xb47a7de35fb63d28ULL,0xf1ccdf62beb77d40ULL,0xe69b27f116ff3d60ULL,0xcdeaf31a9fda6f68ULL,0xbfebfd6f2fe3f68ULL,0x17bfcf78febe7d60ULL,0xd178f6db1fb62f60ULL,0xdba67dc537da2f40ULL,0xebb6ffff9ffe7f60ULL,0xeef77f5f5fb63f60ULL,0x7ffffffbdfff7f68ULL,0xf7fdf7fbfffebf60ULL,0xfab63ff35fbe6d48ULL,0xebffdf7fffbeff60ULL,0xebf7bffbdfffff68ULL,0xe7bf4fffdffeff60ULL,0xf1e6d7fe26befd60ULL,0xf7ba7ff364b23f60ULL,0xfb99fa2b0fbeff40ULL,0xffdfbfbfffbfff40ULL,0x3b7fdfdfbfff3f68ULL,0xfcffffffdfffff60ULL,0xfffefff3ffffff68ULL,0xffbeffffffff7f60ULL,0xffffffbfffdfff68ULL,0xffffffdfffbf7f68ULL,0xf5ff7fffffffff68ULL,0xf9fffffeffbe3f60ULL,0xfebe7fffdfbfff68ULL,0xeffff7feffbeff68ULL,0x3f7fdfdb7fbf7f60ULL,0xffffffffffffff68ULL,0x7dfeb7fbfff6ff60ULL,0xfbfeedfdfefe7f60ULL,0xebf7ff3fdffe7768ULL,0xffff7fdffffe7f60ULL,0xffbfffffffbe7d60ULL,0xff7fff7bffd7bf68ULL,0x1ebc7ffadfb73d60ULL,0xfbd5ef3d9fbf7f60ULL,0xefdbbfcfdffa3668ULL,0x9fbfff7bffb67f68ULL,0x35e606ec37a23d60ULL,0xefecfde3f7bebf60ULL,0xfbfee7be17baff60ULL,0xcfeefc87a7be5540ULL,0xad7fdfcb6bbe7d60ULL,0xd6d5bf6e5f9e7560ULL,0xbeeebf78959e6f68ULL,0x9f4c3f6fc5fa6f60ULL,0xede7ff1adab6bd00ULL,0xeabfdfd63fbebd60ULL,0xfdff7fff7e9e7f68ULL,0xf5ffa77c76fded60ULL,0x5abf2f37f69f6f40ULL,0xeff6ffbf7fbffd68ULL,0xff5f7fdbbef6fd60ULL,0xbebe6f7affbe7d68ULL,0x4e7587294eb27540ULL,0xdbf67f77fffbbf68ULL,0xc9baf6afdfba3d40ULL,0xe8fb9ef3dbbabd60ULL,0x8bbecfaa6eff3f40ULL,0xe677dffbdab62d68ULL,0xdbae3f7afffead40ULL,0xbbfcf76fd7be3d60ULL,0xedebffd99effff60ULL,0xefb28f5f7aa6af60ULL,0xfdfff6f9b7be7f68ULL,0xfffefff3ffffff60ULL,0xffbeff7fdfb7ff60ULL,0xcfdffedffebf7f60ULL,0x2d4e77d8c7965760ULL,0x743d5fec56fe5f68ULL,0x5bef7d615f956f60ULL,0xfb947ff6de922d28ULL,0x97c6ef399abb6f68ULL,0x6b7fdfd7feba6f60ULL,0x94bf766a1e365d60ULL,0xf9c4b75a9696ef60ULL,0xfef9af7fffbafd68ULL,0xedf7ffffdf9e7f60ULL,0xad5ebefe9e9e6d40ULL,0xb6b8cefa53b75d60ULL,0xf0fdff6c7ff76f40ULL,0xcbba6fc5deb67460ULL,0xd9ffefae7f9e6d68ULL,0xedff3fffdff77f60ULL,0xe9dffefcfb77f40ULL,0xd4b2d2fc5e9a7940ULL,0xce06e57a571a6d68ULL,0xfa9063e6169a2d60ULL,0xcc23ebad7d22c60ULL,0x3f3ec76bf6def560ULL,0x75d78ef2b7b2dd40ULL,0xca167c60cb963620ULL,0x5d85a4fbcf1f6940ULL,0x191bdb15479a7760ULL,0xd2815392123d08ULL,0xf4313fc0d6122d60ULL,0x7dcded7266b32d40ULL,0x7983a1150972c00ULL,0x4842000001300040ULL,0x80393642133a9d40ULL,0xa001002002128000ULL,0xcb6c6e46559e7460ULL,0x4272010122d00ULL,0x8022c80004000140ULL,0x20e0804814921d40ULL,0x220050004300400ULL,0x1049827004920e40ULL,0x8924396054842b40ULL,0x40000102020a0020ULL,0x2200000800200ULL,0x488000000200540ULL,0x20004004000040ULL,0x8480d00020a0440ULL,0x190203060aa06040ULL,0x8490300182848c20ULL,0x2110020000920200ULL,0x2006020010802800ULL,0x0,0x802002002000ULL,0x0,0x8adecdc903f62740ULL,0x54d54df016ba1740ULL,0x805b076006b21720ULL,0x0,0x0,0xa000014000c24000ULL,0x6004100000ULL,0xf0706e2026968548ULL,0x4804258000025420ULL,0x3c4a23486a23f60ULL,0x88c2314104123d00ULL,0x202058300860e40ULL,0x5322010006123900ULL,0x4210000e02080440ULL,0xebf4f2b694b60d60ULL,0xeffe1f5feebfff60ULL,0xbefd5dfe9abeef40ULL,0xec15f66d1ed3ed48ULL,0xcfec2e6ad7983d68ULL,0x78c4120ac09e2f00ULL,0x212322e9b922d60ULL,0x3084c1000221420ULL,0x3dff026e5e16ad60ULL,0xf2be355056b62f40ULL,0x4400e34e34a61440ULL,0x695fb2b117df7e60ULL,0x3aaecddacee6ed68ULL,0xf6fe3deddfbfbd60ULL,0xdfb3fb52ce323d60ULL,0x2808240002228840ULL,0xc156fb544f8e6d40ULL,0x8b7cce4e0fba3d40ULL,0x3ed4f5eb9ff36d60ULL,0xe9548e181cd43f60ULL,0xdabe297217bc0560ULL,0xedcedf080eb6ad48ULL,0xec1bbe696fb23540ULL,0x802491202926000ULL,0x20040002000400ULL,0x7882054010b23440ULL,0x4c20406300420ULL,0x52015054800440ULL,0x2000004000000440ULL,0xd350fe084b9a2e40ULL,0x0,0x8b10c1281a8a0460ULL,0x300001000400ULL,0xa0440014900000ULL,0xf1d73e7b43967d68ULL,0xda28ad418bbe7640ULL,0xc080b174121e2560ULL,0xfceceab07962f60ULL,0x2b8e8d9e0aa62d40ULL,0x24cce64096962d40ULL,0x460a551028a2c08ULL,0xc491121006373c40ULL,0xc4c6f0bc4fb26d60ULL,0xa3e9959a1b963d20ULL,0x9475f04197840c40ULL,0xfbcd77627f9a7d40ULL,0x1a8daf20c2882760ULL,0xdbd2f7b253ba7d48ULL,0xefffffdf1fbf7f68ULL,0x1d885d5b97b2ad60ULL,0x5b3e3c4a96ba1f68ULL,0xa110202040001420ULL,0x400800000440ULL,0x80000a1010a00040ULL,0x0,0x420002004100040ULL,0x200010000104000ULL,0x0,0x2004000080400ULL,0x0,0x8084024842102500ULL,0x81002c0102922d20ULL,0x92143b8557b26d40ULL,0x100900100020c20ULL,0x90ca0108120540ULL,0x2024004002001400ULL,0x600500010a0500ULL,0x10020000000440ULL,0x0,0x100001000004040ULL,0x840110001120400ULL,0xa30803400a802440ULL,0x9b1a001006822560ULL,0xcecb5f2c5b902d48ULL,0xffdfffdfdfbfff68ULL,0xf7bb7ffeffffbf68ULL,0x50cc376fc6963d00ULL,0xdffeff6b5f9e7f60ULL },
	{ 0xcba1f7fbf7bc6d7dULL,0x7fff3a85d2a460e5ULL,0x65bccd9ef6ac21d6ULL,0x9cb6e651d2842e4fULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffffeeef7be3bf4ULL,0xdfcba572d6002b77ULL,0x4000400814040040ULL,0xdf7ffefff7be7fffULL,0xff5fdfded73e7fffULL,0xf6fa7fefd79e6bfdULL,0xb377f77cd70e277dULL,0xcfb77ffdc7243fffULL,0xffffffffd7be6fffULL,0xabfeffdbf6242ffcULL,0x1cdf5fe3d634211dULL,0xd5f5ffe9f61e6b95ULL,0xdeae3f76f7842bfdULL,0xcffaf7bdf73c73ffULL,0xefdededbd61e6fffULL,0x96bb6de5d636095eULL,0xf66f36c1d3b26af4ULL,0xda942c74d60c49c4ULL,0xc9d2f37df6b62bfeULL,0x2fdf5fd7f7846fffULL,0xbebf7ff7f6967fdfULL,0xf77fff7fd2b67dffULL,0x7ffe7f73973e3fffULL,0xfbf5fffed7bc3fffULL,0xffffffbfd7be6ffdULL,0x6fbecdf7d61e6336ULL,0xd5f47fecf6a46ffcULL,0x95b96d53f6ac22dfULL,0x9a95d36fc6ae23ddULL,0xdffffffff6be6f7fULL,0xbf5ecfdbf72e6bdfULL,0xf4fffffbd7b67fffULL,0xdfffffebf7b67fffULL,0xffff7fffd7be7fffULL,0xdffffffef7be7fffULL,0xafffffdbd7be6fffULL,0xffffffffd6be6fffULL,0xfffffffff7be7fffULL,0xfefe3ffff6be7fffULL,0xffd6ffefd7be7fffULL,0x7bffffd6f6bc7fffULL,0xfffeffffd6be6fffULL,0xd1effffbf73e2d55ULL,0xffdefdfed68e6fefULL,0xcbffdb3ef6ac6feeULL,0x7fff9fdff6be7bddULL,0xbfbfdffbd7ae6fefULL,0xf5fff7fef6ac7ff7ULL,0xdbfcedf3f6342b75ULL,0xf9dcf77ff6a66adfULL,0xfcfbfef9f63e61feULL,0x2fffef7fd53e77fdULL,0x70949eead28e29ceULL,0xe7feff70d7b67b7dULL,0xffbf6ffff6be7ffdULL,0xe9dede93d72e6fddULL,0xaffe5f5ed7b6367fULL,0xfdfff7ecd62e6fdeULL,0xa94bb772d70474fcULL,0xdf8cff6ff6266fe4ULL,0xd4ced912f63420cfULL,0x6ffcdfdbd6967fdeULL,0xfeffffebd2be6fffULL,0xf1ffbffbd6a46f7eULL,0xdf9f7ffef68661f7ULL,0xdfeeff7ff7ae77ffULL,0xaf7ffedbf7b63fffULL,0x9fbeffffd7ac7fffULL,0xb3fdff70d6b427fdULL,0xfbfe3f7cf23e6dfcULL,0xebf7f73bd62e3bf5ULL,0xfedfbfdbd6be6effULL,0xbbbccfbad7942bedULL,0xd7aaf77bd72e2d5cULL,0x5ffebd60f62e3bcdULL,0xdbf7e73ef23e39fdULL,0xedebbe8ff6a42befULL,0xaffedf77f63677ffULL,0xf0fbff7af72267fdULL,0x9ba2ff73f63e2d5cULL,0xdbbefffef6be7bffULL,0xee4fff3bd72436ffULL,0xaf5a4d5a920c3d86ULL,0xf9f44ef0b62c3969ULL,0x92f6ef6ad60635f5ULL,0xbbc6fe17932e2d54ULL,0x7ddfff2ef7b47ff7ULL,0x4b7fdddbd63e2ae5ULL,0x9c9fd72fd69469efULL,0xdb7eff7bf6be2377ULL,0xdfbfbfffd69c7fefULL,0xeffefffff6be6f7eULL,0x2addbaddb2a62b46ULL,0xb0938febd63473e6ULL,0x95c8dd65d22429c0ULL,0x7a9e3ef1d60e6fd4ULL,0xdbf5fbaad60c79fdULL,0xf7fb32dbd6b43bf7ULL,0xa5185fe356242362ULL,0xf7f1f67cd61e69bdULL,0xd28e7debd6bc7bf5ULL,0xfb8c636ef624637cULL,0xe8c6bab2d2266f6cULL,0xef9e5fbbf62e2f7fULL,0xb4f47642d6042fedULL,0x921a3d70f60c6becULL,0x4b0216ad829c28cdULL,0xc01021c92000134ULL,0x6b4b991bd70625ecULL,0xd614f444d6b42174ULL,0xdffcaf72d62429ffULL,0x7395200dd6046285ULL,0xa500e40286002044ULL,0x8667c10d20e22a4ULL,0x1685601920c0160ULL,0x4800a480d6040180ULL,0xc6902b00c2042095ULL,0x7008b122920428c4ULL,0x600002800ULL,0x8010080000010ULL,0x50000800040ULL,0x20400041000ULL,0x8000822040202000ULL,0x98062748d6142da4ULL,0x2ca9de28f4a47ae0ULL,0x5000a00090004040ULL,0x400c712080942860ULL,0x1d20001100c2001ULL,0x4480001000242860ULL,0xc28180046100ULL,0x0,0x400000c0000ULL,0x2088c002408e28f0ULL,0x304a302540e3880ULL,0xa10b481291841240ULL,0x58314360560420a6ULL,0x918aa718d4804144ULL,0x0,0x42000000200040ULL,0x100080010000020ULL,0x1000040080004000ULL,0x28082000002001ULL,0x2002000040080ULL,0x40c0a000c0002980ULL,0x101207900c2200ULL,0x4c100242080ULL,0x220d0882102d04ULL,0x50242c60d6004154ULL,0xc0c074a4c610201aULL,0x5fde3e39f48c7aefULL,0x8b9857abd4806264ULL,0xfd7dfad0d69e31f4ULL,0xeaaefdc1d7046d7dULL,0xda80b30e44061296ULL,0xce9bbe8bd33e43eaULL,0x247a1272d6042914ULL,0x5536a2f2d62c00dcULL,0x9208256154040220ULL,0xcb28e36ed60c21ddULL,0xefdeba9fd6867d67ULL,0xa31117dad6042514ULL,0xf4fbdce9f6a66b76ULL,0x506f8dd2d60e523eULL,0xeab42facf71e013eULL,0x60dbfa06d61420d9ULL,0x4ededdcad21e6afcULL,0xbcf7dfecd706677dULL,0x616f37ead68e69c4ULL,0xda362f51560c2b5eULL,0xcdcef67ef68c7954ULL,0xabf67edff69e7fefULL,0x5aa4dd292064168ULL,0x1080000010000040ULL,0xf6082f4cf684246cULL,0x8100f5aad00421a8ULL,0x294611d091204870ULL,0x11094fb2d2806064ULL,0x8576e728d6022e5aULL,0x2010000002000ULL,0x800411090000240ULL,0x895232289200254eULL,0xd084360f68440c4ULL,0xb5f58c68f6242ba8ULL,0x4b2e4e2192042e24ULL,0xb2386640c6024156ULL,0x4043200244060080ULL,0xab9ecd8bf234255cULL,0x841214a882242554ULL,0x142243854002041ULL,0x9bb24474d6042157ULL,0x8750f4aa920c694eULL,0x2305d7dad22222d5ULL,0x363e4cfad60e2b9dULL,0xfb77ff73f2b62ff7ULL,0x5ea62fed92166b65ULL,0x8bf7ff7fd6be67f5ULL,0x2a59cecbd6a62dc8ULL,0x81024f6356062343ULL,0xf024872006042562ULL,0x48182cb0c4040090ULL,0x1d0420812043824ULL,0x1010940020c0204ULL,0x400000a082000100ULL,0x406120192006200ULL,0xd887dd2f610234dULL,0x0,0x0,0x120c000080204040ULL,0x100066044041002ULL,0x6b7e7d63d73c2dedULL,0xcf2c55e6d494637dULL,0x100040044000000ULL,0x200000880000010ULL,0x0,0x4020000044000000ULL,0x9280220000002c00ULL,0x10810000100ULL,0x204191c256040066ULL,0x8400080040000ULL,0x1200000000002400ULL,0xd8808b52448c28c0ULL,0xcfd4d627d62469ddULL,0xefdffedbd79c2bffULL,0x3fbe5fffd6ac79efULL,0xde7e9679d216237fULL,0xdf9e3f64d7ae6f7fULL },
	{ 0xe3c3fa77a9f4633cULL,0x77fb3fdb8b9e5baeULL,0x5086036688c43224ULL,0x94fc36a989b4332fULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4480482002840220ULL,0xfcfff7fe8bf677ffULL,0x93bfaf56abfc5b3fULL,0x100042080340220ULL,0xdffbffbfabfe7fffULL,0xffffffdfabfe7bffULL,0xf7fe7eee89be5bffULL,0xfb7def7bab9e7f37ULL,0xeffe7fff8bfd7fefULL,0xfbffffff8bbf7bffULL,0xaf77dedba9de5f2fULL,0x909756ee8b946bb6ULL,0xdbfbf77e89dc5befULL,0xfe323d6389be73e6ULL,0xfdffffff8bbe7f6fULL,0xfeffbf5dabf95f77ULL,0x309b4ee28b9150a3ULL,0xdf7faf7c8bfe7fb3ULL,0xdefe7df78b955b73ULL,0xeb97fa278bfc577dULL,0x6ec78fd509bd437bULL,0xb7ff4fff8bbe77ffULL,0xffffffffabff7fffULL,0xfffe7f738bff7fffULL,0xcbfef73e8bfb4b27ULL,0xe8dbffff8bf45beeULL,0xff9ecffa8bfd7b77ULL,0xb5fd7fea89b657beULL,0xffbfff728bfd7ffeULL,0xffffffffabff7fffULL,0xc7f7fabfabfc5ffeULL,0xbbfdffd78bf46bafULL,0xfcffffffabff7bffULL,0x9feffffb8bfc7bffULL,0xffbe7fff89bf5ffbULL,0xffffffbf8bff7fffULL,0xffffffdfabbf77ffULL,0xfdfff7eb89b67bffULL,0xf3fefffbabfd5fbfULL,0xfefeff7fabbd7fffULL,0xfffff7ff8bfd7fffULL,0xffffffdf8bfe7bffULL,0xbfbb7ffb8bbf7bffULL,0xf7ffff7babbb7beeULL,0xfffe3fff8bbd7ffeULL,0x8bf7fbbb8bfe5336ULL,0xeff77fdeabbc7ffbULL,0x1eba6feb8bfd73efULL,0xffffffffabfd7bffULL,0xfffefff2abfc7ebfULL,0xebf6e77fabfe7ba9ULL,0xfeefbfdb8bfa5b6fULL,0xbfbfdffb8bf87fffULL,0x5ddc7eee89fc57e7ULL,0x93afad32abdc5beaULL,0xfbde9f3e8bff7fafULL,0x8bd6fb1789fc53feULL,0xbbffdf5eabf47b79ULL,0x76b7dee98bbe5f26ULL,0xb9e31d708be8636aULL,0xf7be7f348bbc7ffaULL,0x84e2b71a8be663a6ULL,0x6bde3dd189b61bf2ULL,0x3eff7ffe8b9e63ffULL,0xfdffff77abff7f7fULL,0xfebe6f7e8bde77aaULL,0x8493f287abd30b36ULL,0xcbf79f5b8bf61263ULL,0xc6b3ffd78b9d67a3ULL,0xd7ac37fd8b8d7b7bULL,0xfb722fc58bfc53a1ULL,0x89dffbfe8bbe47f6ULL,0xeff3bfdf8bbc5bfeULL,0x9f5ecfff8bfc7bffULL,0xd7fff6faabbc7fbfULL,0xf2f6ff718bfd72f2ULL,0xcbf7f7beabf55feeULL,0xeffbbfff8bff73eeULL,0x6fbedffd8bfe7ba2ULL,0xd7cfd7eda9fc5bbeULL,0xdbabfc56abfc7bbbULL,0xf79fffff8bfb7fefULL,0xf5d7bfbe8bfe4b7fULL,0x610fde9289740b6aULL,0x10b7666e8bbc42a1ULL,0x93ffef738bb45bbeULL,0xcb867f6f8194737bULL,0xffdeef678b7e6bacULL,0x2afedfd189bc7bbeULL,0xa7fe4ffb89b64fa7ULL,0xfb7daeff8bbc7332ULL,0xeab627678bbd7eeeULL,0xffcfff7fa9dd7bebULL,0x24ffdf728bbc43aeULL,0x9eebefff8bf67fb6ULL,0x93fcf64689800b36ULL,0x7fce772689fd5327ULL,0xc9a1f36e89f85babULL,0x6df3bfdeabbd53a7ULL,0xbdbadfb88bac7b6eULL,0xf676a6fca93853aeULL,0xdf22edc08ba943aaULL,0xf39cfe0689ff42beULL,0xee4fbb5e8b8e53f5ULL,0x7bbfc7fb8bfd4bfbULL,0x7cffbfe6abf87326ULL,0x8a82d620b904188ULL,0x880000200800002ULL,0x4022808098a0209ULL,0xa30c051f89b84390ULL,0x945497e18386032aULL,0xdffdaf7dabbc6bbeULL,0x700e3ca10b90426cULL,0xc77bfb3b8b9363fbULL,0xe990279e09c00a03ULL,0x34abdfe8a3a6233bULL,0xcb7c672809b076beULL,0xca902529811c0a0aULL,0x127028a1800282ULL,0x275cdfd8a9a073fcULL,0x13329c089804923ULL,0x0,0x490c036081b04100ULL,0xd200f92283b003d7ULL,0x8804000602800200ULL,0x8a804d0803940328ULL,0x9000000000100200ULL,0x908240201900201ULL,0x2989e10a0130132cULL,0x800050818a801100ULL,0x203156c00a880211ULL,0x904809000208ULL,0x40200800001ULL,0x200000208400000ULL,0x10100000040200ULL,0x2002008180200202ULL,0x112009688a800222ULL,0xc4220702a0140202ULL,0x201028480a404208ULL,0x4000202800000ULL,0x108100800202ULL,0xc502000b00201ULL,0x4008300002800001ULL,0x4002000001000020ULL,0x100040000000280ULL,0x80010a040200ULL,0x100480201000120ULL,0xf1f514888ab04283ULL,0x0,0xd97bd2ec8bb9723bULL,0x6b7697d7ab9a037fULL,0x1f99cff289b41b6bULL,0xc4df36e88bc47aa0ULL,0x8bbe7e718bfa533bULL,0x80042e0e80840ab1ULL,0x0,0xa020006081004084ULL,0x59f3ae6e8bf47ba9ULL,0xcbbffdf08bbc73aeULL,0xde8d61e78b987322ULL,0xfd7fa230bbc7362ULL,0x2e5e55d089904b72ULL,0x80d0146189385268ULL,0x37eef7758bfc773bULL,0xf59c6f75a1b47bbbULL,0xffc76eb28bbe4bffULL,0xaa971f4188e97aabULL,0x5cd454bb8bb277ffULL,0xf5ddae6c8b9273d9ULL,0x54222d56abb80320ULL,0xc77ed31489d07b2aULL,0x3b57720babb41363ULL,0x921a7ff08bf81337ULL,0x4404814081005800ULL,0x4986304089620002ULL,0xcb9afe3b8bb843bcULL,0xe2ca48db8b305204ULL,0x411878b00b884233ULL,0x1422604001104240ULL,0xa2615a0031c4320ULL,0x8164d22001300360ULL,0xc4b52d889384206ULL,0x5980c9200b880112ULL,0x94f434ea8b805b22ULL,0xcfbe3de08bb84fb7ULL,0xd896b70c8bbc7b70ULL,0xe4e7e01abb9428fULL,0x8077454a8b800386ULL,0x24a0e86909541300ULL,0xb874f77189d94ee7ULL,0xc6852165a18443baULL,0x6840e712ab88603aULL,0xdb3edf5f8bb4436fULL,0x9c5614e589f40322ULL,0xf96faf7eabf97fffULL,0x9cb6057c819057afULL,0xcbfff72eabbc73a7ULL,0xebdddf4b8bf07a67ULL,0x36b85e6b893c7e66ULL,0x9155947289344b20ULL,0xf0024218b805130ULL,0x8844431809d22320ULL,0x41111c90881c122cULL,0x3b384d000986632bULL,0x160067081821462ULL,0x1280280080100e20ULL,0x8002009000044a0ULL,0x0,0x12092180200122ULL,0x2068000800030ULL,0xa586070891c0342ULL,0x810000480a800001ULL,0x102880001004120ULL,0x290d9100ba82320ULL,0x1a40a081900014ULL,0x840240009004002ULL,0x88005240810040c6ULL,0x6044619089114286ULL,0x2c21e1083901328ULL,0x2008500004288ULL,0xc929b00209206360ULL,0xb8102500008063a0ULL,0x4102c12289846322ULL,0x2b73be4da3bc53a2ULL,0xbffffffeabbe7bffULL,0xdcfdfff989fc5befULL,0xfffefffe8bfc77faULL },
	{ 0xc9b2f31b5bd4d1daULL,0xebdf1fdf7bc5d32aULL,0x9baecdcb6ad5c33aULL,0xc77a7ef80bc4c23aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x85ab412a82c4e10aULL,0x7cfff6efebc4cbfaULL,0x9210a571e3d4fb8aULL,0xebb7db7ffbd5fb3aULL,0xcffffffbfbddfb7aULL,0xef7fdfdaebd5fb6aULL,0xbfffff6ffbddc3faULL,0x79beff62fad5d93aULL,0xffdeffeffac4f32aULL,0xffdffffffbddfbfaULL,0xefdf3fdbf9d5f9eaULL,0x909337e3c3c5e392ULL,0xfd7cfffe69d5d0baULL,0xbebe2f5cf905eb9aULL,0xeffffdaefb4dd3aaULL,0x2f5efefbc9cce33aULL,0xb6ba77f5cbc4ebfaULL,0xf6fef77b7bcdd3baULL,0xfefe7f7ffbdcfbfaULL,0xebb7ab7febd4c33aULL,0xefef7edff3d5e3aaULL,0x1fdfede3fbd5e37aULL,0xfff7fffffbddf3baULL,0xfbbe7ffbdbccd17aULL,0xfbfdff6ffbddd37aULL,0xefd3be7b63ddc3faULL,0xefffdff7fadce37aULL,0xf5fff7fffbddf3baULL,0xfffffff3fbddf3baULL,0xffbffffffbdddbfaULL,0xcfdffeabfbd5f3daULL,0xbf5ae7dbf8d5e37aULL,0xfefefffffbccfbfaULL,0xffffb7fdd1d4ebfaULL,0xffff7ffffbddfbbaULL,0xefffffbffbddfbbaULL,0xefdfffdffbc5fbfaULL,0xfffffffffbddfbbaULL,0xffefff7efbddfbfaULL,0xffbefffffbd5fb3aULL,0xdfffff7ffbdddbbaULL,0xeffffffbfbd5fb7aULL,0xbffffffbfbd5fbbaULL,0xffedfffff9ddfbfaULL,0xdfbeefeff9cdc3baULL,0xc1e1fdfeebdce37aULL,0xefdf7e597bd5c37aULL,0xffffdfff7bd5d33aULL,0xf5fefeffebdddb3aULL,0x5dfe7df2fbd5e1baULL,0xfbf7eb7bfbc5c32aULL,0xce6a7e8bf9c4eb2aULL,0xaffecffbfbd5eb7aULL,0x54d4fffc79c5c1aaULL,0xf9fe7762f9dce32aULL,0xebb0770e63d5c32aULL,0xeff7f7937b4dd32aULL,0xebffdfdb7ac5d17aULL,0xf4b6dbe8e1c1d9faULL,0xd3fb2f78f144f3aaULL,0x9f1caff77944c34aULL,0xd8f6ff9e6984d3aaULL,0x6aeefedb79ccd3baULL,0xfefbffebf3d5cb3aULL,0xffeffffbfbc5cb7aULL,0xfafe3f767bd5c3aaULL,0xcf7ff7f3f9c5fb3aULL,0xebffdedff9c5f3baULL,0x3fbb7ffffbdcfbfaULL,0xf5fcbf7dfbdddb6aULL,0xfeae7fe3cbdcfb72ULL,0xcbbdfbffe945fb7aULL,0xffffbfdffbdcdbbaULL,0xeeb6fff3e3dddb3aULL,0xfd7cfa7ff3d4db7aULL,0x8bee7d68d1d5f13aULL,0xbbfcf77f6bc5c37aULL,0xeed3bf9ae9dcfb2aULL,0xfb9e4ff1e1d5eb6aULL,0xf4ddf6eee3d5f3daULL,0xdbbabf70f9d4e33aULL,0xed97f7fde3ddcbfaULL,0xdfdffefff3ddfbbaULL,0xff51d35343dcd27aULL,0x54bf36e051d4c37aULL,0xd3fff756f9d4f378ULL,0xdf957f25f1d4ca3aULL,0xd8c7a2075bc5c10aULL,0x6f065a487244812aULL,0x9df77febc9cce23aULL,0xf95cad78f104f13aULL,0xdf85277c6345d32aULL,0xc0d6d1367ac0c12aULL,0xff57dbca4bc4d1baULL,0xfebcfd9fbdce32aULL,0xd1fdf65bc3d4db1aULL,0xdb1277e4e1c5f08aULL,0x8b46dd2febcce31aULL,0x28d6de137bc4c13aULL,0x1f9e4f05d38cf33aULL,0xadecf77e49d4c30aULL,0x477efde0e34de18aULL,0xf9ffff3ffbddf3aaULL,0x2dd2bb39cb4c734aULL,0x2138cd7bf9c4e08aULL,0x4d70d6ea71d4d10aULL,0x93e1532a305e022ULL,0x99bef5e901d4c32aULL,0xc5809c0002c18152ULL,0x20348584104130aULL,0x949555a951c4933aULL,0xcbaa7fe34144c34aULL,0x82a43e4bda05eb02ULL,0x8002c30040d41002ULL,0xb98cc418044e83aULL,0x10500000404048ULL,0x40060370e0446202ULL,0x2240000044002ULL,0x2080000040200ULL,0x57564970c48302ULL,0x4a240840900c10aULL,0xf04694560bc4c220ULL,0x528a246d7001d11aULL,0x0,0xe232604004d31aULL,0x10200d41c080e100ULL,0x0,0x4880002000ULL,0xb3a1292ad3c4e96aULL,0x1c1080404040ULL,0x6000003000000ULL,0x51400001000200ULL,0x1220a40003040010ULL,0xb024834050001112ULL,0x0,0xa01004000800800aULL,0x0,0x80468248880a102ULL,0x20280200440012ULL,0x41200080042010ULL,0x108000080002000ULL,0x460600001848108ULL,0x888500608080c008ULL,0x0,0x690414706145c208ULL,0x8c1cc4083002012ULL,0x8007c1d1040a00ULL,0x100d82602b81c002ULL,0x4040245040041308ULL,0x8011d2006000c128ULL,0xb61e504005c10aULL,0xf9bdad7beb05f12aULL,0xcc76ab106144d30aULL,0x7cbe75e3c3cdf36aULL,0x519112008044a028ULL,0xdd22e22e0c0c91aULL,0x6022012040c0110aULL,0x543264806084c112ULL,0x10831200080c102ULL,0x52907be6c3d5c30aULL,0xcff3feb5ebddc99aULL,0xebfcdf57e3dceb5aULL,0x10b28e68c1c4c248ULL,0x99ef7d4b5b84c13aULL,0xdfb66ff3eac5f30aULL,0xcdcffebe7bc5d37aULL,0xaadf5fdbf3d5f13aULL,0xc5baf36bd3d5c91aULL,0xdbeff662e1c4f93aULL,0xf6bf1ff2f3c4ebaaULL,0x4a5477a4fb85e00aULL,0xeff39fd6ebd4d9faULL,0x84bb6a7b6ac5d33aULL,0xb7fce76f59c1d32aULL,0x9063e208084e912ULL,0x8880062400d0c102ULL,0x4d470ee8c0c0c82aULL,0x1086b70800dc822ULL,0x2100002880002108ULL,0x44ba00708344430aULL,0x94082040088c110ULL,0x42000000004002ULL,0x1298c420d880c02aULL,0xe18235684184d31aULL,0x89e4ef50e38cd10aULL,0xfaac3f1449d1c2aaULL,0xc1c05203d9045008ULL,0x9100100c0c002ULL,0x47845606040c10aULL,0x31ee4662c104f32aULL,0xc69c3f62e2c4cb0aULL,0xe9e7f3166bcdc33aULL,0xbd7b7dec3d0e37aULL,0x6eb40de372c4d33aULL,0xf975fe7ac1c5fbaaULL,0x3c163ff0f1d5e16aULL,0xdfdbf3eacbc5f33aULL,0xafcfbfdaf9c4f16aULL,0x35aa6fe2dbc4cb5aULL,0xba5cbf09e1d5d9a2ULL,0xb29c2df08144e35aULL,0x8005e38b8144e39aULL,0xaa4d024a8354e13aULL,0xa98ce5359c4431aULL,0x242e6272705dc108ULL,0x720464014350510aULL,0xf83442226144021aULL,0xc40200000458000ULL,0x4488004a00aULL,0x3000088060001000ULL,0x4818081203004108ULL,0x8e2c054e41c5c1d2ULL,0x42400a2150d40012ULL,0x0,0x188153000000ULL,0x0,0x0,0xc44ec0007104c2a0ULL,0x9ede5ad08944e3daULL,0x10984c41e000e02aULL,0x7cc60bc90cf30aULL,0xc406244980c1e138ULL,0xcf6bfb7f4bd5c35aULL,0xaa5f184641d59308ULL,0xf6baeff7ebdcdb7aULL,0xd6ffb7e17bd5c33aULL,0x614655eb01d4c19aULL },
	{ 0xc9e7f2ebc8db76a6ULL,0xb6fa7b53c6dabe60ULL,0xb5bf0f7744caf5beULL,0xf676d67cc6cbb230ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x55c24da3ced3b622ULL,0xfcfffeeececbfff6ULL,0x840820018093b140ULL,0xfbbbf36fcedbf6eeULL,0xdfdfffffcedbfbf6ULL,0xffffbfdfcedbfeeeULL,0xd4bfdfecce8b97fcULL,0x7beff663cec2fbe6ULL,0xdbe076efcec3f7a4ULL,0xefffff3fcedbfff6ULL,0xffffffdbcedbfbeeULL,0x442f5ee7c69ab7e4ULL,0xd17ceff3cedbf66eULL,0xee9427fe4ecad36cULL,0xcfccffb3cec2f3f4ULL,0x6b3fbe5fcec2b3ceULL,0x695dc704e9bb262ULL,0x516df7624ed2b276ULL,0xffbeffffcedbfffeULL,0xe8c5fbebccdbfb32ULL,0xefff9fddcec3ffb4ULL,0x3e9fdeaeced2f370ULL,0x77eff7fececab6e6ULL,0xf5be3d73ccdbfbaeULL,0xdbf0f73e4edabf5cULL,0xffffffdfcedbfff6ULL,0xffbe4fffcecabff6ULL,0xfdf7fefd8ecaff66ULL,0xdffefff2cedbfff6ULL,0xffbeff67cecbffeeULL,0xcffffebdcedaffe6ULL,0xff7fdfdfcedbff7eULL,0xfe77feebcedbfbfeULL,0xff7ffff3cedbfbf4ULL,0xdfddff6fcedbf776ULL,0xefffffbfcedbfffeULL,0xabffffdfcecbfffeULL,0xb7bf7fefcedbfffeULL,0xfffeffffcedbfffeULL,0xfebeffffcedbfff6ULL,0xefdffbbacedbfbe6ULL,0xeffffefdcedbfefeULL,0xb7fb5fffcedbfbb6ULL,0xd7efffeaced3bf7eULL,0xf6ba3f74c6d3f7a2ULL,0xe9def7facecafbb2ULL,0xefdbffdbcecbfb9cULL,0x9ffbcffbcccbf7feULL,0xf777fef7cecbb7eeULL,0xdfbe7fe3cecbfbf4ULL,0xbbfdcf3c4ecbf784ULL,0x6e53ffd3cecbffacULL,0x6f3ecffb8edbbf60ULL,0xfd7576e84edbb7f2ULL,0xdf8aa772ce88b6acULL,0xebb4f7fbcec3fb90ULL,0xdddffea14ec3fbf4ULL,0x8f5ededbcedbbbd8ULL,0x54bff6eec6da7a34ULL,0xa7477fe9cecbfac6ULL,0x99142f344a52bf30ULL,0xd343a73a4ed2fa30ULL,0xabdddfdfced3fbb0ULL,0xb7fbf7ebcecba766ULL,0xfd7fffffcedbfffeULL,0xfa9c4f72cedab762ULL,0xebe7dfbfcedbfe8cULL,0x27ff5fdbcedbfbf4ULL,0xb73a7ff7c6cbfb26ULL,0xddef36fbced3fef2ULL,0xb28c7da1cec3fb84ULL,0xcbe5fbffcccbf6feULL,0xebd69e9fcec3b720ULL,0xb6dfdffbcccbbbeeULL,0xf5e6f6efcedbf3e6ULL,0xfafe3d71cecbf7f2ULL,0xfbf5e77dcecaf7e6ULL,0xcfdbbed9cecbfbb0ULL,0xf53f4dcb4ed2f3fcULL,0xd4f6ffffcedbbab6ULL,0xfbbf7de1ccdb9a24ULL,0x8bb05f3ecedbfae4ULL,0xdd7efebbcedbfff4ULL,0xebbfdfdbcedbff7cULL,0x37ffc4c08acbbac4ULL,0xd9f4b77acecab3b0ULL,0xce3ebf5486da39a0ULL,0xfde7bf9c46cbbfe0ULL,0xee185b494e8ab6e0ULL,0xaee227fbce83b6ecULL,0x59ecf55244dbffb0ULL,0x9eb8277cc6cab77aULL,0xe9aff31bce8bb6e2ULL,0x69fbbfe3cec3ba6eULL,0xbffb6fcbcedaf3f2ULL,0x797f376bcec1db1cULL,0x7aa67d654ccab354ULL,0xe9f3ffae4ec3aaf2ULL,0xe1671ef8c492d76cULL,0x5998dfffcedbfff6ULL,0xb1fffe48cec2a320ULL,0x5fbe76f0ced3fb60ULL,0xf9d4d9374ecafe20ULL,0x4101be12c8c21a20ULL,0xaf9ecfdaced27ef4ULL,0x4104a6c2084232c0ULL,0x840ec40ce829204ULL,0x8b2224664c8ad010ULL,0x48c5a000a022300ULL,0x318893d2c492b04cULL,0x10be76c9c0d39130ULL,0xdbeb657acec3f39cULL,0xb3b61f74c69bf262ULL,0x3ccf3328c8ab660ULL,0x98cc2518a839240ULL,0x86b254e846883004ULL,0x4f41b7720ed9b222ULL,0x56943f2cccc3a220ULL,0x2996d6364ac3fe30ULL,0x846d88884c22122ULL,0x480044c0c002600ULL,0x300c042044121400ULL,0xc1cee062c4820024ULL,0x0,0x200000280001000ULL,0x100009c448808204ULL,0x8000100080001000ULL,0x2a0d604ac22802ULL,0x10404020000ULL,0x5012080040a242ULL,0x200040004000008ULL,0x1004008001000ULL,0x93c8e4524cca7462ULL,0x8000400000020200ULL,0x4000442008004ULL,0x2a04094000121280ULL,0x402000008100ULL,0x1000004004020000ULL,0x12200760ca81a204ULL,0x5000000002002000ULL,0xae504800000a3220ULL,0x80100400010ULL,0x400cf53046929144ULL,0x4a26217a4c8ab5a0ULL,0x8013da014e8aa286ULL,0x2f4afe4b4ed3db00ULL,0x9f325f388ec8fa20ULL,0x46406004001d004ULL,0x46020104004222a0ULL,0x4200410a42c08200ULL,0xcd5c1e4846c2a602ULL,0x77aedffbced3b6ceULL,0xd626b2f4ca83d204ULL,0xcffefff3c6dbbffeULL,0xb880c77c44caad30ULL,0xeef3deb5428bb6b4ULL,0x4a0040000022180ULL,0x56f61eebcc83f3a2ULL,0x8b1a0c024cc10210ULL,0xd980777cc4cb834aULL,0xcefebe238ed3f3faULL,0xbaf7df4ace83b38cULL,0xf4d7fef8ced2fae6ULL,0xdeec7562c8d39900ULL,0xf3baafe6cecbb370ULL,0xa4d6eb3ac69ab686ULL,0xb3db17dbcec3fa72ULL,0xa5f917efcecbf6beULL,0xf9fcf7374ecbfb76ULL,0xfe9e2d77cec3ffb4ULL,0xcffef3734edabba6ULL,0xeaffd8cf0ecaf290ULL,0x400046204822040ULL,0x940c8ca6ccc838a0ULL,0xc84e752286d29230ULL,0x8054932884585014ULL,0x2412368004127020ULL,0x2a8cd840cc2a780ULL,0x3186221886820200ULL,0xfabe6dd24edabe86ULL,0x25c420202087220ULL,0x4819600064a2840ULL,0xa42100e1060292e8ULL,0x26802008200ULL,0x89a47c424680f6b6ULL,0x921a271006493e02ULL,0x4d14e20460a5220ULL,0x2804409b8098d000ULL,0x5424c4a08e81f200ULL,0x81182768ccd21c10ULL,0xcb58af0bc6c2f83cULL,0x6dd6f702c68bf426ULL,0x6a5ff3d7ced2d2eeULL,0x16b37ff2ced3f634ULL,0xd974f64b4edbf1b6ULL,0xefbe2efbcecaf7aeULL,0xcdc7ff3fcecbbfeeULL,0xfffffefbce92f3fcULL,0x94924ea3c6c2af24ULL,0x73c836b8ceca32a4ULL,0x3a866f74cacab652ULL,0xe9f3e33ac2db9666ULL,0x20b394a0c8ad220ULL,0x9a9a9f584edabb94ULL,0xcc152e2c46c38a4cULL,0x68c3840064a2a00ULL,0x9000116044c00204ULL,0x8135004c20240ULL,0x8409cd6044c22204ULL,0x3000000000002020ULL,0x14052040c08804ULL,0x8800000080800000ULL,0x1548e0008401200ULL,0x828080282101204ULL,0x0,0x20004040000004ULL,0x0,0x842020e0642a300ULL,0xad545ac80618f148ULL,0x342005b208c1d200ULL,0x10aea500044222a2ULL,0x529614d0c8d297a4ULL,0x608ae13aca0a9734ULL,0x2591bda46cabfb0ULL,0xfffbffebcedbbaf4ULL,0xd9fffe684ecaf724ULL,0x5bf67bdec6cabb2eULL },
	{ 0xebe4f6ef8d737b9bULL,0xee561edfca473e9bULL,0xcd1adcf016e2b611ULL,0xf5edf6fa8fc76fbbULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9f8adffbeb733ab3ULL,0xfdfff7ffffd37fbbULL,0x8001000000004800ULL,0xeb99f76baf737fbbULL,0xdffefeffdb723bbbULL,0xabe8fd3c9f33a83ULL,0xf4b77efecfd3bfbbULL,0xdd5b37f05a63be29ULL,0xffb67f7f5fd7ffbbULL,0xefffffbfff73ffbbULL,0xefffffdffff7ffbbULL,0xc53a5bff09d35a1bULL,0xf9e4e7eacce2bf1bULL,0x5a973f7adb53bebbULL,0xe9c7f7ffdad37eb1ULL,0xaf5f7d5b49737b9bULL,0x96f2ee2a28733683ULL,0xf56e767d8b737fabULL,0xfebe7fff9f737fabULL,0xebd7ffba7cf6fb1bULL,0xeeffbfffdbf7bf8bULL,0xbfbedff7ed73fbbbULL,0xddffb6ff6bf77a83ULL,0xcbaefff3ca53fa91ULL,0xfbdfff7e2dd37eabULL,0xbfd3ff9b1fd3ff9bULL,0xbebfdf7ffff7ba9bULL,0xf5d6f779ddf3ff93ULL,0xffbffff7dff77b9bULL,0xff8fef6fedd3fe1bULL,0xeefffab609763eb1ULL,0xaf7cdfdfff537fbbULL,0xf4fffffffdf3ffbbULL,0xfffffffbfff77fbbULL,0xfffe7ffbff737fbbULL,0xefffffbbfbd7bbbbULL,0xefffffdffff37fbbULL,0xb5ff7fffcff3ff8bULL,0xfbffffffef77febbULL,0xfebe3fffffd7ffbbULL,0xedfffffffdd7fbbbULL,0xfffffeff1f777fbbULL,0xbfbf7ffffdf37fbbULL,0xd5feffff4bf73f93ULL,0x7bf6ff7f6ff37b99ULL,0xe9f6f37faf737e3bULL,0xefcfbf7b1df37fbbULL,0xedaedffbdc773aa1ULL,0xfdf7fffb6ff37fbbULL,0x7fbefd7239f37bbbULL,0xdbb7ff7fcbf3bf8bULL,0xaef3bf5909733f01ULL,0xefbccffbc9f71a91ULL,0x10f7f6c089c25a81ULL,0xbb1b35704d73beb1ULL,0xe3267b664e52be11ULL,0xcaf6faa71a737bb9ULL,0xfbdecddfcd52fa11ULL,0xd4b5fee90ef27e1bULL,0xdbecf73a0853fb3bULL,0xdeb73f6ead573fabULL,0x41ceff1b0e763b13ULL,0xeffcfd5bcdf33f93ULL,0xc7be5fb78df33a1bULL,0xdff5af7299773fbbULL,0xdebe3ff7bf777f9bULL,0xeedafeff6ff3fa9bULL,0xaf5ffff71e737a3bULL,0xfefb6fff4b73ffbbULL,0xfdff7efbeff37eb3ULL,0xdfde3f7f9b73bfbbULL,0xebfcff620fe33f9bULL,0xebf71fd7cbc35e1bULL,0xb5fedde28df3ff31ULL,0xf77dfffb8bd35aabULL,0x5bfeff6b8b73ffabULL,0xf9f7fb2ebff7bf9bULL,0xffe3fffd5ed3fb91ULL,0xaefe4f5e79677a93ULL,0xf7365efb8bd33b1bULL,0xbff6ff71bdf73b93ULL,0xdfb5efef7dd7febbULL,0xcf5ffe8f0ff7bf93ULL,0xffecd5a0cf73b1bULL,0xf69726e828763a1bULL,0xb5efff744f723fabULL,0xfa872f744c673a9bULL,0xd9c6b3980a521e8bULL,0xeb9bdddaccf73f0bULL,0xb4f6776e99e33b33ULL,0xdffef7fb7ff2ffa1ULL,0xdcbe2f78c9671b9bULL,0xcbc4f777cdf37f03ULL,0x3d76ff5f3b77babbULL,0x77ba5ffdce737a9bULL,0xd5fd1e7b18727b99ULL,0x6db2fd4c48723abbULL,0x8ba3ba7bb9f23bb3ULL,0x64e9fe518e4223b1ULL,0xd054f6158533a8bULL,0xf69e7ee54b53fb01ULL,0x508e6d631a721713ULL,0xf184f7741b537f01ULL,0x4043ae38bb461a13ULL,0xeabce7f98af23b1bULL,0xf576b66f5a52363bULL,0x0,0xc8ec672288c35a01ULL,0x52104000021011ULL,0x20191500492320aULL,0x541404340a023891ULL,0x28ba8f304b42be03ULL,0x8addaf7999d33ebbULL,0xb8cfffce48f27e11ULL,0xc8d29f41c8021a09ULL,0x404124000640600ULL,0x2108020008009901ULL,0x482a2c620d023910ULL,0xe344e1309af35a01ULL,0x210cc21009010901ULL,0x2409642848601201ULL,0x806848000001ULL,0x2040000800800ULL,0x200400000021000ULL,0x0,0x421014000500220ULL,0x9400200000400202ULL,0x280040000804a00ULL,0x0,0x40020000021000ULL,0x0,0x230d669a04202b0ULL,0x0,0x0,0x4c3301008031a80ULL,0x0,0x90102a0008600021ULL,0xb110ac2031413a80ULL,0x926002a8c222000ULL,0x41c1e01200c61a10ULL,0x4010034100425820ULL,0x0,0xdc40324208521001ULL,0x3080200100400280ULL,0x8fc6aa2229d27ea1ULL,0xab9f7d414e733a0bULL,0x22321cfa0e813e01ULL,0x0,0xa214253008701a03ULL,0x700a0040021011ULL,0x480a0000000020bULL,0x43a4d5b22f127f01ULL,0xbe569ff848779a0bULL,0x5bfaff726bf33f93ULL,0x8b95e73d8a43b281ULL,0x81ca1a9146562209ULL,0x180c400060020aULL,0xf8d456fa0dd36e30ULL,0x9ffabd204d12ff9bULL,0xa9a0af2f05523a39ULL,0xc9ca90518dd23aa9ULL,0xa99adfdb38621f91ULL,0x94bddee368537eabULL,0xcdfeef6a4f437f2bULL,0xcb9e3ffcd8437f3bULL,0x4a46eb1259f35a81ULL,0xef7ede5adf527a1bULL,0xbcff7fef5fd33bbbULL,0x497cfe7a0c533eabULL,0xcebe77f39ff3fe8bULL,0xcddfff3fca77fb8bULL,0x5e5e7f5b6bd37f0bULL,0x9a284c1809010f01ULL,0x11248e0000509201ULL,0x4a80256028549281ULL,0x4040e00100021600ULL,0xe7ed0e5298139391ULL,0x613445008421b00ULL,0xf46e32604a461aa9ULL,0x65a27d600d82fa23ULL,0x0,0x8c0b00080031800ULL,0x2cb283d30f62bf01ULL,0x44000021000ULL,0x108542201521800ULL,0xfb00430f8313ba21ULL,0xc2062908224a13ULL,0x23440a4008021901ULL,0x54a2bec60bc37a9bULL,0x96ef7f9c9f21f9bULL,0xcbdfff64eff27e81ULL,0xc8d2eb600d23ff81ULL,0xedfffed12c73fe83ULL,0xbc3f77e95e73fa31ULL,0xf3edf6f44af31a23ULL,0xfabe9f75ce763bbbULL,0xebf3ffeefdf33ab3ULL,0x8adfdec70dc3babbULL,0xdfbfcffb5f573a9bULL,0xf9feb77d4f773ba1ULL,0xff567f6549737e93ULL,0xeb5dff2adbd2fa81ULL,0xa9551bd02a733b0bULL,0x6bbc6777c8b71b1bULL,0x9c9030410912190aULL,0xc1283f40ac836a88ULL,0x920025448002100aULL,0x0,0x212c4ca101533a22ULL,0x6184444008701800ULL,0x1200014001021800ULL,0xc000c11408001e01ULL,0x1870048000681ULL,0x201005008600001ULL,0x8002000101000ULL,0x4800000008000100ULL,0x8404006801710a01ULL,0x9c80018801a00ULL,0xa551d5420a521681ULL,0x0,0x620023200102081aULL,0xd7ae3c7548f71e8bULL,0x5002202208401b00ULL,0x38cc8031a561211ULL,0x95325ef43cf77fbbULL,0xd7fff7ffdbf37fbbULL,0xfaee7d72dcd27e8bULL },
	{ 0xcbfef3bfef7f79deULL,0xaece1e39e37630d6ULL,0x9e7bdce7a6749fcaULL,0xf47f76682e7411deULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xefbceffba77fbddeULL,0xf5ff7f7be77efdceULL,0x0,0xe888d52a207e11caULL,0xcffffebfe67edddeULL,0xa67edfdf627659ceULL,0x7efdfeeee77f7ddeULL,0xbdefff72a67f1fdeULL,0xdf927f7b2e7731daULL,0xffffffffef7fffdeULL,0xefffffdbef7fbfdeULL,0x3ebe7fff6b761ddeULL,0xdd3dfffaee761cdeULL,0xeab03d536976d9deULL,0xebf7f7bfe17ebddaULL,0xabffbfffe37f15deULL,0x33aac7fde37f94ceULL,0xd6edff72e27515ceULL,0xfffe3df4e77e3ddeULL,0xd9cfffbfe77f5ddeULL,0xefffbffb667f97deULL,0xf7bbcffbea7f9edeULL,0xf57dfe6be16f99ceULL,0xd6be3de2eb7e5c5eULL,0xcbece76eaa7f5ddcULL,0xad5bbfffaa773ddeULL,0xffbedf7fe77e5ddeULL,0xd4befef7677e1ddeULL,0xeedfff76e67f94deULL,0xfb8dffeee7765e56ULL,0xc7ffffbbee777fdeULL,0xefffffdfe27f3fd6ULL,0xfcfffefeef7fffdeULL,0xdffff77fef7f37deULL,0xffff7f7fef7ffddeULL,0xffdfffbeee7fdddeULL,0xfbffdfdfe777ffdeULL,0xb6ffdfffef7f5fdeULL,0xfffdff7eef7fbddeULL,0xffbfbfffe7777ddeULL,0xefffffffe77fdddeULL,0xebfffffbe37ffddeULL,0xfdfaddffe377b7caULL,0xfbedff7f677f5ddeULL,0xffde7f7fe77fdfdeULL,0xcbfff7ffae7ffbdeULL,0xefd63e1b6b7e1ddaULL,0x1fbeefebe35e5dceULL,0xfbfffefdef7f97deULL,0xfb3e7de2a67f9fdeULL,0xeba7cf3e2b7f35caULL,0xfbda9653eb7e21caULL,0x8eb7cfe26f765fdeULL,0xd4d1f6e1e77733deULL,0xdbbbbe73a276b1d2ULL,0xfbb4ff7ee87ffddeULL,0xcbf7fe1d6777ddd2ULL,0x2ffedfdba27e91daULL,0xfcf8d5f7ea7419ceULL,0xd9e23f60a26e004aULL,0xbf940b61207299d8ULL,0x6fc5ee82257190d0ULL,0x2bfe59c3227d14ceULL,0x24e7666ea2753cd0ULL,0xddfcfe8a62761d56ULL,0xffbe3f5fef7fb5daULL,0xcfeeff7eaf77ff5eULL,0xabffea7b6e7695deULL,0x97fb7fff677c5fdeULL,0xfdfffffe6f7efcdeULL,0x9f7e2fefe7765dcaULL,0xa9bcdf7a67761adeULL,0xf1de2e2baf7d174aULL,0xadf4dfff6e763944ULL,0xd73fee6b6a7695d6ULL,0x5b8efc43a276b5daULL,0xebf4ff2fe77e35deULL,0xeefbbf5fe77f35deULL,0xbfffcffeef761fd6ULL,0x7bff556b637f91caULL,0xcbffbd70e87695ceULL,0xd9bdffb7aa7700daULL,0x8dfffe25827719d6ULL,0xaa5ee7dfa77c15d6ULL,0xb831df61255a1754ULL,0xfdaff760ee7715c8ULL,0x8b6ebf72e276594eULL,0x8dcfee2e6f7715d0ULL,0xae5fde59e27714caULL,0xbdf25be3e27715daULL,0xf9fcff5bae7e17deULL,0xfbba8f5fe775b9ceULL,0xcbda7f3fa17c5cceULL,0x35f7fdeaf655cdeULL,0x3fbfcffb6776dd4eULL,0x9775f56be37f5dc6ULL,0xbf963df8e67e30deULL,0x6b56ef7ae36414d2ULL,0x6fdb1fe925729f5eULL,0x1f5ccfe1a06e91deULL,0x10eae72e00e11c6ULL,0x4bee6d62a1763cceULL,0xcbd2ebb6e37f17daULL,0xd85b920f4276354eULL,0xa3bc4fb3a23e95d2ULL,0xb4fbf7e1a2775f5eULL,0xab8ca761e0640a5eULL,0xc580a0430074135aULL,0x940800020020006ULL,0x3002810120221900ULL,0xc0201461212e0040ULL,0x21224fc108761146ULL,0xcbbfff6eaf7f9fdeULL,0xefd7f7bf6f76dd5eULL,0x829c5fc16276594cULL,0x8432d303205315c2ULL,0x9d34b6eee27d5858ULL,0x80c2fac627401c8ULL,0x61c6421048460048ULL,0x4400040040000ULL,0x2020000000000048ULL,0x1204010002600040ULL,0x280000200800ULL,0x38ce5a9623632ccULL,0x4032e28002010d0ULL,0x28888c0204200ceULL,0x0,0x84000000000050ULL,0xa0840003a0120400ULL,0x40430a30e0153054ULL,0x944c0000420086ULL,0x404000000080040ULL,0x20080000000240ULL,0xf02480a008600140ULL,0x8154a60e663606d0ULL,0x42000a01411000ULL,0x6668a5406c631d88ULL,0x0,0x8080000000048ULL,0xa004000100400048ULL,0x4820000008ULL,0x102122101000ULL,0x142540020268040ULL,0xca88656226740046ULL,0x6294d6a462361cd6ULL,0x2ffebbcd29773dc8ULL,0x8000442024000040ULL,0x8600000061400ULL,0x782034482c461148ULL,0xc9e3c29663565fc6ULL,0xf4008002010205cULL,0x1ec3c909a07318c6ULL,0x8435f66ba2761856ULL,0x605a253061340842ULL,0x8b80ff3ea377ddd8ULL,0x5cc71ff9667719daULL,0x29344c01226600c0ULL,0x140102060000082ULL,0xb7b23d50271c7dcaULL,0xd7867e6666379954ULL,0xcccfffb12b7635deULL,0x8a4edf8fa167dcdeULL,0x349d7febeb6759deULL,0x9d602610e0543496ULL,0xc0922e1c666000d8ULL,0xfcdfef1aa2765bceULL,0xabffdfdb637fb7deULL,0x8cb37fe3e776365eULL,0x71afdb6aea6f1ddaULL,0xfef63f7f667f3fdaULL,0xefd3fffbe77f9fdeULL,0xaf7ffedf66761956ULL,0x84162124e020014aULL,0xc03a0510ab605c42ULL,0x754e25482764008eULL,0xcae1b28a625e014eULL,0xc0565e01686608d6ULL,0x85b25a3222743848ULL,0x545406f2206249ceULL,0x3aa5f226076a8deULL,0x62908200e22018c0ULL,0x8d4a837123761442ULL,0x29b68d8aa25e684aULL,0x1001704024101000ULL,0x80a0c6123501540ULL,0x8091501000541144ULL,0x8033a22620c1548ULL,0xa90e8940e4540006ULL,0x44b23668227409c2ULL,0xed6afffaab7699deULL,0xfbbf6f7fe37e5ddeULL,0x88d6b62e2b678242ULL,0x8ffd595ba77f10d6ULL,0x3a8057fba07d11deULL,0xd46aff576775995eULL,0x189e1dece37710ceULL,0xeffeef62e7741fdeULL,0x8ede7a1b2f59154eULL,0x3e0e1ff2e76815d6ULL,0xd77db463a25717c6ULL,0x6a14e933e86e02c8ULL,0xa926030e23181450ULL,0x6fdabfd22276bd42ULL,0xf2f5defa27e91daULL,0x84d5d072616710caULL,0x2040000001100ULL,0x8a002546025417ceULL,0x241b68000341d4cULL,0x400c30000301142ULL,0x0,0x4002000011000ULL,0x8300400020620000ULL,0x0,0x5d2158022515ac4ULL,0x92900e80023415c2ULL,0x286e0102241982ULL,0x1200020200600800ULL,0x4042420800080146ULL,0x906001824020006ULL,0x0,0x8418253000721844ULL,0x2943812e14700c2ULL,0x83c862224401caULL,0x2000814022000008ULL,0x8faa27e6666ec8caULL,0xfdfefffbe77f5ddaULL,0xffbc2feee6767dceULL },
	{ 0xd1d7f3ae5e3deedfULL,0xebfbbe519f3cdffcULL,0x3fb95dfb8f1dfff4ULL,0x7596e6ee1b1dcbb3ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffffcff18b3dffffULL,0xbefb5eff9f3dfffcULL,0x0,0xfa19fb3a4e5deffaULL,0xcfdffffb9f7defffULL,0xeff7df5adf3deffaULL,0xfcd7f7ff9f29dd59ULL,0x7dee07689b3ccef6ULL,0xdfbcff7fd71dddffULL,0xffffffbedb7dffffULL,0xabfeff5bcf7dffffULL,0xffbfffebcf7ddffdULL,0xfdf5b77ecf3dcedcULL,0xd2ae7d7f9f7dcdf6ULL,0xeddeff1adb3dcfd8ULL,0x8e57dfdbcb3def7dULL,0xffbb5ffb9f3cefdbULL,0xf57e77c99b3cfedeULL,0xfbbe3f7edf7dcfd9ULL,0x89fef33f9b3dfeddULL,0xefdb3e7f4b7ccff7ULL,0xbebfdff7df38effaULL,0xd5f7f6ffcf7dffd8ULL,0x76fe7d72df7ddfffULL,0xbbf7ef7fdf7ccdffULL,0xc9f3feb60f1defd9ULL,0xfdfedf7ecb5deff9ULL,0xfdfffeffdb7dfffeULL,0x9fbffff2db7cffffULL,0xfff67f7fdf7dffdcULL,0xefdfb6f78f1cdffdULL,0xffffffdfdf3dfffcULL,0xdefffefdce3dffffULL,0xfbeeff7adf7defffULL,0xffffffffdf7ddfffULL,0xfffffffbdf7defffULL,0xefffffdbdf7dffffULL,0xf7ffffffdf7dffdfULL,0xfffffffe9f7dffffULL,0xffde2fffdf1ddffaULL,0xe7f7fb2bdf7dcfbfULL,0xef7ffffbcf7dffffULL,0x9ebbfffb9f3ddfd7ULL,0xffffffffdf7dfffbULL,0xfffe7fffdb5deffdULL,0x89feffaa9b3deffdULL,0xefdbfe79df3dff7dULL,0x9fdfdffa9f2cfdf9ULL,0x95fcbffb8f0dffdcULL,0xdf2e7de38b3cffddULL,0xeae7b72fce3cdff9ULL,0xeefbbfdc9b1dfeffULL,0xc39e8f6a5f7ccf9aULL,0xd4f7bceadb1cdffcULL,0xaffef731df5d5ff4ULL,0xcbf4bbffc91ceef9ULL,0x8dd6ff99cb1cbfffULL,0xf9ea85dacf7dded3ULL,0x36bc766a4714dd9eULL,0xd9eefe07c83ce8c7ULL,0x9bfe2f7ddf3dfbdfULL,0xeee79e22cd1ccf76ULL,0x28db9397431cde17ULL,0xb12e1fe10a54de19ULL,0xf13cf7591e3dc7ddULL,0x9eba2b4fcb7cefd9ULL,0xe7f7f7ef4e7dfffdULL,0xbffffedfd21dddffULL,0xd5befff94f3ddfdeULL,0xfffffffacb7cfeffULL,0xfff6fd6e9f1deeddULL,0xebdffbffdf3dffdfULL,0x4ee23b5d9a4cefd5ULL,0xba331f7acf34cf35ULL,0xd1e3f7d8de19cff8ULL,0x8bc67d61df7ded7eULL,0xfbfeff3f8b7ddfdeULL,0xcef23f7f1f1dfffcULL,0x6f9ecdd75b1deffeULL,0xf49c9efdce3dcefeULL,0x4bbebf73ce3cdfffULL,0xfba3f36f9f3dcfbfULL,0xebdff79f5f3deff9ULL,0x6fffdfdbdb1cfefbULL,0xf4bedcf40b1defddULL,0xddcfef72cb3cced8ULL,0x82b6e372095d4f1cULL,0xfceefb9e1b3df7fdULL,0x7afedf5a1e6ddfdcULL,0x55576ffe8f2cddf6ULL,0xfb1ce772cb38e7dcULL,0xfe962f6a8b2deffdULL,0xe977f76ecf7ccfffULL,0xac7fdea14f3ccfdbULL,0xb59a7ffbdb5ceedfULL,0xf1ffe778df3defdfULL,0xdff62f72571dca9dULL,0xa964da4e8a0dc7f8ULL,0xe61317fb471dc1d4ULL,0xdabec7f31b7dcfd9ULL,0x141ff6ea53348d85ULL,0x46324540820c0504ULL,0xb1f2f3169f6cff1cULL,0x87dbfe3a8f3cdfdfULL,0xbfbaee718f3dccd1ULL,0xe43747e4ca1cdefaULL,0x9beeff73cb3dcf7dULL,0x4284f7b4483d44e8ULL,0x14630041a294010ULL,0x20080552401c0084ULL,0x3c503750cd1c4658ULL,0xc7aebde0130cde5cULL,0xfbffffffdf7dffffULL,0xffffffffdf7dffffULL,0xee7ff5da9b3dfff9ULL,0xb0dadd6b0b1c9ef2ULL,0x59c6d542871cd7baULL,0xdba63deb4f3cedbeULL,0xe943fe520b6c8dd8ULL,0x42114008208440ULL,0xf2be0ff24a3ca6dfULL,0xd406011408640214ULL,0xcc61560c11c0ab0ULL,0x4066480000c00ULL,0x20800080020ULL,0x4f05060120c0450ULL,0x0,0x208250008000040ULL,0xa8806042080ca00aULL,0x230024010c801ULL,0x20049080a0002c0ULL,0x1004500000008418ULL,0x8071280000000210ULL,0x600000000210ULL,0x80d2080081004c40ULL,0x0,0x400002000100080ULL,0x0,0x0,0x800000200008010ULL,0x84000ce120d4020ULL,0x90000a020c0010ULL,0x7000203008002c80ULL,0xb01427611314c094ULL,0x8782f7c48b08cdd0ULL,0x85f1d518f1c4f58ULL,0x16906cd2831c8e79ULL,0xd56fbf548a08c7fdULL,0x84296089204440ULL,0xcb77f72a5a3de736ULL,0x4008005108304a00ULL,0xfebfefff5f7cffbcULL,0xc120d468001dc7d0ULL,0x2ec3960810cdc19ULL,0x5134000609186214ULL,0xea53beef8b3ccfffULL,0xc21206610d00c618ULL,0x14120a02080540ULL,0xdaac00401b01d213ULL,0xeb76672f4f3cdadcULL,0x89c7ff0b8f2cdfd9ULL,0xaf7ecfda5f347f94ULL,0x65b552e50b1c8efcULL,0xc1ae05601339ce18ULL,0x3982656c021ccf93ULL,0x74c3e233cf1dcc97ULL,0x6bffffdecf3dffffULL,0x95ff5fc79a2ccefaULL,0xb9fef73f4f3cffddULL,0xdf90bf5fda3cffdcULL,0xedffffbe9f3ddffdULL,0xeffffedb9f3dffbbULL,0xba074ff35f3cc7dcULL,0x7043332e860cd470ULL,0x62263c620b3ca498ULL,0x8080822282001418ULL,0x4cc72ebcdb5ddebdULL,0x3086c5e5090cec92ULL,0x94bf72709a5c6c88ULL,0x58ba3c521b1c3ffcULL,0x214020e08008099ULL,0xca4393511b35cc9cULL,0x80e4b90021087caULL,0x440114408380280ULL,0x280014008042290ULL,0xc000330686044e98ULL,0x44c27c120a383718ULL,0xb1c428e0a04879eULL,0x14d654ef825c9e16ULL,0xddbebff69f7d6f9cULL,0xfab5bf73cb3cdff6ULL,0x7bd7ee32491de2d6ULL,0x4fdfffdacf3dfffeULL,0xfef4fffbdf7ceff4ULL,0xff7dfefb9b5dfeffULL,0xb2b62e64121c96f8ULL,0x6dcee336853dceb1ULL,0x2940b0550a0ce594ULL,0x9600c508880884e1ULL,0xd709af79935c0f80ULL,0x300234200008cc38ULL,0x88c1eb590b08845aULL,0x5fda3253d3144fb8ULL,0x51804b2109097890ULL,0x9232400008294050ULL,0x0,0x3144802005144084ULL,0x985b3a4b1a1c97d1ULL,0xf99c1994a1d608cULL,0x6e90909c8d0ULL,0x108080004004440ULL,0x6800612401104c88ULL,0x0,0x6490680100c68ULL,0x34005640082c4c08ULL,0x0,0x2300242500009010ULL,0x2004000000008080ULL,0xa0c0400000890d0ULL,0x12a18840080ca486ULL,0x880c74130d28c886ULL,0x48e02e00020c489cULL,0x9019102030c4c24ULL,0xa850e2518b008e51ULL,0x10024fc0000cc8d0ULL,0x97ffbf7f9a1ceff7ULL,0xffbe3df19b3dcfffULL },
	{ 0x8bf7ef5f5dba9ff3ULL,0xebffbfffddbadfffULL,0xdf8feffbfdba0f3eULL,0xd8f6f951582a8fe7ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x284800c88c285600ULL,0xefbefffb5dbacfffULL,0xb4b356ee5daa9f6fULL,0x0,0xf8b1c93e6cba1d62ULL,0xefffffbb7dbadfffULL,0xeffedbdfddbadffbULL,0x74b4aeec1c384702ULL,0xd7feeffbb5bac7f7ULL,0xcbbf6dfa7daa1dfbULL,0xffefffbffdbadfebULL,0x7fde5fdb7dba17b3ULL,0x7ebbfffb8db857ebULL,0x716eff7ae5b8c7a3ULL,0xfff73547ddaa5dfbULL,0xefd7ffeefdba9ffbULL,0xaf9cbeddc5b847a3ULL,0x3fba4ffa75ba57e7ULL,0xddffbffdfdbadf7bULL,0xfefefd7ffdba9ff7ULL,0xebdfe73fddbadfefULL,0xcbd61fdbcdba45eaULL,0xfb8dfd35dbadfabULL,0xf7f6bf7e7dbadf2bULL,0xfff66ff3edba1fa3ULL,0xd376ff3ffdbadfebULL,0xffffdfff5dba5fefULL,0xebbfcf5ffd3a5fe3ULL,0xf6f7f7fffdba5fffULL,0xbbfefff5fdba0fe3ULL,0xfffff7ff5dba5fffULL,0x5ddfffab5d9adfffULL,0xffffdf5f5dbacfffULL,0xf4ff7feefdba5f73ULL,0x77befff37dba8fe3ULL,0xffb67ffffdbadfb3ULL,0xffffffbeddbadfffULL,0xffffdfdffdbadfebULL,0x9fffffef7dba4ffbULL,0xffffff7bddbadfebULL,0xffbfbffd75badffbULL,0xffffffbffdba5fefULL,0xeefeffebfdbacff3ULL,0xfffa5ffaedaadfebULL,0xfffffffffdbadfffULL,0xcffe7df6ddbacfefULL,0xebc7f7fed5ba9f7bULL,0x7fe77ff9fdbadffbULL,0xbbbeedf37dba8f27ULL,0xd1ffb7fbadbadf27ULL,0x4fbe7dc2959a8f23ULL,0xbb9eff3ebd2a1fe3ULL,0xeeebffd3d5ba1de2ULL,0xbfbedfd3fd3ac7ebULL,0xfc7f77fbfdba9febULL,0xaf3ab5f64d3a0fefULL,0xeb2acf7fcd381fe3ULL,0x41cef6020daa4da3ULL,0xf1acb1319b845a0ULL,0x448bc26c149a0522ULL,0xf7fff7fbddbadf73ULL,0xde9c2f7f5dba07ebULL,0xe636d62a311a0762ULL,0x649a1e4185b85523ULL,0x6f25ae50d3a1da3ULL,0xfb2707d8cd2a0defULL,0xd6bf2feefdba87ebULL,0xffddf7a7f8ba8fafULL,0xafdf7f4fdc3a1fe7ULL,0xb6bffffaedba4febULL,0xfffffffef5badfffULL,0xfffe7d77fdbadffbULL,0xfbfffffb7dba9fe3ULL,0xefe77ed37dba1ff3ULL,0x779b4f794dba1733ULL,0xddfffffafdba4ffeULL,0x55ae7df165badd63ULL,0xeb7cf3aae5ba1fe2ULL,0xaec3bfdcbdba5febULL,0xaebfefd77d3a5ff3ULL,0xf632f6fa7dbacf2bULL,0xdfcffde3ecba5fffULL,0xe79cf76e0dbacbefULL,0xdfdfbf9ffdbadfafULL,0xbfdedf9aedbadfa7ULL,0x96f67fe8ddaa0fe7ULL,0xcd7ffffbedaa5fe3ULL,0xfff737757daa4f6aULL,0xdfe6ffbad5bad7a3ULL,0xafffdfdedcba8fa3ULL,0x84ff5f6be5ba5d6eULL,0xf9ff8f5afdbac7abULL,0xdb7caf723d3a9fe3ULL,0xddc9f7f00d3a9debULL,0x2f5feb7b7c8adf33ULL,0xb5bbfff254bad76fULL,0xf57caf98bd3a5f73ULL,0xff567d5e9c9a47f3ULL,0xa904d326e4ba0f43ULL,0xeef3cd9402a096bULL,0x5b1dcd63752a4fabULL,0xd15ffe69112a4d4fULL,0xa63ae5508daa4542ULL,0xfbc4fb5f1d3a4f23ULL,0x47cbbe8f2daa0516ULL,0xa3ccbf16db2c703ULL,0xfdded3efcdbac763ULL,0x9b201c4300288323ULL,0xf1102d24180a0b63ULL,0x810d20331028103ULL,0x8886c810048202b3ULL,0x10b44d40749a0102ULL,0xc5feef77f8aa8febULL,0xfbfffffffdbacfffULL,0xffffffbffdbadfffULL,0x88448483d8100520ULL,0x16105ef104a21b4bULL,0xb0102a0401088303ULL,0xa00242c000a4402ULL,0x6044051a40020b03ULL,0x0,0x91204470881a0562ULL,0x8010200114800100ULL,0x20480000440aULL,0x800202088000500ULL,0x0,0x8008000000100ULL,0x8106206a14921400ULL,0x58bebd2244b85760ULL,0x10408000100ULL,0x0,0x2200000000100400ULL,0x52c008000300ULL,0x10001020000100ULL,0x8000200040000020ULL,0x0,0x800200000000201ULL,0xa000000402ULL,0x10005a000084c03ULL,0x0,0x1044020004001402ULL,0x6000000820000ULL,0x20040004000100ULL,0x1000054000000120ULL,0x1a142712012a88e1ULL,0xd902e609458a1b4bULL,0x6c5bfe3b0dba05ebULL,0x4a444600da80122ULL,0x5b52345af9320737ULL,0xff367f72f53a4de3ULL,0x8c5c17c249815a0ULL,0x47d70a03042819e3ULL,0x91b2cbf435a005e0ULL,0x1404a661452057e2ULL,0x21b04e32053209e0ULL,0x9bf4cf6fdd3a0df3ULL,0xe2fbb3c395aac5c3ULL,0x8a0a4720608201b6ULL,0x550450a201a200eaULL,0xc096244184aa0d02ULL,0x80813c84280500ULL,0x8bf5a80378a88d2bULL,0xe72c3d7413a0f47ULL,0x94b0e4e7552817c3ULL,0xd9eff760ddbadff3ULL,0xb2b137764cba4fbfULL,0xecc3fd2215a815a3ULL,0xaf7e5dcefcaa9fe3ULL,0x383c57e3c49a8f6bULL,0xd9ec7f6afdba9febULL,0xfad77f7f7dbacfe3ULL,0xd9f6db77753a8f2bULL,0xefdfdf543dba9f73ULL,0xbffa4fffcdaadf67ULL,0x91240d00401ac741ULL,0x6a12546005381d0bULL,0xa0e461030c8a094bULL,0x8841a00100a0724ULL,0x823c06612c089921ULL,0x97b356640c1a4743ULL,0x410a051050200740ULL,0x9990eb6854ba0732ULL,0xacde3aa189b88f2bULL,0x3b9e45f2493a8f13ULL,0x14304ed451200132ULL,0x3210503018000700ULL,0xc989ad2c448a1b62ULL,0x143b00204920c00ULL,0x215d11e4028836bULL,0x3564146405a21f43ULL,0x1a6def7b40ba0da0ULL,0xf793be74799a87a7ULL,0x8483c6082d2a0102ULL,0x27fef751fdb245feULL,0x54b5dee34db8c7a3ULL,0x99effe7a1dba8f9bULL,0xd0120742140adf83ULL,0x4e802331080ac9a1ULL,0x6f2936d380502ULL,0x3200249012a0022ULL,0xc260a402a0903ULL,0x839205400c8a0a21ULL,0xd992460314b807a1ULL,0x40160f40013a0d64ULL,0x30b02448480a0106ULL,0xf56e867815badc6bULL,0xcb82144060228107ULL,0xa804802000080920ULL,0x40e23942ad185502ULL,0x2d82810005081223ULL,0x240002600c00010aULL,0x82004400000040aULL,0x4190807001288181ULL,0x8840200004000140ULL,0x20000500800010aULL,0x42444e024821981ULL,0x4082004000020052ULL,0xda166f5c4d3a8f77ULL,0xc9d4d22c402803a3ULL,0x261d48d00c3851e0ULL,0x94857ec455ba15b3ULL,0x91502f604d908721ULL,0x1ca4356504b80703ULL,0xe094b314588a07e1ULL,0x255e265cac1a0d6bULL,0x10a20f4009288303ULL,0x5736976c25b8970bULL,0xfaf67df79d3acff3ULL },
	{ 0xebfbf7bebc9afceeULL,0x3ff77fcbbcb7fdeeULL,0x844c00001260a4ULL,0x974795ceb492eaecULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe8763907b496f2ecULL,0xffffcffbbcf3ffeeULL,0x94da4eec189a5fc6ULL,0x9330a8103432f06cULL,0xc9b6fbfa34f3eeeeULL,0xefdefbbbbc96ffe6ULL,0xef7f5fdb94bbfee6ULL,0x3cfddfe6bc9fffeeULL,0xf547bf723c9afde6ULL,0xffbe3f6eb4f2fce4ULL,0x5fd7febe349efeeeULL,0xebde5fd93cdb7ce6ULL,0xacba7dee94da6eeeULL,0xf16bff72b496feeeULL,0xfb8eaf6e34ff6eeeULL,0xedfef6bbbcd378ecULL,0xfedffff734b3ffeeULL,0xbf32effbbcf3ffe4ULL,0xd8fffffd3cf3feeeULL,0xffbe7dfdb4b6feeeULL,0xebf6ff7a3cffefeeULL,0xffffbffebcf3ffe6ULL,0xd9bdefffb4976eeeULL,0xffbefefdbcffeeeeULL,0xbb7677e834335eacULL,0xdbd5efbf3c9bf6eeULL,0xefe7fffbbcfb7feeULL,0x4f9ecfdfbcfffbe6ULL,0xf6ff77eebc9fefe6ULL,0xffdabf73b49afee6ULL,0xffffffffbcffffeeULL,0xcfffffbfbcbffeeeULL,0xfc5edfdba4b76eeeULL,0xfcf6fefa3cdfefecULL,0xffffff7bbcb7ffe6ULL,0xffbfffffbcdfffeeULL,0xffffffbebcfffeeeULL,0xfbfeffdfbcbbffeeULL,0xbeffffffb4fbffeeULL,0xffffff7fb4bbffeeULL,0xfbff3dff3cffffeeULL,0xeffff7bfb4fb7aeeULL,0x6bffffffb4ffffeeULL,0xffbbfffbb4ffefe6ULL,0xffffffebb4dbefe6ULL,0x7fae7ff3bcdb7eeeULL,0xd9f7ffaf3c31efe6ULL,0xefffbbfab4feeee2ULL,0x8f9edffbbcb7efeeULL,0xd132fef930b6eec2ULL,0xffbe2f73b49ddfe6ULL,0xdbd7efee909b6feeULL,0xffef7fbbbcb3fee6ULL,0x4f7cdff13c335fe4ULL,0xfdfdbfffb49fffeeULL,0xb7887770149b76e0ULL,0x4a896371149ce82aULL,0xcdd6fe8a3c136a66ULL,0xaeddddca34faebeeULL,0xa57fdec0b4d574e0ULL,0xfbeef77a14def7eeULL,0xbbb63ffb94fbedeeULL,0xa352d32aa09ae2e4ULL,0xeb7fdfdab432ffecULL,0x99827dc3a41262ecULL,0xff77a7f89493dfeeULL,0xf91e3f7bb4bb6eecULL,0xcfd3ff30b4f236ecULL,0x6f7fff4334b3ebeeULL,0xb29ecfb3349b6eeaULL,0xf7fdff7fb4f37eeeULL,0xde9ebd66bcb7ffe4ULL,0x69abf76bb0defeeaULL,0x25de975bb4334ceeULL,0xffba6ffb3cf7e7eeULL,0xbc5fefe9bc90dee6ULL,0x7dfeff63bcb3ffeeULL,0xbbddd37eb4d3feeeULL,0xeaebbf8fac9bfeeaULL,0xbf9fdfff34b7ffecULL,0xf5f776fd3cbbfeeeULL,0xbdfbfd72b49cf2ecULL,0xff9cfbef3cb3feeeULL,0xcfcffeb634f2feeeULL,0xaff7cddb3c9bfeeeULL,0xd4feffefb4b77feeULL,0xc4bffd70b49bececULL,0xff8e7f5d34bafaeeULL,0xcddb7fa9349dfaaaULL,0xe3f5f4934326a66ULL,0xfdbef7fbb4bfffeeULL,0xd8fcaf1234137decULL,0xca88272f343264aeULL,0xcff4db3e349afee4ULL,0xaa8fdfd6bcdbeceeULL,0xb71b7fff34b3ece6ULL,0xf1fc9eda34befce4ULL,0xff8e3d6ebcbaf6e6ULL,0xfbf6ff2f3cf3fee6ULL,0x2ef6bed93493eeecULL,0xdfdbcfffb413eceeULL,0xf4ffbe2404bb76eeULL,0xf53a5f5914136eaeULL,0x83d8c17610d35aeeULL,0xcee79e7734b3f26eULL,0x2b166914b411e2e8ULL,0x3067877eb0d3e6e4ULL,0x8d0a7460801270e0ULL,0x2b1a67fdb43272ecULL,0x40003611300048c2ULL,0xa218c200040263e0ULL,0xfef756fbb4b3f7eeULL,0x49a69d60349b26c6ULL,0x8bb66b273cf0e6e4ULL,0xe5f7ff3ebcbaffe6ULL,0x821e5c1734916de0ULL,0x2a800906000ULL,0x90340cc020107140ULL,0xc600204084100084ULL,0x4948d610841303e4ULL,0xa101090200000c6ULL,0x80040020000040ULL,0x140c852c94106a84ULL,0x1004040000123000ULL,0x0,0x80480020808040ULL,0x8008807030000820ULL,0x447cae401490c0e4ULL,0x5bbc6d2234dcece6ULL,0x2b8c600204325ae4ULL,0x480100114002004ULL,0x0,0x4000d04820c2c0c4ULL,0x0,0x0,0x40004a3000828280ULL,0x140010014000880ULL,0x400200000900000ULL,0x8406040910910482ULL,0x8100921000127800ULL,0x2211004001060ULL,0x31c4e0014c86060ULL,0x8020502114110042ULL,0x8008224210101802ULL,0x44000a0000880060ULL,0xc650400d00927ae0ULL,0x0,0x24e0a240009194c2ULL,0x5850936cb480c2e6ULL,0xc290a60fb4123466ULL,0xdb8eff6f14fee6eeULL,0xefff9eff3cbbffeeULL,0xa639640200db64c2ULL,0x420000010000002ULL,0x4cb655d130936fecULL,0xdb5c9b3e34b1e2e4ULL,0x448a13103499c8e4ULL,0x20504440c204004ULL,0x810000014000000ULL,0x1a0ca80418946800ULL,0xc2831a020492e006ULL,0x1efe1924b66ae4ULL,0x6bdf5b17b4b776ecULL,0x5420d6e194b7eae2ULL,0x93ef6f7994d37ee4ULL,0xffb7afffbc937eeeULL,0x87cb1a049c56a8ULL,0xef7ddf5b3cf37feeULL,0xbe7f3ffb3cbff6e6ULL,0xf15ef64234ba6ee4ULL,0xdbb20f6298dcfee0ULL,0x7f0e7631493eee4ULL,0xafffdfc93cff76e6ULL,0xffee5dfbbcbbfee6ULL,0x50f0970034ab4860ULL,0x281e30220084e4e8ULL,0x9878162308462e2ULL,0xae8e32cb94b236e4ULL,0x3eb7cd6314327ce8ULL,0x348e34a8148a6cc6ULL,0x20383d60242202e4ULL,0x100803000808084ULL,0x46031e2800b27884ULL,0x20af8b613400feecULL,0x684066110b2ec40ULL,0x80040420000042a8ULL,0xd18020010091e028ULL,0xc45ff88d24986ae6ULL,0xf29ed5d3b49feee4ULL,0xbcb292ecb4b2ffe2ULL,0x37820f70b8b2e460ULL,0xf6a62e77b4b67a6eULL,0xc8dfb633b49f7aeeULL,0x4f7edcdb34937ee6ULL,0x75b3ffeb34d8feeeULL,0xd1febfe9bcf2efeaULL,0x8022478809062a0ULL,0x8a4120210006cc0ULL,0xc402a0a008662c6ULL,0x80031000901000ULL,0xc5448e6b3480e5e4ULL,0x82002c4010114440ULL,0xc3e2f36a349372eeULL,0x480a320508106860ULL,0x18aac4104307cc0ULL,0xc110146814886c00ULL,0x4140010000400ULL,0x100800000100020ULL,0x840140010000142ULL,0x5394d40b41266e0ULL,0x410202e01ce04400ULL,0x5020004010104800ULL,0x610010c00400ULL,0x840c00014000800ULL,0x800010010010000ULL,0x10042060000220c4ULL,0x2000002000020004ULL,0x180000500104084ULL,0x6866000a4223020ULL,0x8f384d50001368e6ULL,0xd0ba5fc134f06ee4ULL,0xc888a7203480e4e0ULL,0xab042730b491d8e8ULL,0x2183c0010093a240ULL,0x39cfaff9b4ddcbe0ULL,0x18244f7084ba42e6ULL,0xd1901472141125c4ULL,0xdb9667ff3cb3eec6ULL },
	{ 0xc9effbbf721c77faULL,0xeecabad37214dfe0ULL,0x33095c90701014b8ULL,0x95d9f7f2b014d6f1ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x244c02e00142b42ULL,0xeffffefbf034feffULL,0x9fbfcfdff2fdd7feULL,0x1890a48c50b49ec8ULL,0x91188f70105cd498ULL,0xebb3ff3f929cfffbULL,0xcfd7f7effa75b76eULL,0xbb17dd579a34d6ceULL,0xfc3f7efdf2d4fefaULL,0xdd43f7cbd01df6f8ULL,0xd6a13fe6f01dbeeeULL,0xcdd7ff167055f6e1ULL,0xaf9fdddff0b0f7f8ULL,0xb4875feeb0143ce2ULL,0xf768f764585c54daULL,0xfe9f27fef2d4d6faULL,0x795ed73d90b4bfcbULL,0xaaff3f5dfafcff76ULL,0xfe9befbbd274fefaULL,0xd5fdffdbd2bdf7f2ULL,0xfffe7f7ef0d5fffaULL,0xcbfffbefda7ddffbULL,0xfe56ff5eb83df9eaULL,0xb73fffe7983cf7daULL,0xd577f6fbd2fdffebULL,0xdbf23f0260b45e6fULL,0xdbf4e73e7074fef5ULL,0xa7f3bffbf2fcf7f8ULL,0xefffefd3d29cbffeULL,0xfc77f7fdf2fdb6faULL,0xbdfcff77f255fffaULL,0xffb7effff2fcffffULL,0xcfdfffdffabdfefaULL,0xeedfef18da7cfecaULL,0xf4fffef6d0fdfffeULL,0xffffff7bfadddffbULL,0xffbe7ffff8f57ffaULL,0xffdffffffadcfffeULL,0xffffffdffafdfeffULL,0xbfff7ffffa7cfffeULL,0xfbfffffff8ddffffULL,0xffffbffff8fcfffbULL,0xefffff7ef8bcfffbULL,0xfffeffdffaf5fefbULL,0xfffbfffbfafddfffULL,0xfd7efffffafdfeebULL,0xffde7773fadd7ffeULL,0xcbfff33ff01d7ff3ULL,0x7febfadffa3d7efeULL,0xffdfcffbf2f5dffeULL,0xb4b67a79f014d6faULL,0x7ffeffe2b8fddffaULL,0xf998fb17ba65b66aULL,0x5ecb1e5f50d597ebULL,0xaebfdfd6d874fffcULL,0xf5fff7ff7a9ddffaULL,0x9bfae74750b4feebULL,0x81201364501c1482ULL,0xcce3fa29506cd6c8ULL,0x5f9ecfdbf87cfff3ULL,0xf4f54779f03cf6f1ULL,0xfbffaf7b521cd7faULL,0xfbb7af76b0357febULL,0xefe6ea1e50dc1feaULL,0x6f5fdfdbf0fcdfffULL,0x957bdde3123cd2caULL,0xdf48af66281c5eefULL,0xd7be3fe498f556f8ULL,0xe1c6df3e501c97ceULL,0xe8ff7e595075df72ULL,0x979ffff7185dbef3ULL,0xf9fcf67ef0dc97e6ULL,0xf7febf72d0bc77ffULL,0xfba8f7aa00bcf6cbULL,0x6fdf3adff014d6faULL,0xfabe6dfbf2d4fffaULL,0x9cbe766f5abcd6e8ULL,0xdbbcbd63f8f45efaULL,0xeb95e73e3ab05f68ULL,0xdbdbbf7ffaddffffULL,0xfffd5fddd8fdbcffULL,0x51d63ffad294f7c3ULL,0xd7eaef7072b4ffe2ULL,0xfefe7f7eda5cfeccULL,0xdcd2fe8fb25d97efULL,0x9eb7cf5af09df6f8ULL,0xf6f5fef97254f6f6ULL,0xffffff6afa94feeaULL,0xfa947efed815f7ecULL,0xc9d6f3faf06cf7eaULL,0xaffeffdb701df6ebULL,0xd6bf7ffbd0f4ffebULL,0xd97faf6cf09cf7faULL,0x7e3a2d70708cd6ebULL,0xcf7fff7ff27dd6efULL,0xabd6ff5fd0d4fe6dULL,0xfefbfffff0fcfff9ULL,0xd73eff7ff2f556faULL,0xe73e3d76f8d4bfeaULL,0xebf7ff1ef2f4fdfbULL,0x7fef5fd1f8fcd67aULL,0x9bbacffb82f5fe6aULL,0xdd7eb23a7075f7eeULL,0x65ce75625055b6eaULL,0x851e7300005b5eaULL,0xaec63edb701597e2ULL,0xffb1cdf3f81c7f7aULL,0x30b40e42d034d4a2ULL,0x85b827409258d440ULL,0x5306f3209031cce8ULL,0x8141901098041480ULL,0xc08a48001020ULL,0xb29286ecb2a0d560ULL,0x5388a060f0103462ULL,0x8bac0a443050d6e8ULL,0xed16e7476a10d77aULL,0x6f085f1170151aa8ULL,0x30202de0500092c8ULL,0x5204000080001140ULL,0xf4961efa70153ee8ULL,0xc0d09201288c1168ULL,0x69041e80501068c0ULL,0x90050010801080ULL,0x8000100010000080ULL,0x0,0x300001200004240ULL,0x2a04915050180481ULL,0x4c101a040055468ULL,0xc01070603020d4a0ULL,0x4002000000240ULL,0x0,0x400300000008440ULL,0x120405010140240ULL,0x0,0x0,0x420082261214a040ULL,0x4100840000000480ULL,0x0,0x0,0x0,0x20000210000080ULL,0x1080000004000ULL,0x8488000040406ULL,0x150000240042ULL,0x712ca2482010dee3ULL,0x32000e0430101628ULL,0xe10048e0120086c0ULL,0x110000100800ULL,0x11024010001082ULL,0x4842a500501022c0ULL,0x1a14046020d45020ULL,0xd9d6e73fb0bc73faULL,0xedd7de035a3c76faULL,0x3698df42b0d4d4f0ULL,0xd510c86230405662ULL,0x5bae47e23090def8ULL,0xc9dcf316f2b45ef8ULL,0xde4a3b5f7a1d95e3ULL,0x11c040009016e8ULL,0x4f15660509499e2ULL,0x2210100000240ULL,0xc900550c20941d68ULL,0xe74db6011201f768ULL,0xab4a07ca301c47e8ULL,0x50800e6c001416d2ULL,0x917d266ad098f662ULL,0xe2beaf96f095d6ebULL,0xc9d6af6350b8fee5ULL,0xebfbdfdff2fcffebULL,0xf5bffffffafc5efcULL,0xb7fca6f2129cdecaULL,0x3ac635a270103ef0ULL,0xe9c6eb33f0a0f4eaULL,0xa87fdfc8906456c6ULL,0x3fbfdffb32ddd6feULL,0x53a817fad0b116e0ULL,0x83027001000424c6ULL,0xa1fdf32ef88d36eaULL,0xbfd7ff9fd01cd6eeULL,0x4bbe8df73274d7eeULL,0xb427b221523054d0ULL,0xc0022552f2501068ULL,0xdbc5f324d05c168cULL,0x8863b3b93074d6e7ULL,0xef3fddf5d2d0fee8ULL,0x90900e48088c1242ULL,0xa88254230511432ULL,0x83301b1312e0d142ULL,0x5dd7be85f034f7ebULL,0xa2eaca9080145c8eULL,0x5cf37fee1294bfdaULL,0xddbaa75a5894f648ULL,0xcb9c267e201c3768ULL,0x88e2f2c3f255b6e8ULL,0xebfcdddff05dd6feULL,0x84bb6eff123c5fdcULL,0xd189e752b000d5e2ULL,0x80860340103412caULL,0x4080000000040002ULL,0x8060600000c1248ULL,0x1620044010000480ULL,0x13488768d024f5e8ULL,0x1466861d020f440ULL,0x4381ebbb9098b54cULL,0x23473f4bd834bdc6ULL,0x18e59301034f452ULL,0x6000002000110030ULL,0x4024300050040082ULL,0x91a1c07040c190e0ULL,0x2082180050101208ULL,0x90c600000941202ULL,0x1d100204000102aULL,0x602400b0000020ULL,0xc00000440ULL,0x8423f1200141c8aULL,0x4400000000400ULL,0x680104150180481ULL,0x40400000001008ULL,0x800822000040480ULL,0x85888401509810a0ULL,0x800200001400ULL,0xc000466012009440ULL,0x20026010001082ULL,0xcc942870001496d8ULL,0xc1b0f96a1058d7c0ULL,0x96ed3c6504d15e6ULL,0x93304fd3325cd0e2ULL,0x140d448805416c6ULL,0xf7962ff5b0dcb7f8ULL },
	{ 0xdbaff77fd9efb35aULL,0xccdffe97d9f5b76aULL,0xbdde6f24d96dab5cULL,0xdd74b270d9fc2304ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe9d4ef3ef9ffae6eULL,0xffdfffeff9edbbfaULL,0xbfbfcffbf9ffbbfeULL,0xbef7fefff9fdaf5eULL,0xdbeccf22f9ed2b4eULL,0xe804496a516c0372ULL,0xefd7febbd9ffbfdaULL,0xbff5fdfd9fe235eULL,0xf7ff7ffff9fdaffeULL,0xd967ef74d974ab4aULL,0x4fbb7fffd9e42bd6ULL,0xefdffebbf1ed275eULL,0x2ddcdf58d9e92a0aULL,0xb4bf1deff1edbfeaULL,0xd0d6afebd9dc267eULL,0xfab72cbad9e8b3f8ULL,0xeff7ff3ef1efb3deULL,0xeffeffdaf9fdab7eULL,0xbfbfe7fbf1fdb3faULL,0xd1f9f7fbf1ed3bdaULL,0xfbbefff5f9cdaafaULL,0xebcef3efe1ff227aULL,0xefcf3f7bf1edafdeULL,0x95dfdfeff9efaffaULL,0xff7bfb7bd9ed3bdcULL,0xdfbe3d72d15faef2ULL,0xdbffefbff1fdbfeeULL,0xfcfbbfdbd9ef2fdeULL,0xffbffffff9fdbb7eULL,0xfdf6fffff9efa3fcULL,0xf7f6fd72f9fdaf7eULL,0xebfbf7bff9efbfdeULL,0xefcffebbf9ffa7faULL,0xeb9fcfdbd1fda37eULL,0xfcfff6fff9fdbffeULL,0xffffffffd9fdaffeULL,0xffb7fffff9ffbb7aULL,0xfddfffbef9ffabfeULL,0xefdfffdff9fdbffeULL,0xffbfffefd9fdbffeULL,0xf9fffff7d1fdbbfeULL,0xffbe3ffff9fdbfdeULL,0xfffff7fff9efaffeULL,0xeffffffff9ffbbfeULL,0xfffffffbf9edbffeULL,0xfdf7ff7bf1eeabfeULL,0x7ff47ff4f9dfaa7aULL,0xcbdeff7ff9efa3ecULL,0xefff5fffd9fdaf5eULL,0xf79fdeeff1bdaa6eULL,0xf57ef8fcf964bbdeULL,0xfffe6df3d9edabfeULL,0xdbfeff7ed8fcbf58ULL,0xa8f35aa9c1c42bf8ULL,0x6ffccfebf9edab5eULL,0xf5f5ffeed9fdabdcULL,0xdfaefff3f9efabfaULL,0x79867f09d1fda05aULL,0xd7e3b1bd0ef23c2ULL,0xeffedfdfd9efbffeULL,0x9efddf61f9efb35eULL,0xf5ef5f70f9f4a64aULL,0xfbde73bdd18dbbceULL,0xeddfffb3b9ed2f5aULL,0xefdfffdff9ff3f5eULL,0x74d73fefd9ecbfceULL,0xc9edf77af1e5a7f8ULL,0xcaa6af70b9eca34aULL,0xedebf3f9d1ecaa7eULL,0x79db9e81c169a5e6ULL,0xb7bbefe2f9cca7deULL,0xddfffffbd9ef3b6eULL,0xdffe7d67d1cf2ad6ULL,0xebd7fb6bd98b3966ULL,0xefdf3f53d1ffa36eULL,0xdb9c6fbef16d2b52ULL,0xffb7f27cd1f4baeeULL,0x6ffead72d9fd2272ULL,0xd39cff67d177a36aULL,0xefefbffbf97fabfeULL,0xf7be4f73d9dfbffeULL,0x75eefffbd9efabfeULL,0x9ffe5f72f1e4b25aULL,0xdbb6ff3ff9fcab1eULL,0xdfd6febbe1fc2776ULL,0xe35fdf5bf9ed3b4aULL,0xf5fff7faf9efaf7eULL,0xdfefff7bf9dda6daULL,0xfe963d7ed1cfb7d2ULL,0xcd5ffe59d1eead5aULL,0xaa025d5bf1e5af6aULL,0xfcfd7fefe9efaf7eULL,0xd97be7cad1e5323eULL,0x91263de0e0cc3d54ULL,0xcfa9f72bd1dfbf6aULL,0xadfefed3d1eda2e2ULL,0x36be7ffbf1f7abfaULL,0xfd7d7729d1e5334aULL,0xda9e7f61c9cd2adeULL,0xebddfba6d8fdaf4aULL,0x67e73b59f0dc2b18ULL,0x1a28877bc9890ad8ULL,0xb1827ee9d9442124ULL,0x158e3f61e9a5b2f0ULL,0x39c4ef2fe9cc82faULL,0xbedbbbebf9afb2eaULL,0x20244ab1d18c3b40ULL,0x943142bcd0cd3204ULL,0x201e0500c0802006ULL,0x7106641d90970868ULL,0x84883a01c12c2352ULL,0x0,0x3009496850c48348ULL,0xdd3835c2d9e7a37cULL,0x7a9c2d7499cca148ULL,0xda422f02d140b750ULL,0xe88cc5880442a6aULL,0x8531d66188e40368ULL,0xd570062c918c204aULL,0x90ac2d60c06d2060ULL,0x80daffb0c9612742ULL,0xcde3e50c8cda3d0ULL,0x82048000402204ULL,0x8048260a88c12644ULL,0xa001c441480310ULL,0x4be2f12ee9cfa14aULL,0xe8800204402600c0ULL,0x80804010808040ULL,0x80100408d1800000ULL,0x12000010040ULL,0x1000010000080200ULL,0x0,0x489400a80650c40ULL,0xbcf4deffd9f82c06ULL,0x90002d6091900800ULL,0xfae9322fe1df2b6aULL,0x0,0x2100001ec1012040ULL,0x25b100e890900000ULL,0x9000000040000010ULL,0x4020005000640400ULL,0x0,0x0,0x402306880002946ULL,0xa210096000c40700ULL,0x50920c21808d2040ULL,0x1006238c10000e0ULL,0xaf52b94ed9d4aa46ULL,0x121400a81500108ULL,0x20248240008c2208ULL,0x3630205800cc231aULL,0x897ad768d9edb27aULL,0xa8c632dbd8643096ULL,0x35061962c1e42214ULL,0xb3f7f66579e5835aULL,0x6d2cfdf0d9f5a272ULL,0xc1d1804ac1692344ULL,0x7ffbb7d3f9e5abdaULL,0x2d3e6f44c84d3948ULL,0x10521caad1c03900ULL,0xc2b80d2061452362ULL,0x88243041910d824aULL,0xd6c2f631d9e9a3c0ULL,0x10b24152c10e8240ULL,0x74906eefd1e7bb6aULL,0x91884572d1703202ULL,0x92947631d84f234aULL,0xeee8db1af9d52a78ULL,0xffddffdff9ffbefaULL,0x9cbfffeff9efaffeULL,0xb1fde66ad9cd2b5aULL,0xd2867dc249e4b34cULL,0xcb2dde04d9f125e6ULL,0x661f9a36e94823a2ULL,0x6082e5b416d2258ULL,0x1dc4744cd148b4a0ULL,0x2140bc2061ee2342ULL,0xf124612b6124a042ULL,0x2941980041002042ULL,0x6ea8583d9750b58ULL,0xf0f596d0e0e0af50ULL,0x91800f20e01c2080ULL,0xba00490090058048ULL,0x9129188612040ULL,0xea3c4fe2c9fd26c0ULL,0x3085504648448160ULL,0x1064701c12c2008ULL,0x81880521d9802000ULL,0xcec99613c10caa56ULL,0x6f1acf8bc9fd2342ULL,0xf473f6ecf9cd2772ULL,0x8508724ac164b1c0ULL,0x82a46f0cd140a2caULL,0x6682e82ae1c11e5aULL,0xebfeff52f1eda25aULL,0x1cb35fa9d1c53162ULL,0x80100060c0002080ULL,0xd024253240483306ULL,0x8900122051022040ULL,0x650821a88e02140ULL,0x411420491880202ULL,0x1764372af0acaa08ULL,0x2f4e7f6df9ef234aULL,0x2894c20e51010252ULL,0xa7dededef16da9eaULL,0x1800cffaf8ec224eULL,0x40108ec8c100a042ULL,0x2c07401c0102212ULL,0x8004142000400700ULL,0x8801ca02c8242000ULL,0x8004b50c0a42244ULL,0x4260480b0900200ULL,0x81882e00c94c2042ULL,0x200180002000ULL,0xc029020c0252000ULL,0x44000002004ULL,0x54000508c0510288ULL,0x8200000040008ULL,0x1000020000008040ULL,0x802102080012004ULL,0x154100051401100ULL,0x10300142c1410002ULL,0x9080000080800800ULL,0x90844601c0808880ULL,0xc5f07b12d9c320f2ULL,0x60efa71c1e42d44ULL,0x8a048880840940ULL,0x121546080412250ULL,0xdbde3f70d1e5ab7cULL },
	{ 0xcbfdfffedd7efdb2ULL,0x66d73e7b9fbafff0ULL,0x8a9eeff352c8bf26ULL,0xd0b3fed4508af720ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebffcf7fdffefff2ULL,0xafffbfdbdffcf772ULL,0xe7fbcf779aa8ff60ULL,0xb4dffe67dffcff32ULL,0xdffeed639abefdf4ULL,0x4001c020cd0c2820ULL,0x1bffbe1fdf2eedf4ULL,0xefffdfdbdfbeff36ULL,0xfedffffedfecfd76ULL,0x856ef7fbdfaee7a2ULL,0xff9c3f7e9ffebd32ULL,0xdfeeff1bdb7cbf64ULL,0x6bde97db1a3cefa0ULL,0xf6ffdfef9fbcffe6ULL,0xb1bda7729abce5e2ULL,0xfefe3df75afcfd70ULL,0xeffffffbdffef7f4ULL,0xffffffffdffeffe6ULL,0xffbfeffb9ffcffd4ULL,0xff7fb7ffdffcff32ULL,0xfaf66f6c9afeff00ULL,0xe9d5f7b1dd78f744ULL,0xcedf9f6f1dfaf532ULL,0x3fbf4ffadffcf7b2ULL,0xbff7f7efdf3efff6ULL,0x49fe3f72daecfd30ULL,0xdbdecf7f9fbeffe2ULL,0xefdbffff9afcf752ULL,0xbf9fcffbdffeff60ULL,0xf4ff37e69becfd32ULL,0xcfbefd73df7efdf2ULL,0xfba9f73fdf7efdf2ULL,0xd9cefe379ff8ad56ULL,0xf77fdfdedffefd72ULL,0xf4ff7feedffeffb6ULL,0xdfffff7bdffefff6ULL,0xffff7fffdffef7f6ULL,0xeffffffbdffefd76ULL,0xffffffdbdffcfdf4ULL,0xffbfffffdffcffe4ULL,0xffffff7edffeedb4ULL,0x5eae2ff31bf8ffe0ULL,0xeff7ffbddffeef76ULL,0xefdeffdfdafefff6ULL,0xbfbbeff3dfbcef52ULL,0xd5e7bf6bdd7cff30ULL,0xfe7e7ffbd9feff70ULL,0xab87f7fbdaa8ffe6ULL,0xffffbfffdffcfff6ULL,0x9fb3cfd8ddec7536ULL,0xf7fefefedbbef754ULL,0xcfbe7ff29ffae520ULL,0xdbd1ef6eda3cffb6ULL,0xea7b97dbdbf8ef32ULL,0xbfffdffbdffcfd36ULL,0xfdfff7ffdffeffe6ULL,0xdbbfbf73db30ffe4ULL,0xd5a43eef1a2eff22ULL,0x47c7b6bfcf08ff20ULL,0xafffdfdfdbfefd32ULL,0xf2fbeee09decf770ULL,0xc16e7fe0da24ffd0ULL,0xfbd7767ddfbcf7c0ULL,0xeddffffe9f6cfd76ULL,0x2f7fffd39feefdb2ULL,0xb4b7d063db7cfc00ULL,0xfffeaf7fdffef7e6ULL,0xff9e2ffcdd3eb562ULL,0x4f6ff73a5ffebf00ULL,0xecde7e5b59ecf784ULL,0xb4bedffadb78fff0ULL,0xf7efffebdffcfff2ULL,0xffa67de49ff8f7e0ULL,0xd9dffaff5abcefe2ULL,0x6d5b5e154ff8ef20ULL,0x369fcffbd978ef12ULL,0xd595f167d9e8f722ULL,0x5ffa7d73dfacad36ULL,0x8b16af261afce730ULL,0xffffbf9fdff8ffb6ULL,0xffffffffdffceff2ULL,0xfcf7f6eddffeb7b2ULL,0xbbee66639ba8fd24ULL,0xffb0f7dfdff8fd32ULL,0xc9d7fea7dbf8fdd2ULL,0xbffedfde1afaff92ULL,0xd4bfffffdaeeffd2ULL,0xf1efef78dffee7f2ULL,0xb3962f444a68ed20ULL,0xcdc3eb9b98ecf510ULL,0xe9ecd5d31b6cf530ULL,0x9ebffffbdffeff70ULL,0x9f4ae7625a7cf570ULL,0x12a22d75982c6660ULL,0xe9dffb1bd8fcaf32ULL,0xa88f9fc29a2ce5f2ULL,0x9edfefefdfbc7572ULL,0x9d9df7fa18f8af70ULL,0xfff67f77dfecff66ULL,0xebdfeb5fdfecfde6ULL,0x1cce3b791fbcfd62ULL,0x16decfe38034e7e0ULL,0xfdf4df7217b49570ULL,0x58ae5df392bea774ULL,0xe8c4830e10209560ULL,0xa06ab4d94aa0a540ULL,0x292c445843382f20ULL,0x44b80ad010088d24ULL,0x4220240015020400ULL,0x4000820e00200440ULL,0x0,0x838a881a08086702ULL,0x5000008049084000ULL,0x3a21208830d400ULL,0xdb966f3ed548b710ULL,0x6fdff32edfecff66ULL,0x4ed85a1b97b4b5a4ULL,0x24bb526a8f482102ULL,0xfbff9feacff4fd40ULL,0x8280204005000600ULL,0x80c0a00108404000ULL,0x9ea4100d00d100ULL,0x24a35f604d206602ULL,0x20800002000ULL,0x80010005000000ULL,0x2002080000400ULL,0x448080000880400ULL,0x1208400000ULL,0x840020200804a000ULL,0x8000040000402000ULL,0x0,0x0,0x0,0x74d5b6ed9032ddd2ULL,0x9300b4509088a590ULL,0x4000001002000100ULL,0x0,0x0,0xb0b8d46812205732ULL,0xb90a37224d6cec22ULL,0x5aa0065088ec6000ULL,0xe0218000ca002402ULL,0x2340c0000086100ULL,0x1024004010001404ULL,0x840000001080000ULL,0xce02280048887000ULL,0xc312084888a524ULL,0x6b7e7ac3d80ce5f2ULL,0x989a4db3db7cf702ULL,0x120150a90284050ULL,0xd1822860d82c0430ULL,0xa104eb225f00b412ULL,0x4e4b284a9800f540ULL,0x643d7429824a0a2ULL,0x9918b94181202f00ULL,0x8bcef5425b3eb520ULL,0xe9d0f6769f6ef1a0ULL,0x58437e98d8c8fd10ULL,0xfbecfd9d8f8f730ULL,0x84f69aeacd70b560ULL,0xdba484325d3c7502ULL,0x10040d2008344400ULL,0x1cda2a115bcbd70ULL,0xabe88f5bdb3c6d60ULL,0xa09454689d200764ULL,0x1b6675799f8c2d30ULL,0xce345feadaaab672ULL,0xedfeef7adafc6566ULL,0xab7ddf1fdf7cff76ULL,0xf5bfddef9fe4ed30ULL,0xd17c766cdfbcafc0ULL,0xfa9e0b9a8868e724ULL,0xe37cd7784f7ce522ULL,0x650505a18a89c40ULL,0x103844611080ad10ULL,0xd0b0d758d8fce706ULL,0xcb9e155c98aca5f0ULL,0x8100649180002740ULL,0x2a10b2c3112c4930ULL,0xdccaef709fa0f570ULL,0x34ac88405200a560ULL,0x3ab024609dfc3720ULL,0xb32242029800ef30ULL,0x9841968398243452ULL,0x7c8edb438a6ce500ULL,0xd45552ead824b524ULL,0x834a7c410f24fd04ULL,0xca24d32ecac8ed84ULL,0xc7e6fe3bddccf520ULL,0xaf7f9f5b9b68cf56ULL,0x94ffdfe4daacf7d2ULL,0xd8882c4309308d42ULL,0xdb806c661d740520ULL,0x81e4c81602aca700ULL,0xa362575a58a4ffb0ULL,0x14a2fc4058a8bf60ULL,0x1300080081002800ULL,0x98262410302402ULL,0x4482000005242000ULL,0x9740502c9482520ULL,0x36104f619a602400ULL,0x13287f2812bc9602ULL,0xff1e3576dd646570ULL,0x8ccb004a202520ULL,0x4046023991080522ULL,0xa5b446618b84ad40ULL,0x4d0330108008d02ULL,0x4400244000a88500ULL,0x2880402110303400ULL,0x464826e912ace500ULL,0x3b2e290090686404ULL,0x4880000a080ULL,0x0,0x9404014040002600ULL,0xc7c1ba311d24fda4ULL,0x210845ca00300606ULL,0x2000400000404ULL,0x20024808402400ULL,0x9a82674ed02cb590ULL,0x1402020890002400ULL,0x0,0x90000000004100ULL,0x100886a00804d524ULL,0x0,0xc14208109022a400ULL,0x4080002000100ULL,0x600002810008810ULL,0x1200050600884500ULL,0x7a92adf49af23742ULL },
	{ 0xc1d6fb7bbfa65573ULL,0x4c4f3b7197f64f63ULL,0xfffddffb1ffbef70ULL,0x4732ea630ac77d63ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfbfcff7eb7ef5f63ULL,0xfffffbdf3fdf7f63ULL,0xafff4f7a9fde6f62ULL,0xf5f77febbfff7f63ULL,0xcf3efff1bfdfff61ULL,0x6882c02a36926f70ULL,0xbbd7fe6f97f77f73ULL,0xfbbedfdbb7ffff63ULL,0xfeff7fffb7f6ff63ULL,0xd9ffff7abfde5f73ULL,0xa38261801264d60ULL,0xfdc6f7fb37f77f61ULL,0xe53575aa3f0e961ULL,0xfebf7f6b3fbfff62ULL,0xfff5bffc1fde7f61ULL,0xdfb67fdebf9e7d61ULL,0xefdfff3fbfdeef73ULL,0x7dffffffbffe5f73ULL,0xffbffff99bef7f63ULL,0xfff4f7ff3fcf7f73ULL,0x5fd6fd758f86ed70ULL,0xebf5f73e86df4d60ULL,0xe85f9bd72b9f0b60ULL,0xdffbcffb87ff5f71ULL,0xfdf7bfff9ffe7d63ULL,0x6bfe7d6096aaef73ULL,0xfbd4f7373fff7f63ULL,0xbfdb7eef8fae7f63ULL,0x7fff8ff3bfde6f63ULL,0xfdf6ddeb3fdf7f71ULL,0xffffff73b7eeff63ULL,0xfb92ff7f9fdfdd63ULL,0xaff7da730bdf5f62ULL,0x7fffdf5fafbf3f60ULL,0xbeff7fe7bfdf7f73ULL,0xfbefff7b9fffff71ULL,0xffbeff7fbffeff73ULL,0xffffffff97ff7f63ULL,0xefffdfdfbffe7f73ULL,0xfdfffffbbfde7f63ULL,0xfbffff55bfffff73ULL,0xdfff2fff9fff7f73ULL,0xefffff3fbffeff73ULL,0xfffffeffb7f6ff63ULL,0xb77f7fff3fbfdf63ULL,0xfdff37e93fe76f62ULL,0x70ae2fe111e65d63ULL,0xc314f52c0bc74b62ULL,0xefffba5f37f6ff63ULL,0xdf9ecfffbffeff63ULL,0xf5f6fceb8fff6d71ULL,0xd2be7df01fa75d62ULL,0xfba7cbb6839e6f61ULL,0x8f7bba779f9e5f61ULL,0xfffeffff9fdfff63ULL,0xfdfffeffbfdfff63ULL,0x8fae75731fd6ff62ULL,0xdf14ffba1bf67d72ULL,0xcffeffcf1f9f7f73ULL,0xfb5edfdb0ffe5f61ULL,0xf4d67eb011d35961ULL,0xbb6de772bfd7ff61ULL,0xe7b63f569797fb62ULL,0xcdc2bf1c8a8f4d61ULL,0xef7cdd5a078edf72ULL,0x945e5bef8b068f60ULL,0xf1ffbff33bfeff61ULL,0xffbf3fff9fae7f71ULL,0xc7adf7bb17be5f62ULL,0x2b5afbde1fe65d73ULL,0x959edffb3f9e5561ULL,0xf3cfdf7f9fde4f73ULL,0xfffcfdf43fee7d71ULL,0xe1f6f71fbfaf5d71ULL,0x4db7e1313a65d63ULL,0xdd8fcfcb8fcfff62ULL,0xf55cbce29fe6ef62ULL,0x5bfe3ff00fdf5f63ULL,0xfb93fb2d8f9e7f72ULL,0xffdfbffbbfff7f62ULL,0xfffeff7f8ffe7f63ULL,0xf5fff7fe1efedf71ULL,0xb1bbb7738fae7f60ULL,0xff96ff6f9bf67d61ULL,0xcfc7ff8b9fee7f72ULL,0x6ffedfdfbeafdf63ULL,0x76fcfeee9fd7df62ULL,0x9d78df7aaa8bef63ULL,0xdfd1bf3f13de7e73ULL,0xe7e6ef3a97d6ed70ULL,0xbfffff5b9bfeef71ULL,0xf7ffffef9fdfff60ULL,0xf18cf7fd07df5f61ULL,0xff3e29528b965771ULL,0xcfd7ff2b87d6dd72ULL,0xbb7edfdf37d75f63ULL,0x3feeeffb15c67f61ULL,0xd7ecbe793bde6f62ULL,0x782e7fe413de4f63ULL,0x8bf4ff7e3fd6ff63ULL,0xefe7a1c51ef84d61ULL,0x30aa979b2b02ef61ULL,0x8410780180160960ULL,0xb7383d6807065541ULL,0x854c20020404961ULL,0x202221200004061ULL,0x7bbec98617c27673ULL,0xcd173eb4939a5960ULL,0x83a2bd409a8ecc11ULL,0x4100008182300140ULL,0x8102a00001120340ULL,0x20004604000900ULL,0xc0501e8002122540ULL,0x30c276000970920ULL,0x88a42d2a03868b20ULL,0xebc47606b7875863ULL,0xa1e164183842920ULL,0x8090560881021d40ULL,0xfdffeffe37fe7f73ULL,0x92a00d0200960d01ULL,0x800c20020000060ULL,0x810601000e4421ULL,0x1322040801124120ULL,0x4000001000010800ULL,0x80010020000020ULL,0x88818a0000084861ULL,0x800000400000900ULL,0xb90205128b86a400ULL,0x91c0b62c02085473ULL,0x40000404000ULL,0x42001020000ULL,0x200010800000ULL,0x0,0x4876161001845822ULL,0x280000010800000ULL,0x100000100000900ULL,0x0,0x1200000900ULL,0x0,0x24000000020020ULL,0xfa9a6f3e17df6d61ULL,0xc7d6fabe3ec66d60ULL,0x63f49fd812feef70ULL,0x343c57fb07962d60ULL,0x99f6ee6e96d67d62ULL,0xcaba2d5204725d63ULL,0xc924c30f0d9a0f60ULL,0x82ae3ecb15865f70ULL,0x360259c0978a4a61ULL,0x15142e1887135940ULL,0xbb040400061a0960ULL,0x69d2dd9ab3c26762ULL,0xea52125004921940ULL,0xc97c588074e4921ULL,0xb4a8ba7d1fc2df71ULL,0x7f7e7f7393beff63ULL,0xfbd6f36eb7ae7f62ULL,0xeeea3a19a2da4f73ULL,0x6ab8cf5b8e2e6d72ULL,0x4503268829a2f60ULL,0xc3bfaf709fd4fd71ULL,0xc59c2ffe8e964f61ULL,0x4fcbee1b2b8f3b60ULL,0xb01d5603828641ULL,0xf455e6a812ee6760ULL,0x91aae7c396cefd60ULL,0xfb1067203ed26f61ULL,0xf5ce162824067961ULL,0xa33e5f5e87cf4f61ULL,0x7fff7fedbfdf6f73ULL,0xf2eaa7608e865f61ULL,0xceb265eb1fd67f60ULL,0x8184f3ff0f071d60ULL,0x8564a0402165d20ULL,0x400200010000100ULL,0x5c44f4548fc66740ULL,0xc104c22021a0d40ULL,0x8b00830a03842660ULL,0xace50b6816864d60ULL,0xbee00c200ac66141ULL,0xd4f4f46b1fbeef60ULL,0x4a3e0d4aa08b4f21ULL,0x838c420706002d40ULL,0xe0a220b80864c40ULL,0x1e945108140a0d60ULL,0x74beb6eb17d76f62ULL,0x9710ef40178ecd60ULL,0xfb967bebae56fd71ULL,0xefeffebbbfdefd73ULL,0x30abd9d21fed6d71ULL,0xf5f7f6e9be5e7d71ULL,0x9d5c8f78868e6f63ULL,0xf6943effb6567d61ULL,0xefdfef2c1feeef60ULL,0x2f597fdb07d35f63ULL,0xb4b25caf86caef60ULL,0x516cbe0c178ecf61ULL,0xea3e3c6e00a64962ULL,0xcc646008a50e500ULL,0x888e9eda1fc6d463ULL,0xb0bb6ca011035f60ULL,0x6480e010900ULL,0x5c462d6a1e824b22ULL,0x4380c31283c20c40ULL,0x2050129703825562ULL,0x72180380040a4061ULL,0x418b32001921550ULL,0x802004202084001ULL,0x81010802042020ULL,0x66026000b884040ULL,0xb3c4d0806c26d40ULL,0x8010364101a06c41ULL,0x1218400001025400ULL,0x8024214305025900ULL,0x47ca3c1680c64f03ULL,0x410000000002800ULL,0x1484eae010814d70ULL,0x8500240001021140ULL,0xf797a7ee1bce5773ULL,0x9846c2140742d140ULL,0x8008000404000ULL,0x120000000020040ULL,0x8020044000064402ULL,0x20004000004001ULL,0x0,0x160064803022920ULL,0x420660000820860ULL,0xf120060004000d40ULL,0x22187c140a5e6922ULL },
	{ 0xcb7af3b6d79c7d33ULL,0x67ef1fd9f6bc3de0ULL,0xffffeffbff9c7fe3ULL,0xf77efefef7bc77f1ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40ae284890144380ULL,0xbbfdef36f69c6cf8ULL,0xe7cbdbfb779e7ffbULL,0x7b9ccfbbf79c6ff2ULL,0x73ffffffffbc7fe8ULL,0x753af563d7947fc8ULL,0x5000c00003300420ULL,0xdfdffebfffbe7febULL,0x8fdfdfdbff307d2aULL,0x74ffffed7fbc7efbULL,0xf9fef77bffbc7fe0ULL,0xffb4ff7edfbc7fe0ULL,0xdfb7bb9e97b43fe0ULL,0x7bffd7d9f5be7de0ULL,0xd4be577e7e3c6c22ULL,0xf9eefffedfbc7fe9ULL,0xdfa73f71ff3c55e8ULL,0xdbfffbbaf7be7ff0ULL,0xfbfffffff7be7fe0ULL,0x77b7dfeefebe7df8ULL,0xd1e8bfeff5bc6ff8ULL,0xeffeffffffbc7da0ULL,0xebeef3eefebe7ffbULL,0xeeffffdaf7be7df0ULL,0xbfbfefeaf6bc7fe1ULL,0xfffffe7dffbc7ffaULL,0xdffe7dfbdfbc7feaULL,0xfbdff73fdebc6ff8ULL,0x6fcbdfdfd7be7fe0ULL,0xfefecffbdfbc7ef8ULL,0xf5f7f7fdd7be7feaULL,0xfabefd72d7bc7fe8ULL,0xffbffffeffbe3fe3ULL,0xdfcffefff7b67fe3ULL,0xfbfedfdbffbe7ffbULL,0xf6fffffff7be7febULL,0xffffff7bffbe7dfbULL,0xffbefffeffbe7ffaULL,0xefffffbfffbe7feaULL,0xebffdfdfffbe7ffaULL,0xbfffdfffffbe6ff9ULL,0xffffff7bdfbc7feaULL,0xffbf3fffffbe7ffbULL,0xeff7ffffffbe7ffbULL,0xffffffffffbc7febULL,0xffbfeffbf7be7deaULL,0x9c6df76fd6947fe8ULL,0xd2563c32d4982721ULL,0x6b86eb269b1c5520ULL,0x215e9749d4986f60ULL,0xb5ffcffbffbc7cf0ULL,0xfdffd6fff7bc7ffaULL,0xd3beffebffb67fcaULL,0xc394e3bed7bc7d68ULL,0xffeb5fdfffbc7febULL,0xfffffffbffbe7ffbULL,0xf5ffffffffbe7ffbULL,0xdcfefd52d6b46ce2ULL,0xebadf7fade9c6fe2ULL,0xc9defe13f7bc6fc8ULL,0xbbdfdddddfbe7779ULL,0xf6b36feef7bc6fe9ULL,0xbbea7771d69c6d6aULL,0xdf2467ecd79677c8ULL,0x4ec5ff96f6946de3ULL,0xab767e10d79c2c60ULL,0x9c57feebbfb47760ULL,0xe1ff3e72d2bc7ee1ULL,0xd6b63d77df3c7661ULL,0xeff7f782f7bc3debULL,0xee9ffe9bd7bc6df0ULL,0x8dbadff2dd1c3de8ULL,0xf5eff7f7f7be7de9ULL,0xdfb67fdfffb46fe2ULL,0xebf7f73ff7be3fe0ULL,0xcef67afef3b42deaULL,0x9737dff7d79c7d68ULL,0xf843b7f1d4243f73ULL,0xd5ee3f78f69c7f20ULL,0xd3deff3bdfbc7fa8ULL,0xeae7bfdef7b67df0ULL,0xe69fdffbfebc7de1ULL,0xf5f6feffff9c7fe1ULL,0xdd9efdf1f7b67ff8ULL,0xfb9eef7bf7962fe8ULL,0xcfcfff9fdbbe7efbULL,0xab76ccded61c3de9ULL,0xdafe7ed6f7343ceaULL,0x996f7f71f4167fe9ULL,0xfba435b4761474b3ULL,0xf1f7dfdad7bc7d62ULL,0xaffdff57ff3e3ff9ULL,0xbdbffffbdfbe7fe1ULL,0xffffff6bdfbc6fe1ULL,0xf8fe77fbdebc7de9ULL,0xceffff2fd63c65fbULL,0x6affffcdd7be35e3ULL,0xf43b57c9d3bc55e8ULL,0xd3edde34d09c7ef0ULL,0xbb8e7f61d53c5fc2ULL,0xcbf6f1fbd7943d69ULL,0x61d33ec1f6842f68ULL,0x19bad75ad61e3de8ULL,0xf497b6eb93bc6d60ULL,0x108aad4074220420ULL,0x8b246165d21c6002ULL,0xb9000e0554040460ULL,0xeffdc5eaf63e4f60ULL,0x3090802014001460ULL,0x6a02c4284b42408ULL,0xb8002f1c56107cc0ULL,0x3b80008000468ULL,0xa480140d4903c40ULL,0x34a646a8941c1640ULL,0x99af7360d49c3dc9ULL,0x30842760908428c8ULL,0xcfc2a71d93342de8ULL,0x4be89d43d2146dc2ULL,0x5ef1d289ab42d68ULL,0x1000700240100420ULL,0xd8242d61c6945580ULL,0x4800c304020c1480ULL,0x21bb64b57922c60ULL,0x12104820c0044400ULL,0x200608003c0400ULL,0x74b23de4d6a421daULL,0x28b6890840168ULL,0x8523080542401c0ULL,0x80c80a02982400ULL,0x0,0x4200240050002050ULL,0x5000c02844843000ULL,0x800000100100000ULL,0x0,0x400a044851c0000ULL,0xc800000002003400ULL,0x81000080000100ULL,0x0,0x0,0x1002000c0000ULL,0x0,0x844070002882400ULL,0x46208003141400ULL,0x26999dd0d7160720ULL,0x90507256040488ULL,0x8100002040020408ULL,0x0,0x40078e0250182c80ULL,0x80011d080a42400ULL,0x92284ce3de3c26c0ULL,0xc575b67beebc2ca2ULL,0x4116103041900458ULL,0x2005c3e0403454c8ULL,0x3570161093302841ULL,0x8f988809d3043ca2ULL,0x883e222002282c00ULL,0xdfbe3d79d1bc7de8ULL,0x93decb32d4b43ec1ULL,0x7cde3e5fd7be3d73ULL,0x6b1fdfd7ff3e7ce8ULL,0x249e3a28d0346e40ULL,0xcfbe7d71f1146df8ULL,0xd78ddb79d6bc35e1ULL,0x87c4a20ad5b82d20ULL,0x94e05cb5e0e3ce0ULL,0xbe75def9f73c3fe8ULL,0xa1cbff68f7bc6de8ULL,0x3024770dc103f09ULL,0xc4a8d99042802c48ULL,0x8bdeff57d7b66df0ULL,0x24bfffebf73c77e2ULL,0xf1e6aef8f5bc2f63ULL,0xde9eaf7effbc7fe1ULL,0xe9f6d71af69e77caULL,0x1c011a4882942500ULL,0x16317c42d68257c0ULL,0x5045f5b9d53c36c0ULL,0x82c67d6092987700ULL,0x8845d260f02c1500ULL,0x888180104302401ULL,0x34f54ce4fa2c3da0ULL,0xf5f6f66efebc3dc3ULL,0xba7aad709e3c5728ULL,0x820043104c800440ULL,0xcf6ab61af21e6709ULL,0x1884c518500c2c61ULL,0x747376eae1bc7cd8ULL,0xdd801a22838c3401ULL,0xfaa26b7ecf363c82ULL,0xefdfff73f7bc2febULL,0x6a7ece53dfbc7df8ULL,0x56d4def0f4bc6d48ULL,0x596edff9d7be6d63ULL,0xffd67e67d53c2fa0ULL,0xefc7cf7576b43cf1ULL,0xab4cdfd176bc65c1ULL,0x8b725e6bd6bc7ea0ULL,0xd158d7f2f4ba2fc0ULL,0x9284020002140480ULL,0x8584a001448c00c0ULL,0x24a1012f4900040ULL,0x412681486300440ULL,0x92501600d4900062ULL,0xfee2e548f33427c3ULL,0xc1f7d37bd51e7de1ULL,0x2044084812002800ULL,0x18aa4d5bf69474c0ULL,0x413042e844822840ULL,0x504152091002c00ULL,0x20410000020ULL,0x40020008000c0400ULL,0xc9c4e20d6903440ULL,0x1104bc291b42440ULL,0x8e82244016182500ULL,0x4808208280144080ULL,0xe864021c14b01440ULL,0x98880490904100ULL,0x40447ea0168c2460ULL,0x9b40042050040462ULL,0xa210230470902c10ULL,0x440e00080340400ULL,0x10400040000002ULL,0x3808010113000500ULL,0x6004100000ULL,0x8057164100400ULL,0x81c6054b92241560ULL,0x6010904a502005c8ULL,0x108650d0100810ULL,0x140152894040860ULL,0xb2407d4286b43441ULL },
	{ 0x41a3cea2ab88233bULL,0xecdf9edb8b9f2f7bULL,0xcefeefeb9bbc2b7fULL,0xdcfffe7f9b9d2f1fULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x76bc35f88bf52a5fULL,0xfbdef77e9bfb2fffULL,0xa4621ffacaa32317ULL,0x3bbde7d99ba2232cULL,0xfcffffffdbfe2bbfULL,0xfffffdf3cbff2f7fULL,0x548422f90bbc0bb9ULL,0xdeefffbfbbfe27ffULL,0xbbbfffdfbbfd2fbfULL,0xf7bfdfeabbff273fULL,0xbd7fff76bbfe2fffULL,0xdfb6ffffeb9c2bffULL,0xdeffbf9fcbff2fbfULL,0xefdffedbcbdf239fULL,0x952797f78bf023aeULL,0xf8ffff7aabdc27fbULL,0xdaaebf67fbfb2796ULL,0xedffff779bdb2f7fULL,0xefffffdfebfe2f3fULL,0xbfb4fffb9bec2bffULL,0xf7fbf76dfbdc273eULL,0xe53a6d7dfbcd2f4eULL,0xcbeff77ffbfd277fULL,0xffff3f47fbdc275fULL,0x3fa9cff39ba72bf7ULL,0xf77cff7dbbfe23ffULL,0xdffe7ffacbfe2f3fULL,0xf9f5ff778bda2337ULL,0xffdfbf7bfbbf2fbfULL,0xffffcffbdbfa27bfULL,0xfff7ffefdbf82f3fULL,0xffbeff93dbd927ffULL,0xfbbceffffbab2f7fULL,0xeffffffffbda2bffULL,0xbbffdddedbfc277fULL,0xf4f7f6eebbff2f7fULL,0xfbffff7bbb9b2f7fULL,0xffbf7ffffbff2bdfULL,0xffffffbe9b9e2f77ULL,0x6fffffdffbfe2ffeULL,0xdfbffffbdbba27ffULL,0xff7efffffbff2fbfULL,0xfffffffffbfe27ffULL,0xefffffffcbfe2f7fULL,0xffffdfdfbbff27ffULL,0x3ffbffdfabbf279eULL,0xd3cdfdff833c2b66ULL,0xc5ca3b4093b223a7ULL,0x89bcc2061bc027c5ULL,0xe3523a6213182744ULL,0x2dffd7fb9bb82327ULL,0xf57ffefefbb627dfULL,0xcf367df39bd0277fULL,0xab92f76e9bba232fULL,0xfbfbb7dfbbfd2fffULL,0xfffffffffbff2fffULL,0xfdffffffbbff2fffULL,0xdfbebff3fbba2f5fULL,0xfb5e7ffe8bff279bULL,0x5ffebf13bbe6272fULL,0x1acccddbdbca23bfULL,0x76ddf5e9cbc1271eULL,0xb94fef62b39023ceULL,0x7c3437f40bc2226dULL,0x8943efa98b70274eULL,0x2fdadfd99bdd2756ULL,0xbdb77f62a990233fULL,0xff6eff79cbe82384ULL,0xdea0af74dbac2717ULL,0xc9d4f7baab9c23b7ULL,0x3afefbdbbb9b2b1fULL,0xd7bb7ffbdbdc23efULL,0xd9f5f7fcbbf42a1fULL,0xdbdebb74cbe2270fULL,0xc9d3fb6e8bba2f3fULL,0x2c923f088b58230aULL,0xd7facdd1cb9d275eULL,0xfdf5ff7f9bbe2e7eULL,0xdfbe7fe28b9f235fULL,0xabdcef689bf22f1eULL,0xfbf7ffffcbdc23afULL,0xff5f7f2b9bba27bfULL,0xddfff7eefbb423bfULL,0x9ffefd67ab73272fULL,0xcd9effefbbc9272fULL,0xcfcfff1fdbdc2f7fULL,0xafdeff5ebb8e23ffULL,0xf4f7deee8b9f273fULL,0x11c2a57ac3a8231eULL,0xff9e3f7d9bd8273fULL,0xeffbff188b9f277fULL,0xebf6dfdbbbfe2b7fULL,0xb6fbfffffbf623dfULL,0xeb7fef76abfa2737ULL,0xfebe2fffdbff2ffeULL,0xcbf7ff9eeb9b235fULL,0x6b5ffbfb8bd602dfULL,0xb79fcfeaebba274bULL,0xd57ca4d08bd92f0bULL,0x4e963737abc4221eULL,0xe392fbbd8bce2fafULL,0xdd6f5ed6eb982776ULL,0x9b9a6ffa8b9627eaULL,0xb716be42ab8022b6ULL,0x200012082000204ULL,0x8af1af7e9bbc223cULL,0x80426c5151922225ULL,0x3ebd55e1cbd02746ULL,0x34b7f6e883a02f7fULL,0xd9aa2c508ba82218ULL,0x56ac20738b8822aaULL,0xb00880b83822402ULL,0xa9d40c008bc00047ULL,0xe4b65ca8ab1a020aULL,0xf96cee7b8bb42b5eULL,0x8b0cab3083162608ULL,0x23c8e7328b4426c2ULL,0x810cd409114264cULL,0xf216ea9b3e230dULL,0x4004001009800001ULL,0x9c964d25bb84035eULL,0xc0802926a104232aULL,0x1800030002000004ULL,0x508205a003360200ULL,0xb6040200138302a6ULL,0x0,0x2002000000104ULL,0x60230009000030cULL,0x800804008080000aULL,0x1000200580800041ULL,0xc4001009000000cULL,0x0,0x0,0x0,0x8060022002802208ULL,0x1082c0182820041ULL,0x8800631003802028ULL,0x200200800002ULL,0x0,0x0,0x0,0x80004000002001ULL,0x2200000002001ULL,0xa31ccdc4aa800345ULL,0x100030a8400000ULL,0x100040080000008ULL,0xa0008000200ULL,0x628882802020ULL,0x1e04050200220cULL,0xbf395ffab2852f9eULL,0x80500082042600ULL,0x20004001100000ULL,0x8bf9da080b802fa2ULL,0xee7f9bdbabee2bf3ULL,0xcf9fcffd8bfe2ffaULL,0x9220204082802106ULL,0xe66225608292221cULL,0xfbd0eb348b992feeULL,0xcffab6f39ab62715ULL,0xe7be5ff7fbf0263eULL,0xd567f2ee8bd12abbULL,0xfffefd72ebd82e3fULL,0xd98837a68b92233eULL,0x2ae22b2800306ULL,0x39f5d12c2a0270aULL,0x74f85a68cb9a2797ULL,0xdbe677f2cb98232fULL,0xcbc4277389f8236dULL,0x96c3ae128b8023e6ULL,0xa37f5edacb8c233fULL,0x94bb587b9b9a267fULL,0xd5a4a7739b182754ULL,0xf7a62ff6dbb2271fULL,0x7d6ab33a3962331ULL,0xaad7fedf8bd42f1fULL,0x241850460ac02503ULL,0xf101d76e1b94220fULL,0x50438208310200cULL,0xa836422603902223ULL,0x46823e1083802113ULL,0x80800002800000ULL,0xc104d06a8b8c035dULL,0x621a55728b8023ddULL,0x810020802810002ULL,0x48519a100195020aULL,0xae29482803b02a44ULL,0x9656035180222eULL,0x1220d600b182500ULL,0xf1a0be1e0b96226fULL,0xcdc6be1dc3be221eULL,0x8abf17d3abd22795ULL,0x9dbf7fea9bb6227fULL,0x93cb7f63cbd0232fULL,0x83346ce482a82b16ULL,0xcffbffbefbfd2fffULL,0xbfdfffdf8bba236fULL,0x9d9efbfbcbcd2b46ULL,0xf9bf72688b8827b7ULL,0xc220045282800319ULL,0x8947a61492942206ULL,0x29c31c8a0bc0030fULL,0x2004d4a09080328ULL,0x1044aec88a882220ULL,0x648a7de19b89225eULL,0xe1bffaf78ab82f3eULL,0xec562e3389c82f73ULL,0x511ccf138bd8272fULL,0x1082068481022206ULL,0x4142010100a04ULL,0x3084400400880290ULL,0x8801a0013000006ULL,0xc11c0f7082802385ULL,0x90466888410a0aULL,0x8080244009802200ULL,0x400024882002608ULL,0x4c49b01203902228ULL,0xc05448a400300ULL,0x400000480000002ULL,0x2944932083a0242fULL,0x800003440201000eULL,0x4000021000800a02ULL,0x10054001800002ULL,0x0,0x2482000280000cULL,0xc89805568b802329ULL,0x4000440801002308ULL,0x8452080009080008ULL,0x40144600101034aULL,0xd1f657209392033fULL,0xb2166765cacd2306ULL },
	{ 0xc3f1f6a3d914d7aeULL,0xeef7bedfeb99f39eULL,0x7bb95d72cb94532eULL,0xd43d77636b0cc1eaULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfebe1ff9eb35dbbeULL,0xdb56ff3efb2dd1faULL,0xffeafaf6fbace3aeULL,0xafbfdffffbbcf7faULL,0xfdf7fefeebb5dbfeULL,0xffffffe3fb3cffeeULL,0xfefe3f6febb5d72eULL,0xffffbefffbbdfbbeULL,0xeffedfdbdbb5fffeULL,0xfcfffeffebbcd3feULL,0xed76ff6bebbdd7aaULL,0xef95bee4ebb4d33eULL,0xdfcfffbddbbdfffeULL,0xfbffffdbfbb4fffeULL,0xdcfe7dfaeb35d5beULL,0x99cdff49fb145faaULL,0xbab2afcef9b4f7feULL,0xcfffffbffb3cd3baULL,0xfdffbff7fbbcfbfeULL,0x1dbafffbfb94d7feULL,0xdd7e64fbfb05d7beULL,0xf79effeed38df9feULL,0xdbffff7fcbbdd7feULL,0x6fcb3e7deab9f3aeULL,0xbebffffff835f7beULL,0xfdfdfeffebb9db7eULL,0x5fee7de0e325d37eULL,0xfbf4f7bfe9bdddfeULL,0xeecfbffbfba4f7eeULL,0xfbbeeffffbb5f7feULL,0xfff6feeef3b4d72eULL,0xcfbebf72e335d7eaULL,0xf795fb7fd98df7aeULL,0x5ffbffbffb85d9feULL,0xffffdfdffbbddffeULL,0xf6bffffaebbdf7feULL,0xdfefbfffdbadd3eeULL,0xfbbeffddfbb5fbfaULL,0xaff7ffbbf3b5dfbeULL,0xabfedfdffbbdf3eeULL,0xbef7ffffebb5dd2eULL,0xfd7fe7ffebbdfffeULL,0xfebfbf7feb35d7feULL,0xeffffffffbbdfffeULL,0xefff7feffbbdf7baULL,0xb7be7ffbfb34576eULL,0xf13d7ef9d335d33aULL,0x7ff43c7183a4d28eULL,0xc954d3b7c315d7aeULL,0x88c2124dc11450aaULL,0xbebdffdbeb09dfbeULL,0xfff7bffff33cfb3eULL,0x6bfe7d7ad98c7d2aULL,0xb381ebb6eb84d33aULL,0xeef73ffff934d33eULL,0xffbfeffbfbbcdfbeULL,0xf5f7ffeffbbdf7beULL,0xbbbfadf0518cd32eULL,0xb6acffeee981f5faULL,0xe7d7fe19f30dd3aeULL,0x3a96cb9feb24f7aaULL,0xf473f664dba4f33eULL,0xe5fdaf79e33dc3eeULL,0xfb99cf7adb24db2aULL,0xd9dfdb2cfa31d31eULL,0x2bd3f95beb04d18eULL,0xbcbf7feafb3c539eULL,0xd33e7fe0fb24e12aULL,0xdef57ff6cba5d7beULL,0xefc7f71ffb1dd3feULL,0xe9ffdfdeeb24d1baULL,0xdcbbfff2fb3c532eULL,0xf5edfffff9b4df3eULL,0xf6b67dffebb4d3beULL,0xebfefafffbadd72aULL,0xb5c21200d900d12aULL,0x9e9556ece891d7baULL,0xfdf7f7fbfb35fbbeULL,0x7fbe3e73eb21d3beULL,0xfbdcfb7ef984dfeeULL,0xafd7bffbeb39dfeeULL,0x8bbe7fffeb98d39eULL,0x35b5ffebfb3cf33aULL,0xdbeeff71d334db3eULL,0xdff6fff6f335f7beULL,0xd9cffe13fbbdd33eULL,0x631edd1beb25d5aaULL,0xb4f65cedeb84f96aULL,0x592cf846831079f2ULL,0xcbb26e6fdb00d19eULL,0xf6d2fe1ad114d1baULL,0x6f7ddcdfe99cd72eULL,0xbdf75ff7f0b5f7beULL,0xfbbce7f1e935f37eULL,0xdea62fedfb05fb2eULL,0xedfff7bcfb34ffaeULL,0xf9ff9fd1d3a4c7feULL,0xbcbe4dfbfb94d7aeULL,0xc77cbf77f325f32cULL,0x8af6bdfce9b4c39eULL,0x83ba61026801f326ULL,0xc5db3673e985e5b6ULL,0xdc0def8a31447aeULL,0xd3208361eb20c13aULL,0x468803a0811481b2ULL,0xd819eb2ad398c3a6ULL,0x48c136845005c1a2ULL,0x820a0010802c4120ULL,0x35750e568830539eULL,0x87340f608a204206ULL,0xd4204042c104d000ULL,0x4880021482004086ULL,0x2082010ed100d008ULL,0x3454a880148b0eULL,0xd3feefa0fb84f18eULL,0xfb363f76fb84d18eULL,0x8188c6035a80c10aULL,0x4f52cb190b28d39aULL,0x301203214900e102ULL,0x9108b645820c412eULL,0xf3106b3c8badd18aULL,0xa8263620b044206ULL,0x610016d089044096ULL,0x4ab4cee4384c18aULL,0x94450240e208410eULL,0x200000400006000ULL,0x0,0x0,0x200081428080c006ULL,0x24020004000ULL,0x4000008008004000ULL,0x102010440108200ULL,0x0,0x420080000004ULL,0x2840004000ULL,0x80020040c0085004ULL,0x100042001104200ULL,0x0,0x0,0x4000200040002ULL,0x0,0x84050000000ULL,0x10100000004080ULL,0x208050040004008ULL,0x64004124000813aULL,0x0,0xa00000000008020ULL,0x908430243045008ULL,0x6000000000aULL,0x1100465080001102ULL,0x604452026000052aULL,0x0,0xc00048005000ULL,0x2d4793c0c820f7a4ULL,0x1e9eccdaeb3453f6ULL,0x446cf2504100c532ULL,0x21c64e204304506aULL,0xebd4c766db19d3a6ULL,0xaccfbedfcb84d3aaULL,0x26fedf737b15f3faULL,0x3af637eeeb95d36aULL,0xd3b6b573e185f1b2ULL,0xee922546d3b0f21aULL,0xc153fb224300c3eeULL,0xeeffd7deeba4f3baULL,0x249c46686b24c20aULL,0xd1ac06685100e120ULL,0x1ab4662ec8a8e9a2ULL,0xd5eb5e3ceb30d1beULL,0x7b5ed3dac8b9d37aULL,0x1c826dc47300e20eULL,0x6020a731e101c20cULL,0xf2b62675e105d5baULL,0xa1084205c104d32cULL,0x2f5f5249c314d06aULL,0x3120050201006160ULL,0x91ded475c025933aULL,0x6e847463e3a4d1aaULL,0xe8e5d7aafb10dbe6ULL,0x44636885301c000ULL,0xb0804468a9044086ULL,0x842622f18930c30aULL,0x40262c20d820412eULL,0xd394ef4e5335d386ULL,0xab425a8373045726ULL,0x9e01eb48e924e10aULL,0x414124684924c32aULL,0xda3844116110e220ULL,0xfb147e3ac810777eULL,0x4fff3eb3db2cdf7aULL,0xeb7a9fcbfbb5f36aULL,0xf0395ee8c9a5d32eULL,0xd6e276ad325d7b6ULL,0xde2e7b74c820533eULL,0xcddfffcdebbdfffeULL,0xecfed94bf9b0fb8eULL,0xb4f752c361b5d7a4ULL,0x116ae568c994c366ULL,0x363c3ffdc8a1d17aULL,0x8020d081888c4226ULL,0x66feda5acb28490aULL,0x1104350890c5800ULL,0x110e3502c2a1c020ULL,0xa3800544f0218204ULL,0xc9f4d2a2e924d18aULL,0x66953e2c5984c1a0ULL,0xaa98ef528a30572eULL,0xa42320604a04802aULL,0x560c20724100c13aULL,0x91840326c900d326ULL,0xd9e9b710e81dc31eULL,0x820e07005000011aULL,0x810000040008000ULL,0xc0ba1c00d130c000ULL,0x88240400010041a8ULL,0xdd0b902cb04f206ULL,0x30243008000536eULL,0x1018526cc110d106ULL,0x20e4468d0040464ULL,0x80050401006080ULL,0x0,0x200000800004008ULL,0x100744a68135c30cULL,0xd94c34e9db05912aULL,0x97a5096ae9a0c00aULL,0x8800020100040122ULL,0x2c7e9ac85200d12aULL,0x37c347915304c18aULL,0xdd94f6035b90c3a8ULL,0xa122f546905d38aULL },
	{ 0x8980d2a6085291a0ULL,0x25e6df418b0fb360ULL,0x2c304ff082d6b260ULL,0xddffba5ecbd7b3e4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffe7ffbcbdff3feULL,0xfb95f73e82dff7eaULL,0xef6bfffd8bdff37aULL,0xefffdff78bdff7eeULL,0x9ef7f7eb8bd6f77cULL,0xfbbeff758bdeffeaULL,0xdbbf637f8bd7ff2eULL,0xdfe7febf8bdffeeeULL,0xbbfedfdb8bdff3e2ULL,0xf6b3ff6b8bd6fb7eULL,0xbdffff7f8bd7f3faULL,0xdeb5ef7ecb5ff7eeULL,0xfff7fbbecbdffbeeULL,0x2fdad7918b56fbe8ULL,0xb4ef77fe8ad6bbaaULL,0xf5fdef728bd7e77aULL,0xdbbfb77e8bd6f7faULL,0xeffff77fcadff7feULL,0xefff7f6f8adef3eeULL,0xd69befff8ac7febaULL,0xf9ffbff7cadfffeeULL,0xffbc7ffe8ad7fb6aULL,0x89e6ff3f89dffe4aULL,0xef55375b8253fea0ULL,0x67fb4ff3ca5ef334ULL,0xf5fff77e8bdff3feULL,0xdfde7ff382d6fbeaULL,0xfbbdffbf4bdef2eeULL,0xefdbbedbcacebfaeULL,0xffbfdff689dff3eeULL,0xb5ff2ff282dfb76aULL,0xdfeaa561c2defadeULL,0xfdffff6b8bdffffaULL,0xffefbe9bcbdff3feULL,0xffffdfdf8bdffffeULL,0xfffff6efcbdffffeULL,0xffffff7fcbdeffeeULL,0xffbffffecbdef7eeULL,0xffffffffcbdffffeULL,0x6ffeffdf8bdff3eeULL,0xbfffdffbcadffbeeULL,0xfdfde7ff8bdffbeaULL,0xfebf3f3fcbdffbeeULL,0xeff6ffffcadefbeeULL,0xffe75ffbcbd6ffeeULL,0xffb9fffe8adef7faULL,0xf7efbffc8bd7f772ULL,0x77b277f4894ef36aULL,0x8ba3efbd8ac7f36aULL,0xef7ade7b89c6b3e4ULL,0x5dbf5fe38adff3c2ULL,0xfef7fffb8bdefff8ULL,0xdfee7f628bdfbfe2ULL,0x88f5eb7acb47fbe6ULL,0xcfc3db7a8bcef3e2ULL,0x67ffdfb3c35ff7f6ULL,0xfefcd6ff8bd6ff3eULL,0xb3d7b7d04347f75eULL,0xfbbe6f6f8bdefffaULL,0xcf5bb8128b8ff2d2ULL,0x2e5fdfcb0bcfd3daULL,0xd42df6f68a5ef36aULL,0xdd7fbfeb8bd7f7eeULL,0xffbc7fbc83d7bb66ULL,0xcddff7b689cff7e0ULL,0xabdd57dac9de3aeeULL,0xb5f9ff6a8b5ef7eeULL,0x9f6efe7e88deb3e2ULL,0xff9e6f768bd7ba6eULL,0xeffffe7f8ad7fff6ULL,0xeefedf57cbd6fbe6ULL,0x9ffb6effc1def3feULL,0xf5fff7facbdefbfeULL,0xfbe6fde18ac6f2faULL,0xc1bafeea8246f352ULL,0x2ff7dad9cbcdf7eeULL,0x3b9ecfeeca5ef6feULL,0xfffdf77dcbdffefeULL,0x7bb61f638b5ef3e2ULL,0xe9c6e73f8ad7f2f2ULL,0xeef3ffdf8ad7f3e0ULL,0xefbfefe3cadfffeeULL,0x7dff3fefcbd7fbe6ULL,0x9ffefd61cadfb3f4ULL,0xfb2e6ee78ad7fbd6ULL,0xe6d7f223cbc7fb76ULL,0xffffdfdecbc7fb76ULL,0x9cfe7ffc825fb3eaULL,0xcd1e95008ad79340ULL,0xdfb667d68b46f2a2ULL,0xdac3fa1cc3c7e3f6ULL,0x2afedfde8bdef3aaULL,0x921617e28bccb74aULL,0xdd6ce66e8ad5fb30ULL,0xdeae36768a46feeaULL,0xcdfcf33acbc6bbecULL,0xa9fddfdf0bd7a362ULL,0x3eb84ff889d6f3eeULL,0x9cccb57ac2c7a154ULL,0xcfb02d5c8ad7f7e8ULL,0xebd6f32f88c7f6e2ULL,0xaf5236178a86a350ULL,0x1309479e88c65300ULL,0x36e3f67749dfb306ULL,0x1aa2144002d68222ULL,0xae1a6260047826aULL,0x8d0128082845a00ULL,0x6c9645240a538360ULL,0x800551e100503002ULL,0xfbf2bf718ad7bbf2ULL,0xffbefffecadbfbe6ULL,0xa97e961b8adfb702ULL,0xab5ccbce88d7bb72ULL,0x5010be5680d0f0a0ULL,0xdca4aff088d6d226ULL,0xba943f0580d7f6b4ULL,0xc92687108082f920ULL,0x1fb5cfdf89d6d36eULL,0x302a580080462240ULL,0x5222bf408202e3e2ULL,0x8800010081001200ULL,0x81409c218854300aULL,0x241811c101540250ULL,0xf7bbffee49d7b3feULL,0x412a89508006f062ULL,0x200000008020ULL,0xc00000008020ULL,0x0,0x0,0x911400800800f380ULL,0x0,0xc00020200ULL,0x0,0x480000440000ULL,0x4002025002060212ULL,0x8000000000020002ULL,0x4000001000081000ULL,0x0,0x0,0x0,0x810000018000ULL,0x0,0x4000000400005000ULL,0x50000208c00200ULL,0x420204a01828202ULL,0x20024002042200ULL,0x4100001000200ULL,0x8004000400028220ULL,0x40001008000200ULL,0xc90809230887f6a2ULL,0x4020846002089060ULL,0x0,0x20300008ad100ULL,0xac60b4880c7a390ULL,0x51c087081524242ULL,0x28646000460202ULL,0xfbea7d708ad7f7a4ULL,0xc397ff7d0adfb7aaULL,0xcee3ba2182d6726aULL,0xffaecffb8bd7b67aULL,0xf5d49aeac8c4f7e4ULL,0xd95efd728bddf3b0ULL,0x9894ffa281dbd308ULL,0xcfcbba1f8adffaeeULL,0x337ddfd2cadff36cULL,0x80c8d6600a86a2f0ULL,0x8a3ce4628354a320ULL,0x7ae02f2780d6d360ULL,0xee43e78883cbfb8aULL,0x834830db88d6f0a0ULL,0x38f71ead8a97b626ULL,0xd1ef174a0b563372ULL,0xd393372ecadebfeeULL,0x854df09480d0b782ULL,0x2e5e781b8817f3a0ULL,0x9e3347b38bc2faa2ULL,0x80e53e6a084db340ULL,0x128629c000c0864aULL,0x6ac6610b8842f22eULL,0x25c2da08c8879204ULL,0xd2b80d2000568240ULL,0x9974f6cb8b80fba0ULL,0x2ba3f0088d6a26aULL,0x8ad4e228084e1042ULL,0xc08008008500200ULL,0x7bac471789c7e260ULL,0x842650610b529240ULL,0x58006c41800ab380ULL,0xfa0269128206bbc0ULL,0x4fc3fa1b8ad3f762ULL,0x1bf9df5a8adffbeaULL,0x34254dea8a9fd2a2ULL,0x416ebf6289d7f6e2ULL,0xf19077758bdcb720ULL,0xffcef7eecb56bfe6ULL,0xeaf6dedfcbdebe76ULL,0x75937fee82c7b37aULL,0x43ba7778824ff3c2ULL,0x80004c500042a120ULL,0x810020030808d020ULL,0x888380900028322ULL,0x802247520a54b222ULL,0x802193088a05a300ULL,0x852005410a023220ULL,0xc879872f89d6f338ULL,0x8511b410056d068ULL,0x5f104dea8b84a602ULL,0xc121200008541008ULL,0xeafe256083dcf328ULL,0x4102e02009048222ULL,0x9682360b8ad6f122ULL,0x620dc2d20047a238ULL,0x208012018241c008ULL,0xc00afd608844a74aULL,0x48a740a28802b744ULL,0xae673a9b08ce9318ULL,0x24b108000a180ULL,0x7475f4e38a46f3a8ULL,0x851001001280ULL,0x800010002000200ULL,0x80c80000848000ULL,0x802000000000600ULL,0x1c1400020086a060ULL,0xdbcbb662cac6bb4eULL,0x18861d6888c4f120ULL,0x210b20a80560214ULL,0x2010848600561042ULL,0x4a2050280c4a102ULL,0xc02404c288044200ULL,0x7a84fd350b46a3caULL },
	{ 0x8a503662ccc3ae00ULL,0xcefe1fdccce3dfb3ULL,0x3bbeeffacee73fa3ULL,0xd567b6ffccc3be96ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfdeefff3edd7ffbfULL,0xf3d6ef2fedc3fe3cULL,0xcb7a1fdbced3bfb6ULL,0xefbdcffacffafb9cULL,0xbdb67fedcdcffbbfULL,0xbebebf736dd2bf3cULL,0xfbbfff7fefcfffbdULL,0xceca3e0fede1fa22ULL,0xeb9ddf5bede3febbULL,0x37ffdee34de7bfbdULL,0xff667ffecde3ff9dULL,0xfbbb2f7f6ddebea9ULL,0xcffffffecdcffebcULL,0xea3ffbcee9f27fb6ULL,0x1cb27f61efd3bf1aULL,0xfbebfffbcfdfffbbULL,0xcaafae7bccc32bb2ULL,0xefffffbeedc7bf83ULL,0xffffffdfefd7ffbdULL,0x8d3bd77bcdf7de97ULL,0xffffffffeff7ffbdULL,0x7ffeff76edd3feb8ULL,0xfbfef7fbcfc73fbbULL,0xebdb3edbcde3eab6ULL,0xbffb4fbaedc73f91ULL,0x95fff6fbedd27ebdULL,0x4ceeff63cc9b5e9aULL,0xebe4f77f4cdb3b3fULL,0xedf3fffeedcbffbbULL,0xefffcfdfcfcfffbfULL,0xfcff77ffcdffffb7ULL,0xfffefdf3efd7ff3fULL,0xffbfff7fedf7be9fULL,0xfff7ffbfefdbffbbULL,0xafdfefdbefcfbfbbULL,0xf4f7feffcddbffb5ULL,0xf3ffaf77eeebfb37ULL,0xffb77f7fefd77ebeULL,0xcdffffbbcfd3bf3eULL,0xffffffdfcff3ffbfULL,0xffdeffefcfc33fbfULL,0xfbffff7bced7fabaULL,0xfeffeffeeddbffbdULL,0xffdfffffcdffff3eULL,0x7ffeded3cdd37a92ULL,0xc4bbd73aecaa22b5ULL,0xf73cfed8afc33e99ULL,0xbef6ffffcdc3eebbULL,0xe9bef76ecfdf7ab9ULL,0x7fee36f32dd3bf99ULL,0xbfffcffacddbff35ULL,0xff7ff7dbcfe7ffb9ULL,0x5bfefdf3cddb7f37ULL,0xcb82673beee3cea3ULL,0xfecbfeffeedb7e86ULL,0xe7bf4fd2cfc3bf1bULL,0xfdf5fffbcdfbfeb8ULL,0xebfeaff74dc77fa2ULL,0xe9a67f77cfc6eb8cULL,0xbec7ba22cdda7ba7ULL,0xafdfdbd2ccd37f3aULL,0xf4fefef2efcbbf88ULL,0xdf7faff2cdd73ebbULL,0xdb937ba76cd33f85ULL,0xbed3ff1e6dc69ebfULL,0xa9ff9d5a4ed77b94ULL,0x34bf64facec21f30ULL,0xf95aff7bcdf73f29ULL,0xfebe6f7fedcf7ea0ULL,0xdfdef7f24fc7ff9bULL,0x6bffff536ddffe9bULL,0xb7bbfff7eef3febfULL,0xd1e99d746df23a8dULL,0xfbae2d7dcc877f92ULL,0x8b36f2afced33a2aULL,0xffffdfdbccf7bebeULL,0xffb9ceeeccef7eb3ULL,0xf5dfffffcefbffbbULL,0x5ffe7f73cdfb7bbcULL,0xfbfeff6fefffffbfULL,0xc9fffebfedf73f38ULL,0xb7ffcdfbedf7bf39ULL,0xfcbffffdcfe3ffa6ULL,0xdfbeff62cfdf7fb8ULL,0xf2beffd76ed37fb9ULL,0xcdebff3304d7beabULL,0xbfffef5fefe7fbb2ULL,0x7efaf66f4fc76e13ULL,0xa9ceaf716cd7bf9aULL,0xeb347faf6dd23f9bULL,0xefcfee77ccc7efbaULL,0xe77fda51cfc32e19ULL,0x1c931ff1edebbe83ULL,0x9b25a55accc23b2cULL,0x93b52d420d895888ULL,0x41c6d3ac4cd03e90ULL,0x4edfffd36cf3bf3cULL,0x8e5e5f69c8d37eb0ULL,0xd9ddfe7b8fe33a28ULL,0x7eee7dfcced6feb4ULL,0xd1d1f7768daeab03ULL,0xaa760ede4cc13fa0ULL,0x30b46c374c643e30ULL,0xd9efa7f90da53b9aULL,0xbe7c6d706ce2fe20ULL,0x5152f6018c973a19ULL,0x60891a8980533211ULL,0x292c55d2cc923804ULL,0x5490f4f108831b8bULL,0x800150240020024ULL,0xcaac2f5bc9d39ab0ULL,0x0,0x1e0431348d41b00ULL,0x190610c002282ULL,0xcced0c60cda21e92ULL,0x3e5c2e60ccc23e12ULL,0x2481688080c20202ULL,0xa055c00c4901220ULL,0x80100821c6802000ULL,0x20080000101000ULL,0x9014004808901800ULL,0x1428b604dc02a12ULL,0xf300040c00212ULL,0x162148fb44d15e24ULL,0xa2b793288e82a100ULL,0x55061a424cc34b10ULL,0x803a1b0745817e00ULL,0x6e428f4205d1fe00ULL,0x3080030144c1aa20ULL,0x9494a244c7c13a20ULL,0x4042081020800ULL,0x3004080004880000ULL,0x5832a0080c62229ULL,0x0,0x7174bece87c32aa8ULL,0x2022086048810880ULL,0x8000538688032810ULL,0x4954f40a8c232238ULL,0x8100001204080220ULL,0x346941400d886e80ULL,0x8080210004012200ULL,0xa9144148c8a6a81ULL,0x4000040008400000ULL,0x28118400c424800ULL,0x14aa4c2805804a20ULL,0x1000004004020000ULL,0x2014008000400010ULL,0x1820004808000ULL,0x50521805813a00ULL,0x8012064140432610ULL,0x8010000004080000ULL,0x8080000801000ULL,0x0,0x8d14e042cc20a01ULL,0x8400004002000ULL,0x9c62b0b80cc00a0bULL,0x80363d5046c3aa36ULL,0xfbb6ef7fcdd7bbb7ULL,0x5c8aa2fb4fcb1a88ULL,0x6f9fc9b14ecb6ab0ULL,0x647776f349d73e3eULL,0xebbe0fe02cd33e92ULL,0xea80eb0ee4d292adULL,0xe5bf28bcdcb7f21ULL,0xb9ecf414cc3af91ULL,0x10104ee088818800ULL,0xc5aa150289c70b10ULL,0xdf906f760dc1eb00ULL,0xeda6ed2a0cc39a12ULL,0xe35b92daeddbbe90ULL,0xfeff5fbbefc7bebaULL,0x9078df5444c72213ULL,0xde910f52c4c28eaaULL,0x704e0044052120aULL,0xeb0038d34d622918ULL,0xb62dd2f1ccc23e30ULL,0x3020d6294983f804ULL,0x9c047d604ec30610ULL,0xc994d362ece96b14ULL,0xc492274480523210ULL,0x369df77844d63e38ULL,0xd464bb7447d15e80ULL,0xb0865018ce3fa18ULL,0xc95442960e931290ULL,0x40308000010202ULL,0x3806404c020118ULL,0x461064d068022e07ULL,0x44a0054088424e80ULL,0xbfa04342ccc0ae10ULL,0xcfdebebb6dd7bebbULL,0x23fddbdaedd7fabeULL,0xd4f4dc720fcd2a80ULL,0xb7bf7f6bacdb7eb1ULL,0xfbb6efefccdefea3ULL,0xd6d7bf0eade77e92ULL,0x2a76dfcccfc22f87ULL,0xf5fb5fe3e5d27e1aULL,0x416004080008ULL,0x2900a0404888201ULL,0xc2805b96e4c32a12ULL,0x840068040040630ULL,0x890005102c002380ULL,0x804022424c012c80ULL,0x806b0000c800211ULL,0x40812885c00800ULL,0x6c6801085823800ULL,0x1d688fd045933a30ULL,0x4810d0002c402a00ULL,0x14840822000ULL,0x180801808400201ULL,0xaa4b90c2c8d25a82ULL,0x2501005040000222ULL,0x10118cc004423a08ULL,0x10a050048e00a02ULL,0x2102242080d01810ULL,0x8152f8020c412280ULL,0x8a004c1a44022231ULL,0x3407f26048c01e12ULL,0x8cfe87604c563b13ULL,0x420060020c024800ULL,0x1808520048600410ULL,0x2004508048022084ULL,0x810000000004800ULL,0xa15706084cc12720ULL,0x56041f2548521310ULL,0x889013600482a600ULL,0x256ec8180de33a08ULL,0x82cfd384c21810ULL,0x6018040410020ULL,0xf3163f674ce66eaaULL },
	{ 0xa1c5e3aae05b99c7ULL,0xc8d23e29ea52b253ULL,0xbede4fcdee7616ebULL,0xf7bef6ff6e5a90ffULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6dfe7ffbef73bdffULL,0xfb76ff3efb77b9dfULL,0xefebffebef7ebfebULL,0xffbfcffbff7fb3cbULL,0x14fdeee86e1b94cfULL,0xdffffff3ff7f15dfULL,0xff9f7f2fef7b9cefULL,0x1ca10110068804bULL,0xefbfffdf6a7e1fefULL,0xc6ffd7e37a7a1ecfULL,0x9fffff7cbe7abfefULL,0xffbe7fff7f7f9fdfULL,0xffdfff6bff7f97efULL,0x8bd57c5a6d571549ULL,0xf4bbffff7f7e14cfULL,0xd5debf66ea5f9dcfULL,0xfeae376e6e72196fULL,0xcbfaffbb6f5fbdfbULL,0xebfefedbff7e37cfULL,0x36934ffb247814dbULL,0xd1ff77fb6f76974fULL,0xff767d7eeb7bb3ebULL,0x8bfffffdaf7637ebULL,0xffff5f69ef4eb5dfULL,0xff9edfebff769bcfULL,0xd47656eeaa6a13cbULL,0x7bbc7df3e27217ffULL,0xebf5ff2e2a7fb9ffULL,0xcffbfe162e7697dfULL,0xbbafcf6eaf72378bULL,0xf5f7ffefff7b3fdfULL,0xffffef727f7fbaffULL,0xfffeffff6f7e97dfULL,0xefdfff1bff76b7dfULL,0xffdfdfdeff7e97ffULL,0xfefefefbef7a9fefULL,0xfb6dcf7af2733fcfULL,0xdbbe7fe57e7f93efULL,0xefffff3fea5ebbcfULL,0xfffedfdffe76bfffULL,0xfffe5fffff7fbbebULL,0xfdfcffffbb771fefULL,0xffbfbfffef7fbbffULL,0xe9feff7eef5abfdbULL,0xefdfdfef7f779affULL,0x4e93cfe16a7eb5cfULL,0xd16697edb65e33c3ULL,0xefd6777fee43bf8fULL,0xc3f3fb97ea7793ebULL,0xefdf7edbfe571bdfULL,0xb7aad7fbeb7796ebULL,0xf5ffb7fe6a57b3efULL,0x9bee7d61aa7297cbULL,0x7bffd22f6a761b8bULL,0xcd5bbfdbfb60b1efULL,0x32bf4edf2b7a14ffULL,0x75f6ffec6f7eb6dbULL,0xc7bebd31a2661c43ULL,0xef86e7ffee539ecbULL,0x8ddbfaaf2b761bcbULL,0x6ffeffcb627397dfULL,0x55f7ffee6f7a90cfULL,0xfdee3ff2626b10cbULL,0xbe3cffe36b7b136bULL,0x4d5fffb8aa66912fULL,0x2bfe7dcbae77385fULL,0x9ddb7fffba7e10ebULL,0x976faff2fe731bd3ULL,0xfebf3f7fae5b97efULL,0xebdfe37ffb7295efULL,0xaffefedfb77ab5efULL,0x94fbeffffe52b5bfULL,0xddfdff7fef7733efULL,0xfbbe7d72ff7e93ffULL,0x8bf7fe67f67e96efULL,0xeed57edffb7f1ccfULL,0xde9dc7eb706a114fULL,0xfffbffffff7e33ffULL,0xdfee7ffbff7eb7ffULL,0xfbffffffff7fbfffULL,0xaffbfffbef7f9fffULL,0x2f3fcfdeef7e1bebULL,0x7df7f7fa7b5b37cfULL,0xdbbfff73bb7f93dfULL,0xdb9effbe665fbddbULL,0xdecfbf3b7f5b1fcbULL,0xeb3fdfdbea5b9fdfULL,0x3cff5eea3b76b0dfULL,0xdf77b7f06f7316cbULL,0xffbf3fff7f7fb7ffULL,0xcfc7ffbe2c733ddbULL,0xabffdfdaae663dc7ULL,0x6457d3ffa34d1bebULL,0x1924c46228461207ULL,0x8c18a760e86a164bULL,0xe9c6f21022629887ULL,0xcc7e9e53fa721debULL,0x558adf3fea72124bULL,0x32ee376d2272b3cbULL,0xfa765efdaa7219c7ULL,0x8b5ff3eeef52951bULL,0xafff7e5b3f7a31eaULL,0x1a2d35a28721182ULL,0xf5d797ebef73159fULL,0x1b2a1f6026629486ULL,0xcbd5b22eea4793cbULL,0x4dabe0b6b42128eULL,0x679cc9316a4238a7ULL,0xacdd22804843300bULL,0x97103c53b052304bULL,0xda90876c22421c49ULL,0xbc4ca022a401681ULL,0x4840010002001101ULL,0x6ee47e82b7e14caULL,0xc1ae2f6869721a8bULL,0xdb06db7c25431bc9ULL,0x54c0a73027429387ULL,0x31c5e5727721d99ULL,0xcf0dd61a25a02cbULL,0x91098000600a00c1ULL,0xea0215832a601801ULL,0x8001e32861619801ULL,0x822a98ca207090a3ULL,0x20884de8ea602201ULL,0x5002016ea2408083ULL,0x1ac27c621a62320bULL,0x82747970b2721122ULL,0x2400000024000000ULL,0x804000421010ULL,0x9000000000421200ULL,0x24052002000046ULL,0x8100820010401048ULL,0x480200002001402ULL,0x100080001400000ULL,0x6000023100480060ULL,0x30000460400a1041ULL,0x4200241200411220ULL,0x4008100100001080ULL,0x2002000000600000ULL,0x0,0x200010000011000ULL,0x1000080000001001ULL,0x4000040024000000ULL,0x4000100020008000ULL,0x2000220000080ULL,0x100046048000001ULL,0x800010000000082ULL,0x4908c00428141ULL,0x2b00181921521088ULL,0x1259302942000bULL,0x8080000020020000ULL,0x200200020008000ULL,0x2004000040000001ULL,0x409080822501800ULL,0x4b3c4600ea60002aULL,0xf46432087052100fULL,0x22542a4429093ULL,0xeb98ed24266014dfULL,0xcc53be7fb352b56bULL,0x28e8b336952144fULL,0xb3700668a77811ceULL,0xf7beef60fe7a93efULL,0xdb91fb9f6e7a11cfULL,0xfadfff1fef7e91cbULL,0xab7fdfde617711ffULL,0xf0c6f5e6eb62948bULL,0xf96e676af37a96d9ULL,0xdb3e7fe7aa6612cbULL,0x8f564f1cef6218e1ULL,0xaef7ff53ea7b90dbULL,0xb7badfeb7f639ddbULL,0xf96df67f787b10c9ULL,0xc1a22f40226810c9ULL,0xefcffb32eb70baafULL,0x1f76cecce47611efULL,0x33384759ac7e12c1ULL,0x5004bc0808661408ULL,0x4f526d696e6213cbULL,0xcbd6df6e6e52b1efULL,0xfadfbfdb6f7b97dbULL,0xbbbcca786e7a1fc8ULL,0x1340d27263721401ULL,0x4aa86d42265213afULL,0x1304033218603001ULL,0x84198a1210609007ULL,0xad32050267441402ULL,0x153d002a2312140bULL,0x978207402000908fULL,0xa62c45682b42104bULL,0xcfc63613a272b1cfULL,0xffedfd72e7a9fcfULL,0xb4befff96f771eafULL,0xbbefeff22f5797dfULL,0x6fd3257662421783ULL,0xdff6eb7fff7b92dfULL,0x21e8cd1a2172b00bULL,0x2a002008000001ULL,0x100040000400008ULL,0x8ec60f600b72b0a3ULL,0xb41c2342a5230a3ULL,0xa2172e4040621501ULL,0x78005000143100bULL,0x52487771285212cfULL,0xf2023220ca420043ULL,0x816000a0600000ULL,0xcec32e402263116eULL,0x3316c7f06856208bULL,0xc424504222609c83ULL,0x0,0x1d0800e02422049ULL,0xe8ea7e52f87e174fULL,0xb884520224c10ceULL,0x484040002080040ULL,0x3054122c001007ULL,0x6208c13c186a1441ULL,0xfc7af3b6a7a12cbULL,0x2144040a0409442ULL,0x109486e0204614c8ULL,0x79ee7f7bf7721c8fULL,0x4a912d34a201118eULL,0x4041a010a2021082ULL,0xaa15c5460260115fULL,0x869556e8017a31caULL,0xd984b4220a503046ULL,0x18000c0124401420ULL,0x260000001049ULL,0x48cc4cc9a76908bULL,0x4caa59872c60978fULL,0xe1228f58226a14c6ULL,0x4294311c64021101ULL },
	{ 0xe9f3fb7309164ed4ULL,0x6ed0be01483ccfd6ULL,0x9a9bcdfb883e2c7cULL,0xf7fdfffbea3efef6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffbe7dfbea7ffff7ULL,0xfbb6ff7feb7dbfffULL,0xeefef7ef4a3c97dfULL,0xffbfef538b3feeffULL,0xf7fe167bea7efff7ULL,0xdffffff5eb3deff4ULL,0xebdf6e2fc21e1dd4ULL,0x43cc1001011e0e30ULL,0xefffcf5e6b7cbfe5ULL,0xfefffeff4b7effffULL,0xfdef7ffbeb7daff7ULL,0xffff7fffe93faeffULL,0xffffef3beb7dbfffULL,0xabffdfdf8b7caffbULL,0xf5bf7ce76a3fefffULL,0xdb67b6f8037f8ff5ULL,0xfab73f7e6b3caef5ULL,0xcfdfff3feb7fbef7ULL,0x6f7f9ef76a3e7cf4ULL,0xbfbb5ff38b3caefaULL,0xf5eff77feb3eecfeULL,0xefbe7dfde93efed9ULL,0xebfdfbbecb7efff7ULL,0xebffdff3e93caff6ULL,0xdfbb4ffbe87eef57ULL,0xff57767c093c89f6ULL,0x7dfe6f7aea3dbfe7ULL,0xfb84db36c87cfbcfULL,0xebdbde7fc97eadf4ULL,0xefbffff2cb7dabfdULL,0xf6ffd7fa693cefdfULL,0xfbfefd734a7eb3f7ULL,0xffbeffffeb3dfef7ULL,0xefffff7feb3ebff7ULL,0xff7fdfdbeb7ffff7ULL,0xf6ffffffc97ebff7ULL,0xd7eff7734b3ccfd7ULL,0xdfffffffcb7fbfffULL,0xffefff7fcb7cbfffULL,0xfbffdfdbc93efff7ULL,0xb7ffffffeb7dfff7ULL,0xff6effe3ca7fbffcULL,0xffbe3fffeb3ffff5ULL,0xeffffffeeb7fffefULL,0x6bff9fdfca7cbfd7ULL,0xbffbdfffc93cafd7ULL,0x51aeb571603e9e55ULL,0xfffc7fff6b7f1fd4ULL,0xabdff36e0b3cadd7ULL,0xffe77ff7487fafefULL,0xbcfecffb6a3eeff4ULL,0xf47fe6f4e93eecf6ULL,0x6ebe7df0687cfff6ULL,0xbb4c6f6ae93dcfd1ULL,0xcc7b1ed9c83ddfccULL,0xaffecffbc83cbcfaULL,0xd4bffeef4b7ed9f5ULL,0xa96e7d61893ebe66ULL,0xdb9cc7e7683c97f4ULL,0xc5c3f958697e8ed2ULL,0xc3dfcd48cb75f9caULL,0xd477ffeb487caeeeULL,0x83222b406820a041ULL,0xfbbe2bf34a1da7dcULL,0xcfd7ff3aeb3eebf5ULL,0xbffefccf683fa0f5ULL,0x14ee5fe2e93dedfbULL,0x9169f74b0b1daccdULL,0xfe3e2fe6e83cbff1ULL,0xcfe6fff64a3ffcf4ULL,0x4f7edf5dea3c0ef4ULL,0xbefbfffaea7caed7ULL,0xf3fbbffa487fa7f7ULL,0xf8e6ed62eb3ca8f1ULL,0xe9dfefefe93ceff1ULL,0x6df77e5ecb3e4b54ULL,0x97fbdfdbeb3cffc7ULL,0xfff7ffff6b7ffffdULL,0xffbefffbeb7ffff7ULL,0xfbffffffeb7fffffULL,0xfffffffbea7effffULL,0x6f9fdf7b4b3ebff3ULL,0xfdd5fffa6a7d15dcULL,0xdfbeff74c83ebeddULL,0xfb96fffae97fdad7ULL,0xedefbef5c93caff7ULL,0x7bdf5f5f693dbef7ULL,0xb4fc7eed683caadfULL,0xb3ecff704a3e9cdfULL,0xdfb67f7fca3ffff7ULL,0x9cc7ff7ec87c3ef0ULL,0xef5fff5a487dfef7ULL,0xbef51feb6a7efcf4ULL,0xff9beeea486c8dddULL,0xcab7aff70a7ebadfULL,0xc984e72d4b36aa4eULL,0x3ff92c38b3ecb40ULL,0x3ab8d9f30a3ebfd5ULL,0xd36dbf39c93d2cd4ULL,0x93beaf7d6a3ebec5ULL,0xe9e7f94f4b3cafdbULL,0xfe971ecbe87c2fc4ULL,0xbbacdf7bcb3e2ff8ULL,0xfddf97f8eb3dbae8ULL,0x2a847403403e0154ULL,0xfbfff37feb7efff4ULL,0x85f7db49093498d4ULL,0x821ec7f30a3fdaf6ULL,0xd530f67f4a5c27f8ULL,0x950221600b1c981fULL,0x8ccb20008382044ULL,0x47c60a0948362644ULL,0x81200f1000080074ULL,0x8054a80900a842ULL,0x412e4b62083c38c4ULL,0xb6140144083e0a94ULL,0x4782a000081c009cULL,0xf25485a083c16d0ULL,0x8022b0180208882ULL,0x106cef71823c88a1ULL,0x4096010048102200ULL,0x20222a260210140ULL,0x1cd60008082050ULL,0x208060008082004ULL,0x40040000008082ULL,0x400040420800a080ULL,0x4000010808200200ULL,0x9126800108246ULL,0x22a80400000c8086ULL,0x4164142a002c84d6ULL,0x780c2d2000389893ULL,0x802000040004ULL,0xc0189140140180ULL,0x8590080000360400ULL,0x502cd4d4087cb09bULL,0xb038042088302800ULL,0x4320007280148c51ULL,0x604220088100a40ULL,0x0,0x0,0x2100060000000ULL,0x0,0x0,0x0,0x80100008002000ULL,0x0,0x8000010008002000ULL,0x80400000100200ULL,0x8580283082c00c0ULL,0x4114060081890dcULL,0x1002100080004ULL,0x200100100020ULL,0x8800200348100440ULL,0x60a3200601000d0ULL,0xb1cf41b413e88d4ULL,0xd1f49a634334d8c3ULL,0x5a2024400010a1d2ULL,0xfbd0ef4e4b7ecbdcULL,0xafeb9880ca3d0df4ULL,0x3e004f804824ac01ULL,0xd4f7576e0a7faff7ULL,0xff247c604b7ca8f6ULL,0xfba8ff3b683edef4ULL,0xbfdffef18b3e9ecfULL,0xeedf5ed6ea3cbff7ULL,0x103e5420081e22d4ULL,0x84dc2571ca148e93ULL,0xffb02747ca1cabd0ULL,0x5fd6f798cb3feee5ULL,0x2dfb1d4b8a3d6bfcULL,0xbc131deb481eb2c0ULL,0xf7b2af20c23dbad5ULL,0xb8b2e762c8188fd0ULL,0xc59ac379801c88b5ULL,0x28675ad1403487d0ULL,0xf4320fffe97c3e90ULL,0x16cd368031c98d4ULL,0xa00276c003e87d0ULL,0x81720b1dca3489c4ULL,0x6a563fdb8a3eb9c6ULL,0xe92cf02091c2ac0ULL,0x53507069c810aac2ULL,0x29402f22603482b4ULL,0xd070672c0924e8d0ULL,0x4000202000240040ULL,0x65064012001091f0ULL,0x8010004200260040ULL,0xc7bc2e60c834e8c0ULL,0xd4003708406c0a50ULL,0xcec23c1bc83cfef0ULL,0xeebfcddbc93eaed4ULL,0x966646e0893e4d54ULL,0x9f7ffff2697fdffdULL,0xdf9a6f3c483dbed6ULL,0xefc7fbbfe83cbbfeULL,0x85b18514a1c1c80ULL,0x9081512009240a40ULL,0xd800001040108c80ULL,0xd485201000341050ULL,0xc050c00000248452ULL,0x8004000080010ULL,0x80324e23411c8ac2ULL,0x0,0x4212310000080050ULL,0x20008200000ULL,0x84cb0a46c03cc8d0ULL,0xb5bfadf3683daeffULL,0xc084166a001ca0d1ULL,0x88042560483c1040ULL,0x8210400000180480ULL,0x8622290600e82d0ULL,0xc90854000349141ULL,0x4200080040ULL,0xa1da209248880ULL,0xeaa5e32c083e94ccULL,0x8972bea06a1c8f52ULL,0x621881c2081c3080ULL,0x549476ca483caef2ULL,0x808d4060001c14c0ULL,0x9a94a364082c8ad6ULL,0x4280400208380210ULL,0x3f40541030c8c80ULL,0x0,0xa0040480002c00c0ULL,0xd4840e66c15c8a83ULL,0x89448b0ae8100a50ULL,0x102ca40240804ULL,0x27e864f34938c9c6ULL,0x44245c2a80150d64ULL,0x6a408c08900a880ULL },
	{ 0x4200600200094d00ULL,0xf9565b575b181f19ULL,0xde9fc76b1b9b5f5aULL,0xdf7ef6eb038ac79bULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfebe7dfb7f9acfdfULL,0xb3c6bf62268b4f5fULL,0xefe7bf696f9add5bULL,0xe7bfcffe379bdfd9ULL,0xfcfef7fd7f1adf5bULL,0xdffbef76579bdfdbULL,0xfbbfffff7f9acf5fULL,0x252300a658acf1bULL,0xbf7fffcf7f9bdfdbULL,0xc7f3f6fb7b8a870bULL,0xfffeff737b9bdf53ULL,0xffbfff7fff9bdfdfULL,0xcdffffbfbf9adfdbULL,0xefdf5adfbb8bdfdbULL,0x3bb7ffffd71adfdbULL,0xd9fbffd42f9b4f5bULL,0xfebc3dfe7b8ad7dbULL,0xefdff7fbbb9bdf5bULL,0x3edfffdf379b4f5bULL,0xefab6feb5f9adfdbULL,0xd7effdfbff9bdf5fULL,0xffbe5fe77b9bdf97ULL,0xcbf5fbff7f9adf1bULL,0xefd73fff6f93cfdbULL,0xfb1fcff30f9ac75bULL,0xfdf7fe7f3f9bdfdbULL,0x6fc87d59369b5f5bULL,0xfbd5ff6f571bdf5bULL,0xefdfffffb39adfd7ULL,0xffbfef377f9bdfdbULL,0xf5fffefb171bdfdbULL,0xfff7bff1eb9bcfdbULL,0xdffdffffdf9bcfdfULL,0xdfdfbf1fff1bdfdbULL,0xefffffdfff1bdfdfULL,0xfefffff8ff9bdf5fULL,0xd97fff7b771bdfdbULL,0xdbb77f7fdf9b5f5bULL,0xcddfffeafb9bcfdfULL,0xef5fdfdfff8bdfdfULL,0xddffffff379bdfdbULL,0xff6fff7b7f9bdf5fULL,0xfebfbfff7f9bdfdfULL,0xcfdfffafff9bdfdfULL,0xfffefefdff9bdfdbULL,0xbfff4ff2ff9bdfdbULL,0x5178b46b33881f5bULL,0xdb967d735b9bcf5bULL,0xebd5db7f378bcfd7ULL,0xeedfffd33b8bdf1bULL,0xbffbcff11708cf5bULL,0xd573ff7d668adf53ULL,0x5e7ef5f3978adf5bULL,0xfbdfef6e731acfdbULL,0xeefbfe327b829fdbULL,0xcffdfff86f9bd7dbULL,0xbfff5fedb78acfdfULL,0xa794bf30030bd71fULL,0xff86e773f21a4f5aULL,0x4dcefed8d703575bULL,0xabf6dfdb6f8b5fdbULL,0xf4f77fefbf9acfdfULL,0xcdedef4b0b8bcb83ULL,0xff9cbf5e1f83df5aULL,0x7d88cf187b824b1bULL,0xebfbffdbff9bdf5fULL,0x35ffdffb5b9b4f5bULL,0x9deefe7a7b12dd5aULL,0x5ebfbf7fbf1bcfcbULL,0xcdf6ff32539add5fULL,0x1b7fff5f979bdf5eULL,0xbfbb477b3e9bdf5bULL,0xdbf5bf7e6d9bc75bULL,0x779e3d756f9bdf5bULL,0x8bd6ff773b8bdf5bULL,0x6fdefc57079acf53ULL,0x9f9dffff3792cf5fULL,0xfffdfffbff9bdfdfULL,0xfffe7ffbfb9bcfdbULL,0xfbffff7fff9bdfdfULL,0xeffffeffff9bdf5fULL,0xffbfdff77f8b4f5fULL,0xfff73eff4f9a5fdbULL,0xbebe7df51f1bcf5bULL,0xfbbfefbedf0bcf5bULL,0xefd7ffbf7f9bdf1bULL,0x3feedfdf4f98d71bULL,0x91fa7eec3e9adfdbULL,0xcfed67fb6b9bdf57ULL,0xdff6ff6f1f9bdf5bULL,0xcdffff3eff1acf1fULL,0x4bfecf5f5f1ad55fULL,0xb59679f7b28adf5bULL,0xddcdc5c2021acf5bULL,0xdeb77fff7f8adf9bULL,0xcddefaee1b1bdfdfULL,0x6fdf5bdc3393d79fULL,0xbebbefff7f935d52ULL,0xd3f0b743538ac31fULL,0xd8b67f741f0bcf9bULL,0xc9effbff179bd75fULL,0x3edbdf43070ad71fULL,0x90975dba0913cd5bULL,0x75b4fef9b31b475eULL,0xcaae3d70db8a4d5bULL,0x9980e716de0a451eULL,0xe463df5812035f1fULL,0x6eb68f927301cf18ULL,0x308254c0618a8101ULL,0x291e24601288050aULL,0x500426080600410aULL,0x58da3fbfba11cd9bULL,0xa0015d400412551aULL,0x62262804015993ULL,0x214fe76aaf9bdf52ULL,0xc0007c0010004d4aULL,0xc849561036104709ULL,0xa1c14020c02c51aULL,0x84d556a82d01431aULL,0x108860203114410ULL,0xc9ac7de05911cf13ULL,0x820000000402ULL,0x400000802000200ULL,0x4914020800a0442ULL,0x440400380800051aULL,0x204456440008113ULL,0x90c2246908010bULL,0x2020000004800ULL,0x8414000100412ULL,0x2800000408ULL,0x40200080000400ULL,0x0,0x0,0x10020000000100ULL,0x41fcf2e8a709555aULL,0x8088d2003924408ULL,0x880c280381014610ULL,0x802100000000381ULL,0x101914020001518ULL,0x2000002002000008ULL,0x2404080020c00ULL,0x0,0x0,0x100400010000400ULL,0x22004000080112ULL,0x24000100400ULL,0x40a9044202020253ULL,0xc10202000112ULL,0x48161a04010a000aULL,0x400002000000012ULL,0x901c50255500455aULL,0x102100010000100ULL,0x100840008020140ULL,0x10040000000100ULL,0xd5bcff34e9a5fdbULL,0xf3c5fe7a521bcd49ULL,0x5f96bef2798a5f5fULL,0xc901eb6253885552ULL,0xf1da064f909346d5ULL,0x104d421403440aULL,0x99d4d6620b08471bULL,0x9fc276435b82175bULL,0xfba5f76f5b02cf1aULL,0x8fdbff231b8acf1bULL,0x6bf6cfca3792df5bULL,0x74b95d61f19b5f1aULL,0xda2ee5721b0a475bULL,0x34b00b74080253d3ULL,0xce7aff5a0e9a4f4bULL,0x83fefed7d31ad79bULL,0x330476a03004f4aULL,0x49e82d084093cf1bULL,0x7a9faf3cd61ad55fULL,0xe342dd2a7b88df53ULL,0xcace5c19211a4f0aULL,0xf7fe5df33a1a0f5bULL,0xf42ea6f18a0a9f5fULL,0x78143434f2020f09ULL,0x83bab221050acd43ULL,0xeff71f573f8acddbULL,0x913ec7e33f925fdbULL,0x5814220818000540ULL,0x87821f405b024f03ULL,0x8180622610080412ULL,0x48c2921101084f09ULL,0x831244a0108d112ULL,0x508056420f004313ULL,0xfb32af72038a5f5bULL,0xf0243f361392055eULL,0x5fc397b63b8ac75bULL,0xab97c7dfcc0a4fdbULL,0x96be7ee8238a8f5fULL,0xfb7377f36b8b5f1fULL,0xdf92f77f5b9ad71bULL,0xefdfff7e971adf5fULL,0xa2109141020b4059ULL,0x0,0x844127000300830bULL,0x1885010001820110ULL,0xc0c0130201824310ULL,0x21080000400ULL,0x8400001000010ULL,0x2404044008020144ULL,0x608e0de083088713ULL,0x808b0020100500ULL,0x8744e00218a0b13ULL,0x469fcde2738adf9bULL,0xd223e6fb4308cf9bULL,0x5d262e6a0118455bULL,0x80d0024204184500ULL,0xe040b412000206d9ULL,0x36808410210a0902ULL,0x308d002012820300ULL,0x2a262c20028a4f9bULL,0xfe2cf37e321bdfdfULL,0x4dd0de10138ac395ULL,0x808014009000102ULL,0x54d5f58efb081793ULL,0xb97cff72571acf53ULL,0xd21461140002cb1fULL,0x1907f66454084759ULL,0x4a80194820080709ULL,0x441415831008c903ULL,0x8849660040080581ULL,0x7c952bda131bcd5bULL,0x8110c9113000cf01ULL,0x44000000080100ULL,0x88434a32008110ULL,0x50120d08000a0b01ULL,0x5e9e77740b0bdf5bULL },
	{ 0xe8ebeb38fcd15ce2ULL,0x6ffd1f4c15a84ee6ULL,0x1e8fddef9dfdace2ULL,0xf537f6eeecfcece2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdebe3dfafdfddde6ULL,0xfb8afbfff5fdfee6ULL,0xeeeadedb54bdefe2ULL,0xafff5ff37db9efe6ULL,0xfd7d77fe3df967e2ULL,0xfffeff73fdfdffe6ULL,0xffffffff7db9ffe6ULL,0x2956f00f1cb9f9c2ULL,0xbfffdfdf7dfdffe6ULL,0xf7f7f7fefdbdffe6ULL,0xfffffffffdfdffe6ULL,0xfffffffffdfdffe6ULL,0xffffffbfbdfdffe6ULL,0xeffedbdfbdfdefe6ULL,0xffb3ffeffdb8ffe6ULL,0xe5edf772bcf97fe6ULL,0xdabeaff6fcb87fe6ULL,0xcff6ff3bb4f9efe2ULL,0x2bfeb7d394f967e2ULL,0xb6aadffb75bd7ee2ULL,0xf9ffaf7a3df8dfe6ULL,0xc7fe7f76bdfccee2ULL,0xd99ffb7bf4f9efe6ULL,0x6ff736dffcbdede2ULL,0xfb2fef7b5cfc6fe6ULL,0xbaeeff6eddbdffe6ULL,0x7dfc6fa39dfd45e4ULL,0xfbd7ff3ff4bcffe6ULL,0xefffbefbfdbdffe6ULL,0x7fffcfebfdb8efe6ULL,0xf5f7f7eefcbcefe6ULL,0x9ffafff1fdb9efe6ULL,0xfbbfffffb5bdefe6ULL,0xeffeff3f3df9ffe6ULL,0xffffdfdff5fdffe6ULL,0xf4fffffdfcfdffe6ULL,0xff6fff737dfdeee6ULL,0xffb7ff7f9dfdefe6ULL,0xfddfdf2ebdf9ffe6ULL,0x6fffdfddf5bd6de0ULL,0xfefff7fbfdbdffe6ULL,0xffffff7bfdf9ffe6ULL,0xdbbcbffeddf9ffe6ULL,0xe7f7ff3efdf9ffe2ULL,0xafdebfeff8f8efe6ULL,0xfebfeff3ddbdffe6ULL,0xfdfff7ff3db9efe6ULL,0xdffe7df5bcb9ffe6ULL,0xebcfff3a7df86fe6ULL,0xffff1eef3dfdffe6ULL,0x8baacffbf49d7ce2ULL,0xf55676fd3cbdeec2ULL,0xdfce7f70b5bd7fe2ULL,0xfbfcff7ffdf9ffe2ULL,0xeefebffffcfc7de2ULL,0x28bccafb34a546e6ULL,0xb4331feb3cb947e6ULL,0x7bfebf72d5bcece6ULL,0xcf01bf6f1db867e2ULL,0xcecbf62b7cd86fe6ULL,0xaabfdeda7db96fe2ULL,0x36f457e0b4b8efe2ULL,0x9b7fff5ab5b9ede6ULL,0xfb9a7f5c9cb0d6c6ULL,0xdcd78fbe35f8efe6ULL,0x8bdefecf74bc7fe6ULL,0xadb6fffb95bd8fe2ULL,0xb97cf667f5fde7e2ULL,0xde9e3ff1d5f9eee2ULL,0xff7cf33195f9cfe6ULL,0xb5e3e69fde8efe6ULL,0xb6b25ffb7db96d62ULL,0xffd2b77a54b9f9e6ULL,0xfbb6bffc74fdefe6ULL,0xcbf2f363fcb94ee6ULL,0x4dff1e819c9c6ce0ULL,0xbffbfffbf5fc5fe2ULL,0xfffffffff4bd7fe6ULL,0xe7b6bf7935f9ffe2ULL,0xfbfffffffdfdefe6ULL,0xfbeffe9ffdb9efe6ULL,0xefbffffffdfd7fe6ULL,0xf7ffffeffdf9ffe6ULL,0xffffbf737dfdffe6ULL,0xffbff7fff4bdefe6ULL,0xcffffea5d4dc7fe2ULL,0xaf5d7fdf7c917fe6ULL,0xf4f77fea74f9e6e2ULL,0x896cbfb8a5396de2ULL,0xffbc6b727dbde7e6ULL,0x8dc7afb0fce85fe6ULL,0x6e5e9cdb94f877e2ULL,0x309e1cef95b8eee2ULL,0x7aaef5e358316ee2ULL,0x9eda2d51dcdd4de4ULL,0x4b8af7a79df94fc0ULL,0xefdffedfbdbdffe2ULL,0xffff6ffbbcbcfde6ULL,0xd1aff37b1c997ee6ULL,0x5bd67f3015bd6fe6ULL,0xebe7fb2f14957ee6ULL,0xeeea3ffbdcf9efe4ULL,0x9799cfd23c99eec2ULL,0xb576f66e7db9fbe2ULL,0x67aa2df8bcb8e7e2ULL,0xfbd4f1343488f7e2ULL,0xae42bf59f491ede4ULL,0xffa64fb33c8ccbc0ULL,0x84f5e6c01cb949e6ULL,0x928d21c011a087c2ULL,0x8180c80405b08480ULL,0x4dca1c1414986d60ULL,0xa08ddd20db8ed00ULL,0xd290746054184fe2ULL,0x98d2104040184ee4ULL,0x5ebc4571141942a0ULL,0x46008330141166c2ULL,0x2bfece5635b966e6ULL,0xa0d194ab10d84d60ULL,0x8001820000200286ULL,0xd286b324cdb86ee0ULL,0xcdd0d5b255a0bbe2ULL,0x6cc318581488a7e0ULL,0xc3914ce11089c842ULL,0x5c449443353803e2ULL,0x4a86041020d802c0ULL,0x800002000006000ULL,0x860a0900106060ULL,0x28b23b82b0808dc2ULL,0x4104007230004484ULL,0x1884040010880282ULL,0x80030500c40042ULL,0x84ea515191382d20ULL,0x864840140143a0ULL,0x95b510080490c2e0ULL,0x0,0x800000c38004000ULL,0x8020000010000400ULL,0x8010170204d00000ULL,0x4001200000ULL,0x200410000002ULL,0x0,0x0,0x0,0x301a02230099c802ULL,0x0,0x8002000110000ULL,0x187900800196000ULL,0x40020a00310040ULL,0x12420000204042ULL,0x5000845040086020ULL,0x4a12040088904540ULL,0x1048020040008c0ULL,0x808408000c000480ULL,0xffffefe3fdfdefe6ULL,0xfff6f7fdf5b97fe6ULL,0x577e7df9bdf96fe6ULL,0xbbf2ef331cf8c6c6ULL,0xa421e0ad0194460ULL,0x814014001282ULL,0x32008269748054e6ULL,0xa50e7d640cf87ee2ULL,0x73b6fb7c30b967e0ULL,0xc277b23ff4f97fe2ULL,0xa3dfdd1b3da9cce0ULL,0xd60516e9f4b567e2ULL,0x9a4f7638c98ace0ULL,0x9abc02460d306a42ULL,0xedc7fe0e34b8cee6ULL,0x3fed9f581c99fec2ULL,0xb6ba4ffb54d9f7e6ULL,0xfbefff7bfdf96fe6ULL,0xa8be2f50c5a8dca0ULL,0x83d9faf03cd8e5e4ULL,0x1e1c024011b98ca0ULL,0xbcfbdfbb9df947e4ULL,0x108400a124901040ULL,0xf12c3f6c3c0847a0ULL,0xcad0db0804b94fc6ULL,0x2cf73adf3db56fe6ULL,0x87364f7254117fe2ULL,0x911cf22bb0b96462ULL,0x1816045045b04820ULL,0x4290c00214802cc0ULL,0x458021000a08042ULL,0x480843f10099e040ULL,0xd052b34204dd7120ULL,0x1be4d6030b8ece0ULL,0xc2304006141158e0ULL,0x7fc6ba3b5db973e4ULL,0xbb1ec7de94d9cdc2ULL,0xc49276f235b0ffe6ULL,0x976ffff37cbcefe6ULL,0xe3b6e60f95b85fe4ULL,0xe957f71a1dfdefe2ULL,0x62040880084420ULL,0x4ae102210848da0ULL,0x80407750009866a0ULL,0x2802d42140027c0ULL,0x89040a0020112040ULL,0x10340010112000ULL,0x688140890842442ULL,0x940070a05a003a0ULL,0x4204e1e0049982a0ULL,0x8124ae0011b0a400ULL,0x651300180b90400ULL,0x20a4c55290984560ULL,0x61365278841175e0ULL,0x6010b53004bc4342ULL,0x8000200008004000ULL,0xc002023090006d20ULL,0x882090010904820ULL,0x34a006ec10e02760ULL,0x9ac020401c096060ULL,0xc18442bc20186ce0ULL,0x4a433681cdb0e6e2ULL,0x104e4000880120ULL,0x84d054e850b9e6e6ULL,0x95be21603c986ec2ULL,0xc0843b5211a1c684ULL,0x48d5c316149975e2ULL,0x4a061001b4011680ULL,0x18155a144b04462ULL,0x4840166084806d00ULL,0x18282f4c14989282ULL,0x8141a80010980aa0ULL,0x6de8090004ea0ULL,0x190c25000306442ULL,0x2150934840980ULL,0xa804758208004ca4ULL },
	{ 0x30ae3289610566aULL,0x6b52db69a80f756aULL,0xb4b5ef533eb6d4eeULL,0xb4f6160d94f650e2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100020000020ULL,0xcbffeffffefef6eeULL,0xcf4bbfecacf2f4eaULL,0x2fbbcf58caf670e8ULL,0xfcfdf6fb3cf6f7ecULL,0xafb6ad73aede57eeULL,0xdfadeb7fb6f7f66eULL,0x6bfefa9fea7ef7eeULL,0xbfffdfdffefe77eeULL,0xffffffffb6fff6eeULL,0xfffffffffefff7eeULL,0xfffffffffefff7eeULL,0xefffffbefefff7eeULL,0xefdeffdfb637f7eeULL,0xb5ff56fbfcde74eeULL,0xf869ffb2fc74f46eULL,0xfebe377a6efff6eeULL,0xdbdbffffa6fff7eeULL,0xebffff57befff7eeULL,0xbfbb6ffea6f6f7eeULL,0xf5fdf6daf45e77ecULL,0xf3b67dec2436f6e6ULL,0xe9effbffbe3ef6eeULL,0xeff7dfddfc7674eeULL,0xbe9f5ffff676f6eeULL,0xbe76feeeb6fe77eeULL,0xeffe7d7bbeb6f6e2ULL,0xfbe7ff3e3ef7f6ecULL,0xefffbffb7efef5eeULL,0xbfbf4ffbfef6f6eeULL,0xf7fffffefeb7d6ecULL,0xdffffff3fedef7eeULL,0xfbfdfffffe7ff7eeULL,0xdfdfff9ff4f6f3eeULL,0xaf5ecfdfb63676eaULL,0xdefffefff676d7eeULL,0xfdfffffba6f7f7eeULL,0xdffefffff6f6f7eeULL,0xfff6fefef4d6f76eULL,0xfb5edfdbfe76f6ecULL,0xbdfffffff4fe77eeULL,0xd9fffffff4fef6eeULL,0xfeb6ff7ee6d5d7ecULL,0xef9fff7fb6b275eeULL,0x5f7ffe5de676766aULL,0xbebf7ffff4b772ecULL,0x77febfef9456f5eeULL,0xfafe7ffefcb6d6ecULL,0xcbd7fbffeefcd6eeULL,0xeef77adaeef6f6eeULL,0x3faf5fd966f6f6eaULL,0xbce67f7bbe1ff66eULL,0x79bc4760663677e6ULL,0xebbfd367b63776eeULL,0x8ca33550632f66aULL,0xafbedde3a47676ecULL,0x54f4d6f1961776eeULL,0xefbefff2f67ef46eULL,0xdfc4e5ffb4325026ULL,0xdfcfb717fcd6f76eULL,0x33ea4d58045574eaULL,0x1cbf64624456d6eeULL,0x913ef773207dd4eaULL,0xc954362ec446f1e4ULL,0xc24be31a64f0146cULL,0xafdf7d93e6bef3ecULL,0xb5bedfe3e632d2ecULL,0xa1ed7779861ef6ecULL,0xde8fbfd0a4f6f2e8ULL,0xc1fedf287436f66eULL,0xa94ecfcfa4b65562ULL,0x9693dff68c1456ceULL,0x1ae31f79ce16546eULL,0xbf9e7d3dc63677eeULL,0xe33eef628a1ef6eaULL,0x2fd6b6a79434f6caULL,0x838ecff016367468ULL,0xd55bbfefacf272ecULL,0xfffe7df3feff77eeULL,0xfbfe6f3ef43676ecULL,0xcac67f7fa4bed16eULL,0xe3bedff3f4b672eaULL,0x75f7d7ffa676f2eaULL,0xcdbfff73bcb6f6ecULL,0xff9ef7fffeb673eaULL,0xdbdff79bb4f6f7eeULL,0xefdfffcbf63ef76cULL,0xfcffddef96f7f7eeULL,0xedefef72e6f776eaULL,0xdf967f7bd654d4eaULL,0xddf7f73ab65554eaULL,0x4a5edfd95416d7eeULL,0x446a57e936b670e8ULL,0xd1e6f75e20d676eaULL,0xd3a92f67a697f26eULL,0xce02f6828405d26eULL,0xafddc9cf240574aeULL,0x9ebcfff9acf7746eULL,0x8179b06224433464ULL,0xeb06bfe6acf677eaULL,0xe9bffb6eec16f4eaULL,0xfff7fedff6fef6eeULL,0xbffffffffefff7eeULL,0xf7fff7eefebff6eeULL,0x64ee6460b4d2566cULL,0xd9d4c76e341e56e8ULL,0x87421e2314cd35e8ULL,0x85fe8576081ef46aULL,0x3020004020000460ULL,0x402953040801428ULL,0x90000100000010e0ULL,0x862e15a0525024ULL,0x2b5047582676766cULL,0x341510218402502eULL,0x812f0448804e12e0ULL,0x3200824a00161202ULL,0x42200008801000ULL,0xefffdfdffef7f7eeULL,0x804881400440a0ULL,0x61025808441408ULL,0x800020000000028ULL,0x2446680814230660ULL,0x2e71e60f614f56aULL,0x2a03510006e208ULL,0xe061176894961682ULL,0x243e0570c004156aULL,0xe3b9ffbfcafe54eeULL,0x40001000001400ULL,0xa000140000ULL,0x48000040002ULL,0x4000a00080101008ULL,0x80800000000240ULL,0x4821280002210e0ULL,0x0,0x0,0x0,0x20020400001402ULL,0x40200004000004ULL,0x21009458044412c4ULL,0x2020000000000140ULL,0x80617240808010a8ULL,0x7c8908607480146cULL,0x308c402b4100028ULL,0x200004000040002ULL,0x4000002000401000ULL,0x2008000000c000ULL,0x3029d0200052e0ULL,0x60804200400a0ULL,0x1000020080001000ULL,0x88000000040080ULL,0x100900800801040ULL,0x44004002000ULL,0x80706080340414e0ULL,0x2150000c060010c4ULL,0xbfbdeffff49e77eeULL,0xf5fffffffefff7eeULL,0x7bfe7e6234d3f46cULL,0x89e00b1b2c46f06cULL,0x43422c4800104028ULL,0x2108080000423004ULL,0x20d966f244b45468ULL,0xffbeadf1f4b7f6eaULL,0xdfa7ff3f9676f2eeULL,0xcf76fb0f047e12aeULL,0xabcecd5afcfcd56eULL,0xf6a686eaf412f2eaULL,0x979b7ffbea4cd46aULL,0xb394057420c235e8ULL,0xccfeef1bb67777eeULL,0xffdfdfdfecb6f6eaULL,0x953f7fafe072f46cULL,0xf5f7f776b67ff6ecULL,0xc2e25614426502eULL,0xa30eca29b443366eULL,0x88811b5b36041660ULL,0xb65aef7ae4fe54ecULL,0x859c05008004524aULL,0x8104394c84541024ULL,0x5393ffe3a4f2366aULL,0x86a137ab43374a4ULL,0x8c4e800012520aULL,0x10080000c000ULL,0xa020c4024401020ULL,0x4010642000068ULL,0xa85000ca400616e8ULL,0x3d78cff9621556eaULL,0xb13148a000535662ULL,0x61e23de0a4a4f56aULL,0x99c46f3a202272acULL,0xe7c6362a161454eeULL,0xb76cd9d23c1e17ecULL,0xd47777e8a6be76e8ULL,0x8c727f6734a6766aULL,0x52b06607a49550a8ULL,0xeef6c201a41d546cULL,0x2807b50a4067026ULL,0x820410200801248ULL,0x59c068408406f0a8ULL,0x2082c632014c0a0ULL,0x4fc080028080d080ULL,0x9530ac38414d08cULL,0xc034904801e4024ULL,0x416885528006d0acULL,0xe02265c016221648ULL,0x81b0f293025d146aULL,0x42c09652a044142aULL,0x992acf91a674546eULL,0xdcdeb6f4e2b7d4ecULL,0x5ba4058044ec542aULL,0x8106e06eb6104264ULL,0x444a10a0000054e8ULL,0x21e0380200410a2ULL,0x2c08000224000ULL,0x4f06500800631e8ULL,0x8a0008001002ULL,0x844bb748a41e34e8ULL,0x3b4205c8e81230a8ULL,0x86307448144110eaULL,0x800205628452520cULL,0x302026400001268ULL,0x746a602041650a4ULL,0x23f51cd18ed635eeULL,0x62e18e820d414e8ULL,0x218c25f20802742aULL,0xd0006824c404302aULL,0x40001000401000ULL,0x210014034000020ULL,0x809804a3201264e0ULL,0x1040a65086220060ULL,0xfe342571f425d468ULL },
	{ 0xa9543516715c07c2ULL,0x802101010000400ULL,0x38beefe371fdafecULL,0xde79ff7df1ffb7faULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd9ddbfda716d03faULL,0x0,0xe2d5d31051bcb2d8ULL,0xcedbffdf71fdb7feULL,0xeffec7da71942ecaULL,0xbdfffffdf1fdbfdfULL,0xaede6df0517faff2ULL,0xfb97fffff1ffbffbULL,0x5fffffbbd1ffbfffULL,0xbfffdfdf71ffbfffULL,0xfffffffff1ffbfffULL,0xfffffffff1ffbfffULL,0xfffffffff1ffbfffULL,0xffffffbfd1ffbfffULL,0xab7ef9dff17dbffaULL,0xebfa5de1717f17dfULL,0xf8fce771717cbfd0ULL,0xfebfff7e71ffbffbULL,0xf7d7dff371fda7d8ULL,0xeffeffdff17daedcULL,0xdfbbfffb517db7f8ULL,0xf1fef7edd1ff8effULL,0xfb9ebf7751dfbfdeULL,0xc1f6fbfd515d87faULL,0xef7f7fff71fda7deULL,0xff9ccfffd1ffbffaULL,0xfcfff67df1ffb7ffULL,0xf7ae7d73d1dfbffaULL,0xfbbbfb7f71debffaULL,0xfe5bb7dbf1f50ffaULL,0xffbfdfffd1febffaULL,0xf7ffdefff17db7feULL,0xffbf7f7071ffb7f2ULL,0xffffef7f71ffbfffULL,0xdfffffafd1fdbffbULL,0xffffdfdbf1ffaffaULL,0xfefffffdf1ffbfdeULL,0xffff7f7b71ffbffaULL,0xdbbe7f7f51ffb7deULL,0xcffaff3b71feb7ddULL,0xefffdfdff1ffbfdfULL,0xbfff7fef71fdb7fbULL,0xfbffff7ff1fda7d8ULL,0xfe8fbffef1fdbffeULL,0xefdfffff71fdaffaULL,0xebdfffdf71573efeULL,0x9ff3fffb717fbfffULL,0xd7fff77ef1fdb7faULL,0xf7227df0d1753efeULL,0xebfff37f71fdbff9ULL,0xfefbbfcf71ffbff8ULL,0xbfff7dd671d63efdULL,0xd4f777fbd17dbfd8ULL,0xffeeffe271ed97efULL,0xebf7773e71fdbfd3ULL,0x8ea92bb71142e10ULL,0x819fcf79d15fa7e2ULL,0xfdd2beea51df8ffaULL,0x17b2ffe1d1f59762ULL,0xfa93e71e115d87d4ULL,0xf5ffe3f5157bfd9ULL,0xafbecb50517d9fd3ULL,0xdef5f6f8517fafc7ULL,0xc7fef77251fcb7f9ULL,0xd29c2d5c5114a7d8ULL,0x8f92fa0070340258ULL,0xa3dfddda51f587fbULL,0xfdbf5fef51fdaffaULL,0xf5ec2e78f1fdaf4aULL,0xbc2cef6e51dc2fd6ULL,0xeffadf3651798ff1ULL,0x223fce6b51571af2ULL,0x8eaa56f0505d9758ULL,0xf5eb37eaf1358fd7ULL,0x5fbebff771dfbff8ULL,0xc3dffbff51febfddULL,0x6fffb65351ffbfdeULL,0xa7f6c9ff5175af79ULL,0xfd7fff6ff17faf7bULL,0xfffe7ffbf1ffbfdeULL,0xbbd6ff77f17fb7fcULL,0xcdfe3e3971ddafdaULL,0xbfbfdf5f51dfbffaULL,0xddfed7ef717eb7daULL,0xfbfffd6371fcafdfULL,0xdbbce62a700ca778ULL,0x9fe79baff1ffb7daULL,0x3fffdf5bf17db6d2ULL,0xf4b6feecd179bff8ULL,0xf9eeff70d15fb7caULL,0xdf9f7fff71ffabdaULL,0xffdfeffe71ffbfffULL,0x69dd5dd9f1fd2ee6ULL,0xeeffdffb717fb7f4ULL,0xf9bda7fe715faf5cULL,0xbeb62f6d514db7daULL,0x4b5bf732506da7d9ULL,0x2ddfbff6517587f2ULL,0x16127ff671ff1ffaULL,0x90fab759317d9f5aULL,0xfbbe77edf1cc87feULL,0xe9fffffe71ffbfffULL,0xfffffffff1ffbffbULL,0xfffffffff1ffbfffULL,0xbe7effff71fcbffcULL,0xf7be3df9f1e5a77eULL,0xdbf7ef7f717fafffULL,0x860e3211d1350ee2ULL,0xe9af4fd170d5376aULL,0x4124271d030aec0ULL,0x41882f11107d0b58ULL,0x10000b0000210242ULL,0x463d80850100290ULL,0x6844b831115244aULL,0xd274e95151b248ULL,0xf0864570d053be41ULL,0xdab6b77cf19dbffeULL,0xa14ed204518cb140ULL,0x1a4ece9b717997c4ULL,0x100010241040000ULL,0x24000011000000ULL,0x400c00a020400680ULL,0x8120060820340200ULL,0x480c1a5c4104a36aULL,0x410000100601ULL,0xf021546990580fd5ULL,0x202012451a4000aULL,0xd9d4df7af01c3ed1ULL,0x2044b948510f036aULL,0x80008a040100090ULL,0x94648440501102c0ULL,0x0,0x840030041400200ULL,0x2046800000580a00ULL,0x8400000200200ULL,0x0,0x100500100180200ULL,0xdc042a025124b152ULL,0x42020110848040ULL,0x727f0bca5130aec2ULL,0x4794be05054bd88ULL,0x5440100080c0ULL,0x0,0x0,0x14000010250ULL,0x1430400000248042ULL,0x9000200111002040ULL,0x40a0000008240ULL,0x4110c00441008080ULL,0x2080b0800044014aULL,0x8083c62150503700ULL,0x20270241081218ULL,0x88043402405c0208ULL,0x200000200100800ULL,0xe006188040411300ULL,0xfeffffeb51fe9fddULL,0xd777fefd71dca7dfULL,0xbb367f32516fb3daULL,0x5306372771592342ULL,0x8b4bbab95167b7f8ULL,0x0,0x2044c010188200ULL,0x13228743103c0f58ULL,0xf3beffed71bcaffeULL,0xc0d6ffb651dda3d8ULL,0xb5ede17718c97d2ULL,0x4b0717951509bd0ULL,0xf38bb7e35159bbd2ULL,0xdb967f7d51f537d8ULL,0xeddeef3f517fa7dbULL,0xafcebd5fb01cbed9ULL,0x9ad05ce8d11cb698ULL,0xb175ff6a71e5b773ULL,0x72961de651e6b6eaULL,0xc4d8f1b7015e0b7aULL,0xbcc89105001b2d0ULL,0xef3c57fa51bcaed3ULL,0x80607648519c8200ULL,0xf6005d7e5104b56bULL,0x6104e00510189645ULL,0x46863880106c0680ULL,0x48140400180ULL,0xc810a22910188ac0ULL,0x44220c60410083caULL,0x2154000811180610ULL,0x8a421b6a5045b3e8ULL,0x91b04b6051749672ULL,0x32566041408142ULL,0xcd22bd42d14423caULL,0x4220418a01940158ULL,0xcfd7be1b514db3f8ULL,0xbb7fdd9b511da7d0ULL,0x14b33aeb717da7caULL,0xc9fe2561714da6caULL,0xdf9e7ffff1fdb7fdULL,0x8a4ce88a50fc8350ULL,0x40161883203026d0ULL,0x2100012040202040ULL,0x502d34e25145a24aULL,0xc000251100610240ULL,0x4848962870341210ULL,0x25312c0208d8240ULL,0x86324f607105a4c2ULL,0x11ae4d00510106f0ULL,0x74b20d4050d42240ULL,0x8970d3e1615d87faULL,0x4881160131508360ULL,0x1591c580101ca5d8ULL,0x30ff1ae1719da652ULL,0x49247621416da3e2ULL,0x8018404010088680ULL,0x4881930151208240ULL,0xe70a4d935120a3c2ULL,0x414200050000480ULL,0x402254040040248ULL,0x4000240d51000640ULL,0x1002121111850040ULL,0x82030000420240ULL,0x41b2f6ff716ca6e8ULL,0x10a074040340a10ULL,0xd314aa134144235aULL,0x8052161b511ca7f2ULL,0xa2561f4a507c33d0ULL,0xa4a37eea7144834aULL,0xd05026d0701586d1ULL,0x96141d4841848661ULL,0x6906c00251d40202ULL,0xa54008081100a040ULL,0x14b37ce8715dae14ULL,0xc810008030000680ULL,0xffb6666c517c97cbULL },
	{ 0x81568b4490219550ULL,0x10884a2310e4a3a2ULL,0x4ce4dff89867b452ULL,0xf77ff76ed2faa572ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb0114ec21a26b544ULL,0xfdfff6efdaedbdf6ULL,0x80010000020ULL,0x6184f13680acad40ULL,0xcdfbff7fdaf2fff2ULL,0xfe374cb29830bd50ULL,0xf1b3befbdafbad66ULL,0x1fb23d509a70fd00ULL,0xdbfeffffdaffed76ULL,0xdfffffffdafffff6ULL,0xffffdfdfdaf6ff76ULL,0xfefffeffdafffff6ULL,0xffffffffdafffff6ULL,0xffbeffffdafffdf6ULL,0xefffffffdafffff6ULL,0x6ffebd5f9afebf76ULL,0xfeff7ffbda2afde6ULL,0xfdffef7a9af3fd54ULL,0xffbf3f7edaf8aff6ULL,0xefdffffedaf6bdf2ULL,0xeddfdecfdaf4fd56ULL,0xfebedf7b98fdff24ULL,0xe6ed376392b2adb0ULL,0xfabe7ff692f2b5d2ULL,0xe9f7fecf98b1bdb6ULL,0xffcf3ffb9af1afb2ULL,0xbfbfcfdbdaf4bdf4ULL,0xf7fff77bdafdfd76ULL,0xdffe3ff39aecf766ULL,0xfbf6ff7fdafbbdf6ULL,0xfefeff7f9afdad46ULL,0xffbfcff3dafdef76ULL,0xfcfffffedaf7ef76ULL,0xf7fffff5daf82f76ULL,0xefbeff6fdaf1ffc4ULL,0xeffffedfdaf1eff6ULL,0xebffdfdbdabcfd70ULL,0xf4fffffcdaf4ff76ULL,0xffefff7fdaf6fdf4ULL,0xfbbb3f7f9af1bd66ULL,0xfdeffbbedaf4ffe6ULL,0xbfbe9fdb9af22f74ULL,0x3eff7de2daf7f770ULL,0xfffffffedaf5fde4ULL,0xdebe3fffdafaff76ULL,0xc9d6f7e79ab7bdd6ULL,0xbeff5e5e9af1f772ULL,0xbebbffffdabefde2ULL,0xfffff7ffdaf6fd56ULL,0xfffe7fffdafffdf6ULL,0xfbdfff7f9afffff6ULL,0x6ffffffedafffff6ULL,0xbbbeffdf9af3fff6ULL,0xd4fe7e7b98e5bde6ULL,0x7fde777bd8b8fd36ULL,0x33b68777d0b2a526ULL,0x9c8b1e49dad0af22ULL,0xee3edff09af0af70ULL,0xbc50aeec98e5bf46ULL,0x96a9bd6512f4ad46ULL,0xdfb2bf7fdae6bd66ULL,0xcffffe6bdaf4fde2ULL,0xebce87d7da51b546ULL,0x76fe57efdaf4bd62ULL,0xc366afe89060afa6ULL,0xdaf72f7bda34bd36ULL,0xefdfff7bdab4e5e2ULL,0xef9edbdfdaf2e5f4ULL,0xfbbf5fe7daf8ef66ULL,0xdbe6ffffdae0ade4ULL,0xfebe3fed92a7fd70ULL,0x877b7b129ab03d54ULL,0x2c76f75f9af0a554ULL,0xb7a0d6fbd8a12d12ULL,0xf5ffbf7bdaefbf66ULL,0xfff63ff7dafb2d56ULL,0xe9d4f31f9282ad40ULL,0x4be61fff98e7bdf2ULL,0xfe9f7ffbdafc3f62ULL,0xf4f7fef6daf4b564ULL,0x56e67d6392a42534ULL,0xfbadfb4e9af46752ULL,0xeffb7fff9affbd62ULL,0xefadcff39ae2af04ULL,0xf5f7ffeadaf0fff2ULL,0xdbbfff71dab5bd66ULL,0xfb9ef73e98f4bcf4ULL,0x1be3bba390a43d76ULL,0xbfff5fdedaadfdd6ULL,0xf6fee7f0dae6b730ULL,0xd76ab7689ad1bd06ULL,0xdf9c7fffd8fcbff6ULL,0xefffff7e9ae1fff6ULL,0xe7fc5f59dae42f24ULL,0xf5ff5df7daf5fde6ULL,0xc93cf77a9a78a766ULL,0xf3302b66d836e570ULL,0x79defe22d8e1eff0ULL,0x75fadbc1d8f4ad74ULL,0xa67b5fe7daf5b5d6ULL,0xb5ffff79da38ac76ULL,0xf7fe3ffddaf7b574ULL,0xfbffffbfdaf6fff4ULL,0xffffffffdafffff6ULL,0xfffffffbdafffff6ULL,0xfefffe7fdaf4fdf6ULL,0x5aa82d60d0602606ULL,0xfbffffffd8f9ef56ULL,0xf3c2374bd2e66534ULL,0xffffdffbdafef5f4ULL,0x54812dea9ae0a574ULL,0x128ac1088404740ULL,0xc989e72692b4bd60ULL,0x200200002004ULL,0x22d6cd4482706700ULL,0xf7f5fc6d9ab4fd64ULL,0xd1ee2c00d0a02406ULL,0x808014892000404ULL,0x10800b1088100000ULL,0x2064c0682a42504ULL,0x94b5243018c0b560ULL,0x91a3a58592a0a542ULL,0xa89a5d719a70dd00ULL,0xa8078a299a80b562ULL,0x1000020980402604ULL,0x1001129090800040ULL,0x810000000400400ULL,0x2184090800000ULL,0x8800110002200540ULL,0x0,0x8282040090008110ULL,0x8c14549092202544ULL,0x40000000400ULL,0x0,0x846109188c82600ULL,0x0,0x48008400000ULL,0x0,0x0,0x300080190000400ULL,0x210c000080008800ULL,0x48418400010ULL,0x1088002000240400ULL,0x0,0x0,0x0,0x250160088408900ULL,0x0,0x4000001010000400ULL,0x0,0x404c00a92002124ULL,0x10090000000ULL,0x108024018c00000ULL,0x4c04520698c00000ULL,0x910d70890222402ULL,0x340c0002800100ULL,0xb6ee4df398e9ad56ULL,0xd7fff6fbdaf4fd76ULL,0x7b767ff098b4fde2ULL,0x9b56eb2e9a6fad74ULL,0x884a520198c01570ULL,0x8316d9e11024ad72ULL,0x50563ae090b0b414ULL,0x9b2235079030bc50ULL,0xfb40ff2d90bc2400ULL,0xeff7f60e90bead64ULL,0x8d1c5b529ae0a500ULL,0x84f676e3d2b1fd20ULL,0xfd7f7fe2dae5bfe4ULL,0xfa9e7f24dae0b700ULL,0x8ddbcf1a9ae5b524ULL,0x377f5fdbdaf7ed64ULL,0xfbff7feb9ab5ede0ULL,0xfffdfffb92f4bdf4ULL,0xfef67fedd2f0bdf6ULL,0xc79af71e92a83d32ULL,0xa956ee4a9af0a560ULL,0x45a66df49a262550ULL,0xd2f93de0da70a710ULL,0x42163c71d230a552ULL,0xc803b2a292202500ULL,0xc20a1090000520ULL,0xc89a4da2dab02500ULL,0x15d3b67a9a30f520ULL,0x709e3c5192209d80ULL,0x110e2208080a902ULL,0xe6c3db88daf0ad64ULL,0x2322ce909ac82d30ULL,0x107a56e0d0802560ULL,0x11aa3d628230b546ULL,0xc9106f6edae53544ULL,0xcdc7ee098af0ad46ULL,0x75f5edfd8d0ed10ULL,0xdefc76fd9873b7c2ULL,0xddc6f530d2a42d74ULL,0xd7bc7fd792f1bf50ULL,0x41d5c52898f82c02ULL,0xa441a4002602342ULL,0x451000000242002ULL,0x80652c4080a48c80ULL,0x8000202010000030ULL,0x8970c8039a70b400ULL,0x856566098402160ULL,0x16405dd14264b500ULL,0x8118002080008504ULL,0xbe363704d0e6e540ULL,0xc8767f7e9a67a7e6ULL,0x60920c02d0202120ULL,0x80c6d40d0202514ULL,0x81515e299ac0afd0ULL,0x48960d309a622540ULL,0x8121e9028031b960ULL,0xc1da3ec198a8a594ULL,0x1980c0000242104ULL,0x71a702689a240d00ULL,0x1000004010800000ULL,0xa85e5229808a092ULL,0x1800080000800ULL,0x81080c039080a540ULL,0x847172680a60af00ULL,0x8c00438810303100ULL,0xc8102e7282602700ULL,0x1c7fe02d0a4b540ULL,0x6254514b90e02766ULL,0x84d47b22d022a904ULL,0x4100801010208010ULL,0x82a2054290202404ULL,0xc163f6a19a326d60ULL,0x62121042c8e40000ULL,0x499c520c8782040ULL,0x70458550d220b500ULL,0xfa9e3f4e90b2ad72ULL },
	{ 0x8020e01429084e63ULL,0x210190c20d442640ULL,0xb09b4fe0fd287562ULL,0xf6776ffe8dda6d72ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb4ffffeeed7eff63ULL,0xf5f6f7ff8ffe7f73ULL,0x201000000000a000ULL,0x4000408004000ULL,0xeedbbfff3dde7d61ULL,0x7bbccff9fb5eff62ULL,0xb57e4eae0fdaef72ULL,0xbffefdf3ffeeff62ULL,0xfb9677fdddfe7f73ULL,0x5fdffebfffeeff73ULL,0xbfffdfdbfffeff73ULL,0xfeffffffeffefd73ULL,0xfffffffffffeff73ULL,0xffbfffffdffeff73ULL,0xddeffffeefdeff63ULL,0x6fffdfd3dffeef73ULL,0x34b76df64fe86f43ULL,0xfbefff6beffe7f63ULL,0xfeb76f660f7e7f61ULL,0xefdfef376fde6763ULL,0x3f7fffff4fdeff63ULL,0xbfbbdffbcf7ef763ULL,0xffddf7de9ddeff63ULL,0xfb567febbe5e6d62ULL,0xe213e24b49ca7160ULL,0x6ff76f7fdfea6772ULL,0xffffcffbffdeff62ULL,0xfffff779658e7f63ULL,0xfafe7dfb1f7eff63ULL,0xfbd4ef264f5eff73ULL,0xfffbfffbefdcf763ULL,0xdfffcffb4fdaf771ULL,0x55ffffefcfde7f63ULL,0xfffe7f73effeff73ULL,0xffaff775efcefd72ULL,0xffffff7fffdeff63ULL,0xefffdfdfdffeff73ULL,0xfefffffbdf7eff63ULL,0xcfeffff2dd7e6f63ULL,0xff9a2f7ecf5e7f63ULL,0xef9ffe3eefbaff63ULL,0xfffffddfffdeff73ULL,0xf5ff5feb0fda7753ULL,0xf7fefffecbfeff62ULL,0xdebeafffdf7a7f73ULL,0xcfffff3f7f6a6f63ULL,0xeaffdfffcfcefd63ULL,0xf7f367fa4dee6f61ULL,0xf7bcb77b9dce6f73ULL,0xddfe7fff5fceff73ULL,0xfbfffffffffeff73ULL,0xfffffffffffeff73ULL,0xffbedffb0f5e7f73ULL,0xb7b6f2feeffc7f63ULL,0xeffeff624f4e7f63ULL,0xebc5ef17ad4ef760ULL,0x8df3bf4de95a6f62ULL,0x73be87120860ff62ULL,0xd3c5c742054aec70ULL,0x3faeff62adde6f73ULL,0xfbb7ff7f5fea7773ULL,0xcdfefe7fdfdc7f63ULL,0x2bdf4ddf817a7763ULL,0x1ef55eeeaf467763ULL,0x9f6fb65a57aa6f63ULL,0xff9c6f7dddcaf773ULL,0xefdfff93ed3e6d60ULL,0xa9da5d12cd5e6d63ULL,0xb4ff77eb8dfe7f63ULL,0x99ecff72976a7773ULL,0xfeaeff7e7fee7d63ULL,0x49cff77ecd7e6f73ULL,0xa3ffdb5b39ae5f72ULL,0xf4ba5ff2ed5e7f63ULL,0xd5fdf7e9a1fafd62ULL,0xfb767764ad6e6d62ULL,0xabebfeffef3efd63ULL,0xabfe3fefbfca6563ULL,0xb7bbfff3af7eff63ULL,0x94f7f7eccdfc6560ULL,0x7f766db3edcae563ULL,0xdbbcfb2f29fa7761ULL,0xedcbff7f8fde6f63ULL,0xefbedffb4fee7f73ULL,0xd7fff6fbeffe7f63ULL,0xdf3eff71cd3e7f73ULL,0xe314fbff4b5c7762ULL,0xedffbe1bbdfe7d62ULL,0xefefdfde2fcc5d63ULL,0xdcbec4f0ae4c6562ULL,0x914aa7702dba6763ULL,0xdeb46b7e8dee6d73ULL,0xe5f6fb3bedde7f73ULL,0x2ff7dd59c7de2f60ULL,0x143fcef36f5ae562ULL,0xc17db7f2fdce7f63ULL,0xdcb6177ea79c6563ULL,0x6edce7627848f563ULL,0xa37b3bd35c46ff63ULL,0xbfbf7f7a8d5e7f63ULL,0xd3fffffdaf3ef763ULL,0xdfb62ff6ed5efd63ULL,0xfbffffbfeffeff73ULL,0xfffffffffffeff73ULL,0xffbfdfffffeeff73ULL,0xfff6defffdea7f63ULL,0x13ee6f614c1ae761ULL,0xebc4f96a0d9c5b63ULL,0xac5fbb134f4e6d73ULL,0xffbfdffbeddeff73ULL,0x74d1fefb0dfa6f53ULL,0x4a3a2c620bca4622ULL,0xfc84ff1569ca7463ULL,0x1000020000000820ULL,0x69dd99540cbc6142ULL,0xf6ffffefdfdeff63ULL,0xa06d272424ce4760ULL,0x808d08480b064442ULL,0x8120800000040421ULL,0x1131658a1882d22ULL,0x10824d0015020140ULL,0x5000900000000740ULL,0x3c950d54cc486560ULL,0x2114118800080d62ULL,0x36e10038d427b41ULL,0x466005000940ULL,0x2024000008400820ULL,0x6080052005003100ULL,0x80800002008a0040ULL,0xa12020302004740ULL,0x0,0x2890122012084700ULL,0x4082044000880622ULL,0xc894275200085663ULL,0x30a0084002100ULL,0x200020000000400ULL,0x548008040000ULL,0x108150004080560ULL,0xc2102b1201124d33ULL,0x8000200008040000ULL,0x5859f553050e5c23ULL,0x84848060094a0540ULL,0x42304089026000ULL,0x0,0x4002c00001001800ULL,0x0,0x0,0x0,0x0,0x84200249881000ULL,0x806200448004412ULL,0x0,0x0,0x8a04112c00984160ULL,0x82c0653424cae161ULL,0x28c216d904dc4162ULL,0x75b9cfdb01ce5673ULL,0xb4e7f26fcf3aef62ULL,0xf946de21fcee762ULL,0xf2a8e30605084573ULL,0xccc27e3dcdde6f72ULL,0xe14c64c008a4960ULL,0x1cf35ee68f1e6772ULL,0xf5be3f63cd18f563ULL,0x8995636f5f4c7f62ULL,0xefcfffbbbf587773ULL,0xaffedf1ecfdeff71ULL,0x94b5d6e5cfca7c42ULL,0xb63eb77109da7d62ULL,0xf7bebf6b2d6e6d63ULL,0x4cdeffa48d447571ULL,0xebfddfdb9b5a7f72ULL,0xfeffffffefdef763ULL,0xfffffffe7ffeff73ULL,0xfebf2f7fff4eff23ULL,0x4ddb97209c407741ULL,0xbbffffdfdffe7f73ULL,0xffaa4dfb1fde7763ULL,0x6030832009004e60ULL,0xa1001b004204143ULL,0x8044280080182160ULL,0xc8c27e048aa86e73ULL,0x24a3480000426500ULL,0xc414126007100540ULL,0xd4aa4d402cea0561ULL,0xc208610008884c60ULL,0x8c212174a080d60ULL,0x1236065a4bfe4562ULL,0xf7b7f46acd8a6543ULL,0xbb7efff2df66ff71ULL,0xf692ef6b59567f73ULL,0xefeebfb7fd6e7772ULL,0xafded7d21f5e4d23ULL,0xe47474e84ccc7542ULL,0x6bae8d622e48e761ULL,0xcebd3f7e07def560ULL,0xe042405004084760ULL,0x14420004100140ULL,0x10100201080900ULL,0x40200004001ULL,0x82820b40830a2400ULL,0x7140128a040ULL,0x2441630010a0550ULL,0x92428108484540ULL,0x109544800044560ULL,0x52083d541c04a560ULL,0xeaf5f3ef8dbe7f62ULL,0x42c2122d4c586120ULL,0x1899cdca494a4f60ULL,0xb5a7f26e399aed61ULL,0x4a0604200c080550ULL,0x43c4e5320c886f61ULL,0x6e22802484c5160ULL,0x20c480020082020ULL,0x24008040000ULL,0xc9982442c0ca6541ULL,0xdbc0677e4e5e7560ULL,0x944168845006c32ULL,0x31ad2d2c80e2722ULL,0xf4932ea40fda6542ULL,0x812002480008a430ULL,0x2b02c00050e0160ULL,0xec48de9c882c7473ULL,0xa5a180901483320ULL,0x34a0c50000342f60ULL,0x800025020c40e100ULL,0x4220041601084b40ULL,0x4002802808004402ULL,0x0,0x808010200084021ULL,0xd410905549487220ULL,0xc39e2dd0494a6d62ULL },
	{ 0x8000930803908000ULL,0xc01f6810960d20ULL,0xbfffffe2dfdefce9ULL,0x7df6a6f656bf3ca0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffff6feb7fff7ee0ULL,0x7ef67e6a5a9e7ee0ULL,0x200c0044020408ULL,0xca90430201c02c40ULL,0xffffbefffffeffa9ULL,0xffffcf7adb9e7d69ULL,0xa4d757e917933e68ULL,0xbbfcfff3cbff7fe0ULL,0xff9f7ff7fbfefe68ULL,0xdffbffbfffffffe1ULL,0xffffffdfffffffe9ULL,0xf7fffeffdfffffe9ULL,0xffffff7bffffffe8ULL,0xffffffffffff7fe9ULL,0x5bc7f7aa479e7c40ULL,0xefffffdffbffffe8ULL,0xfffedfffdfbfff69ULL,0xde7efffedf9efea9ULL,0xfefe2fff7ebfffa1ULL,0xcddffffe5ffe7de9ULL,0xaa5ededfdb976da8ULL,0xdfc3dfbb5ffa7e68ULL,0xd1efbf734ebe7c28ULL,0xed1a7feecef73da0ULL,0xe38cf32c4a262ca0ULL,0xfff73f9fdfb7ffe1ULL,0xffbfcffbdfb67ea9ULL,0xddfff7fe5f9f7fa0ULL,0xeffeff73ffbffca0ULL,0xfb95ef7efefe2ea9ULL,0xfeeffffffffe7ea0ULL,0xeffecffbdf9eefa8ULL,0xb5fff7efdb9e7da9ULL,0xfffeff737f9f7e28ULL,0xf9beffff66ff7fa9ULL,0xcfffff7f5fffffa8ULL,0xffffdfdfdfbf7fa1ULL,0xfdfffeef7ffebee9ULL,0xbf7cb77a5fdf7da0ULL,0xdbb66ffe6f9f2ca8ULL,0xeffeef9afebf6fa0ULL,0xeffeffdbfffffee0ULL,0xbffbdfefdfbfffa9ULL,0xf9fefffedfdeffa8ULL,0xfeb6bf77ffbe7fa0ULL,0xebf7ffdfffff7fe8ULL,0xcbffff7ffb9e3ee9ULL,0x9fbfffff5bfebfe0ULL,0xf3ffb7e351de7fa1ULL,0xfffebd7fefffffe8ULL,0xebfffffffffffee9ULL,0xffffffffffffffe9ULL,0xfffffffffffeffe9ULL,0xddfff6ff5ffeffa0ULL,0xfbaa7d61f7d6ec20ULL,0xabb5f3d6de9e6c80ULL,0x6e43377fcadf3e01ULL,0x2a1bcccecad63d80ULL,0x15e7f66cd6b6bda0ULL,0xdbfe7f726f9f7ea1ULL,0xfeb7fffeffdeffe1ULL,0xdf7bfeffff9f7fe9ULL,0xef5bdfd7ffbe6da9ULL,0xf4b6fdfbdf9ebfa9ULL,0xfbf5f77bdb94fea0ULL,0xfa34ef7e5fbdbda8ULL,0xef63ff1e5ebf3f60ULL,0xebacdfdf73de7fa9ULL,0x649ef3eb5296fee0ULL,0xd5ae27f04fdd3ce0ULL,0xfebe2f7fce9e7c60ULL,0x8dc5e837ce9f3c20ULL,0xeffbffffdefe6fe0ULL,0xbf9b676a7ed66ea1ULL,0xf779b6ff4e9fbfe0ULL,0xfbfe2f7dfb9e7ce0ULL,0xc3bbfa150bbe3e80ULL,0xfffe7f7b7f9f3d81ULL,0xfbbecf7a569fbde0ULL,0xd4f77efefa973ea0ULL,0xdbfaa5727e9ee680ULL,0xf9fdfbbe7ffeef80ULL,0xddffbf79d6bf7d80ULL,0xffffdfff7fbebfe8ULL,0xfff7feeedbfebf28ULL,0xdbfe7e737b9e7ee0ULL,0xdfa6ffdc579e7e01ULL,0xcdf6febfdbde3da9ULL,0xe272cc8fc3be7c80ULL,0xd4dd7eefde9f7fa1ULL,0xdbfeaf7bd7d67de0ULL,0xffb4be77dbde7f81ULL,0xfde6eb2beffe7da0ULL,0x3bfeffdac79e7d61ULL,0xffa25f6f1e9f6da0ULL,0xf170664f1f9e2d20ULL,0xffbe3fdfdf977fe8ULL,0x8084f3324e366ca0ULL,0xd946fffbc7b27ea0ULL,0x9fb7cfffdfdf7f88ULL,0xffffffffdfbffde0ULL,0xdbf67dfecfff3fe0ULL,0xebfefb2ffffe7ea0ULL,0xffffffdfffbf7ee9ULL,0xffbfdfffffffffe9ULL,0xf5f6fbf9d79e3de1ULL,0xeab675630a9e6ea0ULL,0xcbd2c36a43863c08ULL,0x48c2fec947963bc0ULL,0xa7bfdff3dbdefc41ULL,0xff1fe9c3be6f80ULL,0xc1a4ff6052ae3d80ULL,0xca85d50045943481ULL,0x4a92da02e286e488ULL,0xebffdddbfff6bfe8ULL,0xd5bebffc5bb67680ULL,0x5101004002902400ULL,0x519c585240e02e80ULL,0xe402f0a090fc4c80ULL,0x1415d4850970ca0ULL,0x1d069bc78a8daf80ULL,0x91d6bf425e907480ULL,0xa8121e040147080ULL,0x182c041509e2d80ULL,0xb04040080480ULL,0xb0824f8030d64400ULL,0x6000924081860800ULL,0x4810001000100408ULL,0x0,0x8042000200802041ULL,0x2010000822020ULL,0x8041226910840200ULL,0x10041110820800ULL,0x3950c77210bd2c61ULL,0x5892a2850862260ULL,0x20c4802ca03ac00ULL,0x247156699b8b3900ULL,0x20080000800200ULL,0x0,0x50000010800000ULL,0x8080a00820a00ULL,0x8010000a000000ULL,0xa04202082400ULL,0x80284003080400ULL,0x0,0x480000800040ULL,0x0,0x0,0x0,0x4d0000000810840ULL,0xc20a1900901420ULL,0x80040002000400ULL,0xc110a0124a002d80ULL,0x774e77788797bc20ULL,0x303e90a12ac3620ULL,0x2fd71bdeec9e6ea8ULL,0x1fbdfffb3b977ea0ULL,0xf1f7d5fed7fe2ea0ULL,0xe2ae75211f982d08ULL,0xf880db3cc29e64e0ULL,0xcfc6be036edc7d60ULL,0x124e0210865600ULL,0x456086f451862ea1ULL,0x932a7d61d79e2ee1ULL,0x8306fd2857b72c60ULL,0x89d7d6b24e963f60ULL,0x6fff5f41ca9efea0ULL,0xf4f677666b9e2da1ULL,0xfb6cfff0179676a0ULL,0xc2b03e1363b27c20ULL,0x4e433eba52d67620ULL,0xaa3a17583bae7da0ULL,0xd47b7fefdebfbc20ULL,0xfdced7dc9a9eace0ULL,0xf6b22f64569ea981ULL,0xe158f5e8569ffda8ULL,0x871c7a004b82aea0ULL,0xbef7eff253bf7ee8ULL,0xd175fee95bd66f28ULL,0x805041100b902000ULL,0x4958d00266163d00ULL,0x84e1600419c04c0ULL,0x88b0868801962061ULL,0xd085b66992962920ULL,0xa8604425d930c20ULL,0x8b004120000d2880ULL,0xa5403a071f141900ULL,0xa6ffcf7233be3e20ULL,0x75e55efacf963fa0ULL,0xc5920f7062d66ca0ULL,0xe8a967ff59963fe8ULL,0xcdfff619dfbe7ca0ULL,0x2bf61fd63bfe7ee8ULL,0xf4bf5ff4df9a3f21ULL,0x4b0efd2102963c20ULL,0xc3953d674e3a7c21ULL,0x2002000900000ULL,0x9ec0940400b0d80ULL,0x20b3626a4e167c00ULL,0x990c0528820d2c88ULL,0x8280058100062d20ULL,0x8044bf260e942d28ULL,0x10034040002c00ULL,0x4004fb006322c00ULL,0x0,0x286bc40163cbd01ULL,0x89a4f20a82fe2d20ULL,0x2cc20aa1008a3c20ULL,0x8b9cdfd144b63e01ULL,0xb467f7d843be7f00ULL,0x40885040128c2620ULL,0x80e4c300029a3c20ULL,0xc0e3ba10de8c0c00ULL,0x2000008040000400ULL,0x22026110802800ULL,0x50207502c6100480ULL,0xc9beef6f5fd63f20ULL,0x84c3380000903d40ULL,0x7086811214123c20ULL,0x90284f62169c3c01ULL,0x800740102826c00ULL,0x888073200e9c0400ULL,0xeb46a204cb1a3ca0ULL,0x38c1d821b882c29ULL,0x1010140004100000ULL,0x44020840040080ULL,0xa11406104206a880ULL,0x80f13a433e6420ULL,0xa00004042000480ULL,0x290052010960080ULL,0x2004000800200ULL,0xd2025952229e3e20ULL },
	{ 0x8013198094224aULL,0xbffa5fc8a3146b53ULL,0x9fbdcffed7bc67ffULL,0xf7efb67ab7942bfeULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x480000040200ULL,0xbbadefebb7b4276eULL,0xb530f77c961c2b65ULL,0xffbebffbb39e6f7dULL,0x600082004004ULL,0xeffbbffff7be6fffULL,0x7fbd5ffad6a6275fULL,0xfdff7ffed79c2f5eULL,0xfbfeff73d7b667ffULL,0xd60caba0c6a4635fULL,0xeffffebff7be6fffULL,0xffffffdff7be6fffULL,0xfffffffff7be6fffULL,0xfbffff7fd7366b7fULL,0xffb7bffff7be677fULL,0xef97ffbaf7aa236eULL,0xaa5f5f5fd7b66747ULL,0xbebf7ffff6b46fffULL,0xd9f6f77ef79c67ffULL,0xfebe3ffff7b46f7fULL,0xedfffb3ff6be677fULL,0xbffffbdbf696635cULL,0x9fbb4ff2f7b423ffULL,0xf93dfcf9d73e2726ULL,0xda0e3f62d68623eeULL,0x88c7fb76d7042363ULL,0xeff7bbdbf784277fULL,0xbfffdd5bd7b66bfeULL,0xfdf7bff7d79667feULL,0xdfbe7d79f7b42b77ULL,0xfbfeff3ef79e67feULL,0xee6ff7dfd73c6f7fULL,0x7fbedffbf6bc67ffULL,0xfbf7f7eff7bc23feULL,0xfbfe7d77f7bc67feULL,0xf89ee77ff7b64b7eULL,0xeffffebfd79667ffULL,0xff7fdfdff7b66fffULL,0xf5feffefd79467f7ULL,0xdd7f7ff3f6b46fdfULL,0x9fb67bfbd2b62fdeULL,0xcddfff2e973e23d7ULL,0xfbfeffdbf7b46b7fULL,0xdebfdffff6b4635eULL,0xf7ffff7cd7be2ffeULL,0xfebf3ffff7be6f7fULL,0xffffff3fd7bc6f7fULL,0xaffeff8ef7bc2fffULL,0xbbbb7feff69e2fffULL,0xf9fd36faf6b66befULL,0xe7defffad7b26b7eULL,0xfbfffffff7be6fffULL,0xfffffffff7be6fffULL,0xfbbffffff6b46b7fULL,0xdd7dfe7ad69c6f5eULL,0x5e24ad4092302f44ULL,0xe990e32e93046f2eULL,0xb2fa9e11d7346a20ULL,0x2592d903930a23ceULL,0xfdffb7ffd2b66fefULL,0xcd2efc73f7362b7fULL,0xff9f7ffff7b66fffULL,0xdfdffebff7b4677fULL,0x63ca9fdfd7342b76ULL,0xdcb776ecd6b6274eULL,0xb7a8b749f1342f7eULL,0xdb94ffe4d78422d7ULL,0xaef7f793f70627dcULL,0xab525e17b2326b67ULL,0x24b23f6bb2a06f4fULL,0xdbedbec496b60bc6ULL,0xeebf3ff1d6946756ULL,0xe19af39ef2b42356ULL,0x976bedbf616237cULL,0xfab3fdbaf6b263dfULL,0xf1f57feed734275eULL,0xbfec3f7fd69e2f5fULL,0xebdcfbbcb684234dULL,0x6fd7be19f61a2b6fULL,0x9fadcf9bf5a42f62ULL,0xf4f7f7fff7366fefULL,0x134c2d60d6042765ULL,0xebe7e76bd7362742ULL,0x7fdfffdff71c6fffULL,0xffbecfffd6be2fefULL,0xddfefeeff790676fULL,0xebbeff70f5942f76ULL,0xfb8777eff794636fULL,0xca32feaad414235eULL,0x3be64d48d6006b44ULL,0x74d6eeecf6a82bc6ULL,0xd9ae3df7f7104b64ULL,0xdab67fffb69e4b6eULL,0xeccefb6dd4b02773ULL,0xebffd7dbf6a66f5dULL,0xb6b7dffbd786675eULL,0x99ceff77d78a234fULL,0x7a9a2f4ef4942746ULL,0x8fd5b624d5a42a4eULL,0x2fffbec9d6be63cfULL,0xbf9fff7bd7b4636cULL,0xb5cdd56bd0bc6b7eULL,0x7ffa2f52f6946f7eULL,0xc9feffeed7be2ffeULL,0xeffffffff7be6fffULL,0xbffffffbf78e6fffULL,0x75f6f6ffd7bc2fdeULL,0x53aeac6193a8235eULL,0x98f1cb3ed710275eULL,0xedcfb15ab3a4276eULL,0x7f080b39b38c274fULL,0x84bf0278b3144316ULL,0xb96f7e73b2bc0b78ULL,0xfe98bdfef716615eULL,0x5fcbfe2bd6342f6eULL,0xef7fdfdff7b46f7eULL,0x642c4ad0d6102350ULL,0xd37f2f63c2126716ULL,0x9a8c436ec5842b42ULL,0x4688c62093000292ULL,0x80e0fd0c2146202ULL,0x420000040200ULL,0x4860ce62d7002342ULL,0x848043081900292ULL,0x8003804180002b40ULL,0x148080082202041ULL,0x9292024890a02a42ULL,0xfa741c0191b00252ULL,0x88200d4080042040ULL,0xc2045766f2042671ULL,0x4de20141c3802102ULL,0x8804180002400ULL,0x140010200040ULL,0xa200d0110002440ULL,0x8004000080000008ULL,0x0,0x44210200040ULL,0x0,0x40200080002000ULL,0x0,0x4044001091000200ULL,0x18038480104100ULL,0x240411600600004cULL,0x1200a18081002302ULL,0x30060010000300ULL,0x8400000002001ULL,0x0,0x1001000010000020ULL,0x0,0x0,0x2000800202000ULL,0x800180000040ULL,0x20204800200034cULL,0x100052004000048ULL,0x100048400080004cULL,0x8902e938d6002158ULL,0xa652320fc680245eULL,0x8338cf51d2a62746ULL,0x94fd777c929027efULL,0xe2627d42d6002334ULL,0xa9d3772fd3386b5fULL,0xeaf3ba8fd6a4225dULL,0xcb3dcfded3942fd7ULL,0x305326e996a4214eULL,0xdbbbe766f294275cULL,0xfffeff77f79c2fffULL,0xfddefe33f7be6f5fULL,0xb9265992d3860366ULL,0xbcd77e6af6be6f7fULL,0x9f38ff42b2142f47ULL,0xca3077f656842f66ULL,0xcfe2f3b4d598234fULL,0xa2dddf5cd7342b55ULL,0xbdfa5fffb21a2fcdULL,0xfdfcbfeeb3bc2f7fULL,0xc2a73fe6f7342b4eULL,0x8590a313d3b02257ULL,0x1008020180000050ULL,0x361a6c2252882356ULL,0x50a0120cb4080348ULL,0x201401208000620cULL,0x8112120010200340ULL,0x3527840d4a022d8ULL,0x39a6f7ab3842f66ULL,0x8450023491102020ULL,0x43222f6283100316ULL,0x8b040464e6040216ULL,0x2c5e224ac6900254ULL,0x39ecfd7d70e4b5eULL,0xbd77fe6bf3bc2bd6ULL,0xcb37fd72f7302f47ULL,0xf7107771d6380a4eULL,0x4ecedf13f69023ffULL,0xeadecf5fd4b4276fULL,0x5420066993046107ULL,0x37826de093942346ULL,0xdaa82f61d2042307ULL,0xc800c11392044160ULL,0x800010080040000ULL,0x2cb356888084260eULL,0x9018a07453a00146ULL,0x8a024098004020eULL,0xab57f02892b442d4ULL,0x108024044800040ULL,0x8012408190002048ULL,0x1084550083242202ULL,0x91040820c200001cULL,0xe986ff2bf6982736ULL,0xa0420650c004220cULL,0x1a2bcf9bd6a42b56ULL,0xd7feb6efc7b427c7ULL,0x80c050070082200ULL,0xc202b44080900b48ULL,0x67d27ecff7b82f17ULL,0x0,0x11202000900240ULL,0x2100010000200ULL,0xc110432cd610011eULL,0xcb42b2a2f0a02d5bULL,0xa000c80290800042ULL,0xd2148802800106ULL,0x31c26752b10c4336ULL,0xab9c6f3fd736631aULL,0x4282c20002102700ULL,0x3471040908c0262ULL,0xb04028a0004010ULL,0x10402646b0000b10ULL,0x90042530c4046602ULL,0xe4a3f114f71423a7ULL,0x1029880a2000104ULL,0x84a2004090006205ULL,0x2002024012000020ULL,0x1000a4042020002ULL },
	{ 0xcbd7c35b890473eaULL,0x27447a5a8ba852dcULL,0xcfbdefcbfbf9fffeULL,0xf57ff6ecdbfd53efULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6af77fdffbb4d73fULL,0xdfbedfdbfbfe7fffULL,0xdd70aa70d9b87b78ULL,0xfe3e7df3cbfb5bfbULL,0xbbe57f7fdb2557eeULL,0xcccffebbcbfa5fefULL,0xefbfcff2fbc47ffeULL,0xf5b67eee8bdd736fULL,0xdbfaff71f9b2d7bfULL,0xdfaeff7ecbfc57bcULL,0xffffffbffbffffffULL,0xffffffdffbffffffULL,0xf7ff7ffffbffffffULL,0xfbffef73dbfd77feULL,0xdb367f7e7bf8777fULL,0xedffffffebec77afULL,0xefffffdbfbf7dbfbULL,0xb7ffffebdbfbffbfULL,0xd9fdff7ef9b5f7afULL,0xfefe6fffebff73ffULL,0xffffff7ffbf8fbffULL,0xffdfff4ecbfa57aaULL,0xb6bbfffbebb5ff3eULL,0xffedbff8cbfcdfbdULL,0xdee67f67dbf579efULL,0xfbdeeb4f8bffc3ffULL,0x7ffebbfbfbf17b6bULL,0xbfefcffbfbf1d7bfULL,0xfff7f6faeba94fbfULL,0xbfbefff3cbfaffbeULL,0xfbfdff3fcbff7fffULL,0xeefbbffbcbfe777eULL,0xffbffffedbfcd3ffULL,0xf5ffdeeff9bfffffULL,0x9ffe37f2fbd87b7eULL,0xff36ffebf9f673feULL,0x4fdfffbffbf67fffULL,0xff7fdfdffbfff3ffULL,0xf7f7fefffbb75ffeULL,0xddeffff3fbe1ff3eULL,0xffbf7f7cfbfc5ffbULL,0xcfd7ff7ffbff53beULL,0xcbfeffdefb91f3ffULL,0xf0bff5fbdbb867bdULL,0xfb7faf76fbfdd3ffULL,0xfe9e3f7fe9f5533fULL,0xffffff3cebfe57bfULL,0xadffdf5bfbf3f7afULL,0xbbba6797db75d33eULL,0xf7fca7d5cbf457b6ULL,0xffb67ffdc3b6fbbfULL,0xcbfefffffbfffbffULL,0xffff7ffffbffffffULL,0xbfbfeffefbdf7bffULL,0xf4f6fefffbfd5bfbULL,0x468633608b9a1329ULL,0xdb56f72cdbb9772fULL,0x5cc7fb11e3a0f634ULL,0x6a8c85989e05321ULL,0x74fd3ffcfbbcd3beULL,0xbafc7d55dbdd7727ULL,0xff9effbffbb67ffbULL,0xcfdffffbdbfefffcULL,0x8a1ecdd78ab9777eULL,0xf7b7ff6df9fd737aULL,0xf9eebf30d9e3577aULL,0xdebc2f7e5b9247e6ULL,0xcbd7faa6e9b1536eULL,0x6ff65b9bf9dc537cULL,0xd5defb6bebb44b6eULL,0xf3fcff74eb96dffcULL,0xfeb52ffef9f273fdULL,0xefdff727dbfed3ebULL,0xaffe5ec7e9dc73afULL,0xfeba777efbb4d3feULL,0xf5eff7e7eb94d32fULL,0xd9562d7069a6f265ULL,0xc994f2ffd95d456eULL,0xebdf5fd1fbf6f17cULL,0xb29ccedfc9b47332ULL,0xdcf7feffebfe737fULL,0x6fae7c61dbb453fcULL,0xdbe6cf5cd9f073a7ULL,0xee6bbefbf2a4ffffULL,0xbf3fdf52db9473aeULL,0xfddffefffbf775bfULL,0x9d3aef60e96cd32cULL,0xab40655f8bcc633aULL,0xc9fef6bffbe45feeULL,0x2a76db8d89f80363ULL,0x54fef6e8c9ec5369ULL,0xd5ca3d72c8e41374ULL,0xd29c7e06c997333eULL,0xaf8eff8afbf05b08ULL,0x696f5dcbc9f072e7ULL,0x3c3779f309e1637eULL,0x7d6cff72f9b9fb3eULL,0xdedc2fd45bd557aaULL,0x8defe73acba0d3b7ULL,0xe955d74feb827963ULL,0x23284ff2c3805b2dULL,0xf7fdaf7febf87bf9ULL,0xfef46dfffbb953faULL,0xc3ffffffcbbaf3bdULL,0xffdf3ff3fbf577ffULL,0xbfbf7df7ebf75fbfULL,0xf7fffefefbfc7fffULL,0x32aebf61899853afULL,0x9bd0493ee8ac5221ULL,0x4c4313a84bc85312ULL,0x8c25056a8919032eULL,0x90802ee8dae8510cULL,0xc96c2a0083b0130cULL,0xe0b4e3028bd80a00ULL,0x4d82b380518440f1ULL,0x959c6d888e4132fULL,0x9482162401044764ULL,0x8940d430e8e8130dULL,0x899047e0c100580cULL,0x10250000000ULL,0x21010485b004000ULL,0x10b944e003924268ULL,0x96082400d1001040ULL,0x9082110261800118ULL,0x184120409504000ULL,0x800200080200000ULL,0x8222002849004040ULL,0x8010055048801201ULL,0x202041051000060ULL,0x200100001200ULL,0x66c603c9c9c0710eULL,0x20805080001120ULL,0x20204101044020ULL,0x220020214801130aULL,0x20008000200ULL,0x0,0x0,0x40040001000008ULL,0x8061204001906000ULL,0x8100200003004008ULL,0x0,0x800080001000008ULL,0x20040001000020ULL,0x2003200b004000ULL,0x89b3337f88b47bfcULL,0x4201420080001328ULL,0x10024480141020ULL,0x8140020c0101102ULL,0x9b18521389d0431eULL,0x200200040000000ULL,0x8180920501100328ULL,0x80202018124c028ULL,0x500cd9b619042a1ULL,0x4200042000200200ULL,0xa047180086442a0ULL,0x8000f122e904c32cULL,0xbeebbfd3dbf07fa9ULL,0x1189015a09944320ULL,0x92f7a757eac97f37ULL,0x11387d2288b946b2ULL,0x99d4af6dfb59534eULL,0x6deb38b3ebb45b35ULL,0xbf9f4ddbcbf0dbfcULL,0x71121eefe130636eULL,0xd9bef674dbf877f9ULL,0xef9ff77febfff72fULL,0x495bff13fbff73faULL,0xefdccfcbdb907ffaULL,0xd471a6f48b7f73a2ULL,0xb34037d3eb837a0fULL,0xcba23fdd8b805337ULL,0xc42bc639e9b047bfULL,0xeafad35beb95d33aULL,0x95fb5ffbdbf27b7dULL,0xd3ccaa604b99f20aULL,0xfe803d6e8b81dfb3ULL,0xc7c2e61edbd01256ULL,0x463304980b0132dULL,0x2924e4009b80338ULL,0x91b4166b88c10f3aULL,0x24426402006520ULL,0xc9f6e60afb38436eULL,0x66985fd2c950432eULL,0xacde55f8fbf152a2ULL,0x52642ec8ba07770ULL,0x2d7e7d63dbb4737bULL,0x184e02201806159ULL,0x820818a104200ULL,0xfdf8c733e9f0730cULL,0x65fef76cdba05f28ULL,0xffb9b741dbe8d332ULL,0xcbf6f36ffbfa7bddULL,0xcfefff7efbfff7b7ULL,0xab9ecfdbebfbe67fULL,0x241166841305200ULL,0x81ce57408b644228ULL,0xdb9e7f37fbd45b38ULL,0x6140b08082105110ULL,0x7b3ccddafb94537dULL,0xaba456a0b146228ULL,0x91200c4489200330ULL,0x928c2b0080a00b18ULL,0x50c20880201200ULL,0x8188cac2010043e8ULL,0x1a13031b09206268ULL,0x8008140000800020ULL,0xd5005d66c9124b24ULL,0xca91c312e1b01329ULL,0x88d61fd109d5f223ULL,0x8c87d7e1ebb9d13eULL,0xf1f5ff7b8bf9e33aULL,0x5b2a0f028bd0236aULL,0xc3126420408470a1ULL,0xc8cb9e67eb84562bULL,0x2110040001204200ULL,0xc0105000c0301000ULL,0x883a6541c1301220ULL,0x49226612d0842a2cULL,0x84f22290ca00510cULL,0x8504d4080a00121ULL,0x588061100102ULL,0x9222a50088c44d20ULL,0xcb903f3cc97853eeULL,0x4481221251060128ULL,0xee02dc538b6a4304ULL,0x24c3d76a49a072a9ULL,0x3204a660ca00610cULL,0x9602084062840229ULL,0x8983b0048b980228ULL,0xc0e2d94261a41631ULL,0xf23a4fc3595a438aULL,0x50283610c3b10002ULL,0x800084201800060ULL },
	{ 0xeac5f2e589cdd22eULL,0x24ce1ff281ced346ULL,0xb7ffdfffcbd6fbfeULL,0xfcb6f3ea8bc5e35aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x85bfa8083c6916aULL,0xffffffffebdffbfeULL,0xfbbecf538bd7cb6aULL,0x3dc0927983d6f972ULL,0x796e77f3cac7c34eULL,0xfbddff3fabddfb6aULL,0xeacfbfdb8bdfc33eULL,0xbf9acf788ad7e14eULL,0xf5dfdfefcbd7e36aULL,0xdfbffff3ebcdc33eULL,0xfbdfffffc3d7db6eULL,0xfffffeffebdffbfeULL,0xffffffdfebdffbfeULL,0xffffffffcbd5d3feULL,0xbdffef7fcbdffb7aULL,0xefbd7fffa9cff3eaULL,0xfdebff3eebdffbfeULL,0xffdfdf5f83dfda7aULL,0xfcbf6ffbc9d7e37eULL,0xff7ffffaebcfebfeULL,0xfeb6bffecbdfd3eaULL,0xefdeffffe3cfe3feULL,0x7f7affd583d5f33aULL,0x3dfffdfeabcfe1ceULL,0xffffffffabddf37eULL,0xffd6ffe6abd5d3eeULL,0xebfeff6febdff33aULL,0xfeffff7ccbc5f33eULL,0xbfbfcf7bcbd7e3faULL,0xf7fff7ffebc7f2faULL,0xfffeffebebdffbbeULL,0xfbfefffeebdfebfeULL,0xcffbbffbcbd7fbeeULL,0xfffddfffcbdffbeeULL,0x54f6f7ffebcffb6aULL,0xcffefff3ebd5d36aULL,0xffbeffffcbcffbfeULL,0xcff7ffdfebd7d36aULL,0xff7ffddecbddeb7eULL,0xdcfedeefebcdc37aULL,0xfffe77f3ebc7c3eeULL,0xdfbf3f7fcbcdd3faULL,0xffffdf9aebdfd3eaULL,0xead6dedbebd5f3eaULL,0xbefedfebcb5efb3aULL,0xdfffff7febcff36aULL,0xdea53f7fe1defb7aULL,0xedffdfebebd7d3faULL,0xaeffffedebc5cb8aULL,0xaf9acf3a42c6c32aULL,0xf7fd7f79a3cff96eULL,0xffbe3fffebddf3eaULL,0xebffffffebddf3feULL,0xefffbfdfcbdffbfeULL,0xebfffebcbd5d366ULL,0xdcf3fefb89d6f3faULL,0x57be5d7283c6e3f6ULL,0xdb14ef2c89dfebe2ULL,0x62cfb6096985c16aULL,0x2fbec97a4147c36aULL,0x22056eac1d0e30aULL,0xffffff73abd4c1aaULL,0xe7b9f7ffcbd5c3faULL,0xffcfffbfcbc7fb6aULL,0x1b7e5f5bcbc5d1faULL,0xf57a7af589d6f39aULL,0xffbef772abc5db3aULL,0x9b247776cb44935eULL,0x69dffb6ee3c5f3eaULL,0xaffedf5bc3d5e36eULL,0xffffdfffebd7f36eULL,0xdbfdfedccbc7e36aULL,0xca9d2ffecbd7f37aULL,0xe75bffbe8bd5c36eULL,0x8f5f5f4ac3c4d33eULL,0xbeb55dfecaded37eULL,0xd6e535f383d5f06eULL,0xdbee7f73e9d5eb2eULL,0xdb81d20f63d5c24aULL,0x6f46065a42d7836aULL,0xbcbacd6b0bc5c22aULL,0xfcfdd3facbd6f25eULL,0xffbe7ff0a3c5d36aULL,0xfbc4f79ee9c5fb6aULL,0xaed9fac7abd4f16eULL,0xefbe4bbe8bd7f30aULL,0xbfffbeebebd5f32eULL,0xcfaefd732bddc33aULL,0xd7be7f33a8c7d02eULL,0xcb4cbe85c0cdca6eULL,0x8afc9fdb89d6d3beULL,0xd4b9e6e2a384eaacULL,0x90cbf79c905d10aULL,0xcb90767fa0c5c33eULL,0xcac7ea00ab86ca32ULL,0xaff75ccf83d6d346ULL,0xd6bbdaebe1d7ca4aULL,0xdb6f377bebd7c3feULL,0xceb67d72a985e23aULL,0xef76e77acbc5fbcaULL,0x7bfedffbe3d7ea6aULL,0x339afff2eac5c11aULL,0x93d2a7d2c9c7a25aULL,0xff86f77eebddcb6aULL,0x8bd6fb3bebdfd3faULL,0x6b5fbffbc3d5f3feULL,0x1fbccffb63cdc3eaULL,0xf7f7f7ffebddf37eULL,0x5bee6de24bdcc36aULL,0x8a0283228250b020ULL,0x8febaa5aeb0de30aULL,0xe21e4f0acb85d14aULL,0x50801c8003540082ULL,0xd1702a4000558208ULL,0x421261244981c08aULL,0x4a43ba908205f116ULL,0x71e4d9768c5c202ULL,0x4a3562400c18028ULL,0xc0a6b6608140e12aULL,0xdf9e3f58e9c5fbe6ULL,0x6c033120bc4411aULL,0x2106dcd00041812aULL,0x905e410350434aULL,0x5342067a8350ea40ULL,0xa0344f00018122ULL,0xc100881040018100ULL,0xdc33e488bcdc87aULL,0x22082e98003a142ULL,0x940835380805c12aULL,0x3000280000440002ULL,0x851d1aa8a00b322ULL,0x80580a4700848228ULL,0x1498859309d58262ULL,0xc621667d0bd5c20aULL,0x70ce74602384c132ULL,0xca02616e23504312ULL,0x4ccb1e344354c092ULL,0x15046000444008ULL,0x5000000000040002ULL,0xcd4cf52282c7fb62ULL,0x1000080c8080a000ULL,0xa1c4ca2103d4020aULL,0xebc60bd3abc4c39eULL,0x401000000808000ULL,0x480805202881c002ULL,0x0,0x500206000800800aULL,0x204080208440008ULL,0x4000102100540200ULL,0x804000000102ULL,0x0,0x40a006080404808ULL,0x0,0x18600008058000ULL,0x1100050082001002ULL,0x640001f88a04a822ULL,0x100001001004000ULL,0xae7fbed9abcfe34aULL,0x75bb63fa83c4d31aULL,0xdec3f67f8ac5937aULL,0x5a3e7c33c3c6fb4eULL,0x89f5c70ecb84e362ULL,0xe83bb028287c322ULL,0xad9edf52ebdce36aULL,0xb4d57efa49cec37aULL,0xf7bf7f76a3dfdbeaULL,0xfeb2ff6fcbd5db7aULL,0x6fc7ae8b8bd6f33aULL,0xfe7edfdbe9d7e2eaULL,0xb4f6dfed09d4c37aULL,0x9bcdfb7acbc5c22aULL,0xf336fb7cc3cffbfaULL,0xc1cffe9bc1cfe16eULL,0x9ded5d3cad5d32aULL,0x9cb67d99e9c4c366ULL,0x73eeb76ac1c7c2daULL,0x9faa3f478b07e27aULL,0xc9f3a03603c7832aULL,0x100607408050080aULL,0x81364ce28154c33aULL,0x0,0xa8366472e285d122ULL,0xc002a0200a440002ULL,0xc5d33652c104f252ULL,0x1283459e81c5ca2aULL,0x472ffe6a8ac4c11aULL,0x922c0d82830fe192ULL,0x200400800008108ULL,0xd40323140c1817aULL,0xf9264d24c244d34aULL,0x9cf77fec8bc5cb5eULL,0xd8aef567c3c4fb3aULL,0xdb9eefbae9c7f36eULL,0x8fd7fe2febdde35aULL,0x6bff5fd78bd5d34eULL,0x10a404488080d002ULL,0xcb4d6400815cf222ULL,0xdbfcaf7e83d7d25aULL,0x40d2f3298144e312ULL,0x27fe5fdb8bdcf32aULL,0xb8020d938105c00aULL,0xa51a0a4088c4c816ULL,0x50100408240ULL,0x5300e31c83c4c30eULL,0x208001100c4801aULL,0x8080300038040ULL,0x40001004000ULL,0x48023c30080bc072ULL,0xdb00f32ac8c1c00eULL,0x5b5f1e534bc5c30aULL,0x4aaccfe3ebd5c36aULL,0xcd5754c1e005e942ULL,0x490008000104000aULL,0xc0c1280105c302ULL,0x8cc3ea0041d1c132ULL,0x0,0x5584305ac887c152ULL,0x938c9771c14dc37aULL,0x10230581405802ULL,0x44ac2100103c1e0ULL,0x1008014001008208ULL,0x4a801024040ULL,0x90c35200184c302ULL,0xda200c42e2c1c82aULL,0xbc8008080a100ULL,0x2a8ecd4a8984e322ULL,0x9c90f5ab82c4f30aULL,0x90100e0009008202ULL,0x4e8826448145c052ULL,0x0,0x75e5e5888c6cb4aULL,0x93b27f7281c6f3ceULL,0x8200480008104ULL,0x76b2246083069332ULL },
	{ 0xc082401b47020848ULL,0x6800060440828600ULL,0xe7fffff2cfdb7eeeULL,0xc122a2a806814e00ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe3d7f6eecfcb7efcULL,0x7effbe9f4fdbfeeeULL,0xb8badf7bcec3ffe4ULL,0xf4efb369cdcb5ee4ULL,0xefb67f73cedb7e7eULL,0xfbf7ff7fcfdbfef4ULL,0xcfdffe7fcedb37b4ULL,0x6fbfcfdbcfdbfefcULL,0xf5ffffefcfd3fff8ULL,0xbfb8aff1cfcb7eacULL,0xfbbffffecfcbfefeULL,0xfffffffb4fdb7ffeULL,0xffffdfdfcfdbfffeULL,0xbcfffefccfdbfffcULL,0xfff77f7acfdbffe0ULL,0xffbe7effc6dbff36ULL,0xffffffffcfdbffbeULL,0xffffff5fcfdbfff4ULL,0xffbfffef4fdbfeb4ULL,0xffffffffcfdbfeb8ULL,0xfabe2fff4ec37fb0ULL,0xcf8fffbb4dc3beb4ULL,0x2efff9d7c4d3fee2ULL,0x9d3b777bcecbfeb0ULL,0xffffffffcfd3feb4ULL,0xffbe7ff94ddb7ebeULL,0xeb4effffcfdbffb0ULL,0xeffbbfdfcfdb7af4ULL,0xb7bfffbbc7cbfefeULL,0xfdfef47dc7dbfeeeULL,0xfffe7ffbcfdbfefcULL,0xdbdfff7fcfdbfffeULL,0x8fdfffdfcfcb7efcULL,0xfbbfcffbcdcbfba0ULL,0xfff7feffcfcbfbf6ULL,0xfbffff77ced3fefeULL,0xdfbfffffcddb7eeeULL,0xdfdffef9cfdbffb6ULL,0xeb5e9f5b47c37aa4ULL,0xd6bffefccfcb7bf0ULL,0xfbff37fb4fcbffe6ULL,0xffbeff7f4fdbff2eULL,0xdfd7df3fc79b3ebcULL,0xffdfffdfcfdbfeb6ULL,0xbd7ffded4ddb7ee0ULL,0xd143ff58cfd3bf22ULL,0xdeb62f72ccc27ee2ULL,0xc9fedbf6cfda7ebaULL,0xfdfeff4f4edbfee0ULL,0xbf3a7fefcec36eb4ULL,0xd2e7f7ffcdd2ff74ULL,0xffbe7feecfdb7eb4ULL,0xebfff7f7cfdbfaa0ULL,0xffffffff4fdbfffaULL,0xdebfdfebcfda7f24ULL,0x96d696f04d9afe10ULL,0xfffe6ff24fd2fef4ULL,0xfb187756c0dbbf64ULL,0x4f4b22eb4cda5284ULL,0x4e775f704fcafec6ULL,0xf6d55fe74ed3bef0ULL,0xdbfeff724fd2febcULL,0xffd2ff7f87db3ee0ULL,0xdfdebebbcfcbffacULL,0xaf4edac2c7d3bfe2ULL,0x50fe56e8cdc1da18ULL,0xfd6ffff6cfdbbeb4ULL,0x9b181f7e86db7ad2ULL,0xaff7fffacfdbfff8ULL,0xcbdbd7c7cecb7aacULL,0x9dbf5ffdcfdbfea0ULL,0xf9ecbe784fcb7e24ULL,0xfebfbf7fcfcb7fe0ULL,0xe9fcfab68fdb3ea0ULL,0x1ddedede4fd87e28ULL,0xffffdfffcfdbfffeULL,0xf557f76fcddafe86ULL,0x7a8636a64cc8d6b4ULL,0xabc588588f88faa0ULL,0x4ec7be0b0c8af622ULL,0x9db9cfeeccd2fe22ULL,0xf97ff97fccdbfeb0ULL,0xff2eede2c7dbfaa6ULL,0xdbd6efaecfcbbfbaULL,0xacdfffffcdcb7fb4ULL,0xafffcf6a4fd37eb0ULL,0xfdf7f7facfcb7f30ULL,0xfffefff3cfcb7ebeULL,0xdbbccf7e4fc37ea0ULL,0x4dcfbfb547db7ae4ULL,0x9bffcf5b4fdafec0ULL,0xd437f4ed0fc36a72ULL,0x91e3b76247cb7ea4ULL,0x9a8c667e47cafa6aULL,0xcdd7fe37cedafff6ULL,0xebdfde5fced3feecULL,0x5c977feb46db76a6ULL,0xd16ca7e04f8b2fa0ULL,0xfa982f624ec3feb8ULL,0xe3cbfff78fc37ea0ULL,0x22d113db45dafaa4ULL,0xb21f4ffbcfcafe82ULL,0x58c4a7400cc23a42ULL,0xdf1e3d62c6d37aacULL,0xc9b7fbbf4fd37a8cULL,0xcefb3fdf4f1b7fa4ULL,0x9f3ddffac7d3faf6ULL,0xdfdbf7e8cfdbfeb6ULL,0x5a7275728edb7b20ULL,0xe993ff2a8ed27aa2ULL,0x8c18928b84825220ULL,0xde7757f605c3feb8ULL,0x11ee0400c4822030ULL,0x1236091044c02830ULL,0x6808640944c25e20ULL,0xc419b62041d33860ULL,0x47921f124cc32604ULL,0x84b966604d822e34ULL,0x1048ca60c98a9a04ULL,0xd79c633c0dcb9e00ULL,0x800020440022080ULL,0xe12f90141cb9a30ULL,0x44a2109400a00ULL,0x9700664040827112ULL,0x529e3feec7cafaacULL,0xcf50f01a8188be02ULL,0xb58024202c1f282ULL,0x91bac7cb4783beb0ULL,0x2173276b0dc9fa00ULL,0xc4122db2c4c21e04ULL,0x81b3f7774dcb3fb2ULL,0xc86020000407080ULL,0x410440001400a00ULL,0x810000020200ULL,0x14000002080ULL,0x800202804881000ULL,0x42220040000410ULL,0x403c04a04382ca00ULL,0x844808081200ULL,0x8062400020222c0ULL,0x100400000101000ULL,0x602200008010a00ULL,0xa80014001485800ULL,0x404000040000400ULL,0x26004005002880ULL,0x20404800000ULL,0x40210804420200ULL,0x0,0x0,0x0,0x4000200001000800ULL,0x8000300284802000ULL,0x273100046800604ULL,0x8092048a44828044ULL,0x98c10c6047807800ULL,0xaa8444040410634ULL,0x801810000410800ULL,0x2e50100804404a08ULL,0xbeef5da8c48abb64ULL,0xbd9d74ef4fcbbea0ULL,0x7b647d71cfcbfab4ULL,0xfbd5f97acfda7aecULL,0x7fdbbbebcfdbfba4ULL,0xbf7f5f2dcfc26fa0ULL,0xdcfbf7ffcecbfeb4ULL,0xfbfe7d73cedbfff4ULL,0xfa8effffcfd37f7aULL,0xdfefbe174fcb7ea4ULL,0x2a7f1fca4fcbbaaeULL,0x549c94f4c593faa8ULL,0xd9eebf534dcb3f40ULL,0xdaa475a70f8b3aa0ULL,0x4dc0821846901288ULL,0x2bfe5f894dd3feb0ULL,0x15ac22e0468a6e50ULL,0xd9082f6a4ec3bf28ULL,0xee84277f4dc9bea0ULL,0xcf92fb124ec3f2bcULL,0xb5a7b408ec93aa8ULL,0xb201050840c81610ULL,0x49645242412600ULL,0xce080d0406022620ULL,0x88c5f32d45c3ff0cULL,0xef621ab8cdcb3b8cULL,0x11847a8068a3222ULL,0xb5d676edcac3bbb0ULL,0xc2360d104fd27a20ULL,0x8b18ea2f4ec132b0ULL,0x8001120904c17080ULL,0x1a16cc2906c27e28ULL,0x211656ec46d33e40ULL,0x6fb6ff70ceca7e34ULL,0xeb8fd7bdcf8bfae2ULL,0xcfdfffd74fcb7ff4ULL,0x131adfdb0ecb7e40ULL,0xf4225ee4478b38a2ULL,0x968f2424dc3bfa0ULL,0x8b966f3f43cbfea0ULL,0x40861b0e06c0f810ULL,0x8e204e0004123600ULL,0x1656d703ce822310ULL,0x4900f4438481a620ULL,0xc828040248420204ULL,0x538cd21186c27b00ULL,0x421001040a1000ULL,0x1641f20d825a00ULL,0x8058773206923142ULL,0x4008010200404810ULL,0xd160a32f0cc232a0ULL,0xa812160106ca2a90ULL,0x4c0b5ff34f82ae60ULL,0xb4f576614dcbfe44ULL,0x7beef522cfc3ee30ULL,0xc320c10a40c14a20ULL,0x44e220802402222ULL,0x401801000085000ULL,0x75fd36eb4fca7ffeULL,0x9faabf774edbff3cULL,0xcb88762d4cc236b0ULL,0x4343b22304c32e20ULL,0xa9d9bd3c6c12e80ULL,0x41c482044820000ULL,0x8d45ee0846537aa4ULL,0x9a862c4048c1be20ULL,0x98c1d282004222a0ULL,0x2b54cec00788ba28ULL,0x88ac923405b3a84ULL,0xc50cb670c6826ea0ULL,0x9abe64704dca3aa0ULL,0xd80cdb0c07c8aa80ULL,0xc9559b4ac0c3faa0ULL,0xb5faefef4cd37ea0ULL,0x18d03440448a7804ULL,0xc324f563ccdb3f30ULL },
	{ 0x8810120020021088ULL,0xc5517cc8a721a87ULL,0xfffffffffa76bbbfULL,0xff7ffffffa76bbbfULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebfff3ffea729a9fULL,0x7f7f3ffbea57ba9fULL,0xf77eca76a25612abULL,0xf57ffe7aea7233bbULL,0xfbbe7f73fa76bbbfULL,0xfbfef73ef873ba97ULL,0x5edbbfdb2a773bbfULL,0xeb9ecffbe872b39fULL,0xd5fffffefa77babfULL,0xdfbbff77fa57babfULL,0xffbfff7ffa73bbbbULL,0xffffbfffba73babfULL,0xfbffdfdfea573b9fULL,0xf6fdfffbea73bbbfULL,0xbbaef77bea733a9fULL,0xcbbe7f7fa872b3bbULL,0xcfdfff3fea73bbbfULL,0xabfffbdfea77bbbfULL,0xbfffdfffba53bb9bULL,0xddeefffbfa76baafULL,0xffbf2f76ba563babULL,0xcfdff7f3ba72bb8fULL,0xbfdfffdfea729bbfULL,0xfebfeff9ea563b8fULL,0xfbffff6be876bbafULL,0xfffeff7eea52bbbfULL,0xebf6fbbfba773bbbULL,0x4c5f7fdfea76bb8fULL,0xbfdfeff3fa72bbbfULL,0xddf6f7ffea77bbafULL,0xfffebff3fa77bbbfULL,0xebfff76ffa73bbbfULL,0xeefffffbfa77bbbfULL,0x97fadfdbaa721b8fULL,0xfffff7fefa53bbbfULL,0xfff7ff72fa56bbbfULL,0xefbeff7fea72ba8fULL,0xfdbbe11aa73ba9bULL,0xbfffdfdbfa529a9fULL,0xb6f5ffe9e877bbbfULL,0xfb6fff7ff8763b8fULL,0xdff4fbff2a733bbfULL,0xeff7ffbeba77bb8fULL,0xefff5f5fea77b3abULL,0x95d7ff7ffa73baafULL,0xf97caf73fa773a9fULL,0xfefe3dfeba769bbfULL,0xcfdbff7baa73bb9fULL,0x6fffde5b6a569abbULL,0xbdbaffffaa429a9fULL,0xd7fcff7faa773bafULL,0xfffeff7fba73bbafULL,0xfbf7fffffa77bbbfULL,0xfeff7ff1fa5733bfULL,0xf7befffb6a773bbbULL,0x9673f7307a723283ULL,0xde367dd3ea563a8bULL,0x9bf4ce36f852338bULL,0xf44b1e19a8621a87ULL,0x2c8ddfe3a85212abULL,0xb4b85ff8aa729a9bULL,0xeffeff72ea77bbafULL,0xffe2af7ffa73bbbbULL,0xddf7febfea72babfULL,0xeb9fdfdb6876bb8bULL,0x5cbff6f66a429b8fULL,0xfdffff63f873bbafULL,0xf79ffffefa72babfULL,0xfdf7ffeee877bbbfULL,0x6dfedfc97873bb93ULL,0xb5fe7ff3ea539aabULL,0xfb7ebfe2fa739a93ULL,0xf7bfbfffaa77babfULL,0xe9fff73ea876b2bfULL,0xfbf5fed72a721bbbULL,0x1ffbeff73a523bbfULL,0x91c3bf6f2a621a8bULL,0x794636000840b227ULL,0xcbfefbfeb252babfULL,0x46461b38a8623abbULL,0xb49fcfeb6a501bbbULL,0xd97fe66ee8569bafULL,0xdfbebd729a763a0fULL,0xeb7eea3eba523a9fULL,0xcff79e8fa872bb8eULL,0xe3becff13a723b9fULL,0x789667d8a8529a1fULL,0xffedff73fa72ba8fULL,0xf7bfff7f28763b9fULL,0xc7e4f30fb2709aa7ULL,0x9f9fffd37a729a9fULL,0x12bf7fe7f8563393ULL,0x10cf7e2aa52b2abULL,0xb3947fe52a72aa8fULL,0xd8d29b0de803928fULL,0x2af897d22a62338eULL,0xd4b01d72e0461a23ULL,0xf17efe7aea73ba8bULL,0xd282b65eaa523397ULL,0x8fd4f37dea701a8fULL,0xaed7dfdfaa62ba8fULL,0x34f74fe1a8721a09ULL,0x91793f23e8523b8fULL,0xfaae3de22852ba9fULL,0xd292c22ea8721bafULL,0x65df1e8db8779a97ULL,0xbbaccf736872bbbfULL,0x9ff6b6fce872ba9fULL,0xdbbaff72ea5233bfULL,0xc3d0f34eaa52b209ULL,0xab7bff41aa7212b7ULL,0x43c0561a0721221ULL,0x8110290a101901ULL,0xce9d2de26a46139fULL,0x523c37e928601babULL,0x8140b0800043021bULL,0x2b2a0f4a2a52018fULL,0xd437a06ca242829bULL,0x332a8b2888621aa1ULL,0x210640100121201ULL,0x40c24a00b0021000ULL,0x8a5c7a4aa0621887ULL,0x147615e0a8220382ULL,0x90000208a0020000ULL,0x4a20f40804222a2ULL,0xe757f776e8663bb3ULL,0xbb4fbb4b28511b0bULL,0x801801022002108aULL,0x41002038806202a2ULL,0x59002d94a8629801ULL,0x4000100000000202ULL,0x320100021011ULL,0x1080840a88620080ULL,0x810080020000ULL,0x0,0x840420c10011282ULL,0x363104002500282ULL,0x20120000008ULL,0x102080c0a8020882ULL,0x4010201008000082ULL,0x200100100002ULL,0x0,0x2100400000021082ULL,0x40048000400021ULL,0x40000001001ULL,0x0,0x10200021000ULL,0x800080100001001ULL,0x48040000001ULL,0x0,0x0,0x0,0x140024020020201ULL,0x2010000010200ULL,0x8000004000000201ULL,0x514040600001ULL,0x6000400200ULL,0xa14010880a022011ULL,0x37fb7ceeaa72b29fULL,0xf7fff6fffa52b28fULL,0xfffe7d79ba72babfULL,0xd384effeea72ba9fULL,0xc94bf6f068721abbULL,0xbbb6cdd1fa529abbULL,0xd6775ffce873babbULL,0xdbfabf72ea7332bfULL,0xfa3fe7d728779b8fULL,0xcbd7fa1bea52ba9fULL,0x7ffedfdbea523babULL,0xfefefffde877bbabULL,0x39ae6769ba72ba8bULL,0x8bb41d7e2a621a9fULL,0xd9d2fb2ee253b2afULL,0x6b10dcdb68621b8fULL,0x35bb5fda7a521abfULL,0x7b4fc66a28723b2fULL,0x82b63ff8ea521abbULL,0xc14cf82d6a529a9fULL,0xfe7fbf5be0563a9fULL,0xb3a84fd3c873129fULL,0x905c5eec28023a81ULL,0xe27e6f64ba72ba23ULL,0x8b27eb43ba52138fULL,0x604f5e90a8622a8bULL,0xb8354ea228529a0bULL,0x1142844020031201ULL,0xeb342ed2e852329fULL,0x88c10b1288028213ULL,0x8256b2c2aa42128bULL,0xe63c4efb2a72310fULL,0x53c035aa8121289ULL,0x97062fe3027312b3ULL,0xfaccaf36aa729bb3ULL,0xffeef23b6076b2afULL,0xafdfffdeba773a9fULL,0x74b757e96852989bULL,0x2848627320421292ULL,0xa60feeaa852329aULL,0x84dc00000402226ULL,0x40001020000080ULL,0x8810144820129082ULL,0x412012422800030bULL,0x2862060880202b0ULL,0x4010020080020220ULL,0x0,0x10000202000001ULL,0x2032420a410203ULL,0x8886350020609a87ULL,0xd800982028129081ULL,0x89523c13a8109a8bULL,0x393883b2aa12120fULL,0x543174d188561a96ULL,0x8418008000282ULL,0x8910270a280212a8ULL,0x800101620420200ULL,0x2c002008400210ULL,0xfff77e7cea7733b7ULL,0xdfa5fd5768763b99ULL,0xc020004e80620213ULL,0xa94ff6010a721a83ULL,0x82861a00441221ULL,0x300696812053920aULL,0x8d6a7df2ea731b9bULL,0x52800942a8021980ULL,0x90d00000521200ULL,0x2c411c4008431a01ULL,0x24f846400850988bULL,0xd310cff1e842209fULL,0xdf022966b8421aa3ULL,0xe7aef7b62a56ba9bULL,0xebfffadbe8529aafULL,0xd7fa6fe3ea723abbULL,0xdb74f77f2a773a8bULL,0x421025020840082eULL },
	{ 0xe804ab0263741840ULL,0x88c2381002620140ULL,0xbffb6ffb227c17dcULL,0xff35a67ee7355ef4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82026704403718e0ULL,0xe9fffbffe77f7dfcULL,0xfcd6ff5f677f5efcULL,0x141dc393a23c1e5cULL,0x79ffffffef761f5cULL,0xcb6e7df3677f1ff4ULL,0x8bfcf77fe77409ccULL,0xcfdffff9ef7f5ff4ULL,0xbfbf6ffbe77d1ffcULL,0xf7fff6ffef777ffcULL,0xfffe7ff3ef7f7ff8ULL,0xff9eff7fe77e3fdcULL,0xafffff7fef775fdcULL,0xabbfff5ee77f3cfcULL,0xd4f7dffd637f5ffcULL,0xdffffff76f7f17dcULL,0xdfbf7fff6f7f5d7cULL,0xdf7bff3be77f5ff4ULL,0xaadeffdbe77d1ffcULL,0xffffffff67767ff4ULL,0xf9ffff7e677f1ffcULL,0xdebf2dffeb761efcULL,0xdfefffffe37f3ffcULL,0x7ff1d59e2761c6cULL,0xb7beeffbe37f3ffcULL,0xf5ffbff9ef771dfcULL,0xff6e7ff7e3773fecULL,0xabd2fb7be37f5cdcULL,0xebde3e77ef7779ccULL,0xbfbbeffbe7765fdcULL,0xfdffff7fef7f5ff4ULL,0xfffefff2ef7f3ffcULL,0xfbf7ffffef7e1fd4ULL,0xffffbffe637f1ffcULL,0xbfbfdffbef7f7fecULL,0xfdf6ffffe77f5fdcULL,0x97afffd1e97f5ddcULL,0xffb7fe7de87e1bdcULL,0x81cffe1bee761cfcULL,0xebfeffdfe77f1df4ULL,0x76fffeff6f7f5fdcULL,0xfb67ff7b6f7f1ffcULL,0x9a8cefefeb767ff8ULL,0xfb57ef3feb7f5adcULL,0xffdfff5fe77e3ff4ULL,0xbcff7fefe77f1fdcULL,0xf57dfffbe77f3efcULL,0xfeb6bf7fe77f5ffcULL,0x89d1ff2c6b7c1ed0ULL,0xedbfdf5be77e7ffcULL,0xb61bfff2677c12dcULL,0x977fbfffeb765ddcULL,0xdfbcbd726b757bf4ULL,0xe9fefff7ef7f3fdcULL,0xcfee7fcec07f1f70ULL,0xedfbcfeb6f7f1fdcULL,0x942f3e6be37e3cc8ULL,0x6afeedf9ec743fdcULL,0x8114e73ca5771b54ULL,0xeffbbad7eb7f1fe0ULL,0x8ffdcf71653f7fd0ULL,0x75fff5ebe7773fd8ULL,0xbfb6fff0ef7e5ffcULL,0xffbeff7feb775ffcULL,0xefe7fe1ee7765fd8ULL,0xb18cf5aa77c1eccULL,0xfc7736e663761ee4ULL,0xb3af5768e2661ddcULL,0xd3bfbf77eb7f5fd4ULL,0xc7f7f77f637e15d4ULL,0x1f5f5f53a5763fc4ULL,0xbc1fc7fbe77f5ff0ULL,0xdd3cff73037f3edcULL,0xb2810561627c14c0ULL,0x4fdbeb3a437616ccULL,0xbffdff5bea7f7ddcULL,0x982a5dfee3771bc4ULL,0x52ee9d7466775a58ULL,0xfe566d5d637e4e74ULL,0xabcef7e6af7f0b4cULL,0x6fd7bed56c7f1e54ULL,0xbfff6ffb67755c7cULL,0xf26b76f06d7717dcULL,0xfffe6f71e97f5dccULL,0xcbe4e71f63341bc8ULL,0xfedeb779637d3cd4ULL,0x2ebedffb6f770edcULL,0xd5be76ee63763dccULL,0xeffe7f74e77f1decULL,0xfbaff77de57f1dfcULL,0x8b4fb72f67765cd0ULL,0x2bdddfd7677c1cfcULL,0xbeb4dccce3741f40ULL,0xc566abe863741bf4ULL,0x438d372d687c1764ULL,0xc9c6eb9e67741ac0ULL,0xebe4de4a62767dd0ULL,0x3cd6d6ea2d523d48ULL,0xf7cfe7f308571d4cULL,0xced9236ee6365fd0ULL,0xebf2f23ee1765ddcULL,0x6aff7add647f5d4cULL,0xbebfcfff603f37d4ULL,0xf97d2779c2773954ULL,0x7e9e6df3ef7c1ddcULL,0xc1f2e77a677617c0ULL,0xefd73e5b6b7519ccULL,0x9bafdfbba6753fdcULL,0x95fdfa73623c5ad4ULL,0xdebe6ff26f7f1fdcULL,0x8b42bb24657714ccULL,0xe0f83c1b6a7e18f0ULL,0xa3824d82663508d4ULL,0x1080066001600800ULL,0x812011222041080ULL,0x9210090c64551044ULL,0x4421a08035408a8ULL,0x67071d10627e084cULL,0x2461142022301800ULL,0x8a845000184a40ULL,0xd801b6d237b19e8ULL,0xc053fab064541454ULL,0x9be1df5a257b1dd8ULL,0x15f55f8340761e44ULL,0x9318300000340044ULL,0xc004354200384840ULL,0x6100d528806c4844ULL,0x67d81e81e070484cULL,0x8da223e0026d1948ULL,0x200460000000ULL,0x5a04101202140284ULL,0x4080800400520040ULL,0x4002000000240000ULL,0x8210030040500048ULL,0xb094b51042344c80ULL,0x98c0d60e0347944ULL,0x4020000001400ULL,0xa02320000240854ULL,0x82880060102000ULL,0x3060246280104c04ULL,0x1402d0000180340ULL,0x2b0220080110844ULL,0x818c9002e0280840ULL,0x2110049080100844ULL,0x5090940080210840ULL,0x2004022000000ULL,0xb39c35b0033418c4ULL,0x8120c60a40141184ULL,0x10040200302800ULL,0xa040100000ULL,0x0,0x280010400200a40ULL,0x200020000000140ULL,0x10020024000000ULL,0x0,0x8242001100a00ULL,0x82045c260524c20ULL,0x804002224000050ULL,0x0,0xb7f74ef263764ad4ULL,0xfffffffeef777fdcULL,0xfffefff3e77f7ffcULL,0xdbdef76f637e5dfcULL,0xef73ff5be6644d48ULL,0x8e9fdff86b771dd4ULL,0xf767f6efeb7f3fe4ULL,0xeff6fd73eb777ddcULL,0x598ecf0c0074245cULL,0xcdcf7efee3741bfcULL,0x7fdfd99be37f3ffcULL,0xb6d15e2a627e14d4ULL,0x81df6f7def767cdcULL,0xf2bc3f7e66760fd8ULL,0xcdf6eab26e7635c8ULL,0xaa6dc443627c1f50ULL,0xa570dfe3655636ccULL,0xeb67bf7b6b7e5de8ULL,0xc1be2dfd677f3ed4ULL,0x4982e523027e6864ULL,0x777d5f59e27e5ddcULL,0xab324fabc07c0b4cULL,0xdcf8261327760f40ULL,0x80a6fb6342343e50ULL,0x8852230e615200dcULL,0x6cc31fd0e3665dc0ULL,0x34bd06302c46144cULL,0x8053040400721800ULL,0x951a3d7100741d5cULL,0x3bd7e3b66a7f1f48ULL,0x88438b60c0760cd0ULL,0xe884550022c0848ULL,0xf5777feee6767f5cULL,0xbd6fd70e3561ffcULL,0xcfff77de6b741dd8ULL,0xcfaefe02275d1fdcULL,0x3bafdfdbe7761adcULL,0x74ba76e9e25c48d8ULL,0x41670320c0150ac4ULL,0x12903a6c80521cd0ULL,0x8840000100460244ULL,0x2062080824460000ULL,0x18100c02a5040840ULL,0x810230606c501448ULL,0x8000020000001100ULL,0x484052640360250ULL,0x60001000800ULL,0x2000010000240000ULL,0x8050040060101000ULL,0x8100044002240040ULL,0xa845c00000621148ULL,0x420e16500076007cULL,0x41d4c0362343820ULL,0xd496701463340884ULL,0x4484646260361c44ULL,0x82560020001915c0ULL,0x408680901240a40ULL,0x401d80160101144ULL,0x36b4c468017a1944ULL,0x407af51027351050ULL,0x904710083102840ULL,0xfd0320a08741c4cULL,0x814d05e425210c0ULL,0x24118060407d0848ULL,0x494e2570006e1d5cULL,0x5000020200440004ULL,0x80430a0940360048ULL,0x4002d84122700080ULL,0x2c30d36963340e64ULL,0x612e87f8aa563cc0ULL,0xd414a23060764884ULL,0xccc6f3ede67f5dc8ULL,0xbfdeffdf42765f5cULL,0x8902ce5a674615d0ULL,0x2044260000580248ULL,0x7f962d1023762ca4ULL },
	{ 0xd3f7f31c3f2cd5d1ULL,0x65c2b21a0a28274aULL,0xb91247533b05f3d1ULL,0xdcfef7d09b09efd3ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf7be2fff3f29ffdbULL,0xdbbafb3b2f0defb9ULL,0x4ada3a151b0cc393ULL,0x9ebfdfe3be6cc5dbULL,0xf16bfe7e3e2df7fbULL,0xdeaefd729bade7dbULL,0xd3d5e36faa2dc7f9ULL,0xefdbfb7fbf6dffdbULL,0xffffdfffbfedffffULL,0xf5fdf6ef3fedfffbULL,0xfbfffff1bb6dffdbULL,0xffb7fffdbfa8ffdfULL,0x6fdfff2f2f8dfff9ULL,0xfb5bdfdf9e2df7dbULL,0xbcff5eebbf6dcfd3ULL,0xff6ef7ea176dd7feULL,0xffff7fff1beceff7ULL,0xffcffffabfecd7ffULL,0x6fffffdf3bedf7dbULL,0xfffb7ffb3f6defffULL,0xcd7fa77b3f6de5faULL,0xffbffffdbf2cf7ffULL,0xcfdff7b53b6defdfULL,0xef7fff4f1e6cffdfULL,0xbffbfffb3fecffdfULL,0xf7efb7fa3f2defdbULL,0xfbbe7ff70faddffbULL,0xd992ff3b1fedfd72ULL,0xeffeb65a3b29f7d0ULL,0xffffcdff3a6df7dbULL,0xffffffffbf6dffffULL,0xfffefffb9fedfffdULL,0xfbfde77f8fadfffbULL,0xbefffeffbbedeffbULL,0xafffdffbbfedffdfULL,0xf4fff7fb9fedfffbULL,0xfabe7f722b2dffdbULL,0xfeb7fffc3f6deffbULL,0xcfc3fe9d3eadefd5ULL,0xaaff5fdbbf6df7ddULL,0xf5ff7ffb3facffdbULL,0xbb6f7f7a0b2dfffbULL,0xffb46dfb3beccf59ULL,0xffcfffbc8badf7ffULL,0xabffffdfbbedffdfULL,0xf4dff5e3af8df7fbULL,0xfffefffdbe2dfffdULL,0xfae77fef3fecffdbULL,0xe5c9f2192e2cc7d9ULL,0xea7f3e583f6def9fULL,0xbfff6ff99a29cfdbULL,0xfdffb77bbfeddfffULL,0xfffe7fffbf6dffffULL,0xebffffffbf6ddfffULL,0x8fdf3fdf1e6cd7f9ULL,0x1fbfefdb3f4cdfffULL,0xf7f766f81badffdbULL,0x7cfe3ed21faceddfULL,0xd8f4e70e2ca1fd53ULL,0x6b5fb34b9a29cf7dULL,0x6ebfef433ea0d7dbULL,0xddffff638feddf5dULL,0xffffbff6bfedfffdULL,0xfbffff7f3fede7dfULL,0xfde6ff3fadecd79fULL,0xaeffcddb3a2de757ULL,0xf5f7f6e93f0cf7dfULL,0xefeb7d631e4cfff3ULL,0xffdf7fff9e6dffd3ULL,0x89f3fe8bbf297f58ULL,0xbdf5d4f1e2dffdbULL,0x84fb3f639bccc790ULL,0xfbfe67eebbadfd99ULL,0xbaa33f6b880ca5aaULL,0xc9feff3e8a09e75aULL,0xb6c3a0dba4ce559ULL,0xb6fe5fbb8b2cd69fULL,0xb33aff441a0ca71eULL,0xdef86070080cc71fULL,0x8992db270b2cc753ULL,0x2cf3bed83bacc5f9ULL,0xde1d77de2b2cd7fbULL,0xf0f4f6f31f65f7daULL,0xceaefd4a8b0defd8ULL,0xdbb6eb3c1bcdeb39ULL,0x4fdf3a5f3f6de5f9ULL,0x8fbecbf58a6de7daULL,0x78fd6ffc9a8ca7dbULL,0xd9f63d773e89ff9bULL,0xfbd7fb6f3b6ccffbULL,0x8fcffe2fbeaccfdfULL,0x219207da2f2843c9ULL,0xf6f6a4e60b2ce7dbULL,0xc778a7393b04e598ULL,0xba902f4c0f28c3d0ULL,0x8fc5fe039e0dc7d9ULL,0x6adf5bc2174cc4d1ULL,0xda121f680e6de6d8ULL,0x7955ff531a6ca7ddULL,0xeeb42f7e0a2c97d7ULL,0xebdcda0c1e6cf5b3ULL,0x6a5fdfcd1f4cc5dbULL,0xa7b64ffb3b2dcf13ULL,0xf7ddb7ec1f2ddff0ULL,0xef9e3d741f2cffdbULL,0xe0d6f36a8b6c45dcULL,0xefce3e9e9a0de7dfULL,0xb5afcffbba8dfddfULL,0xf4f6f6eb8e8df7bfULL,0x64ba4df29b2c85d3ULL,0xd950a922070c4d50ULL,0xcffa36218a2ddfd9ULL,0x611430802081194ULL,0x897252ea000dc5ebULL,0x808c04221a08a20dULL,0xc1104012000c0504ULL,0x6c3fe100a89cb91ULL,0x3b3c4e9b8909f57bULL,0x34ea56e83a8c875bULL,0x55188082548ULL,0x21a0f18090c4e04ULL,0x4989c2300408d599ULL,0xa8500c4a1b08845dULL,0x24b64c841f0c4e11ULL,0x995cfc76142cc5c7ULL,0x369e26740f01c892ULL,0x2100c210020891d1ULL,0x226c8046008c8113ULL,0xb1007520a808580ULL,0xd107167a0e08cebbULL,0xd2b02db21b00cf51ULL,0x40021ca00300d018ULL,0x266b1dc8024d4919ULL,0x262601280a08849fULL,0x2400102012080010ULL,0x84080040500ULL,0x8220526000009618ULL,0x4a12d9020c84d2ULL,0x805cd400300c013ULL,0x6480a280100ULL,0x2018050002004b00ULL,0x0,0x0,0x1844010a0108440cULL,0x30f054ca004c8511ULL,0x69fa2e601b2cd891ULL,0x480c093002208418ULL,0x0,0x8994c400b204899ULL,0x2108200000ULL,0x5100004008000440ULL,0x20040000800100ULL,0x0,0x80800000008002ULL,0x818240281a00a191ULL,0x0,0x901408600508c101ULL,0x9c2200300808519ULL,0x20068221000c8512ULL,0xbef9ffebbfadefdfULL,0xffffffffbfedffdfULL,0xfffefff3bfedffffULL,0xfbffff7fbfedffffULL,0xeffbffdf2dadd7faULL,0x97becfebbf8cdddbULL,0xfcb75eee3aedfbd7ULL,0xc33a7d620b8dcfdbULL,0xca0cfe2e1b0cc3dbULL,0x87daff1b882c84f3ULL,0xefffdfddba2df7fbULL,0xf4b6feee1e6cffbeULL,0xc55b77720a2cefc9ULL,0xdb120b76128cc7daULL,0xcdcfc32a0b88edd9ULL,0xef5a1e5a930895d9ULL,0x5a2dd6b206c87b0ULL,0x108c004008208441ULL,0xcc902354004dc9a3ULL,0x8f201710280ca580ULL,0x480008000040ULL,0x3f0e1ff72e2cd792ULL,0x491cf66918a8b3dbULL,0xb31429349a29cdc3ULL,0xe389f30a3b4d5dd1ULL,0x64d884410c0c4450ULL,0xc00a4560200c8311ULL,0xb302a6780a4cff9dULL,0x459c3570120dd191ULL,0xcb82c3cf3a08b71fULL,0x815bb26a000cc793ULL,0xc30c4c301b08c1daULL,0x7755d2693e0d6ff7ULL,0xa51a9d100e4ce3b8ULL,0xe82befff1b2ddfdbULL,0xd7dfbe320e6de7ddULL,0xab9edd930a6cc7f1ULL,0xb5bcc6682e29e759ULL,0x892d0f609b087410ULL,0x84006008002408ULL,0x2c04004084004ULL,0x8114012040500ULL,0x4025b0129086380ULL,0x91085a081208cc86ULL,0xa02351000008598ULL,0x0,0x10040000080080ULL,0x212003401b20c002ULL,0xf0bca6112a09c753ULL,0xdb947f320b64f7dbULL,0x8940804006004c00ULL,0x2e1317448b09c701ULL,0x52818f81804c9491ULL,0xdd67b6ed0acced3bULL,0x50160430286c2104ULL,0x254000000088181ULL,0x2c200800a002ULL,0x901e5009086480ULL,0xf09147620a0c1719ULL,0x803424751a0c03c0ULL,0x8020021004084080ULL,0x801ac02080d6500ULL,0x831a5c822a08a706ULL,0xd0228af00909c381ULL,0x20401200000c010ULL,0x0,0xe4c602300c08c798ULL,0x8140041800054408ULL,0x361f57721a25cdd9ULL,0x50e0354002089017ULL,0xd5963b489a2c8771ULL,0xe057d77f0a6cf7f3ULL,0x18c50e0500248591ULL,0x9a02048a0c886144ULL,0x1080000001000010ULL,0x3a002c00082c8010ULL },
	{ 0xc1c6d30d419ac82bULL,0xa7de3f5f05f64ceeULL,0x9a1ec5520dbec436ULL,0x959034681dc6e452ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfbb63dfc5dbacfeaULL,0xcba5f36b4db65f27ULL,0x26fe9e5addfaefe6ULL,0xc49ecfb041bc67faULL,0xf7ffff6dddbe7f6bULL,0xfffe7ff3ddfaffefULL,0xfbb5ff9e5daadfebULL,0xffffffdfddfafffbULL,0xffbfffffddbeefffULL,0xf5f5fffacdfe7fe7ULL,0xdbfe7f775dba5ff7ULL,0xffbdff7f5dfa6fffULL,0xc7c6bf975dbe5f2bULL,0x8b3dcfdf4d9e5df2ULL,0xfeb7defbdddecfeaULL,0xdbcffff3d5bedfeaULL,0xf3b4e73f5dbafdf7ULL,0xfdf7fffeddfa5fffULL,0xcbfffb57d5faed7bULL,0x3f93ffeb5dbe4ff3ULL,0xffffbd705dbacf7fULL,0xfebe2ffa9dfedf67ULL,0xcbcffe2acdaadf67ULL,0x2b5ffefbddba5ffbULL,0xf6bacffb4dfa4ff3ULL,0xd97dfd524dbcc733ULL,0xfff67f5c5dfecfeaULL,0xeb97ffeb9dba4dfbULL,0x7f7b36dbddbe576aULL,0xd2bdcfb7cdda6f67ULL,0xfffdf7ffcdbadfeeULL,0xdfbefffbddbe7fffULL,0xfbf7ffefddfacf7fULL,0xeffffffb5dbe5ff7ULL,0xefbfdffb5dfeefffULL,0x75fffffbcdfacfebULL,0xdffeffe2ddba7d6eULL,0xff9aff6f5dbafffbULL,0xcfeffe3bcdbad7efULL,0xffffffdbddfedf7fULL,0xf6fff6e6d5fa7ffaULL,0xfffe7ffbddbedfe3ULL,0xf3bdff3f5dbe5dfaULL,0xfbf7ef3fcdfa7f72ULL,0x7fffffdfddfecffaULL,0xf7baffeb5cbacf6aULL,0xff6fbff8ddaecfffULL,0xdb9e2bf708faefeaULL,0xcfc7fe379dfa5e62ULL,0xa5dbdb5bd4da5f6bULL,0xb2bb6fee453a07eaULL,0xfdfcffffddbe4f3bULL,0xfffe7fffddfeffffULL,0xcbfffbffddbeffffULL,0x2f7f3f5fcdb6cffaULL,0xdbbfcffacdbe7fffULL,0xd0deb67b4dba6362ULL,0x92e677e88da8cf5aULL,0xf1958f26d13a0eb7ULL,0x2f5a93db5dfa4de2ULL,0xefbeef5a55ba5ef7ULL,0xdddff77adcbadf6aULL,0xfffffff7ddfeffffULL,0xffffffffddfefffeULL,0xcfefbfbb1dbeff67ULL,0xb97ecf92c5be5f4eULL,0x34ca5feec5fa7f61ULL,0x39e6676a4d3a6f5bULL,0xfe3c2f7b54decf67ULL,0x474bee3a54dae5e0ULL,0xaedfdfd35dbadffbULL,0x34a46761d9aa4f23ULL,0xb9e85e6808f20feaULL,0x3eb43f7dcdb8dfebULL,0xd7dbf2b3ddfe8f67ULL,0x5ddffee1458acb3bULL,0xb118cef241e2cd42ULL,0x944c9dd00dfccb03ULL,0xd7a0fd44158add65ULL,0x8ba2db0e4daac9e2ULL,0x6bf71ffa0dfe5ff3ULL,0xafbfcfdbcddedfebULL,0x4c7f07ec49b8cdeaULL,0x6eae2de35d3a5dfeULL,0xfbf7ff7fddbe7fe7ULL,0x5ff6badfddfacf6bULL,0xafbecf7a59bacd6fULL,0xdc7457e04cb8c732ULL,0xfbb6bf744dae6ba3ULL,0xffa8f76f0dbae7eaULL,0xcff7fe3f5dfacfffULL,0xaf5edf5a4dba9feaULL,0x8b65eed8dba470bULL,0x5fea777b05badf76ULL,0x9fb22d1fc5a85d62ULL,0xc58fea3a0dba5fe6ULL,0xebfcdbc95dea5fe2ULL,0xbcb65efb0dbaeffaULL,0xdbfff7e2cdbaed6aULL,0xfe902f5e9cbacf63ULL,0xc7d3df3849ea4f6bULL,0xaaffdfc64dba7f62ULL,0xbd325d7e44fa4fe1ULL,0xf16e87704dfa4de6ULL,0xdbbe2dfc95f25feaULL,0xebb5ffe29dfedfebULL,0x2f5f1e4105fec1e2ULL,0xfb9ceffb4dbe7feaULL,0xfef776ff9dbeff7bULL,0x4fa62d60c48acd7bULL,0xdbb67206100a4ff2ULL,0xc1c1d21244d68544ULL,0x30e4c104c82d942ULL,0x48010000400ULL,0x618ac30d0c24c60ULL,0x191c22400c2000eULL,0x87c6fe924dac5f03ULL,0xa0005e1a002203e2ULL,0x54a7a4a28dea0da2ULL,0x8541ac01cdd20746ULL,0x181252041888122ULL,0x840021008080320ULL,0x950040840084d20ULL,0xb190150240840122ULL,0x802032604c820960ULL,0x4000280200000320ULL,0x804022030d804002ULL,0x1006025001280002ULL,0x38214bc081b84642ULL,0x1088266300baca21ULL,0x4a84266c14885c22ULL,0xc982db1a8dfa5fe2ULL,0x40020000020800ULL,0x0,0x8000000040000100ULL,0x402625c88c1044e2ULL,0x8c20004100012ULL,0x0,0x208e090045100120ULL,0x110406245800102ULL,0x414002005000102ULL,0x800030400900400ULL,0xa02300305a80340ULL,0x41440110003004a2ULL,0x843132c20cdced62ULL,0x80760758508a4d02ULL,0x1200200018004400ULL,0x8152260344a40112ULL,0x2100000080000400ULL,0x0,0x60001200000ULL,0x0,0x102820008020140ULL,0x0,0x80003100002080aULL,0x80f00c409c800502ULL,0xf2002dccc5928d03ULL,0x400061b000224922ULL,0x60800e0000c24840ULL,0x9fbddfe64dbadf2fULL,0xffffffffddfeffffULL,0xfffefff3ddfeffefULL,0xfbf6ff7fddfa4dffULL,0xffdbbfffcdbeffeaULL,0xff7fdfff0dbaff73ULL,0xb475f679c5badfa7ULL,0x2ab22f02458ad96eULL,0xf298e72e4ca8d9e2ULL,0xceefb73b1d9aed22ULL,0x6b3fcf57cc9a4f64ULL,0x3cbb56ac4dfcc52aULL,0x1b2e3f614c925c6aULL,0xd3a427631dfa4f62ULL,0xc4cae92a0cb8cb2bULL,0x7d6df5adcbadf6eULL,0x35fed5e94cba4f2bULL,0x636aa730cdcc17a3ULL,0x1a3435b349ba49e2ULL,0xc200009404a0ULL,0x4d765c94549a7dbeULL,0x30aaefd05d0a45e2ULL,0x456d152350884d4bULL,0x9061b605c8a4f00ULL,0xc080711149080102ULL,0x6842c844080a41e2ULL,0x41c4ca019a44520ULL,0x143022f9c8ba47a2ULL,0x401c2c70810a0762ULL,0x8a9e531bc8280720ULL,0xa4e2322f04ba5d42ULL,0xf1c5e780dea6342ULL,0xdc90762a5db45f20ULL,0xcf28b4015498ed2aULL,0xfbbe7fef45badd6eULL,0x94ff6df4cb85d32ULL,0x7b7a97dbc5be5faaULL,0xfeb7f6f6ddaa5f7fULL,0xc9d6456a0c824942ULL,0x8b810e448caccd00ULL,0x2004400401000901ULL,0x10c40000000301ULL,0x400804600ca04002ULL,0x4042256088084d20ULL,0x44082df00cf08d22ULL,0xc080401004820044ULL,0x101008000100ULL,0x84280460049a1404ULL,0x8085b5350c387d20ULL,0x33de7f654dea7963ULL,0x88002a4201004422ULL,0x20d61af2919a6c03ULL,0x28cc36884a21645ULL,0xc64264a05fa7de2ULL,0x90011044001100ULL,0x8102e2201d104442ULL,0x48021f8000926740ULL,0x610e4640498a0320ULL,0xf0f210015d8ac522ULL,0x5280204444281302ULL,0x200400000004020ULL,0x86803a000490d842ULL,0xa618d19254a87f23ULL,0x80816840000302ULL,0x220857204002412aULL,0xc8082c4ac4a80760ULL,0xa81c200480801a2ULL,0x8000014200020804ULL,0x80d699f305fae3e1ULL,0xfcdc04320898cb21ULL,0xdbb48b4a4db85332ULL,0x811263284d026902ULL,0x41484800d00000aULL,0x8200040841000102ULL,0x41000000020040ULL,0x4c4c2040019a0422ULL },
	{ 0x952f06a143066c2ULL,0xaf11266a74102ceeULL,0x4212cefa32107ae4ULL,0x90fd2e3312125ac6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf7be3f77361377e6ULL,0x944c26d341047e4ULL,0xaf66364e263272eaULL,0xbb3c4de2720067e4ULL,0xf7fffffbf633ffeeULL,0xfffefdf3f633ffeeULL,0xfbeeffbff633ffe6ULL,0xfffffffff633ffeeULL,0xefbfcffff633ffeeULL,0x7dbdff6bb4337fe4ULL,0xdbaabff776336fecULL,0xff9bfffff633ffeeULL,0xcfcfff7f3633feecULL,0xaf5ed7dbb612fde6ULL,0xf5bef7f93433fee6ULL,0xfffeef77b613ffeeULL,0xfbf77f7df632feecULL,0x9fdfff3b7633f7eeULL,0xebdeffdf3433fee4ULL,0xa57e2aeb32126be4ULL,0xd1aaf7f3743367eeULL,0xfeb02f6e34127eeeULL,0xcbf6ff39341377e4ULL,0xede7fedf7433fbe4ULL,0x7fb247ee3631dfeeULL,0xfdfebdf91213f3eeULL,0xfbb665f516326ee6ULL,0xebf7fb7a763377eeULL,0xfffebfd6b6337eeeULL,0xbffbdff3f632ffeeULL,0xf7ffb7fdb433ffeeULL,0x4a847583943374e6ULL,0xbad6e338063366ecULL,0xef7fb7dfb63376ecULL,0xbfffcfdff433feeeULL,0xf5fbffefb633ffeeULL,0xdbbfbd71b4136eeeULL,0xfbbdfbef7433f7eeULL,0xaf4ffb9b96137deeULL,0x3ffedfdff63377e6ULL,0xf7fffeeff633e7e6ULL,0xf7eff77a3633ffeeULL,0xf7b67ff7f433feeeULL,0xcdddff3ab6333fe4ULL,0xe7dfdfdc363377e6ULL,0xfffff7fff631ffe6ULL,0x3379affa663372e4ULL,0x7ebe3ff73613eeeeULL,0xaefc772a363375e6ULL,0xbf7fbed7f612fb66ULL,0x9fbb7f7e3633e7e6ULL,0xfffeffebb633ffeeULL,0xfffefffeb633ffeeULL,0xfbfffbffb633ffeeULL,0x2f7f1b4af61363e6ULL,0x199f6efff433eee4ULL,0xd5f6fedeb412f76cULL,0xffbe7df336327feeULL,0xdb56ef6714337744ULL,0xe35b360ff4117be6ULL,0x8effcf7af012eee6ULL,0xf5fff7e9b433ffeeULL,0xfffefff7b633ffeeULL,0xffbfffffb633ffeeULL,0xebefffdbb433feeeULL,0x2ecddfdf3613e6eeULL,0x54d5b6f0361076e6ULL,0x97ae3779f430f6e0ULL,0xcb7e7b7f3432d3e6ULL,0xe8d7ff0d141246e6ULL,0xe3fcce583233f4c6ULL,0x40b657a1301366c6ULL,0xd594047410131e86ULL,0xd0b12c7080106eecULL,0x8bf0d36d703376e6ULL,0xef9f2e589433f6e2ULL,0xa5223162a212e6e4ULL,0xfffffffff633ffeeULL,0xdb5ef7f03613efe6ULL,0xc976ebb2f63366eaULL,0xefff5f5fb413f7e6ULL,0xcfbfdffbb6327feeULL,0x1200100810104002ULL,0xcab648780013f5e8ULL,0xebf4e777b6327ee8ULL,0xdfdf7f7f7633e7eeULL,0x7ff9dffb76337fe6ULL,0x757556ee36124eeeULL,0x9ebaec7114136eecULL,0xfa94fb2ab413e7eeULL,0x4fd7bfbf363357eeULL,0x3aa7ef40163342e6ULL,0x96d377eb0431e7e8ULL,0x91f9f773b413f7e6ULL,0xfb91673734333ee6ULL,0xf5ceef3a363377c6ULL,0xea7f5fd296337dc6ULL,0x3eaedfe5041377e4ULL,0xb948f7fa34132ee6ULL,0x9b84047424127660ULL,0xec84d0266412eee4ULL,0xfffeffda34127eeeULL,0x9e327ff164324ee4ULL,0xf0be3739a632e6c4ULL,0x86166469d6333366ULL,0x8bffe77eb433f6eeULL,0x4cf71f5ff6326aeaULL,0xba3d7eebb631f6e6ULL,0xb4fff7eef633eeeeULL,0xdbf62de1b433ece8ULL,0xfbd6f68636132762ULL,0xc67b3613563346ecULL,0xca928d821201e7c4ULL,0x305110a412014042ULL,0x33eefe636430f660ULL,0x50808958160106e0ULL,0x458101814012268ULL,0x8f9cccda6610e764ULL,0x1084548aa01002ecULL,0xcd4ef573a41236ecULL,0xce062dc442002660ULL,0xc492f132a41166e4ULL,0xa5cfcd2b413326cULL,0x34a62d6400122ec0ULL,0xd1d0a5c830106fc0ULL,0xf392271c30116c42ULL,0x400014a020022060ULL,0xa2900f40141236c4ULL,0xb0910d6230116de2ULL,0x50002081120006c0ULL,0x5ed0025e54136ee6ULL,0xac0802210000ce4ULL,0x6443158804314284ULL,0x2100c0c010000568ULL,0xd5fa7eee9633f764ULL,0x4528014010304ac0ULL,0x8900000002020240ULL,0x9b5ab297b43262ecULL,0x2920029814000a44ULL,0x451506010022aecULL,0x4d8c4c22843374acULL,0x4000400000000084ULL,0xd2e00a16026a44ULL,0x214400000104480ULL,0xf2b854d332126462ULL,0x2056296010064ULL,0xc010a21c0410d444ULL,0x200030800101060ULL,0x0,0x0,0x0,0x202000000022000ULL,0x30a010010040ULL,0x849105110114282ULL,0x2220002830000960ULL,0x5020240410010242ULL,0xc082004020204ULL,0x5088f0900030d4e8ULL,0xc000040214000804ULL,0x8498e2a2302344c6ULL,0xfffff7fff633ffeeULL,0xfffefff3f633ffeeULL,0xfbb7ff2ff411f4ecULL,0xfffa7efff43377e6ULL,0xffbfdffff633feeeULL,0xf0ef96f8363376ecULL,0xdf96b770301366e8ULL,0xdfa7f77eb633fceeULL,0xdddbbf97b633f7e6ULL,0xabdedbce3613e6e4ULL,0xf5a464f0361376c4ULL,0x31e7a7f1b4336ee2ULL,0xca906e71140226eeULL,0xebc2fba536335ee4ULL,0xab3d0fde361357e6ULL,0x9fce57ebb4135fe4ULL,0xcbabfd613633ffe4ULL,0xd286ef7b36336aecULL,0x89c0101014001368ULL,0xae50a0d8221260e4ULL,0x5bb86d73363267ecULL,0x81080642121042c2ULL,0x5a0e096030114c60ULL,0x8191b04004323164ULL,0x14110000020c000ULL,0x90a10400041054a8ULL,0x5534f67814316ee4ULL,0xd2e2ff1101377e6ULL,0x8b90e7323013f2c6ULL,0xaf63304b14307eeeULL,0x2f165f53b43362e6ULL,0x996e79e3a41376e6ULL,0x92222d4054014466ULL,0xc798fff664126ee6ULL,0xdfcfff0bb633fee6ULL,0x8bdbfd58b4307dc4ULL,0xf6ffd6ecb413feeeULL,0xf3b8ff72b433f6e6ULL,0xd6842078301164e0ULL,0x46c3ac024602d164ULL,0xa15e144084326046ULL,0x5304bc0142200aaULL,0x80402600000322a0ULL,0xc82004421010c402ULL,0x8140b1014011cee0ULL,0x500042000004c0ULL,0x2204c042004060ULL,0xc506a346262224e0ULL,0xdbe47de23631fec0ULL,0x8006000c04220244ULL,0x6e460299343172a0ULL,0x8b18ee52201347c4ULL,0x7c4c8661202226e4ULL,0x20900050401062a0ULL,0x10000000280ULL,0xe008300001104a0ULL,0x26e6cf4012134ae0ULL,0xc23240141324e2ULL,0xb045860042061c4ULL,0x8584616890121aa4ULL,0x8e7a7e1b001123eeULL,0xff7f1f5616336fc6ULL,0x4000001000204000ULL,0xdaa2542b0136264ULL,0xca02414a1412246cULL,0x104108010100022ULL,0x740401000122480ULL,0xc0148170000460ULL,0x9a844002942044e2ULL,0x8a16016c103154e0ULL,0x0,0x0,0x80460004102004ULL,0x930815230321482ULL,0xcd000000001202a4ULL },
	{ 0x1e30060245072ULL,0x7a471adac01c96aaULL,0x89104e717004f0e8ULL,0xd574b77bd0949feaULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf7be6ffdf8dcf6faULL,0x4880731a103086c4ULL,0xebe7135bf0b4f1c0ULL,0xfebb4fb7f0dcd7faULL,0xf7fffffff8dcfffbULL,0xfffefff3f8fcffffULL,0xfbfffffff8fcffffULL,0xfffbfffff8dcfffeULL,0xafffd7dbf8dcd7feULL,0xa4fec6e950947ee8ULL,0xdbbefd73f8fcb7fbULL,0xff97fffcf0fcbff2ULL,0xcdfeffbff0dcfefbULL,0xef7fffdbf8fcdef6ULL,0xfeffdfef78dcdffaULL,0xfbffbf7a7054bedeULL,0xfbb66734f0d4b7e8ULL,0xcddfff3a78bcbefbULL,0xeefedf57f094f7feULL,0x18cd5581e0a43068ULL,0xd9fbef7af0bcf6faULL,0x86b7bf7e50b4fefbULL,0xefcef23af08cb7eaULL,0xe81efe8138e496e0ULL,0x97daeffbf09c3fd8ULL,0xb0d7ffec7854bffaULL,0xff667fe2f8fcf6eaULL,0xdbd7ffbff89cbef9ULL,0xbdfefe7bf8d4f7feULL,0xffbfaffbf8bcbffeULL,0xfdfffffbf8dc7eefULL,0x4fb6bffbd8109fddULL,0xfb7eff6ff0dcbdfaULL,0xeffbbefff8f4f6faULL,0xffffdffbf8fcbffaULL,0xfdf7ffe5f8dcbfeeULL,0xbffafd71f0fcb4eeULL,0xfbccefadd838bfd9ULL,0xafdefea370bcbcfbULL,0xffffdfd7f8bcb5fbULL,0xf6fdf675f8fc37feULL,0xffeaffe578dcf7faULL,0xdfbe7f7db8fcfefaULL,0xdfdbffbef09cbc79ULL,0xc7ffff5bd094bedaULL,0xbfff7ff3709cb7fbULL,0xfffffffff8fcbfffULL,0xdfbf3ffff09cbeffULL,0xedeff71ff8fc9ffbULL,0x5b5f7f73f85c74f8ULL,0x95bb7ffff0bcb7f2ULL,0xf5fff7fbf89cfffbULL,0xfffe7ffff89cfeffULL,0xcbffff7ff8dcbffaULL,0xdfda3fd7f8f4d7f9ULL,0xbf3ffffff8f4feeeULL,0xfdf7fefb70d4f7f9ULL,0x6dfcbff0d09cbfdeULL,0xcbfff72ff03cb6faULL,0xcfdbfaff709cbffeULL,0x47f7dff970d4fefaULL,0xdcff36eef0d4f6faULL,0xfffefff3f0fcffffULL,0xffbefffff8fcfffbULL,0xefdfbe03f0f4fdfbULL,0xefdecdded09c7efaULL,0xf09c56eed0d430daULL,0xdb7dbc71f8f496ecULL,0xffe66f6fd8fcbed7ULL,0xfcdfe31ff0dcb7faULL,0x6bfa2d5bd0fc1eccULL,0xb43b5ed3d02cf0caULL,0x11cda578f08494e0ULL,0x76a02c66d08474daULL,0xc157da2570d416b2ULL,0x9511fcaf09c9498ULL,0x96964d62d0949e89ULL,0xfffffffbf0dcdebbULL,0x600c041140a03468ULL,0x102000040040000ULL,0xeffe3fdfd05cfefdULL,0xcf9ecdfff8d477feULL,0xf424f2c8d01c1c8cULL,0x7bbe65faf0dcfcfbULL,0xf9c5eb1ff09cb8b8ULL,0xebc3bebdf0949fefULL,0xff3fdff3f0d4fefbULL,0xfdffffeef09cfefbULL,0xfbbebf65d0dcfcffULL,0xf38ceb6f509cd7d8ULL,0xcfffff6ff8dcfffdULL,0x2bf9dad9609c9ef8ULL,0xd4747434505c9698ULL,0xcd7337eb709cbe76ULL,0xffbc6fec70b4a7e9ULL,0x68c6633af0909eb1ULL,0x1ffcdfdb709cb5eeULL,0x9cbb5f7ef0149febULL,0xd9faaf60781cf3eaULL,0xdea32d6ed8b4fed2ULL,0xcd96f70ed0d49ca1ULL,0x2f7ffadbf09c1fe6ULL,0x172eff7370b43efaULL,0x906e3ff750f4b6feULL,0xdf3e7fffd09cbc5fULL,0xc9f6fffaf01cb7eeULL,0xeacb92b3109096c8ULL,0xacfd4feff0a4f4e2ULL,0xdd76f65d70dcfceeULL,0xdbae7c437010bcfcULL,0xdb937b3f5090d499ULL,0xcb53b69b7094b4eaULL,0x103cc772701417e2ULL,0x90332eeed064b7c0ULL,0x9a88254370348842ULL,0x48c9001a00141496ULL,0x495af29238d037eaULL,0xe0eea93700494a8ULL,0xc4b4dec8d0541686ULL,0x124052020143c40ULL,0xcb86f771f0c45428ULL,0x8210a202500014a8ULL,0x20801035040108aULL,0x841152735084b4c8ULL,0xdbd4f652900cd7ceULL,0xfc971750d04094c4ULL,0x89527a1398949e82ULL,0x885a120a300420c2ULL,0x820401000108840ULL,0x1014160290501008ULL,0xc0600210148000ULL,0x8106020450141400ULL,0x8002300280801008ULL,0x992c4d33103497c5ULL,0x8002022010100000ULL,0x9085c2000140e0cULL,0x1000c10010840080ULL,0x431208b0141022ULL,0x40d4c20402014c2ULL,0x304226c04070b150ULL,0x8001012070941000ULL,0xba14d73050141482ULL,0x49061220101028c0ULL,0xb34091bf0b894e0ULL,0x252104205010a028ULL,0x875200901802ULL,0x880046010148000ULL,0x280a000d0001000ULL,0x0,0x0,0x1000000400100010ULL,0x40100100010ULL,0x10840000080ULL,0x5d21230809062ULL,0xa000040000ULL,0x100a41240c41080ULL,0x32f02640301016e2ULL,0x5802000101808ULL,0xa806020030001428ULL,0x83344a90500014a8ULL,0xfffffffff8fcffffULL,0xdffeff73f8dcfeffULL,0x739be39370bcbdf2ULL,0xefdf7fddf0fcbefaULL,0xfdbf7fddf0dcbefbULL,0xbdf5feee50dcf7d0ULL,0xcfb7bf50d094b6faULL,0xeaa5d36ff01c16abULL,0x4fda3e13f014b62bULL,0x29dfd8d7789c1778ULL,0x14ae76cef07c96f8ULL,0xc1ecff22d01cb4d8ULL,0xdaa43f2eb014f1caULL,0x47cff20678943672ULL,0xa79bd9d670bcbedeULL,0xb0535fe1584c30eaULL,0x13c82d0070941168ULL,0x4e972db0500ca6f4ULL,0x8764fb74701417eaULL,0x688bba78705c32f0ULL,0x971a456370bc94f8ULL,0x520074685050d8c8ULL,0x0,0x102a30b103012d0ULL,0x400000800002040ULL,0x128c050c708c14a8ULL,0x631146a608434d0ULL,0x8006800041402ULL,0xc00040001000ULL,0xaff31884f09cd07aULL,0x26364f5250d4bc50ULL,0x79fc56eef0dcbff8ULL,0x85aced72b89c94abULL,0xf88ddf3ad09c1a7aULL,0xb53deabf034b6f8ULL,0xebdfcfdad094feffULL,0x9c957664d0dc55faULL,0xc9ef7f71709cfdfaULL,0xc08e2d40d000b042ULL,0xd986e326d0dcb8f8ULL,0x623fd489401cacc5ULL,0x8493de61509495c8ULL,0x6568f432401ca262ULL,0x8004056c10141480ULL,0x89c6e208f09c3ce2ULL,0x802000010008000ULL,0x483a0f83d014b4d8ULL,0x3108f648509480c0ULL,0xc1807800501012a2ULL,0x7004e132208c1568ULL,0x8142060a90009448ULL,0x8e0b4f9a901c1ecaULL,0xb75eaeeaf09cbce2ULL,0x42554100141682ULL,0x80040108004034c0ULL,0x16422a014084b0c0ULL,0x230004000101498ULL,0xc4545cec78d4b649ULL,0x40250000805140ULL,0xc1102111580096c8ULL,0xc47ef5a3f08cb6d0ULL,0x6a5bcc5ab094b4eaULL,0x14010020000c0000ULL,0xd280200060002020ULL,0x80100000000c0000ULL,0xc00010100000ULL,0x2000008000000401ULL,0x524000001242ULL,0x1060d41008501810ULL,0xc202245000c41480ULL,0x80000200040002ULL,0x808040000080ULL,0xa212416848109850ULL,0x0,0xb10098008001403ULL },
	{ 0x88000e2080048848ULL,0xa9e140188e80370ULL,0xf4b067a090eba380ULL,0xa5ff76e9f16eaf5eULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffe7ffff9ffaffeULL,0xebd4f777d96cb71cULL,0x23ef1fd7f8ffbb5aULL,0xf6fffffad9fbbbfcULL,0xfff7bffdf9efbffeULL,0xfffefffbf9ffbffeULL,0xfbfffffff9ffbffeULL,0xffffbfffd9ffbbfeULL,0xef3fdf76f9ffab7eULL,0xf5f716eed1e5bafaULL,0xbb3eff7399ebaff6ULL,0xffbdfffff9efbbfeULL,0xeffefebfd9ffbb7eULL,0xfffedfdfd9ffbffeULL,0xffffd7efd9ffbf56ULL,0xe97ceffab9edabdaULL,0xff93ff7fd9efba6eULL,0xcddfffb999eeb37eULL,0xfbfeffd9d9ffbb7eULL,0x941e716798eda3d2ULL,0xf17c6ffbf9ebb77eULL,0xf6be6f34f96dbf7aULL,0xcdcbf7ffd9dbab76ULL,0x3ddcfbd5d1f5af5eULL,0xb7b36ff691fcb376ULL,0xd1f097d0b1632b52ULL,0xfabe3ffd98eebb5eULL,0xebd3ff3b90cebf5cULL,0xf7ee3e7dd9fdabeeULL,0xbeb4effbd9f9bb5eULL,0x97fff6fff9efab7eULL,0xfbd6b7f0d9ffa36aULL,0x8ad06a0a90e52368ULL,0xaffffbff91eebb5eULL,0xffffdffff9ffbffeULL,0xf4ff76eed9ebbf7aULL,0x3fbaff73f1eebb7aULL,0xfd8fffffd9ffbf7eULL,0xcec7bf3fd9e7bbfaULL,0xbbdfdfdfd9fdaf7aULL,0xf6f5def6f9fdbffeULL,0xffffff73f9efbffeULL,0xdeff27ffd9edabb8ULL,0xfff7ffbfd9fdbf7aULL,0x2bffdfdf98ebafc2ULL,0xbffffffff9fdbffeULL,0xfffffffff9ffbffeULL,0xfebe3ff7f9ffaffeULL,0xebf7fe36d17fabeaULL,0xeff7fe5ff9e7bf6aULL,0xbffbfffff9ffbbfaULL,0xfffeb7fbd9ffab7aULL,0xdffe7dfbd9fdbffeULL,0xcbdefb8fd9edbbfaULL,0x87df3ed6d1ffb7faULL,0xffd7cfe7d1ebbbfaULL,0xf5f7f6fbb9edbf5aULL,0x653e5d78d9feab7eULL,0xd9dcf72ed9efbb76ULL,0x8efbb35fd1ef237aULL,0x93bc4f7ab9e9b338ULL,0xf777f6fed96b2b76ULL,0xfffefff3f9ffbf7eULL,0xffbffffff9efbffeULL,0xdc7e68fd1ebbf68ULL,0xaf5f5b5e99edbafaULL,0x749cb70099482212ULL,0xf5fff771d1ebabfaULL,0xf2bc7f7fd9f58f5aULL,0xcd8fdf0e91edaedaULL,0x6bdfff5ed1f9ae7eULL,0xe826593a0964a75aULL,0x11aad730d1188ac2ULL,0x4aa0213a908ba350ULL,0x8b71f60290f3bb60ULL,0x840d288904c0308ULL,0xf6fa47e391fdbb7eULL,0x14ed72ebd1a80a72ULL,0x4b0677e2514a225aULL,0xc282e942d9d0a142ULL,0xe04e9801d178aa72ULL,0xb3ba5f3a91e5374eULL,0xcf33f472f0e12378ULL,0x6cb475f1d9cbaf7aULL,0xfba6d33ff9fb2778ULL,0xaff3dcabd958af74ULL,0xffbbdffbf9f5a37eULL,0x77f176ead1eaabb2ULL,0xeb77f5f3d9e7abdaULL,0xfa9aff3dd9efaffeULL,0xaffffffff96fbf7aULL,0xe9df59d2e1aeab4eULL,0x743ff0e5d16dbf5cULL,0x9fadf7f0d12cab7aULL,0xfbb47f74d1e8aaeaULL,0xccdee31c916aab4aULL,0xaa9fdfdbd9e32f56ULL,0x149f59e290d00372ULL,0xf54cbf66f1dfa77aULL,0x76be3747f9ebab52ULL,0x6694f2a491df034aULL,0xab7fde7bd9d9ab7aULL,0x93335ffe99ebab52ULL,0xb1eebff3d1e9321aULL,0xffd67fffd96fab7aULL,0xea87f76ad9fdbb72ULL,0x69d3b6d491df8a1aULL,0x27fdcfc3d17faafaULL,0xf576f6fff14fbb7aULL,0xda6c7df3d9f9a3daULL,0xcb4cef7fd1c8b3f6ULL,0x63c0bc92e968aa04ULL,0x2808c54351252048ULL,0xcad2a2c019650062ULL,0xc51c78708861a140ULL,0x8818c30a99409002ULL,0x82388218618060ULL,0x92364e00b9e62254ULL,0x21a8c02819410270ULL,0x8188204780c18540ULL,0x8a804344b1052248ULL,0xe88560299c02016ULL,0x1d0800200240208ULL,0x1e05340a00a2260ULL,0x4008e453996488a0ULL,0x3800060409480208ULL,0x6060488402004ULL,0x6218140010082256ULL,0x1010240090010040ULL,0x4040900020080200ULL,0xd302200091048908ULL,0x800108810008810ULL,0xc0121e92d8480e18ULL,0x1409058211000052ULL,0x30122e00b8009252ULL,0xa5343d70d983b064ULL,0x9010eb2008480742ULL,0x8cc8108010402414ULL,0x84880019400000ULL,0x1080248088440008ULL,0x820e0160a8012244ULL,0x1004242088002800ULL,0x2106420291400258ULL,0xe3dff96f1f8036aULL,0x30106740a1c08688ULL,0x8246c21d1802042ULL,0x4110220481010230ULL,0x41010801400102ULL,0x0,0x0,0x0,0x8000010000008800ULL,0x0,0x46000888000000ULL,0x8000002000000042ULL,0x4028046988a22044ULL,0xa20d40d0011060ULL,0xa0040002000ULL,0x845a0e48d0402e30ULL,0xfcb3cfe399fbaa58ULL,0xddfff7fff9efbbfeULL,0xfffefffbf9ffbffeULL,0xfbdfffffd9ffbffeULL,0xefeb3bdbf9cfab7eULL,0xbfbecfdbf9ffbb72ULL,0xf5f73ee9f9ebbffaULL,0xa6baad70d9d9ab7aULL,0x508c9adcd9408b7aULL,0x4d52d288d963a27cULL,0x26d8d5cb914fbb5cULL,0x94972420d1401200ULL,0x81b627b09160ab42ULL,0xdb9c7e7f91febffcULL,0x71dfe8229022bbf2ULL,0xeb7fdfdbd9eeba5eULL,0x20260a428040234cULL,0x82484c010ac8208ULL,0x4a963e2298209a50ULL,0x478e70e4d1c00b2aULL,0x2a7018c01100b298ULL,0x41b0e552d164925aULL,0x9048b55098648a10ULL,0xc21e0c31b0082220ULL,0x8100e2402104220aULL,0x0,0xc68090830040ULL,0xd46034fcd9c6a90cULL,0x8002340080012214ULL,0x8220010001000160ULL,0xfda1e53d1b5a34aULL,0x2fff5ff9d1efbbfeULL,0x95f337ebb1e5b35aULL,0xd4a52470f961b3d0ULL,0xdba86f2ed9eda75eULL,0xcfcafb7f99f5bfdeULL,0x3bfedfdbd97fab7aULL,0xd0bf7feed969ab7eULL,0xddfe7ff2d9e9b7f2ULL,0xf6b6afd490faaf76ULL,0xe0defe2490d0ab52ULL,0xca3e3603d9e7b24cULL,0xf4da5d6bf1f5ab50ULL,0x51010e400100214cULL,0x62b82de2b01c9a98ULL,0x100800000208000ULL,0x0,0x8b2245e3f12d3210ULL,0xf5968770d1abb360ULL,0x2880026280692a00ULL,0xb83bb1ed966a352ULL,0x60779251d9e02708ULL,0x449888209180002aULL,0xf547726ad149a3f2ULL,0x1024a000d0402200ULL,0x8000100000002004ULL,0xc6c238ebd1f42358ULL,0x619840800120ab46ULL,0x941754ebc1f4b352ULL,0x2954502250222262ULL,0x4082619590a42038ULL,0x246fa00b1632b40ULL,0x821e044290492362ULL,0x145064e0b0a82200ULL,0x440210400210ULL,0x20000520c0800000ULL,0x8002080100008a08ULL,0x120080100000ULL,0x2cd0302098c02004ULL,0xc000081010021000ULL,0x4202008000402204ULL,0x8000100041000000ULL,0x416000010800010ULL,0x8000012000060048ULL,0x24048000002212ULL,0x240080100000ULL },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfff6fff7fffef767ULL,0xcba7efb7cdfcff66ULL,0xabdedfdbebcee563ULL,0x96dfdff3e36c6561ULL,0xfffffffbfffeff67ULL,0xfffefffbfffeff67ULL,0xfbfffffffffeff67ULL,0xcffffffffffeff63ULL,0xbfffcffbdfbeff66ULL,0x74f7feed5ae47766ULL,0xfffeff62cffe7d63ULL,0xffbffffffffeff67ULL,0xfff6ff9fdbfef567ULL,0x7bfedf5faffe7f63ULL,0xf6fbdfe42bfcfd67ULL,0xfdeeef7e0bfe7763ULL,0xffd5ffffebfe7767ULL,0x4fcfff3ecff87f24ULL,0xaf7ffed7cfeafd67ULL,0xb7bf5ff87f7afd60ULL,0xfbfd77fc9bfeff63ULL,0xdabfbff5bbca7762ULL,0xcb57d7b65bc27762ULL,0xdb5bfe74cbbc7727ULL,0xb88c7b90a886664ULL,0xd7edfffddbfc7763ULL,0xfffe7ffddfee6723ULL,0xf9cff7fa4feef563ULL,0xebfedfffdffc7767ULL,0xbfdfffff7feef763ULL,0xddfefffa8bfeff67ULL,0x7bbe3df35bbef760ULL,0xcb95b75ecfee7720ULL,0xbfd3bbd35ffef763ULL,0xfffffffbfffeff67ULL,0xf5fffffefbfc7767ULL,0xfbbfff73dbfe5767ULL,0x78c5e7960bea7605ULL,0x6ad78f17facc6561ULL,0x7fffdfdfdbfef767ULL,0xfcf5b7f99b6cf567ULL,0xdfffffeedffcf766ULL,0xffbeffeddf7e7562ULL,0xeffffffffffeff67ULL,0x63fdcf5bdf6e7767ULL,0xfcfe7feffbfcf767ULL,0xfffffffffffeff67ULL,0xfebe3ff7fbfeff67ULL,0xadd3f46b5b7eef66ULL,0x2f6ebfcefdca7663ULL,0xaef6dffbdfeef747ULL,0xf874bffe4ffcf566ULL,0xffbebf7defeeff67ULL,0xe9deef7fdffe7763ULL,0x6f579f51da7c6561ULL,0x9fffcfffcffcfd67ULL,0xf1dfbee9effc7760ULL,0xcf267d798dec6f67ULL,0xd9d5c33e9d287f65ULL,0xeeefbbdfdffe7d63ULL,0xeeffcfe9cbf47761ULL,0x7f779afb4bee7762ULL,0xfbfffff6fffcff63ULL,0xfab07f7b8fee6f21ULL,0xcd87ff8e4f4aed20ULL,0x246088b48086140ULL,0x50f4beeedfce7762ULL,0xc5cbb769593e6723ULL,0xd900833c49286760ULL,0x6dc7ec288ffe7564ULL,0xa3ffd3dbcfee7763ULL,0x7df37debefc87763ULL,0xfbf5d7f29dfc7f67ULL,0xe40c2522090a6d41ULL,0x8f28f2b048486704ULL,0x6b7fd7d59bfe7764ULL,0x3bb54f38c96c6d65ULL,0xc8e1134805884140ULL,0x5200745d47203720ULL,0xe2b1fb6e8bf87521ULL,0xa8f7ce1292746507ULL,0x3ebb6cdb49287760ULL,0xfc7ff2fff9fa7761ULL,0xeefefd7afffefd63ULL,0xebdce776ff7a7563ULL,0xfffbbfffdffcf763ULL,0x67bedff39ffe7f67ULL,0x9d7f7fe85fbc7767ULL,0xe39aff71ddfa6543ULL,0xfb9ff7a6dbee7743ULL,0x1df39f39dbee7562ULL,0xabff0fcefdfc7767ULL,0x409686e102786461ULL,0x91487e5109a86f63ULL,0x8d3c2f7d19fa7f60ULL,0xcdf2bf3a9b3afd66ULL,0xaadedf5a59667c42ULL,0x8e871f6a8f6c6d40ULL,0xdb643e5a5d7c6f26ULL,0xf2be7d72fbaae767ULL,0xeecaf36ac8bcf546ULL,0x2b5e7bde4fec6562ULL,0xbf7867f98dfa6742ULL,0xd44cf6f4efa82f60ULL,0xe3b67d6edbfa7661ULL,0x6acce32a09be6562ULL,0x6fff1ffb5b6e7e27ULL,0x68e8f8111aa0f60ULL,0xde7e7eeb7ffc7563ULL,0x8d74edf3098c4f65ULL,0xbbb4ef2f4a6c7767ULL,0x8e4b5fd94fe87f66ULL,0xbbbe5f5d2feaff63ULL,0xa477fae2cb6c6d06ULL,0x9f76fc70cf68e763ULL,0xea8a640000bc5664ULL,0xcfc6ff8213ee7765ULL,0x3f3f3352dd60fe63ULL,0xb0f2dfc7db4cfd63ULL,0x994b2770ed68e422ULL,0xca081e6a85406e03ULL,0xedd5ea3f0fea6763ULL,0xcef23f4789ea4560ULL,0x8bb16fbb0aec7763ULL,0x7964767390bc2764ULL,0xda3e7673da2c3746ULL,0xaffeffbacbeaff67ULL,0xac771b5bb9627765ULL,0xb43b24440d424162ULL,0x54bc2e065f406722ULL,0x56f875544040f462ULL,0xcbbae97b8fae7760ULL,0x606806d128600660ULL,0x2e8dc5411b425463ULL,0x9525006a09ec2c60ULL,0x0,0x0,0x800020000080020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4200004001ULL,0x4008000300000ULL,0x1800000004040ULL,0x84080101200c00ULL,0x0,0x0,0x0,0x80400000404000ULL,0x0,0x0,0x10020000000600ULL,0x0,0x102000000000060ULL,0x0,0x800000801020000ULL,0x100860801380400ULL,0x2806440118a0500ULL,0x8bbbbb27152ce520ULL,0x8000520002000460ULL,0x80c00040a000ULL,0x1100000000000060ULL,0x444000404060ULL,0x400004000000400ULL,0x4c30045003000ULL,0x0,0x2100004002ULL,0x0,0x0,0x0,0x0,0x0,0x84c008084100ULL,0x0,0x0,0x24000000400ULL,0x0,0x20020000000400ULL,0x718086021801520ULL,0x8a80f00041403000ULL,0x8082191005646002ULL,0xa4b6c67083883660ULL,0x3074720589426743ULL,0x21200d0042246101ULL,0x110030d03422400ULL,0xc0c619510d646e00ULL,0xd912cdda9b2e7562ULL,0x6496d0e0112eec00ULL,0xa31eb63008747665ULL,0x9392ef7c5bfae541ULL,0xc5ce328f4fb84562ULL,0xa3dfdfdbdbfcff62ULL,0xb3bc5460db487722ULL,0x2006056102080760ULL,0xd2b43f32d97c6d62ULL,0x4c59941001884f40ULL,0xef7ef401d8387764ULL,0x308a4741013a4720ULL,0x8110420109005003ULL,0x3800050001004921ULL,0x0,0x0,0x148936e85b026503ULL,0xbffe97f9c9ee7762ULL,0x0,0x0,0x0,0x0,0x40100000200040ULL,0x2010000000000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000200000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8188662286f64d20ULL,0xffdeffff9fbe7f79ULL,0xcbb5ff33969f7d70ULL,0xe49a3c8a8a0e3d20ULL,0xbfbfffff8ff6ef60ULL,0xffffffff9ffeff79ULL,0xfffe7ffb9fffff69ULL,0xfbffffff9fffff79ULL,0xffffffff9eff7f69ULL,0xffbfdf4a1ebe7d60ULL,0xf5fff66e9ed76f69ULL,0xfffffff39fbeef70ULL,0xdfe7ff7f8fffef71ULL,0xcfffffbf1ffe7f70ULL,0xfffedfdf9fdffd61ULL,0xf4fbf6ec97bf7d68ULL,0x9beef77a9ebefd68ULL,0xffbfff751fb7cd60ULL,0xdde7ffba1efe0f60ULL,0x2f7fdf5b9bfecf61ULL,0xf4be5fa196bb6f60ULL,0xddddf7fa97b76f69ULL,0xbfaf67769ff6ed79ULL,0xcbc1f7779bb5ae61ULL,0xa5fededf9fbf3f68ULL,0x162b4fe500f67f61ULL,0xf2fef7fa9fde2570ULL,0xffeeff771bbf6f68ULL,0x98ff62e07d30d40ULL,0xf7ffbfdf9fb6df69ULL,0xbe7ecf7b0ffe6f68ULL,0xd6f7f7fd97ffed69ULL,0xefee7deb12be7f61ULL,0x3bd5e33a9ef84d68ULL,0xef77fbdf9efe6f71ULL,0xffbfcfff9fffff61ULL,0xfddfbfef0ffffd61ULL,0xd7fffd721ebe4d70ULL,0x7aaf77fc97e76f71ULL,0x4ddffe1b8ef67d60ULL,0xabfcdddf9ed64f70ULL,0xfcb8f6ee96faef68ULL,0xfbefa7718a9fbf78ULL,0xfbbc7f6f9bde0d68ULL,0xefdfffba1fbf2f79ULL,0xafdecfdb9fbe6d70ULL,0xbdff7ff79afe7f60ULL,0xffffff7f9fffff79ULL,0xfebf3f7f9ebe7f70ULL,0xedfed39f15b61d61ULL,0xeb6bdfdf83de2f60ULL,0xfffffff79f9f7d69ULL,0xfdfdaeca9ebe6f60ULL,0xfff6ffff1e9e6f61ULL,0xebffff7b9ffe6f78ULL,0xf9f71edf9ef6df68ULL,0xbbbdcffd1ebe3f78ULL,0xc177566201974e60ULL,0x4bfe2df31fda7d70ULL,0xcbf47f2b16d6ef60ULL,0xeecf9f530ffa3f70ULL,0xbfbfff5b1fff7f68ULL,0xdd75366d86ba0f78ULL,0xf7bbff7287bf7d71ULL,0xdb9cf5ff86bf2f79ULL,0x8787ee0b1e964d68ULL,0x89df9d521b9fed68ULL,0x9081f06c10826820ULL,0x19f4be0916960540ULL,0xc9cca77c1e926f70ULL,0xc4e8fa1382960d40ULL,0xffffffdf9fffff68ULL,0xf7f6fdff9fb7ff60ULL,0xfb4ca76e12d63c60ULL,0xffbeef7f9ffeff61ULL,0x8ff6f73f1ef6dd79ULL,0x49deddc30e1a2d70ULL,0xfb27d3094920f60ULL,0x95efb74a07f7ff70ULL,0xff2c7e6416b62f20ULL,0xe9ffff2f97f66d60ULL,0xafc61afb9adeed70ULL,0xc1b647f00ada2d70ULL,0xf4d3f6f99e974f70ULL,0x3b3e4f730fff7f60ULL,0xfbdfef2e9ef7ff60ULL,0xefffbbfb9bbeef71ULL,0x3f9fcfcf97b66f61ULL,0xbdf7566e16fa6f60ULL,0xbebebfc212f63d68ULL,0xcb22f77517be5f61ULL,0xeddfff3f1eb62f78ULL,0xab72ffdb0696ef61ULL,0xd4f576f583be9d71ULL,0x7daa27e20e360560ULL,0xcab753fe9f960d68ULL,0xfde6ea120f964d70ULL,0xaeffdf5a1fb65560ULL,0x9c9e67e316be0770ULL,0xdd397e520e9cc561ULL,0xfe9e3d6c07f74d60ULL,0xc956db3e969f2d70ULL,0xafefdbdd0eff6f70ULL,0xf6be5ffb9ffedd60ULL,0x860dd56286926f40ULL,0x8eb2654216df3d70ULL,0xd3c5f7fb9e964f69ULL,0xef63ee8049e4960ULL,0xf9fc36b17b64520ULL,0xd77ff77816ba5f60ULL,0x7edc35611e972f40ULL,0xeb9ee73d979e6d78ULL,0xcdeb7ef396356d61ULL,0xda9ecff2029e5f60ULL,0x1477d66a06b48d60ULL,0x639ca51012961f20ULL,0xdabee6269a9e0d20ULL,0x9dfffe1e9eff6f60ULL,0x8fe1cf9b97fbaf60ULL,0xf4d6f2ee96b6ed60ULL,0xd5ea37e197b64f61ULL,0xdfce75fa87fe3f68ULL,0xd167a7360e375d60ULL,0xc98edd9b069e2f60ULL,0xe5977eea8f366d60ULL,0x68ccf7531fbf0f61ULL,0x12828364832a2d60ULL,0x68c6cb2a08860f60ULL,0xf59fa570e320550ULL,0x36915fba022a4d60ULL,0x65efbf6a16be6d60ULL,0xdab65c0c9abc0d20ULL,0x6bd4a373169e4f61ULL,0xeedf3fc486bedf60ULL,0x9a12944906d22b40ULL,0x1430f13780a32f28ULL,0x8200000000060000ULL,0x800010000040020ULL,0x0,0x0,0x0,0x0,0x202000000900000ULL,0x2100000000140ULL,0x0,0x4002000000420ULL,0x0,0x8000200000000820ULL,0x6000000800040ULL,0x0,0x810003a100c12820ULL,0x8000004000000408ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2000420000060140ULL,0x40020000004040ULL,0x1010210000021820ULL,0x9520024a82900d50ULL,0x20052000842800ULL,0x2004010400200540ULL,0x8020108000120100ULL,0x0,0xc110000000000410ULL,0x0,0x0,0x200210800000ULL,0x0,0x4000001000010800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1104800a000400ULL,0x0,0x9200004000020460ULL,0xb40200100000d10ULL,0x8204080100062820ULL,0x888840880020d00ULL,0x14ac0240001a1020ULL,0x1422e4002810c40ULL,0x1280010000900820ULL,0x12100e02080c20ULL,0xa6fedff306afaf68ULL,0x1d76386d1a8e2f40ULL,0xcb3e3f56863bad60ULL,0xc92eff588616bd68ULL,0xc9dab6db060e6d68ULL,0x2fffffd717b76f70ULL,0x90b614e0139a8d61ULL,0x14abf0805860d60ULL,0x8094019010060d01ULL,0x4bc292340e320d40ULL,0xa1460210020a0c20ULL,0x20f4002002408ULL,0x28420000020160ULL,0x400004000020800ULL,0x80000100020800ULL,0x0,0x10332be184220d40ULL,0xc0911061801a0920ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21000100100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd3be6f22f7be7ffeULL,0xfffe7fffffbe7feeULL,0xcbf7fb3ffe9c7ddfULL,0x6fff1f7dc79e35a6ULL,0xffffcffff73e7ffeULL,0xfffffeffff9e7fffULL,0xfffefffbff9e7ffeULL,0xfbfeffffff9e7ffeULL,0xeffbffffffbe7ffeULL,0xbf7fdf7fd79e7fecULL,0xfcf7fef7debe77eeULL,0xffbefff3ff9c7fefULL,0xff97ffffff9e7dffULL,0xcfffffffff9e7fbeULL,0xeadecfdfffae7fecULL,0xf6ffdeeff7be7fffULL,0xd73faf7edf9e7fc7ULL,0xff9f7ffed73e7fe6ULL,0xffc6ff97f7962fe7ULL,0xebfeffd2d71e1decULL,0xddbf5f2bf68e67dcULL,0xff7eff7fff9e7fe7ULL,0xdfff3f7ef7b47fcfULL,0xebf7c306f7be6e6cULL,0x63d7bbd3fe8c6fd7ULL,0x5ebf6df7d79e776eULL,0xdffdef63d79c77efULL,0xfbfef7fff6bc7fceULL,0xebb6f3dfd79e6fceULL,0xbefebef9de1e7deeULL,0xdffdcfffd79e7f4dULL,0xfdffffebffbe7ffeULL,0x4fde3f03f53c57eeULL,0xebffff7ffe9e3ff6ULL,0xeedbbffbfebe7ffcULL,0xffffefffffbe7fe7ULL,0xccf7fffddf363fe6ULL,0x8bfeff73f3be7ffdULL,0xdb87ef3fff3e2765ULL,0x8feffe0bdf9c7fe2ULL,0x2bdedfdaf69e75cfULL,0xdff57eecdfbe07e6ULL,0xbffffffbffbc776cULL,0xff3e7e7bf6bc7745ULL,0xeff7f6bbf73c7766ULL,0xabffdfd3d78e3fd4ULL,0x9cbfdcefcb9e3597ULL,0xf9eebf7e769c3fe6ULL,0xcab63ff0fe0c3f1dULL,0xc9d3d722d49c15c0ULL,0xadeeffffd71e75c4ULL,0xb7fbeffbdf9e27e4ULL,0xbdfdb67bffbe3df7ULL,0xff3e3fecf79c37e4ULL,0xebf6ff7fff9e7fcfULL,0xfddfffdfd6be3fdfULL,0x3fffeff3f61e7717ULL,0xbcfe7673d5be67e3ULL,0x6bdaaf61da2c67e6ULL,0xfbdee77fff9e2fffULL,0xfefafffbfe9e6ffeULL,0x2bbecf5af79c23efULL,0xc1d3c772f61c6726ULL,0xfbf6fdf2deae36f6ULL,0xfbb6ff7ef7be37edULL,0xcdefb894f694355dULL,0x8a5c5c9adb0e4724ULL,0xb0d51ca4560c3544ULL,0xe9fefe6ad68c3c46ULL,0x80182f78d30c6124ULL,0xc84bf7b9d69e6dcdULL,0x8f5ffe51ff9c3ffeULL,0xe7fe5fdbfe3e7de6ULL,0x944df57ffaae36ccULL,0xe9b70458d60e15ceULL,0xc1c3fffbf7942f62ULL,0xea965dddfe8a6544ULL,0x97b14db2fa822dceULL,0xfced3624799a3f05ULL,0xc20630c0b2043404ULL,0xdbfefaaeff9e75eeULL,0x74e623f0cc942cc3ULL,0xcdd8ddf9d79c7d34ULL,0x17e7776fff9e3d74ULL,0x7dbefd73f71c6fc7ULL,0xf9f6ff3ff6862fedULL,0xeecbfebbf71e77d4ULL,0x7fffd7dad7343fd4ULL,0xf0f21eebf68e2fecULL,0xdbf43d70fe1a440eULL,0xef9eeffdf68e3f7eULL,0x8a5effbff79e3fdfULL,0xbf7fdfdbfe1e37e6ULL,0xd49c70adf6827f66ULL,0xb1e7376af7b61de6ULL,0x6fb43fffd49e3fecULL,0xe1d6df3bd79e2365ULL,0xa7f9d7dbf2ac7fe6ULL,0x14ba4d8656a4294cULL,0xa32c3eeb571a3d54ULL,0xb69e3f60d7b67f7dULL,0xcfcafb02d79c3fe6ULL,0x3adfffcff78c25f6ULL,0xb7fecdfef28c7fffULL,0x61382e21d60c03c2ULL,0xbb367f7dfe3c55eeULL,0xc1e6f267f6842d7fULL,0xa6db9693d79c0664ULL,0xf19f4bf29d9c3d44ULL,0xd1e3f629d21e6cbfULL,0xcfbeff72d2167da4ULL,0xabd0f33ed69c2da7ULL,0xc85336bbf71e35c4ULL,0xa9ae4d68f6967464ULL,0x29525bfbda8c7de4ULL,0x9f88fff0d6021d4eULL,0xfbf2ff7eb2966d4eULL,0x4b863223de006c0dULL,0x3a8f5b9cd2907f15ULL,0x953706e05a0e3e26ULL,0x53e72d75d6a672deULL,0x7aae5feedf9c6dc3ULL,0x845ff00bb3822c4cULL,0xedfdd757f2ae3de3ULL,0x446355e2f2ac23e4ULL,0xf3b87ff3d59e6dacULL,0x870e0868d6944164ULL,0x29c4d228d18421a0ULL,0xc173800120c0064ULL,0x90026cc0402428d1ULL,0x214d3719c4044140ULL,0x53423d20c6802444ULL,0x8bb3e223d6044b84ULL,0x825e1042f1140184ULL,0xa100542408005d4ULL,0x84002260800c1400ULL,0x0,0x0,0x0,0xc000080040000ULL,0x0,0x20040000002800ULL,0x0,0x200010000020ULL,0x0,0x0,0x0,0x0,0x0,0x800080040000080ULL,0x0,0x2004080000100ULL,0x4020244044802000ULL,0x0,0x0,0x0,0x0,0x804000020000400ULL,0x0,0x2100010000040ULL,0x18428202002904ULL,0x0,0x44322541500e0480ULL,0x8b104143560000a6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200100200040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000010000200ULL,0x0,0x0,0x22441b200010cULL,0x22162540008c0ULL,0x28161050b01c0440ULL,0xb0024f49d6001606ULL,0x40060040000184ULL,0xcdfe3d62f614694cULL,0xb990e524de8c25c4ULL,0xc8e3af38deac6fc5ULL,0x10074030080500ULL,0x90b17269d6102646ULL,0xdf9b7760f68e35d5ULL,0xd704fc60d32631deULL,0xfdcef23bf7bc3eedULL,0xeefedfd8f3be7fceULL,0x78b07420d39029c4ULL,0x804f2450d2862422ULL,0x48042491d0807450ULL,0x4012168044042880ULL,0x20d3905890a82544ULL,0x91c0a052040182ULL,0x8110006210002a20ULL,0x0,0x0,0x0,0x841052e836102e04ULL,0x9228dc14da0414c4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10a0844000000ULL,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x21000000300ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd3beef728bf577ffULL,0xfff6ffffcbfe7fffULL,0xebfbffffcbfd7ffaULL,0xff7f5f7bcb9e6fbfULL,0xbfffffffabff6fffULL,0xbfffff7fabfd7f77ULL,0xffbe7ff3abff67fbULL,0xfbfffbffebff6fffULL,0xeffbffffcbff77fbULL,0xffbfffff8bfd6fffULL,0xbdf6ffff8bfd6fefULL,0xf7feef77cbbe6bfbULL,0xdb97ff7e8bbd6fdbULL,0xcfeffe6f8bbd7afaULL,0x3bbedfdb8bbe66bfULL,0xfcf7feffebff7fffULL,0xdbefee7fcbfb7bffULL,0xf7b77e7debf86ee3ULL,0xc7d2fffacbbf7babULL,0xafffffdbabfe67f7ULL,0xfebf777b8bf967ffULL,0xf7fcfffaabfd67ffULL,0xeabf2f6fcbbc77f9ULL,0xcfffdf39eb9573ebULL,0x2bd6dedf8bfb63bfULL,0xbef26f7eab9c6ffbULL,0xf7fdfffeebfa77bfULL,0xdfde7fa68b7c677bULL,0xfafffbfe8bdc73fbULL,0xefdf3fffcbbf7fffULL,0xf7dffffe8bf46fffULL,0xdfffffffcbff7ffbULL,0xdfae3dfbcbff73ffULL,0xfb97eb248be86bbbULL,0x6edf9eafa9db6b37ULL,0xdeffff7febff77ffULL,0x75f7fffaa3be6fffULL,0xdffeeff38bdb7ffbULL,0xfb7afb7f8b9a73baULL,0xcffeffbbcbb77b7bULL,0xebfeff5f8bfa77dfULL,0xf7fdf76febfe67fbULL,0xdcf5ff7ee999639fULL,0xff9f7ff7cbdf7693ULL,0x4fd6ffbfab9a6ff9ULL,0xabffdfdfeb9863f3ULL,0xf5f7dfebebdc67ffULL,0xf5cff7f98bf07ffaULL,0xfeb65f6e8bf673fbULL,0xaffff7efebbc6fffULL,0xaff7dbfc8bc46ffeULL,0xbefb5fee8bfb73faULL,0x7d7df77f8bd472efULL,0xfebe3ded8bf373ffULL,0xf3fefbbaabf477ffULL,0x7def3fd9e9fc767aULL,0xbdfcfd7b89997ffbULL,0xf6bef7f88bf07bfeULL,0xfffa7f708bfa63ecULL,0xdba4fe2fcba863a2ULL,0xfeefbfcfcb946ffbULL,0x7fbfcff2abfe732bULL,0x70ffffeb89fe777bULL,0xdffeffd2c99667bfULL,0x5fac5f7f8b326643ULL,0xc0fbf22389bc67e9ULL,0x2b5b1c528b806227ULL,0xf4b4f6680b4a7b25ULL,0x2627708b186043ULL,0xcdb2777e8b7763eaULL,0xc082e3ba8bdc3f3aULL,0xa7bcd89d09a16743ULL,0xb4fb0fe58be977d3ULL,0x596e2d02aa902735ULL,0xf6342fde8af06ff0ULL,0xe181d7264bc866f8ULL,0x3f8edecba1b463b3ULL,0x8ba2cfb88ab82213ULL,0x9967b4708b982773ULL,0xdb8e7c718bda2b52ULL,0xc9c1f77b8bb8637cULL,0x4ccbdc558b8c62e2ULL,0x569acf7aebb1737bULL,0xf4e3b7fdabb86ff3ULL,0x5f9e7ff14bfd73ebULL,0xbbf4cf3eabb6667bULL,0xafd3bfc38bc56bfbULL,0xeabe6ff3cbd86f7bULL,0xb7fdb76acbb567efULL,0xfffef5b1e9b8633bULL,0xe2b6f2ca8bd863aaULL,0xefdffa93ebb27bf7ULL,0xabdddddf8b9467bfULL,0x34b749e6c1902f72ULL,0xcbdc2770899077e9ULL,0xdf303f6689b56baeULL,0xcc4fef18a91872c9ULL,0x3b2f5c1ee1847fbbULL,0xbcf65f2e8bae6b77ULL,0xd9ffee68abf962ebULL,0xf2a63e62a9a07748ULL,0xc1f6fa2589d223daULL,0xe9feffdb8bb4335bULL,0x9f9beffea99467ffULL,0xd330b64b8bb872baULL,0x7ea67d698b96233bULL,0xe90ef3aa8bdc3f5bULL,0x3e823e9b49d87ab1ULL,0xb4abcdc08bd06b19ULL,0x746f72ea8b08636fULL,0x6b5ee5e1899462ffULL,0xf92cdb2681b43232ULL,0xefefdac3c3b4673bULL,0x49ecd89c820264eULL,0x50d534428b802279ULL,0x2a80f4001900374ULL,0x5954013209e46660ULL,0xca5a32a9ab3e77f1ULL,0x2977dfd38bb266ffULL,0xf7feffea8bfb673fULL,0xd9a5ae62c9386309ULL,0x77bebf6aab9c36bdULL,0x868aa2288b5c620bULL,0xa5a15ca03815a62ULL,0x3d1c03e28b81666aULL,0xfee42f148bd02339ULL,0xdaeab76e83a06e3aULL,0xc26646268b902224ULL,0x18ec84181326a02ULL,0x9f225f60090063e4ULL,0x9180240003104208ULL,0x10a8034009202401ULL,0x88c4010c82802708ULL,0xaf4809488bc0222aULL,0xfa1d6dda89dc62baULL,0x1496b9810908628aULL,0x1004000000000208ULL,0x600009000000ULL,0x2020080000040ULL,0x0,0x2802000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x88c88008a480202ULL,0x10304cc28aa82003ULL,0x1080000000800001ULL,0xa00080000008ULL,0x0,0x0,0x204000001200ULL,0x824a80012003ULL,0x60900c420b880209ULL,0x4000400080000040ULL,0x48000014a002000ULL,0x21000000a000000ULL,0x400100100002300ULL,0x0,0x0,0x0,0x0,0x0,0x100020809420000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000000008400000ULL,0x479231540b0062a8ULL,0x100a28401900070ULL,0x85214050bd86000ULL,0x198310281282213ULL,0xd261fe398931677eULL,0x1647462aba6033aULL,0x5ad16706a93063b3ULL,0xa03bacc88806203ULL,0xa98df1a0bd14327ULL,0x1c41d2280b840328ULL,0x95186d50818061f8ULL,0xc30c777fab3662feULL,0xcc1ff3beb982743ULL,0xbfbeffdbebff6fffULL,0xd0d5f6fe8bf42f13ULL,0x4192a543093063abULL,0x8f1a24240bd01338ULL,0x80000221a2800000ULL,0x6a5e1d5aa9d04261ULL,0xa122056a43027760ULL,0x0,0x0,0x0,0x0,0x8830434080042710ULL,0x9564350609c8272aULL,0x4800000002800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd7beef33fbfdd77eULL,0xfffefffbebfddffeULL,0xebfdff7be9eddfbeULL,0xe8ff7ff5fbf5d9feULL,0xfffffffffbf5dffeULL,0xffffb7ffe3fddbbeULL,0xfffefffbfbf5d7feULL,0xfbfdfffffbf5dffeULL,0xcffffefafbe5cd7eULL,0xefbffffffbfdd77eULL,0xfffffffffbfddffeULL,0xdfbffff7fbfddffeULL,0xfe87fb7aebf5c76aULL,0xdfdfbefffbfddf7aULL,0xebb7df5ae9d5d37eULL,0xfcffffebfbfdd5beULL,0xc1bf3f7ee9f5d37eULL,0xdfbf7e7efb75d33eULL,0xfff7bf3de1d5d33aULL,0xebffdddbcb7dd3faULL,0xf4bf77eddbf5d37aULL,0xfffdb77ef9d5d3feULL,0xfebf6fe7f9f4dd2eULL,0xd5efeeb4c96cd53aULL,0xae7e5dcfc9c4d53aULL,0xb6fbffebfadcd96aULL,0xfdfff7fbfbfddfbeULL,0x75fe6f7dfbd4c3baULL,0xebf7ff3fd0fddb3eULL,0xefffff7ffb75df7eULL,0xffbecffbfbf5dffaULL,0xfffffffffbfddbfaULL,0xfffefffbfbfdd7faULL,0x9bf4f5f6f964c77aULL,0xaf7fbfdfabdcd73eULL,0xefbfdffffbfddffeULL,0xf5bff7ebebd5db7eULL,0xf6feff51caf5d22eULL,0xfba67f7fcb7cd73aULL,0xedfeff7df3fddbbaULL,0xabdfffdffbfdd32eULL,0xf6f5dffefbd4d36eULL,0xfff6ff7afbddc37eULL,0xfbb6f77dcbfdd32eULL,0xedc7ff3fd1f5dd3eULL,0x63ffd7d7e3f4cb3aULL,0xbcbf7fedf34c5fbeULL,0xf9edff72cbf4d77eULL,0xdc3e2fead175d13aULL,0xc5f7ffa6fb5cd7faULL,0xeeaffaebf1fdd32eULL,0xb7fffffffbf5d77eULL,0xf37dfff9f3f5ddbeULL,0xfff67f7dcbf4df6eULL,0xebfdff36ebf5d13eULL,0x7bff1f1fd3edd90aULL,0xb7bfcfdbd1f5db7aULL,0xb42edae8c354c35aULL,0x69ba6c51fbdcd1baULL,0x93ccfb16f1f4573aULL,0xffdb7fadfb74c35eULL,0xc892c57af3f8c102ULL,0x7465d76bd3d4d07aULL,0xdafefd63fb54d3faULL,0xfb9ce737ebfcd72aULL,0xc9c7bf03c165c7feULL,0x27d6cf57f1c4d12eULL,0x86f6d6ebebfdd16eULL,0x81eefb63cb6c50baULL,0xfa121f62fb64c32aULL,0x5802810a6048514aULL,0x2b1edf54c0f5c23aULL,0x94173fe38b3dd52eULL,0x736f075ee164d96eULL,0x7ebe3dd5e935d91eULL,0xc1bdf626db2053beULL,0x940081001484082ULL,0x94f15ee0434c511aULL,0xf7eda65dc0c5c93aULL,0x52822de5c155c376ULL,0xe9a9c130c8d45168ULL,0xa8163e455b45c32aULL,0xb29de762d354c33eULL,0xd5fef6d5f9c4d30aULL,0x7bfe7f63f3d4ddeeULL,0xf1d5d36cf354d3beULL,0xadd7baabdbc5d13eULL,0x7fbfefd1ebd5d3beULL,0x14f67fceebe4c13eULL,0xcb2cfd33f1d0d16eULL,0xfbda6f2ee330d7a8ULL,0xe6eff7bbebe4d36eULL,0x33ceddd8db64c15aULL,0x347edeede344d3faULL,0xf10fa7f3d9d4db3aULL,0xfe32273cdac1c90aULL,0xc9c0cb3cf1c4411eULL,0xccfddfdbcbfc511aULL,0xd7fe7de7d175d3faULL,0xf93c6e807045c31aULL,0xface257551c4d37aULL,0xcdd2b16bc944d3faULL,0x2ff7b43d1d0df7aULL,0x9fbbf7efebdddb1aULL,0x536b376acbdc533eULL,0xa7143d3dd1e4df6eULL,0xeb97eb26f3fcdb3aULL,0x6ec4bfefcbe4d05aULL,0x8592ffe24bdc51daULL,0xdef7f2f7ebf4c76eULL,0x10c3d32e844d31aULL,0xd964fb0ed041d10aULL,0xf5d39a59794cc63aULL,0xe7dccf51d3d4573eULL,0x14fc0e6048ccc122ULL,0x8f867d6181ecd3daULL,0xffb4e372dbdcd32aULL,0x595236026105c72aULL,0xbfafdf92e1d4d33aULL,0xf4f26c68d8ccd52aULL,0x9945ad76c0409136ULL,0x92b62f25d8545358ULL,0xc843ee225041c11aULL,0xafdecb59c0c5d02aULL,0xa9b24da680a1c12aULL,0x5d687756e1f4da1eULL,0xfabe3f76fb55c3aeULL,0xccc99618e8a1c01aULL,0xab5184650444132ULL,0x622cf52c004d11aULL,0x508026688104930aULL,0xefdc61a3c0f5d01aULL,0xda50630f8075931aULL,0x21410628f0e0d008ULL,0xc09446fac20ddb12ULL,0xd0d5e6cccb54d3beULL,0x2010000040020ULL,0x0,0x0,0x0,0x0,0x8080000004008ULL,0x0,0x0,0x0,0x0,0x0,0x4002080001000ULL,0xc0001040000ULL,0x60000000004002ULL,0x90000040004000ULL,0x8080000000400008ULL,0x0,0x0,0x1012000000802aULL,0x0,0x8008a84041041200ULL,0x80010000400002ULL,0x200200001004000ULL,0x2100000000008020ULL,0x2010050000000ULL,0x840e000480219012ULL,0x12005000000120ULL,0x0,0x2c110090044000aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x420000003000000ULL,0x24000002000002ULL,0x0,0x2220240085000ULL,0x24040000000ULL,0x0,0x0,0x8200000080000004ULL,0x0,0x0,0x0,0x0,0x821449c08044d358ULL,0x8bc68a20cb55d102ULL,0x28622048f204d202ULL,0x104041280001104ULL,0x95e9fffbdb14d7beULL,0xb2c3522e0e5c52aULL,0xd9897e1dd1c453deULL,0xc3c03ec77155c322ULL,0x8fee4f136345d12aULL,0x64f6f6f9f17043baULL,0xcbee7f73fbfcd33aULL,0xcb9aefbff37ccb3aULL,0xaefefa3bebf5c12aULL,0xffffdfdbfbfddffeULL,0xfe15b7e7c374c76aULL,0xd77e93eaf1add23aULL,0xff4a7b76f3c45b4eULL,0x86d28a204154c102ULL,0xeb7ec9c3ebc4d31eULL,0x2420101040120ULL,0x1008e2455000d108ULL,0x0,0x0,0x0,0x30404048001002ULL,0x9202031081240a02ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1080000080000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x93beef7382dffbe6ULL,0xfffe7fffcadffff6ULL,0xebffffffcadffff6ULL,0xfffffeffc2d7fff6ULL,0x7fbfcff3cad6fff6ULL,0xfefff773cadffff6ULL,0x7fbe7ff3cadffef6ULL,0xfbffef2fcacfffe6ULL,0xfad7bedfcadff7f6ULL,0xfeb77ff9cacff7f6ULL,0xfdfffffecacffff6ULL,0xdfbfff73cacfff76ULL,0xfbadeb3f8adefaa6ULL,0xefceff3ecadeffe6ULL,0xabffdfdf4adff3f6ULL,0xe5a255fcc8c3ff34ULL,0xd97ee77ac0defbd2ULL,0x7abfff7fcac3f7a6ULL,0x94cdff3bcacbf6e0ULL,0x6b7f97d3ca86f3e2ULL,0xbfffd2facacbf3e4ULL,0xfbdfff6ccadbfff6ULL,0xefd63d6ec2d3f3e0ULL,0xeff6ffb9cadbde74ULL,0xe957de55cadaf2f4ULL,0x3fbfe7fac2c7f7f4ULL,0xd3bdff77cac6f3e6ULL,0xef967d93c2d6ffb6ULL,0xabfdf27fcac6f766ULL,0xcbe75edecacefb20ULL,0xf6ffffefcad7fff6ULL,0xffffffffcadffff6ULL,0xfffefffbcadffff6ULL,0xfbd5db7fc8dffb34ULL,0xefe7ffffcad7fff2ULL,0xffbfdfffcacffff6ULL,0x7cfff6facad7fbe4ULL,0x1f1827f1c2dfb262ULL,0xfe99ff7fcacbffe6ULL,0xeff7ef3fcadefff6ULL,0xeb7ddfdfcad6fee6ULL,0xb6fdf669cac7bff4ULL,0x9bafe767cadbf7f0ULL,0xdbb4ffffcadabf70ULL,0xffffffbecad7ffe6ULL,0xef5bfb584a96f366ULL,0x1dbf5fafcacef2e6ULL,0xbbebf77acadfb3f6ULL,0xfe9b1f74c0d2fb54ULL,0xea86fff6cad7f766ULL,0xa3fedeffcacfff62ULL,0xbfb32fefcadffef2ULL,0xf37cbe73c896ffb2ULL,0xffbcf7fcc8d3f7f4ULL,0xa9f7eb67c8d7f776ULL,0xe25d1ffbc2c7eaa6ULL,0x9fbfdff34adbb6f2ULL,0xffd5ff73cadbf7b2ULL,0x7bfe7d73cadff6f6ULL,0xdbdeefeec2c2ffb0ULL,0x6e8e9f4fc246f726ULL,0x2abecd93cacfb6a4ULL,0x75f7d7fccacef7f6ULL,0xd7faef71cad6f7f6ULL,0xf39cf75dcad6bf26ULL,0x8bc7aa17c8dbf7c0ULL,0x9a4fdd51cac2f2f4ULL,0xdcfdf6fbcadff776ULL,0x9d7f677bca57fff6ULL,0xd3022f05c8d6f324ULL,0xc080a73bca83b024ULL,0x2845ddc202c7a252ULL,0x964b57ea8a41ab60ULL,0xd13eefc38acafb22ULL,0x4a842fe68a45f6a2ULL,0xc286f0fac2c5ae22ULL,0x156740042402a24ULL,0xb6924ff2024bfec0ULL,0xd064fdec48cafe22ULL,0x8b107544c8d233b4ULL,0x8934bbb9cad3f774ULL,0xfe5e36a9ca4abfb4ULL,0xb8abcd6acac6f6a6ULL,0xf17ed7d8cac7eff6ULL,0x5bb63f73cac3f7e6ULL,0xdb76fb37cac2f722ULL,0xeb139fcecadbb9e2ULL,0x339ecfbacad7be36ULL,0x1ff3feeecaceffb6ULL,0x9bb4bd71ca5bfbb6ULL,0xfbbefd2bc2d3f3d2ULL,0xe7fffbffcadffbe6ULL,0xaf1fdfceca43fb36ULL,0xf4f7b6e54043f3a6ULL,0xfe627d71c8d67ba6ULL,0xff76ae70cacfb3f2ULL,0x5dc3fbbac251b594ULL,0xa9f6dfd0c0cfffe6ULL,0xadb66de24acbbff4ULL,0x91e5b5c2c893b382ULL,0xbaae2d68ca43bba4ULL,0xc100d665c242a320ULL,0x6f9e1ed5c8dfbba2ULL,0xc7ba5debc2dbbbf2ULL,0xfdfddefecadef7e6ULL,0xbf86f57c8203b760ULL,0xe8eeffabcadafbf6ULL,0xcdde1a4cca13fbe6ULL,0xc69bcffbca92bf66ULL,0x9276deebcad7f336ULL,0xffba6f71cad7fb76ULL,0xd914e5244842de24ULL,0xcefb3b49c2cfffa0ULL,0x1e576f5a4acb9360ULL,0x777157efc8cfff66ULL,0x76be8ff0ca47fbf6ULL,0xd38669e28acef720ULL,0x8be7de3ac2d3fab6ULL,0xbbdfdfdbcadffff6ULL,0xfef774e0cac7b3e2ULL,0x996c6e3a8883f0e2ULL,0xbb042900c08bb326ULL,0xcec1df2a8a9ef2e2ULL,0x24001520200a062ULL,0x2daaf16b8051f380ULL,0x884e706240c58234ULL,0x963a3de8ca43fef4ULL,0xc0c278034a4a3482ULL,0x245bcc190843a722ULL,0x34b264b18295b222ULL,0xd33c44bb4a02f2e0ULL,0x2b467c6088ce1a06ULL,0xf156eb2748dbbeb0ULL,0xad4e020e8842ba24ULL,0x5f084a300cbb6e0ULL,0xd585d264c2d6f120ULL,0x0,0x44000040000020ULL,0x0,0x0,0x0,0x0,0x0,0x80800008400000ULL,0x0,0x0,0x0,0x0,0x100800000018000ULL,0x0,0x0,0x0,0xc2040000010ULL,0x0,0x6000000008020ULL,0x0,0x0,0x0,0x0,0x2300000028120ULL,0x84c846818201a220ULL,0x404a81000862022ULL,0x48027440c2002a04ULL,0x202c00000400e00ULL,0x0,0x0,0x0,0x10001000002000ULL,0x0,0x4842101042005204ULL,0x0,0x8000002000020020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x820000008020ULL,0x0,0x0,0x0,0x3c5e0960c8125204ULL,0xe390e72a48caf722ULL,0xebfa34d6cac3bf70ULL,0x110bc00002a3c0ULL,0xf0f5f65ecac7f3b6ULL,0xdfc42d02484b36d4ULL,0xc174234fc042ab44ULL,0x8bd39647c0cbfba0ULL,0xe03e8ea3ca46db72ULL,0x14f62ee042daf230ULL,0x9bb40d6188c3eab0ULL,0xdb9c7ff4cad6fff6ULL,0x4d5afa1d8842fbe2ULL,0xafdecfd3cad3baf0ULL,0xf6ffffeacacfff76ULL,0xdd8e5d7a48c6aef6ULL,0xdf9ffffbcacfb3f6ULL,0xc1ceb62e4a42ab96ULL,0xad4355c7c247fbe6ULL,0x30801d0182840a04ULL,0x8c19a2124285d306ULL,0x0,0x0,0x21040000004ULL,0x0,0x50002002080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200000008200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x6000109000002022ULL,0x1002008000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb204a70144026811ULL,0xff36ffffcecabb99ULL,0xfbffffffefffff9bULL,0xefffbfffefffffbbULL,0xff9edffbefebff93ULL,0xfef77ef9cfd2bbb1ULL,0x7ffefffbcfdbffb9ULL,0xfbdfed2fafc7ff39ULL,0xfffbbfdfcfc3bfb3ULL,0x7ffedfffcef7fb99ULL,0xfdfcff7aefdafb9bULL,0xbfbffdf2eedbfb93ULL,0xddbefbff8fdffa92ULL,0xcecfff1f8de3fe91ULL,0x2a5e17de0fd1fe93ULL,0xb6bfd6efafd37f91ULL,0xffaffff3afebbebbULL,0xff966ff38ec3b219ULL,0xcdc7e7178ec3ff89ULL,0x2b9e7d5b0ff2fa93ULL,0xfcfe4ff3afd3bb91ULL,0xdf3eff7aefd7fbb3ULL,0xfe16fff58cca7e91ULL,0xcffbffbfefd7ffbbULL,0x67dffedaadcb3f89ULL,0xccaadf7f8cda7bbbULL,0xd6efff69afc33f83ULL,0xfbeeff66cdeb6f9bULL,0xabc74b2d8dd3fa01ULL,0x96fbbf53eec33b39ULL,0xf7fbffffaff3bfb3ULL,0xffffffffefdfffbbULL,0xfffefff3cfffffbbULL,0x9bf6cb0ecfd6bb91ULL,0xcaeebe50cee23a81ULL,0xffbfef78efe3bfb9ULL,0xfcff5fff8fc3ff93ULL,0xd4995df2e5c33a29ULL,0xfab7e7feefd3fe91ULL,0xdffffeffefff7fbbULL,0x2fffffdfefdf5b9bULL,0xf6bffff6eedbff39ULL,0xf7f7bf5a0cf3baabULL,0xfb98777dcff77bbbULL,0xfddefe92cfd3feb3ULL,0xeffeffdfcff3ffb9ULL,0xbcfbfbef8f92faa1ULL,0xfdf7fff8cfbbbba1ULL,0xfeaeaff92ee3ff93ULL,0xcff3f31b8fe3eba9ULL,0x2b191fdf0ee17f01ULL,0xb3badf7b4fd3fe99ULL,0xdf7dfffbefd7ffbbULL,0xdffe757b0ff77a89ULL,0xe9dcfb3a2dd79ab3ULL,0x6f77beffcfd7bab9ULL,0x9f3f7ffa8cc2ba29ULL,0xf75dbef9cdf2daabULL,0xefbc6ff3cdc3fabbULL,0xfbdceb4fcecbbf9bULL,0xff4aff5fcfc2fb9bULL,0x4f9fcbffade7fb9bULL,0xfcff327f8dc33bb1ULL,0xbbbebf738fe7feb3ULL,0xeb36e76fafc73e93ULL,0x8dc3f61b8ec3ba81ULL,0x2b1cdd940c961e81ULL,0x76b6f6794fb21f9bULL,0xd9ef7f7a8f97df93ULL,0xbbb829064cf69281ULL,0xd1c6e9168cc2ba81ULL,0x4388df920d036a11ULL,0xa4b9f5e24f823e91ULL,0xf1ff66aa0f8a2f11ULL,0x9cb6bffdce8aff1bULL,0xa9d6c20f0fc97f1aULL,0x1674108e801901ULL,0x841034684a819b91ULL,0xf7ffbfffefffffbbULL,0xca887c6d0ef23a11ULL,0x89c1e10a0c927b01ULL,0x6c7e3bcb8cda9ab1ULL,0x2ebdeff94dc5da8bULL,0xf9ffffff8fc7be9bULL,0xfff6ff73afcbff9bULL,0xfbdeef2eadd3fe3bULL,0xc9d39e5be9674b83ULL,0xdfbfdf5f8ffbff9bULL,0xfdb35ffecf837f91ULL,0xffecfdf2aefbbe99ULL,0xf9bced2605d1ba08ULL,0x8efff70a0dc3bf0bULL,0x8bdfdf9b8fe2bb89ULL,0xd4f776ed8fc2be83ULL,0xdae6d66ab42bb01ULL,0x9fba6fdf8ec6bf1bULL,0xfcaef3358ec78b39ULL,0xe3f76432f8fba03ULL,0xffd65f608df3db83ULL,0x537aff788ceaa38bULL,0xfaaeaf63c6c77a13ULL,0xc592c60604c69601ULL,0xeffefa4feef7b2bbULL,0x94937fea44cfba13ULL,0x722cf7798fc2ae91ULL,0x582c65620cc29e89ULL,0xebfffa3e2cfbfe0bULL,0xedf3be7e4fcebf13ULL,0xef3fdf76efcabf11ULL,0xf5e7f67f0fd9ba9bULL,0xf3b47d70cfc17b8bULL,0xa9c5ef9809a18e83ULL,0x8a533b1acda35a81ULL,0x6e16efc84cd37f99ULL,0x84de57a045c4be21ULL,0x9ffefdd48ff3ba99ULL,0xe2866e2a4ec09b11ULL,0xdbdafa79ce9bfe99ULL,0xeb1fdfdb8fdffab9ULL,0xf5ffdeeeedcffe9bULL,0x9fed2d528ccfa333ULL,0xd7100e22cc0a1881ULL,0x80a23e088ac39b01ULL,0x104c0008808001ULL,0x89120fa38fa20800ULL,0x914aa74085c23a11ULL,0xd285031249010a03ULL,0xcddbba848eaa3901ULL,0xaf49f74a0dabfa89ULL,0x8b8cdae8ec22e21ULL,0x90e3077a4a239b93ULL,0xd2820c1288c31ab0ULL,0xabd5eb368e92b891ULL,0x49801b0503431a01ULL,0x26d0d5890d821801ULL,0x911e1aea0d833e98ULL,0x0,0x0,0x8020000000010200ULL,0x0,0x0,0x8000040000008001ULL,0x0,0x40001001000800ULL,0x0,0x2020000004020000ULL,0x0,0x0,0x0,0x0,0x24000020008ULL,0x804000000000600ULL,0x0,0x100000100001800ULL,0x0,0x0,0x0,0x0,0x10808000100ULL,0x800010000004800ULL,0x1000070080030800ULL,0x9012028005422a00ULL,0x1a00250008401a81ULL,0xd200000602800281ULL,0xc0081006002000ULL,0x0,0x40a0045008000630ULL,0x8200008000100ULL,0x9000016809801a01ULL,0x0,0x0,0x1220280c012280ULL,0x8421001001800ULL,0x804020800010202ULL,0x0,0x44000802400201ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x18c820008023101ULL,0xf6bc7fd7cdd3fa1bULL,0x890632f00c90db09ULL,0xaf77b3de6f82fe9bULL,0x9013054a0c007801ULL,0xd1f5a02e0d8b7e09ULL,0x5fbefc488c82ffa0ULL,0xa1d4c7260c527a11ULL,0xe9c7d757ed621e89ULL,0x5929c9d3aed2be01ULL,0xf5f676fccfc39e99ULL,0xbf2aff73efc2aa83ULL,0xe9d9ae6e682226b9ULL,0xcfd7ff3bcff3bb93ULL,0xafddcfdaceea3bb1ULL,0x7eff5ffbcfdebe9bULL,0xffef7772cbd7fb8bULL,0xfbf67ffecffbbb91ULL,0x6944a50a0d4a2e03ULL,0x2f3f7ed2aee3fb21ULL,0x302809010c062288ULL,0x5260aa5005816800ULL,0x80010000000201ULL,0x2000406008c01201ULL,0x0,0x0,0x240208000110ULL,0x0,0x200000001800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x14000100002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6f5dedcb60311150ULL,0xbffbffff663117d4ULL,0xffffffff677717deULL,0xfffefffdff7717ffULL,0xfbffffffff7e97ffULL,0xffffffffff7f97ffULL,0x9fffdfffff7f15d9ULL,0xf5f756fd3f7e11deULL,0xf7be3ffb7e7e14dfULL,0xfbcfeffff77b17dfULL,0xffffffffff7f97feULL,0xbb3fcfffe67397feULL,0xf5d7deeb737a15feULL,0xffbdff73e27e95feULL,0xfebfff6e367e14feULL,0xdffffe2fe67615ffULL,0xebdecd5f735a15feULL,0xf4fefcf0607f95ebULL,0xd9acef78a25611daULL,0xefbaef77657b91dfULL,0x8fdfff8baf7b17fbULL,0x2afeffdb6b7714deULL,0x77ff57eb7a7e95d8ULL,0xf9cff77be57f16daULL,0xd6b627f7aa7615dfULL,0xedefffbbb67a17fbULL,0xa1dedc53337611caULL,0x3cba4fffa27a93d2ULL,0xddfcb652245316feULL,0xebc63ffbf27a155eULL,0xcbfff63a385e93ceULL,0xfff717daef7717faULL,0xfffbfffbef7f97deULL,0xfdfffff9f37f15ffULL,0xfffe7dfae77395efULL,0xfb7ef77f677f95deULL,0x475f7fc3a05e95deULL,0xef1fdf76ba7215dfULL,0xd5b6f6ed237615ffULL,0xdffc6df36f7297daULL,0xfbbfffffb77e11ceULL,0x9ddffebd377717ffULL,0xbbdfffdf277795f7ULL,0xf4bffff2af7f15faULL,0xd1febf6aa77691dfULL,0xdf95bf6ce37294ceULL,0xfdffffbfe77717dfULL,0xef5ed7d1237a94ffULL,0xdcfbffef737e956aULL,0xddfffff8e47691deULL,0xfebeff7ff27797dfULL,0xefe7f3bfef7717fbULL,0x6fffffdba27e1777ULL,0x3ff2d77e247715deULL,0xf0ed37e3627a12c8ULL,0xff9e3fff6f7f15feULL,0xc9fff77fe77f10eeULL,0xe9ff7ffbef7f95deULL,0xdf3deffbea7615dfULL,0xddf7be7ef27b17faULL,0x5fee7debbb5e94fbULL,0xf9d4bf3f627312ffULL,0xce7bb7f76c7b15d7ULL,0xbfbfdfdafe7f17deULL,0xfe94d6eaa77a11dbULL,0xdfbaed52267a96deULL,0xdb9f7f6fe67610eeULL,0x49cffb0ae15017faULL,0xab77dfdb6373167aULL,0x3af97660a272115aULL,0x8090210002001642ULL,0x40080d00004c004aULL,0x405196a000621549ULL,0xabdcd65f297314deULL,0xf5b75febe356154eULL,0x9feda5ea357415ceULL,0xc6a61fe6327615c7ULL,0x8949f91262761450ULL,0x2fd84a1c816215c3ULL,0x9d885fb9207410caULL,0xf1f7b766e067105bULL,0xcfb63ffe6f7f125fULL,0xebf6fe7fef7e15eeULL,0xffffbfffe37b95feULL,0x3fbcefffef7f17ffULL,0xfcf6feffae7396dfULL,0xfbfeff73ff7f95efULL,0xfb70fbf7677f10feULL,0xc9dbbfdfb77a93fbULL,0xf778cd7e7364107fULL,0x9ddf16e86a7b91dbULL,0xcfbefd636d5617deULL,0xdb9ff7752276916eULL,0xcff3ff93ee7795ffULL,0xeff65d87a77713deULL,0xf4b6d6a6ef7215deULL,0xfdff6e74237e94eeULL,0xbeb12d06b27615d3ULL,0xd9dffab2676f91cbULL,0xeb6ffdd9e27a15faULL,0x8c971f6a7a7295deULL,0xd1681fb8b15414caULL,0x848a2949204114cfULL,0x2596f73e617515e8ULL,0x6dff5edfff7714dfULL,0xb6f35ffa277e13ceULL,0xd0fc777bb07a11ceULL,0xab265fe5207313c9ULL,0xcac6f37e327495deULL,0x6fe77ed3247e11dfULL,0xf92cd42a05614cfULL,0xd53776f0267b91ceULL,0xdf9e2f71277314caULL,0xc395f110227200ceULL,0x424b9e4b225811f2ULL,0x8b3e4c7c287295d2ULL,0x75be37c6a27710dfULL,0xdb2abf60e17315deULL,0x82a06c720174114aULL,0x8973fa2be61817d2ULL,0xaffecfdfe77795caULL,0xb7ff7fe6ef7795deULL,0xd92db672227a11dbULL,0x8a04622464601058ULL,0xc000500224600050ULL,0x8a775ed0307210ebULL,0x651106000601050ULL,0xa46ca708007b1458ULL,0x8a04212902531446ULL,0xefeff7fbf77b15d4ULL,0x6e567fd9a26b11deULL,0x30106b02007411c0ULL,0x8000025060421400ULL,0x80004000500000ULL,0x4190f20368481668ULL,0x2115000440006ULL,0x883c6ca070721404ULL,0x9c75d311a07317feULL,0x0,0x8000080022000000ULL,0x0,0x804000200040ULL,0x0,0x100020000600000ULL,0x0,0x0,0xc00000600000ULL,0x0,0x8404000820121000ULL,0x20620062000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xc0000001100ULL,0x0,0x0,0x0,0x0,0xc480000a2000002ULL,0x8002083200501100ULL,0x4424142820620040ULL,0xc427540a21811c2ULL,0x44020810101002ULL,0x0,0x2400000000400008ULL,0x0,0x0,0x8040022000000ULL,0x2100002000002ULL,0x10040000001010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000002000100040ULL,0x0,0xcf167eb7277f10e3ULL,0x895fda28227014d0ULL,0x8b4e0e4022721060ULL,0x7fbfcefb675317deULL,0xfaf7f67c6a7f97eeULL,0xdbeefd63237e97ddULL,0xe354f73ea37210dbULL,0x4edb3e7b2f4217dfULL,0xb3c4754207315c0ULL,0xb467dec2637a11deULL,0x8dbe1ff0a67a905aULL,0xf0207366235215e0ULL,0x4edbfadfa77f90eeULL,0xffffdfdbe57717dfULL,0xffffffffef7f95ffULL,0xfdffaf7fad7f17dfULL,0xdffe7fe6377e14eeULL,0xc47b3b0860721256ULL,0xa1dedcda257e15e6ULL,0x2300422a040024eULL,0x8040040000600000ULL,0x0,0x8000100000200040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000020020000ULL,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x100640288002800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4040200004004800ULL,0x7aa53ff60b7dccdfULL,0xefffffffdf7defffULL,0xffffffffdf7fefffULL,0xffffffffdf7fefffULL,0xffffffffdf7fefffULL,0xfffeffffdf7fefffULL,0xfbffffffdf7fefffULL,0xfeffbfffdf3feef6ULL,0x1f9eff734b3febbfULL,0xdff5f7fedf7eeff7ULL,0x56ae2d7b5e3fefe1ULL,0xfbfff72fdf7fefffULL,0xffffffffdf7fefffULL,0xffbedfff9f7febf7ULL,0xfd7ff6631f3ecff4ULL,0xd7f6ff715b3fefffULL,0xf9bdb77e9e3dcef2ULL,0xcdeefe3b8f3cebecULL,0xfb588dcf863dea96ULL,0x54b773ea0e1fe2cfULL,0x8ffbf77e9e3dedfdULL,0xcbbe7fff1f3fe8ddULL,0xdff7f3ab9f3fcbddULL,0xffdfdfdb9b3eeffeULL,0xbdb77dfbdf3feffdULL,0xffffffeedf7eeff6ULL,0xdaf6fff55b3fefd6ULL,0xf9e4d39c4a7e24fcULL,0x3b5efed14b3d6cbcULL,0xbfb2ffff5f7deff2ULL,0xff5fbf6bcf1fcdfeULL,0x7fbe7dfdc53ecfe6ULL,0x8baedb7f1e3f63dcULL,0x2fdf3ffb8b3eefb7ULL,0xffffcfebcb7eeef6ULL,0xf7fff6790b3feff6ULL,0x5ffe7ff3df7eeffdULL,0xfbf6fff7cf7feffdULL,0xeff3ffffcf7ecfffULL,0xffffdfeadf7feff7ULL,0x8f7d57694d3deaf4ULL,0xdbfe75735f7ccefeULL,0xbfd4e76e4f3defdbULL,0xefdeb73fdf3fefffULL,0x3ffddfcb9f7feadcULL,0xf4be7fe7cf7eefd2ULL,0xfbffb77a5f7deefeULL,0xdfb7bf7f9e3eeefdULL,0xfdf7ff2a5b7eefddULL,0xefff5741da1e2ff4ULL,0x9d3f7ff2cf3defd6ULL,0xf5fbeff2df7eeffeULL,0xeebe3c769b7ec8d4ULL,0xe7dffb1f5b7fedfdULL,0xffffdfdd5b3cabffULL,0xb63b4ffe163ece95ULL,0xd0ffb75f1e7eeaedULL,0xfbde7f534f3feedcULL,0xeb95f73fdf3feffeULL,0x6edf3eff8f7feafeULL,0x59bdcf71c83c4ab4ULL,0x99ffb76e1b37cfcfULL,0x6fbe77e18f3eeeefULL,0xbbfcff7fdf3deff4ULL,0xfb53be8b453d6cd4ULL,0xa7dccc7f9f3dcac6ULL,0xe73f76f81f3c4f9cULL,0xcbaaf7430f2fefd8ULL,0xde80ff764b3fc3c7ULL,0xdad7ffa59a3fcf80ULL,0xe1bfddd14f3ee08cULL,0xb4df5fb24a3fecfeULL,0x8986094009262800ULL,0xb000252c120fc090ULL,0xdcb90aa002368384ULL,0xaec4d560b3dce56ULL,0xb5badde71b7ee7c4ULL,0xdd9ce7744a7c22c4ULL,0xc6b0055e5a1ca7eaULL,0x383e14a1a05e4c0ULL,0x286faa030a3c29c4ULL,0xb4b2f6c0ca3caad4ULL,0xd17dfe6e4e3de7d0ULL,0xffde7fb2df3defffULL,0xc9c3f3ae4f1eede7ULL,0xefffffdfdf7fefffULL,0xffbfffffdf3eefffULL,0xbd7c76ae0f38cadfULL,0xdffc7f738c3ccbd7ULL,0xf386fb7ecb3fcfe6ULL,0xcecfbefecf7dcf89ULL,0x3e3ccd584d34ce80ULL,0x91de14fc4a1eeafcULL,0x9edeebb45a3c24dcULL,0xfbbeffff5b7caff7ULL,0xeb7ffe3fc634cf97ULL,0x8f7fde5e5e3ee9b7ULL,0x77feffed0b7e8fd7ULL,0x8dced74a4a7cce95ULL,0x2bb06f26de7c4c08ULL,0xc7d2f53a4b1eacd5ULL,0xee06d5470b5defaaULL,0x35ba4fe30e3ceddeULL,0xf76c876a033c6cb4ULL,0xff8805580a34ce81ULL,0x82f08330413ec8b2ULL,0x69da9ed15e3fc2b0ULL,0x15fa07f3ca3dedb5ULL,0xfffe96fbdf3c4eacULL,0x42b63d034b30ca82ULL,0x8120f358ca354ac0ULL,0x48001e48810c8c82ULL,0xb3b24f52071ece14ULL,0xc417e47042362b84ULL,0x9abe04025a3d4e94ULL,0xcb34ef088a36cac2ULL,0x843b2300b3c6840ULL,0x3fbcff99cf3deafdULL,0xfcf47eed1f2cede1ULL,0xbffeff739f3fefefULL,0xf716637b5f3deffcULL,0x1db5c0b011fca95ULL,0xbbdfdfdfdb7eeef5ULL,0xb7f9d7e4df3fcf9fULL,0xf9ffaf7c4f1defefULL,0x119253358838aadaULL,0xcdc59410482ca093ULL,0x245b13c0032cc1d6ULL,0x77475fab8b3cea92ULL,0xc102a04a411e4880ULL,0x2ca605f84b144fa8ULL,0xc976f729cb36aac6ULL,0x2b4621c84a31c1a2ULL,0x1a8241ac4b346a00ULL,0xd1b137401b2ce0d3ULL,0xef842c500508c8d2ULL,0x90481040c214000ULL,0xff21940083481c5ULL,0x110909309306280ULL,0x60265067081e82b2ULL,0x40200000040100ULL,0x0,0x8100080000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc0000004800ULL,0x804040000008012ULL,0x100800080010ULL,0x0,0x210000000040400ULL,0x0,0x0,0x0,0x0,0x0,0x20400080004ULL,0x0,0x4904000001800c0ULL,0x9000000000080080ULL,0x57982e000a1cea89ULL,0x800010000040100ULL,0x11000000c0000ULL,0x0,0x0,0x18200008002040ULL,0x4000200012000000ULL,0x850000102080210ULL,0x24000080004ULL,0x100100100020ULL,0x100000200080020ULL,0x0,0x0,0x0,0x204004004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x30405248240200ULL,0x88bffe3a4e3eea98ULL,0x6f5f37fb1f3def9cULL,0xbcbe5fefcb3fced6ULL,0x5df6f77d5f3eeeb5ULL,0x2c9e5df1071ceff4ULL,0xcbd5e72a8f3e69c4ULL,0xe0c7da5b1b3d6c80ULL,0x2bbe5f5cdf3eebbeULL,0xb5feffef4f3feffeULL,0xdfbdef75cf3feed8ULL,0xefb0376f4b7fcfd7ULL,0xdecbff9f4f3febf7ULL,0xebffcfdfdf7fecffULL,0xfefffeffdf7fefffULL,0xffffaf7fdf3fefffULL,0xfebebe7e5f3ceedaULL,0xdcd73faac41ee854ULL,0x2d4cdbd64d3eecb0ULL,0x4000030180280804ULL,0x24c008280040ULL,0x0,0x0,0x50000008002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2020080000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41014240108082ULL,0x0,0x0,0x400400004004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4e45021412000e0ULL,0xfbed7fe151bacfffULL,0xffffbfffd9bacfffULL,0xfffffffffbbadfffULL,0xfffffffffbbadfffULL,0xfffffffffbbadfffULL,0xfffffffffbbadfffULL,0xfffefffffbbadfffULL,0xebfffffffbbadfffULL,0xfffffffffb8adffbULL,0x9f9bcd7bfbba5fffULL,0xf9fff7f3ab3adffbULL,0xfefefdfbdbba8ffbULL,0xfbffffbefbbadffbULL,0xfffffffffbbadfffULL,0xa7bedff7c9ba5ffbULL,0x6d14136bc93805baULL,0xcdbeefe15b9acf37ULL,0xffa6efffdbbadf7bULL,0xc96ab639088a07f7ULL,0xab8e5b58d30885fbULL,0xf0f7ffe7798a9ff7ULL,0xe9faef69199acf7fULL,0xc2a27fa691ba87b2ULL,0xce47ef29eb8acdfbULL,0x2bdfffdfd9ba5ffbULL,0xdfff5ff3fbba4ffbULL,0x592caf7871029bfbULL,0xfab7bf7ffb9a1f7bULL,0xcbd2e764499acf73ULL,0x7edf3e52db0acfb3ULL,0xb7bf77f253b29ffbULL,0xf9efb4f17aba9ffbULL,0xfb967d77f99acf73ULL,0x2800e33a28180712ULL,0x7fdf17e7eb9a8f7bULL,0xfff9cffb9bbadffbULL,0xb7bc7e6e3bba17fbULL,0x5ede7d61e99adffbULL,0xcbf3ff7ff3bacf7bULL,0xfe5b7edb4baadf7bULL,0xe71ecff7ca8adfbbULL,0xbc77feead99acffbULL,0xffbefd7319b2dffbULL,0xff8dffffebbacd7bULL,0xefdffe3f78ba8dfbULL,0xaf5fdfdfbb9a8fffULL,0xf5ffdeebcbbad7ffULL,0xd9eff7fbebbacffbULL,0xcabebf57410a4f97ULL,0x4d57ffbec93acffbULL,0xefffffdbfbbadffbULL,0xbebfffeff91acfffULL,0xe97ef769012a5ff2ULL,0xfa3c2f3ed28a8d7bULL,0xeffffbff5bb2dfffULL,0xade969dd32a075fULL,0xffb3df7a61aa5f5bULL,0xf4efbf5bf93a9fbaULL,0xfffefffff9bacff7ULL,0xabddf77e59bacffbULL,0xeeef7beb739adff3ULL,0x5fbfdf70f99897fbULL,0xf577e7ffdb2adff7ULL,0xfff67ffbebbacffbULL,0x7bd4c72a0b8a4fbbULL,0x2cd3329bd08a0d29ULL,0x5bb8453c288171aULL,0x7df6b778d91acfbbULL,0xddbaff71393a5df2ULL,0x7bbbff2e593adfffULL,0x3fcef61f89ba0f2bULL,0xa4c15db288a4d02ULL,0xd4ecd6a143aa05faULL,0xdd96206662020b10ULL,0x1c30a345a08007a1ULL,0x2858500011821110ULL,0xab1e4ed1f2924fb1ULL,0x3ff33f6b4918d7fbULL,0xf9269fded11a8d73ULL,0xced235d8808a1d93ULL,0x20c4f3f4020a89baULL,0x4bcb1a5903a8c9faULL,0x849bde93413a85ebULL,0xdbfe97094aaa9ffbULL,0x7f8a7de2cb9217f2ULL,0xebc7fee7f9badf7fULL,0xfffffffffbbadfffULL,0xfffeeffffbbadffbULL,0xdfb7feed1aba8fffULL,0xcb4ebfea839a9ff9ULL,0xdbf3e71a619a8ff3ULL,0xe6fbffebcbbacffbULL,0x3b9e4f5a993a5f7dULL,0xa5fe16ec90aa857bULL,0xdfee7d71fbbadfffULL,0xfbf6f26ffb3a0fffULL,0xcfcbd4b1c9084d1bULL,0x8f6f590023129df1ULL,0xf7bbffff799adf7fULL,0xbb6fb77bd13a4f7bULL,0xdd367f5e131acddbULL,0x88d1baa0113804faULL,0xb75ccfdba01a87efULL,0xdcde76efdb9a4fffULL,0xf5bcaf64f30a5dfbULL,0x8ba6266ca008dd3aULL,0xa167c008c1184f23ULL,0x29a9a4f6912054aULL,0x37b54feb6bba5ff2ULL,0x9cf79668309a4f69ULL,0xcf92f6306918057aULL,0xcb95ebd37b2a8d7fULL,0x40c2081020001110ULL,0x838a0f9a1b28577aULL,0x15ef94e0909a0522ULL,0xcdbefd62d89a4ff2ULL,0xca10430e802805eaULL,0xa043ba98819a1453ULL,0xaf9ecfe3399acffbULL,0xfcf7ffe9ebb897fbULL,0xfffeff727b9adfffULL,0xddbce3e7689a9fdbULL,0x8a579a0a41088712ULL,0xabbcdfdbd12addfbULL,0xf4c97ee5d11a9f73ULL,0xb15b95710b08957bULL,0xcb1e0ca8290a81b2ULL,0x816ac0088908050aULL,0x2e939fc600928fbbULL,0xb6160460019a0d42ULL,0xd600d171408a052aULL,0xca082530c11a052aULL,0x41c68c0401089513ULL,0x24644c4008000f63ULL,0x878b6772091a8fafULL,0xe77cd7f903328fbbULL,0xfffe7ff3f99a9fffULL,0x6956bb8e001a0f1bULL,0x4e5938d399880dbbULL,0x363fd5e2a03a8de9ULL,0xfefcde7f7918cffbULL,0x0,0x8000010001000002ULL,0x0,0x1000010000080020ULL,0x0,0x0,0x0,0x0,0x0,0x100004000000000aULL,0x0,0x4002000000408ULL,0x0,0x0,0x0,0x0,0x8000010002000010ULL,0x0,0x108000000000402ULL,0x400000000440ULL,0x0,0x8000200000100400ULL,0x802000000210ULL,0x0,0x3000084000820800ULL,0x40100000080100ULL,0x28b4b5028b02ddf8ULL,0x0,0x0,0x0,0x0,0x800440201200040ULL,0xcb022a6240124f13ULL,0x488208000108403ULL,0x0,0x2004000000020800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000402000010ULL,0x0,0x0,0x0,0xc1804b56901a0ca6ULL,0x8090f843cb228d30ULL,0xff5e9ffecb8acffbULL,0x2bbacfdac1ba5eb3ULL,0xfddfb671eb9a8f7bULL,0x47be2de051ba5ffbULL,0xcbf4dfaad9bacfebULL,0xc94b9eb9411a4fbfULL,0xf7bcdfffc93acfffULL,0xfdf7fffafb9adf7bULL,0xcfbbfff3fba8dffbULL,0xb8ccefff7a8a1ff7ULL,0xc77ffb33e9921f7bULL,0xbbbfdb5e3a98cf7bULL,0xfcfffffffbbadfffULL,0xfffffffffbbadffbULL,0x8a34af2a910a1f5bULL,0x800020000000402ULL,0xe6c10009200160ULL,0x8002000080000400ULL,0x800000200000100ULL,0x0,0x0,0x0,0x0,0x108802000104400ULL,0x680c353041284c20ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x210000000004400ULL,0x0,0x40211000100000aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbbeefddb41336eeULL,0xfffff7efbcfbffe6ULL,0xfffffffabdffffeeULL,0xffbfbffebdfeffe6ULL,0xefffffff3dbfffeeULL,0xfffffffbbdffffeeULL,0xffbfffffbdffffeeULL,0xffffffffbdffffeeULL,0xfffeffffbdffffeeULL,0xebffffffbdffffeeULL,0xffffffffb5fffeeeULL,0xffbefff33cff6eeeULL,0xd7f6ff7d34bbfee6ULL,0xf4aeff6334ff7eeeULL,0xcb56e3bebcfa7feeULL,0xcf7b3efbb5be5feeULL,0x8fbedffbbcbff7eeULL,0xd1f7f6ea0dbaffe6ULL,0xdffffff33dbfefe6ULL,0xfb9ff3af35ffffe6ULL,0xcfdffe3734fb6feeULL,0xab5f7bdbb4bb6dc6ULL,0xb4f5defa9cfbffe0ULL,0xfba77f6334d25ee6ULL,0xfabe3ff7a4fbffe6ULL,0xffffff7eb4b7ffeeULL,0xefffdfdf3cbf7beeULL,0xbfffdffb3cbb7feeULL,0xb97eefe19c9b776aULL,0xf6962bfa1cbfede6ULL,0xfcd3dfac94fbfae6ULL,0xaffe9fd73cbfffe2ULL,0xb712fffab4d34fe6ULL,0xf5feff799cfaffecULL,0xa2da7d6eb0bee6c6ULL,0x8ae43fee303a6ae6ULL,0xfefefadb35faf7e6ULL,0x3f91eff01dbbdae4ULL,0x96f2f7ee1cba7fc6ULL,0x73be7f631cbefee6ULL,0xfbdceb3a34feffe6ULL,0xaed3fbdb349b5de6ULL,0xa79ddffb35fb7ee6ULL,0xd7dfdfff34fbfaeeULL,0xfffeedf33cbbfec6ULL,0xfeb9f77f34bb7ee6ULL,0xefffffffbdbfffeeULL,0xbbdffddbb4ffffe6ULL,0xf7f6fefb3dbffeeeULL,0xf7ffaffa3dffffeeULL,0xffb77f7eb4bbfbeaULL,0xfddfffbd3dffffe6ULL,0xefffffdfbdffffe6ULL,0xffb7ffe735bcfeeeULL,0xff7def7fb5f77beeULL,0xf69d2f461cffe8eaULL,0x4752df913cb672eaULL,0x23cbbafb341375e6ULL,0xbffa7ffbbdffffe2ULL,0xf1ff7ffb15bb78eeULL,0xfffefffc3dfeffe6ULL,0xf9fef30d94bbeeeeULL,0xeed73efdb4bde7e6ULL,0xffbffdffb5bbfee6ULL,0xd5fef6f334dbfeecULL,0x5bae7f6314be7ee6ULL,0xf3d4e73e34d854e2ULL,0x8fe21e10349a03c0ULL,0xfeffef533c9ffeeeULL,0xfd77fffdbdfb7fe6ULL,0xeb71ff7315b8f7e2ULL,0xfa95ff2b3cbaffe4ULL,0xaf5ff33e1dfb7fe0ULL,0xa392dd5b3c987ce2ULL,0xb6ddd6ebbcbbdfc6ULL,0x4922af420c1250e6ULL,0xebd07e360cd3fa62ULL,0xc8db7330bcbc7ee4ULL,0x8ef95fdab4b7fbe6ULL,0xfe7ff7e3b5feffe6ULL,0xf7c6c6761cf2ffe4ULL,0xda3a1759adb95b42ULL,0xae87f76e94bb5346ULL,0x7bdede498dfe72eeULL,0x808ace1a849874c0ULL,0xa0301c2a84da6100ULL,0xf7862f6c34f8be84ULL,0x8b7ff3aa3080eec6ULL,0xefffffdebdffffeeULL,0x7fbfefffbdff7fe6ULL,0xd76e77fb35fb7bc2ULL,0xeb3e6d7a34bf7fe6ULL,0x9b26ef2e34fb3ee6ULL,0x4edbbada84f37fe6ULL,0x3aa1cf48148bc86eULL,0xf5f77dfe3cbeefeeULL,0xfffffdf1b5fb7fe6ULL,0xbdad6fff14ba7fe4ULL,0xff9efaabb5f3faecULL,0xc99699511cbd6cc4ULL,0xf4f3766685bfe2e4ULL,0x9fefad7a9cf3eeeeULL,0xbf9d2f6e0cbaf3c6ULL,0xcb53fbba2cf87bc0ULL,0xe1ff9ac49c1a74e6ULL,0x94d67df294bcf6e6ULL,0xb9ef8f6835b3efe4ULL,0x4a963f75053933e6ULL,0xfebfff7e3cbaffc2ULL,0xe99fdc80499f4c0ULL,0x20aacde100bb07c6ULL,0x904d2461042226e0ULL,0x21025f0209038c4ULL,0xa9e4470604927bc6ULL,0x4d14f4a04126ae4ULL,0x539d2e01cb9e2aaULL,0x90c0d25220946cc0ULL,0x92e4d6005b337c4ULL,0x2bd6c71e3cb8ffe0ULL,0xcc62321f3c9d17c0ULL,0x2ebdcc5b1c3876eeULL,0xf5fefffbb4b3fbe6ULL,0x93ae2f4034a7aae2ULL,0xda222fcf3487fa68ULL,0xccf6961f149b6ee6ULL,0xffffffdbbdffffeeULL,0xb77d56e19cbbeeeeULL,0x7d7eff6a3cdfeee6ULL,0xdb12fe7b149a6ec6ULL,0xc8d6f612048aefa4ULL,0x26e134004aa4840ULL,0xe6f677e424937de2ULL,0xae0260140006c0ULL,0x663a2de10cd85020ULL,0x8bd5be04348e25c0ULL,0xa376163024d36ac4ULL,0xeb046e820bb7142ULL,0xc706ac6a14ba7de0ULL,0xa4886c5d10ae2640ULL,0x8082000f0482e060ULL,0x2d4030018c123484ULL,0xf7ff57f23cf9fec2ULL,0xec58a2b104db6860ULL,0x0,0x0,0x0,0x200000280ULL,0x0,0x200010000110000ULL,0x0,0x0,0x0,0x0,0x8010000000101000ULL,0x0,0x0,0x0,0x0,0x0,0x80010000808000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x12000000280ULL,0x0,0x10040000102000ULL,0x0,0x2010251010030262ULL,0xc004652010132560ULL,0x8000020204004020ULL,0x0,0x0,0x2004000110000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8010048000c84000ULL,0x40162d7214d90402ULL,0x182e2621cc8dd40ULL,0xeed33c523c127ee4ULL,0x49bec9eb1cd8fde2ULL,0xf477eec63cbaeee6ULL,0x6b3a2d6110137ee6ULL,0xbb95bf213dbb5ae2ULL,0x8ec3b6381c3377eeULL,0xcbdd5fb0dba39eeULL,0x7575c7ea3ca6fd46ULL,0x1c6cbf3034bb7fc2ULL,0xf980276f3cb86f60ULL,0xc0c3fe2934bb74e6ULL,0xcbdfcf5e3dfb7bc6ULL,0xfeffffffbdfeffe6ULL,0xfdefa7ea3cbb7fe6ULL,0x4000241011280400ULL,0xc0000101000ULL,0x48081134102040ULL,0x8032082000d42280ULL,0x700100000c80ULL,0x8040000004040ULL,0x0,0x0,0x48000000480ULL,0x84684d200c2803e2ULL,0x0,0x0,0x0,0x0,0x8001000010000400ULL,0x0,0x50000000280000ULL,0x0,0x1082010c900000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfde7ffdff0dc1efbULL,0x7fffdfdbfe7cd7feULL,0xbcbf6dea74d47cfeULL,0xfdeffff3fefdfefeULL,0xffbfbffff6fdffffULL,0xfffffffffefdffeeULL,0xffffffdffefdffffULL,0xfebbfffffefdffffULL,0xfffffffffefdffffULL,0xfffefffffefdfffeULL,0xebfffffff4fdffffULL,0xffffbfdffefdffffULL,0xbbbfeffbb6d45ffeULL,0x97f7ffeffe5477ecULL,0x77befd71f4f5f7eeULL,0xfbdeef7cfe9d76eeULL,0xadda76f7f4d4f6eeULL,0xdbbfdffffefcd7ffULL,0xfcb777feb6f154eaULL,0xbbfe75f3dc1dd66bULL,0xfbb4f73ff630d67cULL,0x4a5ff711bcb476e2ULL,0x8e7f57d83e3557eaULL,0xf4f7fffef4fddffeULL,0x9beef761bc74767bULL,0xfff2bf7efc3cff7fULL,0xdfffffbefefdffefULL,0x6fdfdfdefcbc7ee6ULL,0x9e376febbc15d5edULL,0xd9f9f77e5c5157fcULL,0x9bc77f7ff0d01efeULL,0xcff7ffbf36fd7effULL,0xb67ef7cab4515efaULL,0xd7dbf7fe0efddffaULL,0xf4fdbffcf670fffcULL,0xffb62df1b4b0f6eaULL,0xe990d64476b4147aULL,0x2fffbe4bf454dceeULL,0x819c4f33b01c5ee8ULL,0xfdfffffff6fcd6fbULL,0xdefe6db2f615d6fdULL,0x895feb2d7cf4dcfeULL,0xa6debbfbfc7cd6ebULL,0xf7fe5fcbb65ddeffULL,0x75f5f7eff4f4fefcULL,0xbbbeff7276b4fefeULL,0xfbbcfb7efcbcd6ffULL,0xeffffffffe3dfffeULL,0x7f77dfdfb4f5def7ULL,0xf5fffeeffcf4ffedULL,0xe97f7f76f6dddfffULL,0x9b9c7637de3cd4eeULL,0xfff7ff5afcddf6ffULL,0xaffededf5c9dfefeULL,0x95af7ff39cbcfedeULL,0xd1f2bf7ef450d6edULL,0xdebd2f77b65cddfaULL,0x6ffff77fb6fcf6eeULL,0x2e763f5b7cf4d4edULL,0xb21b4fe39255d7eeULL,0xd26337ff14f55cb6ULL,0xf29a2f7fe6b4def4ULL,0xeb8dfb22e63d5f6cULL,0x6f7f3fdffefd5ceeULL,0x9dbedf7bbcdc5e6fULL,0xd1f7b7ff3cbcfeeeULL,0x7b9ebd617474dcfaULL,0x8988c30d843c662bULL,0xffdbbefd363c5fffULL,0x893fcf532c54de7aULL,0xf5beecec8c197e78ULL,0x85e72f708494d4feULL,0xe226f76ea63cdffaULL,0xabc7b68aa0f4dc7aULL,0xab5bdddb28905cabULL,0xe4f7fee024b454f8ULL,0x85cf7e62dcc4f1eaULL,0xb2102b70a05c1d68ULL,0xddd79f1eb4d474eaULL,0xa9fcfecb54b5f673ULL,0xecc277e8b490546dULL,0xf7e8ff756c14d6faULL,0xe93b75ca434f4b8ULL,0x7c6ee7ab4f4d6e7ULL,0xaf7fffc5d6fc5effULL,0x88bb4fdb841c3c9eULL,0x7544543bb0145ee2ULL,0xd6802f4074214568ULL,0xcbf4e5b7d4947c7aULL,0x6ed75efd749dd6feULL,0x97bddffbb6f55ceeULL,0x9c557e0d04547668ULL,0xdbe5e72d69c1ceaULL,0xbbcdef7e96f474c8ULL,0xceaa13bf3c5c15e9ULL,0xbeadef4ad01d54eaULL,0xf6df3ef476fcd6eaULL,0xbffffdf2f6fcfefeULL,0xfaac6d7af474fefeULL,0xdef67e8b7434767eULL,0x6bd8c9d3a434d46eULL,0x11b27ec6e094def2ULL,0xf96f57695654d4eaULL,0xdfbe7f725c94dcbaULL,0x5fdbdbb0b4f47e7fULL,0x6fdfdf5174107d7eULL,0x14df7fff543dfefaULL,0xdb7367f796d57cfcULL,0xeab5ab77765c76fcULL,0xc4d3d04a3615766dULL,0x5c7effce34bc767eULL,0x881c5efaacc4142eULL,0xb025b561b67453e8ULL,0xb3265e70e01056e8ULL,0x794eff2afc9cd6eeULL,0xe84711fe303c42b2ULL,0xc958d61541474acULL,0xd0e27761408414e2ULL,0xca86d80d2151268ULL,0xb995df30d434de76ULL,0x4b5336c9cc5c1ab6ULL,0xafdb4f68e43456eeULL,0xbc73f7eff69457eeULL,0x9ba7ad70fc545826ULL,0x43a442a716155566ULL,0x8848120194041402ULL,0xffffffdff6bdf7eeULL,0xb4fd7ee5f494f6e0ULL,0x696e2f761cb45cffULL,0x96162d60649054a4ULL,0x2e84630a34b854adULL,0x6f77dfdefcdc16feULL,0xa9300e8212981240ULL,0xb03eeff5a8dc56aaULL,0x8ba225c8561c06eeULL,0xefdffbbef614defdULL,0x2877de74b4145f6eULL,0x1420e16000905441ULL,0x910584cc14903c88ULL,0x3aec2e6466a0146eULL,0x8a80010040004420ULL,0x621a02813010c420ULL,0xb71c0cfb2e30d5feULL,0x97d29f58a4c456eeULL,0x404000000480ULL,0x0,0x0,0x0,0x0,0x0,0x4100004000004ULL,0x204000620001028ULL,0x8000240000020ULL,0x0,0x0,0x0,0x0,0x0,0x808440000401c0ULL,0x0,0x0,0x2004000000100200ULL,0x0,0x0,0x0,0x0,0x8100a00000101802ULL,0x0,0x0,0x8010000000001080ULL,0x2010000020000020ULL,0xc00008001000ULL,0x0,0x8090080001028ULL,0x1000000204001000ULL,0x9000000010000480ULL,0x81a008161044140aULL,0x0,0x14020000020ULL,0x0,0x8400000002000040ULL,0x280000000000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x7efe667da0f55ef2ULL,0xc880c37b441d11eaULL,0x69d6fd9e961c1ccaULL,0x7eb86ff226dd5cffULL,0x551eb6f896541eefULL,0x77b65d526634f7eeULL,0xc21ce330b41416ebULL,0xae038e10e01015f8ULL,0xe8c5f3a3494143aULL,0x25b6d5f8a64cf4eaULL,0xc3280d00281c502aULL,0x4020462216104440ULL,0x8d6af0ea763474faULL,0xa7aa5fdaba14defcULL,0xfcfffefffedddeffULL,0x89fd37723415d6efULL,0x560c10dc10151ca8ULL,0x88436e040634d2d6ULL,0x8861a0100045488ULL,0x2c7e4cec001cd6c6ULL,0x1088a0002000c020ULL,0x8c212014001402ULL,0x0,0x10020000040002ULL,0x8010402000181400ULL,0x20004000104000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80420000004040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffa6ffffd9cfa77aULL,0xffffff3fd1efb3faULL,0xffffffddf9efb3faULL,0xe7ff7ffff9edb7faULL,0xfdefffebf9edb3faULL,0xfebf3ff3d9cfb7faULL,0xffcfeb7ff1edb3daULL,0xfdffbfdbd1e4b3daULL,0xfffffffff9efb7faULL,0xfffffffff9efb7faULL,0xfffefffff9efb7faULL,0xebfffffff9edb7faULL,0xffdfffdbd9edb7f2ULL,0xbffeeffff9efb7faULL,0xddfffffff1ecb7faULL,0x6ffe7f71f1cfb75aULL,0xfbfeff7ed9efb7daULL,0x8dc3bb9bf1efb3d8ULL,0xaffefffbd9edb3eaULL,0xd4ff77fef1e493faULL,0xffbeff71f94fb3faULL,0xfab7f75ed1edb7faULL,0xedcfbf1fd9eda3caULL,0x6b9fdfdfe9edb77aULL,0xdcfffefff9cda3daULL,0xd9deef7af9cda2daULL,0xfbbfff7fd9efb7faULL,0xd7d3fe2bf1e1a3e2ULL,0x1d9c9d6f9eca3c8ULL,0x34ffdfe9f9efb3eaULL,0xe96eff7bd86c23daULL,0xfff67f7fd1edb7daULL,0x6ba3f7bbd9eda6faULL,0xfbf6cfebf9ce23faULL,0x96b07b68d0c023e8ULL,0xfdfff77bd1eca37aULL,0xd7fe3d64f0edb7c8ULL,0xd9cc7736e8ed23e2ULL,0xeeb6974b51c5a35aULL,0xa02cdcf158652368ULL,0xd577d6fdf0eeb350ULL,0xffbe6de299ed975aULL,0xfbd7ef36f9e5b3daULL,0xe253bbdfb1cda7c8ULL,0x6fbecffbf1edb3faULL,0x77f7ffeff8e7b7d0ULL,0xd7bffff3f1edb3faULL,0xcbb9f37ef0ed33faULL,0x4fdffffbf9edb3faULL,0xefdedfdbf1efb3eaULL,0xf4f7f7eff9edb7faULL,0x59ceff73f1efb3faULL,0xdbb8ff7df96db3faULL,0x8dfe7f1ec9c5b2daULL,0x2ff6dfdbf1e5a2eaULL,0x3cb37fefd9edb342ULL,0x35edae48e1658360ULL,0xf59dbbf0d1eca6f8ULL,0xc9ffff22d9edb3daULL,0x8ed6dfdfd9e5b7daULL,0xbf774de8796db2caULL,0xd4ffbedad16923d2ULL,0xd5cc7d7170ec26d0ULL,0x8bdcf6fed1c5a3faULL,0xffe796dbf1e9a3c8ULL,0xb9bfcfeff1efb65aULL,0xfa74b6fa91c9a3f0ULL,0x4d9c6fe2d94e32daULL,0xbbe4bf3fd1c1a65aULL,0x9e7f7e53f9c5a3daULL,0x33384fd1d1403688ULL,0xe15f66e8f16d27d2ULL,0x33fefd70d9ed93d0ULL,0xffdfef6ed9e7b67aULL,0x88fefe03c1cd237aULL,0x229e5fdbd1cd365aULL,0xfcf72474d9cca548ULL,0x984a7540c1e4a2c0ULL,0xefbb7571d1efb372ULL,0x4dcfb2afd0e9b6c2ULL,0xef5e4e9ed1c5b16aULL,0xf6b37bfa9165a34aULL,0xf12eba6ad1c687eaULL,0xf696256cd1cd2358ULL,0x8188f127602ca252ULL,0x9d6d6456184a2faULL,0xcfe24deac1cd32daULL,0x81092622104091c0ULL,0xf28c0b5041a483c2ULL,0xa9f0fae2d9edb3c2ULL,0x5fee5fdff9ecb3faULL,0xb7f94febf1c6b7faULL,0xcf502671d1c01678ULL,0x7bae7d7bf9eda34aULL,0xfbfffb3ff9efa7faULL,0xccfbffdaf9efa6faULL,0xbbbeefdff9efb2d2ULL,0xf777f248f1c5b352ULL,0xbf7afc42f965b3d2ULL,0xdfbd7f7ff96db3daULL,0xc2c3daebf1efb35aULL,0xae7fdfded9cfb77aULL,0x14b77ee1f14fa3f2ULL,0x8c40ac607867b640ULL,0xbbb67fb6f9eeb6caULL,0xcf57db0bf147a7daULL,0x8f5cc9c3d96da3d2ULL,0x8cb3f7ead96da2faULL,0x9ae5ff40c8cfb3d2ULL,0xdf9f176451c5a6d2ULL,0x8ff8f33cf1c4a3f2ULL,0x69fa6ee469c722f2ULL,0x829287bbd1c5234aULL,0xd2adbe78594923c2ULL,0x8a06302671442358ULL,0x8a83b2a25948834aULL,0x62889672d129a248ULL,0x989194d0d0013040ULL,0x9474ee09f96623caULL,0x100020080800000ULL,0x2944c13cd10ca2c0ULL,0x8950935a802d224aULL,0xcf12cf70d0e5b3c8ULL,0x7db6fefdf1af23e8ULL,0x89149f30794d13c2ULL,0xd914b33bc84da2c8ULL,0x5fc33e18f8e523d0ULL,0x2fdc8b9bd160b742ULL,0x38d97169d960b2e2ULL,0x1fb6fffaf1ecb3faULL,0xf014b31c98c1a240ULL,0x2edef0aad12432c2ULL,0x48581f42c1a91348ULL,0x2c1146e051cc129aULL,0x813c002091002148ULL,0x56a28f06d140a3c2ULL,0xafdfffaef1efb7faULL,0x2b480940d940a242ULL,0x262ace62f1c41348ULL,0x154425c8d12c1002ULL,0x8a8c25c9c14b2362ULL,0xab4cd10191e50348ULL,0x79731283f1a58242ULL,0x15fa8f8e61eda2daULL,0xc4643ece51c4b362ULL,0x8040000000140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000400000400200ULL,0x300000200040ULL,0x0,0x0,0x80004000080200ULL,0x0,0x0,0x0,0x0,0x0,0x200040040000ULL,0x0,0x0,0x0,0x4000200010000080ULL,0x100800000040008ULL,0x0,0x0,0x41000040000080ULL,0x0,0x8900e820510481c0ULL,0x80000200000140ULL,0x0,0x4000100050000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000080000040ULL,0x0,0x4000400040008ULL,0x0,0x0,0xbba63de3f949a2c8ULL,0xc9aff13bf1cfb3caULL,0xebcf8ff9d1e5b2d8ULL,0x8dbadfdaf1eda35aULL,0x480526b5940a3c8ULL,0x6bda7d70f1cdb2caULL,0xca2acb0ad9463352ULL,0x8ac2db1ed96583c0ULL,0x69ae4c6041618368ULL,0x84c616a5c125b248ULL,0x93820740884ca740ULL,0xc99be1bbf94ca2d0ULL,0xc9dfb39ff9edb34aULL,0xab7fffdbd9efa7faULL,0xded7fefef9edb3faULL,0x11342101d041a000ULL,0xd2ae2f6ed9c9b37aULL,0x21050000000ULL,0x28091880114c0202ULL,0x4b857ced1052148ULL,0x8040a710186003c0ULL,0x40802840500022c0ULL,0x0,0x108000040000080ULL,0x160b5ec0d16903c2ULL,0x510080082200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000000240ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80800000082000ULL,0x0,0x120010080082000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x37ece7619aacef24ULL,0xffb7fffbdbbcfff6ULL,0xffffff3fd9befdf6ULL,0x6a7bb7cfdab8bdb4ULL,0xa0041c0002800ULL,0x0,0x8880620c0002004ULL,0x800020008000002ULL,0xefffffdfdbbeff66ULL,0xffffffffdbbefff6ULL,0xffffffffdbbefff6ULL,0xfffe7fffdbbeff76ULL,0xfbffffffdabefff6ULL,0xeefffffbdbb8ffb6ULL,0xffbfdfffdbbcfff2ULL,0xf9eff7f3d9bcef76ULL,0x7ffe3ff39bbefff6ULL,0xbbfdff7e9abef5b2ULL,0xeeebbbc89aa8eda6ULL,0xcfbfcffbdabced36ULL,0x7582fef4dab86d24ULL,0xfd7fbfa09abefd34ULL,0xdc84ef74d828e570ULL,0xcdd3feb1d8b4fde4ULL,0xabdfce4bdabcedf4ULL,0xd4bdffeddabeedb6ULL,0xf9f7ff73db2ccf62ULL,0xffbeffffdbb8ff26ULL,0x58d7db1adabebf60ULL,0xabcff3db9a38ed64ULL,0x4fbecf7258bebf74ULL,0xd1fdaf52cab86f64ULL,0xdbbe7f72dbaaff74ULL,0xffffffff9bbeffe4ULL,0xffffffc9dbb8bfd6ULL,0xb4a7ffef9ab8fd66ULL,0xfdf5bfeedbbeff24ULL,0xefbeef72dbbcf572ULL,0xeb84f23adab8e564ULL,0xe9d6ffd2db8afd24ULL,0x4bbecffadbacfd32ULL,0x74fcf2f0da38ed22ULL,0xff7e2df2dabcedf0ULL,0xfbffff3bdabefd62ULL,0x8eda3a9e1aace5e6ULL,0xbfffdffbdabceff6ULL,0xf5ff76fbdbb8e5d4ULL,0xffbeff30d8beed76ULL,0xdebeeff6dbbcef54ULL,0xfffffebfdbbcff76ULL,0xefffdfdfdbbcfff6ULL,0xf7fffffbdabcfd76ULL,0xcbeeff62dbb8ff76ULL,0xdb96bf069ab87710ULL,0xc7c2ee3bc220ef44ULL,0xe6cd4f59db2cefd6ULL,0xb4cf57ed9888efe2ULL,0xddeebf49983cf574ULL,0xfeb6bf6e9aacfff6ULL,0xedfef73ad2b86f74ULL,0xebefffcfdbbcfd66ULL,0xbb337ffada884f36ULL,0xf9edbf748ab8ed60ULL,0xff4efe2ed8bc2770ULL,0xa3b6f7afdabce7e0ULL,0x7effffdbdabeed60ULL,0x3fb7cffa9ab8ff14ULL,0xb475fe7a9b88e596ULL,0xd7be7df1dbbcef32ULL,0xcbd5efe3d9bced36ULL,0xf3c0165bd8b2a502ULL,0xc6bf4fe85bb8ed54ULL,0x17ebef6098a4fd22ULL,0xffdefd709bbcffb2ULL,0xebf3e757d8baed66ULL,0x8ddffea7d83cef72ULL,0x1b58cfdad8b47d42ULL,0x94b77eeddab83562ULL,0x8ce5e75399a89db2ULL,0xf69d2e535b88f700ULL,0xe8f1d31099a8cd30ULL,0x4975525110b8e510ULL,0xb5fb3fe39aaced44ULL,0xd51991505aa8be20ULL,0xf6b225609ba8ce90ULL,0xef80f6a399386c24ULL,0x8deb613912cad40ULL,0xbf3b7f60d9b8bf74ULL,0x325d967ad82c6420ULL,0x2bba7d609aa8ef20ULL,0xab9ffff212bcef54ULL,0x7ed39ed298a0ef92ULL,0x9a9c67e7d832cdf0ULL,0xf4a7546dd8be6550ULL,0xffbe7df9dabeeff4ULL,0xebd6f30edabcb5f4ULL,0xacefbaccd8bee574ULL,0xa9becde39aa8ffa2ULL,0xf9d777feda3af566ULL,0xfaefec72dab8fd64ULL,0xca8c636312a0e564ULL,0xde5232a0d238c500ULL,0xffdadddfda3ced96ULL,0x35ff56e4d828cdb0ULL,0xdfefff7b9ab6edf0ULL,0xdfa17c77dab8dda2ULL,0xefffffb7dbbcfff6ULL,0x2ed7dfca9a38ff70ULL,0xb4fbf6e7dabce7e6ULL,0xd1fcffcbc8acefc4ULL,0xfa9e3fee9ab86506ULL,0xedfbffbedabcffe6ULL,0x6eff3ecf9834ef66ULL,0x14bbf7e0dbb84766ULL,0x927bd7314ab86526ULL,0xf7f27f7f9bbcffc6ULL,0x8152e10a802c4c20ULL,0x4dfc7e0eca282f46ULL,0xfebd4fa898bafd14ULL,0x1700f115d2886534ULL,0x4dba2ff09ab8ae24ULL,0xda9ee76f92b4e566ULL,0x8e4abb5e92bc2f22ULL,0xfbb2cff35bbaef22ULL,0xdcffbef6dbbcfd74ULL,0xfff8fdb2d8287526ULL,0xfab4ff7f1a2efd44ULL,0xaa47da0cc928f506ULL,0x216c759da20a592ULL,0x77e676e2c3ac6d62ULL,0xdf66a77299a8edf0ULL,0x5e8603cc90aa2502ULL,0xa00fe0ce92bc8d00ULL,0x4806000942002504ULL,0xba84a100280760ULL,0x814b85700024f560ULL,0xba1a07709222c722ULL,0x403dd72890362402ULL,0x673ef450103a6544ULL,0x266b47e09020fd62ULL,0xd576f4dedab86746ULL,0x9a8e496c90b02d14ULL,0xcb27a21e932a2772ULL,0x86f02a53c0286f26ULL,0x88becf9f98a0d570ULL,0xb673f6d4d0bafd60ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x880000c4000a100ULL,0x0,0x0,0x404000000008100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8104102408206100ULL,0x950310680284c00ULL,0x0,0x74d49e4803a82724ULL,0x0,0x400402020000d100ULL,0x142200000248800ULL,0x480048000000ULL,0x4554d62192ac2502ULL,0x4c388d230838ed62ULL,0x2988401e11888580ULL,0x81e392101a247500ULL,0xa409ddca9188ed22ULL,0x161f72e113b84760ULL,0x9a48e7721820e552ULL,0x13a4676249280720ULL,0xf9cbd3bb08acc560ULL,0x3a46f81b18884702ULL,0xb5ba55f3903c6520ULL,0xd50087080a244504ULL,0x72b6af7408ac6726ULL,0xd180f79319b86542ULL,0x800100240000ULL,0x1c10ee6808a40540ULL,0x1024206100882700ULL,0xf6ac696cd8b8ad06ULL,0xc1bbfe719ab82524ULL,0x46e79670d2909d10ULL,0x9bbac7da9020bd32ULL,0x708c16c24a282724ULL,0x638e7d7298ba6522ULL,0xdb94cb22daa88f30ULL,0xcc82be2999b885a0ULL,0x62bec7c3daa06536ULL,0x6665547110acad64ULL,0x9d8e2c0399384500ULL,0x80a12848006402ULL,0xffeffabfdabaefb6ULL,0xafdedfdad8aeefb6ULL,0xdcf7feedd0aeff32ULL,0xa76e277adbacc722ULL,0x9aa1cb3edaaae500ULL,0x0,0x284311011020b400ULL,0x46410841aa06400ULL,0x212058200a204400ULL,0x0,0x6000000000002004ULL,0x0,0x80400000008800ULL,0x88302990240120ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x102054040202000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x108000004001ULL,0x0,0x0,0x8000000001020000ULL,0x0,0x0,0x0,0x0,0x0,0xf0e4d44621064c62ULL,0xf9c63773bd746d63ULL,0xa3b27f6c7ef6ff62ULL,0xefdfffbedfdeff73ULL,0x0,0x10001000000140ULL,0xa00000100100ULL,0x800020000800200ULL,0xefdffffffffeff73ULL,0xfffffffffffeff73ULL,0xbfffffffffdeff73ULL,0xffffff7bfffeff73ULL,0xfffefffffffeff73ULL,0xebf7fffffffeff63ULL,0xefffffdfdffe7f73ULL,0xafffdfdf57d6ff73ULL,0xd7fff6ff97f67f61ULL,0xfffe7ff35fdeff63ULL,0xfbf7fb3f5fdeff73ULL,0xcffffb7bfffeff72ULL,0x6fdf5ffeddfef771ULL,0x9dfff6edd7deff60ULL,0x8b9aef715f7e6562ULL,0xf9af7f7f8d4efc63ULL,0x4bdffe9b4fee6f62ULL,0xab3fcfd11bce7572ULL,0xb6f7dfff7dfef570ULL,0xfbc7f7fb5dde7f63ULL,0xfb3efffbbdfe7d63ULL,0xcd47ff3e1fe67e73ULL,0xaffddedfcffe7f73ULL,0x9aff5fcb9bbe7562ULL,0x99ed8fe84d967f63ULL,0xfb9e3ff10ddeef61ULL,0xc3dff3beffca7571ULL,0x2c5efcc84d86fd61ULL,0xbffffdfddddeff73ULL,0xf5f6beef77e6f771ULL,0xf7b67ff9dfd67d63ULL,0x69cdf4380dceff70ULL,0x5c133e8400ce6862ULL,0x9f9b455b5fdef562ULL,0xf5f777fa5bd66f72ULL,0x6f7c6fdb3dc6eb63ULL,0xdbc5cf3eafae7763ULL,0xce5bbbffddf6ff60ULL,0xffbffffffffeff73ULL,0x95fd5ee9bff67d71ULL,0xdb38af731df25760ULL,0xf99eefb775de7f63ULL,0xfffffffffffeff73ULL,0xefffffdffffeff73ULL,0xfefff7e5fdd67563ULL,0xcd6f3ff79dd6ff62ULL,0xdbb43a3b25d67f62ULL,0xce07eef345da6770ULL,0xcfceddb0d467530ULL,0xfcf7757acfeef763ULL,0x97fa2fdd8bc67d63ULL,0xdf863fff95967763ULL,0xdae7fdca5cc6ff61ULL,0xab9efedbddd6f362ULL,0xae786ffeddc67f61ULL,0x96ecb57b11967d61ULL,0xfb7e6f60cdd66561ULL,0xcbf7b3ae9df67561ULL,0x2f5f3ed3d9e67d61ULL,0x3fbfedfc4fee7f70ULL,0x76fff67d7fdeff63ULL,0xfffefffbdffeff63ULL,0xfbdaff7f9fce7f62ULL,0xc243f75b91963f60ULL,0x4bbe4f7b4fe66761ULL,0xf777d6ea3fc67d60ULL,0x9a3ced734d627c70ULL,0xdeb2e7ff7dc67773ULL,0xe5d6bfbb49d67f71ULL,0x8e7fdf5bdfe6af60ULL,0xf43edf3cd5c26d63ULL,0xdbedfe7c59da7f73ULL,0x4f143c6e81963960ULL,0xcdd3f6201d4e4f60ULL,0xc37c455dc31e7561ULL,0x57f777efddde7773ULL,0xb5d4ae60ddc66742ULL,0x44b20fdf49d67140ULL,0xcedbfe365e386561ULL,0x2887ee15a52c7d61ULL,0x343bffefd1b67f63ULL,0x570cb6710da4e943ULL,0xcffe75fcdd967f63ULL,0xebe7fbfe77fe7d73ULL,0xfcd97ec63dd67162ULL,0x1ffa2d7b1dc65f63ULL,0xdff7b44b0d365f72ULL,0x7eac2d43dad66772ULL,0xb3d4f216076e2560ULL,0x9fcfba2e8d5e3561ULL,0x8cbe4c0a1d967d62ULL,0xfc5fb76a97f2f571ULL,0xebfead771dd2ff60ULL,0xcea4017ea5326560ULL,0x8900928005820340ULL,0xa61ec41a05527760ULL,0xd6bed6f55dde6571ULL,0x81006b6218c27441ULL,0xce8c3f762bc05d71ULL,0x8e5ffebe4f46fd63ULL,0xbf5e7fd39792f553ULL,0x879e77ebc7b66761ULL,0xd96cf7f237f65f63ULL,0xfaec2f7647b6ff63ULL,0xeffe756747dc6f61ULL,0x35c7ff5f45962f70ULL,0x2caa4f2829c63e20ULL,0x188100040802200ULL,0xb2be7df091d2ef41ULL,0xe9fdf6cf6fc6ff63ULL,0xe61614860d864420ULL,0x9a2a479a89563740ULL,0xdcb0f375a9cca570ULL,0x4d467df0238e4e60ULL,0xf254e02209c27140ULL,0xe8bb21f13861570ULL,0xbdff6ff0ddd67d52ULL,0xf57ffefedfdeef72ULL,0x9e9eff720ffe7f71ULL,0xb3806f4d2c44fd70ULL,0xe5c6f68a4dd42f70ULL,0x19bd9d387d2f762ULL,0x94b226ed0fc27f73ULL,0x9d73f7611ed65f73ULL,0xd09e2d443dce4760ULL,0x8067222803065c61ULL,0x6a74174c91a6a660ULL,0x10715d4009c24e60ULL,0xd979f7741df65f60ULL,0x10842e3605821740ULL,0xf44960407024561ULL,0xbf5cda4b09865660ULL,0x87b842f39cd66b60ULL,0xdce65d0210927060ULL,0x6666256011e64f60ULL,0x8101a32208c24a60ULL,0x484a337121d27f60ULL,0x8c4f2100c06b62ULL,0xd434546000824a62ULL,0x0,0x2204400000802200ULL,0x50000001000800ULL,0x120000008400000ULL,0x0,0x0,0x0,0x0,0x0,0x20004000100100ULL,0x22044000404040ULL,0x0,0x80800000000140ULL,0x0,0x0,0x0,0x0,0x0,0x210000002000400ULL,0x0,0x0,0x8222256005024740ULL,0x8840810008400140ULL,0x848004001820100ULL,0x90174d6005920340ULL,0xfdf6d6f894963740ULL,0x344342008164140ULL,0xc9d593008a462c20ULL,0x2c6bbb4010807540ULL,0xa58756b824882761ULL,0x315c5fe2d5c06763ULL,0x94362570d59a6162ULL,0xdb32cf7a878e7d62ULL,0x987978f1dd66d63ULL,0xb2de0dd21dd60160ULL,0x8413fce281c67f43ULL,0xe1642f72c5c67963ULL,0x29626a001c4fd50ULL,0x845802408d45100ULL,0x631ccad2d9c46771ULL,0x1db355e61dc66f60ULL,0xc17df660cd426d70ULL,0x96b41570092c6862ULL,0xe5d67707d7a27760ULL,0x80e1001960300ULL,0x1001406002000550ULL,0xb144942529961d20ULL,0xf6b4a55321c07e21ULL,0xeb90570422125571ULL,0x44e9344428844560ULL,0x8480000003000ULL,0x887ef6e94dc47660ULL,0xfb340f4845806942ULL,0xcac0ca444b967420ULL,0x8b5b93737fd27762ULL,0x92aa052a05264672ULL,0xddbc17f81de65f61ULL,0x9a303c0248c63960ULL,0xd31c5b5c8cd66761ULL,0xdbfef6356dd67773ULL,0xf9ff9ddeddce7f62ULL,0xf6bffce77fde7762ULL,0xdfffff7f3ffeff73ULL,0x8a382f2099d22100ULL,0x0,0x4000404000ULL,0x1404100200864200ULL,0x108000000002000ULL,0x0,0x2004000000024000ULL,0x4400000040020ULL,0x80000800000140ULL,0x8108200005260040ULL,0x4800000000840000ULL,0x0,0x0,0x0,0x200200021000ULL,0x0,0x8000200002000400ULL,0x10100001000800ULL,0x0,0x0,0x0,0x0,0x400200000024000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x420000000820ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xeb7f5fde77bc7ef9ULL,0x0,0x2100000041000ULL,0xdbb97f7ff23f7c6bULL,0xc1a8d90a0b3a2c00ULL,0x1000020a000000ULL,0x0,0x142200030200040ULL,0xfffffffffffe7ffbULL,0xfffffffffffffffaULL,0xffffffffffff7ffbULL,0xfffffffffffffffbULL,0xfffdfffbffff7ffbULL,0xfffeffff7ffefffbULL,0xebffff3fffffbfebULL,0xefffbbfbced43cf2ULL,0xb5accdfb56bebe41ULL,0xd7f7f7fadffd3f7aULL,0x7abe7ff3ffde2ce8ULL,0xf3b7fffff7fe3fc8ULL,0xcfff7ffffefe7ed0ULL,0xaffecffb57ffbfc8ULL,0x7eff7efeeebd6cf8ULL,0xff7e7d73ffbc3ee9ULL,0xeb9dff6eddbe2f99ULL,0x8ff6fe0ffffebde1ULL,0x81fadf5654b23fd8ULL,0xdcb6d6f9fe9fbee1ULL,0x9778ef77ff9d3e60ULL,0xbbfe377f55bc7dc9ULL,0xfff7ff3efffe7dfaULL,0xafdfdfdd769f3ed8ULL,0x34f7d2e0ff943cd0ULL,0xdffdbef777beffc9ULL,0xc29fbf775ff7eee8ULL,0x89d7fa08873c2460ULL,0x1b9ffee376fe3f70ULL,0xf6bbfffadffe2f68ULL,0xf52fb7fbde9e7cf8ULL,0xfffefd72fe9efcd2ULL,0xe19bf31bd79b6fd0ULL,0xe4e7163e729eae60ULL,0x9a9d5fdbf29e3ce0ULL,0xb1ded6e05ea83c40ULL,0xdffe7df37ebffcf8ULL,0xc392ef387fff3c2aULL,0x69c31fbe7ebe3ee8ULL,0xfffffffffffffffbULL,0xf6ffffeb7ede7ff3ULL,0xb5fe7f777ffffe40ULL,0xfbfcfffffffc7deaULL,0xffffffbfffff7febULL,0xffffdfdedffeffebULL,0xf7777eedffbfefe8ULL,0xf57f1f42d2b42c63ULL,0x9382f76c7f3e6c81ULL,0xcdfea61a7aae3de8ULL,0x8b7ddf5756fc2fa0ULL,0xf4177ce6dfb4bfd8ULL,0xf7aaa7524e1e3ca1ULL,0xffbe3ff5dfe5bd68ULL,0xeb53f7bffebf3dcaULL,0xbecdec0c70c2dd8ULL,0x9f7acf33569cfd78ULL,0x89fef470d43c3c40ULL,0x76b62c44da391e42ULL,0xebd7fcaaf79caee8ULL,0xefff3fdbfff6efe9ULL,0xddbe5de0ffdcac7aULL,0xddfef6f8febe3f68ULL,0xffae7fe25ff63fe9ULL,0xc376fb2e50fc7ecaULL,0xfbf69fefffdeeef2ULL,0xffbf4ff7ffb63ce8ULL,0x58fe77fa7fbdbfe9ULL,0x97ba3f41d39e3de0ULL,0xfea5776f5efc2f6bULL,0x9bceda875f3ebecbULL,0xab5edf5a7e1c34cbULL,0x34b75eeb5fbe64c8ULL,0x9bb4cf72f68e2c40ULL,0xfbbc7f66f79c3fc8ULL,0xccdbdb3e5fbe2ee1ULL,0x1fffdf5b771e3eb3ULL,0xfcff5fffdefeef69ULL,0xd9f5aff8ef987de0ULL,0x50a72d7356242cf8ULL,0xc182f53bfbce1c60ULL,0x8080a41ce003400ULL,0x8eb3d6e0c7fab408ULL,0xf5ec7feb7edafd81ULL,0x7b167f75d7fe1ce0ULL,0xc1e7d77fdbc2aec1ULL,0xee52329bde883ed0ULL,0x8bb2ff7b5e1e7f90ULL,0xd57efc4e768e3c69ULL,0xcaecf522f61c2e89ULL,0xfbf7fb3a74af3ecbULL,0xde50135656242ea0ULL,0xad3adf747f1efee0ULL,0xd5f657ebc7b22d68ULL,0xfbbebfd177bc76caULL,0xfda2a3787ecc14d2ULL,0x49d696201a2f3ec0ULL,0xff1ecf1f5ebe7f81ULL,0xf47676e8569cd6c8ULL,0x5360ff71feb43f52ULL,0xe3a83c23c7987460ULL,0x93b4eb124f84accaULL,0x4bffdb5bfe9a3e40ULL,0x85b67f6fdfb4be50ULL,0xd3feb7745f9ffdeaULL,0xf6b63ff6c79c3db0ULL,0x8dfcf42656bc3de2ULL,0x2fc6fb9f57f47df8ULL,0x88bacf5bffb46cd2ULL,0xd119f4796fba2c40ULL,0x4394367efe022e00ULL,0x8804e30442200400ULL,0x49521040421c04a0ULL,0x8e032f4972ec2448ULL,0x9ed0787976900668ULL,0xeeac65c3451c1da0ULL,0xa98ccb5863960da0ULL,0xcc7b12963a140f20ULL,0x703a5fe1573e3469ULL,0xfffeb7feffbf7ffaULL,0xffffff77fffe7ffaULL,0xdb9eaf5e7ffc2f68ULL,0x8772b7024eb43c80ULL,0xa9c8dcf7a8c2e89ULL,0x84d656a8f314ec0aULL,0x987de77bce9a3dd9ULL,0x5104690542aa24c0ULL,0x48022202d2002e50ULL,0xe937f2da54ff7cf0ULL,0x342f63c754ec2ee0ULL,0x6824a771cfb06c40ULL,0xb41f40c6926000ULL,0xc0821600542404c0ULL,0xa9c4df4d52906fd2ULL,0x6b0e4f61dadc6e40ULL,0x94f02410560c2660ULL,0x883e2dd3749e2ce0ULL,0x81019f2000b86440ULL,0x376632514e18ac00ULL,0x19a44b2156304470ULL,0x10682e8d4202e00ULL,0x0,0x0,0x0,0x0,0x801408c4000800ULL,0x84000080400ULL,0x0,0x100800002080000ULL,0x4000010000020800ULL,0x0,0x0,0x0,0x0,0x2000008000c0000ULL,0x0,0x1200000000200400ULL,0x0,0x0,0x0,0x0,0x0,0x1a005d40540014c0ULL,0x8180a3251b240620ULL,0x20004082000000ULL,0xbdbbfffa4bbcfd38ULL,0xb7fffee9ffbfffe9ULL,0x90e4c6346b420e0ULL,0x9154ef4e5c3e2ec3ULL,0xe8ca8689d2c82c50ULL,0xaf8a4e5ac68ebcc8ULL,0xd531fae07f983d01ULL,0x94f8ed534e9c2d78ULL,0xfb741b3a4e9eac6bULL,0xcfd1fe19fed51de0ULL,0xafdfdfde7fbf3efaULL,0xf4355fe2deb62da3ULL,0xdf8fbd735ebf3ce0ULL,0xe3106e2e5e1416e2ULL,0x8141ce085a962c48ULL,0x29b8d9c01afe0c48ULL,0x3dda15eb5e903e48ULL,0xc1240c4a5e902240ULL,0x448c0542110e0e20ULL,0xe0f5f3b247122dd8ULL,0x1400e00c0083c01ULL,0x26800040080ULL,0x90415040c6300480ULL,0x642a7560439824c0ULL,0xab84c37effb82c81ULL,0xc44b3852529c2449ULL,0x108016054040020ULL,0xa86d624a4fde3c00ULL,0x239a2e307e8028d0ULL,0x80542c0012083401ULL,0xaec2366f7fb4bee2ULL,0xeb0afff8de1ffe48ULL,0xfcfe16e8ff9a3e11ULL,0xc138fc5012103401ULL,0xccd2f6684f9d69c0ULL,0xeffffebf7fbe3ff2ULL,0x7ffedfdb56bc2fe2ULL,0xf4fff2eecfb87d6aULL,0xfffff77afffe3ffbULL,0x769e2de00ea42cc0ULL,0xa950e51ada1c2ceaULL,0x2002000002002000ULL,0x14547240120c3e20ULL,0x4120180044a02040ULL,0x0,0x4840c000408e2840ULL,0x0,0x10100000104000ULL,0x8099006212a02600ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x7fbeede3abdc2b9aULL,0x0,0x0,0x0,0x400002008400000ULL,0x0,0x0,0x0,0x2f5f15d8d334672fULL,0x0,0x913ac7228bc12702ULL,0xa0060090100100ULL,0xb50fee220ac02230ULL,0xbfff7f5f9bcc623eULL,0xf7ffffff8bdd2bf2ULL,0xfffffffffbff6fffULL,0xffbffffffbff6fffULL,0xfffff7ff9bdc6bfeULL,0xfdfffefffbff27ffULL,0xffbbfffffbdf6fffULL,0xfffffffffbff6fbfULL,0xfffe7ffffbdc6fffULL,0xcbfeffffbbdf675fULL,0x4e7f7edf9bf463f7ULL,0x40bf6fe28bd26736ULL,0xf3fffefbdbfe2f77ULL,0xff7e7df29b9f6f37ULL,0xfbfdf77efbdf6ffeULL,0xffebfffffbfd6fbfULL,0xabbedfd28bbc63deULL,0xf5fffffe9bfe6fffULL,0xfbaebf759bf5630eULL,0xdfaeff7d9bbe6ff6ULL,0xcfc6f623bbd4233eULL,0xbfdedfdbbbde675fULL,0x74f6f6eebade6befULL,0xffffff7bfbfd2fffULL,0xffb56f7f9bd7277fULL,0xcff4df3a9add6796ULL,0xebfedfd79afe2ff7ULL,0xbcfedff79a99673eULL,0xf1edbfdbbb9c2f5fULL,0xfe973f42bb8023bfULL,0x8183f1a3aad82660ULL,0xaf7bfbebba766f7fULL,0x87bed7f19bdd2b16ULL,0xc97df719dad42b86ULL,0xdefeff7abbef6b27ULL,0xd95fff3edb1d275fULL,0x4eff1f5f9adc2fdfULL,0xcffefffd9bd227cfULL,0xb775d664bb0b26d6ULL,0xefb26fe3dbcf2326ULL,0x9b84a77fda9127aeULL,0xdfcf3fffdbff6f7fULL,0x7fffd7fbbbd66fbfULL,0xfc77fffffbff6fffULL,0xeffead73bbcc2bdfULL,0xefeef77fbbf86f7eULL,0xcedefe7bbbbd6befULL,0xefffdfdb9bd66fbeULL,0xb6b7ffedfadd2f76ULL,0x59fc7773bb9c667dULL,0xdb867f64bb8c2b56ULL,0x8be5f635dbcd2312ULL,0xefffdfdedbbc2fffULL,0xb6ff7febdaff6fffULL,0xddecb6d7fbdf2f27ULL,0xfe962f739b9c6715ULL,0xc021f3061acb2f26ULL,0x3a8ebe5b9bdd0336ULL,0x8e124df3f104222aULL,0xd54ce4f6f281653eULL,0x7df47d7bbbb5233eULL,0xc9f7fa2ffbcd6317ULL,0xefdf5fdb9ad6633fULL,0x3dbddffbfa9c6f3fULL,0xc67677f79b9e678eULL,0x1ffefff0fbdd2796ULL,0xfbfd772ebb9c636eULL,0xedefbf5dbbdc63feULL,0x3f9bcff8aadc2f9fULL,0xdd7ffefbfbd66fafULL,0xdbd6ef7392962f76ULL,0xe5a67f7f9af42f9fULL,0xcffef63bdb9627a7ULL,0xa3dfdfd3abc9275fULL,0x34f7e6f09bcc2b0fULL,0xa4af2768db88210eULL,0xff912f7e9b0c63a7ULL,0xcccff0169b8c270fULL,0xab5ddad39b8a2be0ULL,0x94b57fec8ac0230aULL,0xf2dff51f9bd46b2eULL,0xffbe2f7f8bf46f4eULL,0xcfdfdf07f214672eULL,0xcc0f459ac00036ULL,0x25bacff6a384437eULL,0xc52cbf9adb9c6f54ULL,0x53363ff09aca6b5eULL,0xebdff30adb9c27f2ULL,0x6e66349f93944312ULL,0x82ba8c938bc223c6ULL,0x79fcf67ecbd86b5bULL,0x499a76789bd00b22ULL,0xb98df67cb382490eULL,0xdccb3703bb952726ULL,0xffffdffbbb9c6b3fULL,0x7154e7e39394673eULL,0x3ffefff3fbfc6f1fULL,0xbb966f7e9b57675fULL,0xa3c7f81f9b8c632eULL,0xee74ed839b162b5eULL,0x56d076ed9bdc2f5cULL,0xd72a9f6abb402b76ULL,0xffb26e6e9b802576ULL,0xc3f6e624cbc06724ULL,0x8f7cdc439b886b27ULL,0xacfe7fe79b7c2b26ULL,0xd9eead429ab62bbcULL,0xf8febfe29bc76bffULL,0xddfaf70eba1c6badULL,0xff9e5f4ecb8e6722ULL,0x8f8a6e6db294636eULL,0x70cfa7d9d914234dULL,0x4dc2437f9bd12346ULL,0x8894f3e58b3f6302ULL,0xed4e356793306964ULL,0x799cde89bfc2b22ULL,0x93f3de629a9c2335ULL,0xea343dd08b10235fULL,0x8c8c00882002404ULL,0x81c21af8ba080110ULL,0x8b0c4b069b88223dULL,0xfffffffffbff6fffULL,0xfffffff7fbff6fffULL,0xfeafffffdbdd6fefULL,0xd947af13f2122705ULL,0xef3e6fdfb3ae6bfeULL,0x25b0447493902106ULL,0x8549a3b09acc0b82ULL,0xfb8a4de49bdc6fb3ULL,0x4ac3333892816b32ULL,0x231d5cd892906e67ULL,0x13cc77b92800b4eULL,0xc12c2b4199802322ULL,0x88002578a2802100ULL,0x86cb32449bd06716ULL,0x2e9ddfce9bd96327ULL,0x301a2eb3828422a6ULL,0x9c66ac528b0403c4ULL,0xb816143099804306ULL,0x8174c52ad3002137ULL,0x2c69168092900322ULL,0x1581845013880322ULL,0x81409e29cbc06037ULL,0x0,0x0,0x0,0x0,0x0,0x8010010000004ULL,0x0,0x0,0xc000008000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x801002000002ULL,0x0,0x8800100280002100ULL,0x2200040002000ULL,0x200040000002001ULL,0x204002000004ULL,0xc0410091000020ULL,0x8080891181080314ULL,0x25002020004ULL,0x200010082000000ULL,0x404000080100000ULL,0x4186c5803020106ULL,0xb920f300b99423dcULL,0xabefb3869bcd2137ULL,0xadbedd47ba582b9dULL,0x77f256ef93006bbfULL,0x82a2ff51abf02744ULL,0x9aa6332b93906936ULL,0x45499e09b0a02036ULL,0xaededd5b9b9e6b25ULL,0xb5bb66e09b5d6f1eULL,0x179677a98c46306ULL,0x681436149b900322ULL,0x8c6020aca802113ULL,0x200897da9b702a16ULL,0xb13fc2b08ac00316ULL,0xd168ad44d3002d06ULL,0x2040e2090084204ULL,0x80431080802030ULL,0x301084093800004ULL,0x4005c4000a02204ULL,0x8070ae50c2122714ULL,0xb10592000124ULL,0x8292d76e9a502b47ULL,0x22500420a480202ULL,0x8310021080180210ULL,0x98122a0012010b06ULL,0xd3ac0522931a2304ULL,0xc1f0c75e9bdc2726ULL,0xdccffe9abba02fbeULL,0xa29e5e12d094232fULL,0x75f5536a8ad02316ULL,0x9a0d7de08b8f2736ULL,0x4028e6068b880112ULL,0xcbdab6b3d3922ea2ULL,0x981a0fd2ba946326ULL,0xdcffdffb9bdf6f17ULL,0xffffef7ffbff6f7fULL,0xf680361282850106ULL,0xc09a200880d2002ULL,0x8040182040200ULL,0x42446409200010cULL,0x8058a44080884316ULL,0x80005200b000004ULL,0x6000010850002004ULL,0x0,0x420000800020ULL,0x802a2f3013842206ULL,0x1542d308a900254ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4100008000200ULL,0x0,0x0,0x0,0x0,0x0 },
	{ 0x7ffe7ff3cbb6f7ffULL,0xbbd4ff7ed3b3f76cULL,0x4000040040008000ULL,0x85eb4da07009d10aULL,0x94bb26ace095d32aULL,0x0,0x0,0x0,0x680a08034130c22aULL,0xfcfffffef3b9ffbeULL,0xfffffffff9bfdf7fULL,0xffbffffffbbdf7feULL,0xeffffffffbbfffffULL,0xffffffdffbbdffffULL,0xfefffffffbb7f3ffULL,0xfffffffffbbfffffULL,0xffbffffffbbfffbfULL,0xeffffffbfbbbf7ffULL,0xfffffffffbbef7ffULL,0xbfff7ffff3b5f7ffULL,0xfffffffffbbff7ffULL,0xfffe7ffffbbff7fbULL,0xebd7ffbffb9cd3efULL,0xeffbbefbfba5f3feULL,0xb7ff5ffbf1a4dbaeULL,0xfefffffdf9b7dfbfULL,0xfefefdfbf9b6ffeeULL,0xfbdeff7ff9b6f7ffULL,0xbedb3e1bf3bbd329ULL,0xebbecffeebb7d37fULL,0xf7ff7eebf9bbfbffULL,0xabfeedf3d1b5f7feULL,0xefbef7fff9b9df7fULL,0xcfdff73ff9b0fd7eULL,0x1b1dc35ee999dfefULL,0x54b5fee2d1a9dbbaULL,0xf16eff62d3bdd3ebULL,0xffbf7ffbdbb4d3feULL,0xfb7fff7fe9a5d37fULL,0xefdedfdbe3bdd7feULL,0xfcf7ffe9f9bfdf7aULL,0xfbfffffef9b1ffefULL,0x1e9eaf6cc9b1d3efULL,0xe5efff2ce9b1d77eULL,0xfadfdf5bf3b5d3fcULL,0x9698dd7f8195d32cULL,0xd8efbf7ec184f3ffULL,0xfbfefffef1b7f33fULL,0xcbd5f1eee1b8db3fULL,0xcf4f7bdbf132c7a2ULL,0xbfbf7ffbd9b3ff7fULL,0xff35e6e3e1b5c3aeULL,0xefbeed70e9b9d39eULL,0xf9d4db0ee1a4dbe9ULL,0xeecbfa93c9b1fffeULL,0x8effcf6ff9b1ff2bULL,0x796dff7881b4f73fULL,0xdfbefff3f3b1dbafULL,0xff94ef77cbb1ffafULL,0xeffffb1ffbb4dbfeULL,0x8fdfdfdbf9b0d77fULL,0xd6fffeeff9b7dfffULL,0xfbefff7af98cfb7fULL,0xfa9eeffcebb5fbbfULL,0x3ddefebbfb91d16aULL,0x8bff5f5ffbb7df69ULL,0xbf767eeff1b2d35fULL,0xd7adffedd9b6d7beULL,0xdabf2ff7f9abd73eULL,0xc94affbef1aad33fULL,0xa8acdffde124d3eeULL,0x9d3b4fc661b5d7e0ULL,0x117cc7e461845732ULL,0xffb67546f088dd7cULL,0xe9fff7ffebbaf7efULL,0x3b5b5fd7e9b8f3efULL,0xbdbedd7a7b8c53abULL,0xd5cebf4cf9b0db67ULL,0x5fcc7b42a1b1d32bULL,0xebeed73ff131f32aULL,0xeffbbfdff3b1dbffULL,0xffbfcffff1bfffeaULL,0x14e5fd6ae1b1d32fULL,0x97beaf70f9aad3eeULL,0xffbcfbebe18dd3efULL,0xefffff7fe995d7ffULL,0x6e7fdedacbb5d3afULL,0xb4f5dc616080d33fULL,0xf968f66bd192536aULL,0xdb0e2fffd8b0513fULL,0xeb53ffabfb3277ecULL,0xe3e4c15e190c7aeULL,0x9c9f7fe4e1a5c338ULL,0x78c4f764c190532aULL,0xfeb67f7dcbb7d7f6ULL,0xefd2d675ebb7f239ULL,0xbefede43f9b0d93fULL,0xdffafffff0a9d3bfULL,0xf7fff7fbf3b2d39aULL,0xfffe7fffe19dd9aeULL,0xcbdefbbed934d3eeULL,0xeffe1bdffbb0fbefULL,0xf38bcd9ec9b0f3c8ULL,0xfd75fe71f391d26dULL,0x4daa5652c3acd2edULL,0xb302effd89bdfbbcULL,0xfcf7ffd3f1b4dbfbULL,0x390cdbf0d180cd38ULL,0x7076fe6ba1b1d7baULL,0x9f6eff63f9b0f3aeULL,0x9df4f77ee1a1d368ULL,0xfd4ea2e61b1d35dULL,0xbbcacfdbdbb5f7e5ULL,0x613058f44310e328ULL,0x93423452c182d1beULL,0xf79f2975c195fa6eULL,0xedf3fe9af392f36eULL,0xcfffdfdffbb4ffefULL,0xbdffddbfe1b2d77eULL,0x65eca76ed0a8c37bULL,0xdea63ffed0b1ffbeULL,0xcbf7160ad910733eULL,0x3b7f1f5bd9b8d30dULL,0xbd3366fe89b1d36eULL,0xd37ebffdf991db3eULL,0x7f966774c930ff6cULL,0x8940c360c038c31cULL,0x5fdbbe1ee9b0fba4ULL,0x8c1c7ca4c110f22cULL,0x418050220390c421ULL,0x6386af71f90051e4ULL,0x9100e22ce0204b06ULL,0xca6beb4f8bb0d12fULL,0x8b185f716900c18aULL,0xfffffffbfbb7f7ffULL,0xfffffff7fbbfffffULL,0xebaef67fe9b2f37fULL,0x83c7a732d131d33aULL,0x48dccb5bc181d36fULL,0xd6377eeec9a45b6fULL,0x9b490f5a41b1d201ULL,0xb794ff7ee1a1fb1eULL,0xc4ce6912c3b05308ULL,0x16d84c49c090d108ULL,0x300050002005200ULL,0xdd101c62c109d000ULL,0x998e0724c111d124ULL,0x4f56ea234181c2e9ULL,0x437b1bf969b3c378ULL,0xdb09cf61c0a0d2a8ULL,0xdbfdb7fb53b5f778ULL,0x408672ea4b00c18aULL,0x82d4d432c915f226ULL,0x4c9213505190c238ULL,0x6e3153ae71b1c303ULL,0xf1acf23efb31f322ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400040001000ULL,0x0,0x10001000008020ULL,0x0,0x0,0x0,0x8010000000000108ULL,0x0,0x0,0x0,0x0,0x0,0x8080014000004190ULL,0xc00001100000ULL,0x0,0x92cc51e1021924ULL,0x82104481106040ULL,0x42a214404001c008ULL,0x72069b2451840339ULL,0x87cb1759d3b05b79ULL,0x6b3fcdd4d9345b2cULL,0xb57502905080d10aULL,0x17bafd71e3b8f328ULL,0xda04737a7031f208ULL,0x4142e816e890c020ULL,0x8eb6cf5aebb2737eULL,0xf2a376c8c9b0f229ULL,0x4826a3488930422aULL,0x59af2fe8e1b2937bULL,0x500000000204ULL,0x29c53016101f280ULL,0x14195cd0c090c113ULL,0x9100416901900731ULL,0xc024256888805728ULL,0x8020000000800100ULL,0xa1001c824810c200ULL,0x16334250c0101308ULL,0x504cf3088180813fULL,0x3880050cd0204206ULL,0xc4c00209104240ULL,0x64061000002443a0ULL,0x0,0xca66181901209ULL,0xc84a2460d8205200ULL,0x8bc8e72ae9b4d338ULL,0x81fbfa1b61b5e37fULL,0x83e6c46ea287388ULL,0x10f51f4b81a463ebULL,0xceaeef02e1b4fb2dULL,0x290e72fc191d1faULL,0x895ebe838190cbfeULL,0x2d7f5fd2f3b4dff9ULL,0xfefffef3d3b7db7fULL,0xb3f9ff73f3b6ffefULL,0x1a323c200ba6c320ULL,0x1846d6845190d178ULL,0x10100040000000ULL,0x154214081c00aULL,0x0,0x800b0041809020ULL,0x0,0x0,0x0,0x12a2741e194c2a0ULL,0x4004202000a00300ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x1a70250083040106ULL,0x0,0xffdeffef8bd7f3feULL,0xffffdffb8bd6f3faULL,0x3041c36a03143122ULL,0x0,0x0,0xdfde3f9f82d7fbfaULL,0xbfffdfdf8bd7f3faULL,0xffbeffef8bd6fbfeULL,0xffffffff8bd7fbfeULL,0xffbeffff8bd7fbfaULL,0xffffffff8bd7fbfaULL,0xefffffdb8bd7f3faULL,0xfe7fdfff8bd7fbfeULL,0xfdffffff8bd7fbfeULL,0xfebe2fff8bd7ebfeULL,0xcfffffff8bd7fbfeULL,0xbff7ffff8bd7fbe6ULL,0xbfffeffa8bd7fbfaULL,0xffffbf7f8bd7fbfeULL,0xffbe3ffd8bd6ebfeULL,0xfbfbf7bb8bc7fbfeULL,0xffdfbefb81d7fb5eULL,0x9fbfdffb8bd7ebfeULL,0xfbfff7ef83d7f3faULL,0xfffe7d7b8bd6f3faULL,0xfbfdff7f8bd7f3faULL,0xefff3ac78bd7f3aaULL,0xffbfdfff8bd6fbfeULL,0xf5ffffee8bd7f3f6ULL,0xcfffaf738bd7e31eULL,0xf9a8f76d8b47f1faULL,0xcbd7fe398b86fb22ULL,0xaff9dfdf83d7b8f2ULL,0xfefff7fb8bd6f3faULL,0xdffeff7983c6f3faULL,0xfbb63fef8b56fb32ULL,0xffefff2f8bd7f3feULL,0xcbffdd5b8bc7fbeaULL,0xfdfffbeb8bd7fbfeULL,0xfdfdfe7f8bd6fbfeULL,0xfabe2ff68bd6fb72ULL,0xc1f4eb7f8955d3eaULL,0x2a7e3ec88b96f24aULL,0x34ba65fc8bd4f32aULL,0xf7bca7fc8b46e39eULL,0x3dfe7f738bd6fbfaULL,0xabd5efe683d4b36aULL,0xeef77cde8bd4e35eULL,0xffbbf7f783c6e3f2ULL,0xd53ef7ef8bd7f3faULL,0x6f2e3df98b56fbf2ULL,0xfbddf76682c7e96aULL,0x8de3bf5f8bc6ebeeULL,0xc6184f228984eb16ULL,0x7ef7f6f88257f3e2ULL,0xef7e7d758bd7fbdaULL,0xfb9efb7f83c7ebe2ULL,0xffdffe3f8bd7fbe6ULL,0xffffdfdb83d7fbfaULL,0xf6ffffff8bd7fbfeULL,0xffeffffb8bd7fbfeULL,0xffbe7ffe8bd7fbfeULL,0xcdcfdf1a8ad7b32aULL,0xeadfffcf8bd7f3feULL,0xfddab7af89d7e3e2ULL,0xf97fafd381d7faaaULL,0xdebe3ff68bd7f3eaULL,0xf78df33a8b96f3e2ULL,0x6bfefeff89d6fb7aULL,0x3e294ffa8bd7fb3aULL,0xc57ff77b8356f9a2ULL,0xd6be6f7f8bd7fa26ULL,0xcb93db8489d6eaaaULL,0xdeda7e538bd6abeaULL,0xbbb7cff78b86f3a2ULL,0xbcf63ef58b55fb92ULL,0x5fbcff728bd7d3e2ULL,0xf8d4e6a48a46e38aULL,0xeccbbffb83d6a3faULL,0xa7ff4fb383d7ebfaULL,0xf5f5bea08a56bbaaULL,0xfbb7fd728bd6fbcaULL,0xfa36fb6f8bc7eaaeULL,0xedffbe3983d4fbaaULL,0x6dfe5b8283d7f266ULL,0x75bdd67a83c6e3f2ULL,0x93bcae7a8ad4f21aULL,0xa38a2f6f8317b3e2ULL,0xcfffff9e88d7f3aaULL,0xea9cdecb0bc6e3e2ULL,0x8edf566083d5f342ULL,0xe86d266289c5e14aULL,0xfedeafff83d7b3eeULL,0xeddfffff8bd7fbeaULL,0x6fdedaaf8bc6e3eaULL,0xb779fff38b56f35aULL,0xf55dd7f98bd4a38eULL,0x87b67ff60bd4f3aaULL,0xab85caaf80c4e3e2ULL,0x77ff7ffb89d7f3faULL,0x9abe0fe383c7c36aULL,0xddffa6740a57b37aULL,0x40b6fe518856f31aULL,0x6ad4e37a01c4c33aULL,0x8de39eb98ad6a326ULL,0x13bece638943fb22ULL,0x30b7f4e88bc6c346ULL,0xfbdeff738bd6f3faULL,0xfdb0f77a83c6b8eaULL,0x8de7b93b83c6b332ULL,0x1405c782c2a0c4ULL,0x92937c908bd5c312ULL,0xc06b37e80bd6231aULL,0xd3b02a7683c2e9a2ULL,0xedcee61803c7a1f2ULL,0xe368dbca83c7d162ULL,0x3c9f44ea83c2e9aaULL,0xf5edbffc8bc7ebfaULL,0x5487ae5c8306e932ULL,0xc3a2c73581c4e8a6ULL,0xaff79fff0bd7e37aULL,0xd6ff4df98ad5b3deULL,0xf577b5f18ad7a3baULL,0x48e67d7889c4f2a2ULL,0x104e13380c6c056ULL,0x415a36118384e102ULL,0x8008c44000860200ULL,0xd3519b588203fb6eULL,0xb2e3d028217b3e6ULL,0x855a00388809102ULL,0x8090e8102028122ULL,0x9a1edff383c6fb6aULL,0x3d757feb8ad4f1eeULL,0xedfeff7083c7b3eaULL,0xfd96677d82d6a3fcULL,0xbc2be378a57b3aaULL,0x5a97d9428283e322ULL,0x101256a808078102ULL,0x17cef7a8bc7e26aULL,0xcfaa36078a869062ULL,0x59cfe20280169356ULL,0x1110820941c042ULL,0xc8725fea00d7b3eaULL,0xf118667983d4f34aULL,0xd8910d520844a1baULL,0xcc83f9098ac6a13aULL,0x18d99ec38346e3baULL,0x28324b4a8086a136ULL,0x8037144203d4b0a8ULL,0x928437410044a02aULL,0xcae4e22780c6a3daULL,0x487f1bca035633faULL,0x3fbecffb8243b3e2ULL,0x9140644208149022ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004042008400002ULL,0x840000000020002ULL,0x0,0x0,0x200400020200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x802000040002ULL,0x0,0x1000000200040002ULL,0x420200080504004ULL,0x64943c700256b122ULL,0xb180f82e8346a362ULL,0x8042360f8017f826ULL,0xf9c65728856b126ULL,0x74db2ff48b86c346ULL,0xc220240400018122ULL,0xfbd7ffff8bd7f3faULL,0x9bd7df2f8bd6f3faULL,0xe3ffde5a8357fbe2ULL,0x503c56ea0906e3a2ULL,0x8cbcaf728345f2e6ULL,0xc1082a1009169302ULL,0x4b523e000187e092ULL,0x201211580a549040ULL,0x280504200868222ULL,0x0,0x480004020000a180ULL,0x800081000000a020ULL,0x24781c1b0357e322ULL,0x980c00888809002ULL,0x801207628311f20aULL,0x1b062f6183c5e252ULL,0x800423840180a2a0ULL,0xcc158000001a1a2ULL,0x418c80000858020ULL,0x40020000040002ULL,0x20080080000004ULL,0x898461028340e9caULL,0xb2300304a020ULL,0xb04cb5a83c6c22aULL,0xa0204801c40012ULL,0xc98b2d500ac4a372ULL,0x4290e1740294b122ULL,0xcfcaf61f8ac7f3beULL,0x8091055681c4aa02ULL,0x5c94fefb8b447362ULL,0xdbffef6b8bd6bbf2ULL,0xab102c0081045202ULL,0xdcfae8a8b13f262ULL,0x246001a0113d100ULL,0x200500a002040002ULL,0x480004028000a020ULL,0xc00a080402004ULL,0x0,0x50000000101000ULL,0x40003508040410aULL,0x810478b80286a222ULL,0x4008210001500080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x0,0x190000406800000ULL,0x3bdfb3dacfdbfe10ULL,0xffffdff7cfcbfebdULL,0xf7b7fefa4fdafab4ULL,0xadbe2f72a44722b8ULL,0xffffff7fe7cffebdULL,0xffdfffbfefdffebdULL,0xffffdfdfefcffebdULL,0xfdf7dfffefdf7e3dULL,0xfffffffbefdffebdULL,0xffbfffffefdffeb8ULL,0xdfdfffbbefdbfebdULL,0xefffdfd3edcf7eb9ULL,0xfefbffffefdbfe35ULL,0xffffff7befdf7ebdULL,0xfebe3fffefdf7ebdULL,0xefffffffcfdffeb4ULL,0xffffffffcfcffebdULL,0xfffbfffeefdbfe39ULL,0xffffbfffefdffebdULL,0xfffefff7efdf7e3dULL,0xdbdfffbfcedffabdULL,0x4ef77f3eefdbfe30ULL,0x37ffcdf3cfcbfe84ULL,0xf7bcb77fced3fe99ULL,0xff9efdf26fdbfebcULL,0xfbf4ff7fefdb7eb9ULL,0xefdfbf5fefc77e35ULL,0xefbfeffbcfdffe3dULL,0x3dfff66eedc77eacULL,0x9dbc7fe30ed3f628ULL,0xcbbef77fcfdbfe3cULL,0x4fdffe13cfcbfe8cULL,0xbbffdfdbefdffeadULL,0xf5fb7cfaefdb7e28ULL,0x9f7dff72eedffe31ULL,0xdbb17f7fcdcffeacULL,0xffffff3fced6feb4ULL,0xebffffdfcddbfe3dULL,0xbdffdfff4fdbfebdULL,0xf9febffecfcbfebdULL,0xfa3e3fe7edcafa1cULL,0xebe7fb7acfc62e3dULL,0x6fff9ef54e0d7625ULL,0x3ebfcff3cfde7e35ULL,0xd1ebf57b4d83ee35ULL,0x93de796084d3f618ULL,0x88d4fb3f4dc37eb4ULL,0xaef79bffe793febdULL,0x3bbecfffefdffe3cULL,0xdc76bff94ed33eb8ULL,0x7ff87750cccb7e21ULL,0xd150772e0fcb3e10ULL,0x84c78c100cc2b209ULL,0xefb4cff0e5cbfa31ULL,0xf5ff77eacc8a7eb9ULL,0xe7fef5f2c7cf7e08ULL,0xfb9cffffcfcbfebdULL,0xdfffffbfefcb7ebdULL,0xffffffdfefdffebdULL,0xffffffffefdffebdULL,0xfbeffffacfcffe39ULL,0xffbeffffcfdbfebcULL,0xc954e72e0a185200ULL,0x8b88ff5aeed95234ULL,0xffbefffbeedbfe34ULL,0xbf6dff36c6dbfea1ULL,0x4ab63fe7c6965e95ULL,0xcb74d7bcc4c3beb8ULL,0x1573f58ccc2b620ULL,0xb7ff7fefedcffab0ULL,0x977f7febefdbfe81ULL,0x9214b19084833600ULL,0xeaf3f9aa4fca3e20ULL,0xef7edfcbcfcf7eb4ULL,0xfdbfeff2cfd7fe09ULL,0x9676d4f94ec27a1cULL,0x48bc5ff3ec8b7e80ULL,0xfbecff7ecdcf7a39ULL,0xcefe32b3efcb3e19ULL,0x7f3f5ff1cfdbfeb9ULL,0xf5fef7696cdb7e24ULL,0xf3bcff71eedb7e30ULL,0xdbb6f76ee7cb3e94ULL,0x8cdfbab68dc32e11ULL,0xae96cfd3c6cbfabcULL,0x1007c6a8c6c35220ULL,0xff6fff7be7cffeb9ULL,0xffbe6f2acdc2fe30ULL,0xefffffbeefdffe3dULL,0xbfffdfdfefcffe39ULL,0xff9fd7f3c6c77ebdULL,0xff9faf70cf8afa3dULL,0xfbbe3ff7efcbfa3dULL,0x8bfef39b4edb3e08ULL,0xeddfbd5bced37eb8ULL,0x8fbacffaefc27ab5ULL,0xdcfd379048c36810ULL,0x7d26b7e4ac9de85ULL,0x8257831a4bc11e24ULL,0xadf71b40068ad288ULL,0x378fdfd9c68bfea0ULL,0xddf756fb6dcf7e21ULL,0x7fbe7df3cfd7fea9ULL,0xdbe4f72fcedf7e1cULL,0xffffbfffefcffebdULL,0xafbfdf7ecfcffebdULL,0xf5ffdeffcfcbfe34ULL,0xffedff72cfdbfeb4ULL,0x8396e33f8cc6d6a8ULL,0x5ed67eb2ca437eb5ULL,0x7b78bf99e4c73e3dULL,0xf4b8f763e58f7eb9ULL,0x9b0b64720782de09ULL,0xd2a6067066c23220ULL,0x859dff18cccb1a1cULL,0xa62ff59c5c3fa90ULL,0xbcd17722e582ae08ULL,0x796af478eddbba98ULL,0xfab42dd6cfcb5a3cULL,0xe7ceffeaefcb7e31ULL,0xbf76affbcfd65eb1ULL,0xb772dffacfd7bea8ULL,0x606cbd474d895e24ULL,0xfb8aeff865c72e01ULL,0x88a2dbc24c463684ULL,0x142220040121408ULL,0x1498cd69cccb9201ULL,0x9041f3230c037a09ULL,0xc0b2258084805288ULL,0xd3dff51e8dc37aa8ULL,0xa040004a01814a00ULL,0x2f3ccff0eec8a231ULL,0xfdf5feabcfcb3e0cULL,0xd7af2d6347cbbe19ULL,0xda2eff6c65cf7ea4ULL,0xddf7ffbfcfcbfe34ULL,0xe3ceddcececb9a3cULL,0xfc9d7feccddafe0cULL,0xa5437f63cecbfa31ULL,0x3926704c58b7820ULL,0xebffe72ecfcb7e35ULL,0xab140cd745cb7a05ULL,0x92964feac4c31a14ULL,0x882cf60bc78a7e10ULL,0xd0322dd147ca3a10ULL,0xc102eb1a44483e00ULL,0xa1000a0a45027c10ULL,0xca06fc347cffa14ULL,0xc874008889201ULL,0xf6be3d67cf0a3a14ULL,0xaa2df203854b1a90ULL,0xc47e1a5d85821810ULL,0xb65bc76260c33a30ULL,0xe41323f00ec27a10ULL,0x800040000400200ULL,0x0,0x6000000040200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000200400200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2010314040101090ULL,0x8180807101c48a00ULL,0x4201161080023280ULL,0x809d45c00581be08ULL,0x76b6d47a47c3faa0ULL,0xa63424e048095ea4ULL,0xf8d4e15744caf604ULL,0xe4012880c880600ULL,0x302846e146837224ULL,0x30d4f64ae6cb5a00ULL,0x828a2d4044436a89ULL,0xf2a2bb5fcec69a19ULL,0x892292a140cade30ULL,0x8bef5c5a4ccb7a04ULL,0xf0bf4df6c5d37e38ULL,0x5b602c0a86d03218ULL,0x42222c4041c40e24ULL,0xd872a2020c3aa20ULL,0xa180c250c48a9200ULL,0x12220d4105c20a10ULL,0xd004042480012a0ULL,0x18822460a0c10200ULL,0x480000000400200ULL,0x981b120088885e20ULL,0x24004084800ULL,0x28234284022600ULL,0xabd6756e4dc39a05ULL,0xc842328b4cc35230ULL,0x4482021004002601ULL,0x200001000004800ULL,0x8001000001000800ULL,0x4002258009411800ULL,0x202000040800000ULL,0x742000000005680ULL,0x48180b1184ca3200ULL,0x1040220060000220ULL,0x284d4044cb7a14ULL,0xd700f16a05c62a08ULL,0xdfdfbeb3efcafa19ULL,0xb94a7d0045c21a21ULL,0x54bf5ee6cdc37eb1ULL,0x97fbff6acdd33ebdULL,0x9bbe6f3605c6ee38ULL,0xe8614f2cc4927a05ULL,0x2d2d240805621ULL,0x800010000000600ULL,0x50000800800ULL,0x4206100240081230ULL,0x0,0x4052000a05020818ULL,0x0,0xa025042082840600ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x44000000020200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x1080500a00a00c6ULL,0x0,0xebcfb37f2a4e80c8ULL,0xffbffffffa7fbbefULL,0xf17d7658fa6610ebULL,0xfffefff7fa7fbbffULL,0xfffffffffb7fbbffULL,0xfffffebffb7ebbffULL,0xefffdfdffb7fbbdfULL,0xfffffff9eb7fbbefULL,0xfffefff7bb7fbbffULL,0xffffbf77fb7fbbdfULL,0xefffffbb7977bbcbULL,0xffffffdffb7fbbefULL,0xfffffffffb7fbbffULL,0xfffffffffb7fbbffULL,0xfebfff7ffa7fbbffULL,0xeffffffbbb7ebbffULL,0xffffffdffb7bbbefULL,0x9fbf6fff7a7f9bcfULL,0xffffffffea7fbbdbULL,0xffeefffffa7e9bfbULL,0xcbf5ff7f5a769be7ULL,0x71c32f548a7219cbULL,0xfdbaede2ab7e98cbULL,0xfdfff67bfa7a93fbULL,0xfffc7ffa6b7e19cbULL,0xcbfdef3fea7e92efULL,0xeffbbf7feb72bbcbULL,0xbfbeffffeb7fbbffULL,0x7dff76fa3a763bcfULL,0xb7a8fdf1fb7a1bcfULL,0xdea52f6f3b52bad3ULL,0x4fcbbbdb2a769a87ULL,0xabffdbdf6b7f3befULL,0xdefff6ebea5f1aefULL,0xcadb7ff32a7f9bcfULL,0xefbe6c232b5213c1ULL,0xcfffffbefa7b3aefULL,0xefdfdfdafa7fbbcbULL,0xbffefffbba7f93fbULL,0xdbfdf75ffa7bbbdfULL,0xdfb7ff7f2b7ebbcfULL,0xebf6ffff697a1afbULL,0x1add5e6ffa529aefULL,0x3709cfdb2a57baefULL,0x97fcf72a6856b98fULL,0x9216a32c28621389ULL,0xebecffbbea76b8ebULL,0xeeda7f50aa731b7bULL,0x8c994ff3bb4a1847ULL,0xfdf4f3f069761aefULL,0x7fec3f62ea74b2dfULL,0xdbf7cb1e3b7a3bcfULL,0x47db3795aa539b9bULL,0x7bfdff32b761bcbULL,0xf591feebba7a93cfULL,0xf6bf7d733b521adbULL,0xfffefbfe6a721befULL,0xdffffffffb7fbbefULL,0xffffffdffb7fbbffULL,0x3022c008010282ULL,0x5b7e77e148321b93ULL,0x200200000440000ULL,0x0,0x0,0xf6ff7ffea97fbbcfULL,0xd9ffefea4a7239cbULL,0xeb23f5548733b57ULL,0xcddbdb1ceb7533efULL,0x4ffffedbeb7e11efULL,0xfffb7ffbfb7fbbdfULL,0xb6bfff7c2b7e93dfULL,0x7ffe3ffdea5f9bcfULL,0xcbfbfff7ba7a9befULL,0xeeda1f593a52b9cfULL,0xf3aecd5a2b7713cfULL,0xf56cf77d6a7e9adbULL,0x5d3efdf27b7213ebULL,0xfbf6ff3f6a7f9befULL,0xcf7bffd9ab7e19cfULL,0x23aac5130a6291e9ULL,0x7c7fffec6a7ebaefULL,0xdfbfe7f07a721b8fULL,0xff9f7fed7b721acfULL,0x1bceba867a5292e7ULL,0xac27d1dbaa749bebULL,0x94a55c680a4102efULL,0xdd3cfffbfa7b1bcfULL,0xdffffffffb7e9befULL,0xffffffbffb7ebbffULL,0xffffffdffb7fbbffULL,0xfffffffffb7bbbffULL,0xfffffffb7b7bbaefULL,0xffbe7ffdfb7ebaefULL,0xe522ef3b2a721a8bULL,0xab5e5ac62a5513cfULL,0xb2badf5bbb5a19cfULL,0x5072af5a0a7a13c3ULL,0xf68ebf6cea7e12cfULL,0x897c72eb7278104bULL,0xede1e580a541bcaULL,0xbe99effb485211cbULL,0xf4ffe7787976128fULL,0x4abefc736b7b33cbULL,0xebc7f72e6b7a19dfULL,0xfffffffffb7fbbffULL,0xefbfdffffb7ebbffULL,0x79d45e4aa8679bc9ULL,0xa13ead70287a9bcbULL,0x5b94bf656a7e194bULL,0xcd5a16012a7a3ad3ULL,0x3b7edfdb0a7e1bebULL,0xfadb5ee6ea5a91ebULL,0xd9bdad6b6a7a99cfULL,0xc0142a6e88121082ULL,0x58f8a2a48449883ULL,0x27ea5cd8627a02c3ULL,0xe494d4e1ab461b0fULL,0xf26caf2088521b4bULL,0xa2163d654a62936bULL,0xefeff73aeb571b8bULL,0xefffffddfa7ebbefULL,0x9ffbdffb6b771bdfULL,0xb15d27699b7a92cfULL,0xbab42565ea67b2cbULL,0xdbfa7bfa6a563ad9ULL,0xa80a124148623a81ULL,0x9882eac92a421bdbULL,0x552cf641aa4a10cbULL,0xa8c2de0287a114aULL,0x9c0c7300278024aULL,0xc7c0b7560b629a4bULL,0x26ba91c248421bcbULL,0x9cf41e6eab7693ebULL,0xdbbeff70ea7a9bdfULL,0x4f9c370d2a428acaULL,0x8be3de302a561a87ULL,0x796cdd60a7a10deULL,0x92522644a862028bULL,0xc9273d600a6a1869ULL,0x8bba16714b6a3047ULL,0xe446e6380a631aafULL,0xaa08fd1742581a43ULL,0x426519fea1210c3ULL,0xf89c82720a631acbULL,0xee9e2ff60a5713dbULL,0xc91282382a4018dbULL,0xa7129223285a91c8ULL,0x92aa4e700a671a86ULL,0x911dd22828430a8bULL,0x44d429d00868104bULL,0xe1b0c335aa6e80cbULL,0xb400e0248000949ULL,0x8210000a0a440081ULL,0xd5459016a8621083ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000800000021000ULL,0x0,0x0,0x80004002000001ULL,0x0,0x0,0x4000010000400040ULL,0x0,0xb1107704884b1bcaULL,0xfb80f22ee8781aefULL,0x268290010810190aULL,0x80808800a000102ULL,0x10280709800180aULL,0x100020000600000ULL,0x8090e70c2b723247ULL,0x88c0031120500148ULL,0x20002000000000aULL,0x0,0xfa38e58039711a8eULL,0xc0a0a12e025012eaULL,0xcc4bfe89ba7a1afbULL,0xea744ed34a6303f9ULL,0x96d75efa687211dbULL,0x20c2003400040ULL,0x9010824008029801ULL,0x80c4901088650282ULL,0x8a60104301509283ULL,0x1004040002440002ULL,0x403c2c40a2521048ULL,0x0,0x902400000400282ULL,0x456000820021848ULL,0x10000200080040ULL,0x0,0x0,0x28907202084418c8ULL,0xcc827f183a5a30c7ULL,0x8884d800a52120bULL,0x10400000600000ULL,0x20804d6aa3721040ULL,0x120480000006ULL,0x1000020000000082ULL,0x8810044808621000ULL,0xc9e076408501387ULL,0xbefeff726a54bbe7ULL,0xecf6ef68ea5b9acbULL,0xcfdfbeaf796a93dfULL,0xa25646c2ca76104bULL,0xf4f776ee3a7699cfULL,0xfbffff7aea779befULL,0xfa982d4e28609bbeULL,0xc0032a0389501242ULL,0xa611024008798343ULL,0x4204d000a008041ULL,0x0,0x8000206000500040ULL,0x2840400008280040ULL,0x10520002101040ULL,0x23aa41ed205a13c3ULL,0x12284442a8521002ULL,0x0,0x0,0x0,0x0,0x8000002080000002ULL,0x2004000420000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000101000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x4012001000180084ULL,0xfbd7ff7f437d8ff6ULL,0xfffbfffbe77fdff6ULL,0x7b9ccfffe27fbff6ULL,0xf7ffffffe77f9ff6ULL,0xef7fff73e57eddf6ULL,0xffbeff7fe77fdff6ULL,0xdffffebfe77ffff6ULL,0xafffdfdfe77fbff6ULL,0xf4ffffefe77f9ff6ULL,0xfffeff7b677ffff6ULL,0xff96ffffe77e9df6ULL,0xffdfffffe63ffff6ULL,0xffffffdf677f5ff4ULL,0xfdffffffe77ffff6ULL,0xffffffffe77ffff6ULL,0xfebe7f7fe73fffd6ULL,0xefffffefe77ffff6ULL,0xdfdfffff277fdfd6ULL,0xbd3affffa33d1ff4ULL,0xfffdffff633ffff6ULL,0xd7b66dfb677f9ed6ULL,0xfbf6fbbee77f5ff4ULL,0xe5df1ffee73f5ff4ULL,0xbebfdfebe37ffdd6ULL,0xfdfff66f267e5fd4ULL,0xfff67f73e77e3df4ULL,0xfbecef2fe77f5df4ULL,0xbffab36fe67f5ff4ULL,0xcfbfcfff667fdff6ULL,0xf5ffffefe67ffff6ULL,0xdfbebf72e53fdd56ULL,0xdbae77fd267c13d4ULL,0x4d93d80c00767d40ULL,0x8fb6df57617d9d72ULL,0x312012c1603804f0ULL,0xa54de76861059d56ULL,0xcaae7fbe637fdff6ULL,0xfdf7ffbfe77fbff6ULL,0xff7effdfe37f5ff4ULL,0xd73f7febe77eddf6ULL,0xf9befff6673f9ff6ULL,0xffbeaf7fe67ffff6ULL,0xebd7fb1e277efef6ULL,0xabfe1ccfa6759dd6ULL,0x84eb6758a63cdeb6ULL,0xbbbca371e53efdf2ULL,0xfb96bde6e13e5fd4ULL,0xf176ff7ee37f3df4ULL,0x4ef7be58627f1ff4ULL,0xbfbddf7be37d5ed4ULL,0xdbdff67d673e9fe2ULL,0x7ffeff78e37fbff6ULL,0xfb96ff2f663c3fd4ULL,0xec39adbe23d5ce4ULL,0xaebf4f7aa37d1ed4ULL,0xf4dffefce33e3bd0ULL,0xb7bea570e27f9df2ULL,0xdfbef77ee77f9ed6ULL,0xcfffffbfe37fdff6ULL,0xaf7993da663537d0ULL,0x0,0x0,0x0,0x0,0x200031002140084ULL,0xdfb2ff69673ffcf6ULL,0x104702000111740ULL,0xdabe2fefe37e8de6ULL,0xedffdb5aa73fdbd6ULL,0xefffffffe77fdff6ULL,0xfffbfffbe77efff6ULL,0xf7fff76f677f3fd4ULL,0xf7fefdfd677f9ce6ULL,0xc3f6fb3f63379dd6ULL,0xdefe3fffe67f0fd4ULL,0x7fbdcf42407e9ad2ULL,0xfdfefefa227f9ff6ULL,0xdbfe7fe1e67d3ff4ULL,0xfbc2f76ea63cbbd6ULL,0xcfffbebfe57d99e6ULL,0xf79f4f7f627fdfd6ULL,0xf5fff7fa627fdff6ULL,0xfe3635f2423c9ed2ULL,0xfbaeffaee33e5fd4ULL,0xcd5fd68fe35efdd6ULL,0xa90e4fdf633e7cf4ULL,0xf6b476e8e33e5cf0ULL,0xdffeff72677f5ef4ULL,0xffbfffffe7775fd4ULL,0xffffffffe77fbff6ULL,0xffffffdfe77ffff6ULL,0xffffffffe77ffff6ULL,0xff7fffebe67fbfd6ULL,0xfbbf3fffe77fdff6ULL,0xc7d7bfbb633c9ef6ULL,0xeffffb7f637dfef6ULL,0x3a884fdb603c5fd0ULL,0xd3bcb570203696c6ULL,0x92c662ece0109bc2ULL,0xa1a5fa2f627eb0b6ULL,0x4e461247004f1cd4ULL,0xbe1e7de1667d1dd4ULL,0xf44cb66d663e34f0ULL,0x68963d72a02c5ed0ULL,0xfbfcffffe77e7fd4ULL,0xefffbfbfe57fddf6ULL,0xefffcfd0e677dfd2ULL,0x71b4f6ff23778bd6ULL,0xe3b8fd70613f5fd4ULL,0xdab52b7ee7571fd0ULL,0x49733a0b230e88e2ULL,0xbb2bfb97e2359ed6ULL,0xd62f5cb2822d9bd2ULL,0x1048d610604c1174ULL,0xfaad3f36261e17f4ULL,0x27c6c80e623c14d0ULL,0x27f6fbdb623d97d2ULL,0xb2b06fe9673c1dd0ULL,0x4128b778631d15d4ULL,0xd994a7d1613e38e4ULL,0x8ef01a0422361c84ULL,0xefffffdfe77ffff6ULL,0xfffffffbe77ffff6ULL,0xcbf8ff7bc37c4ec4ULL,0xff162bd462355dd4ULL,0x8093eb6e26143654ULL,0x6a40ca74265d05c4ULL,0x40924c83601515c0ULL,0xd555fc82a5171d50ULL,0x33667f71647e14d0ULL,0x7283eb2e623f98d6ULL,0x194b1aca62141dc0ULL,0xd9a89e3631c0ce4ULL,0x50224eeae13cdd86ULL,0x75befd60677e1df4ULL,0xeaac7ed5623eb7c6ULL,0xde6bfe1ae03cc8c6ULL,0xb673db4ec376cc06ULL,0x901381c060531444ULL,0xede44bc0e2541d74ULL,0x40843460e0101c84ULL,0x8840ae0880553ae4ULL,0x6a7e5190015c3bd4ULL,0xb5f7577b607f3dd4ULL,0xd9e4a460805fd986ULL,0xf8022dd0641ec886ULL,0x8840a062605c80a2ULL,0x4f1c0e0b61185a74ULL,0x31a2069e64764a54ULL,0xd1aeb6ea427e9dc6ULL,0xdf802f10623c0844ULL,0x89dde32fc05b8876ULL,0x49c01a15265e0624ULL,0x36f99fe1247fda76ULL,0x92c0144c65140e04ULL,0x0,0x0,0x0,0x0,0x0,0x30000000001400ULL,0x0,0x8000000000080080ULL,0x2000010000001400ULL,0x0,0x0,0x0,0x802001000800ULL,0x0,0x8000002060000000ULL,0x2004000000080080ULL,0x0,0x0,0x0,0x0,0x240000080004ULL,0x302c51200c14d4ULL,0x0,0x8014ac8403c08a0ULL,0x18886800001c00e0ULL,0x1000060004a4804ULL,0x49260782253c1f40ULL,0x1391c53c646f10d0ULL,0x6828a0000351444ULL,0x3004a0000261c00ULL,0x5193f068233e1694ULL,0x93f4bf70233c1ca4ULL,0xc8183122627c9812ULL,0x4b52364a647e00f4ULL,0xabaa0f1b603f97d6ULL,0x8320040000188242ULL,0x90e9272462361440ULL,0xce80131e623c0010ULL,0x45c6a310421ccc12ULL,0x480000240000ULL,0x15240b4181424800ULL,0x12e0730002c00d0ULL,0x4406000020001480ULL,0x0,0x0,0x440000040100000ULL,0x80002440000c04a0ULL,0xd210772423351d50ULL,0x8800230821101470ULL,0x2540004882040940ULL,0xc23e0562620412d0ULL,0x940c2402001d11c0ULL,0x80200c2222342040ULL,0xc00000100004ULL,0x840000000000240ULL,0x29304f00a056a986ULL,0x1102008000051400ULL,0xc8295c2143760e40ULL,0xfa26e77e627e9bc6ULL,0xdeeffbbfe57fbdd2ULL,0xed7bfbca823e9ff6ULL,0xfcb776efe77c9ff6ULL,0xfdff7f7be67f7fd4ULL,0x93886ef6653e5dd0ULL,0xa610431a223c5cc0ULL,0xa0081182001812b0ULL,0x0,0x0,0x8800280040300800ULL,0xcd6120024648c02ULL,0x206cba2062051080ULL,0x343b47e4241435d4ULL,0x6000003002440000ULL,0x1200041000050044ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0xfffefffbff29ffdbULL,0xfbffff7fffa9ffdbULL,0xffefb6ff5f29ffdbULL,0xffffdffeefa9ffdbULL,0xfdffffff7fa9ffdbULL,0xffbefd71ffa9ffdbULL,0xffffffffffa9ffdbULL,0xffffffbfffa9ffdbULL,0xefffdfdfbfa9ffdbULL,0xf5fff7fd7ba9d7dbULL,0xdf7ffffbff28dfdbULL,0xdfbe7fffefa9efdbULL,0xcddfffab5da9ff9bULL,0xffffffdeffa9ffdbULL,0xffffffffffa9ffdbULL,0xf9ffffffbfa9ffdbULL,0xffbefffdff89ffdbULL,0xefffff7e7fa9ffdbULL,0xaffffef9dfa9f7dbULL,0xb7fedffb9b29ffdbULL,0xfffffffbffa9ffdbULL,0x7ff6ff7ddb29ef5bULL,0xa3f7ff7f7f20efdbULL,0x7a761f5f7b2967d2ULL,0xf79effffbfa9f7dbULL,0xb4fffe7f3fa9ffdbULL,0x6bfa7fe2df28dfdbULL,0xf1d2f72f5b29e79bULL,0xeddebe797a28af5bULL,0xff9fcfdb7b2897dbULL,0x75ffd67e7ba0ffdbULL,0xfbbfaf73db29cfdbULL,0xdf9fe54b0b28d3dbULL,0xefff3e17cb09ff5bULL,0xab5f5fc10b29ef59ULL,0x947176e23a01e79bULL,0xfbee7f731f29ef5bULL,0x79d727673b08ff1bULL,0x8fdbff2f5ba9c7dbULL,0xafffffdbefa8ffdbULL,0xbfffdfe3ff29ffdbULL,0xfdffffff7f28ffdbULL,0xfebe3ffc5b29efd3ULL,0xcdfffe335fa8ff5bULL,0x2deab3e91b01eddbULL,0x8eb377ee1fa1cfdbULL,0x91fc976a0a01f79bULL,0xc7a4ff4423008393ULL,0xebf7f7eeef88cfdbULL,0xeec7b6dabb20ffdbULL,0x3fbcefdb9b29f7d9ULL,0x16f4f67a1aa9c753ULL,0x5afefd71bb08e7d9ULL,0xe8f4b72f4b09dfd3ULL,0xc7db9e93dba0d7d3ULL,0xa79bcfd2df21c71aULL,0x34d576fbffa8575bULL,0x9fbffe631fa9d79bULL,0xfbfafffbffa9ffdbULL,0xefdfffba1fa9ffdbULL,0x200001000000440ULL,0x0,0x0,0x0,0x0,0x0,0x94b376ecfb29e7dbULL,0x0,0x30000008010ULL,0x6fdff7f6bfa9ffdbULL,0xffffffffffa9ffdbULL,0xfffbfffb7fa9ffdbULL,0xd4efffff4fa8efdbULL,0xd7fc7f6ecfa1df9bULL,0xcbd6fabf7f20eb9bULL,0xffdf7edb1b01d791ULL,0x3fffcfdfbe89c5dbULL,0x975cba7b5f21cf4bULL,0xffbe7d737fa8efdbULL,0xb3f7ff3a6b28e7dbULL,0xdbd7f6ffcb28f7d3ULL,0xefffdff32ba8ffd3ULL,0xfdffdeff9f29c75bULL,0xbfffff77ff09dfdaULL,0xffb7ffffbfa9ffdbULL,0xc3d6b61f1b28cf1bULL,0x9b5e5f5a1f08579aULL,0xfcfff6ed7b28efdbULL,0x994ee7697a09e711ULL,0x0,0xdcc77eb79b21fcd9ULL,0xffffffdfffa9ffdbULL,0xffffffffffa9ffdbULL,0xfffef7ffffa8ffdbULL,0xfa9e2ff3ffa9efdbULL,0xefdffffb7b09efdbULL,0xefdeff531f29efdbULL,0x9a3827ea4b008793ULL,0xd538fc7018a9cd5bULL,0xa69a2565d9088753ULL,0xe2e5f3fb3f28c7dbULL,0xe6c7325a3b08cdcbULL,0xb99eebea3b01f753ULL,0xbf7ea76f1b215f59ULL,0xeb767d715b28ef5aULL,0xfbdcff0ebf00df5aULL,0xaed2fa8f7981cf5bULL,0xfffbd7f25f21efcbULL,0x4520b04809084343ULL,0xbb6c6df28900dfdbULL,0xeabc755e0f01dd5aULL,0xdffe339fbf29ffd9ULL,0xff7fdfdbdfa9ffdbULL,0xc4a4d4608f00c61bULL,0xa123ff381a08e7d8ULL,0x571064703e004dd3ULL,0xa4a6c8004b0021d0ULL,0x652d9500921c55aULL,0x91bfcf603f00e3d1ULL,0xffeeff63fb889fdbULL,0x5ebdaffd9f28f75bULL,0xc5dff726df01ef9bULL,0xaffffffe7f29ffdbULL,0xffef7fffbf29ffdbULL,0x83e12c52822885d1ULL,0xc01635d0ac28450aULL,0xbd4c62f3a88df5bULL,0xfeee1f539388d75bULL,0x831ec9522a08a580ULL,0xf07eb6ea0f08cf8bULL,0x309c34604928a180ULL,0xf8a1db3e1a21e7cbULL,0x86a101d0900a0d0ULL,0x2dbecd504b20d599ULL,0x7565ede66a2881d1ULL,0x5deeed624a08cd8bULL,0xce12a7060b087e59ULL,0x4843fb041900cb41ULL,0x379cdfda7f29c79bULL,0xb231256a0b080f4bULL,0x9342bb400820cd41ULL,0x8310eb710b08c751ULL,0xd5e27162b08c313ULL,0x14c5d00a000851aULL,0x1462572809200a93ULL,0xbd78e7a0a989f799ULL,0x88860a624c88cb81ULL,0x8bd67b204408cdd3ULL,0x800808448002100ULL,0x28a44164b08cf1aULL,0xd069f67c8b007f5aULL,0x84082d490708450bULL,0xe8e6e26c0b20fd5bULL,0x2fda92020921f212ULL,0x133858301e205350ULL,0x844010200a204900ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x404000001010ULL,0x0,0x0,0x0,0x0,0x0,0xbac63fc46788d71bULL,0xc8f0c3270a00fe98ULL,0x4200021002080101ULL,0x0,0xd1c0a4126a004f59ULL,0x4806202028080108ULL,0xfb09dd639209dd1bULL,0x40021c101a000540ULL,0x1cdf7a0829e34bULL,0x548800008013ULL,0xedce7d70ba08ef19ULL,0xa08c6f673f21d353ULL,0x4942b2760900c71aULL,0x801e4d433e01ddd1ULL,0xc00086e032001593ULL,0x2e044048008350ULL,0x8800020480008410ULL,0x81d9ed0a86004f1bULL,0x201420008002000ULL,0x0,0xc060ac002a00a119ULL,0x140012000000ULL,0x2004008000080ULL,0x14800000a002010ULL,0xa4466802008451ULL,0xc001200008208010ULL,0x20004001004000ULL,0x100800000008409ULL,0x110008200000ULL,0x90ccb7008a82f19ULL,0x904800000100ULL,0x47c71020087c1ULL,0x26800009090ULL,0x480020201000418ULL,0x294c4543a800140ULL,0x585886e30b0003cbULL,0xb3361c700820271bULL,0xff9be75cdf29d719ULL,0xc7e7ff9f3f28ddd3ULL,0xafffdfdf3fa8dfd3ULL,0xf5767efcef89dfdbULL,0xdfefff73df28ffdbULL,0xd5bc26677b28b79bULL,0xa84d4a1909084553ULL,0x1004460e204010ULL,0x469016700d00478bULL,0xf0c4b55c4909cb11ULL,0xd21001424c284110ULL,0x100e3240b088382ULL,0x121092c2080885c9ULL,0x95f2a5e00ba04948ULL,0x9e75b7f85b884f5bULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24020000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204000014000ULL,0x0,0x0,0x0,0x1000040000000300ULL,0x0,0x0 },
	{ 0xfffefffbfdfeffeeULL,0xfbdff3fe90ba4ceeULL,0xffeffffffdfefde2ULL,0xffbffffffdffffeeULL,0xf5ffffffbdfeffeeULL,0xdfbefd71fdbbedeeULL,0xfffffffffdffffeeULL,0xfffffffffdffffeeULL,0xbb7edfdbfdfffde2ULL,0xf6fdfffdbdfffdeaULL,0xebffffffbdbbfde6ULL,0xffbe7ffffdffdde2ULL,0xfff7ff3efcfefd6eULL,0xffffffdffdffffeeULL,0xfffffffffdffffeeULL,0xf7ffff7ffdfffde6ULL,0xffb6ff7ffdfbfde6ULL,0xcfdeffffbdffeeeaULL,0xbfff7ffff5fecdeeULL,0xffbafffb35fb7deaULL,0xf57fbffffdffedeeULL,0xffdebff7fdbf7d6eULL,0xf9ffffba7cbfff6eULL,0x6edf6edd7dbbff6aULL,0xbffbcffefc9edde2ULL,0xd7df677facfafe66ULL,0xdfea7d73ddfddca2ULL,0xfbe7ff7f5dbafd62ULL,0xdec2bfdb55bbec6aULL,0xfebefffb9dbeeeeaULL,0xb4fc77e83cdadde2ULL,0xb3beef7474f954e2ULL,0xeeb7f6ff3dbb7deaULL,0xcde7de1e74bee562ULL,0x2a165bcf34b06b60ULL,0xb0dd6eea30fbec62ULL,0x8b0e657295bc5f42ULL,0xdba0f7fc5dbad5eaULL,0xefdfffbffdfffeeaULL,0xebfedfdbf5bb7feaULL,0xbdfffff7bcbbfd6eULL,0xfb7e77723dffcd62ULL,0xefbeffff75fe6fe2ULL,0xefd2f3b975baff66ULL,0xe9f7ddc6ddbb4c6eULL,0xffc36feb749efce2ULL,0xfd7db7ea74df5d62ULL,0xdbfeff6e0dbedde2ULL,0xdbdef33f54beffeeULL,0xee971e1f15ba4deaULL,0xdebfdff2b4be6c66ULL,0x9e75fe6994baee66ULL,0x7afe7d7135ea6d66ULL,0x9bf5cf1715ba5d62ULL,0x48621f2e25b0c340ULL,0xfdba4f2bf5aacd6eULL,0xbcf577edddba7d66ULL,0xff7eef7275bffdeaULL,0xcffeffffdcbaffeeULL,0x50000000080002ULL,0x0,0x0,0x0,0x0,0x1020008004000ULL,0x0,0x40000680082000aULL,0x0,0x2004000000820ULL,0xc6c3366105beefe0ULL,0xfffffffffdffefeeULL,0xbb3bdffbbcbbfd6aULL,0xfc7dbd6a349e5ce2ULL,0xfe9675fc34beff6eULL,0xc8d6fbaf5493fd62ULL,0x4ff71ead75bafd66ULL,0xbfdfdda87db9fd62ULL,0xf77efffff5fa5feaULL,0xffbe77f2dcfacf66ULL,0xbbd5f77e74ba4de2ULL,0xefd3fe9bfc9f6d66ULL,0x9eb8ccc924fa7c6aULL,0xd4dffbf73caaff6eULL,0xa335f7333cbbcfe2ULL,0xfebf6f7efdfaed6aULL,0x6cc6d61554a8b462ULL,0x4b9f5b5844d2ede4ULL,0x94fddfeaf4fb6f6eULL,0x0,0x0,0x0,0xffffffdffdffffeeULL,0xfffffffffdffffeeULL,0xfffffffffdffffeeULL,0xfefebffffdffffeeULL,0xfbfffffffdfeffe6ULL,0xfadfff6bbcbedd26ULL,0xa7baddfb34a95de2ULL,0x9dfcbef915bb7d22ULL,0x7136f3e094125462ULL,0x8bf0fb7a99fa4e66ULL,0xbe67be4034fb1de2ULL,0xfba9dffbbcfffd62ULL,0x75befe46d4d84de2ULL,0xbad425c0009a6ce2ULL,0xdbffff3fbdf96de2ULL,0xa1cabedb5dbefde2ULL,0xbbbedcf034bc4fe2ULL,0x7436ffd294ded5e2ULL,0xbfba3d75acdedde0ULL,0xcd97f726f5bb6f62ULL,0xe95716b264b64ee4ULL,0x2f3e97d655bf6fe6ULL,0x92b57eee94badc46ULL,0xe055af38bc92c546ULL,0xda107b7c35ba6442ULL,0x6fc5761ed1b34de0ULL,0x8821051108c0042ULL,0xbfbe4dbbd53b5feaULL,0x996cafee19fb4c4aULL,0xda90277f35ba4deeULL,0xc1e4b3261dbae462ULL,0xefff5fddf5bb7de6ULL,0xfdbfff7f75fefd64ULL,0x423037733db05860ULL,0x7b165d7715bb1c0aULL,0x5940f00245b2a16aULL,0x1ecd5b6038baec62ULL,0x8093cc8310384ca0ULL,0xf93c984a149a5c42ULL,0x33f23d40149b4c2aULL,0x8b89c32414994c40ULL,0x8edabea0349b5c42ULL,0xb884f8345bc0da0ULL,0x1443006491a88d22ULL,0x94bc2d62648a4d62ULL,0xdf96ffe53cba7d62ULL,0xc8471e99f49a4564ULL,0xa960f5a103a4c6aULL,0x9c35206a949b0ca6ULL,0x2568874040926ce4ULL,0xda802be0049a0c46ULL,0x244021200024820ULL,0x94421300a248e2ULL,0x14bb6f694daa6262ULL,0x4148a240149a0c24ULL,0x4aa2a79604aa4162ULL,0x8874218a048e0064ULL,0x8d6b2408498ec60ULL,0xe21065814214842ULL,0xd1380400248ac942ULL,0x2ebe7d6130d7ed62ULL,0x83b1f37304ace160ULL,0x80712460088e842ULL,0x8046c1242a0142ULL,0x2800900000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000200002ULL,0x0,0x0,0x0,0x0,0x212205200040ULL,0x1080000004020ULL,0x0,0x0,0x5d4e7df23cbe5d42ULL,0x5040aa0805a04120ULL,0x41120c0010920042ULL,0x808a20c904000ULL,0x8420217014a85622ULL,0xc80e7de08dba0ce0ULL,0xfb1cf0041cda6542ULL,0x5c2928281aa0c20ULL,0x800210000024802ULL,0x1021014010888802ULL,0x19882d030daac8e2ULL,0x8930034810194d40ULL,0x48d312111cb90402ULL,0x4a5148501c94e500ULL,0xa84ca030026842ULL,0x9182a36880ba644eULL,0xcb9342d814bb4de6ULL,0xe000010814000142ULL,0x2004a00001004a0ULL,0x0,0x4030200004020102ULL,0x24000c000000ULL,0x44000801200200ULL,0x208c00a009060a0ULL,0x30000ce404105560ULL,0x840826000080e040ULL,0x2000400000000aULL,0x820000a08084000ULL,0x200812d7148a2c42ULL,0xc2aa8f7afc80756aULL,0x300300280040ULL,0x102c0d00052049a0ULL,0x8b5a400304846840ULL,0x850368315a86042ULL,0x231806d0b4084d60ULL,0xd0421c68108e4820ULL,0xa218d7a03090cd6aULL,0xd514ba742090cd6aULL,0x4ecfff3bbcaaec62ULL,0xbafffdde3cba6d66ULL,0xb6f7ffeffdbbffe6ULL,0xfdef7773dc9f4ceaULL,0xf6b33ebfb4aaddeeULL,0x8481810808c24840ULL,0x900001800084040ULL,0x1000040000900000ULL,0x412c20300082000eULL,0x22c0011200002ULL,0x503c2001028000aULL,0x6f498d014982462ULL,0x1cb15d7001e8c1e2ULL,0xe875162004b644a2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x480000000100400ULL,0x0,0x0,0x102044010210002ULL,0x8000010080000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0x1280d20e6337ee4ULL,0x8010000000000280ULL,0x2b5395dfe4f2f6eeULL,0xfffeffff76f3ffeeULL,0xfcfffffff6f3f7eeULL,0xdffefff7e4b3feecULL,0xfbfef77ee6b3f7ecULL,0xfffffffff6f3ffeeULL,0xefffdfd7e6f3ffeeULL,0xfef7fefff6d3ffeeULL,0xdfffef7ff6f3f6ecULL,0xffbefffbf6f3f7eeULL,0xeff7ff7ff6f3feeeULL,0x7fffffdff6f3feeeULL,0xfffffffff6f3feeeULL,0xffefbffff4b3f7eeULL,0xfbbfbfffe6f3feeeULL,0xeffff7ffb6f3f7ecULL,0xa7ffdff9d65377eeULL,0x15bb7ffab473feeeULL,0xfdfffefff613ffeeULL,0xfffefffff6f3ffeeULL,0x89f6bf3674127eacULL,0xeed33ed37673fdecULL,0xbdbfedb3f4b13eeeULL,0xf7d7fefae4f3feeeULL,0x5faa7d60e67272ecULL,0xf9dcfeeef4b37feeULL,0xcedbbffbf4d3f7eeULL,0xbeff5ff3f6f3f6eeULL,0xfef7ffef769176e8ULL,0xcfaafde2541336ecULL,0xff1cc37e867277ecULL,0xdd5bffd5b43356eeULL,0xfa3ecf572613f6ecULL,0xe6fef7e456937fecULL,0xfd637772a4b35eecULL,0xff1077f524b3f6ecULL,0xefffffbbb4d37eeeULL,0xabdafdd3a6f276ecULL,0x95bf5fe3d433e7ecULL,0xd8a6a778b473f22cULL,0x7ef3af7a74b2f6ecULL,0xcbdafbfb7473f6eeULL,0x229ecbc36492776eULL,0x87aa5cb48450f2ecULL,0xfdfff77f64f2f7ecULL,0xffbe7ffcf673f7ecULL,0xd9fafb3be633fdecULL,0xafd71e57147277ecULL,0x7ba8affb6470dcecULL,0xc5f3f779749376ecULL,0x93fe7f7134f276eeULL,0xbb96cb1724537eeeULL,0x35936a3625154e0ULL,0x27f3cd007673f6e8ULL,0xfdff7ffff4f3feaeULL,0xdb2ffe50e0b376eeULL,0x4000240000000280ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xccc3ba34b4b270acULL,0x7fffffdf76337fecULL,0xbfbaffdbe4f27aecULL,0xf5fdf66bb4f3f6eeULL,0xff9eff7c3493f6ecULL,0xc9ffffe7f4f1f6eeULL,0x4eff3e2fb6f356ecULL,0x37fcdfb2f6d2f7e4ULL,0xcdffbebe74f27aeeULL,0xca0c6d7034f176e0ULL,0x79cce33ed69317e4ULL,0xebfbfffbb4737eeeULL,0xafefcf796632ffeeULL,0xd5fff6f9f6b3ffe8ULL,0xbf3eef73e61356eeULL,0xfbf7ff7ff633f6ecULL,0x7f5fffbef4b3f6eeULL,0xedffebd2b6b37fe6ULL,0x267005a000102284ULL,0x0,0x0,0x0,0xffbfffdff6f3ffeeULL,0xfffffffff6f3ffeeULL,0xfffffffff6f3ffeeULL,0xfffe7ffff6f3ffeeULL,0xefffff3fe6b3ffeeULL,0x4ff6be8944b3f2acULL,0x6b2c4c560114468ULL,0xd2bf0d6804314262ULL,0xa430b65d243354e8ULL,0xcbdff7fe66f2f6ecULL,0xffdf3f7ff6f1fee8ULL,0xfb9bdfffe4f37eeeULL,0xfd7dee5ed6f376eeULL,0xf7be7c7be6d3f66cULL,0xfb97ef6e66b3ffecULL,0xaffa3eea7432f6ecULL,0xbfbad3f9f61231ecULL,0x55ffd6e836b2d2ecULL,0x93603540443214a4ULL,0xca886d08003141ecULL,0xdcce3f1d8013f6ecULL,0xffdedbd0f6737ee8ULL,0xb6f4fe6006b31ee4ULL,0xbf6ef760845336e6ULL,0xd314c6e8d4627ee8ULL,0x4947431882101668ULL,0x207044000905000ULL,0x19245820010d460ULL,0x486d25213432306cULL,0xffbc1e5a3452f4ecULL,0xc9d6d3f68493f8e4ULL,0x5bfefade941276ecULL,0xb4addffbd47336e8ULL,0xab5cb778665276eeULL,0xaeb63dfd5671f7e0ULL,0x80743b6a8000d568ULL,0x486e5bd8e010d3e8ULL,0x809a0652a4323228ULL,0xc012a25024306668ULL,0x220290000101462ULL,0xab5c431234915aa4ULL,0xe42321b2012e4a4ULL,0x63e5646e613f668ULL,0x41feb1a8003150ecULL,0x4b2e7f82a47372a8ULL,0x5a226462805052c4ULL,0xf8e2e2a8612e4c4ULL,0x6bfe9fd326d1f7ecULL,0xf0310464062276c4ULL,0x41280d2006325044ULL,0x38610450230e040ULL,0x6c77e1a7432f060ULL,0xfa44ded20492302cULL,0x85c017ab22923468ULL,0x5a448420a0326260ULL,0x48bc0570e082142cULL,0x8020000000000280ULL,0x8c480408012c080ULL,0xc100520000006c4ULL,0x4071a40500026684ULL,0xfa186d97147356eeULL,0x885832010002e6c0ULL,0x4e80851148012e0ULL,0x800004000801000ULL,0x100201800010060ULL,0x40100020200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202000104000ULL,0x0,0x60000010020ULL,0x3000780c0102020ULL,0x0,0x539213c0569254e8ULL,0xc9a1832f2492556eULL,0x405b1b57a49076ecULL,0x498880140015020ULL,0x90729700262272e4ULL,0x120064000332000ULL,0xcbeef36ec25254ecULL,0x480201c0008016c0ULL,0x12e0c5240905460ULL,0x84b6124036005460ULL,0x931224402602106cULL,0xca0c20221411d0a4ULL,0x480bfa30e03326e4ULL,0x29332f5914b072e2ULL,0x1080a640040002acULL,0xc72da740209372ecULL,0xc380332680132684ULL,0x41c3fc1060133668ULL,0x14020000020ULL,0x1000800340220020ULL,0x8100300020304020ULL,0x0,0x1800020000020ULL,0x4080000001400ULL,0x0,0xb40dac04a01368e8ULL,0x8800450000114440ULL,0x0,0x2585000103280ULL,0x111165c2c2000360ULL,0x1441000080221000ULL,0x4800000440ULL,0x52000002200000a4ULL,0xee8212ca047372acULL,0x1a2447e140323160ULL,0x54f04dcca09052a8ULL,0xd7136550e47076e8ULL,0xfe84656c241277ecULL,0xc7f7fab9f4f377ecULL,0x33d2dfd124927eecULL,0xf4f2fefdd673f6ecULL,0x99aeb77366737ce4ULL,0x8b80e6c24222aecULL,0x4cc81200001242e0ULL,0xa000000000102000ULL,0x80146006001444ULL,0x30000000000440ULL,0x12080e0000001408ULL,0x4fcffe2ba493546cULL,0xf2d6fec9461157ecULL,0x80bb4c22964236ecULL,0x8400247000911680ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40200000801000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 },
	{ 0xfb6a7f62592c95eaULL,0x8010000000000140ULL,0x2b6b91dff9bcbedfULL,0xfffffffbf9bebffeULL,0xfdfffffef9bebfdfULL,0xfffefff7f9bebfdbULL,0xfbbfff7ef93e3fdeULL,0xffffffffd9bebfffULL,0xefffdfdfd9bcbfdfULL,0xfcfffffff1beb7ffULL,0xfffffffbf9bebfffULL,0xffbe7ffff9bebfffULL,0xefffff3fd9bcb7daULL,0xafffffdff9bebfffULL,0xffffffef79bcb7dfULL,0xfffffff7d1bebfdfULL,0xfefe7fffd9bebfdfULL,0xfdfff7fef99eb7fbULL,0x29dffeff711cbfd2ULL,0xdefb7bf779bc9fdaULL,0xffffbf7ff1bebffeULL,0xdbee7f6651be9fdbULL,0xe9f7ee6ef1bc96cbULL,0x5c931e57d094bfd0ULL,0xbebbcff2519eafc9ULL,0xf575feee513eaefaULL,0xf7be2d7971b41fdbULL,0xfbdff77fd93ebfeeULL,0xaddbfebb71343f77ULL,0xa2bdcffbd91cbedbULL,0xd9d3366b5936befaULL,0x892ea553d11c329aULL,0xdb06fb5f513ca7fbULL,0xcd6aba1ff12c91caULL,0xebbe9ddbd13c9ed8ULL,0x34f4f6aad936becbULL,0xc39a3770d936346aULL,0xfb9e6ff1d19cb65aULL,0xe8627b38d1343546ULL,0xaaffffdbf1bcb7daULL,0x87beed43519c15eeULL,0xfb6fff6f59bcbedaULL,0xfeb63fe7f10cb7ceULL,0xef7dfa7f591497feULL,0x9eeebe941b6b7caULL,0xd3becdbb799cb7d8ULL,0xb1dd57ec719e9ddeULL,0xfffe7dfdd19cbfdeULL,0xaba6fb79d114b5dbULL,0x2ff71ed3790eb5cfULL,0xbbbeeffbf1bcb7dbULL,0xd6bbfef8713eb7daULL,0xdffefff9713cb7dbULL,0xcbd5fe4ed01c9dcdULL,0xafcbb2bf5134bdcaULL,0x7fbe5fd851bebffaULL,0xb4bf3feed03c9edfULL,0x0,0x200000200008040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000001000002ULL,0xefffffff79bebedeULL,0xcfdf7fffd13cbfdaULL,0xbfeefff3f1bcbfc7ULL,0xf7ffffebd9b6bffbULL,0xfffeffeff93cbfdfULL,0xe9f6f6bed834bfd8ULL,0xebd77ec3d1bcbe8fULL,0xefbceffdf1bcbffbULL,0xfdf4effee11ebddbULL,0xcefeed7a59b6b7deULL,0xf97ccb3e51149cfaULL,0xdffbfebbf1bcb6eeULL,0xfffdfff593e97fbULL,0xfdf5f6ffd11cb7daULL,0xdbbaaf75f13cbecaULL,0xdb8eefb7d936b6faULL,0xdfeffe3b71b697ceULL,0xa9ffcfde799ebbfaULL,0x0,0x0,0x0,0x0,0xdfbdffdff9bebfffULL,0xfffffffff9bebfffULL,0xfffffffff9bebfffULL,0xffbe7ffff9bebfffULL,0xfffffffff9bebfffULL,0x2af6bf7f519ca6faULL,0xfffe6f6af1aeb7eaULL,0xd3ddb67951be32eaULL,0xef5e7f67f1b4b5dbULL,0xcbd87306d19ebccbULL,0xdcd6060fd1949ddeULL,0xb9b3cfe15194be0eULL,0x307f77cd509c91daULL,0x7ebe7debd91c9fdeULL,0x9314ee0c01142ccaULL,0xc48200910108490ULL,0xa11616d15114a078ULL,0xf1c4bee9d82092d2ULL,0x3d747f4351148fd3ULL,0xcb29e02b5014a3d8ULL,0xcefefebe71bca2daULL,0xefbfdfdb71beb7daULL,0x24744dc0510cb482ULL,0x5ba6ac3c5030b7c1ULL,0x7fb2567fd01896ceULL,0xc9e7b6aff194b7deULL,0x8255404a110c06e4ULL,0xb0e75dc451163cd6ULL,0xa1fc257a511c2af8ULL,0xd096276fd10435cbULL,0x8007fb63d1149e87ULL,0xaf363cdfd09c87c7ULL,0xb6bbf77cf91cb7d2ULL,0xd16036ec713c96c2ULL,0xcbb237e4f11c11faULL,0x900d33a400426caULL,0x6e0188820048240ULL,0x8f88fffa988c17caULL,0xc153a27a5000a2c0ULL,0x4552fe517034b7d0ULL,0xcb83eb5b5910bcc5ULL,0x5e4b925f71048178ULL,0x3084000100150ULL,0x7464f66840bc9dd2ULL,0xe162fe0419c00daULL,0xc892742251a41148ULL,0x6d33e2a51168ceaULL,0x28106ac240941842ULL,0xb4bfb0e75134b87aULL,0x91aea5205084b088ULL,0x53367267d12435e2ULL,0x8e83e30a911699c8ULL,0xee465e98511ca81aULL,0x85a2022f910696c2ULL,0xf504260851101e80ULL,0x4400000240040000ULL,0x81908001300480c0ULL,0x260004858101200ULL,0x20000428000c0240ULL,0x2272c000e10c8ULL,0x80044110000480ULL,0xca88d32af81c30c2ULL,0x8851840400c1088ULL,0x9080b1051000082ULL,0x8440144019009140ULL,0x0,0x6011000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41000000480ULL,0x200000210000400ULL,0x0,0x200020000001002ULL,0x21020640100005c4ULL,0x5b96a761d00e17caULL,0x830ed36e5194b0d2ULL,0x4050123e5110ac56ULL,0x880810800248048ULL,0xd5d4f62cd13484daULL,0x682084001202342ULL,0xd100b70c081c9a42ULL,0x4ee39804d90434c2ULL,0x622201425104844cULL,0x46c1744d114068aULL,0x4b963832d01492c0ULL,0xcf14d3f851148ec2ULL,0xc8cefa2170341fdaULL,0x94e815ad984109aULL,0xba426060140358ULL,0x8128282191163048ULL,0x1220032040022240ULL,0x29c0c32b580c96d8ULL,0x200780180001142ULL,0x300071250008490ULL,0x90882548d1049048ULL,0x8000020040000080ULL,0x8402200c0001080ULL,0x808005440003088ULL,0x104c8000041082ULL,0x4120853161048663ULL,0x90102d4040248058ULL,0x3808000100c0400ULL,0x265d080040860140ULL,0x1048c2100005c0ULL,0x8b6866891949058ULL,0x80440200100040cULL,0x0,0x494e93d041949252ULL,0x1814ed40500430c2ULL,0x44d1c2c471949748ULL,0xfb28f562d0b6b7c6ULL,0xc3b2420e511ea648ULL,0x8c4bda19d136bcf2ULL,0xabfe87da519cbfdaULL,0xfcb61767f9bcb65aULL,0xff2ff76dd1beba5aULL,0xf694bfd6719eb6ceULL,0x4080c00040003080ULL,0x141901050208050ULL,0x1000010000000140ULL,0x40210009040ULL,0x9a0c2922591cb40aULL,0x45d2bc8a508636d2ULL,0x6f5ef759513cbedbULL,0x171a55f1511c93faULL,0x40140191040008ULL,0x18000002040ULL,0x0,0x0,0x82000000100800ULL,0x0,0x0,0x80800000100800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000c0000ULL,0x0,0x0 },
	{ 0xffef7f63d0e9bf6aULL,0x60000000400ULL,0x408118100010140ULL,0xffbfdffbf8ebfffcULL,0xf7fff6fff8ebfffeULL,0xfffe7f77d8fbfffcULL,0xffbdfffff8f9fffeULL,0xffffffbff8fbfffeULL,0xffffdfdfd8ebbffeULL,0xf4fffffff8f9fffeULL,0xffefff7ff8f9bffeULL,0xffbfff7fd8e9bffeULL,0xfdffffaff8f9bfe6ULL,0xefffdfdfd8fbfff6ULL,0xfcbf7ffff8fbfff6ULL,0xdb7dfffaf8f9ff76ULL,0xffbe7fffd8fbeff2ULL,0xefffffbf98ebbf76ULL,0xcfdf7f7fd8f9eff0ULL,0xbff3dffff8f9bf26ULL,0xf7ffbffab8fbbf72ULL,0xbfb67f6ed8e1af56ULL,0xfbf7ffbbd8e1ff76ULL,0x7fd7ba39d8e9af6eULL,0x3cc95eda98412f74ULL,0xf5f6f7f198eaaf70ULL,0xfffe7de3d8fbaf7eULL,0xfbf7e36ff8e9bfd6ULL,0xcc7bf7ab98e1bff6ULL,0xefffcf7ff8ebaffaULL,0x74f77ffdd8e12f5cULL,0xef9aaf63f861b7e6ULL,0xeba24b8e8849e9c6ULL,0xdffabe13d8f1ef76ULL,0xaffbdedbd8e8ef76ULL,0x74a854c498e2af00ULL,0xf5feff7198e1ef74ULL,0xfbb6f766d8b0fd46ULL,0xefcfff1bd8fbfff2ULL,0x2e7e1fd5f8e0ef70ULL,0xcc977fa998f8fb30ULL,0xa51a5f6290e04720ULL,0xfeb63dded8fbe79aULL,0xc7adf73ed8fbaff4ULL,0xcfffbfcd8c9a704ULL,0xbfff7ffff8e1e77cULL,0xf56fff6bd0c1bf0cULL,0xdffe4f73b852bff2ULL,0x884dc36c98d22718ULL,0x27ff3e5ab8d9bb42ULL,0x8b9eefd3d8c9af7cULL,0xd1fef6edf0e9bf72ULL,0xf2bc6d5398f0dd20ULL,0xf9d4effb98e9ef70ULL,0xeeee7bceb8e96f5aULL,0x1004000000208ULL,0x0,0x0,0x12000002004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb8106f2290702760ULL,0xdffeffbff8fbffdaULL,0xefdfffdfd8f9af7aULL,0xd7bbffffb8fb6b7eULL,0xfdef777ad8abfff2ULL,0x5fae6febb8f9b7f6ULL,0xebf7ff2ff8f1ff76ULL,0x69f2be1be8f1ef46ULL,0x8fb9efdcf8d9eb72ULL,0x9517b6f8f8e9ae70ULL,0xcbbeff72f0693f74ULL,0xfbf7ff7eb8e0f772ULL,0xcec3ffcfd8e0e562ULL,0xdfbeeffeb8fbbf7aULL,0xd5f73feed8f8b77eULL,0xdfbfffe3d8fbbf72ULL,0xfbbcfffef8f9bf7eULL,0x9fd7ff1bd8e9af76ULL,0xa25880dab0e96752ULL,0x0,0x0,0x0,0x0,0x9fa9ffdbf8fbfffeULL,0xfffffffff8fbfffeULL,0xffffff7ff8f9fffeULL,0xffbffffff8fbfffeULL,0xf7f7ffbfd0fbaffeULL,0xe61ffafbd079af70ULL,0xb7fe4ffbd0fbaffeULL,0x91ed756ab840be74ULL,0x7bd657f4d8e1a772ULL,0x699befead0e1bf78ULL,0xb7577ffe90f3bf36ULL,0x2c9eefe398e1af52ULL,0xdbf3d67a90e2af28ULL,0x41fe2dd898e12720ULL,0xfa96cb2e98babd62ULL,0x4dd3fe4c98cb2f70ULL,0x8ab44c70d0e2bf00ULL,0x1c7337e190baade6ULL,0x93b1bd7090f1bd84ULL,0xd384cb3cd8c1a500ULL,0x43de921f90e16774ULL,0xffffffdff8f9fffeULL,0xf7f67fe3d8a0af56ULL,0xcd8a75439001d9e0ULL,0xd3107e3cc029bf6aULL,0xa986e7b7d8402f66ULL,0x67ffd7d8d8ebffd6ULL,0xbff33ff3d8e1ff70ULL,0xf5f48ed0b8c8ef60ULL,0xc2843971c0802b70ULL,0xa5e6d7129820ad70ULL,0x4a3feb4b98a0bf78ULL,0x969f5dba9809e350ULL,0xf0f5befdf8f1f7dcULL,0xd192185098408350ULL,0xa9d4f36e90a8677eULL,0x4864025010000510ULL,0xa0b2042010226706ULL,0xc041a8af88002d20ULL,0x2063120c0006500ULL,0x4bc4e10a3080af20ULL,0x164a3218a8e90330ULL,0x8182c81098009840ULL,0x51a665eed089f752ULL,0x95122c61d063af60ULL,0x8dd0ef6a50a0ef98ULL,0x85e2ca38b0c08e70ULL,0x2f181ed010e1af70ULL,0x7ab4bcd090e3bd50ULL,0xd0acf76098eaefa0ULL,0xcb9637d6b8fbab78ULL,0xc966fba398a0bd40ULL,0x55e065490f0bf00ULL,0x563f672b90623760ULL,0x10ac206290422214ULL,0x808043cb0900a20ULL,0x21c4803090410150ULL,0x2c1e5141d8002214ULL,0x420080000800ULL,0x861850002414ULL,0x5800080200202604ULL,0x8300a10b90212050ULL,0x608000010009100ULL,0x8002000010800000ULL,0x420200000082308ULL,0x0,0x0,0x1104880010800ULL,0x0,0x10100090000000ULL,0x0,0x0,0x40001008400000ULL,0x0,0x2020000000002400ULL,0x8000004080000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48000400200ULL,0xc088a00010882208ULL,0x8a766df198e3a7d4ULL,0x8284e2849049a3b8ULL,0x48c0921098400600ULL,0x12480a8821a040ULL,0xdc30b26498202c30ULL,0xc90ea770b880bf78ULL,0xd2b5e90e8068bf68ULL,0x8d4abfab9060ff22ULL,0x220c74808602204ULL,0x1040040000608200ULL,0x4008af6090c22500ULL,0xd1086d8690492c50ULL,0x8122f28290b02630ULL,0xa289155a1021f560ULL,0x22500000018140ULL,0x8000040000000400ULL,0x4980203990488310ULL,0x82907a03c0c02706ULL,0x2100049300412644ULL,0x11000000202000ULL,0x8066074090a02c04ULL,0x948409e4a8608760ULL,0x808c22008002400ULL,0x401004810002410ULL,0x8400000208ULL,0xd1a8727090a02600ULL,0x1408084080080260ULL,0x200200008080ULL,0x60004c8020200340ULL,0x8094c00230089290ULL,0x400301080002002ULL,0x440224489088a314ULL,0x800656000412244ULL,0xe96a120ff0418ff0ULL,0x12014718904b2648ULL,0xb050126498402c04ULL,0xb35a744110e93762ULL,0xd8822e6590c02c10ULL,0xc6d3be19f8e19f50ULL,0x2b3e87c7b0ea2b5aULL,0xb67474eaf8e2ab56ULL,0x96ef7564b0eaab56ULL,0xfbb02f7a9862b2e4ULL,0x10c600290820010ULL,0x0,0x4822054090c02202ULL,0xa00000010040ULL,0x482006aea0612a42ULL,0x4d8a7737f843a314ULL,0x1cafe3ab820aadaULL,0x66607e2b0e92250ULL,0x9000500090809800ULL,0x0,0x0,0x0,0x401090000000ULL,0x0,0x0,0x0,0x100080000000100ULL,0x80800008000400ULL,0x0,0x18000000210ULL,0x0,0x0,0x0,0x0,0x0,0x200000200002400ULL,0x100800000000100ULL,0x108001140002120ULL,0x4012006a8020a400ULL,0x11200c0208202400ULL,0x22d5000a04500ULL,0xa250430608000520ULL,0x24c4080086085a0ULL,0x22000a0000a500ULL,0x404000000000100ULL,0x440004000000a480ULL },
	{ 0xffeef7f9fffefd63ULL,0xbbd67b6dedee7c63ULL,0x1b9b6b324beeb560ULL,0xffffdffffffeff63ULL,0xfdffffeffffeff73ULL,0xfffffff3dffe7f77ULL,0xfffffffffffe7f77ULL,0xcffffffffffeff73ULL,0xffffdfdffbfc7f77ULL,0xfeffffffffeeff77ULL,0xfbfeff7a7f7eff77ULL,0xbbff7f7dcffc7f63ULL,0xcffffffefbfeff63ULL,0xefffffdffffeff77ULL,0xfffedfebbfee7f67ULL,0xfdf8ef7eff7cf761ULL,0xfea62ff7df6cff66ULL,0xefdbffbf9feeff26ULL,0xeddbdfff8f7ef772ULL,0x86aa5ed38b3cef65ULL,0xfdfffffc7f7eff67ULL,0xfebe7ff66d6a7f66ULL,0xcb53db3e4ffcff72ULL,0xeff7ffdbdb7af763ULL,0x9f3edff3fb6e6d62ULL,0xb374df7c0f7c7f62ULL,0x4bbe5fe36f6e7763ULL,0xdb56cb5ec3e8f743ULL,0xeccaf6feedaefe23ULL,0xef1eccd39b4e7f62ULL,0xf4fffffefffeff73ULL,0xfffefff2cfee6f73ULL,0xff8cfe7ddb3e7d73ULL,0xcffff93f5fec7f63ULL,0xa3de5756db6c7742ULL,0xbeffffeefffe7f73ULL,0xf96a6f41cb48fd72ULL,0xfebd7f66efae7f66ULL,0xcfc7ff3acf3eef60ULL,0xa6decedf4becff67ULL,0xfcfaffe9cbeeed62ULL,0xf96fff7bba68fd73ULL,0xfe1e37ff5ffcf763ULL,0xefdff737efeeef73ULL,0xafdfffeb3bee6f77ULL,0x36f26fead93c7746ULL,0xfe7dbf7fdfee7766ULL,0xef9e3decef6e7567ULL,0xc9fdfbf64a0c7f67ULL,0x8ed67ad94fbc7d60ULL,0x9fdf4df3796c6f63ULL,0xd7c2e67b8ffe7553ULL,0xeffef7f04b6c7f63ULL,0x8bd7cf7fdbecef62ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffdeffedb6e7763ULL,0xfffefffffffeff77ULL,0xff7fffefff6cfd70ULL,0xa70cd7791ffc6572ULL,0xfffdfe7bbafcf563ULL,0xffbe7d7ddffeff73ULL,0xcbf6ffafdfeeff77ULL,0xcfcf7f57ebecf573ULL,0x9f5f7f8abffa6f22ULL,0x9553fef1af3c6d61ULL,0xefbe3ff3ef7e7f63ULL,0xdb36fb63fbeeff77ULL,0xaccb3eb5ef6e7561ULL,0x3b9b4eea9fec7746ULL,0xfcf7bfe9db7e6f67ULL,0xffeed734bfc7f76ULL,0xff94eff77fae7773ULL,0xddfffebfef7eff77ULL,0x8000000200202000ULL,0x0,0x0,0x0,0x0,0x8389df5afffeff67ULL,0xfffffffffffeff77ULL,0xfbfffffffffeff77ULL,0xffbf3fffdffeff77ULL,0xeffffffffffe7f77ULL,0x624efac789fefd60ULL,0x174b7f7f4a7cff63ULL,0xf18b776b4b6ef567ULL,0xc2be6ded10acf475ULL,0xcbcdf33edf78f763ULL,0xcef737aecb4cf562ULL,0xc49fd5f65b487562ULL,0x9a6dd7f8eb88f720ULL,0x7ebe7df2ff7cfd73ULL,0xfa6cd92ae87c6c72ULL,0xcf53be76cf48ff62ULL,0xbf2def792ffe7d63ULL,0xb0365eea0b4c7c62ULL,0xa33eb5a00b666c61ULL,0xdb85b63c883cf420ULL,0x4dc7bb3aedee7e20ULL,0xffffffdfdffeff67ULL,0xb6efc7cc2d6efe63ULL,0xe44410a30e400ULL,0x461226380708e501ULL,0xc8f3f87ac2087561ULL,0x69d85f910f0c6405ULL,0xb4be5fefcf6e7627ULL,0xf9efb7dedffcf766ULL,0x560c20e1833c7522ULL,0xcddb9f1817746746ULL,0xcfffefdfdffee766ULL,0x1fa5dffa9ffce666ULL,0xfdfcfffadfee7f67ULL,0x8be645019ec4564ULL,0xe8d5f37793bcf777ULL,0xa9c39e7a082c4f60ULL,0xd09b8d72cd4e7e61ULL,0x96d476f96d4e7e73ULL,0xc93e2de807bc2c71ULL,0xfbddef6f573e7747ULL,0x200301008084022ULL,0x93b1fd3df60fc43ULL,0x112480640d604c60ULL,0x8eb2b5e393bc6664ULL,0xd3066cf9aa2c7474ULL,0x81a852110d642420ULL,0x8d1c5f580b8e4c62ULL,0x70b552e40b266442ULL,0x492ecb601bfe4e64ULL,0xc28064c020b84564ULL,0xcefefffafbfef726ULL,0xcefcd45348aced62ULL,0xa6bef7a9cafced63ULL,0xda7cb78643fe6162ULL,0xca84041803220d60ULL,0x2004000000000140ULL,0x120000040020ULL,0x0,0x5404208080005610ULL,0x2050404008084062ULL,0x0,0x8002114205001200ULL,0x820000000000410ULL,0x650410280a404400ULL,0x0,0x0,0x0,0x0,0x0,0x4000200002000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4100000000a000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x22000000004001ULL,0x0,0x553e3ef43bcc4e63ULL,0x89853703206c6561ULL,0xe4832a5009441436ULL,0x2020014000007002ULL,0x91f9227a4848ef65ULL,0x28544478390c4c22ULL,0xf3d3fb3e4fbe7f23ULL,0x49e2eb130b226d60ULL,0xb045554083c7440ULL,0x6402000a400100ULL,0x12020c4041042820ULL,0xa10050002084422ULL,0xcad39a293b6ced63ULL,0x25d4f5803465d60ULL,0x2080402000002060ULL,0xa88cfb709b747c00ULL,0x8102142000004140ULL,0x212d001022400ULL,0x50401000402060ULL,0x2810000400ULL,0x9000000000240000ULL,0x400a044209000604ULL,0xec92801744860ULL,0x4044160002cc4420ULL,0x0,0xc2751540438aa542ULL,0x41800a0042006101ULL,0x85128403022c20ULL,0x400806791f402500ULL,0x100400000084000ULL,0x4200805002206440ULL,0x186746081ba0d60ULL,0x8000020402000400ULL,0x45db100189886d60ULL,0x1e78cf120fec7f62ULL,0xf5d097ce9f4c7560ULL,0x900020000a602540ULL,0x0,0x400000808400000ULL,0x82000002000100ULL,0x3020400080404400ULL,0x201008400000ULL,0x0,0x0,0x0,0x4000010000080020ULL,0x9000000002000400ULL,0x10001000000410ULL,0x0,0xc000000000400ULL,0x8050008400820ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000000084000ULL,0x0,0x1800000080100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4d00484202402600ULL,0x210243000884060ULL,0x84836e1087a8e503ULL,0xa3dd58d00708ed70ULL,0x4dc562bcf426407ULL,0xbb6fdf604b7e7d62ULL,0x709e1b304b6e7563ULL,0xc185b1aa4d4a7f66ULL,0x4a154b4057247460ULL,0x801267f3906cfd20ULL,0x30000000002000ULL,0x1a0c072409286622ULL },
	{ 0x382421800221460ULL,0x82061a400e003440ULL,0x2deaede2ab922d48ULL,0xffffeff3ffffff68ULL,0xf7ffffffffffff68ULL,0xdffefff7ffffff68ULL,0xffffffffffffff68ULL,0xefffffffdffffd68ULL,0xffffdfdfffff7f68ULL,0xfeffffffffff7f68ULL,0xffeefffbdfbe7f68ULL,0xffff7f7ff7feff40ULL,0xffffffbfffffff60ULL,0xefffffdfffffff68ULL,0xf7ffffffdfffef68ULL,0xfbfffffcfffffd68ULL,0xfab63ff3ffff7d60ULL,0xeffff69ff7f6ff68ULL,0xbbff7efbdfbe7f60ULL,0x3db14ffa4fb62d60ULL,0xfffffffad7f6fd60ULL,0xfbbe1d610bbe2d60ULL,0xabdafb27d7ca4f48ULL,0x7edfbffeffbf4f68ULL,0x9df6fffbdeb7fe60ULL,0x7ceef6f54fbeed60ULL,0xfbbe7d62fff67d48ULL,0xd936ff360fbf3d60ULL,0xf6e79e4ac39a6f60ULL,0x6faecb7bbfbecf60ULL,0xfdfffffefff6ff68ULL,0xffffff775ffeff68ULL,0xdf9fffff7efe6f60ULL,0x9fe6fab397be0460ULL,0xbff2d558579e6560ULL,0xfcf7fefeffff7f60ULL,0xffffff7ffffeff68ULL,0xdfffff679ff67f60ULL,0xe447ee1a5fb67768ULL,0xfbfdffd1ce9ffd68ULL,0x86dbfeff5fbe7f60ULL,0xf56fffff9fffff68ULL,0xde9caf7e86b71f60ULL,0xcfdaf7b95f96ed60ULL,0x791effdf8ebe5d60ULL,0xffffdffbffbefd60ULL,0xdff9bd6576aa7c60ULL,0xfbb63ffecfdfff68ULL,0xebfffbff77be7f60ULL,0x6e9f9f4c668b3c60ULL,0xf38ec57287f69f60ULL,0xded5f074d59e6f40ULL,0xfffefff3ffffff68ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffbebff6fffe7f68ULL,0xeffffbffffbfff68ULL,0xbfffbfc9ffbefd68ULL,0xb799cffb6fb27f60ULL,0xd1cdf678ff9af760ULL,0x5deeff575f96e760ULL,0xebb7ff7ebeeef760ULL,0xefc3fedf37b62768ULL,0x2798fdfa07d67d48ULL,0x9ff3f6f077be5f40ULL,0xff8ebe7217b6cd60ULL,0xdbfdff2755b76d68ULL,0xeffff207cfffbf68ULL,0x2fbfcdd3c79e2f60ULL,0xf4f5ef6ddefe3d60ULL,0xdf9bef559f97ff60ULL,0xff90f37fcfbf9f40ULL,0xcfdfffbf9ffe7f68ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xffffffffffbfff68ULL,0xffffffffffbfff68ULL,0xdeffbffeffbfff68ULL,0xfbf7f3fddfb6ff68ULL,0xf5fffcbb5b76d60ULL,0x36fbeff997bf6f60ULL,0xd4ffbe7ae7b77d40ULL,0xdd166f77d2b27f60ULL,0xebbffffa0fbfbd60ULL,0x67f63eb3daba8d60ULL,0x30becdea5f9e7f40ULL,0xb6b87efed69aed48ULL,0x7bbe7dfa9ef76f60ULL,0xfaf9fb2effb6bd40ULL,0x6f4ba27947d8ed40ULL,0xcd9ddf521f12bd40ULL,0x54e5356fc39e7e00ULL,0xdfba7d45d6b6ed60ULL,0xc9f2f77757ff7d68ULL,0xcde2e50e07b99d60ULL,0xaf5ace5776967d60ULL,0xf5b7feea1ebeaf40ULL,0xd73a2f397f96a400ULL,0x96b0b67842b62500ULL,0x41890800922120ULL,0x3c65b5905b61d60ULL,0xa01a44a11fb28d68ULL,0x5bec77731ff64f60ULL,0xfc1c2d5917967e40ULL,0x836445e13a902440ULL,0x2bef99da17b67d60ULL,0xc7ac52307b0e500ULL,0xd77e2f60379e3d60ULL,0x380060000200440ULL,0x82b22b4b12923920ULL,0x25511ffa728cbd40ULL,0x9fdfedd35f9aff60ULL,0x80227253140a1568ULL,0x4286190080925b00ULL,0xe204f1030a129540ULL,0x4852161e43a96c40ULL,0x8010191203a08640ULL,0x9c77d1c67aba6d40ULL,0xc49f2524f6986d40ULL,0xefa4b6651fbf7d40ULL,0x1156a6121f906f20ULL,0xd91fcf4315967e60ULL,0xa8741ee244126c20ULL,0x1578846840d26000ULL,0xca00140652d02500ULL,0xc492cb00063e4d60ULL,0x40909202d06000ULL,0x300627e00b124000ULL,0xfb2e77229bb7ed48ULL,0x8a938368161e2d00ULL,0x5800c51a14d71f00ULL,0x8e60808018a5600ULL,0x8010101800200ULL,0xbc201014930820ULL,0x820440008800600ULL,0xc20001800000ULL,0x10040000010800ULL,0x100007000802420ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x842070004944140ULL,0x0,0x0,0x0,0x0,0x0,0x40200000100000ULL,0x4ad7d36a131a2d28ULL,0x4fd23e00ce123d20ULL,0x90100c6083160400ULL,0x156ec6d08b852d00ULL,0x4edc7d6b93b2ef60ULL,0x920cc32613960d00ULL,0x40c393de06b67d60ULL,0x2ffedb5b4e36fd60ULL,0x8050100002082400ULL,0x2010040002000ULL,0x8284016442827440ULL,0x844a721102862d40ULL,0xb7a8fd786126d60ULL,0xb4f5f6e83fff2d60ULL,0xc072444354942700ULL,0x1080254000d23840ULL,0x6e4126924e869500ULL,0x2118c84a0ac61200ULL,0x10206000000560ULL,0x9010000000000900ULL,0x8a080a0a8a020440ULL,0x80a2a80001a24c40ULL,0x23611c4002c40ULL,0x480210000100900ULL,0x8440870004063d00ULL,0x528024041a900900ULL,0xe000e08106b82540ULL,0xe4230102c124540ULL,0x8442401016100500ULL,0x9188062a02021c28ULL,0x40a2055011960900ULL,0x8000a70010120540ULL,0xeca120b1fb83540ULL,0xbecf701e986760ULL,0xb07796fcff9e6f20ULL,0x928aaf11d68cad60ULL,0x0,0x8000100010800000ULL,0x0,0x140000120000ULL,0x9210810000860640ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100001001000400ULL,0x0,0x3000000000020800ULL,0x0,0x0,0x0,0x21000200040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81040402a000400ULL,0x20423420120a0500ULL,0xca206f3053961560ULL,0xef86da0a53a22f20ULL,0x8bbf9c0561e6d40ULL,0x1c9f17e352b37d40ULL,0xac1c266302be3f40ULL,0x40261de052a68540ULL,0xebf4f312179f3d40ULL,0xe8031ec204b84f40ULL,0xb4b3667add947f40ULL,0xd33227191a925560ULL,0x4d0e176286d22560ULL },
	{ 0x8ad43026d6006245ULL,0x675e41b2042df4ULL,0x1b3e4d1bc78c276cULL,0xfffffffed79e7fefULL,0xfffffffff7be7fffULL,0xfffffff3f7be7fffULL,0xffbffffff7be7fffULL,0xefffffffd6be7fffULL,0xfffedf5bf7be3dffULL,0xfefffefff7b67fffULL,0xfffffffbd63e6ffdULL,0xffbe7f7ad32e775eULL,0xefffeffed7be7befULL,0xefffffdfd7be7fffULL,0x97ffdffbf63e7ff7ULL,0xfd7fff7ed73e6ff7ULL,0xffbf3ff7d7be7fffULL,0xe9dff71fd7bc6fffULL,0x2976fed1d62e72ddULL,0xf6f37f7ff7967ffdULL,0xf7f5ffffd61e77efULL,0xfffeff7dd7be7ffcULL,0xebf7e73fd3b46ffdULL,0xe8d75fbed6a67aedULL,0xb6bd1ef2d68c23deULL,0xd5f6fe7fd604634fULL,0x7bbefd31d40c7afdULL,0xbb1ee279f4346147ULL,0x5dd73f5fd2ac2ddfULL,0xefbe5fd0f6ae63deULL,0xfdf7feeed63c7bf7ULL,0xdbfb7e60f716652eULL,0xdabcfe7bf6363ddfULL,0xdec3febf931661beULL,0xb3d4dfdbd79e71ffULL,0xc3fff7e9f7ae7df7ULL,0xd3feffd8d60c2dedULL,0xffbebfe6d68e23efULL,0xe9d7ff3ed69633d4ULL,0x6ff6dddad6346bdfULL,0x3db3dff3c62e6dddULL,0xdffcff53d61e77fdULL,0xdab2af6dd73e635eULL,0x89f3fe37d61e297fULL,0xb7ededed61665f4ULL,0x37abef53d69439cfULL,0xb3efa76a938c3de5ULL,0xfbe67fe5d28e2975ULL,0x8bb6fdead49c2261ULL,0xfcd7ae08d68c697eULL,0x3d812540d4202328ULL,0xd5f3f77bd6be2fe6ULL,0x8084c21440c0020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xffbf7ffff7be6fffULL,0xefeff7fff79e6fffULL,0xefffffdff73e6fffULL,0xfbbfeffbd73e2fffULL,0xc1cd3773d700276eULL,0xdef66d70d6262fceULL,0xebfffffff7be7fffULL,0xefdfffdaf6166fffULL,0x9ebd8debd62e29ddULL,0xc4f7bfebf6062fccULL,0xdea6ef61d68669d7ULL,0xfbd5ff2ed6042fdfULL,0xebdfbff3f6946bedULL,0xefbfcfdbf7b62f7fULL,0xbcfdfeff93be7fffULL,0x9b7effe3d6be6b66ULL,0xfbffff7ff79e6b7eULL,0xa365f683f6a62de6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xb6f7fffff7be6fffULL,0xfffffffff6be7fffULL,0xfefefffff7be7fffULL,0xfffffffef6be3ddfULL,0x3b7fffd9d6ac2e75ULL,0x26eacf7ad6043b7eULL,0xf5fdff7bd79e6fffULL,0xfff6ff73d6be6f7fULL,0x89a2e737d60629dfULL,0x4ad71ffbd7ae29f7ULL,0xcbbadf7ad2ac2475ULL,0x3cf67574d6a62bc6ULL,0x67b63d50d62e65b4ULL,0xcbf6e3eed62c6fdfULL,0xcd4a3e4993042c25ULL,0x97a74d7ad71429f7ULL,0xf51f16ead78c37feULL,0xff3abc71d2a464e5ULL,0xd6a22ffad6202f5bULL,0x8ffa9b1bd2843b7dULL,0xb578491292002a6cULL,0xb4b45ff6f2a42ff3ULL,0xdf6efd42d60c7f97ULL,0xcd083d64560c28ecULL,0x8fdfbb28f5ac6f50ULL,0x25c5d08d40438c8ULL,0xadb6ffbad6002df6ULL,0xdb985d42d220297cULL,0x2c07f582bc01f2ULL,0xa847f3e2820c2900ULL,0x2c3f97d5d7242b5fULL,0x491178ad00c6943ULL,0xd67dac78c60630caULL,0x34021561460c0184ULL,0x81a6c22692143023ULL,0x44e24a1690042149ULL,0x4dd2817256a42952ULL,0x8424282976002816ULL,0x3022d40c400209bULL,0xda80f966d6362972ULL,0xc64b162192143901ULL,0x92c86ad6002854ULL,0x75779aced70629cdULL,0x8cede77252202be5ULL,0xd21a6722c6a439cdULL,0x8583ae00c20c21c1ULL,0x180d81440200d0ULL,0xd0150a60d2022262ULL,0xdafeef6092206759ULL,0x4da41451c60c2892ULL,0x83266202d2064386ULL,0xab5b3217f7104be2ULL,0xc000002080000100ULL,0x910004a890206044ULL,0x768ca651c60e2180ULL,0x9018320d08020c4ULL,0x8840085000043001ULL,0xe024046091140200ULL,0x41284470d20c2052ULL,0xa40112000300ULL,0x40200080100100ULL,0x3540400c0000102ULL,0x808201c0040080ULL,0x0,0x0,0x0,0x0,0x0,0x2844000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcb065141d20020cdULL,0x6054020032002804ULL,0x1d87f7bd7102f45ULL,0x82814046002804ULL,0x8052020482002002ULL,0xd88ced42901c212dULL,0x8b02ce2cd21c216dULL,0x4c5b9a0ed624306eULL,0xa2390e5bd4843960ULL,0xc0e49208c680005aULL,0x803c2420540c0020ULL,0x8400cb4cd6040844ULL,0x9d8b9a40d60c2545ULL,0x981c048c5002401ULL,0x94dbdeffd6a43bdfULL,0x2d680561d4046980ULL,0x8080104800802244ULL,0x2000c21094002160ULL,0x68e7d040f28c20c9ULL,0x92046a30041200ULL,0x22144d0000118ULL,0x4c92050292102642ULL,0x443232054002045ULL,0xa100085080042890ULL,0x0,0x1002000c0000ULL,0x522014080002840ULL,0xa9aea90cc6a02544ULL,0x40080000002400ULL,0x8000002080004000ULL,0x1102000c90842044ULL,0x640a344892061334ULL,0xc0809346d2041114ULL,0x2e8a114ed6006985ULL,0xd0864d21920c4374ULL,0xb49475fad60c395fULL,0x2282c0030043820ULL,0x200000440000080ULL,0x840300042000010ULL,0x0,0x0,0x801002000004ULL,0x880020250000042ULL,0x0,0x0,0x2000010002000100ULL,0x0,0x0,0x10010000800ULL,0x6000030000000ULL,0x490010000020ULL,0x41002b0002000ULL,0x0,0x0,0x0,0x0,0x400002000002800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48945521d6002904ULL,0x42905b7cd61c330cULL,0xd0f082f7024000ULL,0x3fc5f56d680217fULL,0x8922c1c0822c60c0ULL,0xfd7f7f4fd7946dc4ULL,0x40b00533c7002114ULL,0x8005f264d4843069ULL,0xaf1b0bc2d68442eeULL,0x326281c600281bULL,0xb0f09649938c3054ULL,0x10002c00830c2000ULL },
	{ 0x8192412081941120ULL,0xe20211c20b940322ULL,0xfb9e4dff89d47afeULL,0xf7efdfac89dd4377ULL,0xfffffffeabff7bffULL,0xfffffff3abff7fffULL,0xffffffffabfd7fffULL,0xcfffffbf8bff7fffULL,0xffffdfdfabfe5fffULL,0xf6ffffffabff7fffULL,0xfffff77fa9fe7fbfULL,0xff9fff7fabfd7bbfULL,0xe7cbef6f8bfe3339ULL,0xffffffdfabff7fffULL,0xfebe7ff78bfd7b7fULL,0xfbffffff8bdc73feULL,0xfbfebf7f8bf67fefULL,0xe9feffbb8bff53fbULL,0xf5ededa89dc132fULL,0xb6fbffffabfd7bbbULL,0xdffdbf7c8bff7f7eULL,0xb7b23fed83a46fb3ULL,0xebfefbbe89bd53b7ULL,0x68f697de8bba77e6ULL,0x5e8ec5cf8ba81b7fULL,0xfdffbefb8bfd5f7fULL,0xdfeff70ab9d777bULL,0xf9fee3ae8bd47ba2ULL,0xd6fbffdbabf67f7fULL,0x9f9f5f4689ba5398ULL,0xf437eefa8bdc53eaULL,0x8dfc577089fa1ab2ULL,0xcea7f66a8b941f3cULL,0xe986df098bfc63b6ULL,0xbadecfdfab9a53edULL,0x36be7ed88b9b132fULL,0x4d642fd28be2736aULL,0xcbb66fdea1b0132eULL,0xcffefef58bf54bfeULL,0xef7fdfdf8bab73b7ULL,0xb5bf4fe389ef5af2ULL,0xfbfebd3a8ba87be6ULL,0xd6942676abd65372ULL,0x6916d7620bdd6202ULL,0x6c57fbc789b61b36ULL,0xb73b5ff389b273eaULL,0xf0e4ffd8a89c53a7ULL,0xc6beedc6839853e6ULL,0xcad1739eabed6bf6ULL,0x7cef3a998bf95b86ULL,0x5990cfa38ae843b7ULL,0x420000000000300ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6080000800ULL,0xffffffffabff7fffULL,0xffffffffabfe7fffULL,0xefffffffabfc7fffULL,0xffffffffabff7fffULL,0xd72caf7fabf47faeULL,0xb77e7f7d8bf653efULL,0xcbffffef8bfd7bfbULL,0xefdf3fffabde6ff3ULL,0xbfffeff6abff4bd3ULL,0xfdffbf7b8bfe537fULL,0xffbe7f7babfe7bffULL,0xfb97ff3eabf47feeULL,0xefffffffa9fe77ffULL,0xffffdfffabff7fffULL,0xfcdeffee8bb877bfULL,0xffbefd708bff5b7bULL,0xdb962b7f8b9c73a6ULL,0x300400288400002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8675f6fd8bbd37ffULL,0xfbfffffeabff73ffULL,0xfef6efffabfc637bULL,0xffdfffbfabbd5bb7ULL,0x2e7f5fea8bfc0f7fULL,0xb7ba7fff8bdd73abULL,0xd3fdff7babff73ffULL,0xfffefff7abff7fffULL,0x89f5d772abb27fffULL,0xeff77eefabbe7f7eULL,0xffffcfff8bbe6fffULL,0xfd76f6dcabb77beeULL,0xb89c3d6a8b904aceULL,0xd190ef33ab9d7a2aULL,0x4e4f76b28bdf472eULL,0xb73e0d668b9043e6ULL,0x75ac077a81a41a27ULL,0x5bb22d0000140760ULL,0xf7b67b78abbc0f60ULL,0x48defb0588b05a8bULL,0x6b965f5aa9b46b3fULL,0x9070d5e88bb47a00ULL,0xd9eebe4889b953afULL,0x7c2c6c70a0b51b12ULL,0x87d2e91889bd6b42ULL,0xa0b21958894c0b02ULL,0x2c11010088a07281ULL,0xe96c76e38ba4437fULL,0x9cce3e7581b04bbeULL,0xca06e37a89ae1f34ULL,0xa9dfe1d38bbe5bbcULL,0x16b146caa9a04292ULL,0xd23cbd7e8bf25379ULL,0x794653e28ac82222ULL,0x891af66989905278ULL,0xf84a925b09d8037bULL,0xb63b2de283902b33ULL,0xd47476c3abb0027eULL,0xddeefdd3abb45ff7ULL,0xfaa4eb0a8bb05935ULL,0x8e5a1e47893943a2ULL,0x27978dc389fc4bb6ULL,0x550c161d03b142caULL,0xcb8c3d6009826531ULL,0xc890e7ac8b10530eULL,0x8c5f263aa89473e2ULL,0x809e055201221228ULL,0x9cd454020ba24230ULL,0x87b30d20a9d04b02ULL,0x17be064309d0006bULL,0x804002000080020aULL,0x921655908190310aULL,0x84d710a289803223ULL,0x902424c20b907a00ULL,0x4090051181001022ULL,0xd88040068b811122ULL,0x0,0x0,0x6124052009026300ULL,0x1002244000101320ULL,0x40c5a30689841a34ULL,0x4950021081041320ULL,0x3000000000100200ULL,0x8000100001000040ULL,0x0,0x800010000100200ULL,0x0,0x800080000040200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200000100a00ULL,0x800200600a000320ULL,0x7575d0288984303ULL,0x8db46dea8bdd53a6ULL,0x85c02008100420cULL,0x904f24008a04282ULL,0x4380e72189f10b32ULL,0xa9539bdfabfd2afaULL,0x6f9dcbfb89fd7befULL,0x9d97ddec8bdd73e7ULL,0x8884821089401002ULL,0xa00e100000c6280ULL,0xc58a0008105b20ULL,0x92260f480a8c43ecULL,0xbcf8ede789de0a67ULL,0xc5aa8258a1807b22ULL,0x808402000ac80200ULL,0xe567ef1eab9957a2ULL,0x22ced8d0a9f41b42ULL,0x8480dc2082a01320ULL,0x51a6014082a10326ULL,0xc20462220b904208ULL,0xc23e60289d82262ULL,0x2010108008200200ULL,0x422000200802010ULL,0x120d50182815102ULL,0x100161408400380ULL,0x8000820408400000ULL,0x40521668018503a3ULL,0x4680a0800010ULL,0x4035502088c02104ULL,0x2424c881912282ULL,0xfbc4d4268bac7388ULL,0xa64afe40abb8333aULL,0x257247c0897572a6ULL,0x7494dece89be5baaULL,0x4b28b4038b045a62ULL,0x0,0x80100080000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1800001100000ULL,0x0,0x0,0x0,0x0,0x0,0x402000080000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000001000002ULL,0x0,0x0,0x1483620a2204b08ULL,0x12843b2e899853a8ULL,0x2ca6e00098c6630ULL,0xabdeded089b06387ULL,0x289ac7ef89f473f2ULL,0xd37ebf62aabc13b6ULL,0xc2924d528bf04363ULL,0x8072338a89d142a2ULL,0xb48c94089101224ULL,0xf130eff389ff72a6ULL,0x31967c6a8b8c5bb3ULL,0x1000094008800202ULL },
	{ 0xabd4f32acad5e12aULL,0xe2500252810c418aULL,0x80bfd6b2d354c02aULL,0x3c6e66f86185c0aaULL,0xfdfe96f87b4dc30aULL,0xfffffff7fbddfbfaULL,0xfffffffffbddfbfaULL,0xeffffffffbddfbfaULL,0xffffdfdffbddfbfaULL,0xfeffffeffbddfbfaULL,0xdfeeef7bfbc5f3baULL,0xfffc7f3fdbd5f33aULL,0xedeffb3ee1ddd97aULL,0x7fffffdffbcdf37aULL,0xffffffebfbdcfb7aULL,0xd9feeffbf9cdf33aULL,0xfebfbf7ff3ddd33aULL,0xffedfffffbddd3baULL,0xafffff7463d5c34aULL,0xb6fbeffbfbd5db7aULL,0xd5fef77ffbc5f31aULL,0xfbde7f65e1ddcb1aULL,0xe8d7fbafcb55e32aULL,0xcddbbe7fd3d5d3baULL,0x8fbadff3f94dc37aULL,0xf3e7deeee9cdfbbaULL,0xf5bcfd63ebdcd0daULL,0xfbb4ff9e71c5d31aULL,0xcedbb77de3c4c17aULL,0xaeffd7ffe3dde32aULL,0xf09fd6f7f3d4d96aULL,0x813eb7136908c33aULL,0xc79e37367b44d1aaULL,0x9dffbf1369d4c3baULL,0x8bffdfdfd3d4cbeaULL,0xf6f7ffeffbddebfaULL,0x9dee376ac9d4c32aULL,0xdbae777e5bd5c33aULL,0xf7d7ffdf7bcdd3aaULL,0xff5fffdbfbd5fb7aULL,0xd73f46eb5954933aULL,0xd17bab70e3d5d938ULL,0xf4be3f52f9c5ebeaULL,0xc6f87836e15dd26aULL,0xf8c085951c0411aULL,0xbdb7fffa63ddc2e2ULL,0xf8ed767fc2cdf17aULL,0xb392fd69e1dce27aULL,0xdbf0f77eebc5cb3aULL,0x6ffffed3f154db3aULL,0x10020000008008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xfffffffffbddfbfaULL,0xffbffffffbddfbfaULL,0xeffffffffbddf3faULL,0xbdf7fffffbddd3faULL,0xa0baddeb4bd4d3baULL,0xddfdaf7efbddf13aULL,0xfffeffffe3d4f37aULL,0xcbfdff2afbd4f3aaULL,0xe4bb3f8fcb84d82aULL,0x9f7bcfe3d9c4cb3aULL,0x97f5967ce2c5c32aULL,0xfffe7ffafbcdd1baULL,0xfbdeef7ffbcdf1faULL,0xefdfbbfbf8cdf17aULL,0xbe1ccff97bdcc32aULL,0xfdf7ffffdbd5fb7aULL,0xf7feff735bcdd33aULL,0xdae4ef3fc3c4e33aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x20080040001000ULL,0xe4f5f67ffbd5fbfaULL,0xfbfffffafbdccb7aULL,0xffbf7ffffbd5eb3aULL,0xefdfff7ffbc5eb7aULL,0xac5ffed94984c1baULL,0xb7ba5ff2fbc5eb0aULL,0xfbfefffbfbddfb3aULL,0xffdeffe37bcdd3baULL,0xdbc7e7fac3ddd37aULL,0xfcff3ec8f984d33aULL,0xf9adffafb4cf3baULL,0xf57ff2f7f1dcc932ULL,0xef3e3d78db55cbbaULL,0xcbd6e9fedbc4699aULL,0x4bd27ebaf9ccdbeaULL,0xd30e473b61c4d37aULL,0x44f6706ee180cb0aULL,0x338ad0083c4a042ULL,0xafa6db77eb05d30aULL,0x44ff222e954cb42ULL,0xa39a0d10da4ce37aULL,0xb6945763694cc33aULL,0x8546a570c0d0a12aULL,0xb2301f7cd104f91aULL,0x186c8004001d112ULL,0x837e5e5679cdd32aULL,0x3cb85de283d443caULL,0xa9505e2061ccc20aULL,0x4e9e2dffe3d4e35aULL,0x68de12324144911aULL,0x2f908e196004c328ULL,0x4030ae18904411aULL,0xd33cb679ebc4eb0aULL,0xc88e7d72ebd4e30aULL,0xb14f31ec15ce28aULL,0x65c21e1641ccc2baULL,0x1e9f1bc06b88c30aULL,0x845dee50c18de392ULL,0x816a1784144c10aULL,0x2ac064020185c000ULL,0xcdcbffaa6b5dc13aULL,0x281e4fc051c4c11aULL,0x8df45668d185e000ULL,0x9406ad31cb89c10aULL,0xca80af36c3ccc32aULL,0x5eebec341c5d33aULL,0x2271d588344c12aULL,0x1000000400008008ULL,0x984086108084e010ULL,0xb08450804081c008ULL,0x1d1ec0288848832ULL,0x60164e8001d4424aULL,0x79318de743c4c39aULL,0xd80c27906041c02aULL,0x9a96a65160ddc06aULL,0x4850900840c00aULL,0xa76384229c4c002ULL,0x480000008020ULL,0xc860202004002ULL,0x80020000008020ULL,0x8126d14640c4c17aULL,0x6000000018102ULL,0x25940066e0004102ULL,0x10080000040002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100100000008020ULL,0x0,0x0,0x0,0x0,0x0,0x2000002000804000ULL,0x80105008514400c2ULL,0x40e200418004a022ULL,0x5a8a4ff86bc5c13aULL,0x14269042e8804102ULL,0x2a00d0080044202ULL,0xc184fb02c0d0e12aULL,0xfb57dad3f35cc16aULL,0xe9becff1d3c4db3aULL,0x1431ca6000ccc10aULL,0xa520af728bc58b62ULL,0x4296742489c4ca02ULL,0x1c01ab14a848122ULL,0x856414408d4c00aULL,0xa49c60600104c32aULL,0x8041400088040800ULL,0xc80021000104410aULL,0x48988088808e00aULL,0xaf1206428345aa22ULL,0x1d8057e75944c12aULL,0x4000a6686850c102ULL,0x62102a028044a00aULL,0xedf5f5a2f3c5e34aULL,0x12d40a8080a000ULL,0x842954db80cde82aULL,0xd1ca264c6180c30aULL,0x8810058041005008ULL,0x81008a68814c484aULL,0x204020168008122ULL,0xc00090001404040ULL,0x204606c081406102ULL,0x20b80d50f804890aULL,0xb3d67d6ce985d30aULL,0xba8a33c34b40d32aULL,0x9bbecf2b895ce36aULL,0x74be99eae154f36aULL,0xa021042080006108ULL,0x52800000112ULL,0x1800000000108ULL,0x8040000008020ULL,0x0,0x8400000050000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4800000110ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6ae065724054811aULL,0xdacc3f46f140c13aULL,0x89c1e99ae3c4c30aULL,0x8b6e5f5bf081d12aULL,0x4607ae14bd4800aULL,0x700802fa4804c12aULL,0x3632320ac1c00aULL,0x8080000550000000ULL,0x994e4b52e3c5a12aULL,0x9491dcf57955d19aULL,0x34a4a80041d4d118ULL,0x4002000001004000ULL },
	{ 0xe9dcd260029af722ULL,0x6082100104805600ULL,0x82db8be14c1bb298ULL,0xd010566506c22260ULL,0x716fb7fa4e5afffcULL,0xfffffff3cedbfffeULL,0xffbfffffcedbfffeULL,0xefdfff7fcecbfffeULL,0xffffffdbcedbfff6ULL,0xffff7eedcedaffe6ULL,0xedffbffacedbffe6ULL,0xffb9ff77ced3bffeULL,0xfdf7ef1bcecbb3a4ULL,0xebdfffdfcedbfffcULL,0xd5ff7fefced3fff6ULL,0xf92ef7facecbbfb6ULL,0xfbffbf7fcedbfffeULL,0xcfdff77fcecbfbeeULL,0xabd19ed142129276ULL,0xbcbbffffcecbb366ULL,0xf7f7f7daced3f3e2ULL,0xfba67d604ecbfb70ULL,0xe933fbfb8ec2b644ULL,0xeff7bedfcedbfffeULL,0x9fbd4feacedab7e6ULL,0xd5fdfed8cedbfb32ULL,0x7ffefff3cedbb7f2ULL,0xebd6f70e4ec3bbe4ULL,0xafcbbf15ce4a32acULL,0x2fffdfd9cedbbb7cULL,0x3ef77ff84ecbb7c4ULL,0xcfba3531cccaba72ULL,0xe9866f6ecec3af68ULL,0xdfffffbfcedbfefcULL,0xad1ccfcececabaf2ULL,0xf4f37ef8cedaf2e4ULL,0xeb6e7feacec2f7e2ULL,0x9a5a6f2ecec3bbe0ULL,0xdeebffaacedbbadcULL,0x63fe575f4ecbb770ULL,0x7c9e7ff3c4dbf1eeULL,0xa1a5aa588e8a162cULL,0xdebe2f7b8ecbb346ULL,0x785ed74082403e94ULL,0xce7e7f92ced3f6e4ULL,0x95fbd6f3c6c3b724ULL,0xf55fbeeac6dabae6ULL,0x5ab6397ac6c2bba6ULL,0xebd5ffaf4ed9fe7eULL,0x5cc22e1cc00bb266ULL,0x200040040000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400400000101000ULL,0xc94528a06822008ULL,0xffffffffcedbfffeULL,0xfffeffffcedbfffeULL,0xefffffffcedbfbfeULL,0x7fdffe9fcecbf3e4ULL,0xbfbeeffecedbfefaULL,0xfffdfffbcedbbffeULL,0xffbeffffcedbf2faULL,0x6bc4fb3ac693b3eaULL,0xedff1ffeccc3fff4ULL,0x4f9efdfbcedbfef2ULL,0xf7ffe068cecbbef0ULL,0xfcf6bde2cedbffaeULL,0xfbb7f376cedbfffcULL,0xe273beffcedabe60ULL,0xff9adfdb4ecbba70ULL,0x74ff76fecedafb36ULL,0xff7efff3cedbb7feULL,0xdfff7b7fcadbf7e6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe4f7f6efcedbfefeULL,0xdbfff77ecedbfff6ULL,0xffbeffffcecbffeeULL,0xc38ff72fce5bbba6ULL,0xeffededd4e5bfee4ULL,0x873f4ef24ac3f2b6ULL,0xd4fdf7d5cedbff36ULL,0xfd5cfd71cecbb7b4ULL,0xc99ffbe3ced3ffb6ULL,0xbddb3e9a88dab356ULL,0x8bbec7f94cdbf6b6ULL,0xb7b4766844c3f7a2ULL,0xd7febf51cecbbbb6ULL,0xb5eef764ed29e36ULL,0xeedff6ffced3fbf2ULL,0xaf9fcf5acedbbefeULL,0x532748eacec99b14ULL,0x93bead70c6d19bd0ULL,0xcb84eb3e4a43abb0ULL,0xef5ef6a1c6c3bbd6ULL,0x8c0818004022022ULL,0x36e6dfec0edbfb62ULL,0x934c760846c0a220ULL,0xcb9377374edbfdaeULL,0x7fdffcbacedbf7feULL,0x80de7dd20e9b9260ULL,0x4c022f800e42ae02ULL,0xa9bef16bc09bb3f4ULL,0xaebc2d8246d3f192ULL,0xc0b4cd3286929320ULL,0x214680c046828220ULL,0xf4bb6f66c69bb36cULL,0xb1109578c083a660ULL,0xfb9e3ff4cecbfb7cULL,0x8316d3d84a422f34ULL,0x863e5c80868ab742ULL,0xdbbfc9a1ced3bfc0ULL,0x95a2fb594a5bfae4ULL,0x71e55d0ced27a88ULL,0x83a085000040ae20ULL,0xe649b0164e53b512ULL,0x61b41e100e02320aULL,0x4e617e1c483b6e4ULL,0x0,0xfb89eb3ececbf2ccULL,0x2fdeffeec6dbbbeaULL,0xd61574acc93b222ULL,0xa6b736614453b4e2ULL,0x842eadd14a4bbfa2ULL,0x90086564024ab940ULL,0x6e2204808040ULL,0x0,0x2000200081000ULL,0x1b6a24424c023620ULL,0xfa882f55cc83a236ULL,0xe852dbac0ecbb202ULL,0x88d6e84dcecaa43cULL,0x3a200008040a2020ULL,0x2c020080802020ULL,0x81060020c6001020ULL,0xc802dd20c8909244ULL,0xe40000000402200ULL,0x0,0x24080001000ULL,0x0,0x0,0x0,0x0,0x4240800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x420000000008040ULL,0x0,0x800040000400200ULL,0x80dcc00044002404ULL,0x84400060c422200ULL,0x6904c0802422222ULL,0xf3c233f1cec3a670ULL,0x1482f540ce823220ULL,0x8bcce16ac842f7b2ULL,0x9ad3721bc6dafa30ULL,0x26cc0b8484829342ULL,0x10ca884e021020ULL,0x8528a560c282b004ULL,0x48c0020680039700ULL,0x41a60004123640ULL,0x30842974880a220ULL,0xf434a6e2c60bf164ULL,0x2c298084822000ULL,0xa390220d84c31840ULL,0x81cac248c6821608ULL,0xa0c07474e88d300ULL,0x19402904818100ULL,0x7004a20006818200ULL,0x2010020040820002ULL,0xc5c4e212868297a0ULL,0xe7761fdbc693f7a0ULL,0xe1011304688a004ULL,0x91008152c4c3d920ULL,0x220024408021040ULL,0xa820e3124041b2b4ULL,0x422a1004903080ULL,0x147c082809600ULL,0xf1f4ba524ecab6b0ULL,0xdbe0573cccbfc32ULL,0x91d4e72a46cbfb06ULL,0x4a4613580ad2f642ULL,0xa7decf71ced3bf38ULL,0x309505e8c882d226ULL,0x928a250000109200ULL,0x0,0x0,0x0,0x0,0x2100000101000ULL,0x0,0x0,0x0,0x0,0x0,0x18000808000ULL,0x0,0x4880001000ULL,0x9f9e7ffbc6cb862cULL,0x0,0x0,0x0,0x106000000ULL,0x0,0x0,0x0,0xc00004800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x15d265604459b824ULL,0xc3ac6bbfc6d3b83cULL,0xa0cee8214cc89060ULL,0xa08c8c244ca1824ULL,0xa83224eac20a1624ULL,0x913609044840db04ULL,0x8210400084800000ULL,0x201c08000c28250ULL,0x68f6eac00ed2b320ULL,0x34ab4e6280ca9b42ULL,0xe154d6024e422260ULL,0x4020ac10c000d104ULL },
	{ 0x8002d16880769201ULL,0x447004808425b08ULL,0x84989db888265b91ULL,0x16ae579848f61a13ULL,0xfffffffffff7ffbbULL,0xfffffff7fff7ffbbULL,0xfffffffffff7ffbbULL,0xffdfbffffff3ffbbULL,0xfbffdfdfcff7bfbbULL,0xfdfffffb7ff7ffbbULL,0xdfffff6afbf7ff9bULL,0xefb77fffbff7fabbULL,0xefcfffbfcff33b9bULL,0xebffff5f8ff77fbbULL,0xffbffffffff77fbbULL,0xfffffffe8ff33ba3ULL,0xffff3f7fdff77fbbULL,0xedf7f7fb8ff77b9bULL,0xefff7af7dbd39e9bULL,0xbdbbdffbdf777ab9ULL,0xf2f7feeb8fe77f9bULL,0xeadee535ced23f21ULL,0x8393e01783f33a31ULL,0xfe5f5ece8dd37fbbULL,0xcebb57fb4973ff13ULL,0xdefee6fe3df37f8bULL,0xd7fefff99df3ff91ULL,0xdb80ff3e69533e11ULL,0xaffbbe5b9ef7fa1bULL,0x8dbdcff689535ebbULL,0xf5eff62f2d57fbb3ULL,0xb38e3d610a523aa3ULL,0xbba1ff7faad73bbbULL,0xeff7ff3fd9f3ffabULL,0xbfdfedcb0f72ba8bULL,0xfcf7f6f9db633f9bULL,0xd5eff77f5bf3bfabULL,0x59847f3528723291ULL,0xd9d2debd9973ffb3ULL,0x6ffddfdffdf6fb9bULL,0x22fbde710df37aa8ULL,0xc502a7d60de33aa3ULL,0xdeb61e73cde37eb3ULL,0xeffbf6720cd77a8bULL,0xfdffe7bfcf7fab3ULL,0x2f3addfb4ef77eb9ULL,0x7363bffa2dc37f93ULL,0x5fa63c62a9d35b2bULL,0xf9dff73eadd77ebbULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81804d180072100aULL,0x989412120a509e0bULL,0xfffffffffff7ffbbULL,0xfffffffffff7ffbbULL,0xeffeffffeff7ffbbULL,0xfffffffb7ff37eabULL,0xffbbfffffff7fbbbULL,0xfdfcb7fb0f777e9bULL,0xd3befdf308765b81ULL,0x81b0dfb148763e1bULL,0x6ed63e9b8de2bba1ULL,0xbbbfeffb5bf3ffbbULL,0x55f5e47b13f25e33ULL,0xf1b63d512f523a13ULL,0xebfce77eedf33e9bULL,0xcfc3bf93e9773a11ULL,0x65e9cbf119f25e0bULL,0xf5f5ffefdcf77abbULL,0xebbafd730df6be8bULL,0xdc162f602d537aa9ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8c7756ff0ef67e3bULL,0xfb7cfff7ddf3bfb9ULL,0xfeff7ffeeff73fbbULL,0xc1d7f7f70ff33fb1ULL,0xd573f591c777a29ULL,0xb93b47f289531e03ULL,0xd3feff33dcf23fbbULL,0xfd4effe26af7fbbbULL,0x8b86fb8e0ce37fa9ULL,0xe9d3de9a08775b09ULL,0xb3bbcefb6df37e03ULL,0xf744f2680dc6ba9bULL,0xcfbe7ff24bf73e8bULL,0xdb93eb0eabf31e0bULL,0xf8f3b7038877122bULL,0xffffeffffff3fbbbULL,0xc8f7deeaa9773f03ULL,0xdba79f700d733e99ULL,0xc98b6b1bca5333bbULL,0xefdfbefcab777b9bULL,0x350311d200d23e02ULL,0xdefcbfeb85c37e11ULL,0xdd767fcbcb431fabULL,0xebbf7727edf77eb1ULL,0xfffffffefff7ffbbULL,0x2b5c5cd84f532f01ULL,0x10a2de6b09633aa1ULL,0x18dc7f008e023f01ULL,0x8aff17f6f8537f81ULL,0xa396f2f70a72ba21ULL,0xa0c18bcba84232b0ULL,0x86b2f6e80dd37b0bULL,0x412cd51003507e01ULL,0xffbefffffb773f9bULL,0xc9f5e36702f69a8bULL,0x7d2380b18f3123bULL,0x8697c7628b43ba1bULL,0x2594f2214041120bULL,0x2bfa6d510f82391bULL,0x8230f366087019a1ULL,0xebf23a8e48171a1bULL,0x2b34435005c23a01ULL,0x70ccdeeea9c35ea8ULL,0x39045df288072292ULL,0xeeafdb7e8ff77e83ULL,0xfffffffffff7ffbbULL,0x34cc6c045503a03ULL,0x54b0d7ec68711e23ULL,0x54a835088021880ULL,0xc686294008523911ULL,0x90380244402021ULL,0xa25215c608201e83ULL,0x810000000004800ULL,0xd9e9a7d2cf837a01ULL,0x36b6005600631200ULL,0x102401800e03e20ULL,0x998c8c40d827f03ULL,0xa210433a08415b03ULL,0xb474ae4558521a1bULL,0x50823c400012120aULL,0x788850040c02368aULL,0xa46068305463a03ULL,0x804810040020aULL,0x410420290152380aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2080008000010ULL,0x0,0x0,0x0,0x840000000600000ULL,0x0,0x0,0x0,0x0,0x8100000000000280ULL,0x154a00828025f10ULL,0x28513510871060bULL,0x104430282121011ULL,0xb526a26e19735b28ULL,0xd56c3d1009c21b9bULL,0x8301c00208810810ULL,0xf7cfb64b88621a1bULL,0xe38fd7726bf37e8bULL,0x94e3b6e52ad37f83ULL,0xed007c220c03a68bULL,0xda52320638571a01ULL,0x407388a0016920bULL,0x210011960801088aULL,0x24b5f6608e533a8bULL,0x41e2260008131a01ULL,0x10c091400061213ULL,0x8052ca4101c75a00ULL,0x2ed7d95989d2fb2bULL,0x3454556009010a91ULL,0xd040c20800063201ULL,0x8e16312122621213ULL,0xc386fd0601133e31ULL,0x9a9eca0119423a9bULL,0xb7b6cff36fd37ba1ULL,0x919c86020842172aULL,0x2030f45808711b81ULL,0x826440020013160aULL,0x2201084008111201ULL,0xb5878d9408d05a11ULL,0x9540be7181531e02ULL,0x3db66df2ca733babULL,0x8a74e72009521a31ULL,0xbda36419d673a00ULL,0x889f6f7a89f27ab3ULL,0x70e646ec204332b0ULL,0x833829500150580bULL,0x0,0x4000040000000600ULL,0x0,0x0,0x0,0x0,0x0,0x14008000800ULL,0x0,0x0,0x0,0x4000001000600000ULL,0xafbfeeddbe777bbbULL,0xfffffffbfbf77bbbULL,0xd5d57e6b485212a9ULL,0x8024000040020aULL,0x8040042810601202ULL,0x0,0x80040000021000ULL,0x404100200402200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82665422c502283ULL,0xfbb4777239971f8bULL,0xc145de0148523601ULL,0xdf1c4928c232a1ULL,0x5000012000111800ULL,0x0,0xc0820f4809601880ULL,0x10100000010200ULL,0x8b4dd408dd23f22ULL,0x80a2079009600aa0ULL,0x1474bc41cc432311ULL,0x102004000440200ULL },
	{ 0x204200400260800ULL,0x2796df086634a2c6ULL,0x7bb2259a625200ceULL,0xffffdffb6f777fdeULL,0xffffffffef7fbfdeULL,0xdfbffff7ef7ffddeULL,0xffffffffef7fdfdeULL,0xffefffbfef7f3fdeULL,0xffffffdfef7fdfdeULL,0xffffffffef7f7fdeULL,0xffffffffef7fdfdeULL,0xffffffffe777ffdeULL,0xffffff9eef7f3fd6ULL,0xffffdfdfef7779deULL,0xdffffffbef7f3ddeULL,0xfbfcee7a6f771ddeULL,0xfbbe3c732e7ffddeULL,0xcfd5f7bfe477b6deULL,0x6f57fe7ba3679fdaULL,0xb7fbffff6a7f3ddeULL,0xb3ed9753267e15daULL,0x7ffe5c75e177d9ceULL,0xebc7e37b637f34deULL,0x6f7e3ffc63745cdeULL,0xbfadef73a77679deULL,0xb57ffeeea27799deULL,0x77be6f712737ddd6ULL,0xf3f7fb7ee07f9fdeULL,0x8fdbbe5be67f5ddeULL,0x2f3fdb536276fbc8ULL,0xdcf6f7ede27912deULL,0xdbafaff3e3773ddeULL,0xff8f7ffeef7fbddeULL,0xbfdffe9fe677b5deULL,0xfe5ecfdee7361ddeULL,0x71bdf7f2275f17deULL,0xdde8bf76a2779ddeULL,0xfebe2f7e6a7433dcULL,0xc9dafe29677e0c4cULL,0xebff5f5feb7f15deULL,0xf91acf83627e795eULL,0xfb7ff77aeb7f15daULL,0x4a362d2b236a114eULL,0xa96dff332c5f344cULL,0x2f9eda4be67ed8d6ULL,0xaa020fec00774856ULL,0xfb779f7be7757dd6ULL,0x5fae2f70a27e95deULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xbfffffdf6f7f3fdeULL,0xfed7fffbe77f3ddeULL,0xffffffffef7fffdeULL,0xffbfffffef7fffdeULL,0xffffffffef7fffdeULL,0xefd7ffffed7f3fdeULL,0xfffb7fffef7f59ceULL,0xdfffdefe6a777fceULL,0x76aa1de2605fa048ULL,0xc3f6f37ea27a2bcaULL,0xefffbeff6f7f3ddeULL,0xeeb297d366765ddaULL,0xc776f7fee77e5dceULL,0x5afe3df163763aceULL,0xfbffe577e3775fceULL,0xeffbffdfe2765ddeULL,0xbffe4f63e37798caULL,0xfcf5f6fb63779dcaULL,0xdfea6fe1e77679d6ULL,0xd68e216ca1769748ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x200020020000080ULL,0xfc77f66def7fffdeULL,0xdd6eb7f8627f17deULL,0xdf9fbfff2f7f5fd6ULL,0xcfcafff6627f35deULL,0xeddffedf2576b7deULL,0x17a2475e623e81c6ULL,0x83ffffeb2a7f35ceULL,0x884c756e637630ceULL,0xc1e7f22ee174585eULL,0x6afe3ed2e37618deULL,0xff9defff6637f9dcULL,0x975ffeefe775fdceULL,0xddae6c71ef7459c4ULL,0xe1e4f36be67a34deULL,0xe8e67f776076bdc0ULL,0xbfefcddfeb7f5cdeULL,0xf4f3bffbe3771fcaULL,0xe9e4af74e15600ceULL,0xc280c716621c1044ULL,0x8dcaf80fa17e15ceULL,0xa1decd4e26777d5cULL,0xc425c2f4227415d6ULL,0xe9f7bf70e55531c6ULL,0x6e9a25786b5617c6ULL,0xfec7fe3aa27e31ceULL,0x4d7a4a4f654e11ceULL,0xfc974eeb657617d0ULL,0xdfdcbf4ae37c91deULL,0x82140130210600c2ULL,0xc3c1ce0d621610ccULL,0x25100e5c60104844ULL,0x84b36662e65616c4ULL,0x112ea758206608deULL,0xfe3635546a6681caULL,0xe87c600a233711c6ULL,0x9773c8a265515d6ULL,0x81088d9ba36a90deULL,0x142cce54626600c6ULL,0x18253022375804ULL,0xfa54630e64761896ULL,0xe422e482a670040ULL,0xe29add7627765dc2ULL,0x91d2356ea64715caULL,0x84082446423501d6ULL,0xc080d1c226640148ULL,0xab57ef8be21e11ceULL,0x69e5cbdbef6710c8ULL,0xf6b944ec666719caULL,0x98722042001115d6ULL,0x8281000202640858ULL,0x8080080000011082ULL,0xba0e1d162163444ULL,0x4a61003a4304806ULL,0x98a6502231630d6ULL,0xce9f646c6c6248ceULL,0x288a520824104804ULL,0x8520200201010c0ULL,0x823046f0a22400c2ULL,0x952c257aa26608d2ULL,0x4e42557062373940ULL,0x6881812040761048ULL,0x2f46388ae74012d2ULL,0xa45975f12670204eULL,0x5761f6fbe23f84daULL,0x0,0x0,0x0,0x0,0x0,0x280000000000006ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100400000102000ULL,0x0,0x0,0x0,0x3000000000102000ULL,0x0,0x876131022661558ULL,0x32280c81a06a1c40ULL,0xa9561662631605caULL,0x1d746dc22756105cULL,0xb800012000501140ULL,0xc8febfdb627614ceULL,0x8ffe0fe3e676b9deULL,0xd4b477e92e779cceULL,0x70a27222a7e1c48ULL,0x2806208a2040042ULL,0x444a2201a03400c4ULL,0x8b0eca5b2b76d9c8ULL,0x410002000010050ULL,0x8065076062240848ULL,0x22900000a2040000ULL,0x6000000001010ULL,0x800ca4026000082ULL,0x8030546a00725886ULL,0x8000250022001100ULL,0x1ac62962034513c2ULL,0xa156e82420741dd4ULL,0x8c291bdbe2561086ULL,0x479ac3f7e6760ccaULL,0xd1ce754c203cd9d0ULL,0x7d8a2f202a7710ceULL,0xb34e33ba26611d2ULL,0xe23d1dd2e05622c6ULL,0x1dac4d28027c10daULL,0x98f720706257598aULL,0x6aba6572ab76f9ceULL,0x8bdac7ae627711deULL,0xc17b96c16615b14aULL,0x231e5f302a7419d6ULL,0x90982688207425c6ULL,0x8204500200a00caULL,0x0,0x800020000080040ULL,0x0,0x4002022008000ULL,0x0,0x4100000440000ULL,0x0,0x0,0x2000002002000040ULL,0x0,0x0,0x0,0x8160600a84000c8ULL,0x88088020460000ULL,0x113a2700621800ceULL,0xff9c0ff4267c1ddcULL,0x8906837827361740ULL,0xa1080000601148ULL,0x480344000060048ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8204056162540408ULL,0xdb9ebffeaf76dddeULL,0x89f5ce1a6170155cULL,0x12090046a0641100ULL,0x419022a60300d40ULL,0x800000200420000ULL,0x2100021000000ULL,0x940002062040000ULL,0x45523630244610d2ULL,0x828a07f3006615d2ULL,0x8180208220000c8ULL,0x100484060600000ULL },
	{ 0x8020000008200000ULL,0x30684c89002800ULL,0xc1c6cdc2091dced2ULL,0xffffdffbdf7dffffULL,0xfcffffffcf3ddfffULL,0xffbefff3db3dffffULL,0xffbfffffdf7dffffULL,0xdfffffbfdf3dffffULL,0xfbffdfdfcf7dffffULL,0xfffff6fb9f3dffffULL,0xdfff7feb8f7ddfffULL,0xffbfffffdf7dfffeULL,0xefdfffffdf3cfeffULL,0xefdedfdfcf3dfffdULL,0xffffdfffdf3dffffULL,0xffedffff9f3dfeffULL,0xfefebfff5f7dfffdULL,0xc977f7af1f3cfef1ULL,0x7bfeffdf9f3deeffULL,0xffffffffdf7dffffULL,0xf5aeff7fdf7cffdbULL,0x6bd665781e5cc6f4ULL,0xcbd7ff3f8f3ddff2ULL,0xfff73ff71b3dfdfdULL,0xbfaf4fb39a7cccfcULL,0xfdfdfe799b1ceffcULL,0x7dbe3dfa1b149efaULL,0xfbd6d13edf5ccffcULL,0x5ad71fdf4e3cf6d4ULL,0xffbfcd878b1c4eecULL,0xddffffeb8b7deff7ULL,0x9fdffdf2df7cdf99ULL,0xffa67f7fcb5dfdffULL,0xebc7fa331f3cdefcULL,0xfe92dddfcf5ddfd3ULL,0xdcb7deee9f3cfef4ULL,0xedeeffca9f3dffdbULL,0xd7a667f10b7cf7d5ULL,0xc9f7fb4e9b1dffbfULL,0xef1f9edf9f3cdebcULL,0x94f9dfdfcb19fedbULL,0xcff4f762855cdcfbULL,0xfba62d731a1cb6daULL,0xe9d3f6a68b3cdffaULL,0xaad6fbca1e3cf7dbULL,0x9fbedffadf7deeddULL,0xddfe7fae1a29e7feULL,0x520f643162dd596ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6fffffdfcb3ccfffULL,0xf7ffffefdf3dfeddULL,0xffffffffdf3dffffULL,0xfff7bfffdf7dffffULL,0xffffffffdf7dffffULL,0xafffffffdf7cffffULL,0xff6adffb1a7cefdcULL,0xff7ffffbdb5dcfedULL,0x6dfe7df91f0d6f5cULL,0x89227aa98901cc59ULL,0xafffbfdfdf5cffffULL,0xffd8eff9497cfdd7ULL,0xf765f7799b3ccffeULL,0xdfee6dc28b5deffeULL,0xdbf2fb3f8f3defffULL,0xcfdf721fcb3defbaULL,0xffbfdfffcb7dfefeULL,0xf0fff6e88f2dce78ULL,0xefbe6de3db3dcfdeULL,0xf73c7dff9a38ffdfULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa4d5f6e8cb3c9efcULL,0xdbf7bf0e8a2cbedeULL,0xde8f2dfa9f2dffdfULL,0xcff7f777db3cdedfULL,0xfbffdedf8f7dddfeULL,0x97a24f281a0ca5a2ULL,0xd7bfeff69f3ddfdcULL,0xfd7efdfe5f3deff9ULL,0x8b55bb2a0a3ca6bdULL,0x2fc72b588f2cdcf0ULL,0xba9e4d7a4b3de3aaULL,0xf153f67d0f3dffddULL,0x7ffeff6a9a3dded8ULL,0xfb90e7760c1cefceULL,0xd8c2fb699b3cecb8ULL,0xafbf8ffb8a7c9ffcULL,0xfaffffff9f3ddfddULL,0x8bb60e700e4cf799ULL,0xf9b0ff3b9f1cef9eULL,0x8a42fa7a8b2886dbULL,0x2c5fdfc6535ccedbULL,0x52f6b6d6cf3cdef9ULL,0x89fcba5a831c9efeULL,0xea4e0e36921c87d9ULL,0xdfc8f72a8f2dcefaULL,0xc9eeb08f0e3ce5d4ULL,0x636042a0308cc19ULL,0x5028a4c0060ccb11ULL,0xfe182f3f1f08df38ULL,0x40c2220002080010ULL,0x4a7e5b5e1f1dcfdcULL,0x3ca00d08025dc182ULL,0xd1acb7de9f5dcfdcULL,0x98962efe0b4c7c10ULL,0x810004710a000708ULL,0xccd72f571b1cef92ULL,0x841907b3161cce88ULL,0x53a0faf05b1c8eb0ULL,0x54aef56bcb3dcde1ULL,0xbb907f6c0b2ddeffULL,0xc9df18500b0dcd79ULL,0xe338ce534e1de6feULL,0x805174800b200f10ULL,0x40a4258000109291ULL,0xf3a46e728a2c8af4ULL,0xca461007882d45c0ULL,0xa20885c10019cef8ULL,0x56651e8ca146e98ULL,0xb14fb668835dca7aULL,0x8b925a7cc93c0ea4ULL,0x208000000c00ULL,0x4000108212080020ULL,0x2420081004400ULL,0xa37238a28ccd2ULL,0xb7124d0c8b1c4f34ULL,0xe9e6e263830cdc1eULL,0x235c5b47020dcddeULL,0x14990671080ca6d0ULL,0xc0012410080c2084ULL,0x8212150101044890ULL,0xc081000180080410ULL,0x2d0201010c4c6020ULL,0x14d007d10e20ce1aULL,0x24e606000e04c41cULL,0x0,0x20400040004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x401000000008010ULL,0x0,0x0,0x0,0x101000008002ULL,0x0,0x0,0x0,0x2002000001010ULL,0x0,0x0,0x0,0x8045d4106084c00ULL,0x80c20000008442ULL,0x2c1e35c505144ce4ULL,0x1448000080440ULL,0x0,0xa2a42d10081c8274ULL,0x3310c3020b008949ULL,0x99fb961b0f3ddbdbULL,0xfededf7b1f3cfff2ULL,0xc25776ed1f18fad8ULL,0xf376fc718b5dcfdaULL,0x6600440292000c00ULL,0x46c5d330083ca4c1ULL,0x6934f40071dc796ULL,0x98501c80060c4530ULL,0x90cd0288908a810ULL,0x9b0049625a10ec52ULL,0x580a80002180490ULL,0x2105c0ce0a044dc0ULL,0x122010081000c00ULL,0x112824440b28e110ULL,0x223224020108c054ULL,0xd1e69a03da308e58ULL,0xff5edf994f3ddbb6ULL,0x87befef60e18fff8ULL,0xf5e73fe1495caef0ULL,0xdafe6e41c91dcedcULL,0xc987e22fda1d5f98ULL,0x4052135103180b90ULL,0x138ccd910f44dc58ULL,0x848aa2400c20dcd1ULL,0x6b0e2d708810cec3ULL,0x9ba4a72c8865ecd8ULL,0xe2df32db5b1df71cULL,0x30d5a700b0d6f90ULL,0x745f9676c93c8ed6ULL,0x9a08251289215e18ULL,0x0,0x4002000008000080ULL,0x81180008006800ULL,0x0,0x30000000100200ULL,0x4000008000080080ULL,0x10000008002ULL,0x0,0x0,0x140060002080180ULL,0x0,0x0,0x0,0x4c8080040880ULL,0x0,0x4000218000180030ULL,0xc9819ba28b30ce18ULL,0xcdfb7e7b8f28dedaULL,0x1204404002001458ULL,0x0,0x0,0x80000200004800ULL,0x0,0x0,0x44000000c00ULL,0x0,0x0,0x0,0x0,0x0,0xc18c7d224439c4e1ULL,0xfb32ef7a1f6dff96ULL,0x80ece392cb3c0d50ULL,0x804004da81088cd0ULL,0x480230801184080ULL,0x64046881094410ULL,0x0,0xc88020200105cd0ULL,0x203200d6010cc488ULL,0x2049c00e086014ULL,0x920082400a208619ULL,0x9124644012008d42ULL },
	{ 0x41c0b109400001abULL,0x6d14a0024880160ULL,0x4400000000801ULL,0xde94f07d640a01e2ULL,0xfdf7fffffdba9fffULL,0xfffefdf3fdbadfffULL,0xfffffffffdbadfffULL,0xdfdfffbffdbadfffULL,0xefffdfdffdbadfffULL,0xfcfffffbddba5fefULL,0xff6eef73fdbacfffULL,0xffbe7ff7edba5dfbULL,0xefffffbffdbadff7ULL,0xefffffdbe5bacf7fULL,0xfdbf7feb5dbacfebULL,0xdbffff7f7dba9fe3ULL,0xffbfbfef7dbacffbULL,0xefd7ff0ffdba07eaULL,0x7fffff5ffdba5ff3ULL,0xfffbfffffdbadfffULL,0xfffff77bddbadff3ULL,0xffb6ef7ffdbacfebULL,0xe9f7d36b7dba9ffbULL,0xeff75fd675ba1f7fULL,0xbfdbe7d2cdba0562ULL,0xf7bdf6fab99acfefULL,0x5ffe7ff28dbadfebULL,0xd334af275d38cff3ULL,0xebf39ffbfcba5fefULL,0x47ba4ff345aa9da3ULL,0xf17f7f6ce4b807eaULL,0xfbbafd72cdb8cfbbULL,0xfbb5fff6edba4fc7ULL,0xcfdefb9bdd9a8fefULL,0x6ffddfdbc5baddfbULL,0x3ebedee80d2ad5f7ULL,0x5366fde1cdba57ebULL,0xdfbeef7f7d3adfe7ULL,0xd7dfffca5dbac5e5ULL,0x6f1f5ddb4dba1f7bULL,0x5fb717e69dba1f77ULL,0xddefff52bdbacffbULL,0xb2b63f7255ba4deaULL,0xc6d1f9bf55badf63ULL,0xefdade5b8dba55fbULL,0x95fafffeddba5ffaULL,0xd91dffeffcaa0fe3ULL,0x8000200000800100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x184206978d8a054eULL,0x6a5e5f580dba0fffULL,0xdcff7ffeb5ba17f3ULL,0xffffff765dbacfffULL,0xfebeafff7dbacfefULL,0xeffffffffdbacfffULL,0xbffffedbfdba4fe3ULL,0xba320ff86588136bULL,0xd3ee9ff97dba97e3ULL,0x871678609510850bULL,0xabfffffffdba9ffbULL,0xfffffffffdbadfffULL,0xeffffffbfdbad7ebULL,0xd4fcff7c6c9a5f67ULL,0xebbefd63ddba1f63ULL,0xf3e5ff3edd9a8de3ULL,0xff4bfebefd120dbfULL,0xffdf5f524dba4de6ULL,0xf1d7f7ee7d3a4fe7ULL,0xdffaef73ddba4fe7ULL,0xd1842d50502a4fe7ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x101000020800ULL,0xf637ffeffdbadfffULL,0xdb5effb7f49acde7ULL,0xdebfae7765bacbebULL,0xc7fc7537412a87fbULL,0xef7e3dc8cd3adfe3ULL,0x73a09ff984aa0da3ULL,0xf3ffbfff3d9a8fefULL,0x4a6e716261ba0fa3ULL,0xc861aa82041a0c2aULL,0x48d51e5669ba87afULL,0xba300fba45aa04eeULL,0x5adc965b4ca8133aULL,0x57ae3f61813a0763ULL,0xe1d5e746d0aa8f67ULL,0x6ddfbe7eddba4f73ULL,0x7fbaddfba5ba874bULL,0xedf7e6e3ddbadf23ULL,0xfff7e773fdba9dfbULL,0xf9bee71f5dba0fe2ULL,0xb56b7370cba0960ULL,0xad56dc5ac53a1f2aULL,0x1490502411a00701ULL,0x804502200084321ULL,0xde3827342c100bbbULL,0xd6e77f3a9dba1feeULL,0x663d97c04daa0972ULL,0x92917faa05b819a1ULL,0xd4e4a6175daa0fafULL,0xb7166f7cddaa17e2ULL,0x895cd2044cb81d2aULL,0xa771bc854ba5de1ULL,0x973f5ff7d49a5fabULL,0x612cd53acc3a5562ULL,0xfabaef6edd9a5d76ULL,0xeaa5317b5d985d2aULL,0x41d70c5a513a4db2ULL,0xf014e18d1384d20ULL,0x1190c5798c084d22ULL,0xcbac5c73c11205e3ULL,0x800400000000012ULL,0xeeeeb7ff7dbadffbULL,0x2a0c4f500d3805c1ULL,0x90545421a4aa0542ULL,0x85e42d616c8201eeULL,0xda60a32e70ba0726ULL,0x46863e3561b8872bULL,0x89be87524cba0722ULL,0x157b26e4149a0f43ULL,0x91f9a771059a0debULL,0x709c0002c5880cb2ULL,0xddd2468c1c18c741ULL,0x8120343d0100420ULL,0x440080104020102ULL,0x45ece71905b80fa3ULL,0xc2a62e57350a8de3ULL,0x2148592114081501ULL,0xcb7f780194981723ULL,0x92922a68049a80e5ULL,0x168762714280320ULL,0x4122f4610d289181ULL,0xc280e003113a4504ULL,0xae577f50d0920d5aULL,0xca1eeefadd9a4d66ULL,0xf44fa2e97cb81702ULL,0x0,0x0,0x0,0x0,0x0,0x10010001000100ULL,0x0,0x100800001200000ULL,0x0,0x0,0x0,0x0,0x800c0050000420ULL,0x0,0x0,0x0,0x8000020004800000ULL,0x0,0x0,0x0,0x0,0x3020040020000901ULL,0x4182c2620cb285a3ULL,0x60f3b2a24924d22ULL,0x208020100184400ULL,0x5144033945301702ULL,0x8611c024aa4146ULL,0x1d0c10c05a01b41ULL,0xc2d236e3ddba4762ULL,0x22904de48caa072bULL,0x75ae3eac158887bbULL,0xd3ab656135ba47aaULL,0xc890fd089d3805a2ULL,0x49423214a53845e0ULL,0xa4988cca602a456aULL,0x2000002004000100ULL,0x2014000000213ULL,0x3214060000b20442ULL,0x1018720440081a1ULL,0x2ddc65024aa556aULL,0x28064140120502ULL,0x900826a40300623ULL,0xc09229d41c8a0603ULL,0x48cbf32ac5aa8783ULL,0x4bdfd7d2fc8acf73ULL,0xb6376ffb1dbacf43ULL,0xddffbd7ffdaa8febULL,0xfff6adf75dbacfebULL,0xc9f3fb7eddbadfffULL,0xf3571fd6ccaa0f6bULL,0x479a659b343a0d42ULL,0x683602a8052a1b04ULL,0xcb8a3723059a016eULL,0x6204e20604820303ULL,0xee7a515f05ba57a3ULL,0xa58dded3ad0a1de2ULL,0xf5be76fa7d3a97ebULL,0x8801ad1215ba4c21ULL,0x0,0x0,0x10020000080002ULL,0x24000020800ULL,0x0,0x0,0x0,0x0,0x404000004000100ULL,0x8010000000180420ULL,0x0,0x0,0xc00000080100ULL,0x0,0x0,0x0,0x1005a8000000301ULL,0xefdf3bdffdbadffbULL,0xbf1fcdf2a5bacfebULL,0x0,0x0,0x0,0x80000100000002ULL,0x0,0x0,0x0,0x0,0x2004000080001ULL,0x0,0x0,0x678e6f314c3a5723ULL,0xccb27b5ee5aa1f3aULL,0xcde7ca22449a0f33ULL,0x2654d4c82c120f03ULL,0x122a92008a004aULL,0x208360040a014aULL,0x8220c680c801140ULL,0xc49a24030c30570aULL,0xb4eee1cd4820d22ULL,0x811eebf3edaa1da0ULL,0xf0a93428c400157aULL,0x8204406014900400ULL },
	{ 0x902c161201242e4ULL,0x48869a3124b26264ULL,0xc886052f241063ecULL,0x6020000020ULL,0xf5feffffbcf7ffeeULL,0xfffffff3bcffffeeULL,0xffd7ffffbcbbffe6ULL,0xdffffebfbcffffeeULL,0xbf7efddeb4bfffeeULL,0xfefff6ef9cbbeeecULL,0xdffaff733cbbeeeeULL,0xdfb66f5fbc9feeeeULL,0xefe7febebcdeffeeULL,0x6fff7f5bb4bfffeeULL,0xfeff7ffab4db7feeULL,0xfbedf7fbbcdefae6ULL,0xdabfbf7fbcbbfee6ULL,0xefdffe3e3cd8ffe6ULL,0xb95fbec3bc9be6eaULL,0xffbbdffbbcffffeeULL,0xffffbfffbcffffeeULL,0xd39c6f7634daf5eaULL,0xebf7fb3f349b7de6ULL,0x7e7f97db3c92fee2ULL,0xb7fcc9ff34032beeULL,0xdefefffeb4bb6feeULL,0xc7b2ed7a14f665e6ULL,0xfb9bc73fb093ffeeULL,0xf7d3feeeb497f6eeULL,0xe234c7ba3c95e862ULL,0xd4f6fe583496f7e6ULL,0xfbffff72bcbbfeeeULL,0xffb2f77eb4b3feeeULL,0xcedf7e7f3c9ffeeeULL,0x6bdcd9dfb49afeeeULL,0x54bf76dfbc32f4eeULL,0x83befd53bcb2f6eeULL,0x8a9c7f652492feecULL,0xfdcfdbddb4f6edeeULL,0xefffdedbb4d7feeeULL,0x7cbff7ffbcbffeeeULL,0xd5aba7d924ff7aeeULL,0xdfbf7fefb4b376eeULL,0xedf3ff9e24bf7ae4ULL,0x5ffedfdfbc9ddce2ULL,0xffffffffbcbeffeeULL,0x8054586b14b6c2a4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x462e13b4107c68ULL,0xcb9adc9fbcbaf6ecULL,0xbcfd5ff9bcd77eeeULL,0xdfe7ef7bbc9ffeeeULL,0xffffbfffbcf7ffeeULL,0xfff7ff3f9c9fffeeULL,0xef7effdf3c337eeeULL,0xfffbffffbcbfffeeULL,0xf173ff7dbcbf7eeeULL,0xdfb6777e3c3256e6ULL,0xcbd7ffeb3cb37eeeULL,0xefff5ffebcbbffeeULL,0x9fbddffb1cfafeeeULL,0xfdd7a7fd1cbecea6ULL,0xd2f42df03096fee2ULL,0xfa95ff5f34b6eee4ULL,0xc7fbbeb134b2eee6ULL,0xbfbfcff334b3fee4ULL,0xdd77776b14bb77e4ULL,0xdffefff3bcfbfeeeULL,0xffac3ffe34ffffc6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb4f5dee3b4d3e6c6ULL,0xdbf6f7ce349bfeecULL,0xf2be3efebcbbffecULL,0xcfe7f7a3b43376e6ULL,0x29efd9593c9eeae6ULL,0xb6fe4d7b34baf6e4ULL,0xfb7fbf7eb4daebeeULL,0x7a4e7cf3349be3e6ULL,0x8a7cff2434de6fc6ULL,0x95e0e058094e2e0ULL,0xbabec5c21493e6e2ULL,0x97e7f26994fe5eaeULL,0x526c2df08092e3ecULL,0xd3dcff263cbafeeeULL,0xef4bbfdb349afaeeULL,0xa6ae4fea3497ebe6ULL,0x1cfe77ea34b3ebc4ULL,0xcaaddf1b4faffeeULL,0xd480ef7ebc97faacULL,0xcf5bff2214b3fee6ULL,0xf95ec7d23c93f1e6ULL,0xdc76163684bc62e4ULL,0xdaa1c71030d46462ULL,0xc8bd2f3b941377eeULL,0xc182c6082cd84044ULL,0x8b8a5f1114917d64ULL,0x511607630906860ULL,0xcb7ca74b10daeee0ULL,0xca320bea2cfee164ULL,0xccd810600936024ULL,0x890d0a48943024c4ULL,0x813047509c1064a6ULL,0x8d60272934b0faa4ULL,0x1618296004327284ULL,0x8bceb98a18126ceeULL,0x5ddb379b84333aeeULL,0x4a8988a1000c86aULL,0x91a8b6b824da6944ULL,0x240134000020ULL,0x88f0c00408026cc0ULL,0xfbf7bb4cbcbbf866ULL,0x1e0ea7d3303276ecULL,0x94736fe634d6786aULL,0xa7792d55b43276eeULL,0xffadefe734337beeULL,0x3c4a6123010dca0ULL,0xab6e8b4ab4b2e8e4ULL,0x70b6d66834987562ULL,0xd16f656a3c037fe6ULL,0xaaac7c52b4fb6ea4ULL,0x406a12a34b844a0ULL,0xe9e2d8532410e066ULL,0xb132032134c23024ULL,0x59c5f50200d074e0ULL,0xd61e3e7294977aecULL,0x4cc6ee0a94fafca4ULL,0x4d3ce4b34127eecULL,0x86935c3a94137862ULL,0x90001008000002c2ULL,0x51c6f4c214b2fae6ULL,0x89d5fa033093dbc4ULL,0x140000000004040ULL,0x20244ab12c80c2e4ULL,0xfdbcb7e93cdef064ULL,0x0,0x0,0x0,0x4010000080100000ULL,0x0,0x0,0x600004001000ULL,0x0,0x0,0x0,0x0,0x8004000000101000ULL,0x0,0x0,0x0,0x851004300020ULL,0x0,0x0,0x0,0x0,0x0,0x20004004200000ULL,0x5000000000000480ULL,0x88c98000088c820ULL,0x82400224000044ULL,0x4000c000200004c0ULL,0x454e31a0340224c6ULL,0x8080690030004060ULL,0xe86bb2b6b8d1ecc2ULL,0xa79dcf9230a2c8eeULL,0x948af7ea1cdb44e2ULL,0x1eb225412092c6e4ULL,0xefcffb6ab4bffce6ULL,0x9d56f243b4bb7aecULL,0xac288ca0032c1c4ULL,0x1018026684005828ULL,0x60220380240016a0ULL,0x328c040c0c204164ULL,0xc956e40630b6c8eeULL,0x6c1f040092a0e0ULL,0x8019546010106062ULL,0xc948b660b01270e2ULL,0x8256084122042ULL,0x4e08d602308acec0ULL,0x25ef1d014cd66e8ULL,0x9fbbcfebbcb76aeeULL,0xf23b26641cba67e4ULL,0xfe5e6f7d349bfeeeULL,0xcbbcf7efb4b3ffe4ULL,0x6dcf1af5b4b3ebecULL,0x1c8ccd0a301046e2ULL,0xd4a2f2f8049b5aacULL,0xc29abd623492ec66ULL,0x91026b262092e6e0ULL,0xadda169394927864ULL,0x3709075210ba68e4ULL,0x51bcf65ab0967ee2ULL,0x9022a511943134e8ULL,0x0,0x0,0x0,0x0,0x0,0x4020000000280ULL,0x40200080100000ULL,0x0,0x346218a800137484ULL,0x10000244200000c4ULL,0x0,0x0,0x8100000c00000ULL,0x0,0x0,0x0,0x2200a80000028ULL,0xfcc23efb3cb36feeULL,0x98a2474a10946fc2ULL,0x6120ee404126aacULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8908e44310136aecULL,0xaa943f611c81ec62ULL,0x8355e3379491eaeeULL,0x89445b12301378a4ULL,0xa42004c400102aa4ULL,0x80282230802200a4ULL,0xd6863f703432feeeULL,0x4653221224b87060ULL,0x16d4840348816eaULL,0x109acb72bc88e8e4ULL,0xd1a4f6f3341334e6ULL,0x80d600400d020ULL },
	{ 0x6884f20210385442ULL,0x17711e4ff0341a70ULL,0xe36acf7a5034d6dbULL,0x4012023000202140ULL,0xfffff7fff2fdffffULL,0xffbfff73fafcfeffULL,0xffbffffff2fcffffULL,0xfffefebff0fddffeULL,0xefffffdfd8f5deeeULL,0xdffef6f2f2fc577aULL,0xfbffff7afafdfffbULL,0xffb77ffdf2b4fffbULL,0xffcbdf3ebaf4dffbULL,0xef7fffdffafdd7ebULL,0xb4dffeebf29474fdULL,0xf9ffff2bd0d4dfdbULL,0xfeb77f7ff03cffffULL,0xe7fef71e72d497ebULL,0xbbffdfcff23cf7feULL,0xb5bbdffbf8f4f7ebULL,0xfffefffefaf4f7feULL,0x7ffefd76f834bfebULL,0x6b82ea1ad014bfeaULL,0x6fffbffddabdffe8ULL,0x1ebccffad0bc7ceaULL,0xf5f7f6fef2f4d7feULL,0x5966fdd2109cfedbULL,0xfbd5eb6670bcdeebULL,0xecebbff1f2ddd7eaULL,0x8ecdcff9f095b5e8ULL,0xf457dff69abddefeULL,0xfbf6ff63fad8f6eeULL,0xdbbeff7edabc56feULL,0xdfcbbb3ffadcdff9ULL,0xafffdfdbf0b1d7feULL,0xf4f7ffefd875b7f9ULL,0xbfffff7bf8fcfefaULL,0xdbb2ff7f7a54b6ffULL,0xdf4effbfdad1fff1ULL,0xefffffdbb0fcf77bULL,0x7dfe5ff1f035dffaULL,0xfb763f63f2f49debULL,0xfe763f7f3afcf6f8ULL,0xc7f9f2bfda21d5e8ULL,0x3f5f9efbf21ddff7ULL,0xbffbfffffafcffffULL,0xfcd5fefbd07cffffULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8036e361020f648ULL,0xabdfddd7b034d6e9ULL,0xdcbedffcd27cfd6eULL,0xdb7cafb032d5dee6ULL,0xbab8af79d25cdff2ULL,0xefcffb3ff29cdffbULL,0xcf3fda7ff2f4feeaULL,0xd7bbcffb9abdd6faULL,0xf3d6b6f37afcf77bULL,0xd3f6fde232dddefaULL,0xebdfe72e5a7c5eedULL,0xafff7fd39875fffeULL,0xcebbeffbd0fc5feeULL,0xddf6b7fce8dc5efbULL,0x281c3d32300510aaULL,0xc3a3d33fb0b4def2ULL,0xa7ffb7d9d0f4f7eaULL,0xffffdffffafdffffULL,0xf6fffefff8fcbeffULL,0xff3f2d329a74d6d2ULL,0xf9044f3c301cd7e2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe6d5566b3034aff5ULL,0xfffeefeff0b4ffefULL,0xfeb6ff7df03cfffbULL,0xdbe9d7ba105dfffcULL,0xbfff3fdfd2bc7efaULL,0xc5b85e61b210dce0ULL,0xf1ff96fab8fcdf7aULL,0x71dc77ee123cd7cdULL,0x985cb0a701844f4ULL,0x8ed21356d0d536eaULL,0x33bfef5bd09c97caULL,0xfdf7fff5d09d56efULL,0x4fac7ef170d4de6aULL,0xfbd8e526d2d8db5aULL,0xece7ba0a122055c2ULL,0x793eef73725cd7eaULL,0xf4ff1e722054deefULL,0x8ab60f71d2f4df6bULL,0xca94fb5f92dc5fc8ULL,0x2147fbea70fd77eeULL,0x6b39c25cf000c6e0ULL,0x949f7ee6da4dd6e8ULL,0x24b617e3725d1fe8ULL,0x8ad2332c82fc54cdULL,0xc1b3ffa8005c5652ULL,0xc2de189752048663ULL,0xb0374fa2da1c17eeULL,0x865a4004040142cULL,0x8fb2257c505494ceULL,0xc1060b08104074c0ULL,0xaef6cec272d5f36aULL,0xb5aa1dd010745c92ULL,0x934d06608041d5e0ULL,0xcb86f47e1805f0eaULL,0x8153e634c0543ce8ULL,0x66131e0050005640ULL,0x8a1243819004f442ULL,0xf334f4ed42c1b7eaULL,0x651a1d105014fcfcULL,0xcac4cb26d03c5ebdULL,0x21e21e1950a4f0e0ULL,0xa73ccc4750f456faULL,0xe8d416a1501c5492ULL,0x2887d233090bce8ULL,0xffb77d81507856e0ULL,0xa1c4800078205420ULL,0xe14641521035cc69ULL,0x7a76f4e0f0b07cf8ULL,0xc1670d60502564e9ULL,0xc104629350141e82ULL,0xcff7ffbbb0dddfffULL,0xaf56fdcad0c1d4aaULL,0xdd9077fb8254f5daULL,0xf1e4f7faf0345feaULL,0xb09635dd52f4d272ULL,0xfc6201580b85642ULL,0x966325c70141c60ULL,0x1fb86fdb5834b6eeULL,0x98f5377bd004d6eaULL,0xfe9621e09a30bee8ULL,0x89418072d0100640ULL,0x8191224101406d4ULL,0xa0b080415001d2e0ULL,0x40c002000104010ULL,0x0,0x0,0x0,0x100001010000080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x600c2d20500486c0ULL,0xc300404250240460ULL,0x8626200010901482ULL,0x9fb85f3232a05568ULL,0x1404360800785c41ULL,0x252090040240ULL,0x0,0x1846224b0211cdecULL,0x2a100c8170009460ULL,0x2099302208a45642ULL,0x1084c008000100aULL,0xcb9ed77ffa81d7eeULL,0x4d56f228103c76c0ULL,0x11d002d0001401ULL,0x64202561400056ecULL,0x0,0x3946234320484e1ULL,0x4fc297ba30b95ee0ULL,0x1212954a20501471ULL,0x46e104f22141062ULL,0x80f80960b09411eaULL,0x4012020000002140ULL,0xa0dcd010e0001660ULL,0x6937ebc25098f368ULL,0xf2bea7e1f04056e8ULL,0xb1dd966b5a7455e8ULL,0xfff67fe5927cd6daULL,0xebcdfbbfd0f4f7ebULL,0xffde7edff2b4feeeULL,0x9274da110bcd759ULL,0x144db67d327452eaULL,0x69361d32c03d56ebULL,0x528401260010fa58ULL,0xba8a324190307c40ULL,0x9b0b8bd3f2b49968ULL,0x10e077ecdad4da40ULL,0x9b50270050305e64ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4852580040150868ULL,0x0,0x484100202011062ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x204000c0000ULL,0x4000032a50001000ULL,0x0,0x6a10440d044102aULL,0x410626840401668ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8001000000040ULL,0x0,0xd2a20741b084d4eaULL,0x820c62000014de48ULL,0x86cccb02501c13eaULL,0x80cc4f4990345dd0ULL,0x1a404100101410ULL,0x98c125429024124aULL,0x2bc2c70108c1488ULL,0xa4f4a028d010d008ULL,0xb04e8e19701c7478ULL,0x9efe6ff27854fffaULL,0x7145befb1250d6c5ULL,0xc4e22d51d2385668ULL },
	{ 0x8b95f769d9cda25aULL,0xafdf9fd3d1ef22caULL,0x239ccf54f9ed3348ULL,0x81ac786ed1e92904ULL,0xfffffffff9ffbbfeULL,0xffffff73f9febffeULL,0xffcaef7ff9ffbf5eULL,0xefdffebff9fdabdeULL,0xaffecfcbd9ddaf76ULL,0xdcf7fefff9eca77eULL,0xffbeff79d9e7bfeeULL,0xfffffffff9efbffaULL,0xfff3ffbef9efbbfeULL,0xafffffdbf9fdbb7eULL,0xfeffffffe9ef2f5cULL,0xf36ef739f9ddbfd2ULL,0xfeb6fffef1fdb3faULL,0xcdc4f77ad9fd23deULL,0xebfedadbe8f723daULL,0x9fbffffff9ffa7feULL,0xf57df766d9e5bb7eULL,0xfffefffff9efbffeULL,0xebfffb7fe9ef2ffeULL,0xbeef7fd1f1e7a7feULL,0xefffc7ffd9f7aef6ULL,0xd7fff65af9ed3ec6ULL,0x6ffc2f7bd9fcaf4eULL,0xfb74f737f1ed326aULL,0xce6fb6cfd9ed3b56ULL,0xef3edf7ff9fdabe2ULL,0x7df3f6fbd1ffaa7aULL,0xfbfeff71f9fd3ffeULL,0xffbcff7ed9fdbbeeULL,0xcffb3f3bf9edbbeaULL,0xef7fdfdff9efbffeULL,0xf6f5fefef9ddbbf8ULL,0x9f47bf7ef9eda77aULL,0xffbe2f7fd9e5bb7eULL,0xfffff6b6d9cd3efeULL,0xafffde59c9efb7feULL,0xf2bf7defd9e1befeULL,0xf3acae72d97daf4eULL,0xdebe3b68f1dd3b7aULL,0x29d7e288b0ec2f64ULL,0xeffeff7df0fdabdcULL,0xfffffffff9ffbffeULL,0xfdfdfefff9ffbffeULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdecffebef9fdbbf6ULL,0xbbffdf5ff9ffa7faULL,0xba934748d14da368ULL,0xf94c760bd1a6294aULL,0xfefa2f5ff1f7225aULL,0xffdfffeef9dda7faULL,0xefffff5ff9ffbbfeULL,0xbfbbfffff9fdaffeULL,0xd7bfff7bd9fdabfaULL,0xfedefffbd9fc2beeULL,0xf184f60af0c5a3d2ULL,0xeff7defff9edbf5eULL,0xefbdfff3f9edaedcULL,0xf467f2ecf96d2646ULL,0x3feefff3f1f8afe0ULL,0xfbfce72ef9ef236aULL,0xffff3ffff9edaffeULL,0xffbffffff9ffbffeULL,0xfdfffffff9edbfd6ULL,0xcb7af7d0d9fcabc6ULL,0xd2160304e14da244ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf4c5546bf9efbffeULL,0xfffffffff9ffaffeULL,0xfebf6ff7f9ec3bdcULL,0xefc6d3fef1ec23daULL,0x4a6f3fdfd8d7ab5aULL,0xeb2fffed1fdb3faULL,0xd4bc746b99a5ab7eULL,0xdf5aef6ef1cdbad2ULL,0xabf7fb3bd1edbbeeULL,0x6dd21fd9f9ce2edcULL,0x5ebccf5ad16da3eeULL,0x7534a2e619412752ULL,0xbfe7d70d8edae52ULL,0xd130a26fd1ef22c8ULL,0xa439f8ed1ec2e1eULL,0xdfbcdf3ad1fda368ULL,0xd0d7dce0f141a246ULL,0x86967d70c14fba50ULL,0x4a826cccc9ec2376ULL,0x8de6e227f1ec2a94ULL,0x2cdeefcee9412f4cULL,0x90b77ee6f9ed2f52ULL,0x59691252c9382354ULL,0xbb8c2b74c8c53446ULL,0x8091c30ab84c8640ULL,0xeb52974fd9e9a7f2ULL,0x24f5550ae84f1a40ULL,0x19c0c65ae1a52a44ULL,0xd622484eb0d40d40ULL,0xc163f938e16d2a48ULL,0x29611ac8408791e0ULL,0xff934fa6d18f0342ULL,0x24904280808250ULL,0x7c446560d1c0935aULL,0x4024220640202284ULL,0xcecf3359d9cda742ULL,0x811ce7fbc1e70a62ULL,0x9182d35ae9e5234aULL,0x41867570f1ed221aULL,0x30884b02d1aca04aULL,0xed7bbfd6f9f72b4eULL,0x8dbecf73f9feab5eULL,0x3444716298d12062ULL,0x5327d72d0d72948ULL,0xd026e10bc9ec264eULL,0xcb4c3613d15f3a68ULL,0x2306c94ad149220aULL,0xb4b0c644202c2142ULL,0x990ef532c16d224eULL,0xdf962b78d8f4b34aULL,0x88f291ded1c03a98ULL,0x7ba04750d12b2bd0ULL,0x34a20cf791c72040ULL,0xd13074da90e72342ULL,0x1a88094c91862046ULL,0x81f7a3a6c96d2a88ULL,0x2b401ad8514c135aULL,0x4f86d6741cda24cULL,0xd87cde78f1cc825aULL,0x6ff86d73c86aa7c6ULL,0xe9c7f1baf918a34eULL,0x6de13fd8d0cdab76ULL,0x3aa0ce8be0542b4aULL,0x8000200400000270ULL,0x0,0x0,0x0,0x0,0x90434080062148ULL,0x1000001040000080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2800202000ULL,0x1080000000010040ULL,0x0,0x0,0x0,0x0,0x4000200000140ULL,0x71de6079c97da576ULL,0x100c70a99980200ULL,0x204118d011480318ULL,0x2e08e1d0c0cc0600ULL,0xc13270524944209aULL,0x128409e8c1c0a600ULL,0x8000001040001000ULL,0x548000080008840ULL,0x0,0x801401000c0042ULL,0x922006408801a044ULL,0xd0044a00018c2266ULL,0xcc1101188602352ULL,0x9012050098c00218ULL,0x30b21600c060a140ULL,0xa0815091020140ULL,0x0,0x8612308ae064a600ULL,0x400400000400100ULL,0x92025ce3c056a142ULL,0x3104a26a484c2340ULL,0x200000280100000ULL,0x4e80862080450208ULL,0x20d19653c86db148ULL,0x94926eec18ed2142ULL,0x54a68d40d844a044ULL,0x40b62744d1499240ULL,0xebd8fb77f9e7bbecULL,0xaeff3edfd1e5afdeULL,0xf9bdcffbd9ef32daULL,0x848eb6e8d8e1aa44ULL,0x8b022d60c86da20aULL,0x88540420c0830050ULL,0x885032d3e1242228ULL,0x2dee4d69c8e127c2ULL,0x44fe137ac10caa6aULL,0x9b382d42d1c4269eULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1080041000000ULL,0x4008100148401000ULL,0x8010004080002042ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40a1d01240008278ULL,0x5c0800880480340ULL,0x1d83cf6ad9c6235aULL,0x81a6104851641208ULL,0x1008016088002214ULL,0x8000211080010210ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc144b630d8f4aa40ULL,0x8a99736cd11622daULL,0x2acfc10ee149331aULL,0xcb7c9d03d9ecb314ULL,0x0,0x1000800000002002ULL,0x82b92e50d94c2376ULL,0xa9c6d325c869a6e2ULL,0x804494080c80600ULL,0x95b7eff2c96fb3d0ULL,0xd1adb774c1bd2ab8ULL,0x8aa475f4d945a74eULL },
	{ 0x5858f1a018042572ULL,0x6150173ed202ac22ULL,0xa080110000520ULL,0x9fff4f6b9feefd62ULL,0xfffffffbdffeff72ULL,0xffdefd77db6ce7b2ULL,0xffbff7fedbaeff66ULL,0xffdfffbfdbfafff4ULL,0xfbdfffdf9afefff2ULL,0xd5f7ffecdf6cfd72ULL,0xf7efffebdabcf776ULL,0xffbe6f7ddbfcf576ULL,0xeff7fbfbdbfcfdc6ULL,0x6ffefddb9fecedb2ULL,0xb55edbf79a8ee7d4ULL,0xbad876681aecb742ULL,0xfeb43d7fdd7cfde2ULL,0xcdfeffa7df6cf5b4ULL,0xe9dffed3df6cfd62ULL,0x973ffff7ddfcfdb2ULL,0xf1fdf67fdfbcfd60ULL,0xff767fefdff8ff24ULL,0xebf6effedfeeffe2ULL,0xfef69ef358fc7730ULL,0xbabffffbdff8ffa2ULL,0xfff7ff7bdffcff56ULL,0xfffe7f72dfbcf7f2ULL,0xfbffeb3fddeefff6ULL,0xfafbb78b5fecfd36ULL,0x6fbdc7bf9decf562ULL,0xfff4f6fcda4cfd32ULL,0xfffffff3dffcfff6ULL,0xffffffffdffcfff6ULL,0xcfffbfbbdffcfdf2ULL,0xafdfdfdfdffcfd72ULL,0xd4fbfee1986cbf22ULL,0xfffefff1dffee7f2ULL,0xfb907ff99feefda0ULL,0xd979ffbcdbcaf5f2ULL,0x8fffdfdfdf7effe4ULL,0x3cffddebdb7efd00ULL,0xfd7f75bfddf4f7d0ULL,0xffbebfff1bbafff6ULL,0xace3efabd924bfe0ULL,0xabeeffffdf2efff2ULL,0xfebefff7dffafff0ULL,0xfdfdffffdffefff6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4fcfbe349afefdf2ULL,0x8fde575f9feeefb0ULL,0x3ef77ce493b6ad74ULL,0xf9fefffbdabcff46ULL,0xfab77fb7dfbeffb6ULL,0xffffff3edffefd76ULL,0xefffffdfdffefdf6ULL,0xffbfffffdffefff6ULL,0xd51d777add7cb520ULL,0xff9efffb9fbefff4ULL,0xe87fd30edfe8ff20ULL,0xfffebfffdffeff70ULL,0x7fbfcffb9bfeff32ULL,0xc576d77b9d26fc32ULL,0xcf667de31a3ac530ULL,0xeb86e122dbfcef30ULL,0xbfd7fffbdfeeffa0ULL,0xffffffffdffefff6ULL,0xfdd2f7ee9df8ff72ULL,0xa7de6db09fac6f36ULL,0xfebebf7fcfe8fd66ULL,0x41000010000020ULL,0x0,0x0,0x0,0x0,0x0,0x1100000200400ULL,0x1c145090d500a820ULL,0xffffffffdffefff6ULL,0xffbfafffdfbefff6ULL,0xffffff3f9ffeff32ULL,0x6fcffeefdfbcfd72ULL,0x979b7ff7da24f5f0ULL,0xa177bfca1de87440ULL,0xfb9ea760dfbcfd46ULL,0xfbe7ffefdffcff74ULL,0xefdfffffdffaffb6ULL,0xd69fc7fa997cf560ULL,0xda7ad67d9a7cf540ULL,0x29fe3d73dff8ed66ULL,0xdbf4df16dbbef712ULL,0xabe39bdecd68fda2ULL,0xfb3edf3b58fea720ULL,0x37e277684da8ffc2ULL,0x9fbefff2dfecfde2ULL,0xff9e777c5f3eef46ULL,0xcda3fe8bda8cf930ULL,0xabbbcd5a1b7c9f30ULL,0x16bdd768c728ed00ULL,0x8acf27729f40ff20ULL,0xcaa0252e0d40fd00ULL,0xd6c6cb1fddc8bd22ULL,0xa798de909768bd20ULL,0xb013002c95080520ULL,0xa16b777a9d087520ULL,0x5600057c15248570ULL,0x8082270e1068e710ULL,0xb560d4450aa6520ULL,0x92004968008a520ULL,0x915c030a900a8930ULL,0xf38c7d319f2a6530ULL,0x8b3efaa24d24f704ULL,0x84659118228e580ULL,0x99c1fb988ece120ULL,0xd54fa6bdac86504ULL,0xf3e2de2daa8f514ULL,0x89c6e11fd8ac2574ULL,0x58413be090b8a580ULL,0x6b97cb02523ced10ULL,0x40406001000a030ULL,0x2f6aff639872fdf0ULL,0x510651e1976c6400ULL,0x88df3f3b5f28ff60ULL,0xe03c5f409a2cc544ULL,0x1170764858306412ULL,0x814377ccd82ced20ULL,0x8bc8295418e8cd40ULL,0xcc77f3981f08cd62ULL,0xe61a5713d0b8bd00ULL,0x89d317699202bda4ULL,0x97ea27a25a08ad44ULL,0x88514210020a940ULL,0x8a54c2145028ad00ULL,0xa153ba590a68fd00ULL,0x86300462910a2120ULL,0x909406829ac83530ULL,0x16907d514df86520ULL,0xa9f5b1ee1228bd30ULL,0x86880241102095a0ULL,0x938a8d090820fd20ULL,0x400002000000600ULL,0x23c2000083100ULL,0x0,0x80800008400000ULL,0x10008084000ULL,0x8010000000080100ULL,0x40902c5019084402ULL,0x4204e2220038a980ULL,0x581e1098002600ULL,0x0,0x128064048309800ULL,0x8024054012002500ULL,0x72a0244090000740ULL,0x1449c30858042400ULL,0x2fd319db1a702744ULL,0x4b0440000008502ULL,0x100106008000500ULL,0x8080030440243000ULL,0x181d3380ab84500ULL,0x18ce4002303402ULL,0x9e0085748d08b422ULL,0x70f097020a306520ULL,0x650a4c781d603404ULL,0xa1b6215a1d08e510ULL,0x308a9a4008046120ULL,0xa8a8cc10d78c6a2ULL,0x24000002000ULL,0x240452098400400ULL,0x1000139900085d0ULL,0xcc6100c0d700600ULL,0x1e8c4d81986ca460ULL,0x115416699038fc00ULL,0x8021350015202050ULL,0x320a400402884502ULL,0x96612104b880d02ULL,0x114d1930720e180ULL,0x41a0d46102208d40ULL,0x1b600c60983c6522ULL,0x8a0e050491080d04ULL,0x40200000000400ULL,0x0,0x14226f0011288d10ULL,0xd0300e18d8003402ULL,0x8012004001380000ULL,0xcbd4b48fd16aff64ULL,0x812149810009110ULL,0x1abb7fff986cfd34ULL,0xfc7526f118bcf520ULL,0x5e020c42112aad40ULL,0xe9fec36e9a4ae570ULL,0x6ffb16e95abcfd04ULL,0xb7ddcfdbdbecfdf2ULL,0xf4ec76e2da6c7530ULL,0xa2bc2e701af0a510ULL,0x418083069a087520ULL,0x146b1000a284740ULL,0x214481008ac3540ULL,0xd4537a831f6cff80ULL,0xc81a2d22c8ec6522ULL,0x20000202000400ULL,0x0,0x0,0x0,0x0,0x8080000084000ULL,0x800020000880000ULL,0x0,0x1000226218208410ULL,0x0,0x0,0x0,0x0,0x100000248000000ULL,0x0,0xf29025544d402e00ULL,0x8046d2120a784100ULL,0xef5bbefbdfbce5e6ULL,0xc1f4fab12222562ULL,0x0,0x80000008800ULL,0x0,0x0,0x0,0x800010000000a000ULL,0x0,0x0,0x0,0x0,0x0,0x100320000000740ULL,0xea3c6f4c9d722522ULL,0xcaf1da108528e500ULL,0x3d895d01026e500ULL,0x801404a292888900ULL,0x240440290020400ULL,0xda26affe87c82720ULL,0xe99e7e2615a8f562ULL,0x17736401d7ef524ULL,0x1e27e7731ab87720ULL,0x90080761092ccd00ULL,0x7ab67ff35af8ff76ULL },
	{ 0x114e18402262860ULL,0x10064002031c00ULL,0x890c70002818f40ULL,0xffffffffbfff7f73ULL,0xf7fdffffbfffff63ULL,0xf7bffd719ffeff61ULL,0xdfbefbff97db7d61ULL,0xcfffbf6b9fff7f63ULL,0xefdfdfdf9ffe7f63ULL,0xf7b3feed1fdfff62ULL,0x9beffff33fffff73ULL,0xffbeff7fbffe7f63ULL,0xefdfffbf8fdeff63ULL,0xbfecf5a0bd76b60ULL,0x5ffb5feb2fdfdf72ULL,0xd1be777a9f86ed62ULL,0xb2bf3f77b7ceff61ULL,0xefccf73f37ff7f61ULL,0x6f77ffd716bf7f61ULL,0xffbb7ff3bfff5f73ULL,0xf7ecbfeabbc67f63ULL,0xfeb6edfea7ae7f72ULL,0xabdffffe37feff63ULL,0xea4f3fd59f82df61ULL,0xb7dbc7fa0f97ef62ULL,0xd75d3f7b8f9e2d60ULL,0xfffe7df39fde7f61ULL,0xfbffffffbfffff73ULL,0xaaf3fbeeaf8e0f60ULL,0xffffefff3fee7f73ULL,0xe4f617ea37d77f60ULL,0xffffff73bf5eff73ULL,0xffd7ffffbffeff73ULL,0xdfffffbf1ffeff63ULL,0xfbdfdfdfbfde5f72ULL,0x94fd7fff9fae7f73ULL,0xdbfeff7ab5ce5f63ULL,0xdfbf7f7f8fee7f63ULL,0xfedfefffbfdf7f63ULL,0xefffffdfbfffff73ULL,0xffff7fe78fffef60ULL,0x3f6ffffb9ade7f73ULL,0x5e8c2f4e9b8e5d61ULL,0xeff2f3bf3f867f72ULL,0x29fbfad7bf9e6f61ULL,0xb7befff19feeff73ULL,0xffffffff1ffeff73ULL,0xcd4efb629fffed71ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1c000e3f9b964742ULL,0xc292df1893867960ULL,0x943fdfe71b9fcf60ULL,0xf97deffa97cecf62ULL,0xfafebf7fbfff7f73ULL,0xedfffb3ebffe5f71ULL,0xafffffdfbffe7f73ULL,0xffffffffbffeff73ULL,0xf47dbf7b1bde7f63ULL,0x7fbc37d43bc65c60ULL,0xcafbb38f86d86f61ULL,0xe9cf9e539fce4f73ULL,0x7f9fcde39dc67f42ULL,0xd4b2576d3ebe3f60ULL,0xdfae7fe29a8f6d73ULL,0xfbf4e77f1ffeff70ULL,0xefefffff9fffff73ULL,0xafbfdddabdbf5d63ULL,0x7dfffeea8f9e7f63ULL,0xbb9677700b8acf60ULL,0xffbe7f7fbfdeff73ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1c14521a030e5f63ULL,0xffffffffbfffff73ULL,0xfffe7ff7bfae7f73ULL,0xeffffe7f1ffe7f71ULL,0x3fdfffff9fbf5f63ULL,0x93b84ffaa22d6d60ULL,0xf5ecb2f807ce7f73ULL,0xda16ae6601167d71ULL,0xfbfffffebffeff73ULL,0xffcf7fffbfff7f63ULL,0x376c95b332de6d61ULL,0xf1ff366b97eeff63ULL,0x79e626708f922f60ULL,0xcbc6fbbf872b5f60ULL,0x6cc3bfb3169e4b62ULL,0x1b1ecf2b9e972d60ULL,0xbadcd6cc9396ff72ULL,0xa7baff708fdedd73ULL,0x5f9ef77fbf8e5d73ULL,0xaedfda030b033f40ULL,0x21dfddd8ba8e0f71ULL,0xc0d45ff817c67f60ULL,0x493af5aa0b8e6d63ULL,0xe212261c11e25651ULL,0xd1f8fb2a11c25753ULL,0xcb9dffdbab4a7f63ULL,0xa118876317867e61ULL,0xdf50f75c9b232d41ULL,0xfab04f6e921b6d62ULL,0xcddffade939eff43ULL,0xc27fce4b3f97cd62ULL,0x9ab327a61eb24f60ULL,0x91b88d580b0a4753ULL,0xd2366f6e26cefe63ULL,0xc95eb2578346f463ULL,0x6a428a440b020d40ULL,0xbd99cfff1fd6df63ULL,0xf5d4bfe99b972f70ULL,0x7fb67d618b8fff62ULL,0xe914f22f1bd25743ULL,0xed4b1c1923867e62ULL,0xafbc4d2c12e64f71ULL,0xd0b4d6c620cc4962ULL,0x85e46d6006d6d561ULL,0xd796ffffbfcf4d73ULL,0xadd79a0c93d36d40ULL,0x2923d14804025d61ULL,0xa7f0d7e813ce4f63ULL,0xdbefb672ae9a4f61ULL,0xfadc46261e804d70ULL,0x4c4e790a0e124542ULL,0x20d485412120540ULL,0x3ed5257b83126d70ULL,0x10b021500e024102ULL,0x4c0a2d21051a5522ULL,0xeb0053250b8c5571ULL,0xb3d7fada0f7b4d62ULL,0x492c76a209e4f61ULL,0x31299f7122c46e30ULL,0x97300d5403965553ULL,0x8131c3341b924462ULL,0xc4d11c718f424d42ULL,0x88492804124141ULL,0x7038e66035825160ULL,0x1a84340121820e60ULL,0x0,0x0,0x2e0d4004474920ULL,0x1024080004000140ULL,0x8360c6216c0c340ULL,0xc800000200120140ULL,0x20258021ac00000ULL,0x2606000002004441ULL,0x0,0x340988168a0900ULL,0x8300000004060100ULL,0x300250000104140ULL,0x210105c92842820ULL,0x1023000401080260ULL,0x1084a41020000960ULL,0x88242f4004c64460ULL,0xa0d0f00293800640ULL,0x1094866080045962ULL,0xbaa04cae24c60460ULL,0x9000040402802200ULL,0x33a0296000826a43ULL,0xa1a0770a0a1e4d62ULL,0x846095093845120ULL,0x200804800844001ULL,0x9400d0400a060940ULL,0x83280f4200cc4223ULL,0x280c10000800660ULL,0x8e46de4287c44962ULL,0x232249580216ac30ULL,0x4c480040820ULL,0x1082080001001400ULL,0x4400010202122000ULL,0x6000000004001ULL,0x62040a080a040c60ULL,0x10400000840000ULL,0x921ab61611a05541ULL,0x82907a0cb1861a20ULL,0xe0f6cf188a966f00ULL,0xa08084008000d20ULL,0x1004000000000820ULL,0x0,0x7834b7c114834941ULL,0xe5dcd11e31b64f61ULL,0x81c5c801e100140ULL,0x9100471102a28260ULL,0xe07c56a00b980f61ULL,0x9302020c0b425000ULL,0xc397d31f1baa4d73ULL,0x647e1f5e0fd3cf51ULL,0x3f9dfffb9ef6fd72ULL,0xa53277eeb3847f61ULL,0x644a5d2213807f62ULL,0x88848a1c11665160ULL,0xe2411e5c21836f71ULL,0x229a4de1241a0d61ULL,0x243fe660318a5360ULL,0x89086502a2025560ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x800100002008000ULL,0x234481404ca6a60ULL,0x480140000000c20ULL,0x0,0x0,0x0,0x804004000040ULL,0x0,0x0,0x763c3c5d871a4f63ULL,0x8800201906810d40ULL,0x0,0x0,0x0,0x0,0xa0004000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8b3afdc0070e6f61ULL,0xf71c7fed05824763ULL,0x8bc7e31a1f965571ULL,0xf3d694420392ff63ULL,0x8c9b132131ce0560ULL,0x10e876000060860ULL,0x9ac096401e846540ULL,0xf50c66002877c00ULL,0x25fefa1387ce4f62ULL,0x8888c3b21e027140ULL,0xb04410a205426900ULL,0x4294a77c03964563ULL },
	{ 0x8d0c12cd9b24d40ULL,0x80908012002400ULL,0xa6becfe0d79e7d43ULL,0xffbffffff7be7ffbULL,0xfdfffffbf7be7ffbULL,0xffffff77f7be7fe9ULL,0xdfb7ff7ff7be77f9ULL,0xcffffbaef7be7fe9ULL,0xffdfdfdff7bc7ff8ULL,0xf4bfffecffbc3feaULL,0xdbeafff9dfbc7de8ULL,0xfe9e6ffff7167fe8ULL,0xdfffff7fffbe7fe9ULL,0x6b7fdf49f3be3df8ULL,0xfcd67fffddbe7f6bULL,0xff7efd76f69e7df2ULL,0xffb72f7ed79c776bULL,0xcddffe3efebc3fe0ULL,0xebdefbd0d7be7ce9ULL,0xbdbeff7bf4be7d7bULL,0xfffd7ff95db47f68ULL,0x5fe63665ff3e7600ULL,0xebf4fb37f7be75eaULL,0xfedf3ebfd6ae6fe3ULL,0x3dd84ff1c6b83561ULL,0xd73eff7bd5b47ffaULL,0x67be7df1d6be3ceaULL,0xebdef73fffbe3d6aULL,0xfed7fffaddbc6f78ULL,0xffb7effaf6b23f52ULL,0xf5fd3eefefbc35a9ULL,0xffbeff73f7be7ffbULL,0xfb9effffffbe7dfbULL,0xcfeffafbdf9c7d7bULL,0xab61f7d6d59c7de8ULL,0x76dcd6fbd7b43f60ULL,0x9beb5f71d73c7fe2ULL,0xdf9b3b5f77be7ee8ULL,0xdfd7ffbeffbc7deaULL,0xffffffdfffbe7ffbULL,0xfffffffff7be6ffbULL,0x7bebfffdd7b667f2ULL,0xfabe0f56d6947d7aULL,0xe9f7f6a7d69c7dfaULL,0xbfdfdefff73e6fe2ULL,0xf7fbfffbf79e7ffaULL,0xf5fcffffff3c7ce9ULL,0x48480102c10c1480ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42000001000400ULL,0xefdfdf49ffbc7f6aULL,0x17bbdfebd6b47fcbULL,0xfbffff38f7967f63ULL,0xde94b762dc3c7d20ULL,0xf7cbffeeff9c3df3ULL,0x177fff5bf6bc7fc3ULL,0xf79e4ff7dc9e6df0ULL,0xddfdf6faff9e75caULL,0xd286ef6d57bc26f8ULL,0xcb80d22294b465e0ULL,0xef7e1ebfd6bc7478ULL,0xfefddff0de9c3dc2ULL,0xd273f6e9d49e57c0ULL,0x7bbc7d71dd9c7fcaULL,0xfbfffb3fd6be7ffbULL,0xefffbfd7f7b47feaULL,0xff3fcf7fffbc7df0ULL,0x75ff67fff6b47fd0ULL,0x9f003542d73c24d2ULL,0xffbeffffff9c7ffbULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8100000000c0000ULL,0xfdfffffaf7be7ffaULL,0xffbe2ff7f7be7dfbULL,0xaffeffbeff9e7d69ULL,0x2dfededfd71e6cf0ULL,0xe7fefffbfdbc7ffaULL,0xb7fda644d4b477d2ULL,0xd3467f7ff79e7fe9ULL,0xcbf6ffeffebc6ef9ULL,0x56933e7fd6bc05faULL,0xb4fcdefdd3ae65c0ULL,0xd1d7f678d5907df9ULL,0xe97cfd7bf7b67ee3ULL,0xcb76e5bedfb45f6aULL,0xfeebfedbfd1c7feaULL,0xafbfeffbdfbe77eaULL,0xbd3676fdfe8c3dcaULL,0xfab23d71f7bc7fc0ULL,0xdf9eef7edf1e7ee8ULL,0x6fd7fa33f69e6440ULL,0xaf57c90ab3ac2c48ULL,0xc5f5d67156341ce2ULL,0xf16f037897945568ULL,0xedbc3f5bd4ba0d60ULL,0xd8109600d3106d20ULL,0x599641548c2440ULL,0xfd0f032bd6306749ULL,0x212cf60053b06620ULL,0xffb7ffffffb87f69ULL,0xdbf7f2bf47b83ca8ULL,0xa21b1c0bd1040570ULL,0x14ae0da0d2a02dd0ULL,0x1000228d0004050ULL,0x821e7a44b10c2d20ULL,0x954c2285a003c20ULL,0xc6e81e4ef31c2490ULL,0xbdba4fdb569c26f0ULL,0xf5fefffaf69c37f9ULL,0x7fbe3d797fbc7de9ULL,0xf891f32ef7bc34e0ULL,0xc1436e39548621f0ULL,0x2f76f7dbd69c64fbULL,0xf69ff67dde9e57faULL,0xf7bc2f30b1bc07e0ULL,0xebb2fffcf59c7df8ULL,0xe2d5be11d41c1ef8ULL,0x2475cda4e9c1450ULL,0x72c9c4788e9c4460ULL,0x59ac1712d49c1640ULL,0x8aa82628521c2470ULL,0x9ddadf1ef5184430ULL,0x3f3e195877107c30ULL,0x6b8e4d356386cc8ULL,0x102c74051200530ULL,0x180e0c20c6102900ULL,0x6681972652802478ULL,0x95412c67e180620ULL,0x957a7fe3563c7649ULL,0xe1cd96b8f68436f1ULL,0x82200264d1004050ULL,0x8320e132c59a4442ULL,0x23d90c01c0086431ULL,0x10060a90100400ULL,0x10e0245050040490ULL,0x200200000040080ULL,0x48c04a0a870424c1ULL,0x800210080806000ULL,0x0,0x9050864413200468ULL,0x819cad00f4902140ULL,0x0,0x206e003c41c4000ULL,0xa200044a10042980ULL,0x92048460402c14a0ULL,0x1000004040000080ULL,0x0,0x4900b202441400c8ULL,0x800200000408ULL,0x16be06e2919c43c8ULL,0x4100a000a0004400ULL,0xc880294d748c2c82ULL,0x146e020c5106800ULL,0x6b9cd41897b42d20ULL,0x88034946003c00ULL,0x88790f69d5b43f68ULL,0x58c44a44c7180408ULL,0x80402b1c74902421ULL,0x4ece9e38c7263420ULL,0x222e0ec856902c40ULL,0xd1d53278b62c3449ULL,0x500000000100000ULL,0xe850600286342c00ULL,0xc263f6da50a43c61ULL,0xe408441c5104000ULL,0x8610166854062820ULL,0x4800000080000100ULL,0x4200cb2cd5840140ULL,0x2e00042002002ULL,0x10cad814902820ULL,0x50802040400028c8ULL,0x928f805ad0907cc0ULL,0xd02414a2c2045480ULL,0x88e48201f1200680ULL,0x32084c80a1806420ULL,0x86868ef3f01c3da0ULL,0x6080000100ULL,0xb8d10110049c1440ULL,0xcfd4dc2754b43eeaULL,0xc064e08c2800542ULL,0xed08703d40c4ee0ULL,0x5002825010986c68ULL,0x82123d2450000a82ULL,0x8030db8212182d48ULL,0x4964e0bd2246cc0ULL,0x5fbcfffff7bc7dfbULL,0x64f277e9f6b42d6aULL,0x412a3700d7344780ULL,0xe9d4e2aaf59c3d40ULL,0x84cba801d6142578ULL,0x68a848a1dc0c25e0ULL,0xfdbf76fff73c7f4aULL,0x9b3a9d2253143d92ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x81d0140872042170ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe414006813104420ULL,0x8400000080400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x10001000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x5fac4562961c2960ULL,0xd2846880b1906c48ULL,0xc7a8d80ad4982d70ULL,0xaafdded6f61c7de8ULL,0x94ca4168d2045422ULL,0x4a283722d39405c0ULL,0xf126bf609e346fc1ULL,0x638ee731d41c7c48ULL,0x655a9042c4123d08ULL,0x8000061040200400ULL,0x111007a12302460ULL,0x40a2294014900408ULL },
	{ 0x8a0aeb6289fa2289ULL,0x880168008a02210ULL,0x80285eb18a040316ULL,0xffffdffffbfe2fffULL,0xfdfffffbdbff2f7fULL,0xfbfeff73fbfb27ffULL,0xfadcffff9bfe2fdfULL,0x7fdfbfebfbfd2f7fULL,0xe7ffdfdfdbfe277fULL,0xf4fffefccbfe273fULL,0xf9ef27efabf62bdfULL,0xfbde7fffdbfb2fffULL,0xefeffffedbfb237fULL,0xbfff5fdfbbab222fULL,0xfcf7dffb9bb82f77ULL,0xd3feff7e9bfb2fdfULL,0xfebeaff3abbf2ff6ULL,0x8d5ff7ae8af1278fULL,0xf7fefffefbf7233fULL,0x9fbffffe9bbc2fffULL,0xf7efb7fb9bfc2f3eULL,0xdfd67ff0bbf8273fULL,0xc9daf3ae8be227cfULL,0xeff7feffdbfc2fbfULL,0xbf8eff7d8bd823f5ULL,0xec7f2af0b989264eULL,0xffae3fe2cb9e27b7ULL,0xf9a1e3ee83ac27efULL,0xfffbffdffbdf2fdfULL,0x3ffedfab9bb023bfULL,0xfd77f6eacbbc277bULL,0xfffefff7bbde2fbfULL,0xffff7f7febb62fffULL,0xcfd6ff1f8bbe23ffULL,0x6b9ecfdaabde236eULL,0xd63736e89baa276eULL,0xb9feff7bfbbf2bffULL,0xffbe7fff9baa2fffULL,0xeede7f1edbda237fULL,0xffffffdffbdf2f3fULL,0xfffefff3fbdb27afULL,0xfbfffffb9bf2275fULL,0xfeb6bf7fabdf2fffULL,0xedffffbddbfa2f6fULL,0xafeffffbf39e27efULL,0x5ddefffb8bbd2f7fULL,0xd1edfffa9b9f2b3fULL,0xc90ce9408b880316ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe7dfdfcfbbfe237fULL,0xadfe5fef8bda2b9eULL,0xfb3eff7ffbd02b1cULL,0xfeb7ff6e8bae279eULL,0xafdeff2b9bb3236fULL,0x6f57bbdfcbb02f7fULL,0x97bbffff8bd423ffULL,0xf7edb7cebbf827ffULL,0xe38e3d2c8b79237eULL,0xe3bce3238ba32b96ULL,0xe0ce0a568b6422eaULL,0xaa187f70cb082396ULL,0xdd75f2e58b942f78ULL,0xef767f709b992f7fULL,0xfbd7ff77dbd8275fULL,0xaef3bfdacbe62777ULL,0x6fbfcff1abce27eeULL,0x74a797ee8bfc277fULL,0xdbb2f5728bf02f2fULL,0xfffffffffbff2fffULL,0x0,0x0,0x0,0x0,0x200010002800000ULL,0x0,0x0,0x0,0xffffb77b9bf52fdfULL,0xdfb63fefcbf8237eULL,0xedfeffb9939f23efULL,0x2b5e574fdace2716ULL,0xb0024ff3b3980347ULL,0xfcfce5f08b93276bULL,0x867665bb82013eULL,0x63f7e76e8b9e2ff6ULL,0x6e570e89138a2b1fULL,0xd3be1f7a0bb7227bULL,0xacf672a10b5c274eULL,0x59e2b8428380025eULL,0xfbf6ff7fabbf27bfULL,0xffffffffdbbe2f7fULL,0xfebfcffbdbf827ffULL,0x5cd1f27ad39e2b9fULL,0xdfbdbd708b9e27efULL,0xeb903f74bb902224ULL,0x99dffe2f1bf327bfULL,0x8bce8bd089f02639ULL,0x5cf664e0ab9a215cULL,0xd1a6b4508bd1021eULL,0x91004f608b402a49ULL,0xd8d387188b0c0b0aULL,0x8388061383942704ULL,0x36be1fe6c3b2267eULL,0x40ec827083942314ULL,0xdba0245eabfc031eULL,0xcddcf6a28bc22b3fULL,0xfefea5bbbb82f5fULL,0x1ef16d8c8be6270aULL,0xed3cbc72838d2b2fULL,0xdeb0f74682b42b1eULL,0xe012420c93802b0eULL,0x281006848bd82000ULL,0x27fbee61aad12b9fULL,0x910eb05800842b01ULL,0x739c7613bb802f15ULL,0x9b18c30c03920b1dULL,0x6c4a000089802218ULL,0x224e5188c02312ULL,0x499612688a102305ULL,0x903a2df083b22307ULL,0xcf86676f1bbe2b1eULL,0x4ddbfe168baa270fULL,0x234ed3508b9a223bULL,0x3400766119d22305ULL,0xf12e8be08bb42243ULL,0x8b88050ec8c02305ULL,0xca560b0e4a402b04ULL,0x2ffccc59aad02315ULL,0xa434406a8290030dULL,0x8049245012902704ULL,0x6c91245009d00276ULL,0x6042020000100a00ULL,0x2d64200082802002ULL,0x300a014088080a0eULL,0xa1018e9009880a52ULL,0x4080260240202a00ULL,0x1da76504a182207ULL,0xccc23e11c2302205ULL,0x200201124a002200ULL,0x120025848402000ULL,0x848414883002006ULL,0x0,0x8000200240002001ULL,0x0,0x4863d7698bf02f3aULL,0xc590a370a998223aULL,0x2c0001120000ULL,0x9c90920088a80212ULL,0x8000a4001800002ULL,0x4050265001900030ULL,0x4000800000208ULL,0x0,0x808fc808c39c2217ULL,0x2100000000000300ULL,0x420002080800030ULL,0x8000060082002402ULL,0x95a0054002202b0cULL,0xc0c760020bc00365ULL,0x6a7effc309952e63ULL,0x80101a82802014ULL,0x80e0fe42831a2303ULL,0x70a8a400b280022cULL,0x89f4c33a9a8a0156ULL,0xea441fdae19a2641ULL,0x603c42939384022fULL,0x810ca25883102219ULL,0x2c23540a200020eULL,0x8a10410082002100ULL,0x8523212b3902105ULL,0x22ccedad3902124ULL,0x5020c448a1910013ULL,0x20027470a2022401ULL,0x81004b0480802031ULL,0x8002910080002010ULL,0xad76c21a8be0222bULL,0x548068a983b20202ULL,0x805c213881a20b39ULL,0x2195360283000b3cULL,0xc000511389300208ULL,0xd2d4c100b900338ULL,0x1490c6e81100026cULL,0xc02d225883100104ULL,0x761e2d50ab342b43ULL,0x8480100010000304ULL,0x890a01000b000002ULL,0x698155eab3a260dULL,0xd36d87685bd82776ULL,0x816407011a0026cULL,0x8100846002010106ULL,0x7ece3fd9dbf92376ULL,0xbff9dff39bd527feULL,0xd6e5f07b8bb82377ULL,0xcb3215529bc82eb7ULL,0xfac6ab0a8bce22a2ULL,0x50e3b7109ba0270eULL,0xa82c9588a48230dULL,0xe8ffbec09bd82337ULL,0x1088011000000300ULL,0x0,0x0,0x0,0x0,0x0,0x82005200b080202ULL,0x840020082000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8180002000ULL,0x48001100000ULL,0x1000010080002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc1423c4092822705ULL,0x1e4c7d408880230eULL,0xe9d6ee800b99223fULL,0x2fdf5f488a05235fULL,0x9430d7658bc0233fULL,0xdde9a8c40bd02717ULL,0xdabe6ffe938a2b5fULL,0x4120c7b0c3902a37ULL,0x2b0808d2008c2290ULL,0x80107a621a802205ULL,0xf08174688b9c033bULL,0x9426094488402690ULL },
	{ 0x89c8d31c6034c328ULL,0x584313220001c3a6ULL,0xefffffffebadf7aaULL,0xffbfcf7efb3ddbbeULL,0xff7fffeff9bdf7beULL,0xdffffff7fbbcdffeULL,0xff90eb7fe3bdd3aeULL,0x89ff7e15fb30d3aeULL,0xfbffdfdbfbb5df3aULL,0xf4f57feef9b4db2eULL,0xffffff7bfb35dbbeULL,0xfb9677757315f3baULL,0xfcffff1eebb5f7feULL,0x7bceffd7ebaddbbeULL,0x3537eff7ebb5f7beULL,0xdbfbff7afb1dd3feULL,0xfeb6ff7feba5f9beULL,0x61f6e33fc3bdd73aULL,0x6fdffffbfb34fffeULL,0xbebbfffffbb5dffeULL,0xfffdfffffbbdd7feULL,0xfa9e7f6ee935d3beULL,0xcafefb7c5b05f5baULL,0xfff79fd3f314fbbeULL,0xbaeeffbfbbdff6eULL,0x9dd2ff7beb25f3baULL,0xff9e7f7beba4f96eULL,0xfb74fb6ffb21d7beULL,0x4fd23f7fd934df7aULL,0xeff7ddfeebbdd7beULL,0x77f7ffe8f980d16eULL,0xffbeff70db1cd76eULL,0xfab8df7ee0bcfbaeULL,0xcfdbfefff335d9faULL,0xfbdfdfdffbacfbfeULL,0x9638dfe4db21d7a8ULL,0x84200400a004002ULL,0x30041008304200ULL,0x0,0x108141208304260ULL,0xbdff7ffffbbcff2eULL,0xff6cef73fb20d7faULL,0xfebe8f6e5b9ce39eULL,0xafe7ffabfba05f2eULL,0x2fdfffdeeba5d32eULL,0xffbaf57be8b1d37eULL,0x3144162888205320ULL,0x200040001004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2bfeffdafb24d1aaULL,0x75b7ffefeb14ebaeULL,0xb77cef7efb3cd33aULL,0xfe3e377efbb5fbbaULL,0x2dd6f796498dc3aeULL,0x2f9cdacbe38cf5aaULL,0xf7bfcffbebb1d76eULL,0xffbdbffaf185d1baULL,0xf3766d71c3bcd7aaULL,0xca04a066c884d12aULL,0xaaf71ed5cb35f3b6ULL,0x39a1ad82c115d10aULL,0xb01d9268a91453aeULL,0xde9c7df3c394f7e6ULL,0xebe4d7eefbb5dfbaULL,0xaafeb3dfdb35dbfaULL,0xf3bfcfffdbb4d77eULL,0xf5b7f7fbeab1d30eULL,0x4baefd63fb04d1feULL,0xffb7eb7fe31df3aeULL,0x8947e8815838c382ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb97def7ffb34f7feULL,0xfeb22ffee9a5dbbeULL,0xe7fefb2ff3bdddbaULL,0x48ff5f5feb2dd52eULL,0xfdba5feba9b9dbaeULL,0xd377df72fb95d32eULL,0x9316b2954180c11aULL,0xf959d309f0a5d9eaULL,0xe7fc89cfe1a04b0eULL,0x9eb7efcbeb3473eeULL,0x95fd26e88b85c7beULL,0x69c6f77aeb00d9aaULL,0xfbc4ffffebb5f7feULL,0xcffbffffebbddffeULL,0xafffdffbfbbdd7beULL,0xddf7feedfb15d376ULL,0xb98cb5705391d30eULL,0xc3b7d358eb04d17aULL,0xdcbea284110c76aULL,0x9e03fa5b832c7f28ULL,0xb6bf7feecb35fffeULL,0xdba2fdf1e321d36aULL,0xce922f70da00c12eULL,0x101e88050105204ULL,0x8b50df598bb9db9aULL,0x16b7d6bf59305338ULL,0xd3dc74416304d32aULL,0x93992f2ccb34f30eULL,0x142fe3073046122ULL,0xe5ffd59f1a5d30aULL,0x8eaa5f996314433aULL,0x81c03620c110d204ULL,0xfb967dc4cb21eb60ULL,0x6801928052004528ULL,0xce4a3a1fcb04d9b2ULL,0x37aacfdffb25d1aeULL,0xc19c36b9c904f92eULL,0x77fe74f0fb95d39aULL,0x8990e702a914592aULL,0xcd832787104c12aULL,0xa6bf4f628b11d72eULL,0x70d15e43d33ccb3aULL,0xb1bdaf766b04d1baULL,0xfdbcfb67db15cb2aULL,0xc07bff0fd934f72eULL,0x217fdfd28b14d33cULL,0xbcf9d7e86924d73aULL,0xad041f004814836aULL,0xd11066625810d728ULL,0x4c8bc6082304453aULL,0x2f545d806924c7aaULL,0x46bfc0675985c06aULL,0xc118b800892042eaULL,0x86142a45c020c882ULL,0x8ccbfa0ef190c37cULL,0x140014101104200ULL,0x202840844080c008ULL,0x80c4b76d638cc122ULL,0x114b4304004c1a8ULL,0x0,0x10918004000c10aULL,0x2010104242008102ULL,0x4d312a0882052e2ULL,0x22082040000108ULL,0x0,0x8040101b6300c102ULL,0x5fbedf51810457eeULL,0x949406f0d0305b0aULL,0x31a1f20d0105b22ULL,0x8100212441845000ULL,0x4002000000000102ULL,0x480c0000000410aULL,0x100102000002ULL,0x41280420c000f0e0ULL,0x8000010000804000ULL,0x8171c208c900c306ULL,0x11108201040108ULL,0x243213a48800932aULL,0x4120148003200260ULL,0x2044000000102ULL,0x45d3f630a39cd3aeULL,0xebff3d4af2a1f96eULL,0xb43ade73c895d7beULL,0x822419008000d822ULL,0x400430004100c000ULL,0x89e5d0e26884813aULL,0x2ac0808381105182ULL,0x9845820220c31cULL,0xc560e478c325c908ULL,0x43a0560e3045526ULL,0x3314c30cc1a1d11aULL,0xc0080850008000ULL,0x21a1f5ae1044922ULL,0x410468e0100b08ULL,0x240080001000ULL,0x5a56472e8115d3e2ULL,0x244820a4000d00aULL,0x2800cc5a0b80c112ULL,0x748052e0c080d4a4ULL,0x51ed3760e134c3eaULL,0x8010220760118208ULL,0x1443220882107312ULL,0x2bd99fc20304c5aaULL,0x105b1b6bda10521eULL,0x8c4041e0005122ULL,0xc888294c7810c30aULL,0xc058f80a80241122ULL,0x1d5002004624ULL,0x109349088114402aULL,0xc004080000300200ULL,0xb6a47448d320d1eaULL,0x8928904b49045302ULL,0x2e1117c9f390531aULL,0xbbbf4fc6f9a5db86ULL,0x5b662a9db0055a8ULL,0x13007d008220472aULL,0xa84830e8124d322ULL,0xe08b3149c8344326ULL,0x833ecd70f8384396ULL,0xd0b076e80b34d7e2ULL,0x804000a104202ULL,0x8200010440008012ULL,0x0,0x0,0x14040001000ULL,0x0,0xe410040080040126ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x400000000aULL,0x0,0x0,0x40080200000ULL,0x0,0x0,0x1000020000c000ULL,0x0,0x20004000004002ULL,0x0,0x0,0x0,0x1000000480200000ULL,0x0,0x0,0x8100000040020ULL,0x0,0x0,0x6cce7de0e904596aULL,0x5ff673d7cb14db2eULL,0xf7f4f21cc194d32eULL,0x2977d290cb24cb0eULL,0xfc73556aebb1d6aaULL,0x9127a44ad130d30eULL,0xc7952530db38c204ULL,0xa85c5628c325c19eULL,0x6113de80c104c92eULL,0x29042625b104202ULL,0x24c103004102ULL,0x720020124000d00aULL },
	{ 0x4133f32382c6bbe2ULL,0x4786ec0002caf762ULL,0xffffffff8bd7effeULL,0x7fbf5ff28ad7f3feULL,0xd4fdfe6a8adff6a2ULL,0x9b7ef771cbcff7a4ULL,0xef9bf77f89dff774ULL,0xcfd7feab0bdeb76aULL,0xab9affdb8bc6e732ULL,0xf6e3fdecca56ffeeULL,0xffffff7fcbdfffeeULL,0xfedf7fff8bdff3faULL,0xffdfeffe8bdef3feULL,0x6f5edb5f8adff772ULL,0x3dfadbfb8ad7f7e2ULL,0xdb5ffffb8bc7f3eaULL,0xfaf77f7f8bdffbeaULL,0xa9b7d3a58ac6e368ULL,0x33dfffff8bdff2f2ULL,0xbfbbdffbcbdffbeeULL,0xdd7ef77b8ac6fbeeULL,0xffde7f778adef378ULL,0xcbd6fabfcbdff3f6ULL,0xfcefbedd8ad7ffeaULL,0xffffeffb83dff3faULL,0xf113f6e5c3dcfb2eULL,0x49bebdfb8ad6b362ULL,0xbbd7e32dcadefa6eULL,0xeed7fffb8bdef3faULL,0xd7bfddff8ad6f7f6ULL,0xf5f4eefe8bdef3b6ULL,0xffbefdf309dfb3faULL,0xdb8efe6e0bd7f748ULL,0xede7ffbd8bdfbbeaULL,0xe75cdeda8bcdfbb2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x960662201c40248ULL,0xeaa02d5800d4801aULL,0x8750a00400c40218ULL,0xb2f7b00540202ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xae1fdfdacbdeb6eeULL,0xfeff77ffcbdfffeeULL,0xdf75fe6e89ded362ULL,0xfcbc3f7b8a56f2a2ULL,0xeff7ff3e8bc7fafaULL,0xeb5ef75f83dffb7aULL,0x7ffdff7cacff3a4ULL,0xffffffffcbdffffeULL,0xfbfef7f1825bf3e8ULL,0xc187fbe88053f2eaULL,0xeaf7de39025ff7e2ULL,0xbfbdcffb8acfffe6ULL,0xb4f5f26b8acef7eaULL,0x5f7cfd72cbd6f3e6ULL,0xeba4eba7cbd6fa8cULL,0xc5f7fb9bcbd7f7ceULL,0xe6b6df7182c7fb28ULL,0xbcb67fef8adff7eeULL,0xfbbc7f738bdef7feULL,0xebbeff7f8adeffe2ULL,0xabddf72f88defbeeULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1474509a828af184ULL,0xde7fa77ac3d7fb6eULL,0xfffe3fff8a5ff7e2ULL,0xe5e7d7698a56bbe2ULL,0xabde7e5f8ad7bbeaULL,0xbdaa7ff78b5afb6aULL,0xfdbc9ffa88d7bfeeULL,0xcffc75618acebb2aULL,0x88c7fbe0c2daf166ULL,0xefc633db8a57f3e2ULL,0xdf9edffb8b5fffe2ULL,0xf0ffb6694adfb3e4ULL,0xfb9a0d7a8ac7e270ULL,0xfbffff7f8bdfffbeULL,0xeffffffb8bdffffaULL,0xffbccf72cacff7eeULL,0xfcfef7eb0bd7a7eaULL,0xbbf6fd638bcfbb26ULL,0xfab7bfe6c9d6f7ceULL,0xcb521e014a45f2c4ULL,0x6756dfc208c6f3e2ULL,0x56f0f4f280c2a3e0ULL,0xddeabfe28ad7bbb2ULL,0xfabc2f7e08dfd32aULL,0x4cb424888c45280ULL,0xa359df5a88d6f27aULL,0x12936c60c8c4ea24ULL,0x7bad356988c6f220ULL,0xda862fee0ac6a2e2ULL,0x81c3fb2b8ad4f0a6ULL,0xcffe7a9f8ad2f3aeULL,0x9cba4c6a8ad7eaa2ULL,0x204307788084fa34ULL,0x4de0755108c4b362ULL,0x8134190c801afb60ULL,0xcd5f7a8883d6b3c8ULL,0x412c8ff7c0c3a2e6ULL,0xb676f779cad7b3f6ULL,0xffbefdf3cbdffbfeULL,0xeb40cb3e8bc3b3e8ULL,0xebf69e33c2d67be6ULL,0xecbfdff5c2d7f3f6ULL,0x18d2e2e08b83b6c0ULL,0xc4a4a861c85c320eULL,0xd986fffd4adef2a6ULL,0x876afe1180d4fb42ULL,0x2126cb81cb866374ULL,0xb0108240018e1300ULL,0xf2b046428115f394ULL,0xebb80f2a8a4eb1eaULL,0x50ad10888425268ULL,0x2b94cc020044a260ULL,0x64b666e100dbb682ULL,0x962824528205a360ULL,0xcb2e0d0e8a44a240ULL,0x8306efec804ef7a0ULL,0x8362d408802f720ULL,0x1f20de5c830aff00ULL,0x80518fe888846200ULL,0x1a882c02c0559a14ULL,0x0,0x52162080099840ULL,0x280002000ULL,0xc14d66988c48926ULL,0x1024004000022180ULL,0xb184e02008d48208ULL,0xa020800001481240ULL,0x916052a0047a060ULL,0x9252ec0850b3e0ULL,0x9116a4700bd4f210ULL,0xd102030080008824ULL,0xc00002002000ULL,0xa080180d00004ULL,0x440000000008200ULL,0x0,0xcb1632050ace1702ULL,0x4144800000508100ULL,0x85084810880b060ULL,0x82500080401008ULL,0x40000008100ULL,0xdc920a3480067622ULL,0xc9c2fdb182dfba22ULL,0x207a3c84a979204ULL,0x489241608200aa60ULL,0xd5cf36698ac4fac2ULL,0xc3023c100882f220ULL,0xe1b1f2e382c7e3acULL,0xaede3618cac6b344ULL,0x2008652081a9220ULL,0x9602827802527020ULL,0x492e29798b55a668ULL,0xb0002b020052d720ULL,0x205212f800878062ULL,0x18a084880849220ULL,0x1503108a0ac08208ULL,0x80913510024c9620ULL,0x9a9def6c0356b36aULL,0x10475a000807b222ULL,0x6dbfcac08adef2b4ULL,0xd436fbe0c2cef236ULL,0x1b2845b3c244ab06ULL,0x50609c008447220ULL,0x580900008c02220ULL,0x1200928602c5a200ULL,0x141b21224048702cULL,0x4000001000008100ULL,0x4802270002529208ULL,0x8382d300884e720aULL,0x208082a140828704ULL,0x102481c0501004ULL,0x4008f4080a442228ULL,0x0,0x83f2236ec844f126ULL,0x40b21d400096b092ULL,0xbedfddfa8acffb6eULL,0xf772b2740ad6b22aULL,0xc38631a302ceb102ULL,0xe1f483368b5cf226ULL,0x9c778008845f60aULL,0xf6eedee088cd9aa2ULL,0xc0f276e2c2d1b244ULL,0x8108e41280c4e102ULL,0xc100801280025820ULL,0x0,0x0,0x2000002000020200ULL,0x2100000008020ULL,0x1004000000008100ULL,0x0,0x0,0x800010000400002ULL,0x0,0x0,0x0,0x0,0x0,0x40020000020200ULL,0x0,0x0,0x0,0x1001000000040200ULL,0x0,0x0,0x2000400000005000ULL,0x0,0x0,0x0,0x880080002000ULL,0x0,0x140000002000200ULL,0x2000800000008020ULL,0x0,0x49a833020887f380ULL,0x39be5fe4c8debbaeULL,0x61c1a212820c9222ULL,0x83df74983d8b32cULL,0xaf974dea81def2e8ULL,0xd191af5a8a54d666ULL,0x884052000d21240ULL,0xe082014a00a224ULL,0x671d7051025a32a2ULL,0x140bcc43081eb222ULL,0xa24004e202d29260ULL,0xc2a024908202f280ULL },
	{ 0xcbf6edaa4cc7be30ULL,0xabd4fa02cfd3feb6ULL,0xfeebffffedffbfb3ULL,0xff9ecfffcddb7fbcULL,0xf49e7feecdd33eb3ULL,0xffbe7f62cddfbebfULL,0xff9effffcfdfffbaULL,0xcfc2fe3fcdd3fe93ULL,0xffffdfdfeddbff93ULL,0xfcfffef8cfe3fbb2ULL,0xcdb7ff59cdc7febeULL,0xffb07fffeee7af93ULL,0xffe7ff7fcfe3eeb6ULL,0xabffffdfefcbbf3cULL,0xfedfdff7cdf7fe31ULL,0xffeeef7bcdd77eb8ULL,0xfe967feecdeb3ebcULL,0xc5cecf3fcee6ffbcULL,0x2feedbfdcdf73eb7ULL,0xb7baffffedeb6f9aULL,0xf97fb77becdbfe31ULL,0x77be3f764cc3de39ULL,0xcbfdf3ffcffa7bb2ULL,0xafffdfd9c7d7be35ULL,0x5ffccf3bcfd37e92ULL,0xf4ffd6f4cdc3df9aULL,0xd7eeedf1cdd23e9aULL,0xfbb7d33fedeb7ebdULL,0x4cb3899244a2609ULL,0x2fbfcff76fc77ba5ULL,0xf1cc9fefcecbbe34ULL,0xbeba3f57edd77b92ULL,0xddb8fb7fefcbbb3bULL,0xdfffffb7cfcfffbbULL,0xe9411086cec18c10ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x9000000008000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xab4b10da4caa3eb4ULL,0xb6f6d7efcca7beb4ULL,0x9323bf38eecfae9dULL,0x7ebf2ffdcdeb2bb9ULL,0xefe7ffb6efdfdfbbULL,0x6f777e584dc3ca93ULL,0xec8edf7bc6c7fe1aULL,0xffffbfffefcfffbeULL,0xf7f6fc6d4febbf9cULL,0x89f4d17eecde3eb1ULL,0xcc935e03e1931a14ULL,0xb49ac7eecd8b9e90ULL,0x95b6d06c4e036e3aULL,0xfb7e2dfaccc31eb2ULL,0xf397f33dcee71ab0ULL,0xc6fb969a48531abfULL,0xf3f4ffecdd35fb1ULL,0x7fffdffbcfdf7fbaULL,0xfb7fbd61cfe37bbeULL,0xfca25f6cc4d3be34ULL,0x47d29cb204db3a13ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4010002008000100ULL,0x91a9a752ccc35eb8ULL,0xffbe2fffcff3bfb4ULL,0xdf56d39a0fe3ff98ULL,0xeffecedfcef7ffbbULL,0x9f7af7ffcfdbff9aULL,0xb9ddfefbedfbbe9bULL,0xde9265d1cc476e13ULL,0xabdffffecde3eb99ULL,0xedff16d3ccdabe1fULL,0xf7ffcffbcfc7febdULL,0xfcb792654dc3deb1ULL,0x6c3ae5f2cea38f34ULL,0xebfef77ecff7feb8ULL,0xfefbffdbeffbfbbbULL,0xbfbfcfffcff7ffb3ULL,0x74f376d84df3ba91ULL,0xb70a3d628fe72f92ULL,0xfb95e7f7cfcf3bb3ULL,0x845a72094dc26b16ULL,0xf37f5d0aece73b1dULL,0xc4ead5d087c33a18ULL,0x8b4dcf614dc22f1cULL,0xfbab3f7acff3bbb2ULL,0x4106f60249430f01ULL,0x81c5c184cc22eb0ULL,0x209267ed4c583e21ULL,0xb179f769ccdabe9aULL,0x9a9aa7248ce3ab10ULL,0xd96f3fecddfbab3ULL,0x8f5b9b7f8df3bf33ULL,0x846dd01820e10ULL,0x5954a5420c82ae92ULL,0xa0102551c8901930ULL,0x8b40203bc883a390ULL,0x668abc502c062b82ULL,0x86d7cff5edc72f1bULL,0x54f776fa4ecf3a96ULL,0x7ffe7d73ede77fb7ULL,0xf390e73a4ec73a00ULL,0xe64b1f5bcec33e96ULL,0xc3c4d23ecda9b98ULL,0x9c9c46fcc4c31a3cULL,0x251e8e208dc70b88ULL,0xd895ec22c6c1a624ULL,0xd9e2ba786fa63b83ULL,0x2b7f5ec7ccf25bb9ULL,0xf4b876638fe73ab1ULL,0x9163074088620211ULL,0x8800010048000500ULL,0x948f12881c11a10ULL,0x1a6cc9584cc30eb1ULL,0x41a54e60cd2b888ULL,0x530c821801c10a00ULL,0xe0b31d0001426a08ULL,0xf984da01c0423a20ULL,0x17e768465c2ba28ULL,0x961a4d4084c62e00ULL,0x40a50444804800ULL,0xa065440c6826800ULL,0x81a0c21444412e20ULL,0xa44e90c20c8b2e08ULL,0x210620800c29200ULL,0x4010801001020808ULL,0xa24047047820800ULL,0x82816a064c823aa0ULL,0x43720068020282ULL,0x1cacddab24c2a239ULL,0x404002e04c22200ULL,0x828017620dc30a08ULL,0xf00609000d111908ULL,0x1800830044810228ULL,0x200008a44802000ULL,0x1482120048000820ULL,0x80022140c0921000ULL,0x1202030448800302ULL,0x406201000025808ULL,0xa04845524c9034a0ULL,0x0,0x8004056080525a10ULL,0x36800d700ee10288ULL,0x8d80b310c8c10a82ULL,0x95df6e1a4dd03e80ULL,0x80884ea34e002a11ULL,0x11348c3c8e907900ULL,0xd3be27744c426e30ULL,0x2e02a08400290ULL,0x204a6a1b09471b10ULL,0x79ccb204c80123dULL,0x4c5dc3084eb21998ULL,0x4c206d50cd820910ULL,0xcb038b6ea8430ab9ULL,0xe026a4000e0202b0ULL,0xa000c0044820408ULL,0x3524166684c30601ULL,0x438afc1068801211ULL,0x42106708c852130cULL,0x44cb200040c08a20ULL,0x1580a8224c03a80ULL,0x32b0666a8cc3ae99ULL,0x804a274808858f01ULL,0x8084340400454a10ULL,0xc0004002000ULL,0x62ac1a09c4c30014ULL,0xc8ba4de0c4c33290ULL,0x9014060c28420311ULL,0x9d1c037c8eca2a00ULL,0xc354c134ccc3eaa8ULL,0xc52021b24c0a220ULL,0xc81200a80c823880ULL,0x94141e48e8020331ULL,0x800380a0020200ULL,0x6854000048800114ULL,0x2d5ef9ca6cc62e90ULL,0xbb3a6d736fd33ab2ULL,0x1465e044c8c20a90ULL,0x6f3e04f08d833b92ULL,0xb05027248c420a83ULL,0xc1422a1045830e08ULL,0x9ff7ddf24dd23eb8ULL,0x3d7cdac049f3ffb3ULL,0x92202d614c83a704ULL,0x800808000080ULL,0x60001000800ULL,0x0,0x0,0x93408dd00c834f1aULL,0x80004000808000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000240800000ULL,0x0,0x0,0x400200000022000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x30004020000ULL,0x800c80000020212ULL,0x6000000020010ULL,0x0,0x2100040000020ULL,0xfff72e5a4dc3bf91ULL,0xc5c6e73bedd03f1cULL,0xaf7ef4db49c37b9aULL,0x346f56e7cd8aad80ULL,0xdb3ce75bccc36f90ULL,0xde1e1f7e4dc2fab4ULL,0x44a21a8c5020e08ULL,0xab58b24ae8c24b15ULL,0x10186d4bc8c11f04ULL,0x93a0206325c18a10ULL,0x528e1f644cd31b3aULL },
	{ 0xcadcf39eb25214cfULL,0xfbffff7fff7f97dfULL,0xdffbff79f77ab7ebULL,0xbffeffff3756b7fbULL,0xdff77feb2a7e11dfULL,0xffbfe777eb7f17ebULL,0xffffffffff7e9fefULL,0x6feffeffbf7fbfffULL,0xabff7ddbaf7b92d7ULL,0xf6fdffeffe72b56fULL,0x9d7ff7ffef7fb5ffULL,0xff9d7f7ffe7eb3dfULL,0xefffffbfef7eb5dfULL,0xe35ffedabf6611cbULL,0xbdf67ffbef7a97ebULL,0xdb7ef778af7b17cbULL,0xdbb63ffeeb7a93fbULL,0xe7f7fbeb2b5fbfcfULL,0xeffffeff6f7e37dfULL,0x5ebf5febef7f1aefULL,0xd57d3fe5ef6a18cfULL,0xfafe677c2a739787ULL,0xe8f5f73a6f7fbacfULL,0xced35f57ae6f30ebULL,0xa8bfec6b666e1749ULL,0xd57ffe7ebe77b1dbULL,0x6f3eff62e37731d3ULL,0xd8d4eb3e6f76908bULL,0xe6c3be022e7a804aULL,0x75b146f3606215c1ULL,0xf4b7dee8337e12daULL,0xffbfef722f779fdbULL,0xdf8aff7e6e7010cbULL,0x8904a20008601801ULL,0x0,0x0,0x0,0x0,0x400200000000082ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2bb9ded8fa00941bULL,0xffffffeb6f7218efULL,0xf577bf73ea7a177fULL,0xfebe2fefef7e15c7ULL,0xdfd7fb3f2b5e1bebULL,0xaede3ecf7e721bebULL,0x26b9e7e1b24016cbULL,0xbffff7ffff7b15cbULL,0xffbebf737b7713efULL,0xabc5ea336f7311dbULL,0xe5f71f506f4215dbULL,0xbcbbc5ef777216c9ULL,0x93fabe7e227e13e8ULL,0x6f9e7d796a72bfc3ULL,0xe9b47faf6c671f9fULL,0x6cdb9b9b677811fbULL,0x6f7fdfd1ab7ab6cbULL,0xf4fffffcbe7f94cfULL,0xdbbd7ff22e7b97e3ULL,0xfbadfbfe6f5ebdcfULL,0x79d7bf9fef76bbcbULL,0x0,0x0,0x0,0x0,0x0,0x0,0xb491b2eea06315d8ULL,0xd1f975506e5a16cbULL,0xdbd6377c7a723bebULL,0xcfc7ffb33a70bcefULL,0xabffbfc86c769cdbULL,0xdfbe7ffefa76949fULL,0xfbffbffcef77b7dfULL,0x54b633702a70118fULL,0xe2f5d7e6a37a12ceULL,0x37c7be79eb5e9bcfULL,0x28e456aa6421b46ULL,0xd7fdf77aee7697dfULL,0xdffe7d72ff5a91cfULL,0xfbb7ff6efa77b22fULL,0xfff7b7fbea739cefULL,0x6fbfeff9fb779fdfULL,0xfddebe6eee7e91cfULL,0xebddbfe2ea7fb3cbULL,0xfbf6773fea7ebfcfULL,0xe90feab0a6218abULL,0xa1a8ddff464304fULL,0x5a9af77cb7473a8bULL,0x5bae356aa15a90dbULL,0xdabe253ce75214daULL,0x1006760602121941ULL,0x4b6eded1687292c9ULL,0x4877d4a01771740ULL,0xe129ff72aa7a15e7ULL,0xdebe3bffea6394cfULL,0xedf7f7996b5617dfULL,0xe87fffe3fb6318cbULL,0x4da26f3b225211dbULL,0x9a31a776207211ceULL,0x4c166571f44039ebULL,0xd056c3142646116fULL,0xe9d37fe6ba46328fULL,0x37d8ced062571367ULL,0x4075a202c850120bULL,0xd6b40df83a58934bULL,0x99009111c072100fULL,0xcdcbf851ea521fdfULL,0x7b0e6bf3a2461c8eULL,0x340408a18a021087ULL,0x19bc2d46ae708201ULL,0xbea003621a421283ULL,0x1747de0a7c72137bULL,0xdf9fcfdffa7abfefULL,0xf7ff7fef7a7fbfefULL,0x11682f70ab42174bULL,0x3090225410623248ULL,0x8691ca0a7240924fULL,0xa12ccca21529449ULL,0x490026a70101402ULL,0x13f18f48684200cfULL,0xa920550006200a6ULL,0x600000000006ULL,0xae4e20002270914bULL,0x65244800010114eULL,0x4cbc7031a8461186ULL,0x226046000521106ULL,0x8a02432c2a6e0241ULL,0x468600302c421208ULL,0x1c06a280480046ULL,0x202610a2805c0042ULL,0x8200a0008000004aULL,0x108462564421401ULL,0xec42011961501a0fULL,0x26fac72ae06619cfULL,0x118a27d8a6569003ULL,0xcac22d00a246204aULL,0xd68e0e8e287f10c9ULL,0x8c683132686490c1ULL,0x1a06cd4802411549ULL,0x210040220121000ULL,0x81028500a0703000ULL,0x800040000021000ULL,0x808a1a20080a1041ULL,0xa640041a4401602ULL,0x404404000621000ULL,0x4000a00100611001ULL,0x480411f0225092c1ULL,0xf7effa6aeb6a12cbULL,0xa95e7e8b2776b687ULL,0x12134ec022621891ULL,0x94e53e016c4312c9ULL,0xf502510a060204aULL,0x8a00632c486000c3ULL,0x40a4a6c000001ULL,0x800c4fa0225a10e2ULL,0x1522009048400207ULL,0x20121d0003581041ULL,0xe000080021400008ULL,0x9422001a0001492ULL,0x48641004040110fULL,0x34306ae02552108aULL,0x92b4a500244611caULL,0xc79c409e636890e9ULL,0xc3f29913717893cfULL,0x2719d95008621301ULL,0xb03056cc2666108eULL,0xd12c07882a54184bULL,0x10143530a460100bULL,0x8c2e28c234810c1ULL,0xa2340c872472208aULL,0x8434d221e242820bULL,0x1944044c8070000aULL,0xc6922f466e111483ULL,0xc1d7f6302a6b91ebULL,0xc74a0024121082ULL,0x632422822460287ULL,0x35cc35422a531081ULL,0x4f021405104214ebULL,0x8192800b24661200ULL,0x7d70b4b2e7010abULL,0x9fff4ff3ee76314bULL,0x9e9612d1bb6215c1ULL,0x4efe7df2e67236abULL,0x9b9aed2ee56e93dbULL,0x68200180460002ULL,0x23be0c222f42226fULL,0x71f5feeb3276915fULL,0xd22c6521b9701031ULL,0x0,0x50000000402000ULL,0x0,0x40040020000080ULL,0x80982100a2600001ULL,0x280000000021000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000100000600000ULL,0x0,0x44000500000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x201000021000ULL,0x200400000001400ULL,0x40080000400040ULL,0x2b7001caef63159bULL,0x2000002000000082ULL,0x4884054120609590ULL,0xc790a3172b7a16c3ULL,0x65defe287b76150fULL,0x2bfa59d13b6a1049ULL,0x34eadd4c216216cbULL,0xe9fea75aea7eb7cbULL,0xf29e3fb02a52b5dfULL,0xc5e7c3266663914fULL,0x6edb1e9b6e72312fULL,0x120000000400040ULL,0x9dbc350222471180ULL,0x7fb0bd7f3e7e18dfULL },
	{ 0xdbe8b568483dccf0ULL,0xfbffff7fea7fbfffULL,0xfefffeda887ebfffULL,0xcbbedf77ca7ebff7ULL,0xfd77feee8b3edff7ULL,0xfff6ff77c97efef7ULL,0xffbff7ffeb7fffffULL,0xefffffffcb3ffff7ULL,0xabffffdb4b3feafdULL,0xf6ffffffeb3fbeffULL,0xcbeff7f3c97eafddULL,0xdbf4ffff693cbfdfULL,0xffffffbbcb7fbfddULL,0xffffffdfc97fbffdULL,0xb4f3ffe3e93fbfffULL,0xf7ffbf7feb7efefdULL,0xdebfe7ffc93ffff7ULL,0xefdfff7fcb3efde7ULL,0xeffefeffcb7ebeffULL,0xb637ffeb087faefeULL,0xd7d6ff4e693caeffULL,0xffde3f77893feef7ULL,0xe9fef3bb6a7eb7feULL,0xea73df5fca7cfef6ULL,0xaebd4eefea7d3eecULL,0x97fff6ff8b3ebcdfULL,0xefe86c63483cb0dcULL,0x39dc6e2e883c2af4ULL,0x84c3a201681c81c9ULL,0x2185b11683c0dc0ULL,0x54f776e86a7caee2ULL,0x9fbfff77883eabf7ULL,0xdd2c2c48083585d2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1e4e10491e8bc2ULL,0x948bc6eb403e8ec6ULL,0xff3baf76c83c8fddULL,0xfebe3fefea3ebeedULL,0xc7fbf775ca3eeef5ULL,0x5f1e5ae74b1faaedULL,0x923bc7f6eb3c5ef4ULL,0xfdffffffeb3fffffULL,0xfffeff7fcb7effffULL,0xcbfffbffcb7dfeffULL,0xe8434b40801d4f74ULL,0x3fffdfefe23fffd7ULL,0xb6b667fc683edbe7ULL,0xfbbe7d716a3eb6ceULL,0xfbbdef6feb3fbfd6ULL,0xc5cfbf304834bec0ULL,0x843ecf62083ea615ULL,0x15b736ee823c8cf7ULL,0x9fbeff53e87f8df7ULL,0xf7b3ef7e693eafddULL,0x2fdfa2becb3dfbd5ULL,0x10008208208010ULL,0x0,0x0,0x0,0x0,0x0,0xb6931fee483cb2d5ULL,0xfbe4f768017e9fd5ULL,0xffff3ffec97fbdd4ULL,0xedf6dab74b75bc5eULL,0xbbde93df4b3c8ff6ULL,0xfdfeffeb493d9ffeULL,0xf1edb7788a3f8ff8ULL,0x8b9ef93200379dd4ULL,0xa9fef1afe93eacf9ULL,0x8b620f52083eb9a2ULL,0xbfbbffebea3efffeULL,0xd507b0e94934bdd0ULL,0x7ef63d734b3eebd6ULL,0xabdde26ee87dafdcULL,0x88c3f7c24a3e82c5ULL,0xb3bfcff3c97eebf5ULL,0xdd767f7f4a7dbefcULL,0x95baaf50cb3cbef0ULL,0xcf92bf2f883fafffULL,0x58567abac23cccf6ULL,0x2a00f5900934ac83ULL,0xb6b954ee697caac0ULL,0x5d668572e9579a42ULL,0xfa3c67666876a5d7ULL,0x83e5a70a411c4e14ULL,0x148918008143004ULL,0x1c307103422c8980ULL,0xb77aa771c93eabe5ULL,0xf6b6a75e497ea8edULL,0xefd7ef9fcb7faeffULL,0x2003ba99417c8c81ULL,0xb60607f50a2c81d9ULL,0xe16d346a097c8cfdULL,0x94963de1887caffeULL,0x8bc4e332411e9cf1ULL,0xafcfbee8c83eeeffULL,0x8582191b4a34a3c6ULL,0x97b2fb51493c9ae4ULL,0x14a82d70483dabd6ULL,0x4144011a01080810ULL,0x80009610411c2890ULL,0xef9f5fd44b7d4ff4ULL,0x65f6b4ea405e86f2ULL,0xa0343c30003690c1ULL,0xd504ef6c483ca3fcULL,0x1d61604e0115c40ULL,0x969dfd3c83e9dfcULL,0x502e4660c11689c4ULL,0xf1ac677ae91daed7ULL,0x7b9574b00a762954ULL,0x8081080240460000ULL,0x4c9e100143048c90ULL,0x94143ec00826acc6ULL,0x9a042608011c0844ULL,0x88520040092c0800ULL,0xc9c2f620403d0674ULL,0x28bd003001e82c3ULL,0x2421410080108a10ULL,0x3055017080158ec0ULL,0x282c356c410e8941ULL,0x4a006102ca048841ULL,0x1c01e0008248cc0ULL,0x4a1e7780856b005ULL,0x2434263088300854ULL,0x46902070881489c0ULL,0x8300cb08087c0054ULL,0x34716000854820eULL,0x2028400008300260ULL,0x11801463017c08d0ULL,0x803629424817a0c0ULL,0xd2362b648a749ad2ULL,0x8dcec2134a1e2ec0ULL,0x2000901000110040ULL,0x0,0x8000040040160000ULL,0x8190021601248c01ULL,0x6004600600303040ULL,0x80411060200040ULL,0x4048169082c8a94ULL,0xa144a20280108ed0ULL,0x415580a14040cULL,0x69c2a00a082da072ULL,0xc4e5b5a483eead0ULL,0x8092cef9003cacd1ULL,0xdc707761c82e0bd8ULL,0xca946c71083c1e34ULL,0x0,0x66000000088012ULL,0x8030040240140050ULL,0x14be4028880c8e9aULL,0x86001200840ULL,0x80040360401080d0ULL,0x88522e03c82ca84dULL,0x0,0xf08410214a340a08ULL,0x931d152001748844ULL,0xf6029b68ca3ca814ULL,0x2966bf21411e4a74ULL,0x299bdb51083c97d2ULL,0xc49673f9e97eaedfULL,0x42e7f760683c82d3ULL,0x200010008000040ULL,0xc8478321001d81d2ULL,0x800c1417086cc810ULL,0x2304f4080108c45ULL,0xf9d487b8483c82d9ULL,0xff162f340b3daed3ULL,0xcc0a26030811115cULL,0x2ce61289685c8799ULL,0x0,0x20020000200040ULL,0xe0105d44411e8c45ULL,0xeb66a92b421c9696ULL,0x484b3711083c82d4ULL,0x38ac8141882c8c80ULL,0xd8fef676083ead54ULL,0x1eb20570093e8b93ULL,0xc38ec00cc03cbbf4ULL,0xedc2bb9ec83cadf0ULL,0x21344b60003c0084ULL,0xf574fee9893ebef5ULL,0x2a4056241360c40ULL,0x4008000080080ULL,0x0,0x0,0x0,0x44020802141000ULL,0x0,0x0,0x0,0x0,0x0,0x81000001000800ULL,0x10200080040ULL,0x0,0x0,0x8010000000080020ULL,0x0,0x4000400000101000ULL,0x0,0x0,0x0,0x0,0x0,0x21001000800ULL,0x0,0x0,0x200040000101000ULL,0xbb266be3c33c0ce0ULL,0x343920a0917ba41ULL,0x36508340e8100a44ULL,0x14a9010003140a40ULL,0x7b8e566a483c1ef4ULL,0xfe6eae72483e8ff8ULL,0x49c6fc02081c92ddULL,0xaf5ccd5b093cbad1ULL,0xbcbbeee90b3cafd3ULL,0xf3b8bf7f483e82d7ULL,0xfeb22fdfe97eaff2ULL,0xc9eee2aa0b3caa92ULL,0xc3ced9474b3ea8d1ULL,0xb7e2477b493cafe1ULL,0x2071b779693c8cd4ULL,0x694afe7e493cbef5ULL },
	{ 0xffbfffffff9bdf5bULL,0xfbffffffff9bdfdfULL,0xcfefffffef9bdfdfULL,0xbfbfdfff3f9bcfdbULL,0x74d7feeb3a9bdf1bULL,0xbfdeed70b59bcf5bULL,0xfbbcfffe5b19df5bULL,0xffffffbb7f9bdfdfULL,0xbf7fffdb3f9adf1bULL,0xf5b7feec7b99d71bULL,0x6dee77ea5f0b5fdbULL,0xfbf7fffeff9bdfdfULL,0xeff7ff9f7b9bcf5bULL,0xaf5ede9f9213df5bULL,0xb6bfffea9f8adf5bULL,0xfbffff6a7b9adf1bULL,0xffbf2f7f7f1bdfdbULL,0xc3f6ff3bdb03cf5bULL,0xbf5fffff7f9bcf53ULL,0x27be7fffdb9a9d1fULL,0xdffdfff8139adfdfULL,0xfffc7d6feb9acf5fULL,0xe3fef37eaf8adf9bULL,0xfed61f5a8f8bcfdfULL,0x9bbc4ff33b0acb12ULL,0xffbdf67d7f9bdfdbULL,0x4f3eff733f1bc7dfULL,0xcbd4e6ae060a4f5bULL,0xae67080a080bdd93ULL,0x605400083440aULL,0x4412aee80700471aULL,0xdf865f60578ac71aULL,0xf88e0d6a0b8bdfdbULL,0x0,0x0,0x0,0x0,0x200100000081ULL,0x0,0x4000404004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8b3c4c065f18470bULL,0xb7b637eaeb82dfdbULL,0xf97efffa7b9a0f5eULL,0xfefe2fffb39adf5aULL,0xcdf3beb5579acd5fULL,0x21c61a9241084b1bULL,0x400102801004010ULL,0x79baaff2778acf5bULL,0xc7ba7d737f0bdf5bULL,0xe9f7f33fdf8bdf5bULL,0x6bc64e114d02cd0bULL,0x78aef77a231acf17ULL,0xdf68aff1238bdddbULL,0xe7be2d71ce0a4fcbULL,0xebfeff76cb9acfd3ULL,0x9cffff7b7f9adfdbULL,0xa39ecdc93f1b4d5aULL,0xfefff7e9df9bdf5fULL,0xb3b617704f03c15fULL,0xdb9fcf79df03d757ULL,0x4fdeffb6bf8bcfdfULL,0x0,0x0,0x0,0x0,0x0,0x10020000020002ULL,0xbdbb7fff1f9a5f5fULL,0xfbe5a7788e0b5f5bULL,0xffbf2f67d79acfdbULL,0xedd7ff663b1a4fdbULL,0xa17ff6ffbf9a5f5bULL,0xb6bfeffb939a5f5aULL,0xd3f09f52451adf53ULL,0xdbbe7d7d079bcf53ULL,0xabf1c02b0b9a0750ULL,0x66943cf853005f1aULL,0x331e0dc20701d310ULL,0xf6f6776b2e9bcf5fULL,0xeaba2df03e0b5d58ULL,0xcbb0e53f3b0b5b5bULL,0xe6f2b69b4f0acfdbULL,0xffbfefffdf9bcf5fULL,0x35febfed7e9bdf5fULL,0xff1eed77469bdfdbULL,0xfbb7feeebf9b5d13ULL,0x44c62e1c0e924f19ULL,0xebbdddca960847dbULL,0x7411f4209308450aULL,0xbffeb773dd8b5f5bULL,0x48b9073e5b18df13ULL,0xcdc3f009a788d71bULL,0x223114c649020343ULL,0x3cae7b6a130a475aULL,0xfffbbf7f5f99df1bULL,0x588e3f7c7e0ad797ULL,0xeddff72edb03df5fULL,0x29c75fde730bdb16ULL,0x3411288a0702c51cULL,0x616d3588161ac497ULL,0x9a86150c12014d0aULL,0xbc6a22a770acd43ULL,0xeeff1e79178bcf5bULL,0x573ac7d317825f1bULL,0xc40c301a63114581ULL,0x53cc3c60810a9f5bULL,0xc881c53603920446ULL,0xd832a1130089d11ULL,0x19a48b2b13084b03ULL,0x10a0044402008301ULL,0x2010050001000203ULL,0xda32474e1f03491bULL,0x4cc2df110b8a010aULL,0x11844110004d01ULL,0xb4afd06c179a4f11ULL,0x1002e1080100500ULL,0x540008580100870bULL,0x8cf12a8132128f0bULL,0x2a911e521a025d50ULL,0xb4a75020160a4d4bULL,0xdb7e7f73db9a8f0bULL,0x8870760400a071bULL,0xc020100000020b05ULL,0x4432a484802094aULL,0x1b84d0310125c50ULL,0x1015420018a0042ULL,0x812e2c4000020a4bULL,0x9028d280f004302ULL,0x4d5a2e65079a441aULL,0x908564f86b089113ULL,0x945596180103cd1bULL,0x28262c601208cb13ULL,0x8980e34f0e08d71bULL,0xc0022000002000aULL,0x8284cb600a120750ULL,0x181306f06800c311ULL,0xab0434320602464aULL,0x8a02277802085f1aULL,0x180480000004c40ULL,0x83814421482c311ULL,0x80a8526900020bcbULL,0x813c0000200a0140ULL,0x6010052604824609ULL,0x40831300100010aULL,0x2001108000014408ULL,0x344010a422094189ULL,0x5b68254213014310ULL,0xb090474603000542ULL,0xcb8ecb2ebf0ac5dbULL,0x68f602cb0102dd1aULL,0xb5b8d7731393dd52ULL,0x21ac90720a024f50ULL,0xc800ed33118ac597ULL,0x88c00100001c19ULL,0x842200000020801ULL,0x5400442014110ULL,0x185426490103c915ULL,0x5206055017804743ULL,0x8900e2256110c512ULL,0xc16312210308070bULL,0xad01e021098ac101ULL,0x844454000110451aULL,0x8d071500a180701ULL,0xe1a09e27208a5d18ULL,0xcdefbe3f238b4fd9ULL,0xbf7fcfd76d8bd7dfULL,0xd497f6f217025d1eULL,0x8164745f1b0bcf1fULL,0xcc381710890a4f53ULL,0xc05dc20c2111c513ULL,0x2b563c5c0a02c751ULL,0x1107d30e084312ULL,0x990d745c600b4344ULL,0xb3349774928bcf19ULL,0xc948f6aa18880f0bULL,0x64544e4d25185753ULL,0xb49042e103094a91ULL,0x409610008284c3ULL,0xf48e5d613f92575bULL,0x184c006200ac941ULL,0x2cf7ff4f1f8b5fdbULL,0x2b1a6d997d934d42ULL,0x817be778330a8f5fULL,0x4f367d71fb8357d3ULL,0xb914d30ebb824f13ULL,0x895212860282855bULL,0xa57edffb6f9a4f5bULL,0xbcf3b7ef7f9a4f5fULL,0x9bcab7004382cb53ULL,0x200200010000100ULL,0x0,0x0,0x1080000000004800ULL,0x8030040002020810ULL,0x8200000000000002ULL,0x0,0x0,0x0,0x0,0x0,0x80100000100400ULL,0x0,0x0,0x0,0x0,0x802000000440ULL,0x42001000004a10ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000400a02080401ULL,0x95a90800a004719ULL,0x210c8830a080302ULL,0x4420049012024840ULL,0x4bceff627b8a8f4fULL,0xf1302f727a1a0f1bULL,0xc5c6ef3b1b1acfdbULL,0x6fdf58d1be8b5f5aULL,0x96bf7feb5f8bc7dbULL,0xf7f9bf6b9f9bdfdbULL,0xdea4156e669bcf5aULL,0xc9d0f737c892d75bULL,0xefa1c5b0693cf52ULL,0x1a8a07d352180753ULL,0x407c0642130b4f57ULL,0xf6963dd00e1bdf58ULL },
	{ 0xfffefffafdf9ffe6ULL,0xfffdff7ffdb9ffe6ULL,0xeffffffbfdb97fe6ULL,0xffffcffefdf97fe6ULL,0xdfff77ef3df97fe2ULL,0xffbeff7295b9efe6ULL,0xdf9ff77f7cd9ffe2ULL,0xdfdbff9ffdf9ffe6ULL,0xbbfedfdffdbdffe6ULL,0xd67e3fe854b94ee0ULL,0x9f2ff76a34a8efe6ULL,0xdff47f7ffdf96fe6ULL,0xefe7ff9efcf4ffe6ULL,0xa6fd7ed798e8e4e6ULL,0xbfdedfebfcfdf7e6ULL,0xf73ebf5a7cf97fe4ULL,0xdffebff7f5bdffe2ULL,0xeffefffff5bdffe6ULL,0xbfdffedb3dfdffe6ULL,0xffbb6ffef1fd6ee2ULL,0xb17cff7b95fdfde6ULL,0xdffeff7cbdb97fe6ULL,0xcbf1f3bb5cbc6de6ULL,0x3f570fd94db8c6e6ULL,0x98bf6fef55bdf6e6ULL,0xfdfff7fffcf9ffe2ULL,0x5b96ad7bb5b86ce6ULL,0xfbdef33e15f8fee6ULL,0xffc7bffbb5bd4fe6ULL,0xafbf5f7a75b866e2ULL,0x347e06e13db05ee0ULL,0xdffeff72bdf86ee2ULL,0xc104212544306ee0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x69449042098c9e4ULL,0x53f47f21cb965e0ULL,0xfffeff3275b9efe2ULL,0xde9737e6b5b8eee2ULL,0xdfdbd768d4bcffe6ULL,0x4c00e28508086a4ULL,0x400400000280000ULL,0xff6fb77ac4ddefe2ULL,0xdca63d61d49c67e2ULL,0xe9fff36b35b9ffe6ULL,0x63771e5354b97fe0ULL,0x1bbacff21cb9fcc6ULL,0xfd72ff6234d8efe2ULL,0xffbe2df07dbd77e2ULL,0xe324f737dcf96ee2ULL,0xcad3ffde74fdefe6ULL,0xe73fce931cf8eca2ULL,0xf5f776e7ddfcefe6ULL,0xb79c3df025f84ae2ULL,0xb28b673965f05ce2ULL,0xcd67be1234bdffe2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xffff5ffffdfdffe2ULL,0xffffef73bdb97fe6ULL,0xba3e39604df867e0ULL,0xcfffffbf34bdefe2ULL,0xdfcfffdbb4bd7de2ULL,0xd73b5ff3a4b97ee6ULL,0xd3fcbf78f5f925e6ULL,0x5b36bd225db967e2ULL,0xe3dff33a15b9d7c6ULL,0x6fff7e9ab4b8f7e4ULL,0x4a9cd49e34904be2ULL,0x94d6f2e815b04fe0ULL,0x6bbefd71f49cefe6ULL,0xbbc4fa377cb86fe2ULL,0x98ce53dbbdb9dce2ULL,0x2d7fce5195f9c7a0ULL,0xf5d47ff91dbdefe0ULL,0xfbbdef72bcb8e7e2ULL,0xfb97ff7ffdf9ffe6ULL,0xc3feff13349d7de6ULL,0xeb7f9ddffdbcfde6ULL,0x57fc76c8fdb96ce4ULL,0x99647c4354886de4ULL,0xd786efdc95b96f62ULL,0xc8dee20a68956ae2ULL,0x21a134211394822ULL,0x149d549e11b928e2ULL,0x8061a44215b06cc6ULL,0xd33e3f6755b9dfe6ULL,0xc9e6d38481b9eca4ULL,0x2be79e58dcb96fe2ULL,0xb03e0d7341282be2ULL,0xd1e5d6683cbdc6e0ULL,0xfb2e2ce95db9ffe2ULL,0x88c1a0021c905520ULL,0x4ea77e9094b976a0ULL,0x13aec57a34b87ee6ULL,0x8536a4d014117462ULL,0xd02a7dc051b96c62ULL,0xfbd0e33a7c114ee6ULL,0x1080321100106084ULL,0x2b184d18041144c4ULL,0xd52772aa11a8e6e6ULL,0x212ab56084206de0ULL,0x3004012430910440ULL,0x948160011310042ULL,0x800ecc431090c684ULL,0x22a6846014317442ULL,0x126e074051204ea0ULL,0x8280000200282ULL,0x100800000100400ULL,0x8565d0054918520ULL,0x5c045100052032c0ULL,0x1baccc720499c5a0ULL,0x282006011310002ULL,0x8070020c1c084500ULL,0x2159c80a40b14422ULL,0x89407a0149804c2ULL,0x9ce50520453104a2ULL,0x7522084e209867e0ULL,0xca2afb6a3c394fe2ULL,0x6e4115081314c00ULL,0x5010ccc1adf10560ULL,0x1080500c10840942ULL,0x482c0c2211284c22ULL,0x100802000110280ULL,0x26ebfc1f9f046a0ULL,0x69ecbdbbdb987a2ULL,0x9cde7fea85fd7ec2ULL,0x812511010882560ULL,0xd2107848300064c4ULL,0x555280000386840ULL,0x2010485c04988820ULL,0x54400028009020e0ULL,0x880c047214a06660ULL,0x100082401880c002ULL,0xc0000000280ULL,0x10a1442800046c0ULL,0xd9379a870116de0ULL,0x94486021480cae0ULL,0x8004022010004560ULL,0x51ed1bcb4b974e2ULL,0xaffefe9f3cf9dce2ULL,0x9e2a0ff255b9fce6ULL,0xf5fef736f491fde0ULL,0x43e06e510df86940ULL,0xe111c10135a441c6ULL,0xc4601440c006d60ULL,0x32004e0010290120ULL,0x10500100280280ULL,0xc29235102d3046e0ULL,0x4901930200a8e280ULL,0x88603e8111b9e400ULL,0xa801f1a34b94162ULL,0x18506690412800a4ULL,0x1086ad0214b146a2ULL,0x4b82ea33348c7282ULL,0xc2a2122100d90020ULL,0xaaae4c503095e560ULL,0xedf9fe6025b163e6ULL,0x9ceb267635b92ee0ULL,0x4004007000006680ULL,0x2cdedae0ccb925e4ULL,0x41881b1194a80542ULL,0xac74c0a31c996ce4ULL,0x4040a60084101540ULL,0x4a962cf200996aa4ULL,0x8980340210840442ULL,0x860001000042a0ULL,0xa242601485a040ULL,0x949cb053149854e2ULL,0x4814501c54204262ULL,0x802cf00615f061a2ULL,0x2a968a4120ac0960ULL,0x2d3bddb3aca0d7e0ULL,0xfd7ef6faf5b92fe2ULL,0xafbc073034a957e2ULL,0xd144f77835b96f42ULL,0x8ce32a80289964e0ULL,0xeffbddea3cbde7e2ULL,0xf6b76eefb5b9ede6ULL,0x1868ef701c883162ULL,0x2100204000042c0ULL,0x0,0x80010000000042ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000880000ULL,0x0,0x0,0x0,0x300000102000ULL,0x0,0x0,0x20001894c0000a0ULL,0x20200001200000ULL,0x0,0x0,0x0,0x0,0x0,0x14000880000ULL,0x20810000400ULL,0x0,0x0,0x0,0x83486f40501867a2ULL,0xfd1627bb1c9c4d62ULL,0xf453ca2971b02de6ULL,0x2dfed7d1fcb827a6ULL,0x9fdb976b35bc6fe2ULL,0xfbedef6a75fd7fe6ULL,0x1abe1f7575b8fee2ULL,0x42c3d33811fc0462ULL,0x22579ebb35e82be4ULL,0x6d375df234d9cfe4ULL,0xf5fdffe6b1bdefc2ULL,0xfac47d9615b96fe2ULL },
	{ 0xfffffffff6fff7eeULL,0xfbffffffeedff7eeULL,0xdfffffbf76f7f7eeULL,0xfffeff7bfef674eeULL,0xfffff7f7b4fff7eeULL,0xbfbeff77f697f6eeULL,0xffbff77ff4f6f7eeULL,0xdfffee5fc6f677eeULL,0xeafedddbec7e766cULL,0xd4f77efbae5fd6eeULL,0xf9feff6ab6d674eeULL,0xfbbffffffef6f6ecULL,0xefffffbda4f6f4eeULL,0xbfdcfbdb34b776eeULL,0x34b7fdefd6b6f3eeULL,0xfffefff6fe36d4eeULL,0xfffe3f7ff6d6f1eeULL,0xeffeffbffe7ff5eeULL,0xefffffdffefff6eeULL,0xbabbeffffef7f6eeULL,0xe57ebf6aeef676eeULL,0xfffe7f71e6f6f6eeULL,0xcb8efbfe7c5e55eeULL,0xfff6fe5bbe76f4eeULL,0xdbbeeffbfebe76eeULL,0xf5fff7f8feb2f06cULL,0xdfbe357344f676eaULL,0xc9c4531fa43253e0ULL,0x4a667e4f6416d6eeULL,0xd1f5f17bcd45368ULL,0x7596beea6ad6546eULL,0xd7feaf63b4f6f6ecULL,0xd9362fb3e452f466ULL,0x8400201000c1080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0ded192632f768ULL,0xd7e7efba296f6eaULL,0x9b7eed6ae43750a6ULL,0xee9bbf7fb6bff2eeULL,0xeff6f7bf7636f6ecULL,0x0,0x0,0xfbdca77ee6d317ecULL,0xcb8e2f6c1e96736eULL,0xe9f5e10ef496756eULL,0x6a5e7e7be4bd50eeULL,0xbdbfddebb6bc766eULL,0xdcfdd2edb6f6d6eaULL,0xeebe3df2e4d577eaULL,0xf985ffbfcedf55eeULL,0xdeef3efb66a770eeULL,0x73b6c6501607f06cULL,0x7df777ffa6d6f6eeULL,0xd7b9e553be9434eaULL,0xd392fb3aec9ef7eaULL,0xc04bf208267c50e8ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8cfb7571e47f566cULL,0xfffcfffafe7ff6ecULL,0xde9647ebb452766cULL,0xe8426a0e843d54e0ULL,0xafffffc8e436f6eeULL,0xbf32cefed6b6d7ecULL,0xfffdff69fe7ff6eeULL,0xdfe65de2c637f46eULL,0x89f2c77aa43e32acULL,0x265279cc341653e8ULL,0xa9895d214227642ULL,0x10de57e986bed06eULL,0x5a1c2db3243ef3eeULL,0xb9d6db2e9ebe14e6ULL,0x4f6bfe73f47776eeULL,0x8fafdd183c1ed666ULL,0x55fd5279a292d76aULL,0xf3befd70ee36d2eeULL,0xdfffff7ffebff7eeULL,0xcddf9649961456eaULL,0x8436ce0224127068ULL,0x44fc7be8943250ecULL,0xc9025602ac127062ULL,0x7c1e37eb3630d760ULL,0x8ce7e6bf60e714eaULL,0xe05c9c4282375064ULL,0x4044640801214e0ULL,0x4340bf008040148aULL,0x429106162c05d4ecULL,0x81a4d3653444106eULL,0xb5d10c6a0b65268ULL,0x34b2e7ea247d566eULL,0xc8341c019c045468ULL,0x9a94426294b4146aULL,0x881cfe22545c5460ULL,0x69261eb41453f4ecULL,0x25fc4cebb462516aULL,0x0,0x9fae7fe0acdc55eeULL,0xf880cba436b611e6ULL,0x1fdb3a58163677eeULL,0x7b1ed515ac3e52eeULL,0x24322064040e3168ULL,0xb2e2d40169c53c8ULL,0x80002900001510a0ULL,0xc0a3824360411428ULL,0x987cb5a345ef026ULL,0x546626400414d00eULL,0x10ea962ac1614e4ULL,0x9314222400d05268ULL,0x8200668200105068ULL,0x2a1150d1241e7160ULL,0x84233428029410eaULL,0xa570466a80145022ULL,0x2002000104000ULL,0x8000700080001060ULL,0x2358134000525208ULL,0x16905d1a841e7060ULL,0x81a9166094741002ULL,0x710a5a78200e306aULL,0x490cff2a147650eeULL,0x48671a10141470e4ULL,0x88055240134024ULL,0x100820000020ULL,0x100040002000040ULL,0x2154420000160204ULL,0x204eb34ef4226028ULL,0xbb3e4f5326b672eeULL,0x7cd43efc3476366eULL,0xc1322f72be9450e8ULL,0x8a0020340144022ULL,0x85830818034d002ULL,0x9ccd850765410e0ULL,0x167056a0484212eeULL,0x4124ac72001454e8ULL,0x2a080c0084001260ULL,0x40032a32840cd464ULL,0x3cc4d0ae010142aULL,0x41a03461402026cULL,0xb044062204905444ULL,0xba20256094345262ULL,0xe48bf0d0d414d6e8ULL,0xeedf7f5d847775eaULL,0xa3b24ffba87256acULL,0xd231a619bc37f666ULL,0xf676b7d4feb3376cULL,0x81f2080026302aULL,0x2e5a13c82014166aULL,0x8008402080027000ULL,0x1130064000022244ULL,0x41803a4060126020ULL,0x9b64033040d4502aULL,0x48120080120004ULL,0xd04480a1410446cULL,0x510c32840222f26aULL,0x8109f1e014001268ULL,0xc2042ba20414446eULL,0x4955a62f801672caULL,0x2610c100140412c8ULL,0xf4b5dee1e43454aaULL,0x914eb3381436d06aULL,0xc3b23706c254d0e8ULL,0xa177ca18801650eeULL,0x288a975ae811566aULL,0x8cde59a3e251d460ULL,0xc13480602084116aULL,0x5096181024041422ULL,0x4840000000401080ULL,0x825a1104325004ULL,0x2214120010d008ULL,0xd54ca66880d0506aULL,0x5bc43d2014227268ULL,0x28f5bbeec4bc56e8ULL,0xa63e12db8876744aULL,0x37d181001622706aULL,0xd48770718410526aULL,0x21763960a8e6f0aaULL,0xcbd1af2a249c706aULL,0x2bc7de1794df74e4ULL,0xdefffffb8cfe75eeULL,0x7fd7ffeaec36d6aeULL,0x9442c06004410e8ULL,0x1000004000001002ULL,0x0,0x2000810100040042ULL,0x1088000214020004ULL,0x0,0x800000100000c000ULL,0x8020500000041008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20000200000040ULL,0x0,0x0,0x0,0x0,0x0,0x2200000140000ULL,0x480000040008ULL,0x0,0xd1aaad50a01e52e8ULL,0x5b820a7e94325064ULL,0x9976e60e843f6266ULL,0x6ffff9a867650eaULL,0x7c3657e8b45676eeULL,0x9bbcd572a6f6f6e6ULL,0xfab63df32ed4f5e8ULL,0xc5cafbbeecb6d66cULL,0x6bdf4dd08294d642ULL,0xb6fa5ff7a436d66cULL,0xa1bdff7afc16726eULL,0xfb36777de4b7f7eeULL },
	{ 0xfdfffff7f17dbfffULL,0xfffffffff1fdbfffULL,0xfffffffff1fdbfffULL,0xfffffffff1ffbfdfULL,0xf5feffed71f7b7faULL,0xfffefdf2d1febfdaULL,0xffbfff7f71ff9ffbULL,0xeddfff3fd17dbbfaULL,0xafffdfdff09cbfdaULL,0xdef7f6efd1dea7daULL,0xcf6fa76551fcbec8ULL,0xfdb67fffd1ddbefeULL,0xffdfffb971ffbffdULL,0xfe7efed9f1feb7faULL,0xdd3fff6bf1fdb7fcULL,0xfbfffff3517d8ffeULL,0xfebe3dff717daff9ULL,0xee3fff76517f97f6ULL,0xefffffdf51ffb7daULL,0xffbbfffb51ffbff7ULL,0xdad335477150b3f0ULL,0x9a9e7d7dd171bfd2ULL,0xabd77bdff17ca7f1ULL,0xeef7dfde517d83d2ULL,0x5296cff351debf5eULL,0xb455ef7a51f43fdaULL,0x3ba87de351ffa7feULL,0xebc4f32c713ca67eULL,0xeeebb77bf151bfc3ULL,0xffdee7b517d37ddULL,0xf5f676fe515fa7ebULL,0xfefe6e707171a755ULL,0xd006a5ac50948ec0ULL,0x8c02b10a5030a240ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x808807105010b6d1ULL,0xadfe57f3717fa3eaULL,0xc07cefc05075b378ULL,0x5eb3277c515daf7cULL,0xc9f7c33d715bbfdaULL,0x0,0x0,0xf42ca5d0d13ea4d8ULL,0xd282adc850549fd9ULL,0x83e7f77f51dcb7faULL,0xbe57bfdb707fbfd8ULL,0x2dbbc95bf17dbe4eULL,0xd37fe7fe51fe97dbULL,0x7f3c6f7bd1bd96faULL,0xcbcfff3ff1bcbff3ULL,0xcff83a1f71fd3ed8ULL,0x7d9a57d25038bee1ULL,0xfcf7feefd07ebfddULL,0xc5a9af7251dd9ed3ULL,0xff97f77d5171affbULL,0xcfcfbeda71ed275eULL,0x0,0x0,0x0,0x0,0x0,0x0,0xb5fe75e671bdb6daULL,0xdbcfef7371fda7ffULL,0xccba2760205f8ed8ULL,0x8580ca20510010b0ULL,0xfafededf715eb7faULL,0xb19acffbd13e9ef8ULL,0xf27cb62d511b27f0ULL,0x8f965d7940b7bdc4ULL,0xeb3eff70513ca7fdULL,0xbae7b65a51fd8ed1ULL,0x1202070910510ef0ULL,0xbe32eb7051752f47ULL,0x412efce211508e59ULL,0xeb10d33e71dd27faULL,0xbee3bf4e51dda7d8ULL,0xefbfec7e5175a7dbULL,0xf776b6fef1ff9ffcULL,0xfff66f7371d7bfdaULL,0xfbdfefef71bcbffdULL,0xe3d6d64a514da7faULL,0x8b5759523149875aULL,0x5015d5e051100af2ULL,0xe9486f63011d0e64ULL,0x880c070c411806f2ULL,0x52cbe73a71bf27dcULL,0x251d1a10517c0a42ULL,0x42265663101c868cULL,0x914a2700d0240240ULL,0xc2863f7d713f87d0ULL,0xe1d077bb517ca7d0ULL,0x584e0787d1118decULL,0x36385fd151262748ULL,0xd8e13678516526ebULL,0x5f8e2d20d0141ed9ULL,0xcbc034b7519c86faULL,0x6838865270d486c0ULL,0x9a9b2fd9507cab50ULL,0xf566d67a71570b7aULL,0xcfb63f22501caedbULL,0x84000600200340ULL,0x4043043c90140e8cULL,0x31c8ef150fc2f5aULL,0x71db07ea5135b6d0ULL,0x2ba2800005d8250ULL,0x98b4830a51108601ULL,0x2ede0209d10086ceULL,0x1a091c0a410406c9ULL,0xa0c04010080640ULL,0x81bcaf68d11426d6ULL,0x202020400250040ULL,0x40271870040484ULL,0x802180010100010ULL,0x50201a2411081d0ULL,0x31406572418c2660ULL,0x2263c40d0008484ULL,0x4152800050080600ULL,0x2c050eda603482d0ULL,0x2421000000000644ULL,0x50c0301050148c43ULL,0xc6162573503d8650ULL,0x8220c42250648388ULL,0xfd60e00005186d5ULL,0xa0040800418040ULL,0x60d9d669512496deULL,0x1cc7d21502136e0ULL,0x4391431841ec0348ULL,0x60121623410c048aULL,0xb9eef3b71778e5bULL,0x1062e6445035a240ULL,0x32ab700f17c0f44ULL,0xc020c200109c0200ULL,0xc58ef80041dc1f46ULL,0xa842894e415c1202ULL,0x189156e9704582c0ULL,0x9d20044211110643ULL,0x8a080b024106204aULL,0x782864010100dc0ULL,0x8a2c0b4251248250ULL,0xb4020fe060942b58ULL,0x10a0052851102a10ULL,0xa0042720710886c0ULL,0xcfbbfb225075bed8ULL,0xaffffff3f1ffaffeULL,0xb6fb5feaf1acb77eULL,0x91cf34f851d5bfe9ULL,0x2632b564517d87c8ULL,0xcb80eb7151548fd8ULL,0x4412260840062042ULL,0x8010040040000ULL,0x0,0x6812f122514c8260ULL,0xc050490851040358ULL,0x6642861001248048ULL,0xa8a2c948510cb140ULL,0x85f114a271680282ULL,0x2030580040000554ULL,0x7d8df277512087c3ULL,0x952f388514587c8ULL,0x298949cb51501372ULL,0xf5bcfed271fca7daULL,0x986fad52d174aff8ULL,0x92906b2441120dc4ULL,0xc563cb3c7114a743ULL,0xaa1657cb313d9e69ULL,0x703241e6501483f0ULL,0x996cad88514da3c0ULL,0xc9082c1011100c87ULL,0x410a0004004aULL,0x24160401409403c0ULL,0x10904c8410100e80ULL,0x4301f75251248684ULL,0x4822af64513d05eaULL,0xc1d4d32d51dc17caULL,0x696610011134065aULL,0x69f446e507c8f90ULL,0x9089b350115006e2ULL,0x4514396251b53242ULL,0x9144416c51ec82c0ULL,0x2621258701c3240ULL,0x2fbcdfd1515cbafaULL,0x18fff67451bc87cbULL,0xd1ac847091500752ULL,0x0,0x0,0xa04080010100c00ULL,0x0,0x0,0x4801000008ULL,0x0,0x0,0x0,0x0,0x40100002040ULL,0x0,0x480000008040ULL,0x0,0x0,0x0,0x1800040000010ULL,0x204000010000400ULL,0x0,0x0,0x18003001020042ULL,0x0,0x0,0x800010000180000ULL,0x0,0x0,0x4000000200840000ULL,0x21000002040ULL,0x98eccc9511811d0ULL,0x0,0xd28aa77051552e4dULL,0xcea05376519c85ceULL,0x4f44212e11c783caULL,0x2b34c71951b40dc2ULL,0x342a7f49103caac8ULL,0xf9c5afeb717f8ff8ULL,0xb2127d80511186d8ULL,0xc0a8f42851148edaULL,0x2dd03c92513c8a52ULL,0xd6b6f7ebd17fb7eaULL,0x9b7fff6071fdbfd3ULL,0xdf766f77517d8ffaULL },
	{ 0xfffeffefdaffbfe6ULL,0xfbf6f77edaf7bd66ULL,0xcffbffffdafeff74ULL,0xffffff7bdaffffb6ULL,0xf5f7feefdaf4ff46ULL,0xfbf73ff3daf5bdd6ULL,0xfb97ef3f9af0ff44ULL,0xffffffbfd8fdafe6ULL,0xafffdfdedaf7efe6ULL,0xd4f6f6efd87eaf64ULL,0xd9bfbd7d9af86574ULL,0xffbf7ff89af2bfd6ULL,0xcfe7ef1adaf3bf74ULL,0xdbfe5f5b92f4bf36ULL,0xdf3f5ffbd2fbbff6ULL,0xf9acf7ea9af5af74ULL,0xfebe7fefd2f3ff66ULL,0xa7f7ffffdaf0af26ULL,0x7ffffedada7eedf6ULL,0xffbbcfffdafcf514ULL,0xf3fd177e9af22740ULL,0xfbba7fff9afced46ULL,0xe9dffb7d9afcedd6ULL,0xebf61fde926caf42ULL,0xf7dadffb9871ade4ULL,0xf77ef7efdaf8af76ULL,0x9b7e7d739af4adf0ULL,0xfbd6ff6e9af0ad70ULL,0xed6a7f899af8ade0ULL,0xcfbfdffadab5fdf6ULL,0x7ffffeffdafdfff6ULL,0xbbbefdf3da7cff06ULL,0xff9e7ec7dab0b546ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x15b84ce0c2e5ed40ULL,0xc97cfa2998e4af50ULL,0xdca62d4ad8682466ULL,0x840000000300000ULL,0x80080000000400ULL,0xc3857ba92308534ULL,0xffc9f7eb98b0e576ULL,0xc484190010243522ULL,0xcbfcf60692f0bd64ULL,0xfdd37edfdaf1bd36ULL,0xdf9b7dfbd2b7ad66ULL,0xd4ebe6f49a68ad60ULL,0xd1be0d72dac03556ULL,0xd1b9672a9a642740ULL,0xceebbb9fdae1ed36ULL,0x2abad3489ae02d14ULL,0x8860c6e89864ad10ULL,0x5aa6ade092f06f64ULL,0xf59cf33fd8f0bf60ULL,0xcfcfbebbdaf5bf56ULL,0x0,0x401000000004100ULL,0x0,0x0,0x0,0x0,0xb4ffdfe992b4fde4ULL,0xf9ffd76a9af7bcd6ULL,0xfabd37f3d2b0f566ULL,0xf1506a1080b0ad02ULL,0xef77bfdfd2bffdf4ULL,0xfeeb6f739af6bd76ULL,0xb773f77fdae9ad16ULL,0x286a52c903dad44ULL,0xcbc7f72bd2b6a544ULL,0x4bc40e539ab02d40ULL,0x379ad4b290a28460ULL,0x4437fa6ed8e8bf64ULL,0x2144ec609224ac44ULL,0xebd5cb02daf6a504ULL,0xeedf3effdafdbf76ULL,0x7eadcffbdaf4f566ULL,0xf5e3d6eb98e8ed36ULL,0xd33e7d73da34ade2ULL,0xdbee7f36daacad16ULL,0xfffebffb9afcad74ULL,0xf7ecd4a9afd2c42ULL,0xd4befcfadab0acf2ULL,0x454e7961d8e82416ULL,0xcfb0377d98f4bf70ULL,0xdfdffeba9afefd74ULL,0xcadfbb439a64bd64ULL,0x2471742892244500ULL,0x100455e992254514ULL,0xfee72f6fd877c560ULL,0xc752d73892a5dd10ULL,0x4b9fde51dae02d60ULL,0x369af5f1d0a4cd70ULL,0x394454a1d0a43420ULL,0x574e7d47d8a1b5a0ULL,0x8106022480200440ULL,0xe591b4b92402f64ULL,0x573dcef098e6ad70ULL,0x100280d490a40050ULL,0x39de41609002a9f4ULL,0xe9f497099ac0ad44ULL,0x86233c0c860a440ULL,0x26ff5fe0d0b0b5f4ULL,0x7161777a1842edb2ULL,0x88b607729860e534ULL,0xf293fb7e9a76bda0ULL,0x40462a32c020bd40ULL,0xc13c34090a02470ULL,0x1413ce6490d09b00ULL,0xd5af7d63daf1f564ULL,0x968423e09200a5b0ULL,0x8950a2008aa4c500ULL,0x81d835000020a564ULL,0x1003031298a16500ULL,0xf04c470290246550ULL,0x40b33572082c2540ULL,0x848e904c92800560ULL,0x61cfdbf99801a5b2ULL,0x75ae5f2298c0bd40ULL,0x8009b74090f01a40ULL,0x33be6f659ae1c550ULL,0xebc4c1d0d228a4d2ULL,0x814100080082004ULL,0x0,0x100800002004ULL,0x200100200040ULL,0x8000200000200400ULL,0xd4a120088182400ULL,0xcf9fdff39af92574ULL,0x9c545fe01a682514ULL,0x8009a41090110c10ULL,0x4800240000ULL,0x83c28c03c0702202ULL,0x8010014288802122ULL,0xb04c489000a814ULL,0xc1e9adf0d0b0a580ULL,0x8b2eee429000bd74ULL,0x89511f8892a02504ULL,0x2b5505429ae40400ULL,0x1814000080002040ULL,0x91780d0818408510ULL,0x4ab22413d2882d40ULL,0x8fc7f317dae9ad30ULL,0xf777ce19d8e42e40ULL,0xfc327ffb9a7c6774ULL,0xd4ff9f479a66f546ULL,0x76ae7d72d8f0bfc0ULL,0x9de56c89ae0b504ULL,0xa047f81198a42140ULL,0x4080050090000400ULL,0x2008000300000ULL,0x29400d6080612606ULL,0x5884602a00348540ULL,0x842125090202440ULL,0x211648c000202546ULL,0x4907c8890310c40ULL,0x81002f228a002542ULL,0x488771909020ed24ULL,0x49521e101082fd14ULL,0xb220e9392a6ac04ULL,0x34b60de29808f5a6ULL,0xd960ae61d8f46510ULL,0xfabc6974d0a0ed74ULL,0xe443eb028ab0ad46ULL,0xb05a0190102400ULL,0x1000480012000500ULL,0x91f1af1990f42724ULL,0x8834047080092406ULL,0xc054182200000560ULL,0x207c86c310302414ULL,0x90826ca08000ac40ULL,0xf1bc24430220b922ULL,0x44466c20d840a910ULL,0xbd2b13890a1acc4ULL,0x8e820a0092242524ULL,0x16914989d201b400ULL,0x198702892020110ULL,0xc600304acaa02400ULL,0x4a444b26d2a0a514ULL,0xe492b4aa9af4a510ULL,0xcabaef49d8f4b5a0ULL,0x5554b6f2d0403f20ULL,0xc121a60090b081f0ULL,0x8400000004100ULL,0x882120042002500ULL,0x80180000040ULL,0x1000800000000400ULL,0x0,0x0,0x0,0x2100000010000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4100000010000400ULL,0x4020040080002400ULL,0x840000090000000ULL,0x0,0x0,0x0,0x20200080000040ULL,0x4000400000008080ULL,0x0,0x0,0xc010008000400400ULL,0x9d7bafffd234fd56ULL,0xdadfaf1f9ab2bd16ULL,0x8573c7089232add2ULL,0xa3e8c50092101d04ULL,0xb9bad76b92f43560ULL,0xc9eebfa298e12776ULL,0xfbaebf719af4bd74ULL,0xcdd1b616d2b10d60ULL,0x2a5e99d8daa4ad52ULL,0x9cb34febdab3bd22ULL,0xf1fdefefd2b5a5a6ULL,0xffbe7df4d2bdadc4ULL },
	{ 0xffffffff7fdeff73ULL,0xfffffffffffe7f73ULL,0xffdffffffdfe7f73ULL,0xffffdfffdfdeff73ULL,0x7cffffe7effe7f71ULL,0xbbbefff34dfeff73ULL,0xffbdff6fcf7eff63ULL,0xcf56fef3ddfaff73ULL,0xef9ddfdbdfdee773ULL,0xfcf9dfe47dde7f63ULL,0xdfffff738f7e7f63ULL,0xfb9c7f55cfdaff63ULL,0xdfceff2d4d4e6963ULL,0x6f7fffd1cd5e7f63ULL,0xf63b5deb2f4e6d63ULL,0xbbcff2794cda6962ULL,0xdeb27f76cfea7d63ULL,0xffdff3afdffe7f73ULL,0xfffefe1fcfdeff63ULL,0x9fbb7ffacfdeef63ULL,0xf5ec77edcdde6f73ULL,0x7f667d7dcb4ef571ULL,0xe9f6f76ebf8e7d73ULL,0x6af796db4dde7740ULL,0xbefbeddbdf7e7561ULL,0xd7fefe79cd7aed73ULL,0xef1e77e24feee563ULL,0xeaf6e36605587563ULL,0xdedabe1b895c7762ULL,0x67fb4fd24dc87d62ULL,0xb5f7f6e84fce6771ULL,0xbf7a6d722f6e5773ULL,0xfbbdeffdff4e7f62ULL,0x1548204b8106560ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x403302549880000ULL,0xdbf8b7d8c944f603ULL,0x0,0x4000040000000600ULL,0x221086c06824a460ULL,0x37bb5ffbcdecfd62ULL,0xe745a7fbcdd86f61ULL,0x9000372429024d62ULL,0xd017e3a24a6a7560ULL,0x2cf67fd24e5a6f62ULL,0x9fdfddba4dcaff62ULL,0xf42eb4f504d82740ULL,0xbf3e3d300f5e6760ULL,0xebb7f736ed1eff61ULL,0xeefabf561d7a4f63ULL,0xfa9cdf7fcdac6d63ULL,0x8ed33fe0294e6d62ULL,0xb79aef31efce7762ULL,0xd70cfe6cee4c7761ULL,0xcddaf6930fe8ff62ULL,0x0,0x0,0x0,0x0,0x0,0x9f295f1a9d3e6f63ULL,0xf47f5ffb7f5c7d73ULL,0xd9bdbee65fc87d73ULL,0xfaa03717198e4763ULL,0xc004db248a407472ULL,0xeffebfdffffaff73ULL,0xefbdd7ebadfaef70ULL,0xd301b7a1d1007141ULL,0x96b2f56cc97c7f63ULL,0xebddf6e779deff60ULL,0xa9490e516dce7920ULL,0x84130a6009624140ULL,0xddd25ee0450af742ULL,0x6dfe65b20dcc4f62ULL,0xa944f72c050caf61ULL,0x4fdbba9f64daef62ULL,0x20c440808420541ULL,0x55c53e690752ed70ULL,0xdaecbd424efe6d71ULL,0xf7b6fff7cb7e7f72ULL,0xcdde9f3bef4aed62ULL,0xfffeefdfefeeff63ULL,0xd63517681f264760ULL,0xc9afad66fdcce561ULL,0x98892d3219501f20ULL,0xdfffffbeefeeff73ULL,0xebeff7dbcadc6732ULL,0x37fb57e949c8f762ULL,0x910ad670194a6553ULL,0xdafaff7f5f5eff73ULL,0xc8e4d3ee0c6e7962ULL,0x7adf3bcd0fce5d73ULL,0x240a2df207986d62ULL,0x4251773a2b5a6d61ULL,0x569e7b579eec6560ULL,0xab7ff22bec4efc63ULL,0x6ed39fceaf04e462ULL,0x56655568ed422960ULL,0x4244565204002441ULL,0x406e74484040a460ULL,0xf004693e0f4e6522ULL,0xebcb9e7688467563ULL,0x3fbe4d1b1c7a4d63ULL,0x749652ed0a1e7762ULL,0xa001853001022541ULL,0xf400754a01080d60ULL,0x48430e00050a4062ULL,0x6b7f03d0cd4afd61ULL,0x61e494cc4d286460ULL,0x88498202080a7442ULL,0x4012044002040ULL,0x81c0a103809ca120ULL,0x46100001084000ULL,0x1489002003020540ULL,0x2124b622ca0c6540ULL,0xaab0a7495d006d62ULL,0xc990939f49427442ULL,0x28d75b9d5b087d70ULL,0xbd204932842a7761ULL,0x9271a240498e6102ULL,0x7284b22c2d0c6461ULL,0x8e581620b4e2560ULL,0xebeb9e1f27caf562ULL,0x4004c8b10c14a140ULL,0x144800000140ULL,0x2a02004000140ULL,0x260408484000ULL,0x828a010488a502ULL,0x21be4f422f0c6763ULL,0x301406620018a760ULL,0x823a6d7107067761ULL,0xd8b4255209446223ULL,0x94642090848e020ULL,0x2b6e4b58cb0e7560ULL,0x520002001020440ULL,0x4ca6496087007762ULL,0x12200764210a0360ULL,0x4c19d1808484c22ULL,0xa7c115205405442ULL,0x812c30004203441ULL,0x142400000080700ULL,0xfebe7ffeedde7f73ULL,0xe9d7f7ffefce7f73ULL,0xe9f7fede6f5e4573ULL,0x37bf5fed074a7773ULL,0xd9c9f778c50a7143ULL,0xa3220e4018004f70ULL,0x8ac0622ecf407640ULL,0xe6f239466ca86162ULL,0xcf9de971c95e6762ULL,0x4c1070200f027500ULL,0x340640084002140ULL,0xdbbeaf3e8ade6560ULL,0xc9daf219ce2e6761ULL,0xab3e185b880c6661ULL,0xb011014200885622ULL,0x8410014000042421ULL,0x5bafe6230708e761ULL,0x809080b08040540ULL,0xab1edfc387986d61ULL,0x14907ce207285c71ULL,0x882d272110404740ULL,0xcbb067748d746f60ULL,0xc5f2faa949de7f63ULL,0x80c7b4105884532ULL,0x880a3b8107084542ULL,0x804a8b4205026960ULL,0x8b941f1ccd026462ULL,0x10902000020140ULL,0x6908032002085d20ULL,0x42820d030c244140ULL,0x9400814030000960ULL,0xef8616605d2ead60ULL,0x8b2bb8c2cd467541ULL,0x10530642050ca561ULL,0xc98a0510c104741ULL,0x3a226800884560ULL,0x2ab40da019487941ULL,0xcb14e12242886740ULL,0xadc21a100b404e70ULL,0x631e4d5b350c3f60ULL,0x80100ee80d486122ULL,0xdfa8ed324f406743ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000002000000440ULL,0x500000000140ULL,0x0,0x0,0x0,0x4084040000040820ULL,0x9ceca5738bd86f42ULL,0xab12753407cc4761ULL,0xc8d3be9e019aef63ULL,0x8b5f8b5a08d85f63ULL,0x1ce765a06d5e5570ULL,0xf5ccb5618aee6563ULL,0xcd8e35590f1e5760ULL,0xcdcefe77ed1e7d70ULL,0xffffbff6df6c7f73ULL,0x37bbedeb1fce6f62ULL,0xfdf7b4fdcb9e7d60ULL,0x47306c534f88ed40ULL },
	{ 0xffffffffdffef6e8ULL,0xfbffff7fdffeffe9ULL,0xffffffffffff7fa8ULL,0xffffffffffbffda8ULL,0xfdffffefdf9ffde9ULL,0xff2eff73daffbca8ULL,0xddeee776df9e7ca9ULL,0x9ff7febfd7beffa9ULL,0xafdecfdbf7be7f29ULL,0xf7f6fff0febe7de9ULL,0xdfffff73f7bfbe68ULL,0xffbe7ff5dbfe6fa9ULL,0xdfffff7fffbefca1ULL,0xbf7eff5adfdefea1ULL,0x9ebf7767dfbe3f68ULL,0xbbfea772b7de7fa0ULL,0xeebebfeffaecbfa9ULL,0x6bfef7bed7bf7fa1ULL,0xbbdefedf7fb76ce8ULL,0xbfbffffbfede7fa9ULL,0xfffdbffefbfebf68ULL,0x176e3d6c7eda4d28ULL,0xc0b4d16a629e2ce0ULL,0x2aff9bebfe9f7cc0ULL,0xbfbbdffbffbfffa8ULL,0xf4befefddf9f7ca1ULL,0x6dee7d625c9f3ee0ULL,0xf3d4ef5cffbf3ee0ULL,0xdef3bfd11bbefd20ULL,0xf124da35e9d2c20ULL,0xd1c7bfe3d39e7ec0ULL,0x291ef7304b8d3c28ULL,0xfdb4ff7bffbeed28ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21edd8c24616ada0ULL,0x33125b97139e3c60ULL,0x93352e6852862ca0ULL,0xea96bf7a5abefea0ULL,0x83fbf3625fbe3f80ULL,0xabe95adb838fbf68ULL,0x32bc47690a962b00ULL,0xc4d4be4b55b63620ULL,0x82a609e0458416c0ULL,0xa399c70703b6ee21ULL,0xc2625f65ce9f1da0ULL,0x2284e40c4823e00ULL,0x5170afeeeb0e3ca9ULL,0xbfbbaf5657dfbfa0ULL,0xf3a0667e599e54a0ULL,0x907d20232c48d40ULL,0x0,0x0,0x8000040000080400ULL,0x0,0x6000000060000ULL,0xef7fdfdaffbefda1ULL,0x3cbef3eddf9fbda0ULL,0xf9f9f777ffbefda0ULL,0xdab62771cefead20ULL,0xcfca6f31d61c3480ULL,0xfede3f9e7b9f7fe0ULL,0x87aa5ffb56de3ea0ULL,0xd454ac2c1ab6ad28ULL,0xffb67f7d769c7ea0ULL,0xc8cdfb7b56bf7da0ULL,0x4ddb1e9e8adf3c01ULL,0x145d4fc0420c3488ULL,0xd8f082400aec1600ULL,0x572e75f0179e7d20ULL,0xfb878b36579c2e80ULL,0xeecbb6b31abbbf20ULL,0x255e5ed1c5be3c21ULL,0xd17f36ee1b9e2ea0ULL,0x9ffcae31c03e3da0ULL,0xefbbff7ec7be7fe0ULL,0xc4d3b3335bae7ca0ULL,0xbb66cfdbd3d63e69ULL,0xf6bf5fecbede3e21ULL,0x412eb7095ca63ca8ULL,0xdefe3eab53ae2d29ULL,0x88c3ee2a16127c29ULL,0xab3f5d53d38e7c81ULL,0x93ac4ea16923d20ULL,0x5084d0c414907c41ULL,0xdabcaf7ed79efca0ULL,0xcbcdf3ba9bd6af60ULL,0x4657df5642d67e20ULL,0xff8ec5b6f3fcafc0ULL,0xffffbf7b7f9e3da9ULL,0xdfdeed7657de3ce1ULL,0x8904e33a52a23d20ULL,0xbfdabe5d87b7fda0ULL,0xa632d69303b73d00ULL,0x5050a7f411971860ULL,0x90272182261400ULL,0x1a81cb030293bb00ULL,0x8040081200185400ULL,0x8eafdf7bbabfcf60ULL,0x5477b6e87b962ea0ULL,0xce38356347953ee0ULL,0x220840020202a820ULL,0xc823000000e0400ULL,0x84c4082261400ULL,0x7422f0e80236bd01ULL,0xd4a6856012843c20ULL,0xa332250486906200ULL,0x8023aa2110852c00ULL,0x0,0x918c13b141c40a0ULL,0x8078e7004a1e44a0ULL,0x260640420c0480ULL,0x414ef318508c0da0ULL,0x222e50805a840080ULL,0x8284b6a9b15a801ULL,0x32402628522c2400ULL,0x8804f00c07dc2400ULL,0x0,0x1124002800600ULL,0x20280c8b82162c20ULL,0xb18c0c5a55106c08ULL,0xc220251000b12d00ULL,0x180a01202061420ULL,0xe072b010c7be2c80ULL,0x85ad4c404a225400ULL,0x4032926144b60600ULL,0x8010240022104400ULL,0x8000250003000400ULL,0x1080001800000ULL,0x21000120000ULL,0x4a04240860280ULL,0x114633208a842800ULL,0x91100a14038e0000ULL,0x8543e02807902200ULL,0x200542528c1080ULL,0x120424050062020ULL,0xe944426301f82c28ULL,0x8e14030402926c20ULL,0xc351fba61b9e1e40ULL,0xeecbbfeb7eb67c20ULL,0xb7bbeffb5a9e3fc0ULL,0xf0bc9f7ddb9f2c60ULL,0xa720246003823400ULL,0xc395d204028e3c00ULL,0x88576e86629eedc0ULL,0xc5ee48215d1e3c00ULL,0x934d41a0b803e40ULL,0xc1a0d91428434a0ULL,0x49b1e73056962220ULL,0xea4e025558fc6480ULL,0xa23e4e805a963ce0ULL,0x11cc96cb0fd46dc0ULL,0xa033ad11c3b24ca8ULL,0x7910361852de0ca0ULL,0x8e20480002a42680ULL,0x23d38fc15eb72d20ULL,0x14f48ee84e063d80ULL,0x808a60224112c00ULL,0x7a942f07029e3ca0ULL,0xc6d5a816161e7c08ULL,0x2936de56ea962ca0ULL,0x800203ed468c2800ULL,0x210e020000a22420ULL,0x0,0x5154726201883d08ULL,0x15190c810a62c00ULL,0x182490000980440ULL,0x184c968288962c20ULL,0xe042377c1292bc20ULL,0xc182f13a129f1d20ULL,0x8a69800108e2420ULL,0x702441604a080d00ULL,0xe1bcb261ce963f00ULL,0xcaa025c0e38684a0ULL,0x88904364620ab480ULL,0x8d01b01833c2c80ULL,0xa79bcfdf3e9e3c40ULL,0xd1d7d67e43dffda9ULL,0xd801af401796e800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000041000ULL,0x0,0x300000120000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4000800040000400ULL,0xd73f237846d6eac0ULL,0x988c2b64743e0420ULL,0xc9c59a98dfb65c08ULL,0x77fadfdacaafbd20ULL,0x949e83e39bbb4d68ULL,0xe8285780538a6420ULL,0x46bf34f082823e20ULL,0xe9f7ff7edfdeff20ULL,0xeededfd39bb67f28ULL,0x66af67bbc3bf2da0ULL,0xb7feff79da8fbc28ULL,0xcf34bf36e5de7ea0ULL },
	{ 0xdffffffff7be6f57ULL,0xffdffffff7bc2f7fULL,0xfffffffff7be6fffULL,0xffff7ffff7be6b5eULL,0xfcfffefff7966fffULL,0x9ffebf73f7ae6fdfULL,0xffbaffeff79c2fc7ULL,0xffdffe3fd7b46f77ULL,0xbbdf5f5ff6b66b5eULL,0xfff5fff8f73e2fffULL,0xdb5f3fe8f7b66be6ULL,0xfe8f3fffd6ae63cfULL,0xcffffe1ed7942b67ULL,0xeffedfdbf6bc6f77ULL,0xb5b7feeff2bc67f6ULL,0xf5acbe78d6902ff7ULL,0xfe9e2f77f79c6776ULL,0xe7fff7ffd7bc67ffULL,0x495fffd5d286677fULL,0xf7bb7ff2f7b66ffeULL,0xd1bdf67bd7242f57ULL,0x9f8c4d6fd7a4237eULL,0x88b1d38893a06362ULL,0x4fdf3fdbd3866f7fULL,0x5f5dfffab2ac235eULL,0xf7f7f764d696275eULL,0x82a67141d6140b16ULL,0xd38e7b6fd794637eULL,0x80c1b970808003c2ULL,0x8efe4f6b97942976ULL,0x35d4e7ebf30443a6ULL,0xbebf7d75b3b46bdfULL,0xb0846568d48c2b4aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400200001000002ULL,0x2a1fd9c8d3946363ULL,0x14bb7ee391b4214fULL,0x997ebf7bf6bc6f7eULL,0xcf1ae75ff614231eULL,0xc3e7b73ef7ac4b4eULL,0xaef23efbf7b42b5eULL,0x928a051aa280254cULL,0x9765f26cd304272eULL,0x5ba83d51d6042b3fULL,0xfbc6f72cb3a2274eULL,0xeefbbffff79c27ffULL,0x2df84dc3b2042602ULL,0x11db36eed6b62346ULL,0x89492dc1b3046116ULL,0xf905ff7ff7be6fdfULL,0x4c004b1000100ULL,0x0,0x0,0x0,0x0,0x0,0xef5cfed7f7902adeULL,0xb4fa0ffff7b62f37ULL,0x997ee60393002f47ULL,0xc092bf70d6202b66ULL,0xefdffefff7be6bdfULL,0x2ddfee7bf7b4277eULL,0x7793ef73d6142f71ULL,0x8378256ed2342246ULL,0xfa86256ed684296fULL,0xa300fbaa77002216ULL,0x5ee65e5bb2a4670fULL,0xb2191a72d7102306ULL,0xc0f4d612b6166740ULL,0xe2bcecdbf6906b26ULL,0xbb16fb3eb7b8634fULL,0xcc4eb6c1d206636fULL,0x8abefbfd9684234cULL,0x51f796faf7806f5eULL,0xba722d62d2b02acfULL,0xdc022a6a92164304ULL,0xdce38f98b1242f43ULL,0x797ebddcd7966357ULL,0xf6f4f6f7f79e0b1eULL,0xd1ccf600c0102157ULL,0x64b12161802c234cULL,0xe6d6ea1af610236dULL,0xbc9c9c893146725ULL,0x54e9712a92162541ULL,0x810c500ad4040804ULL,0xca8aa670929623cfULL,0xe2c3df88d5242626ULL,0xedff9fdfd792674fULL,0xcfbbffbfd3966f7eULL,0xd97cff7bd6b0277eULL,0xffa67564d7946376ULL,0xc9c0f17bf6962223ULL,0x4052361ae6006304ULL,0x19bc4f92f2842706ULL,0x9e50ae53b2942a7bULL,0xaae4d629216631cULL,0x9bdd8725f2242717ULL,0x18911e9292002127ULL,0x8ffcdf0db6b0265fULL,0xb4355ae290142f1fULL,0xbfae3f20c71623dcULL,0xc00010000200ULL,0x89c2e8008210225dULL,0xa002010041000002ULL,0xb0ba52401620014cULL,0x8d76ef10f6b46342ULL,0x7294028e9014021cULL,0xa87e6b169210271dULL,0x844484012020104ULL,0x88820f2156102604ULL,0x4040a160d4902644ULL,0x72142d61d2842552ULL,0x44001854000802ULL,0x200c0040000214ULL,0x86030e28d5040202ULL,0x88348640d1000316ULL,0x1800a4840802442ULL,0x8166aa0ac3022b12ULL,0xe4200008180005aULL,0x91000c531602206dULL,0x4184006890206070ULL,0x60b82c00c680410eULL,0x210020010000104ULL,0x9402b202d600610cULL,0x536536092144324ULL,0x8450c67992902264ULL,0x2010000000002400ULL,0xa890402080900250ULL,0x0,0x40000810000800ULL,0x202004c010800860ULL,0x9406dc3d40c0a94ULL,0x38082d2dc620025cULL,0x146a6229210264eULL,0x1e5d150b0042a83ULL,0x34d04c28f4106a00ULL,0xa100852a80102704ULL,0x601e2420c2142304ULL,0x8946fa01d194230fULL,0x60f63a7992002f07ULL,0xb382cfa89304036cULL,0xd76ef54bb31c630fULL,0x92a0e2183202147ULL,0xd3007602c2a02016ULL,0xe7fb1fcbb78063d7ULL,0x451cec29130262aULL,0x2024a27952002744ULL,0x400844a2d4100850ULL,0xf9706952f7900146ULL,0xaa071a8392182b4cULL,0xaf9fce5d97b62bf7ULL,0x8cd05660b41002c4ULL,0x843ae531d6b00346ULL,0x2100252010042245ULL,0x4140061092100a4cULL,0xbf9fdfdfd6b02ffcULL,0x74bd54afb2844bdeULL,0x37e650203b40346ULL,0x9ba4af6cd6040b72ULL,0x48070f82d0002a68ULL,0x48177cc3d6942706ULL,0x34bb7e689190093eULL,0x41048648c0802102ULL,0x820002000004ULL,0x402212008200240cULL,0xa8e0943b1104104ULL,0x2a610292004b00ULL,0x4228562292106302ULL,0x1528074102122a02ULL,0x108010000100ULL,0x43de134b91a46143ULL,0x864a8e5292142371ULL,0x8002026040802412ULL,0x930a2d4054042e04ULL,0x900c220c4840302ULL,0x49cb12d093844346ULL,0x8002000000000a00ULL,0x2190566ab0800b42ULL,0xd62e2502c23427ccULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000000800040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10002000000ULL,0x0,0x100010044000000ULL,0x0,0x0,0x200000480000040ULL,0x0,0x0,0x0,0x0,0x102600010000a04ULL,0x0,0x0,0x0,0x97ada523d70a2257ULL,0xcb0832b0d7142a0dULL,0x4fd7bf88f7120356ULL,0x2eff54cb93962762ULL,0x203160609180404aULL,0x59c05402b2042320ULL,0x8a0a0f34c7000306ULL,0xa3deef2ad78c27f7ULL,0xcb961b0fd0802b07ULL,0xa6b3ddfaf7866e57ULL,0xfb77f7fff7b66fdeULL,0xfffefd77d7bc2756ULL },
	{ 0xffaefffbfbfad7fcULL,0xfffffffffbfe7fffULL,0xffdfffffebfff7beULL,0xfffffffdfbfc733dULL,0xfdffffebfbfe7737ULL,0xddfebd67ebdef7beULL,0xfffefffffbfff7ffULL,0xebfffe3bdbdc493bULL,0xebcedddffba87b5fULL,0xf6fdffe8ebf45f3fULL,0xfdefff73fbfaf7feULL,0xdfbf7ffffbbff7ebULL,0xffc7fffffbb4f3bfULL,0xefdf5fd7fbebf7bfULL,0x35b5d7e9fbbcd377ULL,0xd957ecf2d9b157bfULL,0xfebebff7ebb5d37aULL,0xebfeffffcbfdf77fULL,0xabdeffdffbe25b7fULL,0xfda7ffffebdcffffULL,0xfbfdbe7ffbb87fffULL,0xf7fe7ff5fbfed3beULL,0xcbfdff3ffbf5ff7fULL,0x8eff3e61dbb5d337ULL,0x9b9adffbcbb17f3fULL,0xf5fffffeebb5d777ULL,0x4d1e7f73ebd7473eULL,0xcb84ff2ee9e4f1e9ULL,0xaed7fa08dbb0c17eULL,0xd3becdebeabc47fcULL,0xf5d3fefbf194ff2bULL,0xfffeff73fbfe73ffULL,0x240000000208ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81000000800010ULL,0x4cc216008bc46a20ULL,0x7fdebe9febbe537dULL,0x9ebb5fdeebf0f39bULL,0xf1fd27305b00d368ULL,0xdfde7f76fbb67b7bULL,0xcb97fb77cbd0577fULL,0xc1d25b10d9a05241ULL,0x3ffecfe9dbbc77edULL,0x50e156685b804628ULL,0xc39c3560c3b05309ULL,0xcbd4ff6ce0b41338ULL,0xca66ffffcbd957abULL,0x839e4e23cb3052e8ULL,0x50c557fbc93b732aULL,0xcdbeef72dbb8d37eULL,0xffae67fbdbf1737bULL,0x0,0x0,0x0,0x0,0x0,0x0,0xa3fddddafbf45b2dULL,0xf7b357eae8f5fbebULL,0xf0e2a77e8bec7325ULL,0x7eff3f7ffbf053dbULL,0xefffffffcbff73ffULL,0xb5f1fcbebd64f34ULL,0x808a4fffc3b01b5dULL,0xf9fd9ffbfbbe7b7fULL,0xfebe6ffffbb8d37fULL,0xe3b1f31aebdc533aULL,0x4ef8bc51d9c4532fULL,0x303c4f91d1805338ULL,0xccff7fdedbf4333bULL,0x4fbe1dd2f10051b8ULL,0xeb54eb1e7b9163beULL,0xa97fb0db89b47bacULL,0x3fafccdafbb87be8ULL,0xfefdf6fffb907ffeULL,0x9f9cdfd1dab073fcULL,0xefbe777fdbbcfff6ULL,0xcfeeffbedbd5433fULL,0xa17b5ddb891c5362ULL,0xfcfcd6fdfbf9333bULL,0xef2efff3dbb5fbbaULL,0xcfa43770fb9c6de3ULL,0xc8a3fb39c9f8372cULL,0xef48de5e8bb2f3fcULL,0x342b7fc58b94d93cULL,0xdba4e5c279c3033bULL,0xcad8176a83b44be2ULL,0xa9f3dff7f9b5d3efULL,0x2b7dfe73c1927b5eULL,0xb4ba6f6ff93873eeULL,0xfdffff7cfbf4f37eULL,0xf7aefe4efbb257e3ULL,0xebbdfbeffbb4df3eULL,0xcfdffe79fbf8dffeULL,0xba9e8dfecbf477fcULL,0xdcd2c66889b64236ULL,0x1aa47d61e8205328ULL,0x9312834c8848031aULL,0x8cc210c18924436aULL,0x3b08050be8180902ULL,0x86903427100015aULL,0x92b82f5009d2037dULL,0x5d36f3260b887366ULL,0x2146000888014202ULL,0x4b0100a03200ULL,0xf49f7feedb305dfeULL,0x116d41426080413aULL,0x5a9c6a3ccaa46b38ULL,0x80438869c880d094ULL,0xa84f3cc8ea802176ULL,0x2090438d89a00250ULL,0xc000706168400260ULL,0xf20029448bb10b6aULL,0x100001000000110ULL,0xa5089c808a0c310ULL,0x9291056888801332ULL,0xc81e27e0ea891723ULL,0xaa0246269404120ULL,0x80800080000004ULL,0x4204060048001308ULL,0x845000a0c0201120ULL,0xb41e36e889bc066dULL,0xb8c77e2f0b057b8ULL,0xc100410889104a00ULL,0x25a325b090463faULL,0x8804be009600308ULL,0x102c08820020030eULL,0x402a296000800523ULL,0x400004000004080ULL,0x4100900000000328ULL,0x89084713d1a072b0ULL,0x4a074408130530cULL,0x88071678828453a5ULL,0x5cae0e74d9e05b69ULL,0xd5c80682304328ULL,0x87380189c01b22ULL,0x10048001000060ULL,0x2000220169800001ULL,0x4886016240041728ULL,0x741d810d8944129ULL,0xb4a1bc240a052a8ULL,0xb6b0dfc8790df3aaULL,0xd1fc76fb6bf973ebULL,0xcf9e3d72e3b45b27ULL,0xa9a4763febb27739ULL,0x2e7eb650d9f01369ULL,0x3f9a0d1ae9685b38ULL,0xa467722dc99143afULL,0x289e0430cb3352a2ULL,0x8300ef25dbd05320ULL,0xdfdabec78bb55377ULL,0xe48804328940526aULL,0x181108e2e1901010ULL,0x2b0a6540988c370ULL,0xc914073861045720ULL,0x904dd0088805303ULL,0x4192df92c0a01b1cULL,0xf4b4f771f9701916ULL,0xfbfeff7bd9937f7eULL,0xc2165bf449b27309ULL,0x81d6e768d190130bULL,0x95d5c528bf9332fULL,0x2127248094b0162ULL,0xd141b7780980437bULL,0x202008006000ULL,0x4080100040001208ULL,0x246a24162900214ULL,0xa82001a1e0001200ULL,0x124200000d200ULL,0x6406747cd1a01234ULL,0xf006da46e134c364ULL,0xa4e2263089887062ULL,0x20a000a80300118ULL,0xd4053080cb207208ULL,0x15e3410538640a0ULL,0xd285873e612043ecULL,0x4957b3618ba453e1ULL,0x243c4890d0800334ULL,0x109456d78914576cULL,0x0,0x0,0x0,0x6000000000108ULL,0x0,0x0,0x8000100000800001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x402a0003820000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1202106001800121ULL,0xcf92277fcbc57127ULL,0xe5c4c128e9d0772cULL,0x6a90d88aca885211ULL,0x24a0556a8091133aULL,0xe967af38c9d6316eULL,0xdabe2ffee9be7befULL,0xc24ffb6ec93c7b2eULL,0xfa7e9fdbd9b07feeULL,0xf7fb5feb8bfedffaULL,0xd4fdbf7cfbbc7ffeULL,0xfab67feaeb98d16cULL },
	{ 0xeffcff7fabdfe3faULL,0xfffffffeebd7fbfaULL,0xbffffffbebdfe3faULL,0xdbfffffbebcfcbeeULL,0xb5fd5febab5ff37aULL,0xbfbefff3cbdfdbf6ULL,0xffbefff7cbdff3deULL,0xcfc2be1dc184c33aULL,0xeefecbc0abc7e27aULL,0x50bef7ebc3d5e39aULL,0xdbfcf6638bcff11aULL,0xfa9d3f7eebd7e96eULL,0xfdffff3fe3dfcbfaULL,0xeeffffdfcbd7837eULL,0xbfff5f62ebd7e37aULL,0x9177b77c83d4f33aULL,0xfe973f77cbd7f366ULL,0xa7f7b73fc3d5e3e2ULL,0x6fffffdbcbd7e3eeULL,0xb6fb6feba3c5d33eULL,0xfefdbf5fcbcde36aULL,0xebbe3f702b45c32aULL,0xcbdfffff8bdfdbfaULL,0xe8767e7fc1c7e37aULL,0xcf9fd5fbcac5a33aULL,0xd477f666abc5d38aULL,0xef3eff72c9cff37eULL,0x8b912b0ec3d5c016ULL,0xefc7be68e185cb3aULL,0x2bb6694a81d4c31aULL,0xfffdfeffebddc37aULL,0x132087004081c008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x454ce00000c0c02aULL,0x689fbfd98ac4d32eULL,0x87bbdd9e88c7e36aULL,0xf1ffffffe3c7f36aULL,0x6bb675f0e245e372ULL,0x8bd3f36a8b4ce32aULL,0x3edd7fd183cdd32aULL,0x49a5fbfe1c5ca2aULL,0x36d252e861c5c068ULL,0x44422840010cc122ULL,0xdba2ff3ee3c5d22eULL,0x6a5632e201c7826aULL,0x8de94d7903c6c35aULL,0x14b6c6308354c162ULL,0xdbbcff32c9dce2beULL,0xff162f4ec1d4e1daULL,0x0,0x0,0x0,0x0,0x0,0x0,0xeffefed68345ea1eULL,0xbdb3dff38b8fd25aULL,0x99fafe58e3d6c222ULL,0xf2962d77a1c5d12eULL,0xc7d4e336a8c6eb6aULL,0x95e5efd88cdc07aULL,0xffff7fe18bcffa5eULL,0xd32fbd638354f33aULL,0xfffe7ffbebd5c39aULL,0xe984f70fab55db12ULL,0x4f5615198186c306ULL,0x959f686aa3d6db0aULL,0xf6f7fe6febd5fb7eULL,0xd2a43060814de062ULL,0xebbc7742cbccc38eULL,0xae02b8038184c166ULL,0x8fbeebd3c3c5eb3aULL,0xb0c50d600147c382ULL,0xcb9eef7383c5cb7eULL,0xfb9cdbbd83d7db7aULL,0xcbdff786a3d5cb7eULL,0x87e7de5283d4db2aULL,0xf4f6d7eacbcdf132ULL,0xfbfde7fb8bd7fbeeULL,0xd38623504844c33aULL,0x8ca20b092304c102ULL,0x489c5dc1e246c256ULL,0x943a7ecb6944c3aaULL,0x140d7c20104c32aULL,0xca32876e09d4c31aULL,0xcf579f0dc9d5ca9aULL,0xab1fdbd2cbc5c28eULL,0xb7bf4ffbcad5f96eULL,0x89fd7e7fcbcff33aULL,0x7b9e3ff2c3dce32aULL,0xcbd7f6e3a3cec31aULL,0xa573e58c9c4c09eULL,0x8ffa069ec144e322ULL,0x7c7e76e8a345f162ULL,0x29563f3102858122ULL,0xcac043268bc7d2caULL,0xa26aa04103c4c312ULL,0x8001006088004002ULL,0x34ccb6c8c8d5c80aULL,0x89eca72009d1c3a2ULL,0xf11c537883c7f21aULL,0x200000201000080ULL,0xa40841418140c10eULL,0xb01100208884a000ULL,0x2040424000c0800aULL,0x969c3f6c48d58362ULL,0xf044ef270354c352ULL,0x2a43d948810ceb80ULL,0x27802cab4bccc09aULL,0x84ac70600945c028ULL,0xdb146f7dcb5de3caULL,0x480802000408202ULL,0x840032c0145c200ULL,0x21a2be68185d20aULL,0xd21c80218004a20aULL,0x1a8000000102400aULL,0x14cd0204108c0a0ULL,0x60981ca06201c202ULL,0x208000000000102ULL,0x104a40048002ULL,0x4d407601a283d02aULL,0xfbdce33eabc5ebfaULL,0x8013942e2b47c0c2ULL,0xe560d2ac80ec012ULL,0x142108201018208ULL,0x5022056020d4c350ULL,0x0,0x454200a0046820aULL,0x1420802a0444000ULL,0x11026040468040ULL,0x9000006080022040ULL,0xf4906840498dc000ULL,0x480822a88c4c100ULL,0x4df211128354a23aULL,0x4c416630306c152ULL,0x8c040880539060ULL,0xa84062089044302ULL,0x8d012000a040202ULL,0x29dfb6cbc987e34aULL,0xf7ffdffbabd7ebfeULL,0xf7ffffffcbdff17aULL,0xffdefd79abd5eb7aULL,0x88c4e3038b03f126ULL,0xaac71fc88bc4e382ULL,0x30814f718186d150ULL,0x76e772f183c4c23eULL,0x4886756a08c4835aULL,0x200ccb6c01c5c022ULL,0xa051a02280049162ULL,0x8292ca000147c102ULL,0x91736eca09c1c23aULL,0x1934e4308381c32aULL,0xec886421c1d79060ULL,0x1c5821548068052ULL,0xaf36cfdb82d4d33aULL,0x4b27eaf83dff20aULL,0xc8aff560898ff07aULL,0xfe1e5f34e387e27aULL,0xc7eed99b80d7d156ULL,0xa91a5a030154c2b2ULL,0x4818038044400cULL,0x8f0c070a8156c060ULL,0xce02231f820ed026ULL,0xc912dc7a83c4da3eULL,0x6a5ec64882c3ea22ULL,0x2004f00c288c80aULL,0x120c004209060002ULL,0x7404416a4ac4802aULL,0xfb5deb3be384c3b8ULL,0x2c4600000305c202ULL,0x1a8cbd1a0dcd10aULL,0x752c7233c15491faULL,0x21107c708087d142ULL,0x88f0af22c180e32aULL,0x2f42a3408146d30aULL,0x889a042b01448342ULL,0x945636a20b04419aULL,0x100301008000102ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1002000440000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8ba3f77e8bd7f17aULL,0xcf6af71c83c4d34aULL,0xabeecfdbcbd5d37aULL,0xbebf5cf38345ebeaULL,0xd563a26a4b55c00aULL,0xfe962f7fcbcde3faULL,0xedf6eb178bc7a2caULL,0xaf1d5ebf83c7e1daULL,0x92a7c7faa3d4d33aULL,0xf3eadffb8bcdfb3eULL,0xe31a3d78cbc4e2a2ULL },
	{ 0xfffefffbcfdbff3eULL,0xffdebfffcfdbfea6ULL,0xddfbfffbcfdb7efeULL,0xef7effffcfdbfefeULL,0xf5fffff6cfc3fbf2ULL,0xfede7f62c7cb7e36ULL,0xffbffffec7db7fe2ULL,0xcfdfbf334cd2fea4ULL,0xeb9fdfda4fdbfea2ULL,0x9cfed7e74fc37ef2ULL,0xdfffff7bcfdafee6ULL,0xfbb73fdfcfcaffb8ULL,0xeff7ffbacfdb7bf4ULL,0x8b9adf5a4ddb7e60ULL,0xbdbe7fff87cbfecaULL,0xf96cf77ecfdb3fe0ULL,0xff9e7fffcfd3feb2ULL,0x41c7c575cdc27e60ULL,0xbbffffdfcfdbffb6ULL,0x97f9fffacfcb3aa8ULL,0xf5eef7fa4fcafec8ULL,0x7df67d7bcec2faaeULL,0xcbcff3ff4edbfeb2ULL,0xa4c0b21e048af860ULL,0xb7bfcdf94fcbff22ULL,0xb19c3269cfc27a90ULL,0x7f7efdf1cedb7e76ULL,0x7100eb314bc26a20ULL,0xa64bffab4bd8fe32ULL,0xfe9fcffbcfd3ff6aULL,0xf6b37feecfdbfeb2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x79d6e72acedbfef2ULL,0x2212778046c22e20ULL,0x8732dfb10ec03822ULL,0xf9fdbfeb4fd3ffb4ULL,0xd3b67c404cdaf634ULL,0xea00330888c22ea2ULL,0x26fe1f434dc93e82ULL,0x239fcf72078a7a20ULL,0x9419fe204dda3a84ULL,0x1aa42d6045c3aa14ULL,0xa300c1644ec32300ULL,0xc5cbffcb4ec27ff0ULL,0x8f3adfd34fca7af0ULL,0x945707ea47c37a48ULL,0xab12ad524e8a36a0ULL,0xc3b427c447d23a30ULL,0x0,0x0,0x0,0x0,0x0,0x20104002000ULL,0xebde5fdf4ec2fe20ULL,0xd8bb5f7b4ecbbfa0ULL,0xfb7fff6bcfcbfab6ULL,0xceae2de7cecb2aa2ULL,0xc4e6d3364eda7e10ULL,0x41cd59d98dcbba4aULL,0x37195ef347cabf20ULL,0xd1bdff764fca7ebcULL,0xfe3e77774ec32e34ULL,0xcbf5f34f4f9bffbcULL,0xab5a32dacfcafe0cULL,0x393bcde8cfc39a64ULL,0x96b7d4f24fd37a82ULL,0x8e2c2d604dd3d80cULL,0x207e32203c03e80ULL,0x4c02054806881600ULL,0xbb3ecd738bdb7ae0ULL,0xfcf276ffcfd3fe70ULL,0x5f9eff7387c2efa0ULL,0xf316f3f44eca7f22ULL,0xcedbfe3a4fd37b1aULL,0x2bffdfdf0fcbffb8ULL,0xfc3e7cb64fc37f22ULL,0xc5a6e7f34fdbfaa2ULL,0xef9ff77fcecbfef6ULL,0xc7d66f1ac5cbbe90ULL,0x28089e5b0ed27ac0ULL,0xf4984abb4e1a76a0ULL,0xc86c4d400382be22ULL,0xfabc7d7d4ecbee28ULL,0xc9c3de73c7d83ea4ULL,0x6b7ff2dbcedb7e3aULL,0x77bb67fb4fc3fa10ULL,0xd1f6f5fec5d2fe2cULL,0xbfe677ed87db7ba8ULL,0x8885932146c30620ULL,0x461236910c83aa20ULL,0x9339cd4a0fc2ff82ULL,0xf73ed6ef4ed8fea2ULL,0x66125450404223a0ULL,0xdaf0c32645131ff4ULL,0x8a6a336a4c527eaaULL,0x8489cb208fc18a08ULL,0x74b7ffe7cdc3be70ULL,0x861080802020ULL,0xdbbc7baecfcb3ab0ULL,0x952b32306c3a720ULL,0x9de095b4d0a7a30ULL,0x3130d64847032c20ULL,0x9b18a5504ec0a604ULL,0xc7927554cf817a28ULL,0x80a4ae3746c05e80ULL,0x16945846807204ULL,0x4b0ce2306014a08ULL,0x3000b42283822a00ULL,0xf0843e4003400e00ULL,0x20800002020ULL,0x26540e200a432620ULL,0x4199458104c03e00ULL,0xb08d920805c12a20ULL,0x2044004020000ULL,0x100040008001000ULL,0x2206101802422240ULL,0x80c8000002280ULL,0x8410300000103080ULL,0x513a094001830a20ULL,0x150040002005200ULL,0x96f3b25046cbbe24ULL,0x8810c95206037280ULL,0x0,0x80416044921000ULL,0x0,0x144000000020300ULL,0xa802ee9205593a02ULL,0x10e1044004121600ULL,0x3460737387824e08ULL,0xca827fa4460a02a0ULL,0x81462820c4832210ULL,0x48ba1e4ec4c1beb0ULL,0x12400240800220ULL,0x380c00260498b082ULL,0x10000a2000085800ULL,0x85c3ae0a0ec97a80ULL,0x5bda1a71471b3ae2ULL,0x1c7b1fb10cd1faa0ULL,0x993d2f3b4ed2fb86ULL,0x7f367f52c7927ea8ULL,0xc174fb3644db3fb0ULL,0x67ff2ffe0fdbfe90ULL,0xb0a1cde14ed17a04ULL,0x3414c3b847c22eb2ULL,0x24104000020620ULL,0xf145e16e4cd9b2a8ULL,0xc4eafef34dda3a2cULL,0x26050000004038a0ULL,0x147036cd4cc232b4ULL,0x22d8ec0046c87600ULL,0xbb80df7e4793bea2ULL,0x8051fa82024227a0ULL,0xa7d797500fd37b88ULL,0xfcbffeffcecb6af6ULL,0xd9c075404c8ab0a0ULL,0xfe3c2fb6cfdafa78ULL,0x4682eb2e44835f2cULL,0x505a0202812a80ULL,0x1826010a09c09800ULL,0x4021a44205820820ULL,0xcaac3da2c6c8f030ULL,0xcf8ada364dd33a36ULL,0x1944a03444a7910ULL,0x8090046006822640ULL,0x110035018c807200ULL,0x4802250400c30a40ULL,0x8981b2bb0c432f80ULL,0xfcdbfef7458b7828ULL,0x7bf5e9b478bbaacULL,0xd47054e00692ff20ULL,0x421e0db04882eaa4ULL,0x8024636449015a24ULL,0xcb43fdca0edabfb2ULL,0xaf5a499bcfc37e20ULL,0x76f1ffe40fdb7e2aULL,0x2100510044a0800ULL,0x0,0x840000000400010ULL,0x0,0x0,0x0,0x0,0x0,0x100002080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0004800000ULL,0x0,0x120d4101400800ULL,0x0,0x0,0xcbea657f4dcb3eb4ULL,0xddd7ff3fcfdbfafcULL,0x8f5fffd74edbfeb4ULL,0xaef71feb4ecbfe34ULL,0x480446524c003006ULL,0xffa63f73cfdbfea8ULL,0xc1f5c22ccfcabb30ULL,0x4b1d4f1b4cc9fe34ULL,0x3faa6ffbcfcbbe6aULL,0xf37fff5e8edb7beaULL,0xfefe7d7546dbfea6ULL },
	{ 0xfbecff73f872bbb3ULL,0xfffeff7e2a72bbbfULL,0xfffffffbea76bbbfULL,0xfffffffbfa73bbbbULL,0x75ec7efbaa77ba9bULL,0xd3beff703a161bbfULL,0xfedcf77fe8723a9bULL,0xdcd6be232a52bab3ULL,0xfbdf97d6e8773b9bULL,0xf4f2f7edfa529ba7ULL,0xbbffff7bfa76bb9fULL,0xf7a57ffeba533b9fULL,0xffcfffbefa739bbfULL,0xaff6ded3ea529bbfULL,0x3ebfffedea523abbULL,0xf9fffffa6a729b9bULL,0xfffeffffea77bbbfULL,0x87daf20bea723a2bULL,0x2ffeffffaa761bbfULL,0x9fbacffaea763bbfULL,0xf5bfaff5ba523babULL,0xfbfe7fa2ba739abbULL,0xebfdf5be7a763babULL,0xeeff9f7bba723abfULL,0xffbbcfefaa733b8fULL,0xfdfefeeb6a729bb1ULL,0xf3a66f7b6a139a9bULL,0xbbac7737a852bb3fULL,0x4c5276faaa721aa3ULL,0xaffecfc1ea723b8bULL,0x6e0084002a0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xebe6a97ae872128fULL,0x1010420020420001ULL,0x97ba67e7ba563a9fULL,0xd0351c7988421319ULL,0xb6be3f74fa52bb07ULL,0x89b270432012128aULL,0x6cf79fd0e87613bbULL,0xa6bbcfe92a5633abULL,0x9708326028111a8bULL,0x333c3ca008723283ULL,0xebc0d51e087232abULL,0xbed3924b68570aa3ULL,0x158d575338523a93ULL,0xa5cd176a8872328bULL,0xdbaeaf76a86232b1ULL,0xf9842755a843323bULL,0x0,0x0,0x0,0x0,0x0,0x8060004a8020108ULL,0x675c5edaaa72b3abULL,0x866f3ff7aa5213bfULL,0xffffff7ffa77bbbbULL,0xf9bf5fff6a72babfULL,0x8ac2ebbaba701b93ULL,0x225a92da2a521a8bULL,0x8c3f7ef98856ba9fULL,0xd1f586feaa561b8bULL,0x5bb63f5e2856bb9bULL,0x8b84ef3eba703a8fULL,0xefffb7dbfa77b2bbULL,0x8b3c7fdfa87232a7ULL,0xd5d2fbdeea5712bbULL,0xe9ee65f27a5232afULL,0xc30220300a500083ULL,0xa425a8a28701882ULL,0x2d1fcdc2a27212b3ULL,0x15d636aca870923fULL,0xc7bdfd70aa27329eULL,0xeb826d7aaa123b9aULL,0x4edf9cb7a05212bbULL,0xb254d753087212b3ULL,0xd0fff6e0e822831bULL,0xcd8fe7622873bbbfULL,0xfb9e7f7efa521b8fULL,0x69cfdea8e872b28bULL,0xabfef75be063128fULL,0x8cf25afb486238bfULL,0xf2f6b7ec5a739aa1ULL,0xda162665a852321fULL,0x6100c6242a6001a3ULL,0x2d7ffdd8ba529a8fULL,0xbcbb4f09a812139aULL,0x5564a4fc88721b21ULL,0xdf5c2d65a012908eULL,0x695e536eea72b18fULL,0x9d73a95b8523a9fULL,0x8d9a797ab872130bULL,0xbc57b4fa3a72ba8fULL,0xdffe2fe2aa52b28eULL,0xc9e453488852123aULL,0xe8c3f23f7a563a9bULL,0xa5a94c50aa521283ULL,0x5454460a4a001883ULL,0x9102a730a242800fULL,0xc900ef3e88721b9bULL,0xce8ffab0aa531ba3ULL,0x61200421200ULL,0xd01400200870010aULL,0x203861e028722202ULL,0xca9e6f1c38523317ULL,0xcd6ea6822a720293ULL,0x434f1ad0e862220bULL,0xd484b8328721202ULL,0x484c310450401283ULL,0x10a41e020a121901ULL,0x80400008000100ULL,0x940000400021201ULL,0x24000020010ULL,0x9346f2f15840ba81ULL,0xc7b024508a428313ULL,0x8820500080121000ULL,0x8d715d22a42b28bULL,0x62b445e048721a01ULL,0x92f332404a601a01ULL,0x204c020800000bULL,0x2004010208000082ULL,0x409bb619b8421a8fULL,0x86080420028228ULL,0x4000800000600000ULL,0x118a010022121001ULL,0x880400a48101801ULL,0x4600701228321800ULL,0x2c7a050308521a36ULL,0xa0f656cce852123fULL,0x830aad710a523b83ULL,0x60a22f4180523219ULL,0x892ac10aa8121b0aULL,0x2080800128508080ULL,0x43c26c0e8020091ULL,0x8101301060400209ULL,0x1ab4273000663a00ULL,0xc8c07202a8423092ULL,0x206300608413a01ULL,0x818acb52e8423a23ULL,0x1c757fabe8179bb3ULL,0x7bb62f74f85291abULL,0xb5af312a85213aeULL,0x6ecb96d8e8569299ULL,0xb29bccf388771ab0ULL,0x341266e4ea501291ULL,0x80a7d5008001a1fULL,0x8bd3030a42021a07ULL,0x48c9122f1876b281ULL,0x9084c1a0052120bULL,0x61503463a862b890ULL,0x82a0ad6088021a89ULL,0xd3a6fb662a52930bULL,0xc9d2de102a023ba3ULL,0x92c3df5aba729aa3ULL,0xf6fa56ee6a7713bbULL,0x348c322a8522207ULL,0xf995ff6fba533b9fULL,0xe5fbee28e842b23fULL,0x1fde5f81b8721a8aULL,0x34ffd6636a521b0fULL,0xe9ec747a2a761aabULL,0x8608256068038291ULL,0xd1661d7418563331ULL,0x3808968002220bULL,0x108428020100206ULL,0x122056023a523083ULL,0x7fccfd7268731bafULL,0xa885e30a2072ba27ULL,0xa222110521210ULL,0x44048000240000bULL,0xc144176868621301ULL,0x4bc2570a8621321ULL,0xb9b8fbacba52120bULL,0xc63a3d112a5232abULL,0x800014000501200ULL,0x7cf7feeffa72baafULL,0x8000200000061200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x40020000000aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3000000008000080ULL,0x0,0x0,0x0,0x0,0x0,0x1000002010001000ULL,0x0,0x0,0x0,0xa50008600007ULL,0x0,0x40080020020000ULL,0x0,0x0,0x0,0xfbb4777f2877339bULL,0xefcfffaeaa76babfULL,0x8bfedf5daa523a9fULL,0x74fe7fe328773bbbULL,0xfb59fff228763b9bULL,0xff8e3f6e6a739b9fULL,0xcff3f7bd2876b38fULL,0x8b56000a0a62031fULL,0x800843b1e052101fULL,0x70bcbae83252b201ULL,0x4bbc2f432852b38fULL },
	{ 0xffeefffbef7d1ffcULL,0xeddfef7f6f7e1f5cULL,0xefffffdf6f7e17dcULL,0xbffedfffef757ffcULL,0xf5f5f66a6b775dd0ULL,0xdfea7c74417659d8ULL,0xefb47fffe77f1ddcULL,0xdbfbfebba37f2cfcULL,0xfadfdfdf6f7e1fdcULL,0x56fff6ee67775d48ULL,0xf8ffff66637f1fd4ULL,0xfab47fefe77f5efcULL,0xcff7ffbfe7767ddcULL,0xefdfffdbef7f1dfcULL,0xcdf757ffef7f5ee8ULL,0xb17eeff903745fd8ULL,0xf2942fefe37e3fdcULL,0xc5defbff63775fecULL,0x93ff7f4b637d5cd8ULL,0x93bb4ffaab7f5d5cULL,0xd1ecfef8e7761a4cULL,0xfe9ef7f2677679c8ULL,0xeb97dbbae7553f5cULL,0xcfd7bf156a751f48ULL,0x65bacdc9667e15c8ULL,0x956e766eeb7f1d48ULL,0x5abe7d536676345cULL,0xe3c5f33ee7551dd0ULL,0xec6bf79d61773fecULL,0x8bbdefd86f7f1fd0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe8d7df3861751c54ULL,0xa9ce8e886f760070ULL,0xadbaaff3e97f1bdcULL,0x85809640024c1150ULL,0xc15c7e60217610ccULL,0x4b82e93781561cd0ULL,0x606220821511c18ULL,0xbfb96feea53d3fe4ULL,0xddf676e9c3362e64ULL,0x541e7d7163751360ULL,0xf39cef6baf7f1fdcULL,0xfb573ffbed7d1f54ULL,0x2e9f0be163741f4cULL,0xcc36afefe53f1df8ULL,0xc3beffe2e3741fe4ULL,0xff94ef2d687f5bc4ULL,0x0,0x0,0x0,0x0,0x0,0xd6c32619e0761accULL,0x6f5fdfdee3761b4cULL,0xbfd5dda3027f1374ULL,0xddf8f770ae7c3d68ULL,0x76b31fdb2f751ed4ULL,0xc4783600424617c0ULL,0xab0f439140765c44ULL,0x1f9b4fe366764ef4ULL,0xfb6cff7fe3775be8ULL,0xdaa47ded6f7417dcULL,0xe1d7f20e615713d4ULL,0x5c8f9e49a71f1dd4ULL,0x28ecfa262541450ULL,0x708f86e1a6340a84ULL,0xc4ec75e2a4157ed4ULL,0xb8fce34e4742ae8ULL,0xfed236ff627e5e4cULL,0x8bbf1f5a6f7f15dcULL,0x10ef86a821645dc4ULL,0xd2be4d70e2374ac4ULL,0xdbc6f73467741cd8ULL,0xcdeffb3d677e7cd8ULL,0xafdfff53676f5ddcULL,0x77b6fffa677e5f4cULL,0xdfaff773e77f5decULL,0xfb743fdc627e1edcULL,0xdfd3ffbae7745dfcULL,0xafffffdbef7f5ffcULL,0x40606e1024e0240ULL,0x5b4c276ae20658ccULL,0xc2822f592b7619c0ULL,0xe184f3bee3741dc8ULL,0xecf27c9dee7d19ccULL,0x56996f7b6c753fecULL,0x9184b07f42567bd8ULL,0xfe9ee77de37d79fcULL,0x88a8d5aa63741404ULL,0xea1a90c227460048ULL,0x8f084cc265741040ULL,0xd594a271657c195cULL,0x7baefdc163761ff4ULL,0xf222e30e63773348ULL,0x19d0336363520ec8ULL,0xbbbcdf33c0371becULL,0x456246823040cc8ULL,0x9f807550087519c8ULL,0xe12cfe71e33629c0ULL,0xfd2ff2a407e08c0ULL,0xa851cdcfa2741dc8ULL,0x9490e5c026741cc8ULL,0x504016824001050ULL,0xdd22296e21741fc4ULL,0xcc7dd06417631d0ULL,0x244020c80040884ULL,0x404000000100200ULL,0x1302e00005214d0ULL,0x9034873003761644ULL,0x4480612060544a00ULL,0x6000060000000ULL,0x9451140200344a40ULL,0x100812002040050ULL,0x5eec604074027cULL,0x2106200322010050ULL,0x800020240100050ULL,0x8102088102741000ULL,0x1000a64000140084ULL,0x200040000100420ULL,0x8900e020263c0060ULL,0xc983ac100651258ULL,0x80354e20265610d4ULL,0x14042880251808c0ULL,0x80280c0080440940ULL,0xc100713180144a84ULL,0x8cc0301c62361414ULL,0xa01a1fd165661100ULL,0x8030506a21445bc8ULL,0xf1aa2470623403e4ULL,0x50282470230419d4ULL,0x88c31600e0101e00ULL,0x2a42821024521218ULL,0x3080048840170040ULL,0xc9425468414f18c0ULL,0xb012bf1d295e10d4ULL,0xc7d2febae1340e44ULL,0x2952ba04214d1cd8ULL,0x632672023781e48ULL,0xd5bc97026e741440ULL,0x71a23c02a65518d4ULL,0xc37df2bfa37c1b5cULL,0x843325926080158ULL,0x16bfddf7e3763dd0ULL,0xf577137c66777cc8ULL,0x8050026040140ULL,0x80d4e264663c48c4ULL,0xf9d2165b83746ac4ULL,0xe3bb5e72a6741bc4ULL,0x8512d6c800361e40ULL,0x98040560a2700860ULL,0x470dc022253509c0ULL,0x6bd6ff1b6b6719d8ULL,0xa75edf1766791d48ULL,0xd4b67eeda6775fdcULL,0xedec377a41761bf8ULL,0xd7967ffdef775dd4ULL,0xd1a0fba2023c0ed0ULL,0x6a3efd5a62775fc8ULL,0x9c3ac773027e035cULL,0x0,0x880d4001360a50ULL,0x108000100080ULL,0x2900726640000ULL,0x1048a223141000ULL,0x8159a63527631940ULL,0x6802502002081148ULL,0x8344bba663560dccULL,0x49504e1060661408ULL,0x23b28e10243138d0ULL,0x84e3706d224c5cd8ULL,0x108a544002141304ULL,0xc316e32a017c036cULL,0xc1a1a836025e48ecULL,0x198ccd4827141c74ULL,0x947776ec6f6f1dc8ULL,0x200c00001000c8ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40800200800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8080000000040004ULL,0x101000000050ULL,0x0,0x0,0xb3983c506d7518acULL,0xdfc7fe22667419e4ULL,0xeffedf53667e3dc8ULL,0x7cfaf36361760bc0ULL,0xf96c6742e5344a6cULL,0xfabe2fffe77f3fe0ULL,0xefdeff7e67755f5cULL,0xe0dedbd265371cc4ULL,0x34da5753e5763944ULL,0xf5fcf6fee7777fccULL,0x561466f56d751fc0ULL },
	{ 0xeff7ffd31fedefffULL,0xf797fffb3f6cdffdULL,0x4eebbfd91f6ce3f3ULL,0x6defdf789e6defdbULL,0x356d7ee81b0dd57fULL,0xdeb63d50374cc591ULL,0xff94476f3f0ccff9ULL,0xc6dff73b8e6cf7d7ULL,0x2fd7ef5f9aaddddbULL,0x3cf656cc0b0de7dbULL,0xfbf6777b1f29fdf1ULL,0xffb27ffd2f4cd7fbULL,0xc3def7781e6cf7ffULL,0xefdedfd93f2dc7fbULL,0x9ebeffff2becff9fULL,0xdffffffabfadef9dULL,0xf6b67ffebf2dcdd7ULL,0xeffbff7e3e6dcfdfULL,0xefffdffdbf8cffdbULL,0x77b8c5740e4c67c4ULL,0x98525558290ce575ULL,0xe79e3d7cbb29f5d9ULL,0xebbef77e3e2cffdbULL,0xffdebf5b3facf7fdULL,0xbba4fff3278dc79dULL,0xb5ffff7f1fedffdaULL,0x290c0da0032886d3ULL,0xfbd44f3e028dd7b8ULL,0xdee7b7ff3facf7ddULL,0x2200000000080040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcfd2ffef9a2ccfd8ULL,0x365c000088855cULL,0x368aa5f50828a11bULL,0x83e7442082ca5c2ULL,0x5ebe6d731e28ff99ULL,0xcba3f103030d5d10ULL,0x65f91e198e2cc6d8ULL,0xdfdf4fbf1f6dcff9ULL,0xc4322670a84c85c2ULL,0x6bec7fe08b6cedf3ULL,0xc98477170f2cc49bULL,0x4ffbbb5bafacffdbULL,0xffbf57fb3eaddffbULL,0xfdf676f6bfecf7bdULL,0xd7fe6d71ae4dfdf9ULL,0xdeb4bf7b3aad47f6ULL,0x0,0x0,0x0,0x0,0x0,0x8dc7b7b3ab28c79bULL,0xcbf7fd581bac87f9ULL,0x60344de38708d516ULL,0x5ae0fd62040dc6b5ULL,0xaa32af641b2ccddbULL,0xc5c1b6101b00c789ULL,0x211ed7583a0ca75dULL,0x37ddfff71bcdedf2ULL,0xd7dd2ffc0f6dd2d5ULL,0x7290234c1a2ca092ULL,0xbf1f30f234cc598ULL,0x2ec5beca2a2de79bULL,0x980dcf333a05e750ULL,0xc47df471ac20c7d3ULL,0xeffe7dfb3f4dffdbULL,0xf307ce1a2ae5dd9bULL,0xe9213ca1a24cd02ULL,0x863ecc650009cc85ULL,0x4b06644064cc418ULL,0xb3bf3f20a829dfd9ULL,0xcd86f37f3aec9793ULL,0x8a83fea71f2cc7daULL,0x883f5e5a1a28ef5bULL,0x4e617691a89cfebULL,0xffffeffabf4ddfdbULL,0xdebe67dd1e28cfe3ULL,0x6ed3f20e1b0ced91ULL,0x2beedfd30a6dcfd4ULL,0xe4bbd7e92a28efd1ULL,0xa10a252013084990ULL,0xfeffafff1f2deffbULL,0xcffaf7393b0deff1ULL,0xefffee7f3f0defd9ULL,0x99a237fe1eadd55bULL,0x1ddcf7f03e6ccddbULL,0xfe9467641f2cc4d9ULL,0xc083b90b0b28c7ddULL,0xc4da36499a2df713ULL,0x3a89218412000110ULL,0x80cea3381f0cc49bULL,0xf9feec711e2deff9ULL,0xdb08f76f0e0de759ULL,0x4fdbbdf70e0df7dbULL,0x4eb94d402a2887d0ULL,0x10f0ccc20b28c9b0ULL,0x80d840200a2c8401ULL,0xc2807401020885c8ULL,0xc94bef110b2ccdd9ULL,0xa94cc4912014510ULL,0x963524601209c711ULL,0x49642408020c8109ULL,0x4f00634e0a65cf88ULL,0x40820d0204084502ULL,0x308da92080c851aULL,0x4c205020800e881ULL,0x4000a80002095010ULL,0x890005011a048581ULL,0x5244608002c44d00ULL,0x8004002000100ULL,0x3000465220008391ULL,0x90909040282081d1ULL,0xc2aaad422909f190ULL,0x834460462008cbb0ULL,0x4d01a003208071eULL,0x9409042228000381ULL,0x1000004000008010ULL,0x50002000100ULL,0xc044452c1a21c00aULL,0x240021002008302ULL,0x2180291080881d0ULL,0x109004620788cd91ULL,0x52b22c501a28a470ULL,0x92918d64210cc549ULL,0x50d18812a08110ULL,0x6211d45a0a09cdd1ULL,0x1020624008080180ULL,0x102a800200c0510ULL,0x9a8177241a4de55bULL,0x9c2e1140300ccd8ULL,0x248014822880301ULL,0x940600830804ce43ULL,0x5900002000048110ULL,0x91020000208010ULL,0xc5fbf3181e4ce7dcULL,0x2c5f2fdb1e08db9aULL,0x279a67da0fa9cff9ULL,0x71e435299709d6d1ULL,0x7a220d701a08859bULL,0xe9f7f31f3f2ccffbULL,0xcfdf3fd33f2cd5daULL,0x2e124d581a28a199ULL,0x9130a646052cc5c4ULL,0x448c0560002ccd81ULL,0xc015ca28014dc491ULL,0x8cc136200acc871aULL,0x2dd6c7101a2c845aULL,0x94f27e84b888a753ULL,0x9b8bbf719e4dc7dbULL,0x41006479170cc409ULL,0x9fcdeb238309d59bULL,0x1261bd11a0ca3d1ULL,0xfcb4fee91f2cd5d3ULL,0xdb65e77a2eccefdbULL,0xc0bf2e52090ccd53ULL,0x4542a2bc0b2c87d1ULL,0xf1c1384060ccd17ULL,0x991257622a08d379ULL,0x20060000080101ULL,0x89340c7008080703ULL,0xe144d5ac0a08e4d2ULL,0x150160000008180ULL,0x84a9000808181ULL,0x9b38277c3f2cc359ULL,0x280000004004000ULL,0x284c0220600c518ULL,0x86c00a020a214281ULL,0xbf99d9fb1705cf9bULL,0x1560036a8644c40bULL,0x252225200b24d490ULL,0xb98423790c0d7ad1ULL,0x674a7eb80a2c86dbULL,0xc7fe4d693a08a799ULL,0x31dc166b0d2ce6f7ULL,0x14008000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000008001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000080400ULL,0x804020000100ULL,0x0,0x80c00200000c012ULL,0x0,0x0,0x0,0x0,0x108c01a0880c0458ULL,0x44802754020d4729ULL,0xc6dfff9e3f2dd5dcULL,0xaffd725b3e0cffdbULL,0x94e715280a8cef1bULL,0x797efe7fbeacef5bULL,0xcbfe3fff9f8cffdfULL,0xcff6f7bfbf2ceddfULL,0x2f9f7edaaa6ca7dfULL,0xdffeffffaf2de7f9ULL,0xdbf4753d3f4dd7d9ULL,0xc2b075d00b4cc759ULL },
	{ 0xef7ffff7cdbafffbULL,0xebf4ffffddbadff3ULL,0xaffbffdfddbafffbULL,0xdffefffaddba7f2aULL,0xbd76b5ef05bacef3ULL,0x5bf2ff72ddba5de2ULL,0xdb9cfb7e5dbccf73ULL,0xcdffbf941cbe4fabULL,0xbfdedccfddfa77baULL,0xd0f4f6ec159a4f23ULL,0xf5ff7ff29dfe4f3bULL,0xffa677fe5dfedfe7ULL,0xcdf6fc0a5dfa5f33ULL,0xefdfdf538dfa5f53ULL,0xfcfb5fe9d5ba4feaULL,0xf9e6f75f4dbaef6aULL,0xcaae2feed5fedf67ULL,0xeffbffffddfeffebULL,0xfe7fff6f5dbe4f6bULL,0xa6f27f620d385d6bULL,0xd57d367dcd3a5f6aULL,0xff967ff7ddba6ffbULL,0xabfce7be99ba4ff3ULL,0x206f93e1418e616aULL,0xfed95f3a4db84be1ULL,0xd7e6c7740dfe4ff3ULL,0x5b2c3d621daeefbeULL,0xdbd6cd3fddba47e2ULL,0x8ee27ab00dbe4f6bULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280004000100ULL,0xe1cfe93645ba1522ULL,0xab5eb6120df84f63ULL,0x14b29ea99488073aULL,0xbffdbe64c09ae7e7ULL,0xdadabdf155be0d62ULL,0x8152f3614d3a0d2aULL,0xd286501c820540ULL,0x1e894d49509a0da3ULL,0x21ef076c849a1c0aULL,0xfbbe7dfbddba6fe2ULL,0xf394e72a4dba6f62ULL,0xea5ffbdbcdfaef2fULL,0xffbfffffddfeffffULL,0xf5dfffebddbeff6fULL,0xc98eff71dd2a5d63ULL,0x208040418000500ULL,0x0,0x0,0x0,0x0,0x0,0xa932ffaac8becfeaULL,0xea77d7d60dbeed63ULL,0x8010a01220002ULL,0x943025600818472aULL,0xfabe3cc0dc00654aULL,0x48c5ba84541a0da0ULL,0xf5e19c80d9edf21ULL,0x867a535b8dba4d2aULL,0x30c476c04092cfeaULL,0x55027b2094180d60ULL,0x8957aba2c492cd62ULL,0x675dbe504dca01faULL,0xb398cfd215aa17f2ULL,0xf534f6e2cdba07f2ULL,0x6c060fe2049a0641ULL,0xcb84ed7005be7de6ULL,0xf949bab58dbe4f22ULL,0xba3f6f6089beee32ULL,0x4c503e69599a4d22ULL,0xfef72d779dbaf763ULL,0xd31eefefdd9adfa0ULL,0xcfcfff1b4dfaef62ULL,0xbbbddfdf9dbeefeaULL,0xfcfd7ce45dba476bULL,0xdb5ef36acc824f73ULL,0x8e942e6554b865a2ULL,0xd3c7ff9a91facf47ULL,0xc27e9cc345ba61e2ULL,0xfcf7ffefddba5ffeULL,0xd978ee7a55baceabULL,0xffbfbfffddfeffefULL,0xffffffffddfafffaULL,0xcddefe9f5dfe7f62ULL,0xb5c8bff9449add6aULL,0xf1fff57a9dfedd6fULL,0x6e9e37701cfecc62ULL,0xcbc4fea2949acd7aULL,0xd8f39e5e589a4f6bULL,0x1f126f5a55204ee2ULL,0xd7b0b7de0df8f86aULL,0x95bc6d701ce8dd6aULL,0xa350e7000c9c4142ULL,0xe8b96494db8cf6bULL,0xb2e09534dba0d23ULL,0x9450a5ec4cb84901ULL,0xb263da2548a4d2aULL,0x89921271499a470aULL,0xc382b22b892a4da2ULL,0x1ba9075215e8d540ULL,0x94c0926001b00962ULL,0x207602108a4442ULL,0xc3000794488a4104ULL,0x90b08f0a4590174bULL,0x25b1152c1b4cc22ULL,0x12000000a0ULL,0x94c206801a08820ULL,0xd300052204205b00ULL,0x8020600001200002ULL,0xad25050ae0e20ULL,0x320a650201b84022ULL,0x9524017800a8c462ULL,0x2010041008280100ULL,0x100800000080002ULL,0x212114090c420ULL,0x880005a004084120ULL,0x4027040088a4120ULL,0x592514109020b03ULL,0x5000d80044a80100ULL,0xb4e0b0f4ca80320ULL,0xb3a4dfbdcba4d01ULL,0x54d056ea0c9a5d24ULL,0x5e3af512d4a84ca2ULL,0x8904113489c48502ULL,0x9018a01048a080aULL,0x80448120810050aULL,0x3420540000186c03ULL,0x91ad8b700c9a5d20ULL,0x8880672d4caa1961ULL,0x846220000226842ULL,0x4b02101900286862ULL,0x1004048010284600ULL,0x5080000000c80020ULL,0x0,0xcfa7f6175db8c77aULL,0xaf6f1a599d9a4da6ULL,0x16230fb1d8024d12ULL,0x8101169189884500ULL,0x1a9a3e500db807e2ULL,0xa964e3395de21d3eULL,0xa84b1351c9a24d6aULL,0xa5134d928c92e40eULL,0x54b4b8240cfa4cf3ULL,0xc1965fd085ea074aULL,0x59d4c12208bacf00ULL,0x28ca304d48920352ULL,0x2d0dc98049c80b23ULL,0x78123472158ad426ULL,0x53aec450442259f2ULL,0xfd60cd5b5d385ff3ULL,0x7063c0a4c98cd01ULL,0xb287cd1accb20f20ULL,0x47154e008ba4f62ULL,0xe5e7ff6b5c9afd3aULL,0x2bb8ef3509bacfaeULL,0xc673ce1015bacceaULL,0x22879355c4a8642aULL,0x2401497305b8040aULL,0x886c2662c48a0540ULL,0xec9c3cd004e20347ULL,0x861201080c004100ULL,0x241a10d05c9804a0ULL,0x3ea9c7521cfa876eULL,0xaddc140801ba6a03ULL,0x82822a0001220352ULL,0x61842005048a4840ULL,0x69e633f75d2a5f63ULL,0x90060004880002ULL,0x80101408c0820500ULL,0x1e262542dda04721ULL,0xe891420c44a88062ULL,0xbcc7ff5d9d9acd6fULL,0xfb9e8febddba4f6eULL,0x44010200000000aULL,0x3a00243044c0016aULL,0x4000200000100400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10004001000ULL,0x21000000820ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400008004000ULL,0x0,0x0,0x0,0x880820080004c00ULL,0x0,0x0,0xdd04b7600caad1e2ULL,0xdebe3f761dbec56aULL,0xd1237a310daa4b00ULL,0x4a40120904905801ULL,0x4535f7ef0dba4ff7ULL,0xa967224289ba8debULL,0xffbe3f78ddb87fabULL,0xe7bdf3245ca84f73ULL,0xbfdfffd9ddba4f6aULL,0xbcb7dfeeddfadf6aULL,0x552d07790db4f56aULL,0x5d9c6fb20dbadf7aULL },
	{ 0xdbdffffbf6336eeeULL,0xffdfffd7b613feecULL,0xdcfffafab633f6eeULL,0xfffffffff633feeeULL,0x5ce7f6f9f41276ecULL,0xfbbafdf0b0136fe6ULL,0xfb8073d63013e6ecULL,0xc2d6b23f16336ae6ULL,0x295fdfd9b43276e6ULL,0x7cb5f6fdb6337feeULL,0xf7fff77d761366eeULL,0xdffe7f71b43373eeULL,0xc4cf9b39b433f6e4ULL,0x6b70dfcbb63356e4ULL,0x36f777eb34336fe4ULL,0x70eff7f93433f5eaULL,0xee3e3fffb433f7eeULL,0xebd7ffbf3633efeeULL,0xfb5a17f3161336ecULL,0x29a4c60841078e0ULL,0x93a9b64cb61162c2ULL,0xcf7eed62361256e2ULL,0xebbeeffff63367e4ULL,0xcbde5eddf433faeeULL,0x37bdcfdb1233f6eeULL,0x9c734c709412e5eeULL,0xba1c7d327431f0e6ULL,0x9bfecb773433feecULL,0x8020000020000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd48adf2aa413e2e0ULL,0xffdb3a5bb631fbe2ULL,0xb48826e80413ece0ULL,0xf189af3e3432fbecULL,0x46527d60f413e4e4ULL,0xc944cb0b2432626eULL,0xa257ef46943362eaULL,0xbe734748b0104dc4ULL,0xd732e67e9633f7e6ULL,0x26b4bd72141337ecULL,0xdbf0c376161375ecULL,0xcecb7e33b6326de6ULL,0x3feacfdb7013eee6ULL,0xfdf776fbf61373eeULL,0xfd48ff31343376eeULL,0x0,0x0,0x0,0x0,0x0,0x8a1020c20022460ULL,0xeff7e73f3632e6eeULL,0xea5695d83413f7eeULL,0x0,0xd548a070943226e4ULL,0x14900052203056c0ULL,0xc487fcbe343266e6ULL,0x4f5ebfddf410ffeeULL,0xbcb2edfa36323fe0ULL,0x51a47452101366e2ULL,0xfacc6da2a43076eaULL,0xeb52ef2b7432e7eeULL,0xe0c218d032104ce0ULL,0x8f9fcffbb613feeeULL,0x1624f4e5063336e4ULL,0xf5be5fe2b61176e6ULL,0xc1e45f303613f4e4ULL,0xaec132923002e3e4ULL,0xbffdc64214336eeeULL,0x14f5f66a3202e7eeULL,0xab3bbd6034117bc4ULL,0xe3a5f36b3622e6ecULL,0xdfdfffbfb633feeeULL,0x8f775fcb161267eeULL,0xc4f477e2e432c7e4ULL,0xbbdbaff03431d7eeULL,0xfa287c7e3633ee62ULL,0xdec2bf9bb43372e6ULL,0x4e9fce4f54114762ULL,0x9db447ea363370e4ULL,0xfbffff7bb633f7eeULL,0xfebffffff633ffeeULL,0xfffffffff633ffeeULL,0xfffffffff633ffe6ULL,0xb3bf4dfff633cfe6ULL,0xf1edf7f8343377e6ULL,0xfb7a6750f6334ee4ULL,0xc336f3fbd4126de4ULL,0x6ffb1f43861372e6ULL,0x8822864104315260ULL,0x9023223020114260ULL,0x1ae0de0361066c6ULL,0xdb96a37e30133fe6ULL,0xeb5e9f4fb433ffe6ULL,0xe90f414080134064ULL,0x90577ef7b4036fe6ULL,0x12be45725612eae0ULL,0x3830ab2a2433306cULL,0x469b9402b43100aeULL,0xcd249021433f6c4ULL,0x90464064104060ULL,0x0,0x18202140011c060ULL,0x1402a33ab01352e0ULL,0x2410008000104480ULL,0x182040204206040ULL,0x80240a0810000442ULL,0xad1005c0b410c8e4ULL,0x435543240610f5ecULL,0x4004400100104084ULL,0x920007010101042ULL,0x1200832080304280ULL,0x800e354004106144ULL,0x8982c022002246c4ULL,0x556108a141108ecULL,0x1202000040ULL,0x540084e45410c2a0ULL,0x146472000004140ULL,0x8200010404204000ULL,0x7527211801162e6ULL,0xe040102342004c4ULL,0x50001245140007c0ULL,0x4000460001040acULL,0xe880522050026aa0ULL,0x14e120030110060ULL,0x82dcd40160224c6ULL,0x1c946a88102226c2ULL,0x801e564014022ac4ULL,0x40102a1000104084ULL,0x100c0081000e042ULL,0x848445420611c1e4ULL,0x1112871b0012e268ULL,0x0,0x240000110000ULL,0x812081100430c000ULL,0x8540e6232100aeeULL,0xbc234de0c433c2e4ULL,0x24150a203240c2ULL,0x7c48646b361146c4ULL,0xc904a222102242e6ULL,0x6cd7b61d341236ecULL,0x1111059612113440ULL,0x74a68662240227e4ULL,0x4a602549101246c6ULL,0xa184c70634127ee8ULL,0xa2421250200223ecULL,0x4412010270000022ULL,0x1081142610100462ULL,0x63829c30821106e8ULL,0xcf00e743903226e2ULL,0x80c30010006842ULL,0x3c6df98141376eeULL,0x30fb7ef5b4335ee4ULL,0xdffd2f7bb6337eeeULL,0xffb0677672324ee8ULL,0xeedf6eaa543277e4ULL,0x14e3b00241279e4ULL,0x847164e01411666aULL,0xd3b2ac04101310e6ULL,0x4a88044c2420ccc0ULL,0x8142f2291633e544ULL,0xc80a3206101162e0ULL,0xb6410d4034004cc4ULL,0x80702410300064c0ULL,0xde00744e94326de4ULL,0xc9cfebeb363077e4ULL,0x6fde3fd0941373e0ULL,0xfbbfc45ba01366e0ULL,0x2002026030030044ULL,0xc6940030b6204464ULL,0x6124420256114164ULL,0x404338103400e2c6ULL,0xc39e4f6b7432f6eeULL,0x4cd506e0401276e4ULL,0x2010000000000480ULL,0x4200600010110400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2020000000060ULL,0x0,0x108000840000e040ULL,0x1be8f70f032c7eaULL,0xfb0e79fc341277eeULL,0xedebfd18163277ecULL,0x2a7eff8314304ec6ULL,0x8432fcec303367e6ULL,0xffbebf7b1633fae6ULL,0xfe1e7fb7b412ffeeULL,0xfeffffcef433fee4ULL,0x9ed7fefb76136fe6ULL,0x9bb347ee723376eeULL,0xf168b7772633c6e6ULL,0xed12016210336462ULL },
	{ 0xdef67ff9f8dc9ffaULL,0x9fde5ff750dcb7dbULL,0xdffffffff0bcfffaULL,0xf7fffe57709cbef2ULL,0xfc76ffebf09cb6faULL,0xfdfaff61f09cb68aULL,0xdaacff7ef89c96feULL,0x5dd7beb5f094fffeULL,0xa9e6854ab0dc36faULL,0xb4b8cbe4d83c3242ULL,0x9df6b572f0fcbeeeULL,0xffb27f7ff8dc76fbULL,0x4adbf6e8509cfec2ULL,0xaffed9db709c94eaULL,0x96733ee7f894b7e2ULL,0xf869e7d770dcbee8ULL,0xdaa3af7670dcb4f2ULL,0xed9eff3fd0dcfffbULL,0xaa5fee5a70543cf8ULL,0xbeb95fd278d4d7dcULL,0x916db67a6054f8f8ULL,0xfa9e75f570bcb7e2ULL,0xcb4ffffff0949eeaULL,0x65ff15ca5814b66eULL,0xbdfa4913709cbee0ULL,0xfd5efeef70bcb7d8ULL,0xbfbe3d69d0bcbff5ULL,0xb2068b0228405428ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6d4edb96d0b4bdd9ULL,0x8956df52f01cb66aULL,0xa205c2b09c10a0ULL,0x600d97cad014b4c8ULL,0xfd9e7dfdc89496dbULL,0xe1fffffff89c9ffeULL,0x577ffaf0dc3ebbULL,0x3ffdcfc7f814bfefULL,0x4504e008504cb060ULL,0x7f3e7ffbd0bcb7fbULL,0x3b2e53e509c3493ULL,0xeb7bbffff09c9ef9ULL,0x7b3e0d50d08c3092ULL,0xbdd77ffff0dcb7faULL,0x91baef72f01cb4eaULL,0x0,0x0,0x108000020000020ULL,0x0,0x0,0xbb126375d09cb4daULL,0xcfcffe36f0dcfe6cULL,0x200a108140801422ULL,0x8ed57463501c9e4fULL,0xd578fdd0f0541eadULL,0xba3e2f75e8fcf6faULL,0xc9d5f73f509cbfceULL,0x6fff7fd7503cffdfULL,0xbee557ea709c947aULL,0xd1699742f0141688ULL,0x6b0c36245804140aULL,0xc8f1fbbac01c9edfULL,0xeafe7fd1f0b4bff0ULL,0x5fbcff4ef0b4beffULL,0xa95da64050849480ULL,0xe4527de3d004b260ULL,0x8b8ba91b5014be79ULL,0xd6c38b52701434e2ULL,0x61646d7e09c3ae2ULL,0x2c369361503c90d2ULL,0x83b8041170140092ULL,0xd21cdb6ed0b4b6f3ULL,0xcfdbff2bf09cb6bdULL,0x2b1dff537094b6eaULL,0xfcb5fefff8dcfeffULL,0x53efaf7b78dcdebaULL,0xd9a2eed050c4b64aULL,0xa1c6ed0a704c92e8ULL,0xcfff7fdff8dc7679ULL,0xb0927d7778d49fe8ULL,0xfffffff7f8fcffffULL,0xffffbffff8fcfffeULL,0xfffffffff8fcffffULL,0xefffffdff8fcffffULL,0xb77bdffff0dcfffbULL,0x9171ff9378dc9efaULL,0xfacabd6e505cbfd9ULL,0xe9f7ff6fd09cbfdfULL,0xfc1a3779408cf6d2ULL,0x508c903501c1a80ULL,0x17661648509c1c02ULL,0x45826552d09c9480ULL,0x8810c3000050180aULL,0x215b3e03f0b4bae8ULL,0xcabfdfdbf01cb6e8ULL,0x142642e0401816f4ULL,0x8030201010102240ULL,0x40242394600452e2ULL,0x144c62018109c42ULL,0x6f188f1650149692ULL,0x5498fca9504c34aeULL,0x54a657f1f804b7d9ULL,0xd38cee56d80cb5c8ULL,0x34c7ce08701485f6ULL,0x1904a0000140682ULL,0xe3320522f0140422ULL,0x60062070b004d0c0ULL,0x8200052050100490ULL,0xc154f57790b412d2ULL,0x8ee3201d01c9cdeULL,0x42271002000a560ULL,0x803712c00c9080ULL,0xb00c3d45901094f8ULL,0x9d0602b101490f8ULL,0x8860c00b01018a0ULL,0x81264d6008349042ULL,0x0,0xc30c4211701c0030ULL,0xc150c502107414e8ULL,0xa200321e0050160cULL,0x938c0120008094c8ULL,0x20160000101410ULL,0x802100220001022ULL,0xc1854012308414a2ULL,0x8000200040040000ULL,0x22c4801903854c0ULL,0x8c9017620090141cULL,0x7620fc060c4101aULL,0xcb1c6662708496ccULL,0x2c947312501496c8ULL,0x82167e566004b4a8ULL,0x332a4febd89cfecbULL,0x8840060030849040ULL,0x8000250040002480ULL,0x2c00800100c80ULL,0x0,0x8490142050109888ULL,0x8520240400014a6ULL,0x40b62ff0f09434d8ULL,0xc9d0ea22d094b6aeULL,0x2d06214060049068ULL,0xb7a94f895814b7fcULL,0xb7f454ea98d49c85ULL,0x24b560508ca0f8ULL,0x1d07202f0149010ULL,0x2b7a3c84401ca340ULL,0x2228c1850108444ULL,0x90db26c8d09496c0ULL,0x813c6d327094f358ULL,0x88002c21d000b408ULL,0x4842301070043082ULL,0x6039d250101c1743ULL,0xc0995669581c74c2ULL,0x9e862560d030a6d0ULL,0xdf9e777d70dcffeaULL,0xaccaef12f094b578ULL,0x398dfc0781c3e80ULL,0x26200c61501496d0ULL,0xd37dfb68f0f4b6fcULL,0xde942fec085c1e9cULL,0x81c1d000600c0038ULL,0x25187e94901c164bULL,0xb000441018001080ULL,0x0,0xf0844d6a701c31e0ULL,0x81208140700004a0ULL,0xfff6ae93f0a4f6e0ULL,0x9aa4cb25050b890ULL,0x4202060001c0c80ULL,0x47a82f40f084f36aULL,0xfa44e262280094eeULL,0xae03da406014ace0ULL,0x903804e0503494c0ULL,0xc30d6424010374eULL,0x200000008040ULL,0x4840000010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20010000080ULL,0x0,0x0,0x0,0x400004000801000ULL,0x0,0x0,0x0,0x0,0x6006000002644ULL,0x0,0x0,0x0,0x14443443600c54a0ULL,0xdbfff75bf0bcf4feULL,0x4cb42f7c501c9e98ULL,0xedd7f7ba50d4bedaULL,0xea0fda51d034fdeaULL,0xc365642700414acULL,0xdbfeff7a707cb7fbULL,0xdebf2f7ff074b7daULL,0x8ac3fb66701436f8ULL,0x2b76db7cd0b47eceULL,0xbf99cffad09cf4d2ULL,0xd5daf722409416baULL,0xe8b42e7470189efcULL },
	{ 0xfbfeffffd9efbf7eULL,0xfbf6be77d9efabceULL,0xafffffded9ecbbdeULL,0xffffffff99ffbbdeULL,0xf4a677fcd9edba7aULL,0xdbbebf41f1e7a3caULL,0xcf8cdeb799e3af7aULL,0x8fcbf6bbd9efbbfaULL,0x297e5fdfd9e6ba6eULL,0xf477b7f5d97fb952ULL,0xd9cbed66f955aa7eULL,0xfb9cbf7cf1edaf7aULL,0xc163e87690c3abd4ULL,0xab98cdda99c2ba9eULL,0x14bdffebf963bfdeULL,0xf9acbe7ed1fbab7aULL,0xdfbf7ffbd9ffbffeULL,0xefffffb7f9f9bbf6ULL,0x6b5fdbda99d9a77aULL,0x3a2e05f891808b58ULL,0x801a8452d1830042ULL,0xded6fd66d9cf267eULL,0x8b6ee769916fbf5eULL,0xeddb1a51f98f335eULL,0x5795fffe99fbaad6ULL,0xfffdfefef9ffaffeULL,0x49083700112c020aULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x851120010460618ULL,0xa34e5e5091e02766ULL,0x98468431082246ULL,0x50eeb77091840a1aULL,0x4eba7c51b1c1bb34ULL,0xcbefbb2fd9e7364eULL,0x8eff7fc89155bb7aULL,0x6fbfcdffd9e9b7f6ULL,0x9cf3e92d5961231cULL,0x61742f709962abdcULL,0xda64fb2ed9ffabdaULL,0x98c63e0198ed2250ULL,0x9b3ccd7fd938296eULL,0x51f6a6f6d94b3372ULL,0x8202004000409200ULL,0x0,0x0,0x0,0x0,0x0,0xfa14f3fcb1108bfaULL,0x8190850ab008a300ULL,0xa0c448a90802804ULL,0x541611e9d1e82340ULL,0x884745688048ab52ULL,0xf298271799e98e38ULL,0xe0c4d71e91c2b36aULL,0xaff9e099948a3fcULL,0x8dbafebb9964bf4aULL,0x3168b66391900378ULL,0xda260f4ef1ac2278ULL,0x83e7e3683967225cULL,0xe9fe5eddf9fda37eULL,0xff9effdff9ffbf7eULL,0x9473f6fe9965b15eULL,0x751605c0d184b01aULL,0xfaa2cf6af1ed2258ULL,0xf5db92b1b9ec2618ULL,0x291ece93b0e92250ULL,0x7fffbfff99e1277aULL,0xbb9e7c53b16da3d4ULL,0xd91ee3bfb1e7bffcULL,0x8febfebbf9efa7faULL,0x60e484099c00e10ULL,0xbcb646ecd14abf5cULL,0xb3fcb7e0d945b37cULL,0xdb965bf491e6267aULL,0xe1c6cf1bd1413b76ULL,0x2f5c56d9f14eab58ULL,0xf4fe57ebf9f9affeULL,0xfffffffff9ffbffeULL,0xfffffffff9ffbffeULL,0xfffffffff9ffbffeULL,0xfffffffff9ffbffeULL,0xbff36ffbf9fbbf76ULL,0xd5eff5fbf9efbb7eULL,0xef46fbe5d9f8a552ULL,0xebd5f3eff9efa3dcULL,0xaec71fd7d9c9aefaULL,0xaa3a4bfbd9c30d40ULL,0xd494d6f8f861ab7eULL,0x5db20dc8d1212366ULL,0xa8dd660ec9c10542ULL,0xa9ca26c1f0c13252ULL,0xbbbadf7bf9ed277aULL,0x100256c441492342ULL,0x5b6afd535104b23aULL,0xc382230d9000a9e6ULL,0x1cd33a08d1458768ULL,0x2026c80290000268ULL,0x94110460c1440278ULL,0x2204201008802ULL,0x54a0e2c3c8d12e50ULL,0x4342e233d8d2264aULL,0x68117ac099c02212ULL,0x10084a8311248308ULL,0x0,0x800014c01419002ULL,0x81b6d20c9065a908ULL,0x7bc69f55d0c4abc8ULL,0x3000000000010250ULL,0x8001500010810040ULL,0x90442200024a00aULL,0x80100a004024a80aULL,0x4c0081000440308ULL,0x90020d60110c8308ULL,0x42024290238000ULL,0x24041000012040ULL,0x80042502c0002060ULL,0x690100000202142ULL,0x400100000080200ULL,0x0,0x0,0xca02a24391e0bb40ULL,0x0,0x385519c211403306ULL,0xc91d06e9963aa7aULL,0x4b243440c968ab00ULL,0x8020112290401900ULL,0x52d80800412352ULL,0x209815899c10202ULL,0x141d4c82112089e0ULL,0x1141000001200202ULL,0x3692055c9965a14aULL,0x0,0x108401c0000040ULL,0x14d14e64d0613240ULL,0x4140103110022140ULL,0x22a62d60d189aa62ULL,0x9d5c20011a5225eULL,0xcdc36d61d9c03354ULL,0x801043719840a112ULL,0x1410ca0091028860ULL,0x102274800490748ULL,0xcb90eb2ed9c9ab42ULL,0x9401d40c9c20140ULL,0x65686b19030224aULL,0x55dfa4ea91e5ae7aULL,0x208050040002004ULL,0x110036c91a02218ULL,0xca52e71790402652ULL,0x286c01091242044ULL,0x54f1dcc9f168a366ULL,0x8020a64351203210ULL,0xdddf7d3c99cea3deULL,0xe757e116f9ce2a4eULL,0xeb5e7d57d9f9bb58ULL,0x6c3505ba81e5a962ULL,0x840a700c8050048ULL,0xd6a70d63d9eaa23aULL,0x8a20720a1040b394ULL,0x405002042004220eULL,0x100205001080220ULL,0x0,0xb5160c519948230eULL,0x8020100141008140ULL,0x7fff7fd3d9d3bb76ULL,0xad3cffe3d961bb58ULL,0x98b45378d1cbab22ULL,0x110054200602204ULL,0xd1206a229101a8c4ULL,0xcdd35eba91cb0372ULL,0xff1edffbf9ebaf7eULL,0xf034d06891512672ULL,0x800240210001010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x41000040002000ULL,0x0,0x400002000082000ULL,0x0,0x0,0x0,0x0,0x0,0x804010001000ULL,0x0,0x0,0x1000020000002002ULL,0x4400001000008ULL,0x0,0x0,0x20400008100ULL,0x0,0x0,0x3050d67dd1212b66ULL,0xa069777851672332ULL,0xfdb63f76d179abf6ULL,0xe94ff3baf139baf6ULL,0x2bfefe5ed865ab5cULL,0x5c551120c8412062ULL,0xdd6def7a91f7ab7eULL,0xca9eaffbd9ffab5aULL,0xeb55d494d9c58a5aULL,0x25fcd3c9d1718acaULL,0xadb35dfad9efab7eULL,0xf9fdf66fd9ecbe72ULL,0xb090257d9943bc26ULL },
	{ 0xdbf7fd7ecbbe7d67ULL,0xffffffffff7ef767ULL,0xbffbfffb7beef563ULL,0xf6fedf735bec7762ULL,0x90bf5eab09c67562ULL,0xbab84d620d2c7566ULL,0xddf67b63ebe67764ULL,0xd7dffe5bcb7cf560ULL,0x820e494000202520ULL,0x10020a8000000522ULL,0x959605f80a206500ULL,0xc084270000002400ULL,0x1010b1000080420ULL,0xe2400b4800080540ULL,0x8090060000002500ULL,0x41eca06800246500ULL,0xf290af5020206520ULL,0x8000100000000120ULL,0x80000100200400ULL,0x1001041a10200400ULL,0x0,0x5080055000086060ULL,0x100800000000400ULL,0xa100100a80400ULL,0x888c8000002400ULL,0x10506000206500ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8380240108646002ULL,0xcfdbfe36ef4aff63ULL,0xe7f7fecafdaef763ULL,0xafbb6e3a0d4a7563ULL,0xd5aeb34c0d327463ULL,0x66bc3e708bc86560ULL,0xcbe4f36cca646523ULL,0x6886f9c08b425720ULL,0xffdfdffb9bfce767ULL,0x2c5ff6651d786663ULL,0xfbf67ff3edfe7d26ULL,0xfbdceb7ecf767746ULL,0x14000000a000ULL,0x67bfdf76df7ee567ULL,0x6000000120ULL,0xc2858d6051026725ULL,0x0,0x0,0x0,0x0,0x6240101024840ULL,0xbbf6fe197d4c7762ULL,0x440bc10802884700ULL,0xe9e4b1c39487d63ULL,0x3cfa57f6cffc7f61ULL,0x99cf6d02b1487d21ULL,0xffb7277f9fee7f63ULL,0xc976f77b2ffe6627ULL,0x6cda9a5b88606726ULL,0xbb2453a596e7445ULL,0xd0ce370a0a687721ULL,0xef767772abc47d63ULL,0x6987fb2302ac7760ULL,0x42c24e5a4b3a6f42ULL,0x3c97ffdb1b7e4724ULL,0xa7efb2fda8ec6722ULL,0xb8d67de0d5f86f42ULL,0xe3d9d73c89226762ULL,0xc08a0e110b620560ULL,0x2b88cc6cb486360ULL,0x801177e008647720ULL,0x9ada7c4208647764ULL,0xdbd5fb2ed9667f64ULL,0xc5c3fe3adff86741ULL,0xabd697d24ae86562ULL,0x1c1518292b42f120ULL,0x89eca730afec7521ULL,0xffaa7fffa97ef762ULL,0x944f72a9aace503ULL,0xebdcd55d3bf8e763ULL,0xffff7fefdffeff67ULL,0xfffffffedffeff67ULL,0xfefffffffffeff67ULL,0xfffffffffffef767ULL,0x3fdfffddcffe7763ULL,0xbebaffebe9ecf567ULL,0xf37efffb5dec7763ULL,0xfbf67fefafee7f63ULL,0xcbceffa32bfc6d67ULL,0xfddffe7b0fee7723ULL,0x959785eacfea6d63ULL,0xfd56f2f9e76a7762ULL,0x49a6744208007522ULL,0x8a9422360234a500ULL,0xcc018d8802c6001ULL,0x279b04128a2ce520ULL,0x90f4f6fc592e6d25ULL,0x2002857200283460ULL,0x7ea0ab780bb64760ULL,0x6816320209267500ULL,0x234a51d0cf086121ULL,0x1425526b5938a720ULL,0x100a4320a404500ULL,0x99046224484c6120ULL,0xc6c6a33b6a6c6560ULL,0x21a105088006000ULL,0x8280043a02084443ULL,0x0,0x8001800008a020ULL,0x383080a040540ULL,0x24100c000200a600ULL,0x6002000400ULL,0x71a0026a4a202500ULL,0x5a8e380003882602ULL,0x8884130a0a244500ULL,0xc60a0e118b426506ULL,0x80a004e040006701ULL,0xcc44030802004544ULL,0x8146001082160ULL,0x8b82be661f707562ULL,0x8882720918084503ULL,0x1000490008044002ULL,0x4080049801085002ULL,0x4a00e3018a602400ULL,0x9e84277c0ae46503ULL,0xcfcb86110ff46760ULL,0x2f3ec4c088647560ULL,0x462108408402400ULL,0xd93b3e7015347e42ULL,0x4aa832420a246500ULL,0x8c2837a02386722ULL,0xa3c7a0312ac4646ULL,0xd014a380a82742ULL,0x49d34201426062ULL,0x4206200002282500ULL,0x80900001001120ULL,0x14000300000ULL,0x8002000000240000ULL,0x8050180021084820ULL,0x4400000000000400ULL,0x1800000000400ULL,0x0,0x401000000400ULL,0x0,0x804000000000400ULL,0x200400a00080500ULL,0x0,0x0,0x0,0x80304000007001ULL,0x110000000000400ULL,0x800020000000400ULL,0x0,0x400002000000100ULL,0x0,0x0,0x10200000400ULL,0x4400000000400ULL,0x0,0x0,0x246802300500ULL,0x0,0x0,0x10200000080100ULL,0x0,0x10408200000a400ULL,0x40200001020000ULL,0x84c2120202000500ULL,0x12004e0000000540ULL,0x11b2c24408a00560ULL,0x0,0x0,0x0,0x1000018100000740ULL,0x80040000005000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400000408000002ULL,0x0,0x0,0x0,0x100080000200400ULL,0x804000404000ULL,0x0,0x0,0x0,0x0,0x5491f66cf9ee6762ULL,0x71afee730f7e6762ULL,0xffba6fd74f6e7f61ULL,0xcccffff6dffaed63ULL,0xeb7ef7578f36e767ULL,0x3c331fd3313a4765ULL,0xfd6ef5aeefc8e522ULL,0x5a8aaf661b7af760ULL,0x87338b2113667500ULL,0x82c792a36a48a524ULL,0x412777273b83524ULL,0xdcbcb6fbbfac6762ULL,0xf22a7de41bfe7502ULL },
	{ 0xfbeeffff9fff7f69ULL,0xfff6ff778bbe7f71ULL,0xfbdfff5f9bbeef68ULL,0x7ebefff20fb73d70ULL,0xbcd936e68e326d70ULL,0x90e3bfd08e9e2d68ULL,0xd7b4777f0bf2cf78ULL,0x21d7c32a123e6d61ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80400002000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc5dfff388e9e0f60ULL,0x2fcebec996bead61ULL,0x948a5e1b0a18a560ULL,0x90ef2edf17ff8f68ULL,0xd5a279728e264d61ULL,0x8b54e5b612af6840ULL,0x282336210882c20ULL,0x1f934ffa9edecd60ULL,0xb604823004900d68ULL,0xfbfebdf89eba6f69ULL,0xdbf7fbbe8f9e7d68ULL,0x946080003100508ULL,0x879ece5482a7ef68ULL,0x0,0x962c45401aee1520ULL,0x0,0x0,0x0,0x0,0x9eaf5c011920559ULL,0xeee4af6f0ebe2d70ULL,0x10200020100ULL,0xefbfffdf9ffeff79ULL,0xb4b17ff99e942f60ULL,0xbdcdbfc00b2ebd60ULL,0x9eb62d7f93ff2f60ULL,0xced27f1096921d60ULL,0x4b5edfc116b24d60ULL,0x963a7fda8e937f61ULL,0xd9fef7ff9fdebd78ULL,0xfbbebffd0fbf6f69ULL,0x89d6e32a9ab44d28ULL,0x2fdfbe53869e3d60ULL,0x7bbe0d1b1e8a6d60ULL,0xb1f4d75212ce4d61ULL,0xcec475c30eda2d60ULL,0xfbedf72a133e6d28ULL,0x94c9bb4b9e144d60ULL,0xaba8871206c63b20ULL,0x75566907364c60ULL,0x910c0520038e2660ULL,0xdb93eb0d96b6ef68ULL,0xdecffe939ffebd78ULL,0xaffed9d917b67d41ULL,0xd4f356a80f904568ULL,0xf9f8cf6016b66d60ULL,0xdf326e5517f2ed68ULL,0xcb67fb3202fafd60ULL,0xef7edcdf9efe6f68ULL,0xb7367ceb169e6f79ULL,0xdfffffef8fbeef61ULL,0xffffff7f9fffff79ULL,0xffffff7f9fffff71ULL,0xbfffbfdf8ebecd70ULL,0xb7bfef738f9e3f61ULL,0xd1fd77fa8ebe7d68ULL,0xfffc7d7c8ffe9d61ULL,0x8bf7fbfe0f1f4d79ULL,0x6fffba5b9fffef69ULL,0x7fdefff39fb27d50ULL,0x716eb4780e292d60ULL,0xeffeff739ffeff78ULL,0x8010202208120440ULL,0x4dcaba5186002d41ULL,0x9db01d528e9a1d60ULL,0x7090f4dc16b60d60ULL,0x9380020007008540ULL,0xc99ce63f0b2a4d68ULL,0xa28ca20b8e12c160ULL,0x204004008040420ULL,0x50b3065216902140ULL,0x1b8c006a97a34960ULL,0xd38021798f960500ULL,0x90c3ae210a950f20ULL,0x295ac2518a848f60ULL,0x742a1fe30aea6640ULL,0x8041040000260400ULL,0x240001000400ULL,0x8088c00010820c10ULL,0x0,0x401040804100820ULL,0x4020114a00270c40ULL,0x3a84304882840400ULL,0x8a20200a02900500ULL,0x0,0x2200004802080420ULL,0x8001014003860400ULL,0x4ab06c5007b04540ULL,0xab2443221bb22540ULL,0x8800a0002000440ULL,0x84e8c23302bd2860ULL,0x10545ee096160d60ULL,0x10200c2010960000ULL,0x7000292a04361560ULL,0xc1c29a3c040e1960ULL,0x10808d4301ca4960ULL,0x8000046000020060ULL,0x858a215003a80740ULL,0x4300412610840400ULL,0x8848e10c84220940ULL,0x614400010920420ULL,0x346017e904902960ULL,0x8132084001820620ULL,0x98062f80122c0520ULL,0xc51ad3390f863740ULL,0x105002000a002400ULL,0x1010400200020940ULL,0x8074020000122d60ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2010000000020040ULL,0x0,0x80100000040020ULL,0x0,0x0,0x0,0x0,0x2020000060000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4400000000120000ULL,0x0,0x0,0x80010000002800ULL,0x100100902002400ULL,0x0,0x0,0x108000000400ULL,0x0,0x0,0x0,0x0,0x400040a000000ULL,0x800080001000400ULL,0x0,0x0,0x1200000000000900ULL,0x80080000c00000ULL,0x0,0x401000000000440ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2000002000200400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002002002000ULL,0x0,0x0,0x0,0x0,0x90b83cea9abe2d60ULL,0x15e6276b168e5d60ULL,0xfe983f6e96debd61ULL,0xcffeff0e9fbe6f71ULL,0xa9efddd30edf7d60ULL,0x149ad56a87b40d60ULL,0xd87f2d7f9ef2cd60ULL,0x9b970f7481be4d60ULL,0xe4a3f24d1e927568ULL,0x22b5ca4606162c28ULL,0x1530d67292862f60ULL,0xc5ad65fa0ae1ad38ULL,0xcab6757d1e934d68ULL },
	{ 0xfb8ef7f3dfbc3ff6ULL,0xcffcb7fff6847d65ULL,0xaf7fbefb560e3f0fULL,0xdedefe7cf68c3f6dULL,0x54d63e80568e3c9cULL,0xbff03f30d68c3ee5ULL,0x7e8e6f6ad69c67c6ULL,0x182250ac28c2482ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x420000002000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe6c7a212fe0c1600ULL,0xa35f1fcaf2167586ULL,0xb22e436256842a6eULL,0xd06134d556842782ULL,0xdf164734d6b42bf0ULL,0x8352ff78d70c1e42ULL,0xb6cd1b52568c03e4ULL,0x8eaa6ff4d38e370cULL,0xf5fffffaff9e6e8fULL,0x7ffe7ff2df9e77eeULL,0xfb84ff3ff6be7de7ULL,0xa0631086d4147c2cULL,0x2198ce80100c1468ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xdaaef5f0d69e475cULL,0xda9d2c3d5f9c2ec1ULL,0x2144c402d4300b40ULL,0x77567dcbf3141536ULL,0xdc171fe1df963784ULL,0x892c5e6bf6806146ULL,0xffb63ddfffbe3ce2ULL,0xc7d6b778f29c2762ULL,0xaaf3ca41c6ac2546ULL,0xb4bbcd76f6be3ce6ULL,0x55b4b7eaf7be3fc0ULL,0xa2123c7cd6bc3542ULL,0x80805202a2020502ULL,0xebe5daf0da2c2ff0ULL,0xbffedffbd6a637f6ULL,0x5424406ad69e20e0ULL,0x457665b0d31c6502ULL,0xf38ceb32d3b07d42ULL,0xdddfbebbf7b477c4ULL,0xef3fcfd3d79c36efULL,0x841016c052802d44ULL,0xd7ba1f50d3146f85ULL,0xf9b7f7fef79e7feeULL,0xefdef69fd79e77ffULL,0x1a5fddcbf69463c7ULL,0x10563c6ac684014cULL,0x913fff7af7b40b4cULL,0xd3280a2ab2946061ULL,0xc122e52a568401e8ULL,0xaa935edbf3382fc6ULL,0xacbb5feadf3e3ce4ULL,0xff5dfef0d7943db5ULL,0xdff3bf7ff79e7feeULL,0xfdffffffffbe7ff7ULL,0xe1ffdfdff7be3fedULL,0x9fbffffef79c7feeULL,0xf7fdf7faffbe7ff6ULL,0xfeb6f5fef79e3fecULL,0xc9f7ebbed79c3f66ULL,0xedeebe4ed21e77eeULL,0x99bbcf93d79e3595ULL,0xf5e7b67ad61677edULL,0x7bee2ff2ff9e4f54ULL,0xd910ad38b3a00b44ULL,0x800b9820120c2105ULL,0x1cb945ca92943885ULL,0x1080748850000116ULL,0x4536ad71d19c3de4ULL,0x600ce4a8c2182d01ULL,0xa15e8a09d6002d04ULL,0xb5c480292002b06ULL,0x414080005400ULL,0x8417041b1002041ULL,0x204846002000ULL,0x4588d10cc50c404cULL,0x621ad810d69420a2ULL,0x44a040042080ULL,0xd001200410001460ULL,0x480005040000504ULL,0xc107f90e52022d06ULL,0xe8020fc272082505ULL,0x6604640d4902004ULL,0x76854040080ULL,0xb404ec6086102948ULL,0x85806487041464ULL,0x486080092000b00ULL,0x16101396440c1402ULL,0x9194c002c4120402ULL,0x6212fd22fe182f48ULL,0x8b246b3a56822a64ULL,0x40260042000110ULL,0xc44000090100004ULL,0x0,0x22145f20568003c0ULL,0x4844a32050802a84ULL,0x480216a974842840ULL,0xba705fc8f69c2504ULL,0x20ac5068540c2a21ULL,0xe9028a40d11c0570ULL,0x80824450802006ULL,0x84828200049c2420ULL,0xab4f5f52df9e2d06ULL,0x75a4056a92846b26ULL,0x4042a540b680024cULL,0x4220348ec2045410ULL,0x4c0c600e2000490ULL,0x6000000040080ULL,0x820a0601c6024100ULL,0x0,0x0,0x800020000080400ULL,0x0,0x8080000002800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100244000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2040002000004ULL,0x0,0x0,0x0,0x100800044000000ULL,0x840000000001400ULL,0x84020410800840ULL,0x0,0x200001010000100ULL,0x2008010000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500000030000000ULL,0x0,0x0,0x0,0x8104100000ULL,0x0,0x0,0x0,0x800010002000004ULL,0x0,0x44040208040480ULL,0x0,0x2020000800040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800044000000ULL,0x0,0x108044000000ULL,0x80120442d59c0460ULL,0x14543ce9d6162ba6ULL,0x9b2e2ff3d69c25e0ULL,0xbfa5f7ffd79e7f44ULL,0xdffdfb9ed78e6beeULL,0x6ebd1e02d02e1d46ULL,0xe4e15d6dd71c5704ULL,0xc5248460e6140404ULL,0x9ab8a70592bc76c1ULL,0xe7cef732fe0c3daeULL,0x6fdefd5fdf987fecULL,0x920acfc71f143e61ULL,0x80522650901c0400ULL,0xd2322b74f20e2d8cULL },
	{ 0xffdefdf1e9fa73fbULL,0xdbd7bf6681b82b7aULL,0xfffb7fdfebba7fdbULL,0xbffedffd83a777fbULL,0x34d46ee08bcc4377ULL,0xdbdfe5f08be077fbULL,0x8300e584e0102282ULL,0x44000080000010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc4d0f21009c46202ULL,0x6d0f2e5bc39063e2ULL,0x129aa5eb8b907359ULL,0xf77bf67189f673fbULL,0xd28e7d43ab7c2383ULL,0x8193ee038bd03748ULL,0xaff71f5983b262efULL,0x8fce4c08a1902163ULL,0x51a6a0f20bcb236aULL,0xdb6e6e7989f87779ULL,0xf3f4ef670bfe62e7ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4004000000001200ULL,0xfb6fff718bdf62e2ULL,0x10001002000000ULL,0xc8c20c0003036002ULL,0x17c5f00e9286220ULL,0x179e17ebcb8062a8ULL,0xf11cf7528bd06778ULL,0xfffe2f7fabd96efbULL,0xf5e4ee078b956bbbULL,0xfbffffdfcbbd6b7aULL,0x8fba5fdaabd16bffULL,0x8514b46a09186323ULL,0x5c2efdec8bb04bfbULL,0xcae0f303a9c8271bULL,0xbbdfdf48abda62f1ULL,0xfa9fa3cf893067feULL,0x9520f24183b06203ULL,0x528438c089c423b2ULL,0xf9e4d31e8bb163baULL,0xc3ca9efa89b26babULL,0xc8140c1200b03210ULL,0xf276efabb063a7ULL,0xb2162c7049846eebULL,0x9bb4ff7f8bd477fbULL,0xcfffffbfebfc77ffULL,0x6bff9fdbabb26afaULL,0xdcb7527e8bbb620bULL,0x91ffb761e93c6e9bULL,0x8bb41e7d8bbc3211ULL,0xee7eef39abe563ffULL,0x6ffb7fcf0bb967bfULL,0x863efaebcbb56332ULL,0xff7ff77b8bfc73ffULL,0xf2b02f5f8bf06beeULL,0xedd7f7a98bf17777ULL,0xbfd7fedbabbc6373ULL,0x1e9a6feec1307be8ULL,0xdf95ffea8bfe63dfULL,0x7fae7fefabf866b3ULL,0xc9f3eb79ab94669bULL,0x4fd73e16ab310a02ULL,0xd79a97b58bfc2772ULL,0xfdf7f7faabca77fbULL,0xdda06d60c1b47372ULL,0x91d7cb0409344200ULL,0x80c39bbfcb186602ULL,0x22e4850a3100242ULL,0x6444b4708b802403ULL,0x49023401a1902002ULL,0xf914ff0488d86299ULL,0x47d72a12c94065f0ULL,0x3d2c08800a03701ULL,0x70700c728b82220bULL,0x824450088604202ULL,0x9000220303800028ULL,0x8090222280803209ULL,0x984801a2804010ULL,0x200f201220204ULL,0xd0e034000ba00330ULL,0x8802e148080210aULL,0x8780a01908f82201ULL,0x864c00040002301ULL,0x16984128012060a3ULL,0x8454140089806020ULL,0x92240d0c81102351ULL,0x4020000002001ULL,0x21c0082002011ULL,0x288086e8a8d02b02ULL,0x9424702803186261ULL,0x4202259001884220ULL,0x8990822d83b02212ULL,0x108000100200ULL,0x3846240b800218ULL,0x749d18280a806331ULL,0x9200000000006001ULL,0x8b2234070b8c2271ULL,0x44d22e2101940b23ULL,0x35f5d0943982a21ULL,0x804000800001ULL,0x2002032000a03220ULL,0xc20046008002261ULL,0x4e0224440bd20260ULL,0xa943180a8191421eULL,0x9080552a8b004122ULL,0x5040228888882202ULL,0x8e102d1ce9c06388ULL,0x41809010000043a0ULL,0x2a1ab0484b902201ULL,0x1491492888506280ULL,0x10c0110000802121ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x800000808400000ULL,0x0,0x0,0x0,0x0,0x0,0x8100000100200ULL,0x201000002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x204000100200ULL,0x0,0x0,0x140022000800218ULL,0x804001000020ULL,0x2000002000080200ULL,0x2010000080002000ULL,0x0,0x0,0x0,0x8400800001ULL,0x0,0x80802101040220ULL,0x2020000000208ULL,0x0,0x0,0x240000200200ULL,0x0,0x0,0x804080002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x140000010002ULL,0x1000010000000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x910a4b918b884b32ULL,0xd7357ff0abc663e3ULL,0x51e71f694bd06b93ULL,0xdfb3772e8b997373ULL,0xedd7ffababda7b33ULL,0xab7fffdacbba77fdULL,0x43356a308b052e6ULL,0xcbcda56a8b0d6b2bULL,0x5604084402106320ULL,0xcec3d82689996b13ULL,0xef5e1fce8bb463e9ULL,0x15105802418060f8ULL,0x354d52108980422eULL,0xc682032482802618ULL },
	{ 0x4be77f68fb6dcf6eULL,0xfbdeffeffbfcd77eULL,0xe4d8fbfae174d73aULL,0xbe3effffe3f5d7feULL,0x100102c201044128ULL,0x90100d408284912aULL,0x0,0x0,0x0,0x1000008000000108ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44800010d0000004ULL,0xb5df3efbf3e4db26ULL,0x82806f4a41c4832aULL,0x807abf50c054592eULL,0xc1c26f2463c4c20aULL,0x8b84f2f4f3f4d36aULL,0xc8065e864844c11aULL,0xc69c75c240d4c33aULL,0xfddfbe7f7bd5c3baULL,0x44ac7d785145c35aULL,0xfbefcf3ff1e5d77eULL,0x0,0x0,0x0,0x0,0x0,0x0,0x400400040000000ULL,0xd444f27141c5c10aULL,0x712e0a6a83d5d16aULL,0x0,0xca43a644c854c11aULL,0xa3d5dbc9e174d132ULL,0xb43f56ffeb04cfa6ULL,0x90b8e676d1e493beULL,0xfb962da5c3e4c32eULL,0xcfc85631e875910aULL,0xefdfffdbfbd4db3aULL,0x96ae47f6ab45d93aULL,0xd567f7f6e3c4d1aaULL,0xfbbeffecfb7ddbfeULL,0x8bf0aa33fb55c91aULL,0x68cf1fe76984c32aULL,0xbf8fcffbeac5d93eULL,0x88c5d67a83a44336ULL,0x4cd62678c01dc31aULL,0xfb40fd6afbf5d32aULL,0xc4c08030c0c19000ULL,0xa9c884258540202ULL,0xd84025e5d354412aULL,0x933ee376e3c5c36eULL,0xd9bcef7fe3f4c33eULL,0xcfd7bf3fd9c5d33eULL,0xbfdfffd9fb55db3eULL,0x3e7476c4dbd5d3aaULL,0xc57eb77acb78531aULL,0xfb963fffe1f4d33eULL,0xedf6ff3edbfdd1feULL,0x8bdf175bdbe5d3baULL,0xbff9fff9db6cc7faULL,0xd3d8afdaf975d17eULL,0xfbbe3fc7ebc5d9faULL,0xe9d79b1ffbf4d3aaULL,0xefffffdafbfddf2eULL,0xf7ff6ffef3fdd73eULL,0xd771ae7ee365d33aULL,0xd3fe7774f324d13eULL,0xab567fa3dbf0d11eULL,0xa206bac3e3d4da7aULL,0xbbbbdf62e9e5c96eULL,0xf5b5f279e3f4d726ULL,0xeffe75e2f975d252ULL,0xa3d6e3377b4dd30aULL,0xe84200b80240452aULL,0xae9fce67d8f4db3aULL,0x414560e050d806ULL,0xc3eced31ebc5c19eULL,0xc0b5f33183544d36ULL,0xfdefda39f3dddf3eULL,0x218441a210c4202ULL,0x94905274490dc202ULL,0x4949076289044816ULL,0xc80639386105811aULL,0xc8d0a38a6104c088ULL,0x940825bc0f8d108ULL,0x2808010000c00aULL,0xc020a5108105d00aULL,0x2a0045ae104c10eULL,0xc004b21e8b005322ULL,0x811129461041312ULL,0x140868000150c020ULL,0x0,0x240001000080ULL,0x12000000120ULL,0x120040000000ULL,0x18014c4220504042ULL,0x801000000108ULL,0x8c2a2063008080ULL,0x82857002410441baULL,0xe0618e810144837aULL,0x8d9121705984c11aULL,0x100001c000048030ULL,0x840007c04344010aULL,0xc80420028101d000ULL,0x8ce9580003044428ULL,0x30a0d404250c10aULL,0x23116915801c102ULL,0x4c8020468060012aULL,0x40223e480b44c42aULL,0x45c7a7aaebf4d11aULL,0x4184d508070811aULL,0x22821c1005000ULL,0xc13a2468e905d92aULL,0xdaae2d784bccc10aULL,0x4d5e6210ea84d12aULL,0x8564bd08844911aULL,0xa77c4632d961c17aULL,0x90d138356844052aULL,0x0,0x0,0x0,0x0,0x400020004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2004000001000080ULL,0x0,0x0,0x840000000000102ULL,0x0,0x0,0x104002880100102ULL,0x0,0x0,0x0,0x0,0x0,0x4002000000c00000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x20840008000ULL,0x100000240008000ULL,0x0,0x0,0x200000200004008ULL,0x0,0x40201000080ULL,0x1a0100061000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100020000040200ULL,0x0,0x0,0x0,0x10001000040002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20080001000200ULL,0x0,0x0,0x0,0x100800080001000ULL,0x902d9d2e1c48b08ULL,0xf4d434abf3d5d5beULL,0x971a9768c3085d2aULL,0xff74373ee3e5d33aULL,0x6dce720fc1f5d18eULL,0xaa7effdaeb74db3aULL,0xb47e7febdb7dd77aULL,0xf9567e63f30dd1beULL,0xfebf7f75fbedd77eULL,0x814e70388b204b02ULL,0x43dd9ecbd154d33eULL,0x16b66ee2d0e45d3eULL,0x711cdd7bc2f1d52aULL,0xeb065d644b05d23aULL },
	{ 0xeb90b746c040aff0ULL,0xdbf5e72ec0d7ba64ULL,0xdcbfffd7c2dff3e6ULL,0x75b4e4ed88c7b320ULL,0x30904bc28082e220ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb6baff2e4adef3b2ULL,0x206d12680a12b260ULL,0x4b062d758242f3e0ULL,0x69a7b236cac3bb90ULL,0xaeff5ee9cadeb336ULL,0xfabfefafcad7bfe6ULL,0xdfffb67fcacfff26ULL,0x8000050040008220ULL,0x0,0x0,0x0,0x140420040a408200ULL,0x0,0x0,0x0,0x0,0x50957ee1c097ff62ULL,0x112c006a0045f102ULL,0x0,0xcdcec52acac3af64ULL,0x2bf6ce978296f3e6ULL,0xad977fe7c8daf626ULL,0xfb69bfe2ca9af340ULL,0xf7963fe4c04bfb62ULL,0xcbcfd77bcac3fbf4ULL,0x29dafbcf4ac7bb76ULL,0xd7be5ffbca43b7c4ULL,0xf7edf759c2d7bbe2ULL,0xff4e7d79cacfff66ULL,0xeb8673e64ad3fe32ULL,0xefefbf7bcadefae6ULL,0x3b97cd6a48c7af64ULL,0x2041002000408030ULL,0x4fe275db4ad2f2e4ULL,0xd98f7becc2cbbf40ULL,0x81021e99c0d4b8a0ULL,0x71cdf5242c3ab40ULL,0xf0009460c8020e24ULL,0x800aec60484297a0ULL,0xfab1632ec2c2b2f0ULL,0xc9deb2abca42faa4ULL,0x8a4f47cf8287b7e6ULL,0xfffffe6bcadbffe2ULL,0xf9eeb7f8cadbf6e6ULL,0xfe9e7f7dcadbffc4ULL,0xfdd7ffeacacefb36ULL,0xafddffdbc2c7ffe6ULL,0xf4ff7fedc287fb64ULL,0xdffe7e58cad7bff2ULL,0xde94bff7cacbf776ULL,0xcdfdf66f8253fee0ULL,0x215c7f51c847ab36ULL,0x9bbb5f7acac7f7a2ULL,0x95fdffffc8d3f7d2ULL,0xafe6fff0ca5af3b6ULL,0xc889fa23cacafe60ULL,0xefcf3fdbcadbfef4ULL,0xc3ba4f52cad6b236ULL,0x423452680a12b280ULL,0xc7bc3df2c2dffb66ULL,0xd976aa7ec2cdff36ULL,0x8ad33ec3c0c79b72ULL,0x41215168085a022ULL,0x905c54100a40a280ULL,0xe2d648842a3c0ULL,0xfe80cf688ac2fa24ULL,0x88c6ff6f8adef322ULL,0x1bd0db508a92bae2ULL,0x5c9156e68842ee20ULL,0xe220060040868202ULL,0x88945bf840439fb0ULL,0x38cf202880273a0ULL,0x804000880402800ULL,0x4806432c8803220ULL,0x402482c00ac0b002ULL,0x1210213340818620ULL,0x460041b800438202ULL,0x60c22a298a5d5202ULL,0x9405042040c00a00ULL,0x8088240240410220ULL,0x2000040000008220ULL,0x4000800002000200ULL,0x8000000280000004ULL,0x208d01048c00004ULL,0x420400440020204ULL,0x108082000481800ULL,0xb06430c4042a236ULL,0x2002aa00c2005600ULL,0x1004000101000ULL,0xd0501ec088425a10ULL,0x8148fc704a03a222ULL,0x212840c00820ULL,0x4c4f3292ca93f764ULL,0x2b64640a48099504ULL,0x3480142e804ad940ULL,0x14a07408a10b220ULL,0xf685b6404ac0aa14ULL,0x81cc284802a232ULL,0x48009048000220ULL,0x20e24a24002a602ULL,0x503028c040c08010ULL,0xdfb66fffcac3f7e6ULL,0x8022280302485210ULL,0x800080000020040ULL,0x303a0ca34206a0a2ULL,0x53be2dad8216fb22ULL,0x0,0x0,0x0,0x2080000020020ULL,0x20020002000200ULL,0x0,0x0,0x0,0x4040080001000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8400000020100ULL,0x60000008200ULL,0x0,0x0,0x0,0x0,0x806800430040ULL,0x20040000402000ULL,0x0,0x2000008000008200ULL,0x0,0x1400000008400000ULL,0x0,0x0,0x0,0x0,0x0,0x2010002000200ULL,0x0,0x0,0x0,0x80040000002080ULL,0x0,0x4100000081000ULL,0x0,0x0,0x0,0x0,0x0,0x120000020020ULL,0x0,0x0,0x0,0x800400000081000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2200000008100ULL,0x120000000808000ULL,0x0,0x0,0x4002000008100ULL,0x0,0x9ebd4fd1cad3db64ULL,0xfef6fcf7cacfff76ULL,0xb9670d61c893b362ULL,0x8bba77fdca1af6a4ULL,0xbaabfe12cacbf7d4ULL,0x8bdefd5e8ad3f7a6ULL,0xd1b4b33c242fb24ULL,0xb9de75eac2c2fbb0ULL,0xfaff3ff6cadbfff4ULL,0xc94293284054bab2ULL,0xaef9ab1ca5afa04ULL,0x91324c738240abe4ULL,0xd5b13e58c24bfb66ULL,0x53a007c142c3a220ULL },
	{ 0xcbde6f674fd3fba1ULL,0xe3f6d273cccb3a0bULL,0xced3feda8ecab3bbULL,0xeeb9fffa4da1bf13ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x102c70628cca2280ULL,0xc2b61f7448823b13ULL,0xd9c0d16b8ec33a11ULL,0xaffabd5bced6bb01ULL,0x4dbeefdf8fdffa83ULL,0xf4ffabffeddb7fabULL,0x10c27608d023e11ULL,0x0,0x0,0x0,0x80040000000220ULL,0x0,0x0,0x0,0xa005308601201ULL,0xa0056828020392ULL,0x0,0x8000112000022280ULL,0xd8c3b6188e833a11ULL,0xea56de95ce963609ULL,0xfebb1fe7ced3bb89ULL,0xf124f760cfe19f81ULL,0xfabe2c708ef33383ULL,0xc9aac32d0ec3fb91ULL,0x297eecdbcec3ba89ULL,0xb6bb7feb4fdb7f9bULL,0xd47ebdff2cc3de9aULL,0xf6744ffc6cd2be93ULL,0xebf4f73bcdc09f81ULL,0x6e765f5f8ccfbb99ULL,0xff7fdf7eaffbfebbULL,0xd40574690e921899ULL,0x4aa24d618d92b910ULL,0xb1a4db868dea9a09ULL,0xa4213590c0b3b00ULL,0xb1047830e533a01ULL,0x9054bee30e92b891ULL,0xdebffff3afdf7fa9ULL,0xcd92e313adc33b12ULL,0xc58ad2280e8b8399ULL,0x881acf5b0eda3e99ULL,0xfffc5fe0eff3bb93ULL,0xd7b67f63cdd2ba89ULL,0xfe9e3f6c8e8bfb9bULL,0xcec3fe3b8dcbba99ULL,0x2acffd538cc3fe19ULL,0x56975d6aeec622b1ULL,0x97f9ff5feedafe91ULL,0xdebe3f6a0eebbb23ULL,0xcec0fff20df27f31ULL,0x6f5ef95b4ec3ef11ULL,0xdcab4dfb2efbfa99ULL,0x716e366b0f533aa9ULL,0xefb63c7a4443fa29ULL,0xe899fbb88783de01ULL,0x4bfa1e1f8f923881ULL,0x83ceb20c0a2313ULL,0x1444a6620e024a11ULL,0x970a2d418ea2b383ULL,0x3a76631e4ec31799ULL,0x4ce93e832c0033a1ULL,0x288c40986422201ULL,0x1091bec608632381ULL,0x5762b30c4ccba09ULL,0x2a50c06c08416b80ULL,0xcce6d2b50dcafa9bULL,0x200a18520a4002a1ULL,0xc818e06026c23281ULL,0xc14027080c503688ULL,0xbb4170088021e10ULL,0x86114a1009421a91ULL,0xb56fa528e433782ULL,0x4224a050e4b0281ULL,0x0,0x92256886812221ULL,0x4d81f40007820810ULL,0x9cc76f7985cffe19ULL,0x1000080040000001ULL,0xa10050088808080ULL,0x0,0x88400004809008ULL,0x204000008000001ULL,0xa00ce488d809880ULL,0x80320e2188c00b00ULL,0x4808641200521603ULL,0xc9f0635bcd83ee81ULL,0xc862ea200c92bb13ULL,0x2832012080020600ULL,0x4022804208001811ULL,0x2a340008109101ULL,0x401022020a000701ULL,0xd407a2200d01221ULL,0x800080008000010ULL,0xb010188204c00202ULL,0x98602c0009406801ULL,0xa130054786820001ULL,0x7c6ea68ab434e93ULL,0x190ca100e819819ULL,0x90542886820010ULL,0xd240a7180c949a09ULL,0xf5140c7a2e877a93ULL,0xc120820146103001ULL,0x8ccdc028493da09ULL,0x908805894e823081ULL,0xd124a762adc23e38ULL,0x0,0x0,0x0,0x0,0x40800004800ULL,0x0,0x802000800800ULL,0x80080000000600ULL,0x2000200000201ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8002000008000001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000000202ULL,0x0,0x0,0x1080002000001ULL,0x0,0x0,0x0,0x0,0x1800000000280ULL,0x0,0x0,0x0,0x0,0x0,0x800000800004800ULL,0x0,0x0,0x0,0x0,0x6000000020000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10020000000220ULL,0x10100004002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1000000480020000ULL,0x0,0x0,0x60080020000ULL,0x224e51daa5c61a09ULL,0x74d7ffed0cdabe89ULL,0x8071245281810e00ULL,0x8084004000001801ULL,0xc34afeb109833e32ULL,0x494cddda8fc3ff09ULL,0x14162b034e422219ULL,0x936bf772cfd33b9bULL,0xfeb7bf7d8dcbffabULL,0xcb41dfb56cd3baabULL,0xa9f6dcc38cc3fb10ULL,0x2cb267cc0dd23a09ULL,0xc02492190583a800ULL,0x142626600f82a811ULL },
	{ 0x8106f761a25a16ceULL,0xcb76cf362676174eULL,0x4af3ffebe37e17caULL,0xe38dcf7aea7e90fbULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4eb82c50206b15d3ULL,0xcb47f2673a5617ebULL,0xcdef5f5faa7597faULL,0xbabb6ff2a67f15deULL,0x700032e922521512ULL,0x54ae3de0227310c9ULL,0x802000400001ULL,0x242301322101444ULL,0x24024026000000ULL,0x0,0x0,0x0,0x0,0x8284010085010c0ULL,0x20004000600000ULL,0x0,0x281820200101482ULL,0x88c5faa20275144aULL,0xa4fc5c56226414daULL,0xa1bae76a677314deULL,0x556eaf40ee7313eeULL,0xdaab3df3767713daULL,0xcb20f77030529558ULL,0x6d3e3fc820239150ULL,0x1fbbcf79257713faULL,0x71efa77c315f17deULL,0xd0ae316165711459ULL,0xa1e6f387605211deULL,0xef5fff766b7b91feULL,0xb9ba8d58ae6a10c6ULL,0x15cea3e6207a12e2ULL,0xcbfe1d00226c00c2ULL,0x93165b34e27a80caULL,0xaec23c9a6e52154eULL,0x433b4f43227210daULL,0x4c0680020a1140ULL,0x118a25607270114aULL,0xd3822f36e26100cbULL,0xc6dbda30b65f15daULL,0x287e9f5a2a6b115aULL,0xf4b656e1a35416eaULL,0xff7efffab37f15daULL,0xffae6f7faf7b11ceULL,0xddfefe3a3e7f97caULL,0xffdf7fd4737e91fbULL,0x4c75c69a252114bULL,0xddfe7e53777417ffULL,0xdfbe7ffe6f5f95deULL,0xed86e72c7a7b15cbULL,0x29f79edbb27694ceULL,0xa4ba7eeee36214daULL,0x4100056be2521482ULL,0xffbe7fffef7f97ffULL,0xd9f7ff7b6f7f15cfULL,0xeff6bfcb377695caULL,0x687c568325610d0ULL,0x44033790a70000eULL,0x4a4a2460241210d2ULL,0x3a4213e267106caULL,0xc7cbb33fa37a15caULL,0xd4e4000450076ULL,0x79fdfffe657117daULL,0xabbee77326771548ULL,0xdb9d47e6235215eaULL,0xcec9363b387f17caULL,0x331addd3225a11caULL,0x34bd76e5337c154eULL,0xa9c45e68317611daULL,0xca002a4fe008114eULL,0x3909000200240ULL,0x106d14000700140ULL,0x102020200211440ULL,0xd020200200701048ULL,0x9a940f14007210b7ULL,0x4180801800671040ULL,0x87484002121042ULL,0x411000224001100ULL,0x91896400202211c8ULL,0x0,0x100020000000006ULL,0x52023223011048ULL,0x80004e1a62020000ULL,0xd2109c08a142000aULL,0xc70634202150145aULL,0x81e6f3222a52114aULL,0xcce29f15675294ceULL,0x20014060101000ULL,0x836520020521482ULL,0x80082020200000daULL,0xc08053042052000eULL,0x84222e0a0250111aULL,0x304001000521000ULL,0xa4320264207a1048ULL,0x46a41801481148ULL,0x2800a46001a105aULL,0x7018e28206010c9ULL,0x40875900a2621442ULL,0x10220c80a0521080ULL,0xa8c64020700048ULL,0xfeb66f7ee75615daULL,0xa402d20224501080ULL,0xe814f3c2622a1740ULL,0x2000e5b226a0148ULL,0xd544af53237184c9ULL,0x0,0x0,0x8200420000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20860000000ULL,0x4000400021000ULL,0x0,0x100200080040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000046022500000ULL,0x0,0x0,0x0,0x0,0x0,0x4200100040ULL,0x0,0x0,0x0,0x0,0x1000080000021000ULL,0x10001000000050ULL,0x240000021000ULL,0x600000000001100ULL,0x1000010000021000ULL,0x0,0x0,0x0,0x200200101000ULL,0x0,0x8100021000ULL,0x20040008400000ULL,0x0,0x0,0x0,0x0,0x9000000000001100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8000024000080048ULL,0x0,0x0,0x0,0x0,0x81000000000048ULL,0x9763f0b6e5994eaULL,0x62fad7d326729042ULL,0x34b95568677a11faULL,0xc120840020500009ULL,0xb8424106010110aULL,0x4800070e607210c0ULL,0xaa4231db632115caULL,0x14f657e3227b11d5ULL,0xf9ffbff66b7f155eULL,0xfff6af7c207e95dfULL,0xd994e73da05715d6ULL,0x265fce5a024d155eULL,0x44a6ef0275211c8ULL,0x9410880024211140ULL,0x0 },
	{ 0xc9b27fcd063ecf96ULL,0x9bfbdf7a9b3debbdULL,0xc5c3fbf3da3eede1ULL,0x20ca1f8002568682ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xf100011400308a82ULL,0xa9c8f0c21a2d484cULL,0xdfd77f3d1b3dcff7ULL,0x189feffe5b3e69dcULL,0x4556328a922f4950ULL,0xc8386660480c80bfULL,0x0,0xe14212594c176048ULL,0x144a50883de894ULL,0x0,0x0,0x0,0x0,0x3908c98189344ac0ULL,0x0,0x8000200004004000ULL,0xf29401841a0ca0c3ULL,0xcdc3fe8a4d356af4ULL,0xa855d052086cccd1ULL,0xf4f277ea1f3ce7c6ULL,0xf95caf72c93ceef4ULL,0x4e9f3f659f17cbddULL,0xe174f6888d3cececULL,0x816fece1a6e2d44ULL,0xf3220f035a384aa8ULL,0xd1789e421a25c8d0ULL,0xc80674000c314804ULL,0xebfffffe5f7feff7ULL,0x6fffbfffdf7feff7ULL,0x3dbcdff94f3ecefcULL,0x3022564c403c8092ULL,0x200680204020cULL,0x8b00f5aa1f3c4a58ULL,0xaeca935b4b3ceb15ULL,0xc4ab8d6b520dc59dULL,0x51f2dcba0e1ee5e2ULL,0x8c02256089246c80ULL,0xfe902f464e3d4f90ULL,0x894385038a648c42ULL,0xfbf8dd5bce7defaaULL,0xb4fffee5cb7ccef6ULL,0x93eef7790b3fcf86ULL,0xdf127b7c037e4f90ULL,0xfcdc770c973dcfcdULL,0xff7f5d5fdf3fcff8ULL,0x21ba430a0a56a18aULL,0xc964a4021715e302ULL,0xff96af74d73ecd94ULL,0xef60f6e35a7c66d8ULL,0x1d6fed2ca36ac82ULL,0xc925ff28a3d4a64ULL,0xd220a0d01f2d4a38ULL,0xf67e7f75433eaef4ULL,0xcad3e32b5e3fe4b4ULL,0xcff7ff7b8b7ecffdULL,0xbe3b4d52cb5cee80ULL,0xd592f22d5b6ee890ULL,0x5d6e7d72d33e0f84ULL,0xd8c0e72e423fcad3ULL,0xc6fbfeabdf3feeddULL,0xef31c5684b35eac2ULL,0x80be7eec4a7c81f9ULL,0xca8865430b3e8b49ULL,0x828463e0130c4840ULL,0x9020681083680e9ULL,0x1c7655460a36c8e9ULL,0xc41502fcca3c0b64ULL,0xd94ce6731e1d49ccULL,0x80101378045c4cd0ULL,0xa106db1a493c88c5ULL,0x91e1340130dcc9cULL,0x4d3849a04a34ae88ULL,0xc95287401214e590ULL,0x2284002040100ULL,0x811099100e04490cULL,0x7a1202000900c0c0ULL,0x100002634014c0d0ULL,0x100100100080ULL,0x0,0x800820240104800ULL,0x2001088008080088ULL,0x16095791c33f4ac0ULL,0x4c1ca8a8011c4ad0ULL,0x402c48e240348c81ULL,0x8bc496650a1c6f58ULL,0xe94233d19b2c0adcULL,0x101841011216000cULL,0x26140a0000c803ULL,0x60000014000ULL,0x200c00801044004ULL,0x4d2202041144600ULL,0x8c0900180230ULL,0x10034001104200ULL,0x81063180040cc041ULL,0x170028610b64cc80ULL,0x8c80ea0243144c2cULL,0x484458484a1c00d0ULL,0x502a401000a0ULL,0xc10ab4720f24c70bULL,0xba922a790b3dc0deULL,0x2004611412348181ULL,0x87643a580a3ccb81ULL,0x122b47424015c8c6ULL,0x164242a1a35c0b6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000080080ULL,0x0,0x0,0x60000002040000ULL,0x0,0x0,0x0,0x800008000440000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x11000040000800a0ULL,0x0,0x800000800040004ULL,0x0,0x0,0x801001004000ULL,0x0,0x0,0x40000808200000ULL,0x0,0x10800000000c0000ULL,0x820404340184840ULL,0x8000020004004000ULL,0x100008080ULL,0x0,0x0,0x0,0x30000040100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x60008002000ULL,0x0,0x0,0x0,0xc00000080040ULL,0x120012000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcc63e11ca7fcfdaULL,0xad9eccc29f7deed1ULL,0x86fb76600b3febdeULL,0x8246802040d4410ULL,0x91901356163dc216ULL,0x40038150083608c0ULL,0x375eddc18e34cbc2ULL,0x948b5de34e3d4a6cULL,0xb96f2f68d63eefe2ULL,0x3a963fb5cf3dea92ULL,0x8c5a63048b046948ULL,0x4055504c0838a2a1ULL,0xb633fff31f6dcfd8ULL,0x8bad1600e24c194ULL,0x0 },
	{ 0xc3aaec60503a0560ULL,0xeafaf7beda1a4ff0ULL,0x30134a8343088551ULL,0x1012222100188513ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000f0a1413a0c20ULL,0x4e02376c9a0a4d93ULL,0xb6badf7828ba8fffULL,0xdcdca7de8128c73aULL,0x404800000101aULL,0x0,0x1b6a13df61b28febULL,0x1a044c0850024c64ULL,0x20400000000a01ULL,0x0,0x0,0x2501012001804a2ULL,0x5f187bd2711a9dfbULL,0x4040de4089a2c501ULL,0xeffefdf3d9ba0fffULL,0xd1263d66c902dff9ULL,0xeed277baf9bac75fULL,0xffeff57f9889733ULL,0x1cff5debcb8adfbbULL,0x11aa26e820105ffaULL,0x8c8a3862f20a057aULL,0x4dd3faa24b184d32ULL,0xae4e3fd953ba5752ULL,0x482511e098a1552ULL,0xb1bca213490a5f70ULL,0x7a960d4cc11816bbULL,0xabfefbfd53ba5f7eULL,0xecdfffffb9aa9ffbULL,0x1f9dcffbf93adf73ULL,0x9798aa70932a07dfULL,0xb00040000084802ULL,0xcbf4e32ed91a4da6ULL,0xeff3be797a9acfbbULL,0x63ae9d121b0a1f33ULL,0x54d356e488088733ULL,0x3b03d50a118853bULL,0xff34776dcb9a8f73ULL,0x5edbbf9f68ba8fe7ULL,0xeffeffcfe9badffbULL,0xf5b7fee1d8ba8debULL,0xdb7f6d6af33adffbULL,0x9b986ef5500a157aULL,0x7fd7fe3a9a3a9f77ULL,0xe09f95def93a8ff2ULL,0x26be5ee971b81773ULL,0xd5e9ee5702188ffbULL,0xc6862f77d39a1d7eULL,0xcdd6f3d7db9acffbULL,0x4870da61220a8553ULL,0xae1247e1908a8f01ULL,0x61401619a2101550ULL,0x73d62d17f19a1733ULL,0x49d0f29140000fabULL,0x6fdfff70b99a1f77ULL,0xf3ba9dffdb3007fbULL,0xb57f67fcdbba9df2ULL,0x4fb62d704a1807f3ULL,0x9bb7732f3092c77bULL,0xdf4f7ec9783a4febULL,0x2b9ec7df3bb21ffeULL,0x9041daeac21a0f7bULL,0x8288e5b330ba07ebULL,0xc0a5f6e4313a8f43ULL,0x4f03f60400ba4e4bULL,0x8b975d1e112a4d4bULL,0x52233ee80a084b13ULL,0xe81003a882080751ULL,0x3ab06e7b430acd7aULL,0xcfc3ffb359bacf23ULL,0xa850929200029b5bULL,0x88a030000180702ULL,0x2000006200028810ULL,0x480024a080820700ULL,0xa004882080080520ULL,0x485790c30b824d12ULL,0x2020000020000100ULL,0x1143804a189a0d02ULL,0x9010001440020124ULL,0x88904024800a0436ULL,0x0,0x30084c1220108591ULL,0x8102a2c0032202f1ULL,0x85007d40c00a8572ULL,0xabe0f72e6812cfbbULL,0x2100000000201ULL,0x800205001100500ULL,0x9013000480001750ULL,0x0,0x48080722880085b0ULL,0xc658b0a240204f49ULL,0x108480a02080413ULL,0xc000168540080323ULL,0x40230901180402ULL,0x9a00130c13020cf2ULL,0xa15488000018040bULL,0x4a5805599008870bULL,0x80b644e0910845a3ULL,0xd9001602020a041aULL,0x5400004801200702ULL,0xa16404201ba20591ULL,0xe97cfaced39287fbULL,0x18220a8001000302ULL,0x50205004c0080528ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200000000aULL,0x0,0x8100001010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000801000010ULL,0x0,0x0,0x0,0x400010000000a0ULL,0x0,0x1002000000100ULL,0x0,0x0,0x8000000100020040ULL,0x0,0x0,0x9200240200028251ULL,0x0,0x0,0x0,0x44000000120ULL,0x0,0x20000200000081ULL,0x6000080000400ULL,0x0,0x0,0x801000000120ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21230026002054aULL,0x1217c1134b008772ULL,0x72bee6eae28a4ffbULL,0x9d77ef731a9a8ff3ULL,0xfb9e277db19a0f3eULL,0xcdcefe8ec81a8d6bULL,0x61855b588128c7ebULL,0xb0f25e6a99828747ULL,0xd440b40c6008850bULL,0xfeba3f7158ba9d7bULL,0x4fdff37e988a8fdbULL,0xaec4ec498310ddf0ULL,0xb4bb6eee9308df1aULL,0x8806002008010ULL,0x78b61bc8011a0d3aULL },
	{ 0xeff7fe713cb97eeeULL,0xcbc6d53e0c9a73e6ULL,0x8008880004020104ULL,0x4a22500c220264ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2957be0010a84a42ULL,0x4ff216551d3b5fa4ULL,0xb6fa2de035bae760ULL,0x1422a2c40c280240ULL,0x128490204086840ULL,0x2054c524849018a4ULL,0x5eef9e1335bbd7c6ULL,0x38190000aa0004ULL,0x0,0x0,0x0,0xcb47b7333c8b4ec6ULL,0xa34f59cb3cda78e4ULL,0x94767ee03cb94dc2ULL,0xbdd66f633cba5fe6ULL,0x430e2d2104194d80ULL,0x986864004a81280ULL,0x8111581211b066a2ULL,0x36f676f0343a6ee6ULL,0xfffeff77b5ffffeeULL,0x9aae3fe19cbb77e6ULL,0xe9d4f32c34bbdce6ULL,0xa97a7a0b14b84ae6ULL,0x588189934982d80ULL,0xc5380e00049a63e2ULL,0xdb8effe73c977ee6ULL,0xeb1af77b15bbefeeULL,0xecf7bf5fb5bbfdeeULL,0xbfbedf7bb4bbfeeaULL,0x1d5efec034026aeeULL,0x82a86c43353817e2ULL,0x43ccf01e153067eaULL,0x6ecbba3514ba7ee2ULL,0xdf56d7fbbcbaffeaULL,0x148c024000866044ULL,0x93a6273014133ae8ULL,0xee94654035f8fec2ULL,0xcdc7be2d35ff6fa4ULL,0xb877dfdf1dbbfbeeULL,0x76fcfeea9cbee5c6ULL,0xd9c1766e1493dbc6ULL,0xd3346fee35bbfbe2ULL,0xd4fedff53cfbffe6ULL,0xcb17df5a35fbfde6ULL,0xbaa45f2814c25fe6ULL,0xdbeda57b153370eeULL,0x98a6237000ba4aa4ULL,0xe1b282041482f262ULL,0xccd85114a85a40ULL,0x428887a51caac046ULL,0x9240164604127060ULL,0x97a62d451db86fe2ULL,0x1ce7392913b64e6ULL,0x66953bd0941b7a86ULL,0xb6bfcf8c1df36ae6ULL,0xf634c66e3cbbeae6ULL,0x5ab2b65014d337e8ULL,0xb9464e2624f007e4ULL,0xafdfbbd134b3f2e6ULL,0x9f9acf723cb36ee6ULL,0x71264e7c34d3f4c6ULL,0xf2beadf034bb7beeULL,0xc3207f39349850e6ULL,0xdd5adefe8cb370e6ULL,0xad5cd18b1495dc62ULL,0x94b474a8053137e2ULL,0xaf64566a9c196fe6ULL,0xd2a2037e10abeee0ULL,0xf5c7bbbc0cbbfdeeULL,0x9c8a14134911068ULL,0x53c04e4141028a6ULL,0xe0758622b53359c6ULL,0xc6106a42202246e6ULL,0xe0142402208800c4ULL,0x205e9400002801c6ULL,0x200040010000002ULL,0x8000002800101140ULL,0x41000000c000000ULL,0xe020a266009a7066ULL,0x848822481c103400ULL,0x29cc1710cbdc482ULL,0x2000220100120024ULL,0x90c040008394000ULL,0x69a1d12a20325fc4ULL,0x662300604938064ULL,0x4103004001100ULL,0x1022048400022684ULL,0x80e2d701888c560ULL,0x0,0xc4870013c105000ULL,0x200001a00900000ULL,0x50daf6ec3484dbc2ULL,0x41ba2d0010d06042ULL,0xec102c0024884940ULL,0x6056c20000024866ULL,0x880804831200002ULL,0x3c150160101140c6ULL,0x8110420a14909002ULL,0x2082c0200200062ULL,0x8582f62804127ce6ULL,0x7b7713c614d2ffa0ULL,0xa224f491411796aULL,0x84a4b60224f961e4ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4080000280000ULL,0x0,0x0,0x0,0x0,0x8802388000984040ULL,0x44030000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x42000008004000ULL,0x0,0x0,0x2900010000040c4ULL,0x10800000280ULL,0x0,0x208000000002040ULL,0x0,0x24000000020200ULL,0x0,0x0,0x0,0x2008000100002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802000010000002ULL,0x0,0x0,0x0,0x80040000c000000ULL,0x0,0x3d7ef5d215bbefe2ULL,0xe4fff47834bbdbe6ULL,0xdfe7f77934bb5feaULL,0xff9e3f8435faebeeULL,0x88d3ea2a34b666c2ULL,0x289819800093e064ULL,0x94275e663cbbe9e6ULL,0xd81cea121c83e1e4ULL,0xdfa42ff534b979c2ULL,0xaac7fb923c3276e8ULL,0x2ed29fd91dbac1c6ULL,0x1034042101120c4ULL,0x4020208000084480ULL,0xf612273639b16cc2ULL },
	{ 0xcbb6fffffe34f6feULL,0xfbd7b2261cb45edcULL,0x3babefdbf479d77eULL,0xf1f6da6c7e44d6e8ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8b64c4044c13024ULL,0xdfffcff7f8d5d7faULL,0x2416622008901a02ULL,0x0,0x0,0x140380084106020ULL,0x14000204000ULL,0x0,0x0,0x80c00a0001166aULL,0xcbe3b293b6341fe8ULL,0xa61ecadb1e9dd760ULL,0x309164ea10745d5aULL,0xd9bebd4a26bc57f2ULL,0x42802e7c34d01228ULL,0xc54f680284345c62ULL,0x2101403048014c4ULL,0xc9124601c143680ULL,0xfd6dff63f63cf2eaULL,0xfeb62f7f2e9c7fe7ULL,0xefeedb12343cffecULL,0x259bfb7b887c56d9ULL,0x9f2055dcb494d422ULL,0xd48d343b14a414ceULL,0xfff67dfbc4bcdcfdULL,0x89deef3fb6355cebULL,0xeff77bdf74f4fceeULL,0xdfffdffffe9cdefeULL,0xf66ff7fe7e3c5fefULL,0x2023c61101cdd41ULL,0xcbc4d31e167c5c52ULL,0xeee3329be49cdefeULL,0x8b144e7af61df6e8ULL,0x1464074010110c25ULL,0xa12e2d6034157cedULL,0xc984c7e6201cc6faULL,0x9ecbfe8b16ddd7efULL,0x4e5eeb52341cc7f8ULL,0x54fbfee4f4b4f6ecULL,0xb56bfe78a63c1ceeULL,0xe3a7f7dabc71dca8ULL,0xefd6fb9896bc54fcULL,0xefbedfdcf6fd5dfcULL,0xbc5f5b7ab090d678ULL,0xfdce75563494f3eeULL,0xda802d43044516e6ULL,0x438bf76f54b4dc7dULL,0x8008000000c000ULL,0x2000602c0001028ULL,0xe0cfb2792020d6e1ULL,0x429001a030101228ULL,0x2b64662224b0db6cULL,0x22a63ad0303c5ee8ULL,0xefffdf7ff6317efeULL,0xdffcf6fda494fee6ULL,0x6adc7df3c87c5feaULL,0xabf5fb3fb0dc1efbULL,0xeddbfbdfb4f457ffULL,0xeb3c477a2c9c7efcULL,0xf1f7feed343c5ebeULL,0xb3aafd72b414d6aaULL,0x9b325d7414d45dd8ULL,0x89f7a70826b455e9ULL,0x2ffe9dcb545476b8ULL,0xf09ef66cd010d6eaULL,0x15670368604514a0ULL,0x12186e2c4041d424ULL,0xc1526a02b4407420ULL,0xef6d9d034141e66ULL,0x81aa03600414226aULL,0x815046080804d082ULL,0xa42650801c00a2ULL,0xc063e6288004d4aaULL,0x694412050435ca60ULL,0x800030020000028ULL,0x0,0xe50c71210010165aULL,0x80807012042015c0ULL,0x21600001c03c0ULL,0x61e4002201c50a0ULL,0x810000004001000ULL,0x2004e4600200c240ULL,0x4144830e401c442dULL,0x84cab2084010546cULL,0x0,0x62040014040080ULL,0x4180158040001468ULL,0x8640806001040ULL,0x8041061000100e00ULL,0x0,0x100280000020ULL,0x4082010002001042ULL,0xc210224044204164ULL,0xc20040040028ULL,0x118440280101498ULL,0x242442402040102aULL,0x95895f402210c668ULL,0xdba80d46b414f26aULL,0xadf4f2721e35d66eULL,0xe9dede5dbc155efaULL,0x8382c7582418dcaeULL,0xc1521662c4354668ULL,0x0,0x0,0x4000100000100010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1920014001400ULL,0x0,0x0,0x100200c00401000ULL,0x0,0x0,0x0,0x0,0x8001020824002020ULL,0x0,0x0,0x0,0x4010000000104000ULL,0x0,0x20040080000020ULL,0x0,0x0,0x8000000200100010ULL,0x14000000480ULL,0x0,0xa000000000c0040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4008000000240ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4800000401ULL,0x0,0x0,0x0,0x0,0x0,0x208c314000028ULL,0xd4f4fee234dcfee5ULL,0xdfede76bd6bdffeeULL,0xea362f71661469eaULL,0x6846c23ca05954b0ULL,0xaa18ced3805c5ce9ULL,0x3d3a798b7c1c7660ULL,0xb9dcff68fc3c56eeULL,0x82b63d7034705e6cULL,0x106cb3ce404f02eULL,0x184006890400146cULL,0x1a020808400110a0ULL,0x108a72c101c1c0aULL,0x7434277c84e054a8ULL },
	{ 0xb996fffff96eb2faULL,0xeabf3effd18eb2eaULL,0x831c4f30586082d0ULL,0x40009840102010c0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6f7f5bdf51ed236aULL,0xd5fae7ffc9c9276aULL,0xf1fefe7988c9274aULL,0x8504416080002002ULL,0xebfcf36a89c92762ULL,0x4c6232b0a0c52242ULL,0xb6bf4c72f06ca360ULL,0x100476c6f0ec1280ULL,0x0,0x0,0x0,0xeac153b80482442ULL,0x4000967aa06c2200ULL,0x82902d4088c12240ULL,0xffd4eb7bf9c9a758ULL,0xc5cf9630d048a1c2ULL,0x2b4e9fcad1a5b3faULL,0xf0141ca8d100a2caULL,0x91ed677879ecb2daULL,0x12832724602c23d2ULL,0x50060040002140ULL,0x629a1549c1410382ULL,0x973fceebf1ee2342ULL,0xf18e3262d0a533d8ULL,0xf2962ff7f9cd83f2ULL,0xceceb2aaf9c9a3caULL,0xee1ffe4fd145b1eaULL,0x1fb8ffa9d9658758ULL,0x9d5de7f171e4a2daULL,0xd7b67d7ff1e5b3f2ULL,0xfbf7fbbbf9edb3faULL,0x7ededebff1eca7faULL,0xbfbfcffbd9edb7daULL,0xf4fdfbe0d1afb2d2ULL,0xafbe7df3f1eca3eaULL,0xfbd4ffeff9efb3eaULL,0xfec3be17f1aea3faULL,0xf3d494869c5a14aULL,0x65ed161bd1699372ULL,0xe2fead73d9eda3eaULL,0xffbffefff9ccb7daULL,0xcffffe3ff9cda3faULL,0x797ddfdaf1eda2e8ULL,0xd4b2dee4d1c9877aULL,0xdfcfbf7bd9efa3daULL,0xdffefff7d9efb3daULL,0xfdcfff1ed14ca37aULL,0x6dfdbd7d1ccb3eaULL,0x3a32dfc1600d234aULL,0x6864f270d005a34aULL,0x68082c82d108a040ULL,0x8a267f24d085a2d2ULL,0x21730001008a042ULL,0x200040000000210ULL,0x1000006800400208ULL,0x86ac2f45d1c19280ULL,0xabfcfbe3c9efa778ULL,0xa2d73ad0d1452372ULL,0xbb9cdf7bf1afa1faULL,0xf4577679d94da3daULL,0xcbde7f63d14db6daULL,0xf9f5e72ad964b3d8ULL,0xeff7bff9f1efa2faULL,0xefdfdf5ef9eda7faULL,0x7cfe8fefd16cb7faULL,0xb79c3d10e165a25aULL,0x8094f6ecc108a340ULL,0x87e4beabd84da3d0ULL,0x65650c1d0a422d0ULL,0x805ea8710422d0ULL,0x856406007004a040ULL,0x880012228440000ULL,0x8810188120050240ULL,0x229ccb86d06d87c8ULL,0x8404002040440400ULL,0x1220660e0042040ULL,0x601204a2c148a258ULL,0x89c3c72498cc3242ULL,0xa1001c0842002ULL,0x81045c260011340ULL,0x510c802050448148ULL,0x4024077450082262ULL,0x820000400200ULL,0x400340000040148ULL,0x2000200008040ULL,0xd601d0a048440298ULL,0x6420b440408501c0ULL,0x1a0c2751100b1c8ULL,0x60c1fa13400da360ULL,0x6000000240000ULL,0x484040050048000ULL,0x8000004040000010ULL,0x3288400e20a40042ULL,0x180a10361a40038ULL,0x88888c3502610c0ULL,0x441000040008010ULL,0x808305805004a080ULL,0x5884262141013162ULL,0x0,0x1500c00400081c0ULL,0x4054102051048082ULL,0xd300830851200250ULL,0x10010000410000ULL,0x1820732700482d0ULL,0xc022608108482c0ULL,0x2000108100100c0ULL,0x1000800040002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8001001200060ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200000000400100ULL,0x2000400000400200ULL,0x10040000200200ULL,0x10011000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x804000000000210ULL,0x80040000000140ULL,0x8400000000000240ULL,0x0,0x800021400240140ULL,0x0,0x0,0x0,0x0,0x0,0x108014000400340ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100080040008000ULL,0x0,0x0,0x0,0x0,0x8000200040000080ULL,0x0,0x0,0x0,0x0,0x282100000040008ULL,0x0,0x0,0x2000800000010040ULL,0x14a454c8e8c52240ULL,0xc129ff21d9cc23f2ULL,0xc29a3850d1469352ULL,0x95e6fe28d16da3daULL,0x4e9e5fdbf96592d2ULL,0xb9115fe258cca3daULL,0x804205e0d1612240ULL,0x720023c6516413d2ULL,0xedf7ffe5f9edb2daULL,0x142a0402004a042ULL,0x82000000000210ULL,0x140c600008610240ULL,0x92327f71e1acb3c2ULL },
	{ 0xf90c776318acfd74ULL,0xdb4a860cd83aa560ULL,0x480000890000800ULL,0x3022000200160ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48d63f4211b8ef60ULL,0xffffffff9abcff76ULL,0xdff5f5fe983eff24ULL,0xff9e3d75dbbc3d24ULL,0xe9dff73fdbb8fd34ULL,0xcfcb4e7cd8bcef30ULL,0x653a0d4310acb534ULL,0x94bce02f10b0b530ULL,0x4bc074409a300406ULL,0x8010000008000002ULL,0x0,0xfdccd0599b8ef36ULL,0xf46696ecd8a4cdd0ULL,0x51fe7d71d8b8bde2ULL,0xdf56e7fc10baad74ULL,0x2fc4ee0f5aa8cf20ULL,0xbe77cf5cda2eed24ULL,0xf49d56e0d8ac2524ULL,0x9857e470d824ed22ULL,0x80822e4492202414ULL,0x1041020000084000ULL,0x249444308a82102ULL,0x8e127de890a8ad66ULL,0x47cc350812269dc0ULL,0xde975e6d982cf512ULL,0xebd4f7bfd1aaaf36ULL,0x7f58134bdaa4e562ULL,0x9fb3cff39abae736ULL,0x932dfe7cdbacf586ULL,0xdfb60d759abeef56ULL,0xfb9bffbbdab4ed36ULL,0xfedfbfd7dabced72ULL,0x8ffecffbdabcfd22ULL,0xdf75f67f9bbcff76ULL,0xfffe7d72dabcff76ULL,0xb3d5cf26d8369f32ULL,0xdde29281d2b2bd44ULL,0x218e492282888920ULL,0x9150b3ec90b2b614ULL,0x54be3d705aa8afe6ULL,0xdfb0ef7fdabcff36ULL,0xccd7ff2bdbbc6566ULL,0xe5ccdd2da88ed10ULL,0x94b4b4e890a8b594ULL,0xb9ee75f7dabc6f34ULL,0xdebe7ff6dbbaef64ULL,0xe7d6d7bcdab8ffb6ULL,0x4f5a7b498a3c6706ULL,0x44e050a89a248762ULL,0x71a723e202b4a960ULL,0xc8202448012c0140ULL,0xe4d094709824f514ULL,0x61810c09000ac80ULL,0x30006100028e404ULL,0x8c10068888304500ULL,0x9d967ff9da387f34ULL,0xe9d4f72edabaed26ULL,0xcac61ed99aac6d30ULL,0x4faceda3193c6524ULL,0xe0d152e9d820a926ULL,0x58824652802cec80ULL,0xfa45ea66d824fd14ULL,0xcf4277dedaa0ad16ULL,0xb7bf4f689a28ff60ULL,0xeddd1ee899a8fd54ULL,0xb539fd719a8afda2ULL,0xc8d0655818acf524ULL,0x83d4be1f9a2ced32ULL,0xa67406001aac6104ULL,0x2800084000ULL,0xdaa828438808e7a2ULL,0x10042a2448206100ULL,0xcd61a82010b0ad24ULL,0x939cca123a0560ULL,0x2016a1022a100ULL,0x24000000400ULL,0x60102100800089a0ULL,0x50000000004100ULL,0x2d208019200a580ULL,0x122401008300540ULL,0xc1c550002080520ULL,0x46442180184740ULL,0x8180450212084410ULL,0x2e80810a102cd100ULL,0x8904c20d2002400ULL,0xc800c8004400ULL,0x40010000400ULL,0x4000200008100ULL,0x0,0x0,0x0,0x2010000001080000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x28000000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000300000ULL,0x0,0x8028400000008900ULL,0x0,0x0,0x4000100010800000ULL,0x3000240080000800ULL,0x0,0x0,0x0,0x0,0x0,0x100000188000000ULL,0x0,0x0,0x0,0x0,0x80000202000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa010140088208140ULL,0x400101280000400ULL,0x0,0x0,0x12000000100ULL,0x0,0x0,0x0,0x3000200102000500ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000000200400ULL,0x0,0x802000080020ULL,0x0,0x0,0x5004849c9080ac04ULL,0xdd3cac5a98b8ed34ULL,0xf29e77c6daac6f04ULL,0xfff5af3adaa8ff70ULL,0xefffdd5cdbaef776ULL,0xb7f9ff669aaa7f32ULL,0xf9fffe74d9bcff36ULL,0x14b60b401280ac12ULL,0xcfdfffbedab87756ULL,0xfedf9edfdbaaad24ULL,0x3288443888a86400ULL,0x906b8f508a302506ULL,0x4202000084402ULL },
	{ 0x49b445701dc6f141ULL,0xc984010751406960ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8200000080002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa18974e17027540ULL,0xffba7ff3dd76fd73ULL,0x96fcbf791d567562ULL,0x1fd6557025486561ULL,0x895233ba12a0e742ULL,0x19c61a30159a4160ULL,0x43184b480d527140ULL,0x972642e915744161ULL,0x502434c285c67201ULL,0x9bc8fb6e05c67f70ULL,0xcccfb79805b64f70ULL,0x2b9ece0b8d947561ULL,0xc4b4b6ac15b64771ULL,0xe98fed62acde6760ULL,0xdbb6677618d60740ULL,0x4adbbb3f5dde6f62ULL,0x635ed7c79fd86560ULL,0xf4207de4df0e2560ULL,0x102054006840500ULL,0x58042d301080f041ULL,0x180e6b290842461ULL,0x2189e4000060570ULL,0x15a2de401c464572ULL,0xc148e6020d46bd20ULL,0xf6b50f70254e7f70ULL,0xebe5ff2b3dde5f73ULL,0x2fdecfdb1ffe7f71ULL,0x1dba57f37df63f70ULL,0xd12c3f7b5bda7f63ULL,0xdeae2f6f1dc6f763ULL,0xcbf7ffaf5ffeff63ULL,0xeff7fef68f8e7f73ULL,0xa3aecfff0df67763ULL,0xfc74b3f10786ff63ULL,0x428e3f33e5d6fb60ULL,0x2287270407407740ULL,0x4810309401c06c43ULL,0x2eac4f121c067163ULL,0xc14021801820260ULL,0xfebeff70dffe6771ULL,0xbb0c6e7dc5fc6d71ULL,0xcdc2f7afcbde7762ULL,0xa202040201800940ULL,0xd6c634ec0dd64d73ULL,0x95ceff71d7c27f72ULL,0xebb6bf5aafc4ed61ULL,0x8deaff6a45e67571ULL,0x6b3ad3512dc26733ULL,0x94f65643d0863760ULL,0x194c602104c44560ULL,0x6cc0257a47c46641ULL,0x8000000000004002ULL,0x8901f8911862b60ULL,0x8b025f1a070a5662ULL,0x38347188946162ULL,0xda006dcd05de6360ULL,0xc9b5d7231fc66f62ULL,0x6e9a0ecf8aee7761ULL,0xc5fcdda29dd67761ULL,0xd490d84553827671ULL,0x2a0754081a66c61ULL,0xc394e724dbc66561ULL,0x244a389881d6e262ULL,0xd39acf6129ae6461ULL,0xfd7fb7fa2db66763ULL,0xfff26fd30966f563ULL,0xe494eb93450efd61ULL,0xa52b1280c065571ULL,0x20561689d18a6121ULL,0x9030d8c00d406840ULL,0x254e13420d404341ULL,0xc82406721dcc7321ULL,0xc1044c0208cc0060ULL,0x200a189000024062ULL,0x20d2b1692d4a7520ULL,0x8000064244802240ULL,0x2010241000164142ULL,0x10b932009406611ULL,0xd40388042006402ULL,0x200010008400000ULL,0x2065000060402020ULL,0x400004101000a100ULL,0x6180f311c8842f60ULL,0x8426e0081c02410ULL,0x90040200940100ULL,0x4000208101003800ULL,0x0,0x20400000140ULL,0x0,0x80c200000000540ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000080005000000ULL,0x2000026010820020ULL,0x84000000440ULL,0x0,0x0,0x0,0x0,0x8200000000000440ULL,0x10020001000200ULL,0x40001000800040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2008002000400ULL,0x0,0x0,0x0,0x0,0x14004000100ULL,0x8000000002000400ULL,0x0,0x0,0x0,0xc00000000060ULL,0x0,0x0,0x0,0x8000100002000400ULL,0x0,0x0,0x0,0x10001000840000ULL,0x0,0x0,0x0,0x0,0x0,0xc00000020100ULL,0x0,0x0,0x0,0x4800000201024040ULL,0x0,0x0,0x0,0x0,0x0,0x81000000000820ULL,0x0,0x0,0x40520080842000ULL,0x0,0x102000000840000ULL,0x2104800008402000ULL,0x0,0x0,0xc8efa74099fe2360ULL,0xcbac73f70dd67561ULL,0xfff6ffbb9bfeff63ULL,0xe31ddf9b9dd67762ULL,0xbdb77ffa1ddeff61ULL,0xd1aab25801a27262ULL,0x1ae0253201487361ULL,0xe5ffffbfed467b63ULL,0xab123c818d042340ULL,0x608575005426a42ULL,0x8120a64a0dc65652ULL,0x303c09400d464960ULL },
	{ 0x89c6f03313083e08ULL,0x0,0x1441088010000248ULL,0x84b83668d6200e40ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200040000000900ULL,0xee322dd0c70d8c90ULL,0x795ecf285f9c2cc8ULL,0x84363c00c39a0c48ULL,0x10116840562a0c40ULL,0xd63c165206bc2d60ULL,0xcea2ad62542e3dc0ULL,0xdbf1ff73509d7d41ULL,0xf0ea72b1fa903e79ULL,0x9681d71072ee3c42ULL,0x55ad84fe779d3df0ULL,0x13822734d6b02620ULL,0xb1ed724c6167caaULL,0xdec7bf3bfebf3dc9ULL,0xaaffdddf7ebe3ee1ULL,0x268446ef7fbc7d6bULL,0xa448277a56acbed0ULL,0x2182f0144106080ULL,0x8559a1046192880ULL,0x1443602460420c0ULL,0xe4200460061d6840ULL,0x1180f2047e200d60ULL,0xe2b63f7a729d3df8ULL,0x65d3f732d3dd2e80ULL,0x68cf1e5c5bb6ecc8ULL,0xcfadf636fdabd02ULL,0xd19cfd009a843f48ULL,0xfffe7ffd5f1fbfb2ULL,0xebf7f77ffefe6eebULL,0x6fdd9ed37fbe6de8ULL,0xfef9cdfb7fbe2ceaULL,0xde36b5fc5bfe3d70ULL,0xfbeefff3febe7fe3ULL,0xd900e30c02ba2c60ULL,0xcccb3091cf082d00ULL,0xe63c0e447a002cc9ULL,0xe9f5be7ad324bce9ULL,0xdfbc2e76d79b2de0ULL,0xf09e7f07cfb8ec62ULL,0x75ef208c7d02c00ULL,0x3e934d5b4e933c00ULL,0x949776a1fe023f6aULL,0x696ea7625efa6cc8ULL,0x8aa627c41a2c3ec8ULL,0x4004062088010c00ULL,0xcbfc5d40be128c48ULL,0xb8c173e5da9f2d68ULL,0xd70c75718eae2cc9ULL,0x40228f7a479a0c20ULL,0xec18000421014a0ULL,0x47d6be3a4f9cbde0ULL,0xa4ff4760471f2882ULL,0x18d476fb1e0e2de8ULL,0xa7367f7d4ebe1ce0ULL,0xcb57d72357fe2cf3ULL,0xf7ce37dd7fbe2fa3ULL,0x1f9feffbd7ac3eebULL,0xd16526b946ab2d00ULL,0xefbe78687a8c3681ULL,0xdbd4fb6e673c3dd0ULL,0xceeb9e5dd6263d59ULL,0x2fbecf5b7aae3ca9ULL,0xffdfd7ead6bc2de1ULL,0xdf3aa5355eb72f61ULL,0xdc143f91c7dcbd82ULL,0x8046ea00440214b2ULL,0x1080e1020001d00ULL,0x441910a092202c00ULL,0x1042364220202c00ULL,0x8200e20802012c00ULL,0x8856061883053d80ULL,0x810008924e008000ULL,0x408082e0460c9d00ULL,0xd328a2d04f26bc00ULL,0x88023d44e7023c80ULL,0x4100066002282d00ULL,0xa900080270040002ULL,0x2a000000041420ULL,0x11883400120c2c40ULL,0x802040002012800ULL,0x102480041100400ULL,0xc5421c0010a02e08ULL,0x901c048102020c48ULL,0x8021400044061800ULL,0x2100000010800ULL,0x8000004000800040ULL,0x0,0x0,0x0,0x0,0x808000040000010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3804040040082ULL,0x0,0x800010000002001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400840000080ULL,0x10020000010800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500000104000ULL,0x400040a000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8100000020800ULL,0x10040000c0000ULL,0x0,0x0,0x0,0x0,0x0,0x10400080000800ULL,0x8202110010c00ULL,0x88000000000c0000ULL,0x0,0x0,0x0,0x0,0x8010000000002800ULL,0x0,0x0,0x0,0x0,0x0,0x1100800444000408ULL,0x0,0x0,0x0,0x0,0x0,0x1800000002800ULL,0x0,0x6002000400ULL,0xd9fc2f61761ebcf0ULL,0x933673e2579e3e40ULL,0xcffbf7beffde3eebULL,0xaa7ededffefffecaULL,0x85f75de7569d6eebULL,0x1840254658d40400ULL,0xa02240010e00460ULL,0x5000c000020c0000ULL,0x0,0x0,0xd3efef70de993c61ULL,0x9bf65f7c709e1ffaULL },
	{ 0xcae43beefb8f2b2eULL,0x840060000042280ULL,0x83a85ef28a1d670eULL,0xf25ff748bbcd2b26ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x72b62df98bcb6256ULL,0xc3f46622b2104104ULL,0x4dd55e6899ac277bULL,0xb110c6f39a1c2344ULL,0x3126062492812926ULL,0xf5a477a2d300211cULL,0x950812010002700ULL,0xca5aba9392302b3eULL,0xa6b7c77b9a8e6b04ULL,0xf1f77eedbb9c277fULL,0x1b8f5e182900322ULL,0x1a22090002024ULL,0x4eea3fb7dab66767ULL,0xaffffedf9bde6ffeULL,0xb0b2dae8cac42396ULL,0x93ac3f339ad80b12ULL,0xc50000010000100ULL,0x3a57beb3bbec2716ULL,0x9dc7f538ac86734ULL,0x308bd56182906726ULL,0xed6ef73b9b8c2336ULL,0xe4b6a7699b952f36ULL,0xedd6dbae9bb4633fULL,0x6d9c7b499a80277cULL,0x16814fcf8af12713ULL,0xc12d8460e3902302ULL,0xffde7ff48b9c6f2fULL,0xcbfffb7f9bdd2fbfULL,0xfffffffddbd76ff7ULL,0xbfbffffb9bcd27ffULL,0xd5fdfffb9bdf6f76ULL,0x9b6e6d739ad9636eULL,0xb70db2e0ac82323ULL,0x8852301a8b480b00ULL,0xab9ccf989b9a2e32ULL,0x10b596e99a94223dULL,0x8ff677f3bbfe670eULL,0xcbb67f3e9bcc6fb6ULL,0x234682228b062301ULL,0xbb297d189b886723ULL,0xb495feeedbc82d37ULL,0x806c266ada45230fULL,0x82802f4082a02306ULL,0x80020e2090000234ULL,0x44000a0000000ULL,0x3cba5bb799186716ULL,0xfb6eb658d3106be6ULL,0x4884047083100304ULL,0x8152e208ba1c2703ULL,0x12b2819a182324ULL,0x34b272e09b502b1fULL,0x844c5e8299006a53ULL,0xde3025d49ad04b14ULL,0xe888f3b2b3b02356ULL,0xfef7dfdfdb9f633eULL,0xf57eeefb9a946bb5ULL,0x510c24699390253eULL,0x18e5f5283812032ULL,0xd2f4ca62a394031eULL,0xaad992829a9c232cULL,0x3f38cf429bc9637eULL,0xb5d776cb9b5c6b6eULL,0xdbfefd67fbfe22bfULL,0xc49035129b002525ULL,0xc8c632969a8c2716ULL,0x203f17c58adc2222ULL,0x441500200a800210ULL,0x202025729a900322ULL,0x88916810a2042500ULL,0x44cbce089ad90332ULL,0x8a04860912802524ULL,0x414400081040002ULL,0xd100a064ca002282ULL,0x8202300718002704ULL,0x848242839ac00106ULL,0x82100100802300ULL,0x488040081002000ULL,0x2000020000202000ULL,0x840000c00102200ULL,0x802061a288882210ULL,0xe4421e1a81942323ULL,0x20004000040200ULL,0x104506108080020eULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x600000080002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24080040000ULL,0x0,0x0,0x0,0x820000080000008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500009000000ULL,0x14080000100ULL,0x0,0x0,0x0,0x4000001008400000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8008080004000ULL,0x0,0x4080000000002001ULL,0x0,0x0,0x0,0x22000002020000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000000100200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000082000000ULL,0x8000020000000280ULL,0x1080000080200ULL,0x210024090000030ULL,0x0,0x8008200090040004ULL,0xc2947ff81b840326ULL,0xcfdffefd9b9e2fe7ULL,0xefdfffdffa8d6faeULL,0xe5bb47f99ad42b1fULL,0x5108a9008a232212ULL,0x10241000800030ULL,0x0,0x2400000002000100ULL,0x64000010002ULL,0x9350a55093902b07ULL,0xfffe6ff7fbd66f7fULL },
	{ 0xcbd7d7b7ebb7d7ecULL,0x1a537e030190c3b8ULL,0x10240e304000110aULL,0x1240aed8d100c56eULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xde962d6cd9a05319ULL,0xc0dc520c6820c31eULL,0x2e4ab2d981b842f7ULL,0xc0904fb9e0ac4f33ULL,0xd18c921cd081d132ULL,0xbb62ff2c9a8517aULL,0x3a8a800080218200ULL,0xce5bb4dbe325d52eULL,0xafbbcd7ff1b3f37bULL,0x70403f6089b44338ULL,0x92c73d505194d33bULL,0x3004080000200314ULL,0x4741a03280809a18ULL,0x9b9ddfdbcbb2dbfcULL,0x58fcdee0e284f32fULL,0x1342f500e0a5c028ULL,0xb3067a5e1b0db88ULL,0xfdffffb7ebb1d76fULL,0x334efcd8c906d3fcULL,0xf4fa5ffbd1b7d1ecULL,0xc96cff7ae1b6f3feULL,0xdcf62fffe194e3efULL,0x658bf088e9a4d362ULL,0x29413ed001b1c209ULL,0x9eb06ff9e181d73eULL,0xf176ae7af194d7eeULL,0xdb96f9f2d1bddfbfULL,0xabfd5bbff9b0fbefULL,0xb2a6fbd8e1b4d37fULL,0xffbaeffbf9b1dbffULL,0xdf37a6ffdbb4f36cULL,0x2a2c3d60d192d303ULL,0xd9204316019042a5ULL,0x5297bb56e1807379ULL,0x6fbdcfc0e9b6f368ULL,0x7ef57beae3b4e33bULL,0xc1ee6f70e9b0d7f6ULL,0xfa3c7f6af9b5fb6dULL,0xdccff63ff1b0d3eeULL,0xfe752dc37102d3aaULL,0x16eade40c919f122ULL,0xb00002302000c120ULL,0x200406000004288ULL,0x0,0x295ca500101d20cULL,0x1cb7d77bf034d3aaULL,0x20249001200320ULL,0xce30674041a48336ULL,0x8000000100800001ULL,0x115a120b4181c1a8ULL,0x94335a60e331d302ULL,0xf17df66ac9b3d33eULL,0x9f943f21e0a1d369ULL,0xcbaffbb2e1bdd76fULL,0xdd6e7abfc990d33eULL,0x1bbecfff73bbf31aULL,0x844106616808c108ULL,0x4504e56ae19048f2ULL,0x359e012e188534dULL,0xecea9a50e1a0d9aeULL,0x273edffbd9b8f3faULL,0x657cdefaf8b0dfbeULL,0xffeeff73fbb67ffdULL,0xebb2e778c9b0537cULL,0xc8c38e00e0200a68ULL,0x6dff1d5881b26379ULL,0x4408c0a402a04302ULL,0xd14860008024c02aULL,0x2818090088204208ULL,0x480060000000ULL,0x8100a0061020000ULL,0x400100180a00010ULL,0x60e64060884001ULL,0x30a1050280a00360ULL,0x8af002048a204160ULL,0x5002004000ULL,0x200008c40018008ULL,0x2025326189225328ULL,0x81062c4060a0c202ULL,0x3800800000208ULL,0x0,0x8310050001100160ULL,0x14117426e1001844ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x20020000018000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4008000000120ULL,0x0,0x0,0x10040000004008ULL,0x1004000040000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400000840001000ULL,0x0,0x0,0x0,0x0,0x0,0x800100000200ULL,0x0,0x0,0x8000100001000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48000000120ULL,0x801000008008ULL,0x0,0x100000100800001ULL,0x40004000000420ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20002122001c000ULL,0x0,0x0,0x0,0x1000080000008008ULL,0x0,0x0,0x0,0x2000002000200200ULL,0x0,0x10000000108ULL,0x0,0x0,0x0,0xa0250041008128ULL,0x0,0x0,0x44050000000ULL,0x0,0x40008004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000018000ULL,0x0,0x0,0x11e6b540e198d32aULL,0xa11e4eb0e0a0c38dULL,0xe1d3ef1acb84d7eeULL,0xab7a5fdaf1b0d32cULL,0x416dca00001c62eULL,0xf97c2623d114ff2eULL,0x549e357089b2431dULL,0x8180c30e801051ecULL,0x42581068904200ULL,0x10884c8001048168ULL,0x8050020001004020ULL,0x108420000004422aULL },
	{ 0x40f430248084a8a0ULL,0x481e5e3381c6aa62ULL,0x80068000048202ULL,0x10418cc00a40a042ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc6b86d1083d3883aULL,0xc9a1de2683d4f122ULL,0xa750bfd803c6e3e2ULL,0xc9a4d41838622a6ULL,0xc18c78e183c4c19aULL,0xef3a2df88b4431eaULL,0xc0b0e20680c63142ULL,0xe6fa325f0bc62192ULL,0x122e4d450354602aULL,0x84d41ab60bc4c22aULL,0x225528046c140ULL,0xc0000a140204802aULL,0xc8624e0003460102ULL,0x5b9fcedf83d7f3faULL,0x56f6eee18b57f3e2ULL,0x99ef647a8b47a1c2ULL,0xcbdaae378bd6e836ULL,0xa784c21a8bc69262ULL,0x28505d5c89466bdaULL,0x34ca55ca82c1f01aULL,0xf77def5e0bc7a3c2ULL,0x80000d0002028122ULL,0xd5c27b1e88c6c160ULL,0x1a177b548946e2f2ULL,0xf7be6ff28bd7fbe2ULL,0xfff5bffd8bd4fb7aULL,0xdfec7ffb83d7f3beULL,0xc980e82e015671a2ULL,0xafc53e4a83d6b38aULL,0x13bcc77b8bd5f1b6ULL,0x4000910000040012ULL,0x5a80e4083c6c120ULL,0xd8d463008253e342ULL,0xc0c33ab383c7f072ULL,0x4d7cceea80d6b37aULL,0x746276f18356e23aULL,0x97a635600347e268ULL,0xfbbbff7f8bd7fb7aULL,0x7ed7b6a68bc6a3eeULL,0x6ff69bcb8bd6f372ULL,0xb6f7edea8996fb7aULL,0x80010000008040ULL,0x323060408944a308ULL,0xc08206480206a1c2ULL,0x23f09fd78bc4c1f2ULL,0xb7bb7de68bc6eb6aULL,0x9008a01080041a00ULL,0x9804000a00060230ULL,0xa10689090046a222ULL,0x6c90bb4b824623caULL,0xbcfb6ffb8bd7e37aULL,0xdeffdffa8bd7b3beULL,0xffb47fef8bd6e3feULL,0xabf7df3c8bd7bb3aULL,0x6ed53e3881c6eb6aULL,0xf6bf5ffc8bd6f3feULL,0x8487106b8904e122ULL,0xe29214200354c112ULL,0x7ac3410a82c23162ULL,0xe8cb9ed583c6e162ULL,0x2b3c4eef83d6f36aULL,0xfdf7767e83d6eafaULL,0x9fbaaf708bc6eba2ULL,0xde34bb7e8b44f3b2ULL,0xc070a42283049392ULL,0x294d98038a54a220ULL,0xb40100208201a000ULL,0x94c8a30082029322ULL,0x1200000002008000ULL,0x10000402000ULL,0x80005c100102c220ULL,0x48271000154c282ULL,0x20004000008020ULL,0xa02016000460048ULL,0x810192048340208aULL,0x40001000020002ULL,0x840263220355e200ULL,0x1241000000448100ULL,0x402405900844a0a0ULL,0x200000008200ULL,0x60e4081100c68362ULL,0x218080080028244ULL,0x404008000008200ULL,0x0,0x0,0x0,0x10000480004000ULL,0x204000400008ULL,0x0,0x0,0x0,0x0,0x680004000060300ULL,0x8001020800848000ULL,0x0,0x6000000008440000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x81000001000080ULL,0x110002002000ULL,0x0,0x0,0x0,0x4000000c80002000ULL,0x0,0x0,0x9150200001a080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x400004000500000ULL,0x0,0x0,0x0,0x0,0x8800000000000102ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x10020080004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6020002080040204ULL,0x0,0x0,0x0,0x2028400080005000ULL,0x0,0x0,0x20401040000ULL,0x8100001004000ULL,0x0,0x0,0x280000020002ULL,0x200203000000ULL,0x0,0x1000010000020020ULL,0x0,0x0,0x0,0x820000808000ULL,0x0,0x2004000808000ULL,0x0,0x0,0x0,0x800000800008020ULL,0x901110010240a208ULL,0xdb4c7c238ac7a1d2ULL,0x8f4078450352f14aULL,0x8ccdf6a488c6b272ULL,0xaa52044a0144224aULL,0x8000026080003002ULL,0x40420800008032ULL,0x202000040002ULL,0x4040e1200104a220ULL,0x20080000500000ULL,0x800031080501000ULL,0x10a4141000c08060ULL,0x0 },
	{ 0x88c1200a49491a04ULL,0x8c0080080800201ULL,0x2860c72cfc88a00ULL,0xb7c7966bccc1de30ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x52222c464d823804ULL,0xebfefff7ed4f7e3dULL,0xcebe3ed94cc9de38ULL,0x909d8ffa64ca263cULL,0x54e6b35bc7436e30ULL,0x822ad024cc012a0ULL,0x4b14c60ecfc2aa30ULL,0x414614144580d820ULL,0xe6bfcff3cfd7fe90ULL,0xf4f336f885cbfe39ULL,0x852637f28187bea0ULL,0xff2877df44ceb63cULL,0x302f80285c22a00ULL,0x1b3eef5fecdb3631ULL,0xf426f6d0c4cabe94ULL,0x884ae672c793fe94ULL,0xfbbe2b7ec5c20e25ULL,0x41d6d31fedc77e1dULL,0x8ffddfd3edcf7e3dULL,0x9f3e47ebecc67a10ULL,0xf34bf772e7cbfe01ULL,0x9a020a4006422a88ULL,0x580c1200d805a80ULL,0x221220034d402a14ULL,0xb5f229ef4fc2fe88ULL,0xe7cc973bccd2f69dULL,0x7b1ef472ccc32ab5ULL,0xf284eb8ac692b004ULL,0x267311da4dca2a30ULL,0xb89a0d80cec02e90ULL,0x9147bb7a87c37e11ULL,0xfffebdf9e7cbfea4ULL,0xbb102b2fc6cb1608ULL,0x86ea728904ca7e11ULL,0x42a4d15044929810ULL,0x4a6046048c95000ULL,0x8462d22c1c00e00ULL,0xcfb77fff6dcffebdULL,0xd062602a58a3aa9ULL,0x9b76cf5847cffe19ULL,0x94251469a44b1a20ULL,0xa97f2f7504dbba31ULL,0x1614130404033290ULL,0xcc8eb20781cbbe10ULL,0xa910265806c21a08ULL,0xddbfffeee5cbde95ULL,0x508244888c01210ULL,0x94024400420210ULL,0xd821080408102a0ULL,0x6edff6dfecdbfebcULL,0xb7ff6ffdcdcbfeadULL,0xfffffffee6cbfea9ULL,0x3386b461c79b7e39ULL,0x8bf7e36fcfca7eb1ULL,0x6fd2be3fcf4a3eb5ULL,0xb2290ca041cb9e90ULL,0x941606480e4a7220ULL,0xaa122d5049c11e20ULL,0xca16674e67c25e30ULL,0xe253f9532fd35a98ULL,0x3ffacffbcfcbbab0ULL,0xc9f5d7fee7ce7eb4ULL,0xf3f3af70cfc3fa95ULL,0x57a77fe84dcbbe3cULL,0x80e00e1000404e80ULL,0x8a5e8d43cec21e30ULL,0x40420c005002800ULL,0x184c651841c2aa14ULL,0xc00a4ec04e401a8cULL,0x2000421000400a01ULL,0x404c14da46490e14ULL,0x8002000000400200ULL,0x862014080d01800ULL,0x8290000002420208ULL,0xa084c10c04523200ULL,0x0,0x80400000081000ULL,0x1040042544402204ULL,0x52282d8048c23808ULL,0x8812180044805400ULL,0x1020000400800ULL,0x8090040000020010ULL,0x84400c2088c00200ULL,0x0,0x102001000022010ULL,0x0,0x0,0x0,0x0,0x200104800000ULL,0x0,0x60008204005008ULL,0x0,0x0,0x0,0x0,0x42004808081000ULL,0x0,0x0,0x280000006000000ULL,0x10800020008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000020004800000ULL,0x0,0x0,0x100800040000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x810000410000ULL,0x8000000200081000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x24080800000ULL,0x40400000081000ULL,0x0,0x4000000404002000ULL,0x0,0x0,0x0,0x0,0x4000000106000000ULL,0x800080004800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x820000101000ULL,0x0,0x0,0x0,0x0,0x8040000400200ULL,0x4000200000800200ULL,0x4000480800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10100020000200ULL,0x0,0x800010000020010ULL,0x200200080020000ULL,0x0,0x0,0x18c0212204107000ULL,0x880411244280de20ULL,0xc190d82066c00610ULL,0xabf64f4a04cbde00ULL,0x8002000000800200ULL,0x840000000020010ULL,0x4000000204020000ULL,0x0,0x4000402000200ULL,0x10002a40400820ULL,0x4020821044020004ULL,0x1014092008010200ULL },
	{ 0x888012ce0001849ULL,0x0,0x804410008000900ULL,0x5040929200521202ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xe21005260270102bULL,0xebdcef0a4a311b93ULL,0x0,0x0,0x40021000121000ULL,0x9bda26500a58134fULL,0x89207b34ba5210c9ULL,0x2003398000681041ULL,0xacbd556a4852920fULL,0xf577d76b287813ceULL,0xbfbefd737b72bbebULL,0xfaafff7abb5eb3c3ULL,0xffd7e6187a56bbefULL,0x2d08c819424d824bULL,0x10002000001001ULL,0x8832242208621282ULL,0x200108000801ULL,0xdfd7fffdfa7f9befULL,0xebffdfdbf87a9bdbULL,0x8c865d05787a1363ULL,0xdf3dfe7b687f3afbULL,0xf6be3decba7b1bcbULL,0x2646220408702082ULL,0x1b4efacea373120eULL,0x3dfb6ffbbb7a9bf7ULL,0xd52ca756aa669a8bULL,0x5a8e447e5876130fULL,0xa926d3686a5a12dfULL,0x6016160008400980ULL,0x20b86cc0a86e0242ULL,0xb52fbb566a7219dbULL,0xaa8eed68e8723b8bULL,0xfb77ef29295219dbULL,0xe6dab6162b668189ULL,0xe554fa3aa5533cbULL,0x70a0a67c2a561089ULL,0x104196020660000ULL,0xeba6ff37fa633bcfULL,0x8fdffe76fa7eb9efULL,0xe9375edb68721b8bULL,0xa694542348600917ULL,0x412081e000021000ULL,0x8b867622295a9acbULL,0xadd7880a6862901bULL,0x6bfedfd2e8723bcfULL,0x920085410a001107ULL,0x18f0a65008501a8eULL,0x82c0f00214a904aULL,0x81f653180a5810ceULL,0xa7fefa710a57bb4bULL,0xfff94ffbfa729bbfULL,0xf6feafeb2a7a9acbULL,0xe3fe7ff4ea7eb9cbULL,0xfbd7e76e6b7bb3dbULL,0x2e4a30cfaa6313cfULL,0x9ae4bf7187a12d3ULL,0x500a324208101906ULL,0x41180e9008701217ULL,0xb8c4f6288a50a21bULL,0x8b720e23ea0a184bULL,0x8dfecef7aa5e11d9ULL,0xfdfff7fe2b76badfULL,0x9b24ffd22a7f93cfULL,0xee2e5d4ea262934bULL,0x43c2360bba6010cbULL,0x28a4c900320812caULL,0x9231042a28608181ULL,0x2000042210401040ULL,0x12120a000870000aULL,0xa081d6020a4a12c1ULL,0xa10000a221001ULL,0x420504088021003ULL,0xd125160898601b03ULL,0x490004000423000ULL,0x8448222280001c6ULL,0x21478a402248024eULL,0x2004008000100ULL,0x10f0310ba06011c1ULL,0x2261500004000eaULL,0x88e22008501849ULL,0x4480021801400148ULL,0x88c0080b410301ULL,0x0,0x4002100000440202ULL,0x0,0x2002000040000001ULL,0x1000010008400000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8000200008000100ULL,0x0,0x0,0x0,0x80d00040000083ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40000008001ULL,0x0,0x42000008001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x88000020008000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800008000021000ULL,0x0,0x0,0x0,0x0,0x2280008000848ULL,0x800080001400000ULL,0x0,0x8400000000008001ULL,0x0,0x0,0x0,0x0,0x0,0x8080210008500001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000420000010282ULL,0x80000100021000ULL,0x0,0x0,0x3024d420044002eULL,0x83902b0120429041ULL,0x8580a60228129081ULL,0x22a4840820029081ULL,0x22482110501206ULL,0x1140e4174040184bULL,0x42882dace20800c9ULL,0xc080702140529001ULL,0x61215c220721082ULL,0x0,0x825220042004eULL,0x4652484800410202ULL },
	{ 0x8004204002c0000ULL,0x622080003140804ULL,0x8080042001000840ULL,0xa4841020002c00d0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xcaac45f0e13e79c4ULL,0xab82f73902349776ULL,0x64a320ae22c0874ULL,0x8f764f7a81373dd4ULL,0xa054167302353170ULL,0x5b887447401cb5d6ULL,0xc330331f64761454ULL,0x10224000000ULL,0x0,0x0,0x0,0x0,0x9feffeab677f5dd4ULL,0xdd6da9f623d3f64ULL,0x319a5329227c9e92ULL,0xf92a8f7a261f16f0ULL,0xcca66774c23e9ff6ULL,0xe3f4e626e23c1ed4ULL,0x833137ae71c1ab4ULL,0xb792fefbe66e9df2ULL,0x9acc5648c14f1dc0ULL,0xfeb41560223f9f06ULL,0x880031a220380bf0ULL,0x6d53de5b663c5dd4ULL,0x1b985ef0237e3dd4ULL,0x316d5428c03c9c96ULL,0xdceb671661eb5d2ULL,0x6900d730623648c0ULL,0xc8439c0765211440ULL,0x6e2aedc3003cb656ULL,0xbdf55efae67f3fd4ULL,0x100e306240141484ULL,0xe38afd2e65341bc4ULL,0x5fc3da11677417c0ULL,0x6a5aca53e43c1ed4ULL,0x4401c8427240a40ULL,0xc902556b643d10d0ULL,0x20900a0021000240ULL,0xd4ece31b603e82f2ULL,0x2afceccae5181e70ULL,0x810426a000d1484ULL,0xd94476e2006c95b2ULL,0x7cae3fe4e55d5cd0ULL,0x8936d72240346f64ULL,0xb95e5ecc627d96d6ULL,0xb4014062251548c4ULL,0x19fcbf6a665f9e32ULL,0xe6907ff5653d9df6ULL,0xc9f2f97b6337ddd6ULL,0xa5f75edc637d3dc0ULL,0x14c10522340340ULL,0xd4943e6541765940ULL,0x12a64562410c0c34ULL,0x8000093000060080ULL,0xd1d337db623f9dd6ULL,0x6faecffbe33e7df4ULL,0xd477feef635e5bf4ULL,0x2bbc7c72e23f39d4ULL,0xfe84776c83359cf2ULL,0x44807201601821c4ULL,0x32e589222240940ULL,0xa4725160241015d4ULL,0x9188150200240e44ULL,0x80220e00000c1340ULL,0x0,0x44165e00231c09e0ULL,0x426000260040ULL,0x160a640213404c0ULL,0x81002d0100061500ULL,0x2000520080001c00ULL,0x208864640141084ULL,0x80e0c140340800ULL,0xc1401402021c00c4ULL,0x42222520001c02d0ULL,0xa0001000800ULL,0x2940908100380060ULL,0x4010000000080080ULL,0x4100201000041404ULL,0x0,0x0,0x0,0x0,0x1000200000200040ULL,0x0,0x11000060000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x200000800040040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800102000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000200000080060ULL,0x0,0x0,0x0,0x8010024000140104ULL,0x49260200c0350860ULL,0x0,0x0,0x6000000000050ULL,0x80040000101000ULL,0x201000010040ULL,0xc60000000ULL,0x0,0x0,0x0,0x0,0x400004000040400ULL,0x0,0x24000000080040ULL,0x0,0x0,0x0,0x0,0x2000000c0000ULL,0x0,0x48000100200ULL,0x0,0x0,0x0,0x8060002000c1010ULL,0x0,0x1100040100081180ULL,0x991e5860660515d0ULL,0x514cc332663c08c0ULL,0xac1c06ca413c18d0ULL,0x3c975eac025e0fc4ULL,0xd2dc45a0601927e0ULL,0x82022567e6175850ULL,0x950902162101050ULL,0x211218c800309e52ULL,0x2000200080014ULL,0x108a0524215311d0ULL,0x5a63d50631c1150ULL },
	{ 0x6000000004004000ULL,0x40030c0006004310ULL,0x809ea20028008150ULL,0xe075e151e2008513ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x211827220808969bULL,0xe154c1061700cb13ULL,0x8dc1b6030c20e2d1ULL,0x74ecf720f21d118ULL,0x7551b26c1f20de18ULL,0x973ab5450f017a10ULL,0x1004004808204010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xef4f9cdf5d08df53ULL,0xde955fe35f29cf93ULL,0xe92eef73df89cf5bULL,0xda3a263f1ba8cdc9ULL,0xc8c6d31b4f01475aULL,0xc94e1759882095daULL,0xd49c47aafd21c7c9ULL,0xd265bc5b3b89c311ULL,0x7f5e7d7083089753ULL,0x83c7efe20b016d4aULL,0x2e425b684a8173d8ULL,0xcecec415b00ff5bULL,0x5536b37b5b88e39bULL,0x598cbd221f89e599ULL,0x0,0x40080000208000ULL,0x845842210a214309ULL,0x10a5f6eddb01d751ULL,0x9300bc002200054aULL,0xce202f061f00d81aULL,0xc82fa885a208309ULL,0xa16150a22800512ULL,0x16787cc4280885c8ULL,0xbffefff6ff89dfdbULL,0xfff67f7fbfa9ff5bULL,0xddffff3b4fa9efdaULL,0x4b8ef81913084ddbULL,0x351f5bad8f01ff19ULL,0xf9dce73a3f00ed1bULL,0x9a80a32c12088b0bULL,0x65d2c468580885d3ULL,0x9573e56db08d5d3ULL,0xb6b2c7e84b21df5bULL,0xddfdb47aeba9d7dbULL,0xffbefffeffa9ff9bULL,0xe9e2fb327a01cf91ULL,0xdc26a101b21ebd1ULL,0x479dce402f20c75aULL,0x9154b03e1e08db11ULL,0x80b0444001a0cb11ULL,0xb924820e1a00971aULL,0x46d33c600a0095d9ULL,0xbf7ecf7b3b21c793ULL,0xb495f6ffdf21dfdbULL,0x8abe7df11f21f7d1ULL,0xca94b43f3b21eb99ULL,0xcdd8f0a35b29e3d8ULL,0xa842d95007084913ULL,0xf43526e45f28e112ULL,0x4847731e084a43ULL,0x49960f0a1b000b81ULL,0x8000100001000002ULL,0x220000c002008403ULL,0x1423220113000211ULL,0x790c221a8400d518ULL,0x8010257002288110ULL,0x101c00001000450ULL,0x27a155a0400cfc3ULL,0xe28052000008689ULL,0x1a1c001012208b01ULL,0x2004000000100ULL,0x8844026096004700ULL,0x60000001004000ULL,0x8000040800000408ULL,0x1000801000800540ULL,0x44020000100ULL,0x41a000009010ULL,0x0,0x0,0x0,0x0,0x0,0x40001000000300ULL,0x0,0x0,0x0,0x0,0x800000404004000ULL,0x40208000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x840000008000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40004008000200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x101002000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000508040080100ULL,0x84820000c004100ULL,0x0,0x0,0x0,0x0,0x0,0x820000014000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000000402ULL,0x560c22001200011aULL,0xcd42beb01ba0af1bULL,0xeffedddbcb88cf1bULL,0x8cba59401b01db5bULL,0xd301a5608929e553ULL,0xfb1eb7e4be01cf9bULL,0x100800000001010ULL,0xe88a31451a208290ULL,0x1140000a000140ULL,0x3c5556220e01e318ULL,0x480865420a20cadbULL },
	{ 0x81928204220a40ULL,0x0,0xc10066310826440ULL,0x901028d084184420ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x804182000c8c8000ULL,0x2e98744005b84d62ULL,0x4a90600d0c284700ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000020800ULL,0x2004c5254880562ULL,0x5945112090184442ULL,0x9963285414ba0722ULL,0x4018272014880542ULL,0xcfdbff30d5fe5d6aULL,0xfd57fef73daa5f62ULL,0xb5966ded04aac9e4ULL,0x59d40ce604b86d6aULL,0x5eb66df3f5bed562ULL,0xc92aa3781d2a6f42ULL,0x7836178015ba2de2ULL,0xbabafdda5cb8fd6aULL,0xd4f5167a15badc66ULL,0x0,0x0,0x0,0x3422140230280142ULL,0xf415dd7a949b6d42ULL,0xc080040040200060ULL,0xe900601014809002ULL,0x80f2bc111daa7966ULL,0x12969f1254ba0466ULL,0x0,0xedf675f33dfbfdeeULL,0xff9f3ffffdfaffeeULL,0xcfcafa9bddba4f6aULL,0xbd48b1a15b0cc62ULL,0x14077961009acc4eULL,0xc07cae5810836142ULL,0x5a1aa74cf0c94c6aULL,0xc252cbbe34987562ULL,0x4e7e768c85ba0de2ULL,0x2db3effd141b6d62ULL,0xdffffff955bb7f6eULL,0xbbbcfd70f5bb6ce2ULL,0xc167faea148a8c66ULL,0xb7c8ff433493cd64ULL,0x20ccdc14c124d24ULL,0x3545f76d349b8d62ULL,0x4068757015a86c42ULL,0x8800610210d84402ULL,0x7cb9a40649a40e2ULL,0x7f768fd375ff4d6eULL,0xfdf5fffbd5ff7deeULL,0xd1bf79f17dfaf4e2ULL,0xfa44b7fc74cae9e2ULL,0xddab12bbc8a4d60ULL,0x8364d08349e5442ULL,0x91b3666e1c8aec4aULL,0x8246200020826ULL,0x20045000080c800ULL,0x4d02ce0c05b2584aULL,0x2100000808000ULL,0x300055c404c0c9a0ULL,0x91702340141a0442ULL,0x60900004000a000aULL,0x180400100884800ULL,0x86e0050008840e2ULL,0x78066c210b14642ULL,0x1060b01000086462ULL,0x610a92010194c42ULL,0xaa54403630016840ULL,0x18800e9035326440ULL,0x4a94d2210b84e22ULL,0x40064020880040ULL,0x0,0x0,0x0,0x8a08210008a4800ULL,0xfe830a8e040ULL,0x0,0x0,0x4000100000020800ULL,0x101001200000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x48001200000ULL,0x8010000080000400ULL,0x0,0x0,0x0,0x0,0x0,0x84000010000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000040000100400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x108000020004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x62800284800ULL,0x40020000c000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x80108004000ULL,0x0,0x6008005201000ULL,0x200280000400ULL,0x0,0x0,0x0,0x0,0x200400000900000ULL,0x0,0x0,0x0,0x0,0x10400004040ULL,0x0,0x0,0x808000000000a0ULL,0x113a9750901264e0ULL,0xf40625d2948a4c62ULL,0xc40233a14b84402ULL,0x4501404181220ca0ULL,0x480230880888400ULL,0x7804060050894060ULL,0xdca02743449a5ca2ULL,0xc090d68804aa4540ULL,0xac5008000090406aULL,0x620078004b80400ULL,0x90a40140608a0062ULL,0x8008480010000402ULL },
	{ 0x80800000000028ULL,0x2101142c4103060ULL,0x4120042220805020ULL,0x7015d27d020262e0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa080608002202c4ULL,0xdcbb5fe9f65276ecULL,0x7967767b6091d668ULL,0xff264f7b9490f6ecULL,0x88c6eb6e247374aeULL,0x825f5ab40013f68cULL,0xb45056f1a03272a8ULL,0x35c4d03ac222d6ecULL,0x48000402000204c0ULL,0xe845b36a648174e8ULL,0x76c7df4b0431f268ULL,0xf7b10572803336a4ULL,0x2000002000000084ULL,0x0,0x0,0x1100000000040ULL,0x8a3c6f60c4b23aecULL,0xfc327a711433f6ecULL,0xc6a21f40207176a8ULL,0xc7875040567252e8ULL,0x208000000880ULL,0x4080000104000ULL,0x0,0xfffffff3b6f3ffeeULL,0xffb26f54e212f6eeULL,0xe5ce6e0a66113ce8ULL,0x2d179007808054e0ULL,0x109665d68490502cULL,0x806000124200ULL,0xfc72645ed473526cULL,0x8001f80000123484ULL,0x8000114000000ULL,0xa4b047e3c2a353ecULL,0xb7ffbf70a471f7ecULL,0xc6963d31341346ecULL,0xe0f7e33b26f3d5eeULL,0xfffe1e7bb632f7ecULL,0x855ec06680d15468ULL,0x8c549ee86492d4a6ULL,0x70eefd7324b3dee4ULL,0xc365fb3e7453f6eeULL,0xb967968cb4b0d464ULL,0xeb8e4fabd652d2ecULL,0xfef776fff673f6eeULL,0xdfbeff73b6f3f6ecULL,0xfb96ffa3e6f3d6eeULL,0x220623022050d020ULL,0x8a1207486011d6a0ULL,0xbfbbffece6f3f5ecULL,0x22100000004e0ULL,0x3a9402242490d062ULL,0x8c0000204105000ULL,0x0,0x4323028a4102060ULL,0x822210400148cULL,0x2a802441000114e8ULL,0x210020000001540ULL,0x8020094200000068ULL,0x8d40492214105060ULL,0x100cb40300314028ULL,0x560a5920040014acULL,0x8864c22014131640ULL,0x484706011410c084ULL,0x1aacc021610d06cULL,0x81408230240014acULL,0x10001040000020ULL,0x110000002000040ULL,0x2380004b01020ULL,0x14ed5768201246ecULL,0x34aa6fee743272ecULL,0x4002020000020ULL,0x0,0x800220082100040ULL,0x0,0x8400002420022020ULL,0x0,0x100400000000240ULL,0x0,0x51500000823028ULL,0x24010100c0100020ULL,0x20080000020200ULL,0x280000000000028ULL,0x0,0x0,0x48004200000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x804000604801004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10001014000000ULL,0x0,0x0,0x0,0x0,0x8000040080000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6004000004ULL,0x0,0x0,0x0,0x0,0x8200040080001420ULL,0xc000000000040ULL,0x0,0x600000000000028ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x10000200004000ULL,0x0,0x0,0x60000080100000ULL,0x20004480200c240ULL,0x0,0x140280440801460ULL,0xa2843f1e867050e4ULL,0xa00068220082188cULL,0x90800200004440ULL,0x0,0x42a7403611d468ULL,0x688235420092fc88ULL,0xc1320200a0105028ULL,0xa400000800104280ULL,0x402208004022004ULL,0x100840024002020ULL,0xa22a2d68947052ecULL },
	{ 0x8b82f2bb51bcb7daULL,0x60f110c0401c1088ULL,0x430e4332d114b15aULL,0x93a0a210511431c2ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8854526a901c1c8bULL,0xf0be1768903c364aULL,0xc3a4eed6d02c36c6ULL,0x2904900080001002ULL,0x695e76f0599c3fd0ULL,0x1680a7fb519c94d2ULL,0x9837dee950169dccULL,0x90204e0cd1041408ULL,0x8834cf3c9804948bULL,0x0,0x0,0x0,0x0,0x0,0x2842a0004010b050ULL,0x292ec548401e3a5aULL,0x657d4268513226e2ULL,0x755180141812ULL,0x4800100200ULL,0x0,0x0,0x547476e0b19c969eULL,0x262c4080141882ULL,0xd6806f62d0ac34d0ULL,0x8566f22bf01496c2ULL,0x8f61f4158849148ULL,0x41e1a09581c1018ULL,0xd18cb6d8710cb148ULL,0xa102f2a1034b7c0ULL,0x2944032b4104254bULL,0xaa4e08d05110b64bULL,0x91b20ff2519c3692ULL,0x367cbe505994bafaULL,0x6f963f74712ca448ULL,0xc9bff72bd93eb7deULL,0x6edd5f7d593ebeeaULL,0x27becda9d91c33daULL,0xb465f7d471bc9fcaULL,0xcb7e7f6271aeb3daULL,0xcb5ee20a711c2942ULL,0xc8d3f393d114bfdbULL,0xbfbfdfdfd19ebfeaULL,0xb4fddee8511eb6ceULL,0x37befdf3501cbdcfULL,0xda8eeb9fd09cb6d2ULL,0x131b12d114160aULL,0x671c1bdfd81c109aULL,0xf5bfffead91c9fdfULL,0x865004a51101c90ULL,0xd200230c510092c0ULL,0x404a600040149040ULL,0x2202948040048048ULL,0x901504a00104016aULL,0xbea545418ea078ULL,0x83946571699630c8ULL,0x8002c06111101682ULL,0x28075908680412caULL,0xeb2a0fd3513ca7d2ULL,0x51007603511c1cc6ULL,0x822848411014148aULL,0x8152083050101406ULL,0xfc01a0841143242ULL,0x53a8c463511cb4d8ULL,0x10601440401210d0ULL,0x0,0x0,0x0,0x40800000401ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8482420102004c0ULL,0x0,0x1000c30000342240ULL,0x0,0x1000080000140000ULL,0x0,0x8100000010100000ULL,0x0,0x840000000000140ULL,0x0,0x8110040041000012ULL,0x2000040000020040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8040000001002ULL,0x0,0x0,0x0,0x0,0x0,0x1010000210101800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1001000000000240ULL,0x0,0x0,0x0,0x0,0x200040000000140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x82050008001480ULL,0x44040210241048ULL,0x0,0x2020000008040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x50000000000480ULL,0x80100020040ULL,0x400200010000080ULL,0x44e460c01c3350ULL,0x2122a005104804aULL,0x942e22a511c175fULL,0x805000104000aULL,0x8a2434051020440ULL,0xf1a4a260111485c2ULL,0x12000dc040148058ULL,0x4002000840000ULL,0x0,0x1242a0101030c0ULL,0x0,0x9d0c3d687984b1daULL },
	{ 0xe9d2673698d02d66ULL,0x6000015090808800ULL,0x208000008400000ULL,0x1600b24210e12260ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8440260280690700ULL,0x10001080402100ULL,0xcccbb0809820ad14ULL,0x24ed8c088680720ULL,0xb5ba1d8a9061f7f0ULL,0xa043270088809d00ULL,0x288445e0d0803850ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8039400008a100ULL,0x2a809088840ac00ULL,0xa17c1ef09070be50ULL,0x8008600120012242ULL,0x0,0x0,0x8201449300800100ULL,0x36b347e8b8689f38ULL,0x90825500d00a2e00ULL,0x5bbc6df4d861bb76ULL,0x8f8ae72a88d0a744ULL,0x875edb5a90f0bd14ULL,0xc321100d0202c00ULL,0x110a03a098500260ULL,0x5883822400602520ULL,0x8000d000c0009100ULL,0x669e7fddd0e9bff4ULL,0xc806fe088492752ULL,0x91da376bd8e2a302ULL,0xf2de7d7d98f9a7f4ULL,0xfbf7febe98c9bf7eULL,0x6dfe7ed9f8e1b77aULL,0xdddb7fead8eab784ULL,0xd57dff7bf0e9af56ULL,0x9b167dd0b869be50ULL,0xb3c4ef4eb0f0ef72ULL,0x8a72fe5298c2bff0ULL,0xeebfed5b986baf72ULL,0xf4fb367ed0f2e736ULL,0x812a0d208008a690ULL,0xfabc777ed06befd4ULL,0x856f28ac8602c40ULL,0xba415d02d0412650ULL,0x50a00ec0d022ac10ULL,0x800c000808412042ULL,0xb100012080010120ULL,0x8402820890a02500ULL,0x10100010000010ULL,0x1000080000010040ULL,0x4000720890004120ULL,0x4204043398c00320ULL,0x100800080000000ULL,0x844161190602330ULL,0x14824b81a0410770ULL,0x10e01c0098804d00ULL,0x3906504800408160ULL,0x400280000000ULL,0x8ee312539068af20ULL,0x8a00cff99843a1e4ULL,0x2000002008400000ULL,0x0,0x0,0x190000000ULL,0x480000400100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000004080000000ULL,0x80800080100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80040000208000ULL,0x0,0x0,0x4010620202200ULL,0x0,0x0,0x0,0x0,0x20090000000ULL,0x0,0x1800040890000000ULL,0x404000000008800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x401008400000ULL,0x20004000080200ULL,0x100040010000020ULL,0x2004000002400ULL,0x60000010040ULL,0x0,0x0,0x4000800000000120ULL,0x100020080000800ULL,0x0,0x21090000000ULL,0x10002008000400ULL,0x0,0x0,0x0,0x8002000010000010ULL,0x4400008400000ULL,0x0,0x0,0x0,0x0,0x0,0x4000008000400200ULL,0x100100010001000ULL,0x26101e70986085e0ULL,0x1818a0a08400520ULL,0x4110580000418480ULL,0xc92eaa3b880ae10ULL,0xe30c06a290e0b618ULL,0x801401c000489200ULL,0x480000000000100ULL,0x6511c5090e82a04ULL,0x100010000400200ULL,0x8061020080002442ULL,0x4a006102c0c12000ULL },
	{ 0x6084c310003ca420ULL,0x0,0x128002001024001ULL,0x9224160808084f62ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdd6200a2b0c4561ULL,0x29feeb71bb7e6f66ULL,0x1fa04f50cb3c6720ULL,0xb10cb04828f84120ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x3820021080020ULL,0x483815002c84102ULL,0x0,0x0,0x0,0x0,0x1f7fdf5adb3c7762ULL,0x8c95c4e000347660ULL,0x80a25742050a4463ULL,0xd38405a28b042750ULL,0x44a0c0002080120ULL,0x808000000a000ULL,0xb8800de12148ec70ULL,0xc070060ae0006421ULL,0x3a92332009007c61ULL,0x8480c00125405020ULL,0x26111eca2a486520ULL,0x346a7280032c4442ULL,0x73aeb5d2956cff43ULL,0xfef66f649b3ced62ULL,0xc9eef3afdffcf507ULL,0xfef73fd973ce7f61ULL,0xaf3ace8bcb7e6722ULL,0xf4e58279eb7e7721ULL,0x6afea5facbaa6763ULL,0x9bc5cb264a7cfd21ULL,0x37e114605c86e20ULL,0xe6dedeabfb6ce763ULL,0x707373fa2d6c5e61ULL,0x10a8ed0048882570ULL,0x4a08e10212084430ULL,0x68d4110120c8e021ULL,0x200e419a18606f61ULL,0x14115ca29b2c6561ULL,0x108045600524f000ULL,0xcc44240189362000ULL,0x8020000000240000ULL,0x600084000ULL,0x1081042040406001ULL,0x206046022884420ULL,0x4606106402883960ULL,0x0,0x45301a484a402560ULL,0x1600a09109243400ULL,0x9816000082120ULL,0xda8822240a48e544ULL,0x8900420242882400ULL,0x9f11c060a6c6470ULL,0x1108000208084000ULL,0x500000004001ULL,0x0,0x408010000400ULL,0x0,0x2020000002000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80800000084000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1400000000000404ULL,0x0,0x0,0x510201482000ULL,0x1080000402000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x500020000020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x10001000000400ULL,0x4000000001020000ULL,0x0,0x0,0x0,0x0,0x100020010000400ULL,0x0,0x20400000000060ULL,0x0,0x500000404000ULL,0x880101008045000ULL,0x6000000005000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002008400000ULL,0x0,0x0,0x0,0x4844202000880c20ULL,0x4002066808080422ULL,0x80000202000100ULL,0x2001508a02040d60ULL,0x489026800240740ULL,0xd100a4900a8c4c33ULL,0x3840000848042420ULL,0x0,0x4000160000200860ULL,0x0,0x50e012100f000403ULL,0x2001c009004200ULL },
	{ 0x8bbffb6fbeffbd60ULL,0x4c1000002022800ULL,0x420c402000c40408ULL,0xf13efb6a96bf7d60ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4220100104408ULL,0xc5ca37bc5ed62d68ULL,0xff7fdfdfdffeef60ULL,0xb3b85d29079d3d40ULL,0x44000861000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x7c33e0902981440ULL,0x280691081c900540ULL,0xb40ce6641ab23c00ULL,0x1b2c1778539e7d40ULL,0x5186204c17902940ULL,0xe5d7f8a814be5560ULL,0x2a744c1af2b84440ULL,0xb000016044822100ULL,0x602c240046102440ULL,0x82912e65c2ba2d60ULL,0xcf9c13254a1a3d28ULL,0xa6d0bec7069a7d00ULL,0x2eaa6f52fbf22e40ULL,0xf1df57edc7f77f68ULL,0x4aba2502c68e2c60ULL,0xc1047a8514962528ULL,0xefd79f575ebf7d60ULL,0x2498dd5147902f00ULL,0xf575c6624e9a7548ULL,0xdfac2d635fb2b768ULL,0xfbdeef6f6fb67760ULL,0xa7fef9115b330d48ULL,0x6efcdfb00fba6d40ULL,0xd5bfdeefdfde5f60ULL,0xe8183d415d922d20ULL,0x889020000220440ULL,0x2830203101400ULL,0xa142090183302940ULL,0x80909468063a1428ULL,0x100a20400a104400ULL,0x8202010c04125140ULL,0x846aaa0016942140ULL,0x10130020001400ULL,0xa2800005000540ULL,0x100661144120500ULL,0x24024040840400ULL,0x4000209003100400ULL,0xbc9180005200d40ULL,0x289045f902962d20ULL,0x8c01000060800ULL,0x2aa2d400e128940ULL,0x88c00002002900ULL,0x767d510bd61f00ULL,0x1614c548828a2d40ULL,0x1a40020840842420ULL,0x4000008020000400ULL,0x8000100000001400ULL,0x40001000001400ULL,0x0,0x100006200000440ULL,0x0,0x2000042000940000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x800004000a000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x1002000060000ULL,0xa00004100000ULL,0x0,0x0,0x130040204100940ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000040a000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21010800000ULL,0x0,0x0,0x0,0x0,0x8000000000020040ULL,0x0,0x0,0x0,0x0,0x1000040000000440ULL,0x0,0x1000010004100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10800020020ULL,0x0,0x0,0x0,0xa00000840000ULL,0x0,0x0,0x0,0x2000002000021000ULL,0x0,0x0,0x0,0x0,0x1000200010000400ULL,0x0,0x0,0x2c20010840400ULL,0x4040000021000ULL,0x4080a000000ULL,0x4824780000220440ULL,0x9810aa780ef80400ULL,0x80a0010a0a826400ULL,0x281096c810840540ULL,0xc8b8456a05964760ULL,0x9300de0016963548ULL,0x461024900e220c60ULL,0x884493048a002900ULL,0x0,0x92200002001400ULL,0x500002000400ULL,0x4e60040002082540ULL },
	{ 0x6050010882046080ULL,0x800008080002000ULL,0x120000000040080ULL,0x2acbe71d224267dULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xca10211010843041ULL,0xcfc3fff1f23e6bdfULL,0x4fdffec4f71c7e9cULL,0x401080004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6000000080002000ULL,0x88e7f2a2d6b20bdcULL,0xa00ac042d02400e0ULL,0xfd7c7e6ef7346b5cULL,0x99ac7441d0862240ULL,0x9a40c30692246304ULL,0xc9d7eb98f60e7f44ULL,0x4006100002002804ULL,0x142303c0c2026804ULL,0x990e7402c6122a54ULL,0xa302e7292002746ULL,0xa9d6fb02d6046cffULL,0xefffffdfd7b6675fULL,0xb4fa47e2f28625e4ULL,0xd0d07c14d6160160ULL,0x292b44044042c82ULL,0xc920eb1ad2846147ULL,0xa571e50c2802858ULL,0x25ebc8cbd60c6eb4ULL,0xf51dbabfd79c3ae6ULL,0xffbe3ffbd28e6bffULL,0xf3d6af3ff6be6fedULL,0xac16bfb9d2363addULL,0xbfaeff69d72e2fe8ULL,0x9532ff68d60c635fULL,0x291091002501ULL,0xc82cf7ded60c25faULL,0xedeba00d6142d97ULL,0x8a4e2056d514186cULL,0x96f20660540620b1ULL,0x22c042092040014ULL,0x4202600630a00040ULL,0x2200000002800ULL,0x2400000000002001ULL,0x80520000802040ULL,0x1061328c4022501ULL,0xc1902535d6200344ULL,0x4650002050002880ULL,0x29ed8810c2063205ULL,0xb0100ce0d6042bc4ULL,0x22d951d20e2003ULL,0xc002664c82006204ULL,0x900012a9004204cULL,0x8080080200002400ULL,0x20040000041000ULL,0xc804010002020ULL,0x0,0x0,0x800020000002400ULL,0x0,0x880000c0000ULL,0x0,0x80000280002000ULL,0x0,0x0,0x0,0x0,0x0,0x44000010000040ULL,0x4000100000002001ULL,0x0,0x0,0x0,0x8000100080000008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x11c82102000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x722000402ccULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8100000002000ULL,0x0,0x0,0x0,0x0,0x0,0x8000180000100ULL,0x0,0x41000041000ULL,0x300000c0000ULL,0x800840000010ULL,0x0,0x0,0x40830000000ULL,0x0,0x0,0x2200044000000ULL,0x0,0x0,0x4002000800040ULL,0x0,0x4000040000000204ULL,0x0,0x60000040020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10000404002c0040ULL,0x8400662c4002244ULL,0x2090000200002400ULL,0x844321282002115ULL,0xcabddf93562027ccULL,0x1c0c018354043860ULL,0x10d0a904c2222148ULL,0x5480000040022001ULL,0x0,0x0,0x0,0x80150010202004ULL,0x54020946d2002d80ULL },
	{ 0x4000c11201921201ULL,0xa1d2900b81200b24ULL,0x40802800000ULL,0xa401022801204b02ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x45006102881041e0ULL,0x68002620091041a0ULL,0xcfcffbffabf553b7ULL,0x211a1050890a4302ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xeb9c773fabba73fcULL,0xa15c760309d04722ULL,0x90cc9d2a8614266ULL,0x9694756b01b44227ULL,0x8340967981bc6ea2ULL,0x83be6fc381b673abULL,0x8fd7b38289f44b6aULL,0xca4e395381161324ULL,0x42054e003804062ULL,0x886c5f61a9b20339ULL,0x5eb72df289b45f63ULL,0xcf5ff7be8b9d53efULL,0xbfdeffdfabdc7fbeULL,0x1743437289b57a12ULL,0x9069a7768a841233ULL,0x53b27c548af83398ULL,0xc900600180080624ULL,0xeed797008be07f2cULL,0x165d479e8bd4472eULL,0xb6f526500b9045a2ULL,0x4feead6089f44327ULL,0xcbb5ff26a9fc4a67ULL,0xfc43bf4dabda63b0ULL,0x9feecddb8bf8426bULL,0x35246ef8891c1b22ULL,0x428825e689300a68ULL,0xdb8d4fb9abfa2326ULL,0xc9eb7e138bbc57eaULL,0x67164bc88bb8732aULL,0x34a24ce40ad062b0ULL,0x40a0404088080304ULL,0x0,0x5d4608a81a0133cULL,0x1004080001000ULL,0x90040880200100ULL,0x5064a46009025320ULL,0xa010004880010122ULL,0x250020001900001ULL,0x18101000001320ULL,0x16324d44a9d34432ULL,0x601c2888c00302ULL,0x1a32045080811126ULL,0x8021810008c00101ULL,0x0,0x401000008004000ULL,0x1434042480200a08ULL,0x0,0x0,0x0,0x0,0x88008400000ULL,0x2000012080200004ULL,0x0,0x8000100000000100ULL,0x4000402004000ULL,0x200200c008800201ULL,0x10442090a004000ULL,0x9010002080000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x6001000002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800002001000002ULL,0x600000000800100ULL,0x8000040000100200ULL,0x0,0x0,0x100800000100200ULL,0x6000008080000ULL,0x30008080000ULL,0x0,0x0,0x400209000000ULL,0x0,0x202090008006000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000001001020000ULL,0x0,0x0,0x1022080002004208ULL,0x100000802800000ULL,0x400000c880800810ULL,0x8000100000000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000100000800001ULL,0x0,0x0,0x0,0x2042028180000124ULL,0x0,0x41001100000ULL,0x4b822420a10046a0ULL,0x68564010088803abULL,0x5c01c68810426cULL,0x2726a09844323ULL,0xd4acf57088585a86ULL,0x41000040200ULL,0x4020400040120ULL,0x0,0x0,0x404000000040200ULL,0x10412c8090440a2ULL },
	{ 0x2004020000008028ULL,0xeb4e94caf104e91aULL,0x810ea407000c008ULL,0x424200003000102ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x816a2e768984e102ULL,0x2a60c6089442288ULL,0xc9d4dfa34bc5d13aULL,0x80080000008008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xdeb6fb7e7bcdc3daULL,0x64000a40809012ULL,0x392ce1080c58820ULL,0xc6f2f6ecc1c5c33aULL,0xc18e622b514cc01aULL,0x8b200107c0806102ULL,0xfe5fca62f1c4c01aULL,0x6f555fc829ccc3a2ULL,0x10400000000102ULL,0x162a91080c4e01aULL,0xfd36af75c3d5eb7aULL,0x8fe9f777f345f37aULL,0x2e7ebadfd985f33aULL,0xb6674fffebc5eb3aULL,0xf87d066af350f10aULL,0xe6a0356d0b50c10aULL,0xc85aea6ae3c45b0aULL,0xeefe5f56f2c4f13aULL,0x4e9e2fe3cbc4e8a2ULL,0xdc55bf7c4adcd16aULL,0xc04e4c34b0cc110ULL,0xc8c5c770e144e13aULL,0x7433ac343c4431aULL,0x192cb97e1c4f02aULL,0x94d13561c2d5ca22ULL,0x10ae2d704044c30aULL,0xbb953734e354e202ULL,0x48c230000154c300ULL,0xa11b4650c304e10aULL,0xc43272e80104839aULL,0x102020083002000ULL,0x10042a042104c010ULL,0x824430000340404aULL,0x44060000000ULL,0x20802c0007000ULL,0x11ac00607804c000ULL,0x188004606100c100ULL,0x450020078100e022ULL,0x8206100683046012ULL,0xb40044120100c102ULL,0x10016014000c022ULL,0xe18025424304c032ULL,0x8000024401004002ULL,0x60000080002000ULL,0x410038480408962ULL,0x100800000018000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1200480061000200ULL,0x0,0x1400000c000ULL,0x0,0x81c00001000092ULL,0x24404028004002ULL,0x1201004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x100100000000102ULL,0x0,0x0,0x0,0x0,0x0,0x104000001000080ULL,0x0,0x0,0x0,0x0,0x0,0x20180004000ULL,0x0,0x2004400000c40200ULL,0x0,0x0,0x0,0x10040000004002ULL,0x0,0x0,0x4000200050000000ULL,0x0,0x0,0x0,0x40100000040200ULL,0x0,0x0,0x0,0x0,0x1004000000000108ULL,0x2004002000002ULL,0x0,0x0,0x0,0x204000808000ULL,0x8080040004000ULL,0x0,0x0,0x10100001040000ULL,0x80040000500000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2000008001040000ULL,0x2020000000102ULL,0x0,0x4002000008008ULL,0x0,0x0,0x48080004000ULL,0x0,0xc000000804000ULL,0x0,0x0,0x0,0x40020000004008ULL,0x8010404040004108ULL,0x840000000000012ULL,0x480000040008000ULL,0x100801000100400aULL,0x800010000000c000ULL,0xc840240000048032ULL,0x88241fa00a04812aULL,0x5000070881042200ULL,0x2046d51269440182ULL,0xd4b2bbeb23ccc1eaULL,0x1000a258c9042200ULL,0x169401508344400aULL,0x0,0x40e190405081c020ULL,0x98885130288c122ULL,0x4004000000008020ULL,0x0 },
	{ 0x4848410240422190ULL,0x18663291c0cab624ULL,0x669bc5f044c18214ULL,0xd4be50480ed07a00ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc18cd40001030a0ULL,0x2cbec6fa8ecbb2feULL,0xd9fe7f724cdaf628ULL,0xffbc2fffcedbb67cULL,0x384c4f48213b224ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa910358286c02828ULL,0xc306f77d46c0f870ULL,0x88f23e26484aba62ULL,0xa2200b4306c28a50ULL,0x56a61e804403a230ULL,0x20005014400000cULL,0x8a16273ec6d3f620ULL,0xe856af3fc6cab8dcULL,0xe3d6f7dbce8bf6ecULL,0x1d9665ee8ed3b3caULL,0xf17fbfea4edbf7f6ULL,0xa1e87340203f2a0ULL,0xf5d2f232ced3f2f4ULL,0xefffffffcecbff6eULL,0xcbab4fdbc6dbf37eULL,0x10e8cd104ed01014ULL,0x12c6bd40c8d23f84ULL,0x837e738906829300ULL,0x69f77fc24e9ab38eULL,0x99ac55b846c3b2c4ULL,0xf67df671cec3bb94ULL,0x31884d7144c22032ULL,0x535066228a80a2c0ULL,0x8e080040c802220ULL,0x8e9f4f7d8edb3a32ULL,0xa05316e08ed9ba80ULL,0xc426fd404ec20234ULL,0x10886c040e002200ULL,0x8a43c22ac4ca1034ULL,0x24084a904ec0a100ULL,0x345558e4c682b100ULL,0x1207a4100109180ULL,0xda807ac64ec2a300ULL,0x800210080c480000ULL,0x208108140105004ULL,0x2091042008402880ULL,0x100a41242005204ULL,0x428224000cc03200ULL,0x4000400040800000ULL,0x140020004800000ULL,0xa5a24fe08443a724ULL,0x83c16a184c2ab00ULL,0xaa9025504cc38a00ULL,0x800182644482a420ULL,0x6900070804828600ULL,0x200060948008210ULL,0x20340060c4802020ULL,0x0,0x0,0x0,0x0,0x1004044008020240ULL,0x0,0x1000004004020000ULL,0x0,0x0,0x0,0x0,0x8000001040800000ULL,0x0,0x4010000000020040ULL,0x80040002002000ULL,0x0,0x200000200018000ULL,0x200400000400800ULL,0x0,0x0,0x0,0x0,0x100800000020100ULL,0x0,0x0,0x40040000402000ULL,0x0,0x0,0x0,0x80080002000200ULL,0x0,0x0,0x0,0x800020040000020ULL,0x0,0x0,0x200040000002020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000240100121040ULL,0x0,0x0,0x0,0x0,0x0,0x8000000004800000ULL,0x0,0x0,0x0,0x0,0x202000081000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000280000004ULL,0x808100100028240ULL,0x0,0x0,0x820004002000ULL,0x0,0x0,0x200040004000008ULL,0xd4600e802208ULL,0x416040921000ULL,0x82a43b4280c39a02ULL,0xf44c200404222b0ULL,0xe35e7fcfcedaffecULL,0xb510012082c0f220ULL,0x4061081088c00210ULL,0x10000a2040020120ULL,0x140043886001000ULL,0x30cc20844838000ULL,0x14d35b224693b7a0ULL,0x40123800089200ULL,0xb2c044200600f2c0ULL },
	{ 0x364623618723981ULL,0x5c4f0a194472b211ULL,0x86914ea222135a03ULL,0x4010834008700282ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8cd0d26205d12a00ULL,0x7d49afe905c16e20ULL,0x4002001000800ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x600600500060200bULL,0xd3923d925dd27a1bULL,0xc28c864c2d02bb09ULL,0x50420004002001bULL,0x8edc0c4a88327f01ULL,0xd062f6e5c9321e83ULL,0x748a7c00292b291ULL,0xfb34eff75b73ba1bULL,0xefffff8ac9d77b8bULL,0xef5bf7df5ff6be9bULL,0x1cf75ffb6a267bbbULL,0xd37a2fd2ea773a81ULL,0xfa9e2ff338521f33ULL,0x8fdef338ce73fb9bULL,0xeffededf7ff77e9bULL,0xdfb6493a8cd37f9aULL,0x91bc5f610a529b0bULL,0xa73e7de518579a01ULL,0xe8d072040843ba23ULL,0x6bffba1b09f3be81ULL,0x3abae6ca0b337a01ULL,0xddfff6fecff73e9bULL,0xc5087d006d527a81ULL,0xa4c2f2e88771b0bULL,0x8763ba797c53bab3ULL,0x97b9c9705b729a31ULL,0x54b27ee919f23831ULL,0x1008ae504870018bULL,0x9b822b5f1ed33a0bULL,0x4302260204513221ULL,0x6910c11248702001ULL,0x9437344e08527a8bULL,0x900c40600003380bULL,0x102c1008210300ULL,0xf4e4e10e0a235b91ULL,0x2240415840521201ULL,0x430c4ac4c12380bULL,0xc800246000021010ULL,0xc10204000c022100ULL,0x4301f20208721e01ULL,0x8208000001ULL,0x0,0x100c724802521601ULL,0x284340014013200ULL,0x1028b2012423201ULL,0x204012180c103100ULL,0x4105413208d25a01ULL,0x9050326908701001ULL,0xa80020001010a00ULL,0x0,0x80800000000201ULL,0xa00021201ULL,0x20004000001001ULL,0x0,0x8000244001402800ULL,0x2004e02008021090ULL,0x0,0x0,0x0,0x820000101000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x120008000001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x918a410700411221ULL,0x0,0xa102000080420200ULL,0x8010000008000080ULL,0x0,0x0,0x8000800000000202ULL,0x0,0x2008000001800ULL,0x0,0x0,0x0,0x4000000140000001ULL,0x800010000001001ULL,0x0,0x0,0x0,0x0,0x0,0x48000000600ULL,0x0,0x0,0x0,0x80800008000100ULL,0x88001000800ULL,0x8025000008400281ULL,0x828001000800011aULL,0x0,0x42024004012800ULL,0x20060008000810ULL,0x94e682000501a01ULL,0xf4902b6618523691ULL,0x6602d8ea49621b03ULL,0x6755d65a8a761aabULL,0x9585450044d02601ULL,0x932882200812118aULL,0xc000200208600200ULL,0x400200000004800ULL,0xc54108000761608ULL,0xca2554008901900ULL,0x1080100800004a02ULL,0x80004000800200ULL },
	{ 0xabfefa2b6a767ddcULL,0x6431a0024521184ULL,0x2020490000106a40ULL,0x1022001024020000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xb5ddd66be27e1d4eULL,0xaa147d63a276154aULL,0x9924132a02340846ULL,0xc9c18a28624610caULL,0xaa84c93622e0858ULL,0x44840028001010c6ULL,0x8320260002664858ULL,0xb5322f7c6377184cULL,0x96658a0420561484ULL,0x81c514e22661550ULL,0xa8b20448002c0040ULL,0xfffd877227761158ULL,0x4c0207682352905eULL,0xeed3ab3e6676dcdeULL,0xefffffcfe67f5fdeULL,0x959beff9a27e6ddeULL,0x9fdc1738626ddb48ULL,0xeebe3fd02677b5deULL,0xc8d4eb6c626494c6ULL,0x4eb65b5f6434194eULL,0xd2baefe2667e32d0ULL,0xe09676fc207d1dc4ULL,0x82a60d6060761444ULL,0x8858110003542082ULL,0x84ce274862694946ULL,0xefbeef7b676f1fdaULL,0xd176eaa2345cd6ULL,0xe9b4e7f2a6761bcaULL,0xbba46292e76811d2ULL,0x4945a06002411ceULL,0x9b6c050000660040ULL,0x421cae002143146ULL,0x20256060061100ULL,0x4a021a0002162240ULL,0x800000400001080ULL,0x908080924521050ULL,0x8a240022100144ULL,0x100824000260048ULL,0x12a0051062140048ULL,0x8f80d60222741480ULL,0x10c000460000ULL,0x809a05ada208805eULL,0x910cf40e21705c86ULL,0x1840b6422040086ULL,0xa0542366016e0940ULL,0x4084022040008ULL,0x481220000148ULL,0xc0502230606249c0ULL,0x0,0x800002000001100ULL,0x0,0x0,0x10001000100002ULL,0x108000000420000ULL,0x200010000000006ULL,0x0,0x0,0x0,0x0,0x200400000001400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100100000000006ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100002040040000ULL,0x0,0x0,0x1000000260000000ULL,0x0,0x0,0x0,0x2400000000400008ULL,0x8010000000600000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x12014000501040ULL,0x0,0x0,0x200040000000050ULL,0x0,0x0,0x100020040040000ULL,0x0,0x0,0x0,0xa00000000000006ULL,0x800404820440080ULL,0x400a1020000082ULL,0x0,0x0,0x40241000201900ULL,0x5290af66273c7cc6ULL,0xe9d5e636e27e304aULL,0xcb7f5f9aae7fb1deULL,0x18b2c7c1246b1540ULL,0x408740226200c2ULL,0x2020044000241100ULL,0x880900008400140ULL,0x20d21a5a23741ac4ULL,0x400244000101000ULL,0x12000000006ULL,0x4000241000200040ULL },
	{ 0xe331e10bc91c8fb2ULL,0x4090162885104c60ULL,0x21d441008006c58ULL,0x9003024000540300ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x140000000000440ULL,0x8f6c05500a19c6daULL,0x408c16e9890cdc9eULL,0xfabe7df39f1cefffULL,0xf3bef7efcb3cefddULL,0x4dd7feae8f15ffd9ULL,0xe1d5df5e8b2dcfdeULL,0x202080610000cc50ULL,0x85c6a7108f0ce911ULL,0x1a141e6c1b017618ULL,0xcc6fe228c2ccacaULL,0xbfecf5c4b3dc6d7ULL,0x64ce5febcf3cdddeULL,0xb9eaff731e5ddedcULL,0x48c015409146480ULL,0xcfc7f2be0e3ce7d7ULL,0x7ffedecbcb3cffffULL,0x77bbdff91f0dffdfULL,0xb8efbf711e3dfdd7ULL,0x92907d53971c4ef4ULL,0x8282ffba1a3cb691ULL,0xeef6b6cb4b5ccf97ULL,0x31fcd7e1b0ccddaULL,0x9cf796e89f1c6fd4ULL,0xcbaa3de21a3cbf94ULL,0xc1821b145a3ce4daULL,0xea67be3b9238cfd9ULL,0xbbbedfdadb5cefddULL,0xbdd71eed1f7cfd7dULL,0xd3fffd739f3dcf9cULL,0xea86b78e1b2c89f4ULL,0x875382004a38cc50ULL,0x2ff29dd3170ccd19ULL,0x109e7ef2521ccd98ULL,0x9083f942431c4fc0ULL,0x8002a0044104c00ULL,0x44880332120c0010ULL,0x518d3d08109cc51ULL,0x242008a01220c010ULL,0x120825600301c810ULL,0xc48101c001014c40ULL,0x650c800003008e90ULL,0xa6e0001010c8010ULL,0x80100a31022c8518ULL,0x24000008000200ULL,0x428aa5628a0c8cc0ULL,0x1038b20020c9490ULL,0x5080861000108488ULL,0x8a07c48b1a0dc694ULL,0x13682480a34a780ULL,0x20004000000408ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80080004004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20080000100080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100008002000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000100080ULL,0x0,0x0,0x8800000000008002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8200000012000000ULL,0x8000020000100020ULL,0x0,0xa00008010ULL,0x0,0x2050040100c00ULL,0x0,0x0,0x0,0x0,0x404000000c00ULL,0x0,0x0,0x0,0x0,0x0,0x20408000008ULL,0x8000004000001010ULL,0x10000200100200ULL,0x0,0x0,0x0,0x0,0x2020002000100ULL,0x0,0x0,0x0,0x0,0x0,0x8040000008010ULL,0x0,0x418615701a28b418ULL,0xffbfff7fdb7dfffdULL,0xefdfffffdf7dffffULL,0x2a3ed6dc033cce2fULL,0x15051b814510ccb1ULL,0x0,0x2a34be109a20a641ULL,0x8000300082000900ULL,0xab6e1808031447d8ULL,0x20220d811200d810ULL,0x410000390a000640ULL,0x2141088000cc0ULL },
	{ 0x4bc1f5e84128c7a3ULL,0x687796401182052aULL,0x402004100000ULL,0x260040010a0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc0edc12e0db85240ULL,0xe8e3b470ec9a0d6aULL,0xfbccdf33d384ff3ULL,0xad7d33ec6daac3efULL,0xdffeff61c5badfb3ULL,0xfb8ff2190d9acd3bULL,0xc5e2be31fdba9f23ULL,0xbfdfdfda5dba1f63ULL,0x7c9e5ee9d5ba45f2ULL,0xfbea2f7bdd321ff3ULL,0xfaa5660fc11a4f6aULL,0x6146460811220542ULL,0x6944c9404cba0d44ULL,0x247e113101084702ULL,0x880c742029000502ULL,0xd032257005181f81ULL,0x29c6e92acd2847f3ULL,0x2bd3f9c74caa476bULL,0xbfb3df6999aa0ffbULL,0xd3dfb6dbfd329fe3ULL,0xea8c3560549acd33ULL,0x98c001210380402ULL,0xb0da9fdfd0b84773ULL,0xf638d1237db29d2bULL,0xd5e6b1ebdcaa0febULL,0x92120d50f5200701ULL,0xfbd6f562dd3a5d72ULL,0xddca1e0d75a80fa2ULL,0x56bdc7fb293a4ffbULL,0xf4d4feea6cb2cdc3ULL,0xe1382de0158a8f2bULL,0xc3936206048a1ba1ULL,0x802ca2001220345ULL,0xedf4cf4914aa574bULL,0x44f5f76a1528150aULL,0x19045c0000120e01ULL,0x20032c08080303ULL,0x6c0f102041886c3ULL,0x4256135095a21412ULL,0x81005800120001aULL,0x9149ac1410288933ULL,0x489025100100030bULL,0x29448628092881a1ULL,0x401e5000c40015a0ULL,0x400a01632482030fULL,0x314105a0112a454aULL,0x32823e4411284701ULL,0x100c00904081303ULL,0x24613514120094bULL,0x2ab06b5205285303ULL,0xe415c02094980f01ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000002000000100ULL,0x0,0x0,0x800020840080020ULL,0x0,0x0,0x0,0x0,0x12040000020ULL,0xc00000000002ULL,0x0,0x0,0x4000601141020902ULL,0x0,0x10040001200000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10000800000000a0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x11000280000ULL,0x0,0x8000100004400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20080008000100ULL,0x52400080101ULL,0x0,0x0,0x200480800020241ULL,0x0,0x0,0x0,0x800080080000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x9032f84e212a81c3ULL,0x414e7403880a0d97ULL,0xfb567b79489a8fe1ULL,0xfffffffefdbacfffULL,0x2d02d0814c8085b3ULL,0x857a5020702007e0ULL,0x4008b464c1220572ULL,0x2090001000100402ULL,0xa1040214448a0025ULL,0x90023211043809e0ULL,0x11143e40c0a0126ULL,0x3870726849220362ULL,0x804240005000102ULL },
	{ 0x104010280120004ULL,0x4c2104004126020ULL,0x930846121482aaa4ULL,0x97b0a25434d3c6aeULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa220044024106020ULL,0x0,0x0,0x0,0x0,0xec4ca470341278e4ULL,0xc944f52c3c98f6e2ULL,0xcc52b218341176e6ULL,0x11bc4b7534975ee4ULL,0xb0fe57eeb492efe6ULL,0xdf9ffd70b4ba6eeeULL,0xdfb3e37eacd37eecULL,0xc5ffb8bf34d1f2e6ULL,0x28edf8dc8c90f8e6ULL,0xc4b43ee23082ebe6ULL,0xb9ce7ff334fae6eeULL,0xefd0eff7bcb3f6eeULL,0xd95e6f1a34b3fea6ULL,0xa641c1a081042ecULL,0x80050ce004903140ULL,0xd1ae04203411d060ULL,0xca902e323491f06aULL,0xc11bf54914bbfceaULL,0x6e588a48308a46c2ULL,0x200022800004940ULL,0xc4809070088a4ce4ULL,0xdcbe77f3bcb7ece6ULL,0x88991221303222e0ULL,0x80131a4220127044ULL,0x88410022c000ULL,0xd571fc73809aea62ULL,0x5ff675723433ffeeULL,0xdad4f32fb4955ce0ULL,0x1843a6a104d1d0ecULL,0x694c358109cc042ULL,0x90d1dc01141324e4ULL,0x80062d60a09262e4ULL,0xca91260e14905ae2ULL,0x84c28f3290908ce6ULL,0xceb75fde38dde5e6ULL,0x9cde7ee31492c7eeULL,0xc1a6116a34b6e166ULL,0x8040010000002ULL,0x2020000104000ULL,0x6000010000400ULL,0x490002000c040c4ULL,0x84c4022804904aa0ULL,0x40a420f008d26824ULL,0xc842e22804f460eeULL,0x8550000200920260ULL,0x6181742100004eaULL,0x1004201614904000ULL,0x3a60c4c00914062ULL,0x8983c36b14136002ULL,0x41914a15149653c4ULL,0x30290101340264c4ULL,0xf307a4e834b8c4e2ULL,0x1200000020000040ULL,0x0,0x40001000100020ULL,0x200040000220000ULL,0x4008000004800ULL,0x1000c1014200400ULL,0x300400104900020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800000800110000ULL,0x4080100000020080ULL,0x0,0x820000000000280ULL,0x21000000280ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x60000000110000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2105000944000ULL,0x0,0x24020800882040ULL,0x4002000004040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000080000004040ULL,0x0,0x0,0x0,0x8002000900000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000032000022280ULL,0x0,0x200000800022000ULL,0x0,0x0,0x0,0x200001c004022004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x100820200002a0ULL,0x25a617f014bacae6ULL,0x81101c65a0954062ULL,0xa5dfff2a3c96ffe6ULL,0xea7e955214926bc6ULL,0x513512a38926062ULL,0xe5c6a660340044e6ULL,0x40020134041024a0ULL,0x6956677a34126ee4ULL,0x2161800808802e2ULL,0x8010227028026060ULL,0x41003031a400486cULL,0x4c341a4490110462ULL },
	{ 0xa122850008460ULL,0x89e4080152105cfaULL,0x8edc4e03003433c2ULL,0xfc7fb672faf0d66cULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd9eaff741054b6cbULL,0xd3be3564329c7cfaULL,0xc181a243a00454a7ULL,0x0,0x481000810008240ULL,0xe50df060029d74faULL,0x45264dc2f80411e2ULL,0xc386e7fff864f568ULL,0xefefbf5fd05cd7ceULL,0xfabb4d7a30b89fe8ULL,0xf5e7ffec7a74dfeeULL,0xbba6fff2f23cf7efULL,0xf7b5ee3370bdd5fbULL,0xefcffe7bd294f7eaULL,0xa4ffcf4a70d896e9ULL,0xfcf5defb5034bfe9ULL,0x797e7561701536eeULL,0xfff6ff3672d4d7feULL,0xcd7ff73ed8f4deffULL,0xa05bdfda70a0b7e8ULL,0x14a758a8020036c0ULL,0x141264180501218ULL,0x81292080001248ULL,0xebece3be2afc777aULL,0xabde8acb10cc94caULL,0x6200040000404c0ULL,0x8000200010000400ULL,0xe20225854200d462ULL,0x89a1d33a9804548bULL,0x62534b40522410c8ULL,0x35a8cf01301d16e1ULL,0xdb6ca756c8bcd66aULL,0x7ff67df3e05c53f2ULL,0x92800a1c001456c8ULL,0x85017500004d08aULL,0x611467a504d146aULL,0x2011c071100056c0ULL,0x192c6f7002b4f2c0ULL,0xcbb78bbf109c9edeULL,0x494630a28054d6c9ULL,0x4830480020000560ULL,0xb0b4d661125c56c0ULL,0xd94cffb29224b7c2ULL,0x284100000401404ULL,0x50420000000420c0ULL,0x208000000000140ULL,0x48000100200ULL,0x1004444a00801541ULL,0x44082822200c5460ULL,0x44c1e00800941541ULL,0xa8c60980128054c1ULL,0xb41557e3b081b660ULL,0x80092210805c5080ULL,0xa802a00000c4240ULL,0x81224108100406c0ULL,0x10e5010c01080ULL,0x51a81c3804034c2ULL,0x64272024089c1240ULL,0x0,0x8110820218109010ULL,0x0,0x401000004000ULL,0x0,0x0,0x200000200002040ULL,0x808180080801000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x800100000240ULL,0x0,0xc000000001008ULL,0x4000200000000404ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xa00040008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8020050000000ULL,0x20000100010ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000240ULL,0x20040000100200ULL,0x4000019002000040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1000020000000240ULL,0x8010000001008ULL,0x0,0x0,0x8010000000004000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20000801000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x500000004000ULL,0x0,0x0,0x48847520d8101ec0ULL,0xd88c280e38101463ULL,0x6dc5ef9210141fffULL,0x479cd71b50149ff6ULL,0x9cfc5771b2b456eaULL,0x937e264018941f69ULL,0xc88e2cea30bc14feULL,0xc1c4d93e720436eaULL,0xac15388e101c34d8ULL,0x9004b6100416c2ULL,0x9488240450181062ULL,0xb8e6fc2727498e2ULL },
	{ 0xa0c641698081aa68ULL,0x834e0811d1693142ULL,0xa084d93412d236cULL,0xe9e6b678d1eaa3eaULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x14b24ee6f9b42356ULL,0xdb7f1568c16f33caULL,0xe056354c4102a0e2ULL,0x0,0x21e308c8c8602a44ULL,0xa7bedd69d8f9265aULL,0x14317644d9490340ULL,0xc7863d62d9e42544ULL,0xfb6cf72ad9d5bedaULL,0xfeffbafff9fdabf6ULL,0xbffccf5ff9fdab7eULL,0xf7f7f7ffd9efabdeULL,0xd3feff73d1ff27feULL,0xffded7f4d9f5b3feULL,0xdbdfbb1bd9edbbdeULL,0xe3ffdfd7f9ee234aULL,0xfdbdff6af1ffb7deULL,0xfbe7fffff9efa36eULL,0xefbeffe2d9eea25eULL,0xadf1ff1ae0ecab7aULL,0xaf7cfd9d1ad22faULL,0x2008a80082800ULL,0x4024804080412208ULL,0x82122400c0402480ULL,0x81220000c0100010ULL,0x4010000040040000ULL,0x4415ce1f1613202ULL,0x20020040001000ULL,0xde1225dcc8e53244ULL,0x180c0004000a040ULL,0x8ed0061241450278ULL,0x9696c738d0ac2084ULL,0x0,0x12000202000ULL,0x110002088c00000ULL,0x0,0x2128489149442204ULL,0xb4100274f1100210ULL,0xeefa3400d100a864ULL,0x80823040040218ULL,0x214db212d1652278ULL,0x11dd881d1802006ULL,0x84103060c1040a50ULL,0xddfe7f62f1efa3faULL,0x5b1805b0c1152234ULL,0x99c6b20ec14ca362ULL,0x4014525001008252ULL,0x62b0576af1c40282ULL,0x1100a2400144220aULL,0xa000458c08000c2ULL,0x458484200040a354ULL,0x62469203e0892340ULL,0x20100162c0400310ULL,0xc030f501d1908020ULL,0x88903520c0e42b10ULL,0x2042080040808ULL,0xca5e51c8608270ULL,0x5b0c098c1008860ULL,0xe0558260c8448270ULL,0x0,0x0,0x40020000002002ULL,0x0,0x0,0x0,0x0,0x0,0x480080000800ULL,0x2200010800000ULL,0x0,0x2002c0000480200ULL,0x22c0088002004ULL,0x0,0x0,0x40046000240060ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80100000000140ULL,0x0,0x0,0x100080080010000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10100040000080ULL,0x801000400100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000100000008040ULL,0x0,0x10000000210ULL,0x0,0x81000080000000ULL,0x0,0x0,0x0,0x0,0x0,0x800010000400100ULL,0x804000000840000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200004011000000ULL,0x0,0x244840002004ULL,0xc00000000140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x18020000200ULL,0x0,0x0,0x2002880090000210ULL,0x1080140080042008ULL,0x40823500608480d8ULL,0x9bb44f1cd14ca490ULL,0xfc7ff3ed8fd3f5aULL,0xb5a1f5bc9edba42ULL,0xb0b205e0c0f6a348ULL,0xc409710251802142ULL,0xc7b6136cc84d2656ULL,0xa85cca0de1c3b646ULL,0xf617db436144a2eaULL,0xad1275bfe1af2a52ULL,0xdff6bf1fd9cba37eULL,0x18040c0260849010ULL },
	{ 0x8bf5ef6b0b6ce522ULL,0x8c1f3a6b5fa42732ULL,0x3fe8cf93d2e0bdd0ULL,0x99d28e681ae8e530ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xac4c504d084000ULL,0x5fbacff7dfecef62ULL,0x5040365290acc530ULL,0x0,0x0,0x42d9594f1822b520ULL,0x1b280c009000ad20ULL,0xc23cac6815a88500ULL,0x2fbc7dcadfeafd20ULL,0xdaf5ff7edffe7f64ULL,0xfffbfeaddffce7f6ULL,0xe7bfcf6bdf7cfdb2ULL,0xfdffffffdffeff74ULL,0xfffeff73dbfeefe6ULL,0xff7eff7fdffaffe6ULL,0xcdcefabbdffcefb6ULL,0xbcbeffdf9fbced56ULL,0x7cfafef7d9acff12ULL,0xdf2e667ada66e772ULL,0xcd3a0a705ab87542ULL,0xc2c1ba0215ea6940ULL,0x2280814808284520ULL,0x1c3775a4d5ea7f20ULL,0x500c060048240400ULL,0x206142490000030ULL,0xc00000003000ULL,0x10040000000400ULL,0x13008210800b400ULL,0x200100008100ULL,0x5c090000120ULL,0x4000000000002400ULL,0x4202004010084020ULL,0x308028480006500ULL,0x60263090003c4100ULL,0x0,0x50000c00004700ULL,0xcac32ad89a30d500ULL,0x494c1db48a8ad00ULL,0x151270fc8908e724ULL,0x44920d400128a500ULL,0x530065040888c960ULL,0xc00c960c15242412ULL,0xa28004210009010ULL,0x10645dee8024a540ULL,0x2124052a1824f404ULL,0xb20271698305122ULL,0x40c2060a0a802500ULL,0x60000008400000ULL,0x3c395f404008aea0ULL,0xc179050008208580ULL,0x5680024218a06500ULL,0xc482a02090ac8500ULL,0x150a12080968b402ULL,0x840210000000e100ULL,0x50c03770900ac100ULL,0x40802d80520aa500ULL,0x8120210b9024a920ULL,0x60000a8000b400ULL,0xa904a42190acc14ULL,0x2404326d00889500ULL,0x0,0x44000000008100ULL,0x0,0x0,0x48010000020ULL,0x0,0x0,0x0,0x0,0x0,0x8040044010000400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10020000000120ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x400200000400ULL,0x0,0x41021000202400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000000402ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x440000000000600ULL,0x0,0x100000190000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x242010241000ULL,0x0,0x0,0x21000084000ULL,0x0,0x0,0x0,0x0,0x0,0x6040002000ULL,0x0,0x0,0x820000009000ULL,0x18040180002120ULL,0x0,0x9010416090268100ULL,0xcf307774dd76bc10ULL,0x88ede186970a3520ULL,0x685e5dd0cd60a480ULL,0x4376e61412c3f40ULL,0x11a005280a20c900ULL,0x4e9222d885286504ULL,0xcf46e32c50ec7560ULL,0xa008108008086420ULL,0xaa294f019040ad22ULL,0x916827f09a486524ULL,0x101005d00030a440ULL },
	{ 0xdadff7682fd24d61ULL,0xcc31b118a8c4f61ULL,0xbb9aeff317f76d60ULL,0x88505461118e5160ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2fffffdb9fdeff71ULL,0x1ebbcffb9fde5d73ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x43362df0159a5770ULL,0xd940f73205065562ULL,0xed6bb74f17165d73ULL,0x6fbedf62b3de5d62ULL,0x77fd7efcb7ce5f73ULL,0xdfffff731fdeff63ULL,0xdf06f76f0b525563ULL,0xc182c23001061020ULL,0x69a4d5d211961560ULL,0x4cbc5eee8e8e7f61ULL,0x99cad37214964962ULL,0x82387f1607966f70ULL,0xc549e92a9fce4470ULL,0xc2d48c43069a6732ULL,0x14b55e6429864f62ULL,0x90b4a258a7406b62ULL,0x22625c007961640ULL,0x2016002884240ULL,0x0,0x4006a1002060440ULL,0x8050054006020460ULL,0x0,0x10100000020800ULL,0x86c00400a860040ULL,0x8408101020900ULL,0x801002000400ULL,0x800000202002000ULL,0x840000001020000ULL,0x0,0x820a0502a6006461ULL,0x40c634e901a61d61ULL,0x2432741207968d60ULL,0x578e79fa9fd7fb60ULL,0x100800002028020ULL,0x210014c002a1060ULL,0x10c0d8cc324c6121ULL,0x95a03f200a4e7020ULL,0x7600276801424d41ULL,0x481090806080530ULL,0x88040100021400ULL,0x106200010c40ULL,0x0,0x8804043810825160ULL,0x8000000400020800ULL,0x800118002000440ULL,0x84e41108c0420ULL,0x9181746bb28e6460ULL,0xc022054809860740ULL,0x4193800620404023ULL,0x4e0000800064021ULL,0x4180160000a4020ULL,0x140920a06000511ULL,0x200240020020060ULL,0x0,0x140000100100ULL,0x210000000020040ULL,0x4100000802006002ULL,0x582004000500ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004004800404900ULL,0x0,0x0,0x4400010800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2004000020040ULL,0x0,0x0,0x0,0x0,0x100000800040020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000000000140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000160000021140ULL,0x162600640ec6c940ULL,0x8800000000080020ULL,0x8100000000060ULL,0x8020040000000470ULL,0x0,0x0,0x0,0x0,0x0,0x500000000820ULL,0x804000000000401ULL,0x0,0x0,0x20000080a000000ULL,0x0,0x944102202404402ULL,0xcb3065528206fc01ULL,0xee56f2f89e9f2d60ULL,0xa9c5f4213865d62ULL,0xcb9e46501c65160ULL,0x7080a660168a0a40ULL,0xdb80277abac64d72ULL,0xadc5e21186d87b00ULL,0x21c25a9104064963ULL,0x2000cdd107cacf21ULL,0x532c2f28862a6d53ULL,0xb2e344e98861d40ULL },
	{ 0x18ed370c7006408ULL,0x826612c2d42c6f49ULL,0x10c69244082c02ULL,0xb4b62aa072b03480ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4ecbbe30568e38f3ULL,0xefdfffff5fbc7ca8ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x7d7fbfebfebc67f9ULL,0xcdfefd73dfbc7ce8ULL,0x0,0x0,0x8002091342282c00ULL,0xf4f7f7ecdfbe74e8ULL,0x905f2d71d73e6de0ULL,0xdbd0ef76d51c4768ULL,0x4082de3076106cd0ULL,0x255036555e1a6c80ULL,0x80294420d6102e00ULL,0x40200000000408ULL,0x0,0x0,0x60121c1093104460ULL,0x81900cea91a80660ULL,0x0,0x4a00001040820040ULL,0x8000400008000400ULL,0x42081761c9943480ULL,0x1011000291000800ULL,0x0,0x0,0x0,0x8080000190000120ULL,0x8000040800200400ULL,0x845204ec56902420ULL,0x8025042d4000708ULL,0x800620c10000448ULL,0x144040800201400ULL,0x801290000300ULL,0x50bceee8f3102322ULL,0x4964963895b427c8ULL,0xa962d1092146da9ULL,0x2041c010b1100100ULL,0x88901102001400ULL,0x40017e0d4000142ULL,0x1840a30e64100422ULL,0xca802c2244001442ULL,0x100800010000040ULL,0x42000000104000ULL,0x200046012002200ULL,0x120404c040001420ULL,0x71062c6244302449ULL,0x8ed420254200548ULL,0x4a200d50720c242aULL,0x8090800002040480ULL,0x60000002001ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2020080000100ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80181010042001ULL,0x0,0x0,0x8100000040000400ULL,0x0,0x80000104100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20c080081200140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4002004100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x14040020000ULL,0x0,0x0,0x10040044000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x2004000000100000ULL,0x10001010000040ULL,0x8800012490080500ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000000000420ULL,0x0,0xa000000080040000ULL,0x400904044002001ULL,0x9f4e7762f48c35e2ULL,0xeb387d72d71c6500ULL,0x4906919c121c2e20ULL,0x2aa5f548d23c3dc8ULL,0xd4fb7fe9f71c7df8ULL,0x1000504044000002ULL,0xcacc05485d041460ULL,0xc223621a802760ULL,0x15338005e8035d8ULL,0x38800183951c2900ULL,0x122ae578f6106521ULL,0x99c60f60d3942fb0ULL },
	{ 0x8134d778ab902fb5ULL,0x4ecf7e71ebdc273fULL,0x67f8cd62986d229eULL,0x8c60020442802b2cULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x280000008000200ULL,0x10280000008ULL,0x0,0x1004000000000cULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd4f4726882902357ULL,0x8001200082000204ULL,0x0,0x0,0x0,0x804080000010ULL,0xc17bf7f29b9c2b0eULL,0xdfbc6ffeabfe233fULL,0x4cc7ee188ba822f9ULL,0x398de1b91900aceULL,0x2d414600280220fULL,0x4000305283000024ULL,0x0,0x0,0x802200002002005ULL,0x49404a009080208ULL,0xd010060202022604ULL,0x2100002000004ULL,0x0,0x200004000800010ULL,0x0,0x440022001002020ULL,0x0,0x3a0830282a0220fULL,0x40431a1180800238ULL,0x86cb44009a182204ULL,0x2048482000110ULL,0x50a300082000102ULL,0xe30016028a800b06ULL,0x40232a0a2802012ULL,0x837c0a5aabc22032ULL,0x709075e48b08205eULL,0x5126a67293a0023eULL,0xa00362c2002010ULL,0xc4309008a382604ULL,0x4444a82800010ULL,0x9013048282000216ULL,0x990c206acac02407ULL,0x1000000400000208ULL,0x18584001200240cULL,0x40024802002005ULL,0x3011402080000386ULL,0xd008a4009a40030eULL,0x4690574443802006ULL,0x8303ea269b6c260eULL,0x4007101008602600ULL,0xb0144ba08a40021eULL,0x3000064008c10012ULL,0x0,0x0,0x0,0x0,0x202c008c00010ULL,0x200201081000012ULL,0xeb0477f88adc2a4fULL,0x800870001900010ULL,0x940000009400000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100e00201000248ULL,0x800004000002400ULL,0x40200080200ULL,0x0,0x0,0x0,0x0,0x0,0x200102201000020cULL,0x0,0x0,0x0,0x0,0x0,0x0,0x5000000002000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8600000001902000ULL,0x0,0x0,0x0,0x12000000208ULL,0x0,0x0,0x44001000020ULL,0x0,0x0,0x8000400001100000ULL,0x300000000208ULL,0x0,0x804001000008ULL,0x0,0x8400002a0100200ULL,0x2020002000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x200000800002ULL,0x20800000300ULL,0x0,0x0,0x0,0x498c7d628ac22205ULL,0xf314667d8a9a2379ULL,0x4054b28689182308ULL,0x6a5cdadc09f0226bULL,0xfd3ac7ffdb9b22eeULL,0x858d06681bd00114ULL,0xdab02f7c0baa2a1bULL,0xd0e0e522991022acULL,0x0,0x88818149802110ULL,0xd18826da8b84230bULL,0x2890054282a02320ULL },
	{ 0xc96ae90a790cd11aULL,0xd8763e26cb2453a6ULL,0x7b3c7f728ba4d3aeULL,0x80438ee0c184d128ULL,0x40000000120ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4080a7c28a005122ULL,0xdffebf7fd9b5f3beULL,0x46432e188980538cULL,0xbd81f108018572aULL,0xc8960120c321d200ULL,0x8140264003005286ULL,0x88012040008000ULL,0x80800008004000ULL,0x80e0080004006ULL,0x0,0x0,0x800040080200000ULL,0x0,0x8004000000204ULL,0x6000100200ULL,0x0,0x4402004001004120ULL,0x18180104000d080ULL,0x807000c280205182ULL,0x4614c10680041162ULL,0x51ee2648c090d10eULL,0x895ead32d834932aULL,0xf9b8c06a4b04c12aULL,0x8448228c110c0e0ULL,0x212040e83240886ULL,0xccb230b84080c02aULL,0x9021542c120c0e2ULL,0x10050063000000ULL,0x855c000012052e0ULL,0x40100048e0200000ULL,0x810090180205068ULL,0xd1001e48c120c284ULL,0x18001004000ULL,0x100a020000041a0ULL,0x10100000000120ULL,0x101342a0e020c00cULL,0x42c040060001120ULL,0x48bc55d0f2b1d32aULL,0x2002000100200ULL,0x40022400c0004004ULL,0x9299cb23c305cb2eULL,0x1140b724812c5080ULL,0x40001000080ULL,0x0,0x0,0xc00000004002ULL,0x0,0x0,0x11000008020ULL,0x482020000804060ULL,0x1044280005008ULL,0x813040200800410aULL,0x0,0x8000020000084000ULL,0x10800000000aULL,0x2100000000110ULL,0x80000800000120ULL,0x0,0x200160004000ULL,0x4002000000004080ULL,0x0,0x0,0x0,0x0,0x0,0x80800050000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x240380104080ULL,0x0,0x0,0x20080000008008ULL,0x4000800008000002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2800200200ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x20020000000120ULL,0x4002001001040204ULL,0xa02060000120ULL,0x80100080100000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x202100180000908ULL,0x0,0x0,0x94bcd7e2ea14d122ULL,0xda947effa995c1faULL,0xd1b4b10af810c386ULL,0x1ffdfdbfb2dd76aULL,0xbdb7cff2db14d3beULL,0xf9d8af03c915d3feULL,0xde920d72cb34d32eULL,0xc1c4e4210390c3eaULL,0x480000004002ULL,0x123006408804c822ULL,0x100002000100200ULL,0x8000250041004080ULL },
	{ 0xc90875da83599a60ULL,0xefdf7ed3cbd7f3aeULL,0x169d4fb381d7fbaaULL,0x2f62dc282d4f220ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40a061200868260ULL,0x7f56dfdb4ad6b78eULL,0x3cb377a2c89ad224ULL,0x93022748ca025224ULL,0x202004c0a003000ULL,0x800020080000004ULL,0x265091ca8242a3e0ULL,0x140005600200a100ULL,0x0,0x8000400002008000ULL,0x0,0x0,0x20a0800002402000ULL,0x22020000022040ULL,0x1200000008040000ULL,0x0,0x802000000020020ULL,0x6d040000244a020ULL,0x787707f808cfb38aULL,0xa394af008accbad0ULL,0xa210420408123300ULL,0x84716a1080c4e280ULL,0x2a0d9d428854f244ULL,0x400020008000a200ULL,0x108242008442000ULL,0x923a2a128280a706ULL,0xd8c234a80a14b2a2ULL,0x2100c01a8000a200ULL,0x90130de40042a340ULL,0x10e05280a049210ULL,0x3008042108061240ULL,0x84c422000202a220ULL,0x400000800005000ULL,0x0,0x814c12288280a200ULL,0x80c08002220ULL,0x2002080002004ULL,0x4200160280083024ULL,0x10288080003080ULL,0x852ca0418a062a22ULL,0x100080000101000ULL,0x0,0x2200000400002ULL,0x0,0x0,0x8108200000828020ULL,0x80811008081000ULL,0x100800000002002ULL,0x240000004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x840000000400008ULL,0x0,0x0,0x0,0x400000002020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000208001000ULL,0x0,0x8000000200020100ULL,0x0,0x50000020002ULL,0x0,0x0,0x480000840000ULL,0x24000040200ULL,0x4000001000040200ULL,0x4020000008040ULL,0x0,0x0,0x0,0x40800400040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x120000005000ULL,0x0,0x100040000500000ULL,0x0,0x80530208060300ULL,0x0,0x0,0x0,0x0,0x800800c00000ULL,0xa4000c200818020ULL,0x80400000020100ULL,0x0,0x0,0x0,0x0,0xa004400000540200ULL,0x1000142102d02008ULL,0x1060820828c3800ULL,0xdf835f768207bb24ULL,0xc8f2f3298a83bf20ULL,0xaa7cffdb82dff3eaULL,0x25c516e382daf320ULL,0xd12a7f1e8ad6e6e2ULL,0xdaae27eb89c6f6e8ULL,0xa2c2fea60847f362ULL,0xa4a1044020693a0ULL,0x802406e080d4a220ULL,0x4c508008068320ULL,0x4b80030401447220ULL },
	{ 0x8810100080840200ULL,0xeedf3eabcdeb7b95ULL,0xd095c560cc923c14ULL,0x9284b22a4dc3aab0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40001000800ULL,0x0,0x0,0x47d8571348e01fa0ULL,0xbefedf7bcfcf6faaULL,0xdfeeffffedfbffbeULL,0x1a88042000462200ULL,0x100c03228020a10ULL,0x108040000004ULL,0x8008482204880680ULL,0x0,0x82244020014a00ULL,0x10800010800ULL,0x0,0x658c40948c63a00ULL,0x100c2800020610ULL,0x240020000200ULL,0x8800000000808000ULL,0x802000640820010ULL,0x280005c028031280ULL,0x21504678ac0a2a01ULL,0x1200000000400800ULL,0x6b00e6084c622890ULL,0x54c63e1481823e03ULL,0x800e4c0044423824ULL,0x849514620ee8d800ULL,0x12405200f000a82ULL,0x9a205f4c4cca0285ULL,0x6c6000008421200ULL,0x24004801400a00ULL,0x4060156148355b20ULL,0x112026c248111720ULL,0xce102143c9660a20ULL,0x8140901144402220ULL,0x80000100021000ULL,0x0,0x2000102186802000ULL,0x9020014800031201ULL,0x8820320108231900ULL,0x800000101000800ULL,0x204026084020008ULL,0x1128820020400a02ULL,0x0,0x8000010004020000ULL,0x0,0x0,0x0,0x0,0x4000200000000600ULL,0x0,0x100030240400220ULL,0x5000000000000201ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1004000004080000ULL,0x0,0x1800000800800ULL,0x4002000040022020ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x8000000004800000ULL,0x0,0x0,0x0,0x400008000800ULL,0x0,0x0,0x0,0x0,0x0,0x8000000000004800ULL,0x202000000400200ULL,0x0,0x0,0x0,0x40040000020008ULL,0x1000a0040000014ULL,0x8104000004822000ULL,0x6000080020000ULL,0x0,0x0,0x0,0x802000400200ULL,0x0,0x0,0xc14000040420200ULL,0x0,0x0,0x0,0x0,0x0,0x2f1c084422208ULL,0x1ba66f6549c26f81ULL,0x89f4932006533208ULL,0x8ece4fd268414fa1ULL,0x47bb55a60ccbae83ULL,0xc1da6c4905c0aa01ULL,0xfb96afde6de2af98ULL,0xc9d4ff6e6dc60e12ULL,0x94a68010c822890ULL,0x65b2cfe24dd73a95ULL,0xa0ac58420cc38680ULL,0x9bc47d6cc4821e1bULL },
	{ 0xc901c072a2421648ULL,0xebffdfdf7a7a17cfULL,0xbdfbefe3a67e38cbULL,0xdfdfb6c9617b17cfULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xa0d481000029001ULL,0xa5fe5fe2aa7691ebULL,0xd12de7722c7631cbULL,0x20102d21204a10c0ULL,0x4682da80274011d0ULL,0x115000500001ULL,0x89b247fa42508209ULL,0x0,0xa00000000600000ULL,0x54020000600048ULL,0x8010044208501041ULL,0x1481c40102029049ULL,0x8000000000101000ULL,0x43c2020021180ULL,0x20010a2000800bULL,0x800020022000000ULL,0x60060482a4000048ULL,0x2c0200a2001182ULL,0x80424112000140eULL,0x13060002404a1049ULL,0x4dfab2116552155eULL,0x355005466a000149ULL,0x4b1d9c2b2409703ULL,0x43b670925011cbULL,0xd426264408461687ULL,0x8208ca002b101083ULL,0x51c464020423080ULL,0x2080002880500002ULL,0x36f0668b070904aULL,0xcc82338ca852104fULL,0x8010000000001400ULL,0x0,0x820000800401483ULL,0x2000020020000008ULL,0x4800400200ULL,0x44000000080040ULL,0x62020c0000501148ULL,0x80401020400028ULL,0xd0348464220814c1ULL,0x0,0x0,0x0,0x1010140060021000ULL,0x0,0x880000600000ULL,0x210000040000001ULL,0x0,0x0,0x8010000000001001ULL,0x0,0x0,0x0,0xc00000400008ULL,0x0,0x0,0x240000420000ULL,0x0,0x4e0060000082ULL,0x200040022000000ULL,0xc000800008600082ULL,0x0,0x8000000000021000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x120060000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000800000082ULL,0x0,0x200010022000000ULL,0x0,0x0,0x80050202600046ULL,0x0,0x200000440000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xc000020008000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x115020420008ULL,0x0,0x0,0x0,0x800002024000000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x80010002000040ULL,0x2004024000420001ULL,0x84625e02602184aULL,0x111a2f007a021049ULL,0x68c7c044215a10caULL,0x6f9cdbd8ee5411afULL,0xb5f637e0eb72929fULL,0xd8f0bedeab5293dfULL,0xda366ef1e24417cbULL,0x8384d695e7121081ULL,0x8a1adeb3704813cfULL,0xabba477a6e7b12e7ULL,0x9b18246648721141ULL,0x6d125c2164501693ULL },
	{ 0x6902fb22403558d0ULL,0xcc05e8b0a3c8dd0ULL,0x1ba9dff369361ed4ULL,0xd67ff2f80b3dece7ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xd88da57641375e54ULL,0xfebf7fffcb37fbf6ULL,0xcb5bfffec3375dd4ULL,0x0,0x1420424000040c04ULL,0x41409200000c00a0ULL,0x5e04552008048184ULL,0x840000008200000ULL,0x200004000000c00ULL,0x0,0x8020000100400ULL,0x44b00458e030a810ULL,0x80008310082c0100ULL,0xe0c29206603400a4ULL,0xf9fcd8168348236ULL,0x4000206008000c40ULL,0x630a05002348040ULL,0x888801220a340200ULL,0x41038c1c086caa90ULL,0x38160d84611809d4ULL,0xf4351ef4097cabd7ULL,0x85cb5560482e06d0ULL,0x41a325140936980cULL,0x4482700800548094ULL,0x31a5a1041140c54ULL,0x8c923481400c0c50ULL,0xb84000001c00c4ULL,0x8880004400008150ULL,0x44000000000140ULL,0x2a00010e08362080ULL,0x10002000080010ULL,0x0,0x20590001c00c0ULL,0x81508000000c05a0ULL,0x16020000280820ULL,0xba0b820829c840ULL,0x10400000000040ULL,0x0,0x6000240000ULL,0x0,0x4400002040000ULL,0x2860000000ULL,0x1080000000100020ULL,0x0,0x8020000000040004ULL,0x0,0x0,0x0,0x0,0x0,0x8000040008200000ULL,0x0,0x1100000200101140ULL,0x0,0x40080000000140ULL,0x0,0x0,0x2004000100200ULL,0x0,0x8000020000200800ULL,0x0,0xc000000000050ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x80020008000008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x800020000080004ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x20004000200040ULL,0x20080000100080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x120000100400ULL,0x2010e4000040144ULL,0x10000100400ULL,0x1001000008000040ULL,0x0,0x0,0x0,0x0,0x1000040000200800ULL,0x0,0x0,0x0,0x0,0x88008000040ULL,0x0,0x1280000008000340ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x10800240000ULL,0x0,0x0,0x0,0x80e75694b3cacf5ULL,0x8f914bf2082c8cd0ULL,0x50d6f63288398edbULL,0xab5b5e934b4e8dd2ULL,0xb5f65fefc97e8ed4ULL,0xd5fbbfe8eb7edef4ULL,0xda921ffa493d79d4ULL,0xe3d7d79f4a3eaff4ULL,0x900e8a8f023c0874ULL,0xda0f2561483ca7d4ULL,0x501c06c2681c13c0ULL,0x729e2be24b340878ULL },
	{ 0xc022d222138a048bULL,0x4eca1e791b02cf1fULL,0xb6bdcde3361a4f4bULL,0xf5eab762171bdf5bULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x290072180209c589ULL,0x6b177f506998075bULL,0xc134482101024902ULL,0x1002001004000ULL,0x80800220000044aULL,0x8010000000020040ULL,0x8022100000000102ULL,0x6b2e4c431b99c703ULL,0x3028224081100402ULL,0x8242105004803ULL,0xd5a0215b6180871bULL,0x8010d0011008050aULL,0x215e4f8c168a4142ULL,0xe435447233084d91ULL,0x202c65e212080510ULL,0x448534000812c550ULL,0x8010480000080102ULL,0x2002d01080008510ULL,0x1455932920889512ULL,0xa21800101412ULL,0x2010019013004010ULL,0x0,0x0,0x18e38005024b02ULL,0x24000001004000ULL,0x1210000410080508ULL,0x4000230004104400ULL,0x408000000000300ULL,0x8402000200ULL,0x560402ec43080759ULL,0x4000008000008010ULL,0x0,0x6000000080400ULL,0x480000000402ULL,0x4000008000000012ULL,0x8200000008001ULL,0x800000200080400ULL,0x0,0x804000000020800ULL,0x40040002000200ULL,0x200053002020112ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100104004000ULL,0x44000020040ULL,0x0,0xc0104000080101ULL,0x0,0x14000100400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44c00003000500ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x802010000400ULL,0x0,0x80080000000408ULL,0x0,0x488c6771511b45d9ULL,0xd2943e525410df11ULL,0xa943fea22b1acd1bULL,0x4b585dc9081a8715ULL,0x35aff16b530acfdbULL,0x7d7ebedc5719df13ULL,0x4aa6a6661b92df55ULL,0x8c7e130c93029e1bULL,0x4dca5a18cc034741ULL,0xe7beed943935f57ULL,0x95b97601783c352ULL,0x22803b6532028703ULL },
	{ 0xd1c1a22911b86fe2ULL,0xeaf71e723cb96ce6ULL,0xc792dfb3f4b85fe6ULL,0xfe943ee3dc9c66e6ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4000200000280000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1b064d106480e162ULL,0x2065f34235bca022ULL,0x408a15200c116100ULL,0x320a70654900de0ULL,0x4c131004a04a82ULL,0xa0ae0c2410c0e942ULL,0x10336880252004c2ULL,0xc1a225001895e8c4ULL,0x808010420c085042ULL,0x80ce10052049c0ULL,0xe48df4194185ce4ULL,0x4a256e821b94ae2ULL,0x61a3ab7b3cf86ee2ULL,0x909149001c114d20ULL,0x801280230084880ULL,0x6014488410000486ULL,0xa1b27b633091cdc6ULL,0xe4650c0894286fc2ULL,0x90013014084500ULL,0x1c4c60014909440ULL,0x652c360000a86842ULL,0x0,0x1000044020000042ULL,0x1000000401200000ULL,0x8000a0010014440ULL,0x4026000004104800ULL,0x20a0040004980000ULL,0x5605804000a8cc00ULL,0x0,0x0,0x0,0x0,0x400002004001000ULL,0x4800000000200040ULL,0x840000000880000ULL,0x8100000006000ULL,0x0,0x1002000000480ULL,0x0,0xa0000100002ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x44000000004040ULL,0x0,0x0,0x200200001200000ULL,0x0,0x10108030000800ULL,0x0,0x10200000004040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x100800000280ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x80000000000000a0ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x201000280000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x820000000004400ULL,0x0,0x0,0x0,0x200004000000084ULL,0x0,0x10201104005040ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x4000000201200000ULL,0x0,0x10000200280000ULL,0x0,0x18ceb610098c2e0ULL,0xc110333430bc40e6ULL,0x5ed7bab0f499c7e6ULL,0x8730ce1a35bc66c4ULL,0x1046928110282662ULL,0x935cae585d3064e2ULL,0xce982f5254806de0ULL,0x86d637fe4ca873e0ULL,0x202dde8040d4e5e0ULL,0x87ba6e72d4f96fe2ULL,0xd103a67c20b8cec0ULL,0xd38a677c30386ce4ULL },
	{ 0xeafefb6a8454f36eULL,0xa7d33acb0c3e766aULL,0x579dc7fbbeb7f2eeULL,0xa6f5f671b41652eaULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0xaa4d43841556e8ULL,0x8c545010020450caULL,0x1002054084103000ULL,0x829b4204a082502cULL,0x8048119ba4103028ULL,0x86fe5ff1c23355e8ULL,0x34240080020440e8ULL,0x8301e581000236c2ULL,0x4000201002000068ULL,0x812086010023102aULL,0x108284200400e0ULL,0x4511a6ab480146aULL,0x8102a758a2911060ULL,0x6d934400040610ecULL,0x82870f08201e10acULL,0x2400c000000011c0ULL,0x400038000305000ULL,0x28c24004146080ULL,0xa20004800104020ULL,0x100002000104000ULL,0xa000020800000040ULL,0x2020000000010020ULL,0x1110005240401420ULL,0xc004000294000020ULL,0x200200104000ULL,0x84c00e00000c1428ULL,0x95e05deaa21f56eeULL,0xa171de68843612eeULL,0x1102040000021600ULL,0x0,0x44c1001016000440ULL,0x8000002000040008ULL,0x0,0x0,0x12000401000ULL,0x0,0x0,0x0,0x0,0x120000004000ULL,0xc000000c0000ULL,0x8004000000040ULL,0x28000000240ULL,0x0,0x0,0x44000000140000ULL,0x100080000000060ULL,0x0,0x0,0xa00000000001008ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40000808001000ULL,0x0,0x0,0x0,0x0,0x0,0x200004000000240ULL,0x0,0x0,0x0,0x810000104000ULL,0x0,0x480000001400ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40080000004000ULL,0x0,0xc000080001000ULL,0x0,0x0,0x480000020004ULL,0x80010014000000ULL,0xa000000028ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4840040000041082ULL,0x0,0x0,0x0,0x0,0x80003080000c0028ULL,0x0,0x0,0x0,0x8040000000240ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x80000100022000ULL,0x0,0x0,0xa8675420037d6a8ULL,0xff3c7ffe5e37f7eeULL,0xe9f7f9a6f4fe74eaULL,0xea1dd649ac92776cULL,0x2e72577ae21650eaULL,0x9ce187d82092702cULL,0xfe862f7ec07776e8ULL,0xa916f2b2a01553e8ULL,0xa5d0e50041674acULL,0x363e8cb6dc127024ULL,0xd5fe3659a8f7d46eULL,0x2c903724228311e8ULL },
	{ 0x8bda730b5116a7ceULL,0x59d217bd511fb7f6ULL,0x375c4dc9515d85d8ULL,0xb5b6f6f371feafceULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x1400402000c2248ULL,0x2084050100800ULL,0xf194cd3ad130afd1ULL,0xc03090001104824aULL,0x4054450140401ULL,0x70449040400ca340ULL,0x8b08e04000412340ULL,0x91802c4500298240ULL,0xc9c6322751a48680ULL,0x232a495151100ae8ULL,0x2210446850648500ULL,0x8022004240048248ULL,0x3a1e6f2051518e56ULL,0xc983ff5ad006a5c7ULL,0x42121208105102c0ULL,0x401048040140010ULL,0x250cd368d11da048ULL,0x2210061401000258ULL,0x100001000102050ULL,0x44000020040000ULL,0x8040000000240ULL,0x80a0080000400240ULL,0x0,0x201050e0401902c0ULL,0x800e0010100090ULL,0x90944ced6106827aULL,0xf6defaff717597d8ULL,0x0,0x0,0x0,0x2108010001100802ULL,0x0,0x108000000000240ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x8000800010000080ULL,0x0,0x404010008000ULL,0x52010100400ULL,0x0,0x110408040100280ULL,0x8002000000400100ULL,0x90042000108050ULL,0x0,0x0,0x0,0x0,0x10002000100010ULL,0x0,0x4000200050000000ULL,0x0,0xc0000180000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2100000008040ULL,0x0,0x0,0x0,0x2800000140ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x24000000010040ULL,0x24040110180400ULL,0x20440002000ULL,0x0,0x0,0x0,0x0,0x200110000400ULL,0x0,0x0,0x0,0x4886053051102362ULL,0xa9947f74517d2fd2ULL,0xcd2ce73ef0fdabf8ULL,0xe3715d58707daad8ULL,0xb7ff5fef71fdb7f3ULL,0xb1cd150c503ca644ULL,0xd3bf2d7551f40fccULL,0x2946e3a6711027f5ULL,0xaafed6de515da7d3ULL,0x34b26eda6040a7c2ULL,0x536554a850e427e0ULL,0x688e3c3151152178ULL },
	{ 0x35a018c08500ULL,0xcee736139090b956ULL,0xb39acdc3d840a794ULL,0x54e496e8d0f32260ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x201000202000ULL,0x1007c0298c02000ULL,0x600000008800ULL,0x2024810022004ULL,0x8800082090100010ULL,0x40126810a03000ULL,0x85080d8098408850ULL,0x400400a080800040ULL,0xa08486029010a904ULL,0x8008840c8400000ULL,0x14312022d8401800ULL,0x9802824290002042ULL,0xca1aaf06d8e8b042ULL,0xc22a8b8040aa80ULL,0xa20800c2d8100010ULL,0x2401400d0001000ULL,0x0,0x121010d0800000ULL,0x100001000082000ULL,0x0,0x8002c44080400c40ULL,0x4140040090000000ULL,0x200040080000040ULL,0x104000000008080ULL,0x0,0x8012000280000040ULL,0xde7c74e898e9be22ULL,0x0,0x0,0x0,0x0,0x0,0x0,0xa0820000402402ULL,0x0,0x8000108400000ULL,0x0,0x0,0x0,0x0,0x200280000020a800ULL,0x4002000008080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x4400000008080ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x2004010800000ULL,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x21890001000ULL,0x0,0x0,0x804016000202002ULL,0x9b308a6e9850b180ULL,0xcbd7c70ed021b9d6ULL,0xab12165ad8c03212ULL,0x5325f689082ac90ULL,0x812c687290802816ULL,0xa692254d98e0bc44ULL,0x89505228c8c00400ULL,0x2dc21a589863b304ULL,0x428400000202000ULL,0x90ccc7609082ad02ULL,0xe3922d70d078ad14ULL },
	{ 0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0 }
};



